package ca.ualberta.proteus;

import ca.ualberta.cs.homodeller.Homodeller;
import ca.ualberta.cs.homodeller.HomodellerException;
import ca.ualberta.proteus.program.ProteusImpl;
import ca.ualberta.proteus.program.ProteusSettings;
import ca.ualberta.proteus.program.predictSPTM.Organism;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.db.CreateDatabase;
import com.martiansoftware.jsap.*;
import org.biojava.bio.BioException;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.io.*;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: Smontgomerie
 * Date: Oct 2, 2005
 * Time: 3:51:49 PM
 * To change this template use File | Settings | File Templates.
 */
public class Setup
{
    private static String CONFIG_FILE = Config.DEFAULT_FILE;

    // where we get the default config from
    private static String CONFIG_FILE_DEFAULT = CONFIG_FILE + ".default";

    static Config config;
    static Config homodellerConfig;

    static final String testProtein = "VTAEIRLYITEGEVEDYRVFLERLEQSGLE";
    static final String testProtein2 = "VAPTLTARLYSLLFRRTSTFALTIVVGALFFERAFDQGADAIYEHINEGKLWKHIKHKYENK";
    private static Sequence testSequence;
    private static Sequence testSequence2;
    private static final String VERSION = "1.1";
    private static final String CLEAN_FLAG = "clean";
    private static final String HOMODELLER_FLAG = "build_homodeller";
    private static boolean buildHomodeller = false;

    static
    {
        config = Config.getConfig();

        if (new File(CONFIG_FILE).exists())
        {
            config.setConfigFile(CONFIG_FILE);
        }
        else
        {
            config.setConfigFile(CONFIG_FILE_DEFAULT);
        }

        try
        {
            testSequence = new SimpleSequence(ProteinTools.createProtein(testProtein), "", "1EIWA", new SimpleAnnotation());
            testSequence2 = new SimpleSequence(ProteinTools.createProtein(testProtein2), "", "TM", new SimpleAnnotation());
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }
    }

    public static void main(String[] args) throws IOException, JSAPException
    {
        // make sure we're in the proteus directory since some paths are relative to it.
        if (!ensureCurrentDir())
        {
            return;
        }

        // Parse any command-line options
        SimpleJSAP options = new SimpleJSAP(
                "Proteus",
                "Comprehensive Secondary Structure Predictions, Local Version " + VERSION,
                new Parameter[]{
                        new Switch(CLEAN_FLAG, 'c', "clean",
                                "Clean the existing configuration and start fresh."),
                        new Switch(HOMODELLER_FLAG, 'h', "build-homodeller",
                                "Apply this option if you wish to use homodeller"),
                }
        );

        // Clean install?
        final JSAPResult jsapResult = options.parse(args);
        if (jsapResult.getBoolean(CLEAN_FLAG))
        {
            config.setConfigFile(CONFIG_FILE_DEFAULT);
        }

        if (jsapResult.getBoolean(HOMODELLER_FLAG))
        {
            buildHomodeller = true;
        }

        BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));

        // Run the setup program
        boolean passed = doSetup(stdin);

        if (passed)
        {
            // Write the config
            config.setConfigFile(Config.DEFAULT_FILE);

            if (config.isChanged())
            {
                config.writeConfig();
            }

            System.out.println("Proteus install successful.");
        }
        else
        {
            System.out.println("Proteus install failed.");
        }
    }

    public static boolean doSetup(BufferedReader input)
            throws IOException
    {
        System.out.println("*---------------------------------*");
        System.out.println("Welcome to the Proteus Setup Program!");
        System.out.println("*---------------------------------*");

        boolean finished = false;

        findBasePrograms();

        findBioinformaticsPrograms(input);

        setupHomodellerConfig();
        
        setupDefaultValues();


        try
        {
            // This makes it so they only need to download seqsee.db and filtnr
            CreateDatabase.verbose = false;
            CreateDatabase.createDatabases();
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }

        // find NR
        if (!findDatabase(input, "blast.dbname", "Where is the NR database located ", true))
        {
            return false;
        }

        // find PDB
        if (!findDatabase(input, "PDB", "Where is the PDB structure database located ", true))
        {
            return false;
        }

        // find SeqSee
        if (!findDatabase(input, "seqsee_db", "Where is the SeqSee structure database located ", false))
        {
            return false;
        }

        String curdir = getCurDir();

        // Ask for where they want a pdb dir
        if (!config.contains("pdb_dir"))
        {
            String pdb_dir = curdir + "/pdb";
            pdb_dir = queryDirectory(input, "Where do you want to store temporary pdb files? ", pdb_dir);
            config.setStringValue("pdb_dir", pdb_dir);
        }

        formatDB("blast.TMPDB");
        formatDB("blast.BBDB");
        formatDB("blast.SPDB.euk");
        formatDB("blast.SPDB.gram+");
        formatDB("blast.SPDB.gram-");                

        // test BLAST
        List<Sequence> blastResult = testBLAST(testSequence);

        // test Xalign
        testXAlign(curdir, testSequence, blastResult);

        testJNET(curdir);

        testPSIPRED(curdir);

        testTM(curdir);

        if (buildHomodeller)
        {
            buildHomodeller(curdir);
        }

        System.out.println("Testing complete package...");

        ProteusImpl proteus = new ProteusImpl(testSequence);

        proteus.run();
        ScoreAnnotatedSequence proteusResult = proteus.getResult();

        // Try a TM sequence
        System.out.println("Testing transmembrane sequence...");
        ProteusSettings settings = new ProteusSettings();
        settings.setSequence(testSequence2);
        settings.setOrganism(Organism.GRAM_POSITIVE);
        proteus = new ProteusImpl(settings);

        proteus.run();
        ScoreAnnotatedSequence proteusResult2 = proteus.getResult();

        System.out.println(proteusResult2);

        Homodeller.Result homodellerResult = null;

        if (buildHomodeller)
        {
            System.out.println("Testing homodeller...");
            Homodeller homodeller = new Homodeller();
            try {
            	homodellerResult = homodeller.run(testSequence);
            } catch(HomodellerException e) {
            	e.printStackTrace();
            }
        }

        boolean rv = proteusResult != null && proteusResult2 != null;

        if (buildHomodeller)
        {
            rv &= homodellerResult != null;
        }

        return rv;
/*
        RuntimeCatalog persistentProteusLogger = new RuntimeCatalog();

        MainPredictor mainPredictor = new MainPredictor(persistentProteusLogger);
        mainPredictor.run(testSequence, false);

        PredictionDB db = PredictionDBFactory.getDB(DBType.XML);

        ScoreAnnotatedSequence jnetResult = db.get(testSequence, "JNET");
        ScoreAnnotatedSequence psipredResult = db.get(testSequence, "PSIPRED");
        ScoreAnnotatedSequence proteusSecResult = db.get(testSequence, "PROTEUS");

        // test the overlay
        HomologCombiner combiner = HomologCombiner.getInstance();
        combiner.setLogger(persistentProteusLogger);
        combiner.searchAndOverlayHomologs(psipredResult, null);

        // verify that they are in fact in the database
        if (jnetResult != null && psipredResult != null && proteusSecResult != null)
        {
            finished = true;
        }

        return finished;*/
    }

    private static void formatDB(String db)
    {
        // find TM PDB - we're including it in the distro
        String tmpdb = config.getStringValue(db);
        if (new File(tmpdb).exists() && !new File(tmpdb + ".phr").exists())
        {
            // format the DB for them
            String formatdb = config.getStringValue("formatdb.command");
            runProgram(formatdb + " -i " + tmpdb);
        }
    }

    private static void findBioinformaticsPrograms(BufferedReader input)
            throws IOException
    {
        String blastDir = "";
        if (!findProgram("blastall -version", "BLAST"))
        {
            String path = "";
            if (LocateUtil.isLocateInstalled())
            {
                path = LocateUtil.locate("blastall");
            }

            int iteration = 0;
            do
            {
                errorMessage("BLASTALL is not in this system's PATH.  Where is it? [" + path + "]");
                String s = input.readLine();
                if (s != null && !s.equals(""))
                {
                    path = s;
                }
                config.setStringValue("blastall.command", path);
            }
            while (!new File(path).exists() && iteration++ < 3);
            blastDir = new File(path).getParent();
        }
        else
            config.setStringValue("blastall.command", "blastall");

        String blastCmd = "blastpgp";
        if (!blastDir.equals(""))
            blastCmd = blastDir + "/" + blastCmd;
        if (!findProgram(blastCmd + " -version", "PSI-BLAST"))
        {
            errorMessage("PSI-BLAST (blastpgp) is not in this system's PATH.  Where is it?");
            String s = input.readLine();
            config.setStringValue("blastpgp.command", s);
        }
        else
            config.setStringValue("blastpgp.command", blastCmd);

        blastCmd = "makemat";
        if (!blastDir.equals(""))
            blastCmd = blastDir + "/" + blastCmd;
        if (!findProgram(blastCmd, "makemat"))
        {
            errorMessage("makemat is not in this system's PATH.  Where is it?");
            String s = input.readLine();
            config.setStringValue("makemat.command", s);
        }
        else
            config.setStringValue("makemat.command", blastCmd);

        blastCmd = "formatdb";
        if (!blastDir.equals(""))
            blastCmd = blastDir + "/" + blastCmd;
        if (!findProgram(blastCmd, "formatdb"))
        {
            errorMessage("formatdb is not in this system's PATH.  Where is it?");
            String s = input.readLine();
            config.setStringValue("formatdb.command", s);
        }
        else
            config.setStringValue("formatdb.command", blastCmd);

        String clustalw = "clustalw";
        if (!findProgram(clustalw + " in.db", "clustalw"))
        {
            String path = "";

            if (LocateUtil.isLocateInstalled())
            {
                path = LocateUtil.locate("clustalw");
            }

            int iteration = 0;
            do
            {
                errorMessage("ClustalW is not in this system'path PATH.  Where is it? [" + path + "]");
                String s = input.readLine();
                if (s != null && !s.equals(""))
                {
                    path = s;
                }
                config.setStringValue("clustalw.command", path);
            }
            while (!new File(path).exists() && iteration++ < 3);
        }
        else
            config.setStringValue("clustalw.command", clustalw);
    }

    private static void findBasePrograms()
    {
        if (!findProgram("gcc"))
        {
            errorMessage("You are missing the GNU compiler (gcc). Please install before compiling Proteus.");
            System.exit(-1);
        }
        if (!findProgram("make"))
        {
            errorMessage("You are missing the GNU make utility (make). Please install before compiling Proteus.");
            System.exit(-1);
        }
        if (!findProgram("perl -version", "perl"))
        {
            errorMessage("You are missing PERL. Please install before compiling Proteus.");
            System.exit(-1);
        }
    }

    private static boolean ensureCurrentDir()
    {
        File file = new File("proteus.jar");
        if (!file.exists())
        {
            System.out.println("\n** ERROR: You must run setup from within the Proteus directory **\n");
            return false;
        }

        return true;
    }

    private static String queryDirectory(BufferedReader input, String question, String defaultDir)
            throws IOException
    {
        System.out.print(question);
        System.out.print("[" + defaultDir + "]");
        String s = input.readLine();
        if (s != null && s.length() != 0)
        {
            return s;
        }

        return defaultDir;
    }

    private static void testTM(String curdir)
    {

        String hmmerDir = curdir + "/src/hmmer";
        String hmmerCmd = hmmerDir + "/src/hmmpfam";
        config.setStringValue("hmmpfam", hmmerCmd);

        /* Give it 1 minute to complete this */
        if (!findProgram(hmmerCmd, "hmmpfam", 60 * 1000))
        {
            // run the configuration file
            runProgram(hmmerDir + "/configure", "hmmer configure", hmmerDir);

            errorMessage("Trying to make hmmpfam...");

            makeProgram(hmmerDir, true);
            if (findProgram(hmmerCmd, "hmmpfam"))
                System.out.println("hmmpfam successful");
            else
            {
                System.out.println("***Error*** [Cannot run hmmpfam]");
            }
        }

        config.setStringValue("tmhmm", curdir + "/src/TMHMM/bin/tmhmm");

        makeRunnable(new File(config.getStringValue("tmhmm")));
        makeRunnable(new File(curdir + "/src/TMHMM/bin/decodeanhmm"));
        makeRunnable(new File(curdir + "/src/TMHMM/bin/tmhmmformat.pl"));

        /* Give it 1 minute to complete this */
        if (!findProgram(config.getStringValue("tmhmm"), "tmhmm", 60 * 1000))
        {
            System.out.println("***Error*** [Cannot run tmhmm]");
        }

        // make tmbhunt
        String tmbhuntDir = curdir + "/src/tmbhunt";
        String tmbhuntCmd = tmbhuntDir + "/TMB_hunt.pl";

        config.setStringValue("tmbhunt", tmbhuntCmd);

        // check to see if it's there
        if (!findProgram("perl " + tmbhuntCmd, "tmbhunt", 60 * 1000))
        {
            errorMessage("Trying to make tmbhunt...");
            makeProgram(tmbhuntDir, true);
            if (findProgram("perl " + tmbhuntCmd, "tmbhunt"))
                System.out.println("tmbhunt successful");
            else
            {
                System.out.println("***Error*** [Cannot run tmbhunt]");
            }
        }
    }

    private static void testPSIPRED(String curdir)
    {
        String psipredCmd = curdir + "/src/psipred/src/psipred";

        config.setStringValue("psipred.command", psipredCmd);
        config.setStringValue("psipass.command", curdir + "/src/psipred/src/psipass2");
        config.setStringValue("psipred.datadir", curdir + "/src/psipred/data");
        //config.setStringValue("makemat.command", "makemat");

        makeProgram(curdir + "/src/psipred/src", true);
        if (!findProgram(psipredCmd, "psipred"))
        {
            errorMessage("Trying to make psipred...");
            makeProgram(curdir + "/src/psipred/src", true);
            if (findProgram(psipredCmd, "psipred"))
                System.out.println("PSIPRED successful");
            else
            {
                System.out.println("***Error*** [Cannot run PSIPRED]");
            }
        }
    }

    private static void testJNET(String curdir)
    {
        /**------------------ Test JNet ----------------*/
        String jnetCmd = curdir + "/src/jnet/src/jnet";
        config.setStringValue("jnet.command", jnetCmd);

        config.setStringValue("parse_psi.command", curdir + "/src/jnet/parse_psi");
        config.setStringValue("gethmm.command", curdir + "/src/jnet/gethmm");
        config.setStringValue("getpssm.command", curdir + "/src/jnet/getpssm");
        config.setStringValue("msf2jnet.command", curdir + "/src/jnet/msf2jnet");

        makeRunnable(new File(config.getStringValue("parse_psi.command")));
        makeRunnable(new File(config.getStringValue("gethmm.command")));
        makeRunnable(new File(config.getStringValue("getpssm.command")));
        makeRunnable(new File(config.getStringValue("msf2jnet.command")));
        makeRunnable(new File(curdir + "/src/jnet/hmmbuild"));
        makeRunnable(new File(curdir + "/src/jnet/hmmconvert"));

        makeProgram(curdir + "/src/jnet/src", true);
        if (!findProgram(jnetCmd, "jnet"))
        {
            errorMessage("Trying to make jnet...");
            makeProgram(curdir + "/src/jnet/src", true);
            if (findProgram(jnetCmd, "jnet"))
                System.out.println("JNET successful");
            else
            {
                System.out.println("***Error*** [Cannot run JNET]");
            }
        }
    }

    private static void buildHomodeller(String curdir)
    {
        String homodellerCmd = curdir + "/src/homodeller/homodeller";

        config.setStringValue("homodeller", homodellerCmd);
        config.setStringValue("loop.db", curdir + "/data/looppdb.new");
        config.setStringValue("side.db", curdir + "/data/side.db");
        config.setStringValue("phipsi", curdir + "/src/homodeller/phipsi");
        config.setStringValue("simplex", curdir + "/src/homodeller/simplex");

        if (!findProgram(homodellerCmd, "homodeller"))
        {
            errorMessage("Trying to make homodeller...");
            makeProgram(curdir + "/src/homodeller", true);
            if (findProgram(homodellerCmd, "homodeller"))
            {
                System.out.println("Homodeller successful");
            }
            else
            {
                System.out.println("***Error*** [Cannot make Homodeller]");
            }
        }

        if (!findProgram(config.getStringValue("simplex"), "simplex"))
        {
            makeProgram(curdir + "/src/homodeller", true);
            if (findProgram(config.getStringValue("simplex"), "simplex"))
            {
                System.out.println("Simplex successful");
            }
            else
            {
                System.out.println("***Error*** [Cannot run simplex]");
            }
        }
        

        makeProgram(curdir + "/src/gafolder", true);
        if (findProgram(homodellerConfig.getStringValue("gafolder"), "gafolder"))
        {
            System.out.println("Gafolder successful");
        }
        else
        {
            System.out.println("***Error*** [Cannot run Gafolder]");
        }

        // seqsee.db needs a copy in the homodeller directory
        // runProgram("ln -s " + curdir + "/data/seqsee.db src/homodeller/");
    }

    /**
     * Finds a database file
     *
     * @param input
     * @param configKey
     * @param prompt
     * @param runFormatDB
     * @return true if the DB file is found, false otherwise
     * @throws IOException
     */
    private static boolean findDatabase(BufferedReader input, String configKey, String prompt, boolean runFormatDB) throws IOException
    {
        File f;
        int i = 0;
        while (!(f = new File(config.getStringValue(configKey, ""))).exists() || !f.isFile())
        {
            System.out.print(prompt + " [" + config.getStringValue(configKey, "") + "]? ");
            String s = input.readLine();
            if (s != null && !s.equals(""))
            {
                config.setStringValue(configKey, s);

                File file = new File(s);
                if (!file.isFile())
                {
                    System.out.println("That file does not exist or is a directory.  Please enter the database file.");
                    continue;
                }

                if (runFormatDB)
                {
                    if (new File(s).exists() && !new File(s + ".phr").exists())
                    {
                        String formatdb = config.getStringValue("formatdb.command");
                        runProgram(formatdb + " -i " + s);
                    }
                }
            }

            if (i++ > 20)
            {
                // screw it
                return false;
            }
        }

        return true;
    }

    private static String getCurDir()
    {
        return new File("").getAbsolutePath();
    }

    private static List<Sequence> testBLAST(Sequence sequence)
    {

        System.out.print("Testing BLAST-P...");

        // try to see if Blast works
        BlastSettings settings = new BlastSettings();
        settings.setSequence(sequence);
        settings.setBlastType(BlastSettings.BLASTTYPE.BLASTP);

        // test it on the NR set to ensure coverage
        settings.setDatabase(config.getStringValue("blast.dbname"));

        BlastRunner blast = new BlastRunner(settings);
        blast.run();
        blast.cleanupTempFiles();

        List<Sequence> blastResult = blast.getUnalignedResult();
        if (blastResult != null)
        {
            System.out.println("Passed");
        }


        System.out.print("Testing PSI-BLAST...");
        blast = new BlastRunner(settings);
        settings.setBlastType(BlastSettings.BLASTTYPE.PSIBLAST);

        // test it on the PDB set to ensure coverage
        settings.setDatabase(config.getStringValue("PDB"));
        blast.run();
        blast.cleanupTempFiles();

        blastResult = blast.getUnalignedResult();
        if (blastResult != null)
        {
            System.out.println("Passed");
        }
        return blastResult;
    }

    private static void testXAlign(String curdir, Sequence sequence, List<Sequence> blastResult) throws IOException
    {
        /**------------------ Begin Test XAlign ----------------*/
        boolean XAlignOK = false;
        boolean secondTime = false;

        File dir = new File(curdir + "/src/xalign/");
        String[] envp = new String[]{"PWD=" + dir.getAbsolutePath()};
        Runtime.getRuntime().exec("perl makeParms.pl", envp, dir);
        
        //makeProgram(curdir + "/src/xalign/", true);
        while (!secondTime)
        {
            String cmd = curdir + "/src/xalign/xalign_x";
            if (!(new File(cmd).exists()))
            {
                errorMessage("Trying to make XAlign...");
                makeProgram(curdir + "/src/xalign/", true);

                String actualCmd = curdir + "/src/xalign/run_xalign";
                if (findProgram(actualCmd, "XAlign", 3000))
                    XAlignOK = true;
                config.setStringValue("xalign.command", actualCmd);

            }
            else
            {
                XAlignOK = true;
            }

            config.setStringValue("xalign.command", curdir + "/src/xalign/run_xalign");

            if (XAlignOK)
            {
                System.out.print("Testing XAlign...");
                try
                {
                    XAlignRunner xalign = new XAlignRunner(sequence, blastResult);
                    xalign.setAddToDB(false);
                    xalign.setKeepOutput(false);
                    xalign.setUseCache(false);
                    xalign.run();
                    Alignment alignment = xalign.getAlignment();
                    if (xalign.isFinished() && alignment != null)
                    {
                        System.out.println("Passed");
                        break;
                    }
                    else
                    {
                        secondTime = xalignError(curdir);
                    }
                }
                catch (Exception e)
                {
                    secondTime = xalignError(curdir);
                }
            }
        }

        /**------------------ Done XAlign ----------------*/}

    private static boolean xalignError(String curdir)
    {
        boolean secondTime;
        System.out.println("***Error*** [Cannot run XAlign]");
        makeProgram(curdir + "/src/xalign/", true);
        secondTime = true;
        return secondTime;
    }

    private static void setupDefaultValues()
    {
        if (!config.contains("blast.evalue"))
            config.setStringValue("blast.evalue", "0.001");

        if (!config.contains("num_alignments"))
            config.setStringValue("num_alignments", "50");

        config.setStringValue("model1", "model.1");
        config.setStringValue("model2", "model.2");
        config.setStringValue("finalModel", "model.3");
        config.setStringValue("logging", "ERROR");

        String curdir = getCurDir();

        config.setStringValue("GRAM+", curdir + "/data/HMMs/GRAM+.hmm");
        config.setStringValue("GRAM-", curdir + "/data/HMMs/GRAM-.-6seqs.hmm");
        config.setStringValue("EUK", curdir + "/data/HMMs/EUKARYOTIC.hmm");

        config.setStringValue("TMPDB", curdir + "/data/seqsee_pdbtm.db");
        config.setStringValue("blast.TMPDB", curdir + "/data/seqsee_pdbtm.faa");

        config.setStringValue("BBDB", curdir + "/data/seqsee_pdbtm.db");
        config.setStringValue("blast.BBDB", curdir + "/data/seqsee_pdbtm.faa");

        config.setStringValue("SPDB.euk", curdir + "/data/seqsee_spdb_euk.db");
        config.setStringValue("blast.SPDB.euk", curdir + "/data/seqsee_spdb_euk.faa");

        config.setStringValue("SPDB.gram+", curdir + "/data/seqsee_spdb_gram+ve.db");
        config.setStringValue("blast.SPDB.gram+", curdir + "/data/seqsee_spdb_gram+ve.faa");

        config.setStringValue("SPDB.gram-", curdir + "/data/seqsee_spdb_gram-ve.db");
        config.setStringValue("blast.SPDB.gram-", curdir + "/data/seqsee_spdb_gram+ve.faa");

        config.setStringValue("seqsee_db", curdir + "/data/seqsee.db");
        config.setStringValue("PDB", curdir + "/data/seqsee.faa");
    }

    private static void setupHomodellerConfig()
    {
        homodellerConfig = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");

        String curdir = getCurDir();

        homodellerConfig.setStringValue("loop.db", curdir + "/data/looppdb.new");
        homodellerConfig.setStringValue("side.db", curdir + "/data/side.db");
        homodellerConfig.setStringValue("phipsi", curdir + "/src/homodeller/phipsi");
        homodellerConfig.setStringValue("simplex", curdir + "/src/homodeller/simplex");
        homodellerConfig.setStringValue("homodeller", curdir + "/src/homodeller/homodeller");
        homodellerConfig.setStringValue("pdb_dir", curdir + "/pdb");
        homodellerConfig.setStringValue("gafolder", curdir + "/src/gafolder/gafolder");
        homodellerConfig.setStringValue("pepmake", curdir + "/src/gafolder/pepmake");
        homodellerConfig.setStringValue("energyMinimizer", "gafolder");

        homodellerConfig.writeConfig();
    }

    private static boolean makeProgram(String directory, boolean clean)
    {
        try
        {
            Process proc;

            if (clean)
            {
                proc = Runtime.getRuntime().exec("make clean", null, new File(directory));
                try
                {
                    // any error message?
                    StreamHandler errorReader = new
                            StreamHandler(proc.getErrorStream(), "ERROR", System.err);

                    // any output?
                    StreamHandler outputReader = new
                            StreamHandler(proc.getInputStream(), "OUTPUT", System.err);

                    // kick them off
                    errorReader.start();
                    outputReader.start();

                    proc.waitFor();
                }
                catch (InterruptedException e)
                {
                    e.printStackTrace();
                }

            }

            // Do the real make now
            proc = Runtime.getRuntime().exec("make", null, new File(directory));

            try
            {
                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(proc.getErrorStream(), "ERROR", System.err);

                // any output?
                StreamHandler outputReader = new
                        StreamHandler(proc.getInputStream(), "OUTPUT", System.err);

                // kick them off
                errorReader.start();
                outputReader.start();

                proc.waitFor();
                //Thread.sleep(3000);
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }
        }
        catch (IOException e)
        {
            System.out.println("failed.");
            return false;
        }
        System.out.println("success.");
        return true;
    }

    /**
     * Basically runs make in the specified directory
     *
     * @param directory Directory to run make command make
     */
    private static boolean makeProgram(String directory)
    {
        return makeProgram(directory, false);
    }

    /**
     * Do a chmod on this file to make it executable
     *
     * @param file File to make executable
     */
    private static void makeRunnable(File file)
    {
        try
        {
            Runtime.getRuntime().exec("chmod u+x " + file.getAbsolutePath(), null);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }


    private static void errorMessage(String s)
    {
        System.out.println(s);
    }

    private static boolean findProgram(String cmd)
    {
        return findProgram(cmd, null);
    }

    /**
     * @param cmd          - command to run to check existence
     * @param name         - Program name output to stdout (optional).  If null, output 'cmd'
     * @param timeout      - Time in millis to wait for program
     * @param checkExitVal - If we should check the exit value of the program to make sure it's 0
     * @return true if command passes
     */
    private static boolean findProgram(String cmd, String name, final int timeout, boolean checkExitVal)
    {

        if (name == null)
            System.out.print("Checking for " + cmd + "...");
        else
            System.out.print("Checking for " + name + "...");
        try
        {
            final Process proc = Runtime.getRuntime().exec(cmd);

            int exitVal = 0;
            try
            {
                /*InputStream is = proc.getErrorStream();
                int i = 0;
                /*StringBuilder sb = new StringBuilder();
                                
         /*       while ((i = is.read()) != -1)
                {
                    //System.err.print((char) i);
                    //sb.append((char) i);
                }*/
                if (timeout == -1)
                    exitVal = proc.waitFor();
                else
                {
                    Thread thread = new Thread()
                    {
                        public void run()
                        {
                            try
                            {
                                proc.waitFor();
                            }
                            catch (InterruptedException e)
                            {
                                e.printStackTrace();
                            }
                        }
                    };
                    thread.start();

                    thread.join(timeout);
                }

                if (checkExitVal)
                {
                    if (exitVal != 0)
                    {
                        System.out.println("failed(" + exitVal + ").");
                        return false;
                    }
                }
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            /*if ( exitVal != 0 ) {
                System.out.println("failed.");
                return false;
            } */
        }
        catch (IOException e)
        {
            System.out.println("failed.");
            return false;
        }
        System.out.println("passed.");
        return true;
    }

    /**
     * @param cmd  - command to run to check existence
     * @param name - Program name output to stdout (optional).  If null, output 'cmd'
     * @return true if command passes
     */
    private static boolean findProgram(String cmd, String name)
    {
        return findProgram(cmd, name, -1);
    }

    /**
     * @param cmd     - command to run to check existence
     * @param name    - Program name output to stdout (optional).  If null, output 'cmd'
     * @param timeout - Time in millis to wait for program
     * @return if command passes
     */
    private static boolean findProgram(String cmd, String name, int timeout)
    {
        return findProgram(cmd, name, timeout, false);
    }

    /**
     * @param cmd - command to run to check existence
     * @return
     */
    private static boolean runProgram(String cmd)
    {
        return runProgram(cmd, null, null);
    }

    /**
     * @param cmd  - command to run to check existence
     * @param name - Program name output to stdout (optional).  If null, output 'cmd'
     * @return true is command passes
     */
    private static boolean runProgram(String cmd, String name, String dir)
    {
        if (name == null)
            System.out.print("Running " + cmd + "...");
        else
            System.out.print("Running " + name + "...");
        try
        {
            Process proc;
            if (dir != null)
            {
                proc = Runtime.getRuntime().exec(cmd, null, new File(dir));
            }
            else
            {
                proc = Runtime.getRuntime().exec(cmd);
            }

            try
            {
                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(proc.getErrorStream(), "ERROR", System.err);

                // any output?
                StreamHandler outputReader = new
                        StreamHandler(proc.getInputStream(), "OUTPUT", System.err);

                // kick them off
                errorReader.start();
                outputReader.start();

                proc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }
        }
        catch (IOException e)
        {
            System.out.println("failed.");
            return false;
        }
        System.out.println("done.");
        return true;
    }

    /**
     * Wraps the UNIX program "locate"
     */
    static class LocateUtil
    {
        private static Boolean isLocateInstalled = null;

        /**
         * Uses the UNIX program "locate" to try to find a certain program
         *
         * @param program
         * @return the path to a program, or null if locate is not installed
         */
        public static String locate(String program)
        {
            try
            {
                String cmd = "locate " + program;
                Process proc = Runtime.getRuntime().exec(cmd);

                try
                {
                    // any error message?
                    StreamHandler errorReader = new
                            StreamHandler(proc.getErrorStream(), "ERROR", System.err);

                    // any output?
                    StreamHandler outputReader = new
                            StreamHandler(proc.getInputStream(), "OUTPUT");

                    // kick them off
                    errorReader.start();
                    outputReader.start();

                    proc.waitFor();

                    BufferedReader reader = new BufferedReader(new StringReader(outputReader.getInputString()));
                    String line;
                    while ((line = reader.readLine()) != null)
                    {
                        if (line.endsWith(program))
                        {
                            return line;
                        }
                    }
                }
                catch (InterruptedException e)
                {
                    e.printStackTrace();
                }
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }

            return null;
        }

        public static boolean isLocateInstalled()
        {
            if (isLocateInstalled == null)
            {
                isLocateInstalled = findProgram("locate --version", "locate");
            }

            return isLocateInstalled;
        }
    }
}
