package ca.ualberta.proteus.program;

import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.program.predictSPTM.Organism;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.CompletionListener;
import ca.ualberta.proteus.util.Consts;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.util.db.CreateDatabase;
import ca.ualberta.proteus.util.io.StructureDB;
import ca.ualberta.proteus.web.ProteusFormatter;
import com.martiansoftware.jsap.*;
import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.structure.Structure;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: May 10, 2007
 * Time: 9:56:57 PM
 */
public class Proteus
{

    public static final String VERSION = "1.1";
    public static final String FILE_FLAG = "file";
    public static final String OUTPUT_FILE_FLAG = "output";
    public static final String ONLY_TM_FLAG = "only_tm";
    public static final String METHOD_FLAG = "method";
    public static final String VERBOSE = "extended";
    private static final String HOMODELLER = "homodeller";
    private static final String ENERGY_MIN = "energy_min";
    private static final String CHECK_DATABASES = "check_db";
    private static final String CONFIG_FLAG = "config";
    private static final String ORGANISM_FLAG = "ORGANISM_FLAG";
	private static final String IDENTITY_FLAG = "identity";

    public static void main(String[] args) throws BioException, IOException, JSAPException
    {
        JSAPResult options = parseCommandLine(args);

        if (options != null)
        {
            runFromCommandLine(options);
        }

    }

    /**
     * Parses the command line parameters
     *
     * @param args
     * @return JSAPResult containing the options if parsing was successful, null otherwise
     * @throws com.martiansoftware.jsap.JSAPException
     *
     */
    public static JSAPResult parseCommandLine(String[] args) throws JSAPException
    {
        final String METHOD_MSG = "Use this option to limit Proteus to using certain methods.  Valid options are: \n" +
                /*"p, psipred, or 1 = PSIPRED\n" +
                "j, jnet, or 2 = JNET\n" +
                "f, fast or 3 = PSIPRED + JNET + CONSENSUS\n" +
                "a, all or 4 = PSIPRED + JNET + TRANSEC + CONSENSUS\n" +
                */
                "- psipred\n" +
                "- jnet\n" +
                "- fast = PSIPRED + JNET + CONSENSUS\n" +
                "- all = PSIPRED + JNET + TRANSEC + CONSENSUS\n" +
                "e.g. '-method psipred' = run PSIPRED\n" +
                "Not including this option will run all methods, as well as the consensus (original Proteus)\n";

        final String ORGANISM_MSG = "Set the origanism for signal peptide prediction.\n" +
                "1 = GRAM+\n" +
                "2 = GRAM-\n" +
                "3 = EUKARYOTE";

        SimpleJSAP options = new SimpleJSAP(
                "Proteus",
                "Comprehensive Secondary Structure Predictions, Local Version " + VERSION,
                new Parameter[]{
                        new FlaggedOption(OUTPUT_FILE_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'o', "output",
                                "The output file for the predictions.  If a PDB file is generated, it will have the same name with a '.pdb' extension."),
                        new FlaggedOption(METHOD_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'm', "method",
                                METHOD_MSG),
                        new Switch(ONLY_TM_FLAG, 't', "only-tm",
                                "If you only want to run transmembrane predictions"),
                        new FlaggedOption(ORGANISM_FLAG, JSAP.INTEGER_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NO_SHORTFLAG,
                                "organism", ORGANISM_MSG),
                        new Switch(VERBOSE, 'v', "verbose",
                                "Verbose output.  Outputs all possible data to the output files. This is disabled by default."),
                        new Switch(HOMODELLER, 'h', "homology-modelling",
                                "Run homology modelling."),
                        new Switch(ENERGY_MIN, 'e', "energy-minimization",
                                "Run energy minimization."),
                        new Switch(CHECK_DATABASES, JSAP.NO_SHORTFLAG, "check-databases",
                                "Run energy minimization."),
						new FlaggedOption(IDENTITY_FLAG, JSAP.DOUBLE_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'i', "sequence-identity",
									  "Filters the sequences above specified identity."),  
					    new UnflaggedOption(FILE_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY,
                                "The name of the FASTA file containing the query sequence(s)."),
                        new FlaggedOption(CONFIG_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'c', "config",
                                "Path to a valid proteus config file"),

                }
        );

        final JSAPResult config = options.parse(args);

        // if there was a mistake, print the usage and kill the program
        if (options.messagePrinted())
        {
            printUsage(options);
            return null;
        }

        if (config.getBoolean(CHECK_DATABASES, false))
        {
            CreateDatabase.checkDatabases();
            return null;
        }

        if (!new File(config.getString(FILE_FLAG)).exists())
        {
            System.err.println("Input file does not exist\n");
            printUsage(options);
            return null;
        }

        // Set the specified config
        if (config.contains(CONFIG_FLAG))
        {
            Config.setDefaultFile(config.getString(CONFIG_FLAG));
        }

        if (config.contains(METHOD_FLAG))
        {
            final String s = config.getString(METHOD_FLAG);
            if (!s.equals("psipred") && !s.equals("jnet") && !s.equals("fast") && !s.equals("all"))
            {
                System.err.println("Invalid 'method' option \n");
                printUsage(options);
                return null;
            }
        }

        if (config.contains(OUTPUT_FILE_FLAG))
        {
            // try it, make sure we can write output
            try
            {
                new FileOutputStream(config.getString(OUTPUT_FILE_FLAG));
            }
            catch (FileNotFoundException e)
            {
                System.err.println("The file '" + config.getString(OUTPUT_FILE_FLAG) + "' is not writable");
                printUsage(options);
                return null;
            }
        }

        if (config.contains(ORGANISM_FLAG))
        {
            final int flag = config.getInt(ORGANISM_FLAG);
            if (flag < 1 || flag > 3)
            {
                System.err.println(config.getInt(ORGANISM_FLAG) + "' is not a valid organism number");
                printUsage(options);
                return null;
            }
        }

        return config;
    }

    public static void printUsage(SimpleJSAP options)
    {
        System.err.println("Usage: ");
        System.err.println("  " + options.getUsage());
        System.err.println();
        System.err.println(options.getHelp());
    }

    public static void runFromCommandLine(JSAPResult options)
            throws IOException
    {
        List<Sequence> sequences = null;
        File file = new File(options.getString(FILE_FLAG));
        try
        {
            // parse the FASTA file, get a list of sequences
            sequences = ProteusTools.parseFasta(file);
        }
        catch (BioException e)
        {
            System.out.println("** ERROR: The file \"" + file.getName() + "\" does not appear to be a FASTA file **");
            return;
        }

        final List<ScoreAnnotatedSequence> predictions = new ArrayList<ScoreAnnotatedSequence>();
        final List<RuntimeCatalog> loggers = new ArrayList<RuntimeCatalog>();

        final ProteusSettings settings = new ProteusSettings();
        settings.setMethod(getMethod(options));
        boolean runHomodeller = options.getBoolean(HOMODELLER, false);
        settings.setRunHomodeller(runHomodeller);
        settings.setRunOnlyTM(options.getBoolean(ONLY_TM_FLAG, false));

        if (runHomodeller)
        {
            settings.setRunEnergyMinimization(options.getBoolean(ENERGY_MIN, false));
        }

        if (options.getBoolean(ONLY_TM_FLAG, false))
        {
            System.out.println("*** Running only transmembrane predictions ***");
        }
		
		if (options.contains(IDENTITY_FLAG))
        {
            settings.setIdentityThreshold( (Double) options.getObject(IDENTITY_FLAG));
        }

        if (options.contains(ORGANISM_FLAG))
        {
            final int flag = options.getInt(ORGANISM_FLAG);
            if (flag >= 1 && flag <= 3)
            {
                settings.setOrganism(Organism.values()[flag - 1]);
            }
            else
            {
                System.err.println(options.getInt(ORGANISM_FLAG) + "' is not a valid organism number");
                return;
            }
        }

        // run proteus on each sequence
        for (Sequence sequence : sequences)
        {
            settings.setSequence(sequence);

            final ProteusImpl proteus = new ProteusImpl(settings);

            Annotation annotation = sequence.getAnnotation();
            String seqName;
            if (annotation != null && annotation.containsProperty(Consts.SEQUENCE_NAME))
            {
                seqName = (String) annotation.getProperty(Consts.SEQUENCE_NAME);
            }
            else
            {
                // must have used BioJava's parsing??
                seqName = sequence.getName();
            }

            proteus.getLogger().logObject(Consts.SEQUENCE_NAME, seqName);

            proteus.addCompletionListener(new CompletionListener()
            {
                public void onComplete(boolean successful, String message)
                {
                    RuntimeCatalog logger = proteus.getLogger();

                    ScoreAnnotatedSequence sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
                    if (sequence == null)
                    {
                        sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL);
                    }

                    switch (settings.getMethod())
                    {
                        case PSIPRED:
                            sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.PSIPRED);
                            break;
                        case JNET:
                            sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.JNET);
                            break;
                        case TRANSSEC:
                            sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.TRANSSEC);
                            break;
                        case ALL:
                        case PSIPRED_JNET:
                        default:
                            break;
                    }

                    System.out.println("Final sequence with homology:");
                    System.out.println(sequence);

                    predictions.add(sequence);
                    loggers.add(logger);
                }
            });

            proteus.run();
        }

        writeOutput(options, loggers, predictions);
    }

    private static void writeOutput(JSAPResult options, List<RuntimeCatalog> loggers, List<ScoreAnnotatedSequence> predictions) throws FileNotFoundException
    {
        PrintStream stream;
        File pdbFile;

        // was an output option specified?
        if (options.contains(OUTPUT_FILE_FLAG))
        {
            File outputFile = new File(options.getString(OUTPUT_FILE_FLAG));
            pdbFile = new File(options.getString(OUTPUT_FILE_FLAG) + ".pdb");

            try
            {
                stream = new PrintStream(outputFile);
            }
            catch (FileNotFoundException e)
            {
                System.err.println("The file '" + outputFile + "' is not writable");
                return;
            }
        }
        else
        {
            System.err.println("### BEGIN OUTPUT ###");

            stream = System.out;
            pdbFile = new File("1.pdb");
        }

        writeOutput(stream, options, loggers, predictions);

        if (options.getBoolean(HOMODELLER, false))
        {
            writePDBOutput(pdbFile, options, loggers);
        }

    }

    private static void writePDBOutput(File pdbFile, JSAPResult options, List<RuntimeCatalog> loggers)
    {
        try
        {
            PrintWriter writer = new PrintWriter(pdbFile);

            writePDBFile(loggers, writer);
        }
        catch (FileNotFoundException e)
        {
            System.err.println(pdbFile.getAbsolutePath() + " is not writable");
        }
    }

    public static void writePDBFile(RuntimeCatalog logger, PrintWriter writer)
    {
        List<Structure> models = ProteusTools.getFinishedModels(logger);
        for (Structure model : models)
        {
            writePDBFileForModel(logger, writer, model);
        }
    }

    public static void writePDBFileForModel(RuntimeCatalog logger, PrintWriter writer, Structure model)
    {
        List<GappedStructureAnnotatedSequence> allAlignedHomologs = (List<GappedStructureAnnotatedSequence>) logger.getLogObject(Consts.ALL_ALIGNED_HOMOLOG_LIST);
        List<GappedStructureAnnotatedSequence> alignedTMSequences = (List<GappedStructureAnnotatedSequence>) logger.getLogObject(Consts.TM_HOMOLOGS);

        boolean hasNoHomolog = allAlignedHomologs == null || (allAlignedHomologs.size() == 0);
        hasNoHomolog &= alignedTMSequences == null || alignedTMSequences.size() == 0;

        String seqName = (String) logger.getLogObject(Consts.SEQUENCE_NAME);

        if (!hasNoHomolog)
        {
            String pdbText = model.toPDB();

            SimpleDateFormat format = new SimpleDateFormat("DD-MMM-yy");
            String shortName = seqName != null ? seqName.substring(0, Math.min(seqName.length(), 15)) : "";

            writer.println(String.format("REMARK    %-40s%s   %s", "PROTEUS2", format.format(new Date()), shortName));
            writer.println("REMARK   1 " + seqName);
            writer.println(pdbText);
        }
        else
        {
            String msg = seqName + " has no homolog > 35% homologous, so a structure could not be produced";
            String s = "REMARK   1 " + msg;
            writer.println(s);
            System.out.println(msg);
        }
    }

    public static void writePDBFile(List<RuntimeCatalog> loggers, PrintWriter writer)
    {
        for (RuntimeCatalog logger : loggers)
        {
            writePDBFile(logger, writer);
        }
    }

    private static void writeOutput(PrintStream stream, JSAPResult options, List<RuntimeCatalog> loggers, List<ScoreAnnotatedSequence> predictions)
    {
        assert loggers.size() == predictions.size();

        for (int i = 0; i < loggers.size(); i++)
        {
            RuntimeCatalog logger = loggers.get(i);
            ScoreAnnotatedSequence sequence = predictions.get(i);

            // This is all taken from final.jsp, so if that changes this has to too

            ScoreAnnotatedSequence homolOverlay = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
            List<GappedStructureAnnotatedSequence> alignedHomologs = (List<GappedStructureAnnotatedSequence>) logger.getLogObject(Consts.ALIGNED_HOMOLOG_LIST);
            GappedStructureAnnotatedSequence alignedQuery = (GappedStructureAnnotatedSequence) logger.getLogObject(Consts.ALIGNED_QUERY);

            StructureAnnotatedSequence signalPeptide = (StructureAnnotatedSequence) logger.getLogObject(Consts.SIGNAL_PEPTIDE);
            StructureAnnotatedSequence transmembrane = (StructureAnnotatedSequence) logger.getLogObject(Consts.TRANSMEMBRANE);
            StructureAnnotatedSequence finalStructure = (StructureAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);

            if (logger.getLogObject(Consts.TRANSMEMBRANE_HOMOLOG) != null)
            {
                transmembrane = (StructureAnnotatedSequence) logger.getLogObject(Consts.TRANSMEMBRANE_HOMOLOG);
            }

            // From TMPDB search
            Boolean isBBTM = (Boolean) logger.getLogObject(Consts.IS_BBTM);
            Boolean isTM = (Boolean) logger.getLogObject(Consts.IS_TM);
            String tmStr = transmembrane != null ? transmembrane.getStructureAnnotation().seqString() : null;

            isTM = isTM || (tmStr != null && tmStr.toLowerCase().indexOf('t') != -1);

            // Define if this is a transmembrane region by if this contains a "t"
            isTM = finalStructure.getStructureAnnotation().seqString().toLowerCase().indexOf('t') != -1;

            boolean hasSignalPeptide = signalPeptide != null;
            hasSignalPeptide = hasSignalPeptide && signalPeptide.getStructureAnnotation().seqString().toLowerCase().contains("s");

            // Define if it is a beta-barrell by if the final sequence contains a "b", as well as if TMBHunt identified it as such
            boolean hasBetaBarrelRegion = finalStructure.getStructureAnnotation().seqString().toLowerCase().indexOf('b') != -1;

            String s;

            // Print the complete sequence

            if (homolOverlay != null)
            {
                if (alignedHomologs != null)
                {
                    s = ProteusFormatter.printFormattedSequenceWithHomologyIndicatorPlainText(homolOverlay, alignedQuery, alignedHomologs, false, isBBTM);
                }
                else
                {
                    s = ProteusFormatter.printFormattedSequencePlainText(homolOverlay, false, isBBTM, true);
                }
            }
            else
            {
                sequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL);
                if (sequence != null)
                {
                    s = ProteusFormatter.printFormattedSequencePlainText(sequence, false, isBBTM, true);
                }
                else
                {
                    s = "NULL";
                }
            }

            String seqName = (String) logger.getLogObject(Consts.SEQUENCE_NAME);

            stream.println(">" + seqName);
            stream.print(s);

            // Print the signal peptide

            if (hasSignalPeptide)
            {
                s = ProteusFormatter.printFormattedSequencePlainText(signalPeptide.subSequence(1, 60), isTM, isBBTM, true);
            }
            else
            {
                s = "No signal peptide found\n";
            }

            stream.println(">" + seqName + " (Signal peptide - first 60 residues)");
            stream.println(s);

            if (isTM)
            {
                String tmStruct = transmembrane.getStructureAnnotation().seqString();
                tmStruct = tmStruct.replaceAll("[cheCHE]", "-");

                try
                {
                    // Create new sequence with all but the T's blocked out
                    transmembrane = new StructureAnnotatedSequence(transmembrane.seqString(), tmStruct, transmembrane.getName());
                }
                catch (IllegalSymbolException e)
                {
                    e.printStackTrace();
                }

                // If isTM is true isBBTM will be false
                s = ProteusFormatter.printFormattedSequencePlainText(transmembrane, isTM, false, true);
            }
            else if (!hasBetaBarrelRegion)
            {
                s = "No transmembrane region found";
            }
            else
            {
                // Show nothing??
                s = "";
            }

            stream.println(">" + seqName + " (Predicted Transmembrane Structure)");
            stream.println(s);

            if (options.getBoolean(VERBOSE, false))
            {
                // Extended output
                writeExtendedOutput(stream, options, logger);
            }

        }
    }

    /**
     * Writes more output
     *
     * @param stream
     * @param options
     * @param logger
     */
    private static void writeExtendedOutput(PrintStream stream, JSAPResult options, RuntimeCatalog logger)
    {
        ProteusSettings settings = (ProteusSettings) logger.getLogObject(Consts.SETTINGS);
        MethodSettings method = settings.getMethod();

        ScoreAnnotatedSequence combined = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL);
        ScoreAnnotatedSequence psipred = (ScoreAnnotatedSequence) logger.getLogObject(Consts.PSIPRED);
        ScoreAnnotatedSequence jnet = (ScoreAnnotatedSequence) logger.getLogObject(Consts.JNET);
        ScoreAnnotatedSequence homolOverlay = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
        ScoreAnnotatedSequence proteusSequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.TRANSSEC);
        if (proteusSequence == null)
        {
            proteusSequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.PROTEUS);
        }

        String seqName = (String) logger.getLogObject(Consts.SEQUENCE_NAME);

        String s;
        if (method == MethodSettings.ALL || method == MethodSettings.PSIPRED || method == MethodSettings.PSIPRED_JNET)
        {
            s = ProteusFormatter.printFormattedSequencePlainText(psipred, false, false, true);

            stream.println(">" + seqName + "(PSIPRED)");
            stream.println(s);
        }

        if (method == MethodSettings.ALL || method == MethodSettings.JNET || method == MethodSettings.PSIPRED_JNET)
        {
            s = ProteusFormatter.printFormattedSequencePlainText(jnet, false, false, true);

            stream.println(">" + seqName + "(JNET)");
            stream.println(s);
        }

        if (method == MethodSettings.ALL || method == MethodSettings.TRANSSEC)
        {
            s = ProteusFormatter.printFormattedSequencePlainText(proteusSequence, false, false, true);

            stream.println(">" + seqName + "(Internal)");
            stream.println(s);
        }

        if (method == MethodSettings.ALL || method == MethodSettings.PSIPRED_JNET)
        {
            s = ProteusFormatter.printFormattedSequencePlainText(combined, false, false, true);

            stream.println(">" + seqName + "(Jury of Experts)");
            stream.println(s);
        }


        List<GappedStructureAnnotatedSequence> alignedHomologsClone = extractHomologs(logger);
        if (alignedHomologsClone.size() > 0)
        {
            s = ProteusFormatter.printAlignmentPlainText((GappedStructureAnnotatedSequence) logger.getLogObject(Consts.ALIGNED_QUERY), alignedHomologsClone, false, false);
        }
        else
        {
            s = "No homologs were found";
        }

        stream.println(">" + seqName + "(Alignment)");
        stream.println(s);
    }

    /**
     * TODO remove into another class since this code is duplicated in a few places
     *
     * @param logger
     * @return
     */
    private static List<GappedStructureAnnotatedSequence> extractHomologs(RuntimeCatalog logger)
    {
        List<GappedStructureAnnotatedSequence> alignedTMSequences = (List<GappedStructureAnnotatedSequence>) logger.getLogObject(Consts.TM_HOMOLOGS);
        List<GappedStructureAnnotatedSequence> alignedHomologs = (List<GappedStructureAnnotatedSequence>) logger.getLogObject(Consts.ALIGNED_HOMOLOG_LIST);

        List<GappedStructureAnnotatedSequence> alignedHomologsClone = new ArrayList<GappedStructureAnnotatedSequence>();
        if (alignedHomologs != null)
        {
            alignedHomologsClone.addAll(alignedHomologs);
        }


        for (int i = alignedHomologsClone.size() - 1; i >= 0; i--)
        {
            GappedStructureAnnotatedSequence seq = alignedHomologsClone.get(i);
            if (seq.getAnnotation().containsProperty("isTM"))
            {
                alignedHomologsClone.remove(seq);
            }

            // remove any from the list the are supposed to be transmembrane
            if (alignedTMSequences != null)
            {
                for (GappedStructureAnnotatedSequence tmseq : alignedTMSequences)
                {
                    if (tmseq.getName().equals(seq.getName()))
                    {
                        alignedHomologsClone.remove(seq);
                        break;
                    }
                }
            }
        }
        return alignedHomologsClone;
    }

    private static void writeSimpleOutput(PrintStream stream, JSAPResult options, List<ScoreAnnotatedSequence> predictions) throws FileNotFoundException
    {
        StructureDB output = new StructureDB(StructureDB.MODE.WRITE);
        for (ScoreAnnotatedSequence sequence : predictions)
        {
            if (sequence != null)
            {
                try
                {
                    output.addSequence(sequence);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }
        }

        output.writeDB(stream);
    }

    public static MethodSettings getMethod(JSAPResult config)
    {
        final String method = config.getString(METHOD_FLAG);

        if (method != null)
        {
            if (method.equals("p") || method.equals("psipred") || method.equals("1"))
            {
                return MethodSettings.PSIPRED;
            }
            else if (method.equals("j") || method.equals("jnet") || method.equals("2"))
            {
                return MethodSettings.JNET;
            }
            else if (method.equals("f") || method.equals("fast") || method.equals("3"))
            {
                return MethodSettings.PSIPRED_JNET;
            }
        }

        return MethodSettings.ALL;
    }
}
