package ca.ualberta.proteus.program;

import ca.ualberta.cs.homodeller.Homodeller;
import ca.ualberta.proteus.db.model.ProteusSettingsIntf;
import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.program.hmmpfam.HMMPFamRunner;
import ca.ualberta.proteus.program.homodeller.HomodellerImpl;
import ca.ualberta.proteus.program.predictSPTM.Organism;
import ca.ualberta.proteus.program.tmbhunt.TMBHuntRunner;
import ca.ualberta.proteus.program.tmhmm.TMHMMRunner;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.program.transsec.HomologCombiner;
import ca.ualberta.proteus.program.transsec.MainPredictor;
import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.Consts.Homolog;
import ca.ualberta.proteus.util.db.DBType;
import ca.ualberta.proteus.util.db.PredictionDB;
import ca.ualberta.proteus.util.db.PredictionDBFactory;
import ca.ualberta.proteus.util.io.StructureDB;
import org.apache.log4j.Logger;
import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.seq.io.*;
import org.biojava.bio.structure.Structure;
import org.biojava.bio.structure.io.PDBFileParser;
import org.biojava.bio.symbol.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Mar 1, 2005
 *         Time: 9:12:05 PM

 *         <p/>
 *         Main class representing the proteus suite of programs
 *         <p/>
 */
public class ProteusImpl extends Thread
{
    private final double DEFAULT_EVALUE = 1e-7;
    private static Alphabet proteinAlphabet = ProteinTools.getAlphabet();
    public RuntimeCatalog logger = new RuntimeCatalog();

    private Logger log4j = Logger.getLogger(ProteusImpl.class);

    private ProteusSettingsIntf settings;
    private ScoreAnnotatedSequence result;
    
    private final Queue<CompletionListener> completionListeners = new ArrayBlockingQueue<CompletionListener>(10);

    public Date getStartTime()
    {
        return startTime;
    }

    public void setStartTime(Date startTime)
    {
        this.startTime = startTime;
    }

    public Date getEndTime()
    {
        return endTime;
    }

    public void setEndTime(Date endTime)
    {
        this.endTime = endTime;
    }

    private Date startTime;
    private Date endTime;


    /**
     * Make the constructor private
     */
    public ProteusImpl(ProteusSettingsIntf settings)
    {
        this.setSettings(settings);
    }

    /**
     * @param fastaFile- the file name to parse as a FASTA formatted file
     */
    public ProteusImpl(String fastaFile) throws BioException, IOException
    {
        this.setSettings(new ProteusSettings());
        settings.setSequence(readFastaFile(fastaFile));
    }

    /**
     * Reads a FASTA formatted file
     *
     * @param fastaFile
     */
    private static Sequence readFastaFile(String fastaFile) throws BioException, IOException
    {
        // TODO move this somewhere so we can reuse it

        File f = new File(fastaFile);
        if (!f.exists())
        {
            System.err.println("File " + fastaFile + " does not exist.");
            return null;
        }

        BufferedReader reader = new BufferedReader(new FileReader(f));
        SymbolTokenization tokenizer = proteinAlphabet.getTokenization("token");
        try
        {
            FastaFormat fastaFormat = new FastaFormat();

            SequenceBuilderFactory factory = SeqIOTools.getFastaBuilderFactory();
            SequenceBuilder parser = factory.makeSequenceBuilder();

            fastaFormat.readSequence(reader, tokenizer, parser);

            return parser.makeSequence();
        }
        catch (IOException e)
        {
            reader.close();
            reader = new BufferedReader(new FileReader(fastaFile));
            StringBuilder buffer = new StringBuilder();
            String line = reader.readLine().replaceAll("\\s+", "");

            // skip blank lines
            while (line.equals("") && (line = reader.readLine()) != null)
            {
                // skip
            }

            String header = "SEQ";
            if (line != null && line.startsWith(">"))
            {
                header = line;
            }
            else
            {
                buffer.append(line);
            }

            while ((line = reader.readLine()) != null)
            {
                if (line.contains(">"))
                {
                    break;
                }
                buffer.append(line);
            }

            String seqStr = ProteusTools.cleanSequence(buffer.toString());

            SimpleSymbolList newList = new SimpleSymbolList(tokenizer, seqStr);
            return new SimpleSequence(newList, "", header.substring(1), null);
        }
    }

    /**
     * Reads a PDB file
     *
     * @param pdbFile
     */
    private static Structure readPDBFile(String pdbFile) throws IOException
    {
        PDBFileParser pdbReader = new PDBFileParser();
        BufferedReader reader = new BufferedReader(new FileReader(pdbFile));
        return pdbReader.parsePDBFile(reader);
    }


    /**
     * @param fastaFile- the file name to parse as a FASTA formatted file
     * @param pdbFile    - the file name to parse as a PDB file, if the user wishes to supply a file
     *                   for comparison in homodeller and threading modules
     */
    public ProteusImpl(String fastaFile, String pdbFile) throws BioException, IOException
    {
        this.setSettings(new ProteusSettings());
        settings.setSequence(readFastaFile(fastaFile));
        settings.setStructure(new File(pdbFile));
    }

    /**
     * The testSequence object to run on
     *
     * @param sequence
     */
    public ProteusImpl(Sequence sequence)
    {
        this.setSettings(new ProteusSettings());

        settings.setSequence(sequence);
    }

    public ProteusImpl(Sequence sequence, File PDBFile)
    {
        this.settings = new ProteusSettings();
        settings.setSequence(sequence);
        settings.setStructure(PDBFile);
    }

    /**
     * Main run method
     */
    public void run()
    {
        startTime = new Date();

        addCompletionListener(new CompletionListener()
        {
            public void onComplete(boolean successful, String message)
            {
                RuntimeCatalog logger = getLogger();

                result = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
                if (result == null)
                {
                    result = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL);
                }
            }
        });

        String message = "";
        boolean successful = false;
        try
        {
            if (settings.isRunSecondaryStructure())
            {
                StructureAnnotatedSequence signalPeptide = null;
                if ( signalPeptide == null && settings.getOrganism() != null && settings.getOrganism() != null ) 
                {
                	signalPeptide = this.findSignalPeptideRegion(settings.getOrganism());
                	if( signalPeptide == null )
                		signalPeptide = runHMMER();
                }

                StructureAnnotatedSequence tm = runTMHMM();
                StructureAnnotatedSequence transmembraneHomolog = findTransmembraneRegion(signalPeptide);

                assert transmembraneHomolog == null || settings.getSequence().length() == transmembraneHomolog.length();
                assert tm == null || settings.getSequence().length() == tm.length();
                assert signalPeptide == null || settings.getSequence().length() == signalPeptide.length();

                boolean isProteinbbtm = false;
                if (transmembraneHomolog == null || tm == null )
                {
                    // We didn't find a TM homolog, so BLAST against beta-barrel database
                	transmembraneHomolog = findBetaBarrelRegions();
                	
                	// We still didn't find a TM homolog, so let TMBHunt tell us if it's beta-barrel or not
                	if( transmembraneHomolog == null )
                		isProteinbbtm = runTMBHunt();
                	else
                		isProteinbbtm = true;
                	
                    //	 System.out.println("isBBTM : "+isProteinbbtm);
                    logger.logObject(Consts.IS_BBTM, isProteinbbtm);
                }
                else
                {                    
                    logger.logObject(Consts.IS_BBTM, isProteinbbtm);
                }

                Map<String, StructureAnnotatedSequence> sequenceMap = new HashMap<String, StructureAnnotatedSequence>(3);
                sequenceMap.put(Consts.SIGNAL_PEPTIDE, signalPeptide);
                sequenceMap.put(Consts.TRANSMEMBRANE, tm);
                sequenceMap.put(Consts.TRANSMEMBRANE_HOMOLOG, transmembraneHomolog);
                
                logger.logObject(Consts.SIGNAL_PEPTIDE, signalPeptide);
                logger.logObject(Consts.TRANSMEMBRANE, tm);
                logger.logObject(Consts.TRANSMEMBRANE_HOMOLOG, transmembraneHomolog);
                
                boolean isTM = transmembraneHomolog != null;
                logger.logObject(Consts.IS_TM, isTM);
                logger.stepComplete(Consts.Steps.TM_SEARCH);
                // System.out.println("isTM : "+transmembraneHomolog);

                ScoreAnnotatedSequence sequenceToOverlay;
                if (settings.isRunOnlyTM())
                {
                    // for testing purposes only
                    sequenceToOverlay = runOnlyTM();
                }
                else
                {
                    sequenceToOverlay = runSecondaryStructure();
                }

                try
                {
                    HomologCombiner homologCombiner = HomologCombiner.getInstance();
                    homologCombiner.setLogger(logger);

                    //ScoreAnnotatedSequence final_homology = homologCombiner.searchAndOverlayHomolog(sequenceToOverlay, sequenceMap);
                    
					/****/
					ScoreAnnotatedSequence final_homology = homologCombiner.searchAndOverlayHomolog(sequenceToOverlay, sequenceMap, settings.getIdentityThreshold());
					/***/	
					
					
					
					if( transmembraneHomolog == null && isProteinbbtm )
                    	final_homology.setStructureAnnotation(ProteusTools.createStructure(final_homology.getStructureAnnotation().seqString().replaceAll("e", "b")));
                    logger.debug("Combined result:\n " + final_homology);

                    logger.logObject(Consts.FINAL_WITH_HOMOLOGY, final_homology);
                    logger.stepComplete(Consts.Steps.HOMOLOG);
                }
                catch (Exception e)
                {
                    e.printStackTrace();
                }
            }

            if (settings.isRunHomodeller())
            {
                runHomodeller();
            }

            successful = true;
            logger.log("Prediction complete.");
        }
        catch (Throwable t)
        {
            logger.error(t.toString());

            t.printStackTrace();
            message = t.getMessage();
        }

        logger.finish();
        endTime = new Date();

        fireCompletionListeners(successful, message);
    }

    private ScoreAnnotatedSequence runOnlyTM()
    {
        /* Create a new structure with the sequence of our query, but a structure of all coils
        This is a subsitute for actually running a secondary structure prediction, which would take too long
         */
        ScoreAnnotatedSequence sequence = new ScoreAnnotatedSequence(settings.getSequence());

        Symbol[] coilStructure = new Symbol[sequence.length()];
        StructureTools structure = new StructureTools();

        Arrays.fill(coilStructure, structure.getC());

        SymbolList symbolList = new SimpleSymbolList(coilStructure, sequence.length(), structure.getStructure());
        sequence.setStructureAnnotation(symbolList);

        return sequence;
    }

    private ScoreAnnotatedSequence runSecondaryStructure()
    {
        MainPredictor mainPredictor = new MainPredictor(this.logger, settings.getMethod());
        mainPredictor.useStoredNN = true;
        mainPredictor.testing = settings.isTesting();
        mainPredictor.setEvaluating(settings.isEvaluating());

        PredictionDB predictionDB = PredictionDBFactory.getDB(DBType.XML);

        ScoreAnnotatedSequence finalSequence = mainPredictor.predict(settings.getSequence());

        if (settings.getMethod() == MethodSettings.ALL || settings.getMethod() == MethodSettings.PSIPRED_JNET)
        {
            assert finalSequence != null : "Final prediction is null";
        }

        logger.logObject(Consts.FINAL, finalSequence.clone());

        ScoreAnnotatedSequence jnetSequence = predictionDB.get(settings.getSequence(), "JNET");
        logger.logObject(Consts.JNET, jnetSequence);
        ScoreAnnotatedSequence psipredSequence = predictionDB.get(settings.getSequence(), "PSIPRED");
        logger.logObject(Consts.PSIPRED, psipredSequence);
        ScoreAnnotatedSequence proteusSequence = predictionDB.get(settings.getSequence(), "PROTEUS");
        logger.logObject(Consts.PROTEUS, proteusSequence);

        logger.log("Searching for homolog...");
        //List<ScoreAnnotatedSequence> homologSequences = mainPredictor.searchForHomolog(settings.get(), isTM);

        Alignment alignment = (Alignment) logger.getLogObject(Consts.ALIGNMENT);
        if (alignment == null)
        {
            // only null in the case it was in the DB and we didn't run it
            // in which case it should be in our DB
            XAlignDB db = XAlignDB.getInstance();
            alignment = db.get(settings.getSequence());
            logger.logObject(Consts.ALIGNMENT, alignment);
        }

        //if ( homologSequences.size() > 0 ) {
        //StringBuilder sb = new StringBuilder("Homologs found : ");
        //for( ScoreAnnotatedSequence seq : homologSequences)
        //     sb.append(seq.toString()).append("\n");
        // logger.log(sb.toString());
        // logger.logObject(Consts.HOMOLOG_LIST, homologSequences);

        ScoreAnnotatedSequence sequenceToOverlay;
        switch (settings.getMethod())
        {
            case PSIPRED:
                sequenceToOverlay = psipredSequence;
                break;
            case JNET:
                sequenceToOverlay = jnetSequence;
                break;
            case TRANSSEC:
                sequenceToOverlay = proteusSequence;
                break;
            case ALL:
            case PSIPRED_JNET:
            default:
                sequenceToOverlay = finalSequence;
                break;

        }

        return sequenceToOverlay;
    }

    private StructureAnnotatedSequence runTMHMM()
    {
        TMHMMRunner runner = new TMHMMRunner(settings.getSequence());
        try
        {
            return runner.runTMHMM();
        }
        catch (Exception e)
        {
            logger.error(e.getMessage());
        }
        return null;
    }

    private StructureAnnotatedSequence runHMMER()
    {
        // PredictSPTMRunner runner = new PredictSPTMRunner(settings.getSequence(), settings.getOrganism());
        HMMPFamRunner runner = new HMMPFamRunner(settings.getSequence(), settings.getOrganism());

        try
        {
            return runner.runHMMPFam();
        }
        catch (Exception e)
        {
            logger.error(e.getMessage());
        }
        return null;
    }

    private boolean runTMBHunt()
    {
        TMBHuntRunner runner = new TMBHuntRunner(settings.getSequence());

        try
        {
            return runner.runTMBHuntRunner();
        }
        catch (Exception e)
        {
            logger.error(e.getMessage());
        }
        return false;
    }

    private List<Sequence> getTemplateHomologs() {
    	List<GappedStructureAnnotatedSequence> homologs = (List<GappedStructureAnnotatedSequence>)logger.getLogObject(Consts.ALL_ALIGNED_HOMOLOG_LIST);
        List<GappedStructureAnnotatedSequence> alignedTMSequences = (List<GappedStructureAnnotatedSequence>)logger.getLogObject(Consts.TM_HOMOLOGS);

        if (homologs != null && homologs.size() > 0)
        {
        	return this.getBestCoverageHomologs(homologs);
        }
        else if (alignedTMSequences != null && alignedTMSequences.size() > 0)
        {
        	return this.getBestCoverageHomologs(alignedTMSequences);
        }
        
        return new ArrayList<Sequence>();
    }
	
    /**
     * Get only homologs from specified list provide coverage across the query sequence.
     * @param homologs
     * @return
     */
    private List<Sequence> getBestCoverageHomologs(List<GappedStructureAnnotatedSequence> homologs) {
    	GappedStructureAnnotatedSequence alignedQuery = (GappedStructureAnnotatedSequence)logger.getLogObject(Consts.ALIGNED_QUERY);
    	Sequence sequence = settings.getSequence();
    	boolean[] covered = new boolean[sequence.length()];
    	
    	List<Sequence> templateHomologs = new ArrayList<Sequence>();
    	
    	for( GappedStructureAnnotatedSequence homolog: homologs ) {
    		logger.debug(homolog.seqString());
    		int firstNonGap = alignedQuery.viewToSource(homolog.firstNonGap())-1;
    		int lastNonGap = alignedQuery.viewToSource(homolog.lastNonGap())-1;
    		boolean addTemplate = false;
    		for( int i = firstNonGap; i <= (lastNonGap > sequence.length() ? sequence.length()-1 : lastNonGap); i++ ) {
    			if( !covered[i] ) {
    				// current homolog covers the sequence where no previous homolog has covered yet
    				covered[i] = true;
    				addTemplate = true;
    			}
    		}
    		if( addTemplate )
    			templateHomologs.add(homolog);
    	}
    	
    	return templateHomologs;
    }
    
	private void runHomodeller()
	{
		try
		{
			List<Homodeller.Result> results = new ArrayList<Homodeller.Result>();
			if (settings.getStructure() != null)
			{
				Homodeller homodeller = new Homodeller();
				if (settings.getTemplateID() != null && settings.getTemplateID().length() > 0)
				{
					Homodeller.Result result = homodeller.run(settings.getSequence().seqString(), settings.getTemplateID());
					if( result != null )
						results.add(result);
				}
				else
				{
                    // TODO out due to compile errors
//					Homodeller.Result result = homodeller.run(settings.getSequence().seqString(), settings.getStructure());
					/*if( result != null )
						results.add(result);*/
				}
			}
			else
			{
				List<Sequence> templateHomologs = this.getTemplateHomologs();
				
				if (!settings.isRunSecondaryStructure())
				{
					// we didn't run secondary structure, run without homolog
					Homodeller homodeller = new Homodeller();
					Homodeller.Result result = homodeller.run(settings.getSequence());
					if( result != null )
						results.add(result);
				}
				else if (!templateHomologs.isEmpty())
				{
					// homodeller should run with the same homolog as we do our mapping with
                    for( int i = 0; i < templateHomologs.size(); i++ ) {
                    	Sequence homolog = templateHomologs.get(i);
                    	Homodeller homodeller = new Homodeller();
                    	Homodeller.Result result = homodeller.run(settings.getSequence().seqString(), homolog.getName());
                    	if( result != null )
                    		results.add(result);
                    }
                }
                
                // Don't run homodeller if a decent match wasn't found
                // result = homodeller.run(settings.getSequence());
            }

            logger.stepComplete(Consts.Steps.HOMODELLER);

            if( !results.isEmpty() ) {
            	logger.logObject(Consts.HOMODELLER_RESULTS, results);
            }

            if (!results.isEmpty() && settings.isRunEnergyMinimization())
            {
                // Run Enery minimization
                logger.log("Running energy minimization..");

                try
                {
                	List<Structure> minimizedModels = new ArrayList<Structure>();
                	for( int i = 0; i < results.size(); i++ ) {
                		Homodeller.Result result = (Homodeller.Result)results.get(i);
                		Structure minimizedStructure = HomodellerImpl.runEnergyMinimization(result.model);
                		minimizedModels.add(minimizedStructure);
                	}
                    logger.stepComplete(Consts.Steps.ENERGY_MINIMIZATION);
                    logger.logObject(Consts.MINIMIZED_STRUCTURE, minimizedModels);
                }
                catch (Throwable e)
                {
                    logger.error("Minimization failed");
                    logger.error(e.toString());
                }
            }
        }
        catch (Throwable e)
        {
            String returnEmail = Config.getConfig().getStringValue("email");

            logger.logObject(Consts.HOMODELLER_ERROR, e.getMessage());

            e.printStackTrace();

            EmailUtils.sendMessage(returnEmail, returnEmail, "Homodeller error", "Check the server log, there was an error with Homodeller . \n"
                    + "ID = " + this.getID()
                    + "\n"
                    + e.getMessage());
        }
    }

	private StructureAnnotatedSequence findRegion(String database, Homolog type) {
		
        BlastSettings blastSettings = new BlastSettings();
        blastSettings.setSequence(settings.getSequence());
        blastSettings.setDatabase(database);
        blastSettings.setBlastType(BlastSettings.BLASTTYPE.PROTEUS_BLASTPARSER_BLASTP);
        blastSettings.setEvalue(DEFAULT_EVALUE);
		
		
        BlastRunner runner = new BlastRunner(blastSettings);
        runner.run();
		
        List<GappedSequence> homologs = runner.getAlignedResult();
		
		//**********************have to add one function call which selects the homologous sequences based on specified sequence identity********//
		
		if ( settings.getIdentityThreshold() != null ) {
			
			final double identityThreshold = settings.getIdentityThreshold().doubleValue();
			
			homologs = filterByPercentageIdentity (homologs, settings.getSequence(), identityThreshold);
		}
		//**************************************************************************************************************************************//
		
        if (homologs.size() != 0)
        {
            GappedStructureAnnotatedSequence alignedHomolog = createStructurallyAlignedHomologs(homologs, type);
            StructureAnnotatedSequence ungappedSequence = alignedHomolog.getUngappedSequence();
            return ungappedSequence;
        }
        
        return null;
    }
    
	//*******************************************************filtering sequences based on sequence identity*************************************//
	private List<GappedSequence> filterByPercentageIdentity (List<GappedSequence> blastSequences, Sequence sequence, double identityThreshold)
	{
		List<GappedSequence> ReducedBlastSequences = new ArrayList<GappedSequence>();
		
		int SequenceLength = sequence.length();
		int identity = 0;
		double percentageIdentity = 0;
		double avgLength = 0;
		int shortestLength = 0;
		int i, j;
		
		for (GappedSequence blastSequence: blastSequences) {
			
			identity = 0;
			
			String s = blastSequence.seqString();
			
			String blast = blastSequence.seqString().replaceAll("-","").replaceAll("~","");
			String query = sequence.seqString();
			
			int blastSequenceLength = blastSequence.seqString().replaceAll("-","").replaceAll("~","").length();
			
			if (blastSequenceLength > SequenceLength) {
				shortestLength = SequenceLength;
			}
			else{
				shortestLength = blastSequenceLength;
			}
			
			
			for (i = 0, j = 0; i < shortestLength; i++, j++) {
				
				
				if (query.charAt (i) == blast.charAt(j) ){
					
					identity++;
				}	
			}
			
			
			
			
			//avgLength = (SequenceLength + blastSequenceLength)/2;
			
			//PercentageIdentity = identity / avgLength;
			percentageIdentity = ((double) identity / (double) shortestLength) * 100;
			
			System.out.println (s);
			System.out.printf ("\t%f\n",percentageIdentity);
			
			if (percentageIdentity <= identityThreshold){
				
				System.out.println(s);
				ReducedBlastSequences.add(blastSequence);
			}
		}
		
		
		
		
		return ReducedBlastSequences;
		
	}	
	//**************************************************************************************************************************************//
    
    public StructureAnnotatedSequence findBetaBarrelRegions() {
    	String database = Config.getConfig().getStringValue("blast.BBDB");
    	StructureAnnotatedSequence ungappedSequence = this.findRegion(database, Homolog.BetaBarrel);
    	if( ungappedSequence != null )
        	log4j.debug("Beta barrel: " + ungappedSequence.getStructureAnnotation().seqString());
    	
    	return ungappedSequence;
    }
    
    /**
     * Returns an ungapped StructureAnnotatedSequence with our SP homolog mapped to it, if there is a homolog.
     * @return	an upgapped StructureAnnotatedSequence with our SP homolog mapped to it, if there is a homolog;
     * 			<tt>null</tt> otherwise.
     */
    public StructureAnnotatedSequence findSignalPeptideRegion(Organism organism) {
    	String database = null;
    	Homolog signalPeptideType = null;
    	switch(organism) {
    	case GRAM_POSITIVE:
    		database = Config.getConfig().getStringValue("blast.SPDB.gram+");
    		signalPeptideType = Homolog.GramPositiveSignalPeptide;
    		break;
    	case GRAM_NEGATIVE:
    		database = Config.getConfig().getStringValue("blast.SPDB.gram-");
    		signalPeptideType = Homolog.GramNegativeSignalPeptide;
    		break;
    	case EUKARYOTIC:
    		database = Config.getConfig().getStringValue("blast.SPDB.euk");
    		signalPeptideType = Homolog.EukaryoticSignalPeptide;
    		break;
    	}
    	StructureAnnotatedSequence ungappedSequence = this.findRegion(database, signalPeptideType);
    	if( ungappedSequence != null ) {
        	log4j.debug(ungappedSequence.getStructureAnnotation().seqString());
        	ungappedSequence = this.postProcessSignalPeptide(ungappedSequence, organism);
        	if( ungappedSequence != null )
        		log4j.debug(ungappedSequence.getStructureAnnotation().seqString());
        }
    	return ungappedSequence;
    }
    
    /**
     * Analyze the output and modify the signal peptide prediction if necessary.
     * 
     * Port of David Arndt's code in pptBlast.cgi Perl script
     * @param ungappedSequence
     * @return
     */
    private StructureAnnotatedSequence postProcessSignalPeptide(StructureAnnotatedSequence ungappedSequence, Organism organism) {
    	// Set minimum, maximum, and standard signal peptide lengths.
    	// These numbers are based on an examination of a histogram of the distribution of signal peptide
    	// lengths in each of the databases. However, the numbers for Gram+ databases are especially arbitrary
    	// due to the unusual profile of the data.
    	int minimumLength = 15;				// minimum signal peptide length
    	int maximumLength = 39;				// maximum signal peptide length
    	int standardLength = 21;			// standard signal peptide length
    	
    	if( organism == Organism.GRAM_POSITIVE ) {
    		minimumLength = 16;
    		maximumLength = 50;
    		standardLength = 28;
    	}
    	
    	int truncLength = minimumLength/2;	// threshold value for predicting absence of signal peptide vs. deciding to extend the prediction

    	String structure = ungappedSequence.getStructureAnnotation().seqString();
    	int signalPeptideLength = structure.lastIndexOf("s") + 1;
    	if( signalPeptideLength < truncLength ) {
    		return null;
    	}
    	else if( signalPeptideLength < minimumLength || signalPeptideLength > maximumLength ) {
    		int sequenceLength = ungappedSequence.length();
    		if( sequenceLength >= standardLength ) {
    			structure = structure.substring(standardLength);
    			for( int i = 1; i <= standardLength; i++ )
    				structure = "s" + structure;
    		}
    		else {
    			structure = "";
    			for( int i = 1; i <= sequenceLength; i++ )
    				structure += "s";
    		}
    	}
    	
    	while( structure.indexOf("-s") >= 0 )
    		structure = structure.replaceAll("-s", "ss");
    	structure = structure.replaceAll("-", " ").replaceFirst("s ", "sc");
    	ungappedSequence.setStructureAnnotation(ProteusTools.createStructure(structure));
    	return ungappedSequence;
    }
    
    /**
     * Returns a ungapped StructureAnnotatedSequence with our TM homolog
     *
     * @param signalPeptide
     * @return our ungapped query sequence with a TM homolog mapped to it, if there is a homolog.
     */
    public StructureAnnotatedSequence findTransmembraneRegion(StructureAnnotatedSequence signalPeptide)
    {
    	// Basically does a BLAST to find any transmembrane homologs
        // Similar to @see HomologCombiner.searchAndOverlayHomologs
    	
    	StructureAnnotatedSequence ungappedSequence = this.findRegion(Config.getConfig().getStringValue("blast.TMPDB"), Homolog.Transmembrane);
        
        if (ungappedSequence != null )
        {
            logger.log("Sequence is transmembrane");

            // find where the TM region is in the 60-n subsection of the TM sequence
            int TMstartAfter60;
            if ( ungappedSequence.length() > 60 )
            {
                TMstartAfter60 = ungappedSequence.getStructureAnnotation().subStr(60, ungappedSequence.length()).indexOf('t');
            }
            else
            {
                TMstartAfter60 = -1;
            }

            // find where the TM region is in the 0-60 region
            int end = Math.min(ungappedSequence.length(),60);
            int TM60 = ungappedSequence.getStructureAnnotation().subStr(1, end).lastIndexOf('t');

            // find where the signal peptide starts
            int SPstart = signalPeptide == null ? -1 : signalPeptide.getStructureAnnotation().seqString().indexOf('s');
            if (SPstart > -1 && TM60 < SPstart && TMstartAfter60 < 0)
            {
                logger.debug("Removing TM homolog because of overlap with signal peptide");

                // there's a signal peptide in the first 60 chars, and the TM region is only in the first 60 chars.
                return null;
            }

            return ungappedSequence;
        }

        return null;
    }

    /**
     * Returns the index of the first instance of the specified pattern
     *
     * @param pattern  a regular expression in the sequence's alphabet
     * @param sequence a sequence
     * @return the index of the first instance of the specified pattern, or -1 if there isn't one
     */
    private int findFirstInstanceOf(String pattern, SymbolList sequence)
    {
        Pattern p = Pattern.compile(".*(" + pattern + ").*", Pattern.CASE_INSENSITIVE);
        Matcher matcher = p.matcher(sequence.seqString());

        if (matcher.matches())
        {
            return matcher.start(1);
        }


        return -1;
    }
    
    /**
     * Takes alignments of homologs, and gets the best one, and maps it to our query sequence
     *
     * @param homologs
     * @return our query sequence with a TM homolog mapped to it, if there is a homolog.
     */
    private GappedStructureAnnotatedSequence createStructurallyAlignedHomologs(List<GappedSequence> homologs, Homolog type)
    {
        assert homologs.size() > 0 : "Why are you calling me with 0 homologs?";

        GappedSequence alignedQuery = null;

        Config config = Config.getConfig();
        StructureDB db;
        try
        {
            db = new StructureDB(config.getStringValue(type.getStringValue()), StructureDB.MODE.READ);
        }
        catch (IOException e)
        {
        	log4j.debug("Could not get " + type.getStringValue() + " structure database.", e);
            return null;
        }
        
        List<GappedStructureAnnotatedSequence> alignedSequences = new ArrayList<GappedStructureAnnotatedSequence>();
        HashMap<GappedStructureAnnotatedSequence, Double> homologEvalues = new HashMap<GappedStructureAnnotatedSequence, Double>();

        Double bestEvalue = Double.MAX_VALUE;
        GappedStructureAnnotatedSequence bestHomolog = null;
        
        // Sort so QUERY is at the top
        Collections.sort(homologs, new Comparator<GappedSequence>()
        {
            public int compare(GappedSequence o1, GappedSequence o2)
            {
                if ( o1.getName().equals("QUERY") )
                {
                    return -1;
                }
                if ( o1 == o2 )
                {
                    return 0;
                }
                return 1;
            }
        });

        for (GappedSequence seq : homologs)
        {
            String name = seq.getName();
            if (name.equals("QUERY")
                    || (settings.getSequence().getName().equals(name) && alignedQuery == null)) // In case the query sequence is exactly the same as the homolog
            {
                alignedQuery = seq;
                continue;
            }

            GappedStructureAnnotatedSequence homolog = new GappedStructureAnnotatedSequence(seq);

            StructureAnnotatedSequence newSequence = null;
            try
            {
                newSequence = (StructureAnnotatedSequence) db.getSequence(name);
            }
            catch (IllegalIDException e)
            {
                log4j.debug("Can't find " + name + " in " + type.getStringValue() + " database", e);
            }

            if (newSequence != null)
            {
                homolog.createGappedStructureToIncompleteSequence(newSequence);
                alignedSequences.add(homolog);
            }
            else
            {
                log4j.debug("Can't find " + name + " in " + type.getStringValue() + " database");
            }

            if (bestHomolog == null)
            {
                // only assign it so we don't end up not choosing a "best" homolog
                bestHomolog = homolog;
            }

            Annotation annotation = seq.getAnnotation();
            if (annotation != null)
            {
                Double evalue = (Double) annotation.getProperty("e-value");
                if (evalue <= bestEvalue)
                {
                    if ( evalue.equals(bestEvalue) )
                    {
                        // if they're both 0 compare on bits  (score)
                        Integer score = (Integer) annotation.getProperty("score");
                        Integer bestScore = (Integer) bestHomolog.getAnnotation().getProperty("score");

                        if ( score < bestScore )
                        {
                            continue;
                        }
                    }

                    bestEvalue = evalue;
                    bestHomolog = homolog;
                }

                homologEvalues.put(homolog, evalue);
            }
        }

        assert alignedSequences.size() > 0 : "Aligned sequences doesn't have > 0 sequences";

        if (bestHomolog == null)
        {
            logger.error("Did not choose a best " + type + " homolog");
            return null;
        }

        logger.log("Using " + type + " homolog " + bestHomolog.getName());

        GappedStructureAnnotatedSequence gappedQuery = new GappedStructureAnnotatedSequence(alignedQuery);
        gappedQuery.assignAlignedStructure(bestHomolog);

        // TODO can we assign the gappedQuery a name??
        
        switch(type) {
        case Transmembrane:
        case BetaBarrel:
        	logger.logObject(Consts.TM_HOMOLOGS, alignedSequences);
        	logger.logObject(Consts.TM_HOMOLOG_EVALUES, homologEvalues);
        	logger.logObject(Consts.TM_HOMOLOG, bestHomolog);
        	logger.logObject(Consts.ALIGNED_TM_HOMOLOG, gappedQuery);
        	break;
        case GramPositiveSignalPeptide:
        case GramNegativeSignalPeptide:
        case EukaryoticSignalPeptide:
        	logger.logObject(Consts.SP_HOMOLOGS, alignedSequences);
        	logger.logObject(Consts.SP_HOMOLOG_EVALUES, homologEvalues);
        	logger.logObject(Consts.SP_HOMOLOG, bestHomolog);
        	logger.logObject(Consts.ALIGNED_SP_HOMOLOG, gappedQuery);
        	break;
        }

        return gappedQuery;
    }

    /*
    private static void printUsage()
    {
        System.out.println("Usage: ProteusImpl <path to sequence file> [options]");
        System.out.println("Options:");
        System.out.println("-o | -output [file] : The output file for the predictions.");
        System.out.println("-m=[s,j,p] | -method=[psipred,jnet,proteus] : The method for producing predictions.");
    }
*/

    public RuntimeCatalog getLogger()
    {
        return logger;
    }

    /**
     * Add a listener that will be fired when proteus finishes runner
     *
     * @param eventListener
     */
    public void addCompletionListener(CompletionListener eventListener)
    {
        completionListeners.add(eventListener);
    }

    public void fireCompletionListeners(boolean successful, String message)
    {
        synchronized(completionListeners)
        {
            for (CompletionListener completionListener : completionListeners)
            {
                try
                {
                    // if it fails we still need to call the other completion listeners
                    completionListener.onComplete(successful, message);
                }
                catch (Throwable e)
                {
                    logger.error(e.toString());
                    e.printStackTrace();
                }
            }
        }
    }

    public String getID()
    {
        return settings.getID();
    }

    public ProteusSettingsIntf getSettings()
    {
        return settings;
    }

    public void setSettings(ProteusSettingsIntf settings)
    {
        this.settings = settings;
        this.logger.logObject(Consts.SETTINGS, settings);
    }


    public ScoreAnnotatedSequence getResult()
    {
        return result;
    }
}
