package ca.ualberta.proteus.program.hmmpfam;

import ca.ualberta.proteus.program.predictSPTM.Organism;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.ExecUtils;
import ca.ualberta.proteus.util.StreamHandler;
import org.apache.log4j.Logger;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.io.*;
import java.text.ParseException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: Apr 15, 2007
 * Time: 11:37:47 PM
 */
public class HMMPFamRunner
{
    private Sequence sequence;
    private Organism organism = Organism.GRAM_POSITIVE;
    private Logger logger = Logger.getLogger(HMMPFamRunner.class);
    private static final Map<Organism, String> organisms = new HashMap<Organism, String>();
    final static char SIGNAL_PEPTIDE = 'S';
    final static char CLEAVAGE_SITE = 'c';
    final static char NOTHING = ' ';

    final static int PEPTIDE_CUTOFF = 60;

    static
    {
        organisms.put(Organism.GRAM_NEGATIVE, "GRAM-");
        organisms.put(Organism.GRAM_POSITIVE, "GRAM+");
        organisms.put(Organism.EUKARYOTIC, "EUK");
    }

    public HMMPFamRunner(Sequence sequence, Organism organism)
    {
        this.sequence = sequence;
        this.organism = organism;
    }

    public void run()
    {
        runHMMPFam();
    }

    public StructureAnnotatedSequence runHMMPFam()
    {
        try
        {
            // write PDB file to temp file
            File tempFile;
            tempFile = File.createTempFile("sptm", ".in");
            tempFile.deleteOnExit();

            FileOutputStream blastfile;
            blastfile = new FileOutputStream(tempFile);
            try
            {
                // Only operate on the first 60 residues
                String protein;
                if ( sequence.length() > PEPTIDE_CUTOFF)
                {
                    protein = sequence.subStr(1, PEPTIDE_CUTOFF + 1);
                }
                else
                {
                    protein = sequence.seqString();
                }
                SeqIOTools.writeFasta(blastfile, ProteinTools.createProteinSequence(protein, sequence.getName()) );
            }
            catch (IllegalSymbolException e)
            {
                e.printStackTrace();
            }
            blastfile.close();

            StructureAnnotatedSequence gramPositive = this.getSequence(Organism.GRAM_POSITIVE, tempFile, 0.25);
            switch(organism) {
            case GRAM_POSITIVE:
            	break;
            case GRAM_NEGATIVE:
            	if( gramPositive != null && !"".equals(gramPositive.getStructureAnnotation().seqString().trim()) )
            		return this.getSequence(Organism.GRAM_NEGATIVE, tempFile, 1.0);
            	break;
            case EUKARYOTIC:
            	if( gramPositive != null && !"".equals(gramPositive.getStructureAnnotation().seqString().trim()) )
            		return this.getSequence(Organism.EUKARYOTIC, tempFile, 1.0);
            	break;
            }
            
            return gramPositive;
        } catch(IOException e) {
        	e.printStackTrace();
        }

        return null;
    }
    
    private StructureAnnotatedSequence getSequence(Organism organism, File tempFile, double eValue) {
    	String output = null;
    	try {
	    	Config config = Config.getConfig();
	    	String path = config.getStringValue("hmmpfam");
	    	String hmmPath = Config.getConfig().getStringValue(organisms.get(organism));

	        String cmd = path + " -E " + eValue + " " + hmmPath + " " + tempFile;
	
	        Process proc = ExecUtils.exec(cmd, 1000, 30 * 1000); // 30 second timeout
	
	        // any error message?
	        StreamHandler errorReader = new
	                StreamHandler(proc.getErrorStream(), "ERROR");
	
	        // any output?
	        StreamHandler outputReader = new
	                StreamHandler(proc.getInputStream(), "OUTPUT");
	
	        // kick them off
	        errorReader.start();
	        outputReader.start();
	
	        int exitVal = proc.waitFor();
	
	        if (exitVal != 0)
	        {
	            System.err.println(cmd);
	            System.err.println("HMMPFam failed " + errorReader.getInputString());
	            return null;
	        }
	
	        /** Wait for the output reader to completely do its thing **/
	        outputReader.join();
	
	        output = outputReader.getInputString();
	
	        BufferedReader reader = new BufferedReader(new StringReader(outputReader.getInputString()));
	
	        HMMPFamReader hmmReader = new HMMPFamReader();
	
	        Integer cleavageSite = hmmReader.parseSignalOutput(sequence, reader);
	
	        if ( cleavageSite != null )
	        {
	            // ScoreAnnotatedSequence result = new ScoreAnnotatedSequence(sequence);
	
	            // Create the structure
	            char[] c = new char[sequence.length()];
	            Arrays.fill(c, NOTHING);
	            Arrays.fill(c, 0, cleavageSite - 1, SIGNAL_PEPTIDE);
	            c[cleavageSite - 1] = CLEAVAGE_SITE;
	            String structure = new String(c);
	
	            return new StructureAnnotatedSequence(sequence, structure);
	        }
	        else
	        {
	            char[] c = new char[sequence.length()];
	            Arrays.fill(c, NOTHING);
	            
	            String structure = new String(c);
	
	            return new StructureAnnotatedSequence(sequence, structure);
	        }
    	}
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
        catch (ParseException e)
        {
            e.printStackTrace();
        }
        
//      in case of Error
        logger.debug(output);
        
        return null;
    }
}
