package ca.ualberta.proteus.program.homodeller;

import org.biojava.bio.structure.Chain;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.StructureException;
import org.biojava.bio.structure.Atom;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.TreeMap;
import java.util.List;

/***
 *
 * TODO
 * MUST REMOVE ALL TODOS - THEY ARE IMPORTANT
 */

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: Apr 19, 2007
 * Time: 11:22:36 PM
 */
public class CCD
{
    private static Logger logger = Logger.getLogger(CCD.class);

    final static int CA = 1;
    final static int C1 = 2;
    final static int O = 3;
    final static int N = 4; // The N atom following CA.
    final static int CA2 = 5;
    final static int HN = 6; // H atom for N.
    final static int HNp = 7; // The 1st (or only) H atom for Np.
    final static int HNp2 = 8; // 2nd H atom for Np (only for first residue).
    final static int HNp3 = 9; // 3rd H atom for Np (only for first residue).
    final static int HA = 10; // H atom for CA.
    final static int HA2 = 11; // Second alpha hydrogen for Glycine. Note that this has the same value as CB!
    final static int CB = 11;
    final static int CG = 12; //added by Raj..


    final static int ANGLE_TYPE_PHI = 0;
    final static int ANGLE_TYPE_PSI = 1;
    final static int ANGLE_TYPE_OMEGA = 2;


    final static int USE_SIDECHAINS = 1;
    final static int MAX_ATOMS = 35;
    final static int CCD_SEG_LENGTH = 4;
    final static int RAMA_LEN = 36;
    final static double ANGLE_SNAP_MARGIN = 1.0;


    // The number of residues in a segment when we try to adjust its angles with adjustSegmentCCD.
    // This number includes the immobile N-anchor and immobile C-anchor.
    //
    // Different values of CCD_SEG_LENGTH were tested on set of proteins. As CCD_SEG_LENGTH
    // increases, there is steadily more damage to the structure, but Rama and Omega scores improve
    // as well. A CCD_SEG_LENGTH of 8 is a fair amount better than a value of 7, but with
    // CCD_SEG_LENGTH > 8 the improvements in Rama and Omega scores are smaller. Therefore a value
    // of 8 may be a good compromise.
    final static double ANGLE_CHANGE_THRESHOLD = 0.00002;
    private static final boolean DEBUG = true;
    private static boolean adjustOmega = true;

    /*
    * Applies the Cyclic Coordinate Descent (CCD) algorithm to attempt to fix    problematic
    * phi, psi, and omega angles in the peptide segment from indices N_anchor_index to
    * C_anchor_index. The algorithm tries to adjust all phi and psi angles so that they
    * have a score of at least rama_score_threshold. Returns TRUE if the CCD algorithm was
    * successfully applied. If the algorithm fails, the peptide segment is restored to its
    * original configuration and FALSE is returned.
    *
    * This implementation of the CCD algorithm is similar to and based upon the CCD algorithm
    * explained in:
    * Canutescu AA, Dunbrack RL Jr. Cyclic coordinate descent: A robotics algorithm for
    * protein loop closure. Protein Sci. 2003 May;12(5):963-72.
    */
    public static boolean adjustSegmentCCD(Chain chain, int numRes, int N_anchor_index, int C_anchor_index, int rama_score_threshold) throws StructureException
    {
        assert N_anchor_index < C_anchor_index: "N-anchor must be smaller than C-anchor";

        int i, j, whichAngleType;
        int num_passes; // Number of passes of the CCD algorithm through the whole segment.
        boolean closureFailed = false; // Whether the CCD algorithm failed in closing the segment.
        boolean applyRotation;
        // Boolean. Whether to apply the rotation for the current angle (set to FALSE if rotation is by
        // 0.0 degrees, to avoid useless computation).
        double phi, psi, omega;
        double desired_phi, desired_psi, desired_omega;
        // Ideal angles for CCD rotation, unadjusted based on the constraints of peptide geometry.
        double sumOfSquaredDistances;
        // Between current C-terminal residue atom coordinates and original C-anchor atom coordinates.
        // Same as S in the paper by Canutescu and Dunbrack.

        // Original C-anchor atom coordinates.
        double[] target_CA = new double[3];
        double[] target_C1 = new double[3];
        double[] target_N = new double[3];

        // Movable C-terminal residue atom coordinates.
        double[] curr_CA = new double[3];
        double[] curr_C1 = new double[3];
        double[] curr_N = new double[3];

        double[] curr_CA_translated = new double[3];
        double[] curr_C1_translated = new double[3];
        double[] curr_N_translated = new double[3];
        double[] rotationAxis = new double[3];
        double[] vec_O1 = new double[3];
        double[] vec_O2 = new double[3];
        double[] vec_O3 = new double[3];
        // The points on the rotation axis closest to the movable C-anchor atoms.
        // Corresponding to CA, C1, and N, respectively.
        double[] temp = new double[3];
        double[] basePoint = new double[3];
        // Location of the N-terminal atom of the bond along which we are currently rotating.
        double[] vec_r1 = new double[3];
        double[] vec_r2 = new double[3];
        double[] vec_r3 = new double[3];
        double[] vec_r1_unit = new double[3];
        double[] vec_r2_unit = new double[3];
        double[] vec_r3_unit = new double[3]; // Unit vectors corresponding to vec_r1, vec_r2, vec_r3.
        double[] vec_s1 = new double[3];
        double[] vec_s2 = new double[3];
        double[] vec_s3 = new double[3];
        double[] vec_f1 = new double[3];
        double[] vec_f2 = new double[3];
        double[] vec_f3 = new double[3];
        double mag_r1, mag_r2, mag_r3; // The magnitudes of vec_r1, vec_r2, vec_r3, respectively.
        double tan_alpha;
        double alpha_rad; // The alpha angle (in radians) derived from tan_alpha;
        double alpha;
        // Alpha angle in degrees. This is the ideal amount by which we should change the current angle (phi, psi, or omega),
        // before taking into account constraints on those angles by peptide geometry.
        double second_deriv; // Second derivate of S with respect to theta.
        double[][] original_angles = new double[C_anchor_index - N_anchor_index + 1][3];

        //double expectedSumOfSquaredDistances;
        //double mag_f1, mag_f2, mag_f3; // The magnitudes of vec_f1, vec_f2, vec_f3, respectively.

        // DEBUG
        logger.debug(String.format("\nAdjusting residues %d to %d:\n", N_anchor_index, C_anchor_index));

        // Record the initial values for the dihedral angles in case the CCD algorithm does not work and
        // we need to restore the original values.

        // Allocate memory for original_angles.
        /*original_angles = malloc((C_anchor_index - N_anchor_index + 1) * sizeof(double*));
        assert (original_angles != NULL);
        for (j = 0; j < (C_anchor_index - N_anchor_index + 1); j++)
        {
            original_angles[j] = malloc(3 * sizeof(double));
            assert (original_angles[j] != NULL);
        }*/

        ArrayList<Group> residues = chain.getGroups();

        TreeMap<Integer, Angle> angles = HomodellerUtils.getAngles(chain);

        // Store original angles.
        for (i = N_anchor_index; i <= C_anchor_index; i++)
        {
            j = i - N_anchor_index; // j here is the index into original_angles, starting from 0.
            original_angles[j][0] = angles.get(i).phi;
            original_angles[j][1] = angles.get(i).psi;
            original_angles[j][2] = angles.get(i).omega;
        }

        // Record the initial positions of the C-anchor backbone atoms, which will be our target positions.

        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("CA").getCoords(), target_CA);
        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("C").getCoords(), target_C1);
        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("N").getCoords(), target_N);

        // Go through the residues of the current segment and repair any "bad" angles.
        // This will relocate the movable C-terminal residue of the segment.

        setCorrectedAngles(residues, angles, numRes, N_anchor_index + 1, C_anchor_index - 1, rama_score_threshold);

        if (DEBUG)
        {
            logger.debug(String.format("------------ Beginning loop closure ------------\n"));
        }

        // Apply CCD algorithm to iteratively adjust segment angles to bring the movable
        // C-terminal residue to the same position as the original C-anchor.

        // Rebuild the backbone atoms of the current segment of the peptide using the repaired angles.
        rebuildPeptideSegment(residues, angles, N_anchor_index, C_anchor_index, true);

        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("CA").getCoords(), curr_CA);
        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("C").getCoords(), curr_C1);
        VectorMath.copyVector(residues.get(C_anchor_index).getAtom("N").getCoords(), curr_N);

        sumOfSquaredDistances =
                VectorMath.distance(curr_CA, target_CA) * VectorMath.distance(curr_CA, target_CA)
                        + VectorMath.distance(curr_C1, target_C1) * VectorMath.distance(curr_C1, target_C1)
                        + VectorMath.distance(curr_N, target_N) * VectorMath.distance(curr_N, target_N);

        //logger.debug(String.format("target: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", target_CA[0], target_CA[1], target_CA[2],
        //	target_C1[0], target_C1[1], target_C1[2], target_N[0], target_N[1], target_N[2]);
        //logger.debug(String.format("curr: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", curr_CA[0], curr_CA[1], curr_CA[2],
        //	curr_C1[0], curr_C1[1], curr_C1[2], curr_N[0], curr_N[1], curr_N[2]);

        //logger.debug(String.format("   num_passes = %d, sumOfSquaredDistances = %f\n", 0, sumOfSquaredDistances);

        num_passes = 0;
        while (num_passes < 20 || sumOfSquaredDistances > 0.000001)
        {
            // Perform at least a certain number of passes.

            // If loop closure has not succeeded after a certain number of passes, give up and restore the original angles.
            if (num_passes >= 5000)
            {

                closureFailed = true;

                if (DEBUG)
                {
                    logger.debug(String.format("**LOOP CLOSURE FAILED** (sumOfSquaredDistances = %f). Restoring original angles yields C-terminal coordinates as follows:\n", sumOfSquaredDistances));
                }

                // Restore original angles.
                for (i = N_anchor_index; i <= C_anchor_index; i++)
                {
                    j = i - N_anchor_index; // j here is the index into original_angles, starting from 0.
                    angles.get(i).phi = original_angles[j][0];
                    angles.get(i).psi = original_angles[j][1];
                    angles.get(i).omega = original_angles[j][2];
                }

                // Rebuild the current segment of the peptide using the original angles.
                rebuildPeptideSegment(residues, angles, N_anchor_index, C_anchor_index, false);

                // DEBUG
                if (DEBUG)
                {
                    VectorMath.copyVector(residues.get(C_anchor_index).getAtom("CA").getCoords(), curr_CA);
                    VectorMath.copyVector(residues.get(C_anchor_index).getAtom("C").getCoords(), curr_C1);
                    VectorMath.copyVector(residues.get(C_anchor_index).getAtom("N").getCoords(), curr_N);
                    sumOfSquaredDistances =
                            VectorMath.distance(curr_CA, target_CA) * VectorMath.distance(curr_CA, target_CA)
                                    + VectorMath.distance(curr_C1, target_C1) * VectorMath.distance(curr_C1, target_C1)
                                    + VectorMath.distance(curr_N, target_N) * VectorMath.distance(curr_N, target_N);
                    logger.debug(String.format("curr: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", curr_CA[0], curr_CA[1], curr_CA[2],
                            curr_C1[0], curr_C1[1], curr_C1[2], curr_N[0], curr_N[1], curr_N[2]));
                    logger.debug(String.format("Restored sumOfSquaredDistances = %f\n", sumOfSquaredDistances));
                }

                break; // Break out of while loop.
            }

            // Iterate through the segment residues and apply CCD algorithm.

            for (i = N_anchor_index + 1; i < C_anchor_index; i++)
            {

                // Rebuild the backbone of the ith residue of the peptide. In the for loop immediately following,
                // we will adjust the backbone atoms in the ith residue.
                rebuildPeptideSegment(residues, angles, i, i, true);

                for (whichAngleType = 0; whichAngleType < 3; whichAngleType++)
                {
                    // whichAngleType can be 0 (phi), 1 (psi) or 2 (omega).

                    applyRotation = true;

                    // DEBUG
                    //logger.debug(String.format("\n");
                    //logger.debug(String.format("curr: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", curr_CA[0], curr_CA[1], curr_CA[2],
                    //	curr_C1[0], curr_C1[1], curr_C1[2], curr_N[0], curr_N[1], curr_N[2]);

                    // Determine the rotation axis.

                    if (whichAngleType == ANGLE_TYPE_PHI)
                    {
                        VectorMath.difference(residues.get(i).getAtom("CA").getCoords(), residues.get(i - 1).getAtom("N").getCoords(), rotationAxis);
                        VectorMath.unitVector(rotationAxis, rotationAxis);
                        VectorMath.copyVector(residues.get(i - 1).getAtom("N").getCoords(), basePoint);
                    }
                    else if (whichAngleType == ANGLE_TYPE_PSI)
                    {
                        VectorMath.difference(residues.get(i).getAtom("C").getCoords(), residues.get(i).getAtom("CA").getCoords(), rotationAxis);
                        VectorMath.unitVector(rotationAxis, rotationAxis);
                        VectorMath.copyVector(residues.get(i).getAtom("CA").getCoords(), basePoint);
                    }
                    else
                    { // if (whichAngleType == ANGLE_TYPE_OMEGA)
                        VectorMath.difference(residues.get(i).getAtom("N").getCoords(), residues.get(i).getAtom("C").getCoords(), rotationAxis);
                        VectorMath.unitVector(rotationAxis, rotationAxis);
                        VectorMath.copyVector(residues.get(i).getAtom("C").getCoords(), basePoint);
                    }

                    // Calculate vec_O1, vec_O2, vec_O3.

                    VectorMath.difference(curr_CA, basePoint, curr_CA_translated);
                    VectorMath.copyVector(rotationAxis, temp);
                    VectorMath.scale(temp, VectorMath.dotproduct(curr_CA_translated, rotationAxis) / VectorMath.dotproduct(rotationAxis, rotationAxis));
                    VectorMath.summation(temp, basePoint, vec_O1); // Translate vec_O1 into normal peptide coordinate space.

                    VectorMath.difference(curr_C1, basePoint, curr_C1_translated);
                    VectorMath.copyVector(rotationAxis, temp);
                    VectorMath.scale(temp, VectorMath.dotproduct(curr_C1_translated, rotationAxis) / VectorMath.dotproduct(rotationAxis, rotationAxis));
                    VectorMath.summation(temp, basePoint, vec_O2); // Translate vec_O2 into normal peptide coordinate space.

                    VectorMath.difference(curr_N, basePoint, curr_N_translated);
                    VectorMath.copyVector(rotationAxis, temp);
                    VectorMath.scale(temp, VectorMath.dotproduct(curr_N_translated, rotationAxis) / VectorMath.dotproduct(rotationAxis, rotationAxis));
                    VectorMath.summation(temp, basePoint, vec_O3); // Translate vec_O3 into normal peptide coordinate space.

                    // Calculate vec_r1, vec_r2, vec_r3.

                    VectorMath.difference(curr_CA, vec_O1, vec_r1);
                    VectorMath.difference(curr_C1, vec_O2, vec_r2);
                    VectorMath.difference(curr_N, vec_O3, vec_r3);

                    // Calculate vec_r1_unit, vec_r2_unit, vec_r3_unit.

                    VectorMath.unitVector(vec_r1, vec_r1_unit);
                    VectorMath.unitVector(vec_r2, vec_r2_unit);
                    VectorMath.unitVector(vec_r3, vec_r3_unit);

                    // Calculate mag_r1, mag_r2, mag_r3.

                    mag_r1 = VectorMath.distance(curr_CA, vec_O1);
                    mag_r2 = VectorMath.distance(curr_C1, vec_O2);
                    mag_r3 = VectorMath.distance(curr_N, vec_O3);

                    // Calculate vec_s1, vec_s2, vec_s3.

                    VectorMath.crossproduct(vec_r1_unit, rotationAxis, vec_s1);
                    VectorMath.crossproduct(vec_r2_unit, rotationAxis, vec_s2);
                    VectorMath.crossproduct(vec_r3_unit, rotationAxis, vec_s3);

                    // Calculate vec_f1, vec_f2, vec_f3.

                    VectorMath.difference(target_CA, vec_O1, vec_f1);
                    VectorMath.difference(target_C1, vec_O2, vec_f2);
                    VectorMath.difference(target_N, vec_O3, vec_f3);

                    // Calculate tan(alpha).

                    tan_alpha = (VectorMath.dotproduct(vec_f1, vec_s1) * mag_r1 + VectorMath.dotproduct(vec_f2, vec_s2) * mag_r2 + VectorMath.dotproduct(vec_f3, vec_s3) * mag_r3)
                            / (VectorMath.dotproduct(vec_f1, vec_r1_unit) * mag_r1 + VectorMath.dotproduct(vec_f2, vec_r2_unit) * mag_r2 + VectorMath.dotproduct(vec_f3, vec_r3_unit) * mag_r3);

                    alpha_rad = Math.atan(tan_alpha);

                    // Calculate alpha. Use the second derivative test to determine which alpha angle is the right one.

                    second_deriv = 2.0 * (
                            Math.cos(alpha_rad) * VectorMath.dotproduct(vec_f1, vec_r1_unit) * mag_r1 + Math.sin(alpha_rad) * VectorMath.dotproduct(vec_f1, vec_s1) * mag_r1
                                    + Math.cos(alpha_rad) * VectorMath.dotproduct(vec_f2, vec_r2_unit) * mag_r2 + Math.sin(alpha_rad) * VectorMath.dotproduct(vec_f2, vec_s2) * mag_r2
                                    + Math.cos(alpha_rad) * VectorMath.dotproduct(vec_f3, vec_r3_unit) * mag_r3 + Math.sin(alpha_rad) * VectorMath.dotproduct(vec_f3, vec_s3) * mag_r3);

                    if (second_deriv < 0.0)
                    {
                        // Current alpha_rad is incorrect; change it by pi radians.
                        if (alpha_rad >= 0.0)
                        {
                            alpha_rad -= Math.PI;
                        }
                        else
                        {
                            alpha_rad += Math.PI;
                        }
                    }

                    alpha = alpha_rad * 180.0 / Math.PI; // Convert to degrees.

                    // Adjust the angle just in case rounding error has placed it outside the range -180.0 to 180.0.
                    if (alpha < -180.0)
                    {
                        alpha += 360.0;
                    }
                    else if (alpha > 180.0)
                    {
                        alpha -= 360.0;
                    }

                    // DEBUG
                    //logger.debug(String.format("rotationAxis: (%f, %f, %f)\n", rotationAxis[0], rotationAxis[1], rotationAxis[2]);
                    /*
                     logger.debug(String.format("vec_O: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", vec_O1[0], vec_O1[1], vec_O1[2],
                         vec_O2[0], vec_O2[1], vec_O2[2], vec_O3[0], vec_O3[1], vec_O3[2]);
                     logger.debug(String.format("vec_r: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", vec_r1[0], vec_r1[1], vec_r1[2],
                         vec_r2[0], vec_r2[1], vec_r2[2], vec_r3[0], vec_r3[1], vec_r3[2]);
                     logger.debug(String.format("mag_r: %f, %f, %f\n", mag_r1, mag_r2, mag_r3);
                     logger.debug(String.format("vec_s: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", vec_s1[0], vec_s1[1], vec_s1[2],
                         vec_s2[0], vec_s2[1], vec_s2[2], vec_s3[0], vec_s3[1], vec_s3[2]);
                     logger.debug(String.format("vec_f: (%f, %f, %f), (%f, %f, %f), (%f, %f, %f)\n", vec_f1[0], vec_f1[1], vec_f1[2],
                         vec_f2[0], vec_f2[1], vec_f2[2], vec_f3[0], vec_f3[1], vec_f3[2]);
                     logger.debug(String.format("tan_alpha = %f, alpha_rad = %f, alpha = %f\n", tan_alpha, alpha_rad, alpha);
                     */

                    // DEBUG - calculate the expected sumOfSquaredVectorMath.distances that should result from applying this alpha.
                    // See equations (5) in the paper.

                    /*
                     expectedSumOfSquaredVectorMath.distances =
                         mag_r1*mag_r1 + mag_f1*mag_f1
                         - 2.0*mag_r1*Math.cos(alpha_rad)*VectorMath.dotproduct(vec_f1, vec_r1_unit) - 2.0*mag_r1*Math.sin(alpha_rad)*VectorMath.dotproduct(vec_f1, vec_s1)
                         + mag_r2*mag_r2 + mag_f2*mag_f2
                         - 2.0*mag_r2*Math.cos(alpha_rad)*VectorMath.dotproduct(vec_f2, vec_r2_unit) - 2.0*mag_r2*Math.sin(alpha_rad)*VectorMath.dotproduct(vec_f2, vec_s2)
                         + mag_r3*mag_r3 + mag_f3*mag_f3
                         - 2.0*mag_r3*Math.cos(alpha_rad)*VectorMath.dotproduct(vec_f3, vec_r3_unit) - 2.0*mag_r3*Math.sin(alpha_rad)*VectorMath.dotproduct(vec_f3, vec_s3);
                     */

                    // Test the alpha angle and, if necessary, adjust it based on constraints imposed by peptide geometry. Then apply
                    // the alpha angle.

                    if (whichAngleType == ANGLE_TYPE_PHI)
                    {
                        desired_phi = angles.get(i).phi;

                        desired_phi -= alpha;
                        if (desired_phi < -180.0)
                        {
                            desired_phi += 360.0;
                        }
                        else if (desired_phi > 180.0)
                        {
                            desired_phi -= 360.0;
                        }

                        phi = getClosestAllowedAngle(residues, angles, i, numRes, desired_phi, ANGLE_TYPE_PHI);

                        //if (phi != angles.get(i).phi) {
                        if (Math.abs(phi - angles.get(i).phi) > ANGLE_CHANGE_THRESHOLD)
                        {

                            // DEBUG
                            if (DEBUG)
                            {
                                if (desired_phi != phi)
                                {
                                    logger.debug(String.format("%s %d: constraints changed phi: %f --> %f (original = %f)\n", residues.get(i).getPDBName(), i,
                                            desired_phi, phi, angles.get(i).phi));
                                    logger.debug(String.format("   staring at phi = %f (phiindex = %d), psi = %f (psiindex = %d)\n",
                                            angles.get(i).phi, getPhiIndex(angles.get(i).phi), angles.get(i).psi, getPsiIndex(angles.get(i).psi)));
                                    logger.debug(String.format("   ending at phi = %f (phiindex = %d), desired_phi = %f (phiindex = %d)\n",
                                            phi, getPhiIndex(phi), desired_phi, getPhiIndex(desired_phi)));
                                }
                            }

                            alpha = phi - angles.get(i).phi; // Set the constrained alpha value, which will be used to calculate curr_CA etc.

                            angles.get(i).phi = phi;
                            // Set the new angle, which will be used when rebuildPeptideSegment() is called after the
                            // segment is closed.
                        }
                        else
                        {
                            // The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
                            // for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
                            // computation.
                            /*
                               if (DEBUG) {
                                   if (phi != angles.get(i).phi) {
                                       logger.debug(String.format("Did not rotate by %f degrees.\n", fabs(phi - angles.get(i).phi));
                                   }
                               }
                               */
                            applyRotation = false;
                        }
                    }
                    else if (whichAngleType == ANGLE_TYPE_PSI)
                    {
                        desired_psi = angles.get(i).psi;

                        desired_psi -= alpha;
                        if (desired_psi < -180.0)
                        {
                            desired_psi += 360.0;
                        }
                        else if (desired_psi > 180.0)
                        {
                            desired_psi -= 360.0;
                        }

                        psi = getClosestAllowedAngle(residues, angles, i, numRes, desired_psi, ANGLE_TYPE_PSI);

                        //if (psi != angles.get(i).psi) {
                        if (Math.abs(psi - angles.get(i).psi) > ANGLE_CHANGE_THRESHOLD)
                        {

                            // DEBUG
                            if (DEBUG)
                            {
                                if (desired_psi != psi)
                                {
                                    logger.debug(String.format("%s %d: constraints changed psi: %f --> %f (original = %f)\n", residues.get(i).getPDBName(), i,
                                            desired_psi, psi, angles.get(i).psi));
                                    logger.debug(String.format("   staring at phi = %f (phiindex = %d), psi = %f (psiindex = %d)\n",
                                            angles.get(i).phi, getPhiIndex(angles.get(i).phi), angles.get(i).psi, getPsiIndex(angles.get(i).psi)));
                                    logger.debug(String.format("   ending at psi = %f (psiindex = %d), desired_psi = %f (psiindex = %d)\n",
                                            psi, getPsiIndex(psi), desired_psi, getPsiIndex(desired_psi)));
                                }
                            }

                            alpha = psi - angles.get(i).psi; // Set the constrained alpha value, which will be used to calculate curr_CA etc.

                            angles.get(i).psi = psi;
                            // Set the new angle, which will be used when rebuildPeptideSegment() is called after the
                            // segment is closed.
                        }
                        else
                        {
                            // The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
                            // for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
                            // computation.
                            /*
                               if (DEBUG) {
                                   if (psi != angles.get(i).psi) {
                                       logger.debug(String.format("Did not rotate by %f degrees.\n", Math.abs(psi - angles.get(i).psi));
                                   }
                               }
                               */
                            applyRotation = false;
                        }
                    }
                    else
                    { // whichAngleType == ANGLE_TYPE_OMEGA
                        desired_omega = angles.get(i).omega;

                        desired_omega -= alpha;
                        if (desired_omega < -180.0)
                        {
                            desired_omega += 360.0;
                        }
                        else if (desired_omega > 180.0)
                        {
                            desired_omega -= 360.0;
                        }

                        omega = getClosestAllowedAngle(residues, angles, i, numRes, desired_omega, ANGLE_TYPE_OMEGA);

                        //if (omega != angles.get(i).omega) {
                        if (Math.abs(omega - angles.get(i).omega) > ANGLE_CHANGE_THRESHOLD)
                        {

                            // DEBUG
                            if (DEBUG)
                            {
                                if (desired_omega != omega)
                                {
                                    logger.debug(String.format("%s %d: constraints changed omega: %f --> %f (original = %f)\n", residues.get(i).getPDBName(), i,
                                            desired_omega, omega, angles.get(i).omega));
                                }
                            }

                            alpha = omega - angles.get(i).omega; // Set the constrained alpha value, which will be used to calculate curr_CA etc.

                            angles.get(i).omega = omega;
                            // Set the new angle, which will be used when rebuildPeptideSegment() is called after the
                            // segment is closed.
                        }
                        else
                        {
                            // The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
                            // for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
                            // computation.
                            /*
                               if (DEBUG) {
                                   if (omega != angles.get(i).omega) {
                                       logger.debug(String.format("Did not rotate by %f degrees.\n", Math.abs(omega - angles.get(i).omega));
                                   }
                               }
                               */
                            applyRotation = false;
                        }
                    }

                    if (applyRotation)
                    {

                        // Reposition the backbone atoms of the current ith residue according to the
                        // change specified by alpha.
                        if (whichAngleType == ANGLE_TYPE_PHI)
                        {

                            VectorMath.difference(residues.get(i).getAtom("C").getCoords(), basePoint, residues.get(i).getAtom("C").getCoords());
                            VectorMath.difference(residues.get(i).getAtom("N").getCoords(), basePoint, residues.get(i).getAtom("N").getCoords());
                            VectorMath.difference(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords());

                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("C").getCoords(), alpha);
                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("N").getCoords(), alpha);
                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("CA2").getCoords(), alpha);

                            VectorMath.summation(residues.get(i).getAtom("C").getCoords(), basePoint, residues.get(i).getAtom("C").getCoords()); // Translate back into normal peptide coordinate space.
                            VectorMath.summation(residues.get(i).getAtom("N").getCoords(), basePoint, residues.get(i).getAtom("N").getCoords()); // Translate back into normal peptide coordinate space.
                            VectorMath.summation(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords()); // Translate back into normal peptide coordinate space.
                        }
                        else if (whichAngleType == ANGLE_TYPE_PSI)
                        {

                            VectorMath.difference(residues.get(i).getAtom("N").getCoords(), basePoint, residues.get(i).getAtom("N").getCoords());
                            VectorMath.difference(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords());

                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("N").getCoords(), alpha);
                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("CA2").getCoords(), alpha);

                            VectorMath.summation(residues.get(i).getAtom("N").getCoords(), basePoint, residues.get(i).getAtom("N").getCoords()); // Translate back into normal peptide coordinate space.
                            VectorMath.summation(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords()); // Translate back into normal peptide coordinate space.
                        }
                        else
                        { // whichAngleType == ANGLE_TYPE_OMEGA

                            VectorMath.difference(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords());

                            VectorMath.rotateCC(rotationAxis, residues.get(i).getAtom("CA2").getCoords(), alpha);

                            VectorMath.summation(residues.get(i).getAtom("CA2").getCoords(), basePoint, residues.get(i).getAtom("CA2").getCoords()); // Translate back into normal peptide coordinate space.
                        }

                        // Calculate coordinates of movable C-terminal residue directly from alpha.

                        VectorMath.difference(curr_CA, basePoint, curr_CA_translated);
                        VectorMath.difference(curr_C1, basePoint, curr_C1_translated);
                        VectorMath.difference(curr_N, basePoint, curr_N_translated);

                        VectorMath.rotateCC(rotationAxis, curr_CA_translated, alpha);
                        VectorMath.rotateCC(rotationAxis, curr_C1_translated, alpha);
                        VectorMath.rotateCC(rotationAxis, curr_N_translated, alpha);

                        VectorMath.summation(curr_CA_translated, basePoint, curr_CA); // Translate back into normal peptide coordinate space.
                        VectorMath.summation(curr_C1_translated, basePoint, curr_C1); // Translate back into normal peptide coordinate space.
                        VectorMath.summation(curr_N_translated, basePoint, curr_N); // Translate back into normal peptide coordinate space.
                    }

                    // Only for DEBUG
                    /*
                     sumOfSquaredVectorMath.distances =
                         VectorMath.distance(curr_CA, target_CA) * VectorMath.distance(curr_CA, target_CA)
                         + VectorMath.distance(curr_C1, target_C1) * VectorMath.distance(curr_C1, target_C1)
                         + VectorMath.distance(curr_N, target_N) * VectorMath.distance(curr_N, target_N);

                     logger.debug(String.format("sumOfSquaredVectorMath.distances = %f\n", sumOfSquaredVectorMath.distances);
                     */

                }

            }

            // After each pass of the CCD algorithm through the whole segment, measure the sum of squared
            // VectorMath.distances between the current C-terminal residue atom coordinates and original C-anchor
            // atom coordinates. Apply the CCD algorithm to the segment repeatedly until the VectorMath.difference
            // is below a certain tolerance threshold where the two sets of coordinates almost perfectly
            // overlap.

            sumOfSquaredDistances =
                    VectorMath.distance(curr_CA, target_CA) * VectorMath.distance(curr_CA, target_CA)
                            + VectorMath.distance(curr_C1, target_C1) * VectorMath.distance(curr_C1, target_C1)
                            + VectorMath.distance(curr_N, target_N) * VectorMath.distance(curr_N, target_N);

            num_passes++;

            //logger.debug(String.format("   num_passes = %d, sumOfSquaredVectorMath.distances = %f\n", num_passes, sumOfSquaredVectorMath.distances);
        }

        if (!closureFailed)
        {

            if (DEBUG)
            {
                logger.debug(String.format("Loop closed in %d passes.\n", num_passes));
            }

            // Loop closure has succeeded. Math.since our calls above to rebuildPeptideSegment set onlyBackbone
            // to true, the non-backbone atoms are in the wrong positions. Therefore we should rebuild
            // the current peptide segment with onlyBackbone set to false to get all atoms in the right
            // place. In addition, the C-terminal coordinates of the residue at C_anchor_index have been
            // slightly changed as a result of the CCD algorithm, which may slightly distort the structure
            // from the point of this segment to the C-terminus of the entire peptide. Therefore we should
            // also rebuild the portion of the structure after the current segment.

            rebuildPeptideSegment(residues, angles, N_anchor_index, numRes - 1, false);
        }

        return !closureFailed;
    }

/*************************************************************/
/* Rebuilds a segment of the peptide from startIndex to      */
/* endIndex. Assumes that the atoms for the residue at index */
/* (startIndex - 1) have already been set, for example by a  */
/* previous call to buildPeptide. If onlyBackbone is true,   */

    /* only the backbone atoms will be manipulated.              */

    /**
     * *********************************************************
     */
    static void rebuildPeptideSegment(ArrayList<Group> residues, TreeMap<Integer, Angle> angles, int startIndex, int endIndex, boolean onlyBackbone) throws StructureException
    {
        int i;

        if (startIndex <= 0)
        {
            logger.error("Error: rebuildPeptideSegment called with startIndex <= 0\n");
        }
        else if (startIndex > endIndex)
        {
            logger.error("Error: rebuildPeptideSegment called with startIndex > endIndex\n");
        }

        for (i = startIndex; i <= endIndex; i++)
        {
            nextPeptide(residues, angles, i, onlyBackbone);
        }
    }

    /* iteratively assembles the peptide */
    /*static void buildPeptide(List<Group> residues, int numRes) throws StructureException
    {
        int i;

        // TODO why?
        // initResData( & residues[0]); // Need to call this for the first residue before calling nextPeptide.


       3
        for (i = 0; i < numRes; i++)
        {
            Group res = residues.get(i);
            Group prev_res = residues.get(i);
            nextPeptide(res, prev_res, false);
        }
    }
*/
/*************************************************************/
/* Add the next residue to the peptide chain, and carry out  */
/* the appropriate rotations and translations.               */
/* Set onlyBackbone to true to only perform operations on    */

    /* backbone atoms.                                           */

    /**
     * *********************************************************
     */
    static void nextPeptide(ArrayList<Group> residues, TreeMap<Integer, Angle> angles, int index, boolean onlyBackbone) throws StructureException
    {
        double[] CN = new double[3];

        /* if (num > 0)
        {

            // TODO why?
            // copyPosition( & residues[num - 1], res);
        }*/

        Group res = residues.get(index);

        applyTemplate(res);

        // Perform omega rotation for the peptide bond in the current residue.
        // This will potentially result in the relocation of the HN
        // and CA2 atoms in the current residue. This needs to be done
        // when reading residues from the template database, which have omega
        // angles that do not agree with those provided in the angles file. If,
        // on the other hand, PepMake has been given a PDB file, PepMake has
        // called phipsi with the --db option to generate a residue coordinates
        // file that already has the correct omega rotations implicit in the coordinates.
        // However, if we will be adjusting the omega angles (such as with fixAngles())
        // or if PepMake is used as part of another program such as GAFolder, then
        // we may want to create a peptide with omega angles that differ from those
        // of the given PDB file, and the code below will adjust the omega angles of
        // the peptide.

        if (adjustOmega)
        {
            double omega = angles.get(index).omega;
            //int debugNumToPrint = 1;

            //debug
            //if (num == debugNumToPrint) {
            //	outputDebugCoords("coords_for_Grapher/before_omega.txt", res->atoms, true, false);
            //}

            // determine omega currently present in the residue template
            double previous_omega = torsion(res.getAtom("CA").getCoords(), res.getAtom("C").getCoords(), res.getAtom("N").getCoords(), res.getAtom("CA2").getCoords());

            VectorMath.difference(res.getAtom("N").getCoords(), res.getAtom("C").getCoords(), CN);
            VectorMath.unitVector(CN, CN);
            //printf("CN: (%f, %f, %f)\n", CN[0], CN[1], CN[2]); // debug

            // Translate to origin - but only translate the atoms that we will rotate.
            double[] C1_vec = new double[3];
            VectorMath.copyVector(res.getAtom("C").getCoords(), C1_vec);
            //for (i=0; i<res->numAtoms; i++) {
            //	VectorMath.difference(res->atoms[i], C1_vec, res->atoms[i]);
            //}
            VectorMath.difference(res.getAtom("CA2").getCoords(), C1_vec, res.getAtom("CA2").getCoords());
            if (!onlyBackbone)
            {
                VectorMath.difference(res.getAtom("HN").getCoords(), C1_vec, res.getAtom("HN").getCoords());
            }

            //debug
            //if (num == debugNumToPrint) {
            //	outputDebugCoords("coords_for_Grapher/omega_to_origin.txt", res->atoms, true, false);
            //}

            // perform omega rotation
            VectorMath.rotateCC(CN, res.getAtom("CA2").getCoords(), omega - previous_omega);
            if (!onlyBackbone)
            {
                VectorMath.rotateCC(CN, res.getAtom("HN").getCoords(), omega - previous_omega);
            }

            // translate back to original position
            //for (i=0; i<res->numAtoms; i++) {
            //	summation(res->atoms[i], C1_vec, res->atoms[i]);
            //}
            VectorMath.summation(res.getAtom("CA2").getCoords(), C1_vec, res.getAtom("CA2").getCoords());
            if (!onlyBackbone)
            {
                VectorMath.summation(res.getAtom("HN").getCoords(), C1_vec, res.getAtom("HN").getCoords());
            }

            //debug
            //if (num == debugNumToPrint) {
            //	outputDebugCoords("coords_for_Grapher/after_omega.txt", res->atoms, true, false);
            //}

            //debug
            //double resulting_omega = torsion(res->atoms[CA], res->atoms[C1], res->atoms[N], res->atoms[CA2]);
            //printf("Performing omega rotation of %f for residue %d (previous omega = %f, resulting omega = %f)\n", omega, num, previous_omega, resulting_omega);
        }

        rotatePhiPsi(residues, angles, index, true, onlyBackbone);

    }

/*************************************************************/
/* Rotate the current residue by phi & psi. This function    */
/* also moves the residue from its old position (the same    */
/* position as the previous residue) to its new position.    */
/*                                                           */
/* Set rotateChi to true to perform rotation of the chi      */
/* angle. Set onlyBackbone to true to only manipulate        */
/* backhone atoms.                                           */
/*                                                           */
/* This version of the function is a complete reimplemention */

    /* done by David Arndt in October-November 2006.             */

    /**
     * *********************************************************
     */
    static void rotatePhiPsi(List<Group> residues, TreeMap<Integer, Angle> angles, int index, boolean rotateChi, boolean onlyBackbone) throws StructureException
    {
        int i;
        double N_CA_C1_angle;
        double[] N_CA_axis = new double[3]; // Vector parallel to N-CA2 bond in previous residue.
        double[] CA_C1_axis = new double[3]; // Vector parallel to CA-C1 bond in current residue.
        double[] L = new double[3];

        // debug variables
        int debugNumToPrint = 0; // Which residue to print for debug info.
        // int debugOutputPRO = (res - > resName.equals("PRO"));
        boolean debugPrintCB = true;
        double intended_phi, intended_psi;
        int num = 0;

        Group res = residues.get(index);
        Group prev_res = residues.get(index - 1);
        Angle angle = angles.get(index);
        List<Atom> atoms = res.getAtoms();

        //debug
        if (DEBUG)
        {
            intended_phi = angle.phi;
            intended_psi = angle.psi;
            /*   if (num == debugNumToPrint)
            {
                printf("Saving debug info for residue %d.\n", num);
                printf("Intended rotation: phi = %f, psi = %f, chi = %f\n", angle.phi, angle.psi, res - > chi);
            }
            if (num == debugNumToPrint && num > 0)
            {
                outputDebugCoords("coords_for_Grapher/00_prev.txt", prev_res - > atoms, debugPrintCB, debugOutputPRO);
            }
            if (num == debugNumToPrint && num > 0)
            {
                double[] prev_trans[
                MAX_ATOMS]; // Previous residue translated to origin by CA2. For debug only.

                // Translate previous residue to the origin.
                for (i = 0; i < prev_res - > numAtoms; i++)
                {
                    VectorMath.difference(prev_res.getAtom("i").getCoords(), prev_res.getAtom("CA2").getCoords(), prev_trans[i]);
                }

                outputDebugCoords("coords_for_Grapher/01_prev_trans.txt", prev_trans, debugPrintCB, debugOutputPRO);
            }
            if (num == debugNumToPrint)
            {
                outputDebugCoords("coords_for_Grapher/02_init.txt", res - > atoms, debugPrintCB, debugOutputPRO);
            }*/
        }

        double[] C1_vec = res.getAtom("C").getCoords();
        double[] CA_vec = res.getAtom("CA").getCoords();
        VectorMath.difference(C1_vec, CA_vec, CA_C1_axis);
        VectorMath.unitVector(CA_C1_axis, CA_C1_axis);
        if (num > 0)
        {
            VectorMath.difference(prev_res.getAtom("CA2").getCoords(), prev_res.getAtom("N").getCoords(), N_CA_axis);
            VectorMath.unitVector(N_CA_axis, N_CA_axis);
        }
        else
        {
            N_CA_axis[0] = 1.0;
            N_CA_axis[1] = 0.0;
            N_CA_axis[2] = 0.0;
        }

        //debug
        if (DEBUG)
        {
            if (num == debugNumToPrint)
            {

                VectorMath.crossproduct(CA_C1_axis, N_CA_axis, L); // Do this ahead of time for debugging purposes.
                VectorMath.unitVector(L, L);

                // fp_debug = fopen("coords_for_Grapher/03_vec.txt", "w");
                logger.debug(String.format("%f\t%f\t%f\n", 0.0, 0.0, 0.0));
                logger.debug(String.format("%f\t%f\t%f\n", CA_C1_axis[0], CA_C1_axis[1], CA_C1_axis[2]));
                logger.debug(String.format("%f\t%f\t%f\n", 0.0, 0.0, 0.0));
                logger.debug(String.format("%f\t%f\t%f\n", N_CA_axis[0], N_CA_axis[1], N_CA_axis[2]));
                logger.debug(String.format("%f\t%f\t%f\n", 0.0, 0.0, 0.0));
                logger.debug(String.format("%f\t%f\t%f\n", L[0], L[1], L[2]));
                // fclose(fp_debug);
            }
        }

        // Translate current residue to the origin with CA at origin.
        double[] old_CA = new double[3];
        VectorMath.copyVector(CA_vec, old_CA);
        if (!onlyBackbone)
        {
            for (i = 0; i < atoms.size(); i++)
            {
                VectorMath.difference(atoms.get(i).getCoords(), old_CA, atoms.get(i).getCoords());
            }
        }
        else
        {
            for (i = 0; i <= CA2; i++)
            {
                VectorMath.difference(atoms.get(i).getCoords(), old_CA, atoms.get(i).getCoords());
            }
        }

        //debug
        /*  if (debug && num == debugNumToPrint)
        {
            outputDebugCoords("coords_for_Grapher/04_to_origin.txt", res - > atoms, debugPrintCB, debugOutputPRO);
        }*/

        // Rotate the current residue so that its CA-C1 bond is parallel to the N-CA2 bond of the previous residue.
        VectorMath.crossproduct(CA_C1_axis, N_CA_axis, L);
        double[] N_vec = res.getAtom("N").getCoords();
        if (!(L[0] == 0 && L[1] == 0 && L[2] == 0) && num > 0)
        {
            // Only do the rotation if L is non-zero. If it is zero, the bonds are already parallel, and we avoid
            // a call to VectorMath.rotateCC that would set all atom coordinates to (0, 0, 0).

            VectorMath.unitVector(L, L);

            if (!onlyBackbone)
            {
                for (i = 0; i < atoms.size(); i++)
                {
                    VectorMath.rotateCC(L, atoms.get(i).getCoords(), VectorMath.angle(N_CA_axis, CA_C1_axis));
                }
            }
            else
            {
                for (i = 0; i <= CA2; i++)
                {
                    VectorMath.rotateCC(L, atoms.get(i).getCoords(), VectorMath.angle(N_CA_axis, CA_C1_axis));
                }
            }

            // We want the C1-N-CA2 of the previous residue to be in the same plane as CA-C1-N of the current residue,
            // but they may be off slightly, so we need to do a rotation of the current residue to straighten it.

            double[] C1_N_prev = new double[3];
            double[] C1_N_curr = new double[3];
            double[] cp1 = new double[3];
            double[] cp2 = new double[3];
            double[] cp12 = new double[3];

            VectorMath.difference(prev_res.getAtom("N").getCoords(), prev_res.getAtom("C").getCoords(), C1_N_prev);
            VectorMath.difference(N_vec, C1_vec, C1_N_curr);
            VectorMath.crossproduct(C1_N_prev, N_CA_axis, cp1);
            //printf("%f\t%f\t%f\n", cp1[0], cp1[1], cp1[2]);
            VectorMath.unitVector(cp1, cp1);
            VectorMath.crossproduct(C1_N_curr, N_CA_axis, cp2);
            //printf("%f\t%f\t%f\n", cp2[0], cp2[1], cp2[2]);
            VectorMath.unitVector(cp2, cp2);
            VectorMath.crossproduct(cp1, cp2, cp12);
            double C1_N_CA_C1_N_angle = VectorMath.angle(cp1, cp2);
            if (VectorMath.dotproduct(N_CA_axis, cp12) > 0.0)
            {
                //printf("   *VectorMath.dotproduct(N_CA_axis, cp12) > 0.0\n");
                C1_N_CA_C1_N_angle = -C1_N_CA_C1_N_angle;
            }

            //printf("   About to apply rotation by C1-N-CA-C1-N angle: %f\n", C1_N_CA_C1_N_angle);
            if (!onlyBackbone)
            {
                for (i = 0; i < atoms.size(); i++)
                {
                    VectorMath.rotateCC(N_CA_axis, atoms.get(i).getCoords(), C1_N_CA_C1_N_angle);
                }
            }
            else
            {
                for (i = 0; i <= CA2; i++)
                {
                    VectorMath.rotateCC(N_CA_axis, atoms.get(i).getCoords(), C1_N_CA_C1_N_angle);
                }
            }

            // Debug. Test angle after applying rotation.
            /*
           VectorMath.difference(prev_res->atoms[N], prev_res->atoms[C1], C1_N_prev);
           VectorMath.difference(res->atoms[N], res->atoms[C1], C1_N_curr);
           VectorMath.crossproduct(C1_N_prev, N_CA_axis, cp1);
           VectorMath.unitVector(cp1, cp1);
           VectorMath.crossproduct(C1_N_curr, N_CA_axis, cp2);
           VectorMath.unitVector(cp2, cp2);
           VectorMath.crossproduct(cp1, cp2, cp12);
           C1_N_CA_C1_N_angle = angle(cp1, cp2);
           if (VectorMath.dotproduct(N_CA_axis, cp12) > 0.0) {
               //printf("   *VectorMath.dotproduct(N_CA_axis, cp12) > 0.0\n");
               C1_N_CA_C1_N_angle = -C1_N_CA_C1_N_angle;
           }
           printf("   After rotation C1-N-CA-C1-N angle: %f\n", C1_N_CA_C1_N_angle);
           */

        }

        //debug
        /*     if (debug && num == debugNumToPrint)
        {
            outputDebugCoords("coords_for_Grapher/05_N_CA_par_CA_C1.txt", res - > atoms, debugPrintCB, debugOutputPRO);
        }*/

        // We assume that psi == 180 degrees. That is, the R-group of the current residue is oriented in such a way
        // that the bond from the first CA of the current residue leading to N is in the same plane.

        // Before this rotation, the current residue's N-CA-C1 angle (N from the previous residue, CA and C1 from the
        // current residue) is 180 degrees. Rotate the current residue so that this angle is equal to N_CA_C1_angle.
        double[] C1_N_axis_temp = new double[3]; // C1-N axis of current residue
        double[] rotate_axis_temp = new double[3];
        VectorMath.difference(C1_vec, N_vec, C1_N_axis_temp);
        VectorMath.crossproduct(N_CA_axis, C1_N_axis_temp, rotate_axis_temp);
        VectorMath.unitVector(rotate_axis_temp, rotate_axis_temp);

        double[] CA_N = new double[3];
        double[] CA_C1 = new double[3];

        VectorMath.difference(N_vec, CA_vec, CA_N);
        VectorMath.difference(C1_vec, CA_vec, CA_C1);
        N_CA_C1_angle = VectorMath.angle(CA_N, CA_C1);

        assert Math.abs(N_CA_C1_angle - 111) < 5 : "N_CA_C1_angle should be around 111 degrees, was " + N_CA_C1_angle;

        // debug
        if (DEBUG)
        {
            logger.debug(String.format("N_CA_C1_angle = %f\n", N_CA_C1_angle));
        }

        if (!onlyBackbone)
        {
            for (i = 0; i < atoms.size(); i++)
            {
                VectorMath.rotateCC(rotate_axis_temp, atoms.get(i).getCoords(), 180 - N_CA_C1_angle);
            }
        }
        else
        {
            for (i = 0; i <= CA2; i++)
            {
                VectorMath.rotateCC(rotate_axis_temp, atoms.get(i).getCoords(), 180 - N_CA_C1_angle);
            }
        }

        //debug
        /*    if (debug && num == debugNumToPrint)
        {
            outputDebugCoords("coords_for_Grapher/06_after_rotate_N_CA_C1_angle.txt", res - > atoms, debugPrintCB, debugOutputPRO);
        }*/

        // Re-calculate CA_C1_axis for the current residue; it has changes because of rotations of the current residue.
        VectorMath.difference(C1_vec, CA_vec, CA_C1_axis);
        VectorMath.unitVector(CA_C1_axis, CA_C1_axis);

        // Perform psi rotation. Note that we only rotate the backbone atoms after the CA-C1 bond.
        // (We omit rotating C1 because we know it will stay in the same place).
        VectorMath.rotateCC(CA_C1_axis, res.getAtom("O").getCoords(), angle.psi - 180);
        VectorMath.rotateCC(CA_C1_axis, N_vec, angle.psi - 180);
        VectorMath.rotateCC(CA_C1_axis, res.getAtom("CA2").getCoords(), angle.psi - 180);
        if (!onlyBackbone)
        {
            VectorMath.rotateCC(CA_C1_axis, res.getAtom("HN").getCoords(), angle.psi - 180);
        }
        // Also rotate the OXT atom (which should only appear in the last residue).
        for (i = CG; i < atoms.size(); i++)
        {
            if (! atoms.get(i).equals("OXT"))
            {
                VectorMath.rotateCC(CA_C1_axis, atoms.get(i).getCoords(), angle.psi - 180);
                break;
            }
        }

        //debug
        /* if (debug && num == debugNumToPrint)
                {
                    outputDebugCoords("coords_for_Grapher/07_after_psi.txt", res - > atoms, debugPrintCB, debugOutputPRO);
                }
        */
        // Perform phi rotation. Note that we do not rotate Np, HNp, HNp2, or HNp3 (nor CA because we know it
        // will stay in the same place).
        if (!onlyBackbone)
        {
            for (i = C1; i < HNp; i++)
            {
                VectorMath.rotateCC(N_CA_axis, atoms.get(i).getCoords(), angle.phi);
            }
            for (i = HA; i < atoms.size(); i++)
            {
                VectorMath.rotateCC(N_CA_axis, atoms.get(i).getCoords(), angle.phi);
            }
        }
        else
        {
            for (i = C1; i <= CA2; i++)
            {
                VectorMath.rotateCC(N_CA_axis, atoms.get(i).getCoords(), angle.phi);
            }
        }

        //debug
        /*if (DEBUG && num == DEBUGNumToPrint)
        {
            outputDEBUGCoords("coords_for_Grapher/08_after_phi.txt", res - > atoms, DEBUGPrintCB, DEBUGOutputPRO);
        }
*/
        // Perform chi rotation.
        if (rotateChi && !onlyBackbone && !res.getPDBName().equals("GLY") && !res.getPDBName().equals("ALA"))
        {

            double previous_chi;

            if (num > 0)
            {
                double[][] prev_trans = new double[MAX_ATOMS][]; // Previous residue translated to origin with CA2 at origin.

                // Translate previous residue to the origin. (Store result in prev_trans).
                List<Atom> prev_atoms = prev_res.getAtoms();
                for (i = 0; i < prev_atoms.size(); i++)
                {
                    VectorMath.difference(prev_atoms.get(i).getCoords(), prev_res.getAtom("CA2").getCoords(), prev_trans[i]);
                }

                previous_chi = torsion(prev_trans[N], CA_vec, res.getAtom("CB").getCoords(), res.getAtom("CG").getCoords());
                if (DEBUG)
                {
                    logger.debug(String.format("   torsion of %f from:\n", previous_chi));
                    logger.debug(String.format("   N: (%f, %f, %f), ", prev_trans[N][0], prev_trans[N][1], prev_trans[N][2]));
                    logger.debug(String.format("CA: (%f, %f, %f), ", CA_vec[0], CA_vec[1], CA_vec[2]));
                    logger.debug(String.format("CB: (%f, %f, %f), ", res.getAtom("CB").getCoords()[0], res.getAtom("CB").getCoords()[1], res.getAtom("CB").getCoords()[2]));
                    logger.debug(String.format("CG: (%f, %f, %f)\n", res.getAtom("CG").getCoords()[0], res.getAtom("CG").getCoords()[1], res.getAtom("CG").getCoords()[2]));
                }
            }
            else
            {
                previous_chi = torsion(res.getAtom("Np").getCoords(), CA_vec, res.getAtom("CB").getCoords(), res.getAtom("CG").getCoords());
                if (DEBUG)
                {
                    logger.debug(String.format("   torsion of %f from:\n", previous_chi));
                    logger.debug(String.format("   Np: (%f, %f, %f), ", res.getAtom("Np").getCoords()[0], res.getAtom("Np").getCoords()[1], res.getAtom("Np").getCoords()[2]));
                    logger.debug(String.format("CA: (%f, %f, %f), ", CA_vec[0], CA_vec[1], CA_vec[2]));
                    logger.debug(String.format("CB: (%f, %f, %f), ", res.getAtom("CB").getCoords()[0], res.getAtom("CB").getCoords()[1], res.getAtom("CB").getCoords()[2]));
                    logger.debug(String.format("CG: (%f, %f, %f)\n", res.getAtom("CG").getCoords()[0], res.getAtom("CG").getCoords()[1], res.getAtom("CG").getCoords()[2]));
                }
            }

            double[] CA_CB_axis = new double[3];
            VectorMath.difference(res.getAtom("CB").getCoords(), CA_vec, CA_CB_axis);
            VectorMath.unitVector(CA_CB_axis, CA_CB_axis);

            for (i = CG; i < atoms.size(); i++)
            {
                if (!atoms.get(i).equals("OXT"))
                { // Do not rotate the OXT atom (which should only appear in the last residue).
                    VectorMath.rotateCC(CA_CB_axis, atoms.get(i).getCoords(), angle.chi - previous_chi);
                }
            }

            // DEBUG
            if (DEBUG)
            {
                double achieved_chi;
                if (num > 0)
                {
                    double[][] prev_trans = new double[MAX_ATOMS][]; // Previous residue translated to origin with CA2 at origin.

                    // Translate previous residue to the origin. (Store result in prev_trans).
                    List<Atom> prev_atoms = prev_res.getAtoms();
                    for (i = 0; i < prev_atoms.size(); i++)
                    {
                        VectorMath.difference(prev_atoms.get(i).getCoords(), prev_res.getAtom("CA2").getCoords(), prev_trans[i]);
                    }

                    achieved_chi = torsion(prev_trans[N], CA_vec, res.getAtom("CB").getCoords(), res.getAtom("CG").getCoords());
                }
                else
                {
                    achieved_chi = torsion(res.getAtom("Np").getCoords(), CA_vec, res.getAtom("CB").getCoords(), res.getAtom("CG").getCoords());
                }
                logger.debug(String.format("   chi: prev = %f, intended = %f, achieved = %f\n", previous_chi, angle.chi, achieved_chi));
            }
        }

        //DEBUG
        /*  if (DEBUG && num == DEBUGNumToPrint)
        {
            outputDEBUGCoords("coords_for_Grapher/09_after_chi.txt", res - > atoms, DEBUGPrintCB, DEBUGOutputPRO);
        }*/

        // Translate current residue so that its first CA overlaps with CA2 of the previous residue.
        double[] prev_CA2 = new double[3];
        if (num > 0)
        {
            VectorMath.copyVector(prev_res.getAtom("CA2").getCoords(), prev_CA2);
        }
        else
        {
            prev_CA2[0] = 0.0;
            prev_CA2[1] = 0.0;
            prev_CA2[2] = 0.0;
        }
        // List<Atom> atoms = atoms;
        if (!onlyBackbone)
        {
            for (i = 0; i < atoms.size(); i++)
            {
                VectorMath.summation(atoms.get(i).getCoords(), prev_CA2, atoms.get(i).getCoords());
            }
        }
        else
        {
            for (i = 0; i <= CA2; i++)
            {
                VectorMath.summation(atoms.get(i).getCoords(), prev_CA2, atoms.get(i).getCoords());
            }
        }

        //DEBUG
        /*    if (DEBUG && num == DEBUGNumToPrint)
        {
            outputDEBUGCoords("coords_for_Grapher/10_done.txt", res - > atoms, DEBUGPrintCB, DEBUGOutputPRO);
        }*/

        //DEBUG
        if (DEBUG)
        {

            // Determine actual phi and psi achieved.
            if (num == debugNumToPrint && num > 0)
            {

                // Print the atom coordinates used for computation of achieved phi to a file.
                // fp_DEBUG = fopen("coords_for_Grapher/phi_atoms.txt", "w");
                logger.debug(String.format("%f\t%f\t%f\n", prev_res.getAtom("C").getCoords()[0], prev_res.getAtom("C").getCoords()[1], prev_res.getAtom("C").getCoords()[2]));
                logger.debug(String.format("%f\t%f\t%f\n", prev_res.getAtom("N").getCoords()[0], prev_res.getAtom("N").getCoords()[1], prev_res.getAtom("N").getCoords()[2]));
                logger.debug(String.format("%f\t%f\t%f\n", prev_res.getAtom("CA2").getCoords()[0], prev_res.getAtom("CA2").getCoords()[1], prev_res.getAtom("CA2").getCoords()[2]));
                logger.debug(String.format("%f\t%f\t%f\n", CA_vec[0], CA_vec[1], CA_vec[2])); // Should be the same as prev's CA2.
                logger.debug(String.format("%f\t%f\t%f\n", C1_vec[0], C1_vec[1], C1_vec[2]));
                // fclose(fp_DEBUG);

                // Compute achieved phi and psi.

                double phi_achieved = torsion(prev_res.getAtom("C").getCoords(), prev_res.getAtom("N").getCoords(), CA_vec, C1_vec);
                logger.debug(String.format("Acheived phi = %f\n", phi_achieved));

                double psi_achieved = torsion(prev_res.getAtom("N").getCoords(), CA_vec, C1_vec, N_vec);
                logger.debug(String.format("Acheived psi = %f\n", psi_achieved));
            }

            // Print all angle comparisons.
            if (num > 0)
            {
                double phi_achieved = torsion(prev_res.getAtom("C").getCoords(), prev_res.getAtom("N").getCoords(), CA_vec, C1_vec);
                //logger.debug(String.format("Acheived phi = %f\n", phi_achieved);

                double psi_achieved = torsion(prev_res.getAtom("N").getCoords(), CA_vec, C1_vec, N_vec);
                //logger.debug(String.format("Acheived psi = %f\n", psi_achieved);

                double chi_achieved;
                if (num > 0 && !(res.getPDBName().equals("ALA") || res.getPDBName().equals("GLY")))
                { // If this residue has a CG atom.
                    chi_achieved = torsion(prev_res.getAtom("N").getCoords(), CA_vec, res.getAtom("CB").getCoords(), res.getAtom("CG").getCoords());
                    logger.debug(String.format("%d (%s, %d): phi %f - %f = %f, psi %f - %f = %f, chi %f - %f = %f\n", num, res.getPDBName(), atoms.size(),
                            intended_phi, phi_achieved, intended_phi - phi_achieved,
                            intended_psi, psi_achieved, intended_psi - psi_achieved,
                            angle.chi, chi_achieved, angle.chi - chi_achieved));
                }
                else
                {
                    logger.debug(String.format("%d (%s, %d): phi %f - %f = %f, psi %f - %f = %f, chi N/A\n", num, res.getPDBName(), atoms.size(),
                            intended_phi, phi_achieved, intended_phi - phi_achieved,
                            intended_psi, psi_achieved, intended_psi - psi_achieved));
                }
            }

            logger.debug(String.format("\n"));
        }

    }

/*************************************************************/
    /* Calculate torsion angle for four points.                  */
/**
 * *********************************************************
 */
    static double torsion(double[] A, double[] B, double[] C, double[] D)
    {
        double[] X = new double[3];
        double[] Y = new double[3];
        double[] AB = new double[3];
        double[] BC = new double[3];
        double[] CD = new double[3];
        double[] cXY = new double[3];

        double result;

        VectorMath.difference(A, B, AB);
        VectorMath.difference(B, C, BC);
        VectorMath.difference(C, D, CD);

        VectorMath.crossproduct(AB, BC, X);
        VectorMath.crossproduct(BC, CD, Y);

        result = VectorMath.angle(X, Y);
        VectorMath.crossproduct(X, Y, cXY);
        if (VectorMath.dotproduct(BC, cXY) > 0.0)
        {
            result = -result;
        }

        return (result);
    }

/*************************************************************/
/* Apply the saved template of this residue's coordinates.   */
/* We rotate the residue into the canonical form, then swap  */
/* in the new coordinates, and then reverse all of the       */

    /* rotations that were used to canonisize the coordinates    */

    /**
     * @param res**********************************************************
     *
     */
    static void applyTemplate(Group res) throws StructureException
    {
        int i;
        double[] xAxis = {1, 0, 0}; // a unit vector
        double[] T = new double[3]; // a temp vector
        double thetaZ, thetaY, thetaX;
        double[] transVec = new double[3];

        // Translate to origin with CA at origin.
        VectorMath.copyVector(res.getAtom("CA").getCoords(), transVec); // store for later
        List<Atom> atoms = res.getAtoms();
        for (i = 0; i < atoms.size(); i++)
        {
            VectorMath.difference(atoms.get(i).getCoords(), transVec, atoms.get(i).getCoords());
        }

        // compute angle from X axis
        VectorMath.setVector(T, res.getAtom("C").getCoords()[0], res.getAtom("C").getCoords()[1], 0);
        thetaZ = VectorMath.angle(T, xAxis);
        if (res.getAtom("C").getCoords()[1] > 0) thetaZ = 360 - thetaZ;
        // rotate atoms around Z axis
        for (i = 0; i < atoms.size(); i++)
        {
            VectorMath.rotateZ(atoms.get(i).getCoords(), thetaZ);
        }

        // compute angle from X axis
        VectorMath.setVector(T, res.getAtom("C").getCoords()[0], 0, res.getAtom("C").getCoords()[2]);
        thetaY = VectorMath.angle(T, xAxis);
        if (res.getAtom("C").getCoords()[2] < 0) thetaY = 360 - thetaY;
        // rotate atoms around Y-axis
        for (i = 0; i < atoms.size(); i++) VectorMath.rotateY(atoms.get(i).getCoords(), thetaY);

        // compute angle for N
        thetaX = Math.atan(res.getAtom("N").getCoords()[2] / res.getAtom("N").getCoords()[1]);
        thetaX = 180 - (thetaX * 180.0 / Math.PI);
        if (res.getAtom("N").getCoords()[1] < 0) thetaX = 180 + thetaX;
        for (i = 0; i < atoms.size(); i++) VectorMath.rotateX(atoms.get(i).getCoords(), thetaX);

        for (i = 0; i < atoms.size(); i++)
        {
            logger.debug(String.format("%d :  %f %f %f\n",
            	i, atoms.get(i).getCoords()[0], atoms.get(i).getCoords()[1], atoms.get(i).getCoords()[2]));

            char code = 0;
            SideChain sideChain = null;
            try
            {
                code = HomodellerUtils.convert_3code_1code(res.getPDBName());
                sideChain = HomodellerUtils.getSideChain(code, 0);

                res = (Group) sideChain.clone();
            }
            catch (IllegalSymbolException e)
            {
                e.printStackTrace();
            }
            
            // VectorMath.copyVector(res - > template[i], atoms.get(i).getCoords());
        }

        for (i = 0; i < atoms.size(); i++) VectorMath.rotateX(atoms.get(i).getCoords(), -1 * thetaX);
        for (i = 0; i < atoms.size(); i++) VectorMath.rotateY(atoms.get(i).getCoords(), -1 * thetaY);
        for (i = 0; i < atoms.size(); i++) VectorMath.rotateZ(atoms.get(i).getCoords(), -1 * thetaZ);

        for (i = 0; i < atoms.size(); i++)
        {
            VectorMath.summation(atoms.get(i).getCoords(), transVec, atoms.get(i).getCoords());
        }
    }

/*************************************************************/
/* Sets the angles in the given segment of the peptide to    */
/* to the closest allowed corrected angles. For residues     */
/* startIndex to endIndex in the given ResData array, the    */
/* phi and psi angles are adjusted to the closest allowed    */
/* (phi, psi) pair that is in the region class specified by  */
/* rama_score_threshold:                                     */
/*                                                           */
/*   3 = angles adjusted to phi/psi core region              */
/*   2 = angles adjusted to allowed region (or core region   */
/*       if a core region is closer)                         */
/*   1 = angles adjusted to at least the minimally allowed   */
/*       region                                              */
/*                                                           */
/* In addition, omega angles are adjusted to the closest     */
/* acceptable omega angles. This function only sets the phi, */
/* psi and omega member variables of the ResData structs,    */

    /* and does not recalculate any atom coordinates.            */

    /**
     * *********************************************************
     */
    static void setCorrectedAngles(List<Group> residues, TreeMap<Integer, Angle> angles, int numRes, int startIndex, int endIndex, int rama_score_threshold)
    {

        int i;
        double phi, psi, omega;

        logger.debug("Fixing angles for segment:\n");


        for (i = startIndex; i <= endIndex; i++)
        {

            // Fix phi and psi.

            phi = findClosestAllowedPhi(residues, angles, i, numRes, rama_score_threshold);
            psi = findClosestAllowedPsi(residues, angles, i, numRes, rama_score_threshold);

            // debug
            if (DEBUG)
            {
                if (phi != angles.get(i).phi || psi != angles.get(i).psi)
                {
                    logger.debug(String.format("%s %d: fixing phi and psi: (%f, %f) --> (%f, %f)\n", residues.get(i).getPDBName(), i,
                            angles.get(i).phi, angles.get(i).psi, phi, psi));
                    logger.debug(String.format("   indices: (%d, %d) --> (%d, %d)\n", getPhiIndex(angles.get(i).phi), getPsiIndex(angles.get(i).psi),
                            getPhiIndex(phi), getPsiIndex(psi)));
                }
            }

            angles.get(i).phi = phi;
            angles.get(i).psi = psi;

            // Fix omega.

            omega = angles.get(i).omega;

            if (omega < -180.0)
            {
                omega += 360.0;
            }
            else if (omega > 180.0)
            {
                omega -= 360.0;
            }

            if (omega < 170.0 && omega >= 90.0)
            {
                omega = 170.0 + ANGLE_SNAP_MARGIN;
            }
            else if (omega > -170.0 && omega <= -90.0)
            {
                omega = -170.0 - ANGLE_SNAP_MARGIN;
            }
            else if (omega > 10.0 && omega <= 90.0)
            {
                omega = 10.0 - ANGLE_SNAP_MARGIN;
            }
            else if (omega < -10.0 && omega >= -90.0)
            {
                omega = -10.0 + ANGLE_SNAP_MARGIN;
            }

            if (DEBUG)
            {
                if (omega != angles.get(i).omega)
                {
                    logger.debug(String.format("%s %d: fixing omega: %f --> %f\n", residues.get(i).getPDBName(), i, angles.get(i).omega, omega));
                }
            }

            angles.get(i).omega = omega;

        }

    }


    /*
    Determines the closest allowed pair of phi and psi angles of the residue with index resIndex. The algorithm
    starts with the phi and psi angles of residues[resIndex], moves diagonally (or orthogonally) along the
    appropriate Ramachandran table to the closest allowed phi and psi pair, and writes those phi and psi angles
    (in degrees) to the ints pointed to by phi_ptr and psi_ptr. numRes is the number of residues in the peptide.
    */
    static double findClosestAllowedPhi(List<Group> residues, TreeMap<Integer, Angle> angles, int resIndex, int numRes, int min_accpetable_rama_score)
    {

        double original_phi, original_psi;
        double phi, psi;
        int phiindex, psiindex;
        int original_phiindex, original_psiindex;
        int rama_score;
        double dist_phi, dist_psi;
        double curr_dist;
        double best_dist = 1000000.0; // Closest diagonal (or orthogonal) distance from original phi and psi to best_phi, best_psi.
        double best_phi = 999.0;
        double best_psi = 999.0;
        // The phi and psi closest to the original phi and psi but not on the same tile.
        // Initialized to dummy values. As long as each Ramachandran table has at least one tile with a score greater
        // than or equal to min_accpetable_rama_score, these dummy values are guaranteed to be overwritten.

        original_phi = angles.get(resIndex).phi;
        original_psi = angles.get(resIndex).psi;

        original_phiindex = getPhiIndex(original_phi);
        original_psiindex = getPsiIndex(original_psi);

        if (lookupRamaScoreByIndices(residues, resIndex, numRes, original_phiindex, original_psiindex) >= min_accpetable_rama_score)
        {
            // The existing phi and psi angles are already acceptable.
            return original_phi;
        }

        // Calculate the distance between the starting phi and psi and the closest point on every single tile to find
        // the tile that is closest and has an acceptable Rama score. (Yes, this algorithm is a bit crude).
        for (psiindex = 0; psiindex < RAMA_LEN; psiindex++)
        {
            for (phiindex = 0; phiindex < RAMA_LEN; phiindex++)
            {

                if (!(phiindex == original_phiindex && psiindex == original_psiindex))
                {

                    rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, psiindex);

                    if (rama_score >= min_accpetable_rama_score)
                    {

                        // Compute the distance from the original phi and psi to the current phi and psi.

                        // First determine where on the tile we should locate the current phi and psi
                        // so that they are as close as possible to the originals.

                        // Which side of the tile is closer to phi?
                        if (phiindex == original_phiindex)
                        {
                            // Both tiles are in the same column.
                            phi = original_phi;
                        }
                        else
                        {
                            phi = (((double) (phiindex * 10)) - 180.0) + 5.0;
                            // phi is now on the left (more negative) edge of the tile.

                            if (original_phi - phi >= 0.0)
                            {
                                if (original_phi - phi < 180.0)
                                {
                                    // Right side of tile is closer.
                                    phi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Left side of tile is closer.
                                    phi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                            else
                            {
                                if (original_phi - phi > -180.0)
                                {
                                    // Left side of tile is closer.
                                    phi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Right side of tile is closer.
                                    phi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                        }

                        // Which side of the tile is closer to psi?
                        // Note that the calculation for psi differs from that of phi, because of
                        // the layout of ramaTableGeneral etc.!
                        if (psiindex == original_psiindex)
                        {
                            // Both tiles are in the same row.
                            psi = original_psi;
                        }
                        else
                        {
                            psi = (((double) (psiindex * -10)) + 180.0) - 5.0;
                            // psi is now on the upper (more positive) edge of the tile.

                            if (original_psi - psi >= 0.0)
                            {
                                if (original_psi - psi < 180.0)
                                {
                                    // Top side of tile is closer.
                                    psi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Bottom side of tile is closer.
                                    psi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                            else
                            {
                                if (original_psi - psi > -180.0)
                                {
                                    // Bottom side of tile is closer.
                                    psi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Top side of tile is closer.
                                    psi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                        }

                        // Compute the distance.

                        dist_phi = angleDist(original_phi, phi);
                        dist_psi = angleDist(original_psi, psi);

                        curr_dist = Math.sqrt(dist_phi * dist_phi + dist_psi * dist_psi);

                        // Compare it with the closest distance found so far.

                        if (curr_dist < best_dist)
                        {
                            best_dist = curr_dist;
                            best_phi = phi;
                            best_psi = psi;
                        }

                    }

                }

            }
        }

        return best_phi;

    }


    /*
    Determines the closest allowed pair of phi and psi angles of the residue with index resIndex. The algorithm
    starts with the phi and psi angles of residues[resIndex], moves diagonally (or orthogonally) along the
    appropriate Ramachandran table to the closest allowed phi and psi pair, and writes those phi and psi angles
    (in degrees) to the ints pointed to by phi_ptr and psi_ptr. numRes is the number of residues in the peptide.
    */
    static double findClosestAllowedPsi(List<Group> residues, TreeMap<Integer, Angle> angles, int resIndex, int numRes, int min_accpetable_rama_score)
    {

        double original_phi, original_psi;
        double phi, psi;
        int phiindex, psiindex;
        int original_phiindex, original_psiindex;
        int rama_score;
        double dist_phi, dist_psi;
        double curr_dist;
        double best_dist = 1000000.0; // Closest diagonal (or orthogonal) distance from original phi and psi to best_phi, best_psi.
        double best_phi = 999.0;
        double best_psi = 999.0;
        // The phi and psi closest to the original phi and psi but not on the same tile.
        // Initialized to dummy values. As long as each Ramachandran table has at least one tile with a score greater
        // than or equal to min_accpetable_rama_score, these dummy values are guaranteed to be overwritten.

        original_phi = angles.get(resIndex).phi;
        original_psi = angles.get(resIndex).psi;

        original_phiindex = getPhiIndex(original_phi);
        original_psiindex = getPsiIndex(original_psi);

        if (lookupRamaScoreByIndices(residues, resIndex, numRes, original_phiindex, original_psiindex) >= min_accpetable_rama_score)
        {
            // The existing phi and psi angles are already acceptable.
            return original_psi;
        }

        // Calculate the distance between the starting phi and psi and the closest point on every single tile to find
        // the tile that is closest and has an acceptable Rama score. (Yes, this algorithm is a bit crude).
        for (psiindex = 0; psiindex < RAMA_LEN; psiindex++)
        {
            for (phiindex = 0; phiindex < RAMA_LEN; phiindex++)
            {

                if (!(phiindex == original_phiindex && psiindex == original_psiindex))
                {

                    rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, psiindex);

                    if (rama_score >= min_accpetable_rama_score)
                    {

                        // Compute the distance from the original phi and psi to the current phi and psi.

                        // First determine where on the tile we should locate the current phi and psi
                        // so that they are as close as possible to the originals.

                        // Which side of the tile is closer to phi?
                        if (phiindex == original_phiindex)
                        {
                            // Both tiles are in the same column.
                            phi = original_phi;
                        }
                        else
                        {
                            phi = (((double) (phiindex * 10)) - 180.0) + 5.0;
                            // phi is now on the left (more negative) edge of the tile.

                            if (original_phi - phi >= 0.0)
                            {
                                if (original_phi - phi < 180.0)
                                {
                                    // Right side of tile is closer.
                                    phi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Left side of tile is closer.
                                    phi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                            else
                            {
                                if (original_phi - phi > -180.0)
                                {
                                    // Left side of tile is closer.
                                    phi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Right side of tile is closer.
                                    phi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                        }

                        // Which side of the tile is closer to psi?
                        // Note that the calculation for psi differs from that of phi, because of
                        // the layout of ramaTableGeneral etc.!
                        if (psiindex == original_psiindex)
                        {
                            // Both tiles are in the same row.
                            psi = original_psi;
                        }
                        else
                        {
                            psi = (((double) (psiindex * -10)) + 180.0) - 5.0;
                            // psi is now on the upper (more positive) edge of the tile.

                            if (original_psi - psi >= 0.0)
                            {
                                if (original_psi - psi < 180.0)
                                {
                                    // Top side of tile is closer.
                                    psi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Bottom side of tile is closer.
                                    psi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                            else
                            {
                                if (original_psi - psi > -180.0)
                                {
                                    // Bottom side of tile is closer.
                                    psi -= (5.0 - ANGLE_SNAP_MARGIN);
                                }
                                else
                                {
                                    // Top side of tile is closer.
                                    psi += (5.0 - ANGLE_SNAP_MARGIN);
                                }
                            }
                        }

                        // Compute the distance.

                        dist_phi = angleDist(original_phi, phi);
                        dist_psi = angleDist(original_psi, psi);

                        curr_dist = Math.sqrt(dist_phi * dist_phi + dist_psi * dist_psi);

                        // Compare it with the closest distance found so far.

                        if (curr_dist < best_dist)
                        {
                            best_dist = curr_dist;
                            best_phi = phi;
                            best_psi = psi;
                        }

                    }

                }

            }
        }

        return best_psi;

    }


    /*
    * Returns the index into ramaTableGeneral etc. for the given phi angle.
    * Note that getPsiIndex is different!
    */
    static int getPhiIndex(double phi)
    {
        if (phi < -180.0)
        {
            phi += 360.0;
        }
        else if (phi >= 180.0)
        {
            phi -= 360.0;
        }
        return (int) ((180.0 + phi) / 10.0);
    }

    /*
    * Returns the index into ramaTableGeneral etc. for the given psi angle.
    * Note that getPhiIndex is different!
    */
    static int getPsiIndex(double psi)
    {
        if (psi <= -180.0)
        {
            psi += 360.0;
        }
        else if (psi > 180.0)
        {
            psi -= 360.0;
        }
        return (int) ((180.0 - psi) / 10.0);
        // Note: Use -psi because the table goes from 180 to -180 as the index increases.
    }


    /*
    * Returns the Rama score based on phiindex and psiindex (indices into ramaTableGeneral etc.).
    * The other variables are provided only to indicate the type of residue.
    * 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
    */
    static int lookupRamaScoreByIndices(List<Group> residues, int resIndex, int numRes, int phiindex, int psiindex)
    {
        int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).

        //if ( (phi == 0.0) || (psi == 0.0) ) {
        
       /* TODO UNCOMMENT
        if (resIndex == 0)
        {
            // any residue which could not be assigned, like at the end of a chain
            rama_score = 3;
        }
        else if ((resIndex + 1 != numRes) && (residues.get(resIndex + 1).getPDBName().equals("PRO")))
        {
            // The next residue is PRO.
            rama_score = ramaTablePreProline[psiindex][phiindex];
        }
        else if (residues.get(resIndex).getPDBName().equals("GLY"))
        {
            // The current residue is GLY.
            rama_score = ramaTableGlycine[psiindex][phiindex];
        }
        else if (residues.get(resIndex).getPDBName().equals("PRO"))
        {
            // The current residue is PRO.
            rama_score = ramaTableProline[psiindex][phiindex];
        }
        else
        {
            // The current residue is a general residue.
            rama_score = ramaTableGeneral[psiindex][phiindex];
        }*/

        // TODO remove
        return rama_score = 0;
    }


    /*
    * Returns the angle closest to desiredAngle (in the direction of the starting angle) of type
    * angleType that is allowed based on peptide geometry (Ramaschandran plots and omega angle
    * restrictions) for the residue with index resIndex. angleType must be one of ANGLE_TYPE_PHI,
    * ANGLE_TYPE_PSI, ANGLE_TYPE_OMEGA. numRes is the number of residues in the peptide.
    * It is assumed that desiredAngle is in the range from -180.0 to 180.0.
    */
    static double getClosestAllowedAngle(List<Group> residues, TreeMap<Integer, Angle> angles, int resIndex, int numRes, double desiredAngle, int angleType)
    {

        int i;
        double phi, psi, omega;
        double finalAngle; // The final value of the angle of type angleType that we will decide to use.
        int phiindex, psiindex, desAngIndex; // Index into ramaTableGeneral for phi, psi, and desiredAngle.

        int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
        int original_rama_score; // The Rama score for the original angle of residue resIndex.
        int rama_score_from_desired_angle; // The Rama score that would result if we used desiredAngle as given.
        int target_rama_score; // The Rama score that we hope to acheive (or do better than).
        //int minimum_acceptable_rama_score;
        // The minimum Rama score that we are willing to accept if target_rama_score cannot be acheived or surpassed.
        int best_rama_score_so_far; // The Rama score for best_index_so_far.

        int index_increment; // -1 or +1, depending on which direction we want to scan.
        int best_index_so_far; // Index (phiindex or psiindex) that gives the best phi or psi angle so far.
        int elapsed_iter; // Number of scanning iterations that have elapsed.


        if (angleType == ANGLE_TYPE_OMEGA)
        {

            omega = angles.get(resIndex).omega;

            if (omega < -180.0)
            {
                omega += 360.0;
            }
            else if (omega > 180.0)
            {
                omega -= 360.0;
            }

            if (omega >= 90.0 || omega <= -90.0)
            {
                // Original omega is closer to 180 degrees, so the final omega should be
                // located there.

                if (desiredAngle < 170.0 && desiredAngle >= 0.0)
                {
                    finalAngle = 170.0 + ANGLE_SNAP_MARGIN;
                }
                else if (desiredAngle > -170.0 && desiredAngle <= -0.0)
                {
                    finalAngle = -170.0 - ANGLE_SNAP_MARGIN;
                }
                else
                {
                    finalAngle = desiredAngle;
                }
            }
            else
            { // omega < 90.0 && omega > -90.0
                // Original omega is closer to 0 degrees, so the final omega should be
                // located there.

                if (desiredAngle > 10.0)
                {
                    finalAngle = 10.0 - ANGLE_SNAP_MARGIN;
                }
                else if (desiredAngle < -10.0)
                {
                    finalAngle = -10.0 + ANGLE_SNAP_MARGIN;
                }
                else
                {
                    finalAngle = desiredAngle;
                }
            }

        }
        else
        { // angleType == ANGLE_TYPE_PHI || angleType == ANGLE_TYPE_PSI
            phi = angles.get(resIndex).phi;
            psi = angles.get(resIndex).psi;

            phiindex = getPhiIndex(phi);
            psiindex = getPsiIndex(psi);

            original_rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, psiindex);

            if (angleType == ANGLE_TYPE_PHI)
            {
                desAngIndex = getPhiIndex(desiredAngle);
                rama_score_from_desired_angle = lookupRamaScoreByIndices(residues, resIndex, numRes, desAngIndex, psiindex);
            }
            else
            { // angleType == ANGLE_TYPE_PSI
                desAngIndex = getPsiIndex(desiredAngle);
                rama_score_from_desired_angle = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, desAngIndex);
            }

            rama_score = rama_score_from_desired_angle;


            if (rama_score_from_desired_angle >= 2)
            {
                if (original_rama_score == 3)
                {
                    if (rama_score_from_desired_angle == 3)
                    {
                        return desiredAngle;
                    }
                    else
                    { // rama_score_from_desired_angle == 2
                        // move to 3
                        target_rama_score = 3;
                        //minimum_acceptable_rama_score = 3;
                    }
                }
                else
                { // original_rama_score <= 2
                    return desiredAngle;
                }
            }
            else
            {
                if (original_rama_score == 3)
                {
                    // move to 3
                    target_rama_score = 3;
                    //minimum_acceptable_rama_score = 3;
                }
                else if (original_rama_score == 2)
                {
                    // move to 2
                    target_rama_score = 2;
                    //minimum_acceptable_rama_score = 2;
                }
                else if (original_rama_score == 1)
                {
                    // try to move to 2, but settle for 1
                    target_rama_score = 2;
                    //minimum_acceptable_rama_score = 1;
                }
                else
                { // original_rama_score == 0
                    // try to move to 2, but settle for 0
                    target_rama_score = 2;
                    //minimum_acceptable_rama_score = 0;
                }
            }

            // Which direction should we scan to move from desiredAngle towards the
            // original phi or psi?
            // Note that the index_increment values for phi and psi differ!
            if (angleType == ANGLE_TYPE_PHI)
            {
                if (desiredAngle - phi >= 0.0)
                {
                    if (desiredAngle - phi < 180.0)
                    {
                        index_increment = -1;
                    }
                    else
                    {
                        index_increment = 1;
                    }
                }
                else
                {
                    if (desiredAngle - phi > -180.0)
                    {
                        index_increment = 1;
                    }
                    else
                    {
                        index_increment = -1;
                    }
                }
            }
            else
            { // angleType == ANGLE_TYPE_PSI
                if (desiredAngle - psi >= 0.0)
                {
                    if (desiredAngle - psi < 180.0)
                    {
                        index_increment = 1;
                    }
                    else
                    {
                        index_increment = -1;
                    }
                }
                else
                {
                    if (desiredAngle - psi > -180.0)
                    {
                        index_increment = -1;
                    }
                    else
                    {
                        index_increment = 1;
                    }
                }
            }


            elapsed_iter = 0;
            best_index_so_far = desAngIndex;
            best_rama_score_so_far = rama_score_from_desired_angle;
            for (i = desAngIndex; elapsed_iter < RAMA_LEN / 2;
                 i = (i + index_increment + RAMA_LEN) % RAMA_LEN)
            {
                // Scan from desAngIndex towards phiindex (in the direction)

                // Do not scan beyond the original angle.
                if ((angleType == ANGLE_TYPE_PHI && i == phiindex) || (angleType == ANGLE_TYPE_PSI && i == psiindex))
                {
                    if (original_rama_score > best_rama_score_so_far)
                    {
                        best_rama_score_so_far = original_rama_score;
                        best_index_so_far = i;
                    }
                    break;
                }

                if (angleType == ANGLE_TYPE_PHI)
                {
                    rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, i, psiindex);
                }
                else
                { // angleType == ANGLE_TYPE_PSI
                    rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, i);
                }

                if (rama_score >= target_rama_score)
                {
                    // This is the first time we have encountered a tile (sector of the Rama table)
                    // with a Rama score that meets or exceeds our target score. We will therefore
                    // place our angle in this tile.
                    best_rama_score_so_far = rama_score;
                    best_index_so_far = i;
                    break;
                }

                // At this point we have not yet found our target, but we keep track of
                // the best tile we've been in so far. Note that we keep track of only the
                // first instance of a tile with this score that we find.
                if (rama_score > best_rama_score_so_far)
                {
                    best_rama_score_so_far = rama_score;
                    best_index_so_far = i;
                }

                elapsed_iter++;
            }

            // We now know that we should place our angle in the tile with index best_index_so_far.

            if (best_index_so_far == desAngIndex)
            {
                // The proposed desiredAngle is in a good sector already, so we will use the
                // angle without modifying it.
                finalAngle = desiredAngle;
            }
            else
            {
                // We will modify the given desiredAngle by moving it into the tile with index
                // best_index_so_far. Upon which side of that tile should we place the angle?

                if (angleType == ANGLE_TYPE_PHI)
                {
                    finalAngle = (((double) (best_index_so_far * 10)) - 180.0);
                    // finalAngle is now on the left (more negative) edge of the tile.
                }
                else
                { // angleType == ANGLE_TYPE_PSI
                    finalAngle = (((double) (best_index_so_far * -10)) + 180.0);
                    // finalAngle is now on the upper (more positive) edge of the tile.
                }

                if (index_increment == 1)
                {
                    // We have been approaching from the negative direction (phi) or positive
                    // direction (psi), so finalAngle is on the correct side. We will nudge
                    // it inward slightly to prevent possible rounding error that would place
                    // it outside the tile.

                    if (angleType == ANGLE_TYPE_PHI)
                    {
                        finalAngle += ANGLE_SNAP_MARGIN;
                    }
                    else
                    { // angleType == ANGLE_TYPE_PSI
                        finalAngle -= ANGLE_SNAP_MARGIN;
                    }
                }
                else
                { // index_increment == -1
                    // We have been approaching from the positive direction (phi) or negative
                    // direction (psi), so move finalAngle to the other side of the tile. We
                    // will nudge it inward slightly because the value directly on the right
                    // (positive side) (phi) or lower side (negative side) (psi) boundary
                    // actually belongs to another tile, as well as to prevent possible
                    // rounding error that would place it outside the tile.

                    if (angleType == ANGLE_TYPE_PHI)
                    {
                        finalAngle += 10.0; // Each tile is 10.0 degrees on each side.
                        finalAngle -= ANGLE_SNAP_MARGIN;
                    }
                    else
                    { // angleType == ANGLE_TYPE_PSI
                        finalAngle -= 10.0; // Each tile is 10.0 degrees on each side.
                        finalAngle += ANGLE_SNAP_MARGIN;
                    }
                }

            }

        }

        // debug
        /*
      if (finalAngle != desiredAngle) {
          if (angleType == ANGLE_TYPE_PHI) {
              printf("Constraints changed phi: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, phi);
          }
          else if (angleType == ANGLE_TYPE_PSI) {
              printf("Constraints changed psi: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, psi);
          }
          else { // angleType == ANGLE_TYPE_OMEGA
              printf("Constraints changed omega: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, omega);
          }
      }
      */


        return finalAngle;

    } // End getClosestAllowedAngle

/*************************************************************/
/* Returns the magnitude of the difference between the two   */

    /* given angles, with angles measured in degrees.            */

    /**
     * *********************************************************
     */
    static double angleDist(double a1, double a2)
    {

        // Put each angle in the range [-180.0, 180.0).
        while (a1 < -180.0)
        {
            a1 += 360.0;
        }
        while (a1 >= 180.0)
        {
            a1 -= 360.0;
        }

        while (a2 < -180.0)
        {
            a2 += 360.0;
        }
        while (a2 >= 180.0)
        {
            a2 -= 360.0;
        }

        // Adjust angles
        if (a1 - a2 >= 0.0)
        {
            if (a1 - a2 < 180.0)
            {
                // No change required.
            }
            else
            {
                a2 += 360.0;
            }
        }
        else
        {
            if (a1 - a2 > -180.0)
            {
                // No change required.
            }
            else
            {
                a1 += 360.0;
            }
        }

        return Math.abs(a1 - a2);
    }

}
