package ca.ualberta.proteus.program.homodeller;

import ca.ualberta.proteus.util.ProteusTools;

import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.symbol.SymbolList;

/**
 * 
 * User: scott
 * Date: Sep 4, 2006
 * Time: 4:45:28 PM
 *
 */
public class Gap
{
    enum TYPE { INSERTION, DELETION }

    private TYPE type;
    private int start;
    	// The index (starting from 0) in the gapped sequence of the first '-' or '~' character of the gap.
    	// Equal to the index (in biological coordinates starting from 1) in the gapped sequence of the residue
    	// immediately preceding the first '-' or '~' character in the gap.
    private int length;
    	// The number of residue positions between the start and end positions of this gap, mapped on the gapped
    	// alignment. For simple gaps, this is equal to the number of '-' or '~' characters in the gap. For gaps
    	// resulting from the merging of two or more gaps, normal residues between gap characters also count
    	// towards the gap length.
    	//
    	// e.g. the following gap has a length of 7
    	//
		//    query: LRSVMTQGHWPR
		//    datab: LRS---QG--PR
    	//             |       |
    	//          start     end
    	//
    private boolean isMergedGap; // Whether this gap consists of two or more merged subgaps.
    private int NterminalSubgapLength;
    	// If this gap is a merged gap, this value will be the number of '-' or '~' characters in the N-terminal
    	// subgap of this gap.
    private int CterminalSubgapLength;
		// If this gap is a merged gap, this value will be the number of '-' or '~' characters in the C-terminal
		// subgap of this gap.
    
    private GappedSequence databaseSeq;
    private GappedSequence querySeq;
    
    private int preOnAlignment;
    	// The position of the Pre residue on the gapped alignment (in biological coordinates starting from 1).
    	// The Pre and Post residues are the residues with fixed positions on either side of a gap.
    private int postOnAlignment;
    	// The position of the Post residue on the gapped alignment (in biological coordinates starting from 1).
    	// The Pre and Post residues are the residues with fixed positions on either side of a gap.
    private boolean preAndPostCalculated; // Whether the positions of the Pre and Post residues have been calculated.
    
    private static final int MIN_GAP_LENGTH = 3;
    private static final int GAP_SIZE_BEFORE_INSERTION = 5;
    	// The number of residues we want between the Pre and Post residues in the template sequence before
    	// a loop is inserted. For example, if GAP_SIZE_BEFORE_INSERTION is 2:
    	//
    	//    query: LRSVMTQGH
    	//    datab: LRS---QGH
    	//            |     |
    	//           Pre   Post
    	//
    	// In the original template structure, there are only 2 residues (S and Q) between the Pre and Post
    	// residues. We hope that this allows enough space for the loop to be inserted without creating a
    	// scrunched loop.
    
    
    public Gap(GappedSequence databaseSeq, GappedSequence querySeq)
    {
        this.databaseSeq = databaseSeq;
        this.querySeq = querySeq;
        this.preAndPostCalculated = false;
        this.isMergedGap = false;
    }

    public void setStart(int start)
    {
        this.start = start;
        this.preAndPostCalculated = false;
    }

    /**
     * Returns the position of the residue immediately preceding the gap, in biological coordinates starting
     * from 1, mapped onto the gapped alignment.
     */
    public int getStart()
    {
        return start;
    }

    /**
     * Returns the position of the residue immediately following the gap, in biological coordinates starting
     * from 1, mapped onto the gapped alignment.
     */
    public int getEnd()
    {
        return start + length + 1;
    }

    /**
     * Returns the number of residue positions between the start and end positions of this gap, mapped on the
     * gapped alignment.
     */
    public int getLength()
    {
        return length;
    }

    public void setLength(int length)
    {
        assert length > 0 : "Length must be larger than 0";

        this.length = length;
        this.preAndPostCalculated = false;
    }

    public void setType(TYPE type)
    {

        this.type = type;
    }

    public TYPE getType()
    {
        return type;
    }

    /**
     * Gets the start index of the gap mapped onto the ungapped query sequence, but adjusted
     * if the gap is too small. Since the minimum loop size is 3, we'll have to turf one of
     * the residues taken from the template sequence.  This function determines which one to
     * turf.
     *
     * Note: This is very complicated and probably doesn't need to be.  See GapTest for all
     * the details of exactly what this function is trying to do.
     *
     * @return the adjusted start index of the gap, mapped onto the ungapped query sequence
     */
    /*
    public int getAdjustedStartOnQuery()
    {
        if ( isBeginning() )
        {
            return start;
        }

        if ( length >= MIN_GAP_LENGTH )
        {
            if ( ! isBeginning() && ! isEnd() )
            {
                //int rv = querySeq.viewToSource(start);
            	//
                //if ( rv + getAdjustedLength() > querySeq.getSourceSymbolList().length() )
                //{
                //    int overlap = querySeq.getSourceSymbolList().length() - rv - 1;
                //    return querySeq.viewToSource(start - overlap);
                //}
                //else
                //{
                //    return rv;
                //}

                if ( type == TYPE.INSERTION )
                {
                    return querySeq.viewToSource(start - 1);
                }
                else
                {
                    int rv = querySeq.viewToSource(start - 1);

                    int seqlen = querySeq.getSourceSymbolList().length();
                    if ( rv + getAdjustedLength() > seqlen )
                    {
                        return querySeq.viewToSource(seqlen - length); // This makes no sense -- David Arndt
                    }
                    else
                    {
                        return rv;
                    }
                }
            }
            else
            {
                // end
                return  querySeq.viewToSource(querySeq.length() - length);
            }
        }
        else if ( length == 1)
        {
            if ( isEnd() || isBeginning() || start <= 1)
            {
                // if it's at the beginning or end, don't expand the loop, we can make the loop smaller
                //return start;
                return querySeq.viewToSource(start);
            }

            // viewToSource sometimes returns a (-) when it should be (+)
            return Math.abs(querySeq.viewToSource(start - 1));
        }
        else if ( length == 2 )
        {
            // we have to choose which side we're going to turf
            // TODO use querySeq and databaseSeq to figure out which
            //   residue to turf based on similarity.  i.e. on the left is one that matches,
            //   and on the right it doesn't match, so turf the right

            // System.err.println("Gap.getAdjustedStartOnQuery() - Remember to complete this function");


            if ( isEnd() )
            {
                // if it's at the beginning or end, don't expand the loop, we can make the loop smaller
                //return start;
                return querySeq.viewToSource(start);
            }

            //return start -1;
            return querySeq.viewToSource(start -1);
        }

        assert false; // shouldn't happen

        return start;
    }
    */
    
    /**
     * Returns the adjusted start coordinate of the gap mapped onto the ungapped query sequence.
     */
    /*
    public int getAdjustedStartOnQuery()
    {
    	
    }
    */
    
    
    /**
     * Returns the position of the Pre residue of this gap, in biological coordinates starting from 1,
     * mapped onto the gapped alignment.
     */
    public int getPreOnAlignment()
    {
    	if (!this.preAndPostCalculated) {
    		this.calculatePreAndPost();
    	}
    	
    	return this.preOnAlignment;
    }
    
    
    /**
     * Returns the position of the Post residue of this gap, in biological coordinates starting from 1,
     * mapped onto the gapped alignment.
     */
    public int getPostOnAlignment()
    {
    	if (!this.preAndPostCalculated) {
    		this.calculatePreAndPost();
    	}
    	
    	return this.postOnAlignment;
    }
    
    
    /**
     * Returns the position of the Pre residue of this gap, in biological coordinates starting from 1,
     * mapped onto the ungapped query sequence.
     */
    public int getPreOnQuery()
    {
    	// We assume here that the alignment has been adjusted by LoopInserter.mergeGaps() so as to
    	// guarantee that the Pre and Post positions do not have gap characters in either the query or
    	// database sequences.
    	return this.querySeq.viewToSource(this.getPreOnAlignment());
    }
    
    
    /**
     * Returns the position of the Post residue of this gap, in biological coordinates starting from 1,
     * mapped onto the ungapped query sequence.
     */
    public int getPostOnQuery()
    {
    	// We assume here that the alignment has been adjusted by LoopInserter.mergeGaps() so as to
    	// guarantee that the Pre and Post positions do not have gap characters in either the query or
    	// database sequences.
    	return this.querySeq.viewToSource(this.getPostOnAlignment());
    }
    
    
    /**
     * Returns the position of the Pre residue of this gap, in biological coordinates starting from 1,
     * mapped onto the ungapped database sequence.
     */
    public int getPreOnDatabase()
    {
    	// We assume here that the alignment has been adjusted by LoopInserter.mergeGaps() so as to
    	// guarantee that the Pre and Post positions do not have gap characters in either the query or
    	// database sequences.
    	return this.databaseSeq.viewToSource(this.getPreOnAlignment());
    }
    
    
    /**
     * Returns the position of the Post residue of this gap, in biological coordinates starting from 1,
     * mapped onto the ungapped database sequence.
     */
    public int getPostOnDatabase()
    {
    	// We assume here that the alignment has been adjusted by LoopInserter.mergeGaps() so as to
    	// guarantee that the Pre and Post positions do not have gap characters in either the query or
    	// database sequences.
    	return this.databaseSeq.viewToSource(this.getPostOnAlignment());
    }
    
    
    /**
     * Calculates the positions of the Pre and Post residues of this gap.
     */
    private void calculatePreAndPost()
    {
    	//TODO make sure there are enough checks
    	
    	int preIdeal = start - (int)Math.floor(GAP_SIZE_BEFORE_INSERTION/2.0);
    	int postIdeal = this.getEnd() + (int)Math.ceil(GAP_SIZE_BEFORE_INSERTION/2.0);
    		// The term in the second pair of brackets is set to deal with rounding error in case
    		// GAP_SIZE_BEFORE_INSERTION is odd. e.g. if GAP_SIZE_BEFORE_INSERTION is 3, then
    		// (GAP_SIZE_BEFORE_INSERTION - GAP_SIZE_BEFORE_INSERTION/2) is 2.
    	int gappedSeqLength = querySeq.length();
    	System.out.println(preIdeal + ", " + postIdeal);
    	// We're working in biological coordinates starting from 1.
    	
    	if (preIdeal < 1 && postIdeal > gappedSeqLength) {
    		throw new IllegalStateException("Gap is too close to the beginning and to the end of the sequence simultaneously.");
    	}
    	else if (preIdeal < 1) {
    		// The ideal Pre residue is too far left. Shift the Pre and Post residues to the right.
    		
    		int pre = 1;
    		int post = postIdeal + (pre - preIdeal);
    		
    		if (postIdeal > gappedSeqLength) {
        		throw new IllegalStateException("Gap is too close to the beginning and to the end of the sequence simultaneously.");
    		}
    		
    		this.preOnAlignment = pre;
    		this.postOnAlignment = post;
    	}
    	else if (postIdeal > gappedSeqLength) {
    		
    		int post = gappedSeqLength;
    		int pre = preIdeal + (post - postIdeal);
    		
    		if (pre < 1) {
        		throw new IllegalStateException("Gap is too close to the beginning and to the end of the sequence simultaneously.");
    		}
    		
    		this.preOnAlignment = pre;
    		this.postOnAlignment = post;
    	}
    	else {
    		this.preOnAlignment = preIdeal;
    		this.postOnAlignment = postIdeal;
    	}
    	
    	this.preAndPostCalculated = true;
    }
    
    
    /**
     * Returns getAdjustedStartOnQuery + getAdjustedLength
     * 
     * The maximum possible value returned is the length of the query sequence, i.e. in
     * biological coordinates starting from 1, the maximum possible value returned is the
     * coordinate of the final residue in the ungapped query sequence.
     * 
     * @return getAdjustedStartOnQuery + getAdjustedLength
     */
    /*
    public int getAdjustedEndOnQuery()
    {
        return getAdjustedStartOnQuery() + getAdjustedLength();
    }
    */
    
    /**
     * Gets the adjusted start position of the gap, as returned by getAdjustedStartOnQuery(),
     * except that the index is mapped onto the ungapped database sequence rather than the
     * ungapped query sequence.
     * 
     * Note: This method is written under the assumption that the alignment will always be
     * such that the adjusted start index on the query is not aligned with a gap on the
     * database sequence. This requirement may be violated with an alignment such as:
     * 
     *                adjustedStartOnQuery
     *                 |
     *    query: MVQ---VMTRGHLQ
     *    datab: MVQLRS---RGHLQ
     * 
     * TODO: Handle this special case, perhaps by changing the alignment to e.g.
     * 
     *    query: MVQ--VMTRGHLQ
     *    datab: MVQLRS--RGHLQ
     * 
     * Added by David Arndt, Feb 13, 2008.
     * 
     * @return the adjusted start position of the gap mapped onto the ungapped database sequence
     */
    /*
    public int getAdjustedStartOnDatabase()
    {
    	int adjustedStartOnView = querySeq.sourceToView(getAdjustedStartOnQuery());
    	int adjustedStartOnDatabase = Math.abs(databaseSeq.viewToSource(adjustedStartOnView));
    	assert adjustedStartOnView == databaseSeq.sourceToView(adjustedStartOnDatabase);
    	return adjustedStartOnDatabase;
    }
    */
    
    /**
     * Gets the adjusted end position of the gap, as returned by getAdjustedEndOnQuery(),
     * except that the index is mapped onto the ungapped database sequence rather than the
     * ungapped query sequence.
     * 
     * See getAdjustedStartOnDatabase() for an important note.
     * 
     * @return the adjusted end position of the gap mapped onto the ungapped database sequence
     */
    /*
    public int getAdjustedEndOnDatabase()
    {
    	int adjustedEndOnView = querySeq.sourceToView(getAdjustedEndOnQuery());
    	int adjustedEndOnDatabase = Math.abs(databaseSeq.viewToSource(adjustedEndOnView));
    	assert adjustedEndOnView == databaseSeq.sourceToView(adjustedEndOnDatabase);
    	return adjustedEndOnDatabase;
    }
    */
    
    /**
     * Gets the length of the gap, but adjusted if the gap is too small.
     * Since the minimum loop size is 3, we'll have to turf one of the residues
     * taken from the template sequence.  This function determines which one to turf.
     *
     * Note: This is very complicated and probably doesn't need to be.  See GapTest for all the
     * details of exactly what this function is trying to do.
     *
     * @return the adjusted length of the gap
     */
    /*
    public int getAdjustedLength()
    {
        if (type == TYPE.DELETION && (isEnd() || isBeginning()))
        {
            // we don't care about it
            return 0;
        }

        if ( isEnd() || isBeginning() )
        {
            // if it's at the beginning or end, don't expand the loop, we can make the loop smaller
            return Math.min(length, querySeq.getSourceSymbolList().length() - getAdjustedStartOnQuery());
        }

        if ( length >= MIN_GAP_LENGTH )
        {
            if ( type == TYPE.INSERTION )
            {
                // want to replace the gap, plus some surrounding residues (3)
                return Math.min(length + MIN_GAP_LENGTH, querySeq.getSourceSymbolList().length() - getAdjustedStartOnQuery());
            }
        }
        
        return MIN_GAP_LENGTH;
    }
    */
    
    /**
     * Returns the intended number of residues between the Pre and Post residues in the final
     * structure. This is equal to the number of coil residues in the loop that will need to
     * be inserted.
     * 
     * Example of insertion:
     * 
     *    loop:  BBBCCCCCBBB
     *    query: ISGCWVTPYFK
     *    datab: LDSC---KVLN
     *             |     |
     *            Pre   Post
     * 
     * Here the 2 residues C and K between the Pre and Post residues will be replaced by the
     * 5 residues CWVTP. getLoopCoilLength() will return 5. 
     * 
     * 
     * Example of deletion:
     * 
     *            Pre   Post
     *             |     |
     *    query: LDSC---KVLN
     *    datab: ISGCWVTPYFK
     *    loop:  BBBC   CBBB = BBBCCBBB
     *    
     * Here the 5 residues CWVTP will be replaced by the 2 residues C and K. getLoopCoilLength()
     * will return 2.
     * 
     * 
     * Example of deletion involving a merged gap:
     * 
     *            Pre   Post
     *             |     |
     *    query: LDSC-Y-KVLN
     *    datab: ISGCWVTPYFK
     *    loop:  BBBC C CBBB = BBBCCCBBB
     *    
     * Here the 5 residues CWVTP will be replaced by the 3 residues CYK. getLoopCoilLength()
     * will return 3.
     */
    public int getLoopCoilLength() {
    	
    	int loopCoilLength = 0;
    	
    	int pre = this.getPreOnAlignment();
    	int post = this.getPostOnAlignment();
    	
    	// Count the number of residues (excluding gap characters) between the Pre and Post residues
    	// on the query sequence.
    	for (int i = pre + 1; i < post; i++) {
    		
    		// pre, post and i are in biological coordinates starting from 1. For String.charAt(), we
    		// subtract 1 in order to use coordinates starting from 0.
    		if ( !ProteusTools.isGap(this.querySeq.seqString().charAt(i-1)) ) {
    			loopCoilLength++;
    		}
    		
    	}
        
    	return loopCoilLength;
    }
    
    
    public boolean isEnd()
    {
        return start + length >= databaseSeq.length();
    }

    public boolean isBeginning()
    {
        return start == 0;
    }
    
    
    /**
     * If this gap is a simple gap containing of only one segment of consecutive '-' or '~' characters,
     * then the length of this gap is returned. If this gap consists of a merging of two or more subgaps,
     * the number of '-' or '~' characters in the N-terminal subgap of this gap is returned.
     * 
	 * e.g. for the following gap, 3 would be returned
	 *
	 *    query: LRSVMTQGHWPR
	 *    datab: LRS---QG--PR
	 *             |       |
	 *          start     end
     * 
     * @return The length of the N-terminal subgap of this gap.
     */
    public int getNterminalSubgapLength() {
    	if (this.isMergedGap) {
    		return this.NterminalSubgapLength;
    	}
    	else {
    		return this.getLength();
    	}
    }
    
    
    /**
     * If this gap is a simple gap containing of only one segment of consecutive '-' or '~' characters,
     * then the length of this gap is returned. If this gap consists of a merging of two or more subgaps,
     * the number of '-' or '~' characters in the C-terminal subgap of this gap is returned.
     * 
	 * e.g. for the following gap, 2 would be returned
	 *
	 *    query: LRSVMTQGHWPR
	 *    datab: LRS---QG--PR
	 *             |       |
	 *          start     end
     * 
     * @return The length of the C-terminal subgap of this gap.
     */
    public int getCterminalSubgapLength() {
    	if (this.isMergedGap) {
    		return this.CterminalSubgapLength;
    	}
    	else {
    		return this.getLength();
    	}
    }
    
    
    /**
     * Sets the length of the N-terminal subgap of a merged gap, and automatically adjusts the total length of
     * this gap. This function should only be called on a gap that is set at a merged gap; if this gap is not
     * yet set as a merged gap, call setAsMergedGap() instead.
     * 
     * @param NterminalSubgapLength The number of '-' or '~' characters in the N-terminal subgap.
     */
    public void setNterminalSubgapLength(int NterminalSubgapLength) {
    	if (this.isMergedGap) {
    		assert NterminalSubgapLength > 0;
    		this.length += NterminalSubgapLength - this.NterminalSubgapLength;
    		assert this.length > 0;
    		
    		this.NterminalSubgapLength = NterminalSubgapLength;
    	}
    	else {
    		throw new IllegalStateException("Call setAsMergedGap() instead.");
    	}
    }
    
    
    /**
     * Sets the length of the C-terminal subgap of a merged gap, and automatically adjusts the total length of
     * this gap. This function should only be called on a gap that is set at a merged gap; if this gap is not
     * yet set as a merged gap, call setAsMergedGap() instead.
     * 
     * @param CterminalSubgapLength The number of '-' or '~' characters in the C-terminal subgap.
     */
    public void setCterminalSubgapLength(int CterminalSubgapLength) {
    	if (this.isMergedGap) {
    		assert CterminalSubgapLength > 0;
    		this.length += CterminalSubgapLength - this.CterminalSubgapLength;
    		assert this.length > 0;
    		
    		this.CterminalSubgapLength = CterminalSubgapLength;
    	}
    	else {
    		throw new IllegalStateException("Call setAsMergedGap() instead.");
    	}
    }
    
    
    /**
     * Returns true if this is a merged gap consisting of two or more subgaps, false otherwise.
     * @return true if this is a merged gap consisting of two or more subgaps, false otherwise.
     */
    public boolean isMergedGap() {
    	return this.isMergedGap;
    }
    
    
    /**
     * Set this gap to be a merged gap consisting of two or more subgaps, all of the same type (i.e.
     * all deletions or all insertions).
     * 
     * e.g. for the following gap, NterminalSubgapLength should be 3 and CterminalSubgapLength
     * should be 2.
	 *
	 *    query: LRSVMTQGHWPR
	 *    datab: LRS---QG--PR
	 *             |       |
	 *          start     end
     * 
     * @param NterminalSubgapLength The number of '-' or '~' characters in the N-terminal subgap.
     * @param CterminalSubgapLength The number of '-' or '~' characters in the C-terminal subgap.
     */
    public void setAsMergedGap(int NterminalSubgapLength, int CterminalSubgapLength) {
    	this.isMergedGap = true;
    	this.NterminalSubgapLength = NterminalSubgapLength;
    	this.CterminalSubgapLength = CterminalSubgapLength;
    }
    
    
    /**
     * Set this gap to be a simple gap consisting of only one segment of consecutive '-' or '~'
     * characters.
     */
    public void setAsSimpleGap() {
    	this.isMergedGap = false;
    }
    
    
    public String toString()
    {
        return "type = " + type
                + ", start = " + start
                + ", end = " + this.getEnd()
                + ", length = " + length;
    }
    
    
    /**
     * Returns the symbol list representing the sequence where the gap will be replaced in the query,
     * which is the same as the number of residues between the Pre and Post residues (exclusive) in
     * the query sequence.
     *
     * e.g.
     * <pre>
     * DELETION:
     * Query: NAA--AKSE
     * Datab: NAAPKAKSE
     *         |    |
     *        Pre  Post
     *
     * AA in the query would be returned.
     *
     * INSERTION:
     * Query: NAAPKAKSE
     * Datab: NAA--AKSE
     *         |    |
     *        Pre  Post
     *
     * It would return APKA
     * </pre>
     * @return the symbol list representing the sequence where the gap will be replaced in the query
     */
    public SymbolList getQueryGapSequence()
    {
        if ( getLoopCoilLength() == 0 )
        {
            return SymbolList.EMPTY_LIST;
        }
        return querySeq.getSourceSymbolList().subList(getPreOnQuery() + 1, getPostOnQuery() - 1);
    }


}
