package ca.ualberta.proteus.program.homodeller;

import com.martiansoftware.jsap.*;
import ca.ualberta.proteus.biojava.PDBFileParser;
import ca.ualberta.proteus.program.minimization.PepmakeRunner;
import ca.ualberta.proteus.program.minimization.EnergyMinimizer;
import ca.ualberta.proteus.program.minimization.GAFolderRunner;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.util.*;
import org.apache.log4j.Logger;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.structure.AminoAcid;
import org.biojava.bio.structure.Chain;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.Structure;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.xml.sax.SAXException;

import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Mar 1, 2005
 * Time: 10:47:44 PM
 * To change this template use File | Settings | File Templates.
 */
public class HomodellerImpl
{
    private static Logger logger = Logger.getLogger(HomodellerImpl.class);
    private static final String VERSION = "0.7";
    private static final String OUTPUT_FILE_FLAG = "OUTPUT";
    private static final String FILE_FLAG = "FILE";
    private static final String QUERY_ID = "QUERY_ID";
    private static final String TEMPLATE_ID = "TEMPLATE_ID";
    private static final String ENERGY_MINIMIZATION_FLAG = "ENERGY_MINIMIZATION_FLAG";
    private static final String ALIGNMENT_FLAG = "ALIGNMENT";
    private static final String ALIGNMENT_ID_FLAG = "ALIGNMENT_ID";
    // private HomodellerSettings settings;

    public HomodellerImpl()
    {
    }

    /**
     * Gets the PDB structure for the passed template ID
     * @param templateID - PDB ID of the requested structure
     * @return the PDB structure for the passed template ID
     */
    public static Structure readTemplateStructure(String templateID)
    {
        return readTemplateStructure(templateID, true);
    }

    /**
     * Gets the PDB structure for the passed template ID
     * @param templateID - PDB ID of the requested structure
     * @param stripHetAtoms - True to strip any excess atoms
     * @return the PDB structure for the passed template ID
     */
    public static Structure readTemplateStructure(String templateID, boolean stripHetAtoms)
    {
        Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");

        String pathname = config.getStringValue("pdb_dir");

        InputStream in = PDBFileLoader.getPDBFile(templateID, new File[]{new File(pathname)});

        try
        {
            if ( in != null )
            {
                return readTemplateStructure(in, templateID, stripHetAtoms);
            }
        }
        finally
        {
            try
            {
                if (in != null)
                {
                    in.close();
                }
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }

        return null;
    }

    /**
     * Gets the PDB structure for the passed file
     * @param templateID - PDB ID of the requested structure
     * @param in - Stream to read from
     * @return the PDB structure for the passed template ID
     */
    public static Structure readTemplateStructure(InputStream in, String templateID)
    {
        return readTemplateStructure(in, templateID, true);
    }

    /**
     * Gets the PDB structure for the passed file
     * @param templateID - PDB ID of the requested structure
     * @param in - Stream to read from
     * @return the PDB structure for the passed template ID
     * @param stripHetAtoms True to strip any excess atoms
     */
    public static Structure readTemplateStructure(InputStream in, String templateID, boolean stripHetAtoms)
    {
        PDBFileParser pdbReader = new PDBFileParser();

        Structure template;
        try
        {
            template = pdbReader.parsePDBFile(in);
        }
        catch (IOException e)
        {
            logger.error("The PDB file could not be read");
            e.printStackTrace();
            return null;
        }

        if ( template.getPDBCode() == null )
        {
            template.setPDBCode(templateID);
        }

        logger.debug("Found template: " + templateID);

        template = removeExtraChainsFromTemplate(template, templateID);

        if ( stripHetAtoms )
        {
            HomodellerUtils.removeHetatoms(template.getChain(0));
        }

        return template;
    }

    /**
     * Run homodeller first to find a database match, then do the loop
     * insertions
     * @param sequence The sequence to run
     * @return result containing the finished model, or null if an error occured
     */
    public Result run( Sequence sequence )
    {
        return run(sequence.seqString());
    }

    /**
     * Run homodeller first to find a database match, then do the loop
     * insertions
     * @param sequence The sequence to run
     * @return a result containing the finished model, or null if an error occured
     */
    public Result run( String sequence )
    {
        File outputFile = runHomodellerProgram(sequence);

        if ( outputFile == null )
        {
            logger.error("Homodeller output file was null");
            return null;
        }

        List<HomodellerOutputResult> resultList;
        try
        {
            // read homodeller's output to get the input sequence and alignment
            resultList = readInfoFile(outputFile);
        }
        catch (IOException e)
        {
            logger.error("Can't read homodeller's output file");
            return null;
        }


        HomodellerOutputResult outputResult = getBestTemplate(resultList);
        Structure template = readTemplateStructure(outputResult.templatePDBID);

        if ( template == null )
        {
            logger.error("Template structure is null");
            return null;
        }


        if ( ! template.getChain(0).getSequence().equals(outputResult.databaseSeq.seqString()) )
        {
            // for some reason, the database and template sequences don't match up.
            // Recover by re-downloading and getting the alignment ourselves

            logger.info("Template sequence does not match aligned sequence, doing own alignment...");
            return run(sequence, outputResult.templatePDBID);
        }

        Result result = run(outputResult.templatePDBID, template, outputResult.querySeq, outputResult.databaseSeq);

        assert outputResult.querySeq.getSourceSymbolList().seqString().equalsIgnoreCase(sequence) : "Homodeller's query sequence doesn't match the input sequence";

        return result;
    }

    /**
     * Goes through all the results and chooses the best one by looking at sequence
     * similarity and NMR/X-RAY structures, and if ligands are present
     * @param resultList list of results from C-program homodeller's results
     * @return the best result
     */
    private HomodellerOutputResult getBestTemplate(List<HomodellerOutputResult> resultList)
    {
        // choose the best alignment result
        for ( HomodellerOutputResult r : resultList )
        {
            Structure template = readTemplateStructure(r.templatePDBID, false);

            if ( template == null )
            {
                logger.debug("Template for " + r.templatePDBID + " could not be found... skipping");
                continue;
            }

            Chain chain = template.getChain(0);

            String technique = (String) template.getHeader().get("technique");
            if ( technique.contains("NMR") )
            {
                /**
                 * allowed techniques are:
                     ELECTRON DIFFRACTION
                     FIBER DIFFRACTION
                     FLUORESCENCE TRANSFER
                     NEUTRON DIFFRACTION
                     NMR
                     THEORETICAL MODEL
                     X-RAY DIFFRACTION
                 */
                r.technique = HomodellerOutputResult.TECHNIQUE.NMR;
            }
            else if ( technique.contains("X-RAY"))
            {
                r.technique = HomodellerOutputResult.TECHNIQUE.XRAY;
            }
            else
            {
                r.technique = HomodellerOutputResult.TECHNIQUE.NONE;
            }

            List<Group> groups = (List<Group>) chain.getGroups();
            for ( Group g: groups)
            {
                // find the hetatoms
                if ( ! ( g instanceof AminoAcid ) )
                {
                    String pdbCode = g.getPDBCode();
                    if ( pdbCode.equals("CA"))
                    {
                        r.hasLigand = true;
                        break;
                    }
                }
            }

            r.sequenceSimilarity = HomodellerUtils.getSequenceSimilarity(r.databaseSeq.seqString(), r.querySeq.seqString());
        }

        // sort the list, uses the compareTo method in HomodellerOutputResult
        Collections.sort(resultList);

        return resultList.get(0);
    }

    /**
     * Does all the tasks of creating a homology model
     *
     * @param templateID Template ID on which to base the model
     * @param template Template on which to base the model
     * @param querySeq The aligned query sequence
     * @param databaseSeq the aligned database sequence
     * @return a result containing the finished model, or null if an error occured
     */
    private Result run(String templateID, Structure template, GappedSequence querySeq, GappedSequence databaseSeq)
    {
        // fix the file and insert loops
    	//LoopInserter loopInserter = new LoopInserter(template, querySeq, databaseSeq);
    	ca.ualberta.proteus.program.homodeller.loop.LoopInserter loopInserter = new ca.ualberta.proteus.program.homodeller.loop.LoopInserter(template, querySeq, databaseSeq);

        logger.debug("QUERY: " + querySeq.seqString());
        logger.debug("DATAB: " + databaseSeq.seqString());

        Structure completeStructure = loopInserter.run();

        // Fix Carbonyl O atoms using pepmake
        // Use empty options to run plain pepmake with no torsion fixing or 
        PepmakeRunner.Options options = new PepmakeRunner.Options();
        PepmakeRunner pepmake = new PepmakeRunner(completeStructure, options);
        completeStructure = pepmake.runPepmake();

        return new Result(completeStructure, template, querySeq, databaseSeq, templateID);
    }

    /**
     * Run this when you already have a template ID
     * @param sequence The sequence to run
     * @param templateID the template ID on which to base the model
     * @return a result containing the finished model, or null if an error occured
     */
    public Result run( String sequence, String templateID )
    {
        Sequence querySeq;
        try
        {
            querySeq = ProteinTools.createProteinSequence(sequence, "QUERY");
        }
        catch (IllegalSymbolException e)
        {
            logger.error("Incorrect input query sequence ", e);
            return null;
        }

        assert querySeq.seqString().equalsIgnoreCase(sequence) : "Homodeller's query sequence doesn't match the input sequence";

        Structure template = readTemplateStructure(templateID);

        return run(querySeq, template, templateID);
    }

    public Result run( String sequence, String templateID, Alignment alignment )
    {
        Sequence querySeq;
        try
        {
            querySeq = ProteinTools.createProteinSequence(sequence, "QUERY");
        }
        catch (IllegalSymbolException e)
        {
            logger.error("Incorrect input query sequence ", e);
            return null;
        }

        assert querySeq.seqString().equalsIgnoreCase(sequence) : "Homodeller's query sequence doesn't match the input sequence";

        Structure template = readTemplateStructure(templateID);

        return run(querySeq, template, templateID, alignment);
    }

    public Result run( Sequence sequence, Structure template )
    {
        return run(sequence, template, template.getName());
    }

    /**
     * Run homodeller based on a query, a template, and an alignment between the two.
     * The alignment MUST ONLY contain the aligned versions of the query and template sequences
     *
     * @param query     The sequence to run
     * @param template  The template sequence
     * @param alignment  Alignment between the two
     * @return
     */
    public Result run( Sequence query, Structure template , Alignment alignment)
    {
        // TODO verify that sequence and template are the only sequences in alignment

        return run(query, template, template.getName(), alignment);
    }

    private Result run(Sequence querySeq, Structure template, String templateID)
    {
        String templateSeqStr = template.getChain(0).getSequence();
        Sequence templateSequence;
        try
        {
            templateSequence = ProteinTools.createProteinSequence(templateSeqStr, template.getPDBCode());
        }
        catch (IllegalSymbolException e)
        {
            logger.error("Incorrect template sequence", e);
            return null;
        }

        Alignment alignment = getAlignment(querySeq, templateSequence);

        return run(querySeq, template, templateID, alignment);
    }

    /**
     * Run homodeller based on a queryID and an alignment which contains that ID.
     *
     * @param queryID
     * @param alignment
     * @return
     */
    private Result run(String queryID, Alignment alignment)
    {
        if ( alignment.getLabels().size() != 2 )
        {
            throw new IllegalArgumentException("The alignment must contain exactly 2 sequences");
        }

        GappedSequence gappedQuery = null;
        try
        {
            gappedQuery = (GappedSequence) alignment.symbolListForLabel(queryID);
        }
        catch (NoSuchElementException e)
        {
            System.err.println(
                    "The 'id' parameter or the aligment file was not correct.  \n" +
                            "The alignment file must be composed of " +
                            "exactly 2 sequences; one sequence must correspond " +
                    "to the 'id' parameter, and the other a valid PDB ID. \n" +
                            "For example, if the following file was used, the 'id' parameter would be 'QUERY': \n\n" +

                            "CLUSTAL W (1.83) multiple sequence alignment\n\n" +
                            "" +
                    "QUERY           MKEVAKPKKLQERFFIPHEIKVMGRWSTEDVEVKDPSLKPYINLEPRLLPHTHGRHAKKH\n" +
                    "1IQVA           -------------------IKVMGRWSTEDVEVKDPSLKPYINLEPR-------------\n" +
                    "                                   ****************************\n");

            return null;
        }

        assert gappedQuery != null : "Query sequence not found in alignment";

        List list = alignment.getLabels();
        int index = Math.abs(list.indexOf(gappedQuery.getName()) - 1 );  // If QUERY is in 1, index = 0. If 0, index = 1;
        String templateID = (String) list.get(index);

        GappedSequence gappedTemplate = (GappedSequence) alignment.symbolListForLabel(templateID);

        Structure template = readTemplateStructure(templateID);

        return run(templateID, template, gappedQuery, gappedTemplate );
    }

    private Result run(Sequence querySeq, Structure template, String templateID, Alignment alignment)
    {
        GappedSequence gappedQuery;
        if ( alignment.getLabels().contains("QUERY"))
        {
            gappedQuery = (GappedSequence) alignment.symbolListForLabel("QUERY");
        }
        else
        {
            gappedQuery =  (GappedSequence) alignment.symbolListForLabel(querySeq.getName());
        }
        GappedSequence gappedTemplate = (GappedSequence) alignment.symbolListForLabel(template.getPDBCode());

        return run(templateID, template, gappedQuery, gappedTemplate );
    }

    /**
     * Gets an alignment between two sequences.
     *
     * @param querySeq First sequence to align
     * @param templateSequence Second sequence to align
     * @return alignment
     */
    private Alignment getAlignment(Sequence querySeq, Sequence templateSequence)
    {
        List<Sequence> sequences = new ArrayList<Sequence>(2);
        sequences.add(0, querySeq);
        sequences.add(1, templateSequence);

        AlignmentProgramRunner runner;
        Alignment alignment;

        try
        {
            runner = new ClustalWRunner(querySeq, sequences);
            runner.run();
            alignment = runner.getAlignment();
        }
        catch(Exception e)
        {
            runner = new XAlignRunner(querySeq, sequences);
            runner.run();
            alignment = runner.getAlignment();
        }

        assert alignment != null;

        return alignment;
    }

    /**
     * Runs the C-based "homodeller" program
     * @param sequence The sequence to run
     * @return the output file from homodeller, detailing the alignments it found
     */
    private File runHomodellerProgram(String sequence)
    {
        Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");
        String homodellerPath = config.getStringValue("homodeller");
        File homodeller = new File(homodellerPath);
        File dir = homodeller.getParentFile().getAbsoluteFile();
        File tempFile;

        try
        {
            tempFile = File.createTempFile("homodeller_input", ".in");
            tempFile.deleteOnExit();

            //homodeller has trouble with long sequences
            PrintWriter w = new PrintWriter(tempFile);
            w.println(sequence.replaceAll("(\\w{50})", "$1\n"));
            w.close();
        }
        catch (IOException e)
        {
            logger.error("Cannot write temp file", e);
            return null;
        }

        try
        {

            String command = homodellerPath + " " + tempFile.getPath();
            Process proc = Runtime.getRuntime().exec(command, null, dir);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(proc.getErrorStream(), "ERROR", System.err);

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(proc.getInputStream(), "OUTPUT", System.out);

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = proc.waitFor();

            if (exitVal != 0)
            {
                System.err.println("Homodeller failed ");
                return null;
            }

            /** Wait for the output reader to completely do its thing **/
            while (! outputReader.isFinished())
            {
                Thread.sleep(10);
            }

            return new File( tempFile.getAbsolutePath() + ".out" );
        }
        catch (InterruptedException e)
        {
            logger.error("Homodeller Interrupted", e);
        }
        catch (IOException e)
        {
            logger.error("Cannot run homodeller command", e);
        }

        return null;
    }

    /**
     * Read's C-program homodeller's output file to find the alignment and the optimal sequence
     * for using as a template
     * @param infoFile File to read
     * @throws IOException
     * @return a list of alignment results
     */
    public List<HomodellerOutputResult> readInfoFile(File infoFile) throws IOException
    {
        BufferedReader reader = new BufferedReader(new FileReader(infoFile));

        List<HomodellerOutputResult> rv = new ArrayList<HomodellerOutputResult>();
        String line = "";

        while ( line != null )
        {
            HomodellerOutputResult result = new HomodellerOutputResult();

            // skip everthing up to alignment output
            while( (line = reader.readLine()) != null)
            {
                if ( line.contains("*("))
                {
                    break;
                }
            }

            // get the alignment
            StringBuilder query = new StringBuilder();
            StringBuilder database = new StringBuilder();
            while( (line = reader.readLine()) != null)
            {
                if ( line.contains("Title"))
                {
                    String[] strings = line.split(":");
                    result.templatePDBID = strings[1].trim();
                }
                else if ( line.contains("Query Seq:"))
                {
                    query.append(line.substring(11, 61));
                }
                else if ( line.contains("Database.:"))
                {
                    database.append(line.substring(11, 61));
                }
                else if ( line.contains("*("))
                {
                    // reset to the mark
                    reader.reset();

                    // end of the result
                    break;
                }

                reader.mark(2000);
            }

            String queryStr = query.toString().replace(" ", "-");
            String databaseStr = database.toString().replace(" ", "-");

            assert queryStr.length() == databaseStr.length() : "Query and database sequences are not the same length";


            // search backwards to find the end of the alignment
            char[] queryChars = queryStr.toCharArray();
            char[] dbChars = databaseStr.toCharArray();
            int length = queryChars.length;
            for ( int i=queryChars.length -1; i >= 0 ; i-- )
            {
                char queryChar = queryChars[i];
                char dbChar = dbChars[i];

                if ( queryChar != '-' || dbChar != '-')
                {
                    length = i;
                    break;
                }
            }

            queryStr = queryStr.substring(0, length + 1);
            databaseStr = databaseStr.substring(0, length + 1);

            assert queryStr.length() == databaseStr.length() : "Query and database sequences are not the same length";

            // build the sequences and the alignment object
            try
            {
                result.querySeq = ProteinTools.createGappedProteinSequence(queryStr, "QUERY");
                result.databaseSeq = ProteinTools.createGappedProteinSequence(databaseStr, result.templatePDBID);
            }
            catch (IllegalSymbolException e)
            {
                e.printStackTrace();
            }

            rv.add(result);
        }

        return rv;
    }

    private static Structure removeExtraChainsFromTemplate(Structure structure, String templateID)
    {
        //Used for parsing specific chains
        List<Chain> chains = structure.getChains(0);

        if (chains.size() > 1 && templateID.length() == 5)
        {
            String chainID = String.valueOf(templateID.charAt(4));
            List<Chain> chainsToRemove = new ArrayList<Chain>();

            for (Chain chain : chains)
            {
                String name = chain.getName();
                if (name.equals(chainID))
                {
                    logger.debug("Found chain " + templateID.charAt(4));
                }
                else
                {
                    chainsToRemove.add(chain);
                }
            }

            // delete the chains from the model
            for ( Chain c: chainsToRemove)
            {
                chains.remove(c);
            }

        }
        else if ( chains.size() > 1 )
        {
            // we only specified at 4-letter ID, get rid of all but the first chain
            for ( int i=chains.size() - 1; i >= 1; i--)
            {
                chains.remove(i);
            }

            Chain chain = chains.get(0);
            logger.debug("Found chain " + chain.getName());
        }

        assert chains.size() == 1 : "Chains has more than 1 chain";

        if ( structure.nrModels() > 1 )
        {
            structure = new ProteusStructureImpl(structure, true);
        }

        return structure;
    }

    public Result run(Sequence sequence, String templateID)
    {
        return run(sequence.seqString(), templateID);
    }

    public Result run(Sequence sequence, File structure)
    {
        try
        {
            return run(sequence, readTemplateStructure(new FileInputStream(structure), "WEB_SEQUENCE"));
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }

        return null;
    }

    /**
     * A class containing all information from an alignment in the homodeller C-program output file
     */
    public static class HomodellerOutputResult implements Comparable<HomodellerOutputResult>
    {
        public enum TECHNIQUE { NONE, NMR, XRAY }

        public String templatePDBID;
        public GappedSequence querySeq;
        public GappedSequence databaseSeq;
        public Structure template;
        public TECHNIQUE technique = TECHNIQUE.NONE;
        public boolean hasLigand;
        public double sequenceSimilarity;

        public int compareTo(HomodellerOutputResult other)
        {
            final double LIGAND_THRESHOLD = 0.5;
            final double XRAY_THRESHOLD = 0.4;

            // techniques are the same, sort based solely on sequence similarity
            double rv = other.sequenceSimilarity - sequenceSimilarity;

            if ( hasLigand != other.hasLigand )
            {
                if ( hasLigand)
                {
                    // other doesn't have a ligand, and we want a sequence with 50% seq ID and no ligand
                    // to have an advantage over a sequence with 90% seq ID but has a ligand
                    // i.e. give the 50% seq % boost to the one without ligand
                    rv = ( other.sequenceSimilarity + LIGAND_THRESHOLD) - sequenceSimilarity;
                }
                else if ( other.hasLigand )
                {
                    // opposite of above
                    rv = other.sequenceSimilarity - ( sequenceSimilarity + LIGAND_THRESHOLD);
                }
            }
            else if ( technique != other.technique )
            {
                if ( technique == TECHNIQUE.NMR && other.technique == TECHNIQUE.XRAY )
                {
                    // If a sequence determined by X-ray has a sequence id % 20% lower than
                    // one with NMR, use NMR, otherwise use X-ray
                    // i.e. give the 20% seq % boost to the one with X-ray
                    rv = ( other.sequenceSimilarity + XRAY_THRESHOLD) - sequenceSimilarity;
                }
                else if ( technique == TECHNIQUE.XRAY && other.technique == TECHNIQUE.NMR )
                {
                    // opposite of above
                    rv = other.sequenceSimilarity - ( sequenceSimilarity + XRAY_THRESHOLD) ;
                }
            }

            // measures are based on sequence similarities, which are < 1, so
            return (int) (rv * 1000);
        }

        public String toString()
        {

            return  String.format("%.2f", sequenceSimilarity)
                    + ", technique = " + technique
                    + ", ligand = " + hasLigand;
        }
}

    /**
     * A homology modelling result from homodeller
     */
    public static class Result implements Serializable
    {
        // the completed model
        public ProteusStructureImpl model;

        // the template structure used
        public ProteusStructureImpl template;
        public String templateName;
        public GappedSequence query;
        public GappedSequence database;

        public Result(Structure model, Structure template, GappedSequence query, GappedSequence database, String title)
        {
            this.template = new ProteusStructureImpl(template, false);
            this.model = new ProteusStructureImpl(model, false);
            this.database = database;
            this.query = query;
            this.templateName = title;
        }
    }

    public static void main(String[] args) throws JSAPException, BioException, IOException
    {
        JSAPResult options = parseCommandLine(args);

        if ( options != null )

        {
            runFromCommandLine(options);
        }
    }

    /**
     *  Parses the command line parameters
     * @param args command line arguments
     * @return JSAPResult containing the options if parsing was successful, null otherwise
     * @throws com.martiansoftware.jsap.JSAPException
     */
    public static JSAPResult parseCommandLine(String[] args) throws JSAPException
    {
        SimpleJSAP options = new SimpleJSAP(
                "Proteus",
                "Homology Modeling " + VERSION,
                new Parameter[]{
                        new Switch(ENERGY_MINIMIZATION_FLAG, 'e', "energy-minimization",
                                "If you want to run energy minimization"),
                        new FlaggedOption(OUTPUT_FILE_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'o', "output",
                                "The output file for the homology model."),
                        new FlaggedOption(QUERY_ID, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'q', "query_id",
                                "The PDB ID of the query sequence"),
                        new FlaggedOption(TEMPLATE_ID, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 't', "template_id",
                                "The PDB ID of the template sequence."),
                        new UnflaggedOption(FILE_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NOT_GREEDY,
                                "The name of the FASTA file containing the query sequence(s)."),
                        new FlaggedOption(ALIGNMENT_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'a', "alignment",
                                "The name of the alignment file containing the query sequence and the template sequence.  " +
                                        "One sequence must correspond to the 'id' parameter, and the other a valid PDB ID. " +
                                        "The alignment file must be in ClustalW format"),
                        new FlaggedOption(ALIGNMENT_ID_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, JSAP.NO_SHORTFLAG, "id",
                                "Required only if providing an alignment.  Must correspond to the ID of the query sequence in the alignment")
                }
        );

        final JSAPResult config = options.parse(args);

        // if there was a mistake, print the usage and kill the program
        if ( options.messagePrinted() )
        {
            printUsage(options);
            return null;
        }

        if ( ! config.contains(FILE_FLAG) && ! config.contains(QUERY_ID) && ! config.contains(ALIGNMENT_FLAG) )
        {
            System.err.println("****************************************************");
            System.err.println("You must specify either a FASTA file, a PDB query ID, or an alignment file\n");
            printUsage(options);
            return null;
        }

        if ( config.contains(FILE_FLAG))
        {
            if ( ! new File(config.getString(FILE_FLAG)).exists())
            {
                System.err.println("Input file does not exist\n");
                printUsage(options);
                return null;
            }
        }

        if ( config.contains(OUTPUT_FILE_FLAG))
        {
            // try it, make sure we can write output
            try
            {
                new FileOutputStream(OUTPUT_FILE_FLAG);
            }
            catch (FileNotFoundException e)
            {
                printUsage(options);
                return null;
            }
        }

        if ( config.contains(ALIGNMENT_FLAG))
        {
            if ( ! new File(config.getString(ALIGNMENT_FLAG)).exists())
            {
                System.err.println("Alignment file does not exist\n");
                return null;
            }

            if ( ! config.contains(ALIGNMENT_ID_FLAG) )
            {
                System.err.println("Invalid paramaters.  You must specify the 'id' parameter when you provide an alignment, " +
                        "which corresponds to the ID in the alignment for which you wish to create a structure");
                return null;
            }
        }

        return config;
    }

    private static void printUsage(SimpleJSAP options)
    {
        System.err.println("Usage: ");
        System.err.println("  " + options.getUsage());
        System.err.println();
        System.err.println(options.getHelp());
    }

    public static void runFromCommandLine(JSAPResult options)
            throws IOException, BioException
    {
        Chain realChain = null;
        String sequence = null;
        Alignment alignment = null ;

        if ( options.contains(FILE_FLAG))
        {
            // parse the FASTA file, get a list of sequences
            List<Sequence> sequences = ProteusTools.parseFasta(new File(options.getString(FILE_FLAG)));

            if ( sequences.size() > 1 )
            {
                // TODO
                System.out.println("Homodeller only runs on one sequence for now.");
            }

            Sequence seq = sequences.get(0);
            sequence = seq.seqString();
        }
        else if ( options.contains(QUERY_ID))
        {
            String queryID = options.getString(QUERY_ID).toUpperCase();

            // get the real file
            Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");
            File pdbDir = new File( config.getStringValue("pdb_dir"));

            if ( !pdbDir.exists())
            {
                pdbDir.mkdir();
            }

            InputStream in = PDBFileLoader.getPDBFile(queryID, new File[]{pdbDir});
            try
            {
                PDBFileParser pdbReader = new PDBFileParser();
                Structure realFile = pdbReader.parsePDBFile(in);

                // get the sequence
                realChain = realFile.getChain(0);
                HomodellerUtils.removeHetatoms(realChain);

                sequence = realChain.getSequence();
            }
            finally
            {
                try
                {
                    if (in != null)
                    {
                        in.close();
                    }
                }
                catch (IOException e)
                {
                    e.printStackTrace();
                }
            }
        }
        else if ( options.contains(ALIGNMENT_FLAG))
        {
            String alignmentFile = options.getString(ALIGNMENT_FLAG);

            ClustalWParser parser = new ClustalWParser();
            try
            {
                parser.parse(alignmentFile);
            }
            catch (SAXException e)
            {
                e.printStackTrace();
            }

            alignment = parser.getAlignment();
        }
        else
        {
            throw new IllegalStateException("Didn't supply either a file or query ID");
        }

        String templateID = null;
        if ( options.contains(TEMPLATE_ID))
        {
            templateID = options.getString(TEMPLATE_ID);
        }

        // Run it
        HomodellerImpl homodeller = new HomodellerImpl();
        Result result;
        if ( templateID != null && sequence != null )
        {
            result = homodeller.run(sequence, templateID);
        }
        else if ( alignment != null )
        {
            String id = options.getString(ALIGNMENT_ID_FLAG);

            if ( id == null)
            {
                System.err.println("Must provide an 'id' parameter when passing in an alignment.  See --help.");
            }

            result = homodeller.run(id, alignment);

            // Result may be null if the alignment or the id parameter was incorrect
            if ( result == null )
            {
                return;
            }
        }
        else
        {
            result = homodeller.run(sequence);
        }

        Chain resultChain = result.model.getChain(0);
        System.out.println("Similarity %: " + HomodellerUtils.getSequenceSimilarity(result.database.seqString(),  result.query.seqString()) * 100);

        if ( realChain != null )
        {
            // check values
            String resultSequence = resultChain.getSequence();
            if (! resultSequence.equals(sequence))
            {
                int start = sequence.indexOf(resultSequence);

                assert start != -1 : "Result sequence is not a subsequence of the query sequence";

                int end = start + resultSequence.length();

                realChain = new ProteusChainImpl(realChain).subChain(start, end);
            }

            Superpose.RMSD rmsd = Superpose.findRMSD(realChain, resultChain);
            System.out.println("Comparing to " + options.getString(QUERY_ID).toUpperCase() );
            System.out.println(rmsd);
        }

        if ( options.contains(ENERGY_MINIMIZATION_FLAG) && options.getBoolean(ENERGY_MINIMIZATION_FLAG))
        {
            System.out.println("Running enerygy minimization...");

            Structure minimizedStructure = runEnergyMinimization(result.model);

            resultChain = minimizedStructure.getChain(0);

            if ( realChain != null )
            {
                Superpose.RMSD rmsd = Superpose.findRMSD(realChain, resultChain);
                System.out.println("Comparing minimized structure to " + options.getString(QUERY_ID).toUpperCase() );
                System.out.println(rmsd);
            }
        }

        // was an output option specified?
        if ( options.contains(OUTPUT_FILE_FLAG))
        {
            // write the output file
            String file = options.getString(OUTPUT_FILE_FLAG);
            System.out.println("Writing to " + file);
            HomodellerUtils.writeChainToFile(file, resultChain);
        }
        else
        {
            HomodellerUtils.writeChainToStream(System.out, resultChain);
        }
    }

    /**
     * Runs an implementation of energy minimization
     *
     * @param structure the structure to minimize
     * @return a minimized structure
     */
    public static Structure runEnergyMinimization(Structure structure)
    {
        Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");

        String minimizerName = config.getStringValue("energyMinimizer", "gafolder");

        EnergyMinimizer minimizer;
        if ( minimizerName.equals("gafolder"))
        {
            minimizer = new GAFolderRunner(structure);
        }
        else
        {
            PepmakeRunner.Options options = new PepmakeRunner.Options();
            options.fixTorsionAngles = true;

            minimizer = new PepmakeRunner(structure, options);
        }

        return minimizer.getMinimizedStructure();
    }
}
