package ca.ualberta.proteus.program.homodeller;

import ca.ualberta.proteus.util.GenericProgramRunner;
import ca.ualberta.proteus.util.StreamHandler;
import ca.ualberta.proteus.program.transsec.Config;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;

import java.io.*;
import java.util.Timer;
import java.util.TimerTask;

/**
 * A wrapper for homodeller to allow it to run from java
 */
public class HomodellerRunner extends GenericProgramRunner
{
    private HomodellerResults results = null;
    
    private String templateID = "USER"; // default it

    public HomodellerRunner()
    {
        this.results = null;
    }

    public HomodellerRunner(HomodellerResults results)
    {
        this.results = results;
    }

    /**
     * Run homodeller with the given sequence, against a given template
     * @param sequence
     * @param templateFile An InputStream containing a PDB file to be used as our template
     * @return true if homodeller succeeds
     */
    public boolean run(Sequence sequence, InputStream templateFile, boolean doEnergyMinimization)
    {

        /**
         * 1. Write template to PDB file
         * 2. Run Vadar on template
         * 3. Write vadar result to new seqsee.db
         * 4. Run homodeller -f <template> -t <ID of template> -d <new seqsee.db> 
         */


        Config config = Config.getConfig();
        String uniqueID = getUniqueIdentifier();
        String homodellerCmd = config.getStringValue("homodeller.command");
        File homodellerRunDir = new File(homodellerCmd).getParentFile();

        String currentDir = new File("").getAbsolutePath() + "/";
        String fname = currentDir + RUN_DATA_DIR + "/" + uniqueID + ".fa";
        //String templateName  = RUN_DATA_DIR +  "/" + uniqueID + ".pdb";
        String templateName  = homodellerRunDir.getAbsolutePath() + "/pdb/" + templateID + ".pdb";
        ensureRunDataDirExists();

        if ( results != null )
        {
            results.setOutputLogFile(fname + ".out");
        }

        // Write out the template PDB file to homodeller's PDB directory
        try
        {
            FileOutputStream inputfile;
            File f = new File(fname);
            inputfile = new FileOutputStream(f);
            SeqIOTools.writeFasta(inputfile, sequence);
            inputfile.close();

            FileOutputStream templateFileStream;
            templateFileStream = new FileOutputStream(templateName);
            byte buf[] = new byte[1024];
            while ( templateFile.read(buf) != -1 )
            {
                templateFileStream.write(buf);
            }
            templateFileStream.close();


        }
        catch (IOException e)
        {
            System.err.println("Error writing temp file for homodeller input");
            e.printStackTrace();
        }

        // run vadar
        String dbname = currentDir + RUN_DATA_DIR  + "/" + uniqueID;// vadar adds ".main" as suffix        try
        try
        {
            String vadarCmd = config.getStringValue("vadar.command");
            File runDir = new File(vadarCmd).getParentFile();
            vadarCmd = String.format(vadarCmd, templateName, dbname);
            runVadar(vadarCmd, runDir, dbname + ".main");
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        // run homodeller
        try
        {
            String command = homodellerCmd + " -f " + fname + " -t "+ templateID + " -d " + dbname + ".main";

            String outputFileName = fname + ".pdb";
            if ( ! runHomodeller(command, homodellerRunDir, outputFileName) )
                return false;

            if ( results != null )
                results.setOutputFile(outputFileName);

            if ( doEnergyMinimization )
            {
                // Energy min requires a filename ending in .pdb
                String newFileName = outputFileName.replaceFirst(".fa", "");
                File f2 = new File(outputFileName);
                File newFile = new File(newFileName);
                f2.renameTo(newFile);
                runEnergyMinimisation(newFileName);

                if ( results != null )
                    results.setOutputFile(newFileName);
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        return true;
    }

    /**
     * Run homodeller with the given sequence, and let homodeller find the appropriate template
     * @param sequence
     * @param doEnergyMinimization
     */
    public boolean run(Sequence sequence, boolean doEnergyMinimization)
    {
        Config config = Config.getConfig();
        String currentDir = new File("").getAbsolutePath() + "/";
        String uniqueID = getUniqueIdentifier();
        String fname = currentDir + RUN_DATA_DIR + "/" + uniqueID + ".fa";
        ensureRunDataDirExists();

        if ( results != null )
        {
            results.setOutputLogFile(fname + ".out");
            results.setOutputFile(fname);
        }

        try
        {
            FileOutputStream inputfile;
            File f = new File(fname);
            inputfile = new FileOutputStream(f);
            SeqIOTools.writeFasta(inputfile, sequence);
            inputfile.close();
        }
        catch (IOException e)
        {
            System.err.println("Error writing temp file for homodeller input");
            e.printStackTrace();
        }

        try
        {
            String homodellerCmd = config.getStringValue("homodeller.command");

            File runDir = new File(homodellerCmd).getParentFile();

            String command = homodellerCmd + " " + fname;

            String outputFileName = fname + ".pdb";
            boolean succeeded = runHomodeller(command, runDir, outputFileName);

            if ( results != null )
                results.setOutputFile(outputFileName);

            if ( !succeeded)
                return false;

            if ( doEnergyMinimization )
            {
                // Energy min requires a filename ending in .pdb
                String newFileName = outputFileName.replaceFirst(".fa", "");
                File f2 = new File(outputFileName);
                File newFile = new File(newFileName);
                f2.renameTo(newFile);
                runEnergyMinimisation(newFileName);

                if ( results != null )
                    results.setOutputFile(newFileName);
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        return true;
    }

    /**
     * Runs homodeller and handles accumulating the output and results.
     *
     * @param command - The command string used to run homodeller
     * @param runDir - The directory to run it in (should be homodeller's directory)
     * @param outputFileName - the PDB file homodeller's results should be put in
     * @return true if homodeller succeeds
     * @throws IOException
     * @throws InterruptedException
     */
    private boolean runHomodeller(String command, File runDir, String outputFileName) throws IOException, InterruptedException {

        System.err.println(command);

        Process homodellerProc = Runtime.getRuntime().exec(command, null, runDir.getAbsoluteFile() );

        // any error message?
        StreamHandler errorReader = new
                StreamHandler(homodellerProc.getErrorStream(), "ERROR");

        // any output?
        StreamHandler outputReader = new
                StreamHandler(homodellerProc.getInputStream(), "OUTPUT");

        // kick them off
        errorReader.start();
        outputReader.start();

        // any error???
        int exitVal = homodellerProc.waitFor();

        if ( exitVal != 0 )
        {
            // homodeller returned error
            System.err.println("Homodeller failed " + errorReader.getInputString());
            System.err.println(outputReader.getInputString());

            results.setErrorString( outputReader.getInputString());
            return false;
        }
        else if ( ! new File(outputFileName).exists() )
        {
            // homodeller returned positive result, but failed for some reason
            BufferedReader reader = new BufferedReader(new StringReader(outputReader.getInputString()));

            // Gather stats for results
            String line;
            while ( ( line = reader.readLine() ) != null ) {
                if ( line.contains("db_percent="))
                {
                    line = line.replaceFirst(".*db_percent=(.*),.*", "$1");
                    double db_percent = Double.valueOf(line);
                    if ( results != null )
                        results.setPercentHomology(db_percent);
                    results.setErrorString( outputReader.getInputString() );
                    return false;
                }
            }
        }
        else
        {
            // homodeller succeeded
            if ( results != null )
                results.setResultsFile(outputFileName);

            BufferedReader reader = new BufferedReader(new StringReader(outputReader.getInputString()));

            // Gather stats for results
            String line;
            while ( ( line = reader.readLine() ) != null ) {
                if ( line.contains("db_percent="))
                {
                    line = line.replaceFirst(".*db_percent=(.*),.*", "$1");
                    double db_percent = Double.valueOf(line);
                    if ( results != null )
                        results.setPercentHomology(db_percent);
                }
                else if ( line.contains("Opened pdb"))
                {
                    line = line.replaceFirst(".*/(\\w+)\\.pdb.*", "$1");
                    if ( results != null )
                        results.setTemplateID(line);
                }
            }
            return true;
        }

        return false;
    }

    /**
     * Do not use
     */
    public void run()
    {
        throw new IllegalStateException("Not implemented");
    }

    /**
     * Run the energy minimization routine on the specified output file
     * @param filename
     */
    public void runEnergyMinimisation(String filename)
    {
        Config config = Config.getConfig();

        Process energyMinProc = null;
        try
        {
            String energyMinCmd = config.getStringValue("energy_min.command");

            // This command should take the form "<command> %s", where %s is filename
            energyMinCmd = String.format(energyMinCmd, filename);

            File f = new File(config.getStringValue("energy_min.run_dir"));

            System.err.println(energyMinCmd);

            energyMinProc = Runtime.getRuntime().exec(energyMinCmd, null, f.getAbsoluteFile() );

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(energyMinProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(energyMinProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            /* Schedule a 10 minute timeout on this process */
            final HomodellerRunner thisRunner = this;
            Timer timer = new Timer();
            timer.schedule(new TimerTask() {

                public void run()
                {
                    thisRunner.interrupt();
                }
            }, 600000);

            // any error???
            int exitVal = energyMinProc.waitFor();

            // cleanup
           // cleanup(filename);

         /*   if ( exitVal != 0 )
            {
                System.err.println("Homodeller failed " + errorReader.getInputString());
                System.out.println(outputReader.getInputString());
            }
            */
            File progOutputFile = new File(filename.replaceFirst(".pdb", "PmsdPmcgPmbf.pdb"));
            File outputFile = new File(filename.replaceFirst(".pdb", "_minimized.pdb"));

            if ( progOutputFile.exists() )
            {
                cleanupOutputFileFormat(progOutputFile, outputFile);

                if ( results != null )
                    results.setResultsFile(outputFile.getName());
            }

            cleanup(filename.replace(".pdb", ""));
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }

    }

    /**
     * Modify the energy minimized output file so it's a valid PDB file
     * @param inputFile
     * @param outputFile
     */
    private void cleanupOutputFileFormat(File inputFile, File outputFile) throws IOException
    {
        BufferedReader reader = new BufferedReader( new InputStreamReader(new FileInputStream(inputFile)) );
        PrintWriter writer = new PrintWriter(outputFile);
        String line;
        while ( (line = reader.readLine()) != null )
        {
            if ( line.equals("ENDMDL"))
                line = "END";
            writer.println(line);
        }
        reader.close();
        writer.close();
    }

    public void cleanup(String filename)
    {
        File dir = new File(filename).getParentFile();
        String ID = new File(filename).getName();

        File[] files = dir.listFiles();
        for ( File file : files)
        {
            if ( file.getName().contains(ID))
            {
                if ( ! file.getName().contains(".pdb") && ! file.getName().contains(".out") )
                    file.delete();
                if ( file.getName().contains("Pmsd.pdb")
                        || file.getName().contains("PmsdPmcg.pdb")
                        || file.getName().contains("PmsdPmcgPmbf.pdb") )
                    file.delete();
            }
        }
    }

    /**
     * Runs vadar and handles accumulating the output and results.
     *
     * @param command - The command string used to run homodeller
     * @param runDir - The directory to run it in (should be homodeller's directory)
     * @param dbname
     * @throws IOException
     * @throws InterruptedException
     */
    private void runVadar(String command, File runDir, String dbname) throws IOException, InterruptedException {

        System.err.println(command);

        Process vadarProc = Runtime.getRuntime().exec(command, null, runDir.getAbsoluteFile() );

        // any error message?
        StreamHandler errorReader = new
                StreamHandler(vadarProc.getErrorStream(), "ERROR");

        // any output?
        StreamHandler outputReader = new
                StreamHandler(vadarProc.getInputStream(), "OUTPUT");

        // kick them off
        errorReader.start();
        outputReader.start();

        // any error???
        int exitVal = vadarProc.waitFor();

        if ( exitVal != 0 )
            System.err.println("Vadar failed " + errorReader.getInputString());

        BufferedReader reader = new BufferedReader(new FileReader(dbname));
        ByteArrayOutputStream fixedFile = new ByteArrayOutputStream();
        // Fix the output
        String line;
        while ( ( line = reader.readLine() ) != null ) {
            if ( line.startsWith(">"))
            {
                line = ">" + templateID;
            }
            else if ( line.contains("#BRKHAVN ID:"))
                line = "#BRKHAVN ID: " + templateID;

            fixedFile.write((line+ "\n") .getBytes());
        }
        reader.close();

        // write it out
        FileOutputStream out = new FileOutputStream(dbname);
        fixedFile.writeTo(out);
        out.close();
    }

    public void setTemplateID(String templateID)
    {
        this.templateID = templateID;
    }
}
