package ca.ualberta.proteus.program.homodeller;

import ca.smontgomerie.util.Math.MathExt;
import ca.ualberta.proteus.util.ProteusTools;
import org.apache.log4j.Logger;
import org.biojava.bio.structure.*;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.io.SymbolTokenization;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeMap;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Sep 7, 2006
 * Time: 11:10:14 PM
 * To change this template use File | Settings | File Templates.
 */
public class HomodellerUtils
{
    private static final Logger logger = Logger.getLogger(HomodellerUtils.class);

    Alphabet alpha_prot = ProteinTools.getAlphabet();

    private static SymbolTokenization threeLetter, oneLetter;

    static
    {
        try
        {
            Alphabet alpha_prot = ProteinTools.getAlphabet();
            threeLetter = alpha_prot.getTokenization("name");
            oneLetter = alpha_prot.getTokenization("token");
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Returns the cartesian distance between two atoms
     * @param a1
     * @param a2
     * @return the cartesian distance between two atoms
     */
    public static double getDistance(Atom a1, Atom a2)
    {
        double x = a1.getX() - a2.getX();
        double y = a1.getY() - a2.getY();
        double z = a1.getZ() - a2.getZ();
        return Math.sqrt( MathExt.square(x) + MathExt.square(y) + MathExt.square(z) );
    }
    
    /**
     * Returns the magnitude of the difference between the two
     * given angles, with angles measured in degrees.
     * 
     * @param a1 first angle (in degrees)
     * @param a2 second angle (in degrees)
     * @return the magnitude of the difference between the two given angles, with angles measured in degrees
     */
    public static double angleDist(double a1, double a2) {
    	
    	// Put each angle in the range [-180.0, 180.0).
    	while (a1 < -180.0) {
    		a1 += 360.0;
    	}
    	while (a1 >= 180.0) {
    		a1 -= 360.0;
    	}
    	
    	while (a2 < -180.0) {
    		a2 += 360.0;
    	}
    	while (a2 >= 180.0) {
    		a2 -= 360.0;
    	}
    	
    	// Adjust angles 
    	if (a1 - a2 >= 0.0) {
    		if (a1 - a2 < 180.0) {
    			// No change required.
    		}
    		else {
    			a2 += 360.0;
    		}
    	}
    	else {
    		if (a1 - a2 > -180.0) {
    			// No change required.
    		}
    		else {
    			a1 += 360.0;
    		}
    	}
    	
    	return Math.abs(a1 - a2);
    }
    
    
    public static TreeMap<Integer,Angle> getAngles(Chain chain)
    {
        TreeMap<Integer,Angle> rv = new TreeMap<Integer, Angle>();
        ArrayList<Group> groups = chain.getGroups();
        for ( int i=0; i< groups.size(); i++ )
        {
            try
            {
                double phi = 360;
                double psi = 360;
                double chi = 999;


                Group g = groups.get(i);

                if ( ! (g instanceof EmptyResidue) )
                {

                    Atom N = g.getAtom("N");
                    Atom CA = g.getAtom("CA");
                    Atom C = g.getAtom("C");

                    if (i > 0)
                    {
                        // calculate phi
                        Group last = groups.get(i - 1);

                        if ( last instanceof EmptyResidue )
                        {
                            phi = 0;
                        }
                        else
                        {
                            Atom C_1 = last.getAtom("C");

                            // You can also use the Point3 class
                            // http://www.xml-cml.org/cmldom/htmlDoc/jumbo/euclid/Point3Vector.html
                            //phi = Point3.getTorsion(createPoint(C_1), createPoint(N), createPoint(CA), createPoint(C)).getDegrees();
                            phi = TorsionUtils.torsion(C_1, N, CA, C);
                        }
                   }

                    if (i < groups.size() - 1)
                    {
                        // calculate psi
                        Group next = groups.get(i + 1);

                        if ( next instanceof EmptyResidue )
                        {
                            psi = 0;
                        }
                        else
                        {
                            Atom N_1 = next.getAtom("N");

                            // You can also use the Point3 class
                            // http://www.xml-cml.org/cmldom/htmlDoc/jumbo/euclid/Point3Vector.html
                            psi = TorsionUtils.torsion(N, CA, C, N_1);
                        }
                    }

                    // calculate chi
                    Atom CB = null;
                    Atom COG = null;

                    if ( g.hasAtom("CB"))
                    {
                        CB = g.getAtom("CB");

                        // if there is a CB, maybe there's a CG too?
                        // if not, there isn't any point looking for one
                        List<Atom> atoms = g.getAtoms();
                        for ( Atom a: atoms)
                        {
                            if ( a.getName().equals("CG")
                                    ||  a.getName().equals("CG1")
                                    ||  a.getName().equals("SG")
                                    ||  a.getName().equals("OG")
                                    ||  a.getName().equals("OG1"))
                            {
                                COG = a;
                                break;
                            }
                        }
                    }

                    if ( CB != null && COG != null)
                    {
                        chi = TorsionUtils.torsion(N, CA, CB, COG);
                    }

                    Angle a = new Angle();
                    a.setPhi(phi);
                    a.setPsi(psi);
                    a.setChi(chi);

                    rv.put(i + 1, a);
                }
                else
                {
                    rv.put(i + 1, Angle.EMPTY);
                }
            }
            catch (StructureException e)
            {
                rv.put(i + 1, Angle.EMPTY);

                logger.error("Bad PDB structure!");
                logger.error(e);
            }
        }

        return rv;
    }

    /**
     * Removes HETATM residues and ACE residues
     * @param realChain
     */
    public static void removeHetatoms(Chain realChain)
    {
        ArrayList<Group> groups = realChain.getGroups();
        List<Group> removals = new ArrayList<Group>();
        for ( Group g: groups)
        {
            // remove it if it's an unknown amino acid like ACE, or if it's not an Amino Acid group
            if ( ! ( g instanceof AminoAcid ) || ((AminoAcid)g).getAminoType().equals('x'))
            {
                removals.add(g);
            }
        }

        for ( Group g: removals)
        {
            groups.remove(g);
        }
    }

    /**
     * Returns the sequence similarity score for the residue (as a proportion in the range [0.0, 1.0]).
     * TODO possibly augment with the PAM30 similarity matrix, evaluate usefulness of similarity as a measure
     * @param sequence1 - typically the database sequence
     * @param sequence2 - typically the query sequence
     * @return a similarity measure
     *
     */
    public static double getSequenceSimilarity(String sequence1, String sequence2)
    {
        assert sequence1.length() == sequence2.length() : "Can't get similarity of different length sequences";

        char[] chars1 = sequence1.toCharArray();
        char[] chars2 = sequence2.toCharArray();

        double identical = 0;
        int length1 = 0, length2 = 0;
        for(int i=0; i<chars1.length; i++ )
        {
            if ( chars1[i] == chars2[i])
            {
                identical++;
            }

            if (!ProteusTools.isGap(chars1[i]))
                length1++;
            if (!ProteusTools.isGap(chars2[i]))
                length2++;
        }

        // get length of shorter sequence
        int length = Math.min(length1, length2);

        // # of identical matches / length of shorter
        return identical / (double) length;
    }

    /**
     * Writes a chain to a file as a PDB formatted file
     * @param filename
     * @param chain
     */
    public static void writeChainToFile(String filename, Chain chain)
    {
        try
        {
            File f = new File(filename);
            
            // Make sure the directory is there
            File parent = f.getParentFile();
            if (parent != null) {
            	parent.mkdirs();
            }

            writeChainToStream(new FileOutputStream(f), chain);
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Writes a chain to an output stream in PDB format
     * @param out
     * @param chain
     */
    public static void writeChainToStream(OutputStream out, Chain chain)
    {
        StructureImpl tmp = new StructureImpl();
        tmp.addChain(chain);

        // use an overrided writer, Biojava's is broken
        ca.ualberta.proteus.biojava.FileConvert f = new ca.ualberta.proteus.biojava.FileConvert(tmp) ;
        String pdbText = f.toPDB();

        PrintWriter w = new PrintWriter(out);
        w.print(pdbText);
        w.close();
    }

    /**
     * find the side chain with the amino acid 1-letter code and chi angle
     *
     * @param amino_acid acid 1-letter code
     * @param chi        angle
     * @return a side chain
     */
    public static SideChain getSideChain(char amino_acid, double chi)
    {
        SideDB db = SideDB.getInstance();

        List<SideChain> sideChains = db.getSideChains(amino_acid);
        SideChain newSideChain = null;
        for (SideChain c : sideChains)
        {
            if (chi != 999)
            {
                double anglediff = Math.abs(chi - c.getAngle());

                if ((anglediff < 60) || (Math.abs(anglediff - 360) < 60))
                {
                    newSideChain = c;
                    break;
                }
            }
            else
            {
                // I think chi means that this residue doesn't have rotamers, choose the first one
                newSideChain = c;
                break;
            }
        }

        if (newSideChain == null)
        {
            logger.debug("Can't find side chain for residue " + amino_acid);

            // Try our next best guess
            SideChain sc = null;
            double bestAngle = Double.MAX_VALUE;
            for (SideChain c : sideChains)
            {
                if (chi != 999)
                {
                    double anglediff = Math.abs(chi - c.getAngle());

                    if (anglediff < bestAngle)
                    {
                        sc = c;
                        bestAngle = anglediff;
                    }
                }
            }

            newSideChain = sc;
        }

        // Clone it to return a copy so we don't modify this one in our database
        return (SideChain) newSideChain.clone();
    }

    /**
     * convert three character amino acid codes into single character
     * e.g. convert CYS to C
     * @param code3 String to convert
     * @throws org.biojava.bio.symbol.IllegalSymbolException
     *
     * @return 1-character code
     */
    public static Character convert_3code_1code(String code3)
            throws IllegalSymbolException
    {
        Symbol sym = threeLetter.parseToken(code3);
        String code1 = oneLetter.tokenizeSymbol(sym);

        return code1.charAt(0);

    }
}
