package ca.ualberta.proteus.program.homodeller;

import org.apache.log4j.Logger;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.structure.Atom;
import org.biojava.bio.structure.AtomImpl;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.NucleotideImpl;
import org.biojava.bio.structure.io.PDBParseException;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.io.CountedBufferedReader;
import org.biojava.utils.io.RandomAccessReader;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Aug 31, 2006
 * Time: 10:52:44 PM
 * To change this template use File | Settings | File Templates.
 */
public class LoopDBReader
{
    private LoopDB db;

    private Loop current_loop;
    private Group current_group;

    private String[] nucleotides = new String[]{"C", "G", "A", "T", "U", "I", "+C", "+G", "+A", "+T", "+U", "+I"};
    private SymbolTokenization threeLetter;
    private SymbolTokenization oneLetter;

    private boolean lazyParsing = false;

    private static String patternStr = "\\s+(\\d+)" + // #
                "\\s+(\\w+)" +  // # atom name, amino acid code
                "\\s+(\\w+)" +  // # atom type (N, CA, C, O, etc.)
                "\\s+(\\w+)" +  // sec str triplet (BBB, BBC)
                "\\s+(\\w+)" +  // sec str single (B,C, H)
                "\\s+([0-9.-]+)" + // x
                "\\s+([0-9.-]+)" + // y
                "\\s+([0-9.-]+)";  // z
    protected static Pattern pattern = Pattern.compile(patternStr);

    private static Logger logger = Logger.getLogger(LoopDBReader.class);


    public LoopDBReader(LoopDB db)
    {
        this(db, true);
    }

    /**
     *
     * @param db
     * @param lazyParsing True if you want a fast parsing that allows you to go back and read structures later (recommended)
     */
    public LoopDBReader(LoopDB db, boolean lazyParsing)
    {
        this.db = db;
        this.lazyParsing = lazyParsing;

        Alphabet alpha_prot = ProteinTools.getAlphabet();

        try
        {
            threeLetter = alpha_prot.getTokenization("name");
            oneLetter = alpha_prot.getTokenization("token");
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    public void parseDB(File file)
    {
        try
        {
            RandomAccessReader reader = new RandomAccessReader(new RandomAccessFile(file.getAbsolutePath(), "r"));
            CountedBufferedReader buf = new CountedBufferedReader(reader);

            parseDB( buf );
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();  
        }
    }
/*
    private void parseLoopDB(FileInputStream fileInputStream)
    {
        //To change body of created methods use File | Settings | File Templates.
    }*/

    public void parseDB(InputStream inputStream)
    {
        try
        {
            parseDB(new BufferedReader(new InputStreamReader(inputStream)));
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

     public void parseDB(BufferedReader buf)
            throws IOException
    {
        // (re)set structure

        current_group = null;

        logger.debug("Reading Loop DB...");

        try
        {

            String line = buf.readLine();

            // if line is null already for the first time, the bufferede Reader had a problem
            if (line == null)
            {
                throw new IOException("could not parse PDB File, BufferedReader returns null!");
            }

            //System.out.println ("starting to parse:");
            //System.out.println (line);

            while (line != null)
            {
                if ( line.length() > 0)
                {                
                    if (line.startsWith(">"))
                    {
                        pdb_RECORD_Handler(line);
                        if ( buf instanceof CountedBufferedReader)
                        {
                            current_loop.setFileIndex( ((CountedBufferedReader) buf).getFilePointer() );
                        }

                        // set it to inialized if we're not doing lazy parsing
                        current_loop.setInitialized(! lazyParsing);
                    }
                    else /*if (line.matches(patternStr))*/
                    {
                        if ( ! lazyParsing )
                        {
                            if ( line.startsWith("*"))
                            {
                                // parse out comments
                                continue;
                            }

                            pdb_ATOM_Handler(line);
                        }
                    }
                }

                line = buf.readLine();
            }

            // add the last one
            if ( current_group != null )
            {
                current_loop.addGroup(current_group);
            }

            if ( current_loop != null )
            {
                db.addLoop(current_loop);

                // renumber the loop residues and atoms
                if ( current_loop.getInitialized())
                {
                    cleanupLoop(current_loop);
                }
            }
        }
        catch (Exception e)
        {

            e.printStackTrace();
            throw new IOException("Error parsing Loop DB");
        }

        logger.debug("Done");
    }

    private void pdb_RECORD_Handler(String line)
    {
        // add old one the database
        if ( current_group != null )
        {
            current_loop.addGroup(current_group);

            // set group to null so next loop will not add it
            current_group = null;
        }

        if ( current_loop != null )
        {
            db.addLoop(current_loop);

            // renumber the loop residues and atoms
            if ( current_loop.getInitialized())
            {
                cleanupLoop(current_loop);
            }
        }


        current_loop = new Loop();

        String[] strings = line.replace(">", "").split(":");

        current_loop.setLoopNumber(Integer.parseInt(strings[0]));
        current_loop.setLoopLength(Integer.parseInt(strings[1]));

        /* There's two variations on the database
         * 1. >Loop_Num:res_num:loop_variation
         * 2. >Loop_Num:Loop_Length:Loop_Type:aa:ac:an:ao:ab:ca:cc:cn:co:cb:na:nc:nn:no:nb:oa:oc:on:oo:ob:ba:bc:bn:bo:bb
         */
        /*
        try
        {
            // NOTE This isn't used right now
            // current_loop.setVariationNumber(Integer.parseInt(strings[2]));
        }
        catch(NumberFormatException e)
        {

            /* This isn't used

            ExtendedLoopData data = new ExtendedLoopData();
            data.setLoopType(strings[2]);
            ExtendedLoopData.FIELD[] fields = ExtendedLoopData.FIELD.values();

            for ( int i=3; i<strings.length; i++)
            {
                ExtendedLoopData.FIELD field = fields[i - 3];
                data.setData(field, Double.valueOf(strings[i]));
            }

            current_loop.setLoopData(data);

        }
        */

    }

    // Handler methods to deal with PDB file records properly.


    /**
     * Database version of the ATOM record
     *
     *  10      LYS     N   CCC C   10.727   -7.235   -6.494
     *  10      LYS     CA  CCC C   11.312   -8.583   -6.444
     *  10      LYS     C   CCC C   10.444   -9.441   -5.520
     *  10      LYS     O   CCC C    9.434   -8.971   -4.976
     * @param line
     * @throws PDBParseException
     */
    private void pdb_ATOM_Handler(String line)
            throws PDBParseException
    {
        //System.out.println(line);
        Matcher matcher = pattern.matcher(line);
        matcher.find();

        // create new atom
        AtomImpl atom = new AtomImpl();

        String residueNumber = matcher.group(1);
        String groupCode3 = matcher.group(2);

        String fullname;
        /*if ( residueNumber.length() == 3 )
        {*/
            int offset = residueNumber.length() - 3;
            fullname = line.substring(8 + offset, 12 + offset).replace('\t', ' ');
      /*  }
        else
        {
            fullname = line.substring(7, 11);
        }*/
        // TODO this fails when residueNumber is length 1, fix after debugging the loop

        atom.setFullName(fullname);
        atom.setName(matcher.group(3));

        double x = Double.parseDouble(matcher.group(6));
        double y = Double.parseDouble(matcher.group(7));
        double z = Double.parseDouble(matcher.group(8));

        double[] coords = new double[3];
        coords[0] = x;
        coords[1] = y;
        coords[2] = z;
        atom.setCoords(coords);


        Character aminoCode;
        try
        {
            // is it a standard amino acid ?
            aminoCode = HomodellerUtils.convert_3code_1code(groupCode3);
        }
        catch (IllegalSymbolException e)
        {
            // hm groupCode3 is not standard
            // perhaps it is an nucleotide?
            if (isNucleotide(groupCode3))
            {
                //System.out.println("nucleotide, aminoCode:"+aminoCode);
                aminoCode = null;
            }
            else
            {
                // does not seem to be so let's assume it is
                //  nonstandard aminoacid and label it "X"
                logger.error("unknown amino acid " + groupCode3);
                aminoCode = 'x';
            }
        }


        if (current_group == null)
        {

            current_group = getNewGroup(aminoCode, matcher.group(5));

            current_group.setPDBCode(residueNumber);
            current_group.setPDBName(groupCode3);
        }

        // check if residue number is the same ...
        // insertion code is part of residue number
        if (! residueNumber.equals(current_group.getPDBCode()))
        {
            //System.out.println("end of residue: "+current_group.getPDBCode()+" "+residueNumber);
            current_loop.addGroup(current_group);

            current_group = getNewGroup( aminoCode, matcher.group(5));

            current_group.setPDBCode(residueNumber);
            current_group.setPDBName(groupCode3);

        }

        current_group.addAtom(atom);
        //System.out.println(current_group);


    }

    /**
     * The file MUST be the same as originally parsed with
     * @param file
     * @param loop
     * @throws IOException
     */
    public void fillInLoopInfo(File file, final Loop loop) throws IOException
    {
        RandomAccessReader reader = new RandomAccessReader(new RandomAccessFile(file.getAbsolutePath(), "r"));
        CountedBufferedReader buf = new CountedBufferedReader(reader);

        reader.seek(loop.getFilePointer());

        current_loop = loop;

        String line;
        while ( (line = buf.readLine()) != null )
        {
            if ( line.length() == 0 || line.startsWith("*"))
            {
                // continue;
            }
            else if ( line.startsWith(">") )
            {
                if ( current_group != null )
                {
                    current_loop.addGroup(current_group);

                    // set group to null so next one won't pick it up
                    current_group = null;
                }

                buf.close();

                loop.setInitialized(true);

                cleanupLoop(current_loop);

                return;
            }
            else/* if (line.matches(patternStr))*/
            {
                try
                {
                    pdb_ATOM_Handler(line);
                }
                catch (PDBParseException e)
                {
                    e.printStackTrace();
                }
            }
        }

        // if it's the last loop, we will run out of file
        buf.close();
        loop.setInitialized(true);
        cleanupLoop(current_loop);
        return;
    }

    private void cleanupLoop(Loop loop)
    {
        int atomNum =0;
        ArrayList<Group> groups = loop.getGroups();

        for ( int i=0; i<groups.size(); i++)
        {
            Group g = groups.get(i);
            g.getPDBCode();
            List<Atom> atoms = g.getAtoms();
            for ( Atom a: atoms)
            {
                a.setPDBserial(atomNum++);
            }

            g.setPDBCode(String.valueOf(i));
        }
    }


    /* test if the threlettercode of an ATOM entry corresponds to a
     * nucleotide or to an aminoacid
     */
    protected boolean isNucleotide(String groupCode3)
    {

        for (String n : nucleotides)
        {
            //System.out.print("isNucleotide?"+groupCode3.trim()+" "+n);
            if (groupCode3.trim().equals(n))
            {
                //System.out.println("yes!") ;
                return true;
            }
            //System.out.println("no...") ;
        }
        return false;
    }    

    /**
     * initiale new group, either Hetatom or AminoAcid
     */
    private Group getNewGroup(Character aminoCode, String residueType)
    {
        Group group;
        if (aminoCode != null)
        {

            LoopResidueGroup aa = new LoopResidueGroup();
            aa.setAminoType(aminoCode);
            aa.setResidueType(residueType);
            group = aa;
        }
        else
        {
            // it is a nucleotidee
            group = new NucleotideImpl();
        }

        //System.out.println("new group type: "+ group.getType() );
        return group;
    }
}
