package ca.ualberta.proteus.program.homodeller;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.program.minimization.PepmakeRunner;
import ca.ualberta.proteus.util.ProteusTools;
import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.structure.*;
import org.biojava.bio.structure.io.PDBParseException;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.io.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * <p/>
 * Inserts loops in a two-sequence alignment
 * <p/>
 * <br/>
 * eg. Deletion
 * <pre>
 * Query: NA<b>A-AK</b>SE
 * Datab: NA<b>APAK</b>SE
 * </pre>
 * <p/>
 * Replaces the string "AAK" in the query with a 3-length loop (keeping the same residues).  For a deletion, we would go from
 * a 4-length loop down to a 3-length loop.
 * <br>
 * For an insertion, we would go from a 3-length up to a 4-length
 * <pre>
 * Query: NA<b>APAK</b>SE
 * Datab: NA<b>A-AK</b>SE
 * </pre>
 * <br>
 * Here we would replace the residues AAK with APAK, or a 3-length loop to a 4-length.
 * <br>
 * For a longer loop:
 * <pre>
 * DELETION:
 * Query: NA<b>A--AK</b>SE
 * Datab: NA<b>APKAK</b>SE
 * <p/>
 * The string "AAK" in the query would be replaced.
 * <p/>
 * INSERTION:
 * Query: NA<b>APKAK</b>SE
 * Datab: NA<b>A--AK</b>SE
 * <p/>
 * The string "PKA" would be replaced
 * </pre>
 */
public class LoopInserter
{
    //private File incompleteFile;
    private Structure newStructure;
    private Structure template;
    private GappedSequence databaseSeq;
    private GappedSequence querySeq;
    private Map<Integer, Angle> chiAngles;
    private static boolean debug = Config.getConfig().getBooleanValue("debug");
    private static boolean debugSideChains = Config.getConfig().getBooleanValue("debugSideChains");
    private static int debugLoopNum = -1;

    private static final int MERGE_GAPS_THRESHOLD = 0;
    final int MINIMIZATION_MAX_TRIES = 5;
    private static final double MINIMIZATION_DISTANCE = 1.5; /*TODO This is an arbitrary value*/
    private static final double N_O_MAX_DISTANCE = 2.2; 
    private static Logger logger = Logger.getLogger(LoopInserter.class);

    public LoopInserter(Structure template, GappedSequence querySeq, GappedSequence templateSequence)
    {
        this.querySeq = querySeq;
        this.databaseSeq = templateSequence;
        this.template = template;
    }

    /**
     * Main method
     *
     * @return the new structure
     */
    public Structure run()
    {
        try
        {
            assert template.size() == 1 : "Template has too many chains";

            // copy it to the new structure
            newStructure = (Structure) template.clone();

            /*    PhipsiRunner phiPhipsiRunner = new PhipsiRunner(newStructure.toPDB());
                        chiAngles = phiPhipsiRunner.runPhiPsi();
            */

            // TODO - for some reason, phipsi internally screws up the chain,
            // probably because of numbering or something.  So keep using phipsi as
            // external program for now
            HomodellerUtils.removeHetatoms(newStructure.getChain(0));

            chiAngles = HomodellerUtils.getAngles(newStructure.getChain(0));

            // make sure our database sequence matches the template structure
            assert newStructure.getChain(0).getSequence().equals(databaseSeq.getSourceSymbolList().seqString()) :
                    "The database structure and template structure do not match";

            // insert gaps or deletions
            insertTemporaryGaps();

            assert querySeq.getSourceSymbolList().length() == newStructure.getChain(0).getSequence().length();
            
            
            
            if (debug)
            {
                String pdbText = newStructure.toPDB();
                File tempFile = new File("newStructure.pdb");
                //tempFile.deleteOnExit();

                PrintWriter w = new PrintWriter(tempFile);
                w.print(pdbText);
                w.close();
            }

            // do final loop insertions
            insertLoops();

            assert querySeq.getSourceSymbolList().length() == newStructure.getChain(0).getSequence().length();

            // Change the sequence and side chains (except for empty residues).
            modifyResidues();

            // debug
            //System.err.println("\n   Writing newStructure_mod.pdb after modifyResidues()...");
            writeChainToFile("newStructure_mod.pdb", newStructure.getChain(0));
            
            // Remove empty residues at the beginning and end.
            removeBeginningAndEndGaps();
            
            // Remove empty residues in the interior.
            removeInteriorGaps();
            
            // go through and verify side chains and numbering
            cleanupChain();
            
            // debug
            //System.err.println("\n   Writing newStructure_clean.pdb after cleanupChain()...");
            writeChainToFile("newStructure_clean.pdb", newStructure.getChain(0));


            // remove any groups we can't gap right
            //removeAllUnFilledGaps();
            
            // debug
            //System.err.println("\n   Writing newStructure_rug.pdb after removeAllUnFilledGaps()...");
            writeChainToFile("newStructure_rug.pdb", newStructure.getChain(0));


            // Do bump checking
            checkBumps();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        return newStructure;
    }

    /**
     * For right now, verifies that all C=0 bonds are proper.  This is to fix a bug where O atoms were being placed
     * alongside N atoms in an odd formation.  It will first try using a exact 90 degree angle placement of the O-atom,
     * and if that is still improper, will grab a new sidechain example and use the O-coordinate from that sidechain's
     * O atom.
     * Note: taking an example sidechain may be the proper way to do this.
     *
     * In future, this may be able to rotate side chains that bump with other side chains. 
     */
    private void checkBumps()
    {
        ArrayList groups = newStructure.getChain(0).getGroups();

        for (int i = 0; i < groups.size() - 1; i++)
        {
            Group N_Group = (Group) groups.get(i + 1);
            Group C_Group = (Group) groups.get(i);

            try
            {
                Atom N_atom = N_Group.getAtom("N");
                Atom O_atom = C_Group.getAtom("O");
                Atom C_atom = C_Group.getAtom("C");
                Atom CA_atom = C_Group.getAtom("CA");

                // check for bumps on moved fragment, between the N and O atom
                double N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);
                if (N_O_dist < N_O_MAX_DISTANCE)
                {
                    // too close, rotate C=O carboxyl to move away from N

                    // this is distance we want C=O to have
                    double C_O_dist = HomodellerUtils.getDistance(C_atom, O_atom);

                    double[] CA_coords = CA_atom.getCoords();
                    double[] N_coords = N_atom.getCoords();
                    double[] C_coords = C_atom.getCoords();

                    // find the point between the CA and N atom
                    double[] point = new double[3];
                    point[0] = (CA_coords[0] + N_coords[0]) / (double) 2;
                    point[1] = (CA_coords[1] + N_coords[1]) / (double) 2;
                    point[2] = (CA_coords[2] + N_coords[2]) / (double) 2;

                    // find the line between (point) and C, which we will extrapolate to be
                    // the position of the O
                    double[] line = new double[3];
                    line[0] = C_coords[0] - point[0];
                    line[1] = C_coords[1] - point[1];
                    line[2] = C_coords[2] - point[2];

                    // change line's magnitude to the magnitude of the distance between C=O
                    // i.e. |line| = C_O_dist
                    double multiplier = C_O_dist / Superpose.normalize(line);
                    line[0] *= multiplier;
                    line[1] *= multiplier;
                    line[2] *= multiplier;

                    // these are O's new coordinates
                    double O_coords[] = O_atom.getCoords();
                    O_coords[0] = C_coords[0] + line[0];
                    O_coords[1] = C_coords[1] + line[1];
                    O_coords[2] = C_coords[2] + line[2];

                    N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);

                    if ( N_O_dist < N_O_MAX_DISTANCE )
                    {
                        /* The distance still sucks.  Find the appropriate side chain and see if its oxygen atom
                         * will work.
                         */

                        Angle angle = chiAngles.get(i);
                        double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                        // get a suitable side chain
                        Character aminoType = ((AminoAcid) N_Group).getAminoType();
                        SideChain newSideChain = HomodellerUtils.getSideChain(aminoType, chi);
                        
                        // modify its coordinates to our protein
                        superImposeResidue(C_Group, newSideChain);

                        // Set the coordinates of this atom
                        Atom O = newSideChain.getAtom("O");
                        O_atom.setCoords(O.getCoords());
                    }
                }
            }
            catch (StructureException e)
            {
                logger.error("Error", e);
            }
        }
    }

    /**
     * Removes any groups we can't gap right
     */
    /*
    private void removeAllUnFilledGaps()
    {
        List<Gap> gaps = findGaps();
        LoopDB db = LoopDB.getInstance();

        List<Integer> groupsToRemove = new ArrayList<Integer>();

        for (Gap gap : gaps)
        {
            if (gap.getType() == Gap.TYPE.DELETION)
            {
                // don't care about deletions
                continue;
            }

            // Remove any gaps that we couldn't find loops for, and any gaps that are terminus insertions
            List<Loop> loops = db.getLoopsBySize(gap.getLoopCoilLength());
            if (loops.size() == 0 || gap.isEnd() || gap.isBeginning())
            {
                for (int i = gap.getStart(); i < gap.getStart() + gap.getLength(); i++)
                {
                    groupsToRemove.add(i);
                }
            }
        }

        // remove the groups
        ArrayList<Group> groups = newStructure.getChain(0).getGroups();
        for (int i = groupsToRemove.size() - 1; i >= 0; i--)
        {
            // for some reason, sometimes it's one off
            int removeIndex = groupsToRemove.get(i);
            if (removeIndex < groups.size())
            {
                groups.remove(removeIndex);
            }
        }
    }
    */
    
    
    /**
     * Removes empty residues from the beginning and end of newStructure.
     */
    private void removeBeginningAndEndGaps() {
    	
        Chain chain = newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();
    	
        List<Integer> groupsToRemove = new ArrayList<Integer>();
        
        // Find empty residues at the beginning.
        boolean inBeginningGap = true;
        for (int i = 0; i < groups.size() && inBeginningGap; i++) {
        	
        	if (groups.get(i) instanceof EmptyResidue) {
        		groupsToRemove.add(i);
        	}
        	else {
        		inBeginningGap = false;
        	}
        }
        
        // Find empty residues at the end.
        boolean inEndGap = true;
        for (int i = groups.size() - 1; i >= 0 && inEndGap; i--) {
        	
        	if (groups.get(i) instanceof EmptyResidue) {
        		groupsToRemove.add(i);
        	}
        	else {
        		inEndGap = false;
        	}
        }
        
        Collections.sort(groupsToRemove);
        
        // Remove the residues.
        for (int i = groupsToRemove.size() - 1; i >= 0; i--)
        {
            int removeIndex = groupsToRemove.get(i);
            if (removeIndex < groups.size())
            {
                groups.remove(removeIndex);
            }
        }
    }
    
    
    /**
     * Removes empty residues from the interior of newStructure.
     */
    private void removeInteriorGaps() {
    	
        Chain chain = newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();
        
        int firstRealResIndex = -1;
        boolean inBeginningGap = true;
        for (int i = 0; i < groups.size() && inBeginningGap; i++) {
        	
        	if ( !(groups.get(i) instanceof EmptyResidue) ) {
        		firstRealResIndex = i;
        		inBeginningGap = false;
        	}
        }
        
        if (firstRealResIndex == -1) {
        	throw new IllegalStateException("No non-empty residues in structure.");
        }
        
        boolean inEndGap = true;
        for (int i = groups.size() - 1; i >= firstRealResIndex; i--) {
        	
        	if (!inEndGap && groups.get(i) instanceof EmptyResidue) {
        		groups.remove(i);
        	}
        	else {
        		inEndGap = false;
        	}
    	}
    	
    }
    
    
    /**
     * Verifies the numbering scheme on the atoms and checks the atoms in each side chain
     * against a known side chain to avoid any missing atoms
     */
    private void cleanupChain()
    {
        Chain chain = newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();

        SideDB db = SideDB.getInstance();

        int atomNum = 1;
        for (int i = 0; i < groups.size(); i++)
        {
            Group g = groups.get(i);
            g.getPDBCode();

            // get an example of a side chain from our side chain database
            Character aminoType = ((AminoAcid) g).getAminoType();
            List<SideChain> sideChains = db.getSideChains(aminoType);

            // check to see that all atoms are there that are in the example side chain
            if (sideChains.size() > 0)
            {
                // go through each example atom and see if it's also in our residue
                SideChain example = sideChains.get(0);
                List<Atom> atoms = example.getAtoms();

                // see if the sizes are the same
                boolean allAtomsPresent = atoms.size() == g.getAtoms().size();

                if (allAtomsPresent)
                {
                    // check each atom
                    for (Atom a : atoms)
                    {
                        Atom atom = null;
                        try
                        {
                            atom = g.getAtom(a.getName());
                        }
                        catch (StructureException e)
                        {
                            e.printStackTrace();
                        }

                        // crap it's not there, we're missing it
                        if (atom == null)
                        {
                            logger.warn("Residue is missing atom " + a.getName());

                            allAtomsPresent = false;
                            break;
                        }
                    }
                }

                if (!allAtomsPresent)
                {
                    // Somehow this group is missing sidechain atoms
                    // Re-insert this side chain
                    Angle angle = chiAngles.get(i);
                    double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                    // get a suitable side chain
                    SideChain newSideChain = HomodellerUtils.getSideChain(aminoType, chi);
                    // modify its coordinates to our protein
                    superImposeResidue(g, newSideChain);
                    // set the new residue in our chain
                    groups.set(i, (Group) newSideChain);
                    // g needs to be set to this group to be renumbered
                    g = groups.get(i);
                }
            }
            else if (aminoType != 'G')
            {
                logger.warn("Could not get example side chain for residue " + aminoType);
            }

            List<Atom> atoms = g.getAtoms();
            for (Atom a : atoms)
            {
                a.setPDBserial(atomNum++);
            }

            g.setPDBCode(String.valueOf(i + 1)); // residue numbering needs to be 1-based
        }
    }

    /**
     * Find gaps and insert loops into them
     */
    public void insertLoops()
    {
        List<Gap> gaps = findGaps();

        // If there are deletions, then move any fragments that are at the end of the chain
        // to remove the deletions.
        gaps = removeEndGaps(gaps);
        
        // debug
        System.err.println("Before merging gaps:");
        System.err.println("query: " + this.querySeq.seqString());
        System.err.println("datab: " + this.databaseSeq.seqString());
        
        // If there are gaps that are too close to each other, either collapse or merge them.
        gaps = mergeGaps(gaps);
        
        // debug
        System.err.println("After merging gaps:");
        System.err.println("query: " + this.querySeq.seqString());
        System.err.println("datab: " + this.databaseSeq.seqString());

        for (Gap gap : gaps)
        {
            logger.debug("Gap: " + gap);

            try
            {
                if (!gap.isBeginning() && !gap.isEnd())
                {
                    /* If this is a middle loop (not a beginning or an ending), then fill it.
                       If it is a end loop, and it's a deletion, we don't care about it. */
                    try
                    {
                        fillGap(gap);
                    }
                    catch (HomodellerException e)
                    {
                        logger.error("Could not fill loop at " + gap.getStart() + "-" + gap.getEnd());

                        // remove gaps we can't fill
                        // removeGroupsInGap(gap);
                        // Do that later
                    }

                    if (debug)
                    {
                        cleanupChain();
                        writeChainToFile("final.pdb", newStructure.getChain(0));
                    }
                }
                else
                {
                    // don't care about it
                }
            }
            catch (StructureException e)
            {
                e.printStackTrace();
            }
        }
    }
    
    /**
     * Modifies the alignment near the beginning and end of the query to shift any gaps
     * to the ends.
     * 
     * e.g.
     * 
     *    query:    MDRVLSRADKER...
     *    database: ~~SHM-REESLQ...
     *
     * becomes
     * 
     *    query:    MDRVLSRADKER...
     *    database: ~~~SHMREESLQ...
     *  
     */
    /*
    private void removeAlignmentGapsAtEnds() {
    	
    	for () {
    		
    		
    		
    	}
    	
    }
    */
    
    
    /**
     * If there are deletions, then move any fragments that are at the end of the chain
     * to remove the deletions.
     *
     * @param gaps
     * @return
     */
    @SuppressWarnings({"unchecked"})
    private List<Gap> removeEndGaps(List<Gap> gaps)
    {
        Chain chain = newStructure.getChain(0);

        // count backwards to do the end gaps first
        for (int i = gaps.size() - 1; i >= 0; i--)
        {
            Gap gap = gaps.get(i);

            if (gap.getType() == Gap.TYPE.DELETION && !gap.isBeginning() && !gap.isEnd())
            {
                int start = gap.getPreOnQuery();
                int end = gap.getPostOnQuery();
                int length = querySeq.getSourceSymbolList().length();

                //if the end is within 20% or 20 residues of the start of the sequence, then move this gap
                int cutoffStart = (int) Math.max(20, (int) length * 0.20);
                int cutoffEnd = length - cutoffStart;

                // if the last residue in the gap is smaller than the cutoff start...
                Chain fragment = null;
                Atom N_atom = null;
                Atom C_atom = null;
                Atom O_atom = null;
                Group referenceGroup = null;

                if (end <= cutoffStart)
                {
                    if (debug)
                    {
                        writeChainToFile("before_transposed.pdb", chain);
                    }

                    // Get the fragment
                    fragment = new ChainImpl();
                    int j = Math.abs(querySeq.viewToSource(gap.getStart())) - 1;
                    AminoAcid group;
                    // start at the gap start and go backwards to get the fragment
                    while (j >= 0 && !(group = (AminoAcid) chain.getGroup(j)).getAminoType().equals('-'))
                    {
                        fragment.getGroups().add(0, group);
                        j--;
                    }

                    try
                    {
                        N_atom = fragment.getGroup(fragment.getGroups().size() - 1).getAtom("N");
                        referenceGroup = chain.getGroup(gap.getPreOnQuery() + 1); // the next group
                        C_atom = referenceGroup.getAtom("C");
                        O_atom = referenceGroup.getAtom("O");
                    }
                    catch (StructureException e)
                    {
                        e.printStackTrace();
                    }
                }

                // similarly with the end
                if (start >= cutoffEnd)
                {
                    if (debug)
                    {
                        writeChainToFile("before_transposed.pdb", chain);
                    }

                    // Get the fragment
                    fragment = new ChainImpl();
                    int j = Math.abs(querySeq.viewToSource(gap.getEnd())) - 1;
                    AminoAcid group;
                    while ( j < chain.getLength()
                            && !(group = (AminoAcid) chain.getGroup(j)).getAminoType().equals('-'))
                    {
                        fragment.addGroup(group);
                        j++;
                    }

                    try
                    {
                        N_atom = fragment.getGroup(0).getAtom("N");
                        referenceGroup = chain.getGroup(start);
                        C_atom = referenceGroup.getAtom("C");
                        O_atom = referenceGroup.getAtom("O");
                    }
                    catch (StructureException e)
                    {
                        e.printStackTrace();
                    }

                }

                if (fragment != null)
                {
                    assert N_atom != null && C_atom != null : "N or C atom is null";

                    // distance away from C to move the fragment
                    final double N_C_DISTANCE = 1.45;

                    // Move the entire group so the N-terminal of the fragment matches up with the O of the C-terminal
                    double[] C_coords = C_atom.getCoords();
                    double[] N_coords = N_atom.getCoords();
                    double[] transpose = new double[3];

                    // distance in space, d^2 = offset^2 + offset^2 + offset^2 = 3* offset ^ 2
                    // d^2 = 3* offset ^ 2
                    // d = sqrt(3) * offset
                    // offset = d / sqrt(3)
                    
                    double offset = N_C_DISTANCE / Math.sqrt(3);

                    // build transposition matrix
                    for (int k = 0; k < N_coords.length; k++)
                    {
                        transpose[k] = C_coords[k] - N_coords[k] - offset;
                    }

                    // do transposition
                    for (Group g : (List<Group>) fragment.getGroups())
                    {
                        for (Atom a : (List<Atom>) g.getAtoms())
                        {
                            double[] coords = a.getCoords();
                            Superpose.translate(coords, transpose);
                        }
                    }

                    if (debug)
                    {
                        writeChainToFile("transposed.pdb", chain);
                    }

                    //TODO remove
                    // check for bumps on moved fragment, between the N and O atom
                    double N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);
                    if (N_O_dist < 1.5)
                    {
                        // too close, rotate C=O carboxyl to move away from N
                        Atom CA_atom = null;
                        try
                        {
                            CA_atom = referenceGroup.getAtom("CA");
                        }
                        catch (StructureException e)
                        {
                            e.printStackTrace();
                        }

                        // this is distance we want C=O to have
                        double C_O_dist = HomodellerUtils.getDistance(C_atom, O_atom);

                        assert CA_atom != null : "CA Atom is null";

                        double[] CA_coords = CA_atom.getCoords();

                        // find the point between the CA and N atom
                        double[] point = new double[3];
                        point[0] = (CA_coords[0] + N_coords[0]) / (double) 2;
                        point[1] = (CA_coords[1] + N_coords[1]) / (double) 2;
                        point[2] = (CA_coords[2] + N_coords[2]) / (double) 2;

                        // find the line between (point) and C, which we will extrapolate to be
                        // the position of the O
                        double[] line = new double[3];
                        line[0] = C_coords[0] - point[0];
                        line[1] = C_coords[1] - point[1];
                        line[2] = C_coords[2] - point[2];

                        // change line's magnitude to the magnitude of the distance between C=O
                        // i.e. |line| = C_O_dist
                        double multiplier = C_O_dist / Superpose.normalize(line);
                        line[0] *= multiplier;
                        line[1] *= multiplier;
                        line[2] *= multiplier;

                        // these are O's new coordinates
                        double O_coords[] = O_atom.getCoords();
                        O_coords[0] = C_coords[0] + line[0];
                        O_coords[1] = C_coords[1] + line[1];
                        O_coords[2] = C_coords[2] + line[2];
                    }

                    if (debug)
                    {
                        writeChainToFile("transposed_fixed.pdb", chain);
                    }


                    gaps.remove(i);
                }
            }
        }

        return gaps;
    }
    
    
    /**
     * When gaps are too close together, it makes sense to merge them
     *
     * @param gaps The list of gaps to merge
     * @return the merged list of gaps
     */
    /*
    public List<Gap> mergeGaps(List<Gap> gaps)
    {
        for (int i = 0; i < gaps.size() - 1; i++)
        {
            Gap gap = gaps.get(i);
            Gap next = gaps.get(i + 1);

            // if the aligned sequence between them is less than 2 residues, we might as well merge the gaps
            if (next.getAdjustedStartOnQuery() - gap.getAdjustedEndOnQuery() < 2 && gap.getType() == next.getType())
            {
                //if ( gap.isBeginning() && gap.getType() == Gap.TYPE.INSERTION )
                //{
                //    gap.setLength(next.getStart() - gap.getStart());
                //}
                //else
                //{
                gap.setLength(next.getStart() + next.getLength() - gap.getStart());
                //}

                logger.debug("Merging gap (" + gap + ") and (" + next + "): " + gap);

                // remove it
                gaps.remove(next);

                // call this function again on the reduced set of gaps
                return mergeGaps(gaps);
            }
        }

        return gaps;
    }
    */
    
    
    /**
     * For any adjacent pair of gaps of different types (i.e. one is an insertion and the other is
     * a deletion), if the Pre residue of the second gap is to the left of the Post residue of the
     * first gap, "collapse" the gaps by editing the alignment between the query and database
     * sequences. e.g. we have a situation like
	 *
	 *                 gap.getPostOnAlignment()
	 *                  |
	 *    query: MVQ---LRSHMTWHQ
	 *    datab: MVQGAKL---MTWHQ
	 *                |
	 *               next.getPreOnAlignment()
	 *
	 * and we want to alter the alignment so it becomes:
	 *
	 *               new gap.getPostOnAlignment()
	 *                |
	 *    query: MVQ-LRSHMTWHQ
	 *    datab: MVQGAKL-MTWHQ
	 *                |
	 *               new next.getPreOnAlignment()
	 *
	 * Notice that this dislocates the alignment between the two L residues. One or both of
	 * the gaps may disappear. The coordinates of many of the gaps in the gap list will also
	 * need to be adjusted.
     * 
     * This method should be called before calling this.mergeGaps().
     */
    /*
    public List<Gap> collapseGaps(List<Gap> gaps)
    {
    	
        for (int i = 0; i < gaps.size() - 1; i++)
        {
            Gap gap = gaps.get(i);
            Gap next = gaps.get(i + 1);
            
            // If the Pre residue of the next gap is to the left of the Post residue of the current gap,
            // alter the alignment between the two gaps (if they are of different types).
            if (next.getPreOnAlignment() - gap.getPostOnAlignment() < 0)
            {
            	if (gap.getType() != next.getType()) {
            		// In choosing shiftAmount here, note that we limit it to the length of the smallest gap.
            		// So in the case e.g.
            		//
            		//                   gap.getPostOnAlignment()
            		//                    |
            		//    query: MVQ---L-RSHMTWHQ
            		//    datab: MVQGAKLQ---MTWHQ
            		//                 |
            		//                next.getPreOnAlignment()
            		//
            		// shiftAmount would be 1 and we would end up with
            		//
            		//                 far left gap's .getPostOnAlignment()
            		//                  |
            		//    query: MVQ---LRSHMTWHQ
            		//    datab: MVQGAKLQ--MTWHQ
            		//                 |
            		//                new next.getPreOnAlignment()
            		//
            		// Notice that the two remaining gaps are still too close together. This will be resolved
            		// when collapseGaps() is called recursively.
            		
            		int shiftAmount; // Number of residues by which to shift the alignment
            		
            		int shortestGapLength = Math.min(gap.getLength(), next.getLength());
            		shiftAmount = Math.min(shortestGapLength, gap.getPostOnAlignment() - next.getPreOnAlignment());
            		
            		// Remove shiftAmount residue positions in each gap. This is done by altering this.querySeq and
            		// this.databaseSeq.
            		
            		List<Gap> gapPair = new ArrayList<Gap>();
            		gapPair.add(gap);
            		gapPair.add(next);
            		
            		for (int j= 0; j < gapPair.size(); j++) {
            			
            			Gap currentGap = gapPair.get(j);
            			
            			GappedSequence sequenceToAlter;
        				if (currentGap.getType() == Gap.TYPE.INSERTION) {
            				sequenceToAlter = this.databaseSeq;
            			}
            			else if (currentGap.getType() == Gap.TYPE.DELETION) {
            				sequenceToAlter = this.querySeq;
        				}
            			else {
            				assert false; // Should never happen.
            			}
            			
        				sequenceToAlter.removeGaps(currentGap.getStart() + 1, shiftAmount);
        				
        				
        				// Change the lengths of the gaps in our gap pair, and remove any of them that have been closed.
        				
        				if (currentGap.getLength() == shiftAmount) {
        					// Remove the gap from the 'gaps' list.
        					gaps.remove(currentGap);
        				}
        				else {
        					// Change the length of the gap.
        					currentGap.setLength(currentGap.getLength() - shiftAmount);
        				}
        				
            		}
            		
            		// Scan through the list of gaps and adjust their coordinates.
            		
            		int firstGapStart = gap.getStart();
            		
            		for (int j = 0; j < gaps.size(); j++) {
            			
            			Gap currentGap = gaps.get(j);
            			
            			if (currentGap.getStart() > firstGapStart) {
            				// The current gap is to the right of the first gap in our gap pair. Change its
            				// starting position.
            				currentGap.setStart(currentGap.getStart() - shiftAmount);
            			}
            		}
            		
            	}

                // call this function again on the reduced set of gaps
                return collapseGaps(gaps);
            }
            
        }
        
    }
    */

    
    /**
     * For any adjacent pair of gaps of the same type (i.e. both deletions or both insertions), combine
     * them into a single gap.
     * 
     * For any adjacent pair of gaps of different types (i.e. one is an insertion and the other is
     * a deletion), if the Pre residue of the second gap is to the left of the Post residue of the
     * first gap, "collapse" the gaps by editing the alignment between the query and database
     * sequences. e.g. we have a situation like
	 *
	 *                 gap.getPostOnAlignment()
	 *                  |
	 *    query: MVQ---LRSHMTWHQ
	 *    datab: MVQGAKL---MTWHQ
	 *                |
	 *               next.getPreOnAlignment()
	 *
	 * and we want to alter the alignment so it becomes:
	 *
	 *               new gap.getPostOnAlignment()
	 *                |
	 *    query: MVQ-LRSHMTWHQ
	 *    datab: MVQGAKL-MTWHQ
	 *                |
	 *               new next.getPreOnAlignment()
	 *
	 * Notice that this dislocates the alignment between the two L residues. One or both of
	 * the gaps may disappear. The coordinates of many of the gaps in the gap list will also
	 * need to be adjusted.
     * 
     * @param gaps The list of gaps.
     * @return The altered list of gaps.
     */
    public List<Gap> mergeGaps(List<Gap> gaps)
    {
        for (int i = 0; i < gaps.size() - 1; i++)
        {
            Gap gap = gaps.get(i);
            Gap next = gaps.get(i + 1);

            // If the Pre residue of the next gap is to the left of the Post residue of the current gap,
            // either merge the gaps (if they are of the same type), or alter the alignment between the
            // two gaps (if they are of different types).
            if (next.getPreOnAlignment() - gap.getPostOnAlignment() < LoopInserter.MERGE_GAPS_THRESHOLD)
            {
            	if (gap.getType() == next.getType()) {
            		// The gaps are of the same type, so merge them.
            		
            		if ( !(gap.isBeginning() || next.isEnd()) ) {
	                    //if ( gap.isBeginning() && gap.getType() == Gap.TYPE.INSERTION )
	                    //{
	                    //    gap.setLength(next.getStart() - gap.getStart());
	                    //}
	                    //else
	                    //{
	                    gap.setLength(next.getStart() + next.getLength() - gap.getStart());
	                    //}
	
	                    logger.debug("Merging gap (" + gap + ") and (" + next + "): " + gap);
	
	                    // remove it
	                    gaps.remove(next);
            		}
            		else {
            			// Do not merge a beginning or end gap with another gap.
            			return gaps;
            		}
            	}
            	else {
            		// The gaps are of different types, so collapse them by altering the alignment.
            		
            		// In choosing shiftAmount here, note that we limit it to the length of the smallest gap.
            		// For gaps that are merged gaps consisting of two or more subgaps, we use the length of
            		// the smallest subgap among the C-terminal subgap of the left gap and the N-terminal
            		// subgap of the right gap. e.g. in the case of two merged gaps as follows,
            		//
            		//                   gap.getPostOnAlignment()
            		//                    |
            		//    query: MVQ---L-SHMVTWHQRC
            		//    datab: MVQGAKLQ----T-HQRC
            		//                 |
            		//                next.getPreOnAlignment()
            		//
            		// shiftAmount would be 1 and we would end up with
            		//
            		//                 new gap.getPostOnAlignment()
            		//                  |
            		//    query: MVQ---LSHMVTWHQRC
            		//    datab: MVQGAKLQ---T-HQRC
            		//                 |
            		//                new next.getPreOnAlignment()
            		//
            		// Notice that the two remaining gaps are still too close together. This will be resolved
            		// when collapseGaps() is called recursively.
            		//
            		// Also, note that we have closed one of the subgaps of the gap on the left. We will have to
            		// set it to be a simple gap, and reset its length. For the gap on the right, it remains a
            		// merged gap, but its N-terminal subgap has become shorter, and we will have to adjust its
            		// length.
            		
            		try {
	            		int shiftAmount; // Number of residues by which to shift the alignment.
	            		
	            		int shortestGapLength = Math.min(gap.getCterminalSubgapLength(), next.getNterminalSubgapLength());
	            		shiftAmount = Math.min(shortestGapLength, gap.getPostOnAlignment() - next.getPreOnAlignment());
	            		
	            		// Remove shiftAmount residue positions in each gap. This is done by altering this.querySeq and
	            		// this.databaseSeq.
	            		
	            		List<Gap> gapPair = new ArrayList<Gap>();
	            		gapPair.add(gap);
	            		gapPair.add(next);
	            		
	            		for (int j= 0; j < gapPair.size(); j++) {
	            			
	            			Gap currentGap = gapPair.get(j);
	            			
	            			GappedSequence sequenceToAlter;
	        				if (currentGap.getType() == Gap.TYPE.INSERTION) {
	            				sequenceToAlter = this.databaseSeq;
	            			}
	            			else if (currentGap.getType() == Gap.TYPE.DELETION) {
	            				sequenceToAlter = this.querySeq;
	        				}
	            			else {
	            				sequenceToAlter = null;
	            				assert false; // Should never happen.
	            			}
	            			
	        				//sequenceToAlter.removeGaps(currentGap.getStart() + 1, shiftAmount);
	        				
	        				if (j == 0) {
	        					// Remove gaps from the C-terminal end of the gap (in case it is a merged gap).
	        					
	        					sequenceToAlter.removeGaps(currentGap.getEnd() - shiftAmount, shiftAmount);
	        				}
	        				else { // j == 1
	        					// Remove gaps from the N-terminal end of the gap (in case it is a merged gap).
	        					
	        					sequenceToAlter.removeGaps(currentGap.getStart() + 1, shiftAmount);
	        				}
	        				
	        				
	        				// Change the lengths and starting positions of the gaps in our gap pair, and remove
	        				// any of them that have been closed.
	        				
	        				if (!currentGap.isMergedGap()) {
	            				if (currentGap.getLength() == shiftAmount) {
	            					// Remove the gap from the 'gaps' list.
	            					gaps.remove(currentGap);
	            				}
	            				else {
	            					// Change the length of the gap.
	            					currentGap.setLength(currentGap.getLength() - shiftAmount);
	            					
	            					// Change the start position of the gap. (The left gap's starting position does
	            					// not need to be changed).
	            					if (j == 1) {
	            						currentGap.setStart(currentGap.getStart() - shiftAmount);
	            					}
	            				}
	        				}
	        				else {
	        					// The gap is a merged gap.
	        					
	        					if (j == 0) {
	        						if (currentGap.getCterminalSubgapLength() == shiftAmount) {
	        							// The C-terminal subgap has closed. Recompute the gap length and determine whether
	        							// it is still a merged gap.
	        							
	        							int scanStart = currentGap.getStart();
	        							int scanEnd = currentGap.getEnd() - shiftAmount;
	        							
	        							recalculateMergedGapCoords(currentGap, sequenceToAlter, scanStart, scanEnd);
	        						}
	        						else {
	        							// Change the length of the C-terminal subgap (which will also change the total
	        							// length of the gap).
	        							currentGap.setCterminalSubgapLength(currentGap.getCterminalSubgapLength() - shiftAmount);
	        						}
	        					}
	        					else { // j == 1
	        						if (currentGap.getNterminalSubgapLength() == shiftAmount) {
	        							// The N-terminal subgap has closed. Recompute the gap length and determine whether
	        							// it is still a merged gap.
	        							
	        							int scanStart = currentGap.getStart();
	        							int scanEnd = currentGap.getEnd() - shiftAmount;
	        							
	        							recalculateMergedGapCoords(currentGap, sequenceToAlter, scanStart, scanEnd);
	        						}
	        						else {
	        							// Change the length of the N-terminal subgap (which will also change the total
	        							// length of the gap).
	        							currentGap.setNterminalSubgapLength(currentGap.getNterminalSubgapLength() - shiftAmount);
	        							
	                					// Change the start position of the gap. (The left gap's starting position does
	                					// not need to be changed).
	                					currentGap.setStart(currentGap.getStart() - shiftAmount);
	        						}
	        					}
	        					
	        				}
	        				
	            		}
	            		
	            		// Scan through the list of gaps and adjust their start positions. The start positions of the gaps
	            		// in our gap pair have already been adjusted above, so skip these gaps.
	            		
	            		int secondGapStart = next.getStart();
	            		
	            		for (int j = 0; j < gaps.size(); j++) {
	            			
	            			Gap currentGap = gaps.get(j);
	            			
	            			if (currentGap.getStart() > secondGapStart) {
	            				// The current gap is to the right of the gaps in our gap pair. Change its
	            				// starting position.
	            				currentGap.setStart(currentGap.getStart() - shiftAmount);
	            			}
	            		}
	            		
	            	}
            		catch (IllegalSymbolException e) {
            			e.printStackTrace();
            			System.exit(1);
            		}
            	}
            	
                // call this function again on the reduced set of gaps
                return mergeGaps(gaps);
            }
        }

        return gaps;
    }
    
    
    /**
     * Recalculates and resets the coordinates of the given Gap that was previously a merged gap,
     * and which now may or may not be a merged gap. This method is intended to be called after the
     * underlying sequence has been altered, and the gap coordinates need to be adjusted. The
     * parameters scanStart and scanEnd are the bounds in which the gap is expected to lie.
     * 
     * @param gap The Gap that needs its coordinates recalculated.
     * @param sequenceWithGaps The underlying sequence containing the gap(s) that the given gap includes.
     * @param scanStart The coordinate (in biological coordinates starting from 1) of a non-gap
     *                  character in sequenceWithGaps left of the first gap character in the gap.
     * @param scanEnd   The coordinate (in biological coordinates starting from 1) of a non-gap
     *                  character in sequenceWithGaps right of the last gap character in the gap.
     */
    private void recalculateMergedGapCoords(Gap gap, GappedSequence sequenceWithGaps, int scanStart, int scanEnd) {
    	
		int newGapStart = -1; // In biological coordinates starting from 1.
		int newGapEnd = -1; // In biological coordinates starting from 1.
		int newGapLength;
		boolean prevCharWasGap = false;
		int numberOfGapChars = 0;
		int NterminalSubgapLength = 0;
		int CterminalSubgapLength = 0;
		int mostRecentSubgapLength = 0;
		boolean inNterminalGap = false;
		String sequenceString = sequenceWithGaps.seqString();
		
		for (int k = scanStart; k <= scanEnd; k++) {
			char c = sequenceString.charAt(k - 1);
				// - 1 to convert from biological coordinates (starting at 1) to coordinates starting at 0.
			
			if (ProteusTools.isGap(c)) {
				if (newGapStart == -1) {
					newGapStart = k - 1;
						// In biological coordinates starting from 1, this refers to the residue immediately
						// preceding the first gap character.
					inNterminalGap = true;
				}
				
				if (!prevCharWasGap) {
					mostRecentSubgapLength = 1;
				}
				else {
					mostRecentSubgapLength++;
				}
				
				if (inNterminalGap) {
					NterminalSubgapLength++;
				}
				numberOfGapChars++;
				
				prevCharWasGap = true;
			}
			else {
				// Current char is not a gap char.
				
				if (prevCharWasGap) {
					newGapEnd = k;
						// In biological coordinates starting from 1, this refers to the residue immediately
						// following the last observed gap character.
				}
				
				inNterminalGap = false;
				prevCharWasGap = false;
			}
		}
		
		CterminalSubgapLength = mostRecentSubgapLength;
		
		if (newGapStart == -1) {
			throw new IllegalStateException("No gap characters found where expected.");
		}
		else if (newGapEnd == -1) {
			throw new IllegalStateException("No gap end found where expected.");
		}
		
		newGapLength = newGapEnd - newGapStart - 1;
		
		if (newGapLength == numberOfGapChars) {
			// The gap is no longer a merged gap.
			gap.setAsSimpleGap();
			gap.setStart(newGapStart);
			gap.setLength(newGapLength);
		}
		else {
			// The gap is still a merged gap.
			gap.setAsMergedGap(NterminalSubgapLength, CterminalSubgapLength);
			gap.setStart(newGapStart);
			gap.setLength(newGapLength);
		}
    	
    }
    
    
    
    /**
     * Fills a single gap
     *
     * @param gap the gap to fill
     * @throws org.biojava.bio.structure.StructureException
     *                             If anything goes wrong with the structure
     * @throws HomodellerException If an optimal loop structure is not selected
     */
    private void fillGap(Gap gap) throws StructureException, HomodellerException
    {
        Group preResidue = null;
        Group postResidue = null;
        
        LoopDB db = LoopDB.getInstance();
        List<Loop> loops = db.getLoopsBySize(gap.getLoopCoilLength());
        	// The "size" of the loops here is the number of coil residues in the loop, which should be
        	// equal to the number of residues we want to have between the pre and post residues.
        	// The pre and post residues are the two fixed residues spanning the section where we
        	// want to insert the loop of size gap.getLoopCoilLength(). Below we will call getCoilLoop()
        	// on the loop to get the coil residues plus the two flanking non-coil residues. These two
        	// flanking residues will correspond to the pre and post residues.
        
        this.debugLoopNum++;
        
        System.err.println();
        System.err.println("**************************");
        System.err.println("Inserting loop " + this.debugLoopNum + " (size " + gap.getLoopCoilLength() + ")"); // debug
        if (gap.isBeginning()) {
        	System.err.println("This is a beginning gap");
        }
        else if (gap.isEnd()) {
        	System.err.println("This is an end gap");
        }
        System.err.println("**************************");
        System.err.println();
        
        
        if (!gap.isBeginning())
        {
            // Because the chi angles calculation is based on the ungapped template (database),
            // convert the index to the index in the gapped template
        	
        	// debug
        	System.err.println("On alignment:");
        	int preOnAlignment = gap.getPreOnAlignment();
            System.err.print  ("Pre:      ");
            for (int i = 0; i < preOnAlignment - 1; i++) {
            	System.err.print(" ");
            }
            System.err.println("v = " + preOnAlignment);
        	
        	System.err.println("query:    " + querySeq.seqString());
        	System.err.println("database: " + databaseSeq.seqString());
        	System.err.println();
        	
        	System.err.println("On newStructure:");
        	int preOnQuery = gap.getPreOnQuery();
            System.err.print  ("Pre:      ");
            for (int i = 0; i < preOnQuery - 1; i++) {
            	System.err.print(" ");
            }
            System.err.println("v = " + preOnQuery);
        	
        	System.err.println("newStruc: " + newStructure.getChain(0).getSequence());
        	System.err.println();
        	
            preResidue = newStructure.getChain(0).getGroup(gap.getPreOnQuery() - 1);
            	// - 1 because getPreOnQuery() returns biological coordinates starting at 1
            	// whereas getGroup() appears to use coordinates starting from 0.

            //Atom preN_Atom = preResidue.getAtom("C");
        }

        if (!gap.isEnd())
        {
            // Because the chi angles calculation is based on the ungapped template (database),
            // convert the index to the index in the gapped template
        	
        	// debug
        	System.err.println("On alignment:");
        	int postOnAlignment = gap.getPostOnAlignment();
            System.err.print  ("Post:     ");
            for (int i = 0; i < postOnAlignment - 1; i++) {
            	System.err.print(" ");
            }
            System.err.println("v = " + postOnAlignment);
        	
        	System.err.println("query:    " + querySeq.seqString());
        	System.err.println("database: " + databaseSeq.seqString());
        	System.err.println();
        	
        	System.err.println("On newStructure:");
        	int postOnQuery = gap.getPostOnQuery();
            System.err.print  ("Post:     ");
            for (int i = 0; i < postOnQuery - 1; i++) {
            	System.err.print(" ");
            }
            System.err.println("v = " + postOnQuery);
        	
        	System.err.println("newStruc: " + newStructure.getChain(0).getSequence());
        	System.err.println();
        	
            postResidue = newStructure.getChain(0).getGroup(gap.getPostOnQuery() - 1);
        		// - 1 because getPostOnQuery() returns biological coordinates starting at 1
        		// whereas getGroup() appears to use coordinates starting from 0.

            //Atom postC_Atom = postResidue.getAtom("N");
        }
        
        
        Loop bestLoop = getBestLoopForCCD(db, loops, preResidue, postResidue, gap);
        
        
        // Superimpose loop. I'm not sure if this is necessary after the superpositioning in
        // getBestLoopForCCD() -- David Arndt.
        Loop coilLoop = bestLoop.getCoilLoop();
        if (preResidue != null)
        {
            superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
        }
        else
        {
            // pre residue will only be null if it's at the beginning
            assert gap.isBeginning() : "Pre residue is null, not a loop at beginning";

            superImposeLoop(postResidue, coilLoop.getGroup(coilLoop.getGroups().size() - 1), coilLoop);
        }
        
        
        
        //if (debug)
        //{
            writeChainToFile("loop_CCD_" + this.debugLoopNum + ".pdb", bestLoop.getCoilLoop());
        //}

        /*
        Chain rotatedLoop;

        if (!gap.isEnd() && !gap.isBeginning()
                && !(postResidue instanceof EmptyResidue)
                && !(preResidue instanceof EmptyResidue))
        {
            //rotatedLoop = insertLoopInGap(preResidue, postResidue, bestLoop.getCoilLoop());
        	rotatedLoop = coilLoop;
        }
        else
        {
            // should already be rotated correctly
            rotatedLoop = coilLoop;
        }

        //if (debug)
        //{

            writeChainToFile("rotatedLoop.pdb", rotatedLoop);
        //}
        */
        

        //Chain finalLoop = rotatedLoop;
        
        // debug
        System.err.println("\n   Writing newStructure_" + this.debugLoopNum + ".pdb before minimizeDistances()...");
        writeChainToFile("newStructure_" + this.debugLoopNum + ".pdb", newStructure.getChain(0));
        
        if (!gap.isEnd() && !gap.isBeginning())
        {
            // If it's not a beginning or end gap, then we need to minimize the distances between both ends' N-C atoms
        	
        	
        	/************************************************************************/
        	/***               Attempt loop insertion with PepMake                ***/
        	/************************************************************************/
        	
        	System.err.println("Attempting loop insertion with PepMake...");
        	
        	boolean pepmakeSucceeded = false;
        	
        	// Use PepMake to close the loop using Cyclic Coordinate Descent (CCD):
        	
			/*
			Create a structure suitable for input into PepMake.
			*/
			ArrayList<Group> coilLoopGroups = coilLoop.getGroups();
			//Structure pepmakeInputStructure = (Structure) newStructure.clone(); // results in NullPointerException
			Structure pepmakeInputStructure = new StructureImpl();
			
			Chain pepmakeInputChain = new ChainImpl();
			pepmakeInputStructure.addChain(pepmakeInputChain);
			
			ArrayList<Group> groups = newStructure.getChain(0).getGroups();
			for (int i = 0; i < groups.size(); i++) {
				pepmakeInputChain.addGroup(groups.get(i));
			}
			
			
			/*
			// Replace the groups list in the PepMake input structure with a new groups list.
			// This will allow us to modify the groups list in the PepMake input structure
			// without modifying the groups in newStructure.
			ArrayList<Group> pepmakeGroups = pepmakeInputStructure.getChain(0).getGroups();
			//assert(pepmakeGroups == newStructure.getChain(0).getGroups()); // test
			//pepmakeGroups = (ArrayList<Group>) pepmakeGroups.clone();
			//pepmakeInputStructure.getChain(0).setGroups(pepmakeGroups);
			assert(pepmakeGroups != newStructure.getChain(0).getGroups()); // test
			*/
			
			
			// Modify the groups in the PepMake input structure by substituting the coil
			// loop groups. 
			
			ArrayList<Group> pepmakeGroups = pepmakeInputChain.getGroups();
			
			// Modify the groups in the PepMake input structure by substituting the coil
			// loop groups. Coil loop includes the flanking residues, so start at 1 and
			// end at size -1. pepmakeGroups contains empty residues where there are gaps,
			// so we don't need to worry about adding or removing different numbers of
			// residues.
			for (int i = 1; i < coilLoopGroups.size() - 1; i++)
			{
				pepmakeGroups.set(gap.getPreOnQuery() + i - 1, coilLoopGroups.get(i));
					// -1 to convert from biological coordinates starting from 1 (from getPreOnQuery()) to
					// cooordinates starting from 0.
			}
			
			
			/*
			int startIndex = gap.getPreOnQuery() + 1 - 1;
				// Index of the first residue to remove, in coordinates starting from 0.
				// + 1 and  because we do not want to remove the Pre or Post residues themselves.
				// - 1 to convert from biological coordinates starting from 1 to coordinates
				// starting from 0.
			int endIndex = gap.getPostOnQuery() - 1 - 1;
				// Index of the last residue to remove, in coordinates starting from 0.
				// - 1 and  because we do not want to remove the Pre or Post residues themselves.
				// - 1 to convert from biological coordinates starting from 1 to coordinates
				// starting from 0.
			int C_indexShift = 0;
				// We will be removing and adding different numbers of residues just below, so we
				// need to keep track of how many so we can give PepMake the right C-anchor index.
			
			// Remove the groups that we want to replace.
			for (int i = startIndex; i <= endIndex; i++) {
				Group removed = pepmakeGroups.remove(startIndex);
					// Not dependent on i because removing elements from pepmakeGroups shifts indices.
				if (removed != null) {
					C_indexShift--;
				}
				System.err.println("removed " + removed.getPDBName());
			}
			*/
			
			// Add the groups from the coil loop (in reverse order). Coil loop includes the flanking residues, 
			// so start at size - 2, end at 1.
			/*
			for (int i = coilLoopGroups.size() - 2; i >= 1; i--) {
				pepmakeGroups.add(startIndex, coilLoopGroups.get(i));
				C_indexShift++;
				`System.err.println("add");
			}
			*/
			
			// Renumber the residues in the structure given to pepmake, starting from 1. This
			// is only for debugging purposes to prevent display glitches in PyMol when
			// examining the pepmake input and output structures.
			for (int i = 0; i < pepmakeGroups.size(); i++) {
				Group currentGroup = pepmakeGroups.get(i);
				currentGroup.setPDBCode(String.valueOf(i + 1));
			}
        	
	        // debug
	        System.err.println("   Writing pepmakeInput_" + this.debugLoopNum + ".pdb...");
	        writeChainToFile("pepmakeInput_" + this.debugLoopNum + ".pdb", pepmakeInputStructure.getChain(0));
        	
	        // Calculate offset for N- and C-anchors due to presence of empty residues.
	        int anchorOffset = 0;
	        for (int i = 0; i < gap.getPreOnQuery() + 1; i++) {
	        	if (pepmakeGroups.get(i) instanceof EmptyResidue) {
	        		anchorOffset--;
	        	}
	        }
	        
            PepmakeRunner.Options options = new PepmakeRunner.Options();
            options.closeGap = true;
            //options.res_id_N = Integer.parseInt(preResidue.getPDBCode()); // Optional.
            //options.res_id_C = Integer.parseInt(postResidue.getPDBCode());
            options.res_index_N = gap.getPreOnQuery() - 1 + anchorOffset;
            options.res_index_C = gap.getPostOnQuery() - 1 + anchorOffset;
            	// -1 to convert from biological coordinates starting from 1 to coordinates starting
            	// from 0.
            //options.gap_threshold = 0.0001;
            
            
            PepmakeRunner pepmakeRunner = new PepmakeRunner(pepmakeInputStructure, options);
            
            Structure pepmakeOutputStructure = pepmakeRunner.runPepmake();
            
            if (pepmakeOutputStructure != null) {
            	
                // The PepMake input structure may have had empty residues to the right of the current gap.
                // But they were not output to PepMake and so were not read back in. So here we add the
                // empty residues to the PepMake output structure.
            	ArrayList<Group> pepmakeOutputGroups = pepmakeOutputStructure.getChain(0).getGroups();
                for (int i = 0; i < pepmakeGroups.size(); i++) {
                	// pepmakeGroups has the groups from just before being output to PepMake.
                	
                	if (pepmakeGroups.get(i) instanceof EmptyResidue) {
                		pepmakeOutputGroups.add(i, pepmakeGroups.get(i));
                	}
                }
            	
            	// Set newStructure to be the structure read in from PepMake.
            	newStructure = pepmakeOutputStructure;
            	pepmakeSucceeded = true;
            	
    	        // debug
            	System.err.println("Loop insertion with PepMake succeeded!\n");
    	        System.err.println("   Writing pepmakeOutput_" + this.debugLoopNum + ".pdb...");
    	        writeChainToFile("pepmakeOutput_" + this.debugLoopNum + ".pdb", pepmakeOutputStructure.getChain(0));
            }
            else {
            	System.err.println("Loop insertion with PepMake FAILED.\n");
            }
        	
        	
        	if (!pepmakeSucceeded) {
        		
            	/************************************************************************/
            	/***                     Insert loop with Simplex                     ***/
            	/************************************************************************/
        		
        		// PepMake did not succeed in closing the loop. Fall back on Plan B, which is running
        		// Simplex to close the loop.
        		//
        		// This is the old method that was partly responsible for scrunched loops (poor selection
        		// of Pre and Post residues also contributed), but at least it always generates a
        		// structure.
        		
            	System.err.println("Inserting loop with Simplex...");
        		
                bestLoop = getBestLoopForSimplex(db, loops, preResidue, postResidue, gap);
            	
                // Superimpose loop.
                coilLoop = bestLoop.getCoilLoop();
                if (preResidue != null)
                {
                    superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
                }
                else
                {
                    // pre residue will only be null if it's at the beginning
                    assert gap.isBeginning() : "Pre residue is null, not a loop at beginning";

                    superImposeLoop(postResidue, coilLoop.getGroup(coilLoop.getGroups().size() - 1), coilLoop);
                }
                
                
                
                //if (debug)
                //{
                    writeChainToFile("loop_simplex_" + this.debugLoopNum + ".pdb", bestLoop.getCoilLoop());
                //}
                
                
                Chain rotatedLoop;

                if (!gap.isEnd() && !gap.isBeginning()
                        && !(postResidue instanceof EmptyResidue)
                        && !(preResidue instanceof EmptyResidue))
                {
                    rotatedLoop = insertLoopInGap(preResidue, postResidue, bestLoop.getCoilLoop());
                }
                else
                {
                    // should already be rotated correctly
                    rotatedLoop = coilLoop;
                }

                //if (debug)
                //{

                    writeChainToFile("rotatedLoop_" + this.debugLoopNum + ".pdb", rotatedLoop);
                //}
                
                Chain finalLoop = rotatedLoop;
        		
        		finalLoop = minimizeDistances(preResidue, postResidue, rotatedLoop);
        		
				/*
				Final step : insert loop residues into the structure
				Final loop includes the flanking residues, so start at 1, end at size - 1
				*/
				ArrayList<Group> loopGroups = finalLoop.getGroups();
				//ArrayList<Group> groups = newStructure.getChain(0).getGroups();
				for (int i = 1; i < loopGroups.size() - 1; i++)
				{
					groups.set(gap.getPreOnQuery() + i - 1, loopGroups.get(i));
						// -1 to convert from biological coordinates starting from 1 (from getPreOnQuery()) to
						// coordinates starting from 0.
				}
        	}
        	
        }
        else {
        	// 
        	
			/*
			Final step : insert loop residues into the structure
			Final loop includes the flanking residues, so start at 1, end at size - 1
			*/
        	ArrayList<Group> groups = newStructure.getChain(0).getGroups();
        	Chain finalLoop = bestLoop.getCoilLoop();
			ArrayList<Group> loopGroups = finalLoop.getGroups();
			//ArrayList<Group> groups = newStructure.getChain(0).getGroups();
			for (int i = 1; i < loopGroups.size() - 1; i++)
			{
				groups.set(gap.getPreOnQuery() + i - 1, loopGroups.get(i));
					// -1 to convert from biological coordinates starting from 1 (from getPreOnQuery()) to
					// coordinates starting from 0.
			}

        }

        
        // debug
        System.err.println("   Writing newStructure_min_" + this.debugLoopNum + ".pdb after minimizeDistances()...");
        writeChainToFile("newStructure_min_" + this.debugLoopNum + ".pdb", newStructure.getChain(0));
        
    }
    
    
    /**
     * Searches the loop library for the loop that is expected to work best with the Simplex algorithm.
     * 
     * @return the best matching loop
     */
    private Loop getBestLoopForSimplex(LoopDB db, List<Loop> loops, Group preResidue, Group postResidue, Gap gap)
    		throws StructureException, HomodellerException {
    	
        /**
         * preGapAngles - the angle of the residue right before the gap (i.e. the residue we're glueing the gap onto)
         * postGapAngles - the angle of the residue right after the gap
         * prePhi - the angle of phi on the "pre" residue
         * postPhi - the angle of psi on the "post" residue
         */
        Angle preGapAngles;
        Angle postGapAngles;
        double prePhi = 0;
        double prePsi = 0;
        double postPhi = 0;
        double postPsi = 0;
        Atom preCA_Atom = null;
        Atom postCA_Atom = null;
    	
        double CA_CAdist = 0;
        // double CNdist;
        
        if (!gap.isBeginning())
        {
        	
	        preGapAngles = chiAngles.get(gap.getPreOnDatabase());
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPreOnDatabase(), so we do not need to adjust the coordinate.
	        prePhi = preGapAngles.getPhi();
	        prePsi = preGapAngles.getPsi();
	
	        if (!(preResidue instanceof EmptyResidue))
	        {
	            preCA_Atom = preResidue.getAtom("CA");
	        }
	
	        //Atom preN_Atom = preResidue.getAtom("C");
	        
        }
        
        if (!gap.isEnd())
        {
            postGapAngles = chiAngles.get(gap.getPostOnDatabase());
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPostOnDatabase(), so we do not need to adjust the coordinate.
	        postPhi = postGapAngles.getPhi();
	        postPsi = postGapAngles.getPsi();
	
	        if (!(postResidue instanceof EmptyResidue))
	        {
	            postCA_Atom = postResidue.getAtom("CA");
	        }
	
	        //Atom postC_Atom = postResidue.getAtom("N");
        }
        
        if (preCA_Atom != null && postCA_Atom != null)
        {
            CA_CAdist = HomodellerUtils.getDistance(preCA_Atom, postCA_Atom);
            // CNdist = HomodellerUtils.getDistance(preN_Atom, postC_Atom);
        }

        Map<Loop, SimplexLoopScore> scores = new HashMap<Loop, SimplexLoopScore>();
        double maxScore = 0;
        Loop bestLoop = null;

        // it's + 1 because this is the symbolList substring, based at 1
        String querySequence = gap.getQueryGapSequence().seqString();
//        if ( !gap.isEnd() )
//        {
//            querySequence = querySeq.getSourceSymbolList().subStr(gap.getAdjustedStartOnQuery() + 1, gap.getAdjustedEndOnQuery() /*-1 + 1*/);
//        }
//        else
//        {
//            querySequence = querySeq.getSourceSymbolList().subStr(gap.getAdjustedStartOnQuery() + 1, gap.getAdjustedEndOnQuery() -1);
//        }

        /**
         * ** Find the best loop to insert
         */
        for (Loop loop : loops)
        {
            if (!loop.getInitialized())
            {
                db.fillInLoopInfo(loop);
            }

            double loopCA_CAdist = loop.getCA_CADist();// the CA-CA distance
            //double loopNCdist = loop.getNCDist();

            String loopSequence = loop.getCoilSequence();

            // Run phipsi to get the angles from the loop
            /*StructureImpl tmp = new StructureImpl();
            tmp.addChain(loop.getCoilLoop());
            String pdbText = tmp.toPDB();

            PhipsiRunner phiPhipsiRunner = new PhipsiRunner(pdbText);
            TreeMap<Integer, Angle> loopAngles = phiPhipsiRunner.runPhiPsi();
            */

            TreeMap<Integer, Angle> loopAngles = HomodellerUtils.getAngles(loop.getCoilLoop());

            // get the first and last angles that will be "glueing" to the pre/post gap residues
            Angle firstLoopResidueAngle = loopAngles.get(loopAngles.firstKey() + 1);
            Angle lastLoopResidueAngle = loopAngles.get(loopAngles.lastKey() - 1);

            if (firstLoopResidueAngle == null || lastLoopResidueAngle == null)
            {
                // TODO hack - fix loopDB Reader, it dies on single digits
                System.err.println("Phipsi failed - Skipping loop " + loop.getLoopNumber());
                continue;
            }
            double loopPrePhi = firstLoopResidueAngle.getPhi();
            double loopPrePsi = firstLoopResidueAngle.getPsi();
            double loopPostPhi = lastLoopResidueAngle.getPhi();
            double loopPostPsi = lastLoopResidueAngle.getPsi();

            SimplexLoopScore score = new SimplexLoopScore();

            if (!gap.isBeginning())
            {
                assert prePsi != 0;
                assert prePhi != 0;

                score.pre_phi_delta = HomodellerUtils.angleDist(prePhi, loopPrePhi);
                score.pre_psi_delta = HomodellerUtils.angleDist(prePsi, loopPrePsi);
            }

            if (!gap.isEnd())
            {
                assert postPsi != 0;
                assert postPhi != 0;

                score.post_phi_delta = HomodellerUtils.angleDist(postPhi, loopPostPhi);
                score.post_psi_delta = HomodellerUtils.angleDist(postPsi, loopPostPsi);
            }

            //score.NC_dist_delta = Math.abs(CNdist - loopNCdist);
            score.CA_CA_dist_delta = Math.abs(CA_CAdist - loopCA_CAdist);
            score.similarityScore = HomodellerUtils.getSequenceSimilarity(loopSequence, querySequence);

            // the score will be higher the better it gets

            scores.put(loop, score);

            if (score.getScore() > maxScore)
            {
                maxScore = score.getScore();
                bestLoop = loop;
            }

        }

        if (debug)
        {
            if (bestLoop == null)
            {
                throw new HomodellerException("No optimal loop selected");
            }

            writeChainToFile("loop.pdb", bestLoop.getCoilLoop());
        }

        logger.debug("Highest score = " + maxScore);
        SimplexLoopScore bestScore = scores.get(bestLoop);
        //System.out.println("bestDistance = " + bestScore.NC_dist_delta);
        logger.debug("Best pre phi delta = " + bestScore.pre_phi_delta);
        logger.debug("Best pre psi delta= " + bestScore.pre_psi_delta);

        // superimpose the best loop on the working structure
        assert bestLoop != null : "No optimal loop selected";
    	
        return bestLoop;
    }
    
    
    /**
     * Searches the loop library for the loop that is expected to work best with the Cyclic Coordinate
     * Descent (CCD) algorithm, which is what PepMake uses to close gaps.
     * 
     * @return the best matching loop
     */
    private Loop getBestLoopForCCD(LoopDB db, List<Loop> loops, Group preResidue, Group postResidue, Gap gap)
    		throws StructureException, HomodellerException {
    	
        /**
         * preGapAngles - the angle of the residue right before the gap (i.e. the residue we're glueing the gap onto)
         * postGapAngles - the angle of the residue right after the gap
         * prePhi - the angle of phi on the "pre" residue
         * postPhi - the angle of psi on the "post" residue
         */
        Angle preGapAngles;
        Angle postGapAngles;
        double prePhi = 0;
        double prePsi = 0;
        double postPhi = 0;
        double postPsi = 0;
        Atom preCA_Atom = null;
        Atom postCA_Atom = null;
    	
        double CA_CAdist = 0.0;
        // double CNdist;
        
        if (!gap.isBeginning())
        {
        	
	        preGapAngles = chiAngles.get(gap.getPreOnDatabase());
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPreOnDatabase(), so we do not need to adjust the coordinate.
	        prePhi = preGapAngles.getPhi();
	        prePsi = preGapAngles.getPsi();
	
	        if (!(preResidue instanceof EmptyResidue))
	        {
	            preCA_Atom = preResidue.getAtom("CA");
	        }
	
	        //Atom preN_Atom = preResidue.getAtom("C");
	        
        }
        
        if (!gap.isEnd())
        {
            postGapAngles = chiAngles.get(gap.getPostOnDatabase());
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPostOnDatabase(), so we do not need to adjust the coordinate.
	        postPhi = postGapAngles.getPhi();
	        postPsi = postGapAngles.getPsi();
	
	        if (!(postResidue instanceof EmptyResidue))
	        {
	            postCA_Atom = postResidue.getAtom("CA");
	        }
	
	        //Atom postC_Atom = postResidue.getAtom("N");
        }
        
        if (preCA_Atom != null && postCA_Atom != null)
        {
            CA_CAdist = HomodellerUtils.getDistance(preCA_Atom, postCA_Atom);
            // CNdist = HomodellerUtils.getDistance(preN_Atom, postC_Atom);
        }

        Map<Loop, CCDLoopScore> scores = new HashMap<Loop, CCDLoopScore>();
        double maxScore = 0;
        Loop bestLoop = null;

        // it's + 1 because this is the symbolList substring, based at 1
        String querySequence = gap.getQueryGapSequence().seqString();
//        if ( !gap.isEnd() )
//        {
//            querySequence = querySeq.getSourceSymbolList().subStr(gap.getAdjustedStartOnQuery() + 1, gap.getAdjustedEndOnQuery() /*-1 + 1*/);
//        }
//        else
//        {
//            querySequence = querySeq.getSourceSymbolList().subStr(gap.getAdjustedStartOnQuery() + 1, gap.getAdjustedEndOnQuery() -1);
//        }

        /**
         * ** Find the best loop to insert
         */
        for (Loop loop : loops)
        {
            if (!loop.getInitialized())
            {
                db.fillInLoopInfo(loop);
            }

            double loopCA_CAdist = loop.getCA_CADist();// the CA-CA distance
            //double loopNCdist = loop.getNCDist();

            String loopSequence = loop.getCoilSequence();

            // Run phipsi to get the angles from the loop
            /*StructureImpl tmp = new StructureImpl();
            tmp.addChain(loop.getCoilLoop());
            String pdbText = tmp.toPDB();

            PhipsiRunner phiPhipsiRunner = new PhipsiRunner(pdbText);
            TreeMap<Integer, Angle> loopAngles = phiPhipsiRunner.runPhiPsi();
            */

            TreeMap<Integer, Angle> loopAngles = HomodellerUtils.getAngles(loop.getCoilLoop());

            // get the first and last angles that will be "glueing" to the pre/post gap residues
            Angle firstLoopResidueAngle = loopAngles.get(loopAngles.firstKey() + 1);
            Angle lastLoopResidueAngle = loopAngles.get(loopAngles.lastKey() - 1);

            if (firstLoopResidueAngle == null || lastLoopResidueAngle == null)
            {
                // TODO hack - fix loopDB Reader, it dies on single digits
                System.err.println("Phipsi failed - Skipping loop " + loop.getLoopNumber());
                continue;
            }
            double loopPrePhi = firstLoopResidueAngle.getPhi();
            double loopPrePsi = firstLoopResidueAngle.getPsi();
            double loopPostPhi = lastLoopResidueAngle.getPhi();
            double loopPostPsi = lastLoopResidueAngle.getPsi();

            
            CCDLoopScore score = new CCDLoopScore();

            if (!gap.isBeginning())
            {
                assert prePsi != 0;
                assert prePhi != 0;

                score.pre_phi_delta = HomodellerUtils.angleDist(prePhi, loopPrePhi);
                score.pre_psi_delta = HomodellerUtils.angleDist(prePsi, loopPrePsi);
            }

            if (!gap.isEnd())
            {
                assert postPsi != 0;
                assert postPhi != 0;

                score.post_phi_delta = HomodellerUtils.angleDist(postPhi, loopPostPhi);
                score.post_psi_delta = HomodellerUtils.angleDist(postPsi, loopPostPsi);
            }

            //score.NC_dist_delta = Math.abs(CNdist - loopNCdist);
            score.CA_CA_dist_delta = Math.abs(CA_CAdist - loopCA_CAdist);
            score.similarityScore = HomodellerUtils.getSequenceSimilarity(loopSequence, querySequence);
            
            
            // Superimpose the loop so that the loop Pre residue is at the same place as the
            // structure's Pre residue. This will create a gap between the loop's Post residue
            // and the structure's Post residue. Shown schematically:
            //
            // Before superpositioning:
            //
			//	                             Pre          Post                
			//	                              |            |                    
			//	 -----------------------------*            *--------------------
            //
            //            *-------*
            //            |       |
            //      loop Pre     loop Post
            //
            //
            // After superpositioning:
            //
			//	                             Pre          Post                
			//	                              |            |                    
			//	 -----------------------------*-------*    *--------------------
            //                                |       |
            //                          loop Pre     loop Post
            //
            //                                        |----| post_PostDist
            //
            //
            // Then calculate additional loop scores that we hope will help select better loops
            // for CCD.
            Loop coilLoop = loop.getCoilLoop();
            if (preResidue != null && postResidue != null)
            {
                superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
                
                Group loopPostResidue = coilLoop.getGroup(coilLoop.getGroups().size() - 1);
                //Atom loopPostN_Atom = loopPostResidue.getAtom("N");
                Atom loopPostCA_Atom = loopPostResidue.getAtom("CA");
                //Atom loopPostC_Atom = loopPostResidue.getAtom("C");
                
                //Atom postN_Atom = postResidue.getAtom("N");
                postCA_Atom = postResidue.getAtom("CA");
                //Atom postC_Atom = postResidue.getAtom("C");
                
                if (loopPostCA_Atom != null && postCA_Atom != null) {
                	score.post_PostDist = HomodellerUtils.getDistance(loopPostCA_Atom, postCA_Atom);
                }
            }
            else
            {
                // pre residue will only be null if it's at the beginning
                //assert gap.isBeginning() : "Pre residue is null, not a loop at beginning";

                //superImposeLoop(postResidue, coilLoop.getGroup(coilLoop.getGroups().size() - 1), coilLoop);
            }
            
            
            
            
            // the score will be higher the better it gets

            scores.put(loop, score);

            if (score.getScore() > maxScore)
            {
                maxScore = score.getScore();
                bestLoop = loop;
            }

        }

        if (debug)
        {
            if (bestLoop == null)
            {
                throw new HomodellerException("No optimal loop selected");
            }

            writeChainToFile("loop.pdb", bestLoop.getCoilLoop());
        }

        logger.debug("Highest score = " + maxScore);
        CCDLoopScore bestScore = scores.get(bestLoop);
        //System.out.println("bestDistance = " + bestScore.NC_dist_delta);
        logger.debug("Best pre phi delta = " + bestScore.pre_phi_delta);
        logger.debug("Best pre psi delta= " + bestScore.pre_psi_delta);

        // superimpose the best loop on the working structure
        assert bestLoop != null : "No optimal loop selected";
    	
        return bestLoop;
    }
    

    private Chain minimizeDistances(Group preResidue, Group postResidue, Chain loop)
    {
        SimplexRunner runner = new SimplexRunner(preResidue, postResidue, loop);
        Group preLoopResidue = loop.getGroup(1);
        Group postLoopResidue = loop.getGroup(loop.getGroups().size() - 2);

        Chain chainToMinimize = loop;

        try
        {
            double NC_dist = HomodellerUtils.getDistance(preResidue.getAtom("C"), preLoopResidue.getAtom("N"));
            double CN_dist = HomodellerUtils.getDistance(postResidue.getAtom("N"), postLoopResidue.getAtom("C"));

            int num_tries = 0;
            while ((NC_dist > MINIMIZATION_DISTANCE || CN_dist > MINIMIZATION_DISTANCE) && num_tries++ < MINIMIZATION_MAX_TRIES)
            {
                // set simplex's subject chain to minimize
                runner.setLoop(chainToMinimize);

                // minimize it
                chainToMinimize = runner.runSimplex();

                // superpose it on the structure again
                chainToMinimize = insertLoopInGap(preResidue, postResidue, chainToMinimize);

                if (chainToMinimize == null)
                {
                    continue;
                }

                preLoopResidue = chainToMinimize.getGroup(1);
                postLoopResidue = chainToMinimize.getGroup(chainToMinimize.getGroups().size() - 2);

                // re-calculate the distances
                CN_dist = HomodellerUtils.getDistance(preResidue.getAtom("C"), preLoopResidue.getAtom("N"));
                NC_dist = HomodellerUtils.getDistance(postResidue.getAtom("N"), postLoopResidue.getAtom("C"));

                logger.info(num_tries + ": CN_dist = " + CN_dist + ", NC_dist = " + NC_dist);
                if (debug)
                {

                    writeChainToFile("rotatedLoop.pdb", chainToMinimize);
                }
            }
        }
        catch (StructureException e)
        {
            logger.error(e);
        }

        return chainToMinimize;
    }

    public void setQuerySeq(GappedSequence query)
    {
        querySeq = query;
    }

    public void setDatabaseSeq(GappedSequence database)
    {
        databaseSeq = database;
    }

    /**
     * Score for how suitable a loop is for use by the Simplex algorithm.
     * Higher scores are better.
     */
    private class SimplexLoopScore
    {
        double pre_phi_delta;
        double pre_psi_delta;
        double post_phi_delta;
        double post_psi_delta;
        //   double NC_dist_delta;
        double similarityScore; // proportion in the range [0.0, 1.0]
        public double CA_CA_dist_delta;

        /*
         * Get the score for a particular loop
         * This scoring method is based on 4 measures:
         * - Psi/psi angle similarity
         * - phi/phi angle similarity
         * - ca-ca distance
         * - sequence similarity between the loop and the query region we are replacing
         */
        double getScore()
        {
/*
            double denominator = ( pre_phi_delta / 360 + pre_psi_delta / 360
                            +   post_phi_delta / 360 + post_psi_delta / 360  + NC_dist_delta );
*/
            // TODO add back in post delta's
            double denominator = (pre_phi_delta / 360 + pre_psi_delta / 360 + CA_CA_dist_delta);
            double score;
            if (denominator == 0)
            {
                // perfect match!
                score = Double.MAX_VALUE;
            }
            else
            {
                // TODO this is an arbitrary measure formula
                score = similarityScore + 1 / denominator;
            }

            return score;
        }
    }

    
    /**
     * Score for how suitable a loop is for use by the CCD algorithm.
     * Higher scores are better.
     */
    private class CCDLoopScore
    {
        double pre_phi_delta;
        double pre_psi_delta;
        double post_phi_delta;
        double post_psi_delta;
        //   double NC_dist_delta;
        double similarityScore; // proportion in the range [0.0, 1.0]
        public double CA_CA_dist_delta;
        
        double post_PostDist = -1.0; // Distance between Post residue of structure and Post residue of a coil loop.
        //double post_PostAngle = -1.0; // Not yet implemented.
        	// Magnitude of the angle between N-C vector for Post residue of structure and N-C vector of Post
        	// residue of a coil loop. The smaller the value, the more the closely the coil loop Post residue is
        	// parallel to ("points at") the structure Post residue. (I'm not sure what the best measure for this
        	// is, though. -- David Arndt).
        
        /*
         * Get the score for a particular loop
         * This scoring method is based on 4 measures:
         * - Psi/psi angle similarity
         * - phi/phi angle similarity
         * - ca-ca distance
         * - sequence similarity between the loop and the query region we are replacing
         */
        double getScore()
        {
/*
            double denominator = ( pre_phi_delta / 360 + pre_psi_delta / 360
                            +   post_phi_delta / 360 + post_psi_delta / 360  + NC_dist_delta );
*/
            // TODO add back in post delta's
            double denominator = (pre_phi_delta / 360 + pre_psi_delta / 360 + CA_CA_dist_delta);
            double score;
            if (denominator == 0)
            {
                // perfect match!
                score = Double.MAX_VALUE;
            }
            else
            {
                // TODO this is an arbitrary measure formula
            	
                score = similarityScore + (1.0 / denominator);
                
                if (post_PostDist != -1.0) {
                	if (post_PostDist == 0.0) {
                		// Perfect match.
                		score += Double.MAX_VALUE;
                	}
                	else {
                		score += 5.0*(1.0 / post_PostDist);
                	}
                }
            }

            return score;
        }
    }
    
    
    private void writeChainToFile(String filename, Chain chain)
    {
        try
        {
            StructureImpl tmp = new StructureImpl();

            tmp.addChain(chain);
            String pdbText = tmp.toPDB();
            File tempFile = new File(filename);
            //tempFile.deleteOnExit();

            try
            {
                PrintWriter w = new PrintWriter(tempFile);
                w.print(pdbText);
                w.close();
            }
            catch (FileNotFoundException e)
            {
                e.printStackTrace();
            }
        }
        catch (Exception e)
        {
            // Don't want any exceptions
            e.printStackTrace();
        }
    }

    public List<Gap> findGaps()
    {
        String query = querySeq.seqString();
        String database = databaseSeq.seqString();

        List<Gap> gaps = new ArrayList<Gap>();
        for (int i = 0; i < query.length(); i++)
        {
            char q = query.charAt(i);
            char d = database.charAt(i);

            if (ProteusTools.isGap(d))
            {
                // insertion

                Gap gap = new Gap(databaseSeq, querySeq);
                gap.setStart(i);
                gap.setType(Gap.TYPE.INSERTION);

                int gapLength = 0;
                while (i < database.length() && ProteusTools.isGap(database.charAt(i)))
                {
                    gapLength++;
                    i++;
                }

                // if we don't do this, we'll miss a possible gap after
                i--;

                gap.setLength(gapLength);

                gaps.add(gap);
            }
            else if (ProteusTools.isGap(q))
            {
                // deletion
                Gap gap = new Gap(databaseSeq, querySeq);
                gap.setStart(i);
                gap.setType(Gap.TYPE.DELETION);

                int gapLength = 0;
                while (i < database.length() && ProteusTools.isGap(query.charAt(i)))
                {
                    gapLength++;
                    i++;
                }

                // if we don't do this, we'll miss a possible gap after
                i--;

                gap.setLength(gapLength);

                gaps.add(gap);
            }
            else if (ProteusTools.isGap(d) && ProteusTools.isGap(q))
            {
                throw new IllegalStateException("wtf??");
            }
        }

        return gaps;
    }

    /**
     * Modifies the templates residues so that they match the query sequence.
     * <p/>
     * Note: Empty residues are not changed.
     */
    public void modifyResidues()
    {
        assert querySeq.getSourceSymbolList().length() == newStructure.getChain(0).getSequence().length();

        assert chiAngles != null : "Chi angles from Phipsi are null";

        String query = querySeq.getSourceSymbolList().seqString();
        String database = newStructure.getChain(0).getSequence();

        ArrayList<Group> groups = newStructure.getChain(0).getGroups();

        for (int i = 0; i < query.length(); i++)
        {
            char q = query.charAt(i);
            char d = database.charAt(i);

            if (q != d)
            {
                if (q == 'G')
                {
                    // Glycine does not need a side chain, and is not in the database,
                    // so we need to remove all the side chain atoms
                    Group group = groups.get(i);
                    List<Atom> removals = new ArrayList<Atom>();
                    for (Atom a : (List<Atom>) group.getAtoms())
                    {
                        String name = a.getName();
                        if (!(name.equals("CA") || name.equals("C") || name.equals("O") || name.equals("N")))
                        {
                            removals.add(a);
                        }
                    }

                    for (Atom a : removals)
                    {
                        group.getAtoms().remove(a);
                    }

                    try
                    {
                        group.setPDBName("GLY");
                        ((AminoAcid) group).setAminoType('G');
                    }
                    catch (PDBParseException e)
                    {
                        e.printStackTrace();
                    }
                    continue;
                }

                // ** need to modify this residue (chi angles are 1-based)
                // Because the chi angles calculation is based on the ungapped template (database),
                // convert the index to the index in the gapped template
                int index = Math.abs(databaseSeq.viewToSource(i + 1));
                Angle angle = chiAngles.get(index);

                // TODO hack - angle should not be null (fails on 1TADB with template 1GIA)
                double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                SideChain newSideChain = HomodellerUtils.getSideChain(q, chi);

                assert newSideChain != null : "Can't find side chain for residue " + i + " " + q;

                if (!(groups.get(i) instanceof EmptyResidue))
                {
                    superImposeResidue(groups.get(i), newSideChain);
                    
                    // put the new side chain into the model
                    groups.set(i, (Group) newSideChain.clone());
                }
                
                if (debugSideChains)
                {
                    logger.debug("Debugging side chains... (this should be off by default)");
                    cleanupChain();
                    writeChainToFile("finalModified.pdb", newStructure.getChain(0));
                }
            }
        }

        String sequence = newStructure.getChain(0).getSequence();
        String querySequence = querySeq.getSourceSymbolList().seqString();
        
        // Check the sequence. We did not substitute sidechains in empty residues above, so we allow gaps in
        // the new structure's sequence.
        for (int i = 0; i < sequence.length(); i++) {
        	char s = sequence.charAt(i);
        	char q = querySequence.charAt(i);
        	
        	if ( !(s == q || ProteusTools.isGap(s)) ) {
                System.err.println(sequence);
                System.err.println(querySequence);
                throw new IllegalStateException("Wasn't able to modify the side chains to fit");
        	}
        }
    }


    /**
     * Modifies the coordinates of the new side chain from the coordinates in the database
     * to the coordinates from the original residue
     *
     * @param original     the original residue whose coordinates we wish to modify
     * @param newSideChain the side chain whose coordinates need to be modified to the original's postion
     */
    private void superImposeResidue(Group original, SideChain newSideChain)
    {
        try
        {
            // For a superposition, we need two sets of coordinates
            // abc and pqr.  That will give us the translation and transformation of the new coordinates
            double[][] abc = SideChain.getReferenceCoordinates(original);
            double[][] pqr = SideChain.getReferenceCoordinates(newSideChain);

            List<Atom> atoms = newSideChain.getAtoms();

            // do the translation on these atoms
            for (Atom a : atoms)
            {
                double[] coordinates = Superpose.superpose(abc, pqr, a.getX(), a.getY(), a.getZ());
                a.setX(coordinates[0]);
                a.setY(coordinates[1]);
                a.setZ(coordinates[2]);
            }
        }
        catch (StructureException e)
        {
            logger.error("Error getting atom", e);
        }

    }

    /**
     * Modifies the coordinates of the new side chain from the coordinates in the database
     * to the coordinates from the original residue
     *
     * @param original                 the original residue whose coordinates we wish to modify
     * @param representativeNewResidue the side chain whose coordinates need to be modified to the original's postion
     * @param chain                    The loop or chain to superimpose
     * @return a superimposed chain
     */
    private Chain superImposeLoop(Group original, Group representativeNewResidue, Chain chain)
    {
        if (chain.getGroups().size() == 0)
        {
            return chain;
        }

        try
        {
            // For a superposition, we need two sets of coordinates
            // abc and pqr.  That will give us the translation and transformation of the new coordinates
            double[][] abc = SideChain.getReferenceCoordinates(original);
            double[][] pqr = SideChain.getReferenceCoordinates(representativeNewResidue);

            ArrayList<Group> groups = chain.getGroups();

            // do the translation on every atom in every residue in the chain
            for (Group g : groups)
            {
                List<Atom> atoms = g.getAtoms();

                for (Atom a : atoms)
                {
                    double[] coordinates = Superpose.superpose(abc, pqr, a.getX(), a.getY(), a.getZ());
                    a.setX(coordinates[0]);
                    a.setY(coordinates[1]);
                    a.setZ(coordinates[2]);
                }
            }
        }
        catch (StructureException e)
        {
            logger.error("Error getting atom", e);
        }

        return chain;
    }

    /**
     * Uses the coordinates of the pre and post residues (the residues flanking the gap) to calculate
     * the rotation matrix of the given chain, and translates it
     *
     * @param preResidue  residue flanking the gap on the left side (as per sequence)
     * @param postResidue residue flanking the gap on the right side (as per sequence)
     * @param loop        The loop whose coordinates are to be modified
     * @return the loop that is rotated and translated to fit into a gap
     */
    private Chain insertLoopInGap(Group preResidue, Group postResidue, Chain loop)
    {
        if (loop.getGroups().size() == 0)
        {
            throw new IllegalArgumentException("Why are you sending me an empty loop?");
        }

        Atom[] original = new Atom[6];
        Atom[] target = new Atom[6];

        try
        {
            /* target needs to be the N, CA, C atoms from the pre residue, and the same atoms
                from the post residue
             */
            target[0] = preResidue.getAtom("N");
            target[1] = preResidue.getAtom("CA");
            target[2] = preResidue.getAtom("C");
            target[3] = postResidue.getAtom("N");
            target[4] = postResidue.getAtom("CA");
            target[5] = postResidue.getAtom("C");

            /* Original needs to be the same atoms as target, but taken from the loop */
            ArrayList<Group> groups = loop.getGroups();
            Group preLoop = groups.get(0);
            Group postLoop = groups.get(groups.size() - 1);

            original[0] = preLoop.getAtom("N");
            original[1] = preLoop.getAtom("CA");
            original[2] = preLoop.getAtom("C");
            original[3] = postLoop.getAtom("N");
            original[4] = postLoop.getAtom("CA");
            original[5] = postLoop.getAtom("C");

            // do the translation
            Superpose.superpose(original, target, loop);
        }
        catch (StructureException e)
        {
            logger.error("Error getting atom", e);
        }

        return loop;
    }

    /**
     * Modifies the template so that it has the exact number of residues that
     * the query sequence has.  This is done by deleting or inserting residues.
     */
    public void insertTemporaryGaps()
    {
        String query = querySeq.seqString();
        String database = databaseSeq.seqString();

        ArrayList<Group> groups = newStructure.getChain(0).getGroups();

        int index = 0;
        for (int i = 0; i < query.length(); i++)
        {
            char q = query.charAt(i);
            char d = database.charAt(i);

            if (ProteusTools.isGap(d))
            {
                // insertion, insert new residue
                //groups.add(index-1, new EmptyResidue());
                groups.add(index, new EmptyResidue());
                index++;
            }
            else if (ProteusTools.isGap(q))
            {
                // deletion, delete the residue
                groups.remove(index);
            }
            else if (ProteusTools.isGap(d) && ProteusTools.isGap(q))
            {
                throw new IllegalStateException("wtf??");
            }
            else
            {
                index++;
            }
        }

        assert querySeq.getSourceSymbolList().length() == newStructure.getChain(0).getSequence().length() : "insertTemporaryGaps(): Sequence lengths don't match";
    }

/*
    public void readIncompleteFile() throws IOException
    {
        FileInputStream in = new FileInputStream(incompleteFile);

        PDBFileParser pdbReader = new PDBFileParser();

        structure = pdbReader.parsePDBFile(in);

            /*
            Used for parsing specific chains, do we need this??

            int chainIdx = 0;
            char chainID = PDBFileName.charAt(4);
            if (PDBFileName.length() == 5)
            {
                List model = structure.getModel(0);
                for (int i = 0; i < model.size(); i++)
                {
                    Chain chain = (Chain) model.get(i);
                    String name = chain.getName();
                    if (name.equals(chainID))
                    {
                        chainIdx = i;
                        System.out.println("Processing chain " + PDBFileName.charAt(4));
                        break;
                    }
                }
            }
            else
                chainIdx = 0;

            Chain chain = structure.getChain(chainIdx);*/
    /*
    }
*/

    public Structure getNewStructure()
    {
        return newStructure;
    }

    public Structure getTemplate()
    {
        return template;
    }

    public GappedSequence getQuerySeq()
    {
        return querySeq;
    }

    public GappedSequence getDatabaseSeq()
    {
        return databaseSeq;
    }

    public static void main(String[] args)
    {

    }
}
