package ca.ualberta.proteus.program.homodeller;

import org.biojava.bio.structure.AtomImpl;
import org.biojava.bio.structure.io.PDBParseException;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.apache.log4j.Logger;

import java.io.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Sep 3, 2006
 * Time: 9:58:59 PM
 * To change this template use File | Settings | File Templates.
 */
public class SideDBReader extends LoopDBReader
{
    private SideDB db;
    private static Logger logger = Logger.getLogger(SideDBReader.class);

    private static String recordpatternStr = "\\s*(\\w+)" + // #
                "\\s+([0-9.-]+)" + // angle
                "\\s+B\\s+([0-9.-]+)" + // B
                "\\s+C\\s+([0-9.-]+)" +  // C
                "\\s+H\\s+([0-9.-]+)";  // H
    private static Pattern recordPattern = Pattern.compile(recordpatternStr);

    //ATOM    236  N   ALA    30      -3.403  25.827  12.109  1.00  8.26
    protected static String atomPatternStr = "ATOM.*(\\d+)" + // #
                "\\s+(\\w+)"  +  // # atom type (N, CA, C, O, etc.)
                "\\s+(\\w+)" +  // # atom name, amino acid code
                "\\s+(\\d+)" +  // residue # (meaningless)
                "\\s+([0-9.-]+)" + // x
                "\\s+([0-9.-]+)" + // y
                "\\s+([0-9.-]+)" + // z
                ".*";

    protected static Pattern atomPattern = Pattern.compile(atomPatternStr);
    private SideChain current_sidechain;

    public SideDBReader(SideDB db)
    {
        super(null);

        this.db = db;
    }

    public void parseDB(File file)
    {
        try
        {
            BufferedReader buf = new BufferedReader(new FileReader(file));

            parseDB( buf );
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

     public void parseDB(BufferedReader buf)
            throws IOException
    {
        // (re)set structure

        current_sidechain = null;

        try
        {

            String line = buf.readLine();

            // if line is null already for the first time, the bufferede Reader had a problem
            if (line == null)
            {
                throw new IOException("could not parse PDB File, BufferedReader returns null!");
            }

            //System.out.println ("starting to parse:");
            //System.out.println (line);

            while (line != null)
            {
                if (line.startsWith(">"))
                {
                    pdb_RECORD_Handler(line);
                }
                else if (line.matches(atomPatternStr))
                {
                    pdb_ATOM_Handler(line);
                }

                line = buf.readLine();
            }

            // add the last one
            if ( current_sidechain != null )
            {
                db.addLoop(current_sidechain);
            }
        }
        catch (Exception e)
        {

            e.printStackTrace();
            throw new IOException("Error parsing Loop DB");
        }
    }

    private void pdb_RECORD_Handler(String line)
    {
        // add old one the database
        if ( current_sidechain != null )
        {
            db.addLoop(current_sidechain);

            // set group to null so next loop will not add it
            current_sidechain = null;
        }

        current_sidechain = new SideChain();


        Matcher recordMatcher = recordPattern.matcher(line);
        recordMatcher.find();

        current_sidechain.setAminoType( recordMatcher.group(1).charAt(0) );
        current_sidechain.setAngle( Double.parseDouble(recordMatcher.group(2)) );
    }

    // Handler methods to deal with PDB file records properly.


    /**
     * Database version of the ATOM record
     *
     *  ATOM    236  N   ALA    30      -3.403  25.827  12.109  1.00  8.26
     * ATOM    237  CA  ALA    30      -2.289  26.744  11.966  1.00  9.10
     * ATOM    238  C   ALA    30      -1.177  26.372  12.961  1.00  9.20
     * ATOM    239  O   ALA    30       0.006  26.362  12.610  1.00  9.57
     * ATOM    240  CB  ALA    30      -2.768  28.172  12.222  1.00  9.73
     * @param line
     * @throws org.biojava.bio.structure.io.PDBParseException
     */
    private void pdb_ATOM_Handler(String line)
            throws PDBParseException
    {
        //System.out.println(line);
        Matcher matcher = atomPattern.matcher(line);
        matcher.find();

        // create new atom
        AtomImpl atom = new AtomImpl();

        // altLoc
        atom.setFullName(line.substring(12, 16));
        atom.setName(matcher.group(2));

        double x = Double.parseDouble(matcher.group(5));
        double y = Double.parseDouble(matcher.group(6));
        double z = Double.parseDouble(matcher.group(7));

        double[] coords = new double[3];
        coords[0] = x;
        coords[1] = y;
        coords[2] = z;
        atom.setCoords(coords);

        String residueNumber = matcher.group(4);
        String groupCode3 = matcher.group(3);

        Character aminoCode;
        try
        {
            // is it a standard amino acid ?
            aminoCode = HomodellerUtils.convert_3code_1code(groupCode3);
        }
        catch (IllegalSymbolException e)
        {
            // hm groupCode3 is not standard
            // perhaps it is an nucleotide?
            if (isNucleotide(groupCode3))
            {
                //System.out.println("nucleotide, aminoCode:"+aminoCode);
                aminoCode = null;
            }
            else
            {
                // does not seem to be so let's assume it is
                //  nonstandard aminoacid and label it "X"
                logger.error("unknown amino acid " + groupCode3);
                aminoCode = 'x';
            }
        }

        assert aminoCode != null;

        assert aminoCode.equals(current_sidechain.getAminoType()) : "Amino code doesn't match group's";

        assert current_sidechain != null : "Side chain is null";

        // check if residue number is the same ...
        if (! residueNumber.equals(current_sidechain.getPDBCode()))
        {
            current_sidechain.setPDBCode(residueNumber);
            current_sidechain.setPDBName(groupCode3);
        }

        current_sidechain.addAtom(atom);
    }

    /**
     * initiale new group, either Hetatom or AminoAcid
    private SideChain getNewSidechain(Character aminoCode)
    {
        SideChain sideChain = null;
        if (aminoCode != null)
        {

            SideChain aa = new SideChain();
            aa.setAminoType(aminoCode);
            sideChain = aa;
        }

        //System.out.println("new group type: "+ group.getType() );
        return sideChain;
    }*/
}
