/*
 * Created on May 8, 2008
 */
package ca.ualberta.proteus.program.homodeller.loop;

import ca.ualberta.proteus.util.ProteusTools;

import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.Structure;

/**
 *
 *
 * @author Joseph Cruz
 */
public abstract class AbstractGap implements Gap {

	private Logger logger = Logger.getLogger(SimpleGap.class);

	private Type type;
	
	public int compareTo(Gap gap) {
		if( this.getStart() < gap.getStart() )
			return -1;
		else if( this.getStart() == gap.getStart() ) {
			if( this.getEnd() < gap.getEnd() )
				return -1;
			else if( this.getEnd() == gap.getEnd() )
				return 0;
		}
		return 1;
	}
	
	@Override
	public boolean equals(Object o) {
		if( this == o )
			return true;
		else if( !(o instanceof SimpleGap) )
			return false;
		
		Gap g = (Gap)o;
		
		if( this.getStart() != g.getStart() )
			return false;
		else if( this.getLength() != g.getLength() )
			return false;
		else if( this.getType() != g.getType() )
			return false;
		
		return true;
	}
	
	@Override
	public int hashCode() {
		return (this.getStart()*10000) + (this.getLength()*10) + this.type.ordinal();
	}
	
	@Override
	public String toString() {
        return this.getClass().getSimpleName() 
        		+ ", type = " + this.type
        		+ ", start = " + this.getStart()
        		+ ", end = " + this.getEnd()
        		+ ", length = " + this.getLength()
        		+ ", fixed N pos = " + this.getFixedNTerminalPosition()
        		+ ", fixed C pos = " + this.getFixedCTerminalPosition();
    }
	
	public Type getType() {
		return type;
	}
	
	public void setType(Type type) {
		this.type = type;
	}
	
	public int getLength() {
		return this.getEnd() - this.getStart() - 1;
	}
	
	public Group getFixedNTerminalGroup(GappedSequence querySequence, GappedSequence templateSequence, Structure newStructure) {
    	// Because the chi angles calculation is based on the ungapped template (database),
        // convert the index to the index in the gapped template
    	
    	// debug
    	logger.debug("On alignment:");
    	int nPos = this.getFixedNTerminalPosition();
    	
    	StringBuilder sb = new StringBuilder("Pre:      ");
        for (int i = 0; i < nPos - 1; i++)
        	sb.append(" ");
        logger.debug(sb.toString());
        logger.debug("v = " + nPos);
    	
        logger.debug("query:    " + querySequence.seqString());
        logger.debug("database: " + templateSequence.seqString());
        logger.debug("");
    	
        logger.debug("On newStructure:");
        
    	int preOnQuery = querySequence.viewToSource(nPos);
    	sb = new StringBuilder("Pre:      ");
        for (int i = 0; i < preOnQuery - 1; i++)
        	sb.append(" ");
        logger.debug(sb.toString());
        logger.debug("v = " + preOnQuery);
    	
        logger.debug("newStruc: " + newStructure.getChain(0).getSequence());
        logger.debug("");
    	
        return newStructure.getChain(0).getGroup(preOnQuery - 1);
        	// - 1 because preOnQuery is in biological coordinates starting at 1
        	// whereas getGroup() appears to use coordinates starting from 0.
    }
	
	public Group getFixedCTerminalGroup(GappedSequence querySequence, GappedSequence templateSequence, Structure newStructure) {
    	// Because the chi angles calculation is based on the ungapped template (database),
        // convert the index to the index in the gapped template
    	
    	// debug
    	logger.debug("On alignment:");
    	int cPos = this.getFixedCTerminalPosition();
    	
    	StringBuilder sb = new StringBuilder("Post:     ");
        for (int i = 0; i < cPos - 1; i++)
        	sb.append(" ");
        logger.debug(sb.toString());
        
        logger.debug("v = " + cPos);
        
        logger.debug("query:    " + querySequence.seqString());
        logger.debug("database: " + templateSequence.seqString());
        logger.debug("");
    	
        logger.debug("On newStructure:");
        
        int postOnQuery = querySequence.viewToSource(cPos);
        sb = new StringBuilder("Post:     ");
        for (int i = 0; i < postOnQuery - 1; i++)
        	sb.append(" ");
        logger.debug(sb.toString());
        logger.debug("v = " + postOnQuery);
        
    	logger.debug("newStruc: " + newStructure.getChain(0).getSequence());
    	logger.debug("");
    	
    	return newStructure.getChain(0).getGroup(postOnQuery - 1);
    		// - 1 because postOnQuery is in biological coordinates starting at 1
    		// whereas getGroup() appears to use coordinates starting from 0.
    }
	
	/**
     * Returns the intended number of residues between the Pre and Post residues in the final
     * structure. This is equal to the number of coil residues in the loop that will need to
     * be inserted.
     * 
     * Example of insertion:
     * 
     *    loop:  BBBCCCCCBBB
     *    query: ISGCWVTPYFK
     *    datab: LDSC---KVLN
     *             |     |
     *            Pre   Post
     * 
     * Here the 2 residues C and K between the Pre and Post residues will be replaced by the
     * 5 residues CWVTP. getLoopCoilLength() will return 5. 
     * 
     * 
     * Example of deletion:
     * 
     *            Pre   Post
     *             |     |
     *    query: LDSC---KVLN
     *    datab: ISGCWVTPYFK
     *    loop:  BBBC   CBBB = BBBCCBBB
     *    
     * Here the 5 residues CWVTP will be replaced by the 2 residues C and K. getLoopCoilLength()
     * will return 2.
     * 
     * 
     * Example of deletion involving a merged gap:
     * 
     *            Pre   Post
     *             |     |
     *    query: LDSC-Y-KVLN
     *    datab: ISGCWVTPYFK
     *    loop:  BBBC C CBBB = BBBCCCBBB
     *    
     * Here the 5 residues CWVTP will be replaced by the 3 residues CYK. getLoopCoilLength()
     * will return 3.
     */
    public int getLoopCoilLength(GappedSequence sequence) {
    	int loopCoilLength = 0;

    	int nPos = this.getFixedNTerminalPosition();
    	int cPos = this.getFixedCTerminalPosition();
    	
    	// Count the number of residues (excluding gap characters) between the Pre and Post residues
    	// on the query sequence.
    	for (int i = nPos + 1; i < cPos; i++) {
    		// pre, post and i are in biological coordinates starting from 1. For String.charAt(), we
    		// subtract 1 in order to use coordinates starting from 0.
    		if( !ProteusTools.isGap(sequence.seqString().charAt(i-1)) )
    			loopCoilLength++;
    	}
        
    	return loopCoilLength;
    }
}
