/*
 * Created on May 8, 2008
 */
package ca.ualberta.proteus.program.homodeller.loop;

import ca.ualberta.proteus.program.homodeller.HomodellerUtils;
import ca.ualberta.proteus.program.homodeller.Superpose;
import ca.ualberta.proteus.program.homodeller.loop.Gap.Type;
import ca.ualberta.proteus.program.transsec.Config;

import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.structure.Atom;
import org.biojava.bio.structure.Chain;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.StructureException;

import java.io.File;
import java.util.List;

/**
 *
 *
 * @author Joseph Cruz
 */
public class BoundaryGapRemover extends AbstractGapModifier {

	private static final Logger logger = Logger.getLogger(BoundaryGapRemover.class);
	private static final boolean debug = Config.getConfig().getBooleanValue("debug");
	
	private Chain chain;
	
	public BoundaryGapRemover(GappedSequence sequence, Chain chain) {
		super(sequence);
		this.chain = chain;
	}
	
	private boolean isOutsideBoundary(int pos, boolean start) {
		//if the end is within 20% or 20 residues of the start/end of the sequence, then pos is not valid
		int length = this.sequence.getSourceSymbolList().length();
		int cutoff = (int)Math.min(20, (int)length*0.20);
		if( start )
			return pos <= cutoff;
		else
			return pos >= (length - cutoff);
	}
	
	/**
     * If there are deletions, then move any fragments that are at the end of the chain
     * to remove the deletions.
     *
     * @param gaps
     * @return
     */
    @SuppressWarnings({"unchecked"})
    public List<Gap> modifyGaps(List<Gap> gaps) throws Exception {
        // count backwards to do the end gaps first
        for (int i = gaps.size() - 1; i >= 0; i--) {
            Gap gap = gaps.get(i);

            if( gap.getType() == Type.Deletion && !LoopUtil.isBeginning(gap) && !LoopUtil.isEnd(this.sequence, gap) ) {
            	int start = this.sequence.viewToSource(gap.getFixedNTerminalPosition());
            	int end = this.sequence.viewToSource(gap.getFixedCTerminalPosition());
            	
                // if the last residue in the gap is smaller than the cutoff start...
//                Chain fragment = null;
//                Atom N_atom = null;
//                Group referenceGroup = null;

                if( this.isOutsideBoundary(end, true) ) {
                    if( debug )
                    	LoopUtil.writeChainToFile("debug" + File.separator + this.sequence.getName() + "_before_start_transposed.pdb", chain);

//                    // Get the fragment
//                    fragment = new ChainImpl();
//                    int j = Math.abs(this.sequence.viewToSource(gap.getStart())) - 1;
//                    AminoAcid group = null;
//                    // start at the gap start and go backwards to get the fragment
//                    while( j >= 0 && !(group = (AminoAcid)chain.getGroup(j)).getAminoType().equals('-') ) {
//                    	logger.debug("Group = " + group.getAminoType());
//                        fragment.getGroups().add(0, group);
//                        j--;
//                    }
//
//                    N_atom = fragment.getGroup(fragment.getGroups().size() - 1).getAtom("N");
//                    referenceGroup = chain.getGroup(this.sequence.viewToSource(gap.getEnd()) - 1); // the next group
                    
                    this.sequence.removeGaps(gap.getStart()+1, gap.getLength());
                    this.sequence.addGapsInSource(1, gap.getLength());
                    
                    if( i > 0 ) {
                		SimpleGap previousGap = (SimpleGap)gaps.get(i-1);
                		previousGap.setEnd(previousGap.getEnd()+gap.getLength());
                		logger.debug("Previous gap: " + previousGap);
                	}
                	gaps.remove(i);
                }

                // similarly with the end
                if( this.isOutsideBoundary(start, false) ) {
                    if (debug)
                        LoopUtil.writeChainToFile("debug" + File.separator + this.sequence.getName() + "_before_end_transposed.pdb", chain);

//                    // Get the fragment
//                    fragment = new ChainImpl();
//                    int j = Math.abs(this.sequence.viewToSource(gap.getEnd())) - 1;
//                    AminoAcid group;
//                    while( j < chain.getLength() && !(group = (AminoAcid) chain.getGroup(j)).getAminoType().equals('-') ) {
//                        fragment.addGroup(group);
//                        j++;
//                    }
//
//                    N_atom = fragment.getGroup(0).getAtom("N");
//                    int groupIndex = this.sequence.viewToSource(gap.getStart()) - 1;
//                    referenceGroup = chain.getGroup(groupIndex);
                    
                    this.sequence.removeGaps(gap.getStart()+1, gap.getLength());
                    this.sequence.addGapsInSource(this.sequence.length(), gap.getLength());
                    if( i < (gaps.size()-1) ) {
                    	SimpleGap nextGap = (SimpleGap)gaps.get(i+1);
                    	nextGap.setStart(nextGap.getStart()-gap.getLength());
                    	logger.debug("Next gap: " + nextGap);
                    }
                    gaps.remove(i);
                }

//                if (fragment != null) {
//                	this.moveFragment(fragment, referenceGroup, N_atom);
//                	gaps.remove(i);
//                }
            }
        }

        return gaps;
    }
	
    private void moveFragment(Chain fragment, Group referenceGroup, Atom N_atom) throws StructureException {
    	Atom C_atom = referenceGroup.getAtom("C");
    	Atom O_atom = referenceGroup.getAtom("O");
		
    	assert N_atom != null && C_atom != null : "N or C atom is null";

        // distance away from C to move the fragment
        final double N_C_DISTANCE = 1.45;

        // Move the entire group so the N-terminal of the fragment matches up with the O of the C-terminal
        double[] C_coords = C_atom.getCoords();
        double[] N_coords = N_atom.getCoords();
        double[] transpose = new double[3];

        // distance in space, d^2 = offset^2 + offset^2 + offset^2 = 3* offset ^ 2
        // d^2 = 3* offset ^ 2
        // d = sqrt(3) * offset
        // offset = d / sqrt(3)
        
        double offset = N_C_DISTANCE / Math.sqrt(3);

        // build transposition matrix
        for (int k = 0; k < N_coords.length; k++) {
            transpose[k] = C_coords[k] - N_coords[k] - offset;
        }

        // do transposition
        for (Group g : (List<Group>) fragment.getGroups()) {
            for (Atom a : (List<Atom>) g.getAtoms()) {
                double[] coords = a.getCoords();
                Superpose.translate(coords, transpose);
            }
        }

        if (debug)
            LoopUtil.writeChainToFile("debug" + File.separator + this.sequence.getName() + "_transposed.pdb", chain);

        //TODO remove
        // check for bumps on moved fragment, between the N and O atom
        double N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);
        if (N_O_dist < 1.5)
        {
            // too close, rotate C=O carboxyl to move away from N
            Atom CA_atom = null;
            try
            {
                CA_atom = referenceGroup.getAtom("CA");
            }
            catch (StructureException e)
            {
                e.printStackTrace();
            }

            // this is distance we want C=O to have
            double C_O_dist = HomodellerUtils.getDistance(C_atom, O_atom);

            assert CA_atom != null : "CA Atom is null";

            double[] CA_coords = CA_atom.getCoords();

            // find the point between the CA and N atom
            double[] point = new double[3];
            point[0] = (CA_coords[0] + N_coords[0]) / (double) 2;
            point[1] = (CA_coords[1] + N_coords[1]) / (double) 2;
            point[2] = (CA_coords[2] + N_coords[2]) / (double) 2;

            // find the line between (point) and C, which we will extrapolate to be
            // the position of the O
            double[] line = new double[3];
            line[0] = C_coords[0] - point[0];
            line[1] = C_coords[1] - point[1];
            line[2] = C_coords[2] - point[2];

            // change line's magnitude to the magnitude of the distance between C=O
            // i.e. |line| = C_O_dist
            double multiplier = C_O_dist / Superpose.normalize(line);
            line[0] *= multiplier;
            line[1] *= multiplier;
            line[2] *= multiplier;

            // these are O's new coordinates
            double O_coords[] = O_atom.getCoords();
            O_coords[0] = C_coords[0] + line[0];
            O_coords[1] = C_coords[1] + line[1];
            O_coords[2] = C_coords[2] + line[2];
        }

        if (debug)
            LoopUtil.writeChainToFile("debug" + File.separator + this.sequence.getName() + "_transposed_fixed.pdb", chain);
    }
}
