/*
 * Created on May 9, 2008
 */
package ca.ualberta.proteus.program.homodeller.loop;

import ca.ualberta.proteus.program.homodeller.Angle;
import ca.ualberta.proteus.program.homodeller.EmptyResidue;
import ca.ualberta.proteus.program.homodeller.HomodellerException;
import ca.ualberta.proteus.program.homodeller.HomodellerUtils;
import ca.ualberta.proteus.program.homodeller.Loop;
import ca.ualberta.proteus.program.homodeller.LoopDB;
import ca.ualberta.proteus.program.transsec.Config;

import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.structure.Atom;
import org.biojava.bio.structure.Group;
import org.biojava.bio.symbol.SymbolList;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 *
 * @author Joseph Cruz
 */
public class CCDLoopFinder implements LoopFinder {

	private static final Logger logger = Logger.getLogger(CCDLoopFinder.class);
	private static final boolean debug = Config.getConfig().getBooleanValue("debug");
	
	private GappedSequence querySequence;
	private GappedSequence templateSequence;
	private Map<Integer, Angle> chiAngles;
	private LoopDB db;
	private List<Loop> loops;
	private Group preResidue;
	private Group postResidue;
	private Gap gap;
	
	public CCDLoopFinder(GappedSequence querySequence, GappedSequence templateSequence, Map<Integer, Angle> chiAngles, LoopDB db, List<Loop> loops, Group preResidue, Group postResidue, Gap gap) {
		this.querySequence = querySequence;
		this.templateSequence = templateSequence;
		this.chiAngles = chiAngles;
		this.db = db;
		this.loops = loops;
		this.preResidue = preResidue;
		this.postResidue = postResidue;
		this.gap = gap;
	}
	
	/**
     * Searches the loop library for the loop that is expected to work best with the Cyclic Coordinate
     * Descent (CCD) algorithm, which is what PepMake uses to close gaps.
     * 
     * @return the best matching loop
     */
    public Loop getBestLoop() throws Exception {
        /**
         * preGapAngles - the angle of the residue right before the gap (i.e. the residue we're glueing the gap onto)
         * postGapAngles - the angle of the residue right after the gap
         * prePhi - the angle of phi on the "pre" residue
         * postPhi - the angle of psi on the "post" residue
         */
        Angle preGapAngles;
        Angle postGapAngles;
        double prePhi = 0;
        double prePsi = 0;
        double postPhi = 0;
        double postPsi = 0;
        Atom preCA_Atom = null;
        Atom postCA_Atom = null;
    	
        double CA_CAdist = 0.0;
        // double CNdist;
        
        if( !LoopUtil.isBeginning(gap) ) {
        	int preOnDatabase = this.templateSequence.viewToSource(gap.getFixedNTerminalPosition());
        	
	        preGapAngles = chiAngles.get(preOnDatabase);
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPreOnDatabase(), so we do not need to adjust the coordinate.
	        prePhi = preGapAngles.getPhi();
	        prePsi = preGapAngles.getPsi();
	
	        if (!(preResidue instanceof EmptyResidue))
	            preCA_Atom = preResidue.getAtom("CA");
	
	        //Atom preN_Atom = preResidue.getAtom("C");
        }
        
        if( !LoopUtil.isEnd(this.querySequence, gap) ){
        	int postOnDatabase = this.templateSequence.viewToSource(gap.getFixedCTerminalPosition());
        	
            postGapAngles = chiAngles.get(postOnDatabase);
	        	// It appears that chiAngles uses biological coordinates starting from 1, as
	        	// does getPostOnDatabase(), so we do not need to adjust the coordinate.
	        postPhi = postGapAngles.getPhi();
	        postPsi = postGapAngles.getPsi();
	
	        if (!(postResidue instanceof EmptyResidue))
	            postCA_Atom = postResidue.getAtom("CA");
	
	        //Atom postC_Atom = postResidue.getAtom("N");
        }
        
        if (preCA_Atom != null && postCA_Atom != null) {
            CA_CAdist = HomodellerUtils.getDistance(preCA_Atom, postCA_Atom);
            // CNdist = HomodellerUtils.getDistance(preN_Atom, postC_Atom);
        }

        Map<Loop, CCDLoopScore> scores = new HashMap<Loop, CCDLoopScore>();
        double maxScore = 0;
        Loop bestLoop = null;

        // it's + 1 because this is the symbolList substring, based at 1
        String querySeq = SymbolList.EMPTY_LIST.seqString();
        if( this.gap.getLoopCoilLength(this.querySequence) > 0 ) {
        	int preOnQuery = this.querySequence.viewToSource(gap.getFixedNTerminalPosition());
            int postOnQuery = this.querySequence.viewToSource(gap.getFixedCTerminalPosition());
            querySeq = this.querySequence.getSourceSymbolList().subList(preOnQuery + 1, postOnQuery - 1).seqString();
        }

        /**
         * Find the best loop to insert
         */
        for (Loop loop : loops) {
            if (!loop.getInitialized())
                db.fillInLoopInfo(loop);

            double loopCA_CAdist = loop.getCA_CADist();// the CA-CA distance

            String loopSequence = loop.getCoilSequence();

            TreeMap<Integer, Angle> loopAngles = HomodellerUtils.getAngles(loop.getCoilLoop());

            // get the first and last angles that will be "glueing" to the pre/post gap residues
            Angle firstLoopResidueAngle = loopAngles.get(loopAngles.firstKey() + 1);
            Angle lastLoopResidueAngle = loopAngles.get(loopAngles.lastKey() - 1);

            if (firstLoopResidueAngle == null || lastLoopResidueAngle == null)
            {
                // TODO hack - fix loopDB Reader, it dies on single digits
                System.err.println("Phipsi failed - Skipping loop " + loop.getLoopNumber());
                continue;
            }
            double loopPrePhi = firstLoopResidueAngle.getPhi();
            double loopPrePsi = firstLoopResidueAngle.getPsi();
            double loopPostPhi = lastLoopResidueAngle.getPhi();
            double loopPostPsi = lastLoopResidueAngle.getPsi();
            
            CCDLoopScore score = new CCDLoopScore();

            if( !LoopUtil.isBeginning(gap) ) {
                assert prePsi != 0;
                assert prePhi != 0;

                score.pre_phi_delta = HomodellerUtils.angleDist(prePhi, loopPrePhi);
                score.pre_psi_delta = HomodellerUtils.angleDist(prePsi, loopPrePsi);
            }

            if( !LoopUtil.isEnd(this.querySequence, gap) ) {
                assert postPsi != 0;
                assert postPhi != 0;

                score.post_phi_delta = HomodellerUtils.angleDist(postPhi, loopPostPhi);
                score.post_psi_delta = HomodellerUtils.angleDist(postPsi, loopPostPsi);
            }

            //score.NC_dist_delta = Math.abs(CNdist - loopNCdist);
            score.CA_CA_dist_delta = Math.abs(CA_CAdist - loopCA_CAdist);
            score.similarityScore = HomodellerUtils.getSequenceSimilarity(loopSequence, querySeq);
            
            // Superimpose the loop so that the loop Pre residue is at the same place as the
            // structure's Pre residue. This will create a gap between the loop's Post residue
            // and the structure's Post residue. Shown schematically:
            //
            // Before superpositioning:
            //
			//	                             Pre          Post                
			//	                              |            |                    
			//	 -----------------------------*            *--------------------
            //
            //            *-------*
            //            |       |
            //      loop Pre     loop Post
            //
            //
            // After superpositioning:
            //
			//	                             Pre          Post                
			//	                              |            |                    
			//	 -----------------------------*-------*    *--------------------
            //                                |       |
            //                          loop Pre     loop Post
            //
            //                                        |----| post_PostDist
            //
            //
            // Then calculate additional loop scores that we hope will help select better loops
            // for CCD.
            Loop coilLoop = loop.getCoilLoop();
            if (preResidue != null && postResidue != null) {
                LoopUtil.superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
                
                Group loopPostResidue = coilLoop.getGroup(coilLoop.getGroups().size() - 1);
                //Atom loopPostN_Atom = loopPostResidue.getAtom("N");
                Atom loopPostCA_Atom = loopPostResidue.getAtom("CA");
                //Atom loopPostC_Atom = loopPostResidue.getAtom("C");
                
                //Atom postN_Atom = postResidue.getAtom("N");
                postCA_Atom = postResidue.getAtom("CA");
                //Atom postC_Atom = postResidue.getAtom("C");
                
                if (loopPostCA_Atom != null && postCA_Atom != null) {
                	score.post_PostDist = HomodellerUtils.getDistance(loopPostCA_Atom, postCA_Atom);
                }
            }
            
            // the score will be higher the better it gets
            scores.put(loop, score);

            if (score.getScore() > maxScore)
            {
                maxScore = score.getScore();
                bestLoop = loop;
            }

        }

        if (debug) {
            if (bestLoop == null)
                throw new HomodellerException("No optimal loop selected");

            LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_best_loop_ccd.pdb", bestLoop.getCoilLoop());
        }

        logger.debug("Highest score = " + maxScore);
        CCDLoopScore bestScore = scores.get(bestLoop);
        //System.out.println("bestDistance = " + bestScore.NC_dist_delta);
        logger.debug("Best pre phi delta = " + bestScore.pre_phi_delta);
        logger.debug("Best pre psi delta= " + bestScore.pre_psi_delta);

        // superimpose the best loop on the working structure
        assert bestLoop != null : "No optimal loop selected";
    	
        return bestLoop;
    }
}
