/*
 * Created on May 8, 2008
 */
package ca.ualberta.proteus.program.homodeller.loop;

import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.util.ArrayList;
import java.util.List;

/**
 *
 *
 * @author Joseph Cruz
 */
public class GapMerger extends AbstractGapModifier {

	private static final int MERGE_GAPS_THRESHOLD = 0;
	private static final Logger logger = Logger.getLogger(GapMerger.class);
	
	private GappedSequence templateSequence;
	
	public GapMerger(GappedSequence sequence, GappedSequence templateSequence) {
		super(sequence);
		this.templateSequence = templateSequence;
	}
	
	/**
     * For any adjacent pair of gaps of the same type (i.e. both deletions or both insertions), combine
     * them into a single gap.
     * 
     * For any adjacent pair of gaps of different types (i.e. one is an insertion and the other is
     * a deletion), if the Pre residue of the second gap is to the left of the Post residue of the
     * first gap, "collapse" the gaps by editing the alignment between the query and database
     * sequences. e.g. we have a situation like
	 *
	 *                 gap.getPostOnAlignment()
	 *                  |
	 *    query: MVQ---LRSHMTWHQ
	 *    datab: MVQGAKL---MTWHQ
	 *                |
	 *               next.getPreOnAlignment()
	 *
	 * and we want to alter the alignment so it becomes:
	 *
	 *               new gap.getPostOnAlignment()
	 *                |
	 *    query: MVQ-LRSHMTWHQ
	 *    datab: MVQGAKL-MTWHQ
	 *                |
	 *               new next.getPreOnAlignment()
	 *
	 * Notice that this dislocates the alignment between the two L residues. One or both of
	 * the gaps may disappear. The coordinates of many of the gaps in the gap list will also
	 * need to be adjusted.
     * 
     * @param gaps The list of gaps.
     * @return The altered list of gaps.
     */
    public List<Gap> modifyGaps(List<Gap> gaps) throws Exception {
        for (int i = 0; i < gaps.size() - 1; i++) {
            Gap gap = gaps.get(i);
            Gap next = gaps.get(i + 1);

            // If the Pre residue of the next gap is to the left of the Post residue of the current gap,
            // either merge the gaps (if they are of the same type), or alter the alignment between the
            // two gaps (if they are of different types).
            if( next.getFixedNTerminalPosition()-gap.getFixedCTerminalPosition() < MERGE_GAPS_THRESHOLD ) {
            	if (gap.getType() == next.getType()) {
            		// The gaps are of the same type, so merge them.
            		if( !LoopUtil.isBeginning(gap) && !LoopUtil.isEnd(this.sequence, next) ) {
            			// Do not merge a beginning or end gap with another gap.
            			MergedGap mergedGap = new MergedGap(gap, next);
	                    logger.debug("Merging gap (" + gap + ") and (" + next + "): " + gap);
	
	                    // remove it
	                    gaps.set(i, mergedGap);
	                    gaps.remove(next);
            		}
            		else if( LoopUtil.isEnd(this.sequence, next) )
            			return gaps;
            		else
            			continue;
            	}
            	else {
            		continue;
            		// The gaps are of different types, so collapse them by altering the alignment.
            		
            		// In choosing shiftAmount here, note that we limit it to the length of the smallest gap.
            		// For gaps that are merged gaps consisting of two or more subgaps, we use the length of
            		// the smallest subgap among the C-terminal subgap of the left gap and the N-terminal
            		// subgap of the right gap. e.g. in the case of two merged gaps as follows,
            		//
            		//                   gap.getFixedCTerminalPosition()
            		//                    |
            		//    query: MVQ---L-SHMVTWHQRC
            		//    datab: MVQGAKLQ----T-HQRC
            		//                 |
            		//                next.getFixedNTerminalPosition()
            		//
            		// shiftAmount would be 1 and we would end up with
            		//
            		//                 new gap.getFixedCTerminalPosition()
            		//                  |
            		//    query: MVQ---LSHMVTWHQRC
            		//    datab: MVQGAKLQ---T-HQRC
            		//                 |
            		//                new next.getFixedNTerminalPosition()
            		//
            		// Notice that the two remaining gaps are still too close together. This will be resolved
            		// when collapseGaps() is called recursively.
            		//
            		// Also, note that we have closed one of the subgaps of the gap on the left. We will have to
            		// set it to be a simple gap, and reset its length. For the gap on the right, it remains a
            		// merged gap, but its N-terminal subgap has become shorter, and we will have to adjust its
            		// length.
//            		try {
//	            		int gapLength = (gap instanceof SimpleGap)? gap.getLength(): ((MergedGap)gap).getCTerminalSubgapLength();
//	            		int nextLength = (next instanceof SimpleGap)? next.getLength(): ((MergedGap)next).getNTerminalSubgapLength();
//	            		int shortestGapLength = Math.min(gapLength, nextLength);
//	            		
//	            		// Number of residues by which to shift the alignment.
//	            		int shiftAmount = Math.min(shortestGapLength, gap.getFixedCTerminalPosition()-next.getFixedNTerminalPosition());
//	            		
//	            		// Remove shiftAmount residue positions in each gap. This is done by altering this.querySeq and
//	            		// this.databaseSeq.
//	            		
//	            		List<Gap> gapPair = new ArrayList<Gap>();
//	            		gapPair.add(gap);
//	            		gapPair.add(next);
//	            		
//	            		for (int j= 0; j < gapPair.size(); j++) {
//	            			Gap currentGap = gapPair.get(j);
//	            			GappedSequence sequenceToAlter = null;
//	            			
//	            			switch(currentGap.getType()) {
//	            			case Insertion:
//	            				sequenceToAlter = this.templateSequence;
//	            				break;
//	            			case Deletion:
//	            				sequenceToAlter = this.sequence;
//	            				break;
//	            			default:
//	            				assert false; // Should never happen
//	            			}
//	            			
//	        				//sequenceToAlter.removeGaps(currentGap.getStart() + 1, shiftAmount);
//	        				
//	        				if (j == 0) {
//	        					// Remove gaps from the C-terminal end of the gap (in case it is a merged gap).
//	        					sequenceToAlter.removeGaps(currentGap.getEnd() - shiftAmount, shiftAmount);
//	        				}
//	        				else { // j == 1
//	        					// Remove gaps from the N-terminal end of the gap (in case it is a merged gap).
//	        					sequenceToAlter.removeGaps(currentGap.getStart() + 1, shiftAmount);
//	        				}
//	        				
//	        				
//	        				// Change the lengths and starting positions of the gaps in our gap pair, and remove
//	        				// any of them that have been closed.
//	        				if( currentGap instanceof SimpleGap ) {
//	            				if( currentGap.getLength() == shiftAmount ) {
//	            					// Remove the gap from the 'gaps' list.
//	            					gaps.remove(currentGap);
//	            				}
//	            				else {
//	            					SimpleGap sg = (SimpleGap)currentGap;
//	            					
//	            					// Change the end of the gap.
//	            					sg.setEnd(sg.getEnd() - shiftAmount);
//	            					
//	            					// Change the start position of the gap. (The left gap's starting position does
//	            					// not need to be changed).
//	            					if( j == 1 ) {
//	            						sg.setStart(currentGap.getStart() - shiftAmount);
//	            					}
//	            				}
//	        				}
//	        				else {
//	        					// The gap is a merged gap.
//	        					MergedGap mg = (MergedGap)currentGap;
//	        					
//	        					if( j == 0 ) {
//	        						if( mg.getCTerminalSubgapLength() == shiftAmount) {
//	        							// The C-terminal subgap has closed.
//	        							gaps.set(j, mg.getFirst());
//	        						}
//	        						else {
//	        							// Change the end of the C-terminal subgap (which will also change the total
//	        							// length of the gap).
//	        							mg.setEnd(currentGap.getEnd() - shiftAmount);
//	        						}
//	        					}
//	        					else { // j == 1
//	        						if (mg.getNTerminalSubgapLength() == shiftAmount) {
//	        							// The N-terminal subgap has closed.
//	        							gaps.set(j, mg.getSecond());
//	        						}
//	        						else {
//	        							// Change the start position of the gap. (The left gap's starting position does
//	                					// not need to be changed).
//	        							mg.setStart(currentGap.getStart() - shiftAmount);
//	        						}
//	        					}
//	        					
//	        				}
//	        				
//	            		}
//	            		
//	            		// Scan through the list of gaps and adjust their start positions. The start positions of the gaps
//	            		// in our gap pair have already been adjusted above, so skip these gaps.
//	            		
//	            		int secondGapStart = next.getStart();
//	            		
//	            		for (int j = 0; j < gaps.size(); j++) {
//	            			Gap currentGap = gaps.get(j);
//	            			
//	            			if (currentGap.getStart() > secondGapStart) {
//	            				// The current gap is to the right of the gaps in our gap pair. Change its
//	            				// starting position.
//	            				if( currentGap instanceof SimpleGap )
//	            					((SimpleGap)currentGap).setStart(currentGap.getStart() - shiftAmount);
//	            				else
//	            					((MergedGap)currentGap).setStart(currentGap.getStart() - shiftAmount);
//	            			}
//	            		}
//	            	}
//            		catch( IllegalSymbolException e ) {
//            			logger.error("Illegal symbol", e);
//            			throw e;
//            		}
            	}
            	
                // call this function again on the reduced set of gaps
                return this.modifyGaps(gaps);
            }
        }

        return gaps;
    }
}
