/*
 * Created on Apr 22, 2008
 */
package ca.ualberta.proteus.program.homodeller.loop;

import ca.ualberta.proteus.program.homodeller.*;
import ca.ualberta.proteus.program.homodeller.loop.Gap.Type;
import ca.ualberta.proteus.program.minimization.PepmakeRunner;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.util.ProteusTools;
import org.apache.log4j.Logger;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.structure.*;
import org.biojava.bio.structure.io.PDBParseException;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 *
 *
 * @author Joseph Cruz
 */
public class LoopInserter {

	private static final Logger logger = Logger.getLogger(LoopInserter.class);
	private static final boolean debug = Config.getConfig().getBooleanValue("debug");
	
	private static int DEBUG_LOOP_NUM = -1;
	private static final int MINIMIZATION_MAX_TRIES = 5;
    private static final double MINIMIZATION_DISTANCE = 1.5; /*TODO This is an arbitrary value*/
    private static final double N_O_MAX_DISTANCE = 2.2;
    
	private Structure newStructure;
	private Structure template;
	private GappedSequence querySequence;
	private GappedSequence templateSequence;
	
	private Map<Integer, Angle> chiAngles;
	
	public LoopInserter(Structure template, GappedSequence querySequence, GappedSequence templateSequence) {
		this.template = template;
		this.querySequence = querySequence;
		this.templateSequence = templateSequence;
    }
	
	public Structure run() {
        assert template.size() == 1 : "Template has too many chains";

        // copy it to the new structure
        this.newStructure = (Structure)this.template.clone();

        Chain chain = this.newStructure.getChain(0);

        // TODO - for some reason, phipsi internally screws up the chain,
        // probably because of numbering or something.  So keep using phipsi as
        // external program for now
        HomodellerUtils.removeHetatoms(chain);

        chiAngles = HomodellerUtils.getAngles(chain);

        // make sure our database sequence matches the template structure
        assert chain.getSequence().equals(this.templateSequence.getSourceSymbolList().seqString()) :
                "The database structure and template structure do not match";

        List<Gap> gaps = this.findGaps();

        try {
            BoundaryGapRemover gapRemover = new BoundaryGapRemover(this.querySequence, chain);
            gaps = gapRemover.modifyGaps(gaps);

            // debug
            logger.debug("Before merging gaps:");
            logger.debug("query  : " + this.querySequence.seqString());
            logger.debug("tmplate: " + this.templateSequence.seqString());

            // If there are gaps that are too close to each other, either collapse or merge them.
            GapMerger gapMerger = new GapMerger(this.querySequence, this.templateSequence);
            gaps = gapMerger.modifyGaps(gaps);

            // debug
            logger.debug("After merging gaps:");
            logger.debug("query  : " + this.querySequence.seqString());
            logger.debug("tmplate: " + this.templateSequence.seqString());

        } catch(Exception e) {
            logger.error("Could not get structure.", e);
        }

        // insert gaps or deletions
        insertTemporaryGaps();

        if( debug ) {
            LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure.pdb", this.newStructure.getChain(0));
        }

        // do final loop insertions
        insertLoops(gaps);

        assert this.querySequence.getSourceSymbolList().length() == this.newStructure.getChain(0).getSequence().length();

        // Change the sequence and side chains (except for empty residues).
        modifyResidues();

        // debug
        logger.debug("\n   Writing newStructure_mod.pdb after modifyResidues()...");
        if( debug )
            LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure_mod.pdb", newStructure.getChain(0));

        // Remove empty residues at the beginning and end.
        removeBeginningAndEndGaps();

        // Remove empty residues in the interior.
        removeInteriorGaps();

        // go through and verify side chains and numbering
        cleanupChain();

        // debug
        logger.debug("\n   Writing newStructure_clean.pdb after cleanupChain()...");
        if( debug )
            LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure_clean.pdb", newStructure.getChain(0));

        // remove any groups we can't gap right
        //removeAllUnFilledGaps();

        // debug
        logger.debug("\n   Writing newStructure_rug.pdb after removeAllUnFilledGaps()...");
        if( debug )
            LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure_rug.pdb", newStructure.getChain(0));

        // Do bump checking
        checkBumps();

        return this.newStructure;
    }
	
	private void insertTemporaryGaps() {
        String query = this.querySequence.seqString();
        String database = this.templateSequence.seqString();

        Chain chain = this.newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();

        int index = 0;
        for( int i = 0; i < query.length(); i++ ) {
            char q = query.charAt(i);
            char d = database.charAt(i);

            if( ProteusTools.isGap(d) && ProteusTools.isGap(q) ) {
            	//throw new IllegalStateException("wtf??");
            }
            else if( ProteusTools.isGap(d) ) {
                // insertion, insert new residue
                groups.add(index, new EmptyResidue());
                index++;
            }
            else if( ProteusTools.isGap(q) ) {
                // deletion, delete the residue
                groups.remove(index);
            }
            else {
                index++;
            }
        }

        assert this.querySequence.getSourceSymbolList().length() == chain.getSequence().length() : "insertTemporaryGaps(): Sequence lengths don't match";
    }
	
	private List<Gap> findGaps() {
		List<Gap> gaps = new ArrayList<Gap>();
    	
    	String query = this.querySequence.seqString();
    	String database = this.templateSequence.seqString();

    	int queryLength = query.length();
    	int i = 0;
    	while( i < queryLength ) {
    		SimpleGap gap = null;
            char q = query.charAt(i);
            char d = database.charAt(i);
            String seqString = "";
            
            if( ProteusTools.isGap(d) && ProteusTools.isGap(q) ) {	// should never have gap on both query and database sequences
                throw new IllegalStateException("wtf??");
            }
            else if( ProteusTools.isGap(d) ) {	// insertion
            	gap = new SimpleGap();
            	gap.setStart(i);
            	gap.setType(Type.Insertion);
            	
            	int gapLength = 1;
                while( (i+gapLength) < database.length() && ProteusTools.isGap(database.charAt(i+gapLength))) {
                    gapLength++;
                }
                i += gapLength;
                logger.debug("Gap length = " + gapLength + " (i=" + i + ")");
                gap.setEnd(i+1);
                
                seqString = database;
            }
            else if( ProteusTools.isGap(q) ) {	// deletion
            	gap = new SimpleGap();
            	gap.setStart(i);
            	gap.setType(Type.Deletion);
            	
                int gapLength = 1;
                while( (i+gapLength) < database.length() && ProteusTools.isGap(query.charAt(i+gapLength)) ) {
                    gapLength++;
                }
                i += gapLength;
                logger.debug("Gap length = " + gapLength + " (i=" + i + ")");
                gap.setEnd(i+1);
                
                seqString = query;
            }
            else
            	i++;
            
            if( gap != null ) {
            	int gapSizeBeforeInsertion = 6;
            	if( gap.getType() == Type.Deletion )
            		gapSizeBeforeInsertion = Math.max((int)(gap.getLength()*1.3), 6);
            	int nPos = gap.getStart() - (int)Math.floor(gapSizeBeforeInsertion/2.0);
            	int cPos = gap.getEnd() + (int)Math.ceil(gapSizeBeforeInsertion/2.0);
            	logger.debug("nPos = " + nPos + "," + "cPos = " + cPos);
            	
            	if( !LoopUtil.isBeginning(gap) && (nPos < 0 || ((nPos+1) < seqString.length() && ProteusTools.isGap(seqString.charAt(nPos-1)))) ) {
            		while( (nPos < 0 || ProteusTools.isGap(seqString.charAt(nPos-1))) && nPos < cPos && cPos < (seqString.length()-1) ) {
            			nPos++;
            			if( !ProteusTools.isGap(seqString.charAt(cPos + 1)) )
            				cPos++;
            		}
            	}
            	if( gap.getEnd() < seqString.length() && (cPos >= seqString.length() || ((cPos-1) >= 0 && ProteusTools.isGap(seqString.charAt(cPos-1)))) ) {
            		while( (cPos >= seqString.length() || ProteusTools.isGap(seqString.charAt(cPos-1))) && cPos > nPos && nPos > 0 ) {
            			if( !ProteusTools.isGap(seqString.charAt(nPos - 2)) )
            				nPos--;
            			cPos--;
            		}
            	}
            	
            	gap.setFixedNTerminalPosition(nPos);
            	gap.setFixedCTerminalPosition(cPos);
            	gaps.add(gap);
            }
        }
    	
    	return gaps;
    }
	
	/**
     * Find gaps and insert loops into them
     */
    public void insertLoops(List<Gap> gaps) {
        for( Gap gap : gaps ) {
            logger.debug("Gap: " + gap);

            try {
            	if( !LoopUtil.isBeginning(gap) && !LoopUtil.isEnd(this.querySequence, gap) ) {
                    /* If this is a middle loop (not a beginning or an ending), then fill it.
                       If it is a end loop, and it's a deletion, we don't care about it. */
            		fillGap(gap);

            		// remove gaps we can't fill
            		// removeGroupsInGap(gap);
            		// Do that later

                    if (debug) {
                        cleanupChain();
                        LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_final.pdb", newStructure.getChain(0));
                    }
                }
                else {
                    // don't care about it
                }
            }
            catch(Exception e) {
            	logger.error("Could not fill loop at " + gap.getStart() + "-" + gap.getEnd(), e);
            }
        }
    }
    
    /**
     * Fills a single gap. Assumes that the gap is neither beginning nor end gap.
     *
     * @param gap the gap to fill
     * @throws org.biojava.bio.structure.StructureException
     *                             If anything goes wrong with the structure
     * @throws HomodellerException If an optimal loop structure is not selected
     */
    private void fillGap(Gap gap) throws Exception {
        LoopDB db = LoopDB.getInstance();
        List<Loop> loops = db.getLoopsBySize(gap.getLoopCoilLength(this.querySequence));
        	// The "size" of the loops here is the number of coil residues in the loop, which should be
        	// equal to the number of residues we want to have between the pre and post residues.
        	// The pre and post residues are the two fixed residues spanning the section where we
        	// want to insert the loop of size gap.getLoopCoilLength(). Below we will call getCoilLoop()
        	// on the loop to get the coil residues plus the two flanking non-coil residues. These two
        	// flanking residues will correspond to the pre and post residues.
        
        DEBUG_LOOP_NUM++;
        
        logger.debug("");
        logger.debug("**************************");
        logger.debug("Inserting loop " + DEBUG_LOOP_NUM + " (size " + gap.getLoopCoilLength(this.querySequence) + ")"); // debug
        if( LoopUtil.isBeginning(gap) )
        	logger.debug("This is a beginning gap");
        else if( LoopUtil.isEnd(this.querySequence, gap) )
        	logger.debug("This is an end gap");
        logger.debug("**************************");
        logger.debug("");
        
        Group preResidue = gap.getFixedNTerminalGroup(this.querySequence, this.templateSequence, this.newStructure);
        //Atom preN_Atom = preResidue.getAtom("C");

        Group postResidue = gap.getFixedCTerminalGroup(this.querySequence, this.templateSequence, this.newStructure);
        //Atom postC_Atom = postResidue.getAtom("N");
        
        LoopFinder loopFinder = new CCDLoopFinder(this.querySequence, this.templateSequence, this.chiAngles, db, loops, preResidue, postResidue, gap);
        Loop bestLoop = loopFinder.getBestLoop();
        
        // Superimpose loop. I'm not sure if this is necessary after the superpositioning in
        // getBestLoopForCCD() -- David Arndt.
        Loop coilLoop = bestLoop.getCoilLoop();
        LoopUtil.superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
        
        if( debug )
        	LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_loop_CCD_" + DEBUG_LOOP_NUM + ".pdb", bestLoop.getCoilLoop());

        // debug
        logger.debug("\n   Writing newStructure_" + DEBUG_LOOP_NUM + ".pdb before minimizeDistances()...");
        if( debug )
        	LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure_" + DEBUG_LOOP_NUM + ".pdb", newStructure.getChain(0));
        
        // If it's not a beginning or end gap, then we need to minimize the distances between both ends' N-C atoms
        
        ArrayList<Group> groups = newStructure.getChain(0).getGroups();
        	
        /************************************************************************/
        /***               Attempt loop insertion with PepMake                ***/
        /************************************************************************/
        if( !this.runPepmake(gap, coilLoop, groups) ) {
        		
        	/************************************************************************/
        	/***                     Insert loop with Simplex                     ***/
        	/************************************************************************/
        		
        	// PepMake did not succeed in closing the loop. Fall back on Plan B, which is running
        	// Simplex to close the loop.
        	//
        	// This is the old method that was partly responsible for scrunched loops (poor selection
        	// of Pre and Post residues also contributed), but at least it always generates a
        	// structure.
        		
    		loopFinder = new SimplexLoopFinder(this.querySequence, this.templateSequence, this.chiAngles, db, loops, preResidue, postResidue, gap);
            bestLoop = loopFinder.getBestLoop();
            
            // Superimpose loop.
            coilLoop = bestLoop.getCoilLoop();
            if (preResidue != null)
            	LoopUtil.superImposeLoop(preResidue, coilLoop.getGroup(0), coilLoop);
            else {
                // pre residue will only be null if it's at the beginning
                assert LoopUtil.isBeginning(gap) : "Pre residue is null, not a loop at beginning";

                LoopUtil.superImposeLoop(postResidue, coilLoop.getGroup(coilLoop.getGroups().size() - 1), coilLoop);
            }
            
            this.runSimplex(gap, preResidue, postResidue, bestLoop, coilLoop, groups);
    	}

        // debug
        logger.debug("   Writing newStructure_min_" + DEBUG_LOOP_NUM + ".pdb after minimizeDistances()...");
        if( debug )
        	LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_newStructure_min_" + DEBUG_LOOP_NUM + ".pdb", newStructure.getChain(0));
        
    }
    
    private boolean runPepmake(Gap gap, Loop coilLoop, List<Group> groups) {
    	logger.debug("Attempting loop insertion with PepMake...");
    	
    	boolean pepmakeSucceeded = false;
    	
    	// Use PepMake to close the loop using Cyclic Coordinate Descent (CCD):
    	
		/*
		Create a structure suitable for input into PepMake.
		*/
		ArrayList<Group> coilLoopGroups = coilLoop.getGroups();
		//Structure pepmakeInputStructure = (Structure) newStructure.clone(); // results in NullPointerException
		Structure pepmakeInputStructure = new StructureImpl();
		
		Chain pepmakeInputChain = new ChainImpl();
		pepmakeInputStructure.addChain(pepmakeInputChain);
		
		for (int i = 0; i < groups.size(); i++) {
			pepmakeInputChain.addGroup(groups.get(i));
		}
		
		// Modify the groups in the PepMake input structure by substituting the coil
		// loop groups. 
		
		ArrayList<Group> pepmakeGroups = pepmakeInputChain.getGroups();
		
		int preOnQuery = this.querySequence.viewToSource(gap.getFixedNTerminalPosition());
		int postOnQuery = this.querySequence.viewToSource(gap.getFixedCTerminalPosition());
		
		// Modify the groups in the PepMake input structure by substituting the coil
		// loop groups. Coil loop includes the flanking residues, so start at 1 and
		// end at size -1. pepmakeGroups contains empty residues where there are gaps,
		// so we don't need to worry about adding or removing different numbers of
		// residues.
		for (int i = 1; i < coilLoopGroups.size() - 1; i++)
		{
			pepmakeGroups.set(preOnQuery + i - 1, coilLoopGroups.get(i));
				// -1 to convert from biological coordinates starting from 1 (from getPreOnQuery()) to
				// cooordinates starting from 0.
		}
		
		// Renumber the residues in the structure given to pepmake, starting from 1. This
		// is only for debugging purposes to prevent display glitches in PyMol when
		// examining the pepmake input and output structures.
		for (int i = 0; i < pepmakeGroups.size(); i++) {
			Group currentGroup = pepmakeGroups.get(i);
			currentGroup.setPDBCode(String.valueOf(i + 1));
		}
    	
        // debug
		logger.debug("   Writing pepmakeInput_" + DEBUG_LOOP_NUM + ".pdb...");
		if( debug )
			LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_pepmakeInput_" + DEBUG_LOOP_NUM + ".pdb", pepmakeInputStructure.getChain(0));
    	
        // Calculate offset for N- and C-anchors due to presence of empty residues.
        int anchorOffset = 0;
        for (int i = 0; i < preOnQuery + 1; i++) {
        	if (pepmakeGroups.get(i) instanceof EmptyResidue)
        		anchorOffset--;
        }
        
        PepmakeRunner.Options options = new PepmakeRunner.Options();
        options.closeGap = true;
        //options.res_id_N = Integer.parseInt(preResidue.getPDBCode()); // Optional.
        //options.res_id_C = Integer.parseInt(postResidue.getPDBCode());
        options.res_index_N = preOnQuery - 1 + anchorOffset;
        options.res_index_C = postOnQuery - 1 + anchorOffset;
        	// -1 to convert from biological coordinates starting from 1 to coordinates starting
        	// from 0.
        //options.gap_threshold = 0.0001;
        
        PepmakeRunner pepmakeRunner = new PepmakeRunner(pepmakeInputStructure, options);
        Structure pepmakeOutputStructure = pepmakeRunner.runPepmake();
        
        if (pepmakeOutputStructure != null) {
            // The PepMake input structure may have had empty residues to the right of the current gap.
            // But they were not output to PepMake and so were not read back in. So here we add the
            // empty residues to the PepMake output structure.
        	ArrayList<Group> pepmakeOutputGroups = pepmakeOutputStructure.getChain(0).getGroups();
            for (int i = 0; i < pepmakeGroups.size(); i++) {
            	// pepmakeGroups has the groups from just before being output to PepMake.
            	
            	if (pepmakeGroups.get(i) instanceof EmptyResidue) {
            		pepmakeOutputGroups.add(i, pepmakeGroups.get(i));
            	}
            }
        	
        	// Set newStructure to be the structure read in from PepMake.
        	this.newStructure = pepmakeOutputStructure;
        	pepmakeSucceeded = true;
        	
	        // debug
        	logger.debug("Loop insertion with PepMake succeeded!\n");
        	logger.debug("   Writing pepmakeOutput_" + DEBUG_LOOP_NUM + ".pdb...");
        	if( debug )
        		LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_pepmakeOutput_" + DEBUG_LOOP_NUM + ".pdb", pepmakeOutputStructure.getChain(0));
        }
        else {
        	logger.debug("Loop insertion with PepMake FAILED.\n");
        }
        
        return pepmakeSucceeded;
    }
    
    private void runSimplex(Gap gap, Group preResidue, Group postResidue, Loop bestLoop, Loop coilLoop, List<Group> groups) {
    	logger.debug("Inserting loop with Simplex...");
		
        if (debug)
        	LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_loop_simplex_" + DEBUG_LOOP_NUM + ".pdb", bestLoop.getCoilLoop());
        
        Chain rotatedLoop;

        if( !LoopUtil.isBeginning(gap) && !LoopUtil.isEnd(this.querySequence, gap)
        		&& !(postResidue instanceof EmptyResidue) && !(preResidue instanceof EmptyResidue))
            rotatedLoop = insertLoopInGap(preResidue, postResidue, bestLoop.getCoilLoop());
        else // should already be rotated correctly
            rotatedLoop = coilLoop;

        if (debug)
        	LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_rotatedLoop_" + DEBUG_LOOP_NUM + ".pdb", rotatedLoop);
        
        Chain finalLoop = rotatedLoop;
		
		finalLoop = minimizeDistances(preResidue, postResidue, rotatedLoop);
		
		/*
		Final step : insert loop residues into the structure
		Final loop includes the flanking residues, so start at 1, end at size - 1
		*/
		ArrayList<Group> loopGroups = finalLoop.getGroups();
		//ArrayList<Group> groups = newStructure.getChain(0).getGroups();
		for (int i = 1; i < loopGroups.size() - 1; i++) {
			groups.set(this.querySequence.viewToSource(gap.getFixedNTerminalPosition()) + i - 1, loopGroups.get(i));
				// -1 to convert from biological coordinates starting from 1 (from getPreOnQuery()) to
				// coordinates starting from 0.
		}
    }
    
    /**
     * Uses the coordinates of the pre and post residues (the residues flanking the gap) to calculate
     * the rotation matrix of the given chain, and translates it
     *
     * @param preResidue  residue flanking the gap on the left side (as per sequence)
     * @param postResidue residue flanking the gap on the right side (as per sequence)
     * @param loop        The loop whose coordinates are to be modified
     * @return the loop that is rotated and translated to fit into a gap
     */
    private Chain insertLoopInGap(Group preResidue, Group postResidue, Chain loop)
    {
        if (loop.getGroups().size() == 0)
        {
            throw new IllegalArgumentException("Why are you sending me an empty loop?");
        }

        Atom[] original = new Atom[6];
        Atom[] target = new Atom[6];

        try
        {
            /* target needs to be the N, CA, C atoms from the pre residue, and the same atoms
                from the post residue
             */
            target[0] = preResidue.getAtom("N");
            target[1] = preResidue.getAtom("CA");
            target[2] = preResidue.getAtom("C");
            target[3] = postResidue.getAtom("N");
            target[4] = postResidue.getAtom("CA");
            target[5] = postResidue.getAtom("C");

            /* Original needs to be the same atoms as target, but taken from the loop */
            ArrayList<Group> groups = loop.getGroups();
            Group preLoop = groups.get(0);
            Group postLoop = groups.get(groups.size() - 1);

            original[0] = preLoop.getAtom("N");
            original[1] = preLoop.getAtom("CA");
            original[2] = preLoop.getAtom("C");
            original[3] = postLoop.getAtom("N");
            original[4] = postLoop.getAtom("CA");
            original[5] = postLoop.getAtom("C");

            // do the translation
            Superpose.superpose(original, target, loop);
        }
        catch (StructureException e)
        {
            logger.error("Error getting atom", e);
        }

        return loop;
    }
    
    private Chain minimizeDistances(Group preResidue, Group postResidue, Chain loop) {
        SimplexRunner runner = new SimplexRunner(preResidue, postResidue, loop);
        Group preLoopResidue = loop.getGroup(1);
        Group postLoopResidue = loop.getGroup(loop.getGroups().size() - 2);

        Chain chainToMinimize = loop;

        try {
            double NC_dist = HomodellerUtils.getDistance(preResidue.getAtom("C"), preLoopResidue.getAtom("N"));
            double CN_dist = HomodellerUtils.getDistance(postResidue.getAtom("N"), postLoopResidue.getAtom("C"));

            int num_tries = 0;
            while ((NC_dist > MINIMIZATION_DISTANCE || CN_dist > MINIMIZATION_DISTANCE) && num_tries++ < MINIMIZATION_MAX_TRIES) {
                // set simplex's subject chain to minimize
                runner.setLoop(chainToMinimize);

                // minimize it
                chainToMinimize = runner.runSimplex();

                // superpose it on the structure again
                chainToMinimize = insertLoopInGap(preResidue, postResidue, chainToMinimize);

                if (chainToMinimize == null)
                    continue;

                preLoopResidue = chainToMinimize.getGroup(1);
                postLoopResidue = chainToMinimize.getGroup(chainToMinimize.getGroups().size() - 2);

                // re-calculate the distances
                CN_dist = HomodellerUtils.getDistance(preResidue.getAtom("C"), preLoopResidue.getAtom("N"));
                NC_dist = HomodellerUtils.getDistance(postResidue.getAtom("N"), postLoopResidue.getAtom("C"));

                logger.info(num_tries + ": CN_dist = " + CN_dist + ", NC_dist = " + NC_dist);
                if( debug )
                    LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_rotatedLoop.pdb", chainToMinimize);
            }
        }
        catch (StructureException e) {
            logger.error(e);
        }

        return chainToMinimize;
    }
    
    /**
     * Verifies the numbering scheme on the atoms and checks the atoms in each side chain
     * against a known side chain to avoid any missing atoms
     */
    private void cleanupChain() {
        Chain chain = newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();

        SideDB db = SideDB.getInstance();

        int atomNum = 1;
        for (int i = 0; i < groups.size(); i++)
        {
            Group g = groups.get(i);
            g.getPDBCode();

            // get an example of a side chain from our side chain database
            Character aminoType = ((AminoAcid) g).getAminoType();
            List<SideChain> sideChains = db.getSideChains(aminoType);

            // check to see that all atoms are there that are in the example side chain
            if (sideChains.size() > 0)
            {
                // go through each example atom and see if it's also in our residue
                SideChain example = sideChains.get(0);
                List<Atom> atoms = example.getAtoms();

                // see if the sizes are the same
                boolean allAtomsPresent = atoms.size() == g.getAtoms().size();

                if (allAtomsPresent)
                {
                    // check each atom
                    for (Atom a : atoms)
                    {
                        Atom atom = null;
                        try
                        {
                            atom = g.getAtom(a.getName());
                        }
                        catch (StructureException e)
                        {
                            e.printStackTrace();
                        }

                        // crap it's not there, we're missing it
                        if (atom == null)
                        {
                            logger.warn("Residue is missing atom " + a.getName());

                            allAtomsPresent = false;
                            break;
                        }
                    }
                }

                if (!allAtomsPresent && !"GLY".equals(g.getPDBName()) )
                {
                    // Somehow this group is missing sidechain atoms
                    // Re-insert this side chain
                    Angle angle = chiAngles.get(i);
                    double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                    // get a suitable side chain
                    SideChain newSideChain = HomodellerUtils.getSideChain(aminoType, chi);
                    // modify its coordinates to our protein
                    superImposeResidue(g, newSideChain);
                    // set the new residue in our chain
                    groups.set(i, (Group) newSideChain);
                    // g needs to be set to this group to be renumbered
                    g = groups.get(i);
                }
            }
            else if (aminoType != 'G')
            {
                logger.warn("Could not get example side chain for residue " + aminoType);
            }

            List<Atom> atoms = g.getAtoms();
            for (Atom a : atoms)
            {
                a.setPDBserial(atomNum++);
            }

            g.setPDBCode(String.valueOf(i + 1)); // residue numbering needs to be 1-based
        }
    }
    
    /**
     * Modifies the coordinates of the new side chain from the coordinates in the database
     * to the coordinates from the original residue
     *
     * @param original     the original residue whose coordinates we wish to modify
     * @param newSideChain the side chain whose coordinates need to be modified to the original's postion
     */
    private void superImposeResidue(Group original, SideChain newSideChain)
    {
        try
        {
            // For a superposition, we need two sets of coordinates
            // abc and pqr.  That will give us the translation and transformation of the new coordinates
            double[][] abc = SideChain.getReferenceCoordinates(original);
            double[][] pqr = SideChain.getReferenceCoordinates(newSideChain);

            List<Atom> atoms = newSideChain.getAtoms();

            // do the translation on these atoms
            for (Atom a : atoms)
            {
                double[] coordinates = Superpose.superpose(abc, pqr, a.getX(), a.getY(), a.getZ());
                a.setX(coordinates[0]);
                a.setY(coordinates[1]);
                a.setZ(coordinates[2]);
            }
        }
        catch (StructureException e) {
            logger.error("Error getting atom", e);
        }
    }
    
    /**
     * Modifies the templates residues so that they match the query sequence.
     * <p/>
     * Note: Empty residues are not changed.
     */
    public void modifyResidues() {
        assert this.querySequence.getSourceSymbolList().length() == this.newStructure.getChain(0).getSequence().length();

        assert chiAngles != null : "Chi angles from Phipsi are null";

        String query = this.querySequence.getSourceSymbolList().seqString();
        String database = this.newStructure.getChain(0).getSequence();

        ArrayList<Group> groups = this.newStructure.getChain(0).getGroups();

        for (int i = 0; i < query.length(); i++) {
            char q = query.charAt(i);
            char d = database.charAt(i);

            if (q != d) {
                if (q == 'G') {
                    // Glycine does not need a side chain, and is not in the database,
                    // so we need to remove all the side chain atoms
                    Group group = groups.get(i);
                    List<Atom> removals = new ArrayList<Atom>();
                    for (Atom a : (List<Atom>) group.getAtoms()) {
                        String name = a.getName();
                        if (!(name.equals("CA") || name.equals("C") || name.equals("O") || name.equals("N")))
                            removals.add(a);
                    }

                    for (Atom a : removals) {
                        group.getAtoms().remove(a);
                    }

                    try {
                        group.setPDBName("GLY");
                        ((AminoAcid) group).setAminoType('G');
                    }
                    catch (PDBParseException e) {
                        e.printStackTrace();
                    }
                    continue;
                }

                // ** need to modify this residue (chi angles are 1-based)
                // Because the chi angles calculation is based on the ungapped template (database),
                // convert the index to the index in the gapped template
                int index = Math.abs(this.templateSequence.viewToSource(i + 1));
                Angle angle = chiAngles.get(index);

                // TODO hack - angle should not be null (fails on 1TADB with template 1GIA)
                double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                SideChain newSideChain = HomodellerUtils.getSideChain(q, chi);

                assert newSideChain != null : "Can't find side chain for residue " + i + " " + q;

                if (!(groups.get(i) instanceof EmptyResidue)) {
                    superImposeResidue(groups.get(i), newSideChain);
                    
                    // put the new side chain into the model
                    groups.set(i, (Group) newSideChain.clone());
                }
                
                if(debug) {
                    logger.debug("Debugging side chains... (this should be off by default)");
                    cleanupChain();
                    LoopUtil.writeChainToFile("debug" + File.separator + this.querySequence.getName() + "_FROM_" + this.templateSequence.getName() + "_finalModified.pdb", newStructure.getChain(0));
                }
            }
        }

        String sequence = this.newStructure.getChain(0).getSequence();
        String querySequence = this.querySequence.getSourceSymbolList().seqString();
        
        // Check the sequence. We did not substitute sidechains in empty residues above, so we allow gaps in
        // the new structure's sequence.
        for (int i = 0; i < sequence.length(); i++) {
        	char s = sequence.charAt(i);
        	char q = querySequence.charAt(i);
        	
        	if ( !(s == q || ProteusTools.isGap(s)) ) {
                logger.debug(sequence);
                logger.debug(querySequence);
                throw new IllegalStateException("Wasn't able to modify the side chains to fit");
        	}
        }
    }
    
    /**
     * Removes empty residues from the beginning and end of newStructure.
     */
    private void removeBeginningAndEndGaps() {
        Chain chain = this.newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();
    	
        List<Integer> groupsToRemove = new ArrayList<Integer>();
        
        // Find empty residues at the beginning.
        boolean inBeginningGap = true;
        for (int i = 0; i < groups.size() && inBeginningGap; i++) {
        	
        	if (groups.get(i) instanceof EmptyResidue) {
        		groupsToRemove.add(i);
        	}
        	else {
        		inBeginningGap = false;
        	}
        }
        
        // Find empty residues at the end.
        boolean inEndGap = true;
        for (int i = groups.size() - 1; i >= 0 && inEndGap; i--) {
        	
        	if (groups.get(i) instanceof EmptyResidue) {
        		groupsToRemove.add(i);
        	}
        	else {
        		inEndGap = false;
        	}
        }
        
        Collections.sort(groupsToRemove);
        
        // Remove the residues.
        for (int i = groupsToRemove.size() - 1; i >= 0; i--)
        {
            int removeIndex = groupsToRemove.get(i);
            if (removeIndex < groups.size())
            {
                groups.remove(removeIndex);
            }
        }
    }
    
    /**
     * Removes empty residues from the interior of newStructure.
     */
    private void removeInteriorGaps() {
        Chain chain = this.newStructure.getChain(0);
        ArrayList<Group> groups = chain.getGroups();
        
        int firstRealResIndex = -1;
        boolean inBeginningGap = true;
        for (int i = 0; i < groups.size() && inBeginningGap; i++) {
        	
        	if ( !(groups.get(i) instanceof EmptyResidue) ) {
        		firstRealResIndex = i;
        		inBeginningGap = false;
        	}
        }
        
        if (firstRealResIndex == -1)
        	throw new IllegalStateException("No non-empty residues in structure.");
        
        boolean inEndGap = true;
        for (int i = groups.size() - 1; i >= firstRealResIndex; i--) {
        	
        	if (!inEndGap && groups.get(i) instanceof EmptyResidue) {
        		groups.remove(i);
        	}
        	else {
        		inEndGap = false;
        	}
    	}
    }
    
    /**
     * For right now, verifies that all C=0 bonds are proper.  This is to fix a bug where O atoms were being placed
     * alongside N atoms in an odd formation.  It will first try using a exact 90 degree angle placement of the O-atom,
     * and if that is still improper, will grab a new sidechain example and use the O-coordinate from that sidechain's
     * O atom.
     * Note: taking an example sidechain may be the proper way to do this.
     *
     * In future, this may be able to rotate side chains that bump with other side chains. 
     */
    private void checkBumps()
    {
        ArrayList<Group> groups = newStructure.getChain(0).getGroups();

        for (int i = 0; i < groups.size() - 1; i++)
        {
            Group N_Group = groups.get(i + 1);
            Group C_Group = groups.get(i);

            try
            {
                Atom N_atom = N_Group.getAtom("N");
                Atom O_atom = C_Group.getAtom("O");
                Atom C_atom = C_Group.getAtom("C");
                Atom CA_atom = C_Group.getAtom("CA");

                // check for bumps on moved fragment, between the N and O atom
                double N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);
                if (N_O_dist < N_O_MAX_DISTANCE)
                {
                    // too close, rotate C=O carboxyl to move away from N

                    // this is distance we want C=O to have
                    double C_O_dist = HomodellerUtils.getDistance(C_atom, O_atom);

                    double[] CA_coords = CA_atom.getCoords();
                    double[] N_coords = N_atom.getCoords();
                    double[] C_coords = C_atom.getCoords();

                    // find the point between the CA and N atom
                    double[] point = new double[3];
                    point[0] = (CA_coords[0] + N_coords[0]) / (double) 2;
                    point[1] = (CA_coords[1] + N_coords[1]) / (double) 2;
                    point[2] = (CA_coords[2] + N_coords[2]) / (double) 2;

                    // find the line between (point) and C, which we will extrapolate to be
                    // the position of the O
                    double[] line = new double[3];
                    line[0] = C_coords[0] - point[0];
                    line[1] = C_coords[1] - point[1];
                    line[2] = C_coords[2] - point[2];

                    // change line's magnitude to the magnitude of the distance between C=O
                    // i.e. |line| = C_O_dist
                    double multiplier = C_O_dist / Superpose.normalize(line);
                    line[0] *= multiplier;
                    line[1] *= multiplier;
                    line[2] *= multiplier;

                    // these are O's new coordinates
                    double O_coords[] = O_atom.getCoords();
                    O_coords[0] = C_coords[0] + line[0];
                    O_coords[1] = C_coords[1] + line[1];
                    O_coords[2] = C_coords[2] + line[2];

                    N_O_dist = HomodellerUtils.getDistance(N_atom, O_atom);

                    if ( N_O_dist < N_O_MAX_DISTANCE )
                    {
                        /* The distance still sucks.  Find the appropriate side chain and see if its oxygen atom
                         * will work.
                         */

                        Angle angle = chiAngles.get(i);
                        double chi = angle != null ? angle.getChi() : Angle.CHI_DEFAULT;

                        // get a suitable side chain
                        Character aminoType = ((AminoAcid) N_Group).getAminoType();
                        SideChain newSideChain = HomodellerUtils.getSideChain(aminoType, chi);
                        
                        // modify its coordinates to our protein
                        superImposeResidue(C_Group, newSideChain);

                        // Set the coordinates of this atom
                        Atom O = newSideChain.getAtom("O");
                        O_atom.setCoords(O.getCoords());
                    }
                }
            }
            catch (StructureException e)
            {
                logger.error("Error", e);
            }
        }
    }
}
