package ca.ualberta.proteus.program.minimization;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.biojava.PDBFileParser;
import org.apache.log4j.Logger;
import org.biojava.bio.structure.Structure;

import java.io.*;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: Jan 24, 2007
 * Time: 1:32:19 AM
 */
public class GAFolderRunner extends GenericProgramRunner implements EnergyMinimizer
{
    private final static int iterations = 20;
    private final static int timeout = 30; // minutes

    private static Logger logger = Logger.getLogger(GAFolderRunner.class);
    private Structure structure;


    public GAFolderRunner(Structure model)
    {
        this.structure = model;
    }

    public void run()
    {
        runGAFolder();
    }

    public Structure runGAFolder()
    {
        Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");
        String path = config.getStringValue("gafolder");
        try
        {
            // write PDB file to temp file
            File tempFile;
            tempFile = File.createTempFile("gafolder", ".in");
            writeStructureToFile(tempFile);
            tempFile.deleteOnExit();

            File outfile = File.createTempFile("gafolder", ".out");
            outfile.deleteOnExit();

            File dir = new File(path);
            assert dir.exists() : "GAFolder file specified does not exist";

            dir = dir.getParentFile().getAbsoluteFile();

            final String cmd = path + " -it " + iterations +
                    " -pdb " + tempFile.getPath() +
                    " -o " + outfile.getPath();

            //Process proc = Runtime.getRuntime().exec(cmd, null, dir);
            Process proc = ExecUtils.exec(cmd, dir, 1000, timeout * 60 * 1000); 

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(proc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(proc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = proc.waitFor();

            if (exitVal != 0)
            {
                System.err.println(cmd);
                System.err.println("GAFolder failed " + errorReader.getInputString());
                return null;
            }

            /** Wait for the output reader to completely do its thing **/
            while (!outputReader.isFinished())
            {
                Thread.sleep(10);
            }

            String output = outputReader.getInputString();
            logger.debug(output);

            BufferedReader reader = new BufferedReader(new FileReader(outfile));

            try
            {
                PDBFileParser parser = new PDBFileParser();
                return new ProteusStructureImpl( parser.parsePDBFile(reader) );
            }
            catch(Exception e)
            {
                logger.error("Problem parsing GAFolder output file");
                logger.error("Cmd: " + cmd);
            }

        }
        catch (IOException e)
        {
            logger.error("", e);
        }
        catch (InterruptedException e)
        {
            logger.error("", e);
        }
        /*catch (ParseException e)
        {
            logger.error("", e);
        }
*/
        return null;
    }

    private void writeStructureToFile(File tempFile)
            throws FileNotFoundException
    {
        PrintWriter writer = new PrintWriter(tempFile);
        writer.print(structure.toPDB());
        writer.close();
    }

    public Structure getMinimizedStructure()
    {
        return runGAFolder();
    }
}
