package ca.ualberta.proteus.program.minimization;

import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.program.homodeller.Gap;
import ca.ualberta.proteus.biojava.PDBFileParser;
import org.apache.log4j.Logger;
import org.biojava.bio.structure.Structure;

import java.io.*;

/**
 * A Class that abstracts the C-program pepmake.
 */
public class PepmakeRunner extends GenericProgramRunner implements EnergyMinimizer
{
    //private final static int iterations = 20;
    private final static int timeout = 20; // minutes
    
    private static Logger logger = Logger.getLogger(PepmakeRunner.class);
    private Structure structure;
    private Options options;


    /**
     * Runs pepmake on this structure
     * @param model
     * @param options
     */
    public PepmakeRunner(Structure model, Options options)
    {
        if ( options == null )
        {
            throw new IllegalArgumentException("Options parameter can't be null");
        }

        this.structure = model;
        this.options = options;
    }

    public void run()
    {
        runPepmake();
    }

    public Structure runPepmake()
    {
    	/*
        if ( options.gap != null )
        {
            // TODO Hack - clean up by unifying these 2 functions
            runPepmake(options.gap);
        }
        */
    	
        try
        {
            Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");
            String path = config.getStringValue("pepmake");

            // write PDB file to temp file
            File tempFile;
            tempFile = File.createTempFile("pepmake", ".in");
            writeStructureToFile(tempFile);
            tempFile.deleteOnExit();

            File outfile = File.createTempFile("pepmake", ".out");
            outfile.deleteOnExit();

            String cmd = path;
            
            if ( options.fixTorsionAngles )
            {
                cmd +=  " -t ";
            }
            
            if ( options.closeGap ) {
            	if (options.res_index_C == -1) {
            		throw new IllegalArgumentException(
            			"C-anchor residue not specified when closing gap with PepMake.");
            	}
            	
            	cmd += " -g ";
            	if (options.res_index_N != -1) {
            		cmd += options.res_index_N + " ";
            	}
            	cmd += options.res_index_C;
            	
            	if (options.gap_threshold != 0.0d) {
            		cmd += " -gt " + options.gap_threshold + " ";
            	}
            }
            
            cmd += " -pdb " + tempFile.getPath();

            return _runPepmake(path, cmd, outfile);
        }
        catch (IOException e)
        {
            e.printStackTrace();  
        }
        return null;
    }

    /**
     * Sets up pepmake to run based on a specific gap 
     *
     * @param gap
     * @return
     */
    /*
    private Structure runPepmake(Gap gap)
    {
        try
        {
            Config config = Config.getConfig(Consts.HOMODELLER, "config/homodeller.xml");
            String path = config.getStringValue("pepmake");

            // write PDB file to temp file
            File tempFile;
            tempFile = File.createTempFile("pepmake", ".in.pdb");
            writeStructureToFile(tempFile);
            //tempFile.deleteOnExit();

            File outfile = File.createTempFile("pepmake", ".out.pdb");
            //outfile.deleteOnExit();

            String cmd = String.format("%s -g %d %d -pdb %s", path, gap.getPreOnQuery(), gap.getPostOnQuery(),
                    tempFile.getPath());

            if ( options.fixTorsionAngles )
            {
                cmd +=  " -t ";
            }

            return _runPepmake(path, cmd, outfile);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        return null;
    }
    */

    /**
     * Actually runs the pepmake program
     *
     * @param path
     * @param cmd
     * @param outfile
     * @return
     */
    private Structure _runPepmake(String path, String cmd, File outfile)
    {
        logger.debug("Running Pepmake: " + cmd);

        // Total hack
        cmd += " -o " + outfile.getPath();

        try
        {
            File dir = new File(path);
            assert dir.exists() : "pepmake file specified does not exist";

            dir = dir.getParentFile().getAbsoluteFile();
            
            System.err.println(cmd); //debug
            
            //Process proc = Runtime.getRuntime().exec(cmd, null, dir);
            Process proc = ExecUtils.exec(cmd, dir, 1000, timeout * 60 * 1000);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(proc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(proc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = proc.waitFor();

            if (exitVal != 0)
            {
                System.err.println(cmd);
                System.err.println("pepmake failed " + errorReader.getInputString());
                return null;
            }

            /** Wait for the output reader to completely do its thing **/
            while (!outputReader.isFinished())
            {
                Thread.sleep(10);
            }
            
            String output = outputReader.getInputString();
            logger.debug(output);
            
            String error = errorReader.getInputString();
            logger.debug(error);
            
            if (this.options.closeGap && output.contains("Gap closure failed")) {
            	// PepMake ran stably, but could not close the gap due to geometric
            	// contraints.
            	return null;
            }
            else {
	            // Read the file back in
	            BufferedReader reader = new BufferedReader(new FileReader(outfile));
	
	            try
	            {
	                // Parse the new PDB file and return it
	                PDBFileParser parser = new PDBFileParser();
	                return new ProteusStructureImpl(parser.parsePDBFile(reader));
	            }
	            catch (Exception e)
	            {
	                logger.error("Problem parsing pepmake output file");
	                logger.error("Cmd: " + cmd);
	            }
            }

        }
        catch (IOException e)
        {
            logger.error("", e);
        }
        catch (InterruptedException e)
        {
            logger.error("", e);
        }

        return null;
    }

    private void writeStructureToFile(File tempFile)
            throws FileNotFoundException
    {
        PrintWriter writer = new PrintWriter(tempFile);
        writer.print(structure.toPDB());
        writer.close();
    }

    public Structure getMinimizedStructure()
    {
        return runPepmake();
    }

    public static class Options
    {
        public boolean fixTorsionAngles = false;

        /* Use this to call pepmake to fill in a gap */
        //public Gap gap;
        
        /*
         * To close a gap using PepMake, set the following parameters.
         * 
         * This will use the -g option from PepMake, described as follows:
         * 
         * -g [res_index_N] res_index_C
         *   Attempt to close a gap in the structure. e.g.
         * 
         *         -----------------------------N--------    C--------------------
         *                                      |            |                    
         *                                   N-anchor     C-anchor                
         * 
         *   res_index_C must be the index of the C-anchor, i.e. the residue just
         *   after the gap. res_index_N is the index of the N-anchor, which should be
         *   several residues to the left; if omitted, the program will try a few
         *   different N-anchor values. The "index" number is counted from the first
         *   residue in the peptide, which is 0. The N-anchor and C-anchor residues will
		 *   stay in fixed positions, and the residues in between will be adjusted --
		 *   except that slight distortion of backbone geometry may occur at the
		 *   C-anchor, especially if the gap threshold is set high with the -gt option.
		 *   The gap closure algorithm uses the Cyclic Coordinate Descent algorithm.
		 *   If gap closure fails, "Gap closure failed" will be printed to standard
		 *  output.
		 * -gid [res_id_N] res_id_C
		 *   Attempt to close a gap in the structure, using residue IDs to indicate
         *   residues.
         * -gt gap_threshold
         *   Allow the sum of squared distances threshold for gap closure to be relaxed
         *   up to gap_threshold. gap_threshold must be greater than the default strict
         *   threshold of 0.000001.
		 * 
		 * 
		 * PepMake should be given an entire PDB file, and that the residues between
		 * the N- and C-anchor residues must be built onto the N-anchor with proper
		 * backbone geometry.
         */
        public boolean closeGap = false;
        //public int res_id_N = RES_ID_NA; // Optional. RES_ID_NA means not specified.
        //public int res_id_C = RES_ID_NA;
        public int res_index_N = -1; // Optional. -1 means not specified.
        public int res_index_C = -1;
        public double gap_threshold = 0.0d; // 0 means not specified.
        
    }
}
