package ca.ualberta.proteus.program.proteus;

import ca.ualberta.proteus.program.MethodSettings;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.program.transsec.MainPredictor;
import ca.ualberta.proteus.program.transsec.PredictionScore;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.logging.ProteusLogger;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.db.HashSequenceDB;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * A wrapper for the JNET secondary structure predictor
 *
 * # Jnet needs 4 files:

   1. A FASTA format multiple sequence alignment
   2. A HMMER2 profile
   3. A PSIBLAST PSSM profile
   4. A PSIBLAST frequency profile

either a, a+b, or a+b+c+d can be applied.
# To carry out a prediction you will need to do the following:

   1. Take your target sequence (my.seq)
   2. Search sequence against database with PSIBLAST
      blastpgp -j3 -d swissprot -i my.seq > my.blast (psiblast 2.0.3)
      blastpgp -Q my.pos -m6 -b20000 -j3 -d swissprot -i my.seq > my.blast (psiblast 2.0.10)

   3. Fetch the sequences from the BLAST report with your favorite method. Call them my.seqs.
   4. Build a multiple sequence alignment from my.seqs with CLUSTALW (v1.7 or greater):
      clustalw -infile=my.seqs -outfile=my.msf -output=gcg -outorder=input
   5. If you have version 2.0.3 of PSIBLAST run the getfreq and getpssm perl scripts that come with the Jnet distribution on the BLAST report file
      getfreq my.blast
      getpssm my.blast
   6. If you have version 2.0.10 of PSIBLAST, thing are slightly different. You'll need the updated parsers from here. This extra step is due to changes in the PSIBLAST output. Once you have these then run:
      parse_psi -freq my.blast > my.blast.freq
      getpssm my.pos > my.blast.pssm
   7. Run: msf2jnet my.msf (this generates the FASTA file for Jnet)
      gethmm my.msf (this generates a HMM based profile for Jnet)
   8. Run Jnet:
      jnet -p my.msf.fa (for prediction from the alignment only)
      jnet -p my.msf.fa my.msf.hmmprof (for prediction without the PSIBLAST profiles)
      jnet -p my.fa my.msf.hmmprofmy.blast.pssm my.blast.freq (if you have the PSIBLAST profiles)


 */
public class JNETRunner extends SecondaryStructurePredictorRunner
{
    private String pssmFile;
    private String blastOutputFileName;
    private String alignmentFile;
    private ScoreAnnotatedSequence sequence;
    public Sequence querySequence;
    public Alignment alignment;

    ProteusLogger logger = ProteusLogger.getLogger();

    public void run()
    {
        this.finished = false;

        Config config = Config.getConfig();

        String runID = getUniqueIdentifier();
        getFreq(config, runID);
        getAlignment(config, runID);
        getPSSM(config, runID);
        getHMM(config, runID);
        try
        {
            try {
                runJNET(config, runID);
            }
            catch( ProcessNotCompleteException e ) {
                // This is a failsafe hack.  For some reason, the file handle seems to close every so often on JNET
                // so maybe trying it again will work?
                logger.debug("Running JNET Again...");
                try
                {
                    runJNET(config, runID, false);
                }
                catch(ProcessNotCompleteException e1)
                {
                    System.err.println("JNET Failed - try running again");
                }
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }

        cleanup(runID);
    }

    private void cleanup(String runID)
    {
        File f = new File(runID + ".hmm");
        f.delete();
        f = new File(runID + ".pssm");
        f.delete();
        f = new File(runID + ".freq");
        f.delete();                                     
        f = new File(runID + ".dnd");
        f.delete();
        f = new File(runID + ".msf");
        f.delete();
        f = new File(runID + ".align");
        f.delete();
        f = new File(runID + ".align.hmmprof");
        f.delete();
        f = new File("tmp.hmm");
        f.delete();
        f = new File("tmp.prof");
        f.delete();
    }

    private void getFreq(Config config, String runID)
    {
/*        getFreqFromXMLBlast(config, runID);
        if ( true) return;
  */

        String parsepsiCmd= config.getStringValue("parse_psi.command");
        parsepsiCmd += " -freq " + getBlastOutputFile();

        Process getFreqProc;
        try
        {
            getFreqProc = Runtime.getRuntime().exec("perl " + parsepsiCmd);

            // any error message?
            StreamHandler errorReader = new
                            StreamHandler(getFreqProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                            StreamHandler(getFreqProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = 0;
            try
            {
                exitVal = getFreqProc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if ( exitVal != 0 )
                logger.error("JNET - ParsePSI in getFreq failed " + errorReader.getInputString());

            try
            {
                FileOutputStream fos = new FileOutputStream( runID + ".freq");
                fos.write(outputReader.getInputString().getBytes());
                fos.close();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    private String getBlastOutputFile()
    {
        if ( blastOutputFileName == null )
            throw new IllegalStateException("Blast output file is not set - run blast and set this property before running");
        return blastOutputFileName;
    }

    private void getPSSM(Config config, String runID)
    {
        String pssmCmd = config.getStringValue("getpssm.command");
        pssmCmd += " " + getPssmFile();
        Process getPSSMProc;
        try
        {
            getPSSMProc = Runtime.getRuntime().exec("perl " + pssmCmd);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(getPSSMProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(getPSSMProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = 0;
            try
            {
                exitVal = getPSSMProc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if (exitVal != 0)
                logger.error("JNET - GetPSSM failed " + errorReader.getInputString());

            try
            {
                FileOutputStream fos = new FileOutputStream(runID + ".pssm");
                fos.write(outputReader.getInputString().getBytes());
                fos.close();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    private void getHMM(Config config, String runID)
    {
        String getHMMCmd= config.getStringValue("gethmm.command");

        // find the directory of this command
        File jnetFile = new File(getHMMCmd);
        assert jnetFile .exists() : "JNET file specified does not exist";
        File dir = jnetFile.getParentFile().getAbsoluteFile();

        String fileName = runID + ".msf";
        getHMMCmd += " " + fileName;


        Process getHMMProc;
        try
        {
            Runtime runtime = Runtime.getRuntime();
            //runtime.exec("perl " + getHMMCmd, null, dir);

            runtime.exec(dir + "/hmmbuild -F --fast --gapmax 1 --wblosum tmp.hmm " + fileName);
            runtime.exec(dir + "/hmmconvert -F -p tmp.hmm tmp.prof");
            getHMMProc = runtime.exec("perl " + dir + "/hmm2profile tmp.prof");

            // any error message?
            StreamHandler errorReader = new
                            StreamHandler(getHMMProc.getErrorStream(), "ERROR");

            // any output?
            FileOutputStream fos = new FileOutputStream( runID + ".hmm");
            StreamHandler outputReader = new
                            StreamHandler(getHMMProc.getInputStream(), "OUTPUT", fos);

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = 0;
            try
            {
                exitVal = getHMMProc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if ( exitVal != 0 )
                logger.error("JNET - GetHMM failed " + errorReader.getInputString());
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    private void getAlignment(Config config, String runID)
    {
        // TODO cleanup
        try
        {
            FileOutputStream fos = new FileOutputStream(runID + ".align");
            SequenceDB db = new HashSequenceDB();

            // query has to go on top of the file
            GappedSequence query = (GappedSequence) alignment.symbolListForLabel("QUERY");
            SeqIOTools.writeFasta(fos, new SimpleSequence(query, "", query.getName(), query.getAnnotation()));

            for (Iterator sequences = alignment.symbolListIterator(); sequences.hasNext(); )
            {
                Sequence s = (Sequence) sequences.next();
                if (s.getName().equals("QUERY"))
                {
                    continue;
                }
                SymbolList ungapped = ((GappedSequence) s).getSourceSymbolList();
                db.addSequence( new SimpleSequence(ungapped, "", s.getName(), s.getAnnotation()));
            }
            SeqIOTools.writeFasta(fos, db);
            fos.close();
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (IllegalIDException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (BioException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }


        String clustalWCmd = config.getStringValue("clustalw.command");
        clustalWCmd +=  " -infile=" + runID + ".align -outfile=" + runID + ".msf -output=gcg -outorder=input";

        try
        {
            Process clustalWProc = Runtime.getRuntime().exec(clustalWCmd);

            // any error message?
            StreamHandler errorReader = new
                            StreamHandler(clustalWProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                            StreamHandler(clustalWProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = 0;
            try
            {
                exitVal = clustalWProc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if ( exitVal != 0 )
                logger.error("JNET - clustalW failed " + errorReader.getInputString());

        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        String msf2JNETCmd = config.getStringValue("msf2jnet.command");
        msf2JNETCmd += " " +  runID + ".msf";

        try
        {
            Process msf2jnetProc = Runtime.getRuntime().exec(msf2JNETCmd);

            // any error message?
            StreamHandler errorReader = new
                            StreamHandler(msf2jnetProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                            StreamHandler(msf2jnetProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = 0;
            try
            {
                exitVal = msf2jnetProc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if ( exitVal != 0 )
                logger.error("JNET - msf2jnet failed " + errorReader.getInputString());
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        // rename the file
        new File(runID + ".msf.fa").renameTo(new File(runID + ".align"));
    }

    private void runJNET(Config config, String runID) throws IOException, BioException
    {
        runJNET(config, runID, true);
    }

    private void runJNET(Config config, String runID, boolean useAll) throws IOException, BioException
    {
        String jnetCmd = config.getStringValue("jnet.command");

        if ( useAll)
        {
            jnetCmd += String.format(" -z %s %s %s %s", runID + ".align", runID + ".hmm", runID + ".pssm", runID + ".freq" );
        }
        else
        {
            // Failsafe mode, try with just the alignment, at least it shouldn't seg fault
            jnetCmd += String.format(" -z %s ", runID + ".align" );
        }

        Process jnetProc = Runtime.getRuntime().exec(jnetCmd);

        //OutputStream fileout = new FileOutputStream(runID + "jnet.out");

        // any error message?
        StreamHandler errorReader = new
                        StreamHandler(jnetProc.getErrorStream(), "ERROR");

        // any output?
        StreamHandler outputReader = new
                        StreamHandler(jnetProc.getInputStream(), "OUTPUT");

        // kick them off
        errorReader.start();
        outputReader.start();

        int exitVal = 0;
        try
        {
            exitVal = jnetProc.waitFor();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }

        //fileout.close();

        if ( exitVal != 0 )
        {
            logger.error("JNET failed with command: " + jnetCmd);
            logger.error( errorReader.getInputString() + outputReader.getInputString());
        }

        String line;
        StringBuffer newStructure = new StringBuffer();
        StringBuffer sequenceStr = new StringBuffer();
        List<PredictionScore> scoresList = new ArrayList<PredictionScore>();
        StringBuilder jnetOutputFile = new StringBuilder();

        BufferedReader reader = new BufferedReader(new StringReader( outputReader.getInputString() ));
        //BufferedReader reader = new BufferedReader(new FileReader(runID + "jnet.out"));
        //reader.readLine(); // skip first line

        // wait for it to be ready
        while ( ! reader.ready() )
        {
            try
            {
                Thread.sleep(10);
            }
            catch (InterruptedException e)
            {
                e.printStackTrace(); 
            }
        }

        while( ( line = reader.readLine() ) != null ) {
            jnetOutputFile.append(line).append("\n");

            if ( line.contains("PRED"))
                continue;

            line = line.replace('*', ' ');

            final String[] strings = line.split("\\s+");

            char structure = strings[1].charAt(0);
            if ( structure == '-' ) structure = 'C' ;
            newStructure.append( structure );

            sequenceStr.append( strings[0] );

            // build the score
            double d = Double.parseDouble(strings[7]) / (double) 10;
            switch(structure) {
                case 'C':
                    scoresList.add( new PredictionScore( d, 0, 0 ));
                    break;
                case 'E':
                    scoresList.add( new PredictionScore( 0, d, 0 ));
                    break;
                case 'H':
                    scoresList.add( new PredictionScore( 0, 0, d ));
                    break;
            }
        }

        assert( reader.readLine() == null );

        reader.close();

        PredictionScore[] scores = new PredictionScore[scoresList.size()];
        sequence = new ScoreAnnotatedSequence(sequenceStr.toString(), newStructure.toString(), querySequence.getName(), scoresList.toArray(scores));

        if ( sequenceStr.length() != querySequence.length() )
        {
            //logger.error("PSIPRED output: " + psipredOutput);
            logger.error("Output file: " + jnetOutputFile.toString());
            logger.error("Sequence string: " + sequenceStr);
            logger.error("Query sequence:  " + querySequence.seqString());
            System.err.flush();
            logger.debug("Result doesn't match sequence!");
            throw new ProcessNotCompleteException("Result doesn't match sequence!");
        }
    }

    private String getAlignmentFile()
    {
        return alignmentFile;
    }

    public ScoreAnnotatedSequence getPrediction()
    {
        return sequence;
    }

    /**
     * Gets the blast position matrix file name given by the -Q option to 'blastpgp'
     */
    public String getPssmFile()
    {
        if ( pssmFile == null )
            throw new IllegalStateException("Blast output file is not set - run blast and set this property before running");
        return pssmFile;
    }

    /**
     * Sets the blast position matrix file name given by the -Q option to 'blastpgp'
     * and the output file from BLAST
     * You need to set this before running JNET
     * @param pssmFile
     * @param alignmentFile - this file SHOULD have the query sequence as the first sequence in it
     * @param blastOutputFileName
     */
    public void setParams(String pssmFile, String alignmentFile, String blastOutputFileName )
    {
        this.pssmFile = pssmFile;
        this.alignmentFile = alignmentFile;
        this.blastOutputFileName = blastOutputFileName;
    }

    public static void main(String[] args) throws IllegalIDException
    {
        Sequence seq = SeqSeeDB.getTestSequence();
        MainPredictor mp = new MainPredictor(null, MethodSettings.JNET);
        mp.run(seq, false);
    }


    private void getFreqFromXMLBlast(Config config, String runID)
    {
        //String file = getBlastOutputFile();

        List<String> sequences = new ArrayList<String>();
        for ( Iterator<Sequence> it = alignment.symbolListIterator(); it.hasNext(); )
        {
            Sequence seq = it.next();
       //     logger.error(seq.seqString());
            sequences.add(seq.seqString());
        }


        int length = sequences.get(0).length();

        // alignments tend to have long regions that are just gaps, trim it
        int lastNonGap = 0;
        boolean [] allGaps = new boolean[length];   // keeps track of how many all-gap alignments there are
        int numAllGaps = 0;
        // (e.g.
        /*
            That's what I mean

            AAAA-AAA
            AAAA-AAA
            AAAA-AAA
        */
        for ( int i=0; i<length; i++)
        {
            boolean allGap = true;
            for (String sequence1 : sequences)
            {
                char c = sequence1.charAt(i);

                if (c != '-' && c != '~')
                {
                    lastNonGap++;
                    allGap = false;
                    break;
                }
            }

            allGaps[i] = allGap;
            if ( allGap ) numAllGaps++;
        }

        //length -= numAllGaps;

        double[][] map = new double[length][];

        for ( int i=0; i<length; i++)
        {
            map[i] = new double[21];

           // System.out.print( (i+1) + " ");
            for (String sequence1 : sequences)
            {
                char c = sequence1.charAt(i);

            //    System.out.print(c + " ");
                int idx = charToInt(c);
                if (idx < 20)
                    map[i][idx]++;
            }
       //     logger.error();
        }


        try

        {

            PrintWriter fos = new PrintWriter(new FileOutputStream( runID + ".freq") );

            for ( int i=0; i<length; i++)
            {
                // don't print it out if this residue list is all gaps
                if ( allGaps[i] )
                    continue;

                int total = 0;
                for ( int j=0; j<20; j++)
                    total += map[i][j];
                for ( int j=0; j<20; j++)
                    map[i][j] = total != 0 ? map[i][j] * 10 / total : 0;
                for ( int j=0; j<20; j++)
                {
                  //  System.out.printf("%2d ", (int) Math.round(map[i][j]));
                    fos.printf("%2d ", (int) Math.round(map[i][j]));
                }
                fos.println();
               /* logger.error();
                System.out.print(i);
                System.out.print(sequences.get(0).charAt(i) + " ");
                */
            }

            fos.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    private int charToInt(char c)
    {
        switch (c)
        {
            case 'A':
                return 0;
            case 'R':
                return 1;
            case 'N':
                return 2;
            case 'D':
                return 3;
            case 'C':
                return 4;
            case 'Q':
                return 5;
            case 'E':
                return 6;
            case 'G':
                return 7;
            case 'H':
                return 8;
            case 'I':
                return 9;
            case 'L':
                return 10;
            case 'K':
                return 11;
            case 'M':
                return 12;
            case 'F':
                return 13;
            case 'P':
                return 14;
            case 'S':
                return 15;
            case 'T':
                return 16;
            case 'W':
                return 17;
            case 'Y':
                return 18;
            case 'V':
                return 19;
            case 'X':
            case '-':
                return 25;
        }
        throw new IllegalArgumentException( c + " is not a valid character");
    }
}

