package ca.ualberta.proteus.program.thrifty;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.GenericProgramRunner;
import ca.ualberta.proteus.util.ProcessNotCompleteException;
import ca.ualberta.proteus.util.StreamHandler;
import ca.ualberta.proteus.util.XAlignParser;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.Alignment;
import org.xml.sax.InputSource;

import java.io.*;
import java.util.Arrays;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Apr 9, 2005
 *         Time: 6:26:10 PM
 *         To change this template use File | Settings | File Templates.
 */
public class FastXAlignRunner extends GenericProgramRunner
{
    private StructureAnnotatedSequence query;
    private Sequence template;
    private Alignment alignment;
    private Map<String, Map<String, Double>> ppshMap;

    public FastXAlignRunner(StructureAnnotatedSequence query, Sequence template)
    {
        this.query = query;
        this.template = template;
    }

    public Alignment getAlignment()
    {
        if (alignment == null)
        {
            throw new ProcessNotCompleteException("XAlign has not finished running yet or there was an error assigning the alignment!");
        }
        return alignment;
    }

    public void run()
    {
        run(false);
    }

    public void run(boolean isRecursiveCall)
    {
       this.finished = false;

        try
        {
            String fname = getUniqueIdentifier();
            String infile = fname + ".db";

            File file;
            file = writeInputFile(infile);

            String outfile = fname + ".out";

            String xalignCmd = Config.getConfig().getStringValue("xalign.command");

            try
            {
                File xalignProg = new File(xalignCmd);
                assert xalignProg.exists() : "XAlign file specified does not exist";

                File dir = xalignProg.getParentFile().getAbsoluteFile();

                xalignProg = new File("");
                String curdir = xalignProg.getAbsolutePath() + "/";

                infile = curdir + infile;
                outfile = curdir + outfile;

                String[] cmdArgs = new String[]{"perl", xalignCmd, "-d", infile, "-o", outfile, "> tmp.out"};

                Process xalignProc = Runtime.getRuntime().exec(cmdArgs, null, dir);

                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(xalignProc.getErrorStream(), "ERROR");

                // any output?
                StreamHandler outputReader = new
                        StreamHandler(xalignProc.getInputStream(), "OUTPUT");

                // kick them off
                errorReader.start();
                outputReader.start();

                // This is a sequece of commands that has to be sent to XAlign's stdin
                //String xalignCommands = "1\n1\n1\n";
                BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(xalignProc.getOutputStream()));
                writer.write("1\n");
                writer.write("1\n");
                writer.write("1\n");
                writer.write("1\n\n\n");
                //xalignProc.getOutputStream().write(xalignCommands.getBytes());

                try
                {
                    // any error???
                    int exitVal = xalignProc.waitFor();
                   // System.out.println("XAlign ExitValue: " + exitVal);

                    if (exitVal != 0)
                    {
                        System.out.println("XAlign ExitValue: " + exitVal);
                        System.out.println("XAlign failed");
                        System.out.println("Cmd: " + xalignCmd + "\n Dir: " + dir);
                        System.out.println("perl " +  xalignCmd +  " -d "  + infile + " -o " +  outfile);
                        System.out.println(errorReader.getInputString());
                        System.out.println(outputReader.getInputString());

                        if ( !isRecursiveCall)
                        {
                            System.out.println("Running XAlign Again...");
                            run(true);
                        }

                        return;
                    }

                    XAlignParser oParser = new XAlignParser();

                    try
                    {
                        // parse it

                        FileInputStream output = new FileInputStream(outfile);
                        oParser.parse(new InputSource(output));

                        this.alignment = oParser.getAlignment();
                        this.ppshMap = oParser.getPPSHMap();
                    }
                    catch(Exception e)
                    {
                        handleError(xalignCmd, dir, isRecursiveCall);
                        e.printStackTrace();
                    }

                }
                catch (IOException e)
                {
                    handleError(xalignCmd, dir, isRecursiveCall);
                    e.printStackTrace();
                }
            }
            finally
            {
                file.delete();
                new File(outfile).delete();
            }
        }
        catch (Throwable t)
        {
            if ( !isRecursiveCall)
            {
                System.out.println("Running XAlign Again...");
                run(true);
            }

            t.printStackTrace();
        }
        finally
        {
            this.finished = true;
        }
    }

    private void handleError(String xalignCmd, File dir, boolean isRecursiveCall)
            throws IOException
    {
        System.out.println("Error running XAlign");
        System.out.println("Cmd: " + xalignCmd);

        // print out log file
        BufferedReader r = new BufferedReader(new FileReader(dir.getAbsolutePath() + "/xalign.log"));
        String line;
        while ((line = r.readLine()) != null)
        {
            System.out.println(line);
        }
        //System.exit(-1);

        if ( !isRecursiveCall)
        {
            System.out.println("Running XAlign Again...");
            run(true);
        }
    }


    private File writeInputFile(String infile)
            throws IOException
    {
        // write the XAlign input file
        File f = new File(infile);
        f.deleteOnExit();

        PrintWriter pw = new PrintWriter(new FileWriter(f));

        try
        {
            writeSequence(query, pw);
        }
        catch(AssertionError e)
        {
            throw new ProcessNotCompleteException(e);
        }
        try
        {
            writeSequence(template, pw);
        }
        catch(AssertionError e)
        {
            throw new ProcessNotCompleteException(e);
        }

        pw.close();

        return f;
    }

    private void writeSequence(Sequence sequence, PrintWriter pw)
    {
        pw.println(">" + sequence.getName());

        String seqStr = sequence.seqString();
        String structStr = null;

        if (sequence instanceof StructureAnnotatedSequence)
        {
            structStr = ((StructureAnnotatedSequence) sequence).getStructureAnnotation().seqString();
        }

        if ( structStr != null )
        {
            /* HACK: Vadar and my 9-state algorithm differ slightly, i get rid of bad trailing residues, where
             * as Vadar doesn't.  So sometimes we have a 1-residue length difference.
             */
            if ( seqStr.length() != structStr.length() )
            {
                if ( seqStr.length() - structStr.length() == 1)
                {
                    structStr += "c";
                }
            }

            assert seqStr.length() == structStr.length(): "Different lengths of sequence and structure";
        }

        for (int i = 0; i < sequence.length(); i += 50)
        {
            String line = seqStr.substring(i, Math.min(i + 50, seqStr.length()));
           // line = line.replaceAll("(\\w{10})", "$1 ");

            // print the dashes
            char[] c = new char[line.length()];
            Arrays.fill(c, 0, line.length() , '-');
            pw.println(c);

            // then teh sequence
            pw.println(line);

            // then the structure
            if ( structStr != null)
            {
                line = structStr.substring(i, Math.min(i + 50, seqStr.length()));
                line = line.toUpperCase().replace("E", "B");
               // line = line.replaceAll("(\\w{10})", "$1 ");
                pw.println(line);
            }
        }

        pw.println();
    }

    public Map<String, Map<String, Double>> getPpshMap()
    {
        return ppshMap;
    }


}
