package ca.ualberta.proteus.program.thrifty;

import ca.ualberta.proteus.program.homodeller.Angle;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.BlastRunner;
import ca.ualberta.proteus.util.BlastSettings;
import ca.ualberta.proteus.util.Consts;
import ca.ualberta.proteus.util.ProteusTools;
import org.apache.log4j.Logger;
import org.biojava.bio.Annotation;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.impl.SimpleSequenceFactory;
import org.biojava.bio.structure.Structure;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.SimpleAlignment;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.io.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Oct 5, 2006
 * Time: 4:10:11 PM
 * To change this template use File | Settings | File Templates.
 */
public class Thrifty
{
    private StringBuilder originalMatrix;
    private Config config = Config.getConfig(Consts.THRIFTY, "config/thrifty.xml");
    private static Logger logger = Logger.getLogger(Thrifty.class);

    private int GAP_PENALTY = 5;
    private int GAP_EXTENSION_PENALTY = 2;

    public static void main(String[] args)
    {

    }

    /**
     * Run homodeller first to find a database match, then do the loop
     * insertions
     * @param sequence The sequence to run
     * @return result containing the finished model, or null if an error occured
     */
    public Result runThrifty( Sequence sequence, String filename ) throws IOException
    {
        return runThrifty(sequence.seqString(), filename);
    }

    /**
     * Run homodeller first to find a database match, then do the loop
     * insertions
     * @param sequence The sequence to run
     * @return a result containing the finished model, or null if an error occured
     */
    public Result runThrifty( String sequence, String filename ) throws IOException
    {
        // Run Preditor

        PreditorRunner runner = new PreditorRunner();
        runner.setFilename(filename);
        runner.run();
        File resultFile = runner.getResultFile();

        // run homodeller on the
        return runThrifty(sequence, resultFile);
    }

    /**
     * Run homodeller first to find a database match, then do the loop
     * insertions
     * @param sequence The sequence to run
     * @return a result containing the finished model, or null if an error occured
     */
    public Result runThrifty( String sequence, File phiPsiFile ) throws IOException
    {
        Map<Integer, Angle> angles = parsePhiPsi(phiPsiFile);

        assert sequence.length() == angles.size() : "Sequence length and phi psi angles are not the same length";

        StructureAnnotatedSequence structure = AngleConverter.convertToSecondaryStructure(sequence, angles);

        try
        {
            replaceBLAST(false);

            BlastSettings settings = new BlastSettings();
            Sequence structureSeq = new SimpleSequenceFactory().createSequence(
                    structure.getStructureAnnotation(),
                    "", "STRUCTURE", new SimpleAnnotation());
            settings.setSequence(structureSeq);

            String matrix = config.getStringValue("structureMatrix");
            settings.setMatrix(matrix);

            settings.setBlasttype(BlastSettings.BLASTTYPE.PROTEUS_BLASTPARSER_BLASTP);

            settings.setDatabase(config.getStringValue("9-structure.db"));
            settings.setGapPenalty(GAP_PENALTY);
            settings.setGapExtensionPenalty(GAP_EXTENSION_PENALTY);

            BlastRunner blast = new BlastRunner(settings);
            blast.run();

//            List<Sequence> results = blast.getUnalignedResult();
            List<GappedSequence> results = blast.getAlignedResult();

            Collections.sort(results, new Comparator<GappedSequence>(){

                public int compare(GappedSequence o1, GappedSequence o2)
                {
                    Object score1 = o1.getAnnotation().getProperty("score");
                    Object score2 = o2.getAnnotation().getProperty("score");
                    if ( score1 != null
                            && score2 != null )
                    {
                        return (Integer) score2 - (Integer) score1;
                    }
                    return 1000;
                }
            });

            blast.cleanupTempFiles();

            Sequence bestByEvalue = null;
            Sequence bestByScore = null;
            Sequence query = null;
            for ( Sequence s: results)
            {
                Integer score = (Integer) s.getAnnotation().getProperty("score");
                Double evalue = (Double) s.getAnnotation().getProperty("e-value");
                System.out.printf("%s (%d): %s\n", s.getName(), score, s.seqString());

                if ( bestByEvalue == null )
                {
                    bestByEvalue = s;
                }
                else
                {
                    Annotation annotation = bestByEvalue.getAnnotation();
                    if ( evalue != null
                        && annotation.containsProperty("e-value")
                        && (Double) annotation.getProperty("e-value") > evalue)
                    {
                        bestByEvalue = s;
                    }
                }

                if ( bestByScore == null )
                {
                    bestByScore = s;
                }
                else{
                    Annotation annotation = bestByScore.getAnnotation();
                    if (    score != null
                            && annotation.containsProperty("score")
                            && (Integer) annotation.getProperty("score") < score)
                    {
                        bestByScore = s;
                    }
                }

                if ( s.getName().equals("QUERY"))
                {
                    // clone it
                    try
                    {
                        query =  ProteinTools.createGappedProteinSequence(s.seqString(), s.getName());
                    }
                    catch (IllegalSymbolException e)
                    {
                        // you better not see this
                        e.printStackTrace();
                    }
                }

            }

            System.out.println("QUERY: " + query);
            System.out.printf("Best score: %s (%d): %s\n", bestByScore.getName(), (Integer) bestByScore.getAnnotation().getProperty("score"), bestByScore.seqString());
            System.out.printf("Best evalue: %s (%g): %s\n", bestByEvalue.getName(), (Double) bestByEvalue.getAnnotation().getProperty("e-value"), bestByEvalue.seqString());

            Map<String, Sequence> alignmentList = new HashMap<String, Sequence>();
            alignmentList.put(query.getName(), query);
            alignmentList.put(bestByScore.getName(), bestByScore);
            alignmentList.put(bestByEvalue.getName(), bestByEvalue);

            Alignment alignment = new SimpleAlignment(alignmentList);

            for ( Iterator iterator = alignment.symbolListIterator(); iterator.hasNext(); )
            {
                Sequence s = (Sequence) iterator.next();
                System.out.println(s.getName() + ": " + s.seqString());
            }

            alignment = ProteusTools.reduceAlignment(alignment);

            for ( Iterator iterator = alignment.symbolListIterator(); iterator.hasNext(); )
            {
                Sequence s = (Sequence) iterator.next();
                System.out.println(s.getName() + ": " + s.seqString());
            }

            // create small alignments for them all
            for ( Sequence s: results)
            {
                alignmentList.clear();

                alignmentList.put(query.getName(), query);
                alignmentList.put(s.getName(), s);

                alignment = new SimpleAlignment(alignmentList);
                alignment = ProteusTools.reduceAlignment(alignment);

                for ( Iterator iterator = alignment.symbolListIterator(); iterator.hasNext(); )
                {
                    Sequence s2 = (Sequence) iterator.next();
                    System.out.println(s2.getName() + ": " + s2.seqString());
                }
            }


        }
        finally
        {
            replaceBLAST(true);
        }

        return null;
    }

    private LinkedHashMap<Integer,Angle> parsePhiPsi(File phiPsiFile)
    {
        LinkedHashMap<Integer,Angle> rv = new LinkedHashMap<Integer, Angle>();

        try
        {
            BufferedReader reader = new BufferedReader(new FileReader(phiPsiFile));

            String line;
            while ((line = reader.readLine()) != null)
            {
                if ( line.startsWith("Res#"))
                {
                    break;
                }
            }

            while ((line = reader.readLine()) != null)
            {
                Angle angle = new Angle();

                String[] strings = line.split("\\s+");

                angle.setPhi(Double.parseDouble(strings[2]));
                angle.setPsi(Double.parseDouble(strings[4]));

                rv.put(Integer.parseInt(strings[0]), angle);
            }
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        return rv;
    }

    /**
     * @param original True to replace it with the original, false to replace it with the temporary matrix
     */
    private void replaceBLAST(boolean original) throws IOException
    {
        File blastDataDir = new File(System.getenv("BLASTMAT"));

        if ( ! blastDataDir.exists())
        {
            throw new IllegalStateException("Blast data directory does not exist: BLASTMAT = " + System.getProperty("BLASTMAT"));
        }
        String matrix = config.getStringValue("structureMatrix");

        File matrixFile = new File(blastDataDir, matrix);

        if (original)
        {
            // write the original
            PrintWriter out = new PrintWriter(matrixFile);

            assert originalMatrix != null : "Original matrix was never assiged";

            out.print(originalMatrix.toString());
            out.close();
        }
        else
        {
            // write the new one and cache the original
            BufferedReader reader = new BufferedReader(new FileReader(matrixFile));

            originalMatrix = new StringBuilder();
            String line;
            reader.mark(1024);

            line = reader.readLine();

            if ( line == null )
            {
                logger.error(matrixFile + " is empty.. recovering");

                matrix = config.getStringValue("backupStructureMatrix");
                matrixFile = new File(blastDataDir, matrix);
                reader = new BufferedReader(new FileReader(matrixFile));

                logger.info("Using backup matrix file: " + matrixFile);
            }
            else
            {
                reader.reset();
            }

            while ( ( line = reader.readLine()) != null )
            {
                originalMatrix.append(line).append('\n');
            }
            reader.close();

            PrintWriter out = new PrintWriter(matrixFile);

            // create the new matrix
            MatrixEditor editor = new MatrixEditor(originalMatrix);
            editor.modifyTo9StateMatrix();

            // write the matrix
            out.println(editor.toString());
        }
    }

    /**
     * A homology modelling result from homodeller
     */
    public static class Result
    {
        // the completed model
        public Structure model;

        // the template structure used
        public Structure template;
        public String templateName;
        public GappedSequence query;
        public GappedSequence database;

        public Result(Structure model, Structure template, GappedSequence query, GappedSequence database, String title)
        {
            this.template = template;
            this.database = database;
            this.query = query;
            this.model = model;
            this.templateName = title;
        }
    }
}
