package ca.ualberta.proteus.program.tmhmm;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.GenericProgramRunner;
import ca.ualberta.proteus.util.StreamHandler;
import org.apache.log4j.Logger;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.SymbolList;

import java.io.*;
import java.util.Arrays;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Jan 18, 2007
 * Time: 10:50:22 PM
 * To change this template use File | Settings | File Templates.
 */
public class TMHMMRunner extends GenericProgramRunner
{
    private Sequence sequence;
    private Logger logger = Logger.getLogger(TMHMMRunner.class);

    public TMHMMRunner(Sequence sequence)
    {
        this.sequence = sequence;
    }

    public void run()
    {
        runTMHMM();
    }

    public StructureAnnotatedSequence runTMHMM()
    {
        Config config = Config.getConfig();
        String path = config.getStringValue("tmhmm");

        // Keep a record of TMHMM directories present at the start of the process
        // as they may be in use by another process
        File[] originalFiles = new File("").getAbsoluteFile().listFiles(new TMHMM_FileFilter());

        try
        {
            // write PDB file to temp file
            File tempFile;
            tempFile = File.createTempFile("tmhmm", ".in");
            tempFile.deleteOnExit();

            FileOutputStream blastfile;
            blastfile = new FileOutputStream(tempFile);
            SeqIOTools.writeFasta(blastfile, sequence);
            blastfile.close();

            String cmd = "perl " + path + " -short " + tempFile.getPath();

            Process proc = Runtime.getRuntime().exec(cmd);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(proc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(proc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            int exitVal = proc.waitFor();

            if (exitVal != 0)
            {
                System.err.println(cmd);
                System.err.println("TMHMM failed " + errorReader.getInputString());
                return null;
            }

            /** Wait for the output reader to completely do its thing **/
            while (!outputReader.isFinished())
            {
                Thread.sleep(10);
            }

            String phipsiOutput = outputReader.getInputString();

            BufferedReader reader = new BufferedReader(new StringReader(phipsiOutput));

            try
            {
                TMHMMReader hmmreader = new TMHMMReader();
                SymbolList symbolList = hmmreader.parseTMHMMOutput(reader);

                StructureAnnotatedSequence rv = new StructureAnnotatedSequence(sequence);
                rv.setStructureAnnotation(symbolList);

                return rv;
            }
            catch (IOException e)
            {
                logger.error("TMHMM output: \n" + phipsiOutput);
                logger.error("TMHMM error: \n" + errorReader.getInputString());
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        finally
        {
            try
            {
                cleanup(originalFiles);
            }
            catch(Exception e)
            {
                e.printStackTrace();
            }
        }

        return null;
    }

    private void cleanup(File[] originalFiles)
    {
        List<File> list = Arrays.asList(originalFiles);

        File[] files = new File("").getAbsoluteFile().listFiles(new TMHMM_FileFilter());

        // Check to see if the original file list contains each of our files.
        // If a file isn't in the original list, assume we created it and delete it
        for ( File f: files)
        {
            if ( ! list.contains(f))
            {
                f.delete();
            }
        }
    }

    static class TMHMM_FileFilter implements FilenameFilter
    {
        public boolean accept(File dir, String name)
        {
            return name.startsWith("TMHMM_");
        }
    }
}
