package ca.ualberta.proteus.program.transsec;

//import com.sun.org.apache.xerces.internal.parsers.XMLParser;

import com.sun.org.apache.xml.internal.serialize.OutputFormat;
import com.sun.org.apache.xml.internal.serialize.XMLSerializer;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.servlet.ServletContext;
import javax.xml.parsers.*;
import java.io.*;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

/**
 * A class to hold configuration.
 *
 * Note: This should probably be split into two classes - one to parse the config, the other to hold the data
 * <p/>
 * Note: any context loaded from the web context will override the local settings
 */
public class Config extends DefaultHandler
{

    private static Map<String, Config> configMap = null;


    /**
     * The default file to read
     */
    public static final String DEFAULT_FILE = "proteus.config";

    /**
     * The actual file to read, can be reset using the setDefaultFile() method
     */
    private static String FILE = DEFAULT_FILE;

    /**
     * Store the name of the file for THIS config, in case we change it for an individual instance
     */
    private String filename = FILE;

    private final static String DEFAULT = "DEFAULT";

    private Map<String, String> parameters = new TreeMap<String, String>();
    private boolean changed = false;

    /* singleton pattern */
    private synchronized void init(String filename)
    {
        this.filename = filename;

        try
        {
            InputStream istream;

            try
            {
                istream = new FileInputStream(filename);
            }
            catch (FileNotFoundException e)
            {
                istream = null;
            }

            if (istream == null)
            {
                istream = getClass().getResourceAsStream("/" + filename);
            }

            if (istream == null)
            {
                istream = getClass().getResourceAsStream(filename);
            }

            if (istream != null)
            {
                SAXParserFactory spf = SAXParserFactory.newInstance();
                SAXParser parser = spf.newSAXParser();
                parser.parse(istream, this);
            }
            else
            {
                // System.out.println("Can't find file " + filename + " for configuration");
            }
        }
        catch (ParserConfigurationException e)
        {
            e.printStackTrace();
        }
        catch (SAXException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

    }

    /**
     * Get an instance of the default configuration
     *
     * @return
     */
    public static Config getConfig()
    {
        if (configMap == null)
        {
            configMap = new HashMap<String, Config>();
        }

        if (!configMap.containsKey(DEFAULT))
        {
            Config config = new Config();
            config.init(FILE);
            configMap.put(DEFAULT, config);
        }

        return configMap.get(DEFAULT);
    }

    /**
     * Returns the configuration parameter for the parameter name.
     * <p/>
     * Note: use ${var} notation to use the variables of other config params or system properties
     *
     * @param paramName - name of the parameter to return
     * @return the configuration parameter for the parameter name
     */
    public synchronized String getStringValue(String paramName)
    {
        synchronized (parameters)
        {
            //System.out.println(Thread.currentThread().getId() + " accessing parameters - size " + parameters.size());
            if (!parameters.containsKey(paramName))
            {
                System.out.println("Parameter " + paramName + " not found in config file, looking in other config files");

                // look elsewhere
                for (String name : configMap.keySet())
                {
                    Config config = configMap.get(name);
                    if (config.contains(paramName))
                    {
                        return config.getStringValue(paramName);
                    }
                }

                throw new IllegalArgumentException("Parameter " + paramName + " not found in config file");
            }
            String value = parameters.get(paramName);

            /* Parse out a variable */
            if (value.contains("${") && value.contains("}"))
            {
                String var = value.replaceAll("\\$\\{(.*)}", "$1");
                String varValue = System.getProperty(var);
                if (varValue == null)
                {
                    // Maybe it's another config param?
                    varValue = getStringValue(var);
                }
                value = value.replaceAll("\\$\\{.*\\}", varValue);
            }
            return value;
        }
    }

    public boolean contains(String paramName)
    {

        synchronized (parameters)
        {
            return (parameters.containsKey(paramName));
        }
    }

    public synchronized String getStringValue(String paramName, String defaultString)
    {
        synchronized (parameters)
        {
            //System.out.println(Thread.currentThread().getId() + " accessing parameters - size " + parameters.size());
            if (!parameters.containsKey(paramName))
                return defaultString;
            return parameters.get(paramName);
        }
    }

    public synchronized int getIntValue(String paramName, int defaultValue)
    {
        if (!parameters.containsKey(paramName))
        {
            return defaultValue;
        }

        try
        {
            return Integer.parseInt(parameters.get(paramName));
        }
        catch (NumberFormatException e)
        {
            return defaultValue;
        }
    }

    public synchronized int getIntValue(String paramName)
    {
        if (!parameters.containsKey(paramName))
        {
            throw new IllegalArgumentException("Parameter " + paramName + " not found in config file");
        }

        try
        {
            return Integer.parseInt(parameters.get(paramName));
        }
        catch (NumberFormatException e)
        {
            throw new IllegalArgumentException("Parameter " + paramName + " is not a valid integer", e);
        }
    }

    public synchronized String setStringValue(String paramName, String value)
    {
        changed = true;
        return parameters.put(paramName, value);
    }

    public synchronized void startDocument() throws SAXException
    {
        parameters.clear();
    }

    public void endDocument() throws SAXException
    {
        // does nothing
    }

    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException
    {
        synchronized (parameters)
        {
            //System.out.println(Thread.currentThread().getId() + " writing parameters");
            if (qName.equals("param"))
            {
                String name = null, value = null;
                for (int i = 0; i < atts.getLength(); i++)
                {
                    String attribute = atts.getQName(i);
                    if (attribute.equals("name"))
                    {
                        name = atts.getValue(i);
                    }
                    else if (attribute.equals("value"))
                    {
                        value = atts.getValue(i);
                    }
                }

                if (name != null && value != null)
                    parameters.put(name, value);
            }
        }
    }

    public void endElement
            (String
                    uri, String
                    localName, String
                    qName) throws SAXException
    {

    }

    public boolean getBooleanValue(String s)
    {
        String stringValue;
        try
        {
            stringValue = getStringValue(s);
        }
        catch (IllegalArgumentException e)
        {
            stringValue = "false";
        }
        return stringValue.equals("true");
    }

    /**
     * Writes the values of a config file in the config format
     */
    public void writeConfig()
    {
        try
        {
            System.err.print("Writing config...");

            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = null;

            builder = factory.newDocumentBuilder();

            DOMImplementation impl = builder.getDOMImplementation();

            Document doc = impl.createDocument(null, "ProteusConfig", null);

            Element root = doc.getDocumentElement();

            synchronized (parameters)
            {
                for (String key : parameters.keySet())
                {
                    String value = parameters.get(key);

                    Element recordElement = doc.createElement("param");
                    recordElement.setAttribute("name", key);
                    recordElement.setAttribute("value", value);
                    root.appendChild(recordElement);
                }
            }

            // Prepare the DOM document for writing, and the output file
            //Source source = new DOMSource(doc);
//            File file = new File(filename + ".new");
            File file = new File(filename);
            //Result result = new StreamResult(file);

            OutputFormat format = new OutputFormat(doc);
            format.setLineWidth(65);
            format.setIndenting(true);
            format.setIndent(2);
            XMLSerializer serializer = new XMLSerializer(new FileWriter(file), format);
            serializer.serialize(doc);

            // Write the DOM document to file, using a document transformer
            /*Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.transform(source, result);
*/
            System.err.println("Done");
        }
        catch (ParserConfigurationException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    public boolean isChanged()
    {
        return changed;
    }

    public void setConfigFile(String file)
    {
        this.filename = file;
    }

    /**
     * Puts all values from a web context into this configuration instance.
     *
     * @param ctx Context to initialize from.
     */
    public static void initializeFromWebContext(ServletContext ctx)
    {
        Enumeration names = ctx.getInitParameterNames();

        while (names.hasMoreElements())
        {
            String name = (String) names.nextElement();
            String value = ctx.getInitParameter(name);

            getConfig().setStringValue(name, value);
        }

        Enumeration attributeNames = ctx.getAttributeNames();

        while (attributeNames.hasMoreElements())
        {
            String name = (String) attributeNames.nextElement();
            Object value = ctx.getAttribute(name);

            getConfig().setStringValue(name, value.toString());
        }
    }

    /**
     * Specify a particular config instance to read a config from a different file.  This is
     * mainly used for homodeller and thrifty so you could have configuration settings for different
     * versions of proteus, one using  a different version of the configuration.  For example, homodeller
     * might use a different BLAST database than plain vanilla proteus.
     *
     * @param configName Should not be DEFAULT
     * @param filename Filename of the config file to read.
     * @return
     */
    public static Config getConfig(String configName, String filename)
    {
        if (configMap == null)
        {
            configMap = new HashMap<String, Config>();
        }

        if (!configMap.containsKey(configName))
        {
            Config config = new Config();
            config.init(filename);
            configMap.put(configName, config);
        }

        return configMap.get(configName);
    }

    /**
     * Set the default file to read the configuration out of
     *
     * @param filename
     */
    public static void setDefaultFile(String filename)
    {
        FILE = filename;
    }
}
