package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.util.io.StructureDB;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.ChangeVetoException;
import org.joone.engine.*;
import org.joone.engine.learning.TeachingSynapse;
import org.joone.io.MemoryInputSynapse;
import org.joone.net.NeuralNet;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Mar 8, 2005
 * Time: 5:05:52 PM
 * To change this template use File | Settings | File Templates.
 */
public class ConsensusPredictor extends Predictor
{
    public static String TRAINING_SET_FILE = "consensusTrainingSet.txt";
    private static final String TEST_SET = "consensusTestSet.txt";
    private static int LAYER2 = ConsensusTrainingSet.NUM_INPUTS + 2 * ConsensusTrainingSet.NUM_OUTPUTS;
    private double percentageSplit = 0.66;
    private static int numToTrainOn  = 10;
    private static int numToPredict  = 10;
    private static String outputFile  = "consensus.model";
    private static boolean verbose;
    private boolean useStoredNN;
    private List validationSequences;
    private int numToValidate = 20;
    private double bestQ3ScoreSoFar = 0;
    private double bestFilteredQ3ScoreSoFar = 0;
    private boolean doFilter = true;
    private static NeuralNet storedNN = null;
    private static Predictor predictor = null;
    private static final String NN1_MODEL = "model1";
    public static String TRAINING_SET_META_FILE = "consensusTrainingSetMetaData.xml";
    private static String TEST_SET_META_FILE = "consensusTestSetMetaData.xml";
    private static String predictionTestSet = null;
    private ConsensusTrainingSet testSet;
    private static final String NN2_MODEL = "model2";

    public ConsensusTrainingSet buildTrainingSet(List sequences)
    {
        ConsensusTrainingSet trainingSet = new ConsensusTrainingSet();

        trainingSet.storeData = true;
        trainingSet.addInstances(sequences);

        try
        {
            OutputStream file = new FileOutputStream(TRAINING_SET_FILE);
            trainingSet.writeToStream(file);
            trainingSet.saveMetaDataToFile( TRAINING_SET_META_FILE );
        }
        catch (IOException ex)
        {
            System.out.println("Cannot perform output.");
        }

        int numInstances = trainingSet.numInstances();

        System.out.println("Num instances = " + numInstances);

        return trainingSet;
    }

    public void train(List sequences)
    {
        loadTestSet();
        ConsensusTrainingSet trainingSet = new ConsensusTrainingSet();

        try
        {
            FileInputStream fis = new FileInputStream(TRAINING_SET_FILE);
            trainingSet.readFromStream(fis);
        }
        catch (IOException ex)
        {
            System.out.println("Cannot perform output.");
        }

        /* The Layers */
        LinearLayer input = new LinearLayer();
        input.setRows( trainingSet.getNumInputs() );

        SigmoidLayer hidden = new SigmoidLayer();
        hidden.setRows(LAYER2);

        SigmoidLayer output = new SigmoidLayer();
        output.setRows( trainingSet.getNumOutputs() );

        /* The Synapses */
        FullSynapse synapse_IH = new FullSynapse(); /* Input -> Hidden conn. */
        input.addOutputSynapse(synapse_IH);
        hidden.addInputSynapse(synapse_IH);

        FullSynapse synapse_HO = new FullSynapse(); /* Hidden -> Output conn. */
        hidden.addOutputSynapse(synapse_HO);
        output.addInputSynapse(synapse_HO);

        int numInstances = trainingSet.numInstances();

        /* The I/O components */
        MemoryInputSynapse inputStream = new MemoryInputSynapse();
        inputStream.setInputArray(trainingSet.getInputArray());
        inputStream.setAdvancedColumnSelector("1-" + trainingSet.getNumInputs() );
        input.addInputSynapse(inputStream);

        int split = (int) (numInstances * percentageSplit);

        TeachingSynapse trainer = new TeachingSynapse();
        MemoryInputSynapse samples = new MemoryInputSynapse();
        samples.setInputArray(trainingSet.getInputArray());
        trainer.setDesired(samples);

        samples.setAdvancedColumnSelector( String.valueOf(trainingSet.getNumInputs() +1) + "-" + String.valueOf(trainingSet.getNumInputs() + trainingSet.getNumOutputs()) );
        output.addOutputSynapse(trainer);

        nnet = new NeuralNet();
        nnet.addLayer(input, NeuralNet.INPUT_LAYER);
        nnet.addLayer(hidden, NeuralNet.HIDDEN_LAYER);
        nnet.addLayer(output, NeuralNet.OUTPUT_LAYER);
        nnet.setTeacher(trainer);

        Monitor monitor = nnet.getMonitor();
        monitor.setLearningRate(0.005);   // 0.8
        monitor.setMomentum(0.5);         // 0.3
        monitor.setTrainingPatterns(split); /* # of rows in the input file */
        monitor.setValidationPatterns(numInstances - split);
        monitor.setTotCicles(500); /* How many times the net must be trained*/
        monitor.setLearning(true); /* The net must be trained */
        monitor.addNeuralNetListener(this);


        System.out.println("Num instances = " + numInstances);
        PredictionScore instanceDistribution = new PredictionScore();
        for (Iterator it=trainingSet.getOutputInstanceIterator(); it.hasNext();) {
            double[] outputInstance = (double[]) it.next();
            instanceDistribution.helix += outputInstance[ProteusTools.HELIX];
            instanceDistribution.coil += outputInstance[ProteusTools.COIL];
            instanceDistribution.gap += outputInstance[ProteusTools.GAP];
            instanceDistribution.strand += outputInstance[ProteusTools.STRAND];
        }
        System.out.println("Instance distribution: ");
        System.out.println("Helix = " + instanceDistribution.helix);
        System.out.println("Coil = " + instanceDistribution.coil);
        System.out.println("Strand = " + instanceDistribution.strand);
        System.out.println("Gap = " + instanceDistribution.gap);


        nnet.start();
        nnet.getMonitor().Go();
    }

    public ScoreAnnotatedSequence predict(Sequence sequence, NeuralNet localNN, boolean filter){
        return predict(sequence, localNN, filter, null);
    }

    /**
     * You can pass a training set, so that the training set can be pre-computed an stored.   This is mostly for training purposes,
     * when you need to evaluate often
     *
     * @param sequence
     * @param localNN
     * @param filter
     * @param pssm
     * @return
     */
    public ScoreAnnotatedSequence predict(Sequence sequence, NeuralNet localNN, boolean filter, PSSM pssm)
    {
        ScoreAnnotatedSequence result;

        if (localNN == null) {
            if ( useStoredNN && storedNN != null) {
                // we have a stored NN, use it
                localNN = storedNN;
            }
            else
            {
                // we either have no NN, or don't want to use it
                localNN = restoreNeuralNet(outputFile);
                storedNN = localNN;
            }
        }

        if (localNN != null)
        {
            /* We get the first layer of the net (the input layer),
              then remove all the input synapses attached to it and attach a DirectSynapse */
            Layer input = localNN.getInputLayer();
            input.removeAllInputs();
            DirectSynapse memInp = new DirectSynapse();
            input.addInputSynapse(memInp);

            /* We get the last layer of the net (the output layer),
             then remove all the output synapses attached to it and attach a DirectSynapse */
            Layer output = localNN.getOutputLayer();
            output.removeAllOutputs();
            DirectSynapse memOut = new DirectSynapse();
            output.addOutputSynapse(memOut);

            // begin the neural net's thread
            localNN.getMonitor().setLearning(false);
            localNN.start();

            // Create the set that contains the instances of this sequence
            ConsensusTrainingSet set = new ConsensusTrainingSet();
            if ( pssm == null ) {

                // didn't pass a pssm
                result = set.getPredictionResult(sequence, memInp, memOut );
            }
            else {
                // passed a pssm
                result = set.getPredictionResult(sequence, memInp, memOut, pssm );
            }

            if ( verbose) System.out.println("Finished");
        }
        else
        {
            throw new IllegalStateException("Cannot read in neural net!");
        }

        if ( filter ) {
            // filter the result with the attached filters
            for (Iterator it = filters.iterator(); it.hasNext();)
                result = ( (PredictionFilter) it.next()).filter(result);
        }

        if ( !useStoredNN) {
            // Cleanu up our memory
            storedNN = null;
            localNN = null;
        }

        return result;
    }


    public static void main(String[] args) throws IllegalSymbolException, ChangeVetoException, IOException
    {
        boolean train = false;
        boolean predict = false;
        boolean buildTrainingSet = false;
        int low = 0;
        boolean pdbEval = false;
        String pdbFile = null;

        for (int i = 0; i < args.length; i++)
        {
            if (args[i].equals("-train"))
            {
                train = true;
                if ( i < args.length -1 ) {
                    try {
                        TRAINING_SET_FILE = args[++i];
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
                }
                else {
                    printUsage();
                }
            }
            else if (args[i].equals("-predict"))
            {
                predict = true;
                if ( i < args.length -1 )
                    try {
                        numToPredict = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        //printUsage();
                        predictionTestSet = args[i];
                    }
            }
            else if ( args[i].equals("-build"))
            {
                buildTrainingSet = true;
                if ( i < args.length -1 )
                    try {
                        numToTrainOn = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
                if ( i < args.length -1 )
                    try {
                        low = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
            }
            else if (args[i].equals("-verbose")||args[i].equals("-v"))
            {
                verbose = true;
            }
            else if (args[i].equals("-output")||args[i].equals("-o"))
            {
                if ( i < args.length -1 ) {
                    String arg = args[++i];
                    if ( arg.charAt(0) == '-' )
                        printUsage();
                    else
                        outputFile = arg;
                }
            }
            else if ( args[i].contains("-layer"))
            {
                if ( i < args.length -1 )
                    try {
                        LAYER2 = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
            }
            else if ( args[i].contains("-win"))
            {
                if ( i < args.length -1 )
                    try {
                        ConsensusTrainingSet.WINDOW_SIZE = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
            }
            else if ( args[i].contains("-pdbtest")) {
                pdbEval = true;
                pdbFile = args[++i];
            }
            else if ( args[i].equals("-test")) {
                ConsensusPredictor predictor = new ConsensusPredictor();

                // Have to add the filters in this order - use the NN filter, then the reg exp filter
                final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
                predictor.addFilter(ps2Ffilter);

                final SimpleFilter filter = new SimpleFilter();
                predictor.addFilter(filter);

                SeqSeeDB db = SeqSeeDB.getInstance();
                List randomSequences = new ArrayList();
                try
                {
                    randomSequences.add(db.getSequence("1GCLA"));
                }
                catch (BioException e)
                {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }

                predictor.buildTrainingSet(randomSequences);
                return;
            }
            else if ( args[i].equalsIgnoreCase("-buildfromDB"))
            {
                // supply only the database
                StructureDB db = new StructureDB(args[++i], StructureDB.MODE.READ);

                System.out.println("Building from " + args[i]);

                ConsensusPredictor predictor = new ConsensusPredictor();

                // Have to add the filters in this order - use the NN filter, then the reg exp filter
                final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
                predictor.addFilter(ps2Ffilter);

                final SimpleFilter filter = new SimpleFilter();
                predictor.addFilter(filter);

                List<StructureAnnotatedSequence> randomSequences = new ArrayList(db.size());
                for ( SequenceIterator it = db.sequenceIterator(); it.hasNext(); ) {
                    try
                    {
                        randomSequences.add( (StructureAnnotatedSequence) it.nextSequence() );
                    }
                    catch (BioException e)
                    {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
                System.out.println(randomSequences.size() + " sequences found.");

                predictor.buildTrainingSet(randomSequences);
                return;
            }
            else if ( args[i].equals("-eval")) {
                // You have to supply a XML training set, and the database it was created from
                TEST_SET_META_FILE = args[++i];
                StructureDB db = new StructureDB(args[++i], StructureDB.MODE.READ);

                ConsensusPredictor predictor = new ConsensusPredictor();
                predictor.useStoredNN = true;

                // Have to add the filters in this order - use the NN filter, then the reg exp filter
                final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
                predictor.addFilter(ps2Ffilter);

                final SimpleFilter filter = new SimpleFilter();
                predictor.addFilter(filter);
                predictor.eval(db, null);
                return;
            }
        }


        if (train)
        {
            // train on this set of sequences
            ConsensusPredictor predictor = new ConsensusPredictor();

            // Have to add the filters in this order - use the NN filter, then the reg exp filter
            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);

            predictor.train(null);
        }
        else if (buildTrainingSet) {
            // train on this set of sequences
            ConsensusPredictor predictor = new ConsensusPredictor();

            // Have to add the filters in this order - use the NN filter, then the reg exp filter
            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);

            SeqSeeDB db = SeqSeeDB.getInstance();

            List<StructureAnnotatedSequence> randomSequences;
            if ( low == 0 )
                randomSequences = db.getEvaluationSet2(numToTrainOn, 2000, 3000, 7);
            else
                randomSequences = db.getEvaluationSet2(numToTrainOn, low, 3000, 7);

            randomSequences = db.getEvaluationSet(numToTrainOn);
            try
            {
                randomSequences.add((StructureAnnotatedSequence) db.getSequence("1M90C"));
                randomSequences.add((StructureAnnotatedSequence) db.getSequence("1UAQA"));
                randomSequences.add((StructureAnnotatedSequence) db.getSequence("1A2PA"));
                randomSequences.add((StructureAnnotatedSequence) db.getSequence("1CSEE"));
                randomSequences.add((StructureAnnotatedSequence) db.getSequence("1MH8A"));
            }
            catch (BioException e)
            {
                e.printStackTrace();  //TODO
            }

            predictor.buildTrainingSet(randomSequences);
        }
        else if (predict && predictionTestSet != null)
        {
            ConsensusPredictor predictor = new ConsensusPredictor();
            predictor.useStoredNN = true;

            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.useStoredNN = true;
            predictor.addFilter(filter);

            ConsensusTrainingSet testSet = new ConsensusTrainingSet();
            try
            {
                //testSet.readFromStream(new FileInputStream(TEST_SET));
                testSet.readMetaDataFromFile(predictionTestSet);
            }
            catch (IOException e1)
            {
                e1.printStackTrace();
            }


            SeqSeeDB db = SeqSeeDB.getInstance();

            Score score = new Score();
            Score filteredScore = new Score();

            NumberFormat format = NumberFormat.getNumberInstance();
            format.setMaximumFractionDigits(4);

            int i =0;
            for (Iterator it = testSet.getPssmList().keySet().iterator(); it.hasNext(); ) {
                String key = (String) it.next();
                final PSSM pssm = (PSSM) testSet.getPssmList().get( key );

                try
                {
                    StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) db.getSequence(key);

                    if (i < 5)
                        System.out.println("\nStructure: "
                                        + sequence.getStructureAnnotation().seqString().toUpperCase());

                    StructureAnnotatedSequence prediction;
                    prediction = predictor.predict(sequence, null, false, pssm);

                    double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                    double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                    if (i < 5)
                        System.out.println("Without filter: "
                                           + " Q3: " + format.format(q3)
                                           + " SOV: " + format.format(sov)
                                           + prediction);
                    score.q3 += q3;
                    score.sov += sov;

                    // Predict filtered
                    prediction = predictor.predict(sequence, null, true, pssm);

                    q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                    sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                    if (i < 5)
                        System.out.println("Filtered: "
                                           + " Q3: " + format.format(q3)
                                           + " SOV: " + format.format(sov)
                                           + prediction);
                    filteredScore.q3 += q3;
                    filteredScore.sov += sov;

                    i++;
                }
                catch (BioException e1)
                {
                    e1.printStackTrace();
                }
            }

            System.out.println("Number of sequences: " + testSet.getPssmList().size());
            System.out.print("Avg Q3: " + score.q3 / testSet.getPssmList().size());
            System.out.println("\t SOV: " + score.sov / testSet.getPssmList().size());

            System.out.print("Filtered Q3: " + filteredScore.q3 / testSet.getPssmList().size());
            System.out.println("\t SOV: " + filteredScore.sov / testSet.getPssmList().size());

        }
        else if (predict)
        {
            // predict this set of sequences
            ConsensusPredictor predictor = new ConsensusPredictor();

           // final NeuralFilter NFfilter = new NeuralFilter();
           // NFfilter.useStoredNN = true;
           // predictor.addFilter(NFfilter);

            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
           // NFfilter.useStoredNN = true;
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.useStoredNN = true;
            predictor.addFilter(filter);

            SeqSeeDB db = SeqSeeDB.getInstance();
            //List randomSequences = db.getEvaluationSet(numToPredict);
            //List randomSequences = db.getRandomSequences(numToPredict, 3000, 4000);
            List randomSequences = db.getEvaluationSet(numToPredict, 3000, 4000);


            /*List randomSequences = db.getEvaluationSet(0, 1, 3000);
            try
            {
                randomSequences.add(db.get("1EIWA"));
                randomSequences.add(db.get("1RET_"));

            }
            catch (BioException e)
            {
                e.printStackTrace();
            }

            */

            ScoreAnnotatedSequence prediction;
            Score score = new Score();
            Score filteredScore = new Score();
            NumberFormat format = NumberFormat.getNumberInstance();
            format.setMaximumFractionDigits(4);

            for (Iterator it = randomSequences.iterator(); it.hasNext();)
            {
                final StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();
                prediction = predictor.predict(sequence, null, false);

                System.out.println("Sequence: \n" + sequence);

                double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                score.q3 += q3;
                score.sov += sov;

                System.out.println("Prediction: "
                                   + " Q3: " + format.format(q3)
                                   + " SOV: " + format.format(sov)
                                   + prediction);

                prediction = predictor.predict(sequence);

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                filteredScore.q3 += q3;
                filteredScore.sov += sov;

                System.out.println("\nPrediction ** filtered ** : "
                                   + " Q3: " + format.format(q3)
                                   + " SOV: " + format.format(sov)
                                   + prediction);

                prediction = getPredictor().predict(sequence, null, true);

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());


                System.out.println("Single sequence predictor: "
                                       + " Q3: " + format.format(q3)
                                       + " SOV: " + format.format(sov)
                                       + prediction);
                filteredScore.q3 += q3;
                filteredScore.sov += sov;
            }

            System.out.print("Avg Q3: " + format.format( score.q3 / randomSequences.size() ));
            System.out.println("\tSOV: " + format.format( score.sov / randomSequences.size() ));

            System.out.print("Filtered Avg Q3: " + format.format( filteredScore.q3 / randomSequences.size()));
            System.out.println("\tFiltered SOV: " + format.format( filteredScore.sov / randomSequences.size()));
        }
        else if ( pdbEval) {
            ScoreAnnotatedSequence prediction;
            Score score = new Score();
            Score filteredScore = new Score();
            NumberFormat format = NumberFormat.getNumberInstance();
            format.setMaximumFractionDigits(4);

            ConsensusPredictor predictor = new ConsensusPredictor();

           // final NeuralFilter NFfilter = new NeuralFilter();
           // NFfilter.useStoredNN = true;
           // predictor.addFilter(NFfilter);

            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
           // NFfilter.useStoredNN = true;
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.useStoredNN = true;
            predictor.addFilter(filter);

            StructureDB db = new StructureDB(pdbFile, StructureDB.MODE.READ);
            for (SequenceIterator it = db.sequenceIterator(); it.hasNext();)
            {
                StructureAnnotatedSequence sequence = null;
                try
                {
                    sequence = (StructureAnnotatedSequence) it.nextSequence();

                    prediction = predictor.predict(sequence, null, false);

                    System.out.println("Sequence: \n" + sequence);

                    double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                    double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                    score.q3 += q3;
                    score.sov += sov;

                    System.out.println("Prediction: "
                                       + " Q3: " + format.format(q3)
                                       + " SOV: " + format.format(sov)
                                       + prediction);

                    prediction = predictor.predict(sequence);

                    q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                    sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                    filteredScore.q3 += q3;
                    filteredScore.sov += sov;

                    System.out.println("\nPrediction ** filtered ** : "
                                       + " Q3: " + format.format(q3)
                                       + " SOV: " + format.format(sov)
                                       + prediction);
                    }
                catch (BioException e)
                {
                    e.printStackTrace();  //TODO
                }
                catch ( Throwable t ) {
                    t.printStackTrace();  //TODO
                }
            }

            System.out.print("Avg Q3: " + format.format( score.q3 / db.size() ));
            System.out.println("\tSOV: " + format.format( score.sov / db.size() ));

            System.out.print("Filtered Avg Q3: " + format.format( filteredScore.q3 / db.size()));
            System.out.println("\tFiltered SOV: " + format.format( filteredScore.sov / db.size()));
        }
        else
            printUsage();
    }

    private static void printUsage() {
        System.out.println("Usage: <ProgName> (-train <# seq> | -predict <#seq>) [options]" );
        System.out.println("Options: ");
        System.out.println("-train <Training set file> -testSet=<Test set> \t\t Train on a training set, and test on a set");
        System.out.println("-build <Training set file> \t\tBuild a training set");
        System.out.println("-predict <# sequences>\t\tPredict a number of sequences");
        System.out.println("-v, -verbose");
        System.out.println("-layer <# hidden nodes>");
        System.out.println("-o, -output <file>\t\t\tFile to output the training model to (if applicable)");
        System.exit(-1);
    }

    public void cicleTerminated(NeuralNetEvent e)
    {
        // call inherited method, it will do validation on a set of sequences
        super.cicleTerminated(e);

        Monitor mon = (Monitor) e.getSource();
        long c = mon.getCurrentCicle();
        long c1 = mon.getTotCicles() - c;

        System.out.println("Training Error = " + mon.getGlobalError() );

        /*if ( c1 < 5
                || c1 < 20 && c1 % 2 == 0
                || c1 < 25 && c1 % 5 == 0
                || c1 < 200 && c1 % 10 == 0
                || c1 % 20 == 0) // after 200, every 20
        {

        }
        else return;*/

        StructureDB db = SeqSeeDB.getInstance();

      //  if (validationSequences == null)
       //     validationSequences = db.getEvaluationSet(numToValidate, 2000, 3000);

        NeuralNet localNN = nnet.cloneNet();
        localNN.getMonitor().setExporting(false);
        localNN.removeAllListeners();

        try
        {
            db = new StructureDB("pdbstruct.fa", StructureDB.MODE.READ);
        }
        catch (IOException e1)
        {
            e1.printStackTrace();
        }

        eval(db, localNN);

        if ((c1 % 25) == 0)
            saveNeuralNet(outputFile);
    }

    private void eval(SequenceDB db, NeuralNet localNN)
    {
        Score score = new Score();
        Score filteredScore = new Score();

        NumberFormat format = NumberFormat.getNumberInstance();
        format.setMaximumFractionDigits(4);

        if ( testSet == null)
            loadTestSet();

        int i =0;
        for (Iterator it = testSet.getPssmList().keySet().iterator(); it.hasNext(); ) {
            String key = (String) it.next();
            final PSSM pssm = (PSSM) testSet.getPssmList().get( key );

            try
            {
                StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) db.getSequence(key);

                if (i < 5)
                    System.out.println("Structure: "
                                    + sequence.getStructureAnnotation().seqString().toUpperCase() + "\n");

                StructureAnnotatedSequence prediction;
                prediction = predict(sequence, localNN, false, pssm);

                double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                if (i < 5)
                    System.out.println("Without filter: "
                                       + " Q3: " + format.format(q3)
                                       + " SOV: " + format.format(sov)
                                       + prediction);
                score.q3 += q3;
                score.sov += sov;

                // Predict filtered
                prediction = predict(sequence, localNN, true, pssm);

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                if (i < 5)
                    System.out.println("Filtered: "
                                       + " Q3: " + format.format(q3)
                                       + " SOV: " + format.format(sov)
                                       + prediction);
                filteredScore.q3 += q3;
                filteredScore.sov += sov;

                i++;
            }
            catch (IllegalSymbolException e1)
            {
                e1.printStackTrace();
            }
            catch (BioException e1)
            {
                e1.printStackTrace();
            }
        }

        System.out.print("Avg Q3: " + score.q3 / testSet.getPssmList().size());
        System.out.println("\t SOV: " + score.sov / testSet.getPssmList().size());

        System.out.print("Filtered Q3: " + filteredScore.q3 / testSet.getPssmList().size());
        System.out.println("\t SOV: " + filteredScore.sov / testSet.getPssmList().size());

        if ( score.q3 > bestQ3ScoreSoFar ) {
           // StringFormatter format = new Formatter();
           // format.
            System.out.println("Best Q3 score so far! (" + score.q3 / testSet.getPssmList().size() + ")");
            bestQ3ScoreSoFar = score.q3;

            final Date date = new Date();
            saveNeuralNet("bestConsensusNN.model" + date.getDate() + "-" + ( date.getMonth() +1) );
        }

        if ( filteredScore.q3 > bestFilteredQ3ScoreSoFar ) {
           // StringFormatter format = new Formatter();
           // format.
            System.out.println("Best Filtered Q3 score so far! (" + filteredScore.q3 / testSet.getPssmList().size() + ")");
            bestFilteredQ3ScoreSoFar = filteredScore.q3;

            final Date date = new Date();
            saveNeuralNet("bestFilteredConsensusNN.model" + date.getDate() + "-" + ( date.getMonth() + 1 ));
        }
    }

    private void loadTestSet()
    {
        testSet = new ConsensusTrainingSet();
        try
        {
            testSet.readFromStream(new FileInputStream(TEST_SET));
            testSet.readMetaDataFromFile(TEST_SET_META_FILE);
        }
        catch (IOException e1)
        {
            System.err.println("Error loading test set");
            e1.printStackTrace();
        }
    }

    public static Predictor getPredictor() {
        if ( predictor == null ) {
            predictor = new SingleSequencePredictor();
            ((SingleSequencePredictor) predictor).setNeuralNetModel(Config.getConfig().getStringValue(NN1_MODEL));
            ((SingleSequencePredictor) predictor).useStoredNN = true;

            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);
        }
        return predictor;
    }

    /**
     * Main prediction method
     * @param sequence
     * @param alignment
     * @return
     */
    public ScoreAnnotatedSequence predict(Sequence sequence, Alignment alignment)
    {

        List labels = alignment.getLabels();
        assert labels.contains(sequence.getName()) || labels.contains("QUERY"): "Alignment sequences does not contain query sequence";

        ScoreAnnotatedSequence result;

        if ( storedNN == null)
            storedNN = restoreNeuralNet(Config.getConfig().getStringValue(NN2_MODEL));

        if (storedNN != null)
        {
            /* We get the first layer of the net (the input layer),
              then remove all the input synapses attached to it and attach a DirectSynapse */
            Layer input = storedNN.getInputLayer();
            input.removeAllInputs();
            DirectSynapse memInp = new DirectSynapse();
            input.addInputSynapse(memInp);

            /* We get the last layer of the net (the output layer),
             then remove all the output synapses attached to it and attach a DirectSynapse */
            Layer output = storedNN.getOutputLayer();
            output.removeAllOutputs();
            DirectSynapse memOut = new DirectSynapse();
            output.addOutputSynapse(memOut);

            // begin the neural net's thread
            storedNN.getMonitor().setLearning(false);
            storedNN.start();
            
            // Create the set that contains the instances of this sequence
            ConsensusTrainingSet set = new ConsensusTrainingSet();
            PSSM pssm = new PSSM(sequence, alignment, getPredictor());

            result = set.getPredictionResult(sequence, memInp, memOut, pssm);

            if ( verbose) System.out.println("Finished");
        }
        else
        {
            throw new IllegalStateException("Cannot read in neural net!");
        }

        // filter the result with the attached filters
        for (Iterator it = filters.iterator(); it.hasNext();)
            result = ( (PredictionFilter) it.next()).filter(result);

        return result;
    }
}
