package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.util.io.StructureDB;
import org.biojava.bio.Annotation;
//import org.biojava.bio.BlastLikeSearchFilter;//newly added interface
import org.biojava.bio.BioException;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
//import org.biojava.bio.alignment.SequenceAlignment;//newly added
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;

import java.util.*;
import java.io.IOException;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Jan 25, 2007
 * Time: 12:23:03 PM
 * To change this template use File | Settings | File Templates.
 */
public class HomologCombiner
{
    private RuntimeCatalog logger = null;
    public static final double DEFAULT_EVALUE = 1e-7;
	private static boolean verbose = false;
    private static boolean evaluating = false;
    private static HomologCombiner instance;

    HomologCombiner()
    {
    }

    public ScoreAnnotatedSequence searchAndOverlayHomolog(ScoreAnnotatedSequence sequence, Map<String, StructureAnnotatedSequence> sequencesget, Double identity)
    {
        return searchAndOverlayHomologs(sequence, DEFAULT_EVALUE, -1, sequencesget, identity);
    }

    /**
     * Takes a sequence with a prediction on it already, and then overlays the structure of several homologs on
     * top of it to obtain a more accurate prediction
     *
     * @param sequence the sequence that already has a prediction on it
     * @return a sequence with its structure annotation composed of the predicted structure + the homolog structure
     */
    public ScoreAnnotatedSequence searchAndOverlayHomologs(ScoreAnnotatedSequence sequence, double searchEvalue, double minEvalue, Map<String, StructureAnnotatedSequence> otherSequences, Double identity)
    {
        assert this.logger != null : "Did not specify a logger";

        List<GappedSequence> homologs = getHomologs(sequence, searchEvalue, identity);

        logger.logObject("homologs", homologs);

        ScoreAnnotatedSequence result;
        List<GappedStructureAnnotatedSequence> alignedSequences = new ArrayList<GappedStructureAnnotatedSequence>();

        if (homologs.size() == 0)
        {
            System.out.println("No PDB homologs found below threshold (" + searchEvalue + ")");

            // set the result to what we came in with... i.e. a predicted sequence 
            result = sequence;
        }
        else
        {
            double smallestEvalue = Double.MAX_VALUE;
            HashMap<Double, GappedSequence> evalueHomologMap = new HashMap<Double, GappedSequence>();

            GappedSequence alignedQuery = null;

            /** Remove some homologs **/
            for (GappedSequence seq : homologs)
            {
                Annotation annotation = seq.getAnnotation();
                Double homologEvalue = (Double) annotation.getProperty("e-value");
                boolean addToSet = homologEvalue != null && (minEvalue == -1 || homologEvalue > minEvalue);

                if (seq.getName().equals("QUERY"))
                {
                    alignedQuery = seq;
                }

                if (evaluating)
                {
                    if (seq.getName().equalsIgnoreCase(sequence.getName()))
                    {
                        addToSet = false;
                    }
                }
                if (addToSet)
                {
                    evalueHomologMap.put(homologEvalue, seq);

                    if (homologEvalue < smallestEvalue)
                    {
                        smallestEvalue = homologEvalue;
                    }

                    if (verbose)
                    {
                        System.out.println(seq.getName() + ": " + seq.seqString());
                    }
                }
            }

            assert alignedQuery != null : "CAN'T FIND ALIGNED QUERY";

            System.out.println(alignedQuery.getName() + ": " + alignedQuery.seqString());

            Object[] evalues = evalueHomologMap.keySet().toArray();
            Arrays.sort(evalues);

            SeqSeeDB db = SeqSeeDB.getInstance();

            HashMap<GappedStructureAnnotatedSequence, Double> homologEvalues = new HashMap<GappedStructureAnnotatedSequence, Double>();

            Config config = Config.getConfig();
            String tmpdb_file = config.getStringValue("TMPDB");

            StructureDB tmpdb = null;
            try
            {
                tmpdb = new StructureDB(tmpdb_file, StructureDB.MODE.READ);
            }
            catch (IOException e)
            {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                return null;
            }

            /**
             * Create ScoreAnnotatedSequences from the structures in the DB
             */
            for (Object ev : evalues)
            {
                GappedStructureAnnotatedSequence homolog;
                Double evalue = (Double) ev;

                try
                {
                    GappedSequence gappedSequence = evalueHomologMap.get(evalue);
                    String name = gappedSequence.getName();
                    boolean sequenceIsTM = false;

                    if (verbose)
                    {
                        System.out.println("Sequence: " + name + " evalue: " + evalue);
                    }

                    StructureAnnotatedSequence newSequence = null;
                    try
                    {
                        newSequence = (StructureAnnotatedSequence) db.getSequence(name);
                    }
                    catch (BioException e)
                    {
                        System.err.println("Can't find homolog '" + name + "' in seqsee.db, trying TM PDB");

                        try
                        {
                            sequenceIsTM = true;
                            newSequence = (StructureAnnotatedSequence) tmpdb.getSequence(name);
                        }
                        catch (IllegalIDException e1)
                        {
                            System.out.println("Can't find it in TM PDB either");
                            e1.printStackTrace();
                        }

                        //e.printStackTrace();
                    }
                    
                    homolog = new GappedStructureAnnotatedSequence(gappedSequence);

                    if (newSequence != null)
                    {
                        homolog.createGappedStructureToIncompleteSequence(newSequence);
                                                    
                        if (verbose)
                        {
                            System.out.println(homolog);
                        }

                        try
                        {
                            Annotation annotation = homolog.getAnnotation();
                            annotation.setProperty("e-value", evalue);
                            if ( sequenceIsTM )
                            {
                                annotation.setProperty("isTM", true);
                            }
                        }
                        catch (ChangeVetoException e)
                        {
                            e.printStackTrace();
                        }

                        alignedSequences.add(homolog);
                        homologEvalues.put(homolog, evalue);
                    }
                }
                catch(Throwable e)
                {
                    // catch the exception so it doesn't kill the rest of the prediction
                    e.printStackTrace();
                }
            }

            logger.logObject(Consts.HOMOLOG_EVALUES, homologEvalues);

            if (alignedSequences.size() == 0)
            {
                System.out.println("0No PDB homologs found below threshold (" + searchEvalue + ")");
                return null;
            }
            // List<GappedStructureAnnotatedSequence> alignedHomologues = new ArrayList<GappedStructureAnnotatedSequence>();

            // Create a gappedstructure for the query sequence
            GappedStructureAnnotatedSequence gappedQuery = new GappedStructureAnnotatedSequence(alignedQuery);
            gappedQuery.createGappedStructureToIncompleteSequence(sequence);

            List<GappedStructureAnnotatedSequence> alignedHomologuesClone = new ArrayList<GappedStructureAnnotatedSequence>();
            alignedHomologuesClone.addAll(alignedSequences);

            // clone it so we can remove stuff from it and not change it in the logger
            logger.logObject(Consts.ALL_ALIGNED_HOMOLOG_LIST, alignedHomologuesClone);


            logger.logObject(Consts.ALIGNED_QUERY, gappedQuery);

            Symbol[] newStructure = new Symbol[alignedQuery.length()];
            PredictionScore[] newScore = new PredictionScore[alignedQuery.length()];

            /** Assign the structure from the homologs **/
            final int WINDOW = 7;
            final int MIDDLERES = WINDOW / 2;
            String gappedQueryString = gappedQuery.seqString();

            boolean[] alignmentUsed = new boolean[alignedSequences.size()];

            for (int resIndex = 0; resIndex < gappedQuery.length(); resIndex++)
            {
                int[] scores = new int[alignedSequences.size()];
                for (int seqIndex = 0; seqIndex < alignedSequences.size(); seqIndex++)
                {
                    GappedStructureAnnotatedSequence gappedHomolog = alignedSequences.get(seqIndex);
                    String gappedHomologString = gappedHomolog.seqString();

                    // can't map to a gap
                    if (gappedHomologString.charAt(resIndex) != '-'
                            && gappedHomologString.charAt(resIndex) != '~')
                    {
                        // find each sequence's similarity score based on a window size
                        for (int i = resIndex - MIDDLERES; i < resIndex - MIDDLERES + WINDOW; i++)
                        {
                            if ((i >= 0) && (i < gappedQuery.length()))
                            {
                                char homologRes = gappedHomologString.charAt(i);
                                char queryRes = gappedQueryString.charAt(i);

                                if (homologRes == queryRes)
                                {
                                    scores[seqIndex]++;
                                }
                            }
                        }
                    }
                }

                // find the top score
                int topScore = -1, topScoreIndex = -1;
                for (int seqIndex = 0; seqIndex < scores.length; seqIndex++)
                {
                    if (scores[seqIndex] > topScore)
                    {
                        topScore = scores[seqIndex];
                        topScoreIndex = seqIndex;
                    }
                }

                /* if there is a homolog at this residue that scores, and that score is > 2 /7
                  then assign its structure.  Otherwise, assign the predicted structure
                   */
                if (topScoreIndex != -1
                        && topScore >= 2)
                {
                    Symbol homologSymbol = alignedSequences.get(topScoreIndex).getStructureAnnotation().symbolAt(resIndex + 1);
                    newStructure[resIndex] = homologSymbol;
                    alignmentUsed[topScoreIndex] = true;
                    if (topScore >= 5)
                    {
                        newScore[resIndex] = new PredictionScore(homologSymbol, 0.9999);

                    }
                    else if (topScore >= 4)
                    {
                        newScore[resIndex] = new PredictionScore(homologSymbol, 0.9);

                    }
                    else if (topScore >= 3)
                    {
                        newScore[resIndex] = new PredictionScore(homologSymbol, 0.8);
                    }
                    else if (topScore >= 2)
                    {
                        newScore[resIndex] = new PredictionScore(homologSymbol, 0.7);
                    }
                }
                else
                {
                    Symbol querySymbol = gappedQuery.getStructureAnnotation().symbolAt(resIndex + 1);
                    newStructure[resIndex] = querySymbol;
                }
            }

            /*
               Eliminate all aligned sequences that weren't used in the mapping, for output purposes
            */
            extractAlignedSequences(alignedSequences, alignmentUsed);


            result = new ScoreAnnotatedSequence(new SimpleSequence(alignedQuery, "", alignedSequences.get(0).getName(), null));
            result.setStructureAnnotation(new SimpleSymbolList(newStructure, newStructure.length, new StructureTools().getStructure()));
            result.setScores(newScore);

            // go back to a ungapped structure
            result = GappedStructureAnnotatedSequence.createUngappedStructure(sequence, result);
        }

        // combine all the output with the transmembrane and signal peptide sequences
        remapOutput(sequence, result, otherSequences, null);

        // filter the output
        SimpleFilter filter = new SimpleFilter();
        result = filter.filter(result);

        /// copy the annotation
        if (alignedSequences.size() > 0)
        {
            Annotation annotation = alignedSequences.get(0).getAnnotation();
            for (Object key : annotation.keys())
            {
                try
                {
                    result.getAnnotation().setProperty(key, annotation.getProperty(key));
                }
                catch (ChangeVetoException e)
                {
                    e.printStackTrace();
                }
            }


            try
            {
                result.getAnnotation().setProperty("homolog", alignedSequences.get(0).getName());
            }
            catch (ChangeVetoException e)
            {
                e.printStackTrace();
            }
        }

        return result;
    }

    /**
     * Combines the output from all sources in the following hierarchy
     * - Signal peptide
     * - TM homolog
     * - Soluble homolog
     * - TM prediction
     * - 2e structure prediction
     *
     * @param sequence
     * @param result
     * @param otherSequences
     */
    private void remapOutput(ScoreAnnotatedSequence sequence, ScoreAnnotatedSequence result, Map<String, StructureAnnotatedSequence> otherSequences, List<StructureAnnotatedSequence> solubleHomologs)
    {
        StructureAnnotatedSequence transmembrane = null;
        StructureAnnotatedSequence tm_homolog = null;
        StructureAnnotatedSequence signalPeptide = null;

        if (otherSequences != null)
        {
            transmembrane = otherSequences.get(Consts.TRANSMEMBRANE);
            tm_homolog = otherSequences.get(Consts.TRANSMEMBRANE_HOMOLOG);
            signalPeptide = otherSequences.get(Consts.SIGNAL_PEPTIDE);
        }
        SymbolList tm_structure = null;
        SymbolList sp_structure = null;
        SymbolList tm_homolog_structure = null;

        if (transmembrane != null)
        {
            assert transmembrane.length() == result.length() : "Lengths don't match";

            tm_structure = transmembrane.getStructureAnnotation();
        }

        if (signalPeptide != null)
        {
            assert signalPeptide.length() == result.length() : "Lengths don't match";

            sp_structure = signalPeptide.getStructureAnnotation();
        }

        if (tm_homolog != null)
        {
            assert tm_homolog.length() == result.length() : "Lengths don't match";

            tm_homolog_structure = tm_homolog.getStructureAnnotation();
        }

        /* Re-map the scores */
        Symbol symbol;
        StructureTools structureTools = new StructureTools();
        Symbol[] structure = new Symbol[result.length()];

        for (int i = 1; i <= result.length(); i++)
        {
            if (sp_structure != null && !ProteusTools.isGap(symbol = sp_structure.symbolAt(i)) && symbol != structureTools.getT() )
            {
            //    if (symbol != structure.getT())
               // {
                    // map the signal peptide
                    structure[i - 1] = symbol;
                    result.setScore(i, new PredictionScore(symbol, 0.9));
              //  }
            }
            else if (tm_homolog_structure != null && !ProteusTools.isGap(symbol = tm_homolog_structure.symbolAt(i)))
            {
                // map the transmembrane homolog
                structure[i - 1] = symbol;
                result.setScore(i,
                        new PredictionScore(symbol, 0.9));
            }
            else if (tm_structure != null
                    && !ProteusTools.isGap(symbol = tm_structure.symbolAt(i)))
            {
                // if there's no soluble homolog, then map the transmembrane prediction over the
                // ab-initio secondary structure prediction
                structure[i - 1] = symbol;
                result.setScore(i, new PredictionScore(symbol, 0.9));
            }
            else if (result.getScore(i) == null
                    || result.getScore(i).getMax() == 0)
            {
                structure[i - 1] = result.getStructureAnnotation().symbolAt(i);
                result.setScore(i, sequence.getScore(i));
            }
            else if (result.getScore(i).getMax() < sequence.getScore(i).getMax())
            {
                // Ab-initio prediction will override homolog
                result.setScore(i, sequence.getScore(i));
                structure[i - 1] = sequence.getStructureAnnotation().symbolAt(i);
            }
            else
            {
                result.setScore(i, sequence.getScore(i));
                structure[i - 1] = result.getStructureAnnotation().symbolAt(i);
            }
        }

        result.setStructureAnnotation(new SimpleSymbolList(structure, result.length(), result.getStructureAnnotation().getAlphabet()));
    }

    /**
     * Eliminate all aligned sequences that weren't used in the mapping, for output purposes
     */
    private void extractAlignedSequences(List<GappedStructureAnnotatedSequence> alignedSequences, boolean[] alignmentUsed)
    {
        List<GappedStructureAnnotatedSequence> removals = new ArrayList<GappedStructureAnnotatedSequence>();
        for (int i = 0; i < alignedSequences.size(); i++)
        {
            GappedStructureAnnotatedSequence homologSequence = alignedSequences.get(i);
            if (!alignmentUsed[i])
            {
                removals.add(homologSequence);
            }
        }

        // remove sequences
        for (GappedStructureAnnotatedSequence s : removals)
        {
            alignedSequences.remove(s);
        }

        logger.logObject(Consts.ALIGNED_HOMOLOG_LIST, alignedSequences);
        logger.debug("AlignedHomologs has " + alignedSequences.size() + " sequences.");

        /* End elimination */
    }

    private List<GappedSequence> getHomologs(ScoreAnnotatedSequence sequence, double searchEvalue, Double identity)
    {
        List<GappedSequence> homologs = new ArrayList<GappedSequence>();
	List<GappedSequence> reducedBlastSequences = new ArrayList<GappedSequence>();
        Sequence blastedSequence = sequence;
        while (true)
        {
            BlastSettings settings = new BlastSettings();
            settings.setSequence(blastedSequence);
            BlastRunner blastRunner = new BlastRunner(settings);
            settings.setDatabase(Config.getConfig().getStringValue("PDB"));
            settings.setBlastType(BlastSettings.BLASTTYPE.PROTEUS_BLASTPARSER_BLASTP);
            settings.setEvalue(searchEvalue);
            blastRunner.run();

            List<GappedSequence> blastSequences = blastRunner.getAlignedResult();

            blastRunner.cleanupTempFiles();
			if (identity != null) {
				
				final double identityThreshold = identity.doubleValue();
				
				reducedBlastSequences = filterByPercentageIdentity (blastSequences, sequence, identityThreshold);
				blastSequences = reducedBlastSequences;
			}
			
			

            if (blastSequences.size() == 0)
            {
                break;
            }
			
			

            homologs.addAll(blastSequences);
	    //homologs.addAll(reducedBlastSequences);
            blastedSequence = blockOutAlignedRegions(blastedSequence, homologs);

            if (blastedSequence.seqString().replaceAll("X", "").replaceAll("-", "").length() == 0)
            {
                // If all that's left if X's, then don't search anymore (and BLAST will complain)
                break;
            }
        }
        return homologs;
		
		
	}
	
//*******************************************************************************************************************************************	
	private List<GappedSequence> filterByPercentageIdentity (List<GappedSequence> blastSequences, Sequence sequence, double identityThreshold)
	{
	
		List<GappedSequence> ReducedBlastSequences = new ArrayList<GappedSequence>();
		int SequenceLength = sequence.length();
		int identity = 0;
		double percentageIdentity = 0;
		double avgLength = 0;
		int shortestLength = 0;
		int i, j;
		
		for (GappedSequence blastSequence: blastSequences) {
		
		if(blastSequence.getName().equals("QUERY")){
		
			ReducedBlastSequences.add(blastSequence);

		}	
		else{
			identity = 0;
			
			String s = blastSequence.seqString();
			
			String blast = blastSequence.seqString().replaceAll("-","").replaceAll("~","");
			String query = sequence.seqString();	
	
			int blastSequenceLength = blastSequence.seqString().replaceAll("-","").replaceAll("~","").length();
			
			
			if (blastSequenceLength > SequenceLength) {
                               shortestLength = SequenceLength;
                        }
                        else{
                                shortestLength = blastSequenceLength;
                        }
	
			//avgLength = (SequenceLength + blastSequenceLength)/2;

			for (i = 0, j = 0; i < shortestLength; i++, j++) {
			//for (i = 0, j = 0; i < avgLength; i++, j++) {
				if (query.charAt (i) == blast.charAt(j) ){
				
					identity++;
				}	
			}
			
			
			
			//avgLength = (SequenceLength + blastSequenceLength)/2;
			
			//percentageIdentity = ((double)identity / (double)avgLength)* 100;
			percentageIdentity = ( (double) identity / (double) shortestLength) * 100;
			
			System.out.println(s);
			System.out.printf("\t%f\n",percentageIdentity);

			if (percentageIdentity <= identityThreshold){
				
				// add the sequence into the list of sequences
				System.out.println(s);
				ReducedBlastSequences.add(blastSequence);
			}
			
			
		  }//else ends here
		}//forloop ends here	
		
		
		return ReducedBlastSequences;
		
	}	
//********************************************************************************************************************************************	

    /**
     * Takes a sequence with a prediction on it already, and then overlays the structure of several homologs on
     * top of it to obtain a more accurate prediction
     *
     * @param blastedSequence the sequence that already has a prediction on it
     * @return a sequence with its structure annotation composed of the predicted structure + the homolog structure
     */
    private Sequence blockOutAlignedRegions(Sequence blastedSequence, List<GappedSequence> homologs)
    {
        char[] seqString = null;
        List<String> homologStrings = new ArrayList<String>(homologs.size());

        for (GappedSequence homolog : homologs)
        {
            if (!homolog.getName().equals("QUERY"))
            {
                homologStrings.add(homolog.seqString());
            }
            else
            {
                seqString = homolog.seqString().toCharArray();
            }
        }

        assert (seqString != null);

        for (int i = 0; i < seqString.length; i++)
        {
            for (String homolog : homologStrings)
            {
                if (homolog.length() >= i && !ProteusTools.isGap(homolog.charAt(i)) && seqString[i] != '-')
                {
                    seqString[i] = 'X';
                }
            }
        }

        try
        {
            blastedSequence = ProteinTools.createProteinSequence(new String(seqString), blastedSequence.getName());
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }

        return blastedSequence;
    }

    public static HomologCombiner getInstance()
    {
        if (instance == null)
        {
            instance = new HomologCombiner();
        }
        return instance;
    }

    public void setLogger(RuntimeCatalog logger)
    {
        this.logger = logger;
    }
}
