package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.program.MethodSettings;
import ca.ualberta.proteus.program.proteus.JNETRunner;
import ca.ualberta.proteus.program.proteus.PsipredRunner;
import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.util.io.StructureDB;
import ca.ualberta.proteus.util.db.DBType;
import ca.ualberta.proteus.util.db.PredictionDB;
import ca.ualberta.proteus.util.db.PredictionDBFactory;
import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.program.sax.blastxml.BlastXMLParserFacade;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleGappedSequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.seq.io.FastaAlignmentFormat;
import org.biojava.bio.seq.io.FastaFormat;
import org.biojava.bio.symbol.*;
import org.biojava.utils.ChangeVetoException;
import org.joone.engine.*;
import org.joone.engine.learning.TeachingSynapse;
import org.joone.io.MemoryInputSynapse;
import org.joone.net.NeuralNet;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import java.io.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Mar 8, 2005
 *         Time: 9:32:10 AM
 *         To change this template use File | Settings | File Templates.
 */
public class MainPredictor extends Predictor
{
    private static final int LAYER2 = UnifiedTrainingSet.NUM_INPUTS + 2 * UnifiedTrainingSet.NUM_OUTPUTS;
    ConsensusPredictor predictor = null;
    private List validationSequences;
    private int numToValidate = 20;
    private double bestFilteredQ3ScoreSoFar = 0;
    private double bestQ3ScoreSoFar = 0;
    private boolean verbose = false;
    private static int numToEval;
    public boolean useStoredNN = false;
    private static volatile NeuralNet storedNN;
    public boolean testing = false;
    private RuntimeCatalog logger = null;
    public static final double DEFAULT_EVALUE = 1e-7;
    private MethodSettings method = MethodSettings.ALL;
    private static boolean evaluating = false;
    private static final String BLAST_DIR = "blastdata";

//    Logger log4j = Logger.getLogger(MainPredictor.class);

    public MainPredictor(RuntimeCatalog logger) {
        this.logger = logger;
    }

    public MainPredictor(RuntimeCatalog logger, MethodSettings method) {
        this.logger = logger;
        this.method = method;
    }


    /**
     * Main running method
     *
     * @param sequence
     * @param alignment
     */
    public void run(Sequence sequence, Alignment alignment, String pssmFile, String checkpointFile, String blastOutputFileName, boolean homologySearch)
    {
        PredictionDB db;

        db = PredictionDBFactory.getRunTimeDB( evaluating );

        ScoreAnnotatedSequence psipredResult = null;
        ScoreAnnotatedSequence jnetResult = null;
        ScoreAnnotatedSequence proteusSecResult = null;

        if ( method == MethodSettings.ALL || method == MethodSettings.PSIPRED_JNET || method == MethodSettings.PSIPRED )
        {
            psipredResult = runPSIPRED(sequence, pssmFile, checkpointFile);
            if (psipredResult != null) {
                db.add(sequence, "PSIPRED", psipredResult);

                if ( logger != null)
                {
                    logger.logObject(Consts.PSIPRED, psipredResult);
                    logger.stepComplete(Consts.Steps.PSIPRED);
                }
            }
            else {
                if ( logger != null)
                    logger.error("PSIPRED failed");
            }
        }

        if ( method == MethodSettings.ALL || method == MethodSettings.PSIPRED_JNET || method == MethodSettings.JNET )
        {
            if ( logger != null)
                logger.log("Running JNET...");

            jnetResult = runJNET(sequence, alignment, pssmFile, blastOutputFileName);
            if (jnetResult != null) {
                db.add(sequence, "JNET", jnetResult);
                if ( logger != null)
                {
                    logger.logObject(Consts.JNET, jnetResult);
                    logger.stepComplete(Consts.Steps.JNET);
                }
            }
            else {
                if ( logger != null)
                    logger.error("JNET failed");
            }
        }

        if ( method == MethodSettings.ALL || method == MethodSettings.TRANSSEC )
        {
            proteusSecResult = runProteusSec(sequence, alignment);
            if (proteusSecResult != null) {
                db.add(sequence, "PROTEUS", proteusSecResult);
                if ( logger != null)
                {
                    logger.logObject(Consts.PROTEUS, proteusSecResult);
                    logger.stepComplete(Consts.Steps.PROTEUS);
                }
            }
            else {
                if ( logger != null)
                    logger.error("Proteus failed");
            }
        }

        if ( method == MethodSettings.ALL || method == MethodSettings.PSIPRED || method == MethodSettings.PSIPRED_JNET)
        {
            System.out.println("\nPSIPRED:" + psipredResult);
        }

        if ( method == MethodSettings.ALL || method == MethodSettings.JNET )
        {
            System.out.println("\nJNET:" + jnetResult);
        }

        if ( method == MethodSettings.ALL )
        {
            System.out.println("\nInternal:" + proteusSecResult);
        }

        // for the 'fast' version, use psipred for the 3rd prediction
        if ( method == MethodSettings.PSIPRED_JNET )
        {
            proteusSecResult = psipredResult;
        }

        cleanup(pssmFile, checkpointFile, blastOutputFileName);

        if ( method == MethodSettings.ALL || method == MethodSettings.PSIPRED_JNET )
        {
            UnifiedTrainingSet ts = new UnifiedTrainingSet();

            ts.addInstance(sequence, jnetResult, psipredResult, proteusSecResult);
        }
    }

    private ScoreAnnotatedSequence runTransSec(Sequence sequence, Alignment alignment)
    {
        try {
            TranssecRunner runner = new TranssecRunner(sequence, alignment);
            runner.run();
            return new ScoreAnnotatedSequence(runner.getResult());
        }
        catch (Throwable t) {
            t.printStackTrace();
            StringWriter out = new StringWriter();
            PrintWriter writer = new PrintWriter( out );
            t.printStackTrace(writer);
            logger.error(out.toString());
            return null;
        }
    }

    public List<ScoreAnnotatedSequence> searchForHomolog(Sequence sequence)
    {
        return searchForHomolog(sequence, DEFAULT_EVALUE, false);
    }

    public List<ScoreAnnotatedSequence> searchForHomolog(Sequence sequence, boolean isTM)
    {
        return searchForHomolog(sequence, DEFAULT_EVALUE, isTM);
    }

    public List<ScoreAnnotatedSequence> searchForHomolog(Sequence sequence, double evalue)
    {
        return searchForHomolog(sequence, evalue);
    }

    /**
     *
     * @param sequence
     * @param searchEvalue
     * @param isTM
     * @return list of sequences that could be used to overlay
     */
    public List<ScoreAnnotatedSequence> searchForHomolog(Sequence sequence, double searchEvalue, boolean isTM)
    {
        return searchForHomolog(sequence, searchEvalue, -1 , isTM);
    }

    /**
     * Returns a structure annotated sequence that is similar to the query sequence
     * @param sequence
     * @param searchEvalue - The evalue limit to give to BLAST
     * @param minEvalue - The evalue we want to limit it to to avoid finding exact homologs (for testing puposes).  Set to -1 for no limit.
     * @param isTM - set to true if the sequence is a transmembrane sequence
     * @return list of sequences that could be used to overlay
     */
    public List<ScoreAnnotatedSequence> searchForHomolog(Sequence sequence, double searchEvalue, double minEvalue, boolean isTM)
    {
        List<ScoreAnnotatedSequence> result = new ArrayList<ScoreAnnotatedSequence>();
        HashMap<ScoreAnnotatedSequence, Double> homologEvalues = new HashMap<ScoreAnnotatedSequence, Double>();
        boolean finished = false;

        while  ( ! finished )
        {
            BlastSettings settings = new BlastSettings();
            settings.setSequence(sequence);
            BlastRunner blastRunner = new BlastRunner(settings);
            settings.setDatabase( Config.getConfig().getStringValue("PDB"));
            settings.setBlastType(BlastSettings.BLASTTYPE.BLASTP);
            settings.setEvalue(searchEvalue);
            blastRunner.run();

            List<Sequence> homologs = blastRunner.getUnalignedResult();
            logger.logObject("homologs", homologs);

            blastRunner.cleanupTempFiles();

            StructureAnnotatedSequence newSequence = null;

            double smallestEvalue = Double.MAX_VALUE;
            String seqName = null;
            HashMap<Double, Sequence> evalueHomologMap = new HashMap<Double, Sequence>();
            for (Sequence seq : homologs) {
                Annotation annotation = seq.getAnnotation();
                Double homologEvalue = (Double) annotation.getProperty("e-value");
                boolean addToSet = (minEvalue == -1 || homologEvalue > minEvalue);
                if ( evaluating )
                {
                    if ( seq.getName().equalsIgnoreCase(sequence.getName()))
                    {
                        addToSet= false;
                    }
                }
                if ( addToSet)
                {
                    evalueHomologMap.put(homologEvalue, seq);

                    if (homologEvalue < smallestEvalue) {
                        smallestEvalue = homologEvalue;
                        seqName = seq.getName();
                    }
                }
            }

            Object[] evalues = evalueHomologMap.keySet().toArray();
            Arrays.sort(evalues);

            SeqSeeDB db = SeqSeeDB.getInstance();

            // Create ScoreAnnotatedSequences from the structures in the DB
            for(Object ev : evalues) {
                Double evalue = (Double) ev;
                String name = evalueHomologMap.get(ev).getName();
                //System.out.println( "Sequence: " + name + " evalue: " + evalue);

                try
                {
                    newSequence = (StructureAnnotatedSequence) db.getSequence(name);
                    //logger.logObject(Consts.HOMOLOG, newSequence);
                }
                catch (BioException e)
                {
                    e.printStackTrace();
                }
                ScoreAnnotatedSequence homolog = new ScoreAnnotatedSequence(newSequence);
                try
                {
                    homolog.getAnnotation().setProperty("e-value", evalue);
                }
                catch (ChangeVetoException e)
                {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }

                if ( evalue <= 10e-20)
                {
                    homolog.fillScores(9);
                }
                else if ( evalue <= 10e-13)
                {
                    homolog.fillScores(8);
                }
                else if ( evalue <= 10e-7) // not that this will ever happen
                {
                    homolog.fillScores(7);
                }

                result.add(homolog);
                homologEvalues.put(homolog, evalue);
            }

            finished = true;
        }

        if ( result.size() > 0 )
        {
            logger.logObject(Consts.HOMOLOG_LIST, result);
            logger.logObject(Consts.HOMOLOG_EVALUES, homologEvalues);
        }

        return result;
    }

    private void cleanup(String pssmFile, String checkpointFile, String blastOutputFileName)
    {
        File f = new File(pssmFile);
        f.delete();
        f = new File(checkpointFile);
        f.delete();
        f = new File(blastOutputFileName);
        f.delete();
    }

    private ScoreAnnotatedSequence runProteusSec(Sequence sequence, Alignment alignment) {
        return runProteusSec(sequence,  alignment, true);
    }

    private ScoreAnnotatedSequence runProteusSec(Sequence sequence, Alignment alignment, boolean doFilter)
    {
        if ( logger != null )
            logger.log("Running Internal Predictor...");
        if (predictor == null)
        {
            predictor = new ConsensusPredictor();

            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);
        }

        ScoreAnnotatedSequence prediction = predictor.predict(sequence, alignment);

        return prediction;
    }

     private ScoreAnnotatedSequence runPSIPRED(Sequence sequence, String pssmFile, String checkpointFile)
    {
        if ( logger != null)
            logger.log("Running PSIPRED...");
        FastaFormat fastaWriter = new FastaFormat();
        String inputFile = getUniqueIdentifier() + ".in";
        try
        {
            PrintStream os = new PrintStream(inputFile);
            fastaWriter.writeSequence(sequence, os);
            os.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        PsipredRunner psipredRunner = new PsipredRunner();
        psipredRunner.querySequence = sequence;
        psipredRunner.setParams(pssmFile, checkpointFile, inputFile);
        psipredRunner.run();
        ScoreAnnotatedSequence psipredResult = psipredRunner.getPrediction();

        File f = new File(inputFile);
        if ( f != null )
            f.delete();

        assert( psipredResult!= null && sequence.length() == psipredResult.length());
        return psipredResult;
    }

    private ScoreAnnotatedSequence runJNET(Sequence sequence, Alignment alignment, String pssmFile, String blastOutputFileName)
    {
        //TODO HACK THIS OUT
        // TODO JNET needs a -m6 blast output, the blast runner by default outputs XML (-m7)
        // do BLAST
        if ( false)
        {
            BlastSettings settings = new BlastSettings();
            settings.setSequence(sequence);
            if ( testing )
                settings.setDatabase( Config.getConfig().getStringValue("PDB"));
            settings.setBlastType(BlastSettings.BLASTTYPE.PSIBLAST);
            settings.setAlignmentView(BlastSettings.ALIGNMENT_VIEW.BLUNT_ENDS); // TODO THIS IS THE HACK
            settings.setParse(false);


            BlastRunner blastRunner = new BlastRunner(settings);
            blastRunner.run();
            Map<BlastRunner.OUTPUT_FILE_TYPE, String> outputFileMap = blastRunner.getOutputFileMap();
            blastOutputFileName = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.OUTPUT);

            // cleanup unused files
            File f = new File( outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.CHECKPOINT));
            f.delete();
            f = new File( outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.PSSM));
            f.delete();
           // blastRunner.cleanupTempFiles();
        }

        // TODO END HACKING

        // write out alignment
        String alignmentFile = getUniqueIdentifier() + ".align";
        try
        {
            // write out query sequence first
            FileOutputStream os = new FileOutputStream(alignmentFile);
            FastaFormat fastaWriter = new FastaFormat();
            fastaWriter.writeSequence(sequence, new PrintStream(os));

            // then alignment
            FastaAlignmentFormat writer = new FastaAlignmentFormat();
            writer.writeProtein(os, alignment);
            os.close();
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        JNETRunner jnetRunner = new JNETRunner();
        jnetRunner.querySequence = sequence;
        jnetRunner.alignment = alignment;
        jnetRunner.setParams(pssmFile, alignmentFile, blastOutputFileName);
        jnetRunner.run();
        ScoreAnnotatedSequence jnetResult = jnetRunner.getPrediction();

        File f = new File(alignmentFile);
        if (!f.delete())
            System.err.println("Unable to delete " + alignmentFile);
        f = new File(alignmentFile + ".hmmprof");
        if ( f.exists() )
            f.delete();

        f = new File(pssmFile);
        f.delete();
        f = new File(blastOutputFileName);
        f.delete();

        assert(jnetResult.length() == 0 || sequence.length() == jnetResult.length());

        return jnetResult;
    }

    /**
     * Method you can call without an alignment - it will perform a BLAST search to obtain the alignment
     * and continue on
     *
     * @param sequence
     */
    public void run(Sequence sequence, boolean homologySearch)
    {
        Map<BlastRunner.OUTPUT_FILE_TYPE, String> outputFileMap = new HashMap<BlastRunner.OUTPUT_FILE_TYPE, String>();
        List<Sequence> blastSequences = new ArrayList<Sequence>();
        Alignment alignment = null;

        // can we find a stored version of the BLAST file?
        //boolean haveStoredBlastResult = findBlast(sequence.getName(), outputFileMap, blastSequences);
        boolean haveStoredBlastResult = false;
        if ( ! haveStoredBlastResult )
        // TODO Put this back in later
        {
            // do BLAST
            BlastSettings settings = new BlastSettings();
            settings.setSequence(sequence);
            if ( testing ) {
                settings.setDatabase( Config.getConfig().getStringValue("PDB"));
                settings.setBlastType(BlastSettings.BLASTTYPE.PSIBLAST);
            }
            else
                settings.setBlastType(BlastSettings.BLASTTYPE.PSIBLAST);

            settings.setAlignmentView(BlastSettings.ALIGNMENT_VIEW.BLUNT_ENDS);

            if ( logger != null)
                logger.log("Running PSI-BLAST...");

            BlastRunner blastRunner = new BlastRunner(settings);
            blastRunner.run();
            blastSequences = blastRunner.getUnalignedResult();

            List<GappedSequence> alignedResult = blastRunner.getAlignedResult();
            ProteusTools.chooseAlignments(sequence, alignedResult, 50 );



            if (alignedResult.size() == 0)
            {
                /* If no sequences were found, fake the alignment */
                try
                {
                    Sequence s = ProteinTools.createProteinSequence(sequence.seqString(), "QUERY");
                    alignedResult.add(new SimpleGappedSequence(s));
                }
                catch (IllegalSymbolException e)
                {
                    e.printStackTrace();
                }

                alignedResult.add(new SimpleGappedSequence(sequence));
            }

            // create the alignment
            alignment = ProteusTools.createAlignment( alignedResult );
            logger.logObject(Consts.ALIGNMENT, alignment);

            outputFileMap = blastRunner.getOutputFileMap();

            if ( logger != null)
            {
                logger.logObject(Consts.BLAST_SEQUENCES, blastSequences);
                logger.stepComplete(Consts.Steps.BLAST);
            }

        }

        blastSequences.add(sequence);

        String numAlignments = Config.getConfig().getStringValue("num_alignments");
        ProteusTools.chooseAlignments(sequence, blastSequences, Integer.valueOf(numAlignments) + 100 - 1); // leave 100 in there to be reduced later on
        if ( logger != null)
            logger.logObject(Consts.REDUCED_SEQUENCES, blastSequences);

        assert blastSequences.contains(sequence) : "Alignment sequences does not contain query sequence";

        if ( testing && !haveStoredBlastResult )
        {
            saveBlast(sequence, outputFileMap);
        }

        run(sequence, alignment, outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.PSSM),
            outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.CHECKPOINT),
            outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.OUTPUT),
            true);

        //cleanup(outputFileMap.values());
    }

    private void cleanup(Collection<String> outputFileMap)
    {
        for ( String s : outputFileMap ) {
            File f = new File(s);
            if ( ! f.delete() )
            {
                if ( f.exists() )
                {
                    System.out.print(f + " exists...");
                }
                System.err.println("MainPredictor.cleanup(): Error deleting " + f);
            }
        }
    }

    public void train(List sequences)
    {
        PredictionDB db = PredictionDBFactory.getDB();
        UnifiedTrainingSet ts = new UnifiedTrainingSet();

        for (Iterator it = sequences.iterator(); it.hasNext();)
        {

            StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();

            if (!db.contains(sequence))
                run(sequence, false);

            ScoreAnnotatedSequence jnetResult = db.get(sequence, "JNET");
            ScoreAnnotatedSequence psipredResult = db.get(sequence, "PSIPRED");
            //ScoreAnnotatedSequence transsecResult = db.get(sequence, "TRANSSEC");
            ScoreAnnotatedSequence proteusSecResult = db.get(sequence, "PROTEUS");

            // verify that they are in fact in the database
            if (jnetResult == null || psipredResult == null || proteusSecResult == null)
            {
                run(sequence, false);

                jnetResult = db.get(sequence, "JNET");
                psipredResult = db.get(sequence, "PSIPRED");
                proteusSecResult = db.get(sequence, "PROTEUS");
            }

           // if ( transsecResult == null )
            //    continue;
            ts.addInstance(sequence, jnetResult, psipredResult, proteusSecResult);
        }
/*
        try
        {
            ts.writeToStream(new FileOutputStream("unified.ts"));
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
  */
        /* The Layers */
        LinearLayer input = new LinearLayer();
        input.setRows(ts.getNumInputs());

        SigmoidLayer hidden = new SigmoidLayer();
        hidden.setRows(LAYER2);

        SigmoidLayer output = new SigmoidLayer();
        output.setRows(ts.getNumOutputs());

        /* The Synapses */
        FullSynapse synapse_IH = new FullSynapse(); /* Input -> Hidden conn. */
        input.addOutputSynapse(synapse_IH);
        hidden.addInputSynapse(synapse_IH);

        FullSynapse synapse_HO = new FullSynapse(); /* Hidden -> Output conn. */
        hidden.addOutputSynapse(synapse_HO);
        output.addInputSynapse(synapse_HO);

        int numInstances = ts.numInstances();

        /* The I/O components */
        MemoryInputSynapse inputStream = new MemoryInputSynapse();
        inputStream.setInputArray(ts.getInputArray());
        inputStream.setAdvancedColumnSelector("1-" + ts.getNumInputs());
        input.addInputSynapse(inputStream);

        //int split = (int) (numInstances * 0.66);

        TeachingSynapse trainer = new TeachingSynapse();
        MemoryInputSynapse samples = new MemoryInputSynapse();
        samples.setInputArray(ts.getInputArray());
        trainer.setDesired(samples);

        samples.setAdvancedColumnSelector(String.valueOf(ts.getNumInputs() + 1) + "-" + String.valueOf(ts.getNumInputs() + ts.getNumOutputs()));
        output.addOutputSynapse(trainer);

        nnet = new NeuralNet();
        nnet.addLayer(input, NeuralNet.INPUT_LAYER);
        nnet.addLayer(hidden, NeuralNet.HIDDEN_LAYER);
        nnet.addLayer(output, NeuralNet.OUTPUT_LAYER);
        nnet.setTeacher(trainer);

        Monitor monitor = nnet.getMonitor();
        monitor.setLearningRate(0.005);   // 0.8   .005
        monitor.setMomentum(0.9);         // 0.3  .9
        monitor.setTrainingPatterns(numInstances); /* # of rows in the input file */
        monitor.setValidationPatterns(0);
        monitor.setTotCicles(500); /* How many times the net must be trained*/
        monitor.setLearning(true); /* The net must be trained */
        monitor.addNeuralNetListener(this);


        System.out.println("Num instances = " + numInstances);
        PredictionScore instanceDistribution = new PredictionScore();
        for (Iterator it = ts.getOutputInstanceIterator(); it.hasNext();)
        {
            double[] outputInstance = (double[]) it.next();
            instanceDistribution.helix += outputInstance[ProteusTools.HELIX];
            instanceDistribution.coil += outputInstance[ProteusTools.COIL];
            instanceDistribution.gap += outputInstance[ProteusTools.GAP];
            instanceDistribution.strand += outputInstance[ProteusTools.STRAND];
        }
        System.out.println("Instance distribution: ");
        System.out.println("Helix = " + instanceDistribution.helix);
        System.out.println("Coil = " + instanceDistribution.coil);
        System.out.println("Strand = " + instanceDistribution.strand);
        System.out.println("Gap = " + instanceDistribution.gap);


        nnet.start();
        nnet.getMonitor().Go();
    }

    public static void main(String[] arg) throws IOException
    {
        boolean train = false;
        boolean eval = false;
        int numToLoad = -1;
        int numToTrainOn = -1;
        int numToStartAt = -1;

        for (int i = 0; i < arg.length; i++)
        {
            if (arg[i].equals("-load"))
            {
                try
                {
                    numToLoad = Integer.parseInt(arg[++i]);
                }
                catch (NumberFormatException e)
                {
                    //printUsage();
                }
                if (i < arg.length - 1)
                    try
                    {
                        numToStartAt = Integer.parseInt(arg[++i]);
                    }
                    catch (NumberFormatException e)
                    {
                    }
            }
            else if (arg[i].equals("-train")) {
                try
                {
                    train = true;
                    numToTrainOn = Integer.parseInt(arg[++i]);
                }
                catch (NumberFormatException e)
                {
                    //printUsage();
                }
            }
            else if (arg[i].equals("-eval")) {
                 try
                {
                    eval = true;
                    numToEval = Integer.parseInt(arg[++i]);
                }
                catch (NumberFormatException e)
                {
                    printUsage();
                }
                if (i < arg.length - 1)
                    try
                    {
                        numToStartAt = Integer.parseInt(arg[++i]);
                    }
                    catch (NumberFormatException e)
                    {
                    }
            }
            else if (arg[i].equals("-ext") || arg[i].equals("-external")) {
                evaluating = true;
                if ( arg.length == 2 )
                    externalEval(arg[++i]);
                else if ( arg.length == 3 )
                    externalEval(arg[++i], arg[++i]);
                else if ( arg.length == 4 )
                    externalEval(arg[++i], arg[++i], arg[++i]);
                else
                    externalEval(arg[++i], arg[++i], arg[++i], arg[++i]);
                return;
            }
            else if (arg[i].equals("-cleanjnet") )
            {
                StructureDB db = new StructureDB(arg[++i], StructureDB.MODE.READ);
                int start = Integer.parseInt(arg[++i]);
                int end = Integer.parseInt(arg[++i]);
                PredictionDB predictionDB = PredictionDBFactory.getDB(DBType.DB4O);

                for ( int j = start; j< end; j++ ) {
                    try
                    {
                        StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) db.getSequence(i);
                        predictionDB.remove(sequence, "JNET");
                        if ( j % 10 == 0)
                            System.out.println("Record: " + j);
                    }
                    catch (BioException e)
                    {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
                return;
            }
            else if (arg[i].equals("-jnet") ) {
                evaluating = true;

                    jnetEval(arg[++i], Integer.parseInt(arg[++i]), Integer.parseInt(arg[++i]));
                return;
            }

        }

        SeqSeeDB db = SeqSeeDB.getInstance();

        RuntimeCatalog logger = new RuntimeCatalog();
        if (train)
        {
            // train on this set of sequences
            Predictor predictor = new MainPredictor(logger);

            // Have to add the filters in this order - use the NN filter, then the reg exp filter
            final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
            predictor.addFilter(ps2Ffilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);

            //List randomSequences = db.getEvaluationSet(numToTrainOn, 4000, 6000);
            List randomSequences = db.getEvaluationSet2(numToTrainOn, 1, 6000, 87);
            /*List randomSequences = new ArrayList();
            try
            {
                for (int i = 0; i < numToTrainOn; i++)
                    randomSequences.add((StructureAnnotatedSequence) db.get(4000 + i));
            }
            catch (BioException e)
            {
                e.printStackTrace();
            } */

            predictor.train(randomSequences);
        }
        else if (numToLoad > 0)
        {
            MainPredictor mainPredictor = new MainPredictor(logger);
            try
            {
                if (numToStartAt == -1)
                    numToStartAt = 4000;
                for (int i = 0; i < numToLoad; i++)
                    mainPredictor.run((StructureAnnotatedSequence) db.getSequence(numToStartAt + i), false);
            }
            catch (BioException e)
            {
                e.printStackTrace();
            }
        }
        else if ( eval ) {
            MainPredictor mainPredictor = new MainPredictor(logger);
            mainPredictor.useStoredNN = true;
            Score score = new Score();
            Score jnet = new Score();
            Score psipred = new Score();
            Score proteus = new Score();

            PredictionDB predictionDB = PredictionDBFactory.getDB(DBType.DB4O);
            NumberFormat format = NumberFormat.getNumberInstance();
            format.setMaximumFractionDigits(4);

            List<StructureAnnotatedSequence> randomSequences  = db.getEvaluationSet2(numToEval, 11, 6000, 37);
            for ( StructureAnnotatedSequence sequence : randomSequences) {
                //StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) db.get(numToStartAt + i);
                ScoreAnnotatedSequence finalSequence = mainPredictor.predict(sequence);
                ScoreAnnotatedSequence jnetSequence = predictionDB.get(sequence, "JNET");
                ScoreAnnotatedSequence psipredSequence = predictionDB.get(sequence, "PSIPRED");
                ScoreAnnotatedSequence proteusSequence = predictionDB.get(sequence, "PROTEUS");

                double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), jnetSequence.getStructureAnnotation());
                double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), jnetSequence.getStructureAnnotation());
                jnet.q3 += q3;
                jnet.sov += sov;

                System.out.println("JNET:\t"
                                + "Q3:" + format.format(q3)
                                + "\tSOV: " + format.format(sov) );


                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), psipredSequence.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), psipredSequence.getStructureAnnotation());
                psipred.q3 += q3;
                psipred.sov += sov;

                System.out.println("PSIPRED:\t"
                                + "Q3: " + format.format(q3)
                                + "\tSOV: " + format.format(sov) );


                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), proteusSequence.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), proteusSequence.getStructureAnnotation());
                proteus.q3 += q3;
                proteus.sov += sov;

                System.out.println("PROTEUS:\t"
                                + "Q3: " + format.format(q3)
                                + "\tSOV: " + format.format(sov) );

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), finalSequence.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), finalSequence.getStructureAnnotation());
                score.q3 += q3;
                score.sov += sov;

                System.out.println("FINAL:\t"
                                   + "Q3: " + format.format(q3)
                                   + "\tSOV: " + format.format(sov) );
            }

            System.out.println("JNET:\\tt"
                            + "Q3:" + format.format(jnet.q3 / (double) numToEval)
                            + "\tSOV: " + format.format(jnet.sov / (double) numToEval) );

            System.out.println("PSIPRED:\t"
                            + "Q3: " + format.format(psipred.q3 / (double) numToEval)
                            + "\tSOV: " + format.format(psipred.sov / (double) numToEval) );

            System.out.println("PROTEUS:\t"
                            + "Q3: " + format.format(proteus.q3 / (double) numToEval)
                            + "\tSOV: " + format.format(proteus.sov / (double) numToEval) );

            System.out.println("FINAL:\\tt"
                               + "Q3: " + format.format(score.q3 / (double) numToEval)
                               + "\tSOV: " + format.format(score.sov / (double) numToEval) );
        }
        else
        {
            System.out.println("Running on test sequence...");
            MainPredictor mainPredictor = new MainPredictor(logger);
            try
            {
                mainPredictor.run((StructureAnnotatedSequence) db.getSequence("1D3YB"), true);
            }
            catch (BioException e)
            {
                e.printStackTrace();
            }
        }
    }

    private static void externalEval(String filename, String listFile, String start, String end) {
        ExternalEvaluator evaluator = new ExternalEvaluator(filename);
        File f = new File(listFile);
        if ( f.exists() )
            try
            {
                evaluator.run(f, Integer.parseInt(start), Integer.parseInt(end));
            }
            catch (IOException e)
            {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
        else
            throw new IllegalArgumentException("File " + listFile + " does not exist");
    }

    private static void externalEval(String filename, String listFile)
    {
        ExternalEvaluator evaluator = new ExternalEvaluator(filename);
        File f = new File(listFile);
        if ( f.exists() )
            try
            {
                evaluator.run(f);
            }
            catch (IOException e)
            {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
        else
            throw new IllegalArgumentException("File " + listFile + " does not exist");
    }

    private static void externalEval(String filename, String start, String end)
    {
         ExternalEvaluator evaluator = new ExternalEvaluator(filename);
        try
        {
            evaluator.run(Integer.parseInt(start), Integer.parseInt(end));
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    private static void jnetEval(String filename, int start, int end)
    {
        StructureDB db = null;
        try
        {
            db = new StructureDB(filename, StructureDB.MODE.READ);
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        if ( start < 0 )
            start = 0;
        if ( end < 0 )
            end = db.size();

        MainPredictor mainPredictor = new MainPredictor(new RuntimeCatalog(), MethodSettings.JNET);
        mainPredictor.useStoredNN = true;
        PredictionDB predictionDB = PredictionDBFactory.getDB(DBType.DB4O);

        for ( int i = start; i< end; i++ ) {
            try
            {
                StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) db.getSequence(i);

                if (sequence.length() < 3 || sequence.length() > 1000)
                {
                    System.out.println("Skipping " + sequence.getName() + " length: " + sequence.length() );
                    continue;
                }

                System.out.println(sequence.getName());

                if ( predictionDB.contains(sequence, "JNET"))
                {
                    predictionDB.remove(sequence, "JNET");
                }
                // assume that JNET has been cleaned out already!!
                mainPredictor.predict(sequence);
                System.out.println("Record: " + i);
            }
            catch (BioException e)
            {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
            catch( Exception e)
            {
                // do nothing
            }
        }
    }

    /**
     * Does an evaluation of a bunch of sequences
      * @param filename
     */
    private static void externalEval(String filename)
    {
        ExternalEvaluator evaluator = new ExternalEvaluator(filename);
        try
        {
            evaluator.run();
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    /**
     * Takes a sequence with a prediction on it already, and then overlays the structure of several homologs on
     * top of it to obtain a more accurate prediction
     *
     * @param predictedSequence - the sequence that already has a prediction on it
     * @param homologSequences
     * @return a sequence with its structure annotation composed of the predicted structure + the homolog structure
     */
    public ScoreAnnotatedSequence overlayHomolog(ScoreAnnotatedSequence predictedSequence, List<ScoreAnnotatedSequence> homologSequences)
    {
        logger.debug("Overlaying list (size = " + homologSequences.size() + ") of sequences...");

        if ( homologSequences.size() == 1)
        {
            return overlayHomolog(predictedSequence, homologSequences.get(0));
        }

        List<Sequence> seqList = new ArrayList<Sequence>(homologSequences.size() + 1);
        seqList.add(predictedSequence);
        seqList.addAll(homologSequences);

        Alignment alignment;
   /*     try
        {
            XAlignRunner runner = new XAlignRunner(predictedSequence, seqList);
            runner.setAddToDB(false);
            runner.setUseCache(false);
            runner.run();
            alignment = runner.getAlignment();
        }
        catch ( ProcessNotCompleteException e )
     */
        {
            // Try ClustalW ?
            ClustalWRunner runner = new ClustalWRunner(predictedSequence, seqList);
            runner.setAddToDB(false);
            runner.setUseCache(false);
            runner.run();
            alignment = runner.getAlignment();
        }
        SymbolList alignedQuery = alignment.symbolListForLabel(predictedSequence.getName());

        Symbol[] newStructure = new Symbol[ alignedQuery.length() ];
        PredictionScore[] newScore = new PredictionScore[ alignedQuery.length() ];

        List<GappedStructureAnnotatedSequence> alignedHomologues = new ArrayList<GappedStructureAnnotatedSequence>();

        // Create a gappedstructure for the query sequence
        GappedStructureAnnotatedSequence gappedQuery = new GappedStructureAnnotatedSequence(new SimpleSequence(alignedQuery, "", "QUERY", null));
        gappedQuery.createGappedStructure(predictedSequence.getStructureAnnotation());

        // make a list of all the aligned homologues
        for( ScoreAnnotatedSequence homologSequence : homologSequences)
        {
            SymbolList homolog = alignment.symbolListForLabel(homologSequence.getName());

            assert alignedQuery.length() == homolog.length() : "Aligned query does not match aligned homolog";

            // Create a gappedstructure object for the homolog
            GappedStructureAnnotatedSequence gappedHomolog = new GappedStructureAnnotatedSequence(new SimpleSequence(homolog, "", homologSequence.getName(), homologSequence.getAnnotation()));
            gappedHomolog.createGappedStructure(homologSequence.getStructureAnnotation());

            alignedHomologues.add(gappedHomolog);
        }

        List<GappedStructureAnnotatedSequence> alignedHomologuesClone = new ArrayList<GappedStructureAnnotatedSequence>();
        alignedHomologuesClone.addAll(alignedHomologues);

        logger.logObject(Consts.ALIGNED_HOMOLOG_LIST, alignedHomologues);
        // clone it so we can remove stuff from it and not change it in the logger
        logger.logObject(Consts.ALL_ALIGNED_HOMOLOG_LIST, alignedHomologuesClone);
        logger.logObject(Consts.ALIGNED_QUERY, gappedQuery);
        logger.logObject(Consts.ALIGNED_HOMOLOG, alignedHomologues.get(0));

        // Weed out the ones that are overlapping
        List<Sequence> removals = new ArrayList<Sequence>();
        for( int i=0; i<homologSequences.size(); i++)
        {
            GappedStructureAnnotatedSequence homologSequence = alignedHomologues.get(i);
            for ( int j=i+1; j<homologSequences.size(); j++)
            {
                GappedStructureAnnotatedSequence comparedHomoSequence = alignedHomologues.get(j);
                if ( homologSequence.overlaps(comparedHomoSequence, 20))
                {
                    removals.add(comparedHomoSequence);
                }
               /*
                if ( homologSequence.residuesNotOverlapping(comparedHomoSequence) < 20)
                {
                    removals.add(comparedHomoSequence);
                }
                */
            }
        }
        // remove overlapping sequences
        for(Sequence s : removals)
        {
            alignedHomologues.remove(s);
        }

        logger.debug("AlignedHomologs has " + alignedHomologues.size() + " sequences.");

        for( GappedStructureAnnotatedSequence gappedHomolog : alignedHomologues)
        {
            //log4j.debug(gappedHomolog.seqString());

          //  logger.logObject(Consts.ALIGNED_QUERY, gappedQuery);
          //  logger.logObject(Consts.ALIGNED_HOMOLOG, gappedHomolog);

            Symbol gapSymbol = gappedHomolog.getStructureAnnotation().getAlphabet().getGapSymbol();
            String homologString = gappedHomolog.getStructureAnnotation().seqString();
            String queryString = gappedQuery.getStructureAnnotation().seqString();

            for ( int i=0; i < gappedHomolog.length() && i < gappedQuery.length(); i++)
            {
                Symbol homologSymbol = gappedHomolog.getStructureAnnotation().symbolAt(i+1);
                Symbol querySymbol = gappedQuery.getStructureAnnotation().symbolAt(i+1);
                if ( ! homologSymbol.equals(AlphabetManager.getGapSymbol()) && homologString.charAt(i) != '-')
                {
                    // homolog is not showing a gap, use homolog structure
                    newStructure[i] = homologSymbol;
                    newScore[i] = homologSequences.get(0).getScore(1);  // they're all uniform anyway
                }
                else if ( !querySymbol.equals(gapSymbol) && queryString.charAt(i) != '-')
                {
                    if ( newScore[i] == null )
                    {
                        // if the gapped query sequence isn't gapped here, use the predicted structure
                        newStructure[i] = querySymbol;
                        ///newScore[i] = null;
                    }
                }
            }
        }

        ScoreAnnotatedSequence result = null;
        result = new ScoreAnnotatedSequence(new SimpleSequence(alignedQuery, "", predictedSequence.getName(), null));
        result.setStructureAnnotation(new SimpleSymbolList(newStructure, newStructure.length, new StructureTools().getStructure()));
        result.setScores(newScore);

        // go back to a ungapped structure
        result = GappedStructureAnnotatedSequence.createUngappedStructure(predictedSequence, result);

        for ( int i=1; i<=result.length(); i++)
        {
            if ( result.getScore(i) == null || result.getScore(i).getMax() == 0)
            {
                result.setScore(i, predictedSequence.getScore(i));
            }
        }

        SimpleFilter filter = new SimpleFilter();
        result = filter.filter(result);

        /// copy the annotation
        if ( alignedHomologues.size() > 0)
        {
            Annotation annotation = alignedHomologues.get(0).getAnnotation();
            for (Object key : annotation.keys() )
            {
                try
                {
                    result.getAnnotation().setProperty(key, annotation.getProperty(key));
                }
                catch (ChangeVetoException e)
                {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }
            }
        }

        return result;
    }

    /**
     * Takes a sequence with a prediction on it already, and then overlays the structure of a homolog on
     * top of it to obtain a more accurate prediction
     *
     * @param predictedSequence - the sequence that already has a prediction on it
     * @param homologSequence
     * @return a sequence with its structure annotation composed of the predicted structure + the homolog structure
     */
    public ScoreAnnotatedSequence overlayHomolog(ScoreAnnotatedSequence predictedSequence, ScoreAnnotatedSequence homologSequence)
    {
        List<Sequence> seqList = new ArrayList<Sequence>(2);
        seqList.add(predictedSequence);
        seqList.add(homologSequence);

        Alignment alignment;
        try
        {
            XAlignRunner runner = new XAlignRunner(predictedSequence, seqList);
            runner.setAddToDB(false);
            runner.setUseCache(false);
            runner.run();
            alignment = runner.getAlignment();
        }
        catch ( ProcessNotCompleteException e )
        {
            // Try ClustalW ?
            ClustalWRunner runner = new ClustalWRunner(predictedSequence, seqList);
            runner.setAddToDB(false);
            runner.setUseCache(false);
            runner.run();
            alignment = runner.getAlignment();
        }
        SymbolList alignedQuery = alignment.symbolListForLabel(predictedSequence.getName());
        SymbolList homolog = alignment.symbolListForLabel(homologSequence.getName());

        assert alignedQuery.length() == homolog.length() : "Aligned query does not match aligned homolog";

        Symbol[] newStructure = new Symbol[ alignedQuery.length() ];
        PredictionScore[] newScore = new PredictionScore[ alignedQuery.length() ];

        // Create a gappedstructure object for the homolog
        GappedStructureAnnotatedSequence gappedHomolog = new GappedStructureAnnotatedSequence(new SimpleSequence(homolog, "", homologSequence.getName(), null));
        gappedHomolog.createGappedStructure(homologSequence.getStructureAnnotation());

        // And for the query sequence
        GappedStructureAnnotatedSequence gappedQuery = new GappedStructureAnnotatedSequence(new SimpleSequence(alignedQuery, "", "QUERY", null));
        gappedQuery.createGappedStructure(predictedSequence.getStructureAnnotation());

        List<GappedStructureAnnotatedSequence> alignedHomologues = new ArrayList<GappedStructureAnnotatedSequence>();
        alignedHomologues.add(gappedHomolog);

        logger.logObject(Consts.ALIGNED_HOMOLOG_LIST, alignedHomologues);
        logger.logObject(Consts.ALIGNED_QUERY, gappedQuery);
        logger.logObject(Consts.ALIGNED_HOMOLOG, gappedHomolog);

        Symbol gapSymbol = gappedHomolog.getStructureAnnotation().getAlphabet().getGapSymbol();

        String homologString = gappedHomolog.getStructureAnnotation().seqString();
        String queryString = gappedQuery.getStructureAnnotation().seqString();

        for ( int i=0; i < gappedHomolog.length() && i < gappedQuery.length(); i++)
        {
            Symbol homologSymbol = gappedHomolog.getStructureAnnotation().symbolAt(i+1);
            Symbol querySymbol = gappedQuery.getStructureAnnotation().symbolAt(i+1);
            if ( ! homologSymbol.equals(AlphabetManager.getGapSymbol()) && homologString.charAt(i) != '-')
            {
                // homolog is not showing a gap, use homolog structure
                newStructure[i] = homologSymbol;
                newScore[i] = homologSequence.getScore(1);  // they're all uniform anyway
            }
            else if ( !querySymbol.equals(gapSymbol) && queryString.charAt(i) != '-')
            {
                if ( newScore[i] == null )
                {
                    // if the gapped query sequence isn't gapped here, use the predicted structure
                    newStructure[i] = querySymbol;
                    ///newScore[i] = null;
                }
            }
        }


        ScoreAnnotatedSequence result;
        result = new ScoreAnnotatedSequence(new SimpleSequence(alignedQuery, "", predictedSequence.getName(), null));
        result.setStructureAnnotation(new SimpleSymbolList(newStructure, newStructure.length, new StructureTools().getStructure()));
        result.setScores(newScore);

        // go back to a ungapped structure
        result = GappedStructureAnnotatedSequence.createUngappedStructure(predictedSequence, result);

        for ( int i=1; i<=result.length(); i++)
        {
            if ( result.getScore(i) == null || result.getScore(i).getMax() == 0)
                result.setScore(i, predictedSequence.getScore(i));
        }

        SimpleFilter filter = new SimpleFilter();
        result = filter.filter(result);

        try
        {
            result.getAnnotation().setProperty("e-value", homologSequence.getAnnotation().getProperty("e-value"));
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        return result;
    }

    private static void printUsage()
    {
        System.out.println("Usage: <ProgName> (-train <# seq> | -eval <#seq> | -load <#seq> [#startAt]) [options]" );
        System.out.println("Options: ");
        System.out.println("-train <# sequences>\t\tTrain on a number of sequences");
        System.out.println("-eval <# sequences>\t\tEvaluate a number of sequences");
        System.out.println("-load <# sequences> [ID# to start at]");
        System.out.println("-ext <# Name of sequence database>");
        System.exit(-1);
    }

    public void run()
    {
        throw new IllegalArgumentException("Must pass a sequence - use other overloaded method");
    }

    public ScoreAnnotatedSequence predict(Sequence sequence, NeuralNet localNNet, boolean doFilter)
    {
        PredictionDB db;

        db = PredictionDBFactory.getRunTimeDB( evaluating );

        if (  ! evaluating
            && ( ! Config.getConfig().getBooleanValue ("useCache")
            || (!db.contains(sequence) )))
        {
            run(sequence, false);
        }

        ScoreAnnotatedSequence jnetResult = db.get(sequence, "JNET");
        ScoreAnnotatedSequence psipredResult = db.get(sequence, "PSIPRED");
        ScoreAnnotatedSequence proteusSecResult = db.get(sequence, "PROTEUS");
        ScoreAnnotatedSequence transsecResult = db.get(sequence, "PROTEUS");

        // If we aren't predicting all methods, we can't do a consensus!
        if ( method != MethodSettings.ALL && method != MethodSettings.PSIPRED_JNET )
        {
            return null;
        }

        if ( jnetResult == null
                || psipredResult == null
                || ( method != MethodSettings.PSIPRED_JNET && proteusSecResult == null) )
        {
            run(sequence, false);
        }

        if (jnetResult == null )
        {
            jnetResult = db.get(sequence, "JNET");
        }
        if (psipredResult == null )
        {
            psipredResult = db.get(sequence, "PSIPRED");
        }
        if ( proteusSecResult == null )
        {
            proteusSecResult = db.get(sequence, "PROTEUS");
        }
        if ( transsecResult == null )
        {
            transsecResult = db.get(sequence, "PROTEUS");
        }

        // for the 'fast' version, use psipred for the 3rd prediction
        if ( method == MethodSettings.PSIPRED_JNET)
        {
            transsecResult = psipredResult;
            proteusSecResult = psipredResult;
        }

        // verify that they are in fact in the database
        if (jnetResult == null || psipredResult == null || transsecResult == null)
        {
            run(sequence, false);

            if (jnetResult == null )
            {
                jnetResult = db.get(sequence, "JNET");
            }
            if (psipredResult == null )
            {
                psipredResult = db.get(sequence, "PSIPRED");
            }
            if ( proteusSecResult == null )
            {
                proteusSecResult = db.get(sequence, "PROTEUS");
            }
            if ( transsecResult == null )
            {
                transsecResult = db.get(sequence, "PROTEUS");
            }
        }

        // TODO hack, this shouldn't ever happen
        if ( transsecResult == null ) {
            transsecResult = psipredResult;
            System.err.println("TransSec Failed");
            logger.logObject("transsec", transsecResult);
        }

        // load in the neural net if we need to
        if ( localNNet == null ) {
            if ( useStoredNN && storedNN != null) {
                // we have a stored NN, use it
                localNNet = storedNN;
            }
            else
            {
                // we either have no NN, or don't want to use it
                Config config = Config.getConfig();
                String netModel = config.getStringValue("finalModel");
                localNNet = restoreNeuralNet(netModel);
                storedNN = localNNet;
            }
        }

        ScoreAnnotatedSequence result = null;

        if (localNNet != null)
        {
            /* We get the first layer of the net (the input layer),
              then remove all the input synapses attached to it and attach a DirectSynapse */
            Layer input = localNNet.getInputLayer();
            input.removeAllInputs();
            DirectSynapse memInp = new DirectSynapse();
            input.addInputSynapse(memInp);

            /* We get the last layer of the net (the output layer),
             then remove all the output synapses attached to it and attach a DirectSynapse */
            Layer output = localNNet.getOutputLayer();
            output.removeAllOutputs();
            DirectSynapse memOut = new DirectSynapse();
            output.addOutputSynapse(memOut);

            // begin the neural net's thread
            localNNet.getMonitor().setLearning(false);
            localNNet.getMonitor().setSingleThreadMode(true);
            localNNet.start();

            // Create the set that contains the instances of this sequence
            UnifiedTrainingSet set = new UnifiedTrainingSet();
            set.addInstance(sequence, jnetResult, psipredResult, proteusSecResult);

            Symbol[] symbolList = new Symbol[set.numInstances()];

            try
            {
                result = new ScoreAnnotatedSequence(sequence);

                int index = 1;
                for (Iterator it = set.getInputInstanceIterator(); it.hasNext();)
                {

                    // Prepare the next input pattern
                    Pattern iPattern = new Pattern((double[]) it.next());
                    iPattern.setCount(1);

                    // Interrogate the net
                    memInp.fwdPut(iPattern);

                    // Read the output pattern and print out it
                    Pattern pattern = memOut.fwdGet();


                    if (verbose) System.out.print("Output #" + (index) + " = ");
                    double max = -1;
                    int maxIndex = -1;
                    final double[] outputArr = pattern.getArray();
                    NumberFormat formatter = DecimalFormat.getInstance();
                    formatter.setMaximumFractionDigits(3);

                    for (int j = 0; j < outputArr.length; j++)
                    {
                        if (outputArr[j] > max)
                        {
                            max = outputArr[j];
                            maxIndex = j;
                        }

                        if (verbose) System.out.print(" " + formatter.format(outputArr[j]));
                    }

                    // set the score of the chosen output
                    result.setScore(index, outputArr[ProteusTools.COIL], outputArr[ProteusTools.HELIX], outputArr[ProteusTools.STRAND]);

                    // get the symbol for this prediction
                    final Symbol symbol = ProteusTools.convertIntToStructure(maxIndex);
                    symbolList[index - 1] = symbol;

                    if (verbose) System.out.println(" Structure = " + symbol.getName());

                    index++;
                }

                SimpleSymbolList structure = new SimpleSymbolList(symbolList, symbolList.length, new StructureTools().getStructure());

                result.setStructureAnnotation(structure);

                /* Stop training - don't know why this is necessary */
                Pattern stop = new Pattern(new double[set.getInputInstanceSize()]);
                stop.setCount(-1);
                memInp.fwdPut(stop);
                memOut.fwdGet();
            }
            catch (IllegalSymbolException e)
            {
                e.printStackTrace();
            }

            if (verbose) System.out.println("Finished");
        }
        else
        {
            throw new IllegalStateException("Cannot read in neural net!");
        }

        if (doFilter)
        {
            // filter the result with the attached filters
            for (PredictionFilter filter : filters)
                result = filter.filter(result);

            // default filtering
            if ( filters.size() == 0 )
            {
                result = new SimpleFilter().filter(result);
            }
        }


        if ( !useStoredNN ) {
            // Clean up our memory
            storedNN = null;
            localNNet = null;
        }

        return result;
    }

    /**
     * Returns a string that represents a unique ID for a file, based on the current time in milliseconds
     * and thread ID, so you can have many threads concurrently creating the same type of file with unique ID's
     * Note - you should append a suffix to this file name if it is a file name you're using
     *
     * @return a string that represents a unique ID for a file
     */
    protected String getUniqueIdentifier()
    {
        return String.valueOf(System.currentTimeMillis()) + String.valueOf(Thread.currentThread().getId());
    }

    public void cicleTerminated(NeuralNetEvent e)
    {
        // call inherited method, it will do validation on a set of sequences
        super.cicleTerminated(e);

        Monitor mon = (Monitor) e.getSource();
//        long c = mon.getCurrentCicle();
//        long c1 = mon.getTotCicles() - c;

        System.out.println("Training Error = " + mon.getGlobalError());

        SeqSeeDB db = SeqSeeDB.getInstance();

        if (validationSequences == null) {
            //validationSequences = db.getEvaluationSet(numToValidate , 2000, 3000);
            validationSequences = db.getEvaluationSet2(numToValidate , 1000, 6000, 77);
            //validationSequences.addAll( db.getEvaluationSet(numToValidate /2 , 5000, 6000) );
        }

        Score score = new Score();
        Score filteredScore = new Score();

        NeuralNet localNN = nnet.cloneNet();
        localNN.getMonitor().setExporting(false);
        localNN.getMonitor().setSingleThreadMode(true);
        localNN.removeAllListeners();

        NumberFormat format = NumberFormat.getNumberInstance();
        format.setMaximumFractionDigits(4);

        int i = 0;
        for (Iterator it = validationSequences.iterator(); it.hasNext();)
        {
            final StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();

            StructureAnnotatedSequence prediction;
            prediction = predict(sequence, localNN, false);

            double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
            double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
            if (i < 5)
                System.out.println("Without filter: "
                                   + " Q3: " + format.format(q3)
                                   + " SOV: " + format.format(sov)
                                   + prediction);
            score.q3 += q3;
            score.sov += sov;

            prediction = predict(sequence, localNN, true);
            if (i++ < 5)
                System.out.println("Filtered: "
                                   + " Q3: " + format.format(q3)
                                   + " SOV: " + format.format(sov)
                                   + prediction);
            q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
            sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

            filteredScore.q3 += q3;
            filteredScore.sov += sov;
        }

        System.out.print("Avg Q3: " + score.q3 / validationSequences.size());
        System.out.println("\t SOV: " + score.sov / validationSequences.size());

        System.out.print("Filtered Q3: " + filteredScore.q3 / validationSequences.size());
        System.out.println("\t SOV: " + filteredScore.sov / validationSequences.size());

        if (score.q3 > bestQ3ScoreSoFar)
        {
            // StringFormatter format = new Formatter();
            // format.
            System.out.println("Best Q3 score so far! (" + score.q3 / validationSequences.size() + ")");
            bestQ3ScoreSoFar = score.q3;

            final Date date = new Date();
            saveNeuralNet("bestFinalNN.model" + date.getDate() + "-" + ( date.getMonth() +1));
        }

        if (filteredScore.q3 > bestFilteredQ3ScoreSoFar)
        {
            // StringFormatter format = new Formatter();
            // format.
            System.out.println("Best Filtered Q3 score so far! (" + filteredScore.q3 / validationSequences.size() + ")");
            bestFilteredQ3ScoreSoFar = filteredScore.q3;

            final Date date = new Date();
            saveNeuralNet("bestFinalFilteredNN.model" + date.getDate() + "-" + (date.getMonth() + 1));
        }
    }

    public void setLogger(RuntimeCatalog logger)
    {
        this.logger = logger;
    }

    /**
     * Saves intermediate blast files in a directory in case we need to use them again
     * @param sequence
     * @param outputFileMap
     */
    private void saveBlast(Sequence sequence, Map<BlastRunner.OUTPUT_FILE_TYPE, String> outputFileMap)
    {
        String seqName = sequence.getName();

        File dir = new File(BLAST_DIR);
        if ( !dir.exists())
            dir.mkdir();

        try
        {
            String oldFile = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.PSSM);
            File f = new File(oldFile);
            String pathname = oldFile.replaceFirst(".*(\\..*)", BLAST_DIR + "/" + seqName + "$1");
            FileUtils.copyFile(f, new File( pathname ));

            oldFile = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.FREQUENCY);
            if ( oldFile != null )
            {
                f = new File(oldFile);
                pathname = oldFile.replaceFirst(".*(\\..*)", BLAST_DIR + "/" + seqName + "$1");
                FileUtils.copyFile(f, new File( pathname ));
            }

            oldFile = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.OUTPUT);
            f = new File(oldFile);
            pathname = oldFile.replaceFirst(".*(\\..*)", BLAST_DIR + "/" + seqName + "$1");
            FileUtils.copyFile(f, new File( pathname ));

            oldFile = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.CHECKPOINT);
            f = new File(oldFile);
            pathname = oldFile.replaceFirst(".*(\\..*)", BLAST_DIR + "/" + seqName + "$1");
            FileUtils.copyFile(f, new File( pathname ));
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    /**
     * Finds intermediate blast files when we want them again
     * @param seqName
     * @param outputFileMap - not null
     * @param blastSequences - not null
     * @return true if it finds a set of blast files
     */
    private boolean findBlast(String seqName, Map<BlastRunner.OUTPUT_FILE_TYPE, String> outputFileMap, List<Sequence> blastSequences)
    {
        File dir = new File(BLAST_DIR);
        if ( !dir.exists() )
            return false;

        File[] files = dir.listFiles();
        for (File f: files)
        {
            // lose everything after "."
            String fname = f.getName().replaceFirst("\\..*", "");

            if ( fname.equals(seqName) || fname.regionMatches(true, 0, seqName, 0 , 4))
            {
                try
                {
                    String uniqueIdentifier = getUniqueIdentifier();
                    File pssm = new File(BLAST_DIR + "/" + fname + BlastRunner.PSSM_SUFFIX);
                    if ( pssm.exists() )
                    {
                        String newFile = uniqueIdentifier + BlastRunner.PSSM_SUFFIX;
                        FileUtils.copyFile(pssm, new File(newFile));
                        outputFileMap.put(BlastRunner.OUTPUT_FILE_TYPE.PSSM, newFile );
                    }
                    //else
                    //    return false;

                    File chk = new File(BLAST_DIR + "/" + fname + BlastRunner.CHKPOINT_SUFFIX);
                    if ( chk.exists() )
                    {
                        String newFile = uniqueIdentifier + BlastRunner.CHKPOINT_SUFFIX;
                        FileUtils.copyFile(chk, new File(newFile));
                        outputFileMap.put(BlastRunner.OUTPUT_FILE_TYPE.CHECKPOINT, newFile );
                    }
                    //else
                    //    return false;

                    File out = new File(BLAST_DIR + "/" + fname + BlastRunner.OUTPUT_SUFFIX);
                    if ( out.exists() )
                    {
                        String newFile = uniqueIdentifier + BlastRunner.OUTPUT_SUFFIX;
                        FileUtils.copyFile(out, new File(newFile));
                        outputFileMap.put(BlastRunner.OUTPUT_FILE_TYPE.OUTPUT, newFile );

                        Reader fileReader = new FileReader(out);
                        BlastParser parser = new BlastParser();

                        XMLReader oParser = new BlastXMLParserFacade();

                        parser.parseBlastResult(oParser, fileReader, blastSequences);

                        if ( method == MethodSettings.ALL || method == MethodSettings.PSIPRED || method == MethodSettings.PSIPRED_JNET )
                        {
                            if ( pssm.exists() && chk.exists() )
                            {
                                // We have to test Makemat too, sometimes it doesn't work

                                String chkptFile = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.CHECKPOINT);
                                String blastfileName = outputFileMap.get(BlastRunner.OUTPUT_FILE_TYPE.OUTPUT);
                                fname = chkptFile.substring(0, chkptFile.indexOf('.'));
                                fname = fname.replaceFirst(".*/(.*$)", "$1");

                                PsipredRunner psipredRunner = new PsipredRunner();
                                psipredRunner.setParams(pssm.getAbsolutePath(), chk.getAbsolutePath(), blastfileName);

                                if (!psipredRunner.runMakeMat(fname, chk) )
                                    continue;

                                return true;
                            }
                        }

                        return true;
                    }
                    //else
                    //    return false;
                }
                catch (FileNotFoundException e)
                {
                    e.printStackTrace();
                }
                catch (IllegalSymbolException e)
                {
                    e.printStackTrace();
                }
                catch (BioException e)
                {
                    e.printStackTrace();
                }
                catch (IOException e)
                {
                    e.printStackTrace();
                }
                catch (ChangeVetoException e)
                {
                    e.printStackTrace();
                }
                catch (SAXException e)
                {
                    e.printStackTrace();
                }
            }
        }
        return false;
    }

    public void setEvaluating(boolean b)
    {
        evaluating = b;
    }

}
