package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.util.Score;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.util.SimpleFilter;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.joone.engine.*;
import org.joone.engine.learning.TeachingSynapse;
import org.joone.io.MemoryInputSynapse;
import org.joone.net.NeuralNet;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Mar 19, 2005
 * Time: 3:51:51 PM
 * To change this template use File | Settings | File Templates.
 */
public class NeuralFilter extends Predictor implements PredictionFilter
{


    //private final static int WINDOW_SIZE = 15;
    //private final static int INPUT_LAYER = ProteusTools.NUM_STRUCTURES * WINDOW_SIZE;
    private final static int LAYER2 = 60;
    //private final static int OUTPUT_LAYER = ProteusTools.NUM_STRUCTURES;

    private static int numToTrainOn = 100;
    private static int numToPredict = 1;
    private static boolean verbose = false;
    private static String outputFile = "filterNN.model";
    public boolean useStoredNN = false;    // this should only be true when using as a batch predictor
    private static NeuralNet storedNN = null;
    private static List validationSequences = null;
    static final private int numToValidate = 100;
    private double bestQ3ScoreSoFar = 0;
    private static SingleSequencePredictor samplePredictor;
    private static Logger logger = Logger.getLogger("transsec.NeuralFilter");

    public ScoreAnnotatedSequence filter(ScoreAnnotatedSequence unfilteredPrediction)
    {
        return predict(unfilteredPrediction);
    }

    public void train(List predictedStructures, List realStructures)
    {
        assert(predictedStructures.size() == realStructures.size());

        NeuralFilterTrainingSet trainingSet = new NeuralFilterTrainingSet();
        for (int i = 0; i < predictedStructures.size(); i++)
        {
            final ScoreAnnotatedSequence predictedStructure = (ScoreAnnotatedSequence) predictedStructures.get(i);
            trainingSet.addInstance(predictedStructure, predictedStructure.getScores(), (SymbolList) realStructures.get(i));
        }

        /* The Layers */
        LinearLayer input = new LinearLayer();
        input.setRows( trainingSet.getNumInputs() );

        SigmoidLayer hidden = new SigmoidLayer();
        hidden.setRows(LAYER2);

        SigmoidLayer output = new SigmoidLayer();
        output.setRows( trainingSet.getNumOutputs() );

        /* The Synapses */
        FullSynapse synapse_IH = new FullSynapse(); /* Input -> Hidden conn. */
        input.addOutputSynapse(synapse_IH);
        hidden.addInputSynapse(synapse_IH);

        FullSynapse synapse_HO = new FullSynapse(); /* Hidden -> Output conn. */
        hidden.addOutputSynapse(synapse_HO);
        output.addInputSynapse(synapse_HO);

        int numInstances = trainingSet.numInstances();

        /* The I/O components */

        MemoryInputSynapse inputStream = new MemoryInputSynapse();
        inputStream.setInputArray(trainingSet.getInputArray());
        inputStream.setAdvancedColumnSelector("1-" + trainingSet.getNumInputs() );
        // use first split for training
        //inputStream.setFirstRow(0);
        //inputStream.setFirstRow( (int) Math.round( Math.floor(numInstances * percentageSplit) ) );
        input.addInputSynapse(inputStream);


        /* The Trainer and its desired file */
        TeachingSynapse trainer = new TeachingSynapse();
        MemoryInputSynapse samples = new MemoryInputSynapse();
        samples.setInputArray(trainingSet.getInputArray());
        trainer.setDesired(samples);

        samples.setAdvancedColumnSelector( String.valueOf(trainingSet.getNumInputs() +1) + "-" + String.valueOf(trainingSet.getNumInputs() + trainingSet.getNumOutputs()) );
        output.addOutputSynapse(trainer);

        nnet = new NeuralNet();
        nnet.addLayer(input, NeuralNet.INPUT_LAYER);
        nnet.addLayer(hidden, NeuralNet.HIDDEN_LAYER);
        nnet.addLayer(output, NeuralNet.OUTPUT_LAYER);
        nnet.setTeacher(trainer);

        Monitor monitor = nnet.getMonitor();
        monitor.setLearningRate(0.8);
        monitor.setMomentum(0.3);
        monitor.setTrainingPatterns(numInstances); /* # of rows in the input file */
        monitor.setTotCicles(200); /* How many times the net must be trained*/
        monitor.setLearning(true); /* The net must be trained */
        monitor.addNeuralNetListener(this);

        nnet.start();
        nnet.getMonitor().Go();

        trainingSet = null;
    }

    public void train(List sequences)
    {
        throw new IllegalStateException("Invalid");
    }

    public ScoreAnnotatedSequence predict(Sequence sequence, NeuralNet localNN, boolean filter)
    {
        return predict( (ScoreAnnotatedSequence) sequence, localNN, filter );
    }

    public ScoreAnnotatedSequence predict(ScoreAnnotatedSequence sequence, NeuralNet localNNet, boolean filter)
    {
        ScoreAnnotatedSequence result = null;
        try
        {
            result = new ScoreAnnotatedSequence(sequence);

            if (localNNet == null) {
                if ( useStoredNN && storedNN != null) {
                    // we have a stored NN, use it
                    localNNet = storedNN;
                }
                else
                {
                    // we either have no NN, or don't want to use it
                    localNNet = restoreNeuralNet(outputFile);
                    storedNN = localNNet;
                }
            }

            if (localNNet != null)
            {
                /* We get the first layer of the net (the input layer),
                  then remove all the input synapses attached to it and attach a DirectSynapse */
                Layer input = localNNet.getInputLayer();
                input.removeAllInputs();
                DirectSynapse memInp = new DirectSynapse();
                input.addInputSynapse(memInp);

                /* We get the last layer of the net (the output layer),
                 then remove all the output synapses attached to it and attach a DirectSynapse */
                Layer output = localNNet.getOutputLayer();
                output.removeAllOutputs();
                DirectSynapse memOut = new DirectSynapse();
                output.addOutputSynapse(memOut);

                // begin the neural net's thread
                localNNet.getMonitor().setLearning(false);
                localNNet.start();

                // Create the set that contains the instances of this sequence
                NeuralFilterTrainingSet set = new NeuralFilterTrainingSet();
                set.addInstance(sequence.getStructureAnnotation(), sequence.getScores() );

                SimpleSymbolList structure = new SimpleSymbolList(new StructureTools().getStructure());

                int index = 1;
                for (Iterator it = set.getInputInstanceIterator(); it.hasNext();)
                {

                    // Prepare the next input pattern
                    Pattern iPattern = new Pattern((double[]) it.next());
                    iPattern.setCount(1);

                    // Interrogate the net
                    memInp.fwdPut(iPattern);

                    // Read the output pattern and print out it
                    Pattern pattern = memOut.fwdGet();


                    if (verbose) System.out.print("Output #" + (index) + " = ");
                    double max = -1;
                    int maxIndex = -1;
                    final double[] outputArr = pattern.getArray();
                    NumberFormat formatter = DecimalFormat.getInstance();
                    formatter.setMaximumFractionDigits(3);

                    for (int j = 0; j < outputArr.length; j++)
                    {
                        if (outputArr[j] > max)
                        {
                            max = outputArr[j];
                            maxIndex = j;
                        }

                        if (verbose) System.out.print(" " + formatter.format(outputArr[j]));
                    }

                    // set the score for this residue prediction
                    result.setScore(index, outputArr[ProteusTools.COIL], outputArr[ProteusTools.HELIX], outputArr[ProteusTools.STRAND]);

                    // get the symbol for this prediction
                    Symbol symbol = ProteusTools.convertIntToStructure(maxIndex);
                    structure.addSymbol(symbol);

                    if (verbose) System.out.println(" Structure = " + symbol.getName());

                    index++;
                }

                result.setStructureAnnotation(structure);

                /* Stop training - don't know why this is necessary */
                Pattern stop = new Pattern(new double[set.getInputInstanceSize()]);
                stop.setCount(-1);
                memInp.fwdPut(stop);
                memOut.fwdGet();

                if (verbose) System.out.println("Finished");
            }
            else
            {
                throw new IllegalStateException("Cannot read in neural net!");
            }
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();  //TODO
        }


        return result;
    }

    public static void main(String[] args) throws IllegalSymbolException, ChangeVetoException
    {
        SeqSeeDB db = SeqSeeDB.getInstance();

        boolean train = false;
        boolean predict = false;

        for (int i = 0; i < args.length; i++)
        {
            if (args[i].equals("-train"))
            {
                train = true;
                if (i < args.length - 1)
                {
                    try
                    {
                        numToTrainOn = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e)
                    {
                        printUsage();
                    }
                }
                else
                {
                    printUsage();
                }
            }
            else if (args[i].equals("-predict"))
            {
                predict = true;
                if (i < args.length - 1)
                    try
                    {
                        numToPredict = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e)
                    {
                        printUsage();
                    }
            }
            else if (args[i].equals("-verbose") || args[i].equals("-v"))
            {
                verbose = true;
            }
            else if (args[i].equals("-output") || args[i].equals("-o"))
            {
                if (i < args.length - 1)
                {
                    String arg = args[++i];
                    if (arg.charAt(0) == '-')
                        printUsage();
                    else
                        outputFile = arg;
                }
            }
        }

        if (train)
        {
            // train on this set of sequences
            NeuralFilter neuralFilter = new NeuralFilter();

            //List randomSequences = db.getTrainingSet(numToTrainOn);
            //validationSequences = db.getEvaluationSet(numToValidate);
            List randomSequences = db.getTrainingSet(40);
            validationSequences = randomSequences;

            List predictedStructures = new ArrayList();
            List realStructures = new ArrayList();

            samplePredictor = new SingleSequencePredictor();
            samplePredictor.useStoredNN = true;

            for (Iterator it = randomSequences.iterator(); it.hasNext();)
            {
                StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();

                realStructures.add(sequence.getStructureAnnotation());

                StructureAnnotatedSequence predictedSequence = samplePredictor.predict(sequence);
                predictedStructures.add(predictedSequence);

                System.out.println(predictedSequence);
                System.out.println("Real      = " + sequence.getStructureAnnotation().seqString().toUpperCase());
            }

            neuralFilter.train(predictedStructures, realStructures);
        }
        else if (predict)
        {
            // predict this set of sequences
            SingleSequencePredictor predictor = new SingleSequencePredictor();
            predictor.useStoredNN = true;

            final NeuralFilter filter = new NeuralFilter();
            filter.useStoredNN = true;
            predictor.addFilter(filter);
            predictor.addFilter(new SimpleFilter());

            List randomSequences = db.getRandomSequences(numToPredict, 0, 100);

            StructureAnnotatedSequence prediction;
            Score score = new Score();
            Score filteredScore = new Score();

            for (Iterator it = randomSequences.iterator(); it.hasNext();)
            {
                final StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();
                prediction = predictor.predict(sequence, null, false);

                System.out.println("Sequence: \n" + sequence);

                double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                score.q3 += q3;
                score.sov += sov;

                System.out.println("Prediction: "
                                   + " Q3: " + q3
                                   + " SOV: " + sov
                                   + "\n" + prediction);

                // Now do a filtered prediction
                prediction = predictor.predict(sequence);

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                filteredScore.q3 += q3;
                filteredScore.sov += sov;

                System.out.println("\nPrediction ** filtered ** : "
                                   + " Q3: " + q3
                                   + " SOV: " + sov
                                   + "\n" + prediction);
            }

            System.out.print("Avg Q3: " + score.q3 / randomSequences.size());
            System.out.println("SOV: " + score.sov / randomSequences.size());

            System.out.print("Filtered Avg Q3: " + filteredScore.q3 / randomSequences.size());
            System.out.println("Filtered SOV: " + filteredScore.sov / randomSequences.size());
        }
        else
            printUsage();
    }

    public void addFilter(PredictionFilter filter)
    {
        throw new IllegalStateException("Can't add a filter to a NeuralFilter!");
    }

    private static void printUsage()
    {
        System.out.println("Usage: <ProgName> (-train <# seq> | -predict <#seq>) [options]");
        System.out.println("Options: ");
        System.out.println("-train <# sequences>\t\tTrain on a number of sequences");
        System.out.println("-predict <# sequences>\t\tPredict a number of sequences");
        System.out.println("-v, -verbose");
        System.out.println("-o, -output <file>\t\t\tFile to output the training model to (if applicable)");
        System.exit(-1);
    }

    public void cicleTerminated(NeuralNetEvent e)
    {
        // call inherited method, it will do validation on a set of sequences
        super.cicleTerminated(e);

        Monitor mon = (Monitor) e.getSource();
        long c = mon.getCurrentCicle();
        long c1 = mon.getTotCicles() - c;
        if ((c1 < 5)  // for first 25, every 5
                || (c1 < 25 && c1 % 5 == 0)  // for first 25, every 5
                || (c1 < 200 && c1 % 10 == 0)  // up to 200, every 10
                || ((c1 % 200) == 0)) // after 200, every 200
        {
            System.out.println(c + " cycles remaining - Training Error = " + mon.getGlobalError() );
        }

        SeqSeeDB db = SeqSeeDB.getInstance();

        if (validationSequences == null)
            validationSequences = db.getEvaluationSet(numToValidate);

        Score score = new Score();
        Score filteredScore = new Score();

        NeuralNet localNN = nnet.cloneNet();
        localNN.getMonitor().setExporting(false);
        localNN.removeAllListeners();

        for (Iterator it = validationSequences.iterator(); it.hasNext();)
        {
            final StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();

            StructureAnnotatedSequence prediction;
            // predict original prediction
            prediction = samplePredictor.predict(sequence, null, false);

            score.q3 += ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
            score.sov += ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

            // now do it with a filter

            prediction = this.predict(prediction, localNN);

            filteredScore.q3 += ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
            filteredScore.sov += ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
        }

        System.out.print("Avg Q3: " + score.q3 / validationSequences.size());
        System.out.println("\t SOV: " + score.sov / validationSequences.size());

        System.out.print("Filtered Q3: " + filteredScore.q3 / validationSequences.size());
        System.out.println("\t SOV: " + filteredScore.sov / validationSequences.size());

        if ( filteredScore.q3 > bestQ3ScoreSoFar ) {
           // StringFormatter format = new Formatter();
           // format.
            System.out.println("Best Q3 score so far! (" + filteredScore.q3 / validationSequences.size() + ")");
            bestQ3ScoreSoFar = filteredScore.q3;

            final Date date = new Date();
            saveNeuralNet("bestNN.model" + date.getDate() + "-" + date.getMonth());
        }

        if ((c1 % 25) == 0)
            saveNeuralNet(outputFile);
    }
}