package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.util.SimpleFilter;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.symbol.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.apache.log4j.Logger;

import java.text.NumberFormat;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Apr 9, 2005
 * Time: 8:43:24 PM
 * To change this template use File | Settings | File Templates.
 */

/**
 * A class to represent a position-specific scoring matrix
 * In this case, it is specialized to represent a structure matrix
 * TODO - possibly redo this class as a generalized PSSM, and subclass it to deal with specialized alphabets
 */
public class PSSM implements PSSMInterface
{
    // data storage to represent the PSSM
    private double[][] matrix;
    private Alphabet alphabet;
    List<Sequence> sequences;
    HashMap<String, GappedStructureAnnotatedSequence> alignedSequences;

    int PSSMlength;
    int numStates;
    boolean logOdds;
    private Sequence querySequence;
    private GappedStructureAnnotatedSequence gappedQuerySequence;
    private boolean verbose = false;
    private Map<String, Map<String, Double>> ppshMap;

    private Logger logger = Logger.getLogger(PSSM.class);

    public PSSM(Alphabet alphabet, int PSSMlength)
    {
        this(alphabet, PSSMlength, false);
    }

    public PSSM(Alphabet alphabet, int PSSMlength, boolean logOdds)
    {
        this.alphabet = alphabet;
        this.logOdds = logOdds;
        this.sequences = new ArrayList<Sequence>();
        this.alignedSequences = new HashMap<String, GappedStructureAnnotatedSequence>();

        // TODO generify this
        this.numStates = ProteusTools.NUM_STRUCTURES;

        this.PSSMlength = PSSMlength;
    }

    /**
     * Builds a PSSM directly from a sequence and a given alignment
     *
     * @param sequence
     * @param alignment
     */
    public PSSM(Sequence sequence, Alignment alignment, Predictor predictor )
    {
        this(sequence.getAlphabet(), sequence.length());
        this.querySequence = sequence;
        logger.debug("Alignments has " + alignment.getLabels().size() + " sequences");

        for (Iterator it = alignment.symbolListIterator(); it.hasNext();)
        {
            Sequence seq = (Sequence) it.next();
            Sequence unGappedSequence = GappedStructureAnnotatedSequence.removeGaps(seq);
            sequences.add(predictor.predict(unGappedSequence));

            GappedStructureAnnotatedSequence gsym = new GappedStructureAnnotatedSequence(seq);

            if (alignedSequences.containsKey(gsym.getName()))
                throw new IllegalStateException("Aligned sequences already contains this key");

            alignedSequences.put(gsym.getName(), gsym);

            if (verbose)
            {
                System.out.println(gsym.getName() + "\t" + gsym.seqString());
            }

            if ( gsym.getName().equals(querySequence.getName() ) || gsym.getName().equals("QUERY" )) {
                gappedQuerySequence = new GappedStructureAnnotatedSequence(gsym);
                // Map the known structure onto the gapped sequence
                if ( querySequence instanceof StructureAnnotatedSequence )
                    gappedQuerySequence.createGappedStructure(((StructureAnnotatedSequence) querySequence).getStructureAnnotation());
            }
        }

        // modify the structure stored in the aligned sequence with the gapped structure
        for (Sequence s : sequences)
        {
            StructureAnnotatedSequence seq = (StructureAnnotatedSequence) s;
            try
            {
                if (alignedSequences.get(seq.getName()) != null)
                    alignedSequences.get(seq.getName()).createGappedStructure(seq);
                else
                {
                    System.err.println("Aligned sequence " + seq.getName() + " is not in alignedSequences list!");
                }
            }
            catch (IllegalStateException e)
            {
                e.printStackTrace(System.err);
            }
        }

        // now recalculate the matrix
        recalculateMatrix();

        if (verbose)
        {
            for (GappedStructureAnnotatedSequence gappedAlignment : alignedSequences.values())
            {
                GappedSymbolList structureAnnotation = gappedAlignment.getStructureAnnotation();
                if (structureAnnotation != null)
                    System.out.println(gappedAlignment.getName() + "\t" + structureAnnotation.seqString());
                else
                    System.out.println(gappedAlignment.getName() + "\t" + "No structure available");
            }

            System.out.println(this);
        }
    }

    public String toString()
    {
        StringBuilder rv = new StringBuilder();

        // write query sequence
        String query = getGappedQuerySequence().seqString();
        query = query.replaceAll("(.)", "$1    ");
        rv.append("      " + query + "\n");

        for (int i = 0; i < matrix.length; i++)
        {
            double[] row = matrix[i];
            switch (i)
            {
                case ProteusTools.GAP:
                    rv.append("GAP   ");
                    break;
                case ProteusTools.HELIX:
                    rv.append("HELIX ");
                    break;
                case ProteusTools.COIL:
                    rv.append("COIL  ");
                    break;
                case ProteusTools.STRAND:
                    rv.append("STRND ");
                    break;
            }

            for (int j = 0; j < row.length; j++)
                rv.append(String.format("%1.2f ", row[j]));
            rv.append("\n");
        }

        /*if ( querySequence instanceof StructureAnnotatedSequence) {
            query = ((StructureAnnotatedSequence) querySequence).getStructureAnnotation().seqString();
            query = query.replaceAll("(.)", "$1    ");
            rv.append("      " + query + "\n");
        } */

        if (getGappedQuerySequence() instanceof GappedStructureAnnotatedSequence)
        {
            if ( getGappedQuerySequence().getStructureAnnotation() != null ) {
                query = getGappedQuerySequence().getStructureAnnotation().seqString();
                query = query.replaceAll("(.)", "$1    ");
                rv.append("      " + query + "\n");
            }
        }

        return rv.toString();
    }

    public GappedStructureAnnotatedSequence getGappedQuerySequence()
    {
        if (gappedQuerySequence == null ) {

            GappedStructureAnnotatedSequence gappedSeq = alignedSequences.get(querySequence.getName());

            if ( querySequence instanceof StructureAnnotatedSequence ) {
                gappedQuerySequence.createGappedStructure(((StructureAnnotatedSequence) querySequence).getStructureAnnotation());
                System.out.println("Warning!  Having to rebuild gappedQuerySequence - you should probably do this before");
            }
            else {
                System.out.println("Warning!!!!!! Using the predicted query sequence structure");
                new Exception().printStackTrace();
            }

            if (gappedSeq == null)
                throw new IllegalStateException("Gapped query sequence is null - the alignment passed likely doesn't include the query sequence.");
            return gappedSeq;
        }

        return gappedQuerySequence;
    }

    /**
     * Call this to update the PSSM with the specified Alignment.
     * Note: All sequences added to the PSSM prior to this should be in the passed alignment
     * @param alignment
     * @deprecated
     */
    public void update(Alignment alignment)
    {
        throw new IllegalStateException("Not implemented don't call this");
        // Do alignment of all sequences

        // copy the alignment sequences into a map to store them for later
        /* int i = 0;
        for (Iterator it = alignment.symbolListIterator(); it.hasNext();)
        {
            Sequence sequence = (Sequence) it.next();
            GappedStructureAnnotatedSequence gsym = new GappedStructureAnnotatedSequence(sequence);
            if (alignedSequences.containsKey(gsym.getName()))
                throw new IllegalStateException("Aligned sequences already contains this key: " + gsym.getName());

            alignedSequences.put(gsym.getName(), gsym);

            if (verbose)
                System.out.println(gsym.getName() + "\t" + gsym.seqString());
        }

        // modify the structure stored in the aligned sequence with the gapped structure
        for (Iterator it = sequences.iterator(); it.hasNext();)
        {
            StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();
            try
            {
                if (alignedSequences.get(sequence.getName()) != null)
                    alignedSequences.get(sequence.getName()).createGappedStructure(sequence);
                else
                {
                    System.err.println("Aligned sequence " + sequence.getName() + " is not in alignedSequences list!");
                }
            }
            catch (IllegalStateException e)
            {
                e.printStackTrace(System.err);
                //System.err.println("Removing the offending sequence");
                //alignedSequences.remove(sequence.getName());
            }
        }

        recalculateMatrix();

        if (verbose)
        {
            for (Iterator it = alignedSequences.values().iterator(); it.hasNext();)
            {
                GappedStructureAnnotatedSequence gappedAlignment = (GappedStructureAnnotatedSequence) it.next();

                GappedSymbolList structureAnnotation = gappedAlignment.getStructureAnnotation();
                if (structureAnnotation != null)
                    System.out.println(gappedAlignment.getName() + "\t" + structureAnnotation.seqString());
                else
                    System.out.println(gappedAlignment.getName() + "\t" + "No structure available");
            }

            System.out.println(this);
        }
           */
    }

    private void recalculateMatrix()
    {
        // now that we know the alignment of the query sequence, build a PSSM with that length
        this.PSSMlength = getGappedQuerySequence().length();
        this.matrix = new double[this.numStates][];
        for (int k = 0; k < matrix.length; k++)
            this.matrix[k] = new double[this.PSSMlength];

        /* alignedSequences now contains all the GappedStructuredSequences with their predictions
          Now to compile them into the PSSM, by looking at each symbol and incrementing
          it's matrix position.  If the sequence does not span the area (has a leading or trailing space),
          then do not include that in the matrix.
        */
        int[] counts = new int[PSSMlength];
        for (int resindex = 0; resindex < PSSMlength; resindex++)
        {
            for (Iterator it = alignedSequences.values().iterator(); it.hasNext();)
            {
                GappedStructureAnnotatedSequence gappedAlignment = (GappedStructureAnnotatedSequence) it.next();

                GappedSymbolList structureAnnotation = gappedAlignment.getStructureAnnotation();

                if (structureAnnotation != null)
                {
                    Symbol symbol = structureAnnotation.symbolAt(resindex + 1);

                    int stateindex = ProteusTools.convertStructureToInt(symbol);

                    if (gappedAlignment.getStructureAnnotation().firstNonGap() <= resindex + 1
                            && gappedAlignment.getStructureAnnotation().lastNonGap() >= resindex + 1)
                    {
                        matrix[stateindex][resindex]++;
                        counts[resindex]++;
                    }
                }
            }

            // now normalize the arrays
            for (int j = 0; j < matrix.length; j++)
                if (counts[resindex] != 0)
                    matrix[j][resindex] /= (double) counts[resindex];
        }
    }

    public void addSequence(StructureAnnotatedSequence sequence)
    {
        sequences.add(sequence);
    }

    public void addQuerySequence(Sequence sequence)
    {
        //sequence.setName("QUERY");

        sequences.add(sequence);
        querySequence = sequence;
    }

    public double get(int col, int row)
    {
        return matrix[row][col];
    }

    public int length()
    {
        return PSSMlength;
    }

    public void saveToXML(Document doc, Element element)
    {
        Element querySeqElement = doc.createElement("querySequence");
        querySeqElement.setTextContent(querySequence.seqString());
        element.appendChild(querySeqElement);

        Element gappedQuerySeqElement = doc.createElement("gappedQuerySequence");
        gappedQuerySeqElement.setTextContent(getGappedQuerySequence().seqString());
        element.appendChild(gappedQuerySeqElement);

        Element gappedQueryStrucElement = doc.createElement("gappedQueryStructure");
        gappedQueryStrucElement.setTextContent(getGappedQuerySequence().getStructureAnnotation().seqString());
        element.appendChild(gappedQueryStrucElement);

        Element data = doc.createElement("data");
        element.appendChild(data);
        data.setTextContent(this.toString());
    }

    public Sequence getQuerySequence()
    {
        return querySequence;
    }

    public static PSSM readFromXMLNode(Node element)
    {
        String querySequenceStr = "",
                gappedQuerySequenceStr = "",
                gappedQueryStructureStr = "",
                data = "",
                name = "";
        name = element.getAttributes().getNamedItem("name").getNodeValue();

        for (int i = 0; i < element.getChildNodes().getLength(); i++)
        {
            Node node = element.getChildNodes().item(i);
            if (node.getNodeName().equals("querySequence"))
            {
                querySequenceStr = node.getTextContent();
            }
            else if (node.getNodeName().equals("gappedQuerySequence"))
            {
                gappedQuerySequenceStr = node.getTextContent();
            }
            else if (node.getNodeName().equals("gappedQueryStructure"))
            {
                gappedQueryStructureStr = node.getTextContent();
            }
            else if (node.getNodeName().equals("data"))
            {
                data = node.getTextContent();
            }
        }

        PSSM pssm = new PSSM(ProteinTools.getAlphabet(), gappedQuerySequenceStr.length());
        try
        {
            pssm.querySequence = new SimpleSequence(ProteinTools.createProtein(querySequenceStr), "", name, new SimpleAnnotation());
            pssm.gappedQuerySequence = new GappedStructureAnnotatedSequence(new SimpleSequence(ProteinTools.createProtein(gappedQuerySequenceStr), "", name, new SimpleAnnotation()));
            pssm.gappedQuerySequence.setStructureAnnotation(new SimpleGappedSymbolList(ProteusTools.createStructure(gappedQueryStructureStr)));
            //pssm.alignedSequences.put(name, gappedStructureAnnotatedSequence);

            pssm.matrix = new double[pssm.numStates][pssm.PSSMlength];

            String[] lines = data.split("\n");
            for (int i = 1; i < lines.length - 1; i++)
            {
                String[] items = lines[i].split("\\s+");
                for (int j = 1; j < items.length; j++)
                {
                    pssm.matrix[i - 1][j - 1] = Double.parseDouble(items[j]);
                }
            }
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }

        return pssm;
    }

    public boolean equals(Object o)
    {
        if (this == o) return true;
        if (!(o instanceof PSSM)) return false;

        final PSSM pssm = (PSSM) o;

        if (querySequence != null ? !querySequence.seqString().equals(pssm.querySequence.seqString()) : pssm.querySequence != null) return false;
        Sequence gappedSequence = getGappedQuerySequence();
        Sequence gappedSequence2 = pssm.getGappedQuerySequence();
        if (gappedSequence != null ? !gappedSequence.seqString().equals(gappedSequence2.seqString()) : gappedSequence2 != null) return false;

        if (matrix == null && pssm.matrix != null) return false;
        if (pssm.matrix == null && matrix != null) return false;

        // compare the matrices
        for (int i = 0; i < matrix.length; i++)
            for (int j = 0; j < matrix[i].length; j++)
                if (Math.abs(matrix[i][j] - pssm.matrix[i][j]) > 0.02)
                    return false;

        return true;
    }

    public int hashCode()
    {
        return (querySequence != null ? querySequence.hashCode() : 0);
    }

    public void getResult()
    {
        Symbol[] symbolList = new Symbol[length()];
        PredictionScore[] scores = new PredictionScore[length()];
        ScoreAnnotatedSequence result = new ScoreAnnotatedSequence(gappedQuerySequence);

        for ( int i=0; i<length(); i++) {
            PredictionScore score = new PredictionScore();
            score.helix = get(i, ProteusTools.HELIX);
            score.strand = get(i, ProteusTools.STRAND);
            score.coil = get(i, ProteusTools.COIL);
            score.gap = get(i, ProteusTools.GAP);

            // get the symbol for this prediction
            Symbol symbol;
            double max = score.getMax();
            if ( max == score.helix )
                symbol = new StructureTools().getH();
            else if ( max == score.strand )
                symbol = new StructureTools().getE();
            else
                symbol = new StructureTools().getC();

            // if it's a gap, it's really a coil
            if (symbol.equals(new StructureTools().getStructure().getGapSymbol()))
                symbol = new StructureTools().getC();

            symbolList[i] = symbol;
            scores[i] = score;
        }

        SimpleSymbolList structure = new SimpleSymbolList(symbolList, symbolList.length, new StructureTools().getStructure());

        result.setStructureAnnotation(structure);
        result.setScores(scores);

        // ungap this prediction
        result = GappedStructureAnnotatedSequence.createUngappedStructure(querySequence, result);

        double q3 = ProteusTools.q3Score(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), result.getStructureAnnotation());
        double sov = ProteusTools.sovScore(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), result.getStructureAnnotation());

        NumberFormat format = NumberFormat.getNumberInstance();
                    format.setMaximumFractionDigits(4);

        System.out.println("PSSM Consensus: "
                                           + " Q3: " + format.format(q3)
                                           + " SOV: " + format.format(sov)
                                           + result);

        final PsiPass2Filter ps2Ffilter = new PsiPass2Filter();
        ScoreAnnotatedSequence filtered = ps2Ffilter.filter(result);

        q3 = ProteusTools.q3Score(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), filtered.getStructureAnnotation());
        sov = ProteusTools.sovScore(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), filtered.getStructureAnnotation());

        System.out.println("PSSM Consensus Filtered with PSIPASS2: "
                                           + " Q3: " + format.format(q3)
                                           + " SOV: " + format.format(sov)
                                           + filtered);

        final SimpleFilter filter = new SimpleFilter();
        filtered = filter.filter(result);

        q3 = ProteusTools.q3Score(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), filtered.getStructureAnnotation());
        sov = ProteusTools.sovScore(((StructureAnnotatedSequence) querySequence).getStructureAnnotation(), filtered.getStructureAnnotation());

        System.out.println("PSSM Consensus Filtered with Normal filter: "
                                           + " Q3: " + format.format(q3)
                                           + " SOV: " + format.format(sov)
                                           + filtered);

    }
}
