package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.util.Format;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.util.StreamHandler;

import java.io.*;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Mar 28, 2005
 * Time: 11:27:37 PM
 * To change this template use File | Settings | File Templates.
 *
 * Wraps the filtering program psipass2 from PSIPRED
 */
public class PsiPass2Filter implements PredictionFilter
{

    /**
     * Filters a ScoreAnnotatedSequence by passing it to the psipass2 program
     *
     * @param unfilteredPrediction - a raw structure outputted from a machine learning algorithm,
     * with associated scores
     * @return a filtered sequence
     */
    public ScoreAnnotatedSequence filter(ScoreAnnotatedSequence unfilteredPrediction)
    {
        try
        {
            final String fileName = String.valueOf(System.currentTimeMillis());
            PrintWriter writer = new PrintWriter(new FileWriter(fileName + ".ss" ));

            final PredictionScore[] scores = unfilteredPrediction.getScores();
            final char[] structure = unfilteredPrediction.getStructureAnnotation().seqString().toUpperCase().toCharArray();
            PredictionScore score;
            for ( int line=0; line< unfilteredPrediction.length(); line++) {
                writer.print( Format.printf("%4d ", line ) );
                writer.print( Format.printf( "%c ", 'A' ) );
                score = scores[line];

                writer.print(Format.printf("%c   ", structure[line]));

                writer.print(Format.printf("%1.3f  ", score.coil )); // C
                writer.print(Format.printf("%1.3f  ", score.helix )); // H
                writer.print(Format.printf("%1.3f \n", score.strand )); // E
            }
            writer.close();

            Config config = Config.getConfig();
            String datadir = config.getStringValue("psipred.datadir");
            String cmd = config.getStringValue("psipass.command");

            cmd += " " + datadir + "/weights_p2.dat 1 1.0 1.0 " + fileName+ ".ss2 " + fileName + ".ss";
            final Process process = Runtime.getRuntime().exec(cmd);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(process.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(process.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            // any error???
            int exitVal = process.waitFor();

            if ( exitVal != 0 ) {
                System.err.println("Failed filtering with psipass2");
                System.out.println(errorReader.getInputString());
                System.out.println(outputReader.getInputString());
                System.out.println("Cmd: " + cmd);
                cleanup(fileName);
                return unfilteredPrediction;
            }

            FileReader in = new FileReader(fileName + ".ss2");
            BufferedReader reader = new BufferedReader(in);
            String line;
            StringBuffer newStructure = new StringBuffer();
            reader.readLine(); reader.readLine(); // skip 2 lines
            int i=0;
            while( ( line = reader.readLine() ) != null ) {
                // the trim is important because after 1000, there is no space
                final String[] strings = line.trim().split("\\s+");
                newStructure.append( strings[2] );
                unfilteredPrediction.getScores()[i++] = new PredictionScore( Double.parseDouble(strings[3]), Double.parseDouble(strings[4]), Double.parseDouble(strings[5]));
            }

            unfilteredPrediction.setStructureAnnotation( ProteusTools.createStructure(newStructure.toString()));
            in.close();

            cleanup(fileName);
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }

        return unfilteredPrediction;
    }

    private void cleanup(final String fileName) {
        File f = new File(fileName + ".ss");
        f.delete();
        f = new File(fileName + ".ss2");
        f.delete();
    }
}
