package ca.ualberta.proteus.program.transsec.ReducedPredictor;

import ca.ualberta.proteus.program.transsec.NeuralFilter;
import ca.ualberta.proteus.program.transsec.PredictionFilter;
import ca.ualberta.proteus.program.transsec.Predictor;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.Symbol;
import org.biojava.utils.ChangeVetoException;
import org.joone.engine.*;
import org.joone.engine.learning.TeachingSynapse;
import org.joone.io.MemoryInputSynapse;
import org.joone.net.NeuralNet;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Iterator;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Mar 20, 2005
 * Time: 12:42:42 PM
 * To change this template use File | Settings | File Templates.
 */
public class SingleSequenceReducedPredictor extends Predictor
{
    private final static int LAYER2 = 60;
    private final static int LAYER3 = 10;

    private static int numToTrainOn = 100;
    private static int numToPredict = 1;
    private static boolean verbose = false;
    private static String outputFile = "reducedNN.model";
    public boolean useStoredNN = false; // this should only be true when using as a batch predictor
    private NeuralNet storedNN = null;
    private boolean _TESTING = false;

    public void train(List sequences)
    {
        TrainingSet trainingSet = new SingleSequenceReducedTrainingSet();
        for (Iterator it = sequences.iterator(); it.hasNext();)
        {
            trainingSet.addInstance((StructureAnnotatedSequence) it.next());
        }

        try
        {
            OutputStream file = new FileOutputStream("reducedTS.set");
            trainingSet.writeToStream(file);
        }
        catch (IOException ex)
        {
            System.out.println("Cannot perform output.");
        }

        if ( _TESTING ) {
            for (Iterator it = trainingSet.getInputInstanceIterator(); it.hasNext(); ) {
                double[] instance = (double[]) it.next();

                final int BINARY_RES = 5;
                for (int i=0; i< instance.length / BINARY_RES; i++) {
                    double value = 0;
                    if ( instance[i * BINARY_RES + 0] == 1 )
                        value += 0x10;
                    if ( instance[i * BINARY_RES+ 1] == 1 )
                        value += 0x8;
                    if ( instance[i * BINARY_RES+ 2] == 1 )
                        value += 0x4;
                    if ( instance[i * BINARY_RES+ 3] == 1 )
                        value += 0x2;
                    if ( instance[i * BINARY_RES+ 4] == 1 )
                        value += 0x1;
                    System.out.print( ProteusTools.convertToChar( ProteusTools.convertIntToSymbol( (int) value ) ) );
                }
                System.out.println();
            }
        }


        /* The Layers */
        LinearLayer input = new LinearLayer();
        input.setRows(trainingSet.getNumInputs());

        SigmoidLayer hidden1 = new SigmoidLayer();
        hidden1.setRows(LAYER2);

        SigmoidLayer hidden2 = new SigmoidLayer();
        hidden2.setRows(LAYER3);

        SigmoidLayer output = new SigmoidLayer();
        output.setRows( trainingSet.getNumOutputs() );

        /* The Synapses */
        FullSynapse synapse_IH1 = new FullSynapse(); /* Input -> Hidden1 conn. */
        input.addOutputSynapse(synapse_IH1);
        hidden1.addInputSynapse(synapse_IH1);

        FullSynapse synapse_H1_H2 = new FullSynapse(); /* Hidden1 -> Hidden2 conn. */
        hidden1.addOutputSynapse(synapse_H1_H2);
        hidden2.addInputSynapse(synapse_H1_H2);

        FullSynapse synapse_H2O = new FullSynapse(); /* Hidden2 -> Output conn. */
        hidden2.addOutputSynapse(synapse_H2O);
        output.addInputSynapse(synapse_H2O);


        int numInstances = trainingSet.numInstances();

        /* The I/O components */
        MemoryInputSynapse inputStream = new MemoryInputSynapse();
        inputStream.setInputArray(trainingSet.getInputArray());
        inputStream.setAdvancedColumnSelector("1-" + trainingSet.getNumInputs() );

        // use first split for training
        input.addInputSynapse(inputStream);


        /* The Trainer and its desired file */
        TeachingSynapse trainer = new TeachingSynapse();
        MemoryInputSynapse samples = new MemoryInputSynapse();
        samples.setInputArray(trainingSet.getInputArray());
        trainer.setDesired(samples);

        samples.setAdvancedColumnSelector( String.valueOf(trainingSet.getNumInputs() +1) + "-" + String.valueOf(trainingSet.getNumInputs() + trainingSet.getNumOutputs()) );
        output.addOutputSynapse(trainer);

        nnet = new NeuralNet();
        nnet.addLayer(input, NeuralNet.INPUT_LAYER);
        nnet.addLayer(hidden1, NeuralNet.HIDDEN_LAYER);
        nnet.addLayer(hidden2, NeuralNet.HIDDEN_LAYER);
        nnet.addLayer(output, NeuralNet.OUTPUT_LAYER);
        nnet.setTeacher(trainer);

        Monitor monitor = nnet.getMonitor();
        monitor.setLearningRate(0.3);
        monitor.setMomentum(0.2);
        monitor.setTrainingPatterns(numInstances); /* # of rows in the input file */
        monitor.setTotCicles(200); /* How many times the net must be trained*/
        monitor.setLearning(true); /* The net must be trained */
        monitor.addNeuralNetListener(this);

        nnet.start();
        nnet.getMonitor().Go();

        System.out.println("Number of instances: " + numInstances);
    }

    /**
     * Default predict function
     * @param sequence
     * @param localNNet
     * @param doFilter
     * @return
     */
    public ScoreAnnotatedSequence predict(Sequence sequence, NeuralNet localNNet, boolean doFilter)
    {
        ScoreAnnotatedSequence result = null;
        try
        {
            result = new ScoreAnnotatedSequence(sequence);

            if (localNNet == null)
            {
                if (useStoredNN && storedNN != null)
                {
                    // we have a stored NN, use it
                    localNNet = storedNN;
                }
                else
                {
                    // we either have no NN, or don't want to use it
                    localNNet = restoreNeuralNet(outputFile);
                    storedNN = localNNet;
                }
            }

            if (localNNet != null)
            {
                /* We get the first layer of the net (the input layer),
                  then remove all the input synapses attached to it and attach a DirectSynapse */
                Layer input = localNNet.getInputLayer();
                input.removeAllInputs();
                DirectSynapse memInp = new DirectSynapse();
                input.addInputSynapse(memInp);

                /* We get the last layer of the net (the output layer),
                 then remove all the output synapses attached to it and attach a DirectSynapse */
                Layer output = localNNet.getOutputLayer();
                output.removeAllOutputs();
                DirectSynapse memOut = new DirectSynapse();
                output.addOutputSynapse(memOut);

                // begin the neural net's thread
                localNNet.getMonitor().setLearning(false);
                localNNet.start();

                // Create the set that contains the instances of this sequence
                TrainingSet set = new SingleSequenceReducedTrainingSet();
                set.addInstance(sequence);

                SimpleSymbolList structure = new SimpleSymbolList(new StructureTools().getStructure());

                int i = 0;
                for (Iterator it = set.getInputInstanceIterator(); it.hasNext();)
                {

                    // Prepare the next input pattern
                    Pattern iPattern = new Pattern((double[]) it.next());
                    iPattern.setCount(1);

                    // Interrogate the net
                    memInp.fwdPut(iPattern);

                    // Read the output pattern and print out it
                    Pattern pattern = memOut.fwdGet();


                    if (verbose) System.out.print("Output #" + (++i) + " = ");
                    double max = -1;
                    int maxIndex = -1;
                    final double[] outputArr = pattern.getArray();
                    NumberFormat formatter = DecimalFormat.getInstance();
                    formatter.setMaximumFractionDigits(3);

                    for (int j = 0; j < outputArr.length; j++)
                    {
                        if (outputArr[j] > max)
                        {
                            max = outputArr[j];
                            maxIndex = j;
                        }

                        if (verbose) System.out.print(" " + formatter.format(outputArr[j]));
                    }

                    // get the symbol for this prediction
                    final Symbol symbol = ProteusTools.convertIntToStructure(maxIndex);
                    structure.addSymbol(symbol);

                    if (verbose) System.out.println(" Structure = " + symbol.getName());
                }

                result.setStructureAnnotation(structure);

                /* Stop training - don't know why this is necessary */
                Pattern stop = new Pattern(new double[set.getInputInstanceSize()]);
                stop.setCount(-1);
                memInp.fwdPut(stop);
                memOut.fwdGet();

                if (verbose) System.out.println("Finished");
            }
            else
            {
                throw new IllegalStateException("Cannot read in neural net!");
            }

            if (doFilter)
            {
                // filter the result with the attached filters
                for (Iterator it = filters.iterator(); it.hasNext();)
                    result = ((PredictionFilter) it.next()).filter(result);
            }
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();  //TODO
        }

        return result;
    }

    public void cicleTerminated(NeuralNetEvent e)
    {
        // call inherited method, it will do validation on a set of sequences
        super.cicleTerminated(e);

        Monitor mon = (Monitor) e.getSource();
        long c = mon.getCurrentCicle();
        long c1 = mon.getTotCicles() - c;
        if ((c1 < 5)  // for first 25, every 5
                || (c1 < 25 && c1 % 2 == 0)  // for first 25, every 5
                || (c1 < 200 && c1 % 5 == 0)  // up to 200, every 10
                || ((c1 % 200) == 0)) // after 200, every 200
            System.out.println(c + " cycles remaining - Error = " + mon.getGlobalError() );

        if ((c1 % 25) == 0)
            saveNeuralNet(outputFile);        
    }

    public static void main(String[] args) throws IllegalSymbolException, ChangeVetoException
    {
        SeqSeeDB db = SeqSeeDB.getInstance();

        boolean train = false;
        boolean predict = false;

        for (int i = 0; i < args.length; i++)
        {
            if (args[i].equals("-train"))
            {
                train = true;
                if ( i < args.length -1 ) {
                    try {
                        numToTrainOn = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
                }
                else {
                    printUsage();
                }
            }
            else if (args[i].equals("-predict"))
            {
                predict = true;
                if ( i < args.length -1 )
                    try {
                        numToPredict = Integer.parseInt(args[++i]);
                    }
                    catch (NumberFormatException e) {
                        printUsage();
                    }
            }
            else if (args[i].equals("-verbose")||args[i].equals("-v"))
            {
                verbose = true;
            }
            else if (args[i].equals("-output")||args[i].equals("-o"))
            {
                if ( i < args.length -1 ) {
                    String arg = args[++i];
                    if ( arg.charAt(0) == '-' )
                        printUsage();
                    else
                        outputFile = arg;
                }
            }
        }

        if (train)
        {
            // train on this set of sequences
            Predictor predictor = new SingleSequenceReducedPredictor();

            final SimpleFilter filter = new SimpleFilter();
            predictor.addFilter(filter);

            List randomSequences = db.getRandomSequences(numToTrainOn, 1000, 3000);

            predictor.train(randomSequences);
        }
        else if (predict)
        {
            // predict this set of sequences
            SingleSequenceReducedPredictor predictor = new SingleSequenceReducedPredictor();

            final NeuralFilter NFfilter = new NeuralFilter();
            NFfilter.useStoredNN = true;
            predictor.addFilter(NFfilter);

            final SimpleFilter filter = new SimpleFilter();
            predictor.useStoredNN = true;
            predictor.addFilter(filter);

            List randomSequences = db.getRandomSequences(numToPredict, 0, 100);

            StructureAnnotatedSequence prediction;
            Score score = new Score();
            Score filteredScore = new Score();

            for (Iterator it = randomSequences.iterator(); it.hasNext();)
            {
                final StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.next();
                prediction = predictor.predict(sequence, null, false);

                System.out.println("Sequence: \n" + sequence);

                double q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                double sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                score.q3 += q3;
                score.sov += sov;

                System.out.println("Prediction: "
                                   + " Q3: " + q3
                                   + " SOV: " + sov
                                   + "\n" + prediction);

                prediction = predictor.predict(sequence);

                q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());
                sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), prediction.getStructureAnnotation());

                filteredScore.q3 += q3;
                filteredScore.sov += sov;

                System.out.println("Prediction filtered: "
                                   + " Q3: " + q3
                                   + " SOV: " + sov
                                   + "\n" + prediction);
            }

            System.out.print("Avg Q3: " + score.q3 / randomSequences.size());
            System.out.println("SOV: " + score.sov / randomSequences.size());

            System.out.print("Filtered Avg Q3: " + filteredScore.q3 / randomSequences.size());
            System.out.println("Filtered SOV: " + filteredScore.sov / randomSequences.size());
        }
        else
            printUsage();
    }

    private static void printUsage() {
        System.out.println("Usage: <ProgName> (-train <# seq> | -predict <#seq>) [options]" );
        System.out.println("Options: ");
        System.out.println("-train <# sequences>\t\tTrain on a number of sequences");
        System.out.println("-predict <# sequences>\t\tPredict a number of sequences");
        System.out.println("-v, -verbose");
        System.out.println("-o, -output <file>\t\t\tFile to output the training model to (if applicable)");
        System.exit(-1);
    }
}
