package ca.ualberta.proteus.program.transsec;

import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.*;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.util.List;
import java.util.zip.GZIPInputStream;

/**
 * Created by IntelliJ IDEA.
 * User: Smontgomerie
 * Date: Jun 27, 2005
 * Time: 11:13:34 PM
 * To change this template use File | Settings | File Templates.
 */
public class TranssecRunner extends GenericProgramRunner
{
    private Sequence sequence;
    private Alignment alignment;
    private static final String DB_FILE = "xalign.db.gz";
    private StructureAnnotatedSequence result = null;

    public TranssecRunner(Sequence sequence, Alignment alignment)
    {
        this.sequence = sequence;
        this.alignment = alignment;
    }

    public void run()
    {
        String xalign = getXAlignResult(sequence);

        StringBuilder sb = new StringBuilder();
        sb.append("<?xml version=\"1.0\"?>\n");
        sb.append("<resultset>\n");
        sb.append("<row>\n<ID>0</ID>\n");
        sb.append("<sequence>" + sequence.seqString() + "</sequence>\n");

        String prediction;
        if ( sequence instanceof StructureAnnotatedSequence )
        {
            prediction = ((StructureAnnotatedSequence) sequence).getStructureAnnotation().seqString();
        }
        else
            prediction = sequence.seqString().replaceAll("\\w", "-");

        prediction = prediction.toUpperCase();
        prediction = prediction.replace('C', '-');
        prediction = prediction.replace('E', 'B');

        xalign = xalign.replace(sequence.getName(), "QUERY");


        sb.append("<prediction>" + prediction + "</prediction>\n");
        sb.append("<Comment></Comment>\n");
        sb.append("<BLASTResult></BLASTResult>\n");
        sb.append("<Xalign>" + xalign + "</Xalign>\n");
        sb.append("</row></resultset>\n");


        String ID = getUniqueIdentifier() + ".xml";

        try
        {
            FileOutputStream fos = new FileOutputStream(ID);
            fos.write(sb.toString().getBytes());
            fos.close();

            String transsecCmd = Config.getConfig().getStringValue("transsec.command");
            File f = new File(transsecCmd).getParentFile();
            transsecCmd += " " + f.getAbsolutePath() + "/net1.model "
                               + f.getAbsolutePath() + "/net2.model -pc -s 0 0 -x "
                                + new File("").getAbsolutePath() + "/" + ID;

//Process blastProc = Runtime.getRuntime().exec("blastall -p blastp -e 1e-10 -d /home/scott/Bioinformatics/myPDB/myPDBnew -FF -i " + fname);
            Process blastProc = Runtime.getRuntime().exec(transsecCmd, null, f.getAbsoluteFile());

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(blastProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(blastProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            // any error???
            int exitVal = blastProc.waitFor();
            System.out.println("ExitValue: " + exitVal);

            if (exitVal != 0)
            {
                System.out.println("MainPredictor failed");
                System.out.println(errorReader.getInputString());
                return;
            }
            else {
                System.out.println(errorReader.getInputString());
                System.out.println(outputReader.getInputString());
            }

            BufferedReader reader = new BufferedReader(new StringReader(outputReader.getInputString()));

            String line;
            boolean ok = false;
            while ( ( line = reader.readLine() ) != null ) {
                if ( line.contains("STR:")) {
                    ok = true;
                }
                if ( ok && line.contains("PRE:")) {
                    line = line.substring(4);
                    line = line.replace('-', 'C');
                    line = line.replace('B', 'E').trim();
                    break;
                }
            }

            System.out.println("Prediction: " + line);
            SimpleSymbolList struct = ProteusTools.createStructure(line);
            result = new StructureAnnotatedSequence(sequence, struct);
            System.out.println(result);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
    }

    private String getXAlignResult(Sequence sequence)
    {
        XAlignDB db = XAlignDB.getInstance();
        if ( ! db.contains(sequence)) {
            /// Not in XAlign DB
            BlastSettings settings = new BlastSettings();
            settings.setSequence(sequence);            
            settings.setBlastType(BlastSettings.BLASTTYPE.PSIBLAST);

            BlastRunner blastRunner = new BlastRunner(settings);
            blastRunner.run();
            List<Sequence> blastSequences = blastRunner.getUnalignedResult();
            blastSequences.add(sequence);

            String numAlignments = Config.getConfig().getStringValue("num_alignments");
            ProteusTools.chooseAlignments(sequence, blastSequences, Integer.valueOf(numAlignments) + 100 - 1); // leave 100 in there to be reduced later on

            assert blastSequences.contains(sequence) : "Alignment sequences does not contain query sequence";

            XAlignRunner runner = new XAlignRunner(sequence, blastSequences);
            runner.run();

            Alignment alignment = runner.getAlignment();

            alignment = ProteusTools.chooseAlignments(sequence, alignment, runner.getPPSHMap(),
                                ProteusTools.MIN_PPSH_THRESHOLD, ProteusTools.MAX_PPSH_THRESHOLD);

        }

        InputStream in = null;
        try
        {
            in = new FileInputStream(DB_FILE);
        }
        catch (FileNotFoundException e)
        {
        }

        if ( in == null )
            in = getClass().getResourceAsStream("/" + DB_FILE);

        if ( in == null ) {
            in = getClass().getResourceAsStream(DB_FILE);
        }
        if (in == null)
        {
            System.out.println("XAlign database does not exist... .");
        }

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try
        {
            builder = factory.newDocumentBuilder();
            GZIPInputStream gis = new GZIPInputStream(in);
            Document doc = builder.parse(gis);

            Element root = doc.getDocumentElement();

            for (int i = 0; i < root.getChildNodes().getLength(); i++)
            {
                Node node = root.getChildNodes().item(i);

                if (node.getNodeName().equals("record"))
                {
                    String id = node.getAttributes().getNamedItem("ID").getNodeValue();
                    if (id.equals(sequence.getName()))
                        return node.getTextContent();
                }
            }
        }
        catch (ParserConfigurationException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (SAXException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (IOException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        return null;
    }

    public static void main(String[] args) {
        Sequence sequence;
        try
        {
            sequence = SeqSeeDB.getTestSequence();
            Alignment alignment = AlignmentDB.getInstance().get(sequence);
            TranssecRunner runner = new TranssecRunner(sequence, alignment);
            runner.run();
        }
        catch (BioException e)
        {
            e.printStackTrace();  //TODO
        }
    }

    public StructureAnnotatedSequence getResult()
    {
        if ( result == null )
            throw new IllegalStateException("Result is null");
        return result;
    }
}
