package ca.ualberta.proteus.sequence;

import ca.ualberta.proteus.program.transsec.PredictionScore;
import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.BioException;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleGappedSequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.seq.impl.SimpleSequenceFactory;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.*;

import java.util.Iterator;
import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Apr 12, 2005
 *         Time: 8:34:44 PM
 *         <p/>
 *         A class to represent a gapped sequence, with an associated structure annotation
 * @author Scott Montgomerie
 * @see StructureAnnotatedSequence
 * @see ScoreAnnotatedSequence
 */
public class GappedStructureAnnotatedSequence extends SimpleGappedSequence
{
    private GappedSymbolList structureAnnotation;

    public GappedStructureAnnotatedSequence(Sequence sequence)
    {
        super(sequence);

        if (sequence instanceof StructureAnnotatedSequence)
            createGappedStructure((StructureAnnotatedSequence) sequence);
    }

    public GappedStructureAnnotatedSequence(StructureAnnotatedSequence sequence)
    {
        super(sequence);

        createGappedStructure(sequence);
    }

    public void setStructureAnnotation(GappedSymbolList structureAnnotation)
    {
        this.structureAnnotation = structureAnnotation;
        if (structureAnnotation.length() != getSourceSymbolList().length())
            throw new IllegalArgumentException("Passed in a structure that is not the same length as the sequence - use createGappedStructure() instead (" + structureAnnotation.length() + " != " + getSourceSymbolList().length() + " Name = " + getName() + ")");
    }

    public GappedSymbolList getStructureAnnotation()
    {
        return structureAnnotation;
    }

    /**
     * Adds a structure to this sequence
     *
     * @param ungappedStructureAnnotation - a ungapped structure annotation. This structure will be mapped to this object's gapped sequence
     */
    public void createGappedStructure(SymbolList ungappedStructureAnnotation)
    {
        if ( ungappedStructureAnnotation instanceof GappedSymbolList )
        {
            ungappedStructureAnnotation = ((GappedSymbolList) ungappedStructureAnnotation).getSourceSymbolList();
        }
        
        this.structureAnnotation = new SimpleGappedSymbolList(ungappedStructureAnnotation);
        String seqStr = seqString();

        int j = 1;
        for (int i = 1; i <= length(); i++)
        {

            //if ( symbolAt(i).equals(getAlphabet().getGapSymbol()) || symbolAt(i).equals(AlphabetManager.getGapSymbol()) ) {
            if (ProteusTools.isGap( symbolAt(i) ) )
            {
                try
                {
                    structureAnnotation.addGapInSource(j);
                }
                catch (IndexOutOfBoundsException e)
                {
                    e.printStackTrace();
                }
            }
            else
                j++;
        }

        if (structureAnnotation.length() != this.length())
        {
            System.out.println("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
            System.out.println("This:  " + this.getName() + " " + this.seqString());
            System.out.println("Struct:      " + structureAnnotation.seqString());
            System.out.println("Ungapped:      " + ungappedStructureAnnotation.seqString());

            throw new IllegalStateException("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
        }
    }

    /**
     * Adds a structure to this sequence
     *
     * @param structuredSequence - a ungapped sequence with a structure. This structure will be mapped to this object's gapped sequence
     */
    public void createGappedStructure(StructureAnnotatedSequence structuredSequence)
    {
        SymbolList ungappedStructureAnnotation = structuredSequence.getStructureAnnotation();
        this.structureAnnotation = new SimpleGappedSymbolList(ungappedStructureAnnotation);

        int j = 1;
        for (int i = 1; i <= length(); i++)
        {

            //if ( symbolAt(i).equals(getAlphabet().getGapSymbol()) || symbolAt(i).equals(AlphabetManager.getGapSymbol()) ) {
            if (symbolAt(i).equals(getAlphabet().getGapSymbol()))
            {
                try
                {
                    structureAnnotation.addGapInSource(j);
                }
                catch (IndexOutOfBoundsException e)
                {
                    e.printStackTrace();
                }
            }
            else
            {
                j++;
            }
        }

        if (structureAnnotation.length() != this.length())
        {
            System.out.println("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
            System.out.println("This:  " + this.getName() + " " + this.seqString());
            System.out.println("Struct:      " + structureAnnotation.seqString());
            System.out.println("Ungapped:      " + structuredSequence.seqString());
            System.out.println("Ungapped:      " + ungappedStructureAnnotation.seqString());

            throw new IllegalStateException("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
        }
    }

    /**
     * Create a ungapped structure-annotated sequence from an ungapped template sequence, and a structure annotation
     *
     * @param template
     * @param structure
     * @return an ungapped sequence
     */
    public static StructureAnnotatedSequence createUngappedStructure(Sequence template, StructureAnnotatedSequence structure)
    {
        // TODO refactor this so that is removes duplicate code from the overloaded methods
        SymbolList structureSymList = structure.getStructureAnnotation();

        StructureAnnotatedSequence result = null;
        Symbol[] symbolList = new Symbol[template.length()];

        Iterator seqIt = structure.iterator();
        Iterator structIt = structureSymList.iterator();

        int seqIndex = 1;
        int structIndex = 1;

        // Iterate through the sequence, removing any gaps (actually copying non-gap symbols)
        for (; structIt.hasNext() && seqIt.hasNext();)
        {
            Symbol structSymbol = (Symbol) structIt.next();
            Symbol seqSymbol = (Symbol) seqIt.next();

            if (!seqSymbol.equals(structure.getAlphabet().getGapSymbol()))
            {
                symbolList[seqIndex - 1] = structSymbol;
                seqIndex++;
            }

            structIndex++;
        }

        result = new StructureAnnotatedSequence(template);

        SimpleSymbolList newList = new SimpleSymbolList(symbolList, symbolList.length, new StructureTools().getStructure());
        result.setStructureAnnotation(newList);

        return result;

    }

    /**
     * Create a ungapped structure-annotated sequence from an ungapped template sequence, and a structure annotation
     *
     * @param template
     * @param structure
     * @return an ungapped sequence
     */
    public static ScoreAnnotatedSequence createUngappedStructure(Sequence template, ScoreAnnotatedSequence structure)
    {
        SymbolList structureSymList = structure.getStructureAnnotation();
        PredictionScore[] score = new PredictionScore[template.length()];

        ScoreAnnotatedSequence result = null;
        Symbol[] symbolList = new Symbol[template.length()];

        Iterator seqIt = structure.iterator();
        Iterator structIt = structureSymList.iterator();

        int seqIndex = 1;
        int structIndex = 1;

        // Iterate through the sequence, removing any gaps (actually copying non-gap symbols)
        for (; structIt.hasNext() && seqIt.hasNext();)
        {
            Symbol structSymbol = (Symbol) structIt.next();
            Symbol seqSymbol = (Symbol) seqIt.next();

            if (!seqSymbol.equals(structure.getAlphabet().getGapSymbol()) && !seqSymbol.equals(AlphabetManager.getGapSymbol()))
            {
                symbolList[seqIndex - 1] = structSymbol;
                score[seqIndex - 1] = structure.getScore(structIndex);
                seqIndex++;
            }

            structIndex++;
        }

        result = new ScoreAnnotatedSequence(template);
        result.setScores(score);

        SimpleSymbolList newList = new SimpleSymbolList(symbolList, symbolList.length, new StructureTools().getStructure());
        result.setStructureAnnotation(newList);

        return result;
    }

    public static Sequence removeGaps(Sequence sequence)
    {
        String seqStr = sequence.seqString();
        seqStr = seqStr.replaceAll("-", "");

        SymbolTokenization tokenization = null;
        SymbolList newList = null;
        try
        {
            tokenization = ProteinTools.getAlphabet().getTokenization("token");
            newList = new SimpleSymbolList(tokenization, seqStr);
        }
        catch (BioException e)
        {
            e.printStackTrace();  //TODO
        }

        return new SimpleSequence(newList, "", sequence.getName(), sequence.getAnnotation());
    }

    public String toString()
    {
        return "ID = " + getName() +
                "\nsequence  = " + seqString() +
                "\nstructure = " + structureAnnotation.seqString().toUpperCase();
    }

    /**
     * Returns the index of the first non-gap residue in this sequence
     * @return
     */
    @Override
    public int firstNonGap()
    {
        Symbol gap = this.getAlphabet().getGapSymbol();
        Symbol gap2 = new StructureTools().getStructure().getGapSymbol();
        Symbol gap3 = AlphabetManager.getGapSymbol();
        String seqStr = this.seqString();
        for (int i = 1; i < this.length(); i++)
        {
            if (!this.symbolAt(i).equals(gap) && !this.symbolAt(i).equals(gap2) && !this.symbolAt(i).equals(gap3) && seqStr.charAt(i) != 'X')
            {
                return i;
            }
        }
        return this.length();
    }


    /**
     * Returns the index of the last non-gap residue in this sequence
     * @return
     */
    @Override
    public int lastNonGap()
    {
        int thisLastNonGap = 0;
        Symbol gap = this.getAlphabet().getGapSymbol();
        Symbol gap2 = new StructureTools().getStructure().getGapSymbol();
        Symbol gap3 = AlphabetManager.getGapSymbol();
        String seqString = this.seqString();
        for (int i = 1; i < this.length(); i++)
        {
            if (!this.symbolAt(i).equals(gap) && !this.symbolAt(i).equals(gap2)
                    && !this.symbolAt(i).equals(gap3) && seqString.charAt(i) != 'X')
            {
                thisLastNonGap = i;
            }
        }
        return thisLastNonGap;
    }

    /**
     * Returns the number of overlapping residues
     *
     * @param comparedHomoSequence
     * @return the number of overlapping residues
     */
    public int residuesOverlapping(GappedStructureAnnotatedSequence comparedHomoSequence)
    {
        String homologString = comparedHomoSequence.getStructureAnnotation().seqString();
        String queryString = this.getStructureAnnotation().seqString();

        int residues = 0;
        for (int i = 0; i < homologString.length() && i < queryString.length(); i++)
        {
            if (homologString.charAt(i) != '-' && queryString.charAt(i) != '-')
            {
                residues++;
            }
        }
        return residues;
    }

    /**
     * Returns if this sequence overlaps the homolog
     *
     * @param comparedHomoSequence
     * @return true if this sequence overlaps the homolog
     */
    public boolean overlaps(GappedStructureAnnotatedSequence comparedHomoSequence)
    {
        return overlaps(comparedHomoSequence, 0);
    }

    /**
     * Returns if this sequence overlaps the homolog
     *
     * @param comparedHomoSequence
     * @param allowableOverlap,    DO NOT Use set to 0
     * @return true if this sequence overlaps the homolog
     */
    public boolean overlaps(GappedStructureAnnotatedSequence comparedHomoSequence, int allowableOverlap)
    {
        int thisFirstNonGap = 0;
        int thisLastNonGap = 0;
        int otherFirstNonGap = 0;
        int otherLastNonGap = 0;
        assert this.length() == comparedHomoSequence.length() : "Testing overlapping on non-equal-length sequences";

        Symbol gap = this.getAlphabet().getGapSymbol();
        Symbol gap2 = new StructureTools().getStructure().getGapSymbol();

        for (int i = 1; i < Math.min(this.length(), comparedHomoSequence.length()); i++)
        {
            if (!this.symbolAt(i).equals(gap) && !this.symbolAt(i).equals(gap2))
            {
                if (thisFirstNonGap == 0)
                    thisFirstNonGap = i;
                thisLastNonGap = i;
            }

            if (!comparedHomoSequence.symbolAt(i).equals(gap) && !comparedHomoSequence.symbolAt(i).equals(gap2))
            {
                if (otherFirstNonGap == 0)
                    otherFirstNonGap = i;
                otherLastNonGap = i;
            }
        }

        if (thisFirstNonGap + allowableOverlap > otherLastNonGap)
        {
            System.out.println(String.format("%s's first gap (%d) after %s's last gap (%d)", this.getName(), thisFirstNonGap, comparedHomoSequence.getName(), otherLastNonGap));
            return false;
        }
        if (thisLastNonGap + allowableOverlap < otherFirstNonGap)
        {
            System.out.println(String.format("%s's first gap (%d) after %s's last gap (%d)", this.getName(), thisLastNonGap, comparedHomoSequence.getName(), otherFirstNonGap));
            return false;
        }
        //System.out.println( this.getName() + " and " + comparedHomoSequence.getName() + " overlap");
        //System.out.println( this.seqString() + String.format("\nFirst gap = %d Last gap = %d", thisFirstNonGap, thisLastNonGap));
        //System.out.println( comparedHomoSequence.seqString() + String.format("\nFirst gap = %d Last gap = %d", otherFirstNonGap, otherLastNonGap));
        return true;
    }

    public int residuesNotOverlapping(GappedStructureAnnotatedSequence comparedHomoSequence)
    {
        return this.length() - residuesOverlapping(comparedHomoSequence);
    }

    public int percentIdentity(GappedStructureAnnotatedSequence comparedHomoSequence)
    {
        String homologString = comparedHomoSequence.getStructureAnnotation().seqString();
        String queryString = this.getStructureAnnotation().seqString();

        int residues = 0;
        for (int i = 0; i < homologString.length() && i < queryString.length(); i++)
        {
            if (homologString.charAt(i) == queryString.charAt(i))
            {
                residues++;
            }
        }
        return residues;
    }

    /**
     * Maps a complete structured sequence to a portion of that sequence. "This" is the incomplete sequence.
     *
     * @param structuredSequence
     */
    public void createGappedStructureToIncompleteSequence(StructureAnnotatedSequence structuredSequence)
    {
        String ungappedStructureAnnotation = structuredSequence.getStructureAnnotation().seqString();
        String ungappedTemplate = structuredSequence.seqString();

        // take the first 5 letters of the sequence and find out where the
        String completePrefix = structuredSequence.seqString().substring(0, 5);
        String ungappedString = this.seqString().replace("-", "").replace("~", "");
        String incompletePrefix = ungappedString.substring(0, 5);


        int start = ungappedTemplate.indexOf(incompletePrefix);

        if (start == -1)
        {
            throw new IllegalArgumentException(
                    String.format("%s \nis not a substring of \n%s",
                            incompletePrefix,
                            completePrefix));
        }

        // truncate it
        ungappedStructureAnnotation = ungappedStructureAnnotation.substring(start, start + ungappedString.length());

        this.structureAnnotation = new SimpleGappedSymbolList(ProteusTools.createStructure(ungappedStructureAnnotation));

        String thisGappedString = seqString();

        int j = 1;
        for (int i = 0; i < length(); i++)
        {
            char residue = thisGappedString.charAt(i);
            if (residue == '-' || residue == '~')
            {
                try
                {
                    structureAnnotation.addGapInSource(j);
                }
                catch (IndexOutOfBoundsException e)
                {
                    e.printStackTrace();
                }
            }
            else
            {
                j++;
            }
        }

        if (structureAnnotation.length() != this.length())
        {
            System.out.println("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
            System.out.println("This:  " + this.getName() + " " + this.seqString());
            System.out.println("Struct:      " + structureAnnotation.seqString());
            System.out.println("Ungapped:      " + structuredSequence.seqString());
            System.out.println("Ungapped:      " + ungappedStructureAnnotation);

            throw new IllegalStateException("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
        }
    }

    /**
     * Returns a new StructureAnnotatedSequence without gaps.
     *  
     * @return
     */
    public StructureAnnotatedSequence getUngappedSequence()
    {
        SymbolList sequence = this.getSourceSymbolList();
        GappedSymbolList structure = getStructureAnnotation();

        if ( sequence instanceof GappedSymbolList )
        {
            sequence = ((GappedSymbolList) sequence).getSourceSymbolList();
        }

        Symbol[] symbolList = new Symbol[sequence.length()];
        Arrays.fill(symbolList, structure.getAlphabet().getGapSymbol());

        // because structure.firstNonGap() won't work
        int firstNonGap = 0;
        for (int structIndex = structure.firstNonGap(); structIndex<structure.length(); structIndex++)
        {
            Symbol structSymbol = (Symbol) structure.symbolAt(structIndex);                        
            if (!structSymbol.equals(structure.getAlphabet().getGapSymbol())
                    && !structSymbol.equals(AlphabetManager.getGapSymbol())
                    && !structSymbol.getName().equals("[]"))
            {
                firstNonGap = structIndex;
                break;
            }
        }

        int seqIndex = firstNonGap;

        // Iterate through the sequence, removing any gaps (actually copying non-gap symbols)
        for (int structIndex = firstNonGap; structIndex<structure.length(); structIndex++)
        {
            Symbol structSymbol = (Symbol) structure.symbolAt(structIndex);
            //Symbol seqSymbol = (Symbol) seqIt.next();

            if (!structSymbol.equals(structure.getAlphabet().getGapSymbol())
                    && !structSymbol.equals(AlphabetManager.getGapSymbol())
                    && !structSymbol.getName().equals("[]"))
            {
                symbolList[seqIndex - 1] = structSymbol;
                seqIndex++;
            }
        }

        Sequence sequence1 = new SimpleSequenceFactory().createSequence(sequence, "", getName(), getAnnotation());
        SimpleSymbolList structureList = new SimpleSymbolList(symbolList, symbolList.length, structure.getAlphabet());
        StructureAnnotatedSequence result = new StructureAnnotatedSequence(sequence1, structureList);

        /*SimpleSymbolList newList =  new SimpleSymbolList(symbolList, symbolList.length, new StructureTools().getStructure());
        result.setStructureAnnotation( newList );
*/
        return result;
    }

    /**
     * Takes the structure aligned from a gapped sequence and maps it to this sequence,
     * assuming they are part of the same alignment
     * 
     * @param alignedSequence
     */
    public void assignAlignedStructure(GappedStructureAnnotatedSequence alignedSequence)
    {
        String ungappedStructureAnnotation = alignedSequence.getStructureAnnotation().getSourceSymbolList().seqString();
     //    this.structureAnnotation = new SimpleGappedSymbolList(ProteusTools.createStructure(ungappedStructureAnnotation));

        String thisGappedString = seqString();
        String alignedString = alignedSequence.seqString();

        char[] residues  = new char[thisGappedString.length()];

        int alignedSequenceIndex = 0;
        for (int i = 0; i < length(); i++)
        {
            char residue = thisGappedString.charAt(i);
            if (residue != '-' && residue != '~')
            {
            	// gap
            	if( alignedString.charAt(i) == '-' || alignedString.charAt(i) == '~' || alignedSequenceIndex >= ungappedStructureAnnotation.length() )
            	{
            		residues[i] = '-';
            	}
            	else
                {
                    residues[i] = ungappedStructureAnnotation.charAt(alignedSequenceIndex++);
                }
            }
            else
            {
                // this is a gap
                residues[i] = '-';

                // if the aligned sequence has a gap while this sequence doesn't, increase alignedSequenceIndex
                if ( ! ProteusTools.isGap( alignedString.charAt(i)) )
                {
                    alignedSequenceIndex++;
                }
            }
        }

        this.structureAnnotation = ProteusTools.createGappedStructure(new String(residues));

        if (structureAnnotation.length() != this.length())
        {
            System.out.println("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
            System.out.println("This:  " + this.getName() + " " + this.seqString());
            System.out.println("Struct:      " + structureAnnotation.seqString());
            System.out.println("Ungapped:      " + alignedSequence.seqString());
            System.out.println("Ungapped:      " + ungappedStructureAnnotation);

            throw new IllegalStateException("Structure does not match sequence (Structure = " + structureAnnotation.length() + ", sequence = " + this.length());
        }
    }
}