package ca.ualberta.proteus.sequence;

import ca.ualberta.proteus.program.transsec.PredictionScore;
import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.BioException;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleSequenceFactory;

import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Mar 26, 2005
 *         Time: 1:55:15 PM
 *         To change this template use File | Settings | File Templates.
 *         <p/>
 *         A class to represent a sequence, annotated with a structure and a score for each residue.
 * @see PredictionScore
 * @see StructureAnnotatedSequence
 */
public class ScoreAnnotatedSequence extends StructureAnnotatedSequence
{
    PredictionScore[] scores;

    /**
     * Copy constructor, initializes with the sequence, and a null structure annotation
     *
     * @param sequence - the sequence to construct it from
     */
    public ScoreAnnotatedSequence(Sequence sequence)
    {
        super(sequence);
        scores = new PredictionScore[sequence.length()];

        for (int i = 0; i < scores.length; i++)
            scores[i] = new PredictionScore();

        if ( sequence instanceof StructureAnnotatedSequence )
        {
            this.setStructureAnnotation(((StructureAnnotatedSequence) sequence).getStructureAnnotation());
        }
    }

    public ScoreAnnotatedSequence(StructureAnnotatedSequence sequence)
    {
        super(sequence);
        scores = new PredictionScore[sequence.length()];

        for (int i = 0; i < scores.length; i++)
            scores[i] = new PredictionScore();

        this.setStructureAnnotation(sequence.getStructureAnnotation());
    }

    /**
     * Copy constructor, initializes with the sequence, and a null structure annotation
     *
     * @param seqStr
     */
    public ScoreAnnotatedSequence(String seqStr, String structStr, String name, PredictionScore[] scores)
            throws BioException
    {
        super(seqStr, structStr, name);
        this.scores = scores;
    }


    /**
     * @param seqStr
     */
    public ScoreAnnotatedSequence(String seqStr, String structStr, String name, String score)
            throws BioException
    {
        super(seqStr, structStr, name);

        assert score.length() == structStr.length() : "Score string length does not equal structure string";

        this.scores = new PredictionScore[score.length()];
        char[] struct = structStr.toUpperCase().toCharArray();

        convertStringToScore(score, struct);
    }

    private void convertStringToScore(String score, char[] struct)
    {
        for (int i = 0; i < score.length(); i++)
        {
            double d = Double.parseDouble(String.valueOf(score.charAt(i))) / (double) 10;
            switch (struct[i])
            {
                case'C':
                    scores[i] = new PredictionScore(d, 0, 0);
                    break;
                case'E':
                    scores[i] = new PredictionScore(0, d, 0);
                    break;
                case'H':
                    scores[i] = new PredictionScore(0, 0, d);
                    break;
            }
        }
    }

    public ScoreAnnotatedSequence(Sequence sequence, String structStr, String score)
    {
        super(sequence, structStr);


        assert score.length() == structStr.length() : "Score string does not equal structure string";

        this.scores = new PredictionScore[score.length()];
        char[] struct = structStr.toUpperCase().toCharArray();

        convertStringToScore(score, struct);
    }

    /**
     * Set the score at certain position
     *
     * @param pos   (1-based, in line with the sequence object being 1-based)
     * @param score
     */
    public void setScore(int pos, PredictionScore score)
    {
        scores[pos - 1] = score;
    }


    /**
     * Get the score at a certain position
     *
     * @param pos (1-based)
     */
    public PredictionScore getScore(int pos)
    {
        if (scores[pos - 1] == null)
            return new PredictionScore();
        return scores[pos - 1];
    }

    public String toString()
    {
        StringBuffer rv = new StringBuffer(
                "\nsequence  = " + seqString() +
                        "\nstructure = " + getStructureAnnotation().seqString().toUpperCase() +
                        "\nscore     = ");

        if (scores != null)
        {
            // add the scores as int values
            for (PredictionScore score : scores)
            {
                if (score != null)
                    rv.append((int) Math.floor(score.getMax() * 10));
                else
                    rv.append("0");
            }
        }
        else
            rv.append("null");

        return rv.toString();
    }

    public PredictionScore[] getScores()
    {
        return scores;
    }

    public void setScores(PredictionScore[] scores)
    {
        assert (scores.length == this.length());

        this.scores = scores;
    }

    public void setScore(int index, double coil, double helix, double strand)
    {
        setScore(index, coil, helix, strand, 0);
    }

    public void setScore(int index, double coil, double helix, double strand, double gap)
    {
        try
        {
            scores[index - 1] = new PredictionScore(coil, helix, strand, gap);
        }
        catch (ArrayIndexOutOfBoundsException e)
        {
            System.err.println("Index out of bounds - length is " + scores.length);
            throw e;
        }
    }

    public void fillScores(double coil, double helix, double strand, double gap)
    {
        for (int i = 0; i < scores.length; i++)
        {
            scores[i] = new PredictionScore(coil, helix, strand, gap);
        }
    }

    public void fillScores(double max)
    {
        StructureTools structureTools = new StructureTools();
        for (int i = 0; i < scores.length; i++)
        {
            if (getStructureAnnotation().symbolAt(i + 1).equals(structureTools.getC()))
                scores[i] = new PredictionScore(0.9, 0, 0, 0);
            else if (getStructureAnnotation().symbolAt(i + 1).equals(structureTools.getH()))
                scores[i] = new PredictionScore(0, 0.9, 0, 0);
            else if (getStructureAnnotation().symbolAt(i + 1).equals(structureTools.getE()))
                scores[i] = new PredictionScore(0, 0, 0.9, 0);
        }
    }

    public double averageScore()
    {
        double average = 0;
        for (PredictionScore score : scores)
        {
            average += score.getMax();
        }
        return average / (double) scores.length;
    }

    public String getScoresAsStr()
    {
        if (scores == null)
            return "null";

        StringBuffer rv = new StringBuffer();

        // add the scores as int values
        for (PredictionScore score : scores)
        {
            if (score == null)
                return "null";
            else
            {
                int i = (int) Math.floor(score.getMax() * 10);

                // Score high is 9
                i = Math.min(9, i);
                rv.append(i);
            }
        }

        return rv.toString();
    }

    public boolean equals(Object o)
    {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        if (!super.equals(o)) return false;

        final ScoreAnnotatedSequence that = (ScoreAnnotatedSequence) o;

        if (!Arrays.equals(scores, that.scores)) return false;

        return true;
    }


    @Override
    public ScoreAnnotatedSequence clone()
    {
        // clone the sequence
        ScoreAnnotatedSequence sas = new ScoreAnnotatedSequence(super.clone());

        // clone the structure
        char[] seqString = getStructureAnnotation().seqString().toCharArray();
        SymbolList structureAnnotation = new SimpleSequenceFactory().createSequence(
                ProteusTools.createStructure(String.valueOf(seqString)),
                "", "STRUCTURE", new SimpleAnnotation());

        sas.setStructureAnnotation(structureAnnotation);

        // clone the scores
        PredictionScore[] newScores = new PredictionScore[scores.length];

        for (int i = 0; i < scores.length; i++)
            newScores[i] = scores[i].clone();

        sas.setScores(newScores);

        return sas;
    }
}
