package ca.ualberta.proteus.sequence;

import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.*;
import org.biojava.bio.symbol.*;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Feb 28, 2005
 *         Time: 11:46:31 PM
 *         To change this template use File | Settings | File Templates.
 */
public class StructureAnnotatedSequence extends SimpleSequence implements Sequence
{

    private SymbolList structureAnnotation;
    private static Alphabet structureAlphabet = AlphabetManager.alphabetForName("STRUCTURE");

    /**
     * Copy constructor, initializes with the sequence, and a null structure annotation
     *
     * @param sequence - the sequence to construct it from
     */
    public StructureAnnotatedSequence(Sequence sequence)
    {
        super(sequence, "", sequence.getName(), new SimpleAnnotation());
    }

    /**
     * Copy constructor
     *
     * @param sequence - the sequence to construct it from
     */
    public StructureAnnotatedSequence(StructureAnnotatedSequence sequence)
    {
        super(sequence, "", sequence.getName(), new SimpleAnnotation());
        this.setStructureAnnotation(sequence.getStructureAnnotation());
    }

    /**
     * Constructs a structureAnnotatedSequence from a sequence and a symbol list
     *
     * @param sequence
     * @param structure
     */
    public StructureAnnotatedSequence(Sequence sequence, SymbolList structure)
    {
        this(sequence);
        this.setStructureAnnotation(structure);
    }

    /**
     * Constructs a structureAnnotatedSequence from a sequence and a string representation of a structure Annotation
     *
     * @param sequence
     * @param structure
     */
    public StructureAnnotatedSequence(Sequence sequence, String structure)
    {
        this(sequence);
        setStructureAnnotation(new SimpleSequenceFactory().createSequence(
                ProteusTools.createStructure(structure),
                "", "STRUCTURE", new SimpleAnnotation()));
    }

    /**
     * Builds a new sequence annotated with a structure
     * Use this method whenever possible, as it is much more efficient than the string parameter version
     *
     * @param protein
     * @param annotation
     * @param name
     * @throws IllegalSymbolException
     */
    public StructureAnnotatedSequence(StringBuilder protein, StringBuilder annotation, String name) throws IllegalSymbolException
    {

        // create the protein sequence
        super(ProteinTools.createProtein(protein.toString()), "", name, new SimpleAnnotation());

        // replace 'B' with 'E' ( E's are the standard for beta sheets )
       /* for (int i = 0; i < annotation.length(); i++)
            if (annotation.charAt(i) == 'B')
                annotation.setCharAt(i, 'E');
                // But we now need to take into account B for beta-barrell*/

        setStructureAnnotation(new SimpleSequenceFactory().createSequence(
                ProteusTools.createStructure(annotation.toString()),
                "", "STRUCTURE", new SimpleAnnotation()));
    }

    /**
     * Builds a new sequence annotated with a structure
     * Use the string buffer parameter method whenever possible, as it is much more efficient than the string parameter version
     *
     * @param protein
     * @param annotation
     * @param name
     * @throws IllegalSymbolException
     */
    public StructureAnnotatedSequence(String protein, String annotation, String name) throws IllegalSymbolException
    {

        // create the protein sequence
        super(ProteinTools.createProtein(protein), "", name, new SimpleAnnotation());

        // annotation = annotation.replaceAll("B", "E"); // E's are the standard for beta sheets
        // But we now need to take into account B for beta-barrell

        setStructureAnnotation(new SimpleSequenceFactory().createSequence(
                ProteusTools.createStructure(annotation),
                "", "STRUCTURE", new SimpleAnnotation()));
    }

    public Alphabet getStructureAlphabet()
    {
        return structureAlphabet;
    }

    public String seqString()
    {
        return super.seqString();
    }

    public String toString()
    {
        return "ID = " + getName() +
                "\nsequence  = " + seqString() +
                "\nstructure = " + structureAnnotation.seqString().toUpperCase();
    }

    public boolean equals(Object o)
    {
        if (this == o) return true;
        if (!(o instanceof StructureAnnotatedSequence)) return false;

        final StructureAnnotatedSequence structureAnnotatedSequence = (StructureAnnotatedSequence) o;

        if (!structureAnnotation.seqString().equals(structureAnnotatedSequence.structureAnnotation.seqString()))
            return false;
        if (!super.equals(structureAnnotatedSequence)) return false;

        return true;
    }

    public int hashCode()
    {
        return super.hashCode() + structureAnnotation.hashCode();
    }

    public void setStructureAt(int index, Symbol symbol)
    {
        if (index < 1)
        {
            throw new IndexOutOfBoundsException("Index must be > 1");
        }
/*

        char[] seqString = structureAnnotation.seqString().toCharArray();
        seqString[index - 1] = ProteusTools.convertStructureToChar(symbol);
        this.structureAnnotation = new SimpleSequenceFactory().createSequence(
                    ProteusTools.createStructure(String.valueOf(seqString)),
                    "", "STRUCTURE", new SimpleAnnotation() );
*/

        Symbol[] symbols;
        if (structureAnnotation instanceof SimpleSymbolList)
        {
            // FAST way
            symbols = ((SimpleSymbolList) structureAnnotation).getSymbolArray();
            symbols[index - 1] = symbol;
        }
        else if (structureAnnotation instanceof Sequence)
        {
            // Re-create the entire sequence

            char[] seqString = structureAnnotation.seqString().toCharArray();
            seqString[index - 1] = ProteusTools.convertStructureToChar(symbol);
            this.structureAnnotation = new SimpleSequenceFactory().createSequence(
                    ProteusTools.createStructure(String.valueOf(seqString)),
                    "", "STRUCTURE", new SimpleAnnotation());
        }

    }

    public void setStructureAnnotation(SymbolList structureAnnotation)
    {
        this.structureAnnotation = structureAnnotation;
    }

    public SymbolList getStructureAnnotation()
    {
        return structureAnnotation;
    }

    /**
     * Returns a new structureAnnotatedSequence with the specified subsequence
     *
     * @param start
     * @param end
     * @return a new structureAnnotatedSequence
     */
    public StructureAnnotatedSequence subSequence(int start, int end)
    {
        if (end > this.length())
        {
            end = this.length();
        }

        Sequence sequence = new SimpleSequence(this.subList(start, end), "", getName(), getAnnotation());
        SymbolList symbolList = structureAnnotation.subList(start, end);
        return new StructureAnnotatedSequence(sequence, symbolList);

    }


    @Override
    public StructureAnnotatedSequence clone()
    {
        // clone the sequence
        StructureAnnotatedSequence sas = new StructureAnnotatedSequence(super.clone());

        // clone the structure
        char[] seqString = structureAnnotation.seqString().toCharArray();
        SymbolList structureAnnotation = new SimpleSequenceFactory().createSequence(
                ProteusTools.createStructure(String.valueOf(seqString)),
                "", "STRUCTURE", new SimpleAnnotation());

        sas.setStructureAnnotation(structureAnnotation);

        return sas;
    }
}
