package ca.ualberta.proteus.util;

import org.biojava.bio.BioException;
import org.biojava.bio.SimpleAnnotation;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.SimpleAlignment;
import org.biojava.bio.symbol.SimpleSymbolList;
import org.biojava.bio.symbol.SymbolList;

import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Created by IntelliJ IDEA.
 * User: Smontgomerie
 * Date: Apr 19, 2005
 * Time: 2:41:39 PM
 * To change this template use File | Settings | File Templates.
 */
public class AlignmentDB {
    private static final String DB_FILE = "align.db.gz";
    private static final String DB_NAME = "ClustalW results";
    private Map<Sequence, Alignment> alignmentList = new HashMap<Sequence,Alignment>();
    private static AlignmentDB db = null;

    public static AlignmentDB getInstance() {
        if ( db == null ) {
            db = loadDB();
        }
        return db;
    }

    /**
     * Reads in the blast database file
     * @return
     */
    private synchronized static AlignmentDB loadDB() {

        AlignmentDB rv = new AlignmentDB();
        InputStream in = rv.getClass().getResourceAsStream(DB_FILE);
        if ( in == null )
            in = rv.getClass().getResourceAsStream("/" + DB_FILE);
        
        if ( in == null ) {
            try
            {
                in = new FileInputStream(DB_FILE);
            }
            catch (FileNotFoundException e)
            {
            }
        }

        if ( in == null ) {
            System.out.println("Alignment database does not exist... creating a new one.");
            return rv;
        }

        System.out.println("Loading ClustalW Alignment DB...");
        try {
            synchronized(rv.alignmentList) {
                SymbolTokenization tokenization = ProteinTools.getAlphabet().getTokenization("token");
                BufferedReader reader = new BufferedReader( new InputStreamReader( new GZIPInputStream(in) ) );

                // can't use the SeqIOTools readers because this is our own DB format
                String line;
                String name;
                String seqStr;

                while ( (line = reader.readLine()) != null ) {
                    // find beginning of Alignment record
                    if ( line.startsWith(">>")) {
                        Map<String, SymbolList> alignList = new HashMap<String, SymbolList>();

                        line = line.substring(2);
                        String querySeqName = line.substring(0, line.indexOf(':')).trim();
                        String querySeqStr = line.substring(line.indexOf(':') + 1).trim();
                        while ( (line = reader.readLine()) != null && !line.equals("") ) {
                            name = line.substring(0, line.indexOf(':')).trim();
                            seqStr = line.substring(line.indexOf(':') + 1).trim();
                            SymbolList symList = new SimpleSymbolList(tokenization, seqStr);

                            // add to list
                            alignList.put(name, new SimpleSequence(symList, "", name, new SimpleAnnotation()));
                        }

                        // create alignment from the alignment map
                        SymbolList symList = new SimpleSymbolList(tokenization, querySeqStr);
                        SimpleSequence key = new SimpleSequence(symList, "", querySeqName, new SimpleAnnotation());
                        rv.alignmentList.put( key, new SimpleAlignment(alignList) );
                    }
                }
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
        catch (BioException e)
        {
            e.printStackTrace();  //TODO
        }
        System.out.println("Done");
        return rv;
    }

    public boolean contains(Sequence sequence) {
        for (Iterator it = alignmentList.keySet().iterator(); it.hasNext(); ) {
            Sequence seq = (Sequence) it.next();
            if ( seq.getName().equals(sequence.getName()))
                return true;
        }
        return false;
    }

    public Alignment get(Sequence sequence) {
        for (Iterator it = alignmentList.keySet().iterator(); it.hasNext(); ) {
            Sequence seq = (Sequence) it.next();
            if ( seq.getName().equals(sequence.getName()))
                return alignmentList.get(seq);
        }
        return null;
    }

    /**
     * Writes out the contents of the database file
     */
    private synchronized void writeDB() {
        try {
            synchronized(alignmentList) {
                OutputStream out = new GZIPOutputStream(new FileOutputStream(DB_FILE));

                for ( Iterator it = alignmentList.keySet().iterator(); it.hasNext(); ) {
                    Sequence querySequence = (Sequence) it.next();
                    Alignment alignment = alignmentList.get(querySequence);

                    // Write header for the alignment
                    String s = "\n>>" + querySequence.getName() + " : " + querySequence.seqString() + "\n";
                    out.write( s.getBytes() );

                    // write out the alignment
                    for ( Iterator<Sequence> seqIt = alignment.symbolListIterator(); seqIt.hasNext(); ) {
                        Sequence seq = seqIt.next();
                        out.write( (seq.getName() + ": " + seq.seqString() + "\n").getBytes() );
                    }
                }

                out.close();
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Add a sequence to the database
     * @param querySequence - for the query sequence
     * @param alignment - for the query sequence
     */
    public synchronized void add(Sequence querySequence, Alignment alignment) {
      /*  try {
            alignmentList.wait();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }*/
        alignmentList.put(querySequence, alignment);
        writeDB();
    }

    public void clearDB() {
        alignmentList.clear();
        db = null;
    }
}
