package ca.ualberta.proteus.util;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import org.apache.log4j.Logger;
import org.biojava.bio.BioException;
import org.biojava.bio.program.sax.BlastLikeSAXParser;
import org.biojava.bio.program.sax.blastxml.BlastXMLParserFacade;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Apr 9, 2005
 *         Time: 5:04:51 PM
 *         To change this template use File | Settings | File Templates.
 */

public class BlastRunner extends GenericProgramRunner
{
    private List<Sequence> seqList;
    private List<GappedSequence> gappedSeqList;

    private Map<OUTPUT_FILE_TYPE, String> outputFileMap = new HashMap<OUTPUT_FILE_TYPE, String>();
    private String output = null;
    private BlastSettings settings;

    public static final String OUTPUT_SUFFIX = ".out";
    public static final String PSSM_SUFFIX = ".pos";
    public static final String CHKPOINT_SUFFIX = ".chk";

    public static final String BLUNT_END_OUTPUT_FLAG = "-m7";
    public static final String XML_OUTPUT_FLAG = "-m6";

    static final int NUM_SEQUENCES_TO_SHOW = 1000;  // BLAST -b and -v values
    private Logger logger = Logger.getLogger(BlastRunner.class);

    public enum OUTPUT_FILE_TYPE
    {
        PSSM, FREQUENCY, CHECKPOINT, OUTPUT
    }

    public String getOutput() throws IOException
    {
        return output;
    }

    public BlastRunner(BlastSettings settings)
    {
        this.settings = settings;
        finished = false;
    }

    /**
     * Returns a list of sequences that BLAST returns
     */
    public void run()
    {
        if (settings.getBlasttype() == null || settings.getBlasttype() == BlastSettings.BLASTTYPE.NONE)
            throw new IllegalStateException("Blast type not set");
        else if (settings.getBlasttype() == BlastSettings.BLASTTYPE.BLASTP)
            runBlastp();
        else if (settings.getBlasttype() == BlastSettings.BLASTTYPE.PSIBLAST)
            runPSIBlast();
        else if (settings.getBlasttype() == BlastSettings.BLASTTYPE.PROTEUS_BLASTPARSER_BLASTP)
            runBlastPOwnParser();
    }

    public BlastSettings getSettings()
    {
        return settings;
    }

    private void runPSIBlast()
    {
        // if two BLAST processes are created concurrently by multiple threads, we need a unique file name
        String runID = getUniqueIdentifier();
        String fname = runID + ".blast";
        //StreamHandler outputReader = null;
        String blastCmd = null;

        File outputFile = new File(runID + OUTPUT_SUFFIX);

        try
        {
            writeBlastInputFile(fname);

            seqList = new ArrayList<Sequence>();

            Config config = Config.getConfig();
            String blastCmdName = config.getStringValue("blastpgp.command");
            String blastDBName = config.getStringValue("blast.dbname");
            if ( settings.getDatabase()!= null )    // override the DB
                blastDBName = settings.getDatabase();

            String evalue;
            if ( settings.getOverriddenEvalue() != -1 )
                evalue = String.valueOf(settings.getOverriddenEvalue());
            else
                evalue = config.getStringValue("blast.evalue");
//            String otherParams = config.getStringValue("blast.others");

            outputFileMap.put(OUTPUT_FILE_TYPE.PSSM, runID + PSSM_SUFFIX);
            outputFileMap.put(OUTPUT_FILE_TYPE.CHECKPOINT, runID + CHKPOINT_SUFFIX);
            outputFileMap.put(OUTPUT_FILE_TYPE.OUTPUT, outputFile.getAbsolutePath());
            int numCPUs = Runtime.getRuntime().availableProcessors();
            String outputFlag;
            switch ( settings.getAlignmentView() ) {
                case BLUNT_ENDS:
                    outputFlag = XML_OUTPUT_FLAG;
                    break;
                case XML:
                    outputFlag = BLUNT_END_OUTPUT_FLAG;
                    break;
                default:
                    outputFlag = BLUNT_END_OUTPUT_FLAG;
            }

            blastCmd = String.format(blastCmdName + " %s -a %d -b%d -v%d -e %s -j 3 -h 0.001 -Q %s -d %s -i %s -C %s -o %s",
                                                    outputFlag, numCPUs, NUM_SEQUENCES_TO_SHOW, NUM_SEQUENCES_TO_SHOW,
                                                    evalue, outputFileMap.get(OUTPUT_FILE_TYPE.PSSM), blastDBName,
                                                    fname, outputFileMap.get(OUTPUT_FILE_TYPE.CHECKPOINT),
                                                    outputFileMap.get(OUTPUT_FILE_TYPE.OUTPUT));

            blastCmd += " -M " + settings.getMatrix();

            if ( settings.getGapPenalty() != -1 )
                    blastCmd += " -G " + settings.getGapPenalty();
            if ( settings.getGapExtensionPenalty() != -1 )
                    blastCmd += " -E " + settings.getGapExtensionPenalty();

            //Process blastProc = Runtime.getRuntime().exec("blastall -p blastp -e 1e-10 -d /home/scott/Bioinformatics/myPDB/myPDBnew -FF -i " + fname);
            Process blastProc = Runtime.getRuntime().exec(blastCmd);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(blastProc.getErrorStream(), "ERROR");

            // any output?
          /*  FileOutputStream fos = new FileOutputStream(outputFile);
            outputReader = new
                    StreamHandler(blastProc.getInputStream(), "OUTPUT", fos);
            */
            // kick them off
            errorReader.start();
            //outputReader.start();

            // any error???
            int exitVal = blastProc.waitFor();
            System.out.println("ExitValue: " + exitVal);

            if (exitVal != 0)
            {
                System.out.println("BLAST failed");
                System.out.println(errorReader.getInputString());
            }

            /** Wait for the output reader to completely do its thing **/
            /*while ( ! outputReader.isFinished() )
            {
                System.out.println("Waiting for outputReader... (PSI_BLAST)");
                Thread.sleep(100);
            }
            */

            if ( settings.getKeepOutput() )
            {
                BufferedReader reader = new BufferedReader(new FileReader(outputFile));
                StringBuilder builder = new StringBuilder();
                String line;
                while ( (line = reader.readLine()) != null)
                {
                    builder.append(line).append('\n');
                }
                reader.close();

                output = builder.toString();
            }
        }
        catch (IOException e)
        {
            logger.error("BLAST failed - Error running blast, reason: " + e);
            logger.error(blastCmd);
            e.printStackTrace();
            File f = new File(fname);
            f.delete();
            return;
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
        catch(OutOfMemoryError e)
        {
            logger.error("Error running PSI-BLAST - not enough memory");
            throw e;
        }
        finally
        {
            File f = new File(fname);
            f.delete();            
        }

        try {
            /**
             * Create a SAX Parser that takes the native output
             * from blast-like bioinformatics software.
             */
            if ( settings.isParse() )
            {
                switch ( settings.getAlignmentView() )
                {
                    case BLUNT_ENDS:
                    {
                        gappedSeqList = new ArrayList<GappedSequence>();

                        MyBlastParser parser = new MyBlastParser();
                        parser.parseBlastResult(settings.getSequence(), outputFile, gappedSeqList);
                        break;
                    }
                    case XML:
                    default:
                    {
                        XMLReader oParser = new BlastXMLParserFacade();

                        BlastParser parser = new BlastParser();
                        parser.parseBlastResult(oParser, new FileReader(outputFile), seqList);
                        break;
                    }
                }
            }

        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();
        }
        catch (SAXException e)
        {
            e.printStackTrace();  
        }
        catch(OutOfMemoryError e)
        {
            logger.error("Out of memory parsing BLAST result");
            e.printStackTrace();
            throw e;
        }
        finally
        {
            BlastDB.add( settings.getSequence().getName(), seqList );
            finished = true;
        }
    }

    private void runBlastp()
    {
        try
        {

            // if two BLAST processes are created concurrently by multiple threads, we need a unique file name
            String runID = getUniqueIdentifier();
            String fname = runID + ".blast";
            writeBlastInputFile(fname);

            seqList = new ArrayList<Sequence>();

            StreamHandler outputReader = null;

            try
            {
                Config config = Config.getConfig();
                String blastCmdName = config.getStringValue("blastall.command");
                String blastDBName;
                if ( settings.getDatabase()!= null )
                    blastDBName = settings.getDatabase();
                else
                    blastDBName = config.getStringValue("blast.dbname");

                String evalue;
                if ( settings.getOverriddenEvalue() != -1 )
                    evalue = String.valueOf(settings.getOverriddenEvalue());
                else
                    evalue = config.getStringValue("blast.evalue");
//                String otherParams = config.getStringValue("blast.others");
                int numCPUs = Runtime.getRuntime().availableProcessors();

                String blastCmd = blastCmdName + " -p blastp -a " + numCPUs + " -e " + evalue + " -d " + blastDBName + " -FF -i " + fname;

                blastCmd += " -M " + settings.getMatrix();

                if ( settings.getGapPenalty() != -1 )
                    blastCmd += " -G " + settings.getGapPenalty();
                if ( settings.getGapExtensionPenalty() != -1 )
                    blastCmd += " -E " + settings.getGapExtensionPenalty();
                if ( settings.getMaxHits() != -1 )
                    blastCmd += " -b " + settings.getMaxHits() + " -v " + settings.getMaxHits();

                Process blastProc = Runtime.getRuntime().exec(blastCmd);

                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(blastProc.getErrorStream(), "ERROR");

                // any output?
                outputReader = new
                        StreamHandler(blastProc.getInputStream(), "OUTPUT");

                // kick them off
                errorReader.start();
                outputReader.start();

                // any error???
                int exitVal = blastProc.waitFor();
                System.out.println("Blast ExitValue: " + exitVal);

                if (exitVal != 0)
                {
                    System.out.println("BLAST failed");
                    System.out.println(errorReader.getInputString());
                }
            }
            catch (IOException e)
            {
                logger.error("BLAST failed - Error running blast, reason: " + e);
                File f = new File(fname);
                f.delete();
                return;
            }
            finally
            {
                File f = new File(fname);
                f.delete();
            }

            /**
             * Create a SAX Parser that takes the native output
             * from blast-like bioinformatics software.
             */
            XMLReader oParser = new BlastLikeSAXParser();

            ((BlastLikeSAXParser) oParser).setModeLazy();

            /**
             * Dynamically change configuration of the parser
             * in regard of Namespace support. Here,
             * the xml.org/sax/features/namespaces feature is simply "reset"
             * to its default value for SAX2.
             * The xml.org/sax/features/namespaces-prefixes feature is
             * also set to true.  This is to ensure that xmlns attributes
             * are reported by the parser. These are required because we want
             * to configure the XMLEmitter to output qualified names (see below).
             */
            try
            {
                oParser.setFeature("http://xml.org/sax/features/namespaces", true);
                oParser.setFeature("http://xml.org/sax/features/namespace-prefixes",
                        true);

            }
            catch (Exception e)
            {
                //If an illegal conmbination of features is chosen,
                //roll back to default settings. Output a warning,
                //even though this might mess up the output...
                System.out.println("WARNING: ignoring attempt to set illegal " +
                        "combination of parser features");
                System.out.println(e);
            }

           //System.out.println(outputReader.getInputString());

            BlastParser parser = new BlastParser();
            parser.parseBlastResult(oParser, new StringReader(outputReader.getInputString()), seqList);

            if ( settings.getKeepOutput() )
            {
                throw new IllegalStateException("Not implemented - need to write code for reading output file.");
                //output = outputReader.getInputString();
            }
        }
        catch (Throwable t)
        {
            t.printStackTrace();
        }
        finally
        {
            BlastDB.add( settings.getSequence().getName(), getUnalignedResult() );
            finished = true;
        }
    }

    private void runBlastPOwnParser()
    {
        try
        {

            // if two BLAST processes are created concurrently by multiple threads, we need a unique file name
            String runID = getUniqueIdentifier();
            String fname = runID + ".blast";
            writeBlastInputFile(fname);

            gappedSeqList = new ArrayList<GappedSequence>();

            StreamHandler outputReader = null;
            File outputFile = new File(runID + OUTPUT_SUFFIX);

            try
            {
                Config config = Config.getConfig();
                String blastCmdName = config.getStringValue("blastall.command");
                String blastDBName;
                if ( settings.getDatabase()!= null )
                    blastDBName = settings.getDatabase();
                else
                    blastDBName = config.getStringValue("blast.dbname");

                String evalue;
                if ( settings.getOverriddenEvalue() != -1 )
                    evalue = String.valueOf(settings.getOverriddenEvalue());
                else
                    evalue = config.getStringValue("blast.evalue");
//                String otherParams = config.getStringValue("blast.others");
                int numCPUs = Runtime.getRuntime().availableProcessors();

                String blastCmd = blastCmdName + " -m 6 -p blastp -a " + numCPUs + " -e " + evalue + " -d " + blastDBName + " -FF -i " + fname;

                blastCmd += " -M " + settings.getMatrix();

                if ( settings.getGapPenalty() != -1 )
                    blastCmd += " -G " + settings.getGapPenalty();
                if ( settings.getGapExtensionPenalty() != -1 )
                    blastCmd += " -E " + settings.getGapExtensionPenalty();
                if ( settings.getMaxHits() != -1 )
                    blastCmd += " -b " + settings.getMaxHits() + " -v " + settings.getMaxHits();

                Process blastProc = Runtime.getRuntime().exec(blastCmd);

                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(blastProc.getErrorStream(), "ERROR");

                // any output?
                FileOutputStream fos = new FileOutputStream(outputFile);
                outputReader = new
                        StreamHandler(blastProc.getInputStream(), "OUTPUT", fos);

                // kick them off
                errorReader.start();
                outputReader.start();

                // any error???
                int exitVal = blastProc.waitFor();
                System.out.println("Blast ExitValue: " + exitVal);

                if (exitVal != 0)
                {
                    System.out.println("BLAST failed");
                    System.out.println(errorReader.getInputString());
                }

                /** Wait for the output reader to completely do its thing **/
                while ( ! outputReader.isFinished() )
                {
                    System.out.println("Waiting for outputReader... (BlastP)");
                    Thread.sleep(100);
                }

                fos.close();                
            }
            catch (IOException e)
            {
                logger.error("BLAST failed - Error running blast, reason: " + e);
                File f = new File(fname);
                f.delete();
                return;
            }
            finally
            {
                File f = new File(fname);
                f.delete();
            }

            //System.out.println(outputReader.getInputString());

            MyBlastParser parser = new MyBlastParser();
            //parser.parseBlastResult(settings.getSequence(), new StringReader(outputReader.getInputString()), gappedSeqList);
            parser.parseBlastResult(settings.getSequence(), outputFile, gappedSeqList);

            if ( settings.getKeepOutput() )
            {
                throw new IllegalStateException("Not implemented - need to write code for reading output file.");
                //output = outputReader.getInputString();
            }

            outputFile.delete();
        }
        catch( OutOfMemoryError e )
        {
            e.printStackTrace();
            logger.error("BLAST can't allocate any more memory");
        }
        catch (Throwable t)
        {
            t.printStackTrace();
        }
        finally
        {
            BlastDB.add( settings.getSequence().getName(), getUnalignedResult() );
            finished = true;
        }
    }


    private void writeBlastInputFile(String fname) throws IOException
    {
        FileOutputStream blastfile;
        blastfile = new FileOutputStream(fname);
        SeqIOTools.writeFasta(blastfile, settings.getSequence() );
        blastfile.close();
    }

    public List<Sequence> getUnalignedResult()
    {
        if ( ! settings.isParse() )
            throw new IllegalStateException("Can't get a result - BlastRunner is set to not parse its result.");

        /*List<Sequence> seqList = new ArrayList<Sequence>();
        for(Sequence seq: this.seqList)
        {
            Sequence sequence = new SimpleSequence(seq.getSourceSymbolList(), "", seq.getName(), seq.getAnnotation());
            seqList.add(sequence);
        }
        */
        return seqList;
    }


    public List<GappedSequence> getAlignedResult()
    {
        if ( ! settings.isParse() )
            throw new IllegalStateException("Can't get a result - BlastRunner is set to not parse its result.");

        return gappedSeqList;
    }

    /**
     * Just runs a test
     * @param args
     */
    public static void main(String[] args) {
        SeqSeeDB db = SeqSeeDB.getInstance();
        List<Sequence> blastSequences;
        try
        {
            Sequence seq = db.getSequence(35);
            BlastSettings settings = new BlastSettings();
            settings.setSequence(seq);
            settings.setBlasttype(BlastSettings.BLASTTYPE.BLASTP);
            settings.setEvalue(0.01);
            settings.setDatabase(Config.getConfig().getStringValue("PDB"));

            BlastRunner blastRunner = new BlastRunner(settings);
            blastRunner.run();

            blastSequences = blastRunner.getUnalignedResult();

            System.out.println("BlastP results: ");
            for (Sequence sequence : blastSequences)
            {
                System.out.println(sequence.getName() + ": " + sequence.seqString());
            }

            /*settings.setBlasttype(BlastSettings.BLASTTYPE.PSIBLAST);
            settings.setDatabase(Config.getConfig().getStringValue("PDB"));
            blastRunner.run();

            blastSequences = blastRunner.getAlignedResult();

            System.out.println("PSIBLAST results: ");
            for ( Iterator<GappedSequence> iterator = blastSequences.iterator(); iterator.hasNext(); ) {
                Sequence sequence = iterator.next();
                System.out.println(sequence.getName() + ": " + sequence.seqString());
            }
            */

            ClustalWRunner runner = new ClustalWRunner(seq, blastSequences);
            runner.setAddToDB(false);
            runner.setUseCache(false);
            runner.run();
            Alignment alignment = runner.getAlignment();

            // make a list of all the aligned homologues
            for( Sequence homologSequence : blastSequences)
            {
                SymbolList homolog = alignment.symbolListForLabel(homologSequence.getName());

               // Create a gappedstructure object for the homolog
                GappedSequence gappedHomolog = ProteinTools.createGappedProteinSequence(homolog.seqString(), homologSequence.getName());

                System.out.println(gappedHomolog.seqString());
            }
        }
        catch (BioException e)
        {
            e.printStackTrace();  
        }
    }

    public void cleanupTempFiles()
    {
        for ( String s : outputFileMap.values()) {
            File f = new File(s);
            if ( ! f.delete() )
            {
                if ( f.exists() )
                {
                    System.out.print(f + " exists...");
                }
                System.out.println("BlastRunner.cleanupTempFiles(): Error deleting " + f);
            }
        }
    }

    /**
     * BLAST searches produces a lot of output data in separate files, so this method
     * allows access to those intermediate files.  These intermediate files should be deleted
     * when the VM exits
     * @return output file map
     */
    public Map<OUTPUT_FILE_TYPE, String> getOutputFileMap()
    {
        return outputFileMap;
    }
}