package ca.ualberta.proteus.util;

import ca.ualberta.proteus.program.transsec.Config;
import org.biojava.bio.program.sax.ClustalWAlignmentSAXParser;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.db.HashSequenceDB;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.seq.io.SeqIOTools;
import org.biojava.bio.symbol.SimpleAlignment;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * @author Scott Montgomerie
 * Date: Apr 9, 2005
 * Time: 6:26:10 PM
 * To change this template use File | Settings | File Templates.
 */
public class ClustalWRunner
        extends AlignmentProgramRunner
{
    public ClustalWRunner(Sequence query, List blastSequences)
    {
        this.querySequence = query;
        this.blastSequences = blastSequences;
    }

    public void run()
    {
        if ( useCache )
        {
            AlignmentDB alignmentDB = AlignmentDB.getInstance();
            if ( alignmentDB.contains(querySequence))
            {
                this.alignment = alignmentDB.get(querySequence);
                this.finished = true;
                return;
            }
        }

        this.finished = false;

        //make the SAX event adapter that will pass events to a Handler.
        ClustalWAdapter adapter = new ClustalWAdapter();

        try
        {
            SequenceDB db = new HashSequenceDB();
            for (Iterator sequences = blastSequences.iterator(); sequences.hasNext(); )
            {
                db.addSequence( (Sequence) sequences.next() );
            }

            String fname = getUniqueIdentifier();
            String infile = fname + ".db";
            FileOutputStream blastfile = new FileOutputStream(infile);
            SeqIOTools.writeFasta(blastfile, db);
            blastfile.close();

            try
            {
                System.out.println("Running ClustalW on " + blastSequences.size() + " sequences...");
                String clustalCmd = Config.getConfig().getStringValue("clustalw.command");

                //Process clustalProc = Runtime.getRuntime().exec( clustalCmd + " -infile=" + infile);
                Process clustalProc = Runtime.getRuntime().exec( clustalCmd + " " + infile);

                // any error message?
                StreamHandler errorReader = new
                        StreamHandler(clustalProc.getErrorStream(), "ERROR");

                // any output?
                StreamHandler outputReader = new
                        StreamHandler(clustalProc.getInputStream(), "OUTPUT");

                // kick them off
                errorReader.start();
                outputReader.start();

                // any error???
                int exitVal = clustalProc.waitFor();
                System.out.println("ClustalW ExitValue: " + exitVal);

                if (exitVal != 0)
                {
                    System.out.println("ClustalW failed");
                    System.out.println(errorReader.getInputString());
                    System.out.println(outputReader.getInputString());
                    cleanup(fname);
                    HashMap sequenceCollection = new HashMap();
                    sequenceCollection.put(((Sequence)blastSequences.get(0)).getName(), blastSequences.get(0) );
                    alignment = new SimpleAlignment(sequenceCollection);
                    return;
                }

                /**
                 * Create a SAX Parser that takes the native output
                 * from blast-like bioinformatics software.
                 */
                XMLReader oParser = (XMLReader) new ClustalWAlignmentSAXParser();

                /**
                 * Dynamically change configuration of the parser
                 * in regard of Namespace support. Here,
                 * the xml.org/sax/features/namespaces feature is simply "reset"
                 * to its default value for SAX2.
                 * The xml.org/sax/features/namespaces-prefixes feature is
                 * also set to true.  This is to ensure that xmlns attributes
                 * are reported by the parser. These are required because we want
                 * to configure the XMLEmitter to output qualified names (see below).
                 */
                try
                {
                    oParser.setFeature("http://xml.org/sax/features/namespaces", true);
                    oParser.setFeature("http://xml.org/sax/features/namespace-prefixes",
                                       true);

                }
                catch (Exception e)
                {
                    //If an illegal conmbination of features is chosen,
                    //roll back to default settings. Output a warning,
                    //even though this might mess up the output...
                    System.out.println("WARNING: ignoring attempt to set illegal " +
                                       "combination of parser features");
                    System.out.println(e);
                }

                //set the parsers SAX event adapter
                oParser.setContentHandler(adapter);

                // parse it
                oParser.parse(new InputSource(new FileInputStream(fname + ".aln")));

                this.alignment = adapter.getAlignment();
            }
            catch (IOException e)
            {
                System.out.println("Error running ClustalW");
                System.out.println(e);                
            }
            finally {
                // remove all intermediate files
                cleanup(fname);
            }
        }
        catch( Throwable t )
        {
            t.printStackTrace();
        }
        finally {
            if ( addToDB )
            {
                AlignmentDB alignmentDB = AlignmentDB.getInstance();
                alignmentDB.add( querySequence, this.alignment );
            }
            this.finished = true;
        }
    }

    private static void cleanup(String filename)
    {
        File f = new File(filename + ".db");
        f.delete();
        f = new File(filename + ".aln");
        f.delete();
        f = new File(filename + ".dnd");
        f.delete();
    }

}
