package ca.ualberta.proteus.util;

import org.biojava.bio.structure.ChainImpl;
import org.biojava.bio.structure.Group;
import org.biojava.bio.structure.Chain;
import org.biojava.bio.structure.AminoAcid;
import org.biojava.bio.Annotation;

import java.util.List;
import java.util.ArrayList;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: Apr 13, 2007
 * Time: 12:27:25 AM
 */
public class ProteusChainImpl implements Chain
{
    String swissprot_id;
    String name; // like in PDBfile
    ArrayList groups;
    Annotation annotation;

    /**
     * Constructs a ChainImpl object.
     */
    public ProteusChainImpl()
    {
        super();
        
        name = "";
        groups = new ArrayList();
        annotation = Annotation.EMPTY_ANNOTATION;
    }

    /**
     * Copy constructor
     *
     * @param chain
     */
    public ProteusChainImpl(Chain chain)
    {
        this();
        copyFrom(chain);
    }

    private void copyFrom(Chain chain)
    {
        setName(chain.getName());
        setSwissprotId(chain.getSwissprotId());
        setAnnotation(chain.getAnnotation());
        
        for (int i = 0; i < chain.getGroups().size(); i++)
        {
            Group g = (Group) chain.getGroups().get(i);
            addGroup((Group) g.clone());
        }
    }


    /**
     * returns an identical copy of this Chain .
     *
     * @return an identical copy of this Chain
     */
    public Object clone()
    {
        // go through all groups and add to new Chain.
        Chain n = new ChainImpl();

        n.setName(getName());
        n.setSwissprotId(getSwissprotId());
        n.setAnnotation(getAnnotation());
        
        for (int i = 0; i < groups.size(); i++)
        {
            Group g = (Group) groups.get(i);
            n.addGroup((Group) g.clone());
        }

        return n;
    }



    public void setAnnotation(Annotation anno)
    {
        annotation = anno;
    }

    public Annotation getAnnotation()
    {
        return annotation;
    }

    /**
     * set the Swissprot id of this chains .
     *
     * @param sp_id a String specifying the swissprot id value
     * @see #getSwissprotId
     */

    public void setSwissprotId(String sp_id)
    {
        swissprot_id = sp_id;
    }

    /**
     * get the Swissprot id of this chains .
     *
     * @return a String representing the swissprot id value
     * @see #setSwissprotId
     */
    public String getSwissprotId()
    {
        return swissprot_id;
    }


    public void addGroup(Group group)
    {
        // TODO Auto-generated method stub

        groups.add(group);
    }

    /**
     * return the amino acid at position .
     *
     * @param position an int
     * @return a Group object
     */
    public Group getGroup(int position)
    {
        // TODO Auto-generated method stub
        return (Group) groups.get(position);
    }

    /**
     * return an array of all groups of a special type (e.g. amino,
     * hetatm, nucleotide).
     *
     * @param type a String
     * @return an List object containing the groups of type...
     */
    public ArrayList getGroups(String type)
    {
        ArrayList tmp = new ArrayList();
        for (int i = 0; i < groups.size(); i++)
        {
            Group g = (Group) groups.get(i);
            if (g.getType().equals(type))
            {
                tmp.add(g);
            }
        }
        //Group[] g = (Group[])tmp.toArray(new Group[tmp.size()]);
        return tmp;
    }

    /**
     * return all groups of this chain .
     *
     * @return an ArrayList object representing the Groups of this Chain.
     */
    public ArrayList getGroups()
    {
        return groups;
    }


    public int getLength()
    {
        return groups.size();
    }

    public int getLengthAminos()
    {

        ArrayList g = getGroups("amino");
        return g.size();
    }


    /**
     * get and set the name of this chain (Chain id in PDB file ).
     *
     * @param nam a String specifying the name value
     * @see #getName
     */

    public void setName(String nam)
    {
        name = nam;
    }

    /**
     * get and set the name of this chain (Chain id in PDB file ).
     *
     * @return a String representing the name value
     * @see #setName
     */
    public String getName()
    {
        return name;
    }


    /**
     * string representation.
     */
    public String toString()
    {

        String str = "Chain >" + getName() + "< total length:" + getLength() + " residues";
        // loop over the residues
        int i = 0;
        do
        {
            Group gr = (Group) groups.get(i);
            str += gr.toString() + "\n";
            i++;
        }
        while (i < groups.size());


        return str;

    }

    /**
     * get amino acid sequence.
     *
     * @return a String representing the sequence.
     */
    public String getSequence()
    {

        List aminos = getGroups("amino");
        StringBuffer sequence = new StringBuffer();
        for (int i = 0; i < aminos.size(); i++)
        {
            AminoAcid a = (AminoAcid) aminos.get(i);
            sequence.append(a.getAminoType());
        }

        String s = sequence.toString();

        return s;
    }

    public ProteusChainImpl subChain(int start, int end)
    {
        // go through all groups and add to new Chain.
        ProteusChainImpl n = new ProteusChainImpl();

        n.setName(getName());
        n.setSwissprotId(getSwissprotId());
        n.setAnnotation(getAnnotation());

        for (int i = start; i < Math.min(groups.size(), end); i++)
        {
            Group g = (Group) groups.get(i);
            n.addGroup((Group) g.clone());
        }

        return n;
    }
}
