package ca.ualberta.proteus.util;

import ca.ualberta.proteus.biojava.PDBFileParser;
import org.biojava.bio.structure.*;

import java.io.*;
import java.text.DecimalFormat;
import java.util.ArrayList;

/*
 *                    BioJava development code
 *
 * This code may be freely distributed and modified under the
 * terms of the GNU Lesser General Public Licence.  This should
 * be distributed with the code.  If you do not have a copy,
 * see:
 *
 *      http://www.gnu.org/copyleft/lesser.html
 *
 * Copyright for this code is held jointly by the individual
 * authors.  These should be listed in @author doc comments.
 *
 * For more information on the BioJava project and its aims,
 * or to join the biojava-l mailing list, visit the home page
 * at:
 *
 *      http://www.biojava.org/
 *
 * Created on 26.04.2004
 * @author Andreas Prlic
 *
 */

/**
 * Implementation of a PDBStructure. This class
 * provides the data contained in a PDB file.
 * Modified to improve algorithm
 */
public class ProteusStructureImpl extends StructureImpl implements Serializable
{
    // for serialization only
    private String pdb = null;

    public ProteusStructureImpl()
    {        
    }

    /**
     * Copy constructor
     * 
     * @param structure
     */
    public ProteusStructureImpl(Structure structure)
    {
        fromStructure(structure, false);
    }

    /**
     * Copy constructor
     *
     * @param oneModel True if you only want to clone the first model
     * */
    public ProteusStructureImpl(Structure structure, boolean oneModel)
    {
        fromStructure(structure, oneModel);
    }

    private void fromStructure(Structure structure, boolean oneModel)
    {
        // copy structure data
        if (structure.isNmr())
            this.setNmr(true);

        this.setPDBCode(structure.getPDBCode());
        this.setName(structure.getName());
        this.setHeader(structure.getHeader());
        this.setConnections(structure.getConnections());

        // go through each chain and clone chain
        for (int i = 0; i < structure.nrModels(); i++)
        {
            ArrayList<Chain> cloned_model = new ArrayList<Chain>();

            for (int j = 0; j < structure.size(i); j++)
            {
                //System.out.println("getting chain "+i+" "+j);
                Chain current_chain = structure.getChain(i, j);
                Chain cloned_chain = (Chain) current_chain.clone();

                cloned_model.add(cloned_chain);
            }
            this.addModel(cloned_model);

            if ( oneModel)
            {
                break;
            }
        }
    }


    /**
     * create a String that contains the contents of a PDB file
     */
    public String toPDB()
    {
        GroupIterator iter = new GroupIterator(this);

        StringBuilder str = new StringBuilder();
        int i = 0;

        DecimalFormat d3 = new DecimalFormat("0.000");
        DecimalFormat d2 = new DecimalFormat("0.00");

        while (iter.hasNext())
        {
            Group g = (Group) iter.next();
            String type = g.getType();
            String record = "";
            if (type.equals("hetatm"))
            {
                record = "HETATM";
            }
            else
            {
                record = "ATOM  ";
            }

            // format output ...
            AtomIterator aiter = new AtomIterator(g);
            String line = "";

            while (aiter.hasNext())
            {
                i++;
                //if ( i > 40 ) continue ;
                Atom a = (Atom) aiter.next();

                int seri = a.getPDBserial();
                String serial = alignRight("" + seri, 5);
                String fullname = a.getFullName();
                String altLoc = " "; // not supported, yet!
                String chainID = " ";
                String resseq = alignRight("" + g.getPDBCode(), 4);
                String resName = g.getPDBName();
                String x = alignRight("" + d3.format(a.getX()), 8);
                String y = alignRight("" + d3.format(a.getY()), 8);
                String z = alignRight("" + d3.format(a.getZ()), 8);
                String occupancy = alignRight("" + d2.format(a.getOccupancy()), 6);
                String tempfactor = alignRight("" + d2.format(a.getTempFactor()), 6);

                line = record + serial + " " + fullname + altLoc
                        + resName + " " + chainID + resseq
                        + "    " + x + y + z
                        + occupancy + tempfactor;
                str.append(line + "\n");
            }
        }
        return str.toString();
    }

    private String alignRight(String input, int length)
    {
        String str = "";
        int diff = length - input.length();

        for (int i = length; i > 0; i--)
        {
            if (i <= diff)
            {
                str = " " + str;
            }
            else
            {
                int pos = input.length() - length + i - 1;
                //System.out.println(input + " " +pos + " " + input.charAt(pos) );
                str = input.charAt(pos) + str;
            }
        }
        return str;
    }

    private void writeObject(ObjectOutputStream out) throws IOException
    {
        this.pdb = this.toPDB();

        out.defaultWriteObject();
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        in.defaultReadObject();

        PDBFileParser parser = new PDBFileParser();
        StringReader reader = new StringReader(this.pdb);
        Structure structure = parser.parsePDBFile(new BufferedReader(reader));

        this.fromStructure(structure, false);
    }
}
