package ca.ualberta.proteus.util;

import ca.ualberta.proteus.program.transsec.Config;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.Alignment;
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Jun 14, 2005
 * Time: 10:07:31 PM
 * To change this template use File | Settings | File Templates.
 */
public class XAlignDB
{
    private static final String DB_FILE = "xalign.db.gz";
    private static final String DB_NAME = "XAlign results";
    private static Map<String, String> alignmentList = new HashMap();
    private static XAlignDB db = null;
    private static AlignmentDB alignmentDB;

    public static XAlignDB getInstance()
    {
        if (db == null)
        {
            db = loadDB();
        }
        return db;
    }

    /**
     * Reads in the blast database file
     *
     * @return a new filled XALign DB 
     */
    private synchronized static XAlignDB loadDB()
    {

        XAlignDB rv = new XAlignDB();
        InputStream in = rv.getClass().getResourceAsStream(DB_FILE);

        if ( in == null )
            in = rv.getClass().getResourceAsStream("/" + DB_FILE);

        if ( in == null ) {
            try
            {
                in = new FileInputStream(DB_FILE);
            }
            catch (FileNotFoundException e)
            {
            }
        }
        if ( in == null )
        {
            System.out.println("XAlign database does not exist... creating a new one.");
            return rv;
        }

        System.out.print("Loading XAlign DB...");
        try
        {
            synchronized (rv.alignmentList)
            {
                SymbolTokenization tokenization = ProteinTools.getAlphabet().getTokenization("token");
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                DocumentBuilder builder = null;
                try
                {
                    builder = factory.newDocumentBuilder();
                    GZIPInputStream gis = new GZIPInputStream(in);
                    Document doc = builder.parse(gis);

                    Element root = doc.getDocumentElement();

                    for (int i = 0; i < root.getChildNodes().getLength(); i++)
                    {
                        Node node = root.getChildNodes().item(i);

                        if (node.getNodeName().equals("record"))
                        {
                            String id = node.getAttributes().getNamedItem("ID").getNodeValue();
                            alignmentList.put(id, node.getTextContent());
                        }
                    }
                }
                catch (ParserConfigurationException e)
                {
                    e.printStackTrace();  //TODO
                }
                catch (SAXException e)
                {
                    e.printStackTrace();  //TODO
                }

            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (BioException e)
        {
            e.printStackTrace();  //TODO
        }
        System.out.println("Done");
        return rv;
    }

    public boolean contains(Sequence sequence)
    {
        for (Iterator it = alignmentList.keySet().iterator(); it.hasNext();)
        {
            String seqID = (String) it.next();
            if (seqID.equals(sequence.getName()))
                return true;
        }
        return false;
    }

    public Alignment get(Sequence sequence)
    {
        for (Iterator it = alignmentList.keySet().iterator(); it.hasNext();)
        {
            String seqID = (String) it.next();
            if (seqID.equals(sequence.getName()))
            {
                String s = alignmentList.get(seqID);
                XAlignParser parser = new XAlignParser();
                try
                {
                    parser.parse(new InputSource(new StringReader(s)));
                    Alignment alignment = parser.getAlignment();

                    // lazy initialization
                    if ( alignmentDB == null )
                    {
                        alignmentDB = AlignmentDB.getInstance();
                    }

                    alignmentDB.add(sequence, alignment);
                    return alignment;
                }
                catch (IOException e)
                {
                    e.printStackTrace();  //TODO
                }
                catch (SAXException e)
                {
                    e.printStackTrace();  //TODO
                }
            }
        }
        return null;
    }

    /**
     * Writes out the contents of the database file
     */
    private synchronized void writeDB()
    {
        if ( ! Config.getConfig().getBooleanValue("storeResults") ) {
            if ( Config.getConfig().getBooleanValue("testing")  )
                System.out.print("Not writing XAlign DB... change \"storeResults\" to true in config file");
            return;
        }

        System.err.print("Writing XAlign DB...");
        try
        {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            DOMImplementation impl = builder.getDOMImplementation();

            Document doc = impl.createDocument(null, "xalign_database", null);

            Element root = doc.getDocumentElement();
            Date today = new Date();
            Attr date = doc.createAttribute("updated");
            date.setValue(today.toString());
            root.getAttributes().setNamedItem(date);

            synchronized (alignmentList)
            {

                for (Iterator it = alignmentList.keySet().iterator(); it.hasNext();)
                {
                    String seqID = (String) it.next();
                    String record = alignmentList.get(seqID);

                    Element recordElement = doc.createElement("record");
                    recordElement.setAttribute("ID", seqID);
                    recordElement.setTextContent(record);
                    root.appendChild(recordElement);
                }

                // Prepare the DOM document for writing, and the output file
                Source source = new DOMSource(doc);
                GZIPOutputStream gos = new GZIPOutputStream(new FileOutputStream(DB_FILE));
                Result result = new StreamResult(gos);

                // Write the DOM document to file, using a document transformer
                Transformer transformer = TransformerFactory.newInstance().newTransformer();
                transformer.transform(source, result);
                gos.close();
                System.err.println("Done");
            }
        }
        catch (ParserConfigurationException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (TransformerException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (FileNotFoundException e)
        {
            e.printStackTrace();  //TODO
        }
        catch (IOException e)
        {
            e.printStackTrace();  //TODO
        }

    }

    /**
     * Add a sequence to the database
     *
     * @param querySequence - for the query sequence
     * @param inputStream   - for the query sequence
     */
    public synchronized void add(Sequence querySequence, InputStream inputStream)
    {
        StringBuilder builder = new StringBuilder(1024);
        BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));
        String line;
        try
        {
            while ((line = reader.readLine()) != null)
                builder.append(line + "\n");
        }
        catch (IOException e)
        {
            e.printStackTrace();  //TODO
        }
        alignmentList.put(querySequence.getName(), builder.toString());
        writeDB();
    }

    public void clearDB()
    {
        alignmentList.clear();
        db = null;
    }

    public Map<String, Map<String, Double>> getPSSHMap(Sequence sequence)
    {
        for (Iterator it = alignmentList.keySet().iterator(); it.hasNext();)
        {
            String seqID = (String) it.next();
            if (seqID.equals(sequence.getName()))
            {
                String s = alignmentList.get(seqID);
                XAlignParser parser = new XAlignParser();
                try
                {
                    parser.parse(new InputSource(new StringReader(s)));
                    return parser.getPPSHMap();
                }
                catch (IOException e)
                {
                    e.printStackTrace();  //TODO
                }
                catch (SAXException e)
                {
                    e.printStackTrace();  //TODO
                }
            }
        }
        return null;
    }
}
