package ca.ualberta.proteus.util;

import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SimpleAlignment;
import org.biojava.bio.symbol.SymbolList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Jun 11, 2005
 * Time: 7:23:25 PM
 * To change this template use File | Settings | File Templates.
 */
public class XAlignParser
{
    private AttributesImpl oAtts = new AttributesImpl();

    private HashMap<String, StringBuffer> alignmentStrings = new HashMap();
    private HashMap<String, Sequence> alignment = new HashMap();
    private ArrayList<String> seqNameList = new ArrayList();
    private Map<String, Map<String, Double>> ppsmMatrix = new HashMap();

    private State state = State.STARTUP;

    public Map<String,Map<String,Double>> getPPSHMap()
    {
        return ppsmMatrix;
    }

    enum State
    {
        STARTUP, SEQUENCES, PPSH, STATS, MULTI_ALIGNMENT
    };

    /**
     * Initialises internal state
     * Sets namespace prefix to "biojava"
     */
    public XAlignParser()
    {
        state = State.STARTUP;
        //this.setNamespacePrefix("biojava");
    }

    /**
     * Only interested in lines that are part of the alignmentStrings.
     * Returns true if line is in alignmentStrings, false if not.
     *
     * @param poLine a <code>String</code> value
     * @return a <code>boolean</code> value
     */
    private boolean lineIsRelevant(String poLine)
    {

        //blank lines not relevant
        //lines that starts with a space  not relevant (consensus line)
        //lines that start with "CLUSTAL W (" not relevant (title)

        if ((poLine.trim().equals("")) ||
                (poLine.startsWith("***")) ||
                (poLine.contains("Acc#")) ||
                poLine.contains("Pairwise") ||
                poLine.contains("Statistics") ||
                poLine.contains("Multiple") ||
                poLine.contains("Summary")
        )
        {

            //System.out.println("Irrelevant|"+poLine+"|");
            return false;
        }

        //if here,line is part of alignmentStrings, so return true
        return true;
    }

    /**
     * Create a stream from an an InputSource, picking the
     * correct stream according to order of precedance.
     *
     * @param poSource an <code>InputSource</code> value
     * @return a <code>BufferedReader</code> value
     */
    protected BufferedReader getContentStream(InputSource poSource)
    {


        InputSource oSource;
        BufferedReader oContents;
        int iBuffSize = 8192;


        oSource = poSource;

        //Check contents InputSource in order of precedence

        //Highest - Character stream

        if (oSource.getCharacterStream() != null)
        {

            oContents = new BufferedReader(oSource.getCharacterStream(),
                                           iBuffSize);
            return oContents;
        }

        //Next to lowest -  Byte stream


        if ((oSource.getByteStream() != null))
        {

            oContents = new BufferedReader(new InputStreamReader(oSource.getByteStream()),
                                           iBuffSize);

            return oContents;
        }

        //Lowest precedence - System URI

        if ((oSource.getSystemId() != null))
        {

            try
            {
                //URL handles file URI's and URL's, which might be sent in
                //by an XSLT processor.  Handles things like file:\C:\file.txt
                URL IORUrl = new java.net.URL(oSource.getSystemId());
                BufferedInputStream inStr =
                        new BufferedInputStream(IORUrl.openStream());
                // Create input stream
                oContents = new
                        BufferedReader(new InputStreamReader(inStr),
                                       iBuffSize);

                return oContents;
                //catch Malfromed URL and IOException
            }
            catch (MalformedURLException x)
            {
                System.out.println(x.getMessage());
                System.out.println("Couldn't open file");
                System.exit(0);
            }
            catch (IOException x)
            {
                System.out.println(x.getMessage());
                System.out.println("Couldn't open file");
                System.exit(0);
            }
        }

        //should throw an exception to say couldn't get stream
        return null;
    }

    /**
     * Parses an XAlign output file
     */
    public void parse(InputSource poSource)
            throws IOException, SAXException
    {

        BufferedReader oContents;
        String oLine = null;

        //Use method form superclass
        oContents = this.getContentStream(poSource);

        // loop over file
        try
        {
            // loop over file
            oLine = oContents.readLine();
            while (oLine != null)
            {
                //System.out.println(oLine);
                this.interpret(oContents, oLine);
                oLine = oContents.readLine();
            } // end while
        }
        catch (java.io.IOException x)
        {
            System.out.println(x.getMessage());
            System.out.println("Stream read interrupted");
        } // end try/catch

        //at end of stream...

        //at this point, alignmentStrings is parsed, now cycle through
        //and emit elements

        // find the length of the shortest alignment
        int length = 0;
        for ( String seqName : seqNameList )
        {
            String seq = alignmentStrings.get(seqName).toString();
            if ( trimRight( seq ).length() > length)
            {
                length = trimRight( seq ).length();
            }
        }

        for (String seqName : seqNameList)
        {
            String seq = alignmentStrings.get(seqName).toString();
            seq = seq.substring(0, length);

            seq = seq.replace(' ', '-');
            seq = seq.replace('B', 'D'); // for some reason, B is getting in there
            // B represents Asparagine or Aspartic acid, so go with Aspartic acid
            seq = seq.replace('Z', 'E');
            // Z represents Glutamine or Glutamic Acid, choose Glutamic Acid

            GappedSequence sequence = null;
            try
            {
                sequence = ProteinTools.createGappedProteinSequence(seq, seqName);
            }
            catch (IllegalSymbolException e)
            {
                e.printStackTrace();
            }

            alignment.put(seqName, sequence);
        }

        oContents.close();

    }

    /**
     * Describe <code>interpret</code> method here.
     *
     * @param poContents a <code>BufferedReader</code> value
     * @param poLine     a <code>String</code> value
     */
    private void interpret(BufferedReader poContents, String poLine)
    {

        switch (state)
        {
            case STARTUP:
                {
                    oAtts.clear();
                    this.seqNameList.clear();
                    this.alignmentStrings.clear();


                    if (poLine.equals("Sequences"))
                        this.changeState(State.SEQUENCES);
                    break;
                }
            case SEQUENCES:
                {
                    if (this.lineIsRelevant(poLine))
                    {
                        // add this sequence name to the list
                        String[] s = poLine.split("\\s+");
                        this.seqNameList.add(s[2]);
                    }

                    if (poLine.contains("Pairwise"))
                        this.changeState(State.PPSH);

                    break;
                }
            case PPSH:
                {
                    if (this.lineIsRelevant(poLine))
                    {
                        // add this sequence name to the list
                        String[] s = this.splitPPSH(poLine);
                        if ( s[0].trim().length() != 0 ) {
                            // this isn't the first line
                            for (int i=1; i<s.length; i++) {
                                this.ppsmMatrix.get(s[0]).put( this.seqNameList.get(i-1), Double.valueOf(s[i]) );
                            }
                        }
                    }

                    if (poLine.contains("Statistics"))
                        this.changeState(State.STATS);

                    break;
                }
            case STATS:
                if (poLine.contains("Multiple Alignment"))
                    this.changeState(State.MULTI_ALIGNMENT);
                break;
            case MULTI_ALIGNMENT:
                {
                    if (this.lineIsRelevant(poLine))
                    {
                        //build aligment in memory
                        this.appendToAlignment(poLine);
                    }
                }

        }
    }

    private String[] splitPPSH(String poLine)
    {
        try {
            String[] result = new String[ poLine.length() / 9];
            result[0] = poLine.substring(0, 10).trim();
            result[1] = poLine.substring(10, 20).trim();
            for ( int i=19; i< poLine.length(); i+=9 ) {
                result[i / 9] = poLine.substring(i, i+9).trim();
            }
            return result;
        }
        catch (ArrayIndexOutOfBoundsException e) {
            System.err.println("Error parsing Pairwise Percent Sequence Homology");
            e.printStackTrace();
        }
        return null;
    }

    private void changeState(State newState)
    {
        this.state = newState;

        if ( newState == State.PPSH) {
            // load the matrix with new hashmaps
            for( String s : seqNameList) {
                ppsmMatrix.put(s, new HashMap());
            }
        }
    }

    /**
     * Parse a relevant line, and add to alignmentStrings
     *
     * @param poLine a <code>String</code> value
     */
    private void appendToAlignment(String poLine)
    {


        //First token is sequence name
        String seqName = poLine.substring(0, 9).trim();

        if ( ! seqName.equals("Struct"))
        {
            String seq = poLine.substring(11, 61);

            //Update object...

            if (alignmentStrings.get(seqName) == null)
            {
                //Here if on first occurence of this sequence
                //Add to alignmentStrings
                alignmentStrings.put(seqName, new StringBuffer(seq));
            }
            else
            {
                //Here if building up an existing sequence
                StringBuffer tmpSeq = (StringBuffer) alignmentStrings.get(seqName);
                alignmentStrings.put(seqName, tmpSeq.append(seq));
            }
        }
    }

    public static void main(String[] args)
    {
        XAlignParser oParser = new XAlignParser();

        // parse it
        String in;
        if ( args.length < 1 )
            in ="test/xaligntest";
        else
            in = args[0];
        try
        {
            oParser.parse(new InputSource(new FileInputStream(in)));
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        catch (SAXException e)
        {
            e.printStackTrace();
        }


        for ( Iterator iterator = oParser.getAlignment().symbolListIterator(); iterator.hasNext(); )
            System.out.println( ((SymbolList) iterator.next()).seqString() );
        System.out.println( oParser.getPPSH("QUERY", "QUERY") );
    }

    /**
     * Get the alignment that Xalign produced
     * @return
     */
    public Alignment getAlignment() {
        return new SimpleAlignment(alignment);
    }

    /**
     * Gets the percent sequence homology of two sequences
     * @param seq1 - ID of sequence 1
     * @param seq2 - ID of sequence 2
     * @throws IllegalArgumentException if the sequence ID's are not in the sequence list
     * @return
     */
    public double getPPSH(String seq1, String seq2) {
        if ( !ppsmMatrix.containsKey(seq1) || !ppsmMatrix.containsKey(seq1) )
            throw new IllegalArgumentException("Matrix does not contain " + seq1 + " or " + seq2);

        return ppsmMatrix.get(seq1).get(seq2);
    }

    private static String trimRight(String toTrim)
	{
		if( 0 >= toTrim.length())
		{
			return toTrim;
		}

		int i;
		for(i=toTrim.length(); i > 0; --i)
		{
			if( !Character.isWhitespace(toTrim.charAt(i-1)) )
			{
				break;
			}
		}

		return toTrim.substring(0, i);
	}
}
