package ca.ualberta.proteus.util;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.impl.SimpleSequence;
import org.biojava.bio.symbol.Alignment;
import org.biojava.bio.symbol.SimpleAlignment;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.xml.sax.InputSource;

import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Created by IntelliJ IDEA.
 *
 * @author Scott Montgomerie
 *         Date: Apr 9, 2005
 *         Time: 6:26:10 PM
 *         To change this template use File | Settings | File Templates.
 */
public class XAlignRunner
        extends AlignmentProgramRunner
{
    private boolean keepOutput = false;
    private FileInputStream output = null;
    private Map<String, Map<String, Double>> ppshMap;

    public XAlignRunner(Sequence query, List blastSequences)
    {
        assert blastSequences.size() > 0 : "Can't pass in blast sequences < 1";

        this.querySequence = query;
        this.blastSequences = blastSequences;
    }

    public Alignment getAlignment()
    {
        if (alignment == null)
        {
            // TODO - maybe make it recover by waiting for it to finish?
            throw new ProcessNotCompleteException("XAlign has not finished running yet or there was an error assigning the alignment!");
        }
        return alignment;
    }

    public void run()
    {
        run(false);
    }

    public void run(boolean isRecursiveCall)
    {
        XAlignDB xAlignDB = null;
        if (useCache)
        {
            xAlignDB = XAlignDB.getInstance();
            if (useCache && xAlignDB.contains(querySequence))
            {
                this.alignment = xAlignDB.get(querySequence);
                this.ppshMap = xAlignDB.getPSSHMap(querySequence);
                this.finished = true;
                return;
            }
        }

        this.finished = false;

        try
        {
            String fname = getUniqueIdentifier();
            String infile = fname + ".db";

            writeInputFile(infile);

            String outfile = fname + ".out";

            System.out.println("Running XAlign on " + blastSequences.size() + " sequences...");

            String xalignCmd = Config.getConfig().getStringValue("xalign.command");
            File xalignProg = new File(xalignCmd);
            assert xalignProg.exists() : "XAlign file specified does not exist";

            File dir = xalignProg.getParentFile().getAbsoluteFile();

            xalignProg = new File("");
            String curdir = xalignProg.getAbsolutePath() + "/";

            //xalignCmd =  xalignCmd + " -d " + infile + " -o " + outfile;
            //*** this has to be run in XAlign's directory!!
            //xalignCmd += " -d " + infile + " -o " + outfile;
            infile = curdir + infile;
            outfile = curdir + outfile;

            String[] cmdArgs = new String[]{"perl", xalignCmd, "-d", infile, "-o", outfile, "> tmp.out"};

            Process xalignProc = Runtime.getRuntime().exec(cmdArgs, null, dir);

            // any error message?
            StreamHandler errorReader = new
                    StreamHandler(xalignProc.getErrorStream(), "ERROR");

            // any output?
            StreamHandler outputReader = new
                    StreamHandler(xalignProc.getInputStream(), "OUTPUT");

            // kick them off
            errorReader.start();
            outputReader.start();

            // This is a sequece of commands that has to be sent to XAlign's stdin
            String xalignCommands = "1\n1\n1\n";
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(xalignProc.getOutputStream()));
            writer.write("1\n");
            writer.write("1\n");
            writer.write("1\n");
            writer.write("1\n\n\n");
            //xalignProc.getOutputStream().write(xalignCommands.getBytes());

            try
            {
                // any error???
                int exitVal = xalignProc.waitFor();
                System.out.println("XAlign ExitValue: " + exitVal);

                if (exitVal != 0)
                {
                    System.out.println("XAlign failed");
                    System.out.println("Cmd: " + xalignCmd + "\n Dir: " + dir);
                    System.out.println(errorReader.getInputString());
                    System.out.println(outputReader.getInputString());
                    cleanup(fname);

                    if ( !isRecursiveCall)
                    {
                        System.out.println("Running XAlign Again...");
                        run(true);
                    }

                    HashMap<String, SymbolList> sequenceCollection = new HashMap<String, SymbolList>();
                    sequenceCollection.put(blastSequences.get(0).getName(), blastSequences.get(0));
                    alignment = new SimpleAlignment(sequenceCollection);

                    return;
                }

                XAlignParser oParser = new XAlignParser();

                try
                {
                    // parse it

                    output = new FileInputStream(outfile);
                    oParser.parse(new InputSource(output));

                    this.alignment = oParser.getAlignment();

                    addSubjectDescriptionsToSequences();
                }
                catch(Exception e)
                {
                    handleError(xalignCmd, dir, isRecursiveCall);
                    e.printStackTrace();
                }

                // add it to the database
                if (addToDB)
                {
                    output = new FileInputStream(outfile);
                    AlignmentDB.getInstance().add(querySequence, alignment);
                    if (xAlignDB == null)
                        xAlignDB = XAlignDB.getInstance();
                    xAlignDB.add(querySequence, output);
                }

                if (! keepOutput)
                    output = null;

                // remove all intermediate files
                cleanup(fname);

                this.ppshMap = oParser.getPPSHMap();

            }
            catch (IOException e)
            {
                handleError(xalignCmd, dir, isRecursiveCall);
                e.printStackTrace();
            }
        }
        catch (Throwable t)
        {
            if ( !isRecursiveCall)
            {
                System.out.println("Running XAlign Again...");
                run(true);
            }

            t.printStackTrace();
        }
        finally
        {
            this.finished = true;
        }
    }

    private void handleError(String xalignCmd, File dir, boolean isRecursiveCall)
            throws IOException
    {
        System.out.println("Error running XAlign");
        System.out.println("Cmd: " + xalignCmd);

        // print out log file
        BufferedReader r = new BufferedReader(new FileReader(dir.getAbsolutePath() + "/xalign.log"));
        String line;
        while ((line = r.readLine()) != null)
        {
            System.out.println(line);
        }
        //System.exit(-1);

        if ( !isRecursiveCall)
        {
            System.out.println("Running XAlign Again...");
            run(true);
        }
    }

    private void addSubjectDescriptionsToSequences()
    {
        try
        {
            for (Iterator it = alignment.symbolListIterator(); it.hasNext();)
            {
                Sequence seq = (Sequence) it.next();
                for (Sequence blastSequence : blastSequences)
                {
                    if (blastSequence.getName().equals(seq.getName()))
                    {
                        if (blastSequence.getAnnotation().containsProperty("Score"))
                            seq.getAnnotation().setProperty("Score", blastSequence.getAnnotation().getProperty("Score"));
                        if (blastSequence.getAnnotation().containsProperty("e-value"))
                            seq.getAnnotation().setProperty("e-value", blastSequence.getAnnotation().getProperty("e-value"));
                        if (blastSequence.getAnnotation().containsProperty("subjectDescription"))
                            seq.getAnnotation().setProperty("subjectDescription", blastSequence.getAnnotation().getProperty("subjectDescription"));
                    }
                }

            }
        }
        catch (ChangeVetoException e)
        {

        }
    }

    private void writeInputFile(String infile)
            throws IOException
    {
        if (blastSequences.size() < 2)
        {
            // XAlign needs 2 sequences to align
            Sequence o = new SimpleSequence(blastSequences.get(0), "", "DUPLI", null);
            blastSequences.add(o);
        }

        // make sure we have a few uniques
        boolean allowDuplicates = false;
        for (Iterator<Sequence> sequences = blastSequences.iterator(); sequences.hasNext();)
        {
            Sequence seq = sequences.next();
            if (! seq.getName().equals(blastSequences.get(0).getName()))
            {
                allowDuplicates = true;
                break;
            }
        }

        if (!allowDuplicates)
        {
            // XAlign needs 2 sequences to align
            Sequence o = new SimpleSequence(blastSequences.get(0), "", "DUPLI", null);
            blastSequences.add(o);
        }

        // write the XAlign input file
        File f = new File(infile);
        f.deleteOnExit();

        PrintWriter pw = new PrintWriter(new FileWriter(f));
        HashMap<String, String> seqNames = new HashMap<String, String>();
        for (Sequence seq : blastSequences)
        {
            // keep track of names to avoid duplicate entries
            if (seqNames.containsKey(seq.getName()))
            {
                //System.err.println("Duplicate entry!");
            }
            else
            {
                seqNames.put(seq.getName(), null);

                // name can only be 7 characters long
                String name = seq.getName();
                if (name.length() > 7)
                {
                    name = name.substring(0, 6);
                }
                pw.println(">" + name);

                String str = seq.seqString();
                // split it up to appease XAlign
                if (str.length() > 65)
                {
                    // don't want single characters on a line, it might contain only C's or E's
                    int lineBreak = (int) (58 + Math.round(Math.random() * 5));
                    // try to use a random breakup so at least one of the two attempts will work
                    pw.println(str.substring(0, lineBreak));
                    pw.println(str.substring(lineBreak));
                }
                else if (str.length() > 15)
                {
                    int lineBreak = (int) (10 + Math.round(Math.random() * 5));
                    // try to use a random breakup so at least one of the two attempts will work
                    pw.println(str.substring(0, str.length() - lineBreak));
                    pw.println(str.substring(str.length() - lineBreak));
                }
                else
                {
                    pw.println(str);
                }
                pw.println();
            }
        }
        pw.close();
    }

    private void cleanup(String filename)
    {
        File f = new File(filename + ".db");
        f.delete();

        f = new File(filename + ".out");
        if (keepOutput)
        {
            // keep it till the VM exits
            f.deleteOnExit();
        }
        else
        {
            f.delete();
        }
    }

    /**
     * The main method will test the xalign's parsing capabilities
     *
     * @param args
     */
    public static void main(String[] args)
    {

        Sequence sequence = null;
        //sequence = (Sequence) SeqSeeDB.getInstance().getRandomSequences(1).get(0);
        try
        {
            sequence = SeqSeeDB.getInstance().getSequence("1EF5A");
        }
        catch (BioException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        List<Sequence> blastSequences;
        BlastSettings settings = new BlastSettings();
        settings.setSequence(sequence);
        BlastRunner blastRunner = new BlastRunner(settings);
        settings.setBlastType(BlastSettings.BLASTTYPE.BLASTP);
        settings.setEvalue(0.001);
        blastRunner.run();

        /// wait for the thread to complete
        while (!blastRunner.isFinished())
        {
            try
            {
                Thread.sleep(10);
            }
            catch (InterruptedException e)
            {
            }
        }

        blastSequences = blastRunner.getUnalignedResult();

        try
        {
            assert sequence != null;
            sequence.getAnnotation().setProperty("description_line", "QUERY");
        }
        catch (ChangeVetoException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
        blastSequences.add(sequence);

        Alignment alignment = null;

        XAlignRunner runner = new XAlignRunner(sequence, blastSequences);
        runner.setUseCache(false);
        runner.run();

        alignment = runner.getAlignment();

        for (Iterator it = alignment.symbolListIterator(); it.hasNext();)
        {
            Sequence seq = (Sequence) it.next();
            System.out.println(seq.getName() + ": " + seq.seqString());
        }

    }

    public String getOutput() throws IOException
    {
        /*CharBuffer cb = CharBuffer.allocate(20000);
        BufferedReader reader = new BufferedReader(new InputStreamReader(output));
        while (reader.read(cb) != 0)
            ;
        return cb.toString();*/

        StringBuilder sb = new StringBuilder();
        if ( output.getFD().valid() )
        {
            BufferedReader reader = new BufferedReader(new InputStreamReader(output));
            String line;
            while ((line = reader.readLine()) != null)
            {
                sb.append(line);
                sb.append("\n");
            }
            return sb.toString();
        }
        else
            throw new IOException("Output File descriptor is not valid");
    }

    public Map<String, Map<String, Double>> getPPSHMap()
    {
        if (ppshMap == null)
            throw new IllegalStateException("PSSH Map is null!  XAlign didn't return one, or we didn't ask the parser for one.");
        return ppshMap;
    }

    public void setKeepOutput(boolean b)
    {
        keepOutput = b;
    }
}
