package ca.ualberta.proteus.util.db;

import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import ca.ualberta.proteus.util.io.StructureDB;
import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.db.IllegalIDException;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: May 21, 2007
 * Time: 10:48:33 AM
 */
public class CreateDatabase
{
    public static boolean verbose = true;

    public static void main(String[] args) throws IOException, BioException
    {
        createDatabases();
    }

    public static void createDatabases()
            throws IOException, BioException
    {
        Config config = Config.getConfig();

        // input files
        String seqsee_db = config.getStringValue("seqsee_db");
        String seqsee_tmpdb = config.getStringValue("TMPDB");
        String seqsee_bbdb = config.getStringValue("BBDB");
        String seqsee_spdb_euk = config.getStringValue("SPDB.euk");
        String seqsee_spdb_gram_pos = config.getStringValue("SPDB.gram+");
        String seqsee_spdb_gram_neg = config.getStringValue("SPDB.gram-");
        
        // output files
        String pdb = config.getStringValue("PDB");
        String tmppdb = config.getStringValue("blast.TMPDB");
        String bbdb = config.getStringValue("blast.BBDB");
        String spdb_euk = config.getStringValue("blast.SPDB.euk");
        String spdb_gram_pos = config.getStringValue("blast.SPDB.gram+");
        String spdb_gram_neg = config.getStringValue("blast.SPDB.gram-");
        
        writeDatabase(new File(seqsee_db), new File(pdb));
        writeDatabase(new File(seqsee_tmpdb), new File(tmppdb));
        writeDatabase(new File(seqsee_bbdb), new File(bbdb));
//        writeDatabase(new File(seqsee_spdb_euk), new File(spdb_euk));
//        writeDatabase(new File(seqsee_spdb_gram_pos), new File(spdb_gram_pos));
//        writeDatabase(new File(seqsee_spdb_gram_neg), new File(spdb_gram_neg));
    }

    private static void writeDatabase(File in, File out) throws IOException, BioException
    {
        try
        {
            if ( verbose)
            {
                System.out.println("Reading database from " + in.getAbsolutePath());
            }

            // Open it R/W because if we are accessing every sequence in the DB it will be much faster
            StructureDB db = new StructureDB(in, StructureDB.MODE.READWRITE);

            PrintWriter writer = new PrintWriter(out);

            if ( verbose)
                System.out.println("DB has " + db.size() + " sequences");

            while (db.hasNext())
            {
                StructureAnnotatedSequence seq = (StructureAnnotatedSequence) db.nextSequence();
                if (seq.length() > 0)
                {
                    writer.println(">" + seq.getName());
                    //writer.println(seq.getStructureAnnotation().seqString());
                    writer.println(seq.seqString());
                }
            }
            writer.close();
            SeqSeeDB.releaseDB();

            if ( verbose)
                System.out.println("DB is in " + out.getAbsolutePath());

            // Format the new database
            String formatdb = Config.getConfig().getStringValue("formatdb.command");
            runProgram(formatdb + " -i " + out.getAbsolutePath());
        }
        catch (RuntimeException e)
        {
            System.out.println("Error creating database from: " + in.getName());
            throw e;
        }
    }

    private static boolean runProgram(String cmd)
    {
        return runProgram(cmd, null, null);
    }

    /**
     * @param cmd  - command to run to check existence
     * @param name - Program name output to stdout (optional).  If null, output 'cmd'
     * @return true is command passes
     */
    private static boolean runProgram(String cmd, String name, String dir)
    {
        if (name == null)
            System.out.print("Running " + cmd + "...");
        else
            System.out.print("Running " + name + "...");
        try
        {
            Process proc;
            if (dir != null)
            {
                proc = Runtime.getRuntime().exec(cmd, null, new File(dir));
            }
            else
            {
                proc = Runtime.getRuntime().exec(cmd);
            }

            int exitVal = 0;
            try
            {
                InputStream is = proc.getErrorStream();
                int i = 0;
                while ((i = is.read()) != -1) System.err.print((char) i);
                exitVal = proc.waitFor();
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }
        }
        catch (IOException e)
        {
            System.out.println("failed.");
            return false;
        }
        System.out.println("done.");
        return true;
    }

    public static void checkDatabases()
    {
        Config config = Config.getConfig();

        // input files
        String seqsee_tmpdb = config.getStringValue("TMPDB");
        String seqsee_db = config.getStringValue("seqsee_db");

        // output files
        String pdb = config.getStringValue("PDB");
        String tmppdb = config.getStringValue("blast.TMPDB");

        try
        {
            checkDatabase(new File(seqsee_db), new File(pdb));
            checkDatabase(new File(seqsee_tmpdb), new File(tmppdb));
        }
        catch (IOException e)
        {
            e.printStackTrace();  
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }
    }

    /**
     *  Verifies that both databases have the exact same structures in them
     *
     * @param seqsee  The seqsee version of the database
     * @param blast   The blast-able version of the database
     */
    private static void checkDatabase(File seqsee, File blast) throws IOException, BioException
    {
        // Open it R/W because if we are accessing every sequence in the DB it will be much faster
        StructureDB db = new StructureDB(seqsee, StructureDB.MODE.READWRITE);

        List<Sequence> sequences = ProteusTools.parseFasta(blast);
        Map<String, Sequence> sequenceMap = new HashMap<String, Sequence>();
        int missedSequences = 0;
        for (Sequence seq : sequences)
        {
            try
            {
                Sequence sequence = db.getSequence(seq.getName());

                sequenceMap.put(seq.getName(), sequence);
            }
            catch (IllegalIDException e)
            {
                missedSequences++;
                System.out.println( seqsee.getName() + " is missing " + seq.getName());
            }
        }

        // Test the other way
        while ( db.hasNext() )
        {
            Sequence sequence = db.nextSequence();

            Sequence sequenceFromBlast = sequenceMap.get(sequence.getName());

            if ( sequenceFromBlast == null)
            {
                missedSequences++;
                System.out.println( blast.getName() + " is missing " + sequence.getName());
            }
        }

        if ( missedSequences != 0 )
        {
            System.out.println( seqsee.getName() + " is missing " + missedSequences + " sequences");
        }

        System.out.println(seqsee.getName() + " has " + db.size() + " sequences");
        System.out.println(blast.getName() + " has " + sequences.size() + " sequences");
    }
}
