package ca.ualberta.proteus.util.db;

import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import org.biojava.bio.seq.ProteinTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.IllegalSymbolException;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: May 4, 2006
 * Time: 11:08:11 PM
 * To change this template use File | Settings | File Templates.
 */
class SQLPredictionDB implements PredictionDB
{
    static SQLPredictionDB db = null;

    public static SQLPredictionDB getInstance()
    {
        if (db == null)
        {
            db = new SQLPredictionDB();
        }
        return db;
    }

    /**
     * Adds a sequence object to the database
     *
     * @param sequence
     * @throws IllegalArgumentException
     */
    public static void addSequence(Sequence sequence) throws IllegalArgumentException
    {
        if (get(sequence.getName()) != null)
        {
            return;
        }

        final String QUERY = "INSERT INTO PROTEUS.SEQUENCES "
                + " VALUES (null, ?, ? )";


        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setString(1, sequence.getName());
            stmt.setString(2, sequence.seqString());

            stmt.execute();
        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Returns the sequence as a string
     *
     * @param name
     * @return the sequence as a string
     */
    public static Sequence get(String name)
    {
        final String QUERY = "SELECT SEQUENCE FROM PROTEUS.SEQUENCES "
                + " WHERE NAME = ? ";

        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setString(1, name);

            ResultSet resultSet = stmt.executeQuery();

            if (resultSet.next())
            {
                String seqStr = resultSet.getString(1);

                Sequence sequence;
                sequence = ProteinTools.createProteinSequence(seqStr, name);
                return sequence;
            }
        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }

        return null;
    }

    public static int getSequenceID(Sequence sequence)
    {
        return getSequenceID(sequence.getName());
    }

    public static int getSequenceID(String name)
    {
        final String QUERY = "SELECT ID FROM PROTEUS.SEQUENCES "
                + " WHERE NAME = ? ";

        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setString(1, name);

            ResultSet resultSet = stmt.executeQuery();

            if (resultSet.next())
            {
                return resultSet.getInt(1);
            }
        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }

        return -1;
    }

    public void add(Sequence sequence, String method, ScoreAnnotatedSequence prediction)
    {
        int seqID = getSequenceID(sequence);

        // Sequence isn't in the database, better put it there
        if (seqID == -1)
        {
            SQLPredictionDB.addSequence(sequence);
            seqID = getSequenceID(sequence);
        }

        if (prediction == null)
        {
            throw new NullPointerException("Cannot add a null prediction to the database");
        }

        if (contains(sequence, method))
        {
            updateSequencePrediction(seqID, method, prediction);
        }
        else
        {
            insertSequencePrediction(seqID, method, prediction);
        }
    }

    public void remove(ScoreAnnotatedSequence sequence, String method)
    {
        throw new IllegalArgumentException("Not implemented");
    }

    private static void updateSequencePrediction(int seqID, String method, ScoreAnnotatedSequence prediction)
    {
        final String QUERY = "UPDATE PROTEUS.STRUCTURE "
                + " SET STRUCTURE=?, SCORE=?"
                + " WHERE METHOD = ? "
                + " AND SEQUENCE_ID = ?";

        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setString(1, prediction.getStructureAnnotation().seqString());
            stmt.setString(2, prediction.getScoresAsStr());
            stmt.setString(3, method);
            stmt.setInt(4, seqID);

            stmt.execute();

        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }
    }

    private static void insertSequencePrediction(int seqID, String method, ScoreAnnotatedSequence prediction)
    {
        final String QUERY = "INSERT INTO PROTEUS.STRUCTURE "
                + " VALUES ( ?, ?, ?, ? )";

        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setInt(1, seqID);
            stmt.setString(2, method);
            stmt.setString(3, prediction.getStructureAnnotation().seqString());
            stmt.setString(4, prediction.getScoresAsStr());

            stmt.execute();

        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }
    }

    public ScoreAnnotatedSequence get(Sequence sequence, String method)
    {
        int seqID = getSequenceID(sequence);

        if (seqID != -1)
        {
            final String QUERY = "SELECT STRUCTURE, SCORE FROM PROTEUS.STRUCTURE "
                    + " WHERE METHOD = ? "
                    + " AND SEQUENCE_ID = ?";

            try
            {
                Connection con = DBUtils.getConnection();

                PreparedStatement stmt = con.prepareStatement(QUERY);
                stmt.setString(1, method);
                stmt.setInt(2, seqID);

                ResultSet resultSet = stmt.executeQuery();

                if (resultSet.next())
                {
                    String structStr = resultSet.getString(1);
                    String scoreStr = resultSet.getString(2);

                    return new ScoreAnnotatedSequence(sequence, structStr, scoreStr);
                }
            }
            catch (SQLException e)
            {
                e.printStackTrace();
            }
        }
        
        return null;
    }

    public static Sequence getSequenceForID(int ID)
    {
        final String QUERY = "SELECT NAME, SEQUENCE FROM PROTEUS.SEQUENCES "
                + " WHERE ID = ? ";

        try
        {
            Connection con = DBUtils.getConnection();

            PreparedStatement stmt = con.prepareStatement(QUERY);
            stmt.setInt(1, ID);

            ResultSet resultSet = stmt.executeQuery();

            if (resultSet.next())
            {
                String name = resultSet.getString(1);
                String seqStr = resultSet.getString(2);

                Sequence sequence;
                sequence = ProteinTools.createProteinSequence(seqStr, name);
                return sequence;
            }
        }
        catch (SQLException e)
        {
            e.printStackTrace();
        }
        catch (IllegalSymbolException e)
        {
            e.printStackTrace();
        }

        return null;
    }

    public boolean contains(Sequence sequence)
    {
        return get(sequence.getName()) != null;
    }

    public boolean contains(Sequence sequence, String method)
    {
        return get(sequence, method) != null;
    }

    public void remove(StructureAnnotatedSequence sequence, String methodName)
    {
        throw new IllegalArgumentException("Not implemented");
    }
}
