package ca.ualberta.proteus.util.domainImage;

import ca.ualberta.proteus.program.transsec.test.TransSecTest;
import ca.ualberta.proteus.util.io.SeqSeeDB;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.GappedSequence;
import org.biojava.bio.seq.Sequence;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

/**
 * User: Scott Montgomerie
 * Date: Mar 28, 2006
 * Description:
 */
public class ConservedDomainImageGenerator
{
    private List<GappedSequence> alignedSequences;

    final static int LINE_WIDTH = 7;

    /**
     * Creates a new generator
     * @param alignedSequences
     */
    public ConservedDomainImageGenerator(List<GappedSequence> alignedSequences)
    {
        assert alignedSequences != null : "Aligned sequences are null";

        this.alignedSequences = alignedSequences;
    }

    public static void main(String[] args)
    {
        SeqSeeDB db = SeqSeeDB.getInstance();
        try
        {
            Sequence sequence = db.getSequence(0);
            List<GappedSequence> homologousSequences = TransSecTest.getHomologousSequences(sequence);

            ConservedDomainImageGenerator generator = new ConservedDomainImageGenerator(homologousSequences);
            try
            {
                generator.createImage( new FileOutputStream(sequence.getName() + ".png"), 800 );
            }
            catch (IOException e)
            {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
            for(GappedSequence s : homologousSequences)
            {
                System.out.println(s);
            }
        }
        catch (BioException e)
        {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    public void createImage(OutputStream outputStream, int width)
    {
        assert alignedSequences != null : "Aligned sequences are null";

        int height = ( alignedSequences.size() + 1 ) * (10 + LINE_WIDTH);

        // Create a buffered image in which to draw
        BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);

        // Create a graphics contents on the buffered image
        Graphics2D g2d = bufferedImage.createGraphics();

        int[] starts = new int[alignedSequences.size()];
        int[] ends = new int[alignedSequences.size()];

        for (int i=0; i<alignedSequences.size(); i++)
        {
            GappedSequence sequence = alignedSequences.get(i);
            starts[i] = sequence.firstNonGap();
            ends[i] = sequence.lastNonGap();
        }

        // clear the area
        g2d.setColor(Color.WHITE);
        g2d.fillRect(0, 0 , width, height);

        Font f = new Font("Arial", Font.PLAIN, 9);
        g2d.setFont(f);
        FontMetrics fontMetrics = g2d.getFontMetrics();

        final int ascent = fontMetrics.getAscent();
        final String str = "QUERY";
        final int offset = fontMetrics.stringWidth("1WWWW");
        final int vertSpacing = ascent + LINE_WIDTH;

        // draw the string
        g2d.setColor(Color.BLUE);
        g2d.drawString(str, 0, ascent);

        // draw the query line

        g2d.fillRect(offset, 1, width, LINE_WIDTH);

        if ( alignedSequences.size() != 0 )
        {
            GappedSequence sequence = alignedSequences.get(0);
            int length = sequence.length();

            // draw each sequence
            for (int i=0; i<alignedSequences.size(); i++)
            {
                double lengthUnit = (width - offset) / (double) length;
                int start = (int) (starts[i] * lengthUnit);
                int end = (int) (ends[i] * lengthUnit);

                sequence = alignedSequences.get(i);

                g2d.setColor(Color.GREEN);
                g2d.fillRect(offset + start, vertSpacing * ( i + 1 ), end - start, LINE_WIDTH);

                // draw sequence ID
                g2d.setColor(Color.BLACK);
                g2d.drawString(sequence.getName(), 0, (int) (vertSpacing * ( i + 1.5 )));

                // draw sequence start  
                g2d.drawString(String.valueOf(starts[i]), offset + start, (int) (vertSpacing * ( i + 1.5 )));

                // draw sequence end
                String endStr = String.valueOf(ends[i]);
                int endOffset = fontMetrics.stringWidth(endStr);
                g2d.drawString(endStr, offset + end - endOffset, (int) (vertSpacing * ( i + 1.5 )));
            }
        }

        try
        {
            // Save as PNG
            ImageIO.write(bufferedImage, "png", outputStream);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        // Graphics context no longer needed so dispose it
        g2d.dispose();
    }
}
