package ca.ualberta.proteus.util.eval;

import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.program.MethodSettings;
import ca.ualberta.proteus.program.ProteusImpl;
import ca.ualberta.proteus.program.ProteusSettings;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.CompletionListener;
import ca.ualberta.proteus.util.Consts;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.util.Score;
import ca.ualberta.proteus.util.db.PredictionDB;
import ca.ualberta.proteus.util.db.PredictionDBFactory;
import ca.ualberta.proteus.util.io.StructureDB;
import com.martiansoftware.jsap.*;
import org.biojava.bio.BioException;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.symbol.SymbolList;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: Jun 5, 2007
 * Time: 10:01:08 PM
 */
public class EVAEval
{
    private static final String FILE1_FLAG = "Seqsee DB file";
    private static final String FILE2_FLAG = "List of Sequence ID's";
    private static final String START = "start";
    private static final String END = "end";
    private static final String NEW = "new";
    private static boolean predictNew = false;

    public static void main(String[] args) throws IOException, JSAPException
    {
        JSAPResult options = parseCommandLine(args);

        if (options != null)
        {
            run(options);
        }
    }

    private static JSAPResult parseCommandLine(String[] args) throws JSAPException
    {
        SimpleJSAP options = new SimpleJSAP(
                "Proteus",
                "Comprehensive Secondary Structure Predictions, Local Version ",
                new Parameter[]
                        {
                                new FlaggedOption(START, JSAP.INTEGER_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "start",
                                        "Index of stat sequence"),
                                new FlaggedOption(END, JSAP.INTEGER_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'e', "end",
                                        "Index of end sequence."),
                                new UnflaggedOption(FILE1_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY,
                                        "The name of the file containing the sequences in seqsee.db format."),
                                new FlaggedOption(FILE2_FLAG, JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'f', "file",
                                        "The name of the file containing the a list of sequence names."),
                                new Switch(NEW, 'n', "new",
                                        "If the predictor should ignore predictions already in the database.")
                        }
        );

        JSAPResult config = options.parse(args);

        if (options.messagePrinted())
        {
            System.err.println("Usage: ");
            System.err.println("  " + options.getUsage());
            System.err.println();
            System.err.println(options.getHelp());

            return null;
        }

        return config;
    }

    public static void run(JSAPResult options) throws IOException
    {
        List<StructureAnnotatedSequence> list = new ArrayList<StructureAnnotatedSequence>();
        StructureDB db = new StructureDB(options.getString(FILE1_FLAG), StructureDB.MODE.READWRITE /*Not READ, So it's faster and not seeking all the time*/);

        int start = 0, end = db.size();

        if (options.contains(START))
        {
            start = options.getInt(START);
        }
        if (options.contains(END))
        {
            end = options.getInt(END);
        }

        predictNew = options.getBoolean(NEW, false);

        if (start < 0)
        {
            start = 0;
        }
        if (end < 0)
        {
            end = db.size();
        }

        if (options.contains(FILE2_FLAG))
        {
            BufferedReader reader = new BufferedReader(new FileReader(options.getString(FILE2_FLAG)));

            String line;
            int i = 0;
            while ((line = reader.readLine()) != null)
            {
                boolean add = false;
                if (i >= start && i <= end)
                    add = true;

                i++;

                if (add)
                {
                    try
                    {
                        list.add((StructureAnnotatedSequence) db.getSequence(line));
                    }
                    catch (IllegalIDException e)
                    {
                        System.out.println(line + " does not exist");
                    }
                }
            }
        }
        else
        {
            SequenceIterator it = db.sequenceIterator();
            while (it.hasNext())
            {
                try
                {
                    list.add((StructureAnnotatedSequence) it.nextSequence());
                }
                catch (BioException e)
                {
                    e.printStackTrace();
                }
            }
        }

        run(list, start, end);
    }

    private static void run(List<StructureAnnotatedSequence> list, int start, int end)
    {
        // final List<ScoreAnnotatedSequence> predictions = new ArrayList<ScoreAnnotatedSequence>();

        final ProteusSettings settings = new ProteusSettings();
        settings.setMethod(MethodSettings.PSIPRED_JNET);
        settings.setEvaluating(true);

        final Score[] finalScore = new Score[list.size()];
        final List<Score> homologScore = new ArrayList<Score>((int) (list.size() * 0.9));

        final List<Score> jnetScores = new ArrayList<Score>(list.size());
        final List<Score> psipredScores = new ArrayList<Score>(list.size());
        final List<Score> internalScores = new ArrayList<Score>(list.size());

        final PredictionDB predictionDB = PredictionDBFactory.getRunTimeDB(true);

        System.out.printf("Running on %d sequences (%d-%d)", list.size(), start, end);

        // run proteus on each sequence
        int index = 0;
        int numSequences = 0;
        final int[] homologsFound = new int[1];
        for (final StructureAnnotatedSequence sequence : list)
        {
            settings.setSequence(sequence);

            ScoreAnnotatedSequence psipred = predictionDB.get(sequence, "PSIPRED");
            ScoreAnnotatedSequence jnet = predictionDB.get(sequence, "JNET");
            ScoreAnnotatedSequence internal = predictionDB.get(sequence, "PROTEUS");

           /* if (psipred != null && jnet != null && !predictNew)
            {
                if (psipred != null)
                {
                    psipredScores.add(createScore(sequence, psipred));
                }
                if (jnet != null)
                {
                    jnetScores.add(createScore(sequence, jnet));
                }
                if (internal != null)
                {
                    internalScores.add(createScore(sequence, internal));
                }
            }
            else*/
            {
                final ProteusImpl proteus = new ProteusImpl(settings);

                final int index1 = index;

                proteus.addCompletionListener(new CompletionListener()
                {
                    public void onComplete(boolean successful, String message)
                    {
                        RuntimeCatalog logger = proteus.getLogger();

                        ScoreAnnotatedSequence homologSequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
                        ScoreAnnotatedSequence predictedSequence = (ScoreAnnotatedSequence) logger.getLogObject(Consts.FINAL);

                        if (homologSequence != null)
                        {
                            homologsFound[0]++;
                        }

                        ScoreAnnotatedSequence psipred = (ScoreAnnotatedSequence) logger.getLogObject(Consts.PSIPRED);
                        if ( psipred == null)
                        {
                            psipred = predictionDB.get(sequence, "PSIPRED");
                        }
                        ScoreAnnotatedSequence jnet = (ScoreAnnotatedSequence) logger.getLogObject(Consts.JNET);
                        if ( jnet == null)
                        {
                            jnet = predictionDB.get(sequence, "JNET");
                        }
                        ScoreAnnotatedSequence internal = (ScoreAnnotatedSequence) logger.getLogObject(Consts.TRANSSEC);
                        if ( internal == null)
                        {
                            internal = predictionDB.get(sequence, "TRANSSEC");
                        }

                        if (psipred != null)
                        {
                            psipredScores.add(createScore(sequence, psipred));
                        }
                        if (jnet != null)
                        {
                            jnetScores.add(createScore(sequence, jnet));
                        }
                        if (internal != null)
                        {
                            internalScores.add(createScore(sequence, internal));
                        }

                        switch (settings.getMethod())
                        {
                            case PSIPRED:
                                predictedSequence = psipred;
                                break;
                            case JNET:
                                predictedSequence = jnet;
                                break;
                            case TRANSSEC:
                                predictedSequence = internal;
                                break;
                            case ALL:
                            case PSIPRED_JNET:
                            default:
                                break;
                        }

                        System.out.println("Final sequence:");
                        System.out.println(predictedSequence);

                        /* Do a replacement B->E */
                        SymbolList structure = sequence.getStructureAnnotation();
                        String structStr = structure.seqString().toUpperCase().replace("B", "E");
                        sequence.setStructureAnnotation(ProteusTools.createStructure(structStr));

                        finalScore[index1] = createScore(sequence, predictedSequence);

                        if (homologSequence != null)
                        {
                            Score score = createScore(sequence, homologSequence);
                            homologScore.add(score);
                        }

                    }

                });

                try
                {
                    proteus.run();

                    numSequences++;
                }
                catch (Throwable t)
                {
                    t.printStackTrace();
                }

                index++;
            }
        }

        Score avgScore = getAverageScore(Arrays.asList(finalScore));
        Score avgHomologScore = getAverageScore(homologScore);
        Score jnetAvgScore = getAverageScore(jnetScores);
        Score psipredAvgScore = getAverageScore(psipredScores);

        NumberFormat format = NumberFormat.getNumberInstance();
        format.setMaximumFractionDigits(4);
        System.out.println(homologsFound[0] + " sequences had homologs");

        System.out.println("JNET:\t"
                + "Q3: " + format.format(jnetAvgScore.q3 / (double) jnetScores.size())
                + "\tSOV: " + format.format(jnetAvgScore.sov / (double) jnetScores.size()));

        System.out.println("PSIPRED:\t"
                + "Q3: " + format.format(psipredAvgScore.q3 / (double) psipredScores.size())
                + "\tSOV: " + format.format(psipredAvgScore.sov / (double) psipredScores.size()));

        System.out.println("FINAL:\t"
                + "Q3: " + format.format(avgScore.q3 / (double) numSequences)
                + "\tSOV: " + format.format(avgScore.sov / (double) numSequences));

        System.out.println("HOMOLOG:\t"
                + "Q3: " + format.format(avgHomologScore.q3 / (double) homologsFound[0])
                + "\tSOV: " + format.format(avgHomologScore.sov / (double) homologsFound[0]));
    }

    private static Score getAverageScore(Iterable<Score> finalScore)
    {
        Score avgScore = new Score();
        for (Score s : finalScore)
        {
            if (s != null)
            {
                avgScore.q3 += s.q3;
                avgScore.sov += s.sov;
            }
        }
        return avgScore;
    }

    private static Score createScore(StructureAnnotatedSequence sequence, ScoreAnnotatedSequence predictedSequence)
    {
        SymbolList structure = sequence.getStructureAnnotation();
        String structStr = structure.seqString().toUpperCase().replace("B", "E");
        sequence.setStructureAnnotation(ProteusTools.createStructure(structStr));

        Score score = new Score();
        score.q3 = ProteusTools.q3Score(sequence.getStructureAnnotation(), predictedSequence.getStructureAnnotation());
        score.sov = ProteusTools.sovScore(sequence.getStructureAnnotation(), predictedSequence.getStructureAnnotation());

        return score;
    }
}
