package ca.ualberta.proteus.util.io;

import ca.ualberta.proteus.logging.ProteusLogger;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.BioException;
import org.biojava.bio.proteomics.StructureTools;
import org.biojava.bio.seq.FeatureFilter;
import org.biojava.bio.seq.FeatureHolder;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;
import org.biojava.bio.seq.db.IllegalIDException;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.utils.ChangeListener;
import org.biojava.utils.ChangeType;

import java.io.*;
import java.util.*;
import java.util.zip.GZIPInputStream;

/**
 * Created by IntelliJ IDEA.
 * User: scott
 * Date: Jun 6, 2005
 * Time: 11:33:08 PM
 * To change this template use File | Settings | File Templates.
 */
public class StructureDB extends Thread implements SequenceIterator, SequenceDB, Cloneable
{
    private File db_file;
    protected TreeMap<String, Sequence> sequences = new TreeMap<String, Sequence>(); 
    protected TreeMap<Integer, String> idToSequence = new TreeMap<Integer, String>();

    protected Iterator<String> sequenceIterator;
    private static final String DB_NAME = "Vadar-Annotated Secondary Structure Sequence Database";
    
    // TODO comment
    private boolean writeComments = true;

    private static ProteusLogger logger = ProteusLogger.getLogger();
    private MODE mode;
    private int id = 0;
    private StructureDBReader reader;

    public enum MODE
    {
        /*READ will allow the database to be accessed randomly from the file, which is much faster and efficient.
        * READWRITE or WRITE will load the contents of the DB into memory, so that it can be modified.  Use
        * READWRITE for full scans of the database, as that will be much more efficient than constantly
        * looking at the file. */
        READ, READWRITE, WRITE
    }

    /**
     * Creates a new, empty structure database
     */
    public StructureDB(MODE mode)
    {
        this.mode = mode;
    }

    /**
     * Creates a new structure database populated with data from the filename
     */
    public StructureDB(File file, MODE mode) throws IOException
    {
        this.mode = mode;
        this.db_file = file;

        reloadDB();

        sequenceIterator = sequences.keySet().iterator();
    }

    /**
     * Creates a new structure database populated with data from the filename
     */
    public StructureDB(String filename, MODE mode) throws IOException
    {
        this(new File(filename), mode);
    }



    /**
     * Add a placeholder so we know it's in the database, just not loaded
     * @param name
     */
    void addEmptySequence(String name)
    {
        if ( ! sequences.containsKey(name))
        {
            // We're not replacing it so it's a new one, as in the sequence does not exist twice in the maps
            idToSequence.put(id++, name);
        }

        sequences.put(name, null);

        // The maps should be exactly the same size
        assert sequences.size() == idToSequence.size();
    }

    public void clear()
    {
        sequences.clear();
        idToSequence.clear();
        id=0;
    }

    protected void reloadDB()
            throws IOException
    {
        reader = new StructureDBReader(this, mode == MODE.READ);
        InputStream inputStream = findInputStream();
        try
        {
            reader.parseDB(inputStream);
        }
        finally
        {
            inputStream.close();
        }
    }

    /**
     * Copy constructor
     */
    private StructureDB(StructureDB template)
    {
        this.db_file = template.db_file;

        for (String seqName : template.sequences.keySet())
        {
            this.sequences.put(seqName, template.sequences.get(seqName));
        }

        for (Integer id : template.idToSequence.keySet())
        {
            this.idToSequence.put(id, template.idToSequence.get(id));
        }

        sequenceIterator = sequences.keySet().iterator();

        this.mode = template.mode;
        this.reader = template.reader;
        this.id = template.id;
    }

    private InputStream findInputStream()
    {
        InputStream inputStream;
        //FileInputStream in = new FileInputStream(db_file);
        InputStream in = null;
        try
        {
            in = new FileInputStream(db_file);
            logger.debug("Using DB file " + new File("").getAbsolutePath() + "/" + db_file);
        }
        catch (FileNotFoundException e)
        {
            // nada
        }

        if (in == null)
        {
            in = getClass().getResourceAsStream(db_file.getAbsolutePath());

            if (in != null)
            {
                logger.debug("Using DB file " + getClass().getResource(db_file.getAbsolutePath()).getPath());
            }
        }

        if (in == null)
        {
            in = getClass().getResourceAsStream("/" + db_file);

            if (in != null)
            {
                logger.debug("Using DB file " + getClass().getResource("/" + db_file).getPath());
            }
        }

        if (in == null)
        {
            try
            {
                in = new FileInputStream(db_file);
            }
            catch (FileNotFoundException e)
            {
                // nada
            }
        }
        if (in == null)
        {
            throw new IllegalArgumentException(db_file + " is not in the classpath.  try adding \".\"");
        }

        try
        {
            inputStream = new GZIPInputStream(in);
        }
        catch (IOException e)
        {
            // "in" isn't used anymore, we're going to re-open the file another way
            try
            {
                in.close();
            }
            catch (IOException e1)
            {
                e1.printStackTrace();
            }

            // not a GZIP input stream
            inputStream = getClass().getResourceAsStream(db_file.getAbsolutePath());

            if (inputStream == null)
                inputStream = getClass().getResourceAsStream("/" + db_file);

            if (inputStream == null)
            {
                try
                {
                    inputStream = new FileInputStream(db_file);
                }
                catch (FileNotFoundException e1)
                {
                    // nada
                }
            }
        }
        return inputStream;
    }

    public boolean hasNext()
    {
        return sequenceIterator.hasNext();
    }

    public Sequence nextSequence() throws NoSuchElementException, BioException
    {
        return sequences.get(sequenceIterator.next());
    }

    public int size()
    {
        if (mode == MODE.READ)
        {
            // D
            return reader.size();
        }
        return sequences.size();
    }

    public String getDBName()
    {
        return DB_NAME;
    }

    public Sequence getSequence(String s) throws IllegalIDException
    {
        if (mode == MODE.READ)
        {
            // See if it's stored in our memory.
            Sequence rv = sequences.get(s);
            if (rv == null)
            {
                // with trailing _
                rv = sequences.get(s + "_");
            }
            if (rv != null)
            {
                return rv;
            }

            // See if it's stored in the file.
            rv = reader.fetchSequence(db_file, s);

            if (rv == null)
            {
                throw new IllegalIDException("Sequence with ID " + s + " does not exist in " + db_file);
            }

            return rv;
        }

        Sequence rv = sequences.get(s);
        if (rv == null)
        {
            // with trailing _
            rv = sequences.get(s + "_");
        }
        if (rv == null)
        {
            throw new IllegalIDException("Sequence with ID " + s + " does not exist in " + db_file);
        }
        return rv;
    }

    public Sequence getSequence(int ID) throws BioException
    {
        String name = idToSequence.get(ID);
        if (name == null)
        {
            throw new IllegalIDException("Sequence with ID " + ID + " does not exist in " + db_file);
        }

        return getSequence(name);
    }

    public Collection<Sequence> getAllSequences()
    {
        return sequences.values();
    }

    public void addSequence(Sequence sequence)
    {
        assert sequence.getName() != null;

        if ( ! sequences.containsKey(sequence.getName()))
        {
            // We're not replacing it so it's a new one, as in the sequence does not exist twice in the maps
            idToSequence.put(id++, sequence.getName());
        }
        
        sequences.put(sequence.getName(), sequence);

        // The maps should be exactly the same size
        assert sequences.size() == idToSequence.size();
    }

    public boolean containsSequence(String s)
    {
        return sequences.containsKey(s) || sequences.containsKey(s + "_");
    }

    public void removeSequence(String s) throws IllegalIDException
    {
        if (sequences.remove(s) == null)
            throw new IllegalIDException("Sequence with ID " + s + " does not exist in " + db_file);
    }

    public synchronized void writeDB(File file) throws FileNotFoundException
    {
        writeDB(new PrintStream(file));
    }

    public synchronized void writeDB(PrintStream writer) {
    	writeDB(writer, false);
    }
    
    public synchronized void writeDB(PrintStream writer, boolean uppercase)
    {
        if (mode == MODE.READ)
        {
            throw new IllegalStateException("Database was not opened for writing!");
        }


        for (String seqName : sequences.keySet())
        {
            Sequence sequence = sequences.get(seqName);

            if (seqName.length() == 4)
            {
                seqName += "_";
            }

            writer.println(">" + seqName);
            if (writeComments)
            {
                writer.println("#BRKHAVN ID: " + seqName);
                writer.println("#NUM RESIDU: " + sequence.length());
            }
            writer.println();

            String seqStr = sequence.seqString();
            String structStr = null;
            if (sequence instanceof StructureAnnotatedSequence)
            {
                structStr = ((StructureAnnotatedSequence) sequence).getStructureAnnotation().seqString();
                if( uppercase )
                	structStr = structStr.toUpperCase().replaceAll("SC", "Sc");
            }

            for (int i = 0; i < sequence.length(); i += 50)
            {
                assert structStr != null;

                assert structStr.length() == seqStr.length();

                String line = seqStr.substring(i, Math.min(i + 50, seqStr.length()));
                line = line.replaceAll("(\\w{10})", "$1 ");
                writer.println(line);

                line = structStr.substring(i, Math.min(i + 50, seqStr.length()));
                //  line = line.toUpperCase().replace("E", "B");
                line = line.replaceAll("(\\w{10})", "$1 ");
                writer.println(line);
                writer.println();
            }
        }
    }

    public void addChangeListener(ChangeListener changeListener)
    {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public void addChangeListener(ChangeListener changeListener, ChangeType changeType)
    {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public void removeChangeListener(ChangeListener changeListener)
    {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public void removeChangeListener(ChangeListener changeListener, ChangeType changeType)
    {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public boolean isUnchanging(ChangeType changeType)
    {
        return true;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public Set ids()
    {
        return sequences.keySet();  //To change body of implemented methods use File | Settings | File Templates.
    }

    public SequenceIterator sequenceIterator()
    {
        sequenceIterator = sequences.keySet().iterator();
        return this;
    }

    public FeatureHolder filter(FeatureFilter featureFilter)
    {
        return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public static void main(String[] argc) throws IOException, BioException
    {
        if (argc.length == 0)
        {
            StructureDB db = new StructureDB("pdbstruct.fa", MODE.READWRITE);
            logger.debug("DB has " + db.size() + " sequences");

            SequenceIterator it = db;
            while (it.hasNext())
            {
                try
                {
                    logger.debug(it.nextSequence().toString());
                }
                catch (BioException e)
                {
                    e.printStackTrace();
                }
            }

            db.writeDB(System.out);
        }
        else if (argc[0].equals("-size"))
        {
            StructureDB db = new StructureDB(argc[0], MODE.READ);
            System.out.print("DB has " + db.size() + " sequences");
        }
        else
        {
            StructureDB db = new StructureDB(argc[0], MODE.WRITE);
            PrintWriter writer = new PrintWriter("tmPDB");
            System.out.print("DB has " + db.size() + " sequences");

            SequenceIterator it = db;
            while (it.hasNext())
            {
                StructureAnnotatedSequence seq = (StructureAnnotatedSequence) it.nextSequence();
                writer.println(">" + seq.getName());
                //writer.println(seq.getStructureAnnotation().seqString());
                writer.println(seq.seqString());
            }
            writer.close();
        }
    }

    @Override
    public StructureDB clone()
    {
        return new StructureDB(this);
    }

    public DatabaseStatistics getStats() throws BioException
    {
        DatabaseStatistics dbStats = new DatabaseStatistics();
        StructureTools structureTools = new StructureTools();
        reset();
        for (SequenceIterator it = this.sequenceIterator(); it.hasNext();)
        {
            StructureAnnotatedSequence sequence = (StructureAnnotatedSequence) it.nextSequence();
            String sl = sequence.getStructureAnnotation().seqString().toUpperCase();
            for (int i = 0; i < sl.length(); i++)
            {
                char c = sl.charAt(i);
                switch (c)
                {
                    case ProteusTools.HELIX_CHAR:
                        dbStats.helix++;
                        break;
                    case ProteusTools.COIL_CHAR:
                        dbStats.coil++;
                        break;
                    case ProteusTools.STRAND_CHAR:
                        dbStats.strand++;
                        break;
                    default:
                        throw new RuntimeException("Wrong character");
                }
                /* if (s.equals(structureTools.getC()))
                dbStats.coil++;
            else if (s.equals(structureTools.getE()))
                dbStats.strand++;
            else if (s.equals(structureTools.getH()))
                dbStats.helix++;
            else
                throw new RuntimeException("WTF?");
                */
                dbStats.numResidues++;
            }
        }
        dbStats.normalize();
        return dbStats;
    }

    /**
     * Reset the iterator
     */
    public void reset()
    {
        sequenceIterator = sequences.keySet().iterator();
    }

    public static class DatabaseStatistics
    {
        public double coil;
        public double helix;
        public double strand;
        public int numResidues;

        public String toString()
        {
            String result = "";
            result += "Total number of residues = " + numResidues + "\n";
            result += "Helix = " + helix + "\n";
            result += "Coil = " + coil + "\n";
            result += "Strand = " + strand + "\n";
            return result;
        }

        public void normalize()
        {
            double total = coil + helix + strand;
            coil = coil / total;
            helix = helix / total;
            strand = strand / total;
        }
    }


    /**
     * TODO
     *
     * @return
     */
    public boolean isWriteComments()
    {
        return writeComments;
    }

    /**
     * TODO
     *
     * @param writeComments
     */
    public void setWriteComments(boolean writeComments)
    {
        this.writeComments = writeComments;
    }
}