package ca.ualberta.proteus.util.io;

import ca.ualberta.proteus.logging.ProteusLogger;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.io.CountedBufferedReader;
import org.biojava.utils.io.RandomAccessReader;

import java.io.*;
import java.util.HashMap;
import java.util.Map;

/**
 * (C) Zigtag Inc. 2007
 * <p/>
 * User: scott
 * Date: May 17, 2007
 * Time: 11:42:50 AM
 */
public class StructureDBReader
{
    private StructureDB db;
    private boolean lazyParsing;
    private ProteusLogger logger = ProteusLogger.getLogger();


    // Map to store sequence to their file pointers so we don't store the full sequence in the database,
    // which is a waste of memory and time
    protected Map<String, Long> filePointers = new HashMap<String, Long>();


    public StructureDBReader(StructureDB db)
    {
        this(db, true);
    }

    public StructureDBReader(StructureDB db, boolean lazyParsing)
    {
        this.db = db;
        this.lazyParsing = lazyParsing;
    }

    @SuppressWarnings({"PointlessArithmeticExpression"})
    public void parseDB(InputStream inputStream) throws IOException
    {
        if ( !( inputStream instanceof FileInputStream))
        {
            lazyParsing = false;
        }

        logger.log("Reloading structure DB...");
        long currentTime = System.currentTimeMillis();
        db.clear();

        CountedBufferedReader reader = new CountedBufferedReader(new InputStreamReader(inputStream));

        String line;
        StringBuilder proteinString = null;
        StringBuilder annotationString = null;

        String name = null;
        long currentFilePointer = 0;
        while ((line = reader.readLine()) != null)
        {
            if ( ! lazyParsing )
            {
                // strip comments
                if (line.indexOf('#') != -1)
                {
                    line = line.substring(0, line.indexOf('#'));
                }

                // Allow spaces sk
                line = line.trim();
            }

            if (line.length() > 1)
            {

                // get the name.  This also means its the end of the last sequenc
                if (line.charAt(0) == '>')
                {
                    if (proteinString != null && proteinString.length() > 4)
                    {
                        if (!lazyParsing)
                        {
                            // If it's not read-only, then store the sequence
                            endSequence(proteinString, annotationString, name);

                        }
                    }

                    // start again
                    name = line.substring(1).trim();

                    if ( ! lazyParsing )
                    {
                        proteinString = new StringBuilder(100);
                        annotationString = new StringBuilder(100);
                    }
                    
                    if (lazyParsing && inputStream instanceof FileInputStream )
                    {
                        // Only store the file pointers if this is a Read-only database, and if we're reading from a file
                        filePointers.put(name, currentFilePointer);
                        db.addEmptySequence(name);
                    }
                }
                else
                {
                    if (!lazyParsing)
                    {
                        if (proteinString != null && annotationString != null)
                        {
                            // append a line of sequence/annotation
                            proteinString.append(line);
                            annotationString.append(reader.readLine());
                        }
                        else
                        {
                            throw new IOException("Incorrect file format");
                        }
                    }
                }
            }

            currentFilePointer = reader.getFilePointer();                        
        }

        if (proteinString != null && proteinString.length() > 4)
        {
            endSequence(proteinString, annotationString, name);
        }

        logger.log(String.format("Done (%s ms)", System.currentTimeMillis() - currentTime));
    }

    private Sequence endSequence(StringBuilder proteinString, StringBuilder annotationString, String name)
    {
        if (proteinString != null)
        {  // not the first sequence
            despace(proteinString);
            despace(annotationString);
            try
            {
                StructureAnnotatedSequence annotatedSequence = new StructureAnnotatedSequence(proteinString,
                        annotationString,
                        name);

                //  if ( sequences.containsKey(name) )
                //      System.err.println("SeqSee contains duplicate sequence " + name);
                db.addSequence(annotatedSequence);

                return annotatedSequence;
            }
            catch (IllegalSymbolException e)
            {
                System.err.println("Protein with ID " + name + " is invalid");
                e.printStackTrace();
            }
        }

        return null;
    }

    private void despace(StringBuilder proteinString)
    {
        for (int i = 0; i < proteinString.length();)
            if (proteinString.charAt(i) == ' ')
                proteinString.deleteCharAt(i);
            else
                i++;
    }

    /**
     * @param sequenceID
     * @return the sequence fetched from a file, or null if it doesn't exist
     */
    public Sequence fetchSequence(File file, String sequenceID)
    {
        Long filePointer = filePointers.get(sequenceID);

        if( filePointer == null ) {
        	for( String s : filePointers.keySet() ) {
        		if( s.startsWith(sequenceID) )
        			filePointer = filePointers.get(s);
        	}
        }
        
        if (filePointer == null)
        {
            return null;
        }

        try
        {
            RandomAccessReader reader = new RandomAccessReader(new RandomAccessFile(file, "r"));
            CountedBufferedReader buf = new CountedBufferedReader(reader);

            try
            {
                reader.seek(filePointer);

                String line;
                StringBuilder proteinString = new StringBuilder();
                StringBuilder annotationString = new StringBuilder();

                String name = null;
                while ((line = buf.readLine()) != null)
                {
                    // strip comments
                    if (line.indexOf('#') != -1)
                    {
                        line = line.substring(0, line.indexOf('#'));
                    }

                    // Allow spaces sk
                    line = line.trim();

                    if (line.length() > 1)
                    {

                        // get the name.  This also means its the end of the last sequenc
                        if (line.charAt(0) == '>')
                        {
                            if (proteinString.length() > 4)
                            {
                                // store the sequence
                                return endSequence(proteinString, annotationString, name);
                            }
                            name = line.substring(1);
                        }
                        else
                        {
                            // append a line of sequence/annotation
                            proteinString.append(line);
                            annotationString.append(buf.readLine());
                        }
                    }
                }

                if (proteinString != null && proteinString.length() > 4)
                {
                    return endSequence(proteinString, annotationString, name);
                }
            }
            finally
            {
                reader.close();
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }

        return null;
    }

    public int size()
    {
        return filePointers.size();
    }
}
