package ca.ualberta.proteus.web;

import ca.ualberta.proteus.sequence.GappedStructureAnnotatedSequence;
import ca.ualberta.proteus.sequence.ScoreAnnotatedSequence;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.ProteusTools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.io.SymbolTokenization;
import org.biojava.bio.symbol.*;
import org.biojava.bio.BioException;

import java.util.List;
import java.util.ArrayList;

/**
 * Created by IntelliJ IDEA.
 * User: Smontgomerie
 * Date: Jun 22, 2005
 * Time: 10:52:12 PM
 * To change this template use File | Settings | File Templates.
 */
@SuppressWarnings({"StringConcatenationInsideStringBufferAppend"})
public class ProteusFormatter
{
    private static final int SEQ_LEN = 60;
    public static final String BETA_COLOR = "blue";
    public static final String HELIX_COLOR = "red";
    public static final String TM_COLOR = "green";
    public static final String DEFAULT_COLOR = "black";
    public static final String CLEAVAGE_SITE_COLOR = "#8000FF";
    public static final String SIGNAL_COLOR = "#FF00FF";

    public static String printFormattedSequence(StructureAnnotatedSequence sequence, boolean isTM, boolean isBBTM, boolean printScore )
    {
        assert(sequence != null);

        StringBuilder result = new StringBuilder("<font color = \"black\">");
        String _space = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";
        _space = _space.intern();
        for (int j = 1; j <= sequence.length(); j += SEQ_LEN)
        {
            result.append("<code>");

            if (j < 10)
            {
                result.append("&nbsp;&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 100)
            {
                result.append("&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 1000)
            {
                result.append("&nbsp;" + j + " &nbsp;");
            }
            else if (j < 10000)
            {
                result.append(j + " &nbsp;");
            }
            int length = j + SEQ_LEN - 1;

            if (sequence.length() < length)
            {
                length = sequence.length();
            }
            result.append("<font color = \"black\">" + sequence.subStr(j, length) + " " + length + "</font><br>\n");
            result.append(_space);
            result.append("</code>");
            result.append("<code>\n" + formatPrediction(sequence.getStructureAnnotation().subList(j, length), isTM, isBBTM) + "\n</code>\n");

            if (printScore && (sequence instanceof ScoreAnnotatedSequence))
            {
                result.append("<code>" + _space + " ");
                result.append(formatScore(((ScoreAnnotatedSequence) sequence).getScoresAsStr().substring(j - 1, length)));
                result.append("</code><br>\n");
            }
            result.append("<br>\n");
        }

        return result.toString();
    }

    /* Formats a score in color-code HTML */
    private static String formatScore(String score)
    {
        //TODO
        return score;
    }


    private static String getColor(char residue)
    {
        switch (residue)
        {
            case 'B':
                return TM_COLOR;
            case 'E':
                return BETA_COLOR;
            case 'H':
                return HELIX_COLOR;
            case 'T':
                return TM_COLOR;
            case 'S':
                return SIGNAL_COLOR;
            case 'c':
                return CLEAVAGE_SITE_COLOR;                    
            default:
                return DEFAULT_COLOR;
        }

    }


    /**
     * Formats a structure annotation into HTML color-highlighted
     *
     * @param structure
     * @return a string that is HTML-formatted prediction
     */
    public static String formatPrediction(SymbolList structure, boolean isTM, boolean isBBTM)
    {
        try
        {
            String structureStr = null;
            try
            {
                structureStr = structure.seqString();
            }
            catch (Exception e)
            {
                // Fix for "No mapping for symbol []" bug
                structure = cleanSequence(structure);
                structureStr = structure.seqString();
            }

            return formatPrediction(structureStr, isTM, isBBTM);
        }
        catch(Exception e)
        {
            e.printStackTrace();
            return "";
        }
    }

    /**
     * Formats a structure annotation into HTML color-highlighted
     *
     * @param structureStr A string representation of a structure in CHE format
     * @return a string that is HTML-formatted prediction
     */
    private static String formatPrediction(String structureStr, boolean isTM, boolean isBBTM)
    {
        structureStr = uppercaseStructure(structureStr, isBBTM);

        StringBuilder result = new StringBuilder();

        // define starting color
        char firstChar = structureStr.charAt(0);
        result.append("<font color = \"" + getColor(firstChar) + "\">" + (isTM ? getTMResidue(firstChar) : firstChar));

        for (int i = 1; i < structureStr.length(); i++)
        {
            char residue = structureStr.charAt(i);
            char previous = structureStr.charAt(i - 1);
            if (residue != previous)
            {
                result.append("</font><font color = \"" + getColor(residue) + "\">");
            }

            if ( isTM )
            {
                result.append( getTMResidue(residue) );
            }
            else
                result.append(residue);

            if (i % SEQ_LEN == 0)
            {
                //result.append("</font>\n");
                result.append("\n");
            }
        }
        result.append("<br>");
        result.append("</font>\n");

        return result.toString();
    }

    private static String uppercaseStructure(String structureStr, boolean isBBTM)
    {
        structureStr = structureStr.toUpperCase();

        // change cleavage site
        structureStr = structureStr.replace("SC", "Sc");

        if(isBBTM)
        {
        	//structureStr = structureStr.replace("E", "B");
        }
        return structureStr;
    }

    private static char getTMResidue(char residue)
    {
        switch( residue)
        {
            case 'B':
                return 'B';
            case 'E':
                return 'E';
            case 'H':
                return 'H';
            case 'T':
                return 'T';
            case 'C':
                return 'C';
            case 'c':
                return 'c';
            case 'S':
                return 'S';
            case ' ':
                return '-';
            case '~':
            case '-':
                return residue;
            default:
                throw new IllegalArgumentException( residue + " not implemented");
        }
    }

    public static String printSequence(Sequence sequence)
    {
        String seqString = sequence.seqString();
        return printSequence(seqString);
    }

    public static String printSequence(String seqString)
    {
        StringBuilder result = new StringBuilder();
        for (int i = 0; i < seqString.length(); i += SEQ_LEN)
        {
            int length = i + SEQ_LEN;
            if (seqString.length() < length)
            {
                length = seqString.length();
            }
            result.append(seqString.substring(i, length));
            result.append("<br>\n");
        }

        return result.toString();
    }

    public static String formatDescription(String description)
    {
        if (description.indexOf("gi|") == -1)
        {
            return description;
        }
        else
        {
            StringBuilder result = new StringBuilder(description);
            int start, end;
            start = description.indexOf("gi|") + 2;
            end = description.indexOf("|", start + 1);
            String link = description.substring(start + 1, end);
            result.delete(start + 1, end);
            result.insert(start + 1, "<a href=\"http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?db=protein&val=" + link + "\">" + link + "</a>");
            return result.toString();
        }
    }

    /**
     * @param query
     * @param homolog
     * @return a HTML-formatted alignment between two sequences
     */
    public static String printAlignment(GappedStructureAnnotatedSequence query, GappedStructureAnnotatedSequence homolog, boolean isTM, boolean isBBTM)
    {
        String querySeq = query.seqString();
        SymbolList queryStruct = query.getStructureAnnotation();
        String homologSeq = homolog.seqString();
        SymbolList homologStruct = homolog.getStructureAnnotation();
        StringBuilder result = new StringBuilder();
        final String _space = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ";

        // clean up alignment output
        StringBuilder buf1 = new StringBuilder(), buf2 = new StringBuilder();
        for (int i = 0; i < Math.max(querySeq.length(), homologSeq.length()); i ++ )
        {
            char qChar =querySeq.charAt(i);
            char hChar =homologSeq.charAt(i);
            if ((qChar != '-' && qChar != '~' )
                || (hChar != '-' && hChar != '~' ))
            {
                buf1.append(qChar);
                buf2.append(hChar);
            }
        }

        querySeq = buf1.toString();
        homologSeq = buf2.toString();

        for (int i = 0; i < Math.max(querySeq.length(), homologSeq.length()); i += SEQ_LEN)
        {
            int j = i + 1;
            if (i == 0)
            {
                result.append("<b>QUERY</b>\n");
            }
            else
            {
                if (j < 10)
                {
                    result.append("&nbsp;&nbsp;&nbsp;" + j + " &nbsp;");
                }
                else if (j < 100)
                {
                    result.append("&nbsp;&nbsp;" + j + " &nbsp;");
                }
                else if (j < 1000)
                {
                    result.append("&nbsp;" + j + " &nbsp;");
                }
                else if (j < 10000)
                {
                    result.append(j + " &nbsp;");
                }
            }

            int length = i + SEQ_LEN;
            if (querySeq.length() < length)
            {
                length = querySeq.length();
            }
            result.append(querySeq.substring(i, length));
            result.append(" " + length);
            result.append("<br>\n");
            //result.append(_space);
            //result.append( formatPrediction( queryStruct.subList(i+1, length) ) );
            //result.append("<br>\n");
            if (i == 0)
            {
                String name = homolog.getName();
                if (name.length() > 5)
                    name = name.substring(0, 5);
                else if (name.length() < 5)
                    name = name + "&nbsp;";
                result.append("<b>" + name + "</b>\n");
            }
            else
            {
                result.append(_space);
            }
            if (homologSeq.length() < length)
            {
                length = homologSeq.length();
            }
            result.append(homologSeq.substring(i, length));
            result.append("<br>\n");
            result.append(_space);
            result.append(formatPrediction(homologStruct.subList(i + 1, length), isTM, isBBTM) );
            result.append("<br><br>\n");
        }

        return result.toString();
    }

    /**
     * Pads out a string upto padlen with pad chars
     *
     * @param str to be padded
     * @param length            of pad (+ve = pad on right, -ve pad on left)
     * @param pad               character
     */
    public static String pad(Object str, int length, String pad)
    {
        String padding = "";
        int len = Math.abs(length) - str.toString().length();
        if (len < 1)
            return str.toString();
        for (int i = 0; i < len; ++i)
            padding = padding + pad;

        return (length < 0 ? padding + str : str + padding);
    }

    public static String printAlignment(GappedStructureAnnotatedSequence query, List<GappedStructureAnnotatedSequence> alignedHomologs, boolean isTM, boolean isBBTM)
    {
        String querySeq = query.seqString().replace('X', '-');
        String queryStruct = query.getStructureAnnotation().seqString().replace('X', '-');
        StringBuilder result = new StringBuilder();
        final String _space = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; ";

        List<String> alignedHomologStrings = new ArrayList<String>();
        List<String> alignedStructStrings = new ArrayList<String>();

        // clean up alignment output
        for ( GappedStructureAnnotatedSequence sequence : alignedHomologs )
        {
            alignedHomologStrings.add(sequence.seqString().replace('X', '-'));
            alignedStructStrings.add(sequence.getStructureAnnotation().seqString().replace('X', '-'));
        }

        StringBuilder queryStringBuffer = new StringBuilder();
        StringBuilder queryStructStringBuffer = new StringBuilder();
        StringBuilder bufList[] = new StringBuilder[alignedHomologStrings.size()];
        StringBuilder structBufList[] = new StringBuilder[alignedHomologStrings.size()];

        // i iterates through each character, cleaning up gaps that permeate through the alignment
        for (int i = 0; i < querySeq.length(); i ++ )
        {
            boolean done = false;
            for ( int seqID=0; seqID<alignedHomologStrings.size(); seqID++ )
            {
                if ( bufList[seqID] == null )
                    bufList[seqID] = new StringBuilder();
                if ( structBufList[seqID] == null )
                    structBufList[seqID] = new StringBuilder();

                char qChar =querySeq.charAt(i);
                char qStructChar = queryStruct.charAt(i);
                String seqString = alignedHomologStrings.get(seqID);
                String structString = alignedStructStrings.get(seqID);
                if ( i >= seqString.length() )
                {
                    done = true;
                    continue;
                }
                char hChar =seqString.charAt(i);
                char sChar = structString.charAt(i);


                /*
                if ((qChar != '-' && qChar != '~' && qChar != 'X' )
                    || (hChar != '-' && hChar != '~'  && hChar != 'X'))
                {
                */
                    if ( seqID == 0 )
                    {
                        queryStringBuffer.append(qChar);
                        queryStructStringBuffer.append(qStructChar);
                    }

                    bufList[seqID].append(hChar);
                    structBufList[seqID].append(sChar);
                /*
                }
                */
            }
            if (done) break;
        }

        // put fixed strings back in homolog list
        alignedHomologStrings.clear();
        for ( StringBuilder buf: bufList)
        {
            alignedHomologStrings.add(buf.toString());
        }
        alignedStructStrings.clear();
        for ( StringBuilder buf: structBufList )
        {
            alignedStructStrings.add(buf.toString());
        }
        querySeq = queryStringBuffer.toString();

        // output the alignment
        for (int i = 0; i < querySeq.length(); i += SEQ_LEN)
        {
            if (i == 0)
            {
                result.append("<b>QUERY</b>\n");
            }
            else
            {
                // here we want the number of residues, not the length
                int resCount = 0;
                for(int k=0; k < i + 1; k++)
                {
                    if ( ! ProteusTools.isGap(querySeq.charAt(k)))
                    {
                        resCount++;
                    }
                }

                if (resCount < 10)
                {
                    result.append("&nbsp;&nbsp;&nbsp;" + resCount + " &nbsp;");
                }
                else if (resCount < 100)
                {
                    result.append("&nbsp;&nbsp;" + resCount + " &nbsp;");
                }
                else if (resCount < 1000)
                {
                    result.append("&nbsp;" + resCount + " &nbsp;");
                }
                else if (resCount < 10000)
                {
                    result.append(resCount + " &nbsp;");
                }
            }

            int length = i + SEQ_LEN;
            if (querySeq.length() < length)
            {
                length = querySeq.length();
            }
            result.append(querySeq.substring(i, length));

            // here we want the number of residues, not the length
            int resCount = 0;
            for(int k=0; k < length; k++)
            {
                if ( ! ProteusTools.isGap(querySeq.charAt(k)))
                {
                    resCount++;
                }
            }
            result.append(" " + resCount);
            result.append("<br>\n");

            // don't want the query's structure in there
            //result.append(_space);
            //result.append(formatPrediction(queryStruct.substring(i, length).replace('-', '~'), isTM));
            
            //result.append(_space);
            //result.append( formatPrediction( queryStruct.subList(i+1, length) ) );
            //result.append("<br>\n");
            for (int seqId=0; seqId<alignedHomologs.size(); seqId++) {
                //if (i == 0)
                //{
                    String name = alignedHomologs.get(seqId).getName();
                    if (name.length() > 5)
                        name = name.substring(0, 5);
                    else if (name.length() < 5)
                        name = name + "&nbsp;";
                    result.append("<b>" + name + "</b>\n");
                //}
                //else
                //{
                //    result.append(_space);
                //}
                if (alignedHomologStrings.get(seqId).length() < length)
                {
                    length = alignedHomologStrings.get(seqId).length();
                }
                if ( i < length )
                {
                    result.append(alignedHomologStrings.get(seqId).substring(i, length).replace('X', '-'));
                    result.append("<br>\n");
                    result.append(_space);
                    result.append(formatPrediction(alignedStructStrings.get(seqId).substring(i, length).replace('-', '~'), isTM, isBBTM));
                }
            }
            result.append("<br><br>\n");
        }

        return result.toString();

    }

    public static String printAlignmentPlainText(GappedStructureAnnotatedSequence query, List<GappedStructureAnnotatedSequence> alignedHomologs, boolean isTM, boolean isBBTM)
    {
        String s = printAlignment(query, alignedHomologs, isTM, isBBTM);

        // add an extra space as a hack to make it work
        s = " " + convertToPlainText(s);

        return s;
    }

    public static String formatPredictionPlainText(String structureStr, boolean isTM, boolean isBBTM)
    {
        String s = formatPrediction(structureStr, isTM, isBBTM);

        s = convertToPlainText(s);

        return s;
    }

    public static String convertToPlainText(String s)
    {
        // First remove all \n's so it's just HTML
        s = s.replace("\n", " ");
        s = s.replace("&nbsp;", " ");
        // s = s.replace("<br>\n", "\n"); 
        s = s.replace("<br>", "\n");
        s = s.replace("</br>", "\n");
        s = s.replace("<b>", "");
        s = s.replace("</b>", "");
        s = s.replace("<code>", "");
        s = s.replace("</code>", "");
        s = s.replaceAll("<font[^>]*>", "");
        s = s.replace("</font>", "");
        return s;
    }

    public static String printFormattedSequenceWithHomologyIndicator(ScoreAnnotatedSequence sequence, GappedStructureAnnotatedSequence gappedQuery, List<GappedStructureAnnotatedSequence> gappedHomologs, Boolean isTM, Boolean isBBTM)
    {
        assert(sequence != null);

        String query = gappedQuery.seqString();
        List<String> homologs  = new ArrayList<String>(gappedHomologs.size());

        for(GappedStructureAnnotatedSequence s : gappedHomologs)
        {
            if ( ! s.getName().equals("QUERY"))
            {
                String element = s.seqString();
                element = element.replaceAll("X", "-");
                homologs.add( element );
            }
        }

        query = query.replaceAll("X", "-");

        // clean up alignment output
        StringBuilder querySequenceBuffer = new StringBuilder(), homologBuffer = new StringBuilder();

        for ( int seqID =0; seqID < homologs.size(); seqID++ )
        {
            homologBuffer.setLength(0);
            for (int i = 0; i < Math.max(query.length(), homologs.get(seqID).length()); i ++ )
            {
                char qChar =query.charAt(i);
                char hChar = homologs.get(seqID).charAt(i);
                if ((qChar != '-' && qChar != '~' )
                    || (hChar != '-' && hChar != '~' ))
                {
                    if ( seqID == 0 )
                    {
                        querySequenceBuffer.append(qChar);
                    }
                    homologBuffer.append(hChar);
                }
            }
            homologs.set(seqID, homologBuffer.toString() );
        }

        query= querySequenceBuffer.toString();

        StringBuilder overlappingRegions = new StringBuilder();
        for ( int i=0; i<query.length(); i++)
        {
            char character = ' ';
            for (String homolog : homologs)
            {
                if (query.charAt(i) != '-'
                        && query.charAt(i) != '~'
                        && homolog.length() > i 
                        && homolog.charAt(i) != '-'
                        && homolog.charAt(i) != '~' )
                {
                    // sequences overlap
                    character = '*';

                    // at least one sequence overlaps here, break.
                    break;
                }
                else/* if( query.charAt(i) != '-' )*/
                {
                    // query sequence is by itself
                    character = ' ';
                }
            }
            overlappingRegions.append(character);
        }


        StringBuilder result = new StringBuilder("<font color = \"black\">");
        String _space = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";
        _space = _space.intern();
        for (int j = 1; j <= sequence.length(); j += SEQ_LEN)
        {
            result.append("<code>");

            if (j < 10)
            {
                result.append("&nbsp;&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 100)
            {
                result.append("&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 1000)
            {
                result.append("&nbsp;" + j + " &nbsp;");
            }
            else if (j < 10000)
            {
                result.append(j + " &nbsp;");
            }
            int length = j + SEQ_LEN - 1;

            if (sequence.length() < length)
            {
                length = sequence.length();
            }

            // get the correct length of the overlaping region
            String overlapStr = overlappingRegions.substring(j - 1, length);

            // change it to proper HTML
            overlapStr = overlapStr.replaceAll(" ", "&nbsp;");

            result.append("<font color = \"black\">" + overlapStr + " " + length + "</font><br>\n");
            result.append(_space + "&nbsp;");
            result.append("<font color = \"black\">" + sequence.subStr(j, length) + "</font><br>\n");
            result.append(_space);
            result.append("</code>");
            result.append("<code>\n" + formatPrediction(sequence.getStructureAnnotation().subList(j, length), isTM, isBBTM) + "\n</code>\n");

            result.append("<code>" + _space + " ");
            result.append(formatScore(sequence.getScoresAsStr().substring(j - 1, length)));
            result.append("</code><br>\n");
            result.append("<br>\n");
        }

        return result.toString();
    }

    public static String printFormattedSequenceWithHomologyIndicatorPlainText(ScoreAnnotatedSequence sequence, GappedStructureAnnotatedSequence gappedQuery, List<GappedStructureAnnotatedSequence> gappedHomologs, Boolean isTM, Boolean isBBTM)
    {
        assert(sequence != null);

        String query = gappedQuery.seqString();
        List<String> homologs  = new ArrayList<String>(gappedHomologs.size());

        for(GappedStructureAnnotatedSequence s : gappedHomologs)
        {
            if ( ! s.getName().equals("QUERY"))
            {
                String element = s.seqString();
                element = element.replaceAll("X", "-");
                homologs.add( element );
            }
        }

        query = query.replaceAll("X", "-");

        // clean up alignment output
        StringBuilder querySequenceBuffer = new StringBuilder(), homologBuffer = new StringBuilder();

        for ( int seqID =0; seqID < homologs.size(); seqID++ )
        {
            homologBuffer.setLength(0);
            for (int i = 0; i < Math.max(query.length(), homologs.get(seqID).length()); i ++ )
            {
                char qChar =query.charAt(i);
                char hChar = homologs.get(seqID).charAt(i);
                if ((qChar != '-' && qChar != '~' )
                    || (hChar != '-' && hChar != '~' ))
                {
                    if ( seqID == 0 )
                    {
                        querySequenceBuffer.append(qChar);
                    }
                    homologBuffer.append(hChar);
                }
            }
            homologs.set(seqID, homologBuffer.toString() );
        }

        query= querySequenceBuffer.toString();

        StringBuilder overlappingRegions = new StringBuilder();
        for ( int i=0; i<query.length(); i++)
        {
            char character = ' ';
            for (String homolog : homologs)
            {
                if (query.charAt(i) != '-'
                        && query.charAt(i) != '~'
                        && homolog.length() > i
                        && homolog.charAt(i) != '-'
                        && homolog.charAt(i) != '~' )
                {
                    // sequences overlap
                    character = '*';

                    // at least one sequence overlaps here, break.
                    break;
                }
                else/* if( query.charAt(i) != '-' )*/
                {
                    // query sequence is by itself
                    character = ' ';
                }
            }
            overlappingRegions.append(character);
        }


        StringBuilder result = new StringBuilder("");
        String _space = "";
        _space = _space.intern();
        for (int j = 1; j <= sequence.length(); j += SEQ_LEN)
        {
           /* if (j < 10)
            {
                result.append("   " + j + "  ");
            }
            else if (j < 100)
            {
                result.append("  " + j + "  ");
            }
            else if (j < 1000)
            {
                result.append(" " + j + "  ");
            }
            else if (j < 10000)
            {
                result.append(j + "  ");
            }*/
            int length = j + SEQ_LEN - 1;

            if (sequence.length() < length)
            {
                length = sequence.length();
            }

            // get the correct length of the overlaping region
            String overlapStr = overlappingRegions.substring(j - 1, length);

            // change it to proper HTML
            // overlapStr = overlapStr.replaceAll(" ", "&nbsp;");

            result.append(overlapStr + "\n");
            // result.append(_space + " ");
            result.append(sequence.subStr(j, length) + "\n");
            // result.append(_space);
            result.append(uppercaseStructure(sequence.getStructureAnnotation().subList(j, length).seqString(), isBBTM) + "\n");

            //result.append(_space + " ");
            result.append(formatScore(sequence.getScoresAsStr().substring(j - 1, length)));
            result.append("\n\n");
        }

        return result.toString();
    }

    public static String printFormattedSequencePlainText(StructureAnnotatedSequence sequence, boolean isTM, boolean isBBTM, boolean printScore )
    {
        assert(sequence != null);

        StringBuilder result = new StringBuilder();
        String _space = "";
        _space = _space.intern();
        for (int j = 1; j <= sequence.length(); j += SEQ_LEN)
        {

           /* if (j < 10)
            {
                result.append("&nbsp;&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 100)
            {
                result.append("&nbsp;&nbsp;" + j + " &nbsp;");
            }
            else if (j < 1000)
            {
                result.append("&nbsp;" + j + " &nbsp;");
            }
            else if (j < 10000)
            {
                result.append(j + " &nbsp;");
            }*/
            int length = j + SEQ_LEN - 1;

            if (sequence.length() < length)
            {
                length = sequence.length();
            }
            result.append(sequence.subStr(j, length));
            result.append(_space);
            String structure = sequence.getStructureAnnotation().subList(j, length).seqString();
            result.append("\n" + uppercaseStructure(structure, isBBTM) + "\n");

            if (printScore && (sequence instanceof ScoreAnnotatedSequence))
            {
                result.append(((ScoreAnnotatedSequence) sequence).getScoresAsStr().substring(j - 1, length));
                result.append("\n");
            }
            result.append("\n");
        }

        return result.toString();
    }

    /**
     * Verifies that the symbolist will actually tokenize correctly.  Mostly a hack to fix the
     *  No mapping for symbol [] problem.
     * @param symbols
     * @return
     */
    public static SimpleSymbolList cleanSequence(SymbolList symbols)
    {
        Alphabet alphabet = symbols.getAlphabet();
        SymbolTokenization tokenization = null;
        try
        {
            tokenization = alphabet.getTokenization("default");
        }
        catch (BioException e)
        {
            e.printStackTrace();
        }

        List list = new ArrayList(symbols.toList());

        for ( int i=0; i<list.size(); i++)
        {
            Symbol symbol = (Symbol) list.get(i);
            try
            {
                tokenization.tokenizeSymbol(symbol);
            }
            catch (IllegalSymbolException e)
            {
                list.set(i, alphabet.getGapSymbol());
            }
        }

        Symbol[] s = new Symbol[list.size()];
        return new SimpleSymbolList((Symbol[]) list.toArray(s), list.size(), alphabet);
    }
}
