package ca.ualberta.proteus.web;

import ca.ualberta.proteus.db.PersistenceQueue;
import ca.ualberta.proteus.db.model.ProteusSettingsIntf;
import ca.ualberta.proteus.logging.RuntimeCatalog;
import ca.ualberta.proteus.program.ProteusImpl;
import ca.ualberta.proteus.program.ProteusSettings;
import ca.ualberta.proteus.program.transsec.Config;
import ca.ualberta.proteus.sequence.StructureAnnotatedSequence;
import ca.ualberta.proteus.util.CompletionListener;
import ca.ualberta.proteus.util.Consts;
import ca.ualberta.proteus.util.EmailUtils;
import ca.ualberta.proteus.util.ProteusTools;
import ca.ualberta.proteus.web.db.SQLWebHitCounter;
import org.apache.log4j.Logger;
import org.biojava.bio.symbol.SymbolList;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;

public class SubmissionQueue extends Thread
{
    private static Logger logger = Logger.getLogger(SubmissionQueue.class);

    static final int interval = 500;
    static final int PROCESS_TIMEOUT = 60 * 60 * 1000; // One hour process timeout in millis
    private final Queue<ProteusImpl> queue = new LinkedBlockingQueue<ProteusImpl>();

    private static SubmissionQueue instance = new SubmissionQueue();
    private ProteusImpl currentThread = null;

    static PersistenceQueue persistenceQueue = new PersistenceQueue();


    /**
     * Make constructor private
     */
    private SubmissionQueue()
    {
        // Recover
        recover();

        this.setDaemon(true);
        this.start();
    }

    /**
     * Recover if we crashed or were shut down
     */
    private void recover()
    {
        if (persistenceQueue == null)
        {
            persistenceQueue = new PersistenceQueue();
        }
        List<ProteusImpl> list = persistenceQueue.recover();

        if (list != null)
        {
            logger.info("Resuming " + list.size() + " recovered proteus sessions ");

            if (list.size() > 0)
            {
                final Map<String, List<ProteusImpl>> emailMap = new HashMap<String, List<ProteusImpl>>();
                final Map<String, Map<String, StructureAnnotatedSequence>> idMap = new HashMap<String, Map<String, StructureAnnotatedSequence>>();


                SQLWebHitCounter webHitCounter = (SQLWebHitCounter) WebHitCounter.getInstance();
                // Collection<WebHitCounter.WebHit> hits = webHitCounter.get(WebHitCounter.WebHit.Type.SUBMISSION).values();

                // Add all proteus sessions to the queue.
                for (int i1 = 0; i1 < list.size(); i1++)
                {
                    final ProteusImpl proteus = list.get(i1);
                    int percent;
                    if ( (percent = i1 * 100 / list.size()) % 10 == 0 )
                    {
                        System.out.println(percent + "%");
                    }

                    WebHitCounter.WebHit hit = webHitCounter.getHit(Integer.parseInt(proteus.getID()));
                    if (hit != null)
                    {
                        // There's a matching hit for the email in the DBfindf
                        String email = hit.getEmail();
                        if (email != null && hit.getEmail().length() > 0)
                        {
                            // Get a list of proteus runs for this email
                            List<ProteusImpl> proteusList = emailMap.get(hit.getEmail());
                            if (proteusList == null)
                            {
                                proteusList = new ArrayList<ProteusImpl>();
                                emailMap.put(hit.getEmail(), proteusList);
                            }

                            // Add this one to the list
                            proteusList.add(proteus);
                        }

                        final String emailCopy = email;
                        proteus.addCompletionListener(new CompletionListener()
                        {
                            public void onComplete(boolean successful, String message)
                            {
                                RuntimeCatalog logger = proteus.getLogger();

                                StructureAnnotatedSequence finalSequence = (StructureAnnotatedSequence) logger.getLogObject(Consts.FINAL_WITH_HOMOLOGY);
                                if (finalSequence == null)
                                {
                                    finalSequence = (StructureAnnotatedSequence) logger.getLogObject(Consts.FINAL);
                                }

                                Boolean isBBTM = (Boolean) logger.getLogObject(Consts.IS_BBTM);
                                if (isBBTM)
                                {
                                    // Do replacement on structure
                                    SymbolList structure = finalSequence.getStructureAnnotation();
                                    String structStr = structure.seqString().replace("E", "B");
                                    finalSequence.setStructureAnnotation(ProteusTools.createStructure(structStr));
                                }

                                Map<String, StructureAnnotatedSequence> proteusIDs = idMap.get(emailCopy);
                                if (proteusIDs == null)
                                {
                                    proteusIDs = new HashMap<String, StructureAnnotatedSequence>();
                                    idMap.put(emailCopy, proteusIDs);
                                }
                                proteusIDs.put(proteus.getID(), finalSequence);

                                // store session
                                try
                                {
                                    File storageDirectory = new File(Config.getConfig().getStringValue("storageDir"));
                                    if (!storageDirectory.exists())
                                    {
                                        storageDirectory.mkdirs();
                                    }

                                    ProteusSettingsIntf settings = proteus.getSettings();
                                    // HACK : The settings used by the ActiveObjects class aren't serializable.  So build
                                    // a new serializable version
                                    proteus.getLogger().logObject(Consts.SETTINGS, new ProteusSettings(settings));


                                    PersistentProteusResultStorage.store(proteus.getID(), proteus.getLogger(), storageDirectory);
                                }
                                catch (IOException e)
                                {
                                    e.printStackTrace();
                                }
                                catch (ClassNotFoundException e)
                                {
                                    e.printStackTrace();
                                }

                                // TODO this is the problem, we never remove it from emailMap

                                // if the list of sequences for this email is empty, fire off an email
                                List<ProteusImpl> proteusList = emailMap.get(emailCopy);
                                if (proteusList != null)
                                {
                                    // Remove this run from the list
                                    for (int i = 0; i < proteusList.size(); i++)
                                    {
                                        ProteusImpl p = proteusList.get(i);
                                        if (p.getID().equals(proteus.getID()))
                                        {
                                            proteusList.remove(i);
                                        }
                                    }

                                    if (proteusList.size() == 0)
                                    {
                                        // All done for this list of runs for this email, send the message
                                        String msg = MultipleSubmissionHandler.buildMessage(proteusList, proteusIDs, 0, null, null);
                                        String returnEmail = Config.getConfig().getStringValue("email");

                                        EmailUtils.sendMessage(returnEmail, emailCopy, Consts.APPNAME + " done", msg);
                                        System.out.println("Sending email to : " + emailCopy);
                                    }
                                }
                            }
                        });

                        queue.add(proteus);
                    }
                }
            }
        }
    }

    public void run()
    {
        //noinspection InfiniteLoopStatement
        while (true)
        {
            try
            {
                final ProteusImpl proteus = queue.peek();
                if (proteus != null && proteus != currentThread)
                {
                    currentThread = proteus;
                    proteus.start();
                    System.out.println(String.format("Proteus (ID=%2$s) started %1$tb %1$te, %1$tY at %1$tT, %3$d in queue", Calendar.getInstance(), proteus.getID(), queue.size()));

                    proteus.addCompletionListener(new CompletionListener()
                    {
                        /* Called when this Proteus thread finishes */
                        public void onComplete(boolean successful, String message)
                        {
                            if (successful)
                            {
                                System.out.println(String.format("Proteus (ID=%2$s) completed %1$tb %1$te, %1$tY at %1$tT", Calendar.getInstance(), proteus.getID()));
                            }
                            else
                            {
                                System.out.println(String.format("Proteus (ID=%2$s) failed %1$tb %1$te, %1$tY at %1$tT", Calendar.getInstance(), proteus.getID()));
                                System.out.println("Message = " + message);
                            }

                        }
                    });

// Wait for proteus to finish
                    proteus.join(PROCESS_TIMEOUT);

                    synchronized (queue)
                    {
                        // Remove from the queue
                        ProteusImpl completedProteus = queue.remove();

// Remove from the persistent queue
                        persistenceQueue.remove(completedProteus);
                    }
                }

                Thread.sleep(interval);
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }
            catch (Throwable e)
            {
                System.err.println("Exception handled");
                e.printStackTrace();
            }
        }
    }

    /**
     * Add proteus implementation to queue
     *
     * @param proteus Thread to add to queue
     */
    public static void addToQueue(ProteusImpl proteus)
    {
        synchronized (instance.queue)
        {
            boolean added = instance.queue.add(proteus);
            if (!added)
            {
                System.err.println("Error adding to queue");
            }

            logger.info(String.format("Proteus (ID=%2$s) submitted to queue %1$tb %1$te, %1$tY at %1$tT, %3$d in queue", Calendar.getInstance(), proteus.getID(), instance.queue.size()));
        }

        // Persist the queue if possible
        persistenceQueue.addToQueue(proteus);
    }

    /**
     * Add proteus implementation to queue
     *
     * @param proteusList Thread to add to queue
     */
    public static void addToQueue(List<ProteusImpl> proteusList)
    {
        synchronized (instance.queue)
        {
            for (ProteusImpl proteus : proteusList)
            {
                boolean added = instance.queue.add(proteus);
                if (!added)
                {
                    System.err.println("Error adding to queue");
                }

                logger.info(String.format("Proteus (ID=%2$s) submitted to queue %1$tb %1$te, %1$tY at %1$tT, %3$d in queue", Calendar.getInstance(), proteus.getID(), instance.queue.size()));
            }
        }

        // Persist the queue if possible
        persistenceQueue.addToQueue(proteusList);
    }

    public static int find(String id)
    {
        Object[] objects = instance.queue.toArray();
        for (int i = 0; i < objects.length; i++)
        {
            if (objects[i] != null && ((ProteusImpl) objects[i]).getID().equals(id))
            {
                return i;
            }
        }
        return -1;
    }

    /**
     * Returns the length of the queue
     *
     * @return the length of the queue
     */
    public static int length()
    {
        return instance.queue.size();
    }

    public static void load()
    {
        if (instance == null)
        {
            instance = new SubmissionQueue();
        }
    }

    public static void clear()
    {
        logger.info("Clearing queue");

        persistenceQueue.clear();

        instance.queue.clear();
    }

    public static Iterator<ProteusImpl> getEnuerator()
    {
        return instance.queue.iterator();
    }
}
