#!/usr/bin/perl -w

# This is a script to generate the residue template database. This database is
# used by PepMake when it is not given a PDB file and so has to use the default
# residues in the template database.
#
# The script looks through all the PDB files in source_pdb_files directory and
# (partly by using phipsi) compares the chi angle of each residue with the
# target chi angles for each residue type. The target chi angles are set
# in @chi_angles and are derived from the Rotamer Library. Each residue has 3
# possible favored chi angles, except for ALA, GLY and PRO. Although the 3 chi
# angles vary slightly based on the phi and psi angles of the residue, the
# variation appears to be small, so the preferred chi angles for phi = -180
# degrees and psi = -180 degrees were used.
#
# The script finds the residues in the given PDB files that have the closest-
# matching chi angles, saves those residues, and outputs them to the template
# database files (ALA.db, ARG.db, etc.). The data format is that produced by
# phipsi when called with the --db flag; this data format was modified by
# David Arndt in October-November 2006, and so the template database generated
# by this script uses a somewhat different format from the older versions of
# the template database. The previous versions of the template database
# contained several errors and limitations, which prompted the writing of this
# script.
#
# The script outputs a list of the target chi angles, how close the best match
# was, and the source file from which the best match was derived.
#
# Written by David Arndt in November 2006.

use strict;

sub slurp;
sub trim($);

my $pdb_files_dirname = "template_database_creation/source_pdb_files";
my $template_db_dirname = "template_database";
my $rescoords_dirname = "template_database_creation/rescoords_files";
my $reduced_pdb_files_dirname = "template_database_creation/reduced_pdb_files";

my $call_reduce = 1;
	# Whether we should call reduce on the PDB files first; doing so will
	# result in the template residues having hydrogens in the positions where
	# the reduce program places them.

my @res_names = (
	"ALA",
	"ARG",
	"ASN",
	"ASP",
	"CYS",
	"GLN",
	"GLU",
	"GLY",
	"HIS",
	"ILE",
	"LEU",
	"LYS",
	"MET",
	"THR",
	"TRP",
	"SER",
	"PHE",
	"PRO",
	"VAL",
	"TYR"
	);

my %res_name_to_num = ();
for (my $i = 0; $i < scalar(@res_names); $i++) {
	$res_name_to_num{$res_names[$i]} = $i;
}

my @chi_angles = (
	"360 360 360", # ALA
	"60.4 -174.5 -71.2", # ARG
	"62.1 -168.3 -70.9", # ASN
	"61.2 -163.7 -74.1", # ASP
	"64.7 -183.3 -65.7", # CYS
	"63.8 -173.4 -67.3", # GLN
	"59.5 -171.8 -66.6", # GLU
	"360 360 360", # GLY
	"63.2 -171.5 -66.1", # HIS
	"61.1 -168.0 -63.4", # ILE
	"62.9 -171.3 -77.4", # LEU
	"66.2 -171.8 -71.6", # LYS
	"65.2 -176.7 -70.2", # MET
	"61.5 -170.6 -60.3", # THR
	"62.3 -177.2 -68.7", # TRP
	"69.6 -166.0 -63.9", # SER
	"60.5 -175.4 -67.6", # PHE
	"27.0 -25.1 27.0", # PRO
	"65.6 -184.1 -61.8", # VAL
	"61.6 -176.1 -66.5" # TYR
	);

# rescoords_hash uses keys that are the concatenation of the residue
# name and the chi angle index, e.g. "ILE0" or "MET2". Values are the
# entire printouts for that particular residue for the chi angle corresponding
# to the given chi index.
my %rescoords_hash = ();

# Same idea as %rescoords_hash, but the values are the names of the files from which
# the data is derived. Used for debugging and reporting purposes.
my %rescoords_source_file_hash = ();


print(STDERR "Creating template database files... (warnings from the \'reduce\' program may appear below)\n");

opendir(DIR, $pdb_files_dirname) or die "can't opendir $pdb_files_dirname: $!";
while (defined(my $file = readdir(DIR))) {
	next if $file =~ /^\./;     # skip . and .. and invisible files
	
	# Call reduce on the PDB file.
	if ($call_reduce) {
		system("./reduce -build -Quiet ${pdb_files_dirname}/$file > ${reduced_pdb_files_dirname}/${file}H");
	}
	
	# Call phipsi to create residue coordinates files.
	if ($call_reduce) {
		system("./phipsi --db ${reduced_pdb_files_dirname}/${file}H > ${rescoords_dirname}/${file}.rescoords\n");
	}
	else {
		system("./phipsi --db ${pdb_files_dirname}/$file > ${rescoords_dirname}/${file}.rescoords\n");
	}
	
	my $file_contents = slurp("${rescoords_dirname}/${file}.rescoords");
	
	# Separate the data for each residue into different elements in @residues.
	my @lines = split(/\n/, $file_contents);
	my @residues;
	my $rescoords_string;
	my $resnum = -1;
	for (my $i = 0; $i < scalar(@lines); $i++) {
		my $current_line = $lines[$i];
		
		if ($current_line =~ /\b[A-Z]{3}\s+\S+\s+\S+\s+\d+/) {
			
			if ($resnum != -1) {
				$residues[$resnum] = $rescoords_string;
			}
			
			$rescoords_string = $current_line . "\n";
			$resnum++;
		}
		else {
			$rescoords_string .= $current_line . "\n";
		}
		
	}
	$residues[$resnum] = $rescoords_string;
	
	for (my $i = 1; $i < scalar(@residues) - 1; $i++) {
		# Note that we skip the first and last residues.
		
		# Skip residues containing "OXT", i.e. C-terminal residues. Skipping the last
		# element in @residues does not seem to be enough to avoid C-terminal residues.
		if ($residues[$i] =~ /OXT/) {
			next;
		}
		
		if ($residues[$i] =~ /\b([A-Z]{3})\s+(\S+)\s+(\S+)\s+(\d+)/) {
			my $current_res_name = $1;
			my $chi = $2;
			my $N_CA_C1_angle = $3;
			my $num_atoms = $4;
			
			if (!exists($res_name_to_num{$current_res_name})) {
				# Current residue name is non-standard.
				next;
			}
			my $res_index = $res_name_to_num{$current_res_name};
			my $chi_targets_string = $chi_angles[$res_index];
			
			#print "res $current_res_name has targets $chi_targets_string\n";
			
			my @chi_targets = split(/\s+/, $chi_targets_string);
			my $closest_index = -1; # Index (0-2) of which chi angle is closest.
			my $lowest_diff = 1000;
			for (my $j = 0; $j < scalar(@chi_targets); $j++) {
				if (&angle_dist($chi_targets[$j], $chi) < $lowest_diff) {
				
					#$lowest_diff = abs($chi_targets[$j] - $chi);
					$lowest_diff = &angle_dist($chi_targets[$j], $chi);
					$closest_index = $j;
				}
			}
			
			# debug
			#print "   Closest index for $chi is $closest_index.\n";
			#print &angle_dist($chi_targets[$closest_index], $chi) . "\n";
			
			# debug
			#if ($current_res_name =~ /TYR/) {
			#	print "$chi\n";
			#}
			
			my $rescoords_key = $current_res_name . $closest_index;
			
			if (!exists($rescoords_hash{$rescoords_key})) {
				$rescoords_hash{$rescoords_key} = $residues[$i];
				$rescoords_source_file_hash{$rescoords_key} = $file . " " . ($i+1);
			}
			else {
				my $competing_residue = $rescoords_hash{$rescoords_key};
					# The residue that was already saved for this chi angle.
				my $competing_chi;
				if ($competing_residue =~ /\b[A-Z]{3}\s+(\S+)\s+\S+\s+\d+/) {
					$competing_chi = $1;
				}
				
				# If the current residue has a closer chi match, use the current residue.
				if (	&angle_dist($chi_targets[$closest_index], $chi) <
						&angle_dist($chi_targets[$closest_index], $competing_chi) ) {
					
					$rescoords_hash{$rescoords_key} = $residues[$i];
					$rescoords_source_file_hash{$rescoords_key} = $file . " " . ($i+1);
					
					# debug
					#print "Replaced entry for $rescoords_key: $chi_targets[$closest_index]"
					#. " - $chi replaces $chi_targets[$closest_index] - $competing_chi.\n";
				}
				
				
			}
			
		}
	}
	
}
closedir(DIR);

# Output results to files.
print("Res  Target chi  Found chi    |Diff|   Source file\n");
print("===  ==========  ===========  =======  ===========\n");
for (my $i = 0; $i < scalar(@res_names); $i++) {
	my $current_res_name = $res_names[$i];
	my $current_filename = $current_res_name . '.db';
	
	open(FILEWRITE, "> $current_filename");
	
	for (my $chi_index = 0; $chi_index < 3; $chi_index++) {
		my $chi_targets_string = $chi_angles[$i];
		my @chi_targets = split(/\s+/, $chi_targets_string);
		my $target_chi = $chi_targets[$chi_index];
		
		my $rescoord_key = $current_res_name . $chi_index;
		if (exists($rescoords_hash{$rescoord_key})) {
			my $current_rescoord_data = $rescoords_hash{$rescoord_key};
			
			my $found_chi = 'null';
			if ($current_rescoord_data =~ /\b[A-Z]{3}\s+(\S+)\s+\S+\s+\d+/) {
				$found_chi = $1;
			}
			
			# printout to see how close our matches are.
			my $diff = &angle_dist($target_chi, $found_chi);
			my $source_file = $rescoords_source_file_hash{$rescoord_key};
			#print "$current_res_name chi = $target_chi: $found_chi (diff = $diff) (from $source_file)\n";
			printf("%3s %11.1f %12f  %.5f  %s\n", $current_res_name, $target_chi, $found_chi, $diff, $source_file);
			
			print FILEWRITE $current_rescoord_data;
		}
		else {
			#print "No data for chi = $target_chi\n";
			#print "$current_res_name chi = $target_chi: **No data found.\n";
			printf("%3s %11.1f  **No data found\n", $current_res_name, $target_chi);
		}
	}
	
	close FILEWRITE;
}

# Remove temporary files
system("rm -f $rescoords_dirname/*.rescoords");
system("rm -f $reduced_pdb_files_dirname/*.pdbH");

print STDERR "\nResidue template files written to $template_db_dirname\/ directory.\n";



################################################################
##########                 Functions                 ###########
################################################################

# Returns the magnitude of the distance between two angles, in degrees.
sub angle_dist{
	
	(my $a1, my $a2) = @_;
	
	# Put each angle in the range [-180.0, 180.0).
	while ($a1 < -180.0) {
		$a1 += 360.0;
	}
	while ($a1 >= 180.0) {
		$a1 -= 360.0;
	}
	
	while ($a2 < -180.0) {
		$a2 += 360.0;
	}
	while ($a2 >= 180.0) {
		$a2 -= 360.0;
	}
	
	# Adjust angles 
	if ($a1 - $a2 >= 0.0) {
		if ($a1 - $a2 < 180.0) {
			# No change required.
		}
		else {
			$a2 += 360.0;
		}
	}
	else {
		if ($a1 - $a2 > -180.0) {
			# No change required.
		}
		else {
			$a1 += 360.0;
		}
	}
	
	return abs($a1 - $a2);
}

# http://b3thm00n.com/scripting/perl_slurp/
# Read a file into a string.
sub slurp{
        my $file=$_[0];
	
	#passed the absolute path on machine to the file you want to read

	unless (-f $file){ return 0; } #if it does not exist 
		#return false
		
	my $src;

        local *INPUT;
        open (INPUT, $file);
        my $chunk;
        for (;;) {
                read(INPUT, $chunk, 1024); #how much to read per cycle (1k)
                last if (length($chunk) == 0); #if nothing was read then  stop
                $src.=$chunk;
        }
        close INPUT;
	
        return $src;
}

# http://www.somacon.com/p114.php
# Perl trim function to remove whitespace from the start and end of the string
sub trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}
