#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <math.h>
#include <errno.h>
#if !defined(__APPLE__)
#include <wait.h>
#endif
#include <assert.h>
#include <ctype.h>

#include "util.h"
#include "pepmake.h"
#include "potential.h"
#include "ga_potential.h"
#include "ga_potential_scores.h"
#include "rama.h"
#include "rotamer.h"
#include "phipsi.h"
#include "readfileShift.h"
#include "secstruct.h"


static void usage();
static void printHeaderLine(FILE *fp, int use_chem_shift_scores, int computeSecondaryScore);
static void printScores(FILE *fp, int printRawScores, int use_chem_shift_scores, int computeSecondaryScore,
		
		float *raw_chem_shift_coeffs, // array of 6 chemical shift scores
		float raw_thread_potential_score,
		float raw_bump_score, float raw_hcount, float raw_rama_score, float raw_omega_score,
		float raw_chi_score, float raw_rad_gyr_score, float raw_hbenergy, float raw_disulphide_bond_score,
		float raw_secScore,
		
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float scaled_thread_potential_score,
		float scaled_bump_score, float scaled_hcount, float scaled_rama_score, float scaled_omega_score,
		float scaled_chi_score, float scaled_rad_gyr_score, float scaled_hbenergy, float scaled_disulphide_bond_score,
		float scaled_secScore,
		float sum_chem_shift_scores, float sum_GA_Potential, float total_sum);
static void print_weights_info();
static char* bmrb_format_to_shifty_format(char* bmrb_shift_file);


static void usage() {
		printf("Usage: ./potential [-h] [-pw] [-uw] [-w] [-r] [-c] [-pss] [-b] [-py] [-z]\n");
		printf("            [-cs <chem shift file>] [-p]\n");
		printf("            [-s proteus_dir | -sf <secondary structure file>] input.pdb\n");
		printf("\n");
		printf("       [-h]\n");
		printf("          Do not call reduce to add hydrogens before calculating the potential\n");
		printf("          score.\n");
		printf("       [-pw]\n");
		printf("          Print information about the score weights and exit.\n");
		printf("       [-uw]\n");
		printf("          \"Unweighted\": Use weights of 1.0 for all scores. Scores will still\n");
		printf("          be normalized by the number of residues and multiplied by the initial\n");
		printf("          scaling values of +/- 1.0. This option is intended for use when\n");
		printf("          training the score weights.\n");
		printf("       [-w]\n");
		printf("          Print raw scores in addition, i.e. the scores before nomralization\n");
		printf("          and weighting.\n");
		printf("       [-r]\n");
		printf("          Print Rama and Omega score breakdowns by score category.\n");
		printf("       [-c]\n");
		printf("          Print all C-alpha to C-alpha distances.\n");
		printf("       [-pss]\n");
		printf("          Print the secondary structure (computed based on the 3-D structure).\n");
		printf("       [-b]\n");
		printf("          Print detailed list of the assigned secondary structure (computed\n");
		printf("          based on the 3-D structure) and the raw predicted Beta-turn locations.\n");
		printf("       [-py]\n");
		printf("          Output PyMOL command files for displaying the secondary structure\n");
		printf("          assigned by GA Potential.\n");
		printf("       [-z]\n");
		printf("          Print score headers to a file named [input.pdb].header, and print the\n");
		printf("          PDB filename plus the scores to a file named [input.pdb].score.\n");
		printf("       [-cs <chem_shift_file in bmrb format>]\n");
		printf("          Compute chemical shift scores in relation to the given chemical shifts\n");
		printf("          file. The chemical shift file can be in BMRB STAR 2.1 or SHIFTY\n");
		printf("          format.\n");
		printf("       [-p]\n");
		printf("          Use Prosa threading score instead of the normal threading score.\n");
		printf("          (Requires Prosa2003 to be installed with prosa2003 in the user's\n");
		printf("          path.)\n");
		printf("       [-s proteus_dir]\n");
		printf("          Call Proteus to predict secondary structure, and compute secondary\n");
		printf("          structure score in relation to this predicted secondary structure.\n");
		printf("          (Requires Proteus2 to be installed, where proteus_dir is the full path\n");
		printf("          to the directory containing run_proteus. Proteus2 requires BLAST and\n");
		printf("          ClustalW to be installed, with blastall and clustalw in the user's\n");
		printf("          path.)\n");
		printf("       [-sf <secondary structure file>]\n");
		printf("          Compute secondary structure score in relation to the secondary\n");
		printf("          structure given in <secondary structure file>. Overrides -s option.\n");
		printf("          The more divergence between this secondary structure and that computed\n");
		printf("          based on the 3-D structure, the higher the secondary structure score\n");
		printf("          will be. This file can be in one of two formats: (a) a secondary\n");
		printf("          structure assignment for all the residues in the PDB file. e.g.\n");
		printf("          HHHHCCCBBBCCCHHHBBBCCC and so on, where H = Helix, B or E = Beta\n");
		printf("          strand, C = coil, - = unknown. (b) If the content of the secondary\n");
		printf("          structure file begins with a '>' or '#', it will be parsed as a\n");
		printf("          Proteus output file.\n");
		exit(1);
}


int main_GA_potential(int argc, char *argv[]){
	int n = 0; // number of residues in the peptide
	char *PDB_filename;
	ResData *res;
	PDBFile pdb;
	char hydroname[256];
	int i, noHydro = FALSE, fileArg = -1; //, fileArg2 = -1;
	FILE *fp;
	int useUnweightedScores = FALSE;
	int printRawScores = FALSE;
	int printRamaAndOmegaScoreBreakdownFlag = FALSE;
	int print_CA_dists = FALSE;
	int print_SecStr = FALSE; // Whether to print the secondary structure assigned based on the 3-D structure.
	int print_SecStr_list = FALSE;
	int output_PyMOL_files = FALSE;
	int print_to_header_and_score_files = FALSE;
	float raw_chem_shift_coeffs[6];
	float raw_thread_potential_score, raw_bump_score,raw_hcount, raw_rama_score, raw_omega_score,
		raw_chi_score, raw_rad_gyr_score, raw_hbenergy, raw_disulphide_bond_score, raw_sec_score;
	float scaled_chem_shift_scores[6];
	float scaled_thread_potential_score, scaled_bump_score,scaled_hcount, scaled_rama_score, scaled_omega_score,
		scaled_chi_score, scaled_rad_gyr_score, scaled_hbenergy, scaled_disulphide_bond_score, scaled_sec_score;
	float sum_chem_shift_scores, sum_GA_Potential, total_sum;
	int use_chem_shift_scores = FALSE;
	char *bmrb_shift_file = NULL; // Name of chemical shifts file in BMRB STAR 2.1 format.
	char *experimental_shifts_in_shifty_format = NULL; // Name of chemical shifts file in SHIFTY format.
	int format_is_shifty = FALSE;
	char buf[5]; // Temporary, for testing chemical shifts file format.
	use_Prosa = FALSE; // Global variable in ga_potential_scores.c
	computeSecondaryScore = FALSE; // Global variable in ga_potential_scores.c
	use_Proteus = FALSE; // Global variable in ga_potential_scores.c
	proteus_dir = NULL; // Global variable in ga_potential_scores.c
	target_sec_structure = NULL; // Global variable in ga_potential_scores.c
	char *sequence = NULL; // Sequence parsed from the PDB file.
	
	
	if(argc<2)
	{
		usage();
	}
	
	
	/* Add hydrogens */
	for (i = 1; i < argc; i++)
	{
		if (strcmp(argv[i],"-h") == 0)
		{
			noHydro = TRUE;
		}
		else if (strcmp(argv[i],"-pw") == 0) {
			print_weights_info();
			exit(0);
		}
		else if (strcmp(argv[i],"-uw") == 0) {
			useUnweightedScores = TRUE;
		}
		else if (strcmp(argv[i],"-w") == 0) {
			printRawScores = TRUE;
		}
		else if (strcmp(argv[i],"-r") == 0) {
			printRamaAndOmegaScoreBreakdownFlag = TRUE;
		}
		else if (strcmp(argv[i],"-c") == 0) {
			print_CA_dists = TRUE;
		}
		else if (strcmp(argv[i],"-pss") == 0) {
			print_SecStr = TRUE;
		}
		else if (strcmp(argv[i],"-b") == 0) {
			print_SecStr_list = TRUE;
		}
		else if (strcmp(argv[i],"-py") == 0) {
			output_PyMOL_files = TRUE;
		}
		else if (strcmp(argv[i],"-z") == 0) {
			print_to_header_and_score_files = TRUE;
		}
		else if (strcmp(argv[i], "-cs") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				use_chem_shift_scores = TRUE;
				bmrb_shift_file = argv[i];
				check_file_readability(bmrb_shift_file);
			}
			else {
				fprintf(stderr, "Chemical shift file must be given after -cs option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i],"-p") == 0) {
			use_Prosa = TRUE;
		}
		else if (strcmp(argv[i], "-s") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				proteus_dir = argv[i];
				if (computeSecondaryScore == FALSE) {
					// Only use Proteus if a secondary structure file has not been given by the user.
					use_Proteus = TRUE;
					computeSecondaryScore = TRUE;
				}
			}
			else {
				fprintf(stderr, "Path to Proteus directory must be given after -s option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-sf") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				strcpy(secondary_structure_filename, argv[i]);
				check_file_readability(secondary_structure_filename);
				use_Proteus = FALSE; // Do not call Proteus if user has supplied a secondary structure file.
				computeSecondaryScore = TRUE;
			}
			else {
				fprintf(stderr, "Secondary structure file must be given after -sf option.\n");
				usage();
			}
		}
		else if (fileArg == -1)
		{
			fileArg = i;
		}
		/*
		else if (fileArg2 == -1)
		{
			fileArg2 = i;
			use_Proteus = FALSE;
			computeSecondaryScore = TRUE;
			sprintf(secondary_structure_filename, "%s", argv[fileArg2]);
		}
		*/
	}
	
	// Test to see whether the files exist.
	check_file_readability(argv[fileArg]);
	
	/*
	if (computeSecondaryScore && !use_Proteus) {
		if ( (fp = fopen(argv[fileArg2], "r")) != NULL ) {
			fclose(fp);
		}
		else {
			fprintf(stderr, "ERROR: Could not open file %s\n", argv[fileArg2]);
			exit(1);
		}
	}
	*/
	
		
	
	if (!noHydro)
	{
		// Call reduce to add hydrogen atoms to the structure, or to adjust them.
		call_reduce(argv[fileArg], hydroname, TRUE);
		PDB_filename = hydroname;
	}
	else
	{
		PDB_filename = argv[fileArg];
	}
	
	int rc;
	rc = parsePDBFile(PDB_filename,&pdb, '?');
	if (rc != TRUE) {
		fprintf(stderr, "Error (%d): GA potential could not parse PDB file %s.\n", rc, PDB_filename);
		exit(-1);
	}
	
	n=pdb.numRes; // Both values are the number of residues in the peptide.
	
	res=(ResData *)malloc(sizeof(ResData) * n);
	
	PDBtoResData(pdb, res);
	
	
	
	if (use_chem_shift_scores) {
		// Check whether the given chemical shifts file is in BMRB or SHIFTY format.
		fp = fopen(bmrb_shift_file, "r");
		if (fgets(buf, 5, fp) != NULL) {
			// Perhaps the "#" criteria needs improvement in case it misidentifies some BMRB files as SHIFTY (?).
			if (strncmp(buf, " NUM", 4) == 0 || strncmp(buf, "NUM", 3) == 0 ||
				strncmp(buf, " #", 2) == 0 || strncmp(buf, "#", 1) == 0 ||
				strncmp(buf, ">", 1) == 0) {
				format_is_shifty = TRUE;
			}
		}
		fclose(fp);
		
		// Convert the format of the chemical shifts file if necessary.
		if (!format_is_shifty) {
			// Convert BMRB-formatted file to SHIFTY format.
			experimental_shifts_in_shifty_format = bmrb_format_to_shifty_format(bmrb_shift_file);
		}
		else {
			// Use the SHIFTY file given by the user.
			experimental_shifts_in_shifty_format = bmrb_shift_file;
		}
		
		// Perform setup operations for an internal call to ShiftX routines:
		
		read_query_seq(experimental_shifts_in_shifty_format, &experimental_chem_shifts, &num_residues_expt_chem_shifts);
		// Read the experimental chemical shifts into the experimental_chem_shifts MATRIX struct array.
		
		// Also read in the random coil shifts, and subtract them from the experimental chemical shifts.
		// We want to subtract the random coil shifts before computing the correlation scores in get_chem_shift_score(),
		// and we do it here so we only have to do it once.
		
		loadRandomCoilShifts();
		subtract_random_coil_shifts(experimental_chem_shifts, randomCoilShifts, num_residues_expt_chem_shifts);
		
		// Allocate memory for the predicted chemical shifts. The purpose is to allocate this only once
		// in the course of the program and re-use the memory each time new chemical shifts are predicted.
		// In single-threaded code, there there is one copy of such a variable.
		
		Predicted_Chem_Shifts[0] = (MATRIX *) malloc(n*sizeof(MATRIX));
		assert( (Predicted_Chem_Shifts[0]) != NULL );
		
		// Do initializations for shiftx. This should to be called once and only once, before calling calcChemShifts().
		do_shiftx_initializations();
		
		// Get alignment data for use in chemical shift correlation scores.
		get_alignment_data(res, n, experimental_chem_shifts, num_residues_expt_chem_shifts);
	}
	
	if (use_chem_shift_scores || computeSecondaryScore || print_SecStr || print_SecStr_list || output_PyMOL_files) {
		
		// Allocate memory for an array of RESIDUE structs that represent the peptide.
		// This will be passed to calcChemShifts() and secondary_score(), which will use
		// this memory in the shiftx and secondary structure assignment routines. The
		// purpose is so that calcChemShifts() and secondary_score() will copy the atom
		// coordinates from the ResData struct array to the RESIDUE struct array, without
		// re-allocating new memory for a RESIDUE struct array each time one of them
		// is called. (These optimizations are perhaps not so important for potential
		// as the are for gafolder).
		Residues[0] = create_RESIDUE_array_from_ResData_array(res, n);
	}
	
	
	// Load rotamer library into memory. This should be done before calling get_GA_potential_scores().
	read_rama_tables();
	loadRotamerLibrary();
	
	if (computeSecondaryScore) {
		target_sec_structure = (char *)malloc(sizeof(char) * (n + 1));
		sequence = (char *)malloc(sizeof(char) * (n + 1));
		
		get_seq_from_ResData(&res[0], n, sequence);
		if (use_Proteus) {
			call_proteus(sequence, PDB_filename, proteus_dir, secondary_structure_filename);
		}
		parseSecondaryStructureFile(secondary_structure_filename, n, sequence, target_sec_structure);
	}
	
	get_GA_potential_scores_full(res, n, POTENTIAL_MODE_EVALUATE, WHICH_SCORES_ALL,
		PDB_filename, 0, // thread ID does matter because we are not running in parallel
		target_sec_structure, print_SecStr, print_SecStr_list,
		output_PyMOL_files, PDB_filename, printRawScores, useUnweightedScores,
		use_chem_shift_scores, experimental_shifts_in_shifty_format,
		
		&raw_chem_shift_coeffs[0],
		&raw_thread_potential_score, &raw_bump_score, &raw_hcount,
		&raw_rama_score, &raw_omega_score, &raw_chi_score, &raw_rad_gyr_score,
		&raw_hbenergy, &raw_disulphide_bond_score, &raw_sec_score,
		
		&scaled_chem_shift_scores[0],
		&scaled_thread_potential_score, &scaled_bump_score, &scaled_hcount,
		&scaled_rama_score, &scaled_omega_score, &scaled_chi_score, &scaled_rad_gyr_score,
		&scaled_hbenergy, &scaled_disulphide_bond_score, &scaled_sec_score,
		&sum_chem_shift_scores, &sum_GA_Potential, &total_sum);
	
	
  //printf("pot_score=%f, bump_score=%f, hcount=%f, rama_score=%f, radGyr_score=%f, secScore=%ld\n",
  //	 potential_score, bump_score, hcount, rama_score, radGyr_score, secScore); 
  //      RN     TH  BM HY RA RG SS  TH BM HY RA SS HE H% LP HZ HZ RD

  //printf("\nThe component scores for %s is as follows:\n",PDB_filename);
  //printf("=--------------------------------------------\n");
  /*printf("%d 0 0 %f  %f %f %f %f %ld %f %f %f %f %f %f %f %f %f %f %f ", 
	 residue_number, potential_score, bump_score, hcount, rama_score, 
	 radGyr_score, secScore, 
	 potential_score / residue_number, bump_score / residue_number,
	 hcount/ residue_number, rama_score / residue_number, 
	 (float)((float)secScore / (float)residue_number), sumEnergy,
	 (percStruct > 0) ? hperc/100.0/(double)percStruct * residue_number : 1.26 , 
	 linPerc,
	 hperc > 70.0 ? 0 : fabs((hperc - AVGPERCINHBONDS)/STDDEVINHBONDS),
	 fabs((hperc - AVGPERCINHBONDS)/STDDEVINHBONDS), rogdelta);
  for (i = 0; i < 9; i++){
    printf("%d ",ssErrors[i]);
    }*/


  //These were the parameters as they were originally reported by Steve
  /*printf("\nThe component scores for %s are as follows:\n",PDB_filename);
  printf("-----------------------------------------------\n");
  printf("Number of residues  = %d\n", residue_number);
  printf("Threading_potential = %-f,        Bump_score        = %-f\n", potential_score, bump_score);
  printf("Hydrogen bond count = %-f,  Ramachandran  score     = %-f\n", hcount, rama_score);
  printf("Radius of Gyr score = %-f,  Energy due to Hyd bonds = %-f\n", radGyr_score, sumEnergy);
  printf("-----------------------------------------------\n");
  

   //Dr. Wishart suggested these scaling factors for these different parameters on Nov 14, 2006
   
  printf("\nThe component scores for %s after scaling are as follows:\n",PDB_filename);
  printf("-----------------------------------------------\n");
  printf("Threading_potential = %-f,        Bump_score        = %-f\n", potential_score*0.2, bump_score/10.0);
  printf("Hydrogen bond count = %-f,  Ramachandran  score     = %-f\n", hcount*(-1.1), rama_score*10.0);
  printf("Radius of Gyr score = %-f\n", radGyr_score);
  printf("-----------------------------------------------\n");
  */


	// Output format so that the parameters can be read in as a popen  can read from the pipe.:
	//*****************************************
	printHeaderLine(stdout, use_chem_shift_scores, computeSecondaryScore);
	if (print_to_header_and_score_files) {
		char header_filename[256];
		sprintf(header_filename, "%s.%s", argv[fileArg], "header");
		FILE *header_out = fopen(header_filename, "w");
		printHeaderLine(header_out, use_chem_shift_scores, computeSecondaryScore);
		fclose(header_out);
	}
	
	printScores(stdout, printRawScores, use_chem_shift_scores, computeSecondaryScore,
		&raw_chem_shift_coeffs[0], raw_thread_potential_score,
		raw_bump_score, raw_hcount, raw_rama_score, raw_omega_score,
		raw_chi_score, raw_rad_gyr_score, raw_hbenergy, raw_disulphide_bond_score,
		raw_sec_score,
		&scaled_chem_shift_scores[0], scaled_thread_potential_score,
		scaled_bump_score, scaled_hcount, scaled_rama_score, scaled_omega_score,
		scaled_chi_score, scaled_rad_gyr_score, scaled_hbenergy, scaled_disulphide_bond_score,
		scaled_sec_score,
		sum_chem_shift_scores, sum_GA_Potential, total_sum);
	if (print_to_header_and_score_files) {
		char score_filename[256];
		sprintf(score_filename, "%s.%s", argv[fileArg], "score");
		FILE *score_out = fopen(score_filename, "w");
		fprintf(score_out, "%s\t", argv[fileArg]);
		printScores(score_out, printRawScores, use_chem_shift_scores, computeSecondaryScore,
			&raw_chem_shift_coeffs[0], raw_thread_potential_score,
			raw_bump_score, raw_hcount, raw_rama_score, raw_omega_score,
			raw_chi_score, raw_rad_gyr_score, raw_hbenergy, raw_disulphide_bond_score,
			raw_sec_score,
			&scaled_chem_shift_scores[0], scaled_thread_potential_score,
			scaled_bump_score, scaled_hcount, scaled_rama_score, scaled_omega_score,
			scaled_chi_score, scaled_rad_gyr_score, scaled_hbenergy, scaled_disulphide_bond_score,
			scaled_sec_score,
			sum_chem_shift_scores, sum_GA_Potential, total_sum);
		fclose(score_out);
	}
	
	if (printRamaAndOmegaScoreBreakdownFlag) {
		printf("\n");
		printRamaAndOmegaScoreBreakdown(res, n);
	}
	
	if (print_CA_dists) {
		printf("\n");
		print_C_alpha_distances(res, n);
	}

	if (!format_is_shifty) {
		free(experimental_shifts_in_shifty_format);
	}
	free(res);
	if (computeSecondaryScore) {
		free(target_sec_structure);
		free(sequence);
	}
	//unlink(hydroname);
	
	return 0;
}

static void printHeaderLine(FILE *fp, int use_chem_shift_scores, int computeSecondaryScore) {
	fprintf(fp, "#  ");
	if (use_chem_shift_scores) {
		fprintf(fp, "     HA     HN      N     CA     CB     CO");
	}
	if (use_Prosa) {
		fprintf(fp, "   Prosa");
	}
	else {
		fprintf(fp, "  Thread");
	}
	fprintf(fp, "    Bump Hcount    Rama    Omega    Chi    Rad  HBener Disulf");
	if (computeSecondaryScore) {
		fprintf(fp, "  SecStr");
	}
	if (use_chem_shift_scores) {
		fprintf(fp, "  SumCS    SumGA    Total\n");
	}
	else {
		fprintf(fp, "    Total\n");
	}
}

static void printScores(FILE *fp, int printRawScores, int use_chem_shift_scores, int computeSecondaryScore,
		
		float *raw_chem_shift_coeffs, // array of 6 chemical shift scores
		float raw_thread_potential_score,
		float raw_bump_score, float raw_hcount, float raw_rama_score, float raw_omega_score,
		float raw_chi_score, float raw_rad_gyr_score, float raw_hbenergy, float raw_disulphide_bond_score,
		float raw_secScore,
		
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float scaled_thread_potential_score,
		float scaled_bump_score, float scaled_hcount, float scaled_rama_score, float scaled_omega_score,
		float scaled_chi_score, float scaled_rad_gyr_score, float scaled_hbenergy, float scaled_disulphide_bond_score,
		float scaled_secScore,
		float sum_chem_shift_scores, float sum_GA_Potential, float total_sum) {
	
	int i;
	float tmp_cs_score;
	
	// Print raw scores.
	
	if (printRawScores) {
		fprintf(fp, "   ");
		
		if (use_chem_shift_scores) {
			for (i = 0; i < 6; i++) {
				tmp_cs_score = raw_chem_shift_coeffs[i];
				if (tmp_cs_score == 99.0) {
					tmp_cs_score = 0.0;
				}
				
				fprintf(fp, " % 6.2f", tmp_cs_score);
			}
		}
		else {
			//for (i = 0; i < 6; i++) {
			//	fprintf(fp, "%5.2f", 0.0);
			//}
		}
		
		//         "   Thread  Bump Hcount  Rama   Omega    Chi   Rad  HBener Disulf"
		fprintf(fp, " % 7.2f % 7.2f % 6.2f % 7.2f % 8.2f % 6.2f % 6.2f % 7.2f % 6.2f",
			raw_thread_potential_score, raw_bump_score, raw_hcount,
			raw_rama_score, raw_omega_score, raw_chi_score, raw_rad_gyr_score, raw_hbenergy,
			raw_disulphide_bond_score);
		if (computeSecondaryScore) {
			fprintf(fp, " % 7.2f", raw_secScore);
		}
		
		fprintf(fp, "\n");
	}
	
	
	// Print scaled scores.
	
	fprintf(fp, "   ");
	
	if (use_chem_shift_scores) {
		for (i = 0; i < 6; i++) {
			fprintf(fp, " % 6.2f", scaled_chem_shift_scores[i]);
		}
	}
	else {
		//for (i = 0; i < 6; i++) {
		//	fprintf(fp, "%5.2f", 0.0);
		//}
	}
	
	//         "   Thread  Bump Hcount  Rama   Omega    Chi   Rad  HBener Disulf"
	fprintf(fp, " % 7.2f % 7.2f % 6.2f % 7.2f % 8.2f % 6.2f % 6.2f % 7.2f % 6.2f",
		scaled_thread_potential_score, scaled_bump_score, scaled_hcount,
		scaled_rama_score, scaled_omega_score, scaled_chi_score, scaled_rad_gyr_score, scaled_hbenergy,
		scaled_disulphide_bond_score);
	if (computeSecondaryScore) {
		fprintf(fp, " % 7.2f", scaled_secScore);
	}
	
	if (use_chem_shift_scores) {
		fprintf(fp, " % 6.2f % 8.2f % 8.2f\n", sum_chem_shift_scores, sum_GA_Potential, total_sum);
	}
	else {
		fprintf(fp, " % 8.2f\n", total_sum);
	}
}


/*
 * Print information about the score weights.
 */
static void print_weights_info() {
	
	printf("\n");
	printf("Score Weighting\n");
	printf("\n");
	printf("Each of the initial terms in the energy function are transformed through a\n");
	printf("series of steps:\n");
	printf("(1) They are multiplied by +/- 1.0 so that they behave as energy-like terms,\n");
	printf("    i.e. more negative scores indicate a better structure.\n");
	printf("(2) All scores except for the chemical shifts scores and the radius of gyration\n");
	printf("    score are normalized by dividing by the number of residues in the peptide,\n");
	printf("    and then multiplying by 100 to re-scale them.\n");
	printf("(3) All scores are multiplied by weights determined through training or some\n");
	printf("    other means.\n");
	printf("\n");
	printf("Score     Initial scaling  Weight (before scaling chem shift  Weight\n");
	printf("term      (Step 1)         scores w.r.t. GA Potential scores  (Step 3)\n");
	printf("                           using scaling for Thread score)\n");
	printf("========  ===============  =================================  ========\n");
	printf("HA                  % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_HA,       WEIGHT_HA,       WEIGHT_HA       * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("HN                  % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_HN,       WEIGHT_HN,       WEIGHT_HN       * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("N                   % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_N,        WEIGHT_N,        WEIGHT_N        * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("CA                  % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_CA,       WEIGHT_CA,       WEIGHT_CA       * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("CB                  % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_CB,       WEIGHT_CB,       WEIGHT_CB       * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("CO                  % 5.1f                           % 8.3f  % 8.3f\n", INITIAL_SCALING_CO,       WEIGHT_CO,       WEIGHT_CO       * CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD);
	printf("ProSa               % 5.1f                                     % 8.3f\n", INITIAL_SCALING_PROSA,                     WEIGHT_PROSA);
	printf("Thread              % 5.1f                                     % 8.3f\n", INITIAL_SCALING_THREAD,                    WEIGHT_THREAD);
	printf("Bump                % 5.1f                                     % 8.3f\n", INITIAL_SCALING_BUMP,                      WEIGHT_BUMP);
	printf("Hcount              % 5.1f                                     % 8.3f\n", INITIAL_SCALING_HCOUNT,                    WEIGHT_HCOUNT);
	printf("Rama                % 5.1f                                     % 8.3f\n", INITIAL_SCALING_RAMA,                      WEIGHT_RAMA);
	printf("Omega               % 5.1f                                     % 8.3f\n", INITIAL_SCALING_OMEGA,                     WEIGHT_OMEGA);
	printf("Chi                 % 5.1f                                     % 8.3f\n", INITIAL_SCALING_CHI,                       WEIGHT_CHI);
	printf("RadGyr              % 5.1f                                     % 8.3f\n", INITIAL_SCALING_RADGYR,                    WEIGHT_RADGYR);
	printf("HBenergy            % 5.1f                                     % 8.3f\n", INITIAL_SCALING_HBENERGY,                  WEIGHT_HBENERGY);
	printf("Disulph             % 5.1f                                     % 8.3f\n", INITIAL_SCALING_DISULPH,                   WEIGHT_DISULPH);
	printf("SecStr              % 5.1f                                     % 8.3f\n", INITIAL_SCALING_SECSTR,                    WEIGHT_SECSTR);
}


/*
 * Calculates the scaled and weighted GA Potential scores without setting the raw scores.
 */
void get_GA_potential_scores(ResData *residues, int numRes, int mode, int which_scores, char *PDB_filename,
		int thread_id, char *target_sec_structure, int print_SecStr, int print_SecStr_list, 
		int output_PyMOL_files, char *base_PDB_filename, int useUnweightedScores,
		int use_chem_shift_scores, char *experimental_shifts_in_shifty_format,
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float *scaled_thread_potential_score, float *scaled_bump_score, float *scaled_hcount,
		float *scaled_rama_score, float *scaled_omega_score, float *scaled_chi_score, float *scaled_rad_gyr_score,
		float *scaled_hbenergy, float *scaled_disulphide_bond_score, float *scaled_sec_score,
		float *sum_chem_shift_scores, float *sum_GA_Potential, float *total_sum) {
	
	get_GA_potential_scores_full(residues, numRes, mode, which_scores, PDB_filename,
		thread_id, target_sec_structure, print_SecStr, print_SecStr_list,
		output_PyMOL_files, base_PDB_filename, FALSE, useUnweightedScores,
		use_chem_shift_scores, experimental_shifts_in_shifty_format,
		
		NULL,
		NULL, NULL, NULL,
		NULL, NULL, NULL, NULL,
		NULL, NULL, NULL,
		
		scaled_chem_shift_scores,
		scaled_thread_potential_score, scaled_bump_score, scaled_hcount,
		scaled_rama_score, scaled_omega_score, scaled_chi_score, scaled_rad_gyr_score,
		scaled_hbenergy, scaled_disulphide_bond_score, scaled_sec_score,
		sum_chem_shift_scores, sum_GA_Potential, total_sum
		);
	
}


/*
 * Calculates the GA Potential scores for the given peptide (weighted)
 * and sets the values of the score parameters. numRes is the number of
 * residues in the peptide. This function assumes that all atom coordinates
 * are in place, that all dihedral angle variables (phi, psi, omega) are
 * set to their proper values. This function will not make any attempt to
 * call reduce to place the hydrogren atoms.
 *
 * 'mode' can be one of POTENTIAL_MODE_MINIMIZE, POTENTIAL_MODE_EVALUATE.
 * These indicate different purposes for using the GA Potential, which may
 * involve the use of different scores and different weighting of scores.
 *
 * 'which_scores' can be one of WHICH_SCORES_ALL, WHICH_SCORES_CHEM_SHIFT,
 * or WHICH_SCORES_GA_POTENTIAL, indicating whether all or a subset of
 * score should be calculated. If WHICH_SCORES_GA_POTENTIAL is given, the
 * chemical shift scores are set to 0.0, and 'total_sum' is set. If
 * WHICH_SCORES_CHEM_SHIFT is given, the other scores must have been computed
 * already, and 'total_sum' will be set based on the computed sum of chemical
 * shift score plus the parameter 'sum_GA_Potential'.
 *
 * If use_chem_shift_scores is TRUE, chemical shift scores will be
 * calculated. If, in addition, experimental_shifts_in_shifty_format is not
 * NULL, a system call will be made to shiftx. Otherwise, shiftx will be
 * called internally. experimental_shifts_in_shifty_format is the (optional)
 * name of a SHIFTY-formatted chemical shifts file.
 *
 * If use_Prosa is TRUE, the normal threading score will be replaced by
 * the Prosa score, obtained by making a system call to prosa2003. The
 * PDB filename must be given in this case; otherwise this parameter is
 * ignored. This only works if prosa2003 is installed and is in the user's
 * path.
 *
 * Note: PDB_filename is the name of the PDB file that should be used to
 * compute certain scores. base_PDB_filename is only used for naming optional
 * PyMOL command files; the structure of this file is not directly used.
 *
 * If set_raw_scores is TRUE, the variables for the raw scores will also be
 * set.
 */
void get_GA_potential_scores_full(ResData *residues, int numRes, int mode, int which_scores, char *PDB_filename,
		int thread_id, char *target_sec_structure, int print_SecStr, int print_SecStr_list,
		int output_PyMOL_files, char *base_PDB_filename, int set_raw_scores, int useUnweightedScores,
		int use_chem_shift_scores, char *experimental_shifts_in_shifty_format,
		
		// Raw scores
		float *raw_chem_shift_coeffs, // array of 6 chemical shift scores
		float *raw_thread_potential_score, float *raw_bump_score, float *raw_hcount,
		float *raw_rama_score, float *raw_omega_score, float *raw_chi_score, float *raw_rad_gyr_score,
		float *raw_hbenergy, float *raw_disulphide_bond_score, float *raw_sec_score,
		
		// Scaled and weighted scores
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float *scaled_thread_potential_score, float *scaled_bump_score, float *scaled_hcount,
		float *scaled_rama_score, float *scaled_omega_score, float *scaled_chi_score, float *scaled_rad_gyr_score,
		float *scaled_hbenergy, float *scaled_disulphide_bond_score, float *scaled_sec_score,
		float *sum_chem_shift_scores, float *sum_GA_Potential, float *total_sum) {
	
	
	int i;
	float chem_shift_coeffs[6];
	float thread_potential_score=0.0F, bump_score=0.0F, rama_score=0.0F, omega_score=0.0F, chi_score=0.0F;
	float rad_gyr_score=0.0F, hcount=0.0F, disulphide_bond_score=0.0F, sec_score=0.0F;
	int disulphide_bond_count = 0;
	double hbenergy, linPerc, hperc, rogdelta;
	long percStruct = 0L;
	int ssErrors[9];
	int ssResCount; // Number of residues with assignable secondary structure score. Used to normalize the score.
	float chem_shift_coeff_wrt_GA_Potential;
	int sec_str_is_assigned = FALSE;
		// Boolean. Whether the secondary structure has been assigned for the given structure.
	int num_cs_scores; // Number of chemical shift scores available.
	
	*sum_chem_shift_scores = 0.0;
	
		
	if (numRes <= 3) {
		// Chemical shift correlation scores cannot be reliably calculated for peptides with 3 or fewer residues.
		use_chem_shift_scores = FALSE;
	}
	
	
	/*************************************************************************/
	/***                       Calculate the scores                        ***/
	/*************************************************************************/
	
	if (use_chem_shift_scores && (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_CHEM_SHIFT)) {
		
		// Old code that can be used to make an external call to shiftx.
		//if (experimental_shifts_in_shifty_format != NULL) {
		//	// Make external system call to shiftx.
		//	get_chem_shift_score_from_PDB_file(residues, numRes, PDB_filename, experimental_shifts_in_shifty_format, chem_shift_coeffs);
		//		//get_chem_shift_score_from_PDB_file is in readfileShift.c
		//}
		
		// Make internal call to shiftx routines.
		get_chem_shift_score(residues, numRes, Residues[thread_id], Predicted_Chem_Shifts[thread_id], chem_shift_coeffs);
			//get_chem_shift_score is in readfileShift.c
		
		// In the course of calculating the chemical shift score, the secondary structure of the structure was
		// assigned (in the relavant variables in Residues[thread_id]). calcChemShifts() in Main_shiftx.c removed
		// the hydrogen atoms and added its own, which may affect the assigned secondary structure as compared to
		// the hydrogens present in the given ResData array 'residues', but hopefully the difference is not very
		// significant. If the secondary structure score will be computed below, we know that we do not need to
		// re-assign the secondary structure again.
		sec_str_is_assigned = TRUE;
	}
	else {
		for (i = 0; i < 6; i++) {
			chem_shift_coeffs[i] = 0.0F;
		}
	}
	if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
		if (use_Prosa) {
			thread_potential_score = getProsaScore(PDB_filename, thread_id);
		}
		else {
			thread_potential_score = (float) Empirical_Potential2(residues, numRes);
		}
		bump_score    =  get_bump_score(residues, numRes);
		hcount        =  get_hcount(residues, numRes, &hbenergy, &linPerc, &hperc, &disulphide_bond_count);
		rama_score    =  get_rama_value(residues, numRes);
		omega_score   =  getOmegaScore(residues, numRes);
		chi_score     =  getChiScore(residues, numRes);
		rad_gyr_score =  get_radGyr(residues, numRes, &rogdelta);
		disulphide_bond_score = (float) disulphide_bond_count;
		if (computeSecondaryScore || print_SecStr || print_SecStr_list || output_PyMOL_files) {
			memset(ssErrors, 0, sizeof(ssErrors));
			
			//sec_score = (float) secondary_score_from_PDB_file(PDB_filename, target_sec_structure, &percStruct, ssErrors,
			//	print_SecStr, print_SecStr_list, output_PyMOL_files, base_PDB_filename, &ssResCount);
			
			sec_score = (float) secondary_score(residues, numRes, Residues[thread_id], target_sec_structure,
				sec_str_is_assigned, &percStruct, ssErrors, &ssResCount);
			
			print_secondary_structure_info(Residues[thread_id], numRes-1, print_SecStr,
				print_SecStr_list, output_PyMOL_files, base_PDB_filename);
		}
		else {
			sec_score = 0.0F;
		}
	}
	
	/*************************************************************************/
	/***                   Set the raw score parameters                    ***/
	/*************************************************************************/
	
	if (set_raw_scores) {
		if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_CHEM_SHIFT) {
			if (use_chem_shift_scores) {
				for (i = 0; i < 6; i++) {
					raw_chem_shift_coeffs[i] = chem_shift_coeffs[i];
				}
			}
		}
		
		if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
			*raw_thread_potential_score = thread_potential_score;
			*raw_bump_score             = bump_score;
			*raw_hcount                 = hcount;
			*raw_rama_score             = rama_score;
			*raw_omega_score            = omega_score;
			*raw_chi_score              = chi_score;
			*raw_rad_gyr_score          = rad_gyr_score;
			*raw_hbenergy               = hbenergy;
			*raw_disulphide_bond_score  = disulphide_bond_score;
			if (computeSecondaryScore) {
				if (ssResCount != 0) {
					*raw_sec_score = sec_score;
				}
				else {
					*raw_sec_score = 0.0F;
				}
			}
		}
	}
	
	
	/*************************************************************************/
	/***  Normalize the scores by dividing by the number of residues, and  ***/
	/***  multiplying by the constant NORMALIZATION_COUNTER_WEIGHT.        ***/
	/*************************************************************************/
	
	if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
		thread_potential_score *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		bump_score             *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		hcount                 *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		rama_score             *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		omega_score            *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		chi_score              *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		// (Radius of gyration is a global score for the peptide).
		hbenergy               *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		disulphide_bond_score  *= NORMALIZATION_COUNTER_WEIGHT / (float) numRes;
		if (ssResCount != 0) {
			sec_score          *= NORMALIZATION_COUNTER_WEIGHT / (float) ssResCount;
		}
		else {
			sec_score          = 0.0F;
		}
	}
	
	
	/*************************************************************************/
	/***    Adjust the scores by the initial scaling values of +/- 1.0,    ***/
	/***           and apply transformations to certain scores.            ***/
	/*************************************************************************/
	
	if (use_chem_shift_scores && (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_CHEM_SHIFT)) {
		
		num_cs_scores = 0;
		for(i=0; i<6; i++) {
			if(chem_shift_coeffs[i] != 99.0) {
				num_cs_scores++;
			}
		}
		
		// We want to scale the chemical shift scores depending on the number of shifts available. e.g. if
		// there are only HA and HN shifts available, scale those scores by 6/2 = 3.
		float missing_shifts_scaling_ratio = 1.0;
		if (num_cs_scores != 6) {
			if (num_cs_scores == 0) {
				fprintf(stderr, "ERROR: Chemical shifts missing for all chem shift types (HA, HN, N, CA, CB, CO).\n");
				exit(1);
			}
			
			missing_shifts_scaling_ratio = ((float) 6)/((float) num_cs_scores);
		}
		
		for(i=0; i<6; i++) {
			if(chem_shift_coeffs[i] == 99.0) {
				chem_shift_coeffs[i] = 0.00;
				scaled_chem_shift_scores[i] = 0.00;
			}
			else {
				     if (i == 0) { scaled_chem_shift_scores[i] = INITIAL_SCALING_HA * (IDEAL_CS_CORREL_HA - chem_shift_coeffs[i]); } // HA
				else if (i == 1) { scaled_chem_shift_scores[i] = INITIAL_SCALING_HN * (IDEAL_CS_CORREL_HN - chem_shift_coeffs[i]); } // HN
				else if (i == 2) { scaled_chem_shift_scores[i] = INITIAL_SCALING_N  * (IDEAL_CS_CORREL_N  - chem_shift_coeffs[i]); } // N
				else if (i == 3) { scaled_chem_shift_scores[i] = INITIAL_SCALING_CA * (IDEAL_CS_CORREL_CA - chem_shift_coeffs[i]); } // CA
				else if (i == 4) { scaled_chem_shift_scores[i] = INITIAL_SCALING_CB * (IDEAL_CS_CORREL_CB - chem_shift_coeffs[i]); } // CB
				else if (i == 5) { scaled_chem_shift_scores[i] = INITIAL_SCALING_CO * (IDEAL_CS_CORREL_CO - chem_shift_coeffs[i]); } // CO
				
				scaled_chem_shift_scores[i] *= missing_shifts_scaling_ratio;
			}
		}
		
		//scaled_chem_shift_scores[0] = INITIAL_SCALING_HA * chem_shift_coeffs[0]; // HA
		//scaled_chem_shift_scores[1] = INITIAL_SCALING_HN * chem_shift_coeffs[1]; // HN
		//scaled_chem_shift_scores[2] = INITIAL_SCALING_N  * chem_shift_coeffs[2]; // N
		//scaled_chem_shift_scores[3] = INITIAL_SCALING_CA * chem_shift_coeffs[3]; // CA
		//scaled_chem_shift_scores[4] = INITIAL_SCALING_CB * chem_shift_coeffs[4]; // CB
		//scaled_chem_shift_scores[5] = INITIAL_SCALING_CO * chem_shift_coeffs[5]; // CO
		
		for(i=0; i<6; i++) {
			*sum_chem_shift_scores += scaled_chem_shift_scores[i];
		}
	}
	else {
		for(i=0; i<6; i++) {
			scaled_chem_shift_scores[i] = 0.0F;
		}
	}
	
	if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
		if (use_Prosa) {
			*scaled_thread_potential_score = INITIAL_SCALING_PROSA * thread_potential_score;
		}
		else {
			*scaled_thread_potential_score = INITIAL_SCALING_THREAD * thread_potential_score;
		}
		*scaled_bump_score = INITIAL_SCALING_BUMP * bump_score;
		*scaled_hcount = hcount;
		// If the number of hydrogen bonds divided by the number of residues is more than -0.9/-1.1 = 81.8 %, then
		// set the scaled Hcount to match this threshold. Added on January 11, 2007.
		if (*scaled_hcount < -0.82F * NORMALIZATION_COUNTER_WEIGHT) {
			*scaled_hcount = -0.82F * NORMALIZATION_COUNTER_WEIGHT;
		}
		*scaled_hcount *= INITIAL_SCALING_HCOUNT;
		*scaled_rama_score = INITIAL_SCALING_RAMA * rama_score;
		*scaled_omega_score = INITIAL_SCALING_OMEGA * omega_score;
		*scaled_chi_score = INITIAL_SCALING_CHI * chi_score;
		if (rad_gyr_score > 1.0) {
			rad_gyr_score = rad_gyr_score * rad_gyr_score; // Square rad_gyr_score.
		}
		*scaled_rad_gyr_score = INITIAL_SCALING_RADGYR * rad_gyr_score;
		*scaled_hbenergy = INITIAL_SCALING_HBENERGY * hbenergy;
		*scaled_disulphide_bond_score = INITIAL_SCALING_DISULPH * ((float) disulphide_bond_count);
		*scaled_sec_score = INITIAL_SCALING_SECSTR * sec_score;
	}
	
	
	/*************************************************************************/
	/***         Adjust the scores by the weighting coefficients.          ***/
	/*************************************************************************/
	
	if (!useUnweightedScores) {
		if (use_chem_shift_scores && (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_CHEM_SHIFT)) {
			if (use_Prosa) {
				chem_shift_coeff_wrt_GA_Potential = CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_PROSA;
			}
			else {
				chem_shift_coeff_wrt_GA_Potential = CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD;
			}
			
			scaled_chem_shift_scores[0] *= WEIGHT_HA; // HA
			scaled_chem_shift_scores[1] *= WEIGHT_HN; // HN
			scaled_chem_shift_scores[2] *= WEIGHT_N ; // N
			scaled_chem_shift_scores[3] *= WEIGHT_CA; // CA
			scaled_chem_shift_scores[4] *= WEIGHT_CB; // CB
			scaled_chem_shift_scores[5] *= WEIGHT_CO; // CO
			
			*sum_chem_shift_scores = 0.0;
			for(i=0; i<6; i++) {
				scaled_chem_shift_scores[i] *= GLOBAL_COEFF * chem_shift_coeff_wrt_GA_Potential;
				*sum_chem_shift_scores += scaled_chem_shift_scores[i];
			}
		}
		
		if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
			if (use_Prosa) {
				*scaled_thread_potential_score *= WEIGHT_PROSA * GLOBAL_COEFF;
			}
			else {
				*scaled_thread_potential_score *= WEIGHT_THREAD * GLOBAL_COEFF;
			}
			*scaled_bump_score *= WEIGHT_BUMP * GLOBAL_COEFF;
			*scaled_hcount *= WEIGHT_HCOUNT * GLOBAL_COEFF;
			*scaled_rama_score *= WEIGHT_RAMA * GLOBAL_COEFF;
			*scaled_omega_score *= WEIGHT_OMEGA * GLOBAL_COEFF;
			*scaled_chi_score *= WEIGHT_CHI * GLOBAL_COEFF;
			*scaled_rad_gyr_score *= WEIGHT_RADGYR * GLOBAL_COEFF;
			*scaled_hbenergy *= WEIGHT_HBENERGY * GLOBAL_COEFF;
			*scaled_disulphide_bond_score *= WEIGHT_DISULPH * GLOBAL_COEFF;
			*scaled_sec_score *= WEIGHT_SECSTR * GLOBAL_COEFF;
		}
	}
	
	
	/*************************************************************************/
	/***                        Calculate the sum.                         ***/
	/*************************************************************************/
	
	if (which_scores == WHICH_SCORES_ALL || which_scores == WHICH_SCORES_GA_POTENTIAL) {
		*sum_GA_Potential = *scaled_thread_potential_score + *scaled_bump_score + *scaled_hcount +
				*scaled_rama_score + *scaled_chi_score + *scaled_omega_score + *scaled_rad_gyr_score +
				*scaled_hbenergy + *scaled_disulphide_bond_score + *scaled_sec_score;
	}
	
	*total_sum = *sum_chem_shift_scores + *sum_GA_Potential;
	
}



/*
 * Calls the Proteus secondary structure prediction program, which is part of the Proteus2
 * package ( http://www.proteus2.ca/proteus2/ ). Here, "Proteus" means the secondary structure
 * prediction program run by the run_proteus script from the Proteus2 package. This function
 * requires that Proteus2 has been installed on the user's system, and that the parameter
 * 'proteus_dir' contains the name of the directory containing the run_proteus script. This
 * function writes the name of the resulting Proteus output file to the parameter
 * 'proteus_output_filename'.
 */
void call_proteus(char *sequence, char *input_PDB_filename, char* proteus_dir, char *proteus_output_filename) {
	FILE *fasta_fp;
	char fasta_filename[512];
	char proteus_command[512];
	char current_dir[512];
	
	// Determine names of FASTA file and Proteus output file (full path).
	
	getcwd(current_dir, 512);
	
	if (input_PDB_filename[0] == '/') {
		// Assume that input_PDB_filename contains the full path to the input PDB file.
		sprintf(fasta_filename, "%s.fa", input_PDB_filename);
		sprintf(proteus_output_filename, "%s.proteus", input_PDB_filename);
	}
	else {
		sprintf(fasta_filename, "%s/%s.fa", current_dir, input_PDB_filename);
		sprintf(proteus_output_filename, "%s/%s.proteus", current_dir, input_PDB_filename);
	}
	
	// Write FASTA file with the protein sequence.
	
	fasta_fp = fopen(fasta_filename, "w");
	fprintf(fasta_fp, ">tmp_fasta\n");
	fprintf(fasta_fp, "%s\n", sequence);
	fclose(fasta_fp);
	
	// Call Proteus.
	
	if (chdir(proteus_dir) != 0) {
		fprintf(stderr, "ERROR: Could not change directory to %s\n", proteus_dir);
		exit(1);
	}
	
	sprintf(proteus_command, "sh run_proteus -m fast -o %s %s", proteus_output_filename, fasta_filename);
	if (system(proteus_command) < 0 ) {
		fprintf(stderr, "ERROR: Call to Proteus failed.\n");
		exit(-1);
	}
	
	if (chdir(current_dir) != 0) {
		fprintf(stderr, "ERROR: Could not change directory to %s\n", current_dir);
		exit(1);
	}
	
	printf("Proteus file: %s\n", proteus_output_filename);
}


/*
 * Parses the given secondary structure file and writes the secondary structure as a null-terminated string
 * to the array pointed to by target_sec_structure, of length totalRes plus 1 for the NULL terminator.
 *
 * The given secondary structure file can be either in Proteus output format, or a simple string (see
 * usage()). The format is automatically detected. If the parsed secondary structure does not have length
 * totalRes, an error is printed and the program exits.
 */
void parseSecondaryStructureFile(char *secondaryStructureFilename, int totalRes, char *sequence, char *target_sec_structure) {
	
	FILE *ff;
	char buffer[MAX_LINE_LEN];
	char c;
	int i, j, k;
	int prev_line_was_seq = FALSE;
	int non_space_count;
	char *proteus_seq; // Sequence as parsed from Proteus output file.
	char *sec_str; // Secondary structure parsed from file.
	int extra_margin = 200; // Maximum number of extra residues we expect.
	char *pdb_segment;
	char *proteus_segment;
	char pdb_char;
	char proteus_char;
	int num_identical_residues;
	int num_aligned_residues;
	float seq_identity; // Proportional.
	float seq_identity_threshold = 0.3F; // Fail if sequence alignment results in a sequence identity below this proportion.
	
	sec_str = (char *)malloc(sizeof(char) * (totalRes + 1 + extra_margin));
	
	ff = fopen(secondaryStructureFilename,"rt");
	if (ff != NULL) {
		c = fgetc(ff);
		
		rewind(ff);
		
		if (c == '>' || c == '#') {
			// Assume that the secondary structure file is in Proteus output format, and parse it.
			
			proteus_seq = (char *)malloc(sizeof(char) * (totalRes + 1 + extra_margin));
			pdb_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
			proteus_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
			
			i = 0;
			k = 0;
			while ( fgets(buffer, MAX_LINE_LEN, ff) != NULL ) {
				
				if (buffer[0] == '>' || buffer[0] == '#') {
					continue;
				}
				else if (buffer[0] == '\n') {
					continue;
				}
				else if (prev_line_was_seq == TRUE) {
					// Assume current line contains secondary structure info.
					
					j = 0;
					c = buffer[0];
					while (c != '\0') {
						if (c != ' ' && c != '\n') {
							if (i > totalRes + extra_margin) {
								fprintf(stderr, "ERROR: Far too many secondary structure chartacters in secondary structure file %s.\n",
									secondaryStructureFilename);
								fprintf(stderr, "   At least %d residues counted, but %d expected based on PDB file.\n", i+1, totalRes);
								fprintf(stderr, "   Secondary structure file was parsed as a Proteus output file.\n");
								fprintf(stderr, "   Parsed secondary structure begins with:\n");
								fprintf(stderr, "   %*s\n", i, sec_str);
								exit(1);
							}
							
							sec_str[i] = c;
							i++;
						}
						j++;
						c = buffer[j];
					}
					
					prev_line_was_seq = FALSE;
				}
				else {
					// Assume the current line contains sequence info.
					
					j = 0;
					c = buffer[0];
					while (c != '\0') {
						if (c != ' ' && c != '\n') {
							if (k > totalRes + extra_margin) {
								fprintf(stderr, "ERROR: Far too many residues in secondary structure file %s.\n",
									secondaryStructureFilename);
								fprintf(stderr, "   At least %d residues counted, but %d expected based on PDB file.\n", k+1, totalRes);
								fprintf(stderr, "   Secondary structure file was parsed as a Proteus output file.\n");
								fprintf(stderr, "   Parsed secondary structure begins with:\n");
								fprintf(stderr, "   %*s\n", k, proteus_seq);
								exit(1);
							}
							
							proteus_seq[k] = c;
							k++;
						}
						j++;
						c = buffer[j];
					}
					
					prev_line_was_seq = TRUE;
				}
			}
			
			fclose(ff);
			
			sec_str[i] = '\0';
			proteus_seq[k] = '\0';
			
			// Compare proteus_seq with the given sequence (from the PDB file). If they have different lengths,
			// adjust the secondary structure to length totalRes (plus 1 for the null terminator). It is assumed
			// that one sequence will be the substring of the other; otherwise an error will be printed and the
			// program will exit.
			
			if (strcmp(sequence, proteus_seq) == 0) {
				// The PDB-derived and Proteus sequences are identical. Transfer the secondary structure.
				strcpy(target_sec_structure, sec_str);
			}
			else {
				// The PDB-derived and Proteus sequences differ. Call clustalw to perform an alignment, and
				// map the secondary structure from the Proteus file onto the sequence from the PDB file.
				// 
				// clustalw1.83.UNIX.tar.gz was obtained from ftp://ftp.ebi.ac.uk/pub/software/unix/clustalw/
				// See http://www.bioinformatics-toolkit.org/Help/Topics/clustalw.html
				// Citation:
				// Thompson J.D., Higgins D.G., Gibson T.J. "CLUSTAL W: improving the sensitivity of progressive
				// multiple sequence alignment through sequence weighting, position-specific gap penalties and
				// weight matrix choice." Nucleic Acids Res. 22:4673-4680(1994).
				
				// Write FASTA file with the two protein sequences.
				
				ff = fopen("tmp_align.fa", "w");
				fprintf(ff, ">pdb\n");
				fprintf(ff, "%s\n", sequence);
				fprintf(ff, ">proteus\n");
				fprintf(ff, "%s\n", proteus_seq);
				fclose(ff);
				
				// Call ClustalW.
				
				if (system("./clustalw tmp_align.fa > /dev/null") != 0) {
					fprintf(stderr, "ERROR: Call to clustalw failed.\n");
					exit(-1);
				}
				
				// Parse alignment file generated by ClustalW.
				
				ff = fopen("tmp_align.aln", "r");
				
				// Skip first 3 lines
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
				}
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
				}
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
				}
				
				num_identical_residues = 0;
				num_aligned_residues = 0;
				j = 0;
				k = 0;
				while( fgets(buffer, MAX_LINE_LEN, ff) != NULL ) {
					sscanf(buffer, "%*s %s", pdb_segment);
					if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
						fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
					}
					sscanf(buffer, "%*s %s", proteus_segment);
					
					// Compare the two sequences, and map the secondary structure from the Proteus file (sec_str)
					// to the sequence read from the PDB file.
					i = 0;
					pdb_char = pdb_segment[0];
					proteus_char = proteus_segment[0];
					while (pdb_char != '\0') {
						
						if (pdb_char != '-') {
							if (proteus_char != '-') {
								target_sec_structure[j] = sec_str[k];
								j++;
								k++;
								num_aligned_residues++;
								if (pdb_char == proteus_char) {
									num_identical_residues++;
								}
							}
							else {
								// Proteus sequence does not contain a residue that is in the PDB
								// sequence. Add a '-' for the secondary structure for this residue.
								target_sec_structure[j] = '-';
								j++;
							}
						}
						else {
							// PDB sequence does not contain this residue. Skip over the secondary
							// structure for this residue.
							k++;
						}
						
						i++;
						pdb_char = pdb_segment[i];
						proteus_char = proteus_segment[i];
					}
					
					// Skip 2 lines.
					if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
						fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
					}
					if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
						// End of file.
						break;
					}
				}
				
				fclose(ff);
				
				target_sec_structure[j] = '\0';
				
				if (num_aligned_residues != 0) {
					seq_identity = ((float) num_identical_residues)/((float) num_aligned_residues);
				}
				else {
					seq_identity = 0.0F;
				}
				
				if (seq_identity < seq_identity_threshold) {
					fprintf(stderr, "ERROR: Aligment between sequences in PDB file and Proteus output file\n");
					fprintf(stderr, "results in sequence identity of only %4.1f%%!\n", 100.0F*seq_identity);
					fprintf(stderr, "Wrong input files suspected.\n");
					exit(1);
				}
				
				if (j != totalRes) {
					fprintf(stderr, "ERROR: Incorrect number of residues after aligning sequences.\n");
					fprintf(stderr, "   %d residues counted, but %d expected based on PDB file.\n", j, totalRes);
					fprintf(stderr, "   Secondary structure file was parsed as a Proteus output file.\n");
					fprintf(stderr, "   Parsed secondary structure after alignment:\n");
					fprintf(stderr, "   %s\n", target_sec_structure);
					exit(1);
				}
			}
			
			// Post-processing of secondary structure input.
			
			for (i = 0; i < totalRes; i++) {
				c = target_sec_structure[i];
				
				if (c == 'c' || c == 's' || c == 'S') {
					c = 'C';
				}
				else if (c == 'h' || c == 't' || c == 'T') {
					c = 'H';
				}
				else if (c == 'e' || c == 'E' || c == 'b' || c == 'B') {
					c = 'B';
				}
				
				target_sec_structure[i] = c;
			}
			
			free(proteus_seq);
			free(pdb_segment);
			free(proteus_segment);
			
			// debug
			//fprintf(stderr, "%s\n", sequence);
			//fprintf(stderr, "%s\n", target_sec_structure);
		}
		else {
			// Assume that the secondary structure file contains the secondary structure in a simple string,
			// and exactly as required, e.g. HHHHCCCBBBCCCHHHBBBCCC
			i = fread(target_sec_structure, sizeof(char), totalRes, ff);
			target_sec_structure[totalRes] = '\0';
			
			if (i < totalRes) {
				fprintf(stderr, "ERROR: Incorrect number of residues in secondary structure file %s.\n",
					secondaryStructureFilename);
				fprintf(stderr, "   %d residues counted, but %d expected based on PDB file.\n", i, totalRes);
				fprintf(stderr, "   Secondary structure file was parsed as a simple string.\n");
				fprintf(stderr, "   Parsed secondary structure:\n");
				fprintf(stderr, "   %s\n", target_sec_structure);
				exit(1);
			}
			
			// Test whether there are too many non-whitespace characters in the secondary structure file.
			i = fread(buffer, sizeof(char), 64, ff);
			non_space_count = 0;
			if (i > 0) {
				for (j = 0; j < i; j++) {
					if (!isspace(buffer[j])) {
						non_space_count++;
					}
				}
			}
			if (non_space_count > 0) {
				fprintf(stderr, "ERROR: Incorrect number of residues in secondary structure file %s.\n",
					secondaryStructureFilename);
				fprintf(stderr, "   At least %d residues counted, but %d expected based on PDB file.\n",
					totalRes+non_space_count, totalRes);
				fprintf(stderr, "   Secondary structure file was parsed as a simple string.\n");
				fprintf(stderr, "   Parsed secondary structure begins with:\n");
				fprintf(stderr, "   %s\n", target_sec_structure);
				exit(1);
			}
			
			fclose(ff);
			
			// Check for illegal characters, convert e/E to B, and convert case if needed.
			
			for (i = 0; i < totalRes; i++) {
				c = target_sec_structure[i];
				
				if (c == 'c') {
					c = 'C';
				}
				else if (c == 'h') {
					c = 'H';
				}
				else if (c == 'b' || c == 'e' || c == 'E') {
					c = 'B';
				}
				
				if ( !(c == 'C' || c == 'H' || c == 'B' || c == '-')) {
					fprintf(stderr, "ERROR: Illegal character '%c' in secondary structure file %s.\n",
						target_sec_structure[i], secondaryStructureFilename);
					exit(1);
				}
				
				target_sec_structure[i] = c;
			}
		}
	}
	
	free(sec_str);
}


/*
 * Call the program 'reduce' to add or reposition hydrogens on the given PDB file
 * 'input_PDB_filename'.
 *
 * If trim_PDB_file is TRUE, this function opens the PDB file and writes a trimmed PDB
 * file with only the ATOM lines from the first chain or model in the PDB file, since
 * reduce (v3.03 or v3.10 at least) takes a long time to add hydrogens to all models
 * in an NMR ensemble. The trimmed PDB file also omits all hydrogen atoms, in order to
 * solve a problem with reduce and hydrogen atom nomenclature. The trimmed PDB file is
 * then passed to reduce. The name of the reduced PDB file is written to the parameter
 * 'reduced_PDB_filename'.
 *
 * For info about reduce, see http://kinemage.biochem.duke.edu/software/reduce.php and
 * Word, et al.(1999) "Asparagine and glutamine: using hydrogen atom contacts in the choice
 * of sidechain amide orientation" J. Mol. Biol. 285, 1733-1745.
 *
 * input_PDB_filename: Name of input PDB file.
 * reduced_PDB_filename: Pointer to memory to which to write the name of the reduced PDB
 *    file.
 * trim_PDB_file_first: Boolean. Whether to trim the PDB file first.
 */
void call_reduce(char *input_PDB_filename, char *reduced_PDB_filename, int trim_PDB_file_first) {
	char trimmed_PDB_filename[512];
	char command[512];
	
	sprintf(trimmed_PDB_filename, "%s_trim", input_PDB_filename);
	sprintf(reduced_PDB_filename, "%sH", input_PDB_filename);
	
	// Open the PDB file and write a trimmed PDB file with only the first chain or model
	// and without hydrogen atoms. This is a simple way to fix the problem where reduce
	// v3.10 reads in PDB file with hydrogen atoms named according to one version of PDB
	// nomenclature, and adds duplicate hydrogens with different atom nomenclature.
	
	if (trim_PDB_file_first) {
		trim_PDB_file(input_PDB_filename, trimmed_PDB_filename, TRUE);
	}
	else {
		sprintf(trimmed_PDB_filename, "%s", input_PDB_filename);
	}
	
	// Call reduce
	
	sprintf(command, "./reduce -BUILD -Quiet -OLDpdb -DB reduce_wwPDB_het_dict.txt %s > %s",
		trimmed_PDB_filename, reduced_PDB_filename);
	system(command);
}


/****************************************************************
 * Changes a chemical shift file from the bmrb format to the shifty format
 ****************************************************************/
static char* bmrb_format_to_shifty_format(char* bmrb_shift_file) {
  
  char *experimental_shifts_in_shifty_format, *change_bmrb_format_to_shifty_format;
  char *perl_script_to_change_format = "bmrb2shifty.pl";
  
  experimental_shifts_in_shifty_format = (char*)malloc((strlen(bmrb_shift_file)+strlen(".shifty")+1)*sizeof(char));
  assert(experimental_shifts_in_shifty_format != NULL);
  sprintf(experimental_shifts_in_shifty_format,"%s.shifty",bmrb_shift_file);
  
  change_bmrb_format_to_shifty_format = (char*)malloc((strlen("perl ")+strlen(perl_script_to_change_format) + 1 +
						       strlen(bmrb_shift_file) + 1 + strlen(experimental_shifts_in_shifty_format) + 1)*sizeof(char));
  assert(change_bmrb_format_to_shifty_format != NULL); 
  sprintf(change_bmrb_format_to_shifty_format,"perl %s %s %s",perl_script_to_change_format, bmrb_shift_file, experimental_shifts_in_shifty_format);
  
  //printf("call = %s\n", change_bmrb_format_to_shifty_format);
  
  if (system(change_bmrb_format_to_shifty_format) != 0) {
	fprintf(stderr, "ERROR: Call to %s failed.\n", perl_script_to_change_format);
	exit(-1);
  }
  free(change_bmrb_format_to_shifty_format);
  
  return(experimental_shifts_in_shifty_format);
  
}//END **********************************************************
