#ifndef __GA_POTENTIAL_H__
#define __GA_POTENTIAL_H__

#include "pepmake.h"

#define POTENTIAL_MODE_MINIMIZE 10
	// Indicates a mode for running the GA Potential, for the purpose of minimizing
	// a structure in GAFolder.
#define POTENTIAL_MODE_EVALUATE 11
	// Indicates a mode for running the GA Potential, for the purpose of evaluating
	// structure quality, and distinguishing decoy structures from a native structure,
	// or distinguishing an NMR structure from an X-ray structure.

// Constants indicating that only certain scores should be calculated.
#define WHICH_SCORES_ALL          20 // Calculate all scores.
#define WHICH_SCORES_CHEM_SHIFT   21 // Calculate only chemical shift scores.
#define WHICH_SCORES_GA_POTENTIAL 22 // Calculate only GA Potential score (i.e. no chemical shift scores).

#define AVGPERCINHBONDS 76.9
#define STDDEVINHBONDS 8.71563

// Ideal chemical shift correlation coefficients for a good-quality structure.
#define IDEAL_CS_CORREL_HA 0.85
#define IDEAL_CS_CORREL_HN 0.70
#define IDEAL_CS_CORREL_N  0.80
#define IDEAL_CS_CORREL_CA 0.85
#define IDEAL_CS_CORREL_CB 0.70
#define IDEAL_CS_CORREL_CO 0.80


/************************************************************************************************************************
 * The following are the initial scaling factors by which to multiply the chemical shift scores and
 * the GA Potential scores. They should all be -1.0 or +1.0.
 ************************************************************************************************************************/

// Initial scaling for chemcial shift scores
#define INITIAL_SCALING_HA 1.0
#define INITIAL_SCALING_HN 1.0
#define INITIAL_SCALING_N  1.0
#define INITIAL_SCALING_CA 1.0
#define INITIAL_SCALING_CB 1.0
#define INITIAL_SCALING_CO 1.0

// Initial scaling for GA Potential scores
#define INITIAL_SCALING_PROSA     1.0
#define INITIAL_SCALING_THREAD    1.0
#define INITIAL_SCALING_BUMP      1.0
#define INITIAL_SCALING_HCOUNT   -1.0
#define INITIAL_SCALING_RAMA      1.0
#define INITIAL_SCALING_OMEGA     1.0
#define INITIAL_SCALING_CHI       1.0
#define INITIAL_SCALING_RADGYR    1.0
#define INITIAL_SCALING_HBENERGY  1.0
#define INITIAL_SCALING_DISULPH  -1.0
#define INITIAL_SCALING_SECSTR    1.0

/************************************************************************************************************************
 * End initial scaling factors.
 ************************************************************************************************************************/



/************************************************************************************************************************
 * The following set of score weights are based on best separation of decoy and native structures,
 * from Mark Berjanskii on Jan 14, 2008. These weight were used in GAFolder and GA Potential starting
 * with v77.
 * These were adjusted on Mar 25, 2008 so that the initial scaling values are all +/-1.0, so that all these weights
 * should be positive.
 ************************************************************************************************************************/
/*
// Weights for chemcial shift scores
#define WEIGHT_HA 1.00 * 10.0
#define WEIGHT_HN 0.81 * 10.0
#define WEIGHT_N  0.58 * 10.0
#define WEIGHT_CA 0.51 * 10.0
#define WEIGHT_CB 0.82 * 10.0
#define WEIGHT_CO 0.34 * 10.0

// Weights for GA Potential scores
#define WEIGHT_PROSA    0.96 * 1.106037956
#define WEIGHT_THREAD   0.63 * 0.2
#define WEIGHT_BUMP     0.71 * 0.29733875 * 0.5
#define WEIGHT_HCOUNT   0.25 * 0.328715333 * 1.1
#define WEIGHT_RAMA     0.20 * 0.854277296
#define WEIGHT_OMEGA    0.40 * 0.308356185
#define WEIGHT_CHI      0.17 * 0.331839855
#define WEIGHT_RADGYR   0.48 * 0.568204099 * 2.5
#define WEIGHT_HBENERGY 0.82 * 1.99054356 * 0.4
#define WEIGHT_DISULPH  5.50 * 3.0 // Not obtained from training; more or less arbitrary.
#define WEIGHT_SECSTR   1.00 * 3.214686966

#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_PROSA 1.0/0.72
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using Prosa threading score).
#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD 1.0/0.42
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using built-in threading score).
#define GLOBAL_COEFF 1.0
	// Number by which to multiply all GA Potential and chemical shift scores.

#define USE_NEAR_NATIVE_SCORE_THRESHOLD 0
	// Whether to use the NEAR_NATIVE_SCORE_THRESHOLD. Set to 1 for TRUE, 0 for FALSE.
#define NEAR_NATIVE_SCORE_THRESHOLD 0.0
	// If the sum of the GA Potential scores for a structure (exclusing chemical shift scores) is below
	// this value, we consider the structure to be close enough to the native structure that we should
	// calculate the chemical shift scores in addition to help further distinguish the structure quality.
*/
/************************************************************************************************************************
 * End score weight set.
 ************************************************************************************************************************/



/************************************************************************************************************************
 * The following set of score weights are based on those from Mark Berjanskii sent on Jan 26, 2008. The coefficients
 * are based largely on those determined by Jianjun (Ben) Zhou based on linear regression of CASP7 decoys and native
 * structures. These weights were used in GAFolder and GA Potential in v79 through v112 (except v106) (the disulphide
 * bond score weight was increased 10-fold as of v96), and then again starting in v124.
 * These were adjusted on Mar 25, 2008 so that the initial scaling values are all +/-1.0, so that all these weights
 * should be positive.
 ************************************************************************************************************************/

// Weights for chemcial shift scores
#define WEIGHT_HA 10.0
#define WEIGHT_HN  8.1
#define WEIGHT_N   5.8
#define WEIGHT_CA  5.1
#define WEIGHT_CB  8.2
#define WEIGHT_CO  3.4

// Weights for GA Potential scores
#define WEIGHT_PROSA    0.96   * 1.106037956 // Coefficient not available; use the one for Mark's Jan 14, 2008 set.
#define WEIGHT_THREAD   0.456  * 0.2
#define WEIGHT_BUMP     0.246  * 0.29733875 * 0.5
#define WEIGHT_HCOUNT   0.1656 * 0.328715333 * 1.1
#define WEIGHT_RAMA     0.06   * 0.854277296
#define WEIGHT_OMEGA    0.06   * 0.308356185 // Arbitrarily set to a low weight rather than 0 as from the training.
#define WEIGHT_CHI      0.06   * 0.331839855 // Arbitrarily set to a low weight rather than 0 as from the training.
#define WEIGHT_RADGYR   1.000  * 0.568204099 * 2.5
#define WEIGHT_HBENERGY 0.076  * 1.99054356 * 0.4
#define WEIGHT_DISULPH  5.50   * 3.0         // Not obtained from training; more or less arbitrary.
#define WEIGHT_SECSTR   0.073  * 3.214686966

#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_PROSA 1.0/1.0 // guess
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using Prosa threading score).
#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD 1.0/0.82
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using built-in threading score).
#define GLOBAL_COEFF 1.0
	// Number by which to multiply all GA Potential and chemical shift scores.

#define USE_NEAR_NATIVE_SCORE_THRESHOLD 0
	// Whether to use the NEAR_NATIVE_SCORE_THRESHOLD. Set to 1 for TRUE, 0 for FALSE.
#define NEAR_NATIVE_SCORE_THRESHOLD 0.0
	// If the sum of the GA Potential scores for a structure (exclusing chemical shift scores) is below
	// this value, we consider the structure to be close enough to the native structure that we should
	// calculate the chemical shift scores in addition to help further distinguish the structure quality.

/************************************************************************************************************************
 * End score weight set.
 ************************************************************************************************************************/



/************************************************************************************************************************
 * The following set of score weights are based on examination of training results on CASP7 decoys and GAFolder "drifting"
 * decoys (using -xa option with gafolder) by Dr. Wishart and Jianjun (Ben) Zhou on Mar 27, 2008. They were introduced
 * into GAFolder starting with v113 and continuing through v123.
 ************************************************************************************************************************/
/*
// Weights for chemcial shift scores
#define WEIGHT_HA 10.0
#define WEIGHT_HN  8.1
#define WEIGHT_N   5.8
#define WEIGHT_CA  5.1
#define WEIGHT_CB  8.2
#define WEIGHT_CO  3.4

// Weights for GA Potential scores
#define WEIGHT_PROSA    0.1370 // Coefficient not available; set it to 10 times WEIGHT_THREAD.
#define WEIGHT_THREAD   0.0137
#define WEIGHT_BUMP     0.0007
#define WEIGHT_HCOUNT   0.0120
#define WEIGHT_RAMA     0.0306
#define WEIGHT_OMEGA    0.0171
#define WEIGHT_CHI      0.0010
#define WEIGHT_RADGYR   0.3126
#define WEIGHT_HBENERGY 0.0458
#define WEIGHT_DISULPH  0.4950
#define WEIGHT_SECSTR   0.0517

#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_PROSA 1.0/1.0 // guess
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using Prosa threading score).
#define CHEM_SHIFT_COEFF_WRT_GA_POTENTIAL_WITH_THREAD 1.0/0.82
	// Weight of all chemical shift scores with respect to the GA Potential scores (when using built-in threading score).
#define GLOBAL_COEFF 1.0
	// Number by which to multiply all GA Potential and chemical shift scores.

#define USE_NEAR_NATIVE_SCORE_THRESHOLD 1
	// Whether to use the NEAR_NATIVE_SCORE_THRESHOLD. Set to 1 for TRUE, 0 for FALSE.
#define NEAR_NATIVE_SCORE_THRESHOLD 0.0
	// If the sum of the GA Potential scores for a structure (exclusing chemical shift scores) is below
	// this value, we consider the structure to be close enough to the native structure that we should
	// calculate the chemical shift scores in addition to help further distinguish the structure quality.
*/
/************************************************************************************************************************
 * End score weight set.
 ************************************************************************************************************************/



#define NORMALIZATION_COUNTER_WEIGHT 100.0
	// Some scores are normalized by dividing them by the number of residues, and are then
	// adjusted by multiplying them by this number in order to bring them back into a normal
	// range.


/* Function prototypes */
void get_GA_potential_scores(ResData *residues, int numRes, int mode, int which_scores, char *PDB_filename,
		int thread_id, char *target_sec_structure, int print_SecStr, int print_SecStr_list, 
		int output_PyMOL_files, char *base_PDB_filename, int useUnweightedScores,
		int use_chem_shift_scores, char *experimental_shifts_in_shifty_format,
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float *scaled_thread_potential_score, float *scaled_bump_score, float *scaled_hcount,
		float *scaled_rama_score, float *scaled_omega_score, float *scaled_chi_score, float *scaled_rad_gyr_score,
		float *scaled_hbenergy, float *scaled_disulphide_bond_score, float *scaled_sec_score,
		float *sum_chem_shift_scores, float *sum_GA_Potential, float *total_sum);
void get_GA_potential_scores_full(ResData *residues, int numRes, int mode, int which_scores, char *PDB_filename,
		int thread_id, char *target_sec_structure, int print_SecStr, int print_SecStr_list,
		int output_PyMOL_files, char *base_PDB_filename, int set_raw_scores, int useUnweightedScores,
		int use_chem_shift_scores, char *experimental_shifts_in_shifty_format,
		
		// Raw scores
		float *raw_chem_shift_coeffs, // array of 6 chemical shift scores
		float *raw_thread_potential_score, float *raw_bump_score, float *raw_hcount,
		float *raw_rama_score, float *raw_omega_score, float *raw_chi_score, float *raw_rad_gyr_score,
		float *raw_hbenergy, float *raw_disulphide_bond_score, float *raw_sec_score,
		
		// Scaled and weighted scores
		float *scaled_chem_shift_scores, // array of 6 chemical shift scores
		float *scaled_thread_potential_score, float *scaled_bump_score, float *scaled_hcount,
		float *scaled_rama_score, float *scaled_omega_score, float *scaled_chi_score, float *scaled_rad_gyr_score,
		float *scaled_hbenergy, float *scaled_disulphide_bond_score, float *scaled_sec_score,
		float *sum_chem_shift_scores, float *sum_GA_Potential, float *total_sum);
void call_reduce(char *input_PDB_filename, char *reduced_PDB_filename, int trim_PDB_file_first);
void call_proteus(char *sequence, char *input_PDB_filename, char* proteus_dir, char *proteus_output_filename);
void parseSecondaryStructureFile(char *secondaryStructureFilename, int totalRes, char *sequence, char *target_sec_structure);

#endif
