#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include <assert.h>

#include "pdbfile.h"
#include "pepmake.h"
#include "rama.h"
#include "ga_potential_scores.h"
#include "rotamer.h"
#include "util.h"

/* Function prototypes */

static int get_longest_helix_length(int numRes);

/* Global variables */

int use_Prosa;
	// Whether to make an external system call to Prosa (prosa2003) when evaluating a peptide.
	// The Prosa score is used in place of the normal threading score.
int computeSecondaryScore; // Whether to compute a secondary structure score.
int use_Proteus;
	// Whether to call Proteus to predict the secondary structure for use with the secondary structure score.
char *proteus_dir; // Directory in which to call Proteus (run_proteus).
char secondary_structure_filename[512];
	// The name of the secondary structure file, either given by the user or computed by Proteus.
char *target_sec_structure;
	// "target_sec_structure" is the secondary structure assignment read in from a file, e.g.
	// HHHHCCCBBBCCCHHHBBBCCC and so on, where H = Helix, B = Beta strand, - or C = coil.


float get_radGyr(ResData *res, int residue_number, double *delta){ 		
	int i;
	int num_res_with_data = 0; // Number of residues with CA atoms; ususally the same as residue_number.
	//int j;
	//float di;
	//float radGyr=0.0;
	float expected_radGyr=0.0;
	float radGyr_score=0.0;
	float min_radGyr;
	double cx = 0.0, cy = 0.0, cz = 0.0; // The geometric center point of the peptide.
	double observed_radGyr = 0.0;

	for (i = 0; i < residue_number; i++){
		if (res[i].atoms[CA][0] != ATOM_COORD_NA) {
			cx += res[i].atoms[CA][0];
			cy += res[i].atoms[CA][1];
			cz += res[i].atoms[CA][2];
			num_res_with_data++;
		}
	}
	if (num_res_with_data == 0) {
		return 0.0f;
	}
	cx /= num_res_with_data;
	cy /= num_res_with_data;
	cz /= num_res_with_data;
	
	// Compute the mean distance between each CA atom and the center point.
	for (i = 0; i < residue_number; i++){
		if (res[i].atoms[CA][0] != ATOM_COORD_NA) {
			observed_radGyr += sqrt(
				(res[i].atoms[CA][0] - cx) * (res[i].atoms[CA][0] - cx)
				+ (res[i].atoms[CA][1] - cy) * (res[i].atoms[CA][1] - cy)
				+ (res[i].atoms[CA][2] - cz) * (res[i].atoms[CA][2] - cz));
		}
	}
	observed_radGyr /= num_res_with_data;
	
	//expected_radGyr = (float)pow(((double)residue_number ), 0.33) * 2.4;
	expected_radGyr = (float)pow(((double)num_res_with_data ), 0.33) * 2.4;
		// Modified here and above to use the number of residues with data for
		// CA atoms, instead of the number of residues, in case some residues
		// don't have data for CA atoms. David Arndt, Dec 10, 2007.
	
	// Adjust the expected radius of gyration based on the longest expected helix length.
	// Added by David Arndt, March 6, 2008.
	if (target_sec_structure != NULL) {
		int longest_helix_length = get_longest_helix_length(residue_number);
		min_radGyr = ((double) longest_helix_length) * ANGSTROMS_PER_HELIX_RESIDUE / 2.0;
		
		//fprintf(stderr, "%s\n", target_sec_structure);
		//fprintf(stderr, "Longest helix: %d\n", longest_helix_length);
		//fprintf(stderr, "min_radGyr: %lf\n", min_radGyr);
		//fprintf(stderr, "previous expected_radGyr: %lf\n", expected_radGyr);
		//fprintf(stderr, "observed_radGyr: %lf\n", observed_radGyr);
		
		if (min_radGyr > expected_radGyr) {
			expected_radGyr = min_radGyr;
		}
	}
	
	*delta = (observed_radGyr - expected_radGyr);
	//printf("observed = %f & expected = %f andf delta = %f\n", observed_radGyr, expected_radGyr, *delta);
	
	// Old algorithm: Assign score penalty if observed RadGyr is not within 10% of the expected RadGyr.
	// New algorithm: Assign score penalty if observed RadGyr is >= 1.1 * expected RadGyr. (Change made Mar 11, 2008).
	//if (observed_radGyr > 0.9*expected_radGyr && observed_radGyr < 1.1*expected_radGyr) {
	if (observed_radGyr < 1.1*expected_radGyr) {
		radGyr_score = 0.00;
	}
	else {
		radGyr_score = fabs(*delta);
	}
	//printf("first return = %f\n",(observed_radGyr - expected_radGyr) * (observed_radGyr - expected_radGyr));

	/* RM: Commented the bottom portion on Nov 14, 2006. 
	return((observed_radGyr - expected_radGyr) * (observed_radGyr - expected_radGyr));

	for (i=0; i<residue_number-1; i++) {
		for (j=i+1; j<residue_number; j++) {
			double d = distance(res[i].atoms[CA], res[j].atoms[CA]);
			radGyr += (float)(d*d);
		}
	}
	di = (float)residue_number;
	radGyr = (1.0 / ((di+1.0)*(di+1.0))) * radGyr;
	radGyr = sqrt((float)radGyr);
	radGyr_score = radGyr - expected_radGyr;
	radGyr_score = (radGyr_score * radGyr_score);
	//printf("radgyr_score = %f\n",radGyr_score);
	*/
	return (radGyr_score);
}


/*
 * Computes the number of residues in the longest helix in target_sec_structure.
 * target_sec_structure cannot be NULL.
 * numRes: Number of residues in the peptide.
 */
static int get_longest_helix_length(int numRes) {
	int i;
	int longest_helix_length = 0;
	int curr_helix_length = 0;
	
	for (i = 0; i < numRes; i++) {
		if (target_sec_structure[i] == 'H') {
			curr_helix_length++;
		}
		else {
			if (curr_helix_length > longest_helix_length) {
				longest_helix_length = curr_helix_length;
			}
			curr_helix_length = 0;
		}
	}
	
	return longest_helix_length;
}


/*
 * Returns a quality score for the entire peptide based on Ramachandran plots.
 *
 * read_rama_tables() should be called before calling this function.
 */
float get_rama_value(ResData *residues, int numRes) {
	int rama_score; // Rama score for a single residue.
	int quality=0; // Score for entire peptide.
	int i;
	int rama_count[4]={0,0,0,0}; // The number of resdidues with a Rama score of i wil be rama_count[i].
	
	for (i=1;i<numRes-1;i++) { /* we don't have any good values for the first and last residue */
		rama_score = lookupRamaScore(residues, i, numRes);
		rama_count[rama_score]++;
		
		quality += (3 - rama_score)*(3 - rama_score);
	}
	/* for(j=0;j<4;j++)
	{ printf("%d  %d ", j, rama[j]);}
	printf("\n");  */
	return ((float) quality) + 1.0F - (float)(rama_count[3]/numRes);

}       /* get_rama_value */

/*
 * Returns a quality score for the entire peptide based on backbone omega angles.
 *
 * Assumes that the omega angle values are in the range from -180 degrees to +180 degrees.
 */
float getOmegaScore(ResData *residues, int numRes) {
	int i;
	float total_omega_score = 0.0F;
	double omega;
	
	for (i=0;i<numRes-1;i++) { // Omit last residue from consideration.
		omega = residues[i].omega;
		
		// old Omaga score
		/*
		if (!(omega >= 170.0 || omega <= -170.0 || (omega <= 10.0 && omega >= -10.0))) {
			if (omega >= 160.0 || omega <= -160.0) {
				total_omega_score += 1; // somewhat bad
			}
			else {
				total_omega_score += 2; // very bad
			}
		}
		*/
		
		if (omega != ANGLE_NA) {
			if (!(omega >= 167.5 || omega <= -167.5 || (omega <= 15.0 && omega >= -15.0))) {
				if (omega >= 90.0) {
					total_omega_score += (float) (167.5 - omega);
				}
				else if (omega <= -90.0) {
					total_omega_score += (float) (omega - -167.5);
				}
				else if (omega > 0.0) {
					total_omega_score += (float) (omega - 15.0);
				}
				else if (omega < 0.0) {
					total_omega_score += (float) (-15.0 - omega);
				}
			}
		}
	}
	
	return total_omega_score;
}


/*
 * Prints the number of residues with each category of Rama score (3, 2, 1, or 0)
 * and the number of residues with each category of Omega score (1 or 0).
 * Rama 3 = phi/psi core
 * Rama 2 = phi/psi allowed
 * Rama 1 = phi/psi generous
 * Rama 0 = phi/psi diallowed
 * Omega 1 = good
 * Omega 0 = bad
 *
 * read_rama_tables() should be called before calling this function.
 */
void printRamaAndOmegaScoreBreakdown(ResData *residues, int numRes) {
	int rama_score; // Rama score for a single residue.
	int i;
	int rama_count[4]={0,0,0,0}; // The number of resdidues with a Rama score of i wil be rama_count[i].
	double omega;
	int omega_count[2]={0,0}; // The number of resdidues with an acceptable omega score is in omega_count[1].
	
	for (i=1;i<numRes-1;i++) { /* we don't have any good values for the first and last residue */
		rama_score = lookupRamaScore(residues, i, numRes);
		rama_count[rama_score]++;
	}
	
	for (i=0;i<numRes-1;i++) { // Omit last residue from consideration.
		omega = residues[i].omega;
		if (omega == ANGLE_NA) {
			omega_count[1]++;
		}
		else if (!(omega >= 170.0 || omega <= -170.0 || (omega <= 10.0 && omega >= -10.0))) {
			omega_count[0]++;
		}
		else {
			omega_count[1]++;
		}
	}
	
	printf("Rama_3\tRama_2\tRama_1\tRama_0\tOmega_1\tOmega_0\n");
	printf("%d\t%d\t%d\t%d\t%d\t%d\n", rama_count[3], rama_count[2], rama_count[1], rama_count[0],
		omega_count[1], omega_count[0]);
	
}


/*
 * Calculates the Chi score for the given peptide.
 *
 * Each residue has 3 characteristic chi1 angles around which chi angles tend to cluster, and these
 * 3 chi angles also depend on the phi and psi angles. Each of these 3 chi angles has an associated
 * probability. For each residue, we determine which of these 3 "canonical" chi angles is closest to
 * the observed chi angle, and use the corresponding probability of this "canonical" chi angle. Since
 * higher probabilities indicate more common chi angles, we calculate the chi score as -1 times the
 * sum of the chi probabilities for each residue.
 *
 * Added by David Arndt, May 30, 2007.
 */
float getChiScore(ResData *residues, int totalRes) {
	int i;
	float chi_score = 0.0F;
	
	for (i = 0; i < totalRes; i++) {
		if ( !(strcmp(residues[i].resName, "GLY") == 0 || strcmp(residues[i].resName, "ALA") == 0) ) {
			chi_score -= getChiProbability(residues[i].resName, residues[i].phi, residues[i].psi, residues[i].chi);
		}
	}
	
	return chi_score;
}


/*
 * Calculates the number of hydrogen bonds and the sum of the hydrogen bond energies, and
 * returns the number of backbone-to-backbone hydrogen bonds identified.
 */
float get_hcount(ResData *res, int number, double *sumEnergy, double *linPerc, double *hperc, int *disulphide_bond_count) {
	int i, /*j,*/ k;
	int hcount=0;
	int *hbond_residue;
	int hbond_resnum=0;
	float score=0,h_perc;
	
	// variables for debug:
	/*
	int num_hbonds_almost_formed = 0;
	double dist_CN;
	//int almost_hbond_eclipsed;
		// Whether the acceptor group or the donor group of an "almost formed" hydrogen bond also participate in a
		// valid hydrogen bond.
	int m;
	Hbonds *hbond_var;
	int eclipsing_partner;
	*/
	
	// Calculate the locations of the hydrogen bonds.
	
	calculate_hbonds(res, number, sumEnergy, disulphide_bond_count);
	
	// Count the number of hydrogen bonds:
	
	hcount=0; // The number of hydrogen bonds (more specifically the number of Hbonds structs specifying acceptors for valid hydrogen bonds).
	score=0;
	hbond_residue=(int *)malloc(number* sizeof(int));

	for(i=0;i<number;i++){     
		hbond_residue[i] = 0;

		for(k=0;k<2;k++){

			if(res[i].hbonds_acceptor[k].partner!=-1 && res[i].hbonds_acceptor[k].hbond_is_valid) {
				hbond_residue[i]=1;
				
				score+= calc_energy_score(res[i].hbonds_acceptor[k].energy);
				hcount++;
			}
			
			if(res[i].hbonds_donor[k].partner!=-1 && res[i].hbonds_donor[k].hbond_is_valid) {
				hbond_residue[i]=1;
			}
			
			// debug: count "almost formed" hydrogen bonds:
			/*
			if (res[i].hbonds_acceptor[k].partner!=-1 && !(res[i].hbonds_acceptor[k].hbond_is_valid)) {
				num_hbonds_almost_formed++;
				
				j = res[i].hbonds_acceptor[k].partner;
				
				for (m = 0; m < 2; m++) {
					hbond_var = &(res[j].hbonds_donor[m]);
					if (hbond_var->partner == i) {
						break;
					}
				}
				// Now we know m such that res[j].hbonds_donor[m] partners with res[i].hbonds_acceptor[k]
				
				dist_CN = distance(res[i].atoms[C1], res[j].atoms[N]);
				printf("Almost: Acceptor %s %d, Donor %s %d. O-H dist = %lf, C-N dist = %lf.\n", res[i].resName, i,
					res[j].resName, j, res[i].hbonds_acceptor[k].distance, dist_CN);
				
				if ( (res[i].hbonds_acceptor[(k+1)%2].partner!=-1 && res[i].hbonds_acceptor[(k+1)%2].hbond_is_valid) ) {
					// The current "almost formed" hydrogen bond is competing with a valid hydrogen bond.
					eclipsing_partner = res[i].hbonds_acceptor[(k+1)%2].partner;
					printf("   Eclipsed by: Acceptor %s %d, Donor %s %d. O-H dist = %lf.\n", res[i].resName, i,
						res[eclipsing_partner].resName, eclipsing_partner, res[i].hbonds_acceptor[(k+1)%2].distance);
				}
				else if ( (res[j].hbonds_donor[(m+1)%2].partner!=-1 && res[j].hbonds_donor[(m+1)%2].hbond_is_valid) ) {
					// The current "almost formed" hydrogen bond is competing with a valid hydrogen bond.
					eclipsing_partner = res[j].hbonds_donor[(m+1)%2].partner;
					printf("   Eclipsed by: Acceptor %s %d, Donor %s %d. O-H dist = %lf.\n", res[eclipsing_partner].resName, eclipsing_partner,
						res[j].resName, j, res[j].hbonds_donor[(m+1)%2].distance);
				}
				else {
					//almost_hbond_eclipsed = FALSE;
				}
				
				printf("   C: (%lf, %lf, %lf)\n", res[i].atoms[C1][0], res[i].atoms[C1][1], res[i].atoms[C1][2]);
				printf("   O: (%lf, %lf, %lf)\n", res[i].atoms[O][0], res[i].atoms[O][1], res[i].atoms[O][2]);
				printf("   HN: (%lf, %lf, %lf)\n", res[j].atoms[HN][0], res[j].atoms[HN][1], res[j].atoms[HN][2]);
				printf("   N: (%lf, %lf, %lf)\n", res[j].atoms[N][0], res[j].atoms[N][1], res[j].atoms[N][2]);
			}
			*/
		}     
	}
	
	// debug
	//printf("Number of hydrogen bonds almost formed: %d\n", num_hbonds_almost_formed);
	
	//printf("score due to the hydrogen bonds = %g\n", score);
	/*** calculate the hydrogen score according to the number of residues involved and energy distribution
	number of residue = 70% ~ 80%    energy in range -1 ~ -3
	****/ 

	hbond_resnum=0; // The number of residues that participate in at least one hydrogen bond.
	for(i=0; i<number; i++){ 
		if(hbond_residue[i] == 1){
			hbond_resnum++;
		}
	}

	//printf("hbond_resnum = %d and number=%d\n", hbond_resnum, number);

	//h_perc = Percentage of hydrogen bonds 
	h_perc=(float)(hbond_resnum + *disulphide_bond_count)/number*100;
	//fprintf(stderr, "disulphide_bond_count = %d\n", *disulphide_bond_count);
	//fprintf(stderr, "h count = %d  H percentage = %f\n",hcount, h_perc);
	if(h_perc < 70){
		score =(70 - h_perc)*(70 - h_perc);  // SJN; instead of +=
		*linPerc = 70 - h_perc;
	}else{
		score = 0;   //Why is the score = 0 if the % of hydrogen bonds is > 70 ? RM Nov 08, 2006. 
		*linPerc = 0;
	}


	// *sumEnergy /= number;
		// Commented out by David Arndt, Mar 5, 2008. We will normalize the score outside this function.
	*hperc = h_perc;   	

	return hcount;    
	//return score; //Commented by RM Nov 8, 2006.	   
}

/*
 * Calculates the locations of all backbone-to-backbone hydrogen bonds in the given peptide,
 * sets the value pointed to by 'sumEnergy' to the sum of the hydrogen bond energies, and sets
 * the value pointed to be disulphide_bond_count to the number of disulphide bonds between CYS residues.
 * This function will re-calculate the data in all the hbonds_acceptor and hbonds_donor member
 * variables in the ResData array 'res' with length 'number'.
 *
 * The algorthm here is adapted from the algorithm in VADAR. Changes to the VADAR algorithm
 * include:
 *   (1) Fixed bug so PRO residues cannot be donors.
 *   (2) Fixed some possible data inconsistencies (both residues participating in a hydrogen
 *       bond must have Hbond structs referring to the hydrogen bond).
 *   (3) Expanded algorithm to allow a residues to be an acceptor for up to 2 hydrogen bonds
 *       and a donor for up to 2 hydrogen bonds.
 *   (4) Added additional restrictions on the geometry of hydrogen bonds.
 *
 *
 *               hydrogen bond
 *                |
 * Acceptor   C=O...H-N   Donor
 *
 *              |---|     O-H distance is in variable dist_OH
 *              |-----|   O-N distance is in variable dist_ON
 *
 *
 *      Cys   CA-CB-S-S-CB-CA   Cys
 *
 *                  |-|  S-S distance for disulphide bond between two Cysteine residues is stored in variable dist_SS
 */
void calculate_hbonds(ResData *res, int number, double *sumEnergy, int *disulphide_bond_count) {
	int i, j, k;
	int new_acc_idx, new_don_idx;
	int partner_with_which_to_break;
	Hbonds *hbond_var;
	double dist_OH, dist_ON, dist_SS, dist_CN;	
	double hbond_angle, COH_angle;
	Vec3D vec_NH, vec_OC, vec_OH;
	double energy;
	int hbond_is_valid = FALSE;
	
	int debug_print_bad_hbonds = FALSE;
	
	*sumEnergy  = 0.0;
	*disulphide_bond_count = 0;
	  
	for(i=0;i<number;i++){
		for(k=0;k<2;k++){
			res[i].hbonds_acceptor[k].partner=-1;
			res[i].hbonds_acceptor[k].distance=-1;
			res[i].hbonds_donor[k].partner=-1;
			res[i].hbonds_donor[k].distance=-1;
		}
	}

	for (i=0; i<number; i++) {
		
		if (res[i].atoms[O][0] == ATOM_COORD_NA || res[i].atoms[C1][0] == ATOM_COORD_NA) {
			continue;
		}
		
		for (j=0; j<number-1; j++) { //since the N-H bond will be in the same direction as the previous C=O bond, due to planarity of peptide
			
			if (res[j].atoms[N][0] == ATOM_COORD_NA || res[j].atoms[HN][0] == ATOM_COORD_NA) {
				continue;
			}
			
			// Skip residue if the donor would be a Proline. This check avoids
			// a bug where a hydrogen bond might be erroneously called.
			// Added by David Arndt, Mar 27, 2007.
			if (strcmp(res[j].resName, "PRO") == 0) {
				continue;
			}
			
			difference(res[j].atoms[N], res[j].atoms[HN], vec_NH);
			difference(res[i].atoms[O], res[i].atoms[C1], vec_OC);
			difference(res[i].atoms[O], res[j].atoms[HN], vec_OH);
			hbond_angle = angle(vec_NH, vec_OC);
			COH_angle = angle(vec_OC, vec_OH);
			dist_OH = distance(res[j].atoms[HN], res[i].atoms[O]);
			dist_ON = distance(res[j].atoms[N], res[i].atoms[O]);
			dist_CN = distance(res[i].atoms[C1], res[j].atoms[N]);
			
			if( j>i+2 && (strcmp(res[i].resName,"CYS") == 0) && (strcmp(res[j].resName,"CYS") == 0) ){
				
				if (res[j].atoms[CG][0] != ATOM_COORD_NA && res[i].atoms[CG][0] != ATOM_COORD_NA) {
					// This condition should not be satisfied if the structure does not have Sulphur atoms
					// for both CYS residues, such as when a PDB file has no sidechains.
					
					dist_SS = distance(res[j].atoms[CG], res[i].atoms[CG]);
					
					//printf("i = %d, j = %d, dist_SS = %lf\n", i, j, dist_SS);
					//printf("   i (%lf, %lf, %lf), j (%lf, %lf, %lf)\n",
					//	res[i].atoms[CG][0], res[i].atoms[CG][1], res[i].atoms[CG][2],
					//	res[j].atoms[CG][0], res[j].atoms[CG][1], res[j].atoms[CG][2]);
					
					if(dist_SS < 2.5){
						*disulphide_bond_count += 1;
					}
				}
			}

			// debug
			/*
			//if ( !(res[j].atoms[H][0] == 0.0 && res[j].atoms[H][1] == 0.0 && res[j].atoms[H][2] == 0.0) ) {
				printf("(%lf, %lf, %lf) (%lf, %lf, %lf) (%lf, %lf, %lf)\n",
					res[i].atoms[O][0], res[i].atoms[O][1], res[i].atoms[O][2],
					res[j].atoms[H][0], res[j].atoms[H][1], res[j].atoms[H][2],
					res[j].atoms[N][0], res[j].atoms[N][1], res[j].atoms[N][2] );
				if (dist_OH < 4.0 && dist_OH >= 3.5) {
					printf("%s %d to %s %d: %lf *\n", res[i].resName, res[i].resNum, res[j].resName, res[j].resNum, dist_OH);
				}
				else if (dist_OH < 3.5) {
					printf("%s %d to %s %d: %lf OK\n", res[i].resName, res[i].resNum, res[j].resName, res[j].resNum, dist_OH);
				}
				else {
					printf("%s %d to %s %d: %lf\n", res[i].resName, res[i].resNum, res[j].resName, res[j].resNum, dist_OH);
				}
			//}
			*/
			
			if( (dist_OH < dist_ON) && (fabs(COH_angle) > 80) && (fabs(hbond_angle) < 90) && (fabs(j-i)>1) ) {
				// (fabs(COH_angle) > 80) added March 26, 2007 by David Arndt.
				
				if ((dist_OH < 3.5) && (dist_OH > 1.37) && (dist_CN < 4.6) &&
						(dist_OH<(2.5+fabs((double)cos((double)RAD(hbond_angle))))) ) { // I don't think the fabs does anything here.
					// (dist_OH > 1.37) added March 26, 2007 by David Arndt to prevent the H atom from being too close to the O atom.
					// The value "1.37" is approximate.
					// (dist_CN < 4.5) added March 26, 2007 by David Arndt because Dr. Wishart wanted to add it. Changed to 4.6 on
					// March 11, 2008.
					
					// The hydrogen bond is a valid hydrogen bond.
					
					hbond_is_valid = TRUE;
				}
				else if ((dist_OH < 3.5 + 2.0) && (dist_CN < 4.6 + 2.0) ) {
					// We have identified a hydrogen bond that is "almost formed." The purpose in identifying these is
					// so that in the GAFolder energy minimizer we can use the CCD algorithm to try to bring the residues
					// closer together to form the hydrogen bonds properly, to form beta-sheet and alpha-helix structure.
					
					hbond_is_valid = FALSE;
				}
				else {
					// Not a valid or even an "almost formed" hydrogen bond. Skip entirely.
					continue;
				}
				
				
				// Find indices into hbonds_acceptor and hbonds_donor where we can store the
				// newly-found hydrogen bond.
				new_acc_idx = getHbondIndex_GA_Pot(dist_OH, res[i].hbonds_acceptor);
				new_don_idx = getHbondIndex_GA_Pot(dist_OH, res[j].hbonds_donor);
				
				if (new_acc_idx != -1 && new_don_idx != -1) {
					// We want to ensure that every hydrogen bond has a record for an acceptor and a
					// donor. That is, each hydrogen bond should be represented in an Hbonds struct in
					// the acceptor residue and in an Hbonds struct in the donor residue.
					
					// If the new hydrogen bond will replace an existing hydrogen bond in which residue
					// i is the acceptor, remove the link to residue i in the donor residue for this
					// soon-to-be replaced hydrogen bond.
					partner_with_which_to_break = res[i].hbonds_acceptor[new_acc_idx].partner;
					if (partner_with_which_to_break != -1) {
						for (k = 0; k < 2; k++) {
							hbond_var = &(res[partner_with_which_to_break].hbonds_donor[k]);
							if (hbond_var->partner == new_acc_idx) {
								hbond_var->partner = -1;
								hbond_var->distance = -1;
								break;
							}
						}
					}
					
					// If the new hydrogen bond will replace an existing hydrogen bond in which residue
					// j is the donor, remove the link to residue j in the acceptor residue for this
					// soon-to-be replaced hydrogen bond.
					partner_with_which_to_break = res[j].hbonds_donor[new_don_idx].partner;
					if (partner_with_which_to_break != -1) {
						for (k = 0; k < 2; k++) {
							hbond_var = &(res[partner_with_which_to_break].hbonds_acceptor[k]);
							if (hbond_var->partner == new_don_idx) {
								hbond_var->partner = -1;
								hbond_var->distance = -1;
								break;
							}
						}
					}
					
					// Store the information for the new hydrogen bond.
					
					res[i].hbonds_acceptor[new_acc_idx].partner=j;
					res[i].hbonds_acceptor[new_acc_idx].distance=dist_OH;
					res[i].hbonds_acceptor[new_acc_idx].hbond_is_valid = hbond_is_valid;
					
					res[j].hbonds_donor[new_don_idx].partner=i;
					res[j].hbonds_donor[new_don_idx].distance=dist_OH;
					res[j].hbonds_donor[new_don_idx].hbond_is_valid = hbond_is_valid;
					
					if (hbond_is_valid) {
						energy = get_hbond_energy(res[i], res[j]);
						if (energy < 0) { // energy may be 0 or 99 with certain parameters
							res[i].hbonds_acceptor[new_acc_idx].energy = energy;
							res[j].hbonds_donor[new_don_idx].energy = energy;
							
							// Add hydrogen bond energy to the sum.
							*sumEnergy += energy;
						}
					}
					
				}
			}
			
			if (debug_print_bad_hbonds) {
				// Print information about potential hydrogen bonds that do not meet the criteria for being
				// true hydrogen bonds.
				
				//if (fabs(j-i)>1 && ((dist_OH - dist_ON) < 0.5) && (dist_OH < 3.5 + 0.5) && (dist_CN < 4.6 + 0.5)
				//		&& (dist_OH < 2.5 + 1.0 + 0.5) ) {
				if (fabs(j-i)>1 && (dist_OH < 3.5) && ((dist_OH - dist_ON) < 0.5) && (fabs(hbond_angle) < 90 + 90) &&
						(dist_OH < 2.5 + 1.0 + 0.5) ) {
					
					if (!(dist_OH < dist_ON)) {
						printf("%s %d to %s %d: dist_OH >= dist_ON, dist_OH = %lf, dist_ON = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							dist_OH, dist_ON);
					}
					
					if ( !(fabs(COH_angle) > 80)) {
						printf("%s %d to %s %d: COH_angle <= 80, COH_angle = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							COH_angle);
					}
					
					if ( !(fabs(hbond_angle) < 90)) {
						printf("%s %d to %s %d: hbond_angle >= 90, hbond_angle = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							hbond_angle);
					}
					
					if ( !(dist_OH < 3.5)) {
						printf("%s %d to %s %d: dist_OH >= 3.5, dist_OH = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							dist_OH);
					}
					
					if ( !(dist_OH > 1.37)) {
						printf("%s %d to %s %d: dist_OH <= 1.37, dist_OH = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							dist_OH);
					}
					
					if ( !(dist_CN < 4.6)) {
						printf("%s %d to %s %d: dist_CN >= 4.6, dist_CN = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							dist_CN);
					}
					
					if ( !(dist_OH<(2.5+fabs((double)cos((double)RAD(hbond_angle)))))) {
						printf("%s %d to %s %d: dist_OH >= (2.5+fabs((double)cos((double)RAD(hbond_angle)))), dist_OH = %lf, hbond_angle = %lf, cos((double)RAD(hbond_angle)) = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							dist_OH, hbond_angle, cos((double)RAD(hbond_angle)));
					}
					
					float rON, rCH, rOH, rCN;

					rON = distance(res[i].atoms[O],  res[j].atoms[N]);
					rCH = distance(res[i].atoms[C1], res[j].atoms[HN]);
					rOH = distance(res[i].atoms[O],  res[j].atoms[HN]);
					rCN = distance(res[i].atoms[C1], res[j].atoms[N]);
					
					if (rON < 1.5625) {
						printf("%s %d to %s %d: rON < 1.5625, rON = %f\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							rON);
					}
					
					if (rCH < 1.5625) {
						printf("%s %d to %s %d: rCH < 1.5625, rCH = %f\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							rCH);
					}
					
					if (rOH < 1.5625) {
						printf("%s %d to %s %d: rOH < 1.5625, rOH = %f\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							rOH);
					}
					
					if (rCN < 1.5625) {
						printf("%s %d to %s %d: rCN < 1.5625, rCN = %f\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							rCN);
					}
					
					double energy = get_hbond_energy(res[i], res[j]);
					
					// This may still be considered a real hydrogen bond, but if energy is < -0.5, we may
					// ignore it when assigning secondary structure.
					if ( !(energy < -0.5)) {
						printf("%s %d to %s %d: energy >= -0.5, energy = %lf\n",
							res[i].resName, res[i].id, res[j].resName, res[j].id,
							energy);
					}
				}
				
			}
		}
	}
	
}

float calc_energy_score(float energy){
  if(energy < -3)
    { return  (energy+3)*(energy+3);}
  else if(energy > -1)
    return (energy+1)*(energy+1);
  else 
    return 0;
}

/*
 * Comments based on observation by David Arndt:
 * This function returns the index in the given array 'hbonds' that can
 * be used to store information about a new hydrogen bond. The returned
 * index may have previously not held a hydrogen bond, or it may have
 * held a hydrogen bond with an H-O distance greater than the given
 * 'hodist' (the H-O distance of the new hydrogen bond, in angstroms).
 * If the hydrogen bonds that are already in the 'hbonds' array all have
 * closer H-O distances than the new bond, then -1 is returned (the new
 * hydrogen bond is rejected).
 */
int getHbondIndex_GA_Pot(double hodist, Hbonds *hbonds){
	
	int j, maxIdx=0;
	float maximum;
	
	maximum = 0;
	
	for (j=0; j<2; j++) {
		if (hbonds[j].partner == -1) {
			return(j);
		}
		else if (hbonds[j].distance > maximum) {
			maximum = hbonds[j].distance;
			maxIdx = j;
		}
	}

	if (hodist < maximum) {
		return(maxIdx);
	}
	else {
		return(-1);
	}
	
}


/*********************************************************************
 * PURPOSE: to calculate the hbond energy between two residues
 */
float get_hbond_energy(ResData res1, ResData res2){
	float energy;
	float rON, rCH, rOH, rCN;

	rON = distance(res1.atoms[O],res2.atoms[N]);
	rCH = distance(res1.atoms[C1], res2.atoms[HN]);
	rOH = distance(res1.atoms[O], res2.atoms[HN]);
	rCN = distance(res1.atoms[C1],res2.atoms[N]);

	if(rON < 1.5625 || rCH < 1.5625 || rOH < 1.5625 || rCN < 1.5625)
		return 99;

	/* formula from kabsch and sander, 1983
	 * hbond energy should be less than -0.5 kcal/mol
	 * a good hbond has ~ -3.0 kcal/mol
	 */
	energy = 332 * .42 * .20 * ( 1/rON + 1/rCH - 1/rOH - 1/rCN);

	if (energy > 0)
		energy = 0;
																			
	return(energy);
																			
}       /* get_hbond_energy */
 
/**************************************************************************
* Get Bump Score
*************************************************************************/
float get_bump_score(ResData *res, int n)
{
  	int i, j, k, m, atom_index;
	int k_adj, m_adj;
	float bump_score=0;
	int violations=0;
	int atom_index_for_i = -1;
	int atom_index_for_j = -1;
	
	int numAtomsToConsider = 7;
	int atomsToConsider[]= {N, CA, CB, C1, O, CG, -1};
		// List of which atoms to consider for bumps.
		// -1 indicates a special case: look for the C-gamma atom that is
		// not at index CG, such as in THR, ILE, VAL.
	float minDistances[6][6]= {
					{3.6, 3.5, 3.0, 2.8, 2.5, 3.1},
					{3.5, 3.7, 3.5, 3.6, 2.9, 3.4},
					{3.0, 3.5, 3.5, 3.3, 2.9, 3.2}, 
					{2.8, 3.6, 3.3, 3.7, 2.7, 3.3},
					{2.5, 2.9, 2.9, 2.7, 2.9, 2.8},
					{3.1, 3.4, 3.2, 3.3, 2.8, 3.3} };
		// These bump distances were obtained by Jack Liang (June 2007) by analyzing the
		// observed atom-atom distances in a set of 200 (?) high-resolution X-ray structures.


	for (i=0; i<n-1; i++) {
		for (j=i+2; j<n; j++) {
			for ( k=0; k<numAtomsToConsider; k++) {
				
				if ( atomsToConsider[k]== CB && (strcmp(res[i].resName, "GLY") == 0) ) { // "GLY" has no CB, kick out of this situation.
					continue;
				}
				
				if (  // "GLY", "ALA", "SER" and "CYS" have no CG, so kick out of these situations.
					atomsToConsider[k]== CG &&
					( (strcmp(res[i].resName, "GLY") == 0) || (strcmp(res[i].resName, "ALA") == 0) ||
					(strcmp(res[i].resName, "SER") == 0) || (strcmp(res[i].resName, "CYS") == 0) )
					) {
					continue;
				}
				
				if (atomsToConsider[k]== CG && res[i].names[CG][1] != 'C') {
					// The residue does not have a carbon atom in position CG. We only want to count
					// bumps for gamma-atoms that are carbons. This problem may occur with THR residues.
					continue;
				}
				
				if (atomsToConsider[k] == -1) {
					// Look for other C-gamma atom if applicable.
					if (strcmp(res[i].resName, "THR") == 0 || strcmp(res[i].resName, "ILE") == 0 ||
							strcmp(res[i].resName, "VAL") == 0) {
						
						atom_index_for_i = -1;
						for (atom_index = CG+1; atom_index <= res[i].max_atom_index; atom_index++) {
							if (strncmp(&(res[i].names[atom_index][1]), "CG", 2) == 0) {
								atom_index_for_i = atom_index;
								break;
							}
						}
						
						if (atom_index_for_i == -1) {
							continue;
						}
					}
					else {
						continue;
					}
				}
				else {
					atom_index_for_i = atomsToConsider[k];
				}
				
				// Ignore bumps between non-existent atoms.
				if (res[i].atoms[atom_index_for_i][0] == ATOM_COORD_NA) {
					continue;
				}
				
				for (m=0; m<numAtomsToConsider; m++) {
					
					if ( atomsToConsider[m]== CB && (strcmp(res[j].resName, "GLY") == 0) ) { // "GLY" has no CB, kick out of this situation.
						continue;
					}
					
					if ( // "GLY", "ALA", "SER" and "CYS" have no CG, so kick out of these situations.
						atomsToConsider[m]== CG && 
						( (strcmp(res[j].resName, "GLY") == 0) || (strcmp(res[j].resName, "ALA") == 0) ||
						(strcmp(res[j].resName, "SER") == 0) || (strcmp(res[j].resName, "CYS") == 0) )
						) {
						continue;
					}
					
					if (atomsToConsider[m]== CG && res[j].names[CG][1] != 'C') {
						// The residue does not have a carbon atom in position CG. We only want to count
						// bumps for gamma-atoms that are carbons. This problem may occur with THR residues.
						continue;
					}
					
					if (atomsToConsider[m] == -1) {
						// Look for other C-gamma atom if applicable.
						if (strcmp(res[j].resName, "THR") == 0 || strcmp(res[j].resName, "ILE") == 0 ||
								strcmp(res[j].resName, "VAL") == 0) {
							
							atom_index_for_j = -1;
							for (atom_index = CG+1; atom_index <= res[j].max_atom_index; atom_index++) {
								if (strncmp(&(res[j].names[atom_index][1]), "CG", 2) == 0) {
									atom_index_for_j = atom_index;
									break;
								}
							}
							
							if (atom_index_for_j == -1) {
								continue;
							}
						}
						else {
							continue;
						}
					}
					else {
						atom_index_for_j = atomsToConsider[m];
					}
					
					// Ignore bumps between non-existent atoms.
					if (res[j].atoms[atom_index_for_j][0] == ATOM_COORD_NA) {
						continue;
					}
					
					double d=0;
					d = distance( res[i].atoms[atom_index_for_i],  res[j].atoms[atom_index_for_j]);
					
					// Determine distance threshold. For the special case of looking at the other
					// C-gamma atoms, we set k_adj and/or m_adj to the CG column of minDistances[][].
					
					if (atomsToConsider[k] == -1) {
						k_adj = k-1;
					}
					else {
						k_adj = k;
					}
					
					if (atomsToConsider[m] == -1) {
						m_adj = m-1;
					}
					else {
						m_adj = m;
					}
					
					double dist_threshold = minDistances[k_adj][m_adj];
					
					if (d < dist_threshold) {
						
						// debug
						/*
						fprintf(stderr, "Bump: %lf < %lf --> +%lf\n   %3s %d: %d \"%4s\" (%lf, %lf, %lf)\n   %3s %d: %d \"%4s\" (%lf, %lf, %lf)\n",
							d, dist_threshold, (dist_threshold - d),
							res[i].resName, i, atom_index_for_i, res[i].names[atom_index_for_i],
							res[i].atoms[atom_index_for_i][0], res[i].atoms[atom_index_for_i][1], res[i].atoms[atom_index_for_i][2],
							res[j].resName, j, atom_index_for_j, res[j].names[atom_index_for_j],
							res[j].atoms[atom_index_for_j][0], res[j].atoms[atom_index_for_j][1], res[j].atoms[atom_index_for_j][2]
							);
						*/
						
						bump_score += (dist_threshold - d);
						violations++;
						
						// set up second threshold as 70% of the first one for adding extra penalty.
						if (d < dist_threshold * 0.7) {
							
							// debug
							//fprintf(stderr, "   *** %lf < %lf --> +10\n", d, dist_threshold * 0.7);
							
							bump_score += 10;
						}
					}
					
				}// end of for (m)
			}// end of for (k)
		}// end of for (j)
	}// end of for (i)
	
	return bump_score;
}


/*
 * Call Prosa (prosa2003) to obtain the Prosa threading score for the entire protein in
 * the given PDB file. The Prosa threading score is used to replace the normal Thred score
 * in the set of GA Potential scores. Prosa2003 must be installed locally, with the
 * executable prosa2003 in the system's PATH environment variable. As of this writing,
 * Prosa2003 is free for academic researchers, but requires a separate licence for each
 * machine that it is installed on.
 *
 * Set the parameter thread_id to the ID of the current thread if this function will be
 * called in parallel.
 * 
 * Prosa can be downloaded from:
 * http://www.came.sbg.ac.at/typo3/index.php?id=prosa
 */
float getProsaScore(char *PDB_filename, int thread_id) {
	
	FILE *pwd_fp;
	FILE *prosa_cmd_fp;
	FILE *prosa_ana_fp;
	char prosa_cmd_filename[32];
	char prosa_ana_filename[32];
	char buffer[512];
	char current_dir[512];
	char prosa_call[64];
	char PDB_filename_full[512]; // Full path to PDB file
	int resNum; // not used in calculation
	float currentResPairScore;
	float currentResProsaScore;
	float dummy; // not used in calculation
	float totalProsaScore;
	
	// Create filenames with appropriate thread IDs.
	
	snprintf(prosa_cmd_filename, 32, "prosa_command.%d.cmd", thread_id);
	snprintf(prosa_ana_filename, 32, "prosa_energy.%d.ana", thread_id);
	
	// Determine full path to the given PDB file.
	if (PDB_filename[0] == '/') {
		// Assume full path has already been given.
		snprintf(PDB_filename_full, 512, "%s", PDB_filename);
	}
	else {
		// Determine the current working directory
		
		pwd_fp = popen("pwd", "r");
		assert(pwd_fp != NULL && "ga_potential.c:pwd_fp is NULL in getProsaScore");
		while(fgets(buffer, 512, pwd_fp) != NULL) {
			sscanf(buffer,"%s", current_dir);
		}
		pclose(pwd_fp);
		
		// debug
		//printf("Current dir is >%s<\n", current_dir);
		
		snprintf(PDB_filename_full, 512, "%s/%s", current_dir, PDB_filename);
	}
	
	// debug
	//printf("Full PDB path name is >%s<\n", PDB_filename_full);
	
	// Write a Prosa .cmd file
	
	prosa_cmd_fp = fopen(prosa_cmd_filename, "w");
	if (prosa_cmd_fp != NULL) {
		fprintf(prosa_cmd_fp, "pdb_dir = /\n"); // Set pdb_dir to / so we can ignore it.
		fprintf(prosa_cmd_fp, "read pdb %s obj1\n", PDB_filename_full); // Command to read the PDB file
		fprintf(prosa_cmd_fp, "analyse energy *\n");
		fprintf(prosa_cmd_fp, "print energy obj1 prosa_energy.%d\n", thread_id);
		fprintf(prosa_cmd_fp, "quit\n");
		fclose(prosa_cmd_fp);
	}
	else {
		fprintf(stderr, "ERROR: Could not write Prosa command file \"%s\". ga_potential.c\n", prosa_cmd_filename);
		exit(1);
	}
	
	// Run Prosa
	
	sprintf(prosa_call, "prosa2003 %s", prosa_cmd_filename);
	if (system(prosa_call) != 0) {
		fprintf(stderr, "ERROR: Call to Prosa failed.\n");
		exit(-1);
	}
	
	// Parse .ana file to determine the Prosa score for the entire protein. We need to use only half of the
	// Pair score, so that we do not count it twice.
	
	prosa_ana_fp = fopen(prosa_ana_filename, "r");
	if (prosa_ana_fp != NULL) {
		
		// Skip first 2 lines.
		fgets(buffer, 512, prosa_ana_fp);
		fgets(buffer, 512, prosa_ana_fp);
		
		totalProsaScore = 0.0F;
		while(fgets(buffer, 512, prosa_ana_fp) != NULL) {
			sscanf(buffer,"%d %f %f %f", &resNum, &currentResPairScore, &dummy, &currentResProsaScore);
			totalProsaScore += currentResProsaScore - 0.5F*currentResPairScore;
		}
		
		fclose(prosa_ana_fp);
	}
	else {
		fprintf(stderr, "ERROR: Could not open Prosa .ana file. ga_potential.c\n");
		exit(1);
	}
	
	return totalProsaScore;
}


/*
 * Prints the C-alpha to C-alpha distances of the peptide.
 */
void print_C_alpha_distances(ResData *residues, int numRes) {
	int i;
	double dist;
	
	printf("C-alpha to C-alpha distances:\n");
	
	for (i = 1; i < numRes; i++) {
		if (residues[i-1].atoms[CA][0] != ATOM_COORD_NA && residues[i].atoms[CA][0] != ATOM_COORD_NA) {
			dist = distance(residues[i-1].atoms[CA], residues[i].atoms[CA]);
			printf("%lf\n", dist);
		}
	}
	
}
