#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "vec3D.h"
#include "pdbfile.h"
#include "potential_data.h"
#include "potential.h"
#include "pepmake.h"

#define DEBUG 0

/*
 * Note by David Arndt:
 * This function allocates a new DOT struct, puts it in the linked list
 * before the given DOT d, and returns a pointer to the new DOT struct.
 */
static DOT  *CreateDOT(DOT *d) {
  DOT    *new=NULL;
	
  if( (new=(DOT *) malloc( (unsigned) sizeof(DOT) ))==NULL ) {
    fprintf( stderr, "Memory allocation failure...\n" );
    exit(-1);
  }
  memset( new, 0, (unsigned) sizeof(DOT) );
  if( d != NULL ) {
    new->next = d;
    d->last = new;
  }
  return(new);
}

static void  FreeDOT(DOT *d) {
  DOT   *p = d, *next;
  while(p != NULL) {
    next = p->next;
    free(p);
    p = next;
  }
  return;
}


static DOT *Simplify(PDBFile *pdb) {
  DOT             *new=NULL;
  Atom            *atm;
  Vec3D           CAx, CBx, last_CA, u;
  double          x;
  int         		i, n = 0, flag = 0;
  
  for (i=pdb->numRes-1; i>=0; i--) {
  	 n = getFirstAtomOfResidue(pdb, i);
	 n = getNextAtomIndex(pdb, n, " CA ", GET_ATOM_SAME_RES);
	 atm = &pdb->atoms[n];
	 copyVector(atm->point, CAx);

    if (strncmp(atm->resName, "GLY", 3)) {
		 n = getNextAtomIndex(pdb, n, " CB ", GET_ATOM_SAME_RES);
    } else {
		 n = getNextAtomIndex(pdb, n, "1HA ", GET_ATOM_SAME_RES);
    }
	 copyVector(pdb->atoms[n].point, CBx);
//    printf("CB = %lf %lf %lf\n", CBx[0],CBx[1],CBx[2]);

    /* Represent the peptide group */
    if (flag) {
      new = CreateDOT(new);
      new->id = i; //p->n;
      new->type = 'p';
      difference(CAx, last_CA, u);
      x = sqrt(dotproduct(u, u)) / 2.0;
		normalize(u, u);
		scale(u, x);
      summation(last_CA, u, new->r);
      if (DEBUG) {
		   printf("\t\t%lf %lf %lf\n", new->r[0],new->r[1],new->r[2]);
		}
    }

    /* Respresent the side chain */
    new = CreateDOT(new);
    new->id = i; //p->n;
	 new->type = res_str_to_char(atm->resName);  //p->atom->letter;
	 difference(CBx, CAx, u);
	 normalize(u, u);
	 scale(u, 2.4);	
    summation(CAx, u, new->r);
    if (DEBUG) {
	    printf("\t%lf %lf %lf\n", CAx[0],CAx[1],CAx[2]);
	 }
    copyVector(CAx, last_CA);
	 flag = 1;
  }
  return(new);
}

static DOT *Simplify2(ResData *res, int numRes) {
  DOT             *new=NULL;
  Vec3D           CAx, CBx, last_CA, u, Np_CA, C1_CA;
  double          x;
  int         		i, flag = 0;
  
  for (i=numRes-1; i>=0; i--) {
	if (res[i].atoms[CA][0] == ATOM_COORD_NA) {
		continue;
	}
	
	copyVector(res[i].atoms[CA], CAx);
	
    if (strncmp(res[i].resName, "GLY", 3) != 0 &&
			res[i].atoms[CB][0] != ATOM_COORD_NA) {
		copyVector(res[i].atoms[CB], CBx);
    } else {
		// Old code: Does not work if HA or HA2 atoms are absent.
		//CBx[0] = (res[i].atoms[HA][0] + res[i].atoms[HA2][0]) / 2.0;
		//CBx[1] = (res[i].atoms[HA][1] + res[i].atoms[HA2][1]) / 2.0;
		//CBx[2] = (res[i].atoms[HA][2] + res[i].atoms[HA2][2]) / 2.0;
		
		if (res[i].atoms[Np][0] == ATOM_COORD_NA ||
				res[i].atoms[C1][0] == ATOM_COORD_NA) {
			continue;
		}
		
		// New code added by David Arndt, Nov 8, 2007.
		// If the CB atom is not present (either becuase this is a GLY
		// residue, or because the input PDB file does not contain a CB
		// atom for the current residue), set the location of CBx based on
		// the vector summation of the Np-CA and C1-CA unit vectors. The
		// CAx-CBx vector will be normalized and scaled further below.
		difference(res[i].atoms[CA], res[i].atoms[Np], Np_CA);
		normalize(Np_CA, Np_CA);
		difference(res[i].atoms[CA], res[i].atoms[C1], C1_CA);
		normalize(C1_CA, C1_CA);
		summation(Np_CA, C1_CA, CBx);
		summation(CAx, CBx, CBx);
    }
	
//	 printf("CB = %lf %lf %lf\n", CBx[0],CBx[1],CBx[2]);

    // Represent the peptide group 
    if (flag) {
      new = CreateDOT(new);
      new->id = i; //p->n;
      new->type = 'p';
      difference(CAx, last_CA, u);
      x = sqrt(dotproduct(u, u)) / 2.0;
		normalize(u, u);
		scale(u, x);
      summation(last_CA, u, new->r);
	
		if (DEBUG) {
		   printf("\t\t%lf %lf %lf\n", new->r[0],new->r[1],new->r[2]);
		}
    }

    // Respresent the side chain 
    new = CreateDOT(new);
    new->id = i; //p->n;
	 new->type = res_str_to_char(res[i].resName);  //p->atom->letter;
	 difference(CBx, CAx, u);
	 normalize(u, u);
	 scale(u, 2.4);	
    summation(CAx, u, new->r);
	 if (DEBUG) {
	    printf("\t%lf %lf %lf\n", CAx[0],CAx[1],CAx[2]);
	 }
    copyVector(CAx, last_CA);
	 flag = 1;
  }
  return(new); 
}

static int IdxResidue(char res) {
  switch(res) {
    case 'A':    return(0);
    case 'C':    return(1);
    case 'D':    return(2);
    case 'E':    return(3);
    case 'F':    return(4);
    case 'G':    return(5);
    case 'H':    return(6);
    case 'I':    return(7);
    case 'K':    return(8);
    case 'L':    return(9);
    case 'M':    return(10);
    case 'N':    return(11);
    case 'P':    return(12);
    case 'Q':    return(13);
    case 'R':    return(14);
    case 'S':    return(15);
    case 'T':    return(16);
    case 'V':    return(17);
    case 'W':    return(18);
    case 'Y':    return(19);
    case 'p':    return(20);
  }
  fprintf( stderr, "Unknown residue type \'%c\'...\n", res);
  fflush(stderr);  
  exit(1);  
  return -1;
}

static double  Pairwise( char a, char b, double r) {
  int     typeA, typeB, distance;

  if( r <= 2.0 )
  {
     return(0);
  }
  
  /* Classify the radius first */
  if( r <= 10.0 ) {
    distance = 0;
    if( r > 5.0 )  distance++;
    if( r > 6.0 )  distance++;
    if( r > 7.0 )  distance++;
    if( r > 8.0 )  distance++;
    if( r > 9.0 )  distance++;
  } else {
    return(0.0);
  }

  typeA = IdxResidue( a );
  typeB = IdxResidue( b );

  return( Contact[typeA][typeB][distance] +
          Hydrophobicities[typeA][distance] +
          Hydrophobicities[typeB][distance] );
}



static double compute(DOT *simple_protein) {
   DOT  		 *p, *q, *s;
   double    U=0.0, r,f;
   int       i;
   Vec3D     v;
   double    SumResidues=0;
	
   p = q = simple_protein;
	for( i=0; i<10; i++ ) {
      q = q->next;
	}
  	i=0;
	
  	while( p != NULL && q != NULL ) {
   	 s = q;
	    while( s != NULL ) {
   	   difference( s->r, p->r, v);
	      r = sqrt(dotproduct(v, v));
			f = Pairwise(p->type, s->type, r);
      	U += f;
		
		// debug
		//if (f != 0.0) {
		//	fprintf(stderr, "%lf:\t\t%d %c (%lf, %lf, %lf)\t\t%d %c (%lf, %lf, %lf)\t%lf\n",
		//		f, s->id, s->type, s->r[0], s->r[1], s->r[2],
		//		p->id, p->type, p->r[0], p->r[1], p->r[2], r);
		//}
		
			SumResidues += f;
   	   s = s->next;
	   }
		i++;
		if (i==2) {
//		 	printf("Sum of Contact Potential for %c Residue=%g\n", p->last->type, SumResidues);
			i=0;
			SumResidues=0;
		}			
   	p = p->next;
   	q = q->next;
	}
	if (i==1) {
//		printf("Sum of Contact Potential for %c Residue=%g\n",p->last->type,SumResidues);
	}
   return( U );
}


double  Empirical_Potential(PDBFile *pdb) {
   DOT      *simple_protein;
	double 	result;
	
	if (pdb->numRes < 10) {
		// Peptide must be at least 10 reisdues long for the threading potential function.
		return 0.0;
  	}

   simple_protein = Simplify(pdb); 
	result = compute(simple_protein);
   FreeDOT( simple_protein );
   return result;
}


double Empirical_Potential2(ResData *res, int n) {
   DOT      *simple_protein;
	double 	result;
		
	if (n < 10) {
		// Peptide must be at least 10 reisdues long for the threading potential function.
		return 0.0;
  	}

   simple_protein = Simplify2(res, n); 
	result = compute(simple_protein);
   FreeDOT( simple_protein );
   return result;
}

