/*
Functions for rotamer library lookup. These relate to the chi angles of amino acid residues.

Author: David Arndt, May 2007.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 

#include "rotamer.h"
#include "rama.h"


/* Variables */

// All rotamer library data, for all 20 amino acids except GLY and ALA.
RotamerLibData rotamer_lib[18];

/* Function prototypes */
static int getIndexFromResName(char *res_name);


/*
 * Loads the rotamer library data from the data file into memory.
 *
 * The rotamer library file in rotamer_files/bbrotamer.lib.new was generated by Raj Maiti, but
 * he does not remember how he generated it. He also does not remember a paricular phi-psi pair
 * corresponds to, e.g. for phi = -180 and psi = -180, does this mean phi is -180 +/- 5, or
 * -170 to -180, or -180 to -190? Dr. Wishart does not know the answer either, but said it
 * probably doesn't really matter. -- David Arndt, May 30, 2007.
 */
void loadRotamerLibrary() {
	FILE *fp;
	char buffer[128];
	char c;
	char res_name[4];
	double phi, psi;
	int chi_type;
	float prob;
	float avg_chi;
	int res_index;
	int phi_index, psi_index;
	
	fp = fopen("rotamer_files/bbrotamer.lib.new", "r");
	if (fp == NULL) {
		fprintf(stderr, "ERROR: Could not load rotamer library file \"rotamer_files/bbrotamer.lib.new\"\n");
		exit(1);
	}
	
	while ( fgets(buffer, 128, fp) != NULL ) {
		
		c = buffer[0];
		
		if ( !(c == '*' || c == '>' || c == '#') ) {
			sscanf(buffer, "%s %lf %lf %d %f %f", res_name, &phi, &psi, &chi_type, &prob, &avg_chi);
			
			res_index = getIndexFromResName(res_name);
			
			// We use the same relationship between phi and psi and their indices as in rama.c,
			// for consistency and convenience. Note that phi increases as the phi index
			// increases, but psi decreases as the psi index increases. Not that in rama.c
			// the phi and psi angles are divided into 10-degree segments, as they are in the
			// rotamer library, but if they were to be different then we could not use these
			// functions.
			phi_index = getPhiIndex(phi);
			psi_index = getPsiIndex(psi);
			
			rotamer_lib[res_index].prob[phi_index][psi_index][chi_type - 1] = prob;
			rotamer_lib[res_index].avg_chi[phi_index][psi_index][chi_type - 1] = avg_chi;
		}
	}
	
	fclose(fp);
}

/*
 * Returns the index into rotamer_lib for the amino acid with the given 3-letter abbreviation.
 * GLY and ALA are not applicable. Returns -1 if the residue name is not recognized.
 */
static int getIndexFromResName(char *res_name) {
	
	if      (strcmp(res_name, "ARG") == 0) { return 0; }
	else if (strcmp(res_name, "ASN") == 0) { return 1; }
	else if (strcmp(res_name, "ASP") == 0) { return 2; }
	else if (strcmp(res_name, "CYS") == 0) { return 3; }
	else if (strcmp(res_name, "GLN") == 0) { return 4; }
	else if (strcmp(res_name, "GLU") == 0) { return 5; }
	else if (strcmp(res_name, "HIS") == 0) { return 6; }
	else if (strcmp(res_name, "ILE") == 0) { return 7; }
	else if (strcmp(res_name, "LEU") == 0) { return 8; }
	else if (strcmp(res_name, "LYS") == 0) { return 9; }
	else if (strcmp(res_name, "MET") == 0) { return 10; }
	else if (strcmp(res_name, "PHE") == 0) { return 11; }
	else if (strcmp(res_name, "PRO") == 0) { return 12; }
	else if (strcmp(res_name, "SER") == 0) { return 13; }
	else if (strcmp(res_name, "THR") == 0) { return 14; }
	else if (strcmp(res_name, "TRP") == 0) { return 15; }
	else if (strcmp(res_name, "TYR") == 0) { return 16; }
	else if (strcmp(res_name, "VAL") == 0) { return 17; }
	else {
		return -1;
	}
	
}

/*
 * For the given residue name (3-letter amino acid abbreviation) with the given phi and psi
 * angles (in degrees), this function determines the "canonical" chi angle closest to the given
 * chi angle and returns the probability for this "canonical" chi angle.
 */
float getChiProbability(char *res_name, double phi, double psi, double chi) {
	int res_index;
	int phi_index, psi_index;
	int chi_index;
	double chi_dist;
	double closest_chi_dist = 1000.0;
	int closest_chi_index = 0;
	
	if (phi == ANGLE_NA || psi == ANGLE_NA || chi == ANGLE_NA) {
		return 1.0f;
			// Do not penalize the score if an angle is missing because certain atoms are missing.
	}
	
	res_index = getIndexFromResName(res_name);
	if (res_index == -1) {
		// unrecognized amino acid
		return 0.0F;
	}
	phi_index = getPhiIndex(phi);
	psi_index = getPsiIndex(psi);
	
	for (chi_index = 0; chi_index < 3; chi_index++) {
		chi_dist = angleDist((double) rotamer_lib[res_index].avg_chi[phi_index][psi_index][chi_index], chi);
		
		if (chi_dist < closest_chi_dist) {
			closest_chi_dist = chi_dist;
			closest_chi_index = chi_index;
		}
	}
	
	//debug
	//printf("%s %7.2lf %7.2lf %7.2lf, prob = %1.4f\n", res_name, phi, psi, chi, rotamer_lib[res_index].prob[phi_index][psi_index][closest_chi_index]);
	
	return rotamer_lib[res_index].prob[phi_index][psi_index][closest_chi_index];
}


/*
 * Returns the most probable chi angle for the residue with the given name and
 * with the given phi and psi angles.
 */
double getMostProbableChi(char *resName, double phi, double psi) {
	
	int res_index;
	int phi_index, psi_index;
	int chi_index;
	float currProb;
	float highestProb = -1.0;
	int best_chi_index = -1;
	
	res_index = getIndexFromResName(resName);
	if (res_index == -1) {
		// unrecognized amino acid
		return 0.0; // Return chi angle of 0.0.
	}
	phi_index = getPhiIndex(phi);
	psi_index = getPsiIndex(psi);
	
	for (chi_index = 0; chi_index < 3; chi_index++) {
		currProb = rotamer_lib[res_index].prob[phi_index][psi_index][chi_index];
		if (currProb >= highestProb) {
			highestProb = currProb;
			best_chi_index = chi_index;
		}
	}
	
	return (double) rotamer_lib[res_index].avg_chi[phi_index][psi_index][best_chi_index];
}

