/******************************************************************
 * Program:    gafolder.c 
 * Purpose:    To create a pdb structure from a chemical shift file.
 * Input:      A chemical shift file in bmrb format and the Preditor output
               for the same chemical shift file
 * Output:     The output pdb file.
 * To compile: make all
 * Author    : Rajarshi Maiti
 *
 *
 *******************************************************************/
 
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>

#include "gafolder.h"
#include "pdbfile.h"
#include "util.h" 
#include "vec3D.h"
#include "pepmake.h"
#include "minimize.h"
#include "create_constraint.h"
#include "create_constraint_util.h"
#include "simpredShift.h"
#include "readfileShift.h"
#include "preditor_parser.h"
#include "rama.h"
#include "rotamer.h"
#include "ga_potential.h"
#include "ga_potential_scores.h"
#include "memwatch.h"


double Empirical_Potential2( ResData *res, int n);
static void doGAMinimize(char *inputPDBFilename, ResData *seed, int numRes, ResConstraint *constraints,
		int use_SCWRL, char* scwrl_path, char* output_file_name, int num_ensemble_structures, char *ensemble_filename);
static void set_seq_and_angles(ResData *res, int numRes, ANGLE_INFO *angles, int num_angle_res);
static void set_constraints(ResData *res, int numRes, char *sequence, int havePreditorFile,
		ANGLE_INFO *angles_preditor, int num_resd_preditor, ResConstraint **constraints_param);
static void set_angle_constraint_bounds(double central_angle, int angle_type, double preditor_angle, double preditor_error,
		int sec_str_index, double *angle_min, double *angle_max);

/* Constants */
#define DEBUG 0 
#define RAD(x)  x*PI/180
#define CH2INT(x) ( x - '0' )





//*******************************************************************************
static void usage(){
	//printf("Usage: $ ./ga_folder <chem_shift_file in bmrb format> <preditor_output_file> \n");
	//printf("Eg (1)./gafolder bmrb_4663.txt preditor_output_bmrb_4663.txt\n");
	//printf("Eg (2)./gafolder files/part_4663/20_resds.txt files/part_4663/20_resds_pred_output.txt\n");

	printf("Usage: $ ./gafolder [-it <number of iterations>] [-r random_seed]\n");
	printf("                    [-sc scwrl_path] [-p]\n");
	printf("                    [-s proteus_dir | -sf <secondary structure file>]\n");
	printf("                    [-cs <chem shift file>]\n");
	printf("                    [-pred <preditor_output_file>]\n");
	printf("                    [-pdb <input PDB file>] [-o <output PDB file>]\n");
	printf("                    [-e num <ensemble filename>] [-x directory] [-xa directory]\n");
	printf("       GAFolder can start with either (1) a PDB file as the starting structure,\n");
	printf("       (2) a Preditor output file of predicted angles, or (3) both a PDB file\n");
	printf("       and a Preditor output file, from which a starting structure is generated.\n");
	printf("       [-sc scwrl_path]\n");
	printf("          Adjust the side-chains of the initial structure with SCWRL, and keep\n");
	printf("          the changes if they improve the structure. The path to the SCWRL\n");
	printf("          executable must be provided (SCWRL3.0 is expected).\n");
	printf("       [-p]\n");
	printf("          Use Prosa threading score instead of the normal threading score.\n");
	printf("          (Requires Prosa2003 to be installed with prosa2003 in the user's\n");
	printf("          path.)\n");
	printf("       [-s proteus_dir]\n");
	printf("          Call Proteus to predict secondary structure, and compute secondary\n");
	printf("          structure score in relation to this predicted secondary structure.\n");
	printf("          (Requires Proteus2 to be installed, where proteus_dir is the full path\n");
	printf("          to the directory containing run_proteus. Proteus2 requires BLAST and\n");
	printf("          ClustalW to be installed, with blastall and clustalw in the user's\n");
	printf("          path.)\n");
	printf("       [-sf <secondary structure file>]\n");
	printf("          Compute secondary structure score in relation to the secondary\n");
	printf("          structure given in <secondary structure file>. Overrides -s option.\n");
	printf("          The more divergence between this secondary structure and that computed\n");
	printf("          based on the 3-D structure, the higher the secondary structure score\n");
	printf("          will be. This file can be in one of two formats: (a) a secondary\n");
	printf("          structure assignment for all the residues in the PDB file. e.g.\n");
	printf("          HHHHCCCBBBCCCHHHBBBCCC and so on, where H = Helix, B or E = Beta\n");
	printf("          strand, C = coil, - = unknown. (b) If the content of the secondary\n");
	printf("          structure file begins with a '>' or '#', it will be parsed as a\n");
	printf("          Proteus output file.\n");
	printf("       [-cs]\n");
	printf("          Compute chemical shift scores in relation to the given chemical shifts\n");
	printf("          file. The chemical shift file can be in BMRB STAR 2.1 or SHIFTY\n");
	printf("          format.\n");
	printf("       [-x directory]\n");
	printf("          Export a PDB file of the best structure after each iteration, written\n");
	printf("          to the given directory.\n");
	printf("       [-xa directory]\n");
	printf("          Export PDB files of all new structures generated in each iteration,\n");
	printf("          written to the given directory.\n");
	printf("       [-e num <ensemble filename>]\n");
	printf("          Output an ensemble of the 'num' (integer) best structures in the\n");
	printf("          population to the file with name <ensemble filename>. The structures\n");
	printf("          will not be aligned.\n");
	printf("Eg (1)./gafolder -cs bmrb_4663.txt -pred preditor_output_bmrb_4663.txt\n");
	printf("Eg (2)./gafolder -cs files/part_4663/20_resds.txt -pred files/part_4663/20_resds_pred_output.txt\n");
	printf("Eg (3)./gafolder -cs files/test_ubiq_global/bmr5387.str -pdb files/test_ubiq_global/1xqq_a.pdb\n");
	printf("Eg (4)./gafolder -it 20 -pdb myinput.pdb -o myoutput.pdb\n");
	
	exit(-1);
}//******************************************************************************
 

/**************************************************************************
 * Assigns the given chemical shift correlation scores to the given Genome.
 **************************************************************************/
void copy_corr_coeffs(Genome *g, float *corr_coeffs){
  
  int i;
  for(i=0; i<6; i++){
    g->corr_coeffs[i] = corr_coeffs[i];
  }
}//END *********************************************************************


/***************************************************************************
 * We are always evaluating a complete genome, hence flag is always complete.
 * Hence the checks for present and finished are redundant.
 ****************************************************************************/
static void evaluate(Genome *g, int thread_id) {
	
	FILE *tmp_PDB_fp;
	char tmp_PDB_filename[1024];
	
	// If we do not need to evaluate the score for this Genome, return without
	// evaluating the score.
	if (!g->score_evaluation_needed) {
		//fprintf(stderr, "Skipped an evaluation.\n");
		return;
	}
	
	
	// Initialize scores for the current Genome.
	
	g->score = 0.0;
	g->constraint_score = 0.0;
	g->rama_score = 0.0;	
	g->bump_score = 0.0;
	g->hcount = 0.0;

	// Rebuild the peptide. This is needed because the peptide may have been mutated
	// but not yet rebuilt.
	
	buildPeptide(g->res, g->numRes);
	
	// Output a PDB file for the current Genome's structure to disk if needed.
	
	if (use_Prosa) {
		// Since we we will make a system call to Prosa within get_GA_potential_scores,
		// we need to output a PDB file to disk that Prosa will read as input.
		snprintf(tmp_PDB_filename, 1024, "_tmp_pdbfile.%d", thread_id);
		tmp_PDB_fp = fopen(tmp_PDB_filename, "w");
		printPDBAtoms(tmp_PDB_fp, g->res, g->numRes);
		fclose(tmp_PDB_fp);
	}
    
	
	/* CheckConstraints only checks the distance constraints.
	In the previous versions of the minimizer, we had different stages, but in the current version, 
	we only use a single stage, hence we only check for g->stage == 0. 
	=====================================================================*/
	if (g->stage == 0) {
		
		// ************************************************************************ //
		// Get the GA Potential scores and (possibly) the chemical shift scores     //
		// for the current Genome.                                                  //
		// ************************************************************************ //
		
		// Compute only the GA Potential score (not the chemical shift scores):
		
		get_GA_potential_scores(g->res, g->numRes, POTENTIAL_MODE_MINIMIZE,
			WHICH_SCORES_GA_POTENTIAL, tmp_PDB_filename, thread_id,
			target_sec_structure, FALSE, FALSE,
			FALSE, NULL, FALSE,
			haveCSFile, NULL,
			&g->corr_coeffs[0],
			&g->thread_potential_score, &g->bump_score, &g->hcount,
			&g->rama_score, &g->omega_score, &g->chi_score, &g->radGyr_score,
			&g->hbond_energy, &g->disulphide_bond_score, &g->sec_score,
			&g->sum_of_corr_coeffs, &g->sum_components, &g->score);
		
		// If the sum of the GA Potential scores is < NEAR_NATIVE_SCORE_THRESHOLD,
		// compute the chemical shift scores as well.
		
		if (!USE_NEAR_NATIVE_SCORE_THRESHOLD || g->sum_components < NEAR_NATIVE_SCORE_THRESHOLD) {
			
			get_GA_potential_scores(g->res, g->numRes, POTENTIAL_MODE_MINIMIZE,
				WHICH_SCORES_CHEM_SHIFT, tmp_PDB_filename, thread_id,
				target_sec_structure, FALSE, FALSE,
				FALSE, NULL, FALSE,
				haveCSFile, NULL,
				&g->corr_coeffs[0],
				&g->thread_potential_score, &g->bump_score, &g->hcount,
				&g->rama_score, &g->omega_score, &g->chi_score, &g->radGyr_score,
				&g->hbond_energy, &g->disulphide_bond_score, &g->sec_score,
				&g->sum_of_corr_coeffs, &g->sum_components, &g->score);
			
			g->chem_shift_scores_calculated = TRUE;
			//fprintf(stderr, "chem_shift_scores_calculated = TRUE\n"); // debug
		}
		else {
			g->chem_shift_scores_calculated = FALSE;
			//fprintf(stderr, "chem_shift_scores_calculated = FALSE!!!!!!!!!!!!\n"); // debug
		}
		
		//print_corr_coeffs(g); 
		//print_potential_components(g);
		//print_corr_coeffs_and_potential_components(g,stdout,0);
		//printf("g->score = %5.3f\n", g->score);
		//exit(-1);
		
	}
	
	g->score_evaluation_needed = FALSE;
	
}// END evaluate ****************************************************************************



//************************************************************************************
static void doGAMinimize(char *inputPDBFilename, ResData *seed, int numRes, ResConstraint *constraints,
		int use_SCWRL, char* scwrl_path, char* output_file_name, int num_ensemble_structures, char *ensemble_filename) {
	
	Genome *complete_pop[POP_SIZE];
	int i, c, bad_evolve;
	FILE *fpw_output_file_name;
	char tmp_export_filename[512];
	double initialScore = 1000.0; // This value will be overwritten below.
	
	
	allocPopulation(complete_pop, numRes);
	
	
	
	/************************************************************************/
	/*** Evaluate the score of the initial structure and print the score. ***/
	/************************************************************************/
	
	// This is used for the user's information, and it is also used as the initial value of
	// prev_score in the evolve function, to test how much of an improvement is being achieved
	// through minimization. We copy the data to the first Genome in the population so we can
	// pass it to the evaluate function.
	copyMultipleResData(seed, complete_pop[0]->res, numRes, FALSE);
	complete_pop[0]->stage = 0;
	complete_pop[0]->completed = 0;
	complete_pop[0]->constraint_score = 0.0;
	strcpy(complete_pop[0]->most_recent_mutation, "initial");
	complete_pop[0]->score_evaluation_needed = TRUE;
	evaluate(complete_pop[0], 0);
	printf(    "---------------------------------------------------------------------------------------------------------------------------------------------------\n");
	if (use_Prosa) {
		printf("        HA     HN      N     CA     CB     CO   Prosa    Bump Hcount    Rama    Omega    Chi    Rad  HBener Disulf  SecStr  SumCS    SumGA    Total\n");
	}
	else {
		printf("        HA     HN      N     CA     CB     CO  Thread    Bump Hcount    Rama    Omega    Chi    Rad  HBener Disulf  SecStr  SumCS    SumGA    Total\n");
	}
	printf(    "---------------------------------------------------------------------------------------------------------------------------------------------------\n");
	printf("Score of initial structure (run through Reduce, otherwise unmodified):\n");
	print_corr_coeffs_and_potential_components(complete_pop[0], stdout, 0);
	initialScore = complete_pop[0]->score;
	
	// Export intermediate PDB file.
	if (exportBestIntermediateStructures) {
		sprintf(tmp_export_filename, "%s/initial_seed.pdb", export_best_directory);
		printPDBAtomsFilename(tmp_export_filename, seed, numRes);
	}
	if (exportAllIntermediateStructures) {
		sprintf(tmp_export_filename, "%s/initial_seed.pdb", export_all_directory);
		printPDBAtomsFilename(tmp_export_filename, seed, numRes);
	}
	
	
	/************************************************************************/
	/***  Call SCWRL to adjust sidechains, evalutate the structure, and   ***/
	/***  print scores.                                                   ***/
	/************************************************************************/
	
	if (use_SCWRL) {
		call_SCWRL(scwrl_path, seed, numRes, 0);
		
		copyMultipleResData(seed, complete_pop[1]->res, numRes, FALSE);
		complete_pop[1]->stage = 0;
		complete_pop[1]->completed = 0;
		complete_pop[1]->constraint_score = 0.0;
		strcpy(complete_pop[1]->most_recent_mutation, "SCWRL");
		complete_pop[1]->score_evaluation_needed = TRUE;
		evaluate(complete_pop[1], 0);
		printf("Score after adjusting sidechains with SCWRL:\n");
		print_corr_coeffs_and_potential_components(complete_pop[1], stdout, 0);
		
		// Export intermediate PDB file.
		if (exportBestIntermediateStructures) {
			sprintf(tmp_export_filename, "%s/scwrl.pdb", export_best_directory);
			printPDBAtomsFilename(tmp_export_filename, seed, numRes);
		}
		if (exportAllIntermediateStructures) {
			sprintf(tmp_export_filename, "%s/scwrl.pdb", export_all_directory);
			printPDBAtomsFilename(tmp_export_filename, seed, numRes);
		}
		
		// Note:
		// complete_pop[0] has the score and structure from before SCWRL;
		// complete_pop[1] has the score and structure from after SCWRL.
		
		if (compareGenomeScores(complete_pop[0], complete_pop[1]) < 0) {
			// The structure from before SCWRL has better scores, so set the seed back
			// to the structure from before SCWRL.
			copyMultipleResData(complete_pop[0]->res, seed, numRes, FALSE);
			
			printf("Scores after SCWRL are inferior; discarding changes from SCWRL.\n");
		}
		else {
			printf("Scores after SCWRL are improved; retaining changes from SCWRL.\n");
		}
	}
	
	
	/************************************************************************/
	/***  Fix torsion angles, evalutate the structure, and print scores.  ***/
	/************************************************************************/
	
	repairAngles(seed, numRes, target_sec_structure);
	
	// Evaluate the score of the structure after problematic torsion angles have been fixed and print the score, as above.
	copyMultipleResData(seed, complete_pop[0]->res, numRes, FALSE);
	complete_pop[0]->stage = 0;
	complete_pop[0]->completed = 0;
	complete_pop[0]->constraint_score = 0.0;
	strcpy(complete_pop[0]->most_recent_mutation, "repair_angles");
	complete_pop[0]->score_evaluation_needed = TRUE;
	evaluate(complete_pop[0], 0);
	printf("Score after repairing problematic torsion angles:\n");
	print_corr_coeffs_and_potential_components(complete_pop[0], stdout, 0);
	
	// Export intermediate PDB file.
	if (exportBestIntermediateStructures) {
		sprintf(tmp_export_filename, "%s/angles_fixed.pdb", export_best_directory);
		printPDBAtomsFilename(tmp_export_filename, seed, numRes);
	}
	if (exportAllIntermediateStructures) {
		sprintf(tmp_export_filename, "%s/angles_fixed.pdb", export_all_directory);
		printPDBAtomsFilename(tmp_export_filename, seed, numRes);
	}
	
	
	/************************************************************************/
	/***  Initialize the population, and mutate each member of the        ***/
	/***  population (except one) and print the score of the best         ***/
	/***  structure.                                                      ***/
	/************************************************************************/
	
	initPopulation(complete_pop, seed, numRes, constraints);      //minimize.c
	
	// Export all new intermediate PDB files for 0th iteration, which we know from initPopulation() is all
	// of them except the first one.
	if (exportAllIntermediateStructures) {
		for (i = 1; i < POP_SIZE; i++) {
			sprintf(tmp_export_filename, "%s/%04d_%03d.pdb", export_all_directory, 0, i);
			printPDBAtomsFilename(tmp_export_filename, complete_pop[i]->res, complete_pop[i]->numRes);
		}
	}
	
	printf("[DEBUG] Printing log file.\n");	
	char *logfile = "_initial_population_scores.txt";
	FILE *fpw_logfile = fopen(logfile, "w");
	assert(fpw_logfile != NULL && "gafolder.c: fpw_logfile is NULL in doGAMinimize()");
	
	fprintf(fpw_logfile,     "---------------------------------------------------------------------------------------------------------------------------------------------------\n");
	if (use_Prosa) {
		fprintf(fpw_logfile, "        HA     HN      N     CA     CB     CO   Prosa    Bump Hcount    Rama    Omega    Chi    Rad  HBener Disulf  SecStr  SumCS    SumGA    Total\n");
	}
	else {
		fprintf(fpw_logfile, "        HA     HN      N     CA     CB     CO  Thread    Bump Hcount    Rama    Omega    Chi    Rad  HBener Disulf  SecStr  SumCS    SumGA    Total\n");
	}
	fprintf(fpw_logfile,     "---------------------------------------------------------------------------------------------------------------------------------------------------\n");
	for(c=0; c<POP_SIZE; c++){
		//print_pop(complete_pop[c], c, fpw_logfile);
		print_corr_coeffs_and_potential_components(complete_pop[c], fpw_logfile, c);
	}
	fflush(fpw_logfile);
	fclose(fpw_logfile); 


	//if(DEBUG)
	//printf("Before sorting population, com_pop[0]->score = %4.3f\n",complete_pop[0]->score);
	//print_corr_coeffs_and_potential_components(complete_pop[0], stdout, 0);
	sortPopulation(complete_pop);
	//if(DEBUG)      
	//printf("After sorting population, com_pop[0]->score = %4.3f\n",complete_pop[0]->score);
	printf("Best score after initializing all genomes (except one) with minor mutations:\n");
	print_corr_coeffs_and_potential_components(complete_pop[0], stdout, 0);
	printf("\n");
	
	// Export best intermediate PDB file for 0th iteration.
	if (exportBestIntermediateStructures) {
		sprintf(tmp_export_filename, "%s/%04d.pdb", export_best_directory, 0);
		printPDBAtomsFilename(tmp_export_filename, complete_pop[0]->res, complete_pop[0]->numRes);
	}
	

	/************************************************************************/
	/***                       Evolve the population                      ***/
	/************************************************************************/
	
	bad_evolve = evolve(complete_pop, seed, numRes, constraints, initialScore);
	
	
	if (bad_evolve == 1) {
		exit(-1); 
	}
	
	
	/************************************************************************/
	/*** Output the best structure and print the best structure's score.  ***/
	/************************************************************************/
	
    printf("After evolving the structure and sorting the population, the score for the best candidate is %f\n", complete_pop[0]->score); 
    
    complete_pop[0]->completed = 1;
	
	// debug
	//debugOutputRamaScores(complete_pop[0]->res, complete_pop[0]->numRes, "debug_rama_scores_after_minimized.txt");
	
	fpw_output_file_name = fopen(output_file_name,"w");
	assert(fpw_output_file_name != NULL);
	
	// Output the best structure.
	printPDBAtoms(fpw_output_file_name, complete_pop[0]->res, complete_pop[0]->numRes);
	
	fflush(fpw_output_file_name);
	fclose(fpw_output_file_name);
	
	
	/************************************************************************/
	/***  Output an ensemble of the best structures in the population to  ***/
	/***  the ensemble PDB file.                                          ***/
	/************************************************************************/
	
	if (ensemble_filename != NULL) { // NULL means the user does not want to output an ensemble file.
		
		fpw_output_file_name = fopen(ensemble_filename,"w");
		assert(fpw_output_file_name != NULL);
		
		if (num_ensemble_structures == 1) {
			printPDBAtoms(fpw_output_file_name, complete_pop[0]->res, complete_pop[0]->numRes);
		}
		else {
			
			fprintf(fpw_output_file_name,
				"REMARK Structures below are ranked from best GAFolder energy (1) to worst (%d).\n",
				num_ensemble_structures);
			
			for (i = 0; i < num_ensemble_structures; i++) {
				fprintf(fpw_output_file_name, "MODEL% 9d\n", i+1);
				printPDBAtoms(fpw_output_file_name, complete_pop[i]->res, complete_pop[i]->numRes);
				fprintf(fpw_output_file_name, "ENDMDL\n");
			}
		}
		
		fflush(fpw_output_file_name);
		fclose(fpw_output_file_name);
	}
	
	
	/************************************************************************/
	/***                           Free memory                            ***/
	/************************************************************************/
	
	freeJustPopulation(complete_pop);
	
	/*
	for(i=0; i<num_of_blocks; i++){
		//free(com_dcfiles[i]);
	}
	*/

	//free(com_dcfiles);
	
	
}//END doGAMinize ****************************************************************************



//*************************************************************************************
/*
static void create_frag_dcfile(FRAG_RES_LIST frag, char* filename, char* dcfile){
  int counter;
  FILE *fpr, *fpw;
  char buffer[512];
  int resd1, resd2;
  char atom1[4],atom2[4];
  
  fpw = fopen(filename, "w");  assert(fpw != NULL);

  if((fpr = fopen(dcfile,"r"))!=NULL){
        counter = 0;
    while(fgets(buffer, MAX_LINE_LEN, fpr)!= NULL){
      sscanf(buffer,"%s %d %s %d %*s",atom1,&resd1,atom2, &resd2);
      //printf("resd1 = %d %d\n", resd1,resd2);
      if(resd1 >= frag.begin_num && resd2 <= frag.end_num){
	fprintf(fpw,"%s",buffer);
	counter++;
      }
    }
    fclose(fpr);
  }else{
    printf("Cannot open %s for reading\n",dcfile);exit(-1);
  }
  
  fflush(fpw);fclose(fpw);
}// END ******************************************************************************
*/




//**************************************************************************
void create_app_dcfile(char *app_dcfile, char* complete_dcfile, int last_res){

  FILE *fpw,*fpr;
  char buffer[512];
  int first_res, sec_res, line_counter = 0;

  fpw = fopen(app_dcfile,"w");
  if(!fpw){printf("Cannot open %s for writing\n",app_dcfile);exit(-1);}

  fpr = fopen(complete_dcfile,"r");
  if(!fpr){printf("Cannot open %s for reading\n",complete_dcfile);exit(-1);}

  while(fgets(buffer,512,fpr)!=NULL){
    sscanf(buffer,"%*s %d %*s %d %*f %*f\n", &first_res, &sec_res);
    if(first_res < last_res && sec_res <= last_res){
      fprintf(fpw,"%s",buffer);
      line_counter++;
    }
  }
  fclose(fpr);
  fflush(fpw);fclose(fpw);
}//END *********************************************************************



/***********************************************************
 * Parses an output file of torsion angles from
 * phipsi and stores the torsion angles in an array of
 * ANGLE_INFO structs. The member variables 'phi_err',
 * 'psi_err', and 'chi_confidence' in each ANGLE_INFO struct
 * are not set, however.
 *
 * This function also determines the number of residues
 * represented in the phipsi file and allocates an array
 * of the ANGLE_INFO structs at the location pointed to by
 * angles_param.
 ************************************************************/
int parse_phipsi_output(char *infile, ANGLE_INFO **angles_param){
	int num_resds = 0;
	int i;
	//int res_no;
	char resName[4];
	FILE *fpr;
	//int numAtoms; // unused
	ANGLE_INFO *angles;

	char buffer[MAX_LINE_LEN];

	//double phi, psi, omega, chi;
	//char res;
	//ANGLE_INFO *curr, *new_node;
	//curr = new_node = NULL;

	fpr = fopen(infile, "r");
	if(!fpr){
		printf(" Cannot open %s for reading.\n", infile);
		exit(-1);
	}
	
	// Count number of residues in file
	while(fgets(buffer, MAX_LINE_LEN, fpr)!= NULL){
		num_resds++;
	}
	
	rewind(fpr);
	
	*angles_param = (ANGLE_INFO *) malloc(num_resds * sizeof(ANGLE_INFO));
	angles = *angles_param;
	
	i = 0;
	while(fgets(buffer, MAX_LINE_LEN, fpr)!= NULL){  
		
		if (sscanf(buffer, "%d%s%lf%lf%lf%lf%*d",
			&angles[i].res_no, resName, &angles[i].phi, &angles[i].psi, &angles[i].omega, &angles[i].chi) == 6) {
			
			angles[i].res = res_str_to_char(resName);
			
			i++;
		}
		
		//printf("Parsed: >%d< >%s< >%lf< >%lf< >%lf< >%lf< >%d<\n", res_no, resName, phi, psi, omega, chi, numAtoms);
		//printf("Float: %f %f %f %f\n", (float) phi, (float) psi, (float) omega, (float) chi);
		
		/*
		if(num_resds == 0){
			*angles = create_node(res_no, res, (float) phi, (float) psi, (float) omega, (float) chi);
			curr = *angles;
		}else{
			new_node = create_node(res_no, res, (float) phi, (float) psi, (float) omega, (float) chi);
			curr->next = new_node;
			curr = new_node;
			curr->next = new_node->next; // Is this line needed?
		}
		*/
		
	}
	fclose(fpr);
	
	if (i != num_resds) {
		fprintf(stderr, "ERROR parsing phipsi file.\n");
		exit(1);
	}
	
	return(num_resds);
}//END parse_phipsi_output ********************************


/*
 * Copies the residue name and angle information from the given array of ANGLE_INFO structs
 * to the given array of ResData structs.
 */
static void set_seq_and_angles(ResData *res, int numRes, ANGLE_INFO *angles, int num_angle_res) {
	
	int i;
	char *resName;
	
	if (numRes != num_angle_res) {
		fprintf(stderr, "ERROR: numRes != num_angle_res in set_seq_and_angles() in gafolder.c.\n");
		exit(1);
	}
	
	for (i = 0; i < numRes; i++) {
		
		res[i].id = angles[i].res_no;
		
		resName = char_to_res_str(angles[i].res);
		if (strcmp(resName, "FIL") == 0) { // Default value for non-normal residue char.
			fprintf(stderr, "ERROR: Unknown residue \'%c\'.\n", angles[i].res);
			exit(1);
		}
		strncpy(res[i].resName, resName, 4);
		
		if (angles[i].phi == PREDITOR_ANGLE_NA) {
			res[i].phi = 180.0;
		}
		else {
			res[i].phi = angles[i].phi;
		}
		
		if (angles[i].psi == PREDITOR_ANGLE_NA) {
			res[i].psi = 180.0;
		}
		else {
			res[i].psi = angles[i].psi;
		}
		
		if (angles[i].omega == PREDITOR_ANGLE_NA) {
			res[i].omega = 180.0;
		}
		else {
			res[i].omega = angles[i].omega;
		}
		
		if (angles[i].chi == PREDITOR_ANGLE_NA) {
			res[i].chi = 180.0;
		}
		else {
			res[i].chi = angles[i].chi;
		}
		
	}
	
}


/*
 * Allocates memory for 'constraints_param' (resulting in an array of numRes ResConstraint structs)
 * and sets the values of the constraints in each ResConstraint struct.
 *
 * res: Array of numRes ResData structs representing the peptide. It must have its phi, psi, and omega
 *      angles set already.
 * numRes: The number of residues in the peptide = the number of ResData structs in the array pointed
 *      to by res.
 * sequence: A string of length numRes (plus a NULL terminator) with the sequence of the peptide
 *      identical to that found in 'res'. The sequence must be in uppercase.
 * havePreditorFile: Whether a Preditor output file was given by the user.
 * angles_preditor: Pointer to an array of ANGLE_INFO structs representing the data parsed from the
 *      Preditor output file.
 * num_resd_preditor: The number of residues in the Preditor output file = the number of elements in
 *      angles_preditor.
 * constraints_param: Pointer to memory where we want to allocate an array of ResConstraint structs.
 */
static void set_constraints(ResData *res, int numRes, char *sequence, int havePreditorFile,
		ANGLE_INFO *angles_preditor, int num_resd_preditor, ResConstraint **constraints_param) {
	
	int i, j, k;
	ResConstraint *constraints;
	char *preditor_seq;
	
	FILE *ff;
	char buffer[MAX_LINE_LEN];
	char *main_segment = NULL;
	char *preditor_segment = NULL;
	char main_char;
	char preditor_char;
	int num_identical_residues;
	int num_aligned_residues;
	float seq_identity; // Proportional.
	float seq_identity_threshold = 0.3F; // Fail if sequence alignment results in a sequence identity below this proportion.
	
	*constraints_param = (ResConstraint *) malloc(numRes * sizeof(ResConstraint));
	constraints = *constraints_param;
	
	if (havePreditorFile) {
		
		// Extract sequence from the Preditor info. If it does not match exactly with the 'sequence' parameter, then
		// do an alignment. Transfer the information about angle errors from the Preditor information to the array of
		// ResConstraint structs.
		
		preditor_seq = (char *)malloc(sizeof(char) * (num_resd_preditor + 1));
		
		for (i = 0; i < num_resd_preditor; i++) {
			preditor_seq[i] = angles_preditor[i].res;
		}
		preditor_seq[num_resd_preditor] = '\0';
		
		if (strcmp(sequence, preditor_seq) == 0) {
			for (i = 0; i < numRes; i++) {
				set_angle_constraint_bounds(res[i].phi, ANGLE_TYPE_PHI, angles_preditor[i].phi, angles_preditor[i].phi_err,
					i, &constraints[i].phi_min, &constraints[i].phi_max);
				set_angle_constraint_bounds(res[i].psi, ANGLE_TYPE_PSI, angles_preditor[i].psi, angles_preditor[i].psi_err,
					i, &constraints[i].psi_min, &constraints[i].psi_max);
				set_angle_constraint_bounds(res[i].omega, ANGLE_TYPE_OMEGA, angles_preditor[i].omega, 0.0,
					i, &constraints[i].omega_min, &constraints[i].omega_max);
				
				constraints[i].phi_preditor = angles_preditor[i].phi;
				constraints[i].psi_preditor = angles_preditor[i].psi;
				constraints[i].omega_preditor = angles_preditor[i].omega;
			}
		}
		else {
			// The given sequence and Preditor sequences differ. Call clustalw to perform an alignment, and
			// map the angle information from the Preditor file onto the given sequence.
			// 
			// clustalw1.83.UNIX.tar.gz was obtained from ftp://ftp.ebi.ac.uk/pub/software/unix/clustalw/
			// See http://www.bioinformatics-toolkit.org/Help/Topics/clustalw.html
			// Citation:
			// Thompson J.D., Higgins D.G., Gibson T.J. "CLUSTAL W: improving the sensitivity of progressive
			// multiple sequence alignment through sequence weighting, position-specific gap penalties and
			// weight matrix choice." Nucleic Acids Res. 22:4673-4680(1994).
			
			main_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
			preditor_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
			
			// Write FASTA file with the two protein sequences.
			
			ff = fopen("tmp_align.fa", "w");
			fprintf(ff, ">main\n");
			fprintf(ff, "%s\n", sequence);
			fprintf(ff, ">pred\n");
			fprintf(ff, "%s\n", preditor_seq);
			fclose(ff);
			
			// Call ClustalW.
			
			if (system("./clustalw tmp_align.fa > /dev/null") != 0) {
				fprintf(stderr, "ERROR: Call to clustalw failed.\n");
				exit(-1);
			}
			
			// Parse alignment file generated by ClustalW.
			
			ff = fopen("tmp_align.aln", "r");
			
			// Skip first 3 lines
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			
			num_identical_residues = 0;
			num_aligned_residues = 0;
			j = 0;
			k = 0;
			while( fgets(buffer, MAX_LINE_LEN, ff) != NULL ) {
				sscanf(buffer, "%*s %s", main_segment);
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
				}
				sscanf(buffer, "%*s %s", preditor_segment);
				
				// Compare the two sequences, and map the angle information from the Preditor file
				// to the main sequence.
				i = 0;
				main_char = main_segment[0];
				preditor_char = preditor_segment[0];
				while (main_char != '\0') {
					
					if (main_char != '-') {
						if (preditor_char != '-') {
							set_angle_constraint_bounds(res[j].phi, ANGLE_TYPE_PHI, angles_preditor[k].phi, angles_preditor[k].phi_err,
								j, &constraints[j].phi_min, &constraints[j].phi_max);
							set_angle_constraint_bounds(res[j].psi, ANGLE_TYPE_PSI, angles_preditor[k].psi, angles_preditor[k].psi_err,
								j, &constraints[j].psi_min, &constraints[j].psi_max);
							set_angle_constraint_bounds(res[j].omega, ANGLE_TYPE_OMEGA, angles_preditor[k].omega, 0.0,
								j, &constraints[j].omega_min, &constraints[j].omega_max);
							
							constraints[j].phi_preditor = angles_preditor[k].phi;
							constraints[j].psi_preditor = angles_preditor[k].psi;
							constraints[j].omega_preditor = angles_preditor[k].omega;
							
							j++;
							k++;
							num_aligned_residues++;
							if (main_char == preditor_char) {
								num_identical_residues++;
							}
						}
						else {
							// Preditor sequence does not contain a residue that is in the main
							// sequence.
							
							set_angle_constraint_bounds(res[j].phi, ANGLE_TYPE_PHI, PREDITOR_ANGLE_NA, 0.0,
								j, &constraints[j].phi_min, &constraints[j].phi_max);
							set_angle_constraint_bounds(res[j].psi, ANGLE_TYPE_PSI, PREDITOR_ANGLE_NA, 0.0,
								j, &constraints[j].psi_min, &constraints[j].psi_max);
							set_angle_constraint_bounds(res[j].omega, ANGLE_TYPE_OMEGA, PREDITOR_ANGLE_NA, 0.0,
								j, &constraints[j].omega_min, &constraints[j].omega_max);
							
							constraints[j].phi_preditor = PREDITOR_ANGLE_NA;
							constraints[j].psi_preditor = PREDITOR_ANGLE_NA;
							constraints[j].omega_preditor = PREDITOR_ANGLE_NA;
							
							j++;
						}
					}
					else {
						// Main sequence does not contain this residue. Skip over the Preditor info
						// for this residue.
						k++;
					}
					
					i++;
					main_char = main_segment[i];
					preditor_char = preditor_segment[i];
				}
				
				// Skip 2 lines.
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
				}
				if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
					// End of file.
					break;
				}
			}
			
			fclose(ff);
			
			if (num_aligned_residues != 0) {
				seq_identity = ((float) num_identical_residues)/((float) num_aligned_residues);
			}
			else {
				seq_identity = 0.0F;
			}
			
			if (seq_identity < seq_identity_threshold) {
				fprintf(stderr, "ERROR: Aligment between sequences in PDB file and Preditor output file\n");
				fprintf(stderr, "results in sequence identity of only %4.1f%%!\n", 100.0F*seq_identity);
				fprintf(stderr, "Wrong input files suspected.\n");
				exit(1);
			}
			
			if (j != numRes) {
				fprintf(stderr, "ERROR: Incorrect number of residues after aligning sequences.\n");
				fprintf(stderr, "   %d residues counted, but %d expected based on PDB file.\n", j, numRes);
				exit(1);
			}
			
			free(main_segment);
			free(preditor_segment);
		}
		
		free(preditor_seq);
	}
	else {
		// No Preditor file was given by the user.
		
		for (i = 0; i < numRes; i++) {
			set_angle_constraint_bounds(res[i].phi, ANGLE_TYPE_PHI, PREDITOR_ANGLE_NA, 0.0,
				i, &constraints[i].phi_min, &constraints[i].phi_max);
			set_angle_constraint_bounds(res[i].psi, ANGLE_TYPE_PSI, PREDITOR_ANGLE_NA, 0.0,
				i, &constraints[i].psi_min, &constraints[i].psi_max);
			set_angle_constraint_bounds(res[i].omega, ANGLE_TYPE_OMEGA, PREDITOR_ANGLE_NA, 0.0,
				i, &constraints[i].omega_min, &constraints[i].omega_max);
			
			constraints[i].phi_preditor = PREDITOR_ANGLE_NA;
			constraints[i].psi_preditor = PREDITOR_ANGLE_NA;
			constraints[i].omega_preditor = PREDITOR_ANGLE_NA;
		}
	}
	
}


/*
 * Sets the parameters angle_min and angle_max to central_angle +/- the appropritate angle error. The angle error from
 * Preditor (preditor_error) is preferred, but if Preditor does not predict an angle (preditor_angle == PREDITOR_ANGLE_NA,
 * or the angle is not a phi or psi angle), then a default angle error is chosen based on the target secondary structure
 * of the residue. If secondary structure information is not available, the angle error for coil regions is used.
 *
 * angle_type = one of ANGLE_TYPE_PHI, ANGLE_TYPE_PSI, ANGLE_TYPE_OMEGA.
 * sec_str_index = index into target_sec_structure for the residue associated with the given angle.
 */
static void set_angle_constraint_bounds(double central_angle, int angle_type, double preditor_angle, double preditor_error,
		int sec_str_index, double *angle_min, double *angle_max) {
	
	double angle_err;
	double angle_err_default; // Default angle error based on secondary structure information.
	double midpoint; // The angle at the midpoint of the range (angle_min, angle_max).
	double dist_to_min;
	double dist_to_max;
	double dist_to_midpoint;
	double err_magnitude; // Half the magnitude of the difference between angle_min and angle_max.
	double temp;
	//char angle_name[6]; // for debug
	
	// Determine angle_err_default.
	
	if (angle_type == ANGLE_TYPE_PHI || angle_type == ANGLE_TYPE_PSI) {
		if (target_sec_structure != NULL) {
			// Set angle constraints based on secondary structure.
			
			if (target_sec_structure[sec_str_index] == 'C') {
				angle_err_default = PHI_PSI_ANGLE_WINDOW_COIL;
			}
			else if (target_sec_structure[sec_str_index] == 'B') {
				angle_err_default = PHI_PSI_ANGLE_WINDOW_BETA;
			}
			else if (target_sec_structure[sec_str_index] == 'H') {
				angle_err_default = PHI_PSI_ANGLE_WINDOW_HELIX;
			}
			else {
				angle_err_default = PHI_PSI_ANGLE_WINDOW_COIL;
			}
		}
		else {
			angle_err_default = PHI_PSI_ANGLE_WINDOW_COIL;
		}
	}
	else if (angle_type == ANGLE_TYPE_OMEGA) {
		if (target_sec_structure != NULL) {
			// Set angle constraints based on secondary structure.
			
			if (target_sec_structure[sec_str_index] == 'C') {
				angle_err_default = OMEGA_ANGLE_WINDOW_COIL;
			}
			else if (target_sec_structure[sec_str_index] == 'B') {
				angle_err_default = OMEGA_ANGLE_WINDOW_BETA;
			}
			else if (target_sec_structure[sec_str_index] == 'H') {
				angle_err_default = OMEGA_ANGLE_WINDOW_HELIX;
			}
			else {
				angle_err_default = OMEGA_ANGLE_WINDOW_COIL;
			}
		}
		else {
			angle_err_default = OMEGA_ANGLE_WINDOW_COIL;
		}
	}
	else {
		fprintf(stderr, "ERROR: Unrecognized value for parameter angle_type in set_angle_constraint_bounds() in gafolder.c.\n");
		exit(1);
	}
	
	// Set angle_err.
	
	if ( (preditor_angle != PREDITOR_ANGLE_NA) && (angle_type == ANGLE_TYPE_PHI || angle_type == ANGLE_TYPE_PSI) ) {
		angle_err = preditor_error;
	}
	else {
		// Preditor angle not available, so do not trust the Preditor angle error.
		angle_err = angle_err_default;
	}
	
	
	// Set the intial values of angle_min and angle_max.
	
	if (preditor_angle == PREDITOR_ANGLE_NA) {
		*angle_min = central_angle - angle_err;
		if (*angle_min < -180.0) {
			*angle_min += 360.0;
		}
		
		*angle_max = central_angle + angle_err;
		if (*angle_max > 180.0) {
			*angle_max -= 360.0;
		}
	}
	else {
		*angle_min = preditor_angle - angle_err;
		if (*angle_min < -180.0) {
			*angle_min += 360.0;
		}
		
		*angle_max = preditor_angle + angle_err;
		if (*angle_max > 180.0) {
			*angle_max -= 360.0;
		}
	}
	
	
	// Check whether central_angle is within [angle_min, angle_max].
	
	// Determine midpoint angle.
	
	err_magnitude = angleDist(*angle_min, *angle_max)/2.0;
	midpoint = *angle_min + err_magnitude;
	
	// Put midpoint in the range [-180.0, 180.0]
	if (midpoint < -180.0) {
		midpoint += 360.0;
	}
	else if (midpoint > 180.0) {
		midpoint -= 360.0;
	}
	
	// Determine distances to range bounds.
	dist_to_min = angleDist(central_angle, *angle_min);
	dist_to_max = angleDist(central_angle, *angle_max);
	dist_to_midpoint = angleDist(central_angle, midpoint);
	
	// for debug
	/*
	if (angle_type == ANGLE_TYPE_PHI) {
		strcpy(angle_name, "phi");
	}
	else if (angle_type == ANGLE_TYPE_PSI) {
		strcpy(angle_name, "psi");
	}
	else { // angle_type == ANGLE_TYPE_OMEGA
		strcpy(angle_name, "omega");
	}
	*/
	
	// If central_angle falls outside the error range, adjust the error range to include central_angle,
	// plus an extra buffer of size angle_err_default.
	if (dist_to_min < dist_to_max) {
		if (dist_to_midpoint > err_magnitude) {
			temp = *angle_min;
			*angle_min = central_angle - angle_err_default;
			if (*angle_min < -180.0) {
				*angle_min += 360.0;
			}
			
			// debug
			//fprintf(stderr, "%7.2lf <-- (%7.2lf, c %7.2lf, p %7.2lf,            %7.2lf)                   diff = %6.2lf  %5s %d\n",
			//	*angle_min, temp, central_angle, preditor_angle, *angle_max, angleDist(temp, *angle_min), angle_name, sec_str_index);
		}
	}
	else {
		if (dist_to_midpoint > err_magnitude) {
			temp = *angle_max;
			*angle_max = central_angle + angle_err_default;
			if (*angle_max > 180.0) {
				*angle_max -= 360.0;
			}
			
			// debug
			//fprintf(stderr, "            (%7.2lf,            p %7.2lf, c %7.2lf, %7.2lf) --> %7.2lf       diff = %6.2lf  %5s %d\n",
			//	*angle_min, preditor_angle, central_angle, temp, *angle_max, angleDist(temp, *angle_max), angle_name, sec_str_index);
		}
	}
	
}


/*
 * Calls SCWRL on the given PDB file to adjust its sidechains.
 *
 * scwrl_path: Path to the SCWRL executable (SCWRL3.0 expected).
 * input_PDB_filename: Name of input PDB file.
 * output_filename: Pointer to memory to which to write the name of the PDB file with adjusted
 *    sidechains.
 * trim_PDB_file_first: Boolean. Whether to trim the PDB file first so that SCWRL only runs on
 *    the first chain of the PDB file.
 */
void call_SCWRL_on_PDB_file(char *scwrl_path, char *input_PDB_filename, char *output_filename,
		int trim_PDB_file_first) {
	char trimmed_PDB_filename[512];
	char command[512];
	
	sprintf(trimmed_PDB_filename, "%s_trim", input_PDB_filename);
	sprintf(output_filename, "%s_scwrl", input_PDB_filename);
	
	if (trim_PDB_file_first) {
		// Trim the PDB file to save computation time, since we only want to adjust the
		// sidechains of the first chain.
		
		trim_PDB_file(input_PDB_filename, trimmed_PDB_filename, FALSE);
	}
	else {
		sprintf(trimmed_PDB_filename, "%s", input_PDB_filename);
	}
	
	// Call SCWRL
	
	sprintf(command, "%s -i %s -o %s > /dev/null", scwrl_path, trimmed_PDB_filename, output_filename);
	system(command);
}


/*************************************************************/
/*             The main entry point for GAFolder             */
/*************************************************************/
int main_gafolder(int argc, char **argv) {
	int i;
	char *minimized_pdb = NULL;
	char *logfile, *path_name = NULL;
	//char *acfile; // *dcfile, *phipsi_output;
	char *csfile = NULL;
#ifdef THREADS
	int rval;
#endif
	FILE *test_fp;
	
	char *change_bmrb_format_to_shifty_format, *shifty_format;
	char *reducedPDBFilename;
	char *bmrb_shift_file = NULL;
	int format_is_shifty = FALSE;
	char *preditor_output = NULL;
	char *inputPDBFilename = NULL;
	char *perl_script_to_change_format = "bmrb2shifty.pl";
	int use_SCWRL = FALSE;
	char *scwrl_path = NULL;
	//int program_ran_SCWRL = FALSE; // Boolean. Whether SCWRL was run.
	use_Prosa = FALSE; // Global variable in ga_potential_scores.c
	computeSecondaryScore = FALSE; // Global variable in ga_potential_scores.c
	use_Proteus = FALSE; // Global variable in ga_potential_scores.c
	proteus_dir = NULL; // Global variable in ga_potential_scores.c
	target_sec_structure = NULL; // Global variable in ga_potential_scores.c
	haveCSFile = FALSE; // Global variable in minimize.c
	int havePreditorFile = FALSE;
	havePDBFile = FALSE; // Global variable in pepmake.c
	int userHasSpecifiedOutputFilename = FALSE;
	long int random_seed = time(0);
	int num_ensemble_structures; // Number of structures to output to the ensemble file.
	char *ensemble_filename = NULL;
		// Name of the file to which to output an ensemble of structures. NULL means do not write the file.
	exportBestIntermediateStructures = FALSE; // Global variable in minimize.c
	export_best_directory = NULL; // Global variable in minimize.c
	exportAllIntermediateStructures = FALSE; // Global variable in minimize.c
	export_all_directory = NULL; // Global variable in minimize.c
	
	PDBFile pdb; // Not currently used.
	ResData *res; // An array of ResData structs, each representing a residue, that represents the peptide.
	int numRes;
		// The number of residues we decide to use for the peptide, which is the same as the number of
		// ResData structs in the array pointed to by 'res'.
	ResConstraint *constraints = NULL; // Array of numRes ResConstraint structs, one for each residue.
	char *sequence = NULL;
		// String of length numRes (plus NULL terminator) with the sequence as it is found in 'res'.
	ANGLE_INFO *angles_preditor = NULL;
	ANGLE_INFO *angles_phipsi = NULL;
	int num_resd_preditor = -1;
	int num_resd_phipsi = -1;
	int res_iter;
	
	user_spec_iterations = -1; // Global variable in minimize.c

	//time_t start_time = time(NULL);
	clock_t start_clock = clock();
	
	/******************************************************/
	/*****        Parse command-line arguments        *****/
	/******************************************************/
	
	
	if(argc < 3){usage();}
	
	
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-it") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				
				if (sscanf(argv[i], "%d", &user_spec_iterations) != 1) {
					fprintf(stderr, "Number of iterations to perform must be specified after -it option.\n");
					usage();
				}
				
				if (user_spec_iterations < 1) {
					fprintf(stderr, "Number of iterations must be at least 1.\n");
					exit(-1);
				}
			}
			else {
				fprintf(stderr, "Number of iterations to perform must be specified after -it option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-r") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				
				if (sscanf(argv[i], "%ld", &random_seed) != 1) {
					fprintf(stderr, "Random seed (non-negative integer) must be specified after -d option.\n");
					usage();
				}
			}
			else {
				fprintf(stderr, "Random seed (non-negative integer) must be specified after -d option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-sc") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				use_SCWRL = TRUE;
				scwrl_path = argv[i];
			}
			else {
				fprintf(stderr, "Path to SCWRL executable must be given after -sc option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-p") == 0) {
			use_Prosa = TRUE;
		}
		else if (strcmp(argv[i], "-s") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				proteus_dir = argv[i];
				if (computeSecondaryScore == FALSE) {
					// Only use Proteus if a secondary structure file has not been given by the user.
					use_Proteus = TRUE;
					computeSecondaryScore = TRUE;
				}
			}
			else {
				fprintf(stderr, "Path to Proteus directory must be given after -s option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-sf") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				strcpy(secondary_structure_filename, argv[i]);
				check_file_readability(secondary_structure_filename);
				use_Proteus = FALSE; // Do not call Proteus if user has supplied a secondary structure file.
				computeSecondaryScore = TRUE;
			}
			else {
				fprintf(stderr, "Secondary structure file must be given after -sf option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-cs") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				bmrb_shift_file = argv[i];
				check_file_readability(bmrb_shift_file);
				haveCSFile = TRUE;
				//printf("input chemical shifts file = %s\n", bmrb_shift_file);
			}
			else {
				fprintf(stderr, "Chemical shifts file must be specified after -cs option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-pred") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				preditor_output = argv[i];
				check_file_readability(preditor_output);
				havePreditorFile = TRUE;
				//printf("input preditor file = %s\n", preditor_output);
			}
			else {
				fprintf(stderr, "Preditor output file must be specified after -pred option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-pdb") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				inputPDBFilename = argv[i];
				check_file_readability(inputPDBFilename);
				havePDBFile = TRUE;
				//printf("input PDB file = %s\n", inputPDBFilename);
			}
			else {
				fprintf(stderr, "PDB file must be specified after -pdb option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-o") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				minimized_pdb = argv[i];
				userHasSpecifiedOutputFilename = TRUE;
			}
			else {
				fprintf(stderr, "Output filename must be specified after -o option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-e") == 0) {
			if (i + 2 < argc ) { // if arg i is not the last or second-last arg
				i++;
				
				if (sscanf(argv[i], "%d", &num_ensemble_structures) != 1) {
					// The argument is not an integer.
					fprintf(stderr, "First argument after -e option must be an integer > 0.\n");
					usage();
				}
				
				if (num_ensemble_structures < 1) {
					fprintf(stderr, "Number of ensemble structures must be at least 1.\n");
					usage();
				}
				
				if (num_ensemble_structures > POP_SIZE) {
					num_ensemble_structures = POP_SIZE;
				}
				
				i++;
				ensemble_filename = argv[i];
			}
			else {
				fprintf(stderr, "Number of ensemble structures and ensemble filename must be specified after -e option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-x") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				export_best_directory = argv[i];
				exportBestIntermediateStructures = TRUE;
				
				// Create the directory if it does not exist.
				struct stat st;
				if (stat(export_best_directory, &st) != 0) {
					char command[256];
					sprintf(command, "mkdir %s", export_best_directory);
					system(command);
				}
			}
			else {
				fprintf(stderr, "Directory must be specified after -x option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-xa") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				export_all_directory = argv[i];
				exportAllIntermediateStructures = TRUE;
				
				// Create the directory if it does not exist.
				struct stat st;
				if (stat(export_all_directory, &st) != 0) {
					char command[256];
					sprintf(command, "mkdir %s", export_all_directory);
					system(command);
				}
			}
			else {
				fprintf(stderr, "Directory must be specified after -xa option.\n");
				usage();
			}
		}
		else {
			fprintf(stderr, "Urecognized option \'%s\'. (Did you forget the -pdb flag before the PDB file?)\n", argv[i]);
			usage();
		}
	}
	
	if (!(havePreditorFile || havePDBFile)) {
		fprintf(stderr, "Must provide either a Preditor output file of predicted angles, or a PDB file.\n");
		usage();
	}
	
	
	/******************************************************/
	/*****            Multi-threading code            *****/
	/******************************************************/
	
#ifdef THREADS
	
	pthread_mutexattr_t mutat;
	pthread_t thr[NUM_OF_THREADS];
	int thread_nums[NUM_OF_THREADS];
		// thread_nums[i] should hold the value i. Pointers to these ints are passed to pthread_create().
		// See http://bytes.com/forum/thread221217.html
	//  pthread_t thr_test;
	int r;
	
	for (i = 0; i < NUM_OF_THREADS; i++) {
		thread_nums[i] = i;
	}
	
	
	r = pthread_mutexattr_init( &mutat );
	assert( r == 0 );
	
	// For info on this fix for multiple threads on Mac OS X, see
	// http://osdir.com/ml/lisp.ecl.general/2005-01/msg00070.html
	#if defined(__APPLE__)
		r = pthread_mutexattr_settype( &mutat, PTHREAD_MUTEX_ERRORCHECK );
	#else
		r = pthread_mutexattr_settype( &mutat, PTHREAD_MUTEX_ERRORCHECK_NP );
	#endif
	
	assert( r == 0 );
	r = pthread_mutex_init( &thread_mutex, &mutat );
	assert( r == 0 );
	
	pthread_mutex_lock(&thread_mutex);  // lock the thread initially
	pthread_cond_init(&work_ready, NULL);
	pthread_cond_init(&work_done, NULL);
	work_done_flag=work_ready_flag=0;
	// here we intentionally skip thr[0] because thread 0 is the
	// main program.
	for (i=1; i<NUM_OF_THREADS; i++){
		r = pthread_create( &(thr[i]), NULL, worker_thread, (void *) &thread_nums[i] );
		if (r != 0) {
			fprintf(stderr, "ERROR: pthread_create() returned %d\n", r); exit(1);
		}
	}
	
#endif
	
	
	/******************************************************/
	/*****            More initial set-up             *****/
	/******************************************************/
	
	eval_function_ptr = evaluate;
	
	// Print command run by user, so that if GAFolder is run on command-line and output is redirected to a file,
	// there will be a convenient record of the command that was run.
	for (i = 0; i < argc; i++) {
		if (i == argc - 1) {
			printf("%s\n", argv[i]);
		}
		else {
			printf("%s ", argv[i]);
		}
	}
	
	printf("Random seed is %ld\n", random_seed);
	srand48(random_seed); // initialize the drand48() function
	
	// Decide upon a name for the output PDB file if one has not been specified by the user.
	if (!userHasSpecifiedOutputFilename) {
		char *basename; // Name to use as part of output filename. The output filename will be basename plus ".minimized.pdb".
		if (havePDBFile) {
			basename = inputPDBFilename;
		}
		else if (haveCSFile) {
			basename = bmrb_shift_file;
		}
		else {
			basename = preditor_output;
		}
		
		minimized_pdb = (char*)malloc( (strlen(basename)+strlen(".minimized.pdb")+1)*sizeof(char) );  assert(minimized_pdb != NULL);
		sprintf(minimized_pdb,"%s.minimized.pdb",basename);
	}
	
	
	// Load Ramachandran tables and the rotamer library into memory. This should be done before calling repairAngles() or
	// get_GA_potential_scores() or get_rama_value().
	read_rama_tables();
	loadRotamerLibrary();
	
	// For PepMake (predict_N_CA_C1_angles).
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", general_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", gly_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/pro_N_CA_C_angle.txt", pro_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/xpr_N_CA_C_angle.txt", xpr_N_CA_C1_angle_lookup);
	
	
	/****************************************************************************************************/
	/*****          Possibly run external programs on the given PDB file (if one is given)          *****/
	/****************************************************************************************************/
	
	// If a PDB file was given, call reduce to create a reduced version of the PDB file.
	if (havePDBFile) {
		
		// Call Reduce on the PDB file.
		
		reducedPDBFilename = (char*)malloc((strlen(inputPDBFilename)+strlen("H")+1)*sizeof(char));
		//assert(reducedPDBFilename != NULL);
		//sprintf(reducedPDBFilename, "%sH", inputPDBFilename);
		
		call_reduce(inputPDBFilename, reducedPDBFilename, TRUE);
		
		inputPDBFilename = reducedPDBFilename;
		
		//printf("inputPDBFilename: %s\n", inputPDBFilename);
	}
	
	
	/****************************************************************************************************/
	/*****  Create an initial protein seed structure, deciding how to use the information from the  *****/
	/*****  given PDB file (if available) and/or the given Preditor output file (if available).     *****/
	/****************************************************************************************************/
		
	if (havePreditorFile) {
		// Read Preditor file into memory.
		
		num_resd_preditor = parse_preditor_output(preditor_output, &angles_preditor);
	}
	
	if (havePDBFile) {
		// Write angle constraints file based on the torsion angles implicit in the PDB file.
		
		char *phipsiAnglesFilename = createAnglesFile(inputPDBFilename);
		num_resd_phipsi = parse_phipsi_output(phipsiAnglesFilename, &angles_phipsi);
		numRes = num_resd_phipsi;
	}
	else {
		numRes = num_resd_preditor;
	}
	
	// Allocate memory for seed structure.
	res = (ResData *) malloc(sizeof(ResData) * numRes);
	
	if (havePDBFile) {
		set_seq_and_angles(res, numRes, angles_phipsi, num_resd_phipsi);
	}
	else {
		set_seq_and_angles(res, numRes, angles_preditor, num_resd_preditor);
	}
	
	if (!havePDBFile) {
		predict_N_CA_C1_angles(res, numRes);
	}
	
	// If havePDBFile is TRUE, then loadResidueTemplates will load residues from the PDB file
	// with name inputPDBFilename. Above we called reduce on the user-input PDB file and
	// changed inputPDBFilename to refer to the reduced PDB file. If havePDBFile is FALSE,
	// loadResidueTemplates will load residues from the template database. The template
	// database should be pre-built from reduced PDB files (PDB files that have been run through
	// reduce). So in both cases, we end up with a seed that has reduce hydrogens, and keep
	// time-consuming calls to reduce to a minimum.
	loadResidueTemplates(inputPDBFilename, pdb, havePDBFile, TRUE, FALSE, NULL, res, numRes);
	
	repair_O_geometry(res, numRes);
	
	buildPeptide(res, numRes);
	
	// Extract the sequence from the array of ResData structs.
	sequence = (char *)malloc(sizeof(char) * (numRes + 1));
	get_seq_from_ResData(res, numRes, sequence);
	
	// Determine target secondary structure for the secondary structure score.
	if (computeSecondaryScore) {
		target_sec_structure = (char *)malloc(sizeof(char) * (numRes + 1));
		
		if (use_Proteus) {
			call_proteus(sequence, inputPDBFilename, proteus_dir, secondary_structure_filename);
		}
		parseSecondaryStructureFile(secondary_structure_filename, numRes, sequence, target_sec_structure);
	}
	
	set_constraints(res, numRes, sequence, havePreditorFile, angles_preditor, num_resd_preditor, &constraints);
	
	// debug - print constraints
	/*
	for (i = 0; i < numRes; i++) {
		char sec_str_char = '?';
		if (computeSecondaryScore) {
			sec_str_char = target_sec_structure[i];
		}
		fprintf(stderr, "%c %c phi %7.2lf +/- %7.2lf (%7.2lf, %7.2lf), psi %7.2lf +/- %7.2lf (%7.2lf, %7.2lf), omega %7.2lf +/- %7.2lf (%7.2lf, %7.2lf)\n",
			sequence[i], sec_str_char,
			res[i].phi, angleDist(constraints[i].phi_min, constraints[i].phi_max)/2.0, constraints[i].phi_min, constraints[i].phi_max,
			res[i].psi, angleDist(constraints[i].psi_min, constraints[i].psi_max)/2.0, constraints[i].psi_min, constraints[i].psi_max,
			res[i].omega, angleDist(constraints[i].omega_min, constraints[i].omega_max)/2.0, constraints[i].omega_min, constraints[i].omega_max);
	}
	*/
	
	
	/**********************************************************************************/
	/*****  Load chemical shifts and perform related set-up for chemical shifts.  *****/
	/**********************************************************************************/
	
	if (haveCSFile) {
		int pcs_iter;
		
		// Check whether the given chemical shifts file is in BMRB or SHIFTY format.
		
		format_is_shifty = FALSE;
		char buf[5];
		test_fp = fopen(bmrb_shift_file, "r");
		if (fgets(buf, 5, test_fp) != NULL) {
			// Perhaps the "#" criteria needs improvement in case it misidentifies some BMRB files as SHIFTY (?).
			if (strncmp(buf, " NUM", 4) == 0 || strncmp(buf, "NUM", 3) == 0 ||
				strncmp(buf, " #", 2) == 0 || strncmp(buf, "#", 1) == 0 ||
				strncmp(buf, ">", 1) == 0) {
				format_is_shifty = TRUE;
			}
		}
		fclose(test_fp);
		
		// Determine the name of the SHIFTY file.
		
		path_name = get_pathname(bmrb_shift_file);
		
		shifty_format = (char*)malloc((strlen(bmrb_shift_file)+strlen(".shifty")+1)*sizeof(char));
		assert(shifty_format != NULL);
		sprintf(shifty_format,"%s.shifty",bmrb_shift_file);
		//printf("shifty_format = %s\n",shifty_format);
		
		if (!format_is_shifty) {
			// Convert BMRB-formatted file to SHIFTY format.
			
			// Call the conversion script.
			
			change_bmrb_format_to_shifty_format = (char*)malloc((strlen("perl ")+strlen(perl_script_to_change_format) + 1 +
													strlen(bmrb_shift_file) + 1 + strlen(shifty_format) + 1)*sizeof(char));
			assert(change_bmrb_format_to_shifty_format != NULL); 
			sprintf(change_bmrb_format_to_shifty_format,"perl %s %s %s",perl_script_to_change_format, bmrb_shift_file, shifty_format);
			//printf("change command=%s and strlen(command)=%d\n",change_bmrb_format_to_shifty_format, strlen(change_bmrb_format_to_shifty_format));  
			if (system(change_bmrb_format_to_shifty_format) != 0) {
				fprintf(stderr, "ERROR: Call to %s failed.\n", perl_script_to_change_format);
				exit(-1);
			}
			free(change_bmrb_format_to_shifty_format);
			
			csfile = shifty_format;
		}
		else {
			// Use the SHIFTY file given by the user.
			
			csfile = bmrb_shift_file; // It wasn't really in BMRB format.
		}
		
		
		read_query_seq(csfile, &experimental_chem_shifts, &num_residues_expt_chem_shifts);
			// Read the experimental chemical shifts into the experimental_chem_shifts MATRIX struct array.
		
		// Also read in the random coil shifts, and subtract them from the experimental chemical shifts.
		// We want to subtract the random coil shifts before computing the correlation scores in get_chem_shift_score(),
		// and we do it here so we only have to do it once.
		
		loadRandomCoilShifts();
		subtract_random_coil_shifts(experimental_chem_shifts, randomCoilShifts, num_residues_expt_chem_shifts);
		
		// Allocate memory for the predicted chemical shifts. The purpose is to allocate this only once
		// in the course of the program and re-use the memory each time new chemical shifts are predicted.
		// In multi-threaded code, there should be more than one copy of such a variable.
		
		for (pcs_iter=0; pcs_iter<NUM_OF_THREADS; pcs_iter++){	
			Predicted_Chem_Shifts[pcs_iter] = (MATRIX *) malloc(numRes*sizeof(MATRIX));
			assert( (Predicted_Chem_Shifts[pcs_iter]) != NULL );
		}

		// Do initializations for shiftx. This should to be called once and only once, before calling calcChemShifts().
		do_shiftx_initializations();
	}
	else if (havePreditorFile) {
		path_name = get_pathname(preditor_output);
	}
	else { // havePDBFile == TRUE
		path_name = get_pathname(inputPDBFilename);
	}	
	
	// Get alignment data for use in chemical shift correlation scores.
	if (haveCSFile) {
		get_alignment_data(res, numRes, experimental_chem_shifts, num_residues_expt_chem_shifts);
	}
	
	
	/**********************************************************************************/
	/*****  Allocate Residues[] array for use with chemcial shift scores and      *****/
	/*****  secondary structure score.                                            *****/
	/**********************************************************************************/
	
	if (haveCSFile || computeSecondaryScore) {
		// Allocate memory for an array of RESIDUE structs that represent the peptide.
		// This will be passed to calcChemShifts(), which will use this memory in the shiftx
		// routines. The purpose is so that calcChemShifts() will copy the atom coordinates
		// from the ResData struct array to the RESIDUE struct array, without re-allocating
		// new memory for the RESIDUE struct array each time it is called. More than one copy
		// of such a variable should be used if the code is to be multi-threaded.
		for (res_iter=0; res_iter<NUM_OF_THREADS; res_iter++){	
			Residues[res_iter] = create_RESIDUE_array_from_ResData_array(res, numRes);
		}
	}
	
	
	/******************************************************/
	/*****          Run energy minimization           *****/
	/******************************************************/
	
	logfile = (char*)malloc( (strlen(path_name)+strlen("logfile")+1)*sizeof(char) ); assert(logfile != NULL);
	sprintf(logfile, "%slogfile",path_name);
	//printf("logfile = %s\n",logfile);
	
	doGAMinimize(inputPDBFilename, res, numRes, constraints, use_SCWRL, scwrl_path, minimized_pdb,
		num_ensemble_structures, ensemble_filename);
	
	
	/******************************************************/
	/*****         Clean up and exit program          *****/
	/******************************************************/
	
	printf("\nTime taken by the Processor = %g sec and the minimized file is %s.\n\n",
		(double)(clock() - start_clock) / (double)CLOCKS_PER_SEC, minimized_pdb);    
	
	
	
	// Deallocate memory...

	if (havePDBFile) {
		free(inputPDBFilename);
		free(angles_phipsi);
	}
	
	free(res);
	free(constraints);

	if (haveCSFile) {
		int pcs_iter;	
		
		free(experimental_chem_shifts);
		for (pcs_iter=0; pcs_iter< NUM_OF_THREADS; pcs_iter++){
			free(Predicted_Chem_Shifts[pcs_iter]);
		}	
		free(randomCoilShifts);
	}
	free(path_name);
	if (!userHasSpecifiedOutputFilename) {
		free(minimized_pdb);
	}
	free(logfile);

	if (haveCSFile && !format_is_shifty) {
		free(csfile);
	}

	free(sequence);
	if (computeSecondaryScore) {
		free(target_sec_structure);
	}

	if (haveCSFile) {
		free(pred_aln);
		free(expt_aln);
	}

#ifdef THREADS 

	pthread_mutex_destroy(&thread_mutex);
	pthread_cond_destroy(&work_ready);
	pthread_cond_destroy(&work_done);
	fprintf(stderr,"Killing thread");
	for (i=1; i<NUM_OF_THREADS; i++){
		fprintf(stderr," %d",i);
		rval=pthread_cancel(thr[i]);
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"thread_cancel() returned %d\n", rval);

#endif
	
	return (0);
}// END main_gafolder *********************************************************************************



   

