/**************************************************************************
 
  minimize.c
     Generate a PDB file from a file of internal angles and coordinates.
     Uses a minimization algorithm (GA) to evolve a more realistic
     structure.

  Original Date:
     May, 2003

  Last Updated:
     June, 2003

  Author:
     Aaron Davidson @ BTI <adavidson@biotools.com>
     Shan Sundararaj (shan@redpoll.pharmacy.ualberta.ca)

  -------------------------------------------------------------------------
  
  Instructions for adding new Evaluation Functions:
  
  An evaluation function is defined as such:
  		
     doEvaluation(Genome *g) {
        g->score = fittness_of_g; 
  	  } 
  
  To set your function as the one used by this software, just set the global
  function pointer to the name of your new function like so:
  	
	  	eval_function_ptr = doEvaluation;

  In this program, lower scores are considered fitter (i.e., a lower energy),

  A Genome is the data structure that holds the current thing being evaluated
  It is a Genome in the Genetic Algorithm sense.

  Here is an example evaluation that builds the peptide from the Genome,
  and then sets the score to a threading potential of the structure:

    void myDoEvaluation(Genome *g) {
      buildPeptide(g->res, g->numRes);
      g->score = threadingPotential(g->res, g->numRes);
    }
  
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <pthread.h>

#include "util.h"
#include "vec3D.h"
#include "pepmake.h"
#include "minimize.h"
#include "ga_potential.h"
#include "ga_potential_scores.h"
#include "rama.h"
#include "gafolder.h"
#include "memwatch.h"

/* Function prototypes. */
//static void computeDistances(PDBFile *pdb, int n);
static int getAtomIndex(char *a);
static void evaluate(Genome *g, int thread_id);
//static void doEvaluation(Genome *g);
static inline void swap(Genome **pop, int a, int b);
static int changeAngle(double *angle, double change, double angle_min, double angle_max);
static void mutateJump(Genome *g, ResData *seed, int n, ResConstraint *constraints);
static void mutateCreep(Genome *g, ResData *seed, int n, ResConstraint *constraints);
static void mutateSingleResidueWrithe(Genome *g, ResData *seed, int numRes, ResConstraint *constraints);
static int mutateMultipleResidueWrithe(Genome *g, ResData *seed, int numRes, ResConstraint *constraints);
static void mutateGenome(Genome *g, ResData *seed, int n, ResConstraint *constraints);
static void mateGenomes(Genome *g1, Genome *g2, int crossNum);
static void swapDoubles(double *a, double *b);
static void initGenome(Genome *g);


/* a threading potential function (potential.c) */
double Empirical_Potential2( ResData *res, int n);

/* GLOBAL VARIABLES */

//double *gDists; /* an ugly global variable I should remove */
//double *cDists; /* same as gDists, but for carbonyl C */



DistConstraint *distConstraints = NULL; 
//AngleConstraint *angleConstraints = NULL;

DistConstraint *old_distConstraints  = NULL;
//AngleConstraint *old_angleConstraints = NULL;

DistConstraint *present_distConstraints  = NULL;
//AngleConstraint *present_angleConstraints = NULL;

DistConstraint *finished_distConstraints  = NULL;
//AngleConstraint *finished_angleConstraints = NULL;

DistConstraint *complete_distConstraints  = NULL;
//AngleConstraint *complete_angleConstraints = NULL;

char *pdbOut = NULL; // output finished file to PDB
double temp = 5.25; // a mutation temperature
int stage = 0; // the current stage (used by evaluation)
void (*eval_function_ptr)(Genome *, int); // function pointer to use for evaluations

#ifdef THREADS

// GLOBAL VARIABLES FOR THREADING

Genome ** POPULATION_GLOBAL;
int ITERATION;

// END GLOBAL VARIABLES
#endif

int haveCSFile; // Whether the user has specified a chemical shifts file.
int user_spec_iterations;
	// The number of iterations for which the user has requested that we run the program. -1 means that the user
	// has not specified a value, so the program will decide based on other factors when to stop iterating.
int exportBestIntermediateStructures; // Boolean. Whether to export the best intermediate structures in each iteration.
char *export_best_directory; // Directory into which to export the best intermediate structures in each iteration.
int exportAllIntermediateStructures; // Boolean. Whether to export all new intermediate structures in each iteration.
char *export_all_directory; // Directory into which to export all new intermediate structures in each iteration.



//**********************************************************************
/** get the index of the Nth Atom of the given tag type in the pdb file **/
/*
int getNthAtom(PDBFile *pdb, int n, char *tag) {
	int x = 0;	
	int ni = 0;
	while (ni < pdb->numAtoms && x < n+1) {
		ni = getNextAtomIndex(pdb, ni, tag, GET_ATOM_ANY_RES) + 1;
		if (ni != -1) x++;
	}
	//fprintf(stderr,"%dth == %d\n", n, ni-1);
	return ni-1;
}// END **************************************************************************
*/



//**********************************************************************
/** compute the distance matrix of CA atoms in the PDB file **/
/*
static void computeDistances(PDBFile *pdb, int n) {
	int i, j;
	//double radGyr = 0.0;
	gDists = malloc(n*n*sizeof(double));
	cDists = malloc(n*n*sizeof(double));
	
	for (i=0; i<n-1; i++) {
		for (j=i+1; j<n; j++) {
			int c1 = getNthAtom(pdb, i, "CA");
			int c2 = getNthAtom(pdb, j, "CA");
			gDists[j*n + i] = distance(pdb->atoms[c1].point, pdb->atoms[c2].point);
			
			c1 = getNthAtom(pdb, i, "C ");
			c2 = getNthAtom(pdb, j, "C ");
			cDists[j*n + i] = distance(pdb->atoms[c1].point, pdb->atoms[c2].point);
			
			
		}
	}
}// END **************************************************************************
*/




//**********************************************************************
/** get the atom id for a given atom name **/
static int getAtomIndex(char *a) {
	if (strncmp(a, "CA", 2) == 0) {
		return CA;
	} else if (strncmp(a, "CB", 2) == 0) {
		return CB;
	} else if (strncmp(a, "H ", 2) == 0) {
		return HN;
	} else if (strncmp(a, "C ", 2) == 0) {
		return C1;
	} else if (strncmp(a, "O ", 2) == 0) {
		return O;
	} else if (strncmp(a, "N ", 2) == 0) {
		return N;		
	} else if (strncmp(a, "O", 1) == 0) {
		return O;
	} else if (strncmp(a, "N", 1) == 0) {
		return N;
	} else if (strncmp(a, "H", 1) == 0) {
                return HN;
        } else if (strncmp(a, "C", 1) == 0) {
                return C1;
	}
	
	fprintf(stderr, "Unknown Atom Index for %s-- \n", a);
	return 0;
}//END **********************************************************************







//***************************************************************************
/** read in a set of distance contraints that must be satisfied **/
DistConstraint *readDistanceConstraints(char *fname) {
  FILE *fp; // *fpw, *fpr;
  double min, max;
  int i1, i2;
  char n1[3], n2[3];
  DistConstraint *firstDC = NULL;
  DistConstraint *curDC = NULL;
  //char buffer[300];

  
  //fpw = fopen("junk.dc","w");
  /* printf("fname = %s\n",fname);
  fpr = fopen(fname,"r");
  if(!fpr){printf("Cannot open %s for reading\n",fname);exit(-1);}
  while(fgets(buffer,300,fpr)!= NULL){
    printf("buf=%s",buffer);
  }
  fclose(fpr);
  */

  if ((fp = fopen(fname, "r")) != NULL) {
    DistConstraint *dc = malloc(sizeof(DistConstraint));
    assert(dc != NULL);
    int rc = fscanf(fp, "%3s %d %3s %d %lf %lf", n1, &i1, n2, &i2, &min, &max);
    dc->min = min;dc->max = max;	dc->r1 = i1;dc->r2 = i2;dc->a1 = getAtomIndex(n1); dc->a2 = getAtomIndex(n2);
    //fprintf(fpw, "rc was %d, the line was: %s %d %s %d %f %f\n", rc, n1, i1, n2, i2, min, max);
    while(rc == 6){
      if(firstDC == NULL){
	curDC = firstDC = dc;
      }else{
	curDC->next = dc;
	curDC = dc;
      }
      dc = malloc(sizeof(DistConstraint));
      rc = fscanf(fp, "%3s %d %3s %d %lf %lf", n1, &i1, n2, &i2, &min, &max);
      //fprintf(fpw, "rc was %d, the line was: %s %d %s %d %f %f\n", rc, n1, i1, n2, i2, min, max);
      dc->min = min;
      dc->max = max;
      dc->r1 = i1;
      dc->r2 = i2;
      dc->a1 = getAtomIndex(n1);
      dc->a2 = getAtomIndex(n2);
    }
    //rc = fscanf(fp, "%3s %d %3s %d %lf %lf", n1, &i1, n2, &i2, &min, &max);
    free(dc);
    
    //free(dc);
  } else {
    fprintf(stderr, "problem reading dist constraints!\n");
  }
  curDC->next = NULL;
  //fflush(fpw); fclose(fpw);
  return (firstDC);
}// END readDistanceConstraints *************************************************



/* If there are distance constraints, then return a penalty 
   based on the squared distance from the constrain boundary 
   OR, if 99999 is passed as second argument, returns the 
   number of distance constraints currently being used */
//*****************************************************************************************************
double check_presentConstraints(Genome *g, int n){
  double penalty = 0.0;
  int total=0;
  int numconstraints = 0;
  DistConstraint *curDC = present_distConstraints;
  
  //a shortcut to finding the total number of constraints
  if (n == 99999) {
    while (curDC != NULL) {
      curDC = curDC->next;
      penalty += 1.0;
      numconstraints += 1;
    }
    printf("penalty = %g, numc = %d\n",penalty,numconstraints);
    return(penalty);
    //return(10);
  }
  
  //printf("are we here\n");
  while (curDC != NULL) {
    double d = distance(g->res[curDC->r1-g->res[0].id].atoms[curDC->a1], g->res[curDC->r2-g->res[0].id].atoms[curDC->a2]);
    if (d < curDC->min) {
      penalty += fabs(d - curDC->min) ;
      //fprintf(stderr, "%d-%d dist was low (%.3f)-- min is %.3f\n", curDC->r1, curDC->r2,d,curDC->min);
    } else if (d > curDC->max) {
      penalty += fabs(d- curDC->max) ;
      //fprintf(stderr, "%d-%d dist was high (%.3f)-- max is %.3f\n", curDC->r1, curDC->r2,d,curDC->max);	
    }
    
    total++;
    curDC = curDC->next;
  }
  //printf("penalty/total = %f\n",penalty/total);
  return (penalty/total);
}//*****************************************************************************************************




/*
  
*********************************************************/
double checkConstraints(Genome *g, int n, char* flag){
  double penalty = 0.0;
  //double sqrt_total= 0;
  int total=0;
  DistConstraint *curDC;

  //printf("flag = %s and strlen(flag)=%d\n",flag,strlen(flag) );

  if(strcmp(flag,"present")==0){
    curDC = present_distConstraints;
  }else if (strcmp(flag,"finished")==0){
    curDC = finished_distConstraints;
  }else if (strcmp(flag, "complete")==0){
    curDC = complete_distConstraints;
  }else{
    printf("Unknown flag %s.\n",flag);exit(-1);
  }

  //a shortcut to finding the total number of constraints
  if(n == 99999){
    while(curDC != NULL){
      curDC = curDC->next;
      penalty += 1.0;
    }
    return penalty;
  }
  

  while(curDC != NULL){
    //printf("curDC->r1=%d, curdcr2=%d gres[%d] and gres[%d],\n",curDC->r1, curDC->r2, curDC->r1 - g->res[0].id, curDC->r2 - g->res[0].id);
    double d = distance(g->res[curDC->r1 - g->res[0].id].atoms[curDC->a1], g->res[curDC->r2 - g->res[0].id].atoms[curDC->a2]);
    //printf("distance = %4.3g\n",distance);
    if (d < curDC->min) {
      penalty += fabs(d - curDC->min) ;
      //fprintf(stderr, "%d-%d dist was low (%.3f)-- min is %.3f\n", curDC->r1, curDC->r2,d,curDC->min);
    } else if (d > curDC->max) { 
      penalty += fabs(d - curDC->max) ;
      //fprintf(stderr, "%d-%d dist was high (%.3f)-- max is %.3f\n", curDC->r1, curDC->r2,d,curDC->max);	
    }
    
    total++;
    curDC = curDC->next;
  }
  //sqrt_total = sqrt(total);
  return (penalty/total);
  //return(penalty/sqrt_total);
}//*****************************************************************************************************








//*****************************************************************
double checkRMSD(Genome *g, int n) {
        double penalty = 0.0;
        DistConstraint *curDC = distConstraints;
                                                                                                           
        //a shortcut to finding the total number of constraints
        if (n == 99999) {
           while (curDC != NULL) {
                   curDC = curDC->next;
                        penalty += 1.0;
                }
                return penalty;
        }
                                                                                                           
        while (curDC != NULL) {
                double d = distance(g->res[curDC->r1-g->res[0].id].atoms[curDC->a1], g->res[curDC->r2-g->res[0].id].atoms[curDC->a2]);
                if (d < curDC->min) {
                        penalty += fabs(d - curDC->min) ;
                } else if (d > curDC->max) {
                        penalty += fabs(d- curDC->max) ;
                }
                                                                                                           
                curDC = curDC->next;
        }
        return penalty;
}// END ***************************************************



//**********************************************************
/** call the evaluation function **/
static void evaluate(Genome *g, int thread_id) {
  eval_function_ptr(g, thread_id);
}// END  **************************************************






//**********************************************************
/** Evaluate the genome by comparing the distances
    to the original PDB file **/
/*
static void doEvaluation(Genome *g) {
	int i, n = g->numRes;
	int total = 0;
	//float radGyr = 0.0;
	//float radGyr_score;
	//double constraint_score = 0.0;
	double potential_score = 0.0;
	//float real_radGyr = 0.0;
	int j = 0;
	//double di = 0.0;
	//Vec3D vec_nh, vec_oc;


   //real_radGyr = pow((double)n, 0.333) * 2.9;


	g->score = 0.0;
	g->constraint_score = 0.0;

	// build the peptide from the dihedral angles
	buildPeptide(g->res, n, TRUE);

	// for each (n choose 2) residue combination, add the CA distance difference (Aaron)
	// actually, just use the CA-CA, N-CO and N-O distances, and add score if any are < 2.0A apart

	for (i=0; i<n-1; i++) {
		for (j=i+1; j<n; j++) {
			double d = distance(g->res[i].atoms[C1], g->res[j].atoms[C1]);
			if (g->stage >= 0) { 
				g->constraint_score += fabs(cDists[j*n + i] - d);
			}

			d = distance(g->res[i].atoms[CA], g->res[j].atoms[CA]);
			if (g->stage >= 0) {
				g->constraint_score += fabs(gDists[j*n + i] - d);
			}
			total++;
		}
	}	
	g->score = g->constraint_score / (double)total;
		
		
	// if we are in the second stage, we incorporate threading potential
   if (g->stage > 0) {
		// g->score *= 10; // scale the CA dist up
		//fprintf(stderr, "Empirical_Potential is %.3f\n", Empirical_Potential2(g->res, n));
		potential_score = Empirical_Potential2(g->res, n);
		g->score += potential_score / 10.0;
	}
   if (g->completed == 1) {
			fprintf(stderr, "score due to threading energy is %.3f\n", potential_score);
			fprintf(stderr, "total score is %.3f\n", g->score);
	}
}// END ************************************
*/









//***************************************************
/** Swap two genomes in a list **/
static inline void swap(Genome **pop, int a, int b) {
  Genome *t = pop[a];
  pop[a] = pop[b];
  pop[b] = t;
}// END ***********************************************




/*
 * Sort the genomes by ascending score. Since any genome with a sum of GA Potential scores
 * that is < WHICH_SCORES_GA_POTENTIAL does not have chemical shifts calculated, we first
 * sort all structures by the sum of the GA Potential scores, and then sort the structures
 * with chemical shift scores.
 */
void sortPopulation(Genome **pop) {
	int i;
	int flag;
	int end;
	int last_with_cs_scores = -1;
	
	// Sort based on GA Potential scores only.
	flag = TRUE;
	end = POP_SIZE;
	while (flag) {
		flag = FALSE;
		for (i=1; i<end; i++) {
			if (pop[i]->sum_components < pop[i-1]->sum_components) {
				swap(pop, i, i-1);
				flag = TRUE;
			}
		}
		end--;
	}
	
	// Sort the structures with chemical shift scores based on their total scores.
	
	// Find the last structure with chemical shift scores.
	for (i = 0; i < POP_SIZE; i++) {
		if ( !(pop[i]->chem_shift_scores_calculated) ) {
			last_with_cs_scores = i-1;
			break;
		}
		else if (i == POP_SIZE - 1) {
			last_with_cs_scores = i;
		}
	}
	
	if (last_with_cs_scores == -1) {
		// No structures have chemical shift scores.
		return;
	}
	else {
		end = last_with_cs_scores + 1;
	}
	
	// Sort based on total scores.
	
	flag = TRUE;
	while (flag) {
		flag = FALSE;
		for (i=1; i<end; i++) {
			if (pop[i]->score < pop[i-1]->score) {
				swap(pop, i, i-1);
				flag = TRUE;
			}
		}
		end--;
	}
	
}


/*
 * Attempts to change the angle pointed to by 'angle' by the amount given by 'change'. The
 * resulting angle is adjusted so that it is within the range [angle_min, angle_max]. If
 * the angle ended up being changed, TRUE is returned; if not, FALSE is returned.
 */
static int changeAngle(double *angle, double change, double angle_min, double angle_max) {
	
	double temp_angle = *angle;
	double midpoint; // The angle at the midpoint of the range (angle_min, angle_max).
	double dist_to_min;
	double dist_to_max;
	double dist_to_midpoint;
	double err_magnitude; // Half the magnitude of the difference between angle_min and angle_max.
	
	temp_angle += change;
	
	// Put temp_angle in the range [-180.0, 180.0]
	if (temp_angle < -180.0) {
		temp_angle += 360.0;
	}
	else if (temp_angle > 180.0) {
		temp_angle -= 360.0;
	}
	
	// Determine midpoint angle.
	
	err_magnitude = angleDist(angle_min, angle_max)/2.0;
	midpoint = angle_min + err_magnitude;
	
	// Put midpoint in the range [-180.0, 180.0]
	if (midpoint < -180.0) {
		midpoint += 360.0;
	}
	else if (midpoint > 180.0) {
		midpoint -= 360.0;
	}
	
	// Determine distances to range bounds.
	dist_to_min = angleDist(temp_angle, angle_min);
	dist_to_max = angleDist(temp_angle, angle_max);
	dist_to_midpoint = angleDist(temp_angle, midpoint);
	
	// Adjust temp_angle if it falls outside the error range.
	if (dist_to_min < dist_to_max) {
		if (dist_to_midpoint > err_magnitude) {
			temp_angle = angle_min;
		}
	}
	else {
		if (dist_to_midpoint > err_magnitude) {
			temp_angle = angle_max;
		}
	}
	
	
	if (temp_angle == *angle) {
		// Angle did not change.
		return FALSE;
	}
	else {
		// Angle changed.
		*angle = temp_angle;
		return TRUE;
	}
}


/*
 * Randomly changes one dihedral angle to anywhere within its allowed range.
 */
static void mutateJump(Genome *g, ResData *seed, int n, ResConstraint *constraints) {
	int i = randDouble(n); // Random residue.
	double randomChoice; // What type of dihedral angle will we change?
	
	randomChoice = randDouble(3);
	
	if (randomChoice < 1.0) {
		// Mutate phi angle.
		g->res[i].phi = getRandomAngleInInterval(constraints[i].phi_min, constraints[i].phi_max);
		strcpy(g->most_recent_mutation, "jump_phi");
	} else if (randomChoice < 2.0) {
		// Mutate psi angle.
		g->res[i].psi = getRandomAngleInInterval(constraints[i].psi_min, constraints[i].psi_max);
		strcpy(g->most_recent_mutation, "jump_psi");
	}
	else { // if (randomChoice < 3.0)
		// Mutate omega angle.
		g->res[i].omega = getRandomAngleInInterval(constraints[i].omega_min, constraints[i].omega_max);
		strcpy(g->most_recent_mutation, "jump_omega");
	}
	
}


//******************************************************
/** perform a random creep mutation on the dihedral angles **/
static void mutateCreep(Genome *g, ResData *seed, int n, ResConstraint *constraints) {
	int i = randDouble(n); // Random residue.
	double randomChoice; // What type of dihedral angle will we change?
	double angle_change;
	int mutationSucceeded = 0;
	
	while (!mutationSucceeded) {
		
		randomChoice = randDouble(3);
		angle_change = randDouble(4) - 2.0;
		
		if (randomChoice < 1.0) {
			// Mutate phi angle.
			mutationSucceeded = changeAngle(&g->res[i].phi, angle_change, constraints[i].phi_min, constraints[i].phi_max);
			if (mutationSucceeded) {
				strcpy(g->most_recent_mutation, "creep_phi");
			}
		} else if (randomChoice < 2.0) {
			// Mutate psi angle.
			mutationSucceeded = changeAngle(&g->res[i].psi, angle_change, constraints[i].psi_min, constraints[i].psi_max);
			if (mutationSucceeded) {
				strcpy(g->most_recent_mutation, "creep_psi");
			}
		}
		else { // if (randomChoice < 3.0)
			// Mutate omega angle.
			mutationSucceeded = changeAngle(&g->res[i].omega, angle_change, constraints[i].omega_min, constraints[i].omega_max);
			if (mutationSucceeded) {
				strcpy(g->most_recent_mutation, "creep_omega");
			}
		}
	}
	
	
	// Since the phi or psi angle of residue i has changed, re-predict its N-CA-C1 angle.
	// Added by David Arndt, Dec 20, 2006.
	if (!havePDBFile) {
		predict_N_CA_C1_angles(&(g->res[i]), 1);
	}
	
}// END ***************************************************

 



// ********************************************************************
/* Randomly alters the phi angle of a random residue, as well as the psi
 * angle of the previous residue by -1 times the same amount. If the
 * omega angle between the two residues is 180 degrees, and if the
 * CA-C1 bond of the previous residue is parallel to the N-CA bond of
 * the randomly chosen residue, the result will be that the overall
 * peptide chain beyond the mutation point is moved but is not rotated.
 *
 *     psi changed
 *      |
 *   CA---C1
 *         \   peptide bond
 *          \
 *           N---CA
 *             |
 *            phi changed
 */
static void mutateSingleResidueWrithe(Genome *g, ResData *seed, int numRes, ResConstraint *constraints) {
	int i; // Random residue index.
	double angle_change;
	double temp_phi, temp_psi;
	int phi_changed, psi_changed; // Booleans.
	int mutationSucceeded = FALSE;

	while (!mutationSucceeded) {
		
		i = randDouble(numRes-1) + 1; // Do not allow residue index 0.
		angle_change = randDouble(4) - 2.0;
		
		temp_phi = g->res[i].phi;
		temp_psi = g->res[i-1].psi;
		
		phi_changed = changeAngle(&temp_phi, angle_change, constraints[i].phi_min, constraints[i].phi_max);
		psi_changed = changeAngle(&temp_psi, -1.0 * angle_change, constraints[i-1].psi_min, constraints[i-1].psi_max);
		
		if (phi_changed && psi_changed) {
			g->res[i].phi = temp_phi;
			g->res[i-1].psi = temp_psi;
			mutationSucceeded = TRUE;
		}
	}
	
	// Since the phi and psi angles of residues i-1 and i have changed, re-predict their N-CA-C1 angles.
	// Added by David Arndt, Dec 20, 2006.
	if (!havePDBFile) {
		predict_N_CA_C1_angles(&(g->res[i-1]), 2);
	}
	
	strcpy(g->most_recent_mutation, "single_writhe");
	
}// END *********************************************************


/*
 * Performs a multi-residue writhe mutation. Chooses a segment of the peptide, and then
 * randomly chooses one of two modes: (1) Mutate a single angle, and then use CCD to
 * close the loop, or (2) Change all angles within the segment to those specified by
 * the given Preditor file, and then use CCD to close the loop. The segment to be altered
 * is chosen with random starting and ending location if no target secondary structure is
 * available; if target secondary structure is available, one of the coil regions is chosen,
 * with slight random adjustments.
 *
 * Returns TRUE if this function succeeded in mutating the structure, FALSE otherwise.
 */
static int mutateMultipleResidueWrithe(Genome *g, ResData *seed, int numRes, ResConstraint *constraints) {
	int i;
	int mutationSucceeded;
	int N_anchor_index = -1;
	int C_anchor_index = -1;
	int temp = -1;
	int num_loops; // The number of loops of appropriate length for CCD.
	int in_coil;
	int res_count = -1; // Number of residues in a loop, plus bordering residues if present.
	int chosen_loop; // The number of the loop we wish to use. The first one is 0.
	int mode; // Either MODIFY_SEG_MODE_MUTATE or MODIFY_SEG_MODE_PREDITOR.
	int new_N_anchor_index;
	int new_C_anchor_index;
	
	if (numRes < 2*MIN_CCD_SEG_LENGTH) {
		// Peptide is too short for multiple residue writhe mutation.
		return FALSE;
	}
	
	// Choose the two residues that will be the beginning and end of the segment to be modified (the
	// N-anchor and C-anchor). They should both be in the same loop segment.
	
	if (target_sec_structure == NULL) {
		// No target secondary structure, so choose any two residues (except the first one).
		
		N_anchor_index = randDouble(numRes-1) + 1;
		temp = N_anchor_index;
		while (temp == N_anchor_index || abs(N_anchor_index - temp)+1 < MIN_CCD_SEG_LENGTH ) {
			temp = randDouble(numRes-1) + 1;
		}
		
		if (temp > N_anchor_index) {
			C_anchor_index = temp;
		}
		else {
			C_anchor_index = N_anchor_index;
			N_anchor_index = temp;
		}
	}
	else {
		// Scan the target secondary structure to find all loop regions that can accomodate a CCD segment
		// length of sufficient size. Choose a random loop segment. Then choose N and C anchor residues
		// within or bordering the segment.
		
		// Count the number of loops (coil regions) of appropriate length.
		num_loops = 0;
		in_coil = FALSE;
		for (i = 0; i < numRes; i++) {
			
			if (!in_coil && (target_sec_structure[i] == 'C' || target_sec_structure[i] == '-') ) {
				in_coil = TRUE;
				num_loops++;
				if (i <= 1) {
					res_count = i;
				}
				else {
					res_count = 2;
				}
			}
			else if (in_coil) {
				if (target_sec_structure[i] == 'B' || target_sec_structure[i] == 'H') {
					in_coil = FALSE;
					res_count++;
					
					if (res_count < MIN_CCD_SEG_LENGTH) {
						// Loop is not long enough, so it does not count.
						num_loops--;
					}
				}
				else {
					res_count++;
				}
			}
		}
		if (in_coil && res_count < MIN_CCD_SEG_LENGTH) {
			// Final loop is not long enough, so it does not count.
			num_loops--;
		}
		
		//fprintf(stderr, "Counted %d loops.\n", num_loops); // debug
		
		if (num_loops == 0) {
			// Cannot perform multiple residue writhe mutation if no suitable loop regions exist.
			return FALSE;
		}
		
		// Choose a random loop.
		chosen_loop = randDouble(num_loops);
		
		// Find the loop we chose.
		num_loops = 0;
		in_coil = FALSE;
		for (i = 0; i < numRes; i++) {
			
			if (!in_coil && (target_sec_structure[i] == 'C' || target_sec_structure[i] == '-') ) {
				in_coil = TRUE;
				num_loops++;
				if (i <= 1) {
					// We do not allow the first residue to be the N anchor.
					res_count = i;
					N_anchor_index = 1;
				}
				else {
					res_count = 2; // Count the current coil residue and the previous residue.
					N_anchor_index = i-1;
				}
			}
			else if (in_coil) {
				if (target_sec_structure[i] == 'B' || target_sec_structure[i] == 'H') {
					in_coil = FALSE;
					res_count++;
					
					if (res_count < MIN_CCD_SEG_LENGTH) {
						// Loop is not long enough, so it does not count.
						num_loops--;
					}
					else {
						// We've now scanned to the end of the current loop and know it is a
						// valid loop. Is it the right one?
						if (num_loops - 1 == chosen_loop) {
							// Yes, it's the right loop.
							C_anchor_index = i; // N_anchor_index was set above already.
							break;
						}
						
					}
				}
				else {
					res_count++;
				}
			}
		}
		if (C_anchor_index == -1) {
			// We did not find our chosen loop so far, so it must be the C-terminal loop region.
			C_anchor_index = numRes - 1; // N_anchor_index was set above already.
		}
		
		// Now N_anchor_index and C_anchor_index are set to values that would yield the
		// longest possible segment for the current loop.
		
	}
	
	// Randomly vary N_anchor_index and C_anchor_index to go up to 2 more residues into the bordering
	// secondary structure regions, or 1 residue further into the loop.
	
	new_N_anchor_index = N_anchor_index + (((int) randDouble(4)) - 2);
	if (new_N_anchor_index < 1) {
		new_N_anchor_index = 1; // N_anchor_index cannot be 0 or less
	}
	
	new_C_anchor_index = C_anchor_index + (((int) randDouble(4)) - 1);
	if (new_C_anchor_index >= numRes) {
		new_C_anchor_index = numRes - 1;
	}
	
	if (new_C_anchor_index - new_N_anchor_index + 1 >= MIN_CCD_SEG_LENGTH) {
		N_anchor_index = new_N_anchor_index;
		C_anchor_index = new_C_anchor_index;
	}
	
	
	// Randomly choose whether to mutate only a single angle in the segment, or whether to change the
	// angles to those predicted by Preditor.
	if (randDouble(2) < 1.0) {
		mode = MODIFY_SEG_MODE_MUTATE;
	}
	else {
		mode = MODIFY_SEG_MODE_PREDITOR;
	}
	
	
	//fprintf(stderr, "Attempting writhe with indices %d, %d...", N_anchor_index, C_anchor_index); // debug
	
	mutationSucceeded = modifySegmentAngles(g->res, numRes, N_anchor_index, C_anchor_index, -1,
		mode, constraints);
	
	if (mutationSucceeded) {
		if (mode == MODIFY_SEG_MODE_MUTATE) {
			strcpy(g->most_recent_mutation, "multi_writhe");
		}
		else { // mode == MODIFY_SEG_MODE_PREDITOR
			strcpy(g->most_recent_mutation, "preditor_writhe");
		}
		//fprintf(stderr, " success!!!!!!!!!!!!!!!!!!!\n\n"); // debug
	}
	else {
		//fprintf(stderr, " FAILED.\n\n"); // debug
	}
	
	return mutationSucceeded;
}




//**********************************************************
/*
 * Mutate the genome in some way.
 *
 * The algorithm is partly based on:
 * Bayley MJ, Jones G, Willett P, Williamson MP. GENFOLD: a genetic algorithm for
 * folding protein structures using NMR restraints. Protein Sci. 1998 Feb;7(2):491-9.
 */
static void mutateGenome(Genome *g, ResData *seed, int n, ResConstraint *constraints) {
	
	double randomChoice; // Which type of mutation?
	int mutationSucceeded = FALSE; // Whether the peptide was mutated or remains unchanged.
	
	while (!mutationSucceeded) {
		
		randomChoice = randDouble(4);
		
		if (randomChoice < 1.0) {
			mutateJump(g, seed, n, constraints);
			mutationSucceeded = TRUE;
		}
		else if (randomChoice < 2.0) {
			mutateCreep(g, seed, n, constraints);
			mutationSucceeded = TRUE;
		}
		else if (randomChoice < 3.0) {
			mutateSingleResidueWrithe(g, seed, n, constraints);
			mutationSucceeded = TRUE;
		}
		else { // randomChoice < 4.0
			mutationSucceeded = mutateMultipleResidueWrithe(g, seed, n, constraints);
		}
	}
	
	// mutate an extra time
	if (randDouble(4) < 1) mutateGenome(g, seed, n, constraints); 
	
	g->mutated_since_prev_iteration = TRUE;
	g->score_evaluation_needed = TRUE;
	
}// END ****************************************************





//*******************************************************************
/*
 * Mate g1 with g2 using crossNum crossovers. Both g1 and g2 are changed.
 */
static void mateGenomes(Genome *g1, Genome *g2, int crossNum) {
	
	int i, j, k;
	int temp;
	int crossover_locs[MAX_CROSSOVERS]; // List of crossover locations (residue indices).
	int in_crossover;
		// Boolean. Whether the current region should be swapped between the two structures.
	int debug = FALSE;
	
	// Determine the locations of crossNum crossover points within the peptide.
	// The crossovers can only occur at coil residues (or anywhere if no target
	// secondary structure information is available). If a crossover occurs at
	// residue i, it means that the crossover occurs between residues i-1 and i.
	// Hence, 0 is not a valid crossover index. If either residue i-1 or i is a
	// coil residue, a crossover is permitted.
	
	// First make sure that there are at least crossNum coil residues.
	if (target_sec_structure != NULL) {
		temp = 0;
		for (i = 1; i < g1->numRes; i++) {
			if (target_sec_structure[i-1] == 'C' || target_sec_structure[i] == 'C') {
				temp++;
			}
		}
		
		if (temp < crossNum) {
			crossNum = temp;
		}
	}
	
	if (crossNum <= 0) {
		return;
	}
	else if (crossNum > MAX_CROSSOVERS) {
		fprintf(stderr, "ERROR: crossNum > MAX_CROSSOVERS in mateGenomes() in minimize.c.\n"); exit(1);
	}
	
	// Randomly choose crossNum crossover locations. 
	
	j = 0;
	while (j < crossNum) {
		
		i = randDouble(g1->numRes - 1) + 1; // Do not allow i == 0.
		
		if (target_sec_structure == NULL || (target_sec_structure[i-1] == 'C' || target_sec_structure[i] == 'C')) {
			// Residue i is a valid crossover location.
			
			for (k = 0; k < j; k++) {
				if (i == crossover_locs[k]) {
					// The potential crossover location i was already chosen, and we cannot choose it twice.
					continue;
				}
			}
			
			crossover_locs[j] = i;
			j++;
		}
	}
	
	// debug
	if (debug) {
		if (target_sec_structure != NULL) { 
			fprintf(stderr, "%s\n", target_sec_structure);
		}
	}
	
	// Perform the crossovers, changing both genomes g1 and g2.
	
	in_crossover = randDouble(2); // 0 or 1
	
	for (i = 0; i < g1->numRes; i++) {
		
		// Check whether the current location is a crossover location, and switch if necessary.
		if (i != 0) {
			for (j = 0; j < crossNum; j++) {
				if (i == crossover_locs[j]) {
					// Switch between crossover and not crossover.
					in_crossover = !in_crossover;
				}
			}
		}
		
		// If we're in a crossover region, swap the angles of the two structures.
		if (in_crossover) {
			swapDoubles(&g1->res[i].phi, &g2->res[i].phi);
			swapDoubles(&g1->res[i].psi, &g2->res[i].psi);
			swapDoubles(&g1->res[i].chi, &g2->res[i].chi);
			swapDoubles(&g1->res[i].omega, &g2->res[i].omega);
			swapDoubles(&g1->res[i].N_CA_C1_angle, &g2->res[i].N_CA_C1_angle);
			
			if (debug) fprintf(stderr, "X"); // debug
		}
		else {
			if (debug) fprintf(stderr, "-"); // debug
		}
	}
	if (debug) fprintf(stderr, "\n\n"); // debug
	
	strcpy(g1->most_recent_mutation, "mate");
	strcpy(g2->most_recent_mutation, "mate");
	
	g1->mutated_since_prev_iteration = TRUE;
	g2->mutated_since_prev_iteration = TRUE;
	
	g1->score_evaluation_needed = TRUE;
	g2->score_evaluation_needed = TRUE;
	
}// END ***********************************************************


/*
 * Calls SCWRL to adjust the sidechains of the given structure.
 *
 * scwrl_path: Path to the SCWRL executable (SCWRL3.0 expected).
 */
void call_SCWRL(char *scwrl_path, ResData *res, int numRes, int thread_id) {
	
	char initial_PDB_filename[512];
	char scwrl_PDB_filename[512];
	char reduced_PDB_filename[512];
	PDBFile pdb; // Not used here.
	//int rc;
	
	sprintf(initial_PDB_filename, "tmp_for_scwrl_%d.pdb", thread_id);
	
	printPDBAtomsFilename(initial_PDB_filename, res, numRes);
	
	call_SCWRL_on_PDB_file(scwrl_path, initial_PDB_filename, scwrl_PDB_filename, FALSE);
	
	
	// The file produced by SCWRL has no hydrogen atoms, so we need to add them.
	
	call_reduce(scwrl_PDB_filename, reduced_PDB_filename, FALSE);
	
	
	// Read back in the modified PDB file
	
	loadResidueTemplates(reduced_PDB_filename, pdb, TRUE, TRUE, FALSE, NULL, res, numRes);
	buildPeptide(res, numRes);
}


/*
 * Swaps the values pointed to by the two parameters a and b.
 */
static void swapDoubles(double *a, double *b) {
	double temp;
	temp = *a;
	*a = *b;
	*b = temp;
}



/** Initialize a genome **/
static void initGenome(Genome *g) {
	g->stage = 0;
	g->completed = 0;
	g->constraint_score = 0.0;
	strcpy(g->most_recent_mutation, "");
}



//**********************************************
// Allocates memory for the population. Call this before calling
// initPopulation.
void allocPopulation(Genome **pop, int n) {
	int i;
	
	for(i=0; i<POP_SIZE; i++) {
		pop[i] = SAFEMALLOC(sizeof(Genome));
		pop[i]->numRes = n;
		pop[i]->res = SAFEMALLOC(sizeof(ResData)*n);
	}
}



//**********************************************
// Initializes the population. Call allocPopulation before calling this function.
void initPopulation(Genome **pop, ResData *seed, int n, ResConstraint *constraints) {	
	int i;

	for(i=1; i<POP_SIZE; i++) { // Mutate all genomes except the first one.
		
		copyMultipleResData(seed, pop[i]->res, n, FALSE);
		initGenome(pop[i]);
		
		mutateGenome(pop[i], seed, n, constraints); // Added by David Arndt, Dec 22, 2006.
		
		evaluate(pop[i], 0);
		
	}
}// END ******************************************


/*
 * Returns a value that can be used to evaluate which of the two given Genomes has a
 * better score. A negative score indicates that a is superior; a positive score
 * indicates that b is superior; 0.0 means they are the same. A Genome with chemical
 * scores is always better than a Genome without them, since chemical shift scores
 * are only evaluated for better-quality structures (see evaluate()).
 */
float compareGenomeScores(Genome *a, Genome *b) {
	
	float pseudo_score_a = a->score;
	float pseudo_score_b = b->score;
	
	if (!a->chem_shift_scores_calculated) {
		pseudo_score_a += 1000000.0;
	}
	
	if (!b->chem_shift_scores_calculated) {
		pseudo_score_b += 1000000.0;
	}
	
	return (pseudo_score_a - pseudo_score_b);
}



/***************************************************************************
 * evolve: Evolves a population of structures towards the best possible candidate.
 *         The best possible candidate is defined by its score.
 ***************************************************************************/
int evolve(Genome **complete_pop, ResData *complete_seed,  int complete_num_resds,
		ResConstraint *constraints, double initialScore) {
	
	int bottomcopy,x,i;
	int  iter=0, worst;
	double prev_score = initialScore;
	int prev_score_used_chem_shifts = FALSE;
	double abs_diff = 0.00;

	char tmp_export_filename[512];

#ifndef THREADS 
	fprintf(stderr, "Single threaded\n");
#endif

	while(1) {
		iter++; 
		
		/*************************************************************************/
		/***               Check whether we should stop evolving               ***/
		/*************************************************************************/
		
		// We are checking the score of every 20th member of the population. 
		// If these 2 scores are within 0.0001 of each other, we break out of the
		// iteration loop (unless the user has specified the number of iterations).
		if (iter % 20 == 0) {
			
			if ( !(!prev_score_used_chem_shifts && complete_pop[0]->chem_shift_scores_calculated) ) {
				// If the previous iteration's score was calculated without chemical shift
				// scores and the current iteration's scores are calculated with chemical
				// shift scores, the total score may have increased due to the use of chemical
				// shift scores even though the structure has improved. So we only want to
				// stop minimization based on the difference between the round's scores if
				// they are comparable.
				
				abs_diff = fabs(prev_score - complete_pop[0]->score);
				printf("iter = %d, abs_diff = %f\n",iter, abs_diff);
				if( (user_spec_iterations == -1) && (abs_diff < 0.0001) ){
					//printf("abs_diff = %f.... Breaking!!!!!\n",abs_diff); 
					break;
				}
			}
			
			prev_score = complete_pop[0]->score;
			prev_score_used_chem_shifts = complete_pop[0]->chem_shift_scores_calculated;
		}

		// If the user has specified the number of iterations we should perform,
		// break after that many iterations have been performed.
		if ( (user_spec_iterations != -1) && (iter - 1 == user_spec_iterations) ) {
			break;
		}
		
		
		/*************************************************************************/
		/***                        Do initializations                         ***/
		/*************************************************************************/
		
		// Reset the records of the most recent mutations for the best structure.
		strcpy(complete_pop[0]->most_recent_mutation, "---");

		// Reset mutated_since_prev_iteration flags.
		for (i = 0; i < POP_SIZE; i++) {
			complete_pop[i]->mutated_since_prev_iteration = FALSE;
		}
		
		
		/*************************************************************************/
		/***  Copy the best 10 structures and replace the 5 worst structures.  ***/
		/*************************************************************************/
		
		worst = POP_SIZE-1; 

		//What I want is to have structures 0-9 copied to fill spots 35 (worst-14)to
		//44 (worst - 5), and get new structures for spots 45-49 (worst-4 to worst)
		for (i=0; i<10; i++) {
			bottomcopy = worst - (14 - i);
			copyMultipleResData(complete_pop[i]->res, complete_pop[bottomcopy]->res, complete_num_resds, FALSE);
			
			complete_pop[bottomcopy]->score = complete_pop[i]->score;
			complete_pop[bottomcopy]->constraint_score = complete_pop[i]->constraint_score;
			
			mutateGenome(complete_pop[bottomcopy], complete_seed, complete_num_resds, constraints);// july 14, 05
			evaluate(complete_pop[bottomcopy], 0);
			//printf("After mutating the complete genome, c_pop->score = %f\n",complete_pop[bottomcopy]->score);
		}

		for(i = (worst-4); i <= worst; i++){
			copyMultipleResData(complete_seed, complete_pop[i]->res, complete_num_resds, FALSE);
			initGenome(complete_pop[i]);
			
			complete_pop[i]->stage = complete_pop[0]->stage;
			
			//printf("Before evaluating complete_pop[%d] in worst block\n",bottomcopy);
			mutateGenome(complete_pop[i], complete_seed, complete_num_resds, constraints); // Mutate the new structures.
			evaluate(complete_pop[i], 0);
			//printf("After creating some present genome, c_pop->score = %f\n",complete_pop[i]->constraint_score);
		}

		sortPopulation(complete_pop);
		
		
		/*************************************************************************/
		/***                    Mate and mutate structures                     ***/
		/*************************************************************************/
		
		for(i=1; i<worst; i++){
			x = randDouble(POP_SIZE - 1) + 1; // x cannot be 0
			if (compareGenomeScores(complete_pop[i], complete_pop[x]) < 0.0) {
				// i has a better score than x
				
				copyMultipleResData(complete_pop[i]->res, complete_pop[x]->res, complete_num_resds, FALSE); // copy i to x
				mutateGenome(complete_pop[x], complete_seed, complete_num_resds, constraints);
				
			}else if( (randDouble(8) < 1) && (x != 0) && (i != 0) ){     
				// mate 1/8th of the time 
				mateGenomes(complete_pop[i], complete_pop[x], 1+(int)randDouble(MAX_CROSSOVERS)); // changes genomes i and x
			}
		} 
		
		
		/*************************************************************************/
		/***                        Evaluate structures                        ***/
		/*************************************************************************/
		
#ifdef THREADS 

		//set global variables
		POPULATION_GLOBAL=complete_pop;
		ITERATION=iter;
		
		thread_it(complete_pop);

		//unset global variables
		POPULATION_GLOBAL=NULL;

#else 

		for(i=0; i<POP_SIZE; i++){
			evaluate(complete_pop[i], 0);
		}

#endif	

		sortPopulation(complete_pop);
		//printf("complete_pop[0].sum_of_corr_coeffs = %4.3f, [0].sum_components = %4.3f, [0].score = %4.3f\n",
		//   complete_pop[0]->sum_of_corr_coeffs, complete_pop[0]->sum_components, complete_pop[0]->score);
		print_corr_coeffs_and_potential_components(complete_pop[0], stdout, iter);
		
		
		/*************************************************************************/
		/***    Export intermediate structures (if user has so specified).     ***/
		/*************************************************************************/
		
		// Export intermediate PDB file, the best structure in the population for this iteration.
		if (exportBestIntermediateStructures) {
			sprintf(tmp_export_filename, "%s/%04d.pdb", export_best_directory, iter);
			printPDBAtomsFilename(tmp_export_filename, complete_pop[0]->res, complete_pop[0]->numRes);
		}

		// Export intermediate PDB files of all new structures in this iteration. We export
		// only the structures that were mutated, so we don't export any identical structures.
		if (exportAllIntermediateStructures) {
			for (i = 0; i < POP_SIZE; i++) {
				if (complete_pop[i]->mutated_since_prev_iteration) {
					sprintf(tmp_export_filename, "%s/%04d_%03d.pdb", export_all_directory, iter, i);
					printPDBAtomsFilename(tmp_export_filename, complete_pop[i]->res, complete_pop[i]->numRes);
				}
			}
		}
		
		
	}// END while(1) **************
	//printf("Before leaving evolve in minimize.c, the lowest score of complete_pop is %f\n",complete_pop[0]->score);
	
	return(0);
	
}// END evolve()
//==============================================================================

#ifdef THREADS

int wait_for_all(const int id){

	int rval;

	pthread_mutex_lock(&thread_mutex);
	threads_waiting++;
//	fprintf(stderr, "[WORKER %d] threads_waiting=%d\n", id, threads_waiting );
	if (threads_waiting==NUM_OF_THREADS){	// if all threads waiting, broadcast 
		threads_waiting=0;
		pthread_mutex_unlock(&thread_mutex);  // we have to call unlock
		rval=pthread_cond_broadcast(&work_ready);
	} else {
//		fprintf(stderr, "[WORKER %d] waiting\n", id );
		pthread_cond_wait(&work_ready, &thread_mutex);
		pthread_mutex_unlock(&thread_mutex);  // we have to call unlock because the threads
										  // regain the lock once they're awoken
	}

	return 0;

}

int thread_it(Genome **g) {
	// this function splits the evolution across two threads

	int i,id=0; // this is the main thread (id 0)
	int start,finish;

	start=0;
	finish=POP_SIZE/NUM_OF_THREADS;

	fprintf(stderr, "[MASTER] I take %d upto %d\n", start, finish );

	wait_for_all(id);	
	
/*	pthread_mutex_lock(&thread_mutex);
	threads_waiting++;
	pthread_mutex_unlock(&thread_mutex);
	rval=pthread_cond_broadcast(&work_ready);
*/	
	//fprintf(stderr, "[MASTER] rval is %d\n", rval );

	for(i=start; i<finish; i++){
	     evaluate(g[i], 0);
    }

	// we have to wait for all threads so that the master can continue.
	pthread_mutex_lock(&thread_mutex);
	work_done_flag++;
	while (work_done_flag<NUM_OF_THREADS){
//		fprintf(stderr, "[MASTER] I'm waiting\n");
		pthread_cond_wait(&work_done, &thread_mutex); // if all work done is not done, then wait
	}
	work_done_flag=0;
	pthread_mutex_lock(&thread_mutex);

	return 0;
}


void * worker_thread(void * vptr){

	int start,finish;
	float chunk;
	int id = *((int *) vptr);
	
	chunk=POP_SIZE/(float)NUM_OF_THREADS;
	start=chunk*id;
	finish=chunk*(id+1);
	
	fprintf(stderr, "[WORKER %d] I take %d upto %d\n", id, start, finish );

	while (1){

		int i;

		wait_for_all(id);

//		fprintf(stderr, "[WORKER %d] I'm now working on iteration %d\n", id, ITERATION );
		for(i=start; i<finish; i++){
			evaluate(POPULATION_GLOBAL[i], id);
		}
	
//		fprintf(stderr, "[WORKER %d] I'm done\n", id );
		pthread_mutex_lock(&thread_mutex);
//		fprintf(stderr, "[WORKER %d] I'm in the last lock\n", id );
		work_done_flag++;
		if (work_done_flag==NUM_OF_THREADS)	// when all work is done, signal the master if it's waiting
			pthread_cond_signal(&work_done);
//		fprintf(stderr, "[WORKER %d] Unlocking mutex\n", id );
		pthread_mutex_unlock(&thread_mutex);

	}

	return 0;

}

#endif



//**********************************************************************
/** free allocated memory **/
void freePopulation(Genome **pop) {
  //DistConstraint *curDC = distConstraints;
  int i;
  for (i=0; i<POP_SIZE; i++) {
    SAFEFREE(pop[i]->res);
    SAFEFREE(pop[i]);
  }
  //while (curDC != NULL) {
  //  DistConstraint *lastDC = curDC;
  //  curDC = curDC->next;
  //  SAFEFREE(lastDC);
  //}
}// END **********************************************************************






//**********************************************************************
void freeJustPopulation(Genome **pop) {
	int i;
   for (i=0; i<POP_SIZE; i++) {
     	SAFEFREE(pop[i]->res);
     	SAFEFREE(pop[i]);
   }
}// END **********************************************************************


void free_angle_constraints(AngleConstraint *head){
  AngleConstraint *curr;
  curr = head;
  while(head != NULL){
    curr = head;
    head = head->next;
    free(curr);
  }
}


void free_dist_constraints(DistConstraint *head){
  DistConstraint *curr;
  curr = head;
  while(head != NULL){
    curr = head;
    head = head->next;
    free(curr);
  }
}



/**************************************************************
 * Prints the correlation coefficients
 ****************************************************************/
void print_corr_coeffs(Genome *g){
  //printf("HA     HN      N     CA     CB     CO\n");
  printf("%f\t %f\t %f\t %f\t %f\t %f\n",g->corr_coeffs[0], g->corr_coeffs[1], g->corr_coeffs[2], g->corr_coeffs[3],
	g->corr_coeffs[4], g->corr_coeffs[5]);
}//*************************************************************





/**************************************************************
 * Prints the potential components
 **************************************************************/
void print_potential_components(Genome *g){
  printf("thread=%6.3f, bump=%6.3f, hcount=%6.3f, rama=%6.3f, radius=%6.3f, hbond=%6.3f, sum=%6.3f\n", g->thread_potential_score, 
	 g->bump_score, g->hcount, g->rama_score, g->radGyr_score, g->hbond_energy, g->sum_components);
}//************************************************************




/****************************************************************
 * Prints the correlation coefficients and the potential components 
 * in a format that is suitable for plotting in gnuplot
 ****************************************************************/
void print_corr_coeffs_and_potential_components(Genome *g, FILE *fpw, int i){
  
  //             #    HA      HN      N     CA     CB     CO Thread   Bump Hcount  Rama  Omega     Chi   Rad  HBener Disulf SecStr  SumCS  SumGA  Total
  fprintf(fpw,"%3d % 6.2f % 6.2f % 6.2f % 6.2f % 6.2f % 6.2f % 7.2f % 7.2f % 6.2f % 7.2f % 8.2f % 6.2f % 6.2f % 7.2f % 6.2f % 7.2f % 6.2f % 8.2f % 8.2f  %s\n",
	  i, g->corr_coeffs[0], g->corr_coeffs[1], g->corr_coeffs[2],
	  g->corr_coeffs[3], g->corr_coeffs[4], g->corr_coeffs[5],
	  g->thread_potential_score,g->bump_score, g->hcount, g->rama_score, g->omega_score, g->chi_score, g->radGyr_score, g->hbond_energy,
	  g->disulphide_bond_score, g->sec_score,
	  g->sum_of_corr_coeffs, g->sum_components, g->score, g->most_recent_mutation);
  

}//END **********************************************************






// prints only a single member of a population.
// ******************************************************
void print_pop(Genome *pop, int pop_num, FILE *fpw){
  //int i;
  //fprintf(fpw,"\n#====== Genome[%d] info. ====================\n",pop_num);
  //printf("Numresds in the genome = %d\n",pop->numRes);
  //printf("stage= %d, completed=%d\n",pop->stage,pop->completed);
  //printf("score=%4.3f, constr_score=%4.3f\n",pop->score,pop->constraint_score);
  //print_potential_components(pop);
  //print_corr_coeffs(pop);
  //print_corr_coeffs_and_potential_components(pop);
  
  printf("=========================================\n");

}// END print_pop ***************************************
