#ifndef __MINIMIZE_H__
#define __MINIMIZE_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pthread.h>

#include "util.h"
#include "vec3D.h"
#include "pepmake.h"



#define NUM_OF_THREADS 4   // The number of threads we will be using - Cam 
#define POP_SIZE 	     50	  // the population size to use

// Constrain phi and psi angles to +/- these values compared to the originals, with different variability
// allowed for coil, helix, and Beta strand residues.
#define PHI_PSI_ANGLE_WINDOW_COIL   15
#define PHI_PSI_ANGLE_WINDOW_BETA   8
#define PHI_PSI_ANGLE_WINDOW_HELIX  5

// Constrain omega angles to +/- these values compared to the originals, with different variability
// allowed for coil, helix, and Beta strand residues.
#define OMEGA_ANGLE_WINDOW_COIL   5
#define OMEGA_ANGLE_WINDOW_BETA   3
#define OMEGA_ANGLE_WINDOW_HELIX  3

//#define AC_WRITE_FORMAT         "%3s %d %f %f %f %f %f %f %f %f\n"

#define MAX_CROSSOVERS 4 // Maximum number of crossover points when mating genomes.

#define NUM_OVERLAP_RESDS    1
#define RESD_GROWTH_NUM        76  //This number of residues are minimized each step 
                                  // and the structure is grown. To minimize the entire
								  // peptide as one fragment, set this to the number of
								  // residues in the peptide (but this is not necessary
								  // if num_of_resds is used instead of RESD_GROWTH_NUM
								  // as is done currently).

#define ZERO                     0.00 




 
typedef struct Genome {
  int stage;     // 0 for first half of ITER, 1 for second half 
  int completed; // 0 if the genome is still evolving, 1 if the structure has been finished 
                 //and we want to output stats 
  int numRes;    // number of residues in the genome 
  ResData *res;  // an array of residues
  char most_recent_mutation[16];
	// A note about the most recent type of mutation undergone by this Genome. Used for user's info only.
  
  // 6 chemical shift correlation scores:
  
  float corr_coeffs[6];
  
  // GA Potential scores:
  
  float thread_potential_score; // Either the normal threading score or the Prosa threading score.
  float bump_score;
  float hcount;   // Number of hydrogen bonds in the structure, multiplied by a weight.
  float rama_score;
  float omega_score;
  float chi_score;
  float radGyr_score;
  float hbond_energy;
  float disulphide_bond_score;
  float sec_score;
  
  // Scores not currently used:
  
  // float disulphide_bond_score;
  // float hbond_score; // Not currently used in the code, so commented-out. David Arndt, Mar 5, 2007.
  // float rmsd_score; // Not currently used in the code, so commented-out. David Arndt, Apr 19, 2007.
  // double cs_score; // Not currently used in the code, so commented-out. David Arndt, Dec 21, 2006.
  float constraint_score; //score due only to constraint violations; not currently used.
  
  // Sums of scores:
  
  float sum_of_corr_coeffs; // The sum of the 6 chemical shift correlation scores.
  float sum_components; // Sum of the GA Potential scores.
  float score; // The last evaluated score of the genome, equal to the sum of the chemical shift and GA Potential scores.
  
  int mutated_since_prev_iteration;
	// Boolean. If TRUE, indicates that this structure was mutated since the previous minimization iteration.
	// Otherwise FALSE.
  int score_evaluation_needed;
	// Boolean. If TRUE, the current score values do not correspond to the structure in the 'res'
	// member variable of this Genome, so the score needs to be re-evaluated.
  int chem_shift_scores_calculated;
	// Boolean. Indicated whether the chemical shift scores have been calculated. Structures with a
	// sum of GA Potential score >= NEAR_NATIVE_SCORE_THRESHOLD are considered to be too far from the
	// native structure to warrant calculating the chemical shift scores (if
	// USE_NEAR_NATIVE_SCORE_THRESHOLD is TRUE).
  
} Genome;


 
//   constraints that atom a1 in residue r1 
//   must be between min & max angstroms
//   away  from atom a2 in residue r2 
typedef struct DistConstraint {
  double min, max;
  int r1, r2, a1, a2;
  void *next;
} DistConstraint;



//   constraints that residue resid must have 
//   dihedral angles restricted to the given ranges.
typedef struct AngleConstraint {
  char name[4];
  double phiMax, phiMin; 
  double psiMax, psiMin;
  double omegaMax, omegaMin;
  double chiMax, chiMin;
  int resid;
  void *next;
} AngleConstraint;


typedef struct {
  int begin_num, end_num,  num_of_resds;
  int numconstraints;
}FRAG_RES_LIST;





/* GLOBAL VARIABLES */

extern DistConstraint *distConstraints, *old_distConstraints, *present_distConstraints;
extern DistConstraint  *finished_distConstraints, *complete_distConstraints;

extern char *pdbOut; // output finished file to PDB
extern void (*eval_function_ptr)(Genome *, int);
int numconstraints;
extern int haveCSFile; // Whether the user has specified a chemical shifts file.
extern int user_spec_iterations; // The number of iterations for which the user has requested that we run the program. -1 means that the user
						  // has not specified a value, so the program will decide based on other factors when to stop iterating.
extern int exportBestIntermediateStructures; // Boolean. Whether to export the best intermediate structures in each iteration.
extern char *export_best_directory; // Directory into which to export the best intermediate structures in each iteration.
extern int exportAllIntermediateStructures; // Boolean. Whether to export all new intermediate structures in each iteration.
extern char *export_all_directory; // Directory into which to export all new intermediate structures in each iteration.





/* PROTOTYPES */

//int getNthAtom(PDBFile *pdb, int n, char *tag);

DistConstraint *readDistanceConstraints(char *fname);
double check_presentConstraints(Genome *g, int n);
double checkConstraints(Genome *g, int n, char*);
double checkRMSD(Genome *g, int n);

void sortPopulation(Genome **pop);
void call_SCWRL(char *scwrl_path, ResData *res, int numRes, int thread_id);
void allocPopulation(Genome **pop, int n);
void initPopulation(Genome **pop, ResData *seed, int n, ResConstraint *constraints);
float compareGenomeScores(Genome *a, Genome *b);
int evolve(Genome **complete_pop, ResData *complete_seed,  int complete_num_resds,
		ResConstraint *constraints, double initialScore);
void freePopulation(Genome **pop);
void freeJustPopulation(Genome **pop);
void free_angle_constraints(AngleConstraint *head);
void free_dist_constraints(DistConstraint *head);
void print_corr_coeffs(Genome *g);
void print_potential_components(Genome *g);
void print_corr_coeffs_and_potential_components(Genome *g, FILE *fpw, int );
void print_pop(Genome* pop, int,  FILE* );

#ifdef THREADS 
int wait_for_all(const int);
void * worker_thread(void *);
int thread_it(Genome **g); 
	// mutex to control the worker thread - Cam

pthread_mutex_t thread_mutex;
pthread_cond_t work_ready;
pthread_cond_t work_done;
int work_ready_flag;
int work_done_flag;
int threads_waiting; 

#endif

#endif
