#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "minimize.h"
#include "simpredShift.h"
#include "readfileShift.h"
#include "memwatch.h"
#include "main_shiftx.h"
#include "ga_potential.h"

#define DEBUG  0

RESIDUE *Residues[NUM_OF_THREADS]; // Global array of RESIDUE structs, which we will pass to calcChemShifts().

MATRIX *experimental_chem_shifts;
	// Experimentally-measured chemical shifts for the peptide, with random coil shifts already subtracted.
int num_residues_expt_chem_shifts; // Number of residues represented by experimental_chem_shifts.
MATRIX *Predicted_Chem_Shifts[NUM_OF_THREADS]; // Points to memory allocated for the predicted chemical shifts for one peptide.
MATRIX *randomCoilShifts = NULL; // Chemical shifts for a random coil.

int *pred_aln;
int *expt_aln;
	// int arrays containing alignment data for the predicted and experimental chemical shifts. The sequences
	// of the predicted chemical shifts and experimental shifts may differ, so an alignment is computed by
	// get_alignment_data().
	//
	// The format for, for example, expt_aln, is an array of indices into the MATRIX struct of experimental
	// chemical shifts (e.g. experimental_chem_shifts). If expt_aln_data contains (2, 3, 5, ...) then only the
	// chemical shifts for experimental_chem_shifts[2], experimental_chem_shifts[3], experimental_chem_shifts[5], etc.
	// will be used in calculating the chemical shift correlation scores. If pred_aln_data contains (0, 1, 3, ...) then
	// experimental_chem_shifts[2] is matched with the prediced shift with index 0; experimental_chem_shifts[3] is
	// matched with the prediced shift with index 1; and so on. The alignment data from get_alignment_data() only
	// includes pairs of identical residues.
int aln_length;
	// The number of elements in pred_aln and expt_aln with meaningful data.


static float computeChemShiftCorrelations(MATRIX *pred_shifts, MATRIX *expt_shifts,
		int num_resds_pred, int num_resds_expt, float *corr_coeffs);


/*
 * Function description as determined by observation by David Arndt:
 *
 * This function reads in the chemical shifts from the file with name queryinput, allocates a
 * one-dimensional array of MATRIX structs, and loads the chemical shift values into this array.
 * The chemical shifts file must be in shifty format. The function sets the variable 'query' to
 * point to this array, and sets seqlen to the number of residues in the chemical shifts file.
 *
 * Additional comments added below by David Arndt. 
 */
void read_query_seq(char *queryinput, MATRIX **query, int *seqlen){
	int i=0,j=0,PHI=99,PSI=99, AA=99, CA_var=99, NUM=99, CB_var=99,CO_var=99, N_var=99, HN_var=99,HA_var=99,ich,col,len,query_len;
		// Variables PHI, NUM, CA_var, etc. are the column numbers for the indicated type of data.
		// Note: The suffix _var has been added to some of the variables above (and in this funciton) by David Arndt
		// to avoid conflicts with constants defined in pepmake.h. Feb 28, 2007.
	char label[12][12],line[256],ch;
	COLUM *tmp;
	FILE *fpquery;

	fpquery=fopen(queryinput,"r");
	assert(fpquery != NULL);

	tmp = (COLUM *) malloc (50*sizeof(COLUM));
	assert(tmp != NULL && "readfileShift.c: tmp is NULL");

	len= 50;

	// Parse the file header to determine the names of each column of data.

	while(1){

		if((ich=fgetc(fpquery))!=EOF){
			ch=(char)ich;
			i=0;j=0; 
			while(1){
				ich=fgetc(fpquery);
				ch = (char)ich;

				if(ch=='\n'){
					if(j==0) i--;
					else label[i][j]='\0';
					goto OUT;
				}
				else if(ich>64 && ich<91){ // If char is within A-Z (inclusive)
					label[i][j]=ich;
					j++; 
				}
				else{  
					label[i][j]='\0';
					/*    printf("%s\n",label[i]); */
					i++;
					if(j==0) i--;
					else j=0;
				}
			}
		}
		else break;

	}
	OUT:
	col=i+1;

	j=0;

	// Determine the column numbers for each column type.

	for(i=0;i<col;i++){
		if(strcmp(label[i],"NUM")==0 || strcmp(label[i],"#")==0) NUM=i;
		if(strcmp(label[i],"AA")==0 || strcmp(label[i],"RES")==0) AA=i;
		if(strcmp(label[i],"CA")==0) CA_var = i;
		if(strcmp(label[i],"CB")==0) CB_var = i;
		if(strcmp(label[i],"CO")==0 || strcmp(label[i],"C")==0) CO_var=i;
		if(strcmp(label[i],"N") ==0 || strcmp(label[i],"N15")==0)  N_var=i;
		if(strcmp(label[i],"HN")==0 || strcmp(label[i],"H")==0) HN_var=i;
		if(strcmp(label[i],"HA")==0) HA_var = i;
		if(strcmp(label[i],"PHI") ==0)  PHI = i;
		if(strcmp(label[i],"PSI")==0) PSI = i;
	}

	if(NUM==99){
		NUM=0;
		AA++;
		HA_var++;
		CA_var++;
		CB_var++;
		CO_var++;
		N_var++;
		HN_var++;
		PHI++;
		PSI++;
	}
	// printf("%d %d %d %d %d %d %d %d\n",NUM, AA, CA, CB, CO, N, HN, HA);
	i=0;
	// printf("before allocationg\n");

	// Read the main data from the file into 'tmp', which is an array of COLUM
	// structs with one COLUM struct per line. (Yes, it is inappropriately named).
	
	while(1){
		if(i==len){
			tmp = (COLUM *)realloc((COLUM *)tmp,(50+len)*sizeof(COLUM));
			assert(tmp != NULL && "readfileShift.c: tmp is NULL in realloc");
			len+=50;
		} 
		if(fgets(line,256,fpquery)!=NULL){  
			//if(line[2]!='-' && line[0]!='\n'){
			if( !((line[1]=='-' && line[2]=='-') || line[0]=='\n') ){
				// Changed to allow negative residue numbers. David Arndt, Jan 9, 2008.
				
				if(line[1]=='*')
					sscanf(&line[2],"%s %s %s %s %s %s %s %s %s %s",tmp[i].col[0],tmp[i].col[1],
					tmp[i].col[2],tmp[i].col[3],tmp[i].col[4],tmp[i].col[5],tmp[i].col[6],
					tmp[i].col[7],tmp[i].col[8],tmp[i].col[9]);
				else
					sscanf(line,"%s %s %s %s %s %s %s %s %s %s",tmp[i].col[0],tmp[i].col[1],
					tmp[i].col[2],tmp[i].col[3],tmp[i].col[4],tmp[i].col[5],tmp[i].col[6],
					tmp[i].col[7],tmp[i].col[8],tmp[i].col[9]);
					i++;
			}

		}
		else break;
	}
	
	// Allocate an array of MATRIX structs (one per residue) and set the chemical shifts and
	// other variables.
	
	query_len=i; *seqlen=i;
	(*query)=(MATRIX *)malloc((i+1)*sizeof(MATRIX));
	assert( (*query) != NULL );
		// Note by David Arndt: I don't know why memory for (i+1) MATRIX structs is used;
		// i should be enough, right? Unless (i+1) is assumed in some other part of the code,
		// but I can't see where.

	//  printf("read_query_seq,query_len=%d\n",i);
	for(i=0; i<query_len; i++){    
		(*query)[i].num = atoi(tmp[i].col[NUM]);
		(*query)[i].AA = tmp[i].col[AA][0];

		//Added by RM. Sep 15,'06.
		if(HA_var != 99 && HA_var != 100)
			(*query)[i].shifts[0]=(float)atof(tmp[i].col[HA_var]); 
		else (*query)[i].shifts[0]=0;

		if(HN_var != 99 && HN_var != 100)
			(*query)[i].shifts[1]=(float)atof(tmp[i].col[HN_var]); 
		else (*query)[i].shifts[1]=0;

		if(N_var != 99 && N_var != 100)
			(*query)[i].shifts[2]=(float)atof(tmp[i].col[N_var]); 
		else (*query)[i].shifts[2]=0;

		if(CA_var != 99 && CA_var != 100)
			(*query)[i].shifts[3]=(float)atof(tmp[i].col[CA_var]); 
		else (*query)[i].shifts[3]=0;

		if(CB_var != 99 && CB_var != 100)
			(*query)[i].shifts[4]=(float)atof(tmp[i].col[CB_var]); 
		else (*query)[i].shifts[4]=0;

		if(CO_var != 99 && CO_var != 100)
			(*query)[i].shifts[5]=(float)atof(tmp[i].col[CO_var]); 
		else (*query)[i].shifts[5]=0;

		if(PHI != 99 && PHI != 100)
			(*query)[i].shifts[6]=(float)atof(tmp[i].col[PHI]); 
		else (*query)[i].shifts[6]=0;

		if(PSI != 99 && PSI != 100)
			(*query)[i].shifts[7]=(float)atof(tmp[i].col[PSI]); 
		else (*query)[i].shifts[7]=0;

	}
	if(tmp!=NULL) 
		free(tmp);
	fclose(fpquery);
}//*******************************************




/*
  Calls shiftx to create the chemical shift file
*****************************************************************/
char* create_csfile(char *input){
  
  char *call_shiftx, *shiftx_output, *truncate_shiftx_output;
  
  shiftx_output = (char*)malloc((strlen(input)+strlen(".shiftx")+1)*sizeof(char));
  assert(shiftx_output != NULL);
  sprintf(shiftx_output,"%s.shiftx",input);
  //printf("shiftx_out = %s\n", shiftx_output);
  
  
  call_shiftx = (char*)malloc((strlen("./shiftx 1 ")+strlen(input)+strlen(" ")+strlen(shiftx_output)+1)*sizeof(char));
  assert(call_shiftx != NULL);
  sprintf(call_shiftx,"./shiftx 1 %s %s",input,shiftx_output);
  //printf("call_shiftx = %s\n", call_shiftx);
  if (system(call_shiftx) != 0) {
	fprintf(stderr, "ERROR: Call to shiftx failed.\n");
	exit(-1);
  }
  free(call_shiftx);

  truncate_shiftx_output = (char*)malloc((strlen("perl truncshiftx.pl ")+strlen(shiftx_output)+1)*sizeof(char));
  assert(truncate_shiftx_output != NULL);
  sprintf(truncate_shiftx_output,"perl truncshiftx.pl %s",shiftx_output);
  //printf("truncate_shiftx_output = %s\n", truncate_shiftx_output);
  if (system(truncate_shiftx_output) != 0) {
	fprintf(stderr, "ERROR: Call to truncshiftx.pl failed.\n");
	exit(-1);
  }
  free(truncate_shiftx_output);

  return(shiftx_output);
}//*****************************************************************



//*******************************************************************
void print_shifts(MATRIX *sh, int num_resds){
  int i;

  printf("%s",SHIFTS_HEADER);
  printf(" --- --- ------ ------ -------- ------- ------- --------\n");
  for(i=0; i<num_resds; i++){
    printf(" %-4d %c %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f\n",sh[i].num, sh[i].AA, 
	   sh[i].shifts[0], sh[i].shifts[1],  sh[i].shifts[2], sh[i].shifts[3], sh[i].shifts[4], sh[i].shifts[5]); 
  }
}//*******************************************************************


/*
 * Loads the random coil shifts into the variable randomCoilShifts.
 */
void loadRandomCoilShifts() {
	int temp;
	
	read_query_seq(RANDOM_COIL_SHIFTS, &randomCoilShifts, &temp);
}

/*
 * Generates the alignment data that is used in calculation of chemical shift correlation scores.
 *
 * This function first allocates memory for the global variables pred_aln and expt_aln. Then it
 * extracts the sequences from (1) the given array of ResData structs of length numRes, and (2) the
 * experimental chemical shifts in expt_shifts, for num_resds_expt residues. An alignment is done between these two
 * sequences, and the values of the global variables pred_aln, expt_aln, and aln_length are set.
 */
void get_alignment_data(ResData *res, int numRes, MATRIX *expt_shifts, int num_resds_expt) {
	
	int aln_mem; // Amount of memory to allocate for alignment.
	int i;
	int aln_index;
	FILE *ff;
	char buffer[MAX_LINE_LEN];
	char *pred_seq;
	char *expt_seq;
	char *pred_segment;
	char *expt_segment;
	char pred_char;
	char expt_char;
	int pred_index;
	int expt_index;
	int num_identical_residues;
	int num_aligned_residues;
	float seq_identity; // Proportional.
	float seq_identity_threshold = 0.3F; // Fail if sequence alignment results in a sequence identity below this proportion.
	
	if (numRes < num_resds_expt) {
		aln_mem = numRes;
	}
	else {
		aln_mem = num_resds_expt;
	}
	pred_aln = (int *) malloc(sizeof(int) * aln_mem);
	expt_aln = (int *) malloc(sizeof(int) * aln_mem);
	
	pred_seq = (char *)malloc(sizeof(char) * (numRes + 1));
	expt_seq = (char *)malloc(sizeof(char) * (num_resds_expt + 1));
	
	// Extract the sequence from the array of ResData structs.
	get_seq_from_ResData(res, numRes, pred_seq);
	
	// Extract the sequence from the experimental chemical shifts.
	for (i = 0; i < num_resds_expt; i++) {
		expt_seq[i] = expt_shifts[i].AA;
	}
	expt_seq[num_resds_expt] = '\0';
	
	if (strcmp(pred_seq, expt_seq) == 0) {
		// The sequences are identical. Set the alignment data to represent a perfect alignment.
		
		for (i = 0; i < numRes; i++) {
			pred_aln[i] = i;
			expt_aln[i] = i;
		}
		aln_length = numRes;
	}
	else {
		// The sequences are not identical. Perform an alignment with ClustalW.
		
		pred_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
		expt_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
		
		// Write FASTA file with the two protein sequences.
		
		ff = fopen("tmp_align.fa", "w");
		fprintf(ff, ">pred\n");
		fprintf(ff, "%s\n", pred_seq);
		fprintf(ff, ">expt\n");
		fprintf(ff, "%s\n", expt_seq);
		fclose(ff);
		
		// Call ClustalW.
		
		if (system("./clustalw tmp_align.fa > /dev/null") != 0) {
			fprintf(stderr, "ERROR: Call to clustalw failed.\n");
			exit(-1);
		}
		
		// Parse alignment file generated by ClustalW.
		
		ff = fopen("tmp_align.aln", "r");
		
		// Skip first 3 lines
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		
		num_identical_residues = 0;
		num_aligned_residues = 0;
		aln_index = 0;
		pred_index = 0;
		expt_index = 0;
		while( fgets(buffer, MAX_LINE_LEN, ff) != NULL ) {
			sscanf(buffer, "%*s %s", pred_segment);
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			sscanf(buffer, "%*s %s", expt_segment);
			
			// Compare the two sequences, and map the secondary structure from the Proteus file (sec_str)
			// to the sequence read from the PDB file.
			i = 0;
			pred_char = pred_segment[0];
			expt_char = expt_segment[0];
			while (pred_char != '\0') {
				
				if (pred_char != '-') {
					if (expt_char != '-') {
						
						if (pred_char == expt_char) { // Only compare shifts between identical residues.
							pred_aln[aln_index] = pred_index;
							expt_aln[aln_index] = expt_index;
							
							aln_index++;
							num_identical_residues++;
							
							if (aln_index > numRes || aln_index > num_resds_expt) {
								fprintf(stderr, "ERROR: Read too many residues in alignment between predicted and\n");
								fprintf(stderr, "experimental chemical shifts.\n");
								exit(1);
							}
						}
						
						pred_index++;
						expt_index++;
						num_aligned_residues++;
					}
					else {
						// expt sequence does not contain a residue that is in the pred
						// sequence. Skip over the residue in the pred sequence.
						pred_index++;
					}
				}
				else {
					// pred sequence does not contain a residue that is in the expt
					// sequence. Skip over the residue in the expt sequence.
					expt_index++;
				}
				
				i++;
				pred_char = pred_segment[i];
				expt_char = expt_segment[i];
			}
			
			// Skip 2 lines.
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				// End of file.
				break;
			}
		}
		
		fclose(ff);
		
		aln_length = aln_index;
		
		if (num_aligned_residues != 0) {
			seq_identity = ((float) num_identical_residues)/((float) num_aligned_residues);
		}
		else {
			seq_identity = 0.0F;
		}
		
		if (seq_identity  < seq_identity_threshold) {
			fprintf(stderr, "ERROR: Aligment between sequences in PDB file and chemical shifts file\n");
			fprintf(stderr, "results in sequence identity of only %4.1f%%!\n", 100.0F*seq_identity);
			fprintf(stderr, "Wrong input files suspected.\n");
			exit(1);
		}
		
		free(pred_segment);
		free(expt_segment);
	}
	
	free(pred_seq);
	free(expt_seq);
}


/*********************************************************************
 Input:  2 structs of type MATRIX. First one is the chemical shifts
         and the second one is the random coil shifts for the 20 amino
         acids.
 Output: The shifts of the first struct are subtracted by the amount in the 
         second struct. 
**********************************************************************/
void subtract_random_coil_shifts(MATRIX *sh, MATRIX *rc, int num_resds) {
  int i,j;
  float CYS_CB_OXD = 40.97;
  float CYS_CB_RED = 29.35;
  float half_of_diff = 5.81; //11.62/2
  //The difference between these 2 numbers is 11.62

  //difference between each of the reduced and the oxidised versions with the original shift
  float oxd_diff, red_diff; 
	
	// Note: Added by David Arndt Nov 22, 2006: If the original chemical shift is 0.0 (i.e.
	// no measurement), then we do not subtract the random coil shift.
	
  for(i=0; i<num_resds; i++){
    for(j=0; j<20; j++){
      
      //If the residues match, 
      //---------------------------------
      if(sh[i].AA == rc[j].AA){
	//printf("sh[%d].AA = %c & rc[%d].AA = %c\n",i,sh[i].AA,j,rc[j].AA);
	/* We will only be modifying the N, CA and CB shifts.
	   These are shifts[2], [3] and [4]
	   We also need to check for the oxidised or reduced form for Cystein */
	
	//If we have a Cystein, then check whether the shift is closer to the oxidised or reduced form.
	if(sh[i].AA == 'C'){
	  
	  red_diff = fabs(sh[i].shifts[4] - CYS_CB_RED);
	  oxd_diff = fabs(sh[i].shifts[4] - CYS_CB_OXD);
	  
	  if((sh[i].shifts[4] <= CYS_CB_RED) || (red_diff < half_of_diff)){
	    if (sh[i].shifts[4] != 0.0) sh[i].shifts[4]-=CYS_CB_RED; // CB
	  }
	  if((sh[i].shifts[4] >= CYS_CB_OXD) || (oxd_diff < half_of_diff)){
	    if (sh[i].shifts[4] != 0.0) sh[i].shifts[4]-=CYS_CB_OXD; // CB
	  }
	  
	  //if (sh[i].shifts[2] != 0.0) sh[i].shifts[2]-=rc[j].shifts[2]; // N
	  if (sh[i].shifts[3] != 0.0) sh[i].shifts[3]-=rc[j].shifts[3]; // CA
	}else{
	  
	  //if (sh[i].shifts[2] != 0.0) sh[i].shifts[2]-=rc[j].shifts[2]; // N

	  if (sh[i].shifts[3] != 0.0) sh[i].shifts[3]-=rc[j].shifts[3]; // CA
	  if (sh[i].shifts[4] != 0.0) sh[i].shifts[4]-=rc[j].shifts[4]; // CB
	}
	
      }//--------------------------------
       
    }//end for(j..

  }//end for(i..

}//END subtract_coil_shifts ******************************************




/*
 * Predicts the chemical shifts for the PDB file with name PDB_filename (via a system call
 * to shiftx), and also loads the experimental chemical shifts from the file with name
 * chem_shifts_filename. The random coil chemical shifts are loaded into memory and subtracted
 * from both sets of chemical shifts.
 *
 * The sample correlations between the experimental and predicted chemical shift values are then computed.
 * The correlation coefficients are stored in corr_coeffs:
 *
 *    corr_coeffs[0] = correlation coefficient for HA shifts
 *    corr_coeffs[1] = correlation coefficient for HN shifts
 *    corr_coeffs[2] = correlation coefficient for  N shifts
 *    corr_coeffs[3] = correlation coefficient for CA shifts
 *    corr_coeffs[4] = correlation coefficient for CB shifts
 *    corr_coeffs[5] = correlation coefficient for CO shifts
 * 
 * The sum of the correlation coefficients is also returned from this function.
 */
float get_chem_shift_score_from_PDB_file(ResData *res, int numRes, char *PDB_filename, char *chem_shifts_filename, float *corr_coeffs) {
	int temp; 
	int num_resds_in_PDB_file, num_resds_in_csfile;
	char *predicted_chem_shifts_filename;
	MATRIX *predicted_shifts, *experimental_shifts;
	MATRIX *random_coil_shifts;
	float sum_cs_correlations;
	
	predicted_chem_shifts_filename = create_csfile(PDB_filename); // System call to shiftx.
	read_query_seq(predicted_chem_shifts_filename, &predicted_shifts, &num_resds_in_PDB_file);
	read_query_seq(chem_shifts_filename, &experimental_shifts, &num_resds_in_csfile);
	free(predicted_chem_shifts_filename);
	
	read_query_seq(RANDOM_COIL_SHIFTS, &random_coil_shifts, &temp); 
	//print_shifts(random_coil_shifts, temp);
	//print_shifts(experimental_shifts, num_resds_in_csfile);
	//print_shifts(predicted_shifts, num_resds_in_PDB_file);
	subtract_random_coil_shifts(predicted_shifts, random_coil_shifts, num_resds_in_PDB_file);
	//print_shifts(predicted_shifts, num_resds_in_PDB_file);
	subtract_random_coil_shifts(experimental_shifts, random_coil_shifts, num_resds_in_csfile);
	
	free(random_coil_shifts);
	
	// Compute the alignment data
	get_alignment_data(res, numRes, experimental_shifts, num_resds_in_csfile);
	
	sum_cs_correlations = computeChemShiftCorrelations(predicted_shifts, experimental_shifts,
		num_resds_in_PDB_file, num_resds_in_csfile, corr_coeffs);
	
	free(pred_aln);
	free(expt_aln);
	
	return sum_cs_correlations;
}


/*
 * Computes the chemical shift correlation values for the peptide represented by 'resDatas' compared
 * to the experimental chemical shifts.
 *
 * This function calculates the predicted chemical shifts for the peptide represented by 'resDatas',
 * with length numRes. The peptide should already be built, such as by buildPeptide(). The predicted
 * chemical shifts are stored in pred_chem_shifts, which should already be allocated. The random
 * coils shifts (which should already have been loaded by a call to loadRandomCoilShifts()) are then
 * subtracted from the predicted chemical shifts. The global variable experimental_chem_shifts should point
 * to an array of MATRIX structs that hold the experimental chemical shifts, from which the random coil
 * shifts should have been subtracted beforehand (such as in main_gafolder()).
 *
 * The sample correlations between the experimental and predicted chemical shift values are then computed.
 * The correlation coefficients are stored in corr_coeffs:
 *
 *    corr_coeffs[0] = correlation coefficient for HA shifts
 *    corr_coeffs[1] = correlation coefficient for HN shifts
 *    corr_coeffs[2] = correlation coefficient for  N shifts
 *    corr_coeffs[3] = correlation coefficient for CA shifts
 *    corr_coeffs[4] = correlation coefficient for CB shifts
 *    corr_coeffs[5] = correlation coefficient for CO shifts
 * 
 * The sum of the correlation coefficients is also returned from this function.
 */
float get_chem_shift_score(ResData *resDatas, int numRes, RESIDUE *shiftx_residues, MATRIX *pred_chem_shifts, float *corr_coeffs) {
	
	calcChemShifts(resDatas, numRes, shiftx_residues, pred_chem_shifts);
	
	// debug
	//print_shifts(pred_chem_shifts, numRes);
	
	subtract_random_coil_shifts(pred_chem_shifts, randomCoilShifts, numRes);
	
	// debug
	//print_shifts(randomCoilShifts, 20);
	//print_shifts(pred_chem_shifts, numRes);
	
	return computeChemShiftCorrelations(pred_chem_shifts, experimental_chem_shifts, numRes, num_residues_expt_chem_shifts, corr_coeffs);
}


/*
 * Computes the sample correlations between the given predicted chemical shifts and experimental
 * chemical shifts. Both sets of chemical shifts should already have had random coil shifts
 * subtracted from them.
 *
 * pred_shifts = predicted chemical shifts
 * expt_shifts = experimental chemical shifts
 * num_resds_pred = number of residues in pred_shifts
 * num_resds_expt = number of residues in expt_shifts
 * corr_coeffs = array of 6 floats in which the sample correlations will be stored
 *
 * This function also assumes that the alignment data has already been set in the global variables
 * pred_aln, expt_aln, and aln_length by a call to get_alignment_data().
 *
 * The correlation coefficients are stored in corr_coeffs:
 *
 *    corr_coeffs[0] = correlation coefficient for HA shifts
 *    corr_coeffs[1] = correlation coefficient for HN shifts
 *    corr_coeffs[2] = correlation coefficient for  N shifts
 *    corr_coeffs[3] = correlation coefficient for CA shifts
 *    corr_coeffs[4] = correlation coefficient for CB shifts
 *    corr_coeffs[5] = correlation coefficient for CO shifts
 * 
 * The sum of the correlation coefficients is also returned from this function.
 */
static float computeChemShiftCorrelations(MATRIX *pred_shifts, MATRIX *expt_shifts,
		int num_resds_pred, int num_resds_expt, float *corr_coeffs) {
	
	int i,j, num[6];
	float sum=0.00, **x = NULL, **y = NULL, cc[6];
	float exprt_shift, pred_shift;   
	//float xsum[6], ysum[6], xavg[6], yavg[6], cctmp[3]; // For Raj's original correlation algorithm.
	
	int pred_index;
	int expt_index;
	
	
	// debug
	//print_shifts(pred_shifts, num_resds_pred);
	//print_shifts(expt_shifts, num_resds_expt);
	
	/* 
	We are choosing 6 because we have 6 shifts HA, HN, N, CA,CB and CO 
	x[j] and y[j] are arrays that are of size num_resds_pred.So,
	x[0] array will contain the chemical shift for all the HAs, x[1] array will
	contain the shifts for HN and so on. Note that x[] contains the experimental shifts 
	and y[] will contain the predicted shifts from SHIFTX.
	*/
	x = (float**)malloc(6*sizeof(float*));  assert(x != NULL);
	y = (float**)malloc(6*sizeof(float*));  assert(y != NULL);
	for(j=0; j<6; j++){
		num[j] = 0;  cc[j] = 99;
		//xsum[j] = 0; ysum[j] = 0; // For Raj's original correlation algorithm.
		x[j]=(float*)malloc(num_resds_pred*sizeof(float)); assert(x[j] != NULL);
		y[j]=(float*)malloc(num_resds_pred*sizeof(float)); assert(y[j] != NULL);
	}
	
	
	for (i = 0; i < aln_length; i++) {
		
		// Use the data in the pre-computed alignments to determine which residues are aligned to
		// which.
		pred_index = pred_aln[i];
		expt_index = expt_aln[i];
		
		// debug
		//fprintf(stderr, "comparing %d (%c) to %d (%c)\n", pred_index, pred_shifts[pred_index].AA,
		//	expt_index, expt_shifts[expt_index].AA);
		
		// Check whether the two residue types are identical. They should be guaranteed to be
		// identical by the alignment generated by get_alignment_data(), but in case they are
		// not we should give an error message and exit.
		if (pred_shifts[pred_index].AA != expt_shifts[expt_index].AA) {
			fprintf(stderr, "ERROR: Alignment between predicted and experimental chemical shifts\n");
			fprintf(stderr, "leads to comparison of non-identical residues.\n");
			fprintf(stderr, "%c matched with %c.\n", pred_shifts[pred_index].AA, expt_shifts[expt_index].AA);
			exit(1);
		}
		
		for (j=0;j<6;j++) {

			// Only compare the non-zero shifts.
			if(expt_shifts[expt_index].shifts[j] != 0.0 && pred_shifts[pred_index].shifts[j] != 0.0){

				//Using 2 local variables to improve spatial and temporal locality.
				exprt_shift = expt_shifts[expt_index].shifts[j];
				pred_shift =  pred_shifts[pred_index].shifts[j];

				x[j][num[j]] = exprt_shift;
				y[j][num[j]] = pred_shift;

				//xsum[j] += exprt_shift; // For Raj's original correlation algorithm.
				//ysum[j] += pred_shift; // For Raj's original correlation algorithm.

				num[j]++;
			}
		}
		
	}
	
	
	// debug
	// Print out the data.
	/*
	for(j=0; j<6; j++){
		printf("Data for j == %d (n == %d):\n", j, num[j]);
		for (i = 0; i < num[j]; i++) {
			printf("%lf\t%lf\n", x[j][i], y[j][i]);
		}
		printf("\n");
	}
	*/
	
	// debug
	//for(j=0; j<6; j++){
	//	printf("j == %d: n == %d (out of %d residues with predicted shifts, %d residues with experimental shifts)\n", j, num[j], num_resds_pred, num_resds_expt);
	//}

	// Note : This algorithm for calculating the correlation coefficients was replaced by the one just
	// below. David Arndt, Nov 24, 2006.
	//
	// Raj's original algorithm for calculating correlation coefficient:
	// Calculating the Pearson product moment correlation coefficient
	// For the formula used, please see the web page in this directory.
	// (http://en.wikipedia.org/wiki/Correlation_coefficient)
	// I have used the sample correlation Rxy for the correlation coefficient.
	//-----------------------------------------------------------------
	/*
	for(j=0;j<6;j++){
		cctmp[0] = cctmp[1] = cctmp[2] = 0;

		if(num[j] != 0){

			//Finding the average of each of the shifts for each atom type,
			//i.e. finding the average for each of x[0], x[1]...,and  y[0], y[1] and so on. 
			xavg[j] = xsum[j] / num[j]; 
			yavg[j] = ysum[j] / num[j];

			for(n=0; n<num[j]; n++){
				cctmp[0] += (x[j][n] - xavg[j]) * (y[j][n] - yavg[j]);
				cctmp[1] += pow ((x[j][n] - xavg[j]), 2);
				cctmp[2] += pow ((y[j][n] - yavg[j]), 2);
			}
			//Haiyan's code: My version is below: RM. Nov 06, 2006
			//cctmp[1] = (float)sqrt ((double)cctmp[1]);  Haiyan's code.
			//cctmp[2] = (float)sqrt ((double)cctmp[2]);
			//cc[j] = cctmp[0] / (cctmp[1] * cctmp[2]);

			//I divided the cctmp[1] and cctmp[2] by the number of non-zero shifts for the atom type
			// to get the correct formula of the standard deviation.
			cctmp[1] = (float)sqrt ((double)cctmp[1]/num[j]);  
			cctmp[2] = (float)sqrt ((double)cctmp[2]/num[j]);
			//cc[j] = cctmp[0] / (cctmp[1] * cctmp[2] * (num[j] - 1 ));

			//with num[j] - 1 we are getting a correlation of more than 1.So, I dropped the subtration by 1
			// RM Nov 6, 2006
			cc[j] = cctmp[0] / (cctmp[1] * cctmp[2] * num[j]);
		}
		else {
			cc[j] = 99.0; // 99.0 is used to indicate that correlation could not be calculated.
		}
	}//-------------------------------------------------------------------
	*/
	
	// An alternate implementation of a "numerically unstable" sample correlation algorithm.
	// Based on formula from:
	// http://en.wikipedia.org/wiki/Correlation_coefficient
	/*
	double sumxi[6], sumxisquared[6], sumyi[6], sumyisquared[6], sumxy[6];
	for(j=0;j<6;j++){
		sumxi[j] = 0.0;
		sumxisquared[j] = 0.0;
		sumyi[j] = 0.0;
		sumyisquared[j] = 0.0;
		sumxy[j] = 0.0;
		
		if(num[j] != 0){
			for(i=0; i<num[j]; i++){
				sumxi[j] += (double) x[j][i];
				sumxisquared[j] += ((double) x[j][i])*((double) x[j][i]);
				sumyi[j] += (double) y[j][i];
				sumyisquared[j] += ((double) y[j][i])*((double) y[j][i]);
				sumxy[j] += ((double) x[j][i])*((double) y[j][i]);
			}
			
			double r = (num[j] * sumxy[j] - sumxi[j] * sumyi[j]) / 
				(sqrt(num[j]*sumxisquared[j] - sumxi[j] * sumxi[j]) * sqrt(num[j]*sumyisquared[j] - sumyi[j] * sumyi[j]));
			
			cc[j] = (float) r;
		}
		else {
			cc[j] = 99.0; // 99.0 is used to indicate that correlation could not be calculated.
		}
	}
	*/
   
	
	// Calculate the sample correlation coefficients. This algorithm is adapted from
	// http://en.wikipedia.org/wiki/Correlation_coefficient#Computing_correlation_accurately_in_a_single_pass
	// This algorithm is supposed to have good numerical stability, whereas the ones commented-out above
	// "is notorious for its numerical instability" according to the Wikipedia page.
	// -- David Arndt, Nov 24, 2006.
	for(j=0;j<6;j++){
		int N_int = num[j];
		if(N_int != 0){
			double sum_sq_x = 0.0;
			double sum_sq_y = 0.0;
			double sum_coproduct = 0.0;
			double mean_x = x[j][0];
			double mean_y = y[j][0];
			double sweep;
			double delta_x, delta_y;
			double i_double;
			for (i = 1; i < N_int; i++) {
				i_double = (double) i;
				sweep = (i_double - 1.0) / i_double;
				delta_x = x[j][i] - mean_x;
				delta_y = y[j][i] - mean_y;
				sum_sq_x += delta_x * delta_x * sweep;
				sum_sq_y += delta_y * delta_y * sweep;
				sum_coproduct += delta_x * delta_y * sweep;
				mean_x += delta_x / i_double;
				mean_y += delta_y / i_double;
			}
			double N_double = (double) N_int;
			double pop_sd_x = sqrt( sum_sq_x / N_double );
			double pop_sd_y = sqrt( sum_sq_y / N_double );
			double cov_x_y = sum_coproduct / N_double;
			cc[j] = cov_x_y / (pop_sd_x * pop_sd_y);
		}
		else {
			cc[j] = 99.0; // 99.0 is used to indicate that correlation could not be calculated.
		}
	}
	
   
   /*
     So far, we have obtained 6 different correlation coefficients, each for one particular type of shift.
     Copying these correlation coefficients in the array     
     cc[0] = correlation coefficient for HA shifts
     cc[1] = correlation coefficient for HN shifts
     cc[2] = correlation coefficient for  N shifts
     cc[3] = correlation coefficient for CA shifts
     cc[4] = correlation coefficient for CB shifts
     cc[5] = correlation coefficient for CO shifts
   */
   for(i=0; i<6; i++){
     corr_coeffs[i] = cc[i];
   }

   //debug
	//printf("corr_coeffs:\n");
	//for(j=0; j<6; j++){
	//	printf("%lf\n", corr_coeffs[j]);
	//}

   //Calculating the weighted sum of all the chemical shifts.
   //n=0;
   for(j=0; j<6; j++){  

     if(cc[j] != 99.0){
       sum += cc[j];
       //n++;
     }
   }
   //sum = sum/(float)n;
  
   for(j=0; j<6; j++){
     free(x[j]);
     free(y[j]);
   }
   free(x);free(y);
   
   
   return(sum); 
   //return((cc[1]+cc[3])/2); //return the avg of the HN and CA shift
   
}//END get_chem_shift_score_from_PDB_file **********************************************
