#ifndef __PEPMAKE_H__
#define __PEPMAKE_H__

#include "vec3D.h"
#include "pdbfile.h"

#define USE_SIDECHAINS 1
#define MAX_ATOMS 32 /* maximum number of atom slots in a ResData struct */
#define RESIDUE_ID_NOT_AVAILABLE -1000000
	// An impossible residue ID, indicating that the residue ID is not available.
	// (Note that small negative residue IDs are possible).
#define ANGLE_NA -1000000
	// Indicates that an angle is not available.
#define CCD_SEG_LENGTH 8
	// The number of residues in a segment when we try to adjust its angles with adjustSegmentCCD.
	// This number includes the immobile N-anchor and immobile C-anchor.
	//
	// Different values of CCD_SEG_LENGTH were tested on set of proteins. As CCD_SEG_LENGTH
	// increases, there is steadily more damage to the structure, but Rama and Omega scores improve
	// as well. A CCD_SEG_LENGTH of 8 is a fair amount better than a value of 7, but with
	// CCD_SEG_LENGTH > 8 the improvements in Rama and Omega scores are smaller. Therefore a value
	// of 8 may be a good compromise.
#define MIN_CCD_SEG_LENGTH 4
	// The minimum CCD segment length to use, in cases where the CCD length has been changed from the
	// value CCD_SEG_LENGTH.
#define CCD_MAX_ITER 5000
	// Maximum number of CCD iterations to perform before giving up.
#define ANGLE_CHANGE_THRESHOLD 0.00002
	// When performing rotations with the CCD algorithm, any change in an angle that is less than
	// or equal to this value is considered to be an insignificant change, so we do not need to
	// waste time performing the rotation. Using such a threshold is necessary because, due to
	// numerical instability, the algorithm may compute that an angle should change by a small
	// non-zero amount, when in theory there should be no change in the angle.
	//
	// Different values of ANGLE_CHANGE_THRESHOLD were tested on a set of test structures (with a
	// sumOfSquaredDistances threshold of 0.000001):
	//
	// ANGLE_CHANGE_THRESHOLD
	//       0.00001           No change noticed compared to 0.0.
	//       0.00002           No change noticed compared to 0.0.
	//       0.00003           2 of 41 structures had better Rama or Omega scores.
	//       0.00004           Rarely, structures may have better or worse scores.
	//       0.00005+          The higher the value, the more problems outweigh benefits.
	//
	// Using a value other than 0.0 gave a slight performance improvement when the CCD algorithm was
	// less efficient. Now it is hard to measure relative differences in performance. A value greater
	// than 0.0 is recommended to avoid rotations that should theoretically be exactly 0.0, but are
	// barely non-zero due to numerical instability or rounding error.

#define CCD_MODE_CLOSE_GAP 0
	// Indicates we want to run the CCD algorithm to close a gap in the peptide backbone.
#define CCD_MODE_ALIGN_HBONDS 1
	// Indicates we want to run the CCD algorithm to form or optimize hydrogen bonds.

#define MODIFY_SEG_MODE_REPAIR 0
	// Mode for modifySegmentAngles(), indicating that the torsion angles should be repaired.
#define MODIFY_SEG_MODE_MUTATE 1
	// Mode for modifySegmentAngles(), indicating that a random angle should be mutated.
#define MODIFY_SEG_MODE_PREDITOR 2
	// Mode for modifySegmentAngles(), indicating that angles predicted by Preditor should be used.

#define CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD 0.000001
#define MAX_GAP_ALONG_BETA_STRAND 1
	// Maximum number of residues along one strand of a beta sheet that do not have backbone hydrogen
	// bonds. If there are more than this many residues in a row without hydrogen bonds, it is considered
	// to be a gap in the secondary structure.
#define HBOND_ALIGN_FLEXIBLE_REGION_LENGTH 7
	// In hydrogen bond alignment, the number of residues before and/or after a rigid Beta-strand segment
	// that are normally allowed to be flexible.
#define CCD_POINT_TYPE_C_ANCHOR 0
	// CCD point is for the C-anchor.
#define CCD_POINT_TYPE_HBOND_ACC 1
	// CCD point is from the acceptor end of a hydrogen bond.
#define CCD_POINT_TYPE_HBOND_DON 2
	// CCD point is from the donor and of a hydrogen bond.
#define HBOND_IDEAL_O_H_DIST 1.9
	// The ideal distance in Angstroms between the acceptor O atom and donor H atom in an intra-backbone
	// hydrogen bond.
#define HBOND_ALIGN_RMSD_THRESHOLD 0.2
	// When doing hydrogen bond alignment, this value is the RMSD threshold in Angstroms such that once
	// all the alignment points for the hydrogens bonds have an RMSD (between the current and target
	// positions) that is below or equal to this value, the hydrogen bonds are considered to be
	// sufficiently aligned.
#define CA_CA_GAP_THRESHOLD 4.0
	// If two neighboring residues in a sequence have a C-alpha to C-alpha distance greater than this value
	// (in Angstroms), we call a gap between the two residues.


/*
 * Note: Comments based on observation by David Arndt.
 *
 * Represents a hydrogen bond. Both residues forming the hydrogen bond will (normally?) have an
 * Hbonds struct.
 */
typedef struct Hbonds {
	int partner;
		// The index of the residue that forms this hydrogen bond with the current residue (where
		// the "current residue" is the one containing this Hbonds struct). -1 means no partner,
		// i.e. this Hbonds struct does not specify a hydrogen bond.
	//int type;
		// The role the current residue plays in the formation of this hydrogen bond. 1 means
		// ACCEPTOR, 2 means DONOR, -1 means that this Hbonds struct does not specify a hydrogen bond.
	float distance;
		// The distance in angstroms between the O atom and the H atom between which the hydrogen
		// bond is formed. -1 means that this Hbonds struct does not specify a hydrogen bond.
	float energy; // The energy for this hydrogen bond.
	int hbond_is_valid; // TRUE if this is a valid hydrogen bond; FALSE if this is an "almost formed" hydrogen bond.
} Hbonds;


/*
 * Represents a single residue of a peptide.
 */
typedef struct ResData {
	double phi, psi, omega, chi, N_CA_C1_angle;
		// N_CA_C1_angle is the angle formed by the N of the previous residue and the CA and C1 of the current residue.
	int id;
		// The residue number as it appears in the PDB file (or Preditor output file if no PDB file is available).
		// This is NOT the same as the index into an array of ResData structs.
	int max_atom_index;
		// The maximum index into atoms and names of an entry with meaningful data.
	Vec3D atoms[MAX_ATOMS];
		// If fixed_position is FALSE, 'atoms' represents the computed coordinates of each atom of this residue
		// in the overall peptide chain. If fixed_position is TRUE, 'atoms' represents the fixed 3-D coordinates
		// of this residue.
	char names[MAX_ATOMS][5]; // The 4-character names of each atom in this residue.
	Vec3D template[MAX_ATOMS];
		// The coordinates for each atom in this residue, in canonical form. Backbone and sidechain.
		// If PepMake is called without a PDB file, the program will use the default backbone coordinates, as well
		// as the default side-chain coordinates from the template database. If, on the other hand, PepMake is called
		// with a PDB file, then PepMake will call phipsi with the --db flag to generate a residue coordinates file
		// with all residues in canonical form, and then read in this file and save the canonical coordinates in
		// this field.
		// "Canonical form" is: C-alpha at the origin, C on the x-axis, and N below the x-axis (negative y-coordinate)
		// and in the xy plane.
	char resName[4]; // The name of this residue (3-letter residue abbreviation).
	int fixed_position;
		// Boolean. Whether this residue has a fixed position. If it does, then this residue's fixed position
		// will be set in the 'atoms' member variable, and buildPeptide() should not change the position but
		// rather use this residue as a starting position onto which downstream residues (i.e. those closer to
		// the C-terminsu) are built.
	Hbonds hbonds_acceptor[2]; // Hydrogen bonds in which this residue participates as an acceptor.
	Hbonds hbonds_donor[2]; // Hydrogen bonds in which this residue participates as a donor.
	Hbonds hbonds_intended_acc[2];
		// Hydrogen bonds which the program will try to form or optimize, in which this residue participates as an
		// acceptor.
	Hbonds hbonds_intended_don[2];
		// Hydrogen bonds which the program will try to form or optimize, in which this residue participates as a
		// donor. Note: This means that the N-H group for "this" residue ID will be involved in the hydrogen bond,
		// and the coordinates of this N-H group are usually in the previous ResData struct within an array of
		// ResData structs representing a peptide, unless the current ResData is the first one in the array.
} ResData;


// The constraints for a single residue. The idea is to have an array of ResConstraint structs,
// one for each residue in the peptide. The original angles predicted by Preditor are also included.
typedef struct ResConstraint {
	double phi_preditor, phi_min, phi_max;
	double psi_preditor, psi_min, psi_max;
	double omega_preditor, omega_min, omega_max;
} ResConstraint;


/*
 * Represents information about a single point that should be aligned in a CCD algorithm.
 */
typedef struct CCD_alignment_point {
	int resIndex;
		// The index of the residue from which this point is derived. For donor groups in hydrogen bonds,
		// this resIndex refers to the index of the ResData struct with the residue ID of the residue that
		// is involved in the hydrogen bond, although the actual HN atom coordinate comes from the ResData struct
		// with index (resIndex - 1). (Yes, this is stupid and confusing).
	int point_type; // One of CCD_POINT_TYPE_C_ANCHOR, CCD_POINT_TYPE_HBOND_ACC, CCD_POINT_TYPE_HBOND_DON.
	Vec3D current; // The current position of this point.
	Vec3D target; // The target position of this point.
	
	// Various vectors and values involved in the CCD calculations.
	Vec3D curr_translated;
	Vec3D vec_O; // The point on the rotation axis closest to the target point.
	Vec3D vec_r;
	Vec3D vec_r_unit; // Unit vector corresponding to vec_r.
	Vec3D vec_s;
	Vec3D vec_f;
	double mag_r; // The magnitude of vec_r.
} CCD_alignment_point;


// User-specified parameters for gap closure.
typedef struct GAP_CLOSURE_PARAMS {
	int close_gap; // Boolean.
	int N_anchor_res_index;
	int C_anchor_res_index;
	int N_anchor_res_id;
	int C_anchor_res_id;
	double max_threshold;
} GAP_CLOSURE_PARAMS;


/* Global variables */
extern int havePDBFile; // Whether the user has specified an input PDB file
extern int rotateChi; // Whether to adjust chi angles when building a peptide.
extern int ignoreDBBackbone; // whether to ignore the databse backbone atoms

/* Lookup tables for N-CA-C1 angles, for general residues (the 20 standard amino
 * acids, minus Gly and Pro), Gly, and Pro.  These tables are used to lookup the
 * N-CA-C1 angle based on the phi and psi angle.
 * Element [0][0] corresponds to the overall average angle;
 * the first row and column correspond to the lookups for phi and psi
 * individually. To perform an angle lookup, it is advised to use one of the
 * functions lookup_N_CA_C1_angle, lookup_N_CA_C1_angle_from_phi,
 * lookup_N_CA_C1_angle_from_psi, or lookup_N_CA_C1_angle_baseline.
 */
extern double general_N_CA_C1_angle_lookup[37][37];
extern double gly_N_CA_C1_angle_lookup[37][37];
extern double pro_N_CA_C1_angle_lookup[37][37];
extern double xpr_N_CA_C1_angle_lookup[37][37];

#define Np		0 // The N atom preceding CA.
#define CA		1
#define C1		2
#define O		3
#define N		4 // The N atom following CA.
#define CA2		5
#define HN		6 // H atom for N.
#define HNp		7 // 1st H atom for Np (only for first residue).
#define HNp2	8 // 2nd H atom for Np (only for first residue).
#define HNp3	9 // 3rd H atom for Np (only for first residue).
#define HA		10 // H atom for CA.
#define HA2     11 // Second alpha hydrogen for Glycine. Note that this has the same value as CB!
#define CB		11
#define CG		12
// The "format" of the order of these atoms is a bit idiosyncratic. The rationale is so that
// every atom from CB and up will be a side-chain atom, so that any transformations which wish to
// act only on the sidechain can be done on these atoms as a group. An exception is the OXT atom,
// which should only exist for the last residue; the code treats this as a special case. The backbone
// H atoms are given special slots in the list, and will therefore not appear below CG.
//
// Note also that the Np atom is mostly used in the code for the first residue; for other residues,
// data from the previous ResData for indices N and HN are used. When building the peptide chain,
// the program ensures that the CA2 atom of the preceding ResData overlaps with the CA of the
// following ResData. Thus, the N atom from the previous ResData may not end up in exactlly the same position
// as the Np atom of the following ResData -- when a PDB file is given, these two atoms should still
// be in almost the same location (perhaps different from rounding error), but when the template
// database is used these two N atoms may be in noticeably different places. In any case, the N atom
// position is output, and the Np atom is ignored (except for the first residue).
//
// In addition, for GLY residues the second HA atom is kept at index CB. CG stores the first gamma
// atom (if present), whether it is a carbon atom or not.
//
// Any atoms that do not exist for a given residue are given coordinates (ATOM_COORD_NA, ATOM_COORD_NA,
// ATOM_COORD_NA). This is normally the case for the HNp, HNp2, and HNp3 atoms for all residues except
// the first residue. It may also be the case when PDB files are missing certain atoms, e.g. hydrogen
// atoms or sidechain atoms.
//
// -- David Arndt, Nov 16, 2007

/* Function prototypes */

void printPDBAtoms(FILE *out, ResData *residues, int numRes);
void printPDBAtomsFilename(char *output_filename, ResData *residues, int numRes);
int getAtomIndexFromAtomName(char *name);
char *getAtomNameFromAtomIndex(int index);
void outputDebugCoords(char *filename, char *resName, int id, Vec3D *coords, int outputCB, int outputPRO);
void copyPosition(ResData *src, ResData *dst);
void copySingleResData(ResData *src, ResData *dst, int copy_hbonds);
void copyMultipleResData(ResData *src, ResData *dst, int n, int copy_hbonds);
void get_seq_from_ResData(ResData *res, int n, char *seq);
double lookup_N_CA_C1_angle(double whichLookupTable[37][37], double phi, double psi);
double lookup_N_CA_C1_angle_from_phi(double whichLookupTable[37][37], double phi);
double lookup_N_CA_C1_angle_from_psi(double whichLookupTable[37][37], double psi);
double lookup_N_CA_C1_angle_baseline(double whichLookupTable[37][37]);
double convert_char_to_N_CA_C1_angle_deviation(char givenChar);
void read_N_CA_C1_angle_lookup_file(char *filename, double targetArray[37][37]);
void predict_N_CA_C1_angle(ResData *res, int next_res_is_Pro);
void predict_N_CA_C1_angles(ResData *residues, int n);
void debugOutputRamaScores(ResData *residues, int numRes, char *debugFilename);
void repair_O_geometry(ResData *residues, int numRes);
void repairAngles(ResData *residues, int numRes, char *secStr);
int modifySegmentAngles(ResData *residues, int numRes, int N_anchor_index, int C_anchor_index, int rama_score_threshold,
	int mode, ResConstraint *constraints);
void alignHydrogenBonds(ResData *residues, int totalRes);
void readResCoordsFile(ResData *residues, int totalRes, char *resCoordsFilename);
int countResidues(char *fname);
void rebuildPeptideSegment(ResData *residues, int startIndex, int endIndex, int onlyBackbone);
void buildPeptide(ResData *residues, int numRes);
void loadResidueTemplates(char *inputPDBFilename, PDBFile pdb, int usePDBFile, int createResCoordsFile,
		int haveResCoordsFile, char *resCoordsFilename, ResData *residues, int n);
void PDBtoResData(PDBFile pdb, ResData *res);
void readAngles(char *inputAnglesFilename, ResData *residues, int n);
void * createResidueCoordinatesFile(char *inputPDBFilename);
void * createAnglesFile(char *inputPDBFilename);

#endif 
