/**
Functions for Ramachandran table lookup. read_rama_table() and the basic data structure
organization are taken from functions.c in the GA potential (formerly "steve_potential") program.

Author: David Arndt, Jan 2006.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include "pepmake.h"
#include "rama.h"

#define CH2INT(x) ( x - '0' )
#define MAX_LINE_LEN		256

/*
The following Rama tables are indexed as follows:

	rama_score = ramaTableGeneral[psiindex][phiindex];

Note that phiindex ranges from 0 to 35 as phi ranges from -180 to 180,
and psiindex ranges from 0 to 35 as psi ranges from 180 to -180 (the reverse).

For phi angles, phiindex 0 includes -180.0 (inclusive) to -170.0 (exclusive),
and so on until phiindex 35 includes 170.0 (inclusive) to 180.0 (exclusive).
Thus 180.0 is illegal and -180.0 should be used instead.

For psi angles, psiindex 0 includes 180.0 (inclusive) to 170.0 (exclusive),
and so on until psiindex 35 includes -170.0 (inclusive) to -180.0 (exclusive).
Thus -180.0 is illegal and 180.0 should be used instead.

The Rama score can be 0, 1, 2, or 3. 3 is very good, and 0 is very bad (disallowed).
*/
int ramaTableGeneral[RAMA_LEN][RAMA_LEN];
int ramaTableGlycine[RAMA_LEN][RAMA_LEN];
int ramaTableProline[RAMA_LEN][RAMA_LEN];
int ramaTablePreProline[RAMA_LEN][RAMA_LEN];


static void read_rama_table(char *filename, int targetArray[RAMA_LEN][RAMA_LEN]);


/*
 * Reads in the Ramachandran table from the file with the given filename into the
 * given array.
 */
static void read_rama_table(char *filename, int targetArray[RAMA_LEN][RAMA_LEN]) {
  char line[MAX_LINE_LEN], c;
  int cint;
  FILE *fp;
  int i, j;
  
  
  fp = fopen(filename, "r");
  if (fp == NULL) {
    printf("ERROR: cannot find rama table file \"%s\".\n", filename);
    exit(-1);
  }
  
  /* skip over header in rama table */
  c = fgetc(fp);
  while (c == '#') {
    fgets(line, MAX_LINE_LEN-1, fp);
    c = fgetc(fp);
  }
  ungetc(c, fp);
  
  /* read table */
  for (i=0; i<RAMA_LEN; i++)
    for (j=0; j<RAMA_LEN; j++) {
      cint = CH2INT(fgetc(fp));
      while ( (cint<0) || (cint>4))
	cint = CH2INT(fgetc(fp));
      targetArray[i][j] = cint;
    }
  fclose(fp);
  
}       /* read_rama_table */


/*
 * Loads the data from files into ramaTableGeneral, ramaTableGlycine, ramaTableProline, and ramaTablePreProline.
 */
void read_rama_tables() {
	read_rama_table("rama_tables/rama_general.table", ramaTableGeneral);
	read_rama_table("rama_tables/rama_glycine.table", ramaTableGlycine);
	read_rama_table("rama_tables/rama_proline.table", ramaTableProline);
	read_rama_table("rama_tables/rama_pre_proline.table", ramaTablePreProline);
}


/*
 * Returns the index into ramaTableGeneral etc. for the given phi angle.
 * Note that getPsiIndex is different!
 */
int getPhiIndex(double phi) {
	if (phi == ANGLE_NA) {
		return -1;
	}
	else if (phi < -180.0) {
		phi += 360.0;
	}
	else if (phi >= 180.0) {
		phi -= 360.0;
	}
	return (int) ((180.0+phi)/10.0);
}


/*
 * Returns the index into ramaTableGeneral etc. for the given psi angle.
 * Note that getPhiIndex is different!
 */
int getPsiIndex(double psi) {
	if (psi == ANGLE_NA) {
		return -1;
	}
	else if (psi <= -180.0) {
		psi += 360.0;
	}
	else if (psi > 180.0) {
		psi -= 360.0;
	}
	return (int) ((180.0-psi)/10.0);
		// Note: Use -psi because the table goes from 180 to -180 as the index increases.
}


/*
 * Returns the Rama score of the residue with index resIndex.
 * 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
 */
int lookupRamaScore(ResData *residues, int resIndex, int numRes) {
	return lookupRamaScoreByIndices(residues, resIndex, numRes, getPhiIndex(residues[resIndex].phi), getPsiIndex(residues[resIndex].psi));
}


/*
 * Returns the Rama score based on phiindex and psiindex (indices into ramaTableGeneral etc.).
 * The other variables are provided only to indicate the type of residue.
 * 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
 */
int lookupRamaScoreByIndices(ResData *residues, int resIndex, int numRes, int phiindex, int psiindex) {
	
	int isPrePro;
	
	if (resIndex == 0) {
		// any residue which could not be assigned, like at the end of a chain
		return 3;
	}
	else {
		if ( (resIndex+1 != numRes) && (strcmp(residues[resIndex+1].resName, "PRO") == 0) ) {
			// The next residue is PRO.
			isPrePro = TRUE;
		}
		else {
			isPrePro = FALSE;
		}
		
		return lookupRamaScoreAbstractByIndices(residues[resIndex].resName, phiindex, psiindex, isPrePro);
	}
	
}


/*
 * Returns the Rama score for the residue with the given name having the given phi and psi
 * angles. 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
 * isPrePro indicates whether the residue in question is before a Proline residue.
 */
int lookupRamaScoreAbstract(char *resName, double phi, double psi, int isPrePro) {
	return lookupRamaScoreAbstractByIndices(resName, getPhiIndex(phi), getPsiIndex(psi), isPrePro);
}


/*
 * Returns the Rama score based on phiindex and psiindex (indices into ramaTableGeneral etc.).
 * 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
 * resName indicates the residue name, and isPrePro indicates whether the residue in question
 * is before a Proline residue.
 */
int lookupRamaScoreAbstractByIndices(char *resName, int phiindex, int psiindex, int isPrePro) {
	int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
	
	if (psiindex == -1 || phiindex == -1) {
		return 3; // Do not penalize for not having phi or psi angles.
	}
	else if (isPrePro) {
		// The next residue is PRO.
		rama_score = ramaTablePreProline[psiindex][phiindex];
	}
	else if (strcmp(resName, "GLY") == 0) {
		// The current residue is GLY.
		rama_score = ramaTableGlycine[psiindex][phiindex];
	}
	else if (strcmp(resName, "PRO") == 0) {
		// The current residue is PRO.
		rama_score = ramaTableProline[psiindex][phiindex];
	}
	else {
		// The current residue is a general residue.
		rama_score = ramaTableGeneral[psiindex][phiindex];
	}
	
	return rama_score;
}


/*************************************************************/
/* Returns the magnitude of the difference between the two   */
/* given angles, with angles measured in degrees.            */
/*************************************************************/
double angleDist(double a1, double a2) {
	
	// Put each angle in the range [-180.0, 180.0).
	while (a1 < -180.0) {
		a1 += 360.0;
	}
	while (a1 >= 180.0) {
		a1 -= 360.0;
	}
	
	while (a2 < -180.0) {
		a2 += 360.0;
	}
	while (a2 >= 180.0) {
		a2 -= 360.0;
	}
	
	// Adjust angles 
	if (a1 - a2 >= 0.0) {
		if (a1 - a2 < 180.0) {
			// No change required.
		}
		else {
			a2 += 360.0;
		}
	}
	else {
		if (a1 - a2 > -180.0) {
			// No change required.
		}
		else {
			a1 += 360.0;
		}
	}
	
	return fabs(a1 - a2);
}


/*
 * Returns the angle closest to desiredAngle (in the direction of the starting angle) of type
 * angleType that is allowed based on peptide geometry (Ramachandran plots and omega angle
 * restrictions) for the residue with index resIndex. angleType must be one of ANGLE_TYPE_PHI,
 * ANGLE_TYPE_PSI, ANGLE_TYPE_OMEGA. numRes is the number of residues in the peptide.
 * It is assumed that desiredAngle is in the range from -180.0 to 180.0.
 *
 * original_phi and original_psi are the values of the phi and psi angles before any CCD
 * adjustments were made. These values are used here only for the purpose of determining the
 * original Rama quality score for the residue. If angleType is ANGLE_TYPE_OMEGA, original_phi
 * and original_psi are ignored.
 *
 * maxRamaQualityDecrease is the maximum amount by which the Rama score for the new angle is
 * allowed to decrease compared to the original Rama score of originalAngle. For example, suppose
 * that maxRamaQualityDecrease is 1, and the orignal phi angle has a Rama score of 3 (core).
 * The algorithm may then choose a new phi angle with a Rama score of 2 (allowed), but not 1
 * (generous). Regardless of the value of maxRamaQualityDecrease, an angle will never be
 * chosen with a Rama score of 0 (disallowed) unless there are no possible angles with a
 * higher Rama score. When maxRamaQualityDecrease > 0, this function will choose the allowed
 * angle closest to desiredAngle even if the Rama score is worse than the original Rama score.
 * Use a maxRamaQualityDecrease value of 0 to ensure that the Rama score of an angle never
 * becomes worse.
 */
double getClosestAllowedAngle(ResData *residues, int resIndex, int numRes, double desiredAngle, int angleType,
	double original_phi, double original_psi, int maxRamaQualityDecrease) {
	
	int i;
	double phi = 999.0, psi = 999.0, omega = 999.0;
	double finalAngle; // The final value of the angle of type angleType that we will decide to use.
	int phiindex, psiindex, desAngIndex; // Index into ramaTableGeneral for phi, psi, and desiredAngle.
	int isPrePro;
	
	int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
	int original_rama_score; // The Rama score for the original angle of residue resIndex.
	int rama_score_from_desired_angle; // The Rama score that would result if we used desiredAngle as given.
	int target_rama_score; // The Rama score that we hope to achieve (or do better than).
	int best_rama_score_so_far; // The Rama score for best_index_so_far.
	
	int index_increment; // -1 or +1, depending on which direction we want to scan.
	int best_index_so_far; // Index (phiindex or psiindex) that gives the best phi or psi angle so far.
	int elapsed_iter; // Number of scanning iterations that have elapsed.
	
	if (maxRamaQualityDecrease < 0) {
		fprintf(stderr, "ERROR: parameter maxRamaQualityDecrease must be >= 0 for getClosestAllowedAngle() in rama.c\n");
		exit(1);
	}
	
	if (angleType == ANGLE_TYPE_OMEGA) {
		
		omega = residues[resIndex].omega;
		
		if (omega < -180.0) {
			omega += 360.0;
		}
		else if (omega > 180.0) {
			omega -= 360.0;
		}
		
		if (omega >= 90.0 || omega <= -90.0) {
			// Original omega is closer to 180 degrees, so the final omega should be
			// located there.
			
			if (desiredAngle < 170.0 && desiredAngle >= 0.0) {
				finalAngle = 170.0 + ANGLE_SNAP_MARGIN;
			}
			else if (desiredAngle > -170.0 && desiredAngle <= -0.0) {
				finalAngle = -170.0 - ANGLE_SNAP_MARGIN;
			}
			else {
				finalAngle = desiredAngle;
			}
		}
		else { // omega < 90.0 && omega > -90.0
			// Original omega is closer to 0 degrees, so the final omega should be
			// located there.
			
			if (desiredAngle > 10.0) {
				finalAngle = 10.0 - ANGLE_SNAP_MARGIN;
			}
			else if (desiredAngle < -10.0) {
				finalAngle = -10.0 + ANGLE_SNAP_MARGIN;
			}
			else {
				finalAngle = desiredAngle;
			}
		}
		
	}
	else { // angleType == ANGLE_TYPE_PHI || angleType == ANGLE_TYPE_PSI
		phi = residues[resIndex].phi;
		psi = residues[resIndex].psi;
		
		phiindex = getPhiIndex(phi);
		psiindex = getPsiIndex(psi);
		
		//original_rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, psiindex);
		if ( (resIndex+1 != numRes) && (strcmp(residues[resIndex+1].resName, "PRO") == 0) ) {
			isPrePro = TRUE;
		}
		else {
			isPrePro = FALSE;
		}
		original_rama_score = lookupRamaScoreAbstract(residues[resIndex].resName, original_phi, original_psi, isPrePro);
		
		if (angleType == ANGLE_TYPE_PHI) {
			desAngIndex = getPhiIndex(desiredAngle);
			rama_score_from_desired_angle = lookupRamaScoreByIndices(residues, resIndex, numRes, desAngIndex, psiindex);
		}
		else { // angleType == ANGLE_TYPE_PSI
			desAngIndex = getPsiIndex(desiredAngle);
			rama_score_from_desired_angle = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, desAngIndex);
		}
		
		rama_score = rama_score_from_desired_angle;
		
		
		if (rama_score_from_desired_angle >= 2) {
			if (original_rama_score == 3) {
				if (rama_score_from_desired_angle == 3) {
					return desiredAngle;
				}
				else { // rama_score_from_desired_angle == 2
					if (maxRamaQualityDecrease >= 1) {
						return desiredAngle;
					}
					else {
						// move to 3
						target_rama_score = 3;
						//minimum_acceptable_rama_score = 3;
					}
				}
			}
			else { // original_rama_score <= 2
				return desiredAngle;
			}
		}
		else {
			if (original_rama_score == 3) {
				// move to 3
				target_rama_score = 3;
			}
			else if (original_rama_score == 2) {
				// move to 2
				target_rama_score = 2;
			}
			else if (original_rama_score == 1) {
				// try to move to 2, but settle for 1
				target_rama_score = 2;
			}
			else { // original_rama_score == 0
				// try to move to 2, but settle for 0
				target_rama_score = 2;
			}
			
			target_rama_score = target_rama_score - maxRamaQualityDecrease;
			if (target_rama_score < 1) {
				target_rama_score = 1;
			}
			
			if (target_rama_score == 1 && rama_score_from_desired_angle == 1) {
				return desiredAngle;
			}
		}
		
		
		// Which direction should we scan to move from desiredAngle towards the
		// original phi or psi?
		// Note that the index_increment values for phi and psi differ!
		if (angleType == ANGLE_TYPE_PHI) {
			if (desiredAngle - phi >= 0.0) {
				if (desiredAngle - phi < 180.0) {
					index_increment = -1;
				}
				else {
					index_increment = 1;
				}
			}
			else {
				if (desiredAngle - phi > -180.0) {
					index_increment = 1;
				}
				else {
					index_increment = -1;
				}
			}
		}
		else { // angleType == ANGLE_TYPE_PSI
			if (desiredAngle - psi >= 0.0) {
				if (desiredAngle - psi < 180.0) {
					index_increment = 1;
				}
				else {
					index_increment = -1;
				}
			}
			else {
				if (desiredAngle - psi > -180.0) {
					index_increment = -1;
				}
				else {
					index_increment = 1;
				}
			}
		}
		
		
		elapsed_iter = 0;
		best_index_so_far = desAngIndex;
		best_rama_score_so_far = rama_score_from_desired_angle;
		for (i = desAngIndex; elapsed_iter < RAMA_LEN/2;
				i = (i + index_increment + RAMA_LEN) % RAMA_LEN) {
				// Scan from desAngIndex towards phiindex (in the direction)
			
			// Do not scan beyond the original angle.
			if ( (angleType == ANGLE_TYPE_PHI && i == phiindex) || (angleType == ANGLE_TYPE_PSI && i == psiindex) ) {
				if (original_rama_score > best_rama_score_so_far) {
					best_rama_score_so_far = original_rama_score;
					best_index_so_far = i;
				}
				break;
			}
			
			if (angleType == ANGLE_TYPE_PHI) {
				rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, i, psiindex);
			}
			else { // angleType == ANGLE_TYPE_PSI
				rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, i);
			}
			
			if (rama_score >= target_rama_score) {
				// This is the first time we have encountered a tile (sector of the Rama table)
				// with a Rama score that meets or exceeds our target score. We will therefore
				// place our angle in this tile.
				best_rama_score_so_far = rama_score;
				best_index_so_far = i;
				break;
			}
			
			// At this point we have not yet found our target, but we keep track of
			// the best tile we've been in so far. Note that we keep track of only the
			// first instance of a tile with this score that we find.
			if (rama_score > best_rama_score_so_far) {
				best_rama_score_so_far = rama_score;
				best_index_so_far = i;
			}
			
			elapsed_iter++;
		}
		
		// We now know that we should place our angle in the tile with index best_index_so_far.
		
		if (best_index_so_far == desAngIndex) {
			// The proposed desiredAngle is in a good sector already, so we will use the
			// angle without modifying it.
			finalAngle = desiredAngle;
		}
		else {
			// We will modify the given desiredAngle by moving it into the tile with index
			// best_index_so_far. Upon which side of that tile should we place the angle?
			
			if (angleType == ANGLE_TYPE_PHI) {
				finalAngle = (((double) (best_index_so_far*10)) - 180.0);
					// finalAngle is now on the left (more negative) edge of the tile.
			}
			else { // angleType == ANGLE_TYPE_PSI
				finalAngle = (((double) (best_index_so_far*-10)) + 180.0);
					// finalAngle is now on the upper (more positive) edge of the tile.
			}
			
			if (index_increment == 1) {
				// We have been approaching from the negative direction (phi) or positive
				// direction (psi), so finalAngle is on the correct side. We will nudge
				// it inward slightly to prevent possible rounding error that would place
				// it outside the tile.
				
				if (angleType == ANGLE_TYPE_PHI) {
					finalAngle += ANGLE_SNAP_MARGIN;
				}
				else { // angleType == ANGLE_TYPE_PSI
					finalAngle -= ANGLE_SNAP_MARGIN;
				}
			}
			else { // index_increment == -1
				// We have been approaching from the positive direction (phi) or negative
				// direction (psi), so move finalAngle to the other side of the tile. We
				// will nudge it inward slightly because the value directly on the right
				// (positive side) (phi) or lower side (negative side) (psi) boundary
				// actually belongs to another tile, as well as to prevent possible
				// rounding error that would place it outside the tile.
				
				if (angleType == ANGLE_TYPE_PHI) {
					finalAngle += 10.0; // Each tile is 10.0 degrees on each side.
					finalAngle -= ANGLE_SNAP_MARGIN;
				}
				else { // angleType == ANGLE_TYPE_PSI
					finalAngle -= 10.0; // Each tile is 10.0 degrees on each side.
					finalAngle += ANGLE_SNAP_MARGIN;
				}
			}
			
		}
		
	}
	
	
	// Test to make sure that finalAngle is not further away from desiredAngle than the starting
	// angle. This anomaly may arise because finalAngle is adjusted by ANGLE_SNAP_MARGIN, which
	// may move finalAngle too far away.
	if (angleType == ANGLE_TYPE_PHI) {
		if ( angleDist(desiredAngle, finalAngle) > angleDist(desiredAngle, phi) ) {
			finalAngle = phi;
		}
	}
	else if (angleType == ANGLE_TYPE_PSI){
		if ( angleDist(desiredAngle, finalAngle) > angleDist(desiredAngle, psi) ) {
			finalAngle = psi;
		}
	}
	else { // angleType == ANGLE_TYPE_OMEGA
		if ( angleDist(desiredAngle, finalAngle) > angleDist(desiredAngle, omega) ) {
			finalAngle = omega;
		}
	}
	
	
	// debug
	/*
	if (finalAngle != desiredAngle) {
		if (angleType == ANGLE_TYPE_PHI) {
			printf("Constraints changed phi: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, phi);
		}
		else if (angleType == ANGLE_TYPE_PSI) {
			printf("Constraints changed psi: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, psi);
		}
		else { // angleType == ANGLE_TYPE_OMEGA
			printf("Constraints changed omega: %lf --> %lf (original = %lf)\n", desiredAngle, finalAngle, omega);
		}
	}
	*/
	
	
	return finalAngle;
	
} // End getClosestAllowedAngle

/*
Determines the closest allowed pair of phi and psi angles of the residue with index resIndex. The algorithm
starts with the phi and psi angles of residues[resIndex], moves diagonally (or orthogonally) along the
appropriate Ramachandran table to the closest allowed phi and psi pair, and writes those phi and psi angles
(in degrees) to the ints pointed to by phi_ptr and psi_ptr. numRes is the number of residues in the peptide.
*/
void findClosestAllowedPhiPsiPair(ResData *residues, int resIndex, int numRes, int min_accpetable_rama_score, double *phi_ptr, double *psi_ptr) {
	
	double original_phi, original_psi;
	double phi, psi;
	int phiindex, psiindex;
	int original_phiindex, original_psiindex;
	int rama_score;
	double dist_phi, dist_psi;
	double curr_dist;
	double best_dist = 1000000.0; // Closest diagonal (or orthogonal) distance from original phi and psi to best_phi, best_psi.
	double best_phi = 999.0;
	double best_psi = 999.0;
		// The phi and psi closest to the original phi and psi but not on the same tile.
		// Initialized to dummy values. As long as each Ramachandran table has at least one tile with a score greater
		// than or equal to min_accpetable_rama_score, these dummy values are guaranteed to be overwritten.
	
	original_phi = residues[resIndex].phi;
	original_psi = residues[resIndex].psi;
	
	original_phiindex = getPhiIndex(original_phi);
	original_psiindex = getPsiIndex(original_psi);
	
	if (lookupRamaScoreByIndices(residues, resIndex, numRes, original_phiindex, original_psiindex) >= min_accpetable_rama_score) {
		// The existing phi and psi angles are already acceptable.
		*phi_ptr = original_phi;
		*psi_ptr = original_psi;
		return;
	}
	
	// Calculate the distance between the starting phi and psi and the closest point on every single tile to find
	// the tile that is closest and has an acceptable Rama score. (Yes, this algorithm is a bit crude).
	for (psiindex = 0; psiindex < RAMA_LEN; psiindex++) {
		for (phiindex = 0; phiindex < RAMA_LEN; phiindex++) {
			
			if ( !(phiindex == original_phiindex && psiindex == original_psiindex) ) {
				
				rama_score = lookupRamaScoreByIndices(residues, resIndex, numRes, phiindex, psiindex);
				
				if (rama_score >= min_accpetable_rama_score) {
					
					// Compute the distance from the original phi and psi to the current phi and psi.
					
					// First determine where on the tile we should locate the current phi and psi
					// so that they are as close as possible to the originals.
					
					// Which side of the tile is closer to phi?
					if (phiindex == original_phiindex) {
						// Both tiles are in the same column.
						phi = original_phi;
					}
					else {
						phi = (((double) (phiindex*10)) - 180.0) + 5.0;
							// phi is now on the left (more negative) edge of the tile.
						
						if (original_phi - phi >= 0.0) {
							if (original_phi - phi < 180.0) {
								// Right side of tile is closer.
								phi += (5.0 - ANGLE_SNAP_MARGIN);
							}
							else {
								// Left side of tile is closer.
								phi -= (5.0 - ANGLE_SNAP_MARGIN);
							}
						}
						else {
							if (original_phi - phi > -180.0) {
								// Left side of tile is closer.
								phi -= (5.0 - ANGLE_SNAP_MARGIN);
							}
							else {
								// Right side of tile is closer.
								phi += (5.0 - ANGLE_SNAP_MARGIN);
							}
						}
					}
					
					// Which side of the tile is closer to psi?
					// Note that the calculation for psi differs from that of phi, because of
					// the layout of ramaTableGeneral etc.!
					if (psiindex == original_psiindex) {
						// Both tiles are in the same row.
						psi = original_psi;
					}
					else {
						psi = (((double) (psiindex*-10)) + 180.0) - 5.0;
						// psi is now on the upper (more positive) edge of the tile.
						
						if (original_psi - psi >= 0.0) {
							if (original_psi - psi < 180.0) {
								// Top side of tile is closer.
								psi += (5.0 - ANGLE_SNAP_MARGIN);
							}
							else {
								// Bottom side of tile is closer.
								psi -= (5.0 - ANGLE_SNAP_MARGIN);
							}
						}
						else {
							if (original_psi - psi > -180.0) {
								// Bottom side of tile is closer.
								psi -= (5.0 - ANGLE_SNAP_MARGIN);
							}
							else {
								// Top side of tile is closer.
								psi += (5.0 - ANGLE_SNAP_MARGIN);
							}
						}
					}
					
					// Compute the distance.
					
					dist_phi = angleDist(original_phi, phi);
					dist_psi = angleDist(original_psi, psi);
					
					curr_dist = sqrt(dist_phi*dist_phi + dist_psi*dist_psi);
					
					// Compare it with the closest distance found so far.
					
					if (curr_dist < best_dist) {
						best_dist = curr_dist;
						best_phi = phi;
						best_psi = psi;
					}
					
				}
				
			}
			
		}
	}
	
	*phi_ptr = best_phi;
	*psi_ptr = best_psi;
	
}


/*
 * Returns a random angle uniformly distributed in the range [angle_min, angle_max),
 * where the angles are in degrees.
 */
double getRandomAngleInInterval(double angle_min, double angle_max) {
	double temp_angle;
	double range = angleDist(angle_min, angle_max);
	
	temp_angle = angle_min + range*drand48();
	
	// Put temp_angle in the range [-180.0, 180.0]
	if (temp_angle > 180.0) {
		temp_angle -= 360.0;
	}
	
	return temp_angle;
}
