#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "template.h"
#include "util.h"
#include "vec3D.h"
#include "pdbfile.h"
#include "pepmake.h"
#include "rama.h"


/*
 * Reads the best matching template from the database, for the given Residue.
 * 
 * This function sets the 'template', 'names', and 'max_atom_index' variables of
 * the given ResData struct based on its 'resName' and 'chi' variables.
 *
 * If isFirstResidue is TRUE, N-terminal N and H atoms will be added. If
 * isLastResidue is TRUE, a C-terminal O atom (" OXT") will be added.
 */
void readTemplate(ResData *res, int ignoreBackbone, int isFirstResidue, int isLastResidue) {
	int n, i = 0, j, read;
	FILE *fp;
	double chi;
	double N_CA_C1_angle;
	char fname[64];
	char name[5];
	char resName[4];
	double x,y,z;
	int best = 1000;
	char line[64];
	int atom_index;
	int readN, readCA, readCG;
	int debug = FALSE;
	//debug = (strcmp("PRO", res->resName) == 0);
	//debug = (strcmp("ALA", res->resName) == 0);
	
	// initialize
	for (i = 0; i < MAX_ATOMS; i++) {
		setVector(res->template[i], ATOM_COORD_NA, ATOM_COORD_NA, ATOM_COORD_NA);
		res->names[i][0] = (char) NULL;
		res->fixed_position = FALSE;
	}
	
	
	sprintf(fname, "%s%s%s", "template_database/", res->resName, ".db");
	if (debug) printf("CHI = %lf\n", res->chi);
	if ((fp = fopen(fname, "r")) != NULL) {
		int r = 2;
		while (!feof(fp) && r >= 2) {
			
			//printf("***start loop %d\n", n);
			
			// hack to handle two formats
			r = fscanf(fp, "%lf %d", &chi, &n);
			if (r != 2) {
				if (fgets(line, 64, fp) != NULL) {
					r = sscanf(line, "%s %lf %lf %d", resName, &chi, &N_CA_C1_angle, &n);
				}
				//printf("resName is %s\n", resName);
			}
			if (r >= 2) {		
				if (debug) printf("chi=%lf, n=%d\n", chi, n);
				read = FALSE;
				if (angleDist(chi, res->chi) < best) {
					best = angleDist(chi, res->chi);
					read = TRUE;
					res->max_atom_index = 0; // Initialize this here; it will be adjusted below.	
					if (debug) printf("Reading! %d\n", n);
				}
				
				j = CG;
				readN = FALSE;
				readCA = FALSE;
				readCG = FALSE;
				for (i=0; i<n; i++) {
					
					if (fgets(line, 64, fp) != NULL) {
						strncpy(name, &(line[1]), 4);
							// On each line the first char is a space, then next 4 are the atom name.
						name[4] = (char) NULL;
						
						if (read) {
							sscanf(&(line[5]), "%lf%lf%lf", &x, &y, &z);
							
							atom_index = getAtomIndexFromAtomName(name);
							
							if (atom_index == CG) {
								if (readCG == FALSE) {
									readCG = TRUE;
								}
								else {
									// Only the first gamma atom can occupy index CG.
									atom_index = -1;
								}
							}
							
							if (atom_index == -1) {
								j++;
								atom_index = j;
							}
							else if (atom_index == Np) {
								if (readN == FALSE) {
									readN = TRUE;
								}
								else {
									atom_index = N;
								}
							}
							else if (atom_index == CA) {
								if (readCA == FALSE) {
									readCA = TRUE;
								}
								else {
									atom_index = CA2;
								}
							}
							
							strncpy(res->names[atom_index], name, 5);
							res->template[atom_index][0] = x;
							res->template[atom_index][1] = y;
							res->template[atom_index][2] = z;
							
							// Adjust res->max_atom_index.
							if (atom_index > res->max_atom_index) {
								res->max_atom_index = atom_index;
							}
						}
					}
					
				}
			}
		}
		fclose(fp);
	} else {
		fprintf(stderr, "Cannot open %s\n", fname);
	}
	
	// Adjust the residue template so that the psi angle is the standard 180 degrees expected by the
	// rotatePhiPsi function. This is needed because the rotatePhiPsi function was reimplemented by
	// David Arndt in October 2006.
	//   -- Added by David Arndt
	// Further note: David Arndt regenerated the template database with an updated format, so now
	// this adjustment is not needed.
	/*
	//outputDebugCoords("coords_for_Grapher/template_before_adjust.txt", res->template, TRUE, FALSE);
	double assumed_N_CA_CB_diff = 120;
		// Assumed angle from the N-CA-CB bonds projected onto the plane normal to the CA_C1 axis.
		// This means it should ideally be 120 degrees, not 109.5 degrees.
	double CB_torsion = torsion(res->template[CB], res->template[CA], res->template[C1], res->template[N]);
	// Rotate around CA-C1 bond to set psi to 180 degrees. -- Added by David Arndt.
	double psi = CB_torsion - assumed_N_CA_CB_diff;
		// Since we don't know exactly where the previous N atom should be (i.e. we don't know exactly what
		// psi angle is reflected in the residue template), we guess based on the location of the CB atom.
	for (i=Np; i<n; i++) {
		rotateX(res->template[i], 180+psi);
	}
	//outputDebugCoords("coords_for_Grapher/template_after_adjust.txt", res->template, TRUE, FALSE);
	*/
	// End adjust residue template.
	
	if (ignoreBackbone) {
		/* use a generic template rather than the database for 
			the main chain atoms */
		setVector(res->template[CA], 0.0, 0.0, 0.0);
		setVector(res->template[C1], 1.53, 0.0, 0.0);
		setVector(res->template[O], 2.1686, 1.0629, 0.0);
		setVector(res->template[N], 2.0669, -1.2059, 0.0);
		setVector(res->template[CA2], 3.5188, -1.4359, 0.0);
	}
	
	
	// For the first residue, copy in default locations for Np, HNp, HNp2, HNp3.
	// These values were taken by David Arndt from the protein 3TRX, which has an
	// N-terminal MET residue. From informal observations, these values appear to be
	// similar to those of other N-terminal residues, though others have shorter N-CA
	// bond lengths. The data below result in an N-CA bond length of about 1.491,
	// whereas the generic backbone template given above has a N-CA bond length of 1.470,
	// which is closer to the N-CA bond length for other N-terminal residues that were
	// observed.
	if (isFirstResidue) {
		strncpy(res->names[HNp], "1H  ", 5);
		strncpy(res->names[HNp2], "2H  ", 5);
		strncpy(res->names[HNp3], "3H  ", 5);
		setVector(res->template[Np], -0.496989, 1.405916, 0.000000);
		setVector(res->template[HNp], 0.204055, 2.021366, -0.460723);
		setVector(res->template[HNp2], -1.396522, 1.455698, -0.519867);
		setVector(res->template[HNp3], -0.644574, 1.721908, 0.980245);
	}
	
	// For the last residue, copy in an OXT atom. This is done in a semi-hackish way by putting
	// it after all the other residues. Source data were taken by David Arndt from 2spz_a, but perhaps
	// better data could be used.
	if (isLastResidue) {
		setVector(res->template[res->max_atom_index + 1], 2.107932, -1.090784, -0.000000);
		strncpy(res->names[res->max_atom_index + 1], " OXT", 5);
		(res->max_atom_index)++;
	}
	
}
