/**************************************************************************

  sfassembler.c
     Subfragment assembly of protein structures. Generates a PDB file
	 from multiple PDB file subfragments and Preditor angle predictions.

  Date: 
     March, 2008

  Author:
     David Arndt

 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <ctype.h>

#include "sfassembler.h"
#include "util.h"
#include "pepmake.h"
#include "rama.h"
#include "preditor_parser.h"
#include "rotamer.h"
#include "template.h"


/* Function prototypes */

static void assemble_subfragments(char *sequence, int numRes, ANGLE_INFO *angles_preditor, ResData **subfragments,
		int *subfrag_lengths, char **subfrag_filenames, int num_subfragments, int print_alignment,
		ResData *assembled_structure);
static void rank_subfragments(ResData **subfragments, int *subfrag_lengths, int num_subfragments, int *ranked_subfrags);
static void setResDataFromAngleInfo(ResData *res, ANGLE_INFO *angle_info, int numRes, int j);
static void set_C_terminal_psi(ResData *res, ResData *temp);
static void print_alignment_info(char **subfrag_filenames, char **alignment_strings, int num_subfragments,
		char *sequence, int numRes, int havePreditorFile, int preditor_aln_index);
static void load_PDB_files(char *pdb_files_list_filename, ResData ***subfragments, int **subfrag_lengths,
		char ***subfrag_filenames, int *num_subfrags);
static void load_PDB_file(char *inputPDBFilename, ResData **residues, int *numRes);
static int parse_FASTA_file(char *fasta_filename, char **sequence);
static void align_preditor_info_to_target_sequence(char *sequence, int numRes, ANGLE_INFO *angles_preditor,
		int num_resd_preditor, ANGLE_INFO *angles_preditor_aligned);
static void usage();


/*
 * Assembles the subfragments and Preditor angle information (if available) into a single structure.
 * Residues for which no coordinates are available are modeled based on residue templates. Default angle
 * values are supplied where they are absent from the PDB files and the Preditor file. This function
 * does not build the structure, but it sets the data in 'assembled_structure' so that it can be
 * built by a call to buildPeptide().
 *
 * sequence: Final desired sequence of the peptide, in single-residue amino acid abbreviations plus
 *      null terminator. The sequence must be in uppercase.
 * numRes: The number of residues in the final assembled structure.
 * angles_preditor: Array of numRes ANGLE_INFO structs with angle information parsed from the Preditor
 *      file of predicted angles, mapped onto the final peptide sequence.
 * subfragments: Pointer to an array of protein structures representing the subfragments.
 * subfrag_lengths: Array with the lengths (number of residues) of each subfragment.
 * subfrag_filenames: The names of the PDB files from which the subfragments were loaded.
 * num_subfragments: The number of subfragments.
 * print_alignment: Boolean. Whether to print information about the alignment used to assemble the
 *      final structure.
 * assembled_structure: Pointer to array of numRes ResData structs where this function will write
 *      the data for the final assembled structure. The array must already have been allocated.
 */
static void assemble_subfragments(char *sequence, int numRes, ANGLE_INFO *angles_preditor, ResData **subfragments,
		int *subfrag_lengths, char **subfrag_filenames, int num_subfragments, int print_alignment,
		ResData *assembled_structure) {
	
	int i, j, k;
	int frag_index;
	int *ranked_subfrags;
		// ranked_subfrags[0] is the index into subfragments of the "best" subfragment.
		// ranked_subfrags[num_subfragments-1] is the index into subfragments of the "worst" subfragment.
	char *curr_sequence;
	char **alignment_strings;
		// Strings with info about where each subfragment was aligned an which residues were used in the
		// final assembled structure. Used only for printing a summary of the alignment for the user. An
		// additional string at the end is used for noting where the Preditor file was used to fill in
		// information.
	int preditor_aln_index;
		// Index into alignment_strings of the line with Preditor information.
	ResData *curr_subfragment;
	int numRes_subfrag;
	char *position;
	int start_index; // Index in 'sequence' where the sequence of the current subfragment starts.
	int used_this_res;
		// Boolean. Whether we used information from the current residue in the final assembled structure.
	
	char *resName;
	ResData temp;
	int next_res_is_Pro;
	
	
	curr_sequence = malloc( (numRes+1) * sizeof(char) ); assert(curr_sequence != NULL);
	
	// Allocate memory for the alignment strings.
	alignment_strings = malloc( (num_subfragments + 1) * sizeof(char*) ); assert(alignment_strings != NULL);
	for (i = 0; i < num_subfragments + 1; i++) {
		alignment_strings[i] = malloc( (numRes+1) * sizeof(char) ); assert(alignment_strings[i] != NULL);
		
		// Initialize string to all spaces.
		for (j = 0; j < numRes; j++) {
			alignment_strings[i][j] = ' ';
		}
		alignment_strings[i][numRes] = '\0';
	}
	
	preditor_aln_index = num_subfragments;
	
	/******************************************************************/
	/***                   Rank the subfragments                    ***/
	/******************************************************************/
	
	ranked_subfrags = malloc( num_subfragments * sizeof(int) ); assert(ranked_subfrags != NULL);
	rank_subfragments(subfragments, subfrag_lengths, num_subfragments, ranked_subfrags);
	
	
	/******************************************************************/
	/***  Copy the information from each subfragment, in ranked     ***/
	/***  order, to the final assembled structure.                  ***/
	/******************************************************************/
	
	// Note that we do not remove "1H  ", "2H  ", "3H  " and " OXT" atoms from residues that are
	// not at the N- or C-termini in the assembled structure because printPDBAtoms() does not
	// output these atoms for non-terminal residues anyway.
	
	for (i = 0; i < num_subfragments; i++) {
		
		frag_index = ranked_subfrags[i];
		curr_subfragment = subfragments[frag_index];
		numRes_subfrag = subfrag_lengths[frag_index];
		
		get_seq_from_ResData(curr_subfragment, numRes_subfrag, curr_sequence);
		
		// We assume that the sequence of each subfragment will be a proper substring of the
		// final sequence of the assembled peptide. Find the position of the current subfragment
		// in the final sequence.
		
		position = strstr(sequence, curr_sequence);
		if (position == NULL) {
			fprintf(stderr,
				"ERROR: Sequence of PDB file \'%s\' is not a proper substring of the final sequence.\n",
				subfrag_filenames[frag_index]);
			fprintf(stderr, "Desired final sequence of assembled structure:\n%s\n", sequence);
			fprintf(stderr, "Problematic subfragment sequence:\n%s\n", curr_sequence);
			exit(1);
		}
		start_index = (int) (position - sequence); // Pointer arithmetic!
		if (start_index < 0) {start_index = -1 * start_index;}
		
		//fprintf(stderr, "start_index is %d\n", start_index); // debug
		
		
		// Copy the information for the current subfragment to the assembled structure.
		// All residue IDs of the assembled structure should have been initialized to
		// RES_ID_NA, which we use here to indicate that the residue information has
		// not yet been set. If a residue's information has been set with a previous
		// "better" subfragment, we do not want to overwrite it.
		
		for (j = start_index; j < start_index + numRes_subfrag; j++) {
			
			used_this_res = FALSE;
			
			if (assembled_structure[j].id == RES_ID_NA) {
				copySingleResData(&curr_subfragment[j - start_index], &assembled_structure[j], FALSE);
				
				if (j == start_index) {
					// Since the residue is from the first residue of the current subfragment, the phi
					// angle is unavailable. Therefore mark the phi angle as unavailable.
					// (The phi angle of the first residue of the assembled structure will always be
					// unavailable).
					
					if (j != 0) {
						assembled_structure[j].phi = ANGLE_NA;
					}
				}
				
				if (j == start_index + numRes_subfrag - 1) {
					// Since the residue is from the last residue of the current subfragment, the psi
					// and omega angles are unavailable. Therefore mark them as unavailable.
					// (The psi and omega angles of the last residue of the assembled structure will
					// always be unavailable). Also mark the HN atom (which really belongs to the
					// next residue) as unavailable.
					
					if (j != numRes - 1) {
						assembled_structure[j].psi = ANGLE_NA;
						assembled_structure[j].omega = ANGLE_NA;
						
						setVector(assembled_structure[j].template[HN],
							ATOM_COORD_NA, ATOM_COORD_NA, ATOM_COORD_NA);
					}
				}
				
				alignment_strings[frag_index][j] = sequence[j];
			}
			else {
				// The current residue has already been set from another subfragment, but its phi,
				// psi or omega angles may be missing. If we can set the phi or psi angle based on the
				// current subfragment, we will do so. If the psi angle is replaced, some backbone
				// atoms that depend on the psi angle (and C1) are also replaced.
				
				// Check phi angle.
				if (!(j == start_index) && (j != 0) &&
						assembled_structure[j].phi == ANGLE_NA) {
					
					assembled_structure[j].phi = curr_subfragment[j - start_index].phi;
					used_this_res = TRUE;
				}
				
				// Check psi and omega angles.
				if ( !(j == start_index + numRes_subfrag - 1) && (j != numRes - 1) ) {
					
					if (assembled_structure[j].psi == ANGLE_NA) {
						assembled_structure[j].psi = curr_subfragment[j - start_index].psi;
						
						for (k = C1; k <= HN; k++) {
							copyVector(curr_subfragment[j - start_index].template[k], assembled_structure[j].template[k]);
							strncpy(assembled_structure[j].names[k], getAtomNameFromAtomIndex(k), 4);
						}
						
						used_this_res = TRUE;
					}
					
					if (assembled_structure[j].omega == ANGLE_NA) {
						assembled_structure[j].omega = curr_subfragment[j - start_index].omega;
						used_this_res = TRUE;
					}
				}
				
				if (used_this_res) {
					alignment_strings[frag_index][j] = sequence[j];
				}
				else {
					alignment_strings[frag_index][j] = (char) tolower(sequence[j]);
				}
			}
		}
		
		// debug
		//printf("After fragment %d (%d in file):\n", i, frag_index);
		//for (j = 0; j < numRes; j++) {
		//	printf("% 3d: %d\n", j, assembled_structure[j].id);
		//}
		
	}
	
	
	/******************************************************************/
	/***  Fill in any missing residues or angles from the Preditor  ***/
	/***  file (if it is available).                                ***/
	/******************************************************************/
	
	if (angles_preditor != NULL) {
		
		for (j = 0; j < numRes; j++) {
			
			used_this_res = FALSE;
			
			if (angles_preditor[j].res_no != RES_ID_NA) {
				// We have Preditor information mapped to the current residue.
				
				if (assembled_structure[j].id == RES_ID_NA) {
					// The residue is missing from all subfragments; load the default residue template based
					// on the residue name and chi angle from the Preditor file.
					
					setResDataFromAngleInfo(&assembled_structure[j], &angles_preditor[j], numRes, j);
					readTemplate(&assembled_structure[j], ignoreDBBackbone, (j == 0), (j == numRes - 1));
					
					// Mark the N-CA-C1 angle as "not available"; we will set it below.
					assembled_structure[j].N_CA_C1_angle = ANGLE_NA;
					
					alignment_strings[preditor_aln_index][j] = angles_preditor[j].res;
				}
				else {
					// The current residue has already been set from the subfragments, but its phi, psi
					// or omega angles may be missing. Set the phi, psi and omega angles based on the values
					// from the Preditor file. If the psi angle is replaced, some backbone atoms that
					// depend on the psi angle (and C1) are also replaced.
					
					// Check phi angle.
					if ((j != 0) && (assembled_structure[j].phi == ANGLE_NA) &&
							(angles_preditor[j].phi != PREDITOR_ANGLE_NA) ) {
						assembled_structure[j].phi = angles_preditor[j].phi;
						used_this_res = TRUE;
					}
					
					// Check psi and omega angles.
					if ( j != numRes - 1 ) {
						
						if ( (assembled_structure[j].psi == ANGLE_NA) && (angles_preditor[j].psi != PREDITOR_ANGLE_NA) ) {
							assembled_structure[j].psi = angles_preditor[j].psi;
							
							// Load the default residue template and replace some backbone atoms based on the residue
							// template.
							
							setResDataFromAngleInfo(&temp, &angles_preditor[j], numRes, j);
							readTemplate(&temp, ignoreDBBackbone, (j == 0), (j == numRes - 1));
							
							for (k = C1; k <= HN; k++) {
								copyVector(temp.template[k], assembled_structure[j].template[k]);
								strncpy(assembled_structure[j].names[k], getAtomNameFromAtomIndex(k), 4);
							}
							
							used_this_res = TRUE;
						}
						
						if ( (assembled_structure[j].omega == ANGLE_NA) && (angles_preditor[j].omega != PREDITOR_ANGLE_NA) ) {
							assembled_structure[j].omega = angles_preditor[j].omega;
							used_this_res = TRUE;
						}
					}
					
					if (used_this_res) {
						alignment_strings[preditor_aln_index][j] = angles_preditor[j].res;
					}
					else {
						alignment_strings[preditor_aln_index][j] = (char) tolower(angles_preditor[j].res);
					}
				}
			}
		}
		
	}
	
	
	/******************************************************************/
	/***  Fill in any missing residues or angles based on default   ***/
	/***  values.                                                   ***/
	/******************************************************************/
	
	for (j = 0; j < numRes; j++) {
		
		if (assembled_structure[j].id == RES_ID_NA) {
			// The residue is missing from all subfragments; load the default residue template based
			// on the residue name in the full sequence.
			
			resName = char_to_res_str(sequence[j]);
			if (strcmp(resName, "FIL") == 0) {
				fprintf(stderr, "ERROR: Invalid residue \'%c\' in FASTA file.\n", sequence[j]);
				exit(1);
			}
			strncpy(assembled_structure[j].resName, resName, 4);
			
			assembled_structure[j].id = j;
			if (j == 0) {
				assembled_structure[j].phi = 0.0;
			}
			else {
				assembled_structure[j].phi = SFA_DEFAULT_PHI;
			}
			assembled_structure[j].psi = SFA_DEFAULT_PSI;
			if (j == numRes - 1) {
				assembled_structure[j].omega = 0.0;
			}
			else {
				assembled_structure[j].omega = SFA_DEFAULT_OMEGA;
			}
			assembled_structure[j].chi = getMostProbableChi(resName, SFA_DEFAULT_PHI, SFA_DEFAULT_PSI);
			
			if ((j + 1 < numRes) && strcmp(char_to_res_str(sequence[j+1]), "PRO") == 0) {
				next_res_is_Pro = TRUE;
			}
			else {
				next_res_is_Pro = FALSE;
			}
			predict_N_CA_C1_angle(&assembled_structure[j], next_res_is_Pro);
			
			readTemplate(&assembled_structure[j], ignoreDBBackbone, (j == 0), (j == numRes - 1));
			
			if (j == numRes - 1) {
				set_C_terminal_psi(&assembled_structure[j], &assembled_structure[j]);
			}
			
			// debug
			//printf("%3s (%d) %lf %lf\n", assembled_structure[j].resName, assembled_structure[j].id,
			//	assembled_structure[j].chi, assembled_structure[j].N_CA_C1_angle);
			//for (k = 0; k <= assembled_structure[j].max_atom_index; k++) {
			//	printf(" %4s % 11.6lf % 11.6lf % 11.6lf\n", assembled_structure[j].names[k],
			//		assembled_structure[j].template[k][0], assembled_structure[j].template[k][1],
			//		assembled_structure[j].template[k][2]);
			//}
			//printf("\n");
		}
		else {
			// The current residue has already been set from the subfragments or Preditor file,
			// but its phi, psi or omega angles may be missing. Set the phi, psi and omega angles
			// based on the default values. If the psi angle is replaced, some backbone atoms that
			// depend on the psi angle (and C1) are also replaced. If the N-CA-C1 angle is not
			// available (which must be because this residue is based on information from the
			// Preditor file), then the N-CA-C1 angle will be set.
			
			// Check phi angle.
			if (assembled_structure[j].phi == ANGLE_NA) {
				if (j == 0) {
					assembled_structure[j].phi = 0.0;
				}
				else {
					assembled_structure[j].phi = SFA_DEFAULT_PHI;
				}
			}
			
			// Check omega angle.
			if (assembled_structure[j].omega == ANGLE_NA) {
				if (j == numRes - 1) {
					assembled_structure[j].omega = 0.0;
				}
				else {
					assembled_structure[j].omega = SFA_DEFAULT_OMEGA;
				}
			}
			
			// Check psi angle.
			if (assembled_structure[j].psi == ANGLE_NA) {
				
				// Load the default residue template and replace some backbone atoms based on the residue
				// template.
				
				resName = char_to_res_str(sequence[j]);
				if (strcmp(resName, "FIL") == 0) {
					fprintf(stderr, "ERROR: Invalid residue \'%c\' in FASTA file.\n", sequence[j]);
					exit(1);
				}
				strncpy(temp.resName, resName, 4);
				
				temp.id = j;
				temp.phi = assembled_structure[j].phi;
				temp.psi = SFA_DEFAULT_PSI;
				temp.omega = assembled_structure[j].omega;
				temp.chi = getMostProbableChi(resName, temp.phi, temp.psi);
				readTemplate(&temp, ignoreDBBackbone, (j == 0), (j == numRes - 1));
				
				for (k = C1; k <= HN; k++) {
					copyVector(temp.template[k], assembled_structure[j].template[k]);
					strncpy(assembled_structure[j].names[k], getAtomNameFromAtomIndex(k), 4);
				}
				
				
				if (j == numRes - 1) {
					set_C_terminal_psi(&assembled_structure[j], &temp);
				}
				else {
					assembled_structure[j].psi = SFA_DEFAULT_PSI;
				}
				
				
			}
			
			// Set the N-CA-C1 angle.
			if (assembled_structure[j].N_CA_C1_angle == ANGLE_NA) {	
				if ((j + 1 < numRes) && strcmp(char_to_res_str(sequence[j+1]), "PRO") == 0) {
					next_res_is_Pro = TRUE;
				}
				else {
					next_res_is_Pro = FALSE;
				}
				predict_N_CA_C1_angle(&assembled_structure[j], next_res_is_Pro);
			}
		}
		
	}
	
	
	/******************************************************************/
	/***                   Renumber the residues                    ***/
	/******************************************************************/
	
	for (j = 0; j < numRes; j++) {
		assembled_structure[j].id = j+1;
	}
	
	
	/******************************************************************/
	/***             Print alignment info for the user.             ***/
	/******************************************************************/
	
	if (print_alignment) {
		print_alignment_info(subfrag_filenames, alignment_strings, num_subfragments,
			sequence, numRes, angles_preditor != NULL, preditor_aln_index);
	}
	
	
	// debug
	/*
	printf("After setting default values:\n");
	for (j = 0; j < numRes; j++) {
		printf("%3s (%d) %lf %lf\n", assembled_structure[j].resName, assembled_structure[j].id,
			assembled_structure[j].chi, assembled_structure[j].N_CA_C1_angle);
		for (k = 0; k <= assembled_structure[j].max_atom_index; k++) {
			printf(" %4s % 11.6lf % 11.6lf % 11.6lf\n", assembled_structure[j].names[k],
				assembled_structure[j].template[k][0], assembled_structure[j].template[k][1],
				assembled_structure[j].template[k][2]);
		}
	}
	
	printf("After setting default values:\n");
	for (j = 0; j < numRes; j++) {
		printf("%3s %d: % 11.6lf % 11.6lf % 11.6lf % 11.6lf\n",
			assembled_structure[j].resName, assembled_structure[j].id,
			assembled_structure[j].phi, assembled_structure[j].psi,
			assembled_structure[j].omega, assembled_structure[j].chi);
	}
	*/
	
	// debug
	//ResData *debug1 = &assembled_structure[69];
	//ResData *debug2 = &assembled_structure[70];
	
	
	/******************************************************************/
	/***                        Free memory                         ***/
	/******************************************************************/
	
	free(curr_sequence);
	
	for (i = 0; i < num_subfragments + 1; i++) {
		free(alignment_strings[i]);
	}
	free(alignment_strings);
	
	free(ranked_subfrags);
}


/*
 * Rank the subfragments in order of their priority for use in the final structure.
 * Currently we simply rank the longest subfragments are the best ones.
 */
static void rank_subfragments(ResData **subfragments, int *subfrag_lengths, int num_subfragments, int *ranked_subfrags) {
	
	int i, j;
	int frag_index;
	int curr_score;
	int best_score;
	int best_index = -1;
	int skip = FALSE;
	
	for (frag_index = 0; frag_index < num_subfragments; frag_index++) {
		ranked_subfrags[frag_index] = -1;
	}
	
	for (i = 0; i < num_subfragments; i++) {
		
		best_score = -1;
		for (frag_index = 0; frag_index < num_subfragments; frag_index++) {
			
			// Skip this subfragment if it has already been placed in the ranking.
			skip = FALSE;
			for (j = 0; j < i; j++) {
				if (frag_index == ranked_subfrags[j]) {
					skip = TRUE;
					break;
				}
			}
			if (skip) {
				continue;
			}
			
			curr_score = subfrag_lengths[frag_index];
			
			if (curr_score > best_score) {
				best_score = curr_score;
				best_index = frag_index;
			}
			
		}
		
		ranked_subfrags[i] = best_index;
	}
	
}


/*
 * Sets the residue ID, name, and angles of the given ResData struct based on the
 * given ANGLE_INFO struct.
 *
 * j: the index of the residue in the peptide.
 * numRes: the number of residues in the peptide.
 */
static void setResDataFromAngleInfo(ResData *res, ANGLE_INFO *angle_info, int numRes, int j) {
	
	char *resName;
	
	resName = char_to_res_str(angle_info->res);
	if (strcmp(resName, "FIL") == 0) {
		fprintf(stderr, "ERROR: Invalid residue \'%c\' in Preditor file.\n", angle_info->res);
		exit(1);
	}
	
	strncpy(res->resName, resName, 4);
	
	res->id = angle_info->res_no;
	
	if (angle_info->phi != PREDITOR_ANGLE_NA && (j != 0)) {
		res->phi = angle_info->phi;
	}
	else {
		res->phi = ANGLE_NA;
	}
	
	if (angle_info->psi != PREDITOR_ANGLE_NA && (j != numRes - 1)) {
		res->psi = angle_info->psi;
	}
	else {
		res->psi = ANGLE_NA;
	}
	
	if (angle_info->omega != PREDITOR_ANGLE_NA && (j != numRes - 1)) {
		res->omega = angle_info->omega;
	}
	else {
		res->omega = ANGLE_NA;
	}
	
	if (angle_info->chi != PREDITOR_ANGLE_NA) {
		res->chi = angle_info->chi;
	}
	else {
		res->chi = ANGLE_NA;
	}
}


/*
 * Sets the psi angle (not a real psi angle) for the given C-terminal residue res based on
 * the template atoms in 'temp'. The 'temp' residue must have an " OXT" atom.
 */
static void set_C_terminal_psi(ResData *res, ResData *temp) {
	
	int k;
	int OXT_index = -1;
	
	// Just like in the phipsi program, we will set the "psi" angle of the last residue
	// using the " OXT" atom location. The " OXT" atom should have been set by readTemplate().
	
	// Look for " OXT" atom in last residue.
	
	for (k = HN; k <= temp->max_atom_index; k++) {
		if (strncmp(temp->names[k], " OXT", 4) == 0) {
			OXT_index = k;
			break;
		}
	}
	if (OXT_index == -1) {
		// This should never happen.
		fprintf(stderr, "ERROR: Failed to load \" OXT\" atom from template database.\n");
		exit(1);
	}
	
	res->psi =
		torsion(temp->template[Np], temp->template[CA], temp->template[C1], temp->template[OXT_index]);
	
}


/*
 * Prints alignment info for the user.
 */
static void print_alignment_info(char **subfrag_filenames, char **alignment_strings, int num_subfragments,
		char *sequence, int numRes, int havePreditorFile, int preditor_aln_index) {
	
	int i;
	int section_num;
	int num_sections = numRes/PRINT_ALIGNMENT_RESIDUES_PER_LINE;
	if (num_sections*PRINT_ALIGNMENT_RESIDUES_PER_LINE < numRes) {
		num_sections++;
	}
	int start;
	char tmp_string[256];
	char tmp_string2[256];
	
	for (section_num = 0; section_num < num_sections; section_num++) {
		
		start = section_num*PRINT_ALIGNMENT_RESIDUES_PER_LINE;
		
		strncpy(tmp_string, "Assembled", PRINT_ALIGNMENT_LABEL_WIDTH);
		strncpy(tmp_string2, &sequence[start], PRINT_ALIGNMENT_RESIDUES_PER_LINE);
		tmp_string2[PRINT_ALIGNMENT_RESIDUES_PER_LINE] = '\0';
		printf("%-*s: %-*s\n",
			PRINT_ALIGNMENT_LABEL_WIDTH, tmp_string,
			PRINT_ALIGNMENT_RESIDUES_PER_LINE, tmp_string2);
		
		for (i = 0; i < num_subfragments + 1; i++) {
			
			if (!havePreditorFile && i == preditor_aln_index) {
				break;
			}
			
			if (havePreditorFile && i == preditor_aln_index) {
				strncpy(tmp_string, "Preditor", PRINT_ALIGNMENT_LABEL_WIDTH);
			}
			else {
				get_PDB_name_from_path(subfrag_filenames[i], tmp_string);
				tmp_string[PRINT_ALIGNMENT_LABEL_WIDTH] = '\0';
			}
			
			strncpy(tmp_string2, &alignment_strings[i][start], PRINT_ALIGNMENT_RESIDUES_PER_LINE);
			tmp_string2[PRINT_ALIGNMENT_RESIDUES_PER_LINE] = '\0';
			printf("%-*s: %-*s\n",
				PRINT_ALIGNMENT_LABEL_WIDTH, tmp_string,
				PRINT_ALIGNMENT_RESIDUES_PER_LINE, tmp_string2);
			
		}
		printf("\n");
		
	}
	
	printf("Uppercase: Used in final assembled structure.\n");
	printf("Lowercase: Not used in final assembled structure.\n");
	printf("\n");
	
}


/*
 * Parses the file given by the parameter 'pdb_files_list_filename', which must contain
 * the paths of one or more PDB files that represent the subfragments of the final
 * desired structure, loads the PDB structures, and allocates memory for each of the
 * peptides. Memory is allocated for 'subfragments', an array with one ResData* for each
 * PDB file (subfragment). Also allocates 'subfrag_lengths', and array of the number of
 * residues in each subfragment, and 'subfrag_filenames', and array of the names of the
 * PDB files of each subfragment. Sets 'num_subfrags' to the number of subfragments,
 * i.e. the number of PDB files loaded.
 */
static void load_PDB_files(char *pdb_files_list_filename, ResData ***subfragments, int **subfrag_lengths,
		char ***subfrag_filenames, int *num_subfrags) {
	
	FILE *fp;
	char line[MAX_LINE_LEN];
	int num_subfragments = 0;
	int subfrag_index = 0;
	char *curr_PDB_filename;
	
	if ((fp = fopen(pdb_files_list_filename,"r")) != NULL) {
		
		// Count the number of PDB files listed in the file.
		while (fgets(line, MAX_LINE_LEN, fp) != 0) {
			if ( !(line[0] == '\n' || line[0] == '\0') ) {
				num_subfragments++;
			}
		}
		
		if (num_subfragments == 0) {
			fprintf(stderr, "ERROR: No PDB files listed in file \'%s\'.\n", pdb_files_list_filename);
			exit(1);
		}
		
		*subfragments = malloc( num_subfragments * sizeof(ResData *) ); assert(*subfragments != NULL);
		*subfrag_lengths = malloc( num_subfragments * sizeof(int) ); assert(*subfrag_lengths != NULL);
		*subfrag_filenames = malloc( num_subfragments * sizeof(char *) ); assert(*subfrag_filenames != NULL);
		
		// Parse each PDB file individually.
		rewind(fp);
		subfrag_index = 0;
		while (fgets(line, MAX_LINE_LEN, fp) != 0) {
			if ( !(line[0] == '\n' || line[0] == '\0') ) {
				trim(line); // Remove trailing whitespace or newline.
				curr_PDB_filename = line;
				
				load_PDB_file(curr_PDB_filename, &(*subfragments)[subfrag_index], &(*subfrag_lengths)[subfrag_index]);
				
				// Save name of PDB file.
				
				(*subfrag_filenames)[subfrag_index] = malloc( (strlen(curr_PDB_filename) + 1) * sizeof(char) );
				assert((*subfrag_filenames)[subfrag_index] != NULL);
				strcpy((*subfrag_filenames)[subfrag_index], curr_PDB_filename);
				
				subfrag_index++;
			}
		}
		
		*num_subfrags = num_subfragments;
	}
	else {
		fprintf(stderr, "ERROR: Could not open file with list of PDB files \'%s\'.\n", pdb_files_list_filename);
		exit(1);
	}
	
}


/*
 * Loads the peptide (1st chain) from the given PDB file. Allocates memory for the an array of
 * ResData structs representing the peptide at the memory pointed to by 'residues'. Sets 'numRes'
 * to the number of residues in the peptide.
 *
 * This function sets the angles and template coordinates, but does not build the peptide itself
 * (i.e. it does not call buildPeptide()).
 */
static void load_PDB_file(char *inputPDBFilename, ResData **residues, int *numRes) {
	
	char *phipsiAnglesFilename;
	int n = -1;
	PDBFile pdb; // Not currently used.
	//int rc;
	
	phipsiAnglesFilename = createAnglesFile(inputPDBFilename);
	n = countResidues(phipsiAnglesFilename);
	
	/*
	rc = parsePDBFile(inputPDBFilename,&pdb, '?');
	if (rc != TRUE) {
		fprintf(stderr, "Error (%d): Could not parse PDB file \'%s\'.\n", rc, inputPDBFilename);
		exit(-1);
	}
	
	n=pdb.numRes; // Both values are the number of residues in the peptide.
	*/
	
	if (n <= 0) {
		fprintf(stderr, "ERROR: PDB file \'%s\' contains 0 residues.\n", inputPDBFilename);
		exit(1);
	}
	
	*residues = SAFEMALLOC(sizeof(ResData)*n);
	readAngles(phipsiAnglesFilename, *residues, n);
	loadResidueTemplates(inputPDBFilename, pdb, TRUE, TRUE, FALSE, NULL, *residues, n);
	repair_O_geometry(*residues, n);
	
	*numRes = n;
}


/*
 * Parses the sequence from the given FASTA file, allocated memory for the array pointed to by
 * 'sequence', and sets 'sequence' to the parsed sequence (single-residue codes). Returns the
 * number of residues in the sequence.
 */
static int parse_FASTA_file(char *fasta_filename, char **sequence) {
	
	FILE *fp;
	char line[MAX_LINE_LEN];
	int numRes = 0;
	
	if ((fp = fopen(fasta_filename,"r")) != NULL) {
		
		// Count the number of residues.
		while (fgets(line, MAX_LINE_LEN, fp) != 0) {
			if (line[0] == '>') {
				// Header line; skip it.
			}
			else {
				trim(line); // Remove trailing whitespace or newline.
				numRes += strlen(line);
			}
		}
		
		// Allocate memory for the sequence.
		*sequence = malloc( (numRes + 1) * sizeof(char) ); assert(*sequence != NULL);
		
		// Load the sequence.
		(*sequence)[0] = '\0';
		rewind(fp);
		while (fgets(line, MAX_LINE_LEN, fp) != 0) {
			if (line[0] == '>') {
				// Header line; skip it.
			}
			else {
				trim(line); // Remove trailing whitespace or newline.
				strcat(*sequence, line);
			}
		}
		
		return numRes;
	}
	else {
		fprintf(stderr, "ERROR: Could not open FASTA file \'%s\'.\n", fasta_filename);
		exit(1);
	}
	
}


/*
 * Aligns the sequence from the Preditor file, from 'angles_preditor', onto the target sequence
 * and sets the values in 'angles_preditor_aligned' according to this alignment. The result will
 * be that angles_preditor_aligned[i] has the Preditor angles mapped to the residue at sequence[i],
 * or "not available" values if no Preditor info is mapped for the residue.
 *
 * This function assumes that memory for 'angles_preditor_aligned' has already been allocated.
 *
 * sequence: The final target sequence as a string of single-residue amino acid abbreviations
 *      plus NULL terminator. The sequence must be in uppercase.
 * numRes: Number of residues in 'sequence'.
 * angles_preditor: Pointer to an array of num_resd_preditor ANGLE_INFO structs as parsed from
 *      the Preditor file.
 * num_resd_preditor: Number of residues in the Preditor file = the number of ANGLE_INFO structs
 *      pointed to by 'angles_preditor'.
 * angles_preditor_aligned: Pointer to an array of numRes ANGLE_INFO structs that this function
 *      will fill in with data.
 */
static void align_preditor_info_to_target_sequence(char *sequence, int numRes, ANGLE_INFO *angles_preditor,
		int num_resd_preditor, ANGLE_INFO *angles_preditor_aligned) {
	
	int i, j, k;
	char *preditor_seq;
	
	FILE *ff;
	char buffer[MAX_LINE_LEN];
	char *main_segment = NULL;
	char *preditor_segment = NULL;
	char main_char;
	char preditor_char;
	int num_identical_residues;
	int num_aligned_residues;
	float seq_identity; // Proportional.
	float seq_identity_threshold = 0.3F; // Fail if sequence alignment results in a sequence identity below this proportion.
	
	
	// Extract sequence from the Preditor info. If it does not match exactly with the 'sequence' parameter, then
	// do an alignment. Transfer the information to 'angles_preditor_aligned'.
	
	preditor_seq = (char *)malloc(sizeof(char) * (num_resd_preditor + 1));
	
	for (i = 0; i < num_resd_preditor; i++) {
		preditor_seq[i] = angles_preditor[i].res;
	}
	preditor_seq[num_resd_preditor] = '\0';
	
	if (strcmp(sequence, preditor_seq) == 0) {
		for (i = 0; i < numRes; i++) {
			angles_preditor_aligned[i] = angles_preditor[i];
		}
	}
	else {
		// The given sequence and Preditor sequences differ. Call clustalw to perform an alignment, and
		// map the angle information from the Preditor file onto the given sequence.
		// 
		// clustalw1.83.UNIX.tar.gz was obtained from ftp://ftp.ebi.ac.uk/pub/software/unix/clustalw/
		// See http://www.bioinformatics-toolkit.org/Help/Topics/clustalw.html
		// Citation:
		// Thompson J.D., Higgins D.G., Gibson T.J. "CLUSTAL W: improving the sensitivity of progressive
		// multiple sequence alignment through sequence weighting, position-specific gap penalties and
		// weight matrix choice." Nucleic Acids Res. 22:4673-4680(1994).
		
		main_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
		preditor_segment = (char *)malloc(sizeof(char) * MAX_LINE_LEN);
		
		// Write FASTA file with the two protein sequences.
		
		ff = fopen("tmp_align.fa", "w");
		fprintf(ff, ">main\n");
		fprintf(ff, "%s\n", sequence);
		fprintf(ff, ">pred\n");
		fprintf(ff, "%s\n", preditor_seq);
		fclose(ff);
		
		// Call ClustalW.
		
		if (system("./clustalw tmp_align.fa > /dev/null") != 0) {
			fprintf(stderr, "ERROR: Call to clustalw failed.\n");
			exit(-1);
		}
		
		// Parse alignment file generated by ClustalW.
		
		ff = fopen("tmp_align.aln", "r");
		
		// Skip first 3 lines
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
			fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
		}
		
		num_identical_residues = 0;
		num_aligned_residues = 0;
		j = 0;
		k = 0;
		while( fgets(buffer, MAX_LINE_LEN, ff) != NULL ) {
			sscanf(buffer, "%*s %s", main_segment);
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			sscanf(buffer, "%*s %s", preditor_segment);
			
			// Compare the two sequences, and map the angle information from the Preditor file
			// to the main sequence.
			i = 0;
			main_char = main_segment[0];
			preditor_char = preditor_segment[0];
			while (main_char != '\0') {
				
				if (main_char != '-') {
					if (preditor_char != '-') {
						angles_preditor_aligned[j] = angles_preditor[k];
						
						j++;
						k++;
						num_aligned_residues++;
						if (main_char == preditor_char) {
							num_identical_residues++;
						}
					}
					else {
						// Preditor sequence does not contain a residue that is in the main
						// sequence.
						
						angles_preditor_aligned[j].res_no = RES_ID_NA;
						angles_preditor_aligned[j].res = 'X';
						angles_preditor_aligned[j].phi = PREDITOR_ANGLE_NA;
						angles_preditor_aligned[j].psi = PREDITOR_ANGLE_NA;
						angles_preditor_aligned[j].omega = PREDITOR_ANGLE_NA;
						angles_preditor_aligned[j].chi = PREDITOR_ANGLE_NA;
						angles_preditor_aligned[j].phi_err = 0.0;
						angles_preditor_aligned[j].psi_err = 0.0;
						angles_preditor_aligned[j].chi_confidence = 0.0;
						
						j++;
					}
				}
				else {
					// Main sequence does not contain this residue. Skip over the Preditor info
					// for this residue.
					k++;
				}
				
				i++;
				main_char = main_segment[i];
				preditor_char = preditor_segment[i];
			}
			
			// Skip 2 lines.
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				fprintf(stderr, "ERROR: Problem reading tmp_align.aln file.\n"); exit(1);
			}
			if ( fgets(buffer, MAX_LINE_LEN, ff) == NULL ) {
				// End of file.
				break;
			}
		}
		
		fclose(ff);
		
		if (num_aligned_residues != 0) {
			seq_identity = ((float) num_identical_residues)/((float) num_aligned_residues);
		}
		else {
			seq_identity = 0.0F;
		}
		
		if (seq_identity < seq_identity_threshold) {
			fprintf(stderr, "ERROR: Aligment between sequences in FASTA file and Preditor output file\n");
			fprintf(stderr, "results in sequence identity of only %4.1f%%!\n", 100.0F*seq_identity);
			fprintf(stderr, "Wrong input files suspected.\n");
			exit(1);
		}
		
		if (j != numRes) {
			fprintf(stderr, "ERROR: Incorrect number of residues after aligning sequences.\n");
			fprintf(stderr, "   %d residues counted, but %d expected based on PDB file.\n", j, numRes);
			exit(1);
		}
		
		free(main_segment);
		free(preditor_segment);
	}
	
	free(preditor_seq);
	
}


/*************************************************************/
/* Prints info about how to use the program.                 */
/*************************************************************/
static void usage() {
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "   sfassembler [-p] -f fasta_file -pdb pdb_files_list [-pred preditor_file]\n");
	fprintf(stderr, "   [-o output.pdb]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Arguments:\n");
	fprintf(stderr, " -p\n");
	fprintf(stderr, "   Print information about the alignment used to assemble the structure.\n");
	fprintf(stderr, " -f\n");
	fprintf(stderr, "   FASTA file with full sequence desired in file peptide.\n");
	fprintf(stderr, " -pdb\n");
	fprintf(stderr, "   File with the paths of one or more PDB files, one per line.\n");
	fprintf(stderr, " -pred\n");
	fprintf(stderr, "   Optional file with Preditor-predicted angles.\n");
	fprintf(stderr, " -o output.pdb\n");
	fprintf(stderr, "   Name of output PDB file.\n");
	exit(1);
}


int main_sfassembler(int argc, char **argv) {
	
	int i;
	FILE *fp_write;
	
	char *fasta_filename = NULL;
	int havePreditorFile = FALSE;
	char *preditor_output = NULL;
	char *pdb_files_list_filename = NULL;
	char *outputPDBFilename = "OUTPUT.pdb";
	
	int numRes = 0; // Number of residues in the full sequence.
	char *sequence = NULL; // Full sequence in desired final peptide.
	ANGLE_INFO *angles_preditor = NULL; // Angles from the Preditor file.
	int num_resd_preditor;
		// Number of residues in Preditor file = number of ANGLE_INFO structs in angles_preditor.
	ANGLE_INFO *angles_preditor_aligned = NULL;
		// Angles from the Preditor file aligned onto the final desired sequence of the peptide.
		// This should have numRes ANGLE_INFO structs.
	ResData **subfragments = NULL;
		// Array with a ResData* pointing to each subfragment peptide, one per PDB file
		// for a total of num_subfragments. A peptide is represented by a ResData array.
	int *subfrag_lengths = NULL;
		// Array of num_subfragments ints representing the lengths of each peptide in
		// 'subfragments'.
	char **subfrag_filenames;
		// Array of num_subfragments strings with the PDB filenames of each of the subfragments.
	int num_subfragments; // The number of subfragments, i.e. the number of PDB files given.
	ResData *assembled_structure;
	int print_alignment = FALSE;
	
	havePDBFile = TRUE;
	
	
	if (argc == 1) {
		fprintf(stderr, "SFAssembler: subfragment assembly of protein structures.\n\n");
	}
	
	if (argc < 5) {
		usage();
	}
	
	
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-p") == 0) {
     		print_alignment = TRUE;
		}
     	else if (strcmp(argv[i], "-f") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				fasta_filename = argv[i];
				check_file_readability(fasta_filename);
			}
			else {
				fprintf(stderr, "FASTA filename must be specified after -f option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-pdb") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				pdb_files_list_filename = argv[i];
				check_file_readability(pdb_files_list_filename);
			}
			else {
				fprintf(stderr, "File with list of PDB filenames must be specified after -pdb option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-pred") == 0) {
			if (i + 1 < argc ) { // if arg i is not the last arg
				i++;
				preditor_output = argv[i];
				check_file_readability(preditor_output);
				havePreditorFile = TRUE;
				//printf("input preditor file = %s\n", preditor_output);
			}
			else {
				fprintf(stderr, "Preditor output file must be specified after -pred option.\n");
				usage();
			}
		}
		else if (strcmp(argv[i], "-o") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				outputPDBFilename = argv[i];
			}
			else {
				fprintf(stderr, "Output PDB filename must be specified after -o option.\n");
				usage();
			}
		}
		
	}
	
	if (fasta_filename == NULL) {
		fprintf(stderr, "ERROR: FASTA file with full sequence of final desired peptide must be given.\n");
		exit(1);
	}
	else if (pdb_files_list_filename == NULL) {
		fprintf(stderr, "ERROR: File with list of PDB files must be given.\n");
		exit(1);
	}
	
	
	read_rama_tables();
	
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", general_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", gly_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/pro_N_CA_C_angle.txt", pro_N_CA_C1_angle_lookup);
	read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/xpr_N_CA_C_angle.txt", xpr_N_CA_C1_angle_lookup);
	
	loadRotamerLibrary();
	
	
	numRes = parse_FASTA_file(fasta_filename, &sequence);
	
	// Convert sequence to uppercase.
	for (i = 0; i < numRes; i++) {
		sequence[i] = (char) toupper(sequence[i]);
	}
	
	if (havePreditorFile) {
		// Read Preditor file into memory.
		
		num_resd_preditor = parse_preditor_output(preditor_output, &angles_preditor);
		
		angles_preditor_aligned = malloc( numRes * sizeof(ANGLE_INFO) ); assert(angles_preditor_aligned != NULL);
		
		align_preditor_info_to_target_sequence(sequence, numRes, angles_preditor, num_resd_preditor,
			angles_preditor_aligned);
		
		free(angles_preditor);
	}
	
	load_PDB_files(pdb_files_list_filename, &subfragments, &subfrag_lengths, &subfrag_filenames, &num_subfragments);
	
	
	// Allocate memory for the final assembled structure, and initialize the residue IDs.
	assembled_structure = malloc( numRes * sizeof(ResData) ); assert(assembled_structure != NULL);
	for (i = 0; i < numRes; i++) {
		assembled_structure[i].id = RES_ID_NA;
	}
	
	
	assemble_subfragments(sequence, numRes, angles_preditor_aligned, subfragments, subfrag_lengths, subfrag_filenames,
		num_subfragments, print_alignment, assembled_structure);
	
	
	buildPeptide(assembled_structure, numRes);
	
	fp_write = fopen(outputPDBFilename,"w");
	printPDBAtoms(fp_write, assembled_structure, numRes);
	fclose(fp_write);
	
	printf("The assembled structure is written to %s\n\n", outputPDBFilename);
	
	
	// Free memory:
	
	for (i = 0; i < num_subfragments; i++) {
		free(subfragments[i]);
	}
	free(subfragments);
	
	free(subfrag_lengths);
	
	for (i = 0; i < num_subfragments; i++) {
		free(subfrag_filenames[i]);
	}
	free(subfrag_filenames);
	
	free(sequence);
	free(angles_preditor_aligned);
	free(assembled_structure);
	
	return 0;
}

