/***************************************************************
 * File Name: loop.c
 * Purpose: This file contains routines required to perform 
 *   csipdb gap finding and filling.  
 *
 * Authors: Philip S Luk ~ philipluk@bigfoot.com
 * Refactoring: Moustafa Mourad 
 *          Haiyan Zhang   
 * Date Written : February 01, 2001
 * Last Modified: July 05, 2001
 *
 * Written for Dr. D Wishart, Faculty orf Pharmacy & 
 * Pharmaceutical Sciences at the University of Alberta ~ 2001
 * This file is provided as-is.
 **************************************************************/

#include <math.h>
#include "csipdb.h"
#include "vector.h"
#include "loop.h"
#include "rotatefinal.h"
#include "superpose.h"
#include "time.h"
#include "misc.h"
#define CN_MAX_DIST      2.0
#define THRESHOLD_RMSD   1.85
#define NUM_TRIES 5

float distancecutoff=1.0; 

/* the cutoff value of how much distance is allowed
between the len of loop inserted and the space provided */
/*----------------------------------------------------
 * Step 1: Find gaps
 * Find gaps in query sequence and data base sequence 
 * from the alignment record given and stores gap
 * information in the gap info record pass in. match
 * rank is the ranking of the alignment in scoring.
 * Returns the number of gaps found in total.
 *---------------------------------------------------- */
int find_gaps(ALIGN_REC *align, char *pdbfile_name) {  
  int inqpGap, indpGap, gapCount;

  /* initialize a few things needed for counting gaps */
  inqpGap  = 0;
  indpGap  = 0;
  gapCount = 0;

  for (int i=0; i<align->len; i++) {
    /*** GAPS IN QUERY STRING ***/
    if (align->qpRes[i] == ' ') {
      /* Gap here.. */
      if (inqpGap == 0) {
        /* Gap beginning */
        inqpGap = 1;       /* update gap status */
        gapCount ++;
    	setup_query_gap(align, gapCount, i);
    	
 	 	} 
 	 else {
 	 	gap_continuation(align, gapCount, i);
 	 	
 	 	}
    }
    else {
      /* Not in gap.. */
      inqpGap = 0; /* update gap status */
    
    } 
    /*** GAPS IN DATABASE STRING ***/
    if (align->dpRes[i] == ' ') {
      /* Gap here.. */
      if (indpGap == 0) {
        /* Gap beginning */
        indpGap = 1;                 /* update gap status */
        gapCount ++;
        setup_database_gap(align, gapCount, i);
       
      } 
      else {
      	gap_continuation(align, gapCount, i);
      }
    } 
    else {
      /* Not in gap.. */
      indpGap = 0; /* update gap status */
    }
  }
  align->gapCount = gapCount;
  
  /* -------------------------------------------------------------
   * Step 2: Find indices of pre and post secondar structures
   * Ok, now we need to calculate the indices of the pre and post
   * secondary structures for later fist and elbow calculations
   * -------------------------------------------------------------*/
  for (int i=1;i<=align->gapCount;i++) {
    set_prev_post_fist_struct_index(align,i);
 
  }

  /* Calculate the fist and elbow distances */
  calc_gaps_fist_elbow_dist(align, pdbfile_name);
  return gapCount;
} /* end find_gaps() */



/*----------------------------------------------------------------
 * Get the closest previous secondary structure index and following
 * the gap for fist dist calcuations. Also sets the gap_class of the loop.
 *---------------------------------------------------------------- */
int set_prev_post_fist_struct_index(ALIGN_REC *align, int gap) {
  int didWePassTheGap;
  int didWeEnterTheGap;
  int done,prevdist,postdist;
  char before = 'B';   /* HCorB before the loop (B by default) */
  char after = 'B';  /* HCorB after the loop (again, B by default) */

  /* Initialize variables */
  didWePassTheGap  = 0; //0: No, 1: Yes
  didWeEnterTheGap = 0;
  done = 0; int i = 0;
  
	
	/* grep starting and ending position information of gaps*/
	if(align->gaps[gap].isBeg==1) {
		/* for gaps in the beginning */
		if(align->gaps[gap].type==INSERTION)
		{
			while (align->dsRes[i] == 'C' || align->dsRes[i]==' ' )
				i++;
			align->gaps[gap].post_fInd=i+1;
			after=align->dsRes[i];
			align->gaps[gap].loop_len=i;
			align->gaps[gap].prev_fInd=1;
			if(align->dsRes[i]=='H')
				align->gaps[gap].gap_class = BCH;
			if(align->dsRes[i]=='B')
				align->gaps[gap].gap_class = BCB;

//Moustafa: Not sure what is being done
	//if the tail is too long, we just predict 14 residue loop?
	
	
			if(align->gaps[gap].size > 10){	
				align->gaps[gap].loop_len =align->gaps[gap].loop_len - align->gaps[gap].size + 10;
		
				if(align->gaps[gap].post_fInd - align->gaps[gap].size -1 >=4)
				{
					align->gaps[gap].post_fInd = align->gaps[gap].size + 4 +1;
					align->gaps[gap].loop_len = 10 + 4;	
				}
			}
	
			//added by xiaoli. If the loop is too long, it may overlap with
			//other portion of the structure. So, making the loop shorter
			else
			{	
				if(align->gaps[gap].post_fInd - align->gaps[gap].size -1 >=4)
					{
						align->gaps[gap].post_fInd = align->gaps[gap].size + 4 +1;
						align->gaps[gap].loop_len = align->gaps[gap].size + 4;
						
					}
			}
	
	
	//end of xiaoli
		}
		else 
		{ /* if deletion, do not care about it */
			align->gaps[gap].prev_fInd = 0;
			align->gaps[gap].post_fInd = 0;
			align->gaps[gap].gap_class = -1;
		}
		
	
	}
	else if(align->gaps[gap].isEnd==1)
	{ 
		/* for gaps in the end of sequence */
		if(align->gaps[gap].type==INSERTION)
		{ 
			i=align->len;
	
			while(align->dsRes[i]!='B' && align->dsRes[i] != 'H')
			{i--;}
			align->gaps[gap].prev_fInd=i+1;
			after=align->dsRes[i];
			align->gaps[gap].loop_len=align->len-i-1;
			align->gaps[gap].post_fInd=align->len;
			if(align->dsRes[i]=='H')
				align->gaps[gap].gap_class = HCB;
			if(align->dsRes[i]=='B')
				align->gaps[gap].gap_class = BCB;
		} 
	
		else{ /* if deletion, don't care abt it */
			align->gaps[gap].prev_fInd = 0;
			align->gaps[gap].post_fInd = 0;
			align->gaps[gap].gap_class = -1;
		}
			
			/*
	added by xiaoli. If the loop is too long, it may overlap with
	other portion of the structure. So, making the loop shorter
	
	So,just move prev_fInd closer to the end
			*/
			
			//tail is longer than 10
		if(align->gaps[gap].size > 10){
	
			align->gaps[gap].loop_len = align->gaps[gap].loop_len - align->gaps[gap].size + 10;
				
			if(align->gaps[gap].post_fInd - align->gaps[gap].size - align->gaps[gap].prev_fInd >=4)
			{
				align->gaps[gap].prev_fInd = align->gaps[gap].post_fInd - align->gaps[gap].size-4;
				align->gaps[gap].loop_len = 10 + 4;
				
			}

	
		}
		else{
	
			if(align->gaps[gap].post_fInd - align->gaps[gap].size - align->gaps[gap].prev_fInd >=4)
			{
				align->gaps[gap].prev_fInd = align->gaps[gap].post_fInd - align->gaps[gap].size-4;
				align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd;
				
			}
		}
			
			//end of xiaoli
			
	}
	
	else {
		/* need to figure out the secondary structures */
		//printf("middle1 xiaoli size=%d, loop_len=%d\n",align->gaps[gap].size, align->gaps[gap].loop_len );
		
		while (done == 0) { 
			/* don't ask why its i+1.. i don't know 
			* MOUSTAFA: its i + 1 because all the arrays start at 0, and therefore they are offset by 1,
			* the individual that programmed this wanted all the indeces found for gaps, fist, and elbow positions
			* to correspond to the residue position in the sequence, and not the index in the array. So if a gap is found to
			* beg at index 3 in the array of characters, it really is the 4th residue position in, and therefore they 
			* add "1" to that i where the gap begining was found
			* ALSO NOTE: the align->gaps[] array starts at index 1, and align->gaps[0] never gets anything stored in it*/
			if (align->gaps[gap].beg == i+1) {
				didWeEnterTheGap = 1;
			}
			if (align->gaps[gap].end == i) {
				didWePassTheGap = 1;
			}
			//printf("middle2 xiaoli size=%d, loop_len=%d\n",align->gaps[gap].size, align->gaps[gap].loop_len );
			if (didWeEnterTheGap == 0) {
				if ((align->dsRes[i] == 'B') || (align->dsRes[i] == 'H')) {
		/* Keep updating index if secondary structure is either Beta
			or Helix */
					align->gaps[gap].prev_fInd = i+1; /* update until we get to gap */
					before = align->dsRes[i]; /* hold onto this for gap_classification */
				}
			} else if (didWePassTheGap == 1) { //printf("middle3 xiaoli size=%d, loop_len=%d\n",align->gaps[gap].size, align->gaps[gap].loop_len );
				if ((align->dsRes[i] == 'B') || (align->dsRes[i] == 'H')) {
					/* Hold onto this for loop_len calc, but will update later */
					align->gaps[gap].post_fInd = i +1; /* grab pos right after gap */
		//printf("post postion post_fInd=%d\n",align->gaps[gap].post_fInd );
		if (align->gaps[gap].type == DELETION) {
						/* req loop is loop length minus deletion */
						align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd - 1 - align->gaps[gap].size; //****MOUSTAFA: this statement is problematic, gives potential loop length of 0, is that possible?
		// printf("post postion post_fInd=%d loop_size=%d\n",align->gaps[gap].post_fInd , align->gaps[gap].loop_len);
			align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd - 1;
			align->gaps[gap].post_fInd = align->gaps[gap].post_fInd + floor(align->gaps[gap].size/2);
			align->gaps[gap].prev_fInd = align->gaps[gap].prev_fInd - ceil(align->gaps[gap].size/2);
		// printf("loop length xiaoli = %d\n", align->gaps[gap].loop_len);
			
						if ( align->gaps[gap].loop_len == 0 )
						{
							/* this means that the entire deletion is the template's loop.  We can either expand the loop back by one residue 
							on either side or we can forget about it, which means we
							will have discontinuities in the protein.*/
							//MOUSTAFA: THIS case has not been programmed, what should we do?
							printf("**WARNING**: We have a loop length of 0\n");
							
						}
			
					} else if (align->gaps[gap].type == INSERTION) {
						/* req loop is loop length plus insertion */
						/* since database indices already take into account gaps, */
						/* we don't need to do anything special */
						align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd - 1; 
					} else {
						fprintf(stderr, "Could not determine required loop size for gap!\n");
					}
					done = 1; /* have our info, can leave now */
				}
	
			}
		
			i++;
			if (i == align->len) {


				done = 1;
			}
		} /* end while */
		
		
		/* set the loop classification */ 
	
	set_loop_classification(align, before, after, gap);
	

	} /* end else */

 
  
  
  
  
  int envelopsOtherGap = 0;
  envelopsOtherGap = envelops_other_gap(align, gap);


  if (  envelopsOtherGap == 0 )
  {
    /*
       Not sure why we're doing this... but it looks like
      we're trying to minimize the information loss in big gaps.
      Obviously if there's a big loop we're going to have a pretty 
      bad RMSD in the loop area. So this minimizes how much of the loop
      that we lose.
    */ 
    if(  align->gaps[gap].loop_len>6 
      && align->gaps[gap].isBeg == 0 
      && align->gaps[gap].isEnd == 0)
      { 
        //if(align->gaps[gap].type==DELETION)
        //{  
            prevdist=align->gaps[gap].beg-align->gaps[gap].prev_fInd;
            postdist=align->gaps[gap].post_fInd-align->gaps[gap].end;
      
            if(prevdist>4)
              {
                align->gaps[gap].prev_fInd = align->gaps[gap].beg-4;
                //align->gaps[gap].prev_fInd = align->gaps[gap].beg-4 - ceil(align->gaps[gap].size/2);
                
              }
      
            if(postdist>4)
              {
                align->gaps[gap].post_fInd = align->gaps[gap].end+4;
                //align->gaps[gap].post_fInd = align->gaps[gap].end+4 + floor(align->gaps[gap].size/2);
              }
            if(align->gaps[gap].type==DELETION)
              align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd+1 - align->gaps[gap].size - 2;
            //align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd -1 + align->gaps[gap].size;
            
            if(align->gaps[gap].type==INSERTION)
              align->gaps[gap].loop_len = align->gaps[gap].post_fInd - align->gaps[gap].prev_fInd - 1;
            //}
      }
    }
  
  
  return 0;
} /* end set_prev_post_fist_struct_index() */

/*----------------------------------------------------
 * Reads the PDB file and calculates the fist and
 * elbow distances.
 *---------------------------------------------------- */
int calc_gaps_fist_elbow_dist(ALIGN_REC *align, char *pdbfile_name) {
  
  FILE *pdbF;                   /* PDB database file template */
  ATOM_INFO_REC atom;           /* give to read_pdb to get atom info */
  SVECTOR res_coords[10000];      /* coordinates of residues from pdb file */ 
  int gap;
  int i = 0;
  double dists[25];
  
#ifdef DEBUG
  printf("\nGap calculations using PDB file: %s\n", pdbfile_name);
#endif

  memset(res_coords, 0, 10000 * sizeof(SVECTOR));
  pdbF = open_file(pdbfile_name, "r");
  
  /* Now let us pull info from the PDB file */
  while ( read_pdb(pdbF, &atom) != 0) {
    /* When calculating distances, we use the CA atom's xyz coordinates */
    if ( strcmp(atom.atomName, "N") == 0 ) {
      res_coords[atom.resNum].Nx = atom.x;
      res_coords[atom.resNum].Ny = atom.y;
      res_coords[atom.resNum].Nz = atom.z;
    }
    if ( strcmp(atom.atomName, "C") == 0 ) {
      res_coords[atom.resNum].Cx = atom.x;
      res_coords[atom.resNum].Cy = atom.y;
      res_coords[atom.resNum].Cz = atom.z;
    }
    if ( strcmp(atom.atomName, "CA") == 0 ) {
      res_coords[atom.resNum].Ax = atom.x;
      res_coords[atom.resNum].Ay = atom.y;
      res_coords[atom.resNum].Az = atom.z;
    }
    if ( strcmp(atom.atomName, "O") == 0 ) {
      res_coords[atom.resNum].Ox = atom.x;
      res_coords[atom.resNum].Oy = atom.y;
      res_coords[atom.resNum].Oz = atom.z;
    }
    if ( strcmp(atom.atomName, "CB") == 0 ) {
      res_coords[atom.resNum].Bx = atom.x;
      res_coords[atom.resNum].By = atom.y;
      res_coords[atom.resNum].Bz = atom.z;
    }
  }
  fclose(pdbF);
  /* free(pdbfile_name);*/
  
  /* Now that we have the PDB coordinates for all residues.. */
  /* Let's calculate some fist and elbow distances */
  /* NOTE: Leave out first and last gap for now until I can 
     figure out how to handle */
  align->query_offset=0;
  for (gap=1;gap<=align->gapCount;gap++) {
    if ((align->gaps[gap].isBeg==0) && (align->gaps[gap].isEnd==0)) {

      	align->gaps[gap].fistdist = 
		calc_len(
		 res_coords[align->gaps[gap].prev_fInd - align->query_offset],
		 res_coords[align->gaps[gap].prev_fInd + align->gaps[gap].loop_len + 1
			    - align->query_offset] );
      


      //calculate the distance from ca,c, n, o to ca, c, n, o and put it in the array

      for(i = 0; i < 16; i++){
		dists[i] = 0.0000;
      }
      // get the distances for all the atoms of each fist
      calc_dists(
      			res_coords[align->gaps[gap].prev_fInd - align->query_offset],
		 		res_coords[align->gaps[gap].prev_fInd + align->gaps[gap].loop_len + 1
			    - align->query_offset],dists);
            // now copy the distances into the gap structure
      for(i = 0; i < 16; i++){
		align->gaps[gap].dists[i] = dists[i];
      }
      
   
    } else {
      /* This is a beginning or ending gap.. don't need to calc MOUSTAFA: CASE NOT HANDLED */
      align->gaps[gap].fistdist  = 0;
    }
    if(align->gaps[gap].type==DELETION)
      align->query_offset+=align->gaps[gap].size;
    if(align->gaps[gap].type==INSERTION)
      align->query_offset-=align->gaps[gap].size;
  }

  return 0;
} /* end calc_gaps_fist_elbow_dist */

/*----------------------------------------------------------------
 * Finds where the query string begins with respect to the data
 * base string. Important info for later PDB loop insertion.
 *---------------------------------------------------------------- */
int set_query_database_offset(ALIGN_REC *align) {
  int i;

  align->query_offset =0;
  i = 0;
  while (align->qpRes[i] == ' ') {
    i++;
  }
  align->query_offset = i;
  return 0;
} /* END set_query_database_offset() */






/*-----------------------------------------------
 * Parses looppdb database and finds appropriate
 * matches for all gaps in alignment sequence.
 * match * rank is the ranking of the alignment 
 * in scoring.
 *----------------------------------------------- */
int fill_gaps(ALIGN_REC *align, char *pdbfile_name, char *shifty_home,
	      char *pdbdir, char *simplexFilename) 
{
  int  gap, tmpoffset=0;
  int num, len, gap_class;
  FILE *loopfp;
  char tar_filename[100], buffer[MAX_LINE_LEN],type[5];
  char looppdb_ind[MAX_FNAME_LEN];
  int fail = 1;
  
  int candidate;
  int count = 0;
  CAND candidates[100];
  double cn_nc_dis_sum = 999;
  double best_sum = 999;
  int best_loop = 0;
  
  
  sprintf(tar_filename, "%s.tmp", pdbfile_name);
  sprintf(looppdb_ind, "%s%c%s", shifty_home, '/', LOOPPDB_IND);

  
  align->query_offset=0;
  for (gap=1; gap<=align->gapCount; gap++)
    {
      if ( align->gaps[gap].dontFill ) 
      {
        printf("Skipping Gap: Gap_num= %d\n", gap );
        continue;
      }
      printf("Next Gap: Gap_num= %d\n", gap );
            
      /* CASE: gap is not at the beginning or end gap */
      if ((align->gaps[gap].isBeg == 0) && (align->gaps[gap].isEnd == 0))
        {
          
                
        START:
          align->query_offset=tmpoffset; 
#ifdef DEBUG
          printf("Looking for a candidate for gap %d, loop_len=%d\n", gap,align->gaps[gap].loop_len );
#endif
          /* Open loop index file and search for a candidate for gap */
          /* Index file contains only the data headers of the db entries..
            this was done to significantly speed up the searches */
          loopfp = open_file(looppdb_ind, "r");
		
          //corelation
          //align->gaps[gap].candidate = get_candidates(loopfp, align->gaps[gap]);
          
          //difference sum
          
          
          candidate = get_candidates1(loopfp, align->gaps[gap], candidates);
          
          /* If we have a suitable match, then replace into PDB file */
          // if (align->gaps[gap].candidate >= 0) {
          
          printf("gap length = %d, gap loop leng = %d pre index=%d, post index = %d\n", align->gaps[gap].size, align->gaps[gap].loop_len, align->gaps[gap].prev_fInd, align->gaps[gap].post_fInd );
		
          if(candidate == 0)
					{
            printf("Warning:Couldn't find right size loop in our loop database, ignore this gap\n");//,align->gaps[gap].loop_len=%d \n",align->gaps[gap].loop_len );
          }
          else if (candidate > 0) { 
 
       
            while(fail == 1 && count < NUM_TRIES){ 
            
              //fail = insert_into_pdbfile(align, gap, align->gaps[gap].candidate, pdbfile_name, tar_filename, pdbdir, shifty_home);
              fail = insert_into_pdbfile(align, gap, candidates[count].loop_num, pdbfile_name, tar_filename, pdbdir, shifty_home, &cn_nc_dis_sum, simplexFilename);
              
#ifdef DEBUG
              printf("..Count=%d, fail = %d\n",count, fail);
#endif
              
              if(cn_nc_dis_sum < best_sum)
              {
                best_loop = count;
                best_sum = cn_nc_dis_sum;
              }
              
              count ++;
            }
            
            fail = 1;
            //free(candidates);
            
            if ( count >= NUM_TRIES ){
              
              //superpose the best loop into the pdb file
#ifdef DEBUG
              printf("..Could not find ideal loop, So, we decide to use LOOP NUM %d\n", candidates[best_loop].loop_num);
#endif
              insert_into_pdbfile_1(align, gap, candidates[best_loop].loop_num, 
              pdbfile_name, tar_filename, pdbdir, shifty_home, simplexFilename);

//              insert_into_pdbfile(align, gap, candidates[best_loop].loop_num, 
//              pdbfile_name, tar_filename, pdbdir, shifty_home, simplexFilename);
            }
            
          
            fclose (loopfp);
          } 
          else //CASE: Gap is of length 1 or 2, and candidate = -1
          { 
            printf("sad....................................\n");
            fclose (loopfp);
            refindgap(align, gap,pdbfile_name);
						printf("..Couldn't find any candidates.\n");

            goto START;             
          }
        }
      else//we are at a beginning or end gap
      {
      	
      	
        loopfp = open_file(looppdb_ind, "r");	
	
	while (fgets(buffer, MAX_LINE_LEN, loopfp) != NULL) {
	
	  if (buffer[0] == '>') {
	    //parse_data_line(buffer, &num, &len, type, &elbow, &fist);
	    //parse_data_line1(buffer, &num, &len, type);
	    sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
	    gap_class = converttoint(type);
	    if((align->gaps[gap].loop_len == len) && (align->gaps[gap].gap_class==gap_class))
	      {
		align->gaps[gap].candidate=num;
		//align->gaps[gap].fistdist=fist; 

		fclose(loopfp);
		if(align->gaps[gap].isBeg==1) 
		  { 
		    //printf("this gap is at the begin.\n");
		    //modified by xiaoli
		    insert_beg_into_pdbfile(align, gap, align->gaps[gap].candidate, 
					    pdbfile_name, tar_filename, pdbdir, shifty_home);
		    

		  }
		if(align->gaps[gap].isEnd==1) 
		  {  
		    
		    //printf("this gap is at the end.\n");
		    //modified by xiaoli
		    
		    insert_end_into_pdbfile(align, gap, align->gaps[gap].candidate, 
					    pdbfile_name, tar_filename, pdbdir, shifty_home); 
		  }      
		break;
	      }
	  }
	}
      }
      if(align->gaps[gap].type==DELETION)
	align->query_offset+=align->gaps[gap].size;
      tmpoffset=align->query_offset;
    }
  
  return 0;
} /* end fill_gaps() */

void initPoint(POINT_t* pt)
{
  pt->x = 0;
  pt->y = 0;
  pt->z = 0;
}

int insert_end_into_pdbfile(ALIGN_REC *align, int gap, int candidate,
			    char *src_filename, char *tar_filename,
			    char *pdbdir, char *shifty_home) {
  FILE *srcfp;    /* Existing PDB file */
  FILE *tarfp,*dbp;    /* New PDB file with loops */
  FILE *rot_abcfp, *rot_pqrfp, *rot_prerotfp, *rot_postrotfp;

  char base_pdbfile[100];  /* Name of the pdb file from database */
  char tmp_filename[MAX_FNAME_LEN]; /* temp filename for renaming purposes */

  int start_index, finish_index,init_base;
  POINT_t database;     /* database coordinates */
  POINT_t rotated;     /* rotated coordinates */

  int num, count;
  char atom[MAX_NAME_LEN],dbatom[MAX_NAME_LEN]; 
  char res[MAX_NAME_LEN],rep_res[MAX_NAME_LEN];

  POINT_t a,b,c, p,q,r;    /* data points need for transformation */
  int want_replace;    /* control variable */
  char buffer[MAX_LINE_LEN];   /* holds input lines as we stroll thru file */
  long int position;
  float  tmpx,tmpy,tmpz;
  int atom_num,flag;
  char type[3];                 /* Loop type, eg. BCB, HCB, etc. */
  int len, beginno=0;                 /* Loop number and coil length */
  
  char looppdb_db[MAX_FNAME_LEN];
  int rc;
  
  /* use the new pdb library */
  sprintf(base_pdbfile, "%s/%s.pdb", pdbdir, align->id);
  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);


  start_index=align->gaps[gap].prev_fInd - align->query_offset;;
  finish_index=align->gaps[gap].post_fInd-align->gaps[gap].size-align->query_offset;


#ifdef DEBUG
  printf(" gap in the end , Replacing gap %d in file %s with entry %d from looppdb file %s. start_index=%d, finish_index=%d, loop_len=%d\n",
         gap, src_filename, candidate, looppdb_db, start_index, finish_index,align->gaps[gap].loop_len);
#endif
  
  /* Get rotation reference points */
 
  initPoint(&a);
  initPoint(&b);
  initPoint(&c);
  initPoint(&p);
  initPoint(&q);
  initPoint(&r);

  go_get_me_abc_pqr(src_filename, base_pdbfile, 
                    candidate, start_index, finish_index,
                    align->gaps[gap].loop_len, align->gaps[gap].size,
                    align->gaps[gap].type, align->gaps[gap].prev_fInd, 
                    &a,&b,&c,&p,&q,&r,
                    shifty_home); 
  
  rot_abcfp = fopen(ABCVALUES, "w");
  
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", a.x, a.y, a.z); 
  
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", b.x, b.y, b.z); 
  
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", c.x, c.y, c.z); 
  
  
  fclose(rot_abcfp);
  

#ifdef DEBUG
  //printf( "is Begin %.3f %.3f %.3f\n", q.x, q.y, q.z); 
  //printf( "is Begin %.3f %.3f %.3f\n", p.x, p.y, p.z); 
  //printf( "is Begin %.3f %.3f %.3f\n", r.x, r.y, r.z); 
#endif
  
  rot_pqrfp = open_file(PQRVALUES, "w");
  
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", p.x, p.y, p.z);
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", q.x, q.y, q.z);
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", r.x, r.y, r.z);
  fclose(rot_pqrfp);
  
  sprintf(tmp_filename, "%s.tmp2", src_filename);
  
  rc = rename(src_filename, tmp_filename);
  if (rc != 0) {
    printf("Failed to rename %s with %s", src_filename, tmp_filename);
    return -1;
  }
  srcfp = fopen(tmp_filename, "r");
  if (srcfp == NULL) {
    printf("File %s not found\n", tmp_filename);
    return -1;
  }
  /*
    srcfp = fopen(src_filename, "r");
    if (srcfp == NULL) {
    printf("File %s not found\n", src_filename);
    return -1;
    }
  */
  tarfp = open_file(tar_filename, "w");

  want_replace = 0;count=0;
  /* copy the pdb coords before loop */  
  while (fgets(buffer, MAX_LINE_LEN, srcfp)!=NULL)
    {

      sscanf(buffer, "%*s%*d%s%s%d%*f%*f%*f", atom,res,&num);
      if(num==start_index)
	{ 
	  want_replace=1; goto NEXT1;
	}
      
      if(want_replace==0)
	{
	  /* If we don't want to replace.. 
	     we just dump the buffer directly to the file */
	  fputs(buffer, tarfp);}
      count++;
    }

 NEXT1:  
  /* add by haiyan to replace get_loop_coords() */


  dbp = open_file(looppdb_db, "rb");

  /* search the database for the position to begin the loop */
  while (fgets(buffer, MAX_LINE_LEN, dbp) != NULL) {
    /* All loop entries in database has the format '>num' */
    if (buffer[0] == '>') {
      //parse_data_line(buffer, &num, &len, type, &elbow, &fist);
      //parse_data_line1(buffer, &num, &len, type);
      sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
    }
    if (num == candidate) 
      break;  /* stop looking for loop now */  
  }

  
  position=ftell(dbp);
  fgets(buffer, MAX_LINE_LEN, dbp);
  sscanf(buffer, "%d%s%s%*s%*s%f%f%f", &init_base, res, dbatom, &tmpx, &tmpy, &tmpz);
     
  /* search the position  to the loop beginning */
  fseek(dbp,position,SEEK_SET);

  want_replace=0;num=start_index;
  beginno=init_base+2;


  while(fgets(buffer,MAX_LINE_LEN, dbp)!=NULL)
    { 
      sscanf(buffer, "%d%s%s%*s%*s%f%f%f", 
             &atom_num, rep_res, dbatom, &tmpx, &tmpy, &tmpz); 
      num=atom_num-beginno+start_index;
     
      //printf("xiaoli %d, %d\n",beginno,atom_num );


      if(beginno==atom_num)
	{
	  want_replace=1;
	}       
      if((beginno+align->gaps[gap].loop_len+1)==atom_num)
        {
	  want_replace=0;
	  break;
	}
      /* Here is where we replace stuff in the PDB file */     
      if(want_replace==1)
	{
	  flag=0;
	  
	  
	  /* Yup, we want to start replacing lines */
	  database.x=tmpx;database.y=tmpy;database.z=tmpz;        
	  /* Apply transformation and print result to PDB file */
	  /* Create coordinates to be rotated input file */
	  rot_prerotfp = open_file(PREROTCOORDS, "w");
	  fprintf(rot_prerotfp, "%.3f %.3f %.3f\n", database.x,database.y,database.z);
	  fclose(rot_prerotfp);

	  /* Call Raj's rotate program */
	  superimpose(ABCVALUES,PQRVALUES,PREROTCOORDS,POSTROTCOORDS);

	  
	  rot_postrotfp = open_file(POSTROTCOORDS, "r");
	  fscanf(rot_postrotfp, "%lf%lf%lf", &rotated.x, &rotated.y, &rotated.z);
	  fclose(rot_postrotfp);

	  /* Print new coordinates to file */
  
	  convert(res,mySeq.aaRes[num-1]);
	  if(strcmp(dbatom,"CB")!=0 || mySeq.aaRes[num-1]!='G')
	    {
	      sprintf(buffer,
		      "ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
		      count,dbatom, res, num, rotated.x, rotated.y, rotated.z, src_filename 
		      );
	      //printf("ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
	      //count,dbatom, res, num, rotated.x, rotated.y, rotated.z, src_filename 
			//);
	      
	    pdb_line_replace(buffer, tarfp);
	    count++;} 
	} 
    }

  /* Clean up I/O files for superimpose function */
  remove(ABCVALUES);    remove(PQRVALUES); 
  remove(PREROTCOORDS); remove(POSTROTCOORDS);

  fclose(srcfp);
  fclose(tarfp);

  rc = rename(tar_filename, src_filename);
  if (rc != 0) {
    printf("Failed to overwrite %s with %s", tar_filename, src_filename);
    return -1;
  }

  remove(tmp_filename);

  return 0;
    
}/* End of insert_end_into_pdbfile */
 
/*-----------------------------------------------
 * Replaces the requested gap with the given
 * candidate based on the info found in the alignment
 * record.
 *----------------------------------------------- */
int insert_beg_into_pdbfile(ALIGN_REC *align, int gap, int candidate,
			    char *src_filename, char *tar_filename,
			    char *pdbdir, char *shifty_home) {  
  FILE *tarfp,*dbp;    /* New PDB file with loops */
  FILE *rot_abcfp, *rot_pqrfp, *rot_prerotfp, *rot_postrotfp;

  char base_pdbfile[100];  /* Name of the pdb file from database */
  char tmp_filename[MAX_FNAME_LEN]; /* temp filename for renaming purposes */

  int start_index, finish_index,init_base;
  POINT_t database;     /* database coordinates */
  POINT_t rotated;     /* rotated coordinates */

  int num=0, count;
  char atom[MAX_NAME_LEN], dbatom[MAX_NAME_LEN];
  char res[MAX_NAME_LEN], rep_res[MAX_NAME_LEN];

  POINT_t a,b,c, p,q,r;    /* data points need for transformation */
  int dirty_dirty;    /* flag to insert comments in PDB file */
  int want_replace;    /* control variable */
  char buffer[MAX_LINE_LEN];   /* holds input lines as we stroll thru file */
  long int position;
  float  tmpx,tmpy,tmpz;
  int atom_num,flag;
  char type[3];                 /* Loop type, eg. BCB, HCB, etc. */
  int len, beginno=0;                 /* Loop number and coil length */

  int db_mid_res_no;
  char looppdb_db[MAX_FNAME_LEN];
  int rc;
  int start_residue = 0;
  int endloop = 0;

  /* use the new pdb library */
  sprintf(base_pdbfile, "%s/%s.pdb", pdbdir, align->id);
  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);
  

  
  start_index=align->gaps[gap].size+1;
  finish_index=align->gaps[gap].post_fInd;
  


  printf("insert_begin, tar_filename=%s, src_filename=%s\n",tar_filename,src_filename);
  
#ifdef DEBUG
  printf("..insert loop at the beginning, Replacing gap %d in file %s with entry %d from looppdb file %s.start_index=%d, finish_index=%d loop_len=%d\n",
         gap, src_filename, candidate, looppdb_db, start_index, finish_index,align->gaps[gap].loop_len);
  
  //printf("%s", shifty_home);
#endif

  int offset = 0;
  while(align->dpRes[offset] == ' ')
  {
    offset++;
  }

  /* Get rotation reference points */
  go_get_me_abc_pqr(src_filename, base_pdbfile, 
                    candidate, start_index - offset, finish_index - offset,
                    align->gaps[gap].loop_len, align->gaps[gap].size,
                    align->gaps[gap].type, align->gaps[gap].prev_fInd, 
                    &a,&b,&c,&p,&q,&r,
                    shifty_home); 
  
  
  get_loop_coords(candidate, start_index, "N", &p, res, shifty_home);
  get_loop_coords(candidate, finish_index, "C", &q, res, shifty_home);
  db_mid_res_no = (int)((start_index+finish_index+1) / 2);
  get_loop_coords(candidate, db_mid_res_no, "CA", &r, res, shifty_home);
  
  /* Create the rotation reference input files required for superimpose */
  rot_abcfp = fopen(ABCVALUES, "w");
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", b.x, b.y, b.z); 
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", a.x, a.y, a.z); 
  fprintf(rot_abcfp, "%.3f %.3f %.3f\n", c.x, c.y, c.z); 
  fclose(rot_abcfp);

  rot_pqrfp = fopen(PQRVALUES, "w");
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", q.x, q.y, q.z);
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", p.x, p.y, p.z);
  fprintf(rot_pqrfp, "%.3f %.3f %.3f\n", r.x, r.y, r.z);
  fclose(rot_pqrfp);

  sprintf(tmp_filename, "%s.tmp2", src_filename);
  rc = rename(src_filename, tmp_filename);
  if (rc != 0) {
    printf("Failed to rename %s with %s", src_filename, tmp_filename);
    return -1;
  }
  
  /*
    srcfp = fopen(src_filename, "r");
    if (srcfp == NULL) {
    printf("File %s not found\n", src_filename);
    return -1;
    }
  */
  tarfp = fopen(tar_filename, "w");

  dirty_dirty = 0;
  want_replace = 0;
 
  /* add by haiyan to replace get_loop_coords() */
  dbp = fopen(looppdb_db, "rb");
  /* search the database for the position to begin the loop */
  while (fgets(buffer, MAX_LINE_LEN, dbp) != NULL) {
    //printf("%s", buffer);
    /* All loop entries in database has the format '>num' */
    if (buffer[0] == '>') {
      //parse_data_line(buffer, &num, &len, type, &elbow, &fist);
      sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
      
  
    }    
    if (num == candidate) {
      break;  /* stop looking for loop now */
    }
  }

  position=ftell(dbp);
  fgets(buffer, MAX_LINE_LEN, dbp);
  sscanf(buffer, "%d%s%s%*s%*s%f%f%f", 
	 &init_base, res, dbatom, &tmpx, &tmpy, &tmpz);
  
  /* search the position  to the loop beginning */
  fseek(dbp,position,SEEK_SET);

  num=1; want_replace=0;count=1;

  /* Build the tempory file to contain the beginning loop
    Adding  the rest of the file will happen later */
  if(align->gaps[gap].isBeg==1)
    {
      beginno=init_base+3;
      endloop = beginno + align->gaps[gap].loop_len;
      
      if(align->gaps[gap].size > 10){

	start_residue = align->gaps[gap].size - 10;
	
      }

      while(fgets(buffer,MAX_LINE_LEN, dbp)!=NULL)
	{ 
	  sscanf(buffer, "%d%s%s%*s%*s%f%f%f", &atom_num, rep_res, dbatom, &tmpx, &tmpy, &tmpz); 
	  
	  //printf("%s \n",buffer);
	  
	  num= atom_num-beginno + 1;
	  
	  if(beginno==atom_num)
	    {
	      want_replace=1;
	    }
	  
	  //if((beginno+finish_index)==atom_num)
	  if( endloop == atom_num)  
	    {
	      want_replace=0;
	      break;
	    }

	  /* Here is where we replace stuff in the PDB file */     
	  if(want_replace==1)
	    { 
	      flag=0;
	      /* Yup, we want to start replacing lines */
	      database.x=tmpx;database.y=tmpy;database.z=tmpz;        
	      /* Apply transformation and print result to PDB file */
	      /* Create coordinates to be rotated input file */
	      rot_prerotfp = fopen(PREROTCOORDS, "w");
	      //printf("%.3f %.3f %.3f\n", database.x,database.y,database.z);
	      fprintf(rot_prerotfp, "%.3f %.3f %.3f\n", database.x,database.y,database.z);
	      fclose(rot_prerotfp);
	      
	      /* Call Raj's rotate program */
	      superimpose(ABCVALUES,PQRVALUES,PREROTCOORDS,POSTROTCOORDS);
	      rot_postrotfp = fopen(POSTROTCOORDS, "r");
	      fscanf(rot_postrotfp, "%lf%lf%lf", &rotated.x, &rotated.y, &rotated.z);
	      fclose(rot_postrotfp);

	      /* Print new coordinates to file */
  
	      convert(res,mySeq.aaRes[num + start_residue -1]);

	      if(strcmp(dbatom,"CB")!=0 || mySeq.aaRes[start_residue + num-1]!='G'){
		
		sprintf(buffer,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
			count,dbatom, res, num + start_residue, rotated.x, rotated.y, rotated.z, src_filename);
		
		pdb_line_replace(buffer, tarfp);
		count++; 
	      }
	    } 
	}
    }  

  // Add the remainder of the PDB file to the temporary file
  want_replace=1;
   
  FILE* srcfp = fopen(tmp_filename, "r");
  if (srcfp == NULL) {
    printf("File %s not found\n", tmp_filename);
    return -1;
  }

  offset = 0;
  while(align->dpRes[offset] == ' ')
  {
    offset++;
  }

  while (fgets(buffer, MAX_LINE_LEN, srcfp)!=NULL)
  { 
    
    sscanf(buffer, "%*s%*d%s%s%d%*f%*f%*f", atom,res,&num);
    
    char atom_name[MAX_NAME_LEN];
    char line_msg[MAX_LINE_LEN];
    char aa_name[MAX_NAME_LEN];
    float x, y, z;
    sscanf(&buffer[13],"%3s", atom_name);
    sscanf(&buffer[17],"%3s",aa_name);
    sscanf(&buffer[30],"%f%f%f",&x,&y,&z);
    line_msg[0] = 0;

    if(num  >= finish_index - offset){
      want_replace=0;
    }
    
    if(want_replace==0)
    {
      /* 
          If we don't want to replace.. 
          we just dump the buffer directly to the file 
      */

      fprintf(tarfp, "ATOM %6d  %-3s %s %5d    %8.3f%8.3f%8.3f  1.00  1.00    %s\n",
                          0, atom_name, aa_name, num + offset, x, y, z, line_msg);
      //fputs(buffer, tarfp);
    }
    want_replace=1;
  }
    
    
  /* Clean up I/O files for superimpose function */
  remove(ABCVALUES);    remove(PQRVALUES); 
  remove(PREROTCOORDS); remove(POSTROTCOORDS); 

  fclose(srcfp);
  fclose(tarfp);

  // Replace the existing file
  rc = rename(tar_filename, src_filename);
  if (rc != 0) {
    printf("Failed to overwrite %s with %s", tar_filename, src_filename);
    return -1;
  }
  remove(tmp_filename);
  
  return 0;
} /* END insert_beg_into_pdbfile() */



/*-----------------------------------------------
 * Replaces the requested gap with the given
 * candidate based on the info found in the alignment
 * record.
 *  Returns 1 if it fails, 0 if success, or -1 if error
 *----------------------------------------------- */
int insert_into_pdbfile(ALIGN_REC *align, int gap, int candidate,
                        char *src_filename, char *tar_filename,
                        char *pdbdir, char *shifty_home, double *cn_nc_dis_sum, char *simplexFilename) {

  FILE *tarfp;    /* New PDB file with loops */


  char base_pdbfile[100];  /* Name of the pdb file from database */
 

  int start_index, finish_index;
  char looppdb_db[MAX_FNAME_LEN];
  int rc;
   
  //FILE *rotate = fopen(RESULT_FROM_SUP, "w");
  
  //result_from_sup = RESULT_FROM_SUP;
  float rmsd_before_min = 0;
  int no_of_fist_resd = 0;
  
  double time_diff = 0;
  int need = 1;
  double cn_diff = 0;
  double nc_diff = 0;
  char command[256];
  char temp[256];
  get_pre_post_pdb(align,gap,src_filename);
  get_pre_post_loop_pdbfile(candidate,pdbdir,shifty_home);
  
  /* use the new pdb library */
  sprintf(base_pdbfile, "%s/%s.pdb", pdbdir, align->id);
  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);
  
  start_index  = align->gaps[gap].prev_fInd - align->query_offset;  
  finish_index = start_index + align->gaps[gap].loop_len+1;
  
  printf("tar_filename=%s, src_filename=%s\n",tar_filename,src_filename);
    
  tarfp = fopen(tar_filename, "w");
    
  /* Call Raj's rotate program */
  //superimpose(ABCVALUES,PQRVALUES,PREROTCOORDS,POSTROTCOORDS);
  no_of_fist_resd = 2;
  
  
  //do the GA stuff, for now, we just call simplex
        
#ifdef DEBUG
  printf("Gap in the middle, Replacing gap %d in file %s with entry %d from looppdb file %s\nstart_index=%d, finish_index=%d, loop_len=%d\n",
	 gap, src_filename, candidate, looppdb_db, start_index, finish_index,align->gaps[gap].loop_len);
#endif
  
  char* result_from_sup = call_superpose(PRE_POST_LOOP_PDB,PRE_POST_PDB,no_of_fist_resd, &rmsd_before_min);
  
  time_t start_time = time(NULL);
  time_t finish_time;
  double cn=999;      // cn == distance from gap opening to loop start
  double nc = 999;    // nc == distance from gap close to loop end

  while(need == 1){
    
    if(time_diff >= 5){
      if(cn > CN_MAX_DIST)
	cn_diff = cn - CN_MAX_DIST;
      else
	cn_diff = 0;
      
      if(nc > CN_MAX_DIST)
	nc_diff = nc - CN_MAX_DIST;
      
      else
	nc_diff = 0;
     
      *cn_nc_dis_sum = nc_diff + cn_diff;
      

      break;
      
    }
    else{
      /* Call simplex on the resulting file */
      sprintf (command,"%s/simplex %s %s", simplexFilename,PRE_POST_PDB,RESULT_PDB );
      system(command);
      sprintf (temp, "%s", SIMPLEX_OUTPUT);
      
      // result from superpose is the file name of the resulting superposed file
      /* This is where the loop from the database is super-imposed 
        on the loop gap, from theprevious secondary structure to the
        next secondary structure
      */
      result_from_sup = call_superpose(temp,PRE_POST_PDB,no_of_fist_resd,&rmsd_before_min);
      
      finish_time = time(NULL);
      time_diff = difftime(finish_time, start_time);
      
      /*
        This is where it is determined if the loop matches well enough
        or if it needs to be minimized
      */
      need = simplex_needed(align, gap,result_from_sup,PRE_POST_PDB,rmsd_before_min, &cn, &nc);
	  
      // Output some info
      printf("cn=%4g, nc=%4g, rmsd_before_min=%4g, need=%d\n", cn, nc, rmsd_before_min, need);
    }
  }
 
  
  if(need == 1)
    return 1;
  
  else{
    printf("Running time for simplex: %g sec. insert loop to pdb file\n", time_diff);
    
    insert_loop_pdb(src_filename,tarfp, result_from_sup,start_index,finish_index,align,gap);
    
    fclose(tarfp);
    
    rc = rename(tar_filename, src_filename);
    if (rc != 0) {
      printf("Failed to overwrite %s with %s", tar_filename, src_filename);
      return -1;
    }
    
    
    return 0;
  }
} 
/* END insert_into_pdbfile() */
/*-----------------------------------------------
 * Replaces the requested gap with the given
 * candidate based on the info found in the alignment
 * record.
 *----------------------------------------------- */
int insert_into_pdbfile_1(ALIGN_REC *align, int gap, int candidate,
                        char *src_filename, char *tar_filename,
                        char *pdbdir, char *shifty_home, char *simplexFilename) {

  FILE *tarfp;    /* New PDB file with loops */


  char base_pdbfile[100];  /* Name of the pdb file from database */
 

  int start_index, finish_index;
  char looppdb_db[MAX_FNAME_LEN];
  int rc;
   
  //FILE *rotate = fopen(RESULT_FROM_SUP, "w");
  char *result_from_sup;
  //result_from_sup = RESULT_FROM_SUP;
  float rmsd_before_min = 0;
  int no_of_fist_resd = 0;
  double cn=999;
  double nc = 999;
  
  time_t start_time;
  time_t finish_time;
  double time_diff = 0;
  char command[256];
  char temp[256];

  get_pre_post_pdb(align,gap,src_filename);
  get_pre_post_loop_pdbfile(candidate,pdbdir,shifty_home);
  
  /* use the new pdb library */
  sprintf(base_pdbfile, "%s/%s.pdb", pdbdir, align->id);
  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);
  
  start_index  = align->gaps[gap].prev_fInd - align->query_offset;  
  finish_index = start_index + align->gaps[gap].loop_len+1;
  printf("tar_filename=%s, src_filename=%s\n", tar_filename,src_filename);
    
  tarfp = fopen(tar_filename, "w");
  if (tarfp==NULL)
    {
      printf ("\nCSIPDB: Could not open file %s!\n",tar_filename);
      exit(-1);
    }
 
    
  /* Call Raj's rotate program */
  //superimpose(ABCVALUES,PQRVALUES,PREROTCOORDS,POSTROTCOORDS);
  no_of_fist_resd = 2;
  
        
#ifdef DEBUG
  printf(".. gap in the middle, Replacing gap %d in file %s with entry %d from looppdb file %s.\n",gap, src_filename, candidate, looppdb_db);
#endif
  
  result_from_sup = call_superpose(PRE_POST_LOOP_PDB,PRE_POST_PDB,no_of_fist_resd, &rmsd_before_min);
  
  start_time = time(NULL);
  

  while(simplex_needed(align, gap,result_from_sup,PRE_POST_PDB,rmsd_before_min, &cn, &nc) == 1 && time_diff <= 60){
    printf("cn=%4g, nc=%4g, rmsd_before_min=%4g\n", cn, nc, rmsd_before_min);

    sprintf (command,"%s/simplex %s %s", simplexFilename,PRE_POST_PDB,RESULT_PDB );
    system(command);

    sprintf (temp, "%s", SIMPLEX_OUTPUT);
    result_from_sup = call_superpose(temp,PRE_POST_PDB,no_of_fist_resd,&rmsd_before_min);

    //result_from_sup = call_superpose( SIMPLEX_OUTPUT,PRE_POST_PDB,no_of_fist_resd,&rmsd_before_min);
    finish_time = time(NULL);
    time_diff = difftime(finish_time, start_time);
  }
    
  printf("Running time for simplex: %g sec.\n", time_diff);
  
  insert_loop_pdb(src_filename,tarfp, result_from_sup,start_index,finish_index,align,gap);
  
  //call_superpose(PRE_POST_LOOP_PDB, PRE_POST_PDB,2,&rmsd);
  fclose(tarfp);
  rc = rename(tar_filename, src_filename);
  if (rc != 0) {
    printf("Failed to overwrite %s with %s", tar_filename, src_filename);
    return -1;
  }
  return 0;
} /* END insert_into_pdbfile() */


/*-----------------------------------------------
 * Gets the ABC/PQR coordinates for rotation.
 * Coordinates are based on fist-fist atoms. 
 * PARAMS:
 * -src_filename is where to get current PDB file.
 * -candidate is the loop to insert from loopdb
 * -start/finish indices determine which piece from
 *  current PDB file is to be replaced by loop.
 * -POINT a,b,c,p,q,r are calculated based on the
 *  above information.
 *----------------------------------------------- */
void go_get_me_abc_pqr(char* src_filename, char* base_pdbfile,
                       int loop_candidate,
                       int start_index, int finish_index,
                       int loop_len, int gap_size, 
                       int gap_type, int fist_index,
                       POINT_t* a, POINT_t* b, POINT_t* c,
                       POINT_t* p, POINT_t* q, POINT_t* r,
                       char *shifty_home) {

  FILE* srcfp;      /* Current PDB file */
  FILE* basefp;      /* Base template PDB file */
  char buffer[MAX_LINE_LEN];   /* Line buffer */
  char atom[5], res[5];    /* Data fields from PDB database */
  int  num, count;    /* Data fields from PDB database */
  POINT_t template;    /* Coordinates from current PDB file */
  POINT_t database;    /* Coordinates from PDB database */

  /* New C point calculation variables */
  int     db_mid_res_no;
  int     tp_mid_res_no;

  
  srcfp = fopen(src_filename, "r");
  if (srcfp == NULL) { printf("File %s not found\n", src_filename); return; }

  int foundIndex = FALSE;
  /* Run thru the file and grab the coordinates to calc translation */
  while (fgets(buffer, MAX_LINE_LEN, srcfp) != NULL) {
    sscanf(buffer, "%*s%d%s%*s%d%lf%lf%lf",
           &count, atom, &num, &template.x, &template.y, &template.z);

    if (num == start_index) {
      foundIndex = TRUE;
      if (strcmp(atom,"N") == 0) {
        get_loop_coords(loop_candidate, (num-start_index), atom, &database,res,
			shifty_home);
        /* this is our point A/R - the CA atom of the first fist residue */
        a->x = template.x; a->y = template.y; a->z = template.z;
        p->x = database.x; p->y = database.y; p->z = database.z;
      } 
      else if (strcmp(atom,"O") == 0) {
        get_loop_coords(loop_candidate, (num-start_index), atom, &database,res,
			shifty_home);
        /* this is our point C/T - the O atom of the first fist residue */
        c->x = template.x; c->y = template.y; c->z = template.z;
        r->x = database.x; r->y = database.y; r->z = database.z;
      }
    } 
    else if (num == finish_index) {
      if (strcmp(atom,"C") == 0) {
        get_loop_coords(loop_candidate, (num-start_index), atom, &database,res,
			shifty_home);
        /* this is our point B/S - the CA atom of the last fist residue */
        b->x = template.x; b->y = template.y; b->z = template.z;
        q->x = database.x; q->y = database.y; q->z = database.z;
        break; /* We got our coords, so we can stop reading the file now */

      }
    }
  } /* END _while_ */

  if ( !foundIndex )
    printf("ERROR: Can't find start_index to find coordinates!\n");

  fclose(srcfp);

#define NEW_ABC_PQR
#ifdef NEW_ABC_PQR
 
  /* The C/R point we chose above is not great.. try to find a better one */
  /* So we choose the middle residue of the loop for our point R */
  db_mid_res_no = (int)((loop_len+1) / 2);
  get_loop_coords(loop_candidate, db_mid_res_no, "CA", r, res, shifty_home);

  /* And we go get the middle residue of the old loop for point C */
  basefp = fopen(base_pdbfile, "r");
  if (basefp == NULL) { printf("File %s not found\n", base_pdbfile); return; }

  if (gap_type==INSERTION) { tp_mid_res_no=fist_index+((loop_len-gap_size+1)/2);
  } else { tp_mid_res_no=fist_index+((loop_len+gap_size+1)/2); }
  /* Run thru the file and grab the coordinates to calc translation */
  rewind(basefp);
  while (fgets(buffer, MAX_LINE_LEN, basefp) != NULL) {
    sscanf(&buffer[11],"%4s",atom);
    sscanf(&buffer[22], "%d%lf%lf%lf",
	   &num, &template.x, &template.y, &template.z);
    if ((num == tp_mid_res_no) && (strcmp(atom,"CA")==0)) {
      c->x = template.x; c->y = template.y; c->z = template.z;
      break; /* Got what we wanted.. no need to keep parsing */

     
    }
  } /* END _while_ */
  /* printf("%f %f %f\n", c->x, c->y, c->z);
     
  */ 
  //printf("xiaoli after %s, %s, fiishindex=%d, start=%d, tp_mid_res_no = %d, c = %f, %f, %f\n",base_pdbfile, src_filename,finish_index, start_index, tp_mid_res_no, c->x, c->y, c->z);
  fclose(basefp);

#endif

} /* END go_get_me_abc_pqr() */


/*-----------------------------------------------
 * Goes and fetches the (x,y,z) coordinates from
 * the given loop at the given residue of the 
 * given atom from loop pdb database.
 *----------------------------------------------- */
int get_loop_coords(int loop_no, int res_no, char* atom, 
                    POINT_t *coordinate, char* res,
                    char *shifty_home) {
  FILE *dbp;
  char buffer[MAX_LINE_LEN];

  char type[3];                 /* Loop type, eg. BCB, HCB, etc. */
  int num, len;                 /* Loop number and coil length */
 
  int searching_for_loop;
  int need_init_base;
  int init_base = 0;
  int atom_num;

  float tmpx, tmpy, tmpz;
  char tmpatom[MAX_NAME_LEN];
  char looppdb_db[MAX_FNAME_LEN];

  /* Initialize variables */
  searching_for_loop = 1;
  need_init_base = 1;

  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);

  dbp = fopen(looppdb_db, "r");
  while (fgets(buffer, MAX_LINE_LEN, dbp) != NULL) {
    if (searching_for_loop == 1) {
      /* All loop entries in database has the format '>num' */
      if (buffer[0] == '>') {
        //parse_data_line(buffer, &num, &len, type, &elbow, &fist);
	 sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
	if (num == loop_no) {
          searching_for_loop = 0;   /* stop looking for loop now */
        }
      }
    } else {
      /* Found the loop entry in the loop database */
      sscanf(buffer, "%d%s%s%*s%*s%f%f%f", 
             &atom_num, res, tmpatom, &tmpx, &tmpy, &tmpz);
     
      if (need_init_base == 1) {
        init_base = atom_num;
        need_init_base = 0;
      }
      /* Let's see if this is the set of coordinates that they want */
      /* Have to match atom and residue number: res_no of 0 indicates they
         want the prefiststruct */
      /* NOTE: The reason for +2 is because the database always has 3 residues
         before the coil. If this ever changes, then this also has to change */
      if ((strcmp(tmpatom,atom) == 0) && ((init_base+2+res_no)==(atom_num)) ) {
        break;
      }
    } 
  } /* END while != NULL */
  fclose(dbp);

  coordinate->x = tmpx; coordinate->y = tmpy; coordinate->z = tmpz;

  return 0;
} /* END get_loop_coords() */

/*-----------------------------------------------
 * Replaces line in PDB file, if appropriate.
 *----------------------------------------------- */
int pdb_line_replace(char *buffer, FILE *destp) {

  fputs(buffer, destp);
  fflush(destp);

  return 0;
} /* END pdb_line_replace() */

/*-----------------------------------------------
 * Parses the looppdb database for top candidates
 * to fill given gap. Returns -1 if no suitable
 * candidate was found. 
 *----------------------------------------------- */
int get_candidates(FILE *fp, GAP_INFO_REC gapinfo) {
  char buffer[MAX_LINE_LEN];  /* Temp buffer for reading file */
  char type[3];      /* Loop type, eg. BCB, HCB, etc. */
  int num, len;      /* Loop number and coil length */

  float score;
  int candidate = -1;    /* Best candidate for the gap */
  double score_to_beat;    /* The best score so far */

  double dists[25];
  
  /* Let's just set it as a really high value to start with */
  //score_to_beat = 1000;
  score_to_beat = 0;
  //printf("..xiaoli xiaoli loop pre %d, post %d, len= %d\n", gapinfo.prev_fInd,gapinfo.post_fInd, gapinfo.loop_len);
  
  //added by xiaoli
  if(gapinfo.loop_len == 1 && gapinfo.size > gapinfo.loop_len){
    candidate = -1;
  }
  else{
    while (fgets(buffer, MAX_LINE_LEN, fp) != NULL) {
      if (buffer[0] == '>') {
	
	//parse_data_line(buffer, &num, &len, type, &elbow, &fist);
	parse_data_line1(buffer, &num, &len, type,dists);

	//sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
	/* Lower the score, the closer the fit */
	//score = rank_score(gapinfo, len, type, elbow, fist);
	if(len == gapinfo.loop_len){
	  
	  score = rank_score1(gapinfo, len, type, dists);
	  
	  //if ((score >= 0) && (score < score_to_beat)) {
	  
	  if (score > score_to_beat) {
#ifdef DEBUG
	    printf("..Loop %5d is a candidate;\tSCORE: %f \t %f\n", num, score, score_to_beat);
#endif
	    candidate     = num; 
	    score_to_beat = score;
	  }
	}
      }
    } /* END while != NULL */
    
    
    //if(score_to_beat>distancecutoff)
    //  {
    //	candidate=-1;
    // }
  }
  
  return candidate;
} /* end get_candidates() */



/*

this version choose candidates according to the sum of the distance difference

it will store the top 100 candidates in the candidates array. 

the candidates array are sorted according to the rank score in ascending order.

return value will be the number of the loop in the array. If the gap size is 1 or 2, 

return -1 to try to extend gap.

*/

int get_candidates1(FILE *fp, GAP_INFO_REC gapinfo, CAND candidates[100]) {
  char buffer[MAX_LINE_LEN];  /* Temp buffer for reading file */
  char type[3];      /* Loop type, eg. BCB, HCB, etc. */
  int num, len;      /* Loop number and coil length */

  float score = 0;
  int candidate = 0;    /* Best candidate for the gap */
  double score_to_beat = 999.0;    /* The best score so far */

  double dists[25];
  CAND cand; 
  int count = 0;
  
  // Initialize variables  
  memset(candidates, 0, 100 * sizeof(CAND));
  
  if(gapinfo.loop_len > 31)
    return 0;
  
  else if(gapinfo.loop_len == 1 || gapinfo.loop_len == 2){
		printf("Loop length too short.");
    candidate = -1;
  }
  else{
    
    while (fgets(buffer, MAX_LINE_LEN, fp) != NULL) {
      
      if (buffer[0] == '>') {
	
	parse_data_line1(buffer, &num, &len, type,dists);
	
	if(len == gapinfo.loop_len){
	  
	  score = rank_score2(gapinfo, len, type, dists);
	  
	  
	  if (score < score_to_beat){
	    cand.loop_num = num;
	    cand.score = score;
	    
	    if(count < 99)
            {
	      candidates[count] = cand;
	      count++;
	    }
	    else
            {
              // replace the lowest scoring candidate
	      candidates[99] = cand;
              // resort them
	      qsort(candidates, 100, sizeof(struct CAND), compare_double);
	      score_to_beat = candidates[99].score;
	    }
	    
	  }
	  
	}
      }
    } /* END while != NULL */
    candidate = count;
    
  }
  
  return candidate;
} /* end get_candidates() */


int get_candidates2(FILE *fp, GAP_INFO_REC gapinfo) {
  char buffer[MAX_LINE_LEN];  /* Temp buffer for reading file */
  char type[3];      /* Loop type, eg. BCB, HCB, etc. */
  int num, len;      /* Loop number and coil length */


  int candidate = -1;    /* Best candidate for the gap */
  double score_to_beat;    /* The best score so far */

  double daa, dcc, dnn, dists[25];
  
  /* Let's just set it as a really high value to start with */
  //score_to_beat = 1000;
  score_to_beat = 999;
  //printf("..xiaoli xiaoli loop pre %d, post %d, len= %d\n", gapinfo.prev_fInd,gapinfo.post_fInd, gapinfo.loop_len);
  
  //added by xiaoli
  if(gapinfo.loop_len == 1 && gapinfo.size > gapinfo.loop_len){
    candidate = -1;
  }
  else{
    while (fgets(buffer, MAX_LINE_LEN, fp) != NULL) {
      if (buffer[0] == '>') {
	
	//parse_data_line(buffer, &num, &len, type, &elbow, &fist);
	parse_data_line1(buffer, &num, &len, type,dists);

	//sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
	/* Lower the score, the closer the fit */
	//score = rank_score(gapinfo, len, type, elbow, fist);
	if(len == gapinfo.loop_len){
	  daa = gapinfo.dists[0] - dists[0];
	  dcc = gapinfo.dists[6] - dists[6];
	  dnn = gapinfo.dists[12] -dists[12];
	
	  if(daa < 0.5 && dcc <0.5 && dnn <0.5){
	    candidate = num; 
	    break;
	  }
	}
	
      }
    } /* END while != NULL */
    
    
    //if(score_to_beat>distancecutoff)
    //  {
    //	candidate=-1;
    // }
  }
  
  return candidate;
} /* end get_candidates() */



/*-----------------------------------------------
 * Parses a data header line from the looppdb file
 * for the following fields.
 *----------------------------------------------- */
int parse_data_line(char *buffer, int *num, int *len, char *type, 
                    float *elbow, float *fist) {
  int num_temp, len_temp;
  float elbow_temp, fist_temp;

  sscanf(buffer, ">%d:%d:%3s:%f:%f", &num_temp, &len_temp, type, &elbow_temp, &fist_temp);
  *num = num_temp; *len = len_temp; *elbow = elbow_temp; *fist = fist_temp;

  return 0;
} /* end parse_data_line() */

/*-----------------------------------------------
 * Parses a data header line from the looppdb file
 * for the following fields.
 *----------------------------------------------- */
void parse_data_line1(char *buffer, int *num, int *len, char *type, double dists[]) {
  int num_temp, len_temp;

  float d0, d1, d2, d3, d4, d5, d6, d7, d8, d9, d10, d11, d12, d13, d14, d15, d16, d17, d18, d19, d20, d21, d22, d23, d24;
  
  sscanf(buffer, ">%d:%d:%3s:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f:%f", &num_temp, &len_temp, type, &d0, &d1,&d2,&d3,&d4,&d5,&d6,&d7,&d8,&d9,&d10,&d11,&d12,&d13,&d14,&d15,&d16,&d17,&d18,&d19,&d20,&d21,&d22,&d23,&d24);
  *num = num_temp;
  *len = len_temp;
  
  dists[0] = d0;
  dists[1] = d1;
  dists[2] = d2;
  dists[3] = d3;
  dists[4] = d4;
  dists[5] = d5;
  dists[6] = d6;
  dists[7] = d7;
  dists[8] = d8;
  dists[9] = d9;
  dists[10] = d10;
  dists[11] = d11;
  dists[12] = d12;
  dists[13] = d13;
  dists[14] = d14;
  dists[15] = d15;
  dists[16] = d16;
  dists[17] = d17;
  dists[18] = d18;
  dists[19] = d19;
  dists[20] = d20;
  dists[21] = d21;
  dists[22] = d22;
  dists[23] = d23;
  dists[24] = d24;

} /* end parse_data_line() */

/*-----------------------------------------------
 * Ranks loops from looppdb to in order of 
 * suitability. The lower the returned score,
 * the more suitable the candidate is.
 *----------------------------------------------- */
double rank_score(GAP_INFO_REC gapinfo, int len, char *type, float elbow, float fist) {
  double score;
  double fist_diff;

  fist_diff = gapinfo.fistdist - fist;
  if (fist_diff < 0) { fist_diff*=-1; }

  if (gapinfo.loop_len != len) {
    return -1; 
  }

  /* let's try another ranking scheme */
  score = fist_diff;

  return score;
} /* end rank_score() */
/*-----------------------------------------------
 * Ranks loops from looppdb to in order of 
 * suitability. The lower the returned score,
 * the more suitable the candidate is.
 *----------------------------------------------- */
double rank_score1(GAP_INFO_REC gapinfo, int len, char *type,double dists[]) {
  
  double score;
  //printf("len = %d\ \t %d\n", len, gapinfo.loop_len);
  
  if (gapinfo.loop_len != len) {
    return -1; 
  }
  
  score = cal_coefficient(gapinfo.dists, dists);
  //printf("score = %g\n", score);
  return score;
  
  
} /* end rank_score() */


/*-----------------------------------------------
 * Ranks loops from looppdb to in order of 
 * suitability. The lower the returned score,
 * the more suitable the candidate is.
 *----------------------------------------------- */
double rank_score2(GAP_INFO_REC gapinfo, int len, char *type,double dists[]) {
  
  double score;
  //printf("len = %d\ \t %d\n", len, gapinfo.loop_len);
  
  if (gapinfo.loop_len != len) {//Error handling
    return -1; 
  }
  
  score = cal_sum_dif(gapinfo.dists, dists);
  //printf("score = %g\n", score);
  return score;
  
  
} /* end rank_score() */

/*-----------------------------------------------
 * Ranks loops from looppdb to in order of 
 * suitability. The lower the returned score,
 * the more suitable the candidate is.
 *----------------------------------------------- */
double rank_score3(GAP_INFO_REC gapinfo, int len, char *type,double dists[]) {
  
  double score, daa;
  //printf("len = %d\ \t %d\n", len, gapinfo.loop_len);
  
  if (gapinfo.loop_len != len) {
    return -1; 
  }
  
  daa = 
  
  score = cal_sum_dif(gapinfo.dists, dists);
  //printf("score = %g\n", score);
  return score;
  
  
} /* end rank_score() */



/*-----------------------------------------------
 * Dumps the information given from the gap info record
 * passed in to a file 
 *----------------------------------------------- */
int dump_gaps_to_file(ALIGN_REC *align, char *filename) {
  int i;
  FILE *gapF;                   /* gap information file */

  gapF = open_file(filename,"w");
  /* print header */
  fprintf(gapF,"*****************************************************\n");
  fprintf(gapF, "Finding gaps for %s : %s\n", align->id,
	  align->title);
  fprintf(gapF, "Total alignment sequence length is %d.\n", align->len); 
  fprintf(gapF, "Query alignment sequence offset is %d.\n", align->query_offset); 

  /* dump gap info */
  for (i=1; i<=align->gapCount; i++) {
    switch (align->gaps[i].type) {
    case DELETION:
      fprintf(gapF, "GAP %d:\tType = Deletion\n", i);
      break;
    case INSERTION:
      fprintf(gapF, "GAP %d:\tType = Insertion\n", i);
      break;
    default:
      fprintf(gapF, "GAP %d:\tType = Unknown\n", i);
      break;
    } /* End case */

    switch (align->gaps[i].gap_class) {
    case HCH:
      fprintf(gapF, "\tClass = HCH\n");
      break;
    case BCB:
      fprintf(gapF, "\tClass = BCB\n");
      break;
    case HCB:
      fprintf(gapF, "\tClass = HCB\n");
      break;
    case BCH:
      fprintf(gapF, "\tClass = BCH\n");
      break;
    case BBB:
      fprintf(gapF, "\tClass = BBB\n");
      break;
    case HHH:
      fprintf(gapF, "\tClass = HHH\n");
      break;
    default:
      fprintf(gapF, "\tClass = Unknown\n");
      break;
    } /* End case */

    fprintf(gapF, "\tGap Size       = %d\n", align->gaps[i].size);
    fprintf(gapF, "\tGap Beg        = %d\n", align->gaps[i].beg);
    fprintf(gapF, "\tGap End        = %d\n", align->gaps[i].end);
    fprintf(gapF, "\tisBeg          = %d\n", align->gaps[i].isBeg);
    fprintf(gapF, "\tisEnd          = %d\n", align->gaps[i].isEnd);
    fprintf(gapF, "\tLoop Length    = %d\n", align->gaps[i].loop_len);
    //fprintf(gapF, "\tPrevEInd       = %d\n", align->gaps[i].prev_eInd);
    //fprintf(gapF, "\tPostEInd       = %d\n", align->gaps[i].post_eInd);
    //fprintf(gapF, "\tGap Elbow Dist = %f\n", align->gaps[i].elbowdist);
    fprintf(gapF, "\tPrevFInd       = %d\n", align->gaps[i].prev_fInd);
    fprintf(gapF, "\tPostFInd       = %d\n", align->gaps[i].post_fInd);
    fprintf(gapF, "\tGap Fist Dist  = %f\n", align->gaps[i].fistdist );
    
    if ( ! align->gaps[i].candidate )
      align->gaps[i].candidate = 0;
    fprintf(gapF, "\tCandidate      = %d\n", align->gaps[i].candidate );
      
  } /* end for */
  fprintf(gapF,"*****************************************************\n");

  fflush(gapF);
  fclose(gapF);

  return 0;
} /* end dump_gaps_to_file() */


/*----------------------------------------------------------------
 * calculate all the distance between pre_fist ca, c, N, O and 
 post_fist ca, c, N, O. Put the result in the dists array.
 * Distance = square_root((x2-x1)^2 + (y2-y1)^2 + (z2-z1)^2)
 *---------------------------------------------------------------- */
void calc_dists(SVECTOR beg, SVECTOR end, double dists[]) {
  VECTOR diff;
 
#if 0 
  
  float length;
	
  length = 0.00;
  	  
  diff.x = end.Ax - beg.Ax; 
  diff.y = end.Ay - beg.Ay;  
  diff.z = end.Az - beg.Az;
   
  diff.x = diff.x*diff.x;  diff.y = diff.y*diff.y;  diff.z = diff.z*diff.z;
  length = sqrt(diff.x + diff.y + diff.z);
  
 // return(length);
  
#endif  
  
  

  double len;
//  int i = 0;
#ifdef DEBUG
  //printf("beg %f %f %f\n", beg.Ax, beg.Ay,beg.Az);
  //printf("beg %f %f %f\n", beg.Cx, beg.Cy,beg.Cz);
  //printf("beg %f %f %f\n", beg.Nx, beg.Ny,beg.Nz);
  //printf("beg %f %f %f\n", beg.Ox, beg.Oy,beg.Oz);
 
  //printf("end %f %f %f\n", end.Ax, end.Ay,end.Az);
  //printf("end %f %f %f\n", end.Cx, end.Cy,end.Cz);
  //printf("end %f %f %f\n", end.Nx, end.Ny,end.Nz);
  //printf("end %f %f %f\n", end.Ox, end.Oy,end.Oz);
#endif

  //aa
  diff.x=end.Ax-beg.Ax; diff.y=end.Ay-beg.Ay; diff.z=end.Az-beg.Az;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[0] = len;

  //ac
  diff.x=end.Cx-beg.Ax; diff.y=end.Cy-beg.Ay; diff.z=end.Cz-beg.Az;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[1] = len;

  //an
  diff.x=end.Nx-beg.Ax; diff.y=end.Ny-beg.Ay; diff.z=end.Nz-beg.Az;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[2] = len;

  //ao
  diff.x=end.Ox-beg.Ax; diff.y=end.Oy-beg.Ay; diff.z=end.Oz-beg.Az;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[3] = len;

  //ab
  diff.x=end.Bx-beg.Ax; diff.y=end.By-beg.Ay; diff.z=end.Bz-beg.Az;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[4] = len;
  
  //ca
  diff.x=end.Ax-beg.Cx; diff.y=end.Ay-beg.Cy; diff.z=end.Az-beg.Cz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[5] = len;

  //cc
  diff.x=end.Cx-beg.Cx; diff.y=end.Cy-beg.Cy; diff.z=end.Cz-beg.Cz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[6] = len;
  
  //cn
  diff.x=end.Nx-beg.Cx; diff.y=end.Ny-beg.Cy; diff.z=end.Nz-beg.Cz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[7] = len;

  //co
  diff.x=end.Ox-beg.Cx; diff.y=end.Oy-beg.Cy; diff.z=end.Oz-beg.Cz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[8] = len;

  //cb
  diff.x=end.Bx-beg.Cx; diff.y=end.By-beg.Cy; diff.z=end.Bz-beg.Cz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[9] = len;


  //na
  diff.x=end.Ax-beg.Nx; diff.y=end.Ay-beg.Ny; diff.z=end.Az-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[10] = len;

  //nc
  diff.x=end.Cx-beg.Nx; diff.y=end.Cy-beg.Ny; diff.z=end.Cz-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[11] = len;

  //nn
  diff.x=end.Nx-beg.Nx; diff.y=end.Ny-beg.Ny; diff.z=end.Nz-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[12] = len;

  //no
  diff.x=end.Ox-beg.Nx; diff.y=end.Oy-beg.Ny; diff.z=end.Oz-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[13] = len;

  //nb
  diff.x=end.Bx-beg.Nx; diff.y=end.By-beg.Ny; diff.z=end.Bz-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[14] = len;

  

  //oa
  diff.x=end.Ax-beg.Ox; diff.y=end.Ay-beg.Oy; diff.z=end.Az-beg.Oz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[15] = len;

  //oc
  diff.x=end.Cx-beg.Ox; diff.y=end.Cy-beg.Oy; diff.z=end.Cz-beg.Oz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[16] = len;

  //on
  diff.x=end.Nx-beg.Ox; diff.y=end.Ny-beg.Oy; diff.z=end.Nz-beg.Oz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[17] = len;

  //oo
  diff.x=end.Ox-beg.Ox; diff.y=end.Oy-beg.Oy; diff.z=end.Oz-beg.Oz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[18] = len;
  
  //ob
  diff.x=end.Bx-beg.Ox; diff.y=end.By-beg.Oy; diff.z=end.Bz-beg.Oz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[19] = len;
  
  //ba
  diff.x=end.Ax-beg.Bx; diff.y=end.Ay-beg.By; diff.z=end.Az-beg.Bz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[20] = len;

  //bc
  diff.x=end.Cx-beg.Bx; diff.y=end.Cy-beg.By; diff.z=end.Cz-beg.Bz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[21] = len;

  //bn
  diff.x=end.Nx-beg.Bx; diff.y=end.Ny-beg.By; diff.z=end.Nz-beg.Bz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[22] = len;

  //bo
  diff.x=end.Ox-beg.Bx; diff.y=end.Oy-beg.By; diff.z=end.Oz-beg.Bz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[23] = len;
  
  //bb
  diff.x=end.Bx-beg.Bx; diff.y=end.By-beg.By; diff.z=end.Bz-beg.Bz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;
  len = sqrt(diff.x + diff.y + diff.z);
  dists[24] = len;


  //for(i = 0; i < 25; i++){
	//printf("%g \t", dists[i]);
  	//printf("\n");
  //}
} /* end calc_dists() */


//********************************************************
float calc_dist(SVECTOR beg, SVECTOR end){
  POINT diff;
  float length;
	
  length = 0.00;
  	  
  diff.x = end.Cx - beg.Cx; 
  diff.y = end.Cy - beg.Cy;  
  diff.z = end.Cz - beg.Cz;
   
  diff.x = diff.x*diff.x;  diff.y = diff.y*diff.y;  diff.z = diff.z*diff.z;
  length = sqrt(diff.x + diff.y + diff.z);
  
  return(length);
  
} /* END calc_len() */
/*----------------------------------------------------------------
 * Calculates the length between two points  based on the formula:
 * Distance = square_root((x2-x1)^2 + (y2-y1)^2 + (z2-z1)^2)
 *---------------------------------------------------------------- */
double calc_len(SVECTOR beg, SVECTOR end) {
  VECTOR diff;
  double len;

#ifdef DEBUG
  //printf("%f %f %f\n", beg.Nx, beg.Ny,beg.Nz);
  //printf("%f %f %f\n", end.Cx, end.Cy,end.Cz);
#endif

  diff.x=end.Cx-beg.Nx; diff.y=end.Cy-beg.Ny; diff.z=end.Cz-beg.Nz;
  diff.x=diff.x*diff.x; diff.y=diff.y*diff.y; diff.z=diff.z*diff.z;

  len = sqrt(diff.x + diff.y + diff.z);
  return len;
} /* end calc_len() */

/*----------------------------------------------------------------
 * Calculates the midpoint between two points based on the formula:
 * Midpoint :  x = (x1+x2)/2 y = (y1+y2)/2 z = (z1+z2)/2
 *---------------------------------------------------------------- */
void calc_midpoint(POINT_t a, POINT_t b, POINT_t *mid) {

  mid->x = (a.x+b.x)/2;
  mid->y = (a.y+b.y)/2;
  mid->z = (a.z+b.z)/2;

} /* END calc_midpoint() */


/*----------------------------------------------------------------
 * Reads one atom from the given pdb file, excluding H atoms.
 *---------------------------------------------------------------- */
int read_pdb(FILE *pdbfp, ATOM_INFO_REC *atom) {
  char buffer[MAX_LINE_LEN];
  char tag[MAX_LINE_LEN];

  if (fgets(buffer, MAX_LINE_LEN, pdbfp) != NULL) 
  {
    if(buffer[0]!='>')
    {
/*	sscanf(buffer, "%4s", tag);
        atom->atomName[0] = 0;
        sscanf(&buffer[12], "%4s", atom->atomName);
        atom->resNum = 0;
        sscanf(&buffer[22], "%d", &atom->resNum);
        sscanf(&buffer[30], "%f%f%f",&atom->x, &atom->y, &atom->z);
*/
      sscanf(buffer, "%4s %*d %s %*s %d %f%f%f", 
                    tag, atom->atomName, &atom->resNum, &atom->x, &atom->y, &atom->z);
  
      if (strcmp(tag, "ATOM")==0) {
        return 1;
      } else if (strcmp(tag, "END")==0) {
        return 0;
      } else {
        fprintf(stderr, "ERROR: unrecognized PDB file format! \"%s\"\n", buffer);
        return 0;
      }
    }
  }
  else {
    return 0;
  }

  return 1;
} /* end read_pdb() */

/*----------------------------------------------------------------
 * Finds and fills gaps.
 *---------------------------------------------------------------- */
void do_gap_findin_and_fillin(ALIGN_REC *topAligns, int topAlignNum,
			      PROGRAM_PARAMS params) {

  int  i;
  FILE *pdbtmpfp;

  char dump_file[10];
  char usrfilename[100];
  char pdbfile_name[100];           /* name of PDB database file template */


  /* Kludge.. need to figure out what pdb file name user chose */
  /// TODO This looks like it's reading in variables from a temp file!
  pdbtmpfp  = fopen(params.pdbtmp, "r");
  fscanf(pdbtmpfp, "%s", usrfilename); 
  fclose(pdbtmpfp);

  /* Let's calculate the gaps for all of the alignments */
  for (i=0; i<topAlignNum; i++) {
    sprintf(pdbfile_name,"%s%d.pdb", usrfilename, i+1);
    topAligns[order[i]].query_offset=0;
    /*set_query_database_offset(&topAligns[order[i]]);*/
    find_gaps(&topAligns[order[i]], pdbfile_name);
    topAligns[order[i]].query_offset=0;
    fill_gaps(&topAligns[order[i]], pdbfile_name, params.shifty_home, params.pdbdir, params.simplexFilename); 
    
    sprintf(dump_file, "%s%cgap%d.nfo", params.tmpdir, '/', i+1);
    
 
    dump_gaps_to_file(&topAligns[order[i]],dump_file);
  }
  /* END gap calc stuff */

} /* END do_gap_findin_and_fillin() */


/* END _loop.c_ */

/*
 * ------------------------------------------------------------
 * MOUSTAFA:
 * This Function changes the size of a gap that is size 1 or 2
 * which allows get_candidates1 to act on a larger gap, finding 
 * a larger sized loop
 * ------------------------------------------------------------
*/
void refindgap(ALIGN_REC *align,int gap,char *pdb_filename)
{
  int lengthChange = 0;

  if(align->gaps[gap].type==INSERTION)
    {
      //if (align->gaps[gap].loop_len<(2*align->gaps[gap].size)) {
	if (align->gaps[gap].prev_fInd > 0) {
	  align->gaps[gap].prev_fInd--;
	  lengthChange++;
	}
	align->gaps[gap].post_fInd++;
	lengthChange++;
	align->gaps[gap].loop_len += lengthChange;
	calc_gaps_fist_elbow_dist(align, pdb_filename);
	//} else {
	//distancecutoff+=0.5;
	//}
    }
  if(align->gaps[gap].type==DELETION)
    {
      //printf("happy......................\n");
      //if (align->gaps[gap].loop_len<align->gaps[gap].size) {
	if (align->gaps[gap].prev_fInd > 0) {
	  align->gaps[gap].prev_fInd--;
	  lengthChange++;
	}
	align->gaps[gap].post_fInd++;
	lengthChange++;
	align->gaps[gap].loop_len += lengthChange;
	calc_gaps_fist_elbow_dist(align, pdb_filename);
      

	//} else {
	//distancecutoff+=0.5;
	//}
    }
}

int converttoint(char *type)
{

  if(strcmp(type, "BCH")==0)
    return BCH;
  if(strcmp(type, "BCB")==0)
    return BCB;
  if(strcmp(type, "HCH")==0)
    return HCH;
  if(strcmp(type, "HCB")==0)
    return HCB;
  if(strcmp(type, "BBB")==0)
    return BBB;
  if(strcmp(type, "HHH")==0)
    return HHH;
  return 0;
}



/*-----------------------------------------------
 * Replaces the requested gap with the given
 * candidate based on the info found in the alignment
 * record.
 *----------------------------------------------- */
void get_pre_post_loop_pdbfile(int candidate,char *pdbdir, char *shifty_home) {
  
  FILE *dbp;    /* New PDB file with loops */
  int num;
  char dbatom[MAX_NAME_LEN], res[MAX_NAME_LEN];
  
  char buffer[MAX_LINE_LEN];   /* holds input lines as we stroll thru file */
  long int position=-1;
  float  tmpx,tmpy,tmpz;
  
  int len;                 /* Loop number and coil length */

  char looppdb_db[MAX_FNAME_LEN];
  char type[3];                 /* Loop type, eg. BCB, HCB, etc. */

 
  int init_base;
  
  sprintf(looppdb_db, "%s%c%s", shifty_home, '/', LOOPPDB_DB);
  
  dbp = fopen(looppdb_db, "rb");
  
  /* search the database for the position to begin the loop */
  while (fgets(buffer, MAX_LINE_LEN, dbp) != NULL) {
    
    /* All loop entries in database has the format '>num' */
    if (buffer[0] == '>') {
      //parse_data_line(buffer, &num, &len, type, &elbow, &fist);
      //parse_data_line1(buffer, &num, &len, type);
      sscanf(buffer, ">%d:%d:%3s", &num, &len, type);
    }
    if (num == candidate) 
      break;  /* stop looking for loop now */  
  }
     
  fgets(buffer, MAX_LINE_LEN, dbp);
  sscanf(buffer, "%d%s%s%*s%*s%f%f%f", &init_base, res, dbatom, &tmpx, &tmpy, &tmpz);
      

  int res_num = -1;
  int pre_loop = init_base +2;
  
  int after_loop = pre_loop + len + 1;
  int temp_count = 0;
  
  FILE *target_loop = fopen(PRE_POST_LOOP_PDB, "w");
  
  while(fgets(buffer,MAX_LINE_LEN, dbp)!=NULL){
    
    sscanf(buffer, "%d%s%s%*s%*s%f%f%f", 
	   &res_num, res, dbatom, &tmpx, &tmpy, &tmpz);
    
    //printf("ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
    //     1,dbatom, res, res_num, tmpx, tmpy, tmpz,LOOP_TAR);
    
    if(res_num >after_loop ){
      break;
      
    }
    else if(res_num >= pre_loop && res_num <= after_loop){
      temp_count++;
      
      fprintf(target_loop,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
	      temp_count, dbatom, res, res_num, tmpx, tmpy, tmpz,PRE_POST_LOOP_PDB);
    }
      
  }
  
  fseek(dbp,position,SEEK_SET);
  fclose(target_loop);
  
}

/***************** end   xiaoli *******************************************
    
    grab loop info from the loop database and put it in a file
    loop residue: 3456
    grab residue 234567 from loop database 

**********************************************************************/

/*-----------------------------------------------
 * Replaces the requested gap with the given
 * candidate based on the info found in the alignment
 * record.
 *----------------------------------------------- */
int  get_pre_post_pdb(ALIGN_REC *align,int gap, char *src_filename) {
  FILE *srcfp;    /* Existing PDB file */


  //char tmp_filename[MAX_FNAME_LEN]; /* temp filename for renaming purposes */

  int start_index, finish_index;
  int num, count;
  
  char buffer[MAX_LINE_LEN];   /* holds input lines as we stroll thru file */

  float  tmpx,tmpy,tmpz;
  FILE *pre_post_pdb = fopen(PRE_POST_PDB, "w");
  char atom[MAX_NAME_LEN], res[MAX_NAME_LEN];
  
  num=0;
  start_index  = align->gaps[gap].prev_fInd - align->query_offset;  
  finish_index = start_index + align->gaps[gap].loop_len+1;
  
  srcfp = open_file(src_filename, "r");

  while (fgets(buffer, MAX_LINE_LEN, srcfp) != NULL) {
    
    if(buffer[0] == 'E'){
      break;
    }
    
    
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    //if(num == 35)
    //printf("%%%%%%%%%%%%%%%%%%%%%% = %d, %s, %s,%f, %f, %f\n", count, atom, res,tmpx, tmpy, tmpz);
    
    if(strcmp(atom, "N") == 0 || strcmp(atom, "CA") == 0 || strcmp(atom, "C") == 0 || strcmp(atom, "O") == 0){
      /* Detect if we are in the loop replacement section of file */
      if (num == start_index) {
	
	fprintf(pre_post_pdb,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
		count,atom, res,num, tmpx, tmpy, tmpz,PRE_POST_PDB);
	//printf("start ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
	//count,atom, res,num, tmpx, tmpy, tmpz,PRE_POST_PDB);
      }
      else if (num == finish_index) {
	
	fprintf(pre_post_pdb,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
		count,atom, res,num, tmpx, tmpy, tmpz,PRE_POST_PDB);
      
	//printf("end ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
	//count,atom, res,num, tmpx, tmpy, tmpz,PRE_POST_PDB);
      }
    }
      
  }
  
  fclose(srcfp);
  fclose(pre_post_pdb);
  return 0;
} /* END insert_into_pdbfile() */

void  insert_loop_pdb(char *src_filename, FILE *tarfp,char *superpose, int start_index, int finish_index, ALIGN_REC *align,int gap) {
  
  float  tmpx,tmpy,tmpz;
  int num, count;
    
  char atom[MAX_NAME_LEN], res[MAX_NAME_LEN];
    
  FILE *src = fopen(src_filename, "r");
  FILE *input = fopen(superpose, "r");
  
  
  char buffer[MAX_LINE_LEN];
  
  int atom_num = 0;
  int loop_begin_residue = 0;
  int loop_finish_residue = 0;
  int loop_residue_num = 0;
  char aa;
  char aa_name[3];
  char *seq;
    
  int pos = 0;
  int resCount1 = 0;
  int resCount2 = 0;
  int resCount = 0;

  printf("src_filename=%s,superpose=%s\n",src_filename,superpose );
 
  seq = align->qpRes;
  
  while( pos < (align->gaps[gap]).prev_fInd){
  
    if(seq[pos] != ' ')
      resCount1++;
    pos++;
  }

 
  resCount2 = resCount1;

  while( pos < (align->gaps[gap]).post_fInd){
  
    if(seq[pos] != ' ')
      resCount2++;
    pos++;
  }
 
  if(fgets(buffer, MAX_LINE_LEN,input) != NULL) {
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    loop_begin_residue = num + 1; 
    loop_finish_residue = num + align->gaps[gap].loop_len;
  }
 
  
  
  while (fgets(buffer, MAX_LINE_LEN, src) != NULL) {

    if(buffer[0] == 'E'){
      break;
    }
    
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
   
    if(num <= resCount1){
    
      atom_num++;
      fprintf(tarfp,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",atom_num, atom, res,num, tmpx, tmpy, tmpz,src_filename);

      #ifdef DEBUG
      fflush(tarfp);
      #endif
      //printf("ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",atom_num, atom, res,num, tmpx, tmpy, tmpz,src_filename);
      resCount = num;
      continue;
    }
    
    //loop inserting with insertion

    //else if(num > resCount1 && num < resCount2){
    resCount = resCount + 1;
    
    if(resCount >= resCount1 && resCount < resCount2){ 
    
      //resCount = resCount1+1;
      
      pos = (align->gaps[gap]).prev_fInd;

      while(resCount < resCount2){
		
	while( pos < (align->gaps[gap]).post_fInd){
	  
	  if(seq[pos] != ' '){
	  
	    aa = seq[pos];
	    convert (aa_name,aa);
	    pos++;
	    
	    
	    break;
	  }
	  else{
	    pos++;
	  }
	}
	
	atom_num = get_coords(atom_num,loop_begin_residue,resCount, aa_name,superpose, tarfp, src_filename);
	resCount++;
	
	loop_begin_residue++;
	
      }
  
      // skip the corresponding parts of the src file that we just inserted
      while(num > resCount1 && num < resCount2 && fgets(buffer, MAX_LINE_LEN, src) != NULL){
	sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
	
      }
      
    }
    
    if(num >= resCount2){
      atom_num++;
      fprintf(tarfp,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",atom_num, atom, res,num, tmpx, tmpy, tmpz,src_filename);

      #ifdef DEBUG
      fflush(tarfp);
      #endif
      
      loop_residue_num++;
    }
    
  }
  
  fclose(input);
  fclose(src);
}

/**-----------------------------------------------
 * Goes and fetches the (x,y,z) coordinates from
 * the given loop at the given residue of the 
 * given atom from loop pdb database.

  @return the new atom index after insertion
 **----------------------------------------------- */
int get_coords(int count,int loop_begin, int residue_num, char *res, char *superpose_result, FILE *tmp, char *src_filename) {
  
  FILE *loop;
  char buffer[MAX_LINE_LEN];
  int num;                 /* Loop number and coil length */
  
  float tmpx, tmpy, tmpz;
  char atom[MAX_NAME_LEN];
  int temp_count = count;
 
  loop = fopen(superpose_result, "r");

  while (fgets(buffer, MAX_LINE_LEN, loop) != NULL) {
    
    sscanf(buffer, "%*s%*d%s%*s%d%f%f%f",atom,&num,&tmpx, &tmpy, &tmpz);
 
    if (num == loop_begin) {
      
      temp_count++;
      //printf("ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
      //  temp_count, atom, res, residue_num, tmpx, tmpy, tmpz,src_filename);  
      fprintf(tmp,"ATOM %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f  1.00  1.00 %12s\n",
            temp_count, atom, res, residue_num, tmpx, tmpy, tmpz,src_filename);
  

      #ifdef DEBUG
      fflush(tmp);
      #endif
    }
    
  }
  
  fclose(loop);
  return temp_count;
} 

/*
  Does a distance calculation of the loop start to the gap opening, 
  and loop end to gap closing, to see if it's a good match
*/
int simplex_needed(ALIGN_REC *align, int gap, char *loop_pdb, char *pre_post_pdb, double rmsd, double *cn, double *nc){
  
  float  tmpx,tmpy,tmpz;
  int num, count;
    
  char atom[MAX_NAME_LEN],res[MAX_NAME_LEN], isatom[10];
  POINT_t pt_gap_open, pt_loop_start, pt_loop_end, pt_gap_close;

  // Initialize points
  memset(&pt_gap_open, 0, sizeof(POINT_t));
  memset(&pt_loop_start, 0, sizeof(POINT_t));
  memset(&pt_loop_end, 0, sizeof(POINT_t));
  memset(&pt_gap_close, 0, sizeof(POINT_t));

  
  FILE *loop = fopen(loop_pdb, "r");
  char buffer[MAX_LINE_LEN];
  int loop_begin_residue = 0;
  int loop_finish_residue = 0;
  int pre=0;
  
  if(fgets(buffer, MAX_LINE_LEN,loop) != NULL) {
    
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    loop_begin_residue = num + 1; 
    loop_finish_residue = num + align->gaps[gap].loop_len;
  }

  // go back to the beginning of the file otherwise we might miss an atom
  rewind(loop);
  
  // p2 == coordinates of the N atom of the beginning of the loop
  // p3 == coordinates of the C atom of the end of the loop
  while(fgets(buffer, MAX_LINE_LEN,loop) != NULL) {
    
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    
    // Use the N-terminus of the beginning of the loop
    if(num == loop_begin_residue && strcmp(atom, "N") == 0 )
    {
      pt_loop_start.x = tmpx;
      pt_loop_start.y = tmpy;
      pt_loop_start.z = tmpz;
    }
    // Use the C-terminus of the end of the loop
    else if(num == loop_finish_residue && strcmp(atom, "C") == 0 )
    {
      pt_loop_end.x = tmpx;
      pt_loop_end.y = tmpy;
      pt_loop_end.z = tmpz;
    }
  }

  fclose(loop);

  FILE *fist = fopen(pre_post_pdb, "r");
  
  if (fgets(buffer, MAX_LINE_LEN,fist) != NULL){
    sscanf(buffer, "%*s%d%s%s%d%f%f%f",&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    pre = num;
  }
  
  // go back to the beginning of the fist file
  rewind(fist);

  if(strcmp("C", atom) == 0){
    
    pt_gap_open.x = tmpx;
    pt_gap_open.y = tmpy;
    pt_gap_open.z = tmpz;
  }

  
  // pre_post.pdb stores the two residues that are the open and close of the gap
  // p1 == coordinates of the gap opening residue
  // p4 == coordinates of the gap closing residue
  while (fgets(buffer, MAX_LINE_LEN,fist) != NULL) {
    
    sscanf(buffer, "%s%d%s%s%d%f%f%f",isatom,&count, atom, res,&num,&tmpx, &tmpy, &tmpz );
    
    // Use the C-terminus of the gap opening residue
    if(num == pre && strcmp("C", atom) == 0 && strcmp("ATOM", isatom) == 0) 
    {
      pt_gap_open.x = tmpx;
      pt_gap_open.y = tmpy;
      pt_gap_open.z = tmpz;
    }
    // Use the N-terminus of the gap closing residue
    //else if(num != pre && strcmp("N", atom) == 0 && strcmp("ATOM", isatom) == 0 )
    else if(num == pre && strcmp("N", atom) == 0 && strcmp("ATOM", isatom) == 0 )
    {
      pt_gap_close.x = tmpx;
      pt_gap_close.y = tmpy;
      pt_gap_close.z = tmpz;
    }
  }
  // printf("start residue in the pdb C %g, %g, %g end residue N %g, %g, %g\n", p3.x, p3.y, p3.z, p4.x, p4.y, p4.z);
  
  fclose(fist);

  if ( (pt_gap_open.x == 0 && pt_gap_open.y == 0)
       || (pt_loop_start.x == 0 && pt_loop_start.y == 0)
        || (pt_loop_end.x == 0 && pt_loop_end.y == 0)
        || (pt_gap_close.x == 0 && pt_gap_close.y == 0))
      printf("ERROR: One of the points did not get initialized!! Bad pre_post file or result file!\n (In simplex_needed)");

  // Calculate the euclidean distance between the two sets of points
  double x_dist = pt_gap_open.x- pt_loop_start.x;
  double y_dist = pt_gap_open.y- pt_loop_start.y;
  double z_dist = pt_gap_open.z- pt_loop_start.z;
  double gap_open_to_loop_start = sqrt(pow(x_dist, 2) + pow(y_dist, 2) + pow(z_dist, 2));
  
  *cn = gap_open_to_loop_start;

  x_dist = pt_loop_end.x- pt_gap_close.x;
  y_dist = pt_loop_end.y- pt_gap_close.y;
  z_dist = pt_loop_end.z- pt_gap_close.z;  
  double gap_close_to_loop_end = sqrt(pow(x_dist, 2) + pow(y_dist, 2) + pow(z_dist, 2));
  *nc = gap_close_to_loop_end;
  
  // Check if they are within the limits
  if(*cn <= CN_MAX_DIST && *nc <=  CN_MAX_DIST && rmsd <= THRESHOLD_RMSD)
    return 0;
  else
    return 1;
  
}


double cal_coefficient(double array1[], double array2[]){
  
  double ss_xx = 0, ss_yy = 0, ss_xy = 0, sum_xx = 0, sum_yy = 0, sum_xy = 0, sum_x=0, sum_y = 0, mean_x = 0, mean_y = 0, r_square = 0, score=0;
  int i = 0;
  
  for(i = 0; i < 25; i++){
    //printf("xiaoli xiaoli %g, %g\n",array1[i],array2[i] );
    //printf("xiaoli xiaoli %g\n",r_square );
    sum_xx += array1[i] * array1[i];
    sum_yy += array2[i] * array2[i];
    sum_xy += array1[i] * array2[i];
    sum_x += array1[i];
    sum_y += array2[i];
  }
  
  mean_x = sum_x/25;
  mean_y = sum_y/25;
  
  ss_xx = sum_xx - 25 * mean_x * mean_x;
  ss_yy = sum_yy - 25 * mean_y * mean_y;
  ss_xy = sum_xy - 25 * mean_x * mean_y;

  r_square = (ss_xy * ss_xy)/(ss_xx * ss_yy);
  score = sqrt(r_square);
    return score;
}

double cal_sum_dif(double array1[], double array2[]){
  
  double sum = 0;
  int i = 0;
  
  for(i = 0; i < 25; i++){
    sum += fabs(array1[i]-array2[i]);
    
  }
  return sum;
}




int compare_double (const void *v1, const void *v2)
{
  return  (*(struct CAND *)v1).score - (*(struct CAND *)v2).score;
 
}

/*
 * ---------------------------------------------------------------------------
 * Precondition: the current gaps classification is unknown					 
 * Postcondition: the current gaps classification is set					 
 * Moustafa
 *-----------------------------------------------------------------------------*/
 
void set_loop_classification(ALIGN_REC * align, char before, char after, int gap){
	
	if (before == 'H') {
      if (after == 'H') {
        /* check for special cases ie. HHH or BBB */
        if (align->gaps[gap].loop_len + align->gaps[gap].size == 0) {
          align->gaps[gap].gap_class = HHH;
        } else {
          align->gaps[gap].gap_class = HCH;
        }
      } else if (after == 'B') {
        align->gaps[gap].gap_class = HCB;
      } else {
        align->gaps[gap].gap_class = -1;
      }
    } else if (before == 'B') {
      if (after == 'H') {
        align->gaps[gap].gap_class = BCH;
      } else if (after == 'B') {
        /* check for special cases ie. HHH or BBB */
        if (align->gaps[gap].loop_len + align->gaps[gap].size == 0) {
          align->gaps[gap].gap_class = BBB;
        } else {
          align->gaps[gap].gap_class = BCB;
        }
      } else {
        align->gaps[gap].gap_class = -1;
      }
    } else {
      align->gaps[gap].gap_class = -1;
    } /* end loop gap_classification */
}/* Modified by haiyan  */


/*
 *---------------------------------------------------------------------------------
 * Checks to see if the given gap overlaps another gap. If true, the second,
 * overlapped gap is marked as "dontFill", indicating that no loop will be inserted
 * there.
 *---------------------------------------------------------------------------------- 
 */
int envelops_other_gap(ALIGN_REC * align, int gap){
	
	
	boolean envelopsOtherGap = false;
 	 GAP_INFO_REC* this_gap = &align->gaps[gap];
  for (int j=1;j<=align->gapCount;j++) {
    if ( j == gap ) continue;

    GAP_INFO_REC* tmpgap = &align->gaps[j];
    
    if ( this_gap->dontFill ) continue;

    if ( tmpgap->isBeg == 0 && tmpgap->isEnd == 0 )
    {
      // it's not a end of beginning gap
      if ( this_gap->prev_fInd < tmpgap->beg 
        && this_gap->post_fInd > tmpgap->end )
      {
        envelopsOtherGap = true;
        tmpgap->dontFill = true;
      }
    }
  }
  if(envelopsOtherGap== false)
  	return 0;
  else
  	return 1;
  
}

/*
 * ---------------------------------------------------------------
 * Precondition: a gap has been found in the query sequence, this
 * means that the gap is a deletion. 
 * Postcondition: all of the properties of the deletion gap are stored
 * Moustafa
 * ----------------------------------------------------------------*/

void setup_query_gap(ALIGN_REC * align, int gapCount, int i){
        align->gaps[gapCount].size = 1; /* init size */
        align->gaps[gapCount].beg  = i+1;      /* gap start index in the sequence
        										It is "+1" because it is in an array that starts at 0 */
        align->gaps[gapCount].end  = i+1;      /* init gap end index */
        align->gaps[gapCount].type = DELETION; /* gap is an deletion */
        align->gaps[gapCount].loop_len = 0;
        if (align->gaps[gapCount].beg==1) {
          align->gaps[gapCount].isBeg=1;
        } else {
          align->gaps[gapCount].isBeg=0;
        }
        if (align->gaps[gapCount].end==align->len) {
          align->gaps[gapCount].isEnd=1;
        } else {
          align->gaps[gapCount].isEnd=0;
        }
}
 
/*
 * ---------------------------------------------------------------
 * Precondition: a gap has been found in the database sequence, this
 * means that the gap is an insertion. 
 * Postcondition: all of the properties of the insertion gap are stored
 * Moustafa
 * ----------------------------------------------------------------*/

 void setup_database_gap(ALIGN_REC * align, int gapCount, int i){
 		align->gaps[gapCount].size = 1;         /* init size */
        align->gaps[gapCount].beg  = i+1;       /* gap start index */
        align->gaps[gapCount].end  = i+1;       /* init gap end index */
        align->gaps[gapCount].type = INSERTION; /* gap is an insertion */
        align->gaps[gapCount].loop_len = 0;
        if (align->gaps[gapCount].beg==1) {
          align->gaps[gapCount].isBeg=1;
        } else {
          align->gaps[gapCount].isBeg=0;
        }
        if (align->gaps[gapCount].end==align->len) {
          align->gaps[gapCount].isEnd=1;
        } else {
          align->gaps[gapCount].isEnd=0;
        }
}

/*
 * ---------------------------------------------------------------
 * Precondition: a gap has been found to be one size larger then its previous size
 * Postcondition: increase size of gap
 * Moustafa
 * ----------------------------------------------------------------*/
void gap_continuation(ALIGN_REC * align, int gapCount, int i){
 	align->gaps[gapCount].size ++;  /* incr gap size */
    align->gaps[gapCount].end = i+1;  /* update until we leave gap */
    if (align->gaps[gapCount].end==align->len) {
       align->gaps[gapCount].isEnd=1;
    } 
     else {
          align->gaps[gapCount].isEnd=0;
       	 }       
}

int set_prim_loop_ref_pt (const char * loop, POINT * pt) {
  /***********************************************************
   * Searches the given loop file in PDB format for the
   * primary rotation reference point. Guarantee not to change
   * the contents of the file 'loop'.
   ***********************************************************/

  FILE     * fp;
  PDB_LINE   line;
  
  
  fp = fopen (loop, FILE_READ);

  while (read_pdb_line (fp, &line) != EOF) {
    
	   
    if (is_C_atom (line.atom) == true) {  
      pt->x = line.pt.x;  
      pt->y = line.pt.y;
      pt->z = line.pt.z;
      
     
      break;
    }
  } /* END _while_ */

  fclose (fp);

  return (1);

} /* END set_prim_loop_ref_pt () */


/*Updated Function's to loop.c on July 18, 2006*/
int set_sec_loop_ref_pt (const char * loop, POINT * pt) {
  /***********************************************************
   * Searches the given loop file in PDB format for the
   * secondary rotation reference point. Guarantee not to change
   * the contents of the file 'loop'.
   ***********************************************************/

  FILE     * fp;
  PDB_LINE   line;

  fp = fopen (loop, FILE_READ);

  while (read_pdb_line (fp, &line) != EOF) {
    /* keep updating coords.. we want the last set */
   
    if (is_N_atom (line.atom) == true) { 
      pt->x = line.pt.x;  
      pt->y = line.pt.y;  
      pt->z = line.pt.z;  
     
    }
  } /* END _while_ */

  fclose (fp);

  return (1);

} /* END set_sec_loop_ref_pt () */

//*****************************************************************************
float compare_matrices(float loop_dist[][SIZE_OF_LOOP_DISTMATRIX], float oth_strands[][SIZE_OF_LOOP_DISTMATRIX]){
  int r, c;
  float abs_difference=0.00, sum=0.00, average=0.00;
  int size_sqrd;
	
  size_sqrd = (SIZE_OF_LOOP_DISTMATRIX*SIZE_OF_LOOP_DISTMATRIX) ;  
	
  for(r=0; r<SIZE_OF_LOOP_DISTMATRIX; r++){
      for(c=0; c<SIZE_OF_LOOP_DISTMATRIX; c++){	
	  abs_difference = fabs(loop_dist[r][c] - oth_strands[r][c]);
	  sum+=abs_difference;
	}
    }	
	      
  average = sum/size_sqrd;
  return(average);
	
}

/***********************************************************	
 * compares the scores of the distance matrices  
 ***********************************************************/ 
int compare_scores(const void *p, const void *q){
  
  LOOP_PROP *p1 = ( LOOP_PROP *)p;
  LOOP_PROP *q1 = ( LOOP_PROP *)q;
  
  if( (p1->score) < (q1->score) )
    return (-1);
  else if( (p1->score) == (q1->score) )
    return (0);
  else
    return (1);
  
}/* END  compare_scores */

//*******************************************************************************************
// Check for distance violations when the loop has been inserted..
int evaluate_distance(char *pdbfile, E_ARRAY *e_arr ){

  FILE *fp;
  SVECTOR *whole, *loop_cas;
  float len;
  unsigned int i, j, count, min;
  unsigned int CA_counter, loop_CA_counter, first_coil_counter;
  
  PDB_LINE line;
  
  /* Initialize: */
  count = 0;
  min = 99.00;
  first_coil_counter = 0;
  CA_counter = loop_CA_counter = 0;
  
  fp = fopen(pdbfile, FILE_READ);
  if(!fp){printf("Cannot open %s for reading.. \n",pdbfile);exit(-1);}
   

  /* Calculating the total number of CA atoms of the complete file 
     and the CA atoms of the loop. 
  *********************************************/	  
  while (read_pdb_line (fp, &line) != EOF){
    if (is_ref_atom (line.atom) == true){ 
      CA_counter++;
    }
    if(  (is_ref_atom (line.atom) == true) && (is_coil(line.extra) == true  )   ){	
      loop_CA_counter++; 
    }
  }
  rewind (fp);
  /* printf("CA_counter = %d and residue no = %d\n", CA_counter, line.res_no); 
     printf("loop_CA_counter = %d \n", loop_CA_counter); */	  
  
  
  
  
  whole = (SVECTOR*) malloc(CA_counter*sizeof(SVECTOR) );
  if(!whole){err_print(ERR_MEM_ALLOC, "whole in module loop.c");exit(EXIT_FAILURE);}
	  
  loop_cas   = (SVECTOR*) malloc(loop_CA_counter*sizeof(SVECTOR) );
  if(!loop_cas){err_print(ERR_MEM_ALLOC, "loop_cas in module loop.c");exit(EXIT_FAILURE);}
   
   
  while(read_pdb_line (fp, &line) != EOF){
    if(is_ref_atom (line.atom) == true){ 
      whole[line.res_no-1].Cx = line.pt.x; whole[line.res_no-1].Cy = line.pt.y; whole[line.res_no-1].Cz =line.pt.z;
    }
    if( (is_ref_atom (line.atom) == true) && (is_coil(line.extra) == true) ){
      loop_cas[first_coil_counter].Cx = line.pt.x; 
      loop_cas[first_coil_counter].Cy = line.pt.y;
      loop_cas[first_coil_counter].Cz = line.pt.z;
      
      first_coil_counter++;
    }
  }/* END while */
  fclose(fp);
  
  
  
  for(i=0; i<CA_counter; i++){
      for(j=0; j<loop_CA_counter; j++ ){
	  len = calc_dist(whole[i], loop_cas[j]);
                	   
	  if(len < MIN_CA_CA_DIST && len > 0.00){   
	      /*printf("i = %d,  j = %d, and  len = %4.2f, whole[i].CAx = %4.2f\n",i+1,j+1,len, whole[i].CAx);*/
	    //printf("count = %d ",count); 
	      e_arr->dist_between_atoms[count] = len;
	      /*printf("e_arr->dist_between_atoms[%d] = %4.2f\n",count, len);*/
	      count++; 
	      if(len<min)
		min=len;
	    }
			   
	}
    
    } 
   
  e_arr->no_of_conflicts = count;
   
   
  free(whole);
  free(loop_cas);
   
  return (count);
   
} /* END  evaluate_distance() */


void form_CA_dmatrix_of_loop(float dist_matrix[][SIZE_OF_LOOP_DISTMATRIX], SIX_CA_POINTS six_cas[], int seg_number){
  /***********************************************  
   * Forms the CA_dmatrix of the loop that is to be inserted. 
   *  The pairwise distances between the 3 beginning CA atoms and the 
   *  last 3 CA atoms are calculated and stored in the distance matrix.
   **************************************************************/
	
  dist_matrix[0][0] = get_distance( &(six_cas[seg_number].first_CA), &(six_cas[seg_number].last_CA) );
  dist_matrix[0][1] = get_distance( &(six_cas[seg_number].first_CA), &(six_cas[seg_number].sec_last_CA) );
  dist_matrix[0][2] = get_distance( &(six_cas[seg_number].first_CA), &(six_cas[seg_number].third_last_CA) ); 
	
  dist_matrix[1][0] = get_distance( &(six_cas[seg_number].sec_CA), &(six_cas[seg_number].last_CA) ); 
  dist_matrix[1][1] = get_distance( &(six_cas[seg_number].sec_CA), &(six_cas[seg_number].sec_last_CA) );
  dist_matrix[1][2] = get_distance( &(six_cas[seg_number].sec_CA), &(six_cas[seg_number].third_last_CA) );
	
  dist_matrix[2][0] = get_distance( &(six_cas[seg_number].third_CA), &(six_cas[seg_number].last_CA) );
  dist_matrix[2][1] = get_distance( &(six_cas[seg_number].third_CA), &(six_cas[seg_number].sec_last_CA) );
  dist_matrix[2][2] = get_distance( &(six_cas[seg_number].third_CA), &(six_cas[seg_number].third_last_CA) );
	
}/* END ..form_CA_dmatrix.of loop.*/

void form_dmatrix_of_adj_strands(float dist_matrix[][SIZE_OF_LOOP_DISTMATRIX], SIX_CA_POINTS six_cas[], int seg_number){
  /***********************************************  
   * Forms the CA_dmatrix of the two adjacent strands to the loop 
   *  The pairwise distances between the 3 ending CA atoms of one strand and  
   *  the 3 beginning CA atoms are calculated and stored in the distance matrix.
   **************************************************************/			  
  dist_matrix[0][0] = get_distance( &(six_cas[seg_number-1].third_last_CA), &(six_cas[seg_number+1].third_CA) );
  dist_matrix[0][1] = get_distance( &(six_cas[seg_number-1].third_last_CA), &(six_cas[seg_number+1].sec_CA) );
  dist_matrix[0][2] = get_distance( &(six_cas[seg_number-1].third_last_CA), &(six_cas[seg_number+1].first_CA) );

  dist_matrix[1][0] = get_distance( &(six_cas[seg_number-1].sec_last_CA), &(six_cas[seg_number+1].third_CA) );
  dist_matrix[1][1] = get_distance( &(six_cas[seg_number-1].sec_last_CA), &(six_cas[seg_number+1].sec_CA) );
  dist_matrix[1][2] = get_distance( &(six_cas[seg_number-1].sec_last_CA), &(six_cas[seg_number+1].first_CA) );

  dist_matrix[2][0] = get_distance( &(six_cas[seg_number-1].last_CA), &(six_cas[seg_number+1].third_CA) );
  dist_matrix[2][1] = get_distance( &(six_cas[seg_number-1].last_CA), &(six_cas[seg_number+1].sec_CA) );
  dist_matrix[2][2] = get_distance( &(six_cas[seg_number-1].last_CA), &(six_cas[seg_number+1].first_CA) );

}/* END ..form_CA_dmatrix.of adj strand.*/

/***********************************************************
 * Fetches the given loop from the database.
 ***********************************************************/
char *go_get_me_the_loop (long int file_position, int residues, POINT *first_pt, POINT *last_pt,int no_of_fist_resds){
  
  FILE   *dbfp,*tarfp;
  LOOP_b   line;
  char     buffer[MAX_LINE_LEN];
  PDB_LINE  pdb_line;
  boolean   got_it;
  unsigned int    residue_num, prev_residue_num, middle_residue_num, last_residue_num;
 

  /* initialize */
  got_it = false;
  residue_num = prev_residue_num = middle_residue_num = last_residue_num = 0;
  
  dbfp  = fopen(COIL_DAT, FILE_READ);
  if(!dbfp){printf("Cannot open %s for reading. Exiting..\n",COIL_DAT);exit(-1);}
  
  tarfp = fopen(LOOP_TEMP_FILE, FILE_WRITE);
  if(!tarfp){printf("Cannot open %s for writing. Exiting..\n",LOOP_TEMP_FILE);exit(-1);}
  

  fseek(dbfp, file_position, SEEK_SET);
  
  while (fgets (buffer, MAX_LINE_LEN, dbfp) != NULL) {
    if (is_loop_header(buffer) == false) {
      read_loop_database_line (buffer, &line);
                               
      if (line.type == 'C') {
	residue_num = line.res_no;
	middle_residue_num = residue_num + floor(residues/2);
				    
	prev_residue_num = line.res_no - 1 ;
	last_residue_num = (prev_residue_num + residues + 1 );
	break;
      }
    }
  }
		
  fseek(dbfp, file_position, SEEK_SET);
		
  while (fgets (buffer, MAX_LINE_LEN, dbfp) != NULL){ 
    if (is_loop_header(buffer) == false){ 
      read_loop_database_line (buffer, &line);
	
      if( (line.res_no == prev_residue_num )  ){
	if (is_ref_atom (line.atom) == true){ 
	  assign_pt(first_pt, line.pt);
	}
      }
			
      if( (line.res_no == last_residue_num )  ){
	if (is_ref_atom (line.atom) == true){
	  assign_pt(last_pt, line.pt);
	}
      }

      //  LOOK HERE FOR SELECTING THE ANCHOR RESIDUES	
      //Use the following for only two fist residues..	   
      if(no_of_fist_resds == 2){
	if( (line.res_no >= prev_residue_num ) && (line.res_no <= (prev_residue_num + residues + 1 )) ){
	  create_pdb_line_from_loop_line (&pdb_line, line);
	  write_pdb_line (tarfp, pdb_line);
	}
      }
      //Use the following for all six fist residues..
      if(no_of_fist_resds == 6){
	if((line.res_no >= prev_residue_num -2 ) && (line.res_no <= (prev_residue_num + residues + 1 +2)) ){
	  create_pdb_line_from_loop_line (&pdb_line, line);
	  write_pdb_line (tarfp, pdb_line);
	}
      }
    } /* END if (is_loop_header... */
    else{ 
      break;
    }
  } /* END while (fgets (buffer,.... */ 
                
  fflush(tarfp); fclose (tarfp);
  fclose (dbfp);

  return (LOOP_TEMP_FILE);

} /* END go_get_me_the_loop () */


boolean is_loop_header (char * line) {
  /***********************************************************
   * Returns true if the given line matches the header line
   * pattern of the loop database.
   ***********************************************************/

  if (line[0] == '>') { return true;  }
  else                { return false; }

} /* END is_loop_header () */

int create_pdb_line_from_loop_line (PDB_LINE * pdb, LOOP_b loop) {
  /***********************************************************
   * Takes in a line from the loop pdb database and creates
   * a pdb line data structure.
   ***********************************************************/
  
  /* copy over the info */
  pdb->res_no = loop.res_no;
  strcpy (pdb->res, loop.res);
  strcpy (pdb->atom, loop.atom);
  pdb->pt.x = loop.pt.x;
  pdb->pt.y = loop.pt.y;
  pdb->pt.z = loop.pt.z;

  /* setup the other lines */
  strcpy (pdb->flag, "ATOM"); 
  strcpy (pdb->extra, "COIL");
  pdb->atom_no = 1; 
  pdb->num1    = 1.0; 
  pdb->num2    = 1.0;

  return(1);

} /* END create_pdb_line_from_loop_line () */
//***********************************************************
//* Parses loop database line and parses for PDB line info.
// ***********************************************************/

int read_loop_database_line (char * line, LOOP_b * l) {
  
  int ret;
  
  ret = sscanf (line, LOOP_PDB_LINE_READ_FORMAT,
                &l->res_no,  l->res, l->atom, l->class, &l->type, 
                &l->pt.x, &l->pt.y, &l->pt.z);
  
  return ret;
  
} /* END read_loop_database_line () */



 /***********************************************************
   * Searches the loop database and finds all suitable candidates.
   * Returns -1 if no loop cannot be found.
   ***********************************************************/
LOOP_PROP* find_loop_candidates(int residues,float desired_length,SIX_CA_POINTS six_cas[],int seg_number,
				int *no_of_loops, int *found_loop,char *loop_type){
  
  FILE  *indexfp, *dbfp;
  char   buffer[MAX_LINE_LEN];
  LOOP_h_for_ind   loop_ind;
  
  LOOP_b   line;
  float   dist_diff;
  
  boolean  found_one;
  
  unsigned int  loop_count;
  unsigned int  i; 
 
  float dist_matrix_of_adj_strands[SIZE_OF_LOOP_DISTMATRIX][SIZE_OF_LOOP_DISTMATRIX];
   
  unsigned int  residue_num;
  unsigned int first_residue_num_of_elbow, sec_residue_num_of_elbow, third_residue_num_of_elbow;
  unsigned int fourth_residue_num_of_elbow, fifth_residue_num_of_elbow, sixth_residue_num_of_elbow;
  int use_CA_matrix = 0;     

  LOOP_PROP *loop_prop;
  //LOOP_ARR *loop_arr;
  
  /* Initialize */
  found_one = false;
  loop_count = residue_num = 0;
  first_residue_num_of_elbow = sec_residue_num_of_elbow = third_residue_num_of_elbow = 0;
  fourth_residue_num_of_elbow = fifth_residue_num_of_elbow = sixth_residue_num_of_elbow = 0;
  
 
  dist_diff = 0.00;
  //printf("loop type = %s\n",loop_type);   

  loop_prop  = (LOOP_PROP*)malloc(1*sizeof(LOOP_PROP) );
  if(!loop_prop){err_print(ERR_MEM_ALLOC, "loop_prop in module loop.c");exit(EXIT_FAILURE);}
  
  /* Opening the index file for reading the loop headers 
   * All the loops that have fist-fist distance close to the desired length, are flagged.
   * Their loop number, and the file position of the loop header are stored.
   */
  indexfp = fopen(COIL_IND,FILE_READ);
  if(!indexfp){printf("Cannot open %s for reading\n",COIL_IND);exit(-1);}
  
  while(fgets (buffer, MAX_LINE_LEN, indexfp) != NULL) {

    if(is_loop_header(buffer) == true) {
			
      read_loop_header_for_ind (buffer, &loop_ind);
      //printf("loopind type = %s\n",loop_ind.type);   
      //if( (loop_ind.len == residues) && (strncmp(loop_ind.type, loop_type, strlen(loop_ind.type))==0) ) { 
	 if( loop_ind.len == residues){ 
				 
	dist_diff = fabs(loop_ind.fist_dist - desired_length); 
        
	//if( dist_diff < MIN_LOOP_DIST_DIFFERENCE ){ 		
                          
	loop_prop = (LOOP_PROP*)realloc( (LOOP_PROP*)loop_prop, (loop_count+1)*sizeof(LOOP_PROP) );
	if(!loop_prop){err_print(ERR_MEM_REALLOC, "loop_prop in module loop.c");exit(-1);}
			   			    
	loop_prop[loop_count].loop_num = loop_ind.num;
	loop_prop[loop_count].file_pos = loop_ind.file_pos;
        loop_prop[loop_count].fist_dist = loop_ind.fist_dist;
	loop_prop[loop_count].abs_diff_between_lengths = dist_diff;
	//printf("header = %s", buffer);
	loop_count++;
	//}
      } 
    }    
  }		
  fclose (indexfp);
 
  /* Copying the loop count to the no_of_loops */
  *no_of_loops = loop_count; 
 
  
  if(loop_count == 0){
    printf("NO POSSIBLE LOOPS CAN BE FOUND\n");
    free(loop_prop);
    *found_loop = -1; 
    return(NULL);
  }else{
    *found_loop = 1;
  }
  
  
  
  if(use_CA_matrix == 1){
    //******************************************************************************************
    /* Forming the dmatrix of the strands that are adjacent to the loop */	  
    form_dmatrix_of_adj_strands(dist_matrix_of_adj_strands, six_cas, seg_number);
    
    
    /*******************************************************************************	     
     * So far, we have obtained all the relevant loops and their loop number and  file positions
     *  of the loop headers in the database file.
     * We start to read the loop database by accessing the loop headers by
     *  their file positions. 
     * After reading each loop body, we form the CA distance matrix of that loop. 
     *********************************************************************************/		     
    
    // opening the loop database file for reading
    dbfp  = fopen (COIL_DAT, FILE_READ);
    if(!dbfp){printf("Cannot open %s for reading\n",COIL_DAT);exit(-1);}
    
    for(i=0; i<loop_count; i++){
      fseek(dbfp, loop_prop[i].file_pos, SEEK_SET);
      
      while (fgets (buffer, MAX_LINE_LEN, dbfp) != NULL){ 
	if (is_loop_header(buffer) == false){ 
	  read_loop_database_line (buffer, &line);
	  
	  if (line.type == 'C'){
	    residue_num = line.res_no;
	    
	    first_residue_num_of_elbow = residue_num - 3;
	    sec_residue_num_of_elbow   = residue_num - 2;
	    third_residue_num_of_elbow = residue_num - 1;
	    
	    fourth_residue_num_of_elbow = residue_num + residues;
	    fifth_residue_num_of_elbow  = residue_num + residues + 1;
	    sixth_residue_num_of_elbow  = residue_num + residues + 2;
	    
	    break;
	  } 
	}
      }
      
      /* We are going back to the begining of the loop body, just after the loop header */
      fseek(dbfp, loop_prop[i].file_pos, SEEK_SET);
      
      while (fgets (buffer, MAX_LINE_LEN, dbfp) != NULL){
	if (is_loop_header(buffer) == false){
	  read_loop_database_line (buffer, &line);
	  
	  if( (line.res_no == first_residue_num_of_elbow )  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].first_CA, line.pt);
	    }
	  }
	  if( (line.res_no ==  sec_residue_num_of_elbow)  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].sec_CA, line.pt);
	    }
	  }
	  if( (line.res_no == third_residue_num_of_elbow )  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].third_CA, line.pt);
	    }
	  }
	  if( (line.res_no == fourth_residue_num_of_elbow )  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].third_last_CA, line.pt);
	    }
	  }
	  if( (line.res_no ==  fifth_residue_num_of_elbow)  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].sec_last_CA, line.pt);
	    }
	  }
	  if( (line.res_no == sixth_residue_num_of_elbow )  ){
	    if (is_ref_atom (line.atom) == true){ 
	      assign_pt( &six_cas[seg_number].last_CA, line.pt);
	    }
	  }
	} /* END if (is_loop_header (buffer) == false) */
	
	else{
	  break;
	} 
	
      } /* END while (fgets (buffer, MAX_LINE_LEN, dbfp) != NULL) */ 
      
      /* forming the CA distance matrix of the loops */
      form_CA_dmatrix_of_loop(loop_prop[i].dist_matrix, six_cas, seg_number); 
      loop_prop[i].score  = compare_matrices(loop_prop[i].dist_matrix, dist_matrix_of_adj_strands); 
      
    } /* END for(i=0; i<loop_count; i++)  */
    
    fclose(dbfp);
    
    /* The CA_dist matrix of the loops are compared to the
     * CA dist matrix of the two adjacent strands. Based on the difference,
     *  a score is assigned to each loop. After sorting the loops
     * according to their scores in ascending order, we obtain the best fit loops. */
    
    qsort(loop_prop, loop_count, sizeof(LOOP_PROP), compare_scores);
  }//END if(use_CA_matrix == 1)
  //*****************************************************************************************************

	
  /*
  loop_arr = malloc(loop_count*sizeof(LOOP_ARR));	
  // Copying loop-prop values to loop-array values
  for(i=0; i<loop_count; i++){
    loop_arr[i].loop_num = loop_prop[i].loop_num;
    loop_arr[i].file_pos = loop_prop[i].file_pos;
    loop_arr[i].fist_dist = loop_prop[i].fist_dist;
    loop_arr[i].abs_diff_between_lengths = loop_prop[i].abs_diff_between_lengths;

    if(use_CA_matrix == 1){
      loop_arr[i].score = loop_prop[i].score;
    }else{
      loop_arr[i].score = 0.00;
    }
  }
  
  free(loop_prop);
  */
  return(loop_prop);	
		
}// END find_loop_candidates *********************************************************************************

/***********************************************************
   * Parses loop header and returns loop information.
   ***********************************************************/ 
int read_loop_header_for_ind (char * header, LOOP_h_for_ind * l) {
   int ret;
   //ret = sscanf (header, LOOP_HEAD_READ_FORMAT_FOR_IND,&l->num,  &l->len, l->type, &l->elbow_dist, &l->fist_dist, &l->file_pos);
  ret = sscanf (header, LOOP_HEAD_READ_FORMAT_FOR_IND,&l->num,  &l->len, l->type, &l->fist_dist, &l->file_pos);
  return ret;

} /* END read_loop_header () */
//****************************************************************


//********************************************************************************************
void print_loop_properties(char *loop_information, LOOP_PROP *loop_arr, int no_of_loops,float loop_length, int residues){
  FILE *fpw;
  int i;
  
  fpw = fopen(loop_information,"a");
  if(!fpw){printf("Cannot open %s. Exiting..\n",loop_information);exit(-1);}

  fprintf(fpw,"-- Num of residues = %d\n",residues);
  fprintf(fpw,"-- Desired length = %4.3f\n",loop_length);
  fprintf(fpw,"-- Num of loops found = %d\n\n",no_of_loops);  
  
  fprintf(fpw,"      Loop Number     File    Fist Distance      Diff between loop length     Score based on six CA \n");
  fprintf(fpw,"     in index file    Pos                        and desired length               diff matrix      \n");
  fprintf(fpw,"----------------------------------------------------------------------------------------------\n");
  
  for(i=0;i<no_of_loops;i++){
    fprintf(fpw,"%4d  %8d %10ld %15.3f %22.3f %28.3f\n",i,loop_arr[i].loop_num,loop_arr[i].file_pos,loop_arr[i].fist_dist,loop_arr[i].abs_diff_between_lengths,loop_arr[i].score);
  }
  fflush(fpw);fclose(fpw);
  
}// END ***************************************************************************************** 
/*TODO: Utility functions, should be modulerized to be included in seperate util.h*/

/***********************************************************
 * Read one line from the given pdb file and stores it in 
 * the given variable. Returns the number of items read or
 * EOF if end of file is encountered.
 ***********************************************************/
int read_pdb_line (FILE *fp, PDB_LINE *l) {
  
  int ret;

  ret = fscanf(fp, PDB_READ_FORMAT,
                l->flag,  &l->atom_no, l->atom,  l->res,   &l->res_no, 
                &l->pt.x, &l->pt.y,    &l->pt.z, &l->num1, &l->num2, 
                l->extra);
  sprintf(l->whole_line,PDB_WRITE_FORMAT,l->flag,  l->atom_no, l->atom,  l->res,   l->res_no,
                 l->pt.x,  l->pt.y,    l->pt.z,  l->num1,  l->num2,
                 l->extra);
  return ret;

  } /* END read_pdb_line () */
//************************************************************

/***********************************************************
 * Returns true if given atom is the Nitrogen atom.
 ***********************************************************/
boolean is_N_atom (char * atom) {
  if (strcmp (atom, N_ATOM) == 0) { return true;  }
  else                              { return false; }
} /* END is_N_atom () */


/***********************************************************
 * Returns true if given atom is the Carbon atom.
 ***********************************************************/
boolean is_C_atom (char * atom) {
  if (strcmp (atom, C_ATOM) == 0) { return true;  }
  else                              { return false; }
} /* END is_C_atom () */



/***********************************************************
 * Returns true if given atom is the CB atom.
 ***********************************************************/
boolean is_CB_atom (char * atom){
if( strcmp(atom,CB_ATOM) == 0) { return true; }
else                            {return false;}	
}/* END is_CB_atom () */



/***********************************************************
 * Returns true if given atom is the Oxygen atom.
 ***********************************************************/	
boolean is_O_atom (char * atom){
if (strcmp(atom,O_ATOM) == 0) { return true; }
else                            {return false;}	
}/* END is_O_atom () */

/***********************************************************
 * Returns true if given atom is the rotation reference
 * point atom.
 ***********************************************************/
boolean is_ref_atom (char * atom) {
  if (strcmp (atom, REF_ATOM) == 0) { return true;  }
  else                              { return false; }
} /* END is_ref_atom () */

/***********************************************************
 * Returns true if extra is HELIX
 ***********************************************************/
boolean is_helix (char * extra) {
  if (strcmp (extra, helix) == 0) { return true;  }
  else                              { return false; }
} /* END is_helix () */



/***********************************************************
 * Returns true if extra is COIL
 ***********************************************************/
boolean is_coil (char * extra) {
  if (strcmp (extra, coil) == 0) { return true;  }
  else                              { return false; }
} /* END is_coil () */


int assign_pt (POINT * a, POINT b) {
/***********************************************************
 * Assigns point b to point a;
 ***********************************************************/
 
  a->x = b.x; a->y = b.y; a->z = b.z;

  return 1;

} /* END assign_pt () */

/***********************************************************
 * Write one pdb line to the given pdb using data from 
 * the given variable. Returns the number of items written.
 ***********************************************************/
int write_pdb_line(FILE * fp, PDB_LINE l) {
  
  int ret;
  
  ret = fprintf (fp, PDB_WRITE_FORMAT,
                 l.flag,  l.atom_no, l.atom,  l.res,   l.res_no,
                 l.pt.x,  l.pt.y,    l.pt.z,  l.num1,  l.num2,
                 l.extra);
  
  return ret;
  
} //END write_pdb_line*********************************

/*TODO: Create an error.c for error handling*/
void err_print (char * err_string, char * err_details) {
/***********************************************************
 * Prints an error message based on the error string passed
 * in. (See error.h for complete list of pre-defined error
 * strings.
 ***********************************************************/

  if (err_details != NULL) {
    fprintf (stderr, "ERROR: %s - \'%s\'\n", err_string, err_details);
  } else {
    fprintf (stderr, "ERROR: %s\n", err_string);
  }

} /* END err_print () */

float get_distance(POINT *pt1, POINT *pt2){
/***********************************************************	
 * returns the distance between two points 
***********************************************************/ 
float l1, l2, l3;
	
	l1 = pow((pt1->x - pt2->x),2.00);
	l2 = pow((pt1->y - pt2->y),2.00);
	l3 = pow((pt1->z - pt2->z),2.00);

	return( sqrt (l1+l2+l3) );
} /* END float get_distance() */	
