/* MODULE: sidechain.c */

/***OVERVIEW************************************************
 * File Name     : sidechain.c
 * Purpose       : Side chain insertion module for thrifty.
 * Author        : Philip S Luk ~ philipluk@bigfoot.com
 * Date Written  : July 06, 2001
 * Last Modified : July 13, 2001
 ***END*OVERVIEW********************************************/

/* STANDARD INCLUDES */
#include <stdlib.h>
#include <math.h>
/* END STANDARD INCLUDES */

#include "btimem.h"
#include "sidechain.h"
#include "rotatefinal.h"
#include "csipdb.h"
#include "misc.h"

void initialize (void)
{
  /***********************************************************
   * Initializes the global variables.
   ***********************************************************/
  g_res_count = 0;
} /* END initialize () */

void shut_down_cleanup (void)
{
  /***********************************************************
   * Cleans up global variables.
   ***********************************************************/
  g_res_count = 0;
} /* END initialize () */

int add_residue (char * res_name)
{
  /***********************************************************
   * Adds a new residue to the residue list.
   ***********************************************************/

  if (g_res_count < MAX_RESIDUES-1)
    {
      strcpy (g_res_list[++g_res_count].name, res_name);
      return 1;
    }
  else
    {
      fprintf (stderr, "ERROR: Maximum number of residues allowed is %d\n", MAX_RESIDUES);
      return -1;
    }
} /* END set_res_name () */

int set_res_ref_pt (char label, int res_no, POINT_t pt)
{
  /***********************************************************
   * Sets the residues rotation reference points.
   ***********************************************************/


  if(label=='a')
    g_res_list[res_no].A = pt;
  else if(label== 'b')
    g_res_list[res_no].B = pt;
  else if(label=='c')
    g_res_list[res_no].C = pt;
  else
    return 0;

  return 1;

} /* END set_ref_pt () */

int set_res_atom_count (int res_no, int count)
{
  /***********************************************************
   * Sets the residues atom count. Useful in determining
   * whether or not we need to attach side chains.
   ***********************************************************/

  g_res_list[res_no].num_atoms = count;

  return 1;

} /* END set_res_atom_count () */

boolean is_atom_flag (char * flag)
{
  /***********************************************************
   * Returns true if the given flag is an atom data line 
   * marker; false otherwise.
   ***********************************************************/

  if (strcmp (flag, ATOM_FLAG) == 0)
    {
      return true;
    }
  else
    {
      return false;
    }

} /* END is_atom_flag () */

boolean is_end_flag (char * flag)
{
  /***********************************************************
   * Returns true if the given flag is an end buffer/file 
   * marker; false otherwise.
   ***********************************************************/

  if (strcmp (flag, END_FLAG) == 0)
    {
      return true;
    }
  else
    {
      return false;
    }

} /* END is_end_flag () */

int see_who_needs_a_side_chain (void)
{
  /***********************************************************
   * Run through the list of residues and determine which
   * ones require a side chain.
   ***********************************************************/

  int i;

  for (i=1; i<=g_res_count; i++)
    {
      if (strcmp (g_res_list[i].name, "GLY") == 0)
        {
          // GLY residues do not require side chains
          g_res_list[i].need_sdechn = false;
          continue;
        }
      else if (strcmp (g_res_list[i].name, "ALA") == 0)
      {
        // ALA residues do not require side chains
        // yeah they do - they need a methyl group
        g_res_list[i].need_sdechn = true;
        continue;
      }
      else
        {
          /* Other residues with only 5 atoms would need a side chain */
          /* SCOTT - why?? this makes no sense.  What about atoms that had 6 atoms, and now need 7? */
          /*if (g_res_list[i].num_atoms <= 5)
          {
            g_res_list[i].need_sdechn = true;
          }
          else
*/
          {
            g_res_list[i].need_sdechn = true;
          }
        }   /* END _if_     */
    }     /* END _for_    */

  return 1;

} /* END see_who_needs_a_side_chain () */

int parse_pdb_side_chain_data (char * pdbfile)
{
  /***********************************************************
   * Parses given PDB file for residue information for side
   * chain insertion stuffies.
   ***********************************************************/

  FILE    * pdbfilep;
  int       prev_res_no, atom_count;
  int       atom_no, res_no;
  char      flag[20], atom[5], res[5],line[MAX_LINE_LEN];
  POINT_t   pt;
  
  /* initialize */
  prev_res_no = 0;
  atom_count  = 0;
  
  pdbfilep    = open_file(pdbfile, "r");
  while (fgets(line,MAX_LINE_LEN,pdbfilep)!=NULL)
  {
    sscanf(line,"%s", flag);
    if (is_atom_flag (flag) == true)
      {
        /* check to see if atom info line */
        sscanf(line,"%s %6d %3s %3s %5d %8lf%8lf%8lf %*f %*f %*s",
               flag, &atom_no, atom, res, &res_no, &pt.x, &pt.y, &pt.z);
        if (prev_res_no != res_no)
        {	  
          /* 	 new residue */
          atom_count = 0;				/*  reset atom count for residue */
          prev_res_no = res_no;			  /* set_new_residue */

          if (add_residue (res) < 0)
            {
              fprintf (stderr, "Failed to add residue %d!\n", res_no);
              return -1;
            }
        }  /*   END _if_new_residue_ */

        if (strcmp (atom, A_REF_ATOM) == 0)
        {
          set_res_ref_pt ('a', res_no, pt);
        }
        else if (strcmp (atom, B_REF_ATOM) == 0)
        {
          set_res_ref_pt ('b', res_no, pt);
        }
        else if (strcmp (atom, C_REF_ATOM) == 0)
        {
          set_res_ref_pt ('c', res_no, pt);
        } /*  END _set_reference_pts_ */

        set_res_atom_count (res_no, ++atom_count);	 /* increment the atom count */
      }    /* END _if_atom_ */

  }  /*END _while_ */
  fclose (pdbfilep);
  
  /* setup markers to indicate which residues need a side chain */
  see_who_needs_a_side_chain ();
  return 1;

} /* ENDparse_pdb_side_chain_data () */

void print_res_list ()
{
  /***********************************************************
   * Prints out the residue list. Useful in debugging.
   ***********************************************************/
  int i;

  for (i=1; i<=g_res_count; i++)
    {
      fprintf (stdout, "%3d. RESIDUE: %s NO_ATOMS: %d\n",
               i, g_res_list[i].name, g_res_list[i].num_atoms);
      if (g_res_list[i].need_sdechn == true)
        {
          fprintf (stdout, "     ->A: (%.3f,%.3f,%.3f)\n",
                   g_res_list[i].A.x, g_res_list[i].A.y, g_res_list[i].A.z);
          fprintf (stdout, "     ->B: (%.3f,%.3f,%.3f)\n",
                   g_res_list[i].B.x, g_res_list[i].B.y, g_res_list[i].B.z);
          fprintf (stdout, "     ->C: (%.3f,%.3f,%.3f)\n",
                   g_res_list[i].C.x, g_res_list[i].C.y, g_res_list[i].C.z);
        }
      else
        {
          fprintf (stdout, "     -> SIDE CHAIN UNNECESSARY\n");
        }
    }

} /* END print_res_list () */

boolean need_side_chain (int res_no)
{
  /***********************************************************
   * Returns true if the residue requires us to attach a 
   * side chain; false otherwise.
   ***********************************************************/

  return g_res_list[res_no].need_sdechn;

} /* END need_side_chain () */

boolean side_chain_goes_here (char * atom)
{
  /***********************************************************
   * Returns true if the side chain should go after the
   * given atom; false otherwise. 
   ***********************************************************/
  /* change CB to O since we use N CA C to do the superposition */
  if (strcmp (atom, "O") == 0)
    {
      return true;
    }
  else
    {
      return false;
    }

} /* END side_chain_goes_here () */

char res_str_to_char (char * res_name)
{
  /***********************************************************
   * Converts the residue's 3 character representation to it's
   * single character representation.
   ***********************************************************/

  if      (strcmp (res_name, "ALA") == 0)
    {
      return 'A';
    }
  else if (strcmp (res_name, "ARG") == 0)
    {
      return 'R';
    }
  else if (strcmp (res_name, "ASN") == 0)
    {
      return 'N';
    }
  else if (strcmp (res_name, "ASP") == 0)
    {
      return 'D';
    }
  else if (strcmp (res_name, "CYS") == 0)
    {
      return 'C';
    }
  else if (strcmp (res_name, "GLN") == 0)
    {
      return 'Q';
    }
  else if (strcmp (res_name, "GLU") == 0)
    {
      return 'E';
    }
  else if (strcmp (res_name, "GLY") == 0)
    {
      return 'G';
    }
  else if (strcmp (res_name, "HIS") == 0)
    {
      return 'H';
    }
  else if (strcmp (res_name, "ILE") == 0)
    {
      return 'I';
    }
  else if (strcmp (res_name, "LEU") == 0)
    {
      return 'L';
    }
  else if (strcmp (res_name, "LYS") == 0)
    {
      return 'K';
    }
  else if (strcmp (res_name, "MET") == 0)
    {
      return 'M';
    }
  else if (strcmp (res_name, "PHE") == 0)
    {
      return 'F';
    }
  else if (strcmp (res_name, "PRO") == 0)
    {
      return 'P';
    }
  else if (strcmp (res_name, "SER") == 0)
    {
      return 'S';
    }
  else if (strcmp (res_name, "THR") == 0)
    {
      return 'T';
    }
  else if (strcmp (res_name, "TRP") == 0)
    {
      return 'W';
    }
  else if (strcmp (res_name, "TYR") == 0)
    {
      return 'Y';
    }
  else if (strcmp (res_name, "VAL") == 0)
    {
      return 'V';
    }
  else
    {
      return -1;
    }

} /* END res_str_to_char () */

boolean matching_chi_angle (int res_no, double angle)
{
  /***********************************************************
   * Returns true if the chi angle for the residue matches;
   * false otherwise.
   ***********************************************************/

  /* Here we have to get the chi angle for the given residue
     and see if it matches the given chi angle */
  float anglediff;

  if(g_res_list[res_no-1].chi<360)
    {
      anglediff=(float)fabs((double)g_res_list[res_no-1].chi-angle);

      if((anglediff<60) || (fabs(anglediff-360)<60))
        return true;
      else
        return false;
    }
  else
    return true;

} /* END matching_chi_angle () */

void load_aa_db(char* database)
{
  FILE  * databasep;
  char    buffer[MAX_LINE_LEN];
  int residueIndex;

  for ( int i=0; i<26; i++ ) 
  {
    g_amino_acid_db[i].num_atoms = 0; 
    g_amino_acid_db[i].set = false;
  }

  databasep = fopen (database, "r");
  while (fgets (buffer, MAX_LINE_LEN, databasep) != NULL)
    {

      if (buffer[0] == '>')
        {	/* Look for side chain header lines */
          /* Load up the amino acid DB */
          residueIndex = buffer[1] - 'A';
          g_amino_acid_db[residueIndex].num_atoms = 0;
        }     /* END _if_header_    */
        else if ( strncmp (buffer, ATOM_FLAG, 4) == 0)
        {
          int index = g_amino_acid_db[residueIndex].num_atoms;
          sscanf(buffer,"%*s %*6d %3s %*3s %*5d %*f%*f%*f %*f %*f %*s",
               g_amino_acid_db[residueIndex].atoms[index]);
          g_amino_acid_db[residueIndex].num_atoms++;
        }
    } /* END _while_ */
}

char * fetch_side_chain (char * sdechn_buffer, char * database, int res_no)
{
  /***********************************************************
   * Run through the side chain database and find a suitable
   * candidate. Returns a buffer holding the side chain.
   ***********************************************************/

  FILE  * databasep;
  char    buffer[MAX_LINE_LEN];
  boolean have_candidate;
  boolean finished_reading;
  double  angle;
  int  num=0;
  have_candidate   = false;
  finished_reading = false;
  sdechn_buffer[0] = '\0';		/* clear buffer */

  databasep = fopen (database, "r");
  while (fgets (buffer, MAX_LINE_LEN, databasep) != NULL)
    {

      if (buffer[0] == '>')
        {	/* Look for side chain header lines */
          
          /* If header line, then see if it is a side chain we want */
          if (buffer[1] == res_str_to_char (g_res_list[res_no].name))
            {
              num++;
              sscanf (buffer, ">%*c %5lf", &angle);
              if (matching_chi_angle (res_no, angle) == true)
              {
                /* Found our candidate, so stop searching */
                have_candidate = true;
                break;
              }/* END _if_chi_angle_ */

              if(res_str_to_char (g_res_list[res_no].name)!='P')
              {
                /* Does this make sense: If this residue is not proline, 
                and we've already looked at 3, then all of a sudden we have one?? 
              */
                if(num==3 && have_candidate==false)
                  {
                    have_candidate=true;
                    break;
                  }
              }
              else
                {
                  if(num==2 && have_candidate==false)
                    {
                      have_candidate=true;
                      break;
                    }
                }
            }   /* END _if_candidate_ */
        }     /* END _if_header_    */
    } /* END _while_ */
  if (have_candidate == true)
    {
      /* Now we can push side chain into the buffer */

      while (finished_reading == false)
        {
          if(fgets (buffer, MAX_LINE_LEN, databasep)!=NULL)
            {
              if (buffer[0] == 'A')
                {		/* 'ATOM' */
                  strcat (sdechn_buffer, buffer);
                }
              else
                {
                  finished_reading = true;
                }/* END _if_atom_ */
            }
          else
            {
              finished_reading = true;
            }
        }   /* END _while_ */

    } /* END _if_have_candidate_ */

  fclose (databasep);
  strcat (sdechn_buffer, END_FLAG);	/* insert terminating flag */
  return sdechn_buffer;

} /* END fetch_side_chain () */

int paste_side_chain (FILE * fp, char * buffer, int res_no)
{
  /***********************************************************
   * Places the side chain into the given file.
   ***********************************************************/

  char      flag[5], atom[5], res[5];
  int       atom_no;
  POINT_t   pt;
  char      line_buffer[MAX_LINE_LEN];
  char    * temp_buffer;
  boolean   end_of_side_chain;
  boolean   ok_to_dump_side_chain_to_file;

  /* initialization */
  temp_buffer = bti_malloc (strlen (buffer) + 1);
  strcpy (temp_buffer, buffer);
  strcpy (line_buffer, strtok (temp_buffer, "\n"));
  ok_to_dump_side_chain_to_file = false;
  end_of_side_chain = false;

  while (end_of_side_chain == false)
    {

      sscanf (line_buffer, "%s %6d %3s %3s %*5d %7lf %7lf %7lf %*f %*f\n",
              flag, &atom_no, atom, res, &pt.x, &pt.y, &pt.z);

      if (is_end_flag (flag) == true)
        {
          end_of_side_chain = true;
          break;
        }

      if (ok_to_dump_side_chain_to_file == true)
        {

          /*  if(strcmp(atom,"CB")!=0) add by Haiyan */
          fprintf (fp, "%s %6d  %-3s %3s %5d     %7.3f %7.3f %7.3f %2.2f %2.2f\n",
                   flag, atom_no, atom, res, res_no, pt.x, pt.y, pt.z, 1.00, 1.00);
        } /* END _if_ok_to_dump_ */

      if (side_chain_goes_here (atom) == true)
        {
          /* next time around we can print the side chain */
          ok_to_dump_side_chain_to_file = true;
        } /* END _if_side_chain_goes_here_ */

      strcpy (line_buffer, strtok (NULL, "\n"));  /* setup the next line */
    } /* END _while_ */
  free(temp_buffer);
  return 1;

} /* END paste_side_chain () */

RES_t * get_side_chain_ref_pts (char * buffer, RES_t * sdechn_res)
{
  /***********************************************************
   * Gets the rotation reference points for the side chain.
   ***********************************************************/

  char      flag[MAX_NAME_LEN], atom[MAX_NAME_LEN];
  POINT_t   pt;
  char      line_buffer[MAX_LINE_LEN];
  char    * temp_buffer;
  boolean   end_of_side_chain;

  /* initialization */
  temp_buffer = (char *)bti_malloc ((strlen (buffer) + 1)*sizeof(char));
  if(temp_buffer==NULL)
    {
      printf("Can not malloc temp_buffer\n");
      exit(0);
    }
  strcpy (temp_buffer, buffer);
  strcpy (line_buffer, strtok (temp_buffer, "\n"));
  end_of_side_chain = false;
  while (end_of_side_chain == false)
    {

      sscanf (line_buffer, "%s %*6d %3s %*3s %*5d %7lf %7lf %7lf %*f %*f\n",
              flag, atom, &pt.x, &pt.y, &pt.z);

      if (is_end_flag (flag) == true)
        {
          end_of_side_chain = true;
          break;
        }

      if        (strcmp (atom, A_REF_ATOM) == 0)
        {
          sdechn_res->A.x = pt.x;
          sdechn_res->A.y = pt.y;
          sdechn_res->A.z = pt.z;
        }
      else if (strcmp (atom, B_REF_ATOM) == 0)
        {
          sdechn_res->B.x = pt.x;
          sdechn_res->B.y = pt.y;
          sdechn_res->B.z = pt.z;
        }
      else if (strcmp (atom, C_REF_ATOM) == 0)
        {
          sdechn_res->C.x = pt.x;
          sdechn_res->C.y = pt.y;
          sdechn_res->C.z = pt.z;
        } /* END _set_reference_pts_ */

      strcpy (line_buffer, strtok (NULL, "\n"));  /* setup the next line */
    } /* END _while_ */

  free(temp_buffer);
  return sdechn_res;

} /* END get_side_chain_ref_pts () */

char * calc_side_chain_coords (char * buffer, RES_t sdechn_res, int res_no,
                               char *tmpdir)
{
  /***********************************************************
   * Places the side chain into the given file.
   ***********************************************************/

  FILE    * abcfp, * pqrfp, * infp, * outfp;
  char      flag[MAX_NAME_LEN], atom[MAX_NAME_LEN], res[MAX_NAME_LEN];
  int       atom_no, rnum;
  POINT_t   pt;
  char      line_buffer[MAX_LINE_LEN];
  char    * temp_buffer;
  boolean   end_of_side_chain;
  char abcFilename[MAX_FNAME_LEN];
  char pqrFilename[MAX_FNAME_LEN];
  char inFilename[MAX_FNAME_LEN];
  char outFilename[MAX_FNAME_LEN];

  /* initialization */
  temp_buffer = bti_malloc (strlen (buffer) + 1);
  strcpy (temp_buffer, buffer);
  strcpy (line_buffer, strtok (temp_buffer, "\n"));
  buffer[0] = '\0';			/* clear the buffer */
  end_of_side_chain = false;

  sprintf(abcFilename, "%s%c%s", tmpdir, '/', ABCFILE);
  sprintf(pqrFilename, "%s%c%s", tmpdir, '/', PQRFILE);
  sprintf(inFilename, "%s%c%s", tmpdir, '/', INFILE);
  sprintf(outFilename, "%s%c%s", tmpdir, '/', OUTFILE);

  /* setup superimpose files */
  abcfp = fopen (abcFilename, "w");
  fprintf (abcfp, "%.3f %.3f %.3f\n",
           g_res_list[res_no].A.x, g_res_list[res_no].A.y, g_res_list[res_no].A.z);
  fprintf (abcfp, "%.3f %.3f %.3f\n",
           g_res_list[res_no].B.x, g_res_list[res_no].B.y, g_res_list[res_no].B.z);
  fprintf (abcfp, "%.3f %.3f %.3f\n",
           g_res_list[res_no].C.x, g_res_list[res_no].C.y, g_res_list[res_no].C.z);
  fclose  (abcfp);

  pqrfp = fopen (pqrFilename, "w");
  fprintf (pqrfp, "%.3f %.3f %.3f\n", sdechn_res.A.x, sdechn_res.A.y, sdechn_res.A.z);
  fprintf (pqrfp, "%.3f %.3f %.3f\n", sdechn_res.B.x, sdechn_res.B.y, sdechn_res.B.z);
  fprintf (pqrfp, "%.3f %.3f %.3f\n", sdechn_res.C.x, sdechn_res.C.y, sdechn_res.C.z);
  fclose  (pqrfp);
  /* end of setup superimpose files */

  while (end_of_side_chain == false)
    {

      /* read in original pre-rotation coordinates */
      sscanf (line_buffer, "%s %6d %3s %3s %5d %7lf %7lf %7lf %*f %*f\n",
              flag, &atom_no, atom, res, &rnum, &pt.x, &pt.y, &pt.z);

      if (is_end_flag (flag) == true)
        {
          end_of_side_chain = true;
          break;
        }

      /* setup coordinate values to be rotated file */
      infp = fopen (inFilename, "w");
      fprintf (infp, "%.3f %.3f %.3f\n", pt.x, pt.y, pt.z);
      fclose (infp);
      /* end setup of input file */

      /* call rotation subroutine */
      superimpose (abcFilename, pqrFilename, inFilename, outFilename);

      /* read back rotated coordinate values */
      outfp = fopen (outFilename, "r");
      fscanf (outfp, "%lf %lf %lf\n", &pt.x, &pt.y, &pt.z);
      fclose (outfp);
      /* end of reading output file */

      /* create data line with rotated coordinates and insert into the buffer */
      sprintf (line_buffer, "%s %6d  %-3s %3s %5d     %7.3f %7.3f %7.3f %2.2f %2.2f\n",
               flag, atom_no, atom, res, rnum, pt.x, pt.y, pt.z, 1.00, 1.00);
      strcat (buffer, line_buffer);

      strcpy (line_buffer, strtok (NULL, "\n"));	/* setup the next line */
    } /* END _while_ */

  free(temp_buffer);
  strcat (buffer, END_FLAG);	/* insert terminating flag */
  return buffer;

} /* END calc_side_chain_coords () */

/**
  @param pdbfile is the file to attach the side chain to
  @param chnfile is the output file
  @param database The path to the side.db file
  @param tmpdir  The temporay directory
**/
int attach_side_chains (char * pdbfile, char * chnfile, char * database,
                        char *tmpdir)
{
  /***********************************************************
   * Run through the PDB file and insert the necessary side
   * chains.
   ***********************************************************/

  FILE    * pdbfilep;
  FILE    * newfilep;
  char      sdechn_buffer[MAX_LINE_LEN * MAX_RES_ATOMS];
  int       atom_no, res_no;
  char      flag[MAX_NAME_LEN], atom[MAX_NAME_LEN]; 
  char res[MAX_NAME_LEN],line[MAX_LINE_LEN];
  RES_t     sdechn_res;
  POINT_t   pt;
  char      tempFile[MAX_FNAME_LEN];

  sprintf(tempFile, "%s%c%s", tmpdir, '/', TEMP_FILE);
  pdbfilep    = fopen (pdbfile,   "r");
  newfilep    = fopen (tempFile, "w");

  while (fgets(line,MAX_LINE_LEN,pdbfilep)!=NULL)
    {
      sscanf (line, "%s %6d %3s %3s %5d %7lf %7lf %7lf %*f %*f %*s",
              flag, &atom_no, atom, res, &res_no, &pt.x, &pt.y, &pt.z);
      if(strcmp(atom,"O")==0)
        sprintf(line,"%3s %5d    %8.3f%8.3f%8.3f  1.00 1.00",
                res, res_no, pt.x, pt.y, pt.z);

      if (is_atom_flag (flag) == true)
        {

          if((strcmp(atom,"CB")==0) && (need_side_chain (res_no) == true))
            ;
          else
            fprintf (newfilep, "%s %6d  %-3s %3s %5d    %8.3f%8.3f%8.3f %2.2f %2.2f\n",
                     flag, atom_no, atom, res, res_no, pt.x, pt.y, pt.z, 1.00, 1.00);

          if ((need_side_chain (res_no) == true) 
              && (side_chain_goes_here (atom) == true))
            {
              fetch_side_chain (sdechn_buffer, database, res_no);
              get_side_chain_ref_pts (sdechn_buffer, &sdechn_res);
              calc_side_chain_coords (sdechn_buffer, sdechn_res, res_no, tmpdir);
              paste_side_chain (newfilep, sdechn_buffer, res_no);
            }  /* END _if_need_side_chain_ */
        }   /* END _if_is_atom_flag_    */

    } /* END _while_ */
  //fprintf (newfilep,"ATOM %6d  OT1 %s\n",atom_no+1,line);
  //fprintf (newfilep,"ATOM %6d  OT2 %s\n",atom_no+2,line);
  fprintf (newfilep, "%s\n", END_FLAG);

  fclose (newfilep);
  fclose (pdbfilep);

  rename (tempFile, chnfile);
  /* printf("attach side chain\n"); */
  return 0;

} /* END attach_side_chains () */

/**
  @param filename the output file name
  @param inputFilename the inpt file name
  @param tmpdir the temporary directory
**/
int remove_extra_side_chain_atoms(char* filename, char * inputFilename, char* tmpdir)
{
   /***********************************************************
   * Goes through and removes any atoms that are leftover from 
   * adding other sidechains
   ***********************************************************/

  FILE  * pdbfp;
  FILE  * newfp;
  int     res_no, atom_no;
  char    flag[MAX_NAME_LEN], atom[MAX_NAME_LEN], res[MAX_NAME_LEN];
  POINT_t pt;
  char    tempFile[MAX_FNAME_LEN];
  char    line[MAX_LINE_LEN];

  sprintf(tempFile, "%s%c%s", tmpdir, '/', TEMP_FILE);

  pdbfp = open_file(inputFilename, "r");
  newfp = open_file(tempFile, "w");

  int     atom_counts[MAX_RESIDUES];
  for ( int i=0; i<MAX_RESIDUES; i++)  atom_counts[i] = 0;

  /* Get the atom counts for each residue in the chain */
  while (fgets(line,MAX_LINE_LEN,pdbfp) != NULL)
  {
    sscanf (line, "%s %6d %3s %3s %5d %7lf %7lf %7lf %*f %*f",
            flag, &atom_no, atom, res, &res_no, &pt.x, &pt.y, &pt.z);
    atom_counts[res_no]++;
  } /* END _while_ */


  /* Look to see if any atoms need to remove their extra atoms */
  for ( int i=0; i<MAX_RESIDUES; i++)
    if ( atom_counts[i] != 0 )
    {
      int index = res_str_to_char(g_res_list[i].name ) - 'A';
      if ( g_amino_acid_db[index].num_atoms != atom_counts[i] )
        g_res_list[i].remove_extra_atoms = true;
    }

  /* go back to the beginning of the file */
  rewind(pdbfp);

  char atomsSeen[MAX_RES_ATOMS][5];
  int atomsSeenCount = 0;
  int lastResNo = -1;

  /* Fix any atoms that need to be removed */
  while (fgets(line,MAX_LINE_LEN,pdbfp) != NULL)
  {
    sscanf (line, "%s %*6d %3s %3s %5d %7lf %7lf %7lf %*f %*f",
            flag, atom, res, &res_no, &pt.x, &pt.y, &pt.z);
    
    /* Keep track of when we're changing residues */
    if ( res_no != lastResNo )
      atomsSeenCount = 0;

    lastResNo = res_no;

    if (is_atom_flag (flag) == true)
    {
      boolean printThisAtom = true;

      if ( g_res_list[res_no].remove_extra_atoms )
      {
        printThisAtom = false;
        
        int db_index = res_str_to_char( g_res_list[res_no].name ) - 'A';

        /* See if this atom is in the list of atoms OK for this amino acid */
        for(int i=0; i<g_amino_acid_db[db_index].num_atoms; i++)
        {
          if ( strcmp( g_amino_acid_db[db_index].atoms[i], atom ) == 0 )
          {
            printThisAtom = true;
            break;
          }
        }

        
        if ( printThisAtom ) // this is just an optimization
          /* See if we've already printed this atom, so we remove duplicates */
          for ( int i=0; i<atomsSeenCount; i++)
            if ( strcmp( atomsSeen[i], atom ) == 0 )
            {
              printThisAtom = false;
              break;
            }

      }

      if ( printThisAtom )
      {
        sprintf(atomsSeen[atomsSeenCount++], "%s", atom);
        fprintf (newfp, "%s %6d  %-3s %3s %5d     %7.3f %7.3f %7.3f %2.2f %2.2f\n",
                flag, atom_no, atom, res, res_no, pt.x, pt.y, pt.z, 1.00, 1.00);
        #ifdef DEBUG
        fflush(newfp);
        #endif
      }
    }
  }
  fprintf (newfp, "%s\n", END_FLAG);

  fclose (newfp);
  fclose (pdbfp);

  rename (tempFile, filename);

  return 1;
}

int renumber_atoms (char * filename, char * inputFilename, char *tmpdir)
{
  /***********************************************************
   * Goes through and renumbers the pdbfile atoms correctly.
   ***********************************************************/

  FILE  * pdbfp;
  FILE  * newfp;
  int     atom_count, res_no;
  char    flag[MAX_NAME_LEN], atom[MAX_NAME_LEN], res[MAX_NAME_LEN];
  POINT_t pt;
  char    tempFile[MAX_FNAME_LEN];
  char    line[MAX_LINE_LEN];

  sprintf(tempFile, "%s%c%s", tmpdir, '/', TEMP_FILE);
  /* initialize */
  atom_count = 0;

  pdbfp = fopen (inputFilename, "r");
  newfp = fopen (tempFile, "w");

  while (fgets(line,MAX_LINE_LEN,pdbfp) != NULL)
    {
      sscanf (line, "%s %*6d %3s %3s %5d %7lf %7lf %7lf %*f %*f",
              flag, atom, res, &res_no, &pt.x, &pt.y, &pt.z);

      if (is_atom_flag (flag) == true)
        {
          atom_count ++;   /* increment atom count */
          fprintf (newfp, "%s %6d  %-3s %3s %5d     %7.3f %7.3f %7.3f %2.2f %2.2f\n",
                   flag, atom_count, atom, res, res_no, pt.x, pt.y, pt.z, 1.00, 1.00);
        }   /* END _if_is_atom_flag_    */

    } /* END _while_ */

  fprintf (newfp, "%s\n", END_FLAG);

  fclose (newfp);
  fclose (pdbfp);

  rename (tempFile, filename);

  return 1;

} /* END renumber_atoms () */

void get_chi_angle(int num)
{
  int i;

  for(i=0;i<g_res_count;i++)
    {
      g_res_list[i].chi = chi_struc[num][i].chi;
     
      char res[5]; 
      // SCOTT - this is VERY iffy, check this!
      //printf("SCOTT - IFFY - check this!  sidechain.c 809");
      strncpy(res, chi_struc[num][i].res, 5);
      // either this or 
      // strncpy(res, g_res_list[num][i].name, 5);
      // but I don't think g_res_list is initialized to the correct value


      if( g_res_list[i].chi>400)
      {
        g_res_list[i].chi=-70;
        if(strcmp(res,"PRO")==0)
          g_res_list[i].chi=30;
        if(strcmp(res,"VAL")==0)
          g_res_list[i].chi=175;
        if(strcmp(res,"SER")==0)
          g_res_list[i].chi=70;
        if(strcmp(res,"THR")==0)
          g_res_list[i].chi=70;
      }
    }
}

void do_side_chain_stuff (int topAlignNum, PROGRAM_PARAMS params)
{
  /***********************************************************
   * Attaches side chains for thrifty.
   ***********************************************************/

  FILE * tmpfilep;
  char   usrfile [100];
  char   pdbfile [100];
  char   chnfileTmp [100];
  char   chnfileTmp2[100];
  char   chnfile [100];
  int    i;
  char sidedb[MAX_FNAME_LEN];

  /* Kludge.. need to figure out what pdb file name user chose */
  tmpfilep = fopen (params.pdbtmp, "r");
  fscanf (tmpfilep, "%s", usrfile);
  fclose (tmpfilep);
  /* END kludge */

  sprintf(sidedb, "%s%c%s", params.shifty_home, '/', SIDE_CHAIN_DB);

  load_aa_db(sidedb);

  /* Put in the side chains for the alignments */
  for (i=0; i<topAlignNum; i++)
    {
      initialize();
      sprintf (pdbfile,"%s%d.pdb", usrfile, i+1);
      sprintf (chnfileTmp,"%s.side.tmp", pdbfile);
      sprintf (chnfileTmp2,"%s.side.tmp2", pdbfile);
      sprintf (chnfile,"%s.side", pdbfile);
      parse_pdb_side_chain_data (pdbfile);
      get_chi_angle(i);
      attach_side_chains (pdbfile, chnfileTmp, sidedb, params.tmpdir);
      
      /* This... */
      remove_extra_side_chain_atoms(chnfileTmp2, chnfileTmp, params.tmpdir);
      renumber_atoms (chnfile, chnfileTmp2, params.tmpdir);

      /* or this */
     // renumber_atoms (chnfile, chnfileTmp, params.tmpdir);
    } /* END _for_ */

#ifdef FILDEBUG
  print_res_list ();
#endif

  shut_down_cleanup ();

} /* END do_side_chain_stuff () */

/* END MODULE: sidechain.c */
