/**************************************************************************
  phipsi.c
     Generates a file of internal angles and coordinates from a PDB File

  Date:	
     March, 2003

  Author:
     Aaron Davidson @ BTI
  
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ga_util.h"
#include "vec3D.h"
#include "pdbfile.h"
#include "phipsi.h"

#define MAX_ATOMS 30 /* a maximum number of atoms in a residue */

/* constants for atom indexs for a residue */
#define CA_pp 	0
#define C1_pp 	1
#define O_pp  	2
#define N_pp  	3
#define CA2_pp  4
#define H_pp	5


/*************************************************************/
/* calculate the diffence of the vectors x and y.            */
/*************************************************************/
static void atom_difference(Atom *A, Atom *B, Vec3D dif) {
	difference(A->point, B->point, dif);
}

/*************************************************************/
/* Calculate torsion angle for four atoms                    */
/*************************************************************/
static double torsion(Atom *A, Atom *B, Atom *C, Atom *D) {
	Vec3D X, Y, AB, BC, CD, cXY;
	double result;

	atom_difference(A, B, AB);
	atom_difference(B, C, BC);
	atom_difference(C, D, CD);

	crossproduct(AB, BC, X);
	crossproduct(BC, CD, Y);

   result = angle(X, Y);
   crossproduct(X, Y, cXY);
   if (dotproduct(BC, cXY) > 0.0) {
      result = -result;
	}
   return(result);
}

/*************************************************************/
/* Rotate the atoms in a residue into the template position. */
/*************************************************************/
static double canonicalForm(Vec3D *vecs, int n) {
	int i;
	Vec3D xAxis = {1,0,0}; // a unit vector
	Vec3D T; // a temp vector
	double theta;
	
	//	printf("\n");
	for (i=1; i<n; i++) difference(vecs[i], vecs[CA_pp], vecs[i]);
	difference(vecs[CA_pp], vecs[CA_pp], vecs[CA_pp]);

	// compute angle from X axis
	setVector(T, vecs[C1_pp][0], vecs[C1_pp][1], 0);
	theta = angle(T, xAxis);	
	if (vecs[C1_pp][1] > 0) theta = 360 - theta;
	// rotate atoms around Z axis
	for (i=0; i<n; i++) rotateZ(vecs[i], theta);

	// compute angle from X axis
	setVector(T, vecs[C1_pp][0], 0, vecs[C1_pp][2]);
	theta = angle(T, xAxis);
	if (vecs[C1_pp][2] < 0) theta = 360 - theta;
	// rotate atoms around Y-axis
	for (i=0; i<n; i++) rotateY(vecs[i], theta);

	// compute angle for N
	theta = atan(vecs[N_pp][2] / vecs[N_pp][1]);
	theta = 180 - (theta * 180/PI_GA);
	if (vecs[N_pp][1] < 0) theta = 180+theta;
	for (i=0; i<n; i++) rotateX(vecs[i], theta);

	return 0.0;
}


/*************************************************************/
/* Read out the atoms for the residue starting at n. Compute */
/* local coordinates for the set of atoms.                   */
/*************************************************************/
static void measureResidueBonds(PDBFile *pdb, int n,Vec3D *vecs, char names[MAX_ATOMS][4], int *numAtoms) {
	int i;
	Atom *a;
	int iCA, iC1, iO, iN, iCA2;
	int ni = n;
	int res = pdb->atoms[ni].resNum;

	iCA = getNextAtomIndex(pdb, n++, "CA");
	a = &pdb->atoms[iCA];
	strncpy(names[CA_pp], a->name, 4);
	copyVector(a->point, vecs[CA_pp]);

	iC1 = getNextAtomIndex(pdb, n++, "C ");
	a = &pdb->atoms[iC1];
	strncpy(names[C1_pp], a->name, 4);
	copyVector(a->point, vecs[C1_pp]);

	iO = getNextAtomIndex(pdb, n++, "O ");
	if (iO >= 0) {
		a = &pdb->atoms[iO];
		strncpy(names[O_pp], a->name, 4);
		copyVector(a->point, vecs[O_pp]);
	}

	iN = getNextAtomIndex(pdb, n++, "N ");
	if (iN >= 0) {
		a = &pdb->atoms[iN];
		strncpy(names[N_pp], a->name, 4);
		copyVector(a->point, vecs[N_pp]);
	}

	iCA2 = getNextAtomIndex(pdb, n++, "CA");
	if (iCA2 >= 0) {
		a = &pdb->atoms[iCA2];
		strncpy(names[CA2_pp], a->name, 4);
		copyVector(a->point, vecs[CA2_pp]);
	}

	n = H_pp + 1; // skip this one (hacky)

	// side-chain
	for (i=ni; i<pdb->numAtoms && pdb->atoms[i].resNum == res; i++) {
		if (i != iCA && i != iC1 && i != iO && i != iN && i != iCA2) {
			a = &pdb->atoms[i];
			strncpy(names[n], a->name, 4);
			copyVector(a->point, vecs[n++]);
		}
	}
	*numAtoms = n;
	
	canonicalForm(vecs, n); // do rotations of current coordinates

	//iCA = iC1 = iO = iN = iCA2 = -1; // blah
	// set any unfound atoms to defaults
	if (iCA < 0) setVector(vecs[CA_pp], 0.0, 0.0, 0.0);
	if (iC1 < 0) setVector(vecs[C1_pp], 1.53, 0.0, 0.0);
	if (iO < 0)	setVector(vecs[O_pp], 2.1686, 1.0629, 0.0);
	if (iN < 0)	setVector(vecs[N_pp], 2.0669, -1.2059, 0.0);
	if (iCA2 < 0) setVector(vecs[CA2_pp], 3.5188, -1.4359, 0.0);
	vecs[CA2_pp][2] = 0;	// hack

	// set HN no matter what
	strncpy(names[H_pp], "H ", 4);
	setVector(vecs[H_pp], 1.52, -2.04, 0.0); 
}



/*	extra code for outputting side chain coordinates: */
static void printResidueCoordinates(char *name, int num,  double angles[4],
					Vec3D *template, char names[MAX_ATOMS][4], int numAtoms,FILE *fpw) {
	int i;
	printf("%s \t %lf \t %d\n", name, angles[3], numAtoms);
   for (i=0; i<numAtoms; i++) {
		fprintf(fpw," %4s %9.4f %9.4f %9.4f\n", 
			names[i], template[i][0], template[i][1],  template[i][2]);
	}
}



/*************************************************************/
/* Print the actual line of output for a residue.            */
/*************************************************************/
static void printResidue(char *name, int num, double angles[4], 
					Vec3D *template, char names[MAX_ATOMS][4], int numAtoms, FILE *fpw) {
	fprintf(fpw,"%d  %s  %11.6lf %11.6lf %11.6lf %11.6lf\n",num+1, name, angles[0], angles[1], angles[2], angles[3]);
}



/*************************************************************/
/* get the next atom in the backbone                         */
/*************************************************************/
static int getNextBackboneAtom(int n, PDBFile *pdb) {
	int i;
	for (i=n; i<pdb->numAtoms; i++)	{
		if (n == 0) {
			/* start backbone on an N only */
			if ( (pdb->atoms[i].type == N_TYPE) ) {
				return i;
			}
		} else {
			int oldtype = pdb->atoms[n-1].type;
			int curtype = pdb->atoms[i].type;
			if (curtype != oldtype) {
				if ((curtype == CA_TYPE) || (curtype == C_TYPE) || (curtype == N_TYPE)) {
					return i;
				}
			}
		}
   }
	return -1;
}


/*************************************************************/
/* Get the next atom after index n, of the given type.       */
/* Returns -1 if none found.                                 */
/*************************************************************/
static int getAtomOfType(int n, PDBFile *pdb, int type) {
	int resNum = pdb->atoms[n].resNum;	
	while (pdb->atoms[n].resNum == resNum) {
		if (pdb->atoms[n].type == type) return n;
		if (++n >= pdb->numAtoms) return -1;
	}
	return -1;
}


/*************************************************************/
/* Compute & print the dihedral angles of the PDB file       */
/*************************************************************/
static int dihedralAngles(PDBFile *pdb, int makeDB, FILE *fpw) {
	double angles[4] = {0, 0, 0, 0};
	Vec3D template[MAX_ATOMS];
	char names[MAX_ATOMS][4];
	int numAtoms;
	int atoms[4];
	int a=1, n=0, count=0;

	 
	 /* Read in the first 4 atoms into our rotating array.
	    We start at the first N, to compute PSI */
	 atoms[0] = getNextBackboneAtom(0, pdb);
	 atoms[1] = getNextBackboneAtom(atoms[0]+1, pdb);
	 atoms[2] = getNextBackboneAtom(atoms[1]+1, pdb);
	 atoms[3] = getNextBackboneAtom(atoms[2]+1, pdb);
	 
	while (atoms[n] >= 0 && atoms[(n+3)%4] >= 0) {

		angles[a] = torsion(
				&pdb->atoms[atoms[n]], 
				&pdb->atoms[atoms[(n+1)%4]],
			  	&pdb->atoms[atoms[(n+2)%4]], 
				&pdb->atoms[atoms[(n+3)%4]] );

		// compute N's chi angle
		if (pdb->atoms[atoms[n]].type == N_TYPE) {
			// compute chi latte angle:
			int cbi = getAtomOfType(atoms[n], pdb, CB_TYPE);
			int cgi = getAtomOfType(atoms[n], pdb, CG_TYPE);
			if (cbi > 0 && cgi > 0) {
				angles[3] = torsion(&pdb->atoms[atoms[n]],
							            &pdb->atoms[atoms[(n+1)%4]],
											&pdb->atoms[cbi], &pdb->atoms[cgi]);
			} else {
				angles[3] = 360;
			}
		}

		if (a == 2) { /* last angle */
			measureResidueBonds(pdb, atoms[n], template, names, &numAtoms);

			if (makeDB) {
				printResidueCoordinates(pdb->atoms[atoms[n]].resName, count++, angles, template, names, numAtoms,fpw);
			} else {
				printResidue(pdb->atoms[atoms[n]].resName, count++, angles, template, names, numAtoms,fpw);
			}
			angles[0] = angles[1] = angles[2] = 0;
		}
		atoms[n] = getNextBackboneAtom(atoms[(n+3)%4]+1, pdb);
		
		n = (n+1)%4;  /* rotate the pointer */
		a = (a+1)%3;  /* the angle we will read next */ 
	}

	if (a < 3 && a > 0) {
		measureResidueBonds(pdb, atoms[n], template, names, &numAtoms);
		if (makeDB) {
			printResidueCoordinates(pdb->atoms[atoms[n]].resName, count++, angles, template, names, numAtoms,fpw);		
		} else {
			printResidue(pdb->atoms[atoms[n]].resName, count++, angles, template, names, numAtoms,fpw);
		}
	}

	return 0;
}


//*******************************************************
int call_phipsi(char* input, char* output){
  
  char chain = '?'; // read first chain found
  int db = FALSE;
  PDBFile pdb;
  FILE *fpw;  
  
  fpw = fopen(output, "w");
  if(!fpw){printf("Cannot open %s for writing..exiting..\n",output);exit(-1);}
  parsePDBFile(input, &pdb, chain);
  
  dihedralAngles(&pdb, db, fpw);
  freePDBFile(&pdb);
  fclose(fpw);

  return 0;
  
}// END call_phipsi *************************************



#if 0
/*************************************************************/
/* The main entry point                                      */
/*************************************************************/
int main_phipsi(int argc, char **argv) {
	char chain = '?'; // read first chain found
	int db = FALSE;
 	char *input;
	PDBFile pdb;
	int i;
	
	if (argc < 2) {
		printf("Usage:\n   phipsi [--all_chains] [--db] file.pdb\n");
		printf("Options:\n");
		printf(" --all_chains\n\t Process all chains in the pdb file\n");
		printf(" --db\n\t Print out residue template for the database\n");
		return 0;
	}
	
	for (i=1; i<argc-1; i++) {
		if (argc >= 3) {
			if (strcmp(argv[i], "--all_chains") == 0) {
				chain = '*'; // read all chains
			} else if (strcmp(argv[i], "--db") == 0) {
				db = TRUE;
			}
		}
	}
	
	input = argv[argc-1];

	parsePDBFile(input, &pdb, chain);

	fprintf(stderr, "Loaded %s: %d atoms\n", input, pdb.numAtoms);
	
	dihedralAngles(&pdb, db);
	freePDBFile(&pdb);
	return 0;
}

#endif
