/* MODULE: util.h */
#ifndef __UTIL_H
#define __UTIL_H

/***OVERVIEW************************************************
 * File Name     : util/util.h
 * Purpose       : Generally useful functions and features.
 * Author        : Philip S Luk ~ philipluk@bigfoot.com
 * Date Written  : July 9, 2001
 * Last Modified : July 9, 2001
 ***END*OVERVIEW********************************************/

/* MODULE DEFINITIONS */

  #define helix            "HELIX"
  #define coil             "COIL"
  #define REF_ATOM         "CA"
  #define N_ATOM           "N"
  #define C_ATOM           "C"
  #define CB_ATOM          "CB"
  #define O_ATOM           "O"
  #define GLYCINE          "GLY"
  #define FILE_WRITE	   "w"		/* fopen write macro       */
  #define FILE_READ	   "r"		/* fopen read macro        */
  #define FILE_APPEND	   "a"		/* fopen append macro       */
  #define MAX_LINE_LEN     250		/* Max string length macro */
  #define MAX_SHEET_IN_A_GROUP 20         /* max number of sheets present in a group */
  
  #define PDB_READ_FORMAT  "%4s %6d %3s%4s %5d %7lf %7lf %7lf %lf %lf %8s\n"
  #define PDB_WRITE_FORMAT "%4s %6d  %-3s%4s %5d %11.3f %7.3f %7.3f %5.2f %5.2f %13s\n"
/* END MODULE DEFINITIONS */

/* TYPE DEFINTIONS */
  typedef struct {
    double x; double y; double z;
  } POINT;				/* 3D Coordinate point type */

  typedef enum {
    false, true
  } boolean;

  typedef struct {
    char   flag[5];
    int    atom_no;
    char   atom[5];
    char   res[5];
    int    res_no;
    POINT  pt;
    double num1;
    double num2;
    char   extra[10];
  } PDB_LINE;
  
  
  typedef struct {
    POINT begin;
    POINT end;
    }PTS;
     
     
  typedef struct {
     POINT N_of_first_res, C_of_first_res;
     POINT N_of_last_res,  C_of_last_res;
     }ANCH_PTS;
	
  typedef struct {
     POINT first_CA, sec_CA, third_CA;
     POINT third_last_CA, sec_last_CA, last_CA;
     }SIX_CA_POINTS;	  
     
     
 /* The struct EIGHTPTS_OF_RES holds the N,CA,C & O points
     of the first and last residue of a strand. */    
   typedef struct {
      POINT first_N, first_CA;  
      POINT first_C, first_O;
      
      POINT last_N, last_CA;  
      POINT last_C, last_O;
      }EIGHTPTS_OF_RES; 
      
   typedef struct{
	 double Nx, Ny, Nz, CAx, CAy, CAz, Cx, Cy, Cz, Ox, Oy, Oz;
   }SVECTOR;
      
    
     
  typedef struct {
	  POINT firstpt; POINT secondpt; POINT middlept;
  }LOOP_ANCHPTS;
  
  typedef struct{
  float matrix[4][4];
  }MATRIX;
  
  
 /* Structure defining beta strand */
 /* len = number of residues of the beta strand */ 
 /* column no. from which the beta strand is taken */ 
 /* beginning index of the beta strand */
 /* last index of the beta strand */
 /* Library no. from which the beta strand is taking */
 /* Group no. to which the beta strand belongs */
  
  typedef struct { 
	  int index;
	  int len;   
	  int colno; 
	  int segment_counter;
	  int beg_resd_num;   
	  int last_resd_num;   
	  POINT beg_pt;
	  POINT end_pt; 
	  int lib_no; 
	  int group_no;  
  } BSTRAND_PROP;
  
  
  
  typedef struct {
	 int            no_of_sheets_in_the_group;   
	 BSTRAND_PROP  sheet_arr[MAX_SHEET_IN_A_GROUP];  /* We assume that the max num of beta sheets in a group won't exceed 20 */
	 int             group_no;
  } BSTRAND_GROUP;
  
  
  
  
/* END TYPE DEFINTIONS */

/* GLOBAL VARIABLES */
  const POINT origin_pt;
  /*origin_pt.x = 0.00; origin_pt.y = 0.00; origin_pt.z = 0.00;*/ 
/* END GLOBAL VARIABLES */

/* PROCEDURE DECLARATIONS */

/***********************************************************
 * Appends file 'appendee' to file 'appendor'. Warning, 
 * the appendee file is destroyed.
 ***********************************************************/
int append_file (char * appendor, char * appendee);

/***********************************************************
 * Assigns point b to point a;
 ***********************************************************/
int assign_pt (POINT * a, POINT b);
 
 
 /***************************************************
 * Prints the co-ordinates of the point
 **************************************************/
 void printPOINT(POINT *point);
 
 
 /***************************************************
 * Forms a vector which points from the start to the
 *  end co-ordinates.
 **************************************************/
POINT  formvector(PTS *points);
 

/***************************************************
 * Forms the negative of a vector *
 **************************************************/
POINT make_neg_vector(POINT *some_vector);


/**********************************************
 * Forms the unit vector of some_vector and 
 returns the point co-ords of the unit vector
 **********************************************/
POINT form_unitvector(POINT *some_vector);
 

/************************************************
*  Adds two vectors & returns the co-ords of the
   added vector
**************************************************/
POINT add_twovectors(POINT *vector1, POINT *vector2);
 

/************************************************
*  Returns the midpoint co-ords of the two points 
**************************************************/ 
POINT get_midpoint(POINT *pt1, POINT *pt2);
 
/**********************************************
* Scales some_vector by the factor and returns 
   the co-ords of the resultant point
***********************************************/
POINT scale(POINT *some_vector, float factor);
 
/***********************************************
 * Forms the vector pointing from pt1 to pt2 and
 returns the co-ords of the resultant point
 ***********************************************/
POINT formvector_betn_2pts(POINT *pt2, POINT  *pt1);
 

/********************************************
 Calculates the dot_product of the vectors v1 and v2 
 dot-product = v1.x*v2.x + v1.y*v2.y + v1.z*v2.z 
 **********************************************/
double get_dot_product(POINT *v1, POINT *v2);

 /*******************************************
 Calculates the norm of a vector
 norm = sqrt(v.x^2 + v.y^2 + v.z^2)
 ******************************************/
 double get_norm (POINT *v);
 
 
 /**********************************************
 Calculates the centroid of three points 
 centroid.x = (pt_1.x + pt_2.x + pt_3.x)/3; and so on..
 *********************************************/
 POINT get_centroid(POINT *pt_1, POINT *pt_2, POINT *pt_3);
 
 
 /***********************************************************	
 * Returns the distance between two points 
***********************************************************/ 
 float get_distance(POINT *pt1, POINT *pt2);

 
 
 void assign_first_res(EIGHTPTS_OF_RES *res, POINT **side);
 
 void assign_last_res(EIGHTPTS_OF_RES *res, POINT **side);
 
 
 /***********************************************************
 * Write one pdb line to the given pdb using data from
 * the given variable. Returns the number of items written.
 ***********************************************************/
int write_pdb_line (FILE * fp, PDB_LINE l);

/***********************************************************
 * Read one line from the given pdb file and stores it in 
 * the given variable. Returns the number of items read or
 * EOF if end of file is encountered.
 ***********************************************************/
int read_pdb_line (FILE * fp, PDB_LINE * line);



/***********************************************************
 * Converts the residue's 3 character representation to it's
 * single character representation.
 ***********************************************************/
char res_str_to_char (char * res_name);

/***********************************************************
 * Converts the residue's single character representation
 * to its 3 character representation.
 ***********************************************************/
char * char_to_res_str (char res);

/***********************************************************
 * Returns true if extra is HELIX
 ***********************************************************/
boolean is_helix (char * extra);


/***********************************************************
 * Returns true if extra is COIL
 ***********************************************************/
boolean is_coil (char * extra);


/***********************************************************
 * Returns true if given atom is the rotation reference
 * point atom.
 ***********************************************************/
boolean is_ref_atom (char * atom);

/***********************************************************
 * Returns true if given atom is the Nitrogen atom.
 ***********************************************************/
boolean is_N_atom ( char * atom);

/***********************************************************
 * Returns true if given atom is the Carbon atom.
 ***********************************************************/
boolean is_C_atom ( char * atom);

/***********************************************************
 * Returns true if given atom is the CB atom.
 ***********************************************************/
boolean is_CB_atom ( char * atom);

/***********************************************************
 * Returns true if given atom is the Oxygen atom.
 ***********************************************************/
boolean is_O_atom ( char * atom);

/***********************************************************
 * Returns true if given residue is GLYCINE.
 ***********************************************************/
boolean is_GLY_residue ( char * res);
/* END PROCEDURE DECLARATIONS */

#endif
/* END MODULE: util.h */
