/**************************************************************************
  vec3D.c
     Routines for doing basic 3D vector math.

  Date:	
     March, 2003

  Author:
     Aaron Davidson @ BTI
  
 **************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vec3D.h"
//#include "../util/memwatch.h"

/*************************************************************/
/* Set diff to v1 - v2                                       */
/*************************************************************/
void difference(Vec3D v1, Vec3D v2, Vec3D diff) {
	diff[0] = v1[0] - v2[0];
	diff[1] = v1[1] - v2[1];
	diff[2] = v1[2] - v2[2];
}

/*************************************************************/
/* Set sum to v1 + v2                                        */
/*************************************************************/
void summation(Vec3D v1, Vec3D v2, Vec3D sum) {
	sum[0] = v1[0] + v2[0];
	sum[1] = v1[1] + v2[1];
	sum[2] = v1[2] + v2[2];
}


/*************************************************************/
/* Scales the vector by a scalar value                       */
/*************************************************************/
void scale3d(Vec3D vec, double scalar) {
	vec[0] = vec[0] * scalar;
	vec[1] = vec[1] * scalar;
	vec[2] = vec[2] * scalar;
}

/*************************************************************/
/* Turn the vector into a Unit Vector                        */
/*************************************************************/
void unitVector(Vec3D vec, Vec3D uvec) {
	double mag = sqrt(dotproduct(vec,vec));
	uvec[0] = vec[0]/mag;
	uvec[1] = vec[1]/mag;
	uvec[2] = vec[2]/mag;
}
/*************************************************************/
/* Turn the vector into a Unit Vector                        */
/*************************************************************/
void UnitVector(Vec3D vec) {
	double mag = sqrt(dotproduct(vec,vec));
	vec[0] = vec[0]/mag;
	vec[1] = vec[1]/mag;
	vec[2] = vec[2]/mag;
}

/* calculate the cross product of vectors x and y.           */
/*************************************************************/
void crossproduct(Vec3D a, Vec3D b, Vec3D result) {
   result[0] = (a[1]*b[2]) - (b[1]*a[2]);
   result[1] = (a[2]*b[0]) - (b[2]*a[0]);
   result[2] = (a[0]*b[1]) - (b[0]*a[1]);
} 

/*************************************************************/
/* calculate the dot product of vectors x and y.             */
/*************************************************************/
double dotproduct(Vec3D x, Vec3D y) {
	return ((x[0]*y[0]) + (x[1]*y[1]) + (x[2]*y[2]));
}


/*************************************************************/
/* calculate the angle between two vectors defined by the    */
/* ends of the vectors.                                    */
/*************************************************************/
double angle(Vec3D X, Vec3D Y) {
	double XY = dotproduct(X, Y);
	double Xlen = sqrt(dotproduct(X,X));
	double Ylen = sqrt(dotproduct(Y,Y));
	double theta = acos( XY / (Xlen*Ylen) );
	theta = theta*(180.0/PI_GA); /* convert to degrees */
	return theta;
}

/*************************************************************/
/* rotate r around n by theta counter clocwise               */
/*************************************************************/
void rotateCC(Vec3D n, Vec3D r, double theta) {
	double A, B, C;
	Vec3D q;
   if (theta == 180 || theta == -180) {
      A = -1.0;
      C = 0.0;
   } else {
      theta *= PI_GA/180;
      A = cos(theta);
      C = sin(theta);
   }
   B = (n[0]*r[0]+n[1]*r[1]+n[2]*r[2])*(1.0 - A);
   q[0] = A*r[0] + B*n[0] - C*(r[1]*n[2] - r[2]*n[1]);
   q[1] = A*r[1] + B*n[1] - C*(r[2]*n[0] - r[0]*n[2]);
   q[2] = A*r[2] + B*n[2] - C*(r[0]*n[1] - r[1]*n[0]);
	r[0] = q[0]; r[1] = q[1]; r[2] = q[2];
}


/*************************************************************/
/* Rotate r around the x-axis by theta                       */
/*************************************************************/
void rotateX(Vec3D r, double theta) {
	Vec3D U;
	setVector(U, 1, 0, 0);
	rotateCC(U, r, theta);
}

/*************************************************************/
/* Rotate r around the y-axis by theta                       */
/*************************************************************/
void rotateY(Vec3D r, double theta) {
	Vec3D U;
	setVector(U, 0, 1, 0);
	rotateCC(U, r, theta);
}

/*************************************************************/
/* Rotate r around the z-axis by theta                       */
/*************************************************************/
void rotateZ(Vec3D r, double theta) {
	Vec3D U;
	setVector(U, 0, 0, 1);
	rotateCC(U, r, theta);
}

/*************************************************************/
/* copy src into dst                                         */
/*************************************************************/
void copyVector(Vec3D src, Vec3D dst) {
	dst[0] = src[0];
	dst[1] = src[1];
	dst[2] = src[2];
}

/*************************************************************/
/* Set the points of the vector all to val                   */
/*************************************************************/
void initVector(Vec3D vec, double val) {
	vec[0] = vec[1] = vec[2] = val;
}


/*************************************************************/
/* Set the points of the vector                              */
/*************************************************************/
void setVector(Vec3D vec, double x, double y, double z) {
	vec[0] = x;
	vec[1] = y;
	vec[2] = z;
}


/*************************************************************/
/* Compute the distance between two points                   */
/*************************************************************/
double distance(Vec3D a, Vec3D b) {
	return sqrt( (a[0]-b[0])*(a[0]-b[0]) + 
					 (a[1]-b[1])*(a[1]-b[1]) + 
					 (a[2]-b[2])*(a[2]-b[2]) );
  
}


void normalize3d(Vec3D a, Vec3D c) {
	double mag = sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
  	c[0] = a[0]/mag;
  	c[1] = a[1]/mag;
  	c[2] = a[2]/mag;
}

void makevector (Vec3D a, Vec3D c, Vec3D newvec) {
	setVector( newvec, (a[0] - c[0]), (a[1] - c[1]), (a[2] - c[2]) );
}

void printvector (Vec3D a) {
	printf("%.3f, %.3f, %.3f", a[0], a[1], a[2]);
}
