
/*

	file...: align.c

	These routines perform the fundamental matrix manipulations
	of the dynamic programming model to calculate optimal
	alignments of character strings.

	author.: Robert Boyko
	date...: Nov 29, 1995	
	place..: University of Alberta / Department of Biochemistry

*/

#include "../include/seqsee.h"
#include "../include/align.h"
#include <stdlib.h>

/* procedure calls */
static int insert_mSeq_gap(), insert_dSeq_gap();


/*************************************************************/
/* init Needleman-Wunsch scoring matrix. 
*/
void sm_mat_init(sm, rows, cols, zeroFlag)
SM_REC *sm;
int rows, cols, zeroFlag;
{
	register int i, elements;
	register SM_REC *p1;

/* init scores to zero? */
	if (zeroFlag == TRUE) {
		elements = rows * cols;
		for (i=0; i<elements; i++)
			sm[i].scor = 0;
	}

/* init pointer field in last column of matrix */
	p1 = &sm[cols - 1];
	for (i=0; i<rows; i++) {
		p1->next = NULL;
		p1 += cols;
	}

/* init pointer field in last row of matrix */
	p1 = &sm[(rows - 1) * cols];
	for (i=0; i<cols; i++) {
		p1->next = NULL;
		p1++;
	}

}	/* sm_mat_init */


/*************************************************************/
/* Determine initial scoring matrix for Needleman-Wunsch algorithm
** for 2 sequences. Matrix has no previous values.
*/
void sm_mat_score(wtMat, sm, qLen, dLen, qArr, dArr)
register int *wtMat;		/* amino acid similarity matrix */
SM_REC *sm;					/* scoring matrix */
register int qLen, dLen;	/* lengths of both sequences */
int *qArr, *dArr;			/* amino acids indices */
{
	register int m, k;
	register int *matRow, *dix;
	register SM_REC *smPtr;

	smPtr = sm;
	for (k=0; k<qLen; k++) {			/* calc each row in sm */
		m = dLen;						/* last entry */
		dix = dArr;						/* column index */
		matRow = wtMat + qArr[k];		/* row index for wtMat */ 

		do {
			smPtr->scor = matRow[*dix++]; /* calc each entry in mat row */
			smPtr++;
		} while (--m);

	}

}	/* sm_mat_score */


/*************************************************************/
/* Add to the initial scoring matrix for Needleman-Wunsch algorithm
** for 2 sequences.
*/
void sm_mat_add_score(wtMat, sm, qLen, dLen, qArr, dArr)
register int *wtMat;		/* amino acid similarity matrix */
SM_REC *sm;					/* scoring matrix */
register int qLen, dLen;	/* lengths of both sequences */
int *qArr, *dArr;			/* amino acids indices */
{
	register int m, k;
	register int *matRow, *dix;
	register SM_REC *smPtr;

	smPtr = sm;
	for (k=0; k<qLen; k++) {			/* calc each row in sm */
		m = dLen;						/* last entry */
		dix = dArr;						/* column index */
		matRow = wtMat + qArr[k];		/* row index for wtMat */ 

		do {
			smPtr->scor += matRow[*dix++]; /* calc each entry in mat row */
			smPtr++;
		} while (--m);

	}

}	/* sm_mat_add_score */


/*************************************************************/
/* Finds summation of Needleman-Wunsch matrix given a penalty
** for the number of gaps, gap size, gaps in secondary structure. 
*/
void sm_mat_sum(sm, qLen, dLen, qsScore, dsScore, qPenGroup, dPenGroup, gap, gapSize)
SM_REC *sm;
int	qLen, dLen;
int *qsScore, *dsScore;
int *qPenGroup, *dPenGroup;
int gap, gapSize;
{
	register int diagVal, rowMax, colMax;
	register int x1, m;
	SM_REC *p1, *p2, *rowPtr, *dColumns, *colPtr;
	int	k, sGapPen, rowsGapPen, *colsGapPen;

	/*SCOTT
dColumns = (SM_REC *)emalloc((dLen*sizeof(SM_REC)), "sm_mat_sum");
*/
        dColumns = (SM_REC *)malloc((dLen*sizeof(SM_REC)));

	for (m=0; m<dLen; m++)
		dColumns[m].scor = SUM_SMAT_INIT;

/*SCOTT	colsGapPen = (int *)emalloc((dLen*sizeof(int)), "sm_mat_sum");*/
        colsGapPen = (int *)malloc((dLen*sizeof(int)));
	for (m=0; m<dLen; m++)
		colsGapPen[m] = 0;

	p1 = &sm[(qLen*dLen)-1];
	p2 = p1 - dLen - 1; 
	rowPtr = p1;

	for (k=qLen-1; k>0; k--) {			/* summation of each row */
		m = dLen-1;						/* process dLen-1 columns */
		rowMax = SUM_SMAT_INIT;
		rowsGapPen = 0;
		do {							/* compute each matrix entry */
			diagVal = p1->scor;
			colPtr = &dColumns[m];
			colMax = colPtr->scor;

			/* determine new matrix entry (p2) */
			if ((diagVal >= rowMax) && (diagVal >= colMax)) {
				p2->scor += diagVal;
				p2->next = p1;
			}
			else if (rowMax >= colMax) {
				p2->scor += rowMax;
				p2->next = rowPtr;
			}
			else {
				p2->scor += colMax;
				p2->next = colPtr->next;
			}

			/* find maximum value in current row */
			sGapPen = MIN(dsScore[m-1], dsScore[m]) + 
					  MIN(qsScore[k-1], qsScore[k]);
			if (sGapPen > rowsGapPen) {
				rowMax -= sGapPen - rowsGapPen;
				rowsGapPen = sGapPen;
			}

			x1 = diagVal - gap - qPenGroup[k] - sGapPen;
			if (x1 > rowMax) {
				rowPtr = p1;
				rowMax = x1;
				rowsGapPen = sGapPen;
			}
			rowMax -= gapSize;

			/* find maximum value in current columns */
			if (sGapPen > colsGapPen[m]) {
				colPtr->scor -= sGapPen - colsGapPen[m];
				colsGapPen[m] = sGapPen;
			}

			x1 = diagVal - gap - dPenGroup[m] - sGapPen;
			if (x1 > dColumns[m].scor) {
				colPtr->next = p1;
				colPtr->scor = x1;
				colsGapPen[m] = sGapPen;
			}
			colPtr->scor -= gapSize;

			p2--;
			p1--;

		} while (--m);
		p1--;
		p2--;
	}

	free(dColumns);
	free(colsGapPen);

}	/* sm_mat_sum */


/*************************************************************/
/* Returns pointer to best score in Needleman Wunsch matrix.
*/
SM_REC *find_path_start(sm, qLen, dLen)
SM_REC *sm;
register int qLen, dLen;
{
	register int j, bestScor;
	SM_REC *p1, *bestPtr;

/* find maximum value of first row */
	bestScor = 0;
	p1 = &sm[0];
	for (j=0; j<dLen; j++) {
		if (p1->scor > bestScor) {
			bestScor = p1->scor;
			bestPtr = p1;
		}
		p1++;
	}
/* check to see if there is larger value in the first column */
	p1 = &sm[dLen];
	for (j=1; j<qLen; j++) { 
		if (p1->scor > bestScor) {
			bestScor = p1->scor;
			bestPtr = p1;
		}
		p1 += dLen;
	}

	return(bestPtr);

}	/* find_path_start */

 
/*************************************************************/
/* Parse the summed Needleman-Wunsch matrix finding the optimal
** alignment for two strings. Return the new string length.
*/
int find_best_align(top, sm, mLen, dLen, mGapStr, dGapStr, maxSize)
SM_REC *top, *sm;
int mLen, dLen, maxSize;
char *mGapStr, *dGapStr;
{
	register int len;
	register int nextRow, nextCol, row, col;
	int ix;
	SM_REC *p1;
	char sVal[MAX_LINE_LEN];

	p1 = top;
	row = col = -1;
	len = 0;
	while (p1 != NULL) {
		
		ix = p1 - sm;
		nextRow = ix / dLen;
		nextCol = ix % dLen;

		if ((row + 1) < nextRow)
			len = insert_dSeq_gap(row+1, nextRow, len, mGapStr, dGapStr);

		else if ((col + 1) < nextCol)
			len = insert_mSeq_gap(col+1, nextCol, len, mGapStr, dGapStr);

		mGapStr[len] = TEMPLATE_RES;
		dGapStr[len] = TEMPLATE_RES;
		len++;

		row = nextRow;
		col = nextCol;
		p1 = p1->next;
	}

	if ((row + 1) < mLen)
		len = insert_dSeq_gap(row+1, mLen, len, mGapStr, dGapStr);
	if ((col + 1) < dLen)
		len = insert_mSeq_gap(col+1, dLen, len, mGapStr, dGapStr);

	if (len > maxSize) {
		sprintf(sVal, "Alignment size exceeded, len = %d\n", maxSize);
		show_err(TRUE, "find_best_align", sVal);
	}

	mGapStr[len] = '\0';
	dGapStr[len] = '\0';
	return(len);

}	/* find_best_align */


/*************************************************************/
/* insert proper size gap into the added sequence. 
*/
static int insert_dSeq_gap(startRow, endRow, ix, mGapStr, dGapStr)
register int startRow, endRow, ix;
char *mGapStr, *dGapStr;
{
	register int j;

	for (j=startRow; j<endRow; j++) {
		mGapStr[ix] = TEMPLATE_RES;
		dGapStr[ix] = TEMPLATE_GAP;
		ix++;
	}

	return(ix);

}	/* insert_dSeq_gap */


/*************************************************************/
/* insert proper size gap into current mult align sequence. 
*/
static int insert_mSeq_gap(startCol, endCol, ix, mGapStr, dGapStr)
register int startCol, endCol, ix;
char *mGapStr, *dGapStr;
{
	register int j;

	for (j=startCol; j<endCol; j++) {
		mGapStr[ix] = TEMPLATE_GAP;
		dGapStr[ix] = TEMPLATE_RES;
		ix++;
	}

	return(ix);

}	/* insert_mSeq_gap */
