
/*

	file...: pair_align.c
	purpose: low-level routines for pairwise alignments 

	author.: Robert Boyko
	date...: Nov 29, 1995	
	place..: University of Alberta / Department of Biochemistry

*/

#include "../include/seqsee.h"
#include "../include/align.h"
#include "../include/palign.h"
#include <stdlib.h>

/* static vars */

static SM_REC *sm;

/* procedure calls */

static void calc_pair_align(), save_seq_align();
extern void sm_mat_init(), sm_mat_score(), sm_mat_add_score(), sm_mat_sum();
extern SM_REC *find_path_start();
extern int find_best_align(), pct_start_align(), pct_end_align();
extern int pct_homol_score(), pct_perfect_score();
extern void calc_struct_scores();
extern double scor_pair_align(), scor_align_pct();
extern void codes_to_indices();


/*************************************************************/
/* Performs pairwise alignments
*/
void do_pair_align(pAlign, s0, s1, wtMat, ssMat, gap, gapSize, midPen, endPen, endLen, structVal, saveFlag)
ALIGN_REC *pAlign;
SEQ_REC *s0, *s1;
int *wtMat, *ssMat;
int gap, gapSize, midPen, endPen, endLen, structVal, saveFlag;
{

	int newLen;
	char aGapStr[MAX_ALIGN_LEN], bGapStr[MAX_ALIGN_LEN];
	SM_REC *top;

/* calculate alignment */

	calc_pair_align(s0, s1, wtMat, ssMat, gap, gapSize, midPen, endPen, endLen);
	top = find_path_start(sm, s0->len, s1->len);
	newLen = find_best_align(top, sm, s0->len, s1->len, aGapStr, bGapStr,
		MAX_ALIGN_LEN);

/* save alignment */

	save_seq_align(pAlign, s0, s1, newLen, aGapStr, bGapStr,
		wtMat, structVal, saveFlag);

	free(sm);

}	/* do_align */


/*************************************************************/
/* calc matrix operations for alignment of two sequences. 
*/
static void calc_pair_align(s0, s1, wtMat, ssMat, gap, gapSize, midPen, endPen,
endLen)
SEQ_REC *s0, *s1;
int *wtMat, *ssMat, gap, gapSize, midPen, endPen, endLen;
{
	register int j, aLen, bLen, size;
	int aSeqNums[MAX_SEQ_LEN], bSeqNums[MAX_SEQ_LEN];
	int asSeqNums[MAX_SEQ_LEN], bsSeqNums[MAX_SEQ_LEN];
	int aPenGroup[MAX_SEQ_LEN], bPenGroup[MAX_SEQ_LEN];
	int aStructNums[MAX_SEQ_LEN], bStructNums[MAX_SEQ_LEN];

	/* lengths of the 2 sequences */
	aLen = s0->len;
	bLen = s1->len;

	/* allocate space for Needleman-Wunsch scoring matrix */
	size = aLen * bLen * sizeof(SM_REC);
	//sm = (SM_REC *)emalloc(size, "calc_pair_align: sm");
        sm = (SM_REC *)malloc(size);
	sm_mat_init(sm, aLen, bLen, TRUE);

	/* change sequences into indices for wtMat */
	codes_to_indices(s0->pRes, aLen, aSeqNums, BLANK_POS);
	codes_to_indices(s1->pRes, bLen, bSeqNums, BLANK_POS);
	for (j=0; j<aLen; j++)
		aSeqNums[j] *= WT_MAT_DIM;

	/* there are no grouping penalties for pairwise alignments */
	for (j=0; j<aLen; j++) aPenGroup[j] = 0;
	for (j=0; j<bLen; j++) bPenGroup[j] = 0;

	/* calculate secondary structure gapping penalties */
	calc_struct_scores(s0->sRes, aLen, aStructNums, midPen, endPen, endLen);
	calc_struct_scores(s1->sRes, bLen, bStructNums, midPen, endPen, endLen);

	/* calc initial scoring matrix just for the sequences */
	sm_mat_score(wtMat, sm, aLen, bLen, aSeqNums, bSeqNums);

	/* add to scoring matrix for secondary structure identity */
	if ((s0->sFlag == TRUE) && (s1->sFlag == TRUE)) {
		codes_to_indices(s0->sRes, aLen, asSeqNums, BLANK_POS);
		codes_to_indices(s1->sRes, bLen, bsSeqNums, BLANK_POS);
		for (j=0; j<aLen; j++)
			asSeqNums[j] *= WT_MAT_DIM;
		sm_mat_add_score(ssMat, sm, aLen, bLen, asSeqNums, bsSeqNums);
	}

	/* process our scoring matrix */
	sm_mat_sum(sm, aLen, bLen, aStructNums, bStructNums,
		aPenGroup, bPenGroup, gap, gapSize);

}	/* calc_pair_align */


/*************************************************************/
/* save the alignment of the sequences  
*/
static void save_seq_align(p, s0, s1, newLen, aGapStr, bGapStr, wtMat, aStruct, saveFlag)
ALIGN_REC *p;
SEQ_REC *s0, *s1;
char *aGapStr, *bGapStr;
int newLen, *wtMat, aStruct, saveFlag;
{

/* calc alignment of sequence 1 */
 
	/* SCOTT p->aSeq[0].pRes = (char *)malloc(newLen+1, "save_seq_align"); */
        p->aSeq[0].pRes = malloc(newLen+1);
	str_match_blanks(s0->pRes, s0->len, aGapStr, newLen, 
		TEMPLATE_GAP, p->aSeq[0].pRes);
	if (s0->sFlag == TRUE) {
/*		SCOTT p->aSeq[0].sRes = (char *)emalloc(newLen+1, "save_seq_align");*/
                p->aSeq[0].sRes = malloc(newLen+1);
		str_match_blanks(s0->sRes, s0->len, aGapStr, newLen, 
			TEMPLATE_GAP, p->aSeq[0].sRes);
	}

/* calc alignment of sequence 2 */

/*SCOTT	p->aSeq[1].pRes = (char *)emalloc(newLen+1, "save_seq_align");*/
        p->aSeq[1].pRes = malloc(newLen+1);
	str_match_blanks(s1->pRes, s1->len, bGapStr, newLen, 
		TEMPLATE_GAP, p->aSeq[1].pRes);
	if (s1->sFlag == TRUE) {
/*SCOTT		p->aSeq[1].sRes = (char *)emalloc(newLen+1, "save_seq_align");*/
                p->aSeq[1].sRes = (char *)malloc(newLen+1);
		str_match_blanks(s1->sRes, s1->len, bGapStr, newLen,
			TEMPLATE_GAP, p->aSeq[1].sRes);
	}

/* save length of alignment, percent homology and align score */

	p->num = 2;
	p->len = newLen;
	p->pctScore = scor_align_pct(wtMat, p, newLen);
	p->alignScore = scor_pair_align(p->pctScore, s0->sFlag, s1->sFlag, aStruct);

/* if sequence alignment not needed, then free up space */

	if (saveFlag == FALSE) {
		free(p->aSeq[0].pRes);
		free(p->aSeq[1].pRes);
		if (s0->sFlag == TRUE)
			free(p->aSeq[0].sRes);
		if (s1->sFlag == TRUE)
			free(p->aSeq[1].sRes);
	}

}	/* save_seq_align */
