
/*
	file: parmfile.c
	common parameter reading/checking routines used in SEQSEE 
*/

#include "../include/seqsee.h"

/* procedure calls */
extern void show_err();


/*************************************************************/
/* Find the parameter file for this program. 
*/
FILE *get_parms_file(fname)
char *fname;
{
	FILE *fp;
	char sVal[MAX_LINE_LEN];

	fp = fopen(fname, "r");
	if (fp == NULL) {
		sprintf(sVal, "Cannot open parameter file: %s\n", fname);	
		show_err(TRUE, "get_parms_file", sVal);
	}

	return(fp);

}	/* get_parms_file */


/*************************************************************/
/* search the given file for a keyword
*/
int read_parm_key(fp, pfx, pfxLen, keyword)
FILE *fp;
char *pfx, *keyword;
int pfxLen;
{
	char line[MAX_LINE_LEN];

	while (TRUE) {

		if (EOF == fscanf(fp, "%s", keyword))
			return(FALSE);

		if (0 == strncmp(keyword, pfx, pfxLen))
			break;

		fgets(line, MAX_LINE_LEN-1, fp);
	}

	return(TRUE);

}   /* read_parm_key */


/*************************************************************/
/* read an integer from the parameter file 
*/
int read_parm_int(fp, aVal, lb, ub, keyword)
FILE *fp;
int *aVal, lb, ub;
char *keyword;
{
	int num;
	char sVal[MAX_LINE_LEN];

	fscanf(fp, "%d", &num);

	if (num < lb) {
		sprintf(sVal, "Parm key = '%s' Lower bound exceeded: %d < %d\n",
			keyword, num, lb);
		show_err(TRUE, "read_parm_int", sVal);
		return(FALSE);
	}

	if (num > ub) {
		sprintf(sVal, "Parm key = '%s' Upper bound exceeded: %d > %d\n",
			keyword, num, ub);
		show_err(TRUE, "read_parm_int", sVal);
		return(FALSE);
	}

	*aVal = num;
	return(TRUE);

}   /* read_parm_int */


/*************************************************************/
/* read a double from the parameter file 
*/
int read_parm_double(fp, aVal, lb, ub, keyword)
FILE *fp;
double *aVal, lb, ub;
char *keyword;
{
	double num;
	char sVal[MAX_LINE_LEN];

	fscanf(fp, "%lf", &num);

	if (num < lb) {
		sprintf(sVal, "Parm key = '%s' Lower bound exceeded: %lf < %lf\n",
			keyword, num, lb);
		show_err(TRUE, "read_parm_double", sVal);
		return(FALSE);
	}

	if (num > ub) {
		sprintf(sVal, "Parm key = '%s' Upper bound exceeded: %lf < %lf\n",
			keyword, num, ub);
		show_err(TRUE, "read_parm_double", sVal);
		return(FALSE);
	}

	*aVal = num;
	return(TRUE);

}   /* read_parm_double */


/*************************************************************/
/* read a string from the parameter file
*/
int read_parm_string(fp, input_string)
FILE *fp;
char *input_string;
{

	return(fscanf(fp, "%s", input_string));

}   /* read_parm_string */


/*************************************************************/
/* read a file from the parameter file 
*/
int read_parm_fname(fp, fname, keyword)
FILE *fp;
char *fname, *keyword;
{
	FILE *tmpfp;
	char sVal[MAX_LINE_LEN];

	fscanf(fp, "%s", fname);

	if ((tmpfp = fopen(fname, "r")) == NULL) {
		sprintf(sVal, "Parmkey = '%s' Cannot open file: '%s'.\n",
			keyword, fname);
		show_err(TRUE, "read_parm_fname", sVal);
		return(FALSE);
	}

	fclose(tmpfp);
	return(TRUE);

}   /* read_parm_fname */


/*************************************************************/
/* read a file from the parameter file and return a file pointer
*/
FILE *read_parm_fname2(fp, fname, keyword)
FILE *fp;
char *fname, *keyword;
{
	FILE *datafp;
	char sVal[MAX_LINE_LEN];

	fscanf(fp, "%s", fname);

	if ((datafp = fopen(fname, "r")) == NULL) {
		sprintf(sVal, "Parmkey = '%s' Cannot open file: '%s'.\n",
			keyword, fname);
		show_err(TRUE, "read_parm_fname2", sVal);
	}

	return(datafp);

}   /* read_parm_fname2 */


/*************************************************************/
/* Reads similarity matrix (primary sequence or secondary structure).
*/
int read_matrix_file(fp, wtMat)
FILE *fp;
int	*wtMat;
{
	register int j, count, ch, ix, codes;
	int nd, score, ixWt[WT_MAT_DIM];
	char *rc, *s1, line[MAX_LINE_LEN], sVal[MAX_LINE_LEN];
	char first[MAX_LINE_LEN], aStr[MAX_LINE_LEN];
	extern char *strtok();

	count = 0;
	for (j=0; j<WT_MAT_SIZE; j++)
		wtMat[j] = WT_MAT_INIT;

/* read number of codes */
	nd = fscanf(fp, "%s", aStr);
	codes = atoi(aStr);
	if ((codes < 1) || (codes > WT_MAT_DIM)) {
		sprintf(sVal, "Illegal number of codes: %s", line);
		show_err(TRUE, "read_matrix_file", sVal);
	}

/* skip over label line */
	while (TRUE) {
		rc = fgets(line, MAX_LINE_LEN-1, fp);
		if (rc == NULL) {
			sprintf(sVal, "Unexpected EOF, no amino acid label line found");
			show_err(TRUE, "read_matrix_file", sVal);
		}
		nd = sscanf(line, "%s", first);
		if ((nd == 1) && (first[0] != COMMENT))
			break;
	}

/* read each amino acid line */
	rc = fgets(line, MAX_LINE_LEN-1, fp);
	while (count < codes) {

		if (rc == NULL) {
			sprintf(sVal, "Unexpected EOF at: %s, count = %d", line, count);
			show_err(TRUE, "read_matrix_file", sVal);
		}

		strcpy(aStr, line);
		s1 = strtok(line, " ");
		if ((s1[0] != COMMENT) || (s1[0] != '\n')) {

			if (strlen(s1) > 1) {
				sprintf(sVal, "Illegal amino acid label length: %s", line);
				show_err(TRUE, "read_matrix_file", sVal);
			}

			ch = AA_UC(s1[0]);
			if (FALSE == AA_CVALID(ch)) {
				sprintf(sVal, "Illegal amino acid character: %s", aStr);
				show_err(TRUE, "read_matrix_file", sVal);
			}

			ix = AA_IX(ch);
			ixWt[count] = ix;

			for (j=0; j<=count; j++) {

				s1 = strtok((char *)NULL, " ");
				if (s1 == NULL) {
					sprintf(sVal, "Trouble reading matrix at %s", aStr);
					show_err(TRUE, "read_matrix_file", sVal);
				}

				sscanf(s1, "%d", &score);
				if (score < MIN_WT_MAT_VAL) {
					sprintf(sVal, "Min score violated at %s", aStr);
					show_err(TRUE, "read_matrix_file", sVal);
				}

				if (score > MAX_WT_MAT_VAL) {
					sprintf(sVal, "Max score violated at %s", aStr);
					show_err(TRUE, "read_matrix_file", sVal);
				}

				wtMat[(ixWt[count] * WT_MAT_DIM) + ixWt[j]] = score;
				wtMat[(ixWt[j] * WT_MAT_DIM) + ixWt[count]] = score;
			}
			count++;

			s1 = strtok((char *)NULL, " ");
			if (s1 != NULL) {
				nd = sscanf(s1, "%d", &score);
				if (nd > 0) {
					sprintf(sVal, "Input check, line too long at %s", aStr);
					show_err(TRUE, "read_matrix_file", sVal);
				}
			}

		}

		rc = fgets(line, MAX_LINE_LEN-1, fp);
	}

	return(TRUE);

}	/* read_matrix_file */


/************************************************************/
/* Read hydrophobicity table.
*/
void read_hydro_data(fp, data, scale1, scale2)
FILE *fp;
int	*data;
double *scale1, *scale2;
{
	register int i;
	char label, line[MAX_LINE_LEN];

	/* read in scaling constants */
	if (read_prompt(fp) == FALSE)
		show_err(TRUE, "read_hydro_data", "Parameter expected but not found.");
	fscanf(fp, "%lf", scale1);

	if (read_prompt(fp) == FALSE)
		show_err(TRUE, "read_hydro_data", "Parameter expected but not found.");
	fscanf(fp, "%lf", scale2);

	fgets(line, MAX_LINE_LEN-1, fp);
	fgets(line, MAX_LINE_LEN-1, fp);

	for (i=0; i<CODES; i++) {
		fscanf(fp, "%c", &label);
		fscanf(fp, "%d", &data[i]);
		fgets(line, sizeof(line), fp);
	}

}	/* read_hydro_data */


/*************************************************************/
/* read for next prompt in a data file.
*/
int read_prompt(fp)
FILE *fp;
{
	int ch1, ch2;

	ch2 = 0;
	while (ch2 != EOF) {
		ch1 = ch2;
		ch2 = getc(fp);
		if ((ch1 == DATA_PROMPT) && (ch2 == DATA_PROMPT))
			return(TRUE);
	}

	return(FALSE);

}   /* read_prompt */
