
/*
	file: rseq_rbo.c
	purpose: read sequences in ascii ig format.
*/

#include "../include/seqsee.h" 

/* global vars */

/* procedure calls */
static void read_seq_data(), store_seq_data();
static void read_seq_id_str(), read_seq_title();
static int check_ubuf_line(), check_sbuf_line();
extern void show_err(), to_upper();
extern char *emalloc();


/*************************************************************/
/* Reads one sequence from an input file.
*/
int read_seq_rbo(fp, rSeq, aaMin, aaMax)
FILE *fp;
SEQ_REC *rSeq;
int aaMin, aaMax;
{
	register int ch;
	char uBuf[MAX_RES], pBuf[MAX_RES], sBuf[MAX_RES];

/* look for start of sequence */
	ch = getc(fp);
	while ((ch != EOF) && (ch != PIR_SCH)) {
		ch = getc(fp);
	}
	if (ch == EOF)
		return(EOF);

/* read sequence id code */
	read_seq_id_str(fp, rSeq);

/* read sequence title */
	read_seq_title(fp, rSeq);

/* read amino acid data */
	read_seq_data(fp, uBuf, pBuf, sBuf, aaMax);
	store_seq_data(rSeq, uBuf, pBuf, sBuf, aaMin);

	return(TRUE);

}	/* read_seq_rbo */


/*************************************************************/
/* Reads and verifies the sequence id code of a sequence.
*/
static void read_seq_id_str(fp, rSeq)
FILE *fp;
SEQ_REC *rSeq;
{
	register int len;
	char sVal[MAX_LINE_LEN], aStr[MAX_LINE_LEN];

	if (EOF == fscanf(fp, "%s", aStr)) {
		sprintf(sVal, "Expecting sequence but EOF found.");
		show_err(TRUE, "read_seq_id_str", sVal);
	}

	len = strlen(aStr);
	if (len > MAX_ACC_LEN) {
		sprintf(sVal, "Sequence ID code too long at: %s", aStr);
		show_err(TRUE, "read_seq_id_str", sVal);
	}

	rSeq->acc = strdup(aStr);

	to_upper(aStr, aStr);
	if (strncmp(aStr, "TITLE", 5) == 0)
		rSeq->acc[0] = '\0';

}	/* read_seq_id_str */


/*************************************************************/
/* Reads and verifies the title of a sequence.
*/
static void read_seq_title(fp, rSeq)
FILE *fp;
SEQ_REC *rSeq;
{
	register int i, len;
	int lb, ub, nd, count;
	char sVal[MAX_LINE_LEN], *line, *s1;
	extern char *strtok();

	fgets(sVal, MAX_LINE_LEN-1, fp);
	fflush(stdout);

	len = strlen(sVal);
	for (i=0; i<len; i++) {
		if ((sVal[i] != ' ') && (sVal[i] != '\t'))
			break;
	}

	rSeq->title = strdup(&sVal[i]);
	rSeq->title[strlen(rSeq->title)-1] = '\0';

	/* keep track of residue Id numbers in title */
	for (i=0; i<MAX_RES; i++) {
		rSeq->resId[i] = i;
	}
	count = 1;
	s1 = strtok(sVal, "[");
	while (s1 != NULL) {
		nd = sscanf(s1, "%d-%d]", &lb, &ub);
		if (nd == 2) {
			while (lb <= ub) {
				rSeq->resId[count] = lb;
				count++;
				lb++;
			}
		}
		s1 = strtok((char *)NULL, "[");
	}

	if (i < MAX_RES && sVal[i] == '\n')
        {
		sprintf(sVal, "Missing sequence title at: %s", rSeq->acc);
		show_err(TRUE, "read_seq_title", sVal);
	}

}	/* read_seq_title */


/*************************************************************/
/* Reads one sequence from an input file checking that input
** data is legit.
*/
static void read_seq_data(fp, uBuf, pBuf, sBuf, aaMax)
FILE *fp;
char *uBuf, *pBuf, *sBuf;
int aaMax;
{
	register int i, ch, len, pLen, sLen, uLen;
	char line[MAX_LINE_LEN], newLine[MAX_LINE_LEN];
	char sVal[MAX_LINE_LEN];

	pLen = sLen = uLen = 0;
	ch = getc(fp);
	while ((ch != '\n') && (ch != PIR_SCH) && (ch != EOF)) {

		if (ch != COMMENT) {
			line[0] = ch;
			fgets(&line[1], MAX_LINE_LEN-2, fp);
			len = remove_blanks(line, strlen(line), newLine);
			to_upper(newLine, newLine);

			if (TRUE == check_ubuf_line(newLine, len))
				for (i=0; i<len; i++)
					uBuf[uLen++] = newLine[i];
	
			else if (TRUE == check_sbuf_line(newLine, len))
				for (i=0; i<len; i++)
					sBuf[sLen++] = AA_UC(newLine[i]);
	
			else {
				for (i=0; i<len; i++) {
					ch = AA_UC(newLine[i]);	/* convert to upper case */
					if (AA_CVALID(ch))
						pBuf[pLen++] = ch;
					else {
						sprintf(sVal, "Illegal input line: %s", line);
						show_err(TRUE, "read_seq_data", sVal);
					}
				}
			}
	
			if (pLen > aaMax) { /* too many residues */
				sprintf(sVal, "Sequence has > %d amino acids", MAX_RES);
				show_err(TRUE, "read_seq_data", sVal);
			}
		}

		else 
			fgets(sVal, MAX_LINE_LEN-1, fp);

		ch = getc(fp);
	}

	uBuf[uLen] = '\0';
	pBuf[pLen] = '\0';
	sBuf[sLen] = '\0';
	ungetc(ch, fp);

}	/* read_seq_data */


/*************************************************************/
/* store the sequence data we read making sure it is ok.
*/
static void store_seq_data(rSeq, uBuf, pBuf, sBuf, aaMin)
SEQ_REC *rSeq;
char *uBuf, *pBuf, *sBuf;
int aaMin;
{
	register int i, pLen, sLen, uLen;
	char sVal[MAX_LINE_LEN];

	uLen = strlen(uBuf);
	pLen = strlen(pBuf);
	sLen = strlen(sBuf);

	rSeq->uBuf = emalloc((pLen + 1), "store_seq_data: uBuf");
	rSeq->pRes = emalloc((pLen + 1), "store_seq_data: pRes");
	rSeq->sRes = emalloc((pLen + 1), "store_seq_data: sRes");
	rSeq->sFlag = FALSE;
	rSeq->uFlag = FALSE;
	rSeq->len = pLen;

	if (pLen < aaMin) {
		sprintf(sVal, "Sequence has < %d amino acids.", aaMin);
		show_err(TRUE, "store_seq_data", sVal);
	}
	strncpy(rSeq->pRes, pBuf, pLen);

	if (sLen > 0) {
		if (pLen != sLen) {
			sprintf(sVal, "Primary/secondary structures do not line up.");
			show_err(TRUE, "store_seq_data", sVal);
		}
		rSeq->sFlag = TRUE;
		strncpy(rSeq->sRes, sBuf, sLen);
	}
	else {
		for (i=0; i<pLen; i++)
			rSeq->sRes[i] = CH_UNKNOWN;
	}

	if (uLen > 0) {
		if (pLen != uLen) {
			sprintf(sVal, "Primary/anchor symbols do not line up.");
			show_err(TRUE, "store_seq_data", sVal);
		}
		rSeq->uFlag = TRUE;
		strncpy(rSeq->uBuf, uBuf, uLen);
	}
	else {
		for (i=0; i<pLen; i++)
			rSeq->uBuf[i] = GROUP_NOTHING;
	}

}	/* store_seq_data */


/*************************************************************/
/* decide if an input line from the sequence file is a "user" line. 
*/
static int check_ubuf_line(line, len)
char *line;
int len;
{
	register int i, ch;

	for (i=0; i<len; i++) {
		ch = line[i];
		if ((ch != GROUP_NOTHING) && (ch != GROUPING_CHAR) && 
			((ch < ANCHOR_SCHAR) || (ch > ANCHOR_ECHAR)))
			return(FALSE);

	}

	return(TRUE);

}	/* check_ubuf_line */


/*************************************************************/
/* decide if an input line from the sequence file is a 
** secondary structure line.
*/
static int check_sbuf_line(line, len)
char *line;
int len;
{
	register int i;

	for (i=0; i<len; i++) {
		if (AA_SVALID(line[i]) == FALSE)
			return(FALSE);
	}

	return(TRUE);

}	/* check_sbuf_line */
