
/*
	file: ualign.c
	Basic utility routines for doing alignment algorithm
*/

#include "include/seqsee.h"
#include "include/palign.h"
#include "include/xalign.h"

/* procedure calls */
static int pct_start_align(), pct_end_align();
extern void show_err();


/*************************************************************/
/* Calculates penalty scores for secondary structure.
*/
void calc_struct_scores(sStruct, len, sScores, midPen, endPen, endLen)
register char *sStruct;	/* secondary structure array */
register int len;		/* length of sequence */
register int *sScores;	/* penalty scores returned */
int midPen, endPen;		/* penalty values */
int endLen;
{
	register int i, j, scor;

	for (i=0; i<len; i++) {
		scor = 0;
		if ((sStruct[i] == CH_HELX) || (sStruct[i] == CH_BETA)) {
			scor = midPen;
			for (j=1; j<endLen+1; j++) {
				if ((i >= j) && ((i+j) < len)) {
					if ((sStruct[i-j] == CH_COIL) ||
						(sStruct[i+j] == CH_COIL) ||
						(sStruct[i-j] == CH_UNKNOWN) ||
						(sStruct[i+j] == CH_UNKNOWN) ||
						(sStruct[i-j] == CH_TURN) ||
						(sStruct[i+j] == CH_TURN)) {
						scor = endPen;
						break;
					}
				}
				else
					scor = endPen;
			}
		}
		sScores[i] = scor;
	}

}	/* calc_struct_scores */

 
/*************************************************************/
/* process penalties for anchoring and grouping residues.
** return sum of changes.
*/
void calc_anchor_group_pen(uStr, uLen, code, penGroup, penalty)
char *uStr;
int uLen, *code, *penGroup;
{
	register int j, ch;

	for (j=0; j<uLen; j++) {
	   
		ch = uStr[j];
					
		if ((ch >= ANCHOR_SCHAR) && (ch <= ANCHOR_ECHAR)) {
			code[j] = ANCHOR_POS + ch - ANCHOR_SCHAR;
		}

		else {
			if (ch != GROUP_NOTHING) {
				if ((uStr[j-1] != GROUP_NOTHING) && (j > 0)) {
					penGroup[j] = penalty;
				}
			}
		}
	}

}   /* calc_anchor_group_pen */


/*************************************************************/
/* given multiple aligned sequences, calculate the percent identity
** between the sequences based on the homology matrix.
*/
double scor_align_pct(wtMat, mAlign, len)
int *wtMat;
ALIGN_REC *mAlign;
int len;
{
	register int j, k, ix, startIx, endIx, newLen;
	register int best, homScor;
	char *aRes, *bRes;

/* find alignment region */

	startIx = 0;
	endIx = len;

	for (j=0; j<mAlign->num; j++) {
		ix = pct_start_align(mAlign->aSeq[j].pRes, len);
		startIx = MAX(startIx, ix);
		ix = pct_end_align(mAlign->aSeq[j].pRes, len);
		endIx = MIN(endIx, ix);
	}

	newLen = endIx - startIx + 1;
	mAlign->startAlign = startIx + 1;
	mAlign->endAlign = endIx + 1;

/* calculate our score and the perfect score */ 

	homScor = 0;
	best = 0;
	for (j=0; j<mAlign->num; j++) {
		aRes = &mAlign->aSeq[j].pRes[startIx];
		for (k=j+1; k<mAlign->num; k++) {
			bRes = &mAlign->aSeq[k].pRes[startIx];
			homScor += pct_homol_score(wtMat, aRes, bRes, newLen);
			best += pct_perfect_score(wtMat, aRes, bRes, newLen);
		}
	}

/* return percent homology */

	if (best == 0) {
		return(0.0);
	}
	return(100.0 * homScor / (double)(best));

}	/* scor_align_pct */


/*************************************************************/
/* when ranking aligned sequences, bias those which have
** secondary structure known.
*/
double scor_pair_align(pctScor, sFlag0, sFlag1, structVal)
double pctScor;
int sFlag0, sFlag1, structVal;
{

	if (sFlag0 == TRUE)
		pctScor += structVal;
	if (sFlag1 == TRUE)
		pctScor += structVal;

	return(pctScor);

}	/* scor_pair_align */


/*************************************************************/
/* given two aligned sequences, calculate the score based on
** homology only.
*/
int pct_homol_score(wtMat, aRes, bRes, len)
int *wtMat;
char *aRes, *bRes;
int len;
{
	register int i, scor;
	register char aCh, bCh;

	scor = 0;
	for (i=0; i<len; i++) {
		aCh = aRes[i];
		bCh = bRes[i];
		if (AA_CVALID(aCh) && AA_CVALID(bCh))
			scor += wtMat[(AA_IX(aCh) * WT_MAT_DIM) + AA_IX(bCh)];
	}

	return(scor);

}	/* pct_homol_score */


/*************************************************************/
/* given two aligned sequences, calculate the perfect homology
** score.
*/
int pct_perfect_score(wtMat, aRes, bRes, len)
int *wtMat;
char *aRes, *bRes;
int len;
{
	register int i, scor, aix, bix, aScor, bScor;

	scor = 0;
	for (i=0; i<len; i++) {

		aScor = 0;
		aix = AA_IX(aRes[i]);
		if (AA_IVALID(aix))
			aScor = wtMat[(aix * WT_MAT_DIM) + aix];

		bScor = 0;
		bix = AA_IX(bRes[i]);
		if (AA_IVALID(bix))
			bScor = wtMat[(bix * WT_MAT_DIM) + bix];

		if ((aScor == 0) || (bScor == 0))
			scor += MAX(aScor, bScor);
		else
			scor += MIN(aScor, bScor);
	}

	return(scor);

}	/* pct_perfect_score */


/*************************************************************/
/* find the start of the alignment
*/
static int pct_start_align(aRes, len)
char *aRes;
int len;
{
	register int i;
	char sVal[MAX_LINE_LEN];

	for (i=0; i<len; i++) {
		if (AA_CVALID(aRes[i]))
			break;
	}

	if (i == len) {
		sprintf(sVal, "Program check: len = %d", len);
		show_err(TRUE, "pct_start_align", sVal);
	}

	return(i);

}	/* pct_start_align */


/*************************************************************/
/* find the end of the alignment
*/
static int pct_end_align(aRes, len)
char *aRes;
int len;
{
	register int i;
	char sVal[MAX_LINE_LEN];

	for (i=len-1; i>0; i--) {
		if (AA_CVALID(aRes[i]))
			break;
	}

	if (i == 0) {
		sprintf(sVal, "Program check: len = %d", len);
		show_err(TRUE, "pct_end_align", sVal);
	}

	return(i);

}	/* pct_end_align */


/*************************************************************/
/* utility that changes an array of one letter codes to index form 
*/
void codes_to_indices(inStr, len, outNums, invalidNum)
int len, *outNums, invalidNum;
char *inStr;
{
	register int k, ix;

	for (k=0; k<len; k++) {
		ix = AA_IX(inStr[k]);
		if (TRUE == AA_IVALID(ix))
			outNums[k] = ix;
		else
			outNums[k] = invalidNum;
	}

}	/* codes_to_indices */
