/*
	file...: main.c
	purpose: main program for xalign 

	author.: Robert Boyko
	date...: Nov 21, 1994	
	place..: University of Alberta / Department of Biochemistry

	Modified: JCS, July 18, 1995
*/

#include "include/seqsee.h"
#include "include/align.h"
#include "include/palign.h"
#include "include/xalign.h"


/* static vars */

static ALIGN_REC mar;
static int		mSeqSelect;
static FILE		*outfp;
static char		wtMatFname[MAX_FNAME_LEN];
static int		wtMatrix[WT_MAT_SIZE], ssMatrix[WT_MAT_SIZE];
static int		gapPen, gapSizePen, blankMatch;
static int		midStructPen, endStructPen;
static int		incGapPen, structBias;
static int		printPairsFlag, nearMatchScore, printPctHomFlag;
static int		printOrderFlag, consensusPercent, resLineLen;
static int		asciiFlag, printFormat, multAlignNum;
static SEQ_REC	mySeq[MAX_MULT_ALIGN];

/* procedure calls */

static void do_xalign(), add_align(), init_added_seq(), score_malign();
static void update_malign_seqs(), add_dSeq_malign(), sort_align();
static void set_seq_id_str(), print_sequence(), mprint_seq_set();
static void read_parms_file(), read_parms();
static void check_xalign_parms(), read_mat_parms();
static int  member_of_align(), get_malign_index();

static void calc_align_order(), select_align_order(), input_align_order();
static void init_xalign(), do_pairs(), do_mult_align();
static int  calc_new_malign();
static void save_new_malign();

static void print_pair_align();
static void mprint_stats(), mprint_align();
static void print_run(), mprint_seq_names(), mprint_pct_hom_table();
static void print_seq_id(), print_divide();
static void print_res_line(), print_struct_line();
static void print_consensus_line(), print_match_line();
extern void print_seq_seqsee();

extern void str_match_blanks(), str_scan();
extern void codes_to_indices();
extern FILE *get_parms_file(), *read_parm_fname2();
extern int read_parm_key(), read_parm_int(), read_parm_fname();
extern int read_matrix_file();
extern char *get_date(), *emalloc(), *strtok();
extern int *ecalloc();
extern void show_err(), exit();

extern SM_REC *find_path_start();
extern void sm_mat_score(), sm_mat_add_score(), sm_mat_sum();
extern double scor_align_pct(), scor_pair_align();
extern int find_best_align();
extern void do_pair_align();
extern void sm_mat_init();
extern void calc_struct_scores(), calc_anchor_group_pen();


/*************************************************************/
/* main program
*/
int xalign_main(parmsFname, outputFname)
char *parmsFname, *outputFname;
{

	read_parms(parmsFname);
	do_xalign(outputFname);
	return(0);

}	/* main */


/*
	file: calc.c
	main heuristic algorithm for performing xalign multiple alignment
*/



/*************************************************************/
/* Performs multiple alignment algorithm. 
*/
void do_xalign(char *resultsFname)
{
	int alignOrder[MAX_MULT_ALIGN];
	double pctTable[MAX_MULT_ALIGN][MAX_MULT_ALIGN];


	init_xalign(resultsFname, alignOrder);
	print_run(outfp);

/* do we need to determine pairwise alignments? */
	if ((printPairsFlag == TRUE) || (mSeqSelect <= multAlignNum) ||
		(printPctHomFlag == TRUE)) {
		fprintf(stdout, "\nProcessing pairwise alignments\n");
		fflush(stdout);
		do_pairs(alignOrder, pctTable);
	}

/* print multiple align sequence order */
	mprint_seq_names(outfp, alignOrder);

	if (printPctHomFlag == 1)
		mprint_pct_hom_table(outfp, pctTable, alignOrder);

/* perform mult align algorithm */
	fprintf(stdout, "\nDoing Multiple Alignment");
	fflush(stdout);
	do_mult_align(alignOrder);
	fprintf(stdout, "\nDone\n");

}	/* do_xalign */


/*************************************************************/
/* Before we start the xalign algorithm 
*/
static void init_xalign(fname, alignOrder)
char *fname;
int *alignOrder;
{
	register int i, size;

/* open output file */
	outfp = fopen(fname, "w");

/* determine order of the sequences in multiple alignment */
	input_align_order(alignOrder);

/* unused entries in homology matrix used for anchoring sites */
	for (i=ANCHOR_POS; i<WT_MAT_DIM; i++)
		wtMatrix[i*WT_MAT_DIM + i] = ANCHOR_PEN;

/* init blank position score in homology matrix */
	for (i=0; i<WT_MAT_DIM; i++)
		wtMatrix[i * WT_MAT_DIM + BLANK_POS] = blankMatch;
	for (i=0; i<WT_MAT_DIM; i++)
		wtMatrix[i + WT_MAT_DIM * BLANK_POS] = blankMatch;

/* allocate space for sequences in the mult align record */
	size = multAlignNum * sizeof(ALIGN_SEQ);
	mar.aSeq = (ALIGN_SEQ *)emalloc(size, "init_xalign");

}	/* init_xalign */


/*
	file: malign.c
	main heuristic algorithm for performing xalign multiple alignment
*/



/*************************************************************/
/* Process multiple alignment algorithm. 
*/
static void do_mult_align(orderArr)
int *orderArr;
{
	register int i, seqIx;
	int mGapPen, mGapSizePen;

/* gap penalties to offset the cumulative effect of adding more sequences */
	mGapPen = gapPen;
	mGapSizePen = gapSizePen;

/* insert first sequence into multiple alignment record */
	mar.num = 0;
	seqIx = orderArr[0];
	add_dSeq_malign(seqIx, &mySeq[seqIx], mySeq[seqIx].pRes, mySeq[seqIx].len);
	mar.len = mySeq[seqIx].len;

/* add each successive sequence to the multiple alignment */
	for (i=1; i<multAlignNum; i++) {
		seqIx = orderArr[i];
		add_align(seqIx, &mySeq[seqIx], mGapPen, mGapSizePen);
		mGapPen = gapPen + (gapPen * i * incGapPen * 0.01);
		mGapSizePen = gapSizePen + (gapSizePen * i * incGapPen * 0.01);
	}

/* print results */
	mar.pctScore = scor_align_pct(wtMatrix, &mar, mar.len);
	mprint_stats(outfp, &mar);
	mprint_align(outfp, &mar, orderArr);

}	/* do_mult_align */


/*************************************************************/
/* Add a new sequence to our present multiple alignment.
*/
static void add_align(seqIx, dSeq, mGapPen, mGapSizePen)
SEQ_REC *dSeq;
int seqIx, mGapPen, mGapSizePen;
{
	int elements, newLen;
	int	*dSeqNums, *dsSeqNums;
	int	*dsNums, *msNums;
	int	*dPenGroup, *mPenGroup;
	SM_REC *sm;
	char mGapStr[MAX_ALIGN_LEN], dGapStr[MAX_ALIGN_LEN];

/* allocate space for structures */
	elements = mar.len * dSeq->len;
	sm = (SM_REC *)emalloc((elements * sizeof(SM_REC)), "sm");
	dSeqNums = (int *)ecalloc(dSeq->len, sizeof(int), "dSeqNums");
	dsSeqNums = (int *)ecalloc(dSeq->len, sizeof(int), "dsSeqNums");
	dsNums = (int *)ecalloc(dSeq->len, sizeof(int), "dsNums");
	msNums = (int *)ecalloc(mar.len, sizeof(int), "msNums");
	dPenGroup = (int *)ecalloc(dSeq->len, sizeof(int), "dPenGroup");
	mPenGroup = (int *)ecalloc(mar.len, sizeof(int), "mPenGroup");

/* init structures */
	sm_mat_init(sm, mar.len, dSeq->len, TRUE);
	init_added_seq(dSeq, dSeqNums, dsSeqNums, dsNums, dPenGroup);

/* process our scoring matrix, intense cpu usage */ 
	score_malign(sm, dSeq, dSeqNums, dsSeqNums, msNums, mPenGroup);
	sm_mat_sum(sm, mar.len, dSeq->len, msNums, dsNums,
		mPenGroup, dPenGroup, mGapPen, mGapSizePen);

/* use scoring matrix to determine and save new multiple alignment */
	newLen = calc_new_malign(sm, mar.len, dSeq->len, mGapStr, dGapStr);
	save_new_malign(seqIx, newLen, mGapStr, dGapStr, dSeq);

/* de-allocate data structure memory */
	free(sm);
	free(dSeqNums);
	free(dsSeqNums);
	free(msNums);
	free(dsNums);
	free(mPenGroup);
	free(dPenGroup);

}	/* add_align */


/*************************************************************/
/* init structures needed for adding a sequence to our current
** multiple alignment.
*/
static void init_added_seq(dSeq, dSeqNums, dsSeqNums, dsNums, dPenGroup)
SEQ_REC *dSeq;
int *dSeqNums, *dsSeqNums, *dsNums, *dPenGroup;
{

	codes_to_indices(dSeq->pRes, dSeq->len, dSeqNums, BLANK_POS);
	calc_struct_scores(dSeq->sRes, dSeq->len, dsNums,
		midStructPen, endStructPen, END_STRUCT_LEN);

	if (dSeq->uFlag == TRUE)
		calc_anchor_group_pen(dSeq->uBuf, dSeq->len, dSeqNums,
			dPenGroup, GROUPING_PEN);

	if (dSeq->sFlag == TRUE)
		codes_to_indices(dSeq->sRes, dSeq->len, dsSeqNums, BLANK_POS);

}	/* init_added_seq */


/*************************************************************/
/* calculate the Needleman Wunsch scoring matrix for our 
** present multiple align and added sequence.
*/
static void score_malign(sm, dSeq, dSeqNums, dsSeqNums, msNums, mPenGroup)
SM_REC *sm;
SEQ_REC *dSeq;
int *dSeqNums, *dsSeqNums, *msNums, *mPenGroup;
{
	register int i, j;
	int	*aNums, *mSeqNums, *msSeqNums;
	int uFlag, sFlag;

	aNums = (int *)ecalloc(mar.len, sizeof(int), "aNums");
	mSeqNums = (int *)ecalloc(mar.len, sizeof(int), "mSeqNums");
	msSeqNums = (int *)ecalloc(mar.len, sizeof(int),"msSeqNums");

	for (i=0; i<mar.num; i++) {

		codes_to_indices(mar.aSeq[i].pRes, mar.len, mSeqNums, BLANK_POS);
		calc_struct_scores(mar.aSeq[i].sRes, mar.len, aNums,
			midStructPen, endStructPen, END_STRUCT_LEN);
		for (j=0; j<mar.len; j++)
			msNums[j] += aNums[j];

		uFlag = mySeq[mar.aSeq[i].dix].uFlag;
		if (uFlag == TRUE)
			calc_anchor_group_pen(mar.aSeq[i].uBuf, mar.len, mSeqNums,
				mPenGroup, GROUPING_PEN);

		for (j=0; j<mar.len; j++)
			mSeqNums[j] *= WT_MAT_DIM;

		sm_mat_add_score(wtMatrix, sm, mar.len, dSeq->len, mSeqNums, dSeqNums);

		sFlag = mySeq[mar.aSeq[i].dix].sFlag;
		if ((sFlag == TRUE) && (dSeq->sFlag == TRUE)) {
			codes_to_indices(mar.aSeq[i].sRes, mar.len, msSeqNums, BLANK_POS);
			for (j=0; j<mar.len; j++)
				msSeqNums[j] *= WT_MAT_DIM;
			sm_mat_add_score(ssMatrix, sm, mar.len, dSeq->len, 
				msSeqNums, dsSeqNums);
		}
	}

	free(aNums);
	free(mSeqNums);
	free(msSeqNums);

}	/* score_malign */


/*************************************************************/
/* from our processed scoring matrix, calculate the path 
** to take to form the new multiple alignment. 
** Return the length of this new alignment.
*/
static int calc_new_malign(sm, mLen, dLen, mGapStr, dGapStr)
SM_REC *sm;
int mLen, dLen;
char *mGapStr, *dGapStr;
{
	int newLen;
	SM_REC *top;

	top = find_path_start(sm, mLen, dLen);
	newLen = find_best_align(top, sm, mLen, dLen, mGapStr, dGapStr,
		MAX_ALIGN_LEN);

	return(newLen);

}	/* calc_new_malign */


/*************************************************************/
/* update our new multiple alignment record. 
*/
static void save_new_malign(seqIx, newLen, mGapStr, dGapStr, dSeq)
int seqIx, newLen;
char *mGapStr, *dGapStr;
SEQ_REC *dSeq;
{

	update_malign_seqs(newLen, mGapStr);
	add_dSeq_malign(seqIx, dSeq, dGapStr, newLen);
	mar.len = newLen;

}	/* save_new_malign */


/*************************************************************/
/* update each sequence in the multiple alignment record. 
*/
static void update_malign_seqs(newLen, tStr)
int newLen;
char *tStr;
{

	register int j;
	char aStr[MAX_ALIGN_LEN];

	if (newLen == mar.len)  /* no change in mult align sequences */
		return;

	for (j=0; j<mar.num; j++) {

		strncpy(aStr, mar.aSeq[j].pRes, mar.len);	
		free(mar.aSeq[j].pRes);
		mar.aSeq[j].pRes = (char *)emalloc(newLen, "update_malign_seqs: pRes");
		str_match_blanks(aStr, mar.len, tStr, newLen,
			TEMPLATE_GAP, mar.aSeq[j].pRes);
		str_scan(mar.aSeq[j].pRes, newLen, TEMPLATE_GAP, ' ');

		strncpy(aStr, mar.aSeq[j].sRes, mar.len);	
		free(mar.aSeq[j].sRes);
		mar.aSeq[j].sRes = (char *)emalloc(newLen, "update_malign_seqs: sRes");
		str_match_blanks(aStr, mar.len, tStr, newLen, 
			TEMPLATE_GAP, mar.aSeq[j].sRes);
		str_scan(mar.aSeq[j].sRes, newLen, TEMPLATE_GAP, ' ');

		strncpy(aStr, mar.aSeq[j].uBuf, mar.len);	
		free(mar.aSeq[j].uBuf);
		mar.aSeq[j].uBuf = (char *)emalloc(newLen, "update_malign_seqs: uBuf");
		str_match_blanks(aStr, mar.len, tStr, newLen,
			TEMPLATE_GAP, mar.aSeq[j].uBuf);
		str_scan(mar.aSeq[j].uBuf, newLen, TEMPLATE_GAP, ' ');

	}

}	/* update_malign_seqs */


/*************************************************************/
/* save sequence into our multiple alignment 
*/
static void add_dSeq_malign(seqIx, dSeq, dGapStr, dGapLen)
SEQ_REC *dSeq;
int seqIx, dGapLen;
char *dGapStr;
{
	int ix;
	char aStr[MAX_ALIGN_LEN];

	ix = mar.num;

	mar.aSeq[ix].dix = seqIx;

	strncpy(aStr, dSeq->pRes, dSeq->len);	
	mar.aSeq[ix].pRes = (char *)emalloc((dGapLen + 1), "add_dSeq_malign: pRes");
	str_match_blanks(aStr, dSeq->len, dGapStr, dGapLen,
		TEMPLATE_GAP, mar.aSeq[ix].pRes);
	str_scan(mar.aSeq[ix].pRes, dGapLen, TEMPLATE_GAP, ' ');

	strncpy(aStr, dSeq->sRes, dSeq->len);	
	mar.aSeq[ix].sRes = (char *)emalloc((dGapLen + 1), "add_dSeq_malign: sRes");
	str_match_blanks(aStr, dSeq->len, dGapStr, dGapLen,
		TEMPLATE_GAP, mar.aSeq[ix].sRes);
	str_scan(mar.aSeq[ix].sRes, dGapLen, TEMPLATE_GAP, ' ');

	strncpy(aStr, dSeq->uBuf, dSeq->len);	
	mar.aSeq[ix].uBuf = (char *)emalloc((dGapLen + 1), "add_dSeq_malign: uBuf");
	str_match_blanks(aStr, dSeq->len, dGapStr, dGapLen,
		TEMPLATE_GAP, mar.aSeq[ix].uBuf);
	str_scan(mar.aSeq[ix].uBuf, dGapLen, TEMPLATE_GAP, ' ');

	mar.num++;

}	/* add_dSeq_malign */




/*
	file: mpairs.c 
	This module is executed when we need to do pairwise alignments.
	This can be for ordering sequences when the user does not know
	the order of the sequences in the multiple alignment or it can
	be when the user wants printouts of the pairwise alignments.
*/


/*************************************************************/
/* Performs pairwise alignments of the input sequences. 
*/
static void do_pairs(orderArr, pctTable)
int *orderArr;
double pctTable[MAX_MULT_ALIGN][MAX_MULT_ALIGN];
{
	register int i, j, count, size, multAlignPairs;
	ALIGN_REC *topAligns;

	multAlignPairs = (multAlignNum * (multAlignNum-1)) / 2;
	size = multAlignPairs * sizeof(ALIGN_REC);
	topAligns = (ALIGN_REC *)emalloc(size, "do_pairs");

	count = 0;
	for (i=0; i<multAlignNum; i++) { 	/* find each alignment pair */
		for (j=0; j<i; j++) {
			size = 2 * sizeof(ALIGN_SEQ);
			topAligns[count].aSeq = (ALIGN_SEQ *)emalloc(size, "do_pairs");
			topAligns[count].aSeq[0].dix = i;
			topAligns[count].aSeq[1].dix = j;
			do_pair_align(&topAligns[count], &mySeq[i], &mySeq[j], wtMatrix,
				ssMatrix, gapPen, gapSizePen, midStructPen, endStructPen,
				END_STRUCT_LEN, structBias, printPairsFlag);
			pctTable[i][j] = pctTable[j][i] = topAligns[count].pctScore;
			count++;
			putc(PAIR_ALIGN_CHAR, stdout);
			fflush(stdout);
		}
	}

	sort_align(topAligns, 0, count - 1, TRUE);

	if (mSeqSelect < multAlignNum)
		select_align_order(orderArr, multAlignPairs, topAligns);
	else if (mSeqSelect == multAlignNum)
		calc_align_order(orderArr, multAlignPairs, topAligns);

	if (printPairsFlag == TRUE) {
		fprintf(outfp, "\n");
		fprintf(outfp, "Pairwise Alignments\n");
		fprintf(outfp, "*******************\n\n");
		for (i=0; i<multAlignPairs; i++)
			print_pair_align(outfp, &topAligns[i]);
	}

}	/* do_pairs */


/*************************************************************/
/* Insert alignment record into list of best alignments 
*/
static void sort_align(data, left, right, descendFlag)
ALIGN_REC *data;
int left, right, descendFlag;
{
	register int i, j;
	double x;
	ALIGN_REC tmp;

	i = left;
	j = right;
	x = data[(left + right) / 2].alignScore;

	do {
		if (descendFlag == FALSE) {
			while ((data[i].alignScore < x) && (i < right)) i++;
			while ((x < data[j].alignScore) && (j > left)) j--;
		}
		else {
			while ((data[i].alignScore > x) && (i < right)) i++;
			while ((x > data[j].alignScore) && (j > left)) j--;
		}
		if (i <= j) {
			tmp = data[i];
			data[i] = data[j];
			data[j] = tmp;
			i++; j--;
		}
	} while (i <= j);

	if (left < j)   sort_align(data, left, j, descendFlag);
	if (i < right)  sort_align(data, i, right, descendFlag);

}	/* sort_align */



/*
	file: newprint.c
	This file contains high level print routines for seqsee user interface. 
*/



/*************************************************************/
/* prints one sequence
*/
static void print_sequence(fp, aSeq, mar, ix) 
FILE *fp;
SEQ_REC *aSeq;
ALIGN_REC *mar;
int ix;
{
	register int i;
	char pSeq[MAX_ALIGN_LEN], sSeq[MAX_ALIGN_LEN], uSeq[MAX_ALIGN_LEN];
	SEQ_REC rSeq;

	strncpy(pSeq, mar->aSeq[ix].pRes, mar->len);
	strncpy(sSeq, mar->aSeq[ix].sRes, mar->len);
	strncpy(uSeq, mar->aSeq[ix].uBuf, mar->len);
	pSeq[mar->len] = '\0';
	sSeq[mar->len] = '\0';
	uSeq[mar->len] = '\0';

	rSeq.title = (char *)emalloc(strlen(aSeq->title), "print_sequence");
	strcpy(rSeq.title, aSeq->title);
	rSeq.len = mar->len;
	rSeq.pRes = pSeq;
	rSeq.sRes = sSeq;
	rSeq.uBuf = uSeq;

	print_seq_seqsee(fp, &rSeq);

}	/* print_sequence */




/*
	file: order.c 
	purpose: to order the sequences to be added to the multiple
	alignment.
*/


/*************************************************************/
/* computer determines order for adding sequences for multiple alignment 
*/
static void calc_align_order(orderArr, multAlignPairs, topAligns)
int *orderArr, multAlignPairs;
ALIGN_REC *topAligns;
{
	register int i, count, first, second;

	count = 0;
	for (i=0; i<multAlignPairs; i++) {

		first = topAligns[i].aSeq[0].dix;
		second = topAligns[i].aSeq[1].dix;

		if ((mySeq[first].sFlag == FALSE) && (mySeq[second].sFlag == TRUE)) {
			first = topAligns[i].aSeq[1].dix;
			second = topAligns[i].aSeq[0].dix;
		}

		if (member_of_align(orderArr, count, first) == FALSE)
			orderArr[count++] = first;

		if (member_of_align(orderArr, count, second) == FALSE)
			orderArr[count++] = second;

		if (count == multAlignNum)
			break;
	}

	if (printOrderFlag == FALSE)
		set_seq_id_str(orderArr);

}	/* calc_align_order */


/*************************************************************/
/* user selects sequence to which we apply the multiple alignment
** algorithm to.
*/
static void select_align_order(orderArr, multAlignPairs, topAligns)
int *orderArr, multAlignPairs;
ALIGN_REC *topAligns;
{
	register int i, count, first, second;

	count = 0;
	orderArr[count++] = mSeqSelect;

	for (i=0; i<multAlignPairs; i++) {

		first = topAligns[i].aSeq[0].dix;
		second = topAligns[i].aSeq[1].dix;

		if (first == mSeqSelect)
			orderArr[count++] = second;

		if (second == mSeqSelect)
			orderArr[count++] = first;

		if (count == multAlignNum)
			break;

	}

	if (printOrderFlag == FALSE)
		set_seq_id_str(orderArr);

}	/* select_align_order */


/*************************************************************/
/* the order of the sequences in the alignment is the order in
** which they were read in.
*/
static void input_align_order(orderArr)
int *orderArr;
{
	register int i;

	for (i=0; i<multAlignNum; i++) {
		orderArr[i] = i;
	}

	if ((mSeqSelect > multAlignNum) || (printOrderFlag == TRUE))
		set_seq_id_str(orderArr);

}	/* input_align_order */


/*************************************************************/
/* Set up those sequences which the user did not supply an ID.
*/
static void set_seq_id_str(orderArr)
int *orderArr;
{
	register int i, ix;


	for (i=0; i<multAlignNum; i++) {
		ix = orderArr[i];
		if (mySeq[i].acc == NULL) {
			mySeq[i].acc = (char *)emalloc(8, "set_seq_id_str:acc");
			if ((ix+1) >= 100)
				sprintf(mySeq[i].acc, "Seq%d", ix+1);
			else
				sprintf(mySeq[i].acc, "Seq_%d", ix+1);
		}
	}

}	/* set_seq_id_str */


/*************************************************************/
/* determine if a sequence id is found in our current list.
*/
static int member_of_align(orderArr, len, x)
int *orderArr;
register int len, x;
{
	register int i;
	 
	for (i=0; i<len; i++) {
		if (orderArr[i] == x)
			return(TRUE);
	}
			 
	return(FALSE);
	  
}   /* member_of_align */


/*
	file: parms.c
	read in parameter file for xalign 
*/




/*************************************************************/
/* obtain the parameter file and read it. 
*/
static void read_parms(fname)
char *fname;
{
	FILE *parmsfp;

	parmsfp = get_parms_file(fname);
	if (parmsfp == NULL)
		exit(1);

	read_parms_file(parmsfp);
	fclose(parmsfp);

}	/* read_parms */


/*************************************************************/
/* Read parameters from the parameter file.
*/
static void read_parms_file(parmsfp)
FILE *parmsfp;
{
	FILE *fp;
	int i, len, okFlag[PARM_NUM], fileType;
	char keyword[MAX_LINE_LEN], *aStr, sVal[MAX_LINE_LEN];
	char dataFile[MAX_FNAME_LEN];

	/* assume no parms read */
	for (i=0; i<PARM_NUM; i++)
		okFlag[i] = FALSE;
	resLineLen = 0;
	printPairsFlag = FALSE;
	printFormat = TRUE;
	asciiFlag = FALSE;
	
	len = strlen(PARM_PREFIX);
	while (TRUE == read_parm_key(parmsfp, PARM_PREFIX, len, keyword)) {

		/* read data file format */
		if (0 == strcmp(keyword, KEY_DATA_FORMAT)) {
			okFlag[DATA_FORMAT] = read_parm_int(parmsfp, &fileType,
				1, 5, keyword);
		}

		/* read alignment data file, but process later */
		if (0 == strcmp(keyword, KEY_DATA_FILE)) {
			okFlag[DATA_FILE] = read_parm_string(parmsfp, dataFile);
		}

		/* read sequence to align to in multiple align */
		if (0 == strcmp(keyword, KEY_SEQ_SELECT)) {
			okFlag[SEQ_SELECT] = read_parm_int(parmsfp, &mSeqSelect,
				1, MAX_MULT_ALIGN + 2, keyword);
			mSeqSelect--;
		}

		/* read amino acid homology matrix and scoring parms */
		if (0 == strcmp(keyword, KEY_SCOR_PARMS)) {
			fp = read_parm_fname2(parmsfp, wtMatFname, keyword);
			okFlag[SCOR_PARMS] = (fp == NULL? FALSE: TRUE);
			if (fp != NULL) {
				read_mat_parms(fp, okFlag);
				fclose(fp);
			}
		}

		/* read parm that biases sequences which have secondary structure */
		if (0 == strcmp(keyword, KEY_STRUCT_BIAS))
			okFlag[STRUCT_BIAS] = read_parm_int(parmsfp, &structBias,
				MIN_STRUCT_BIAS, MAX_STRUCT_BIAS, keyword);

		/* Read cumulative alignment penalty */
		if (0 == strcmp(keyword, KEY_INC_GAP_PEN))
			okFlag[INC_GAP_PEN] = read_parm_int(parmsfp, &incGapPen,
				MIN_INC_GAP_PEN, MAX_INC_GAP_PEN, keyword);

		/* Read print pairwise alignments flag */
		if (0 == strcmp(keyword, KEY_PRINT_PAIRWISE))
			okFlag[PRINT_PAIRWISE] = read_parm_int(parmsfp, &printPairsFlag,
				0, 1, keyword);

		/* Read print percent homology table flag */
		if (0 == strcmp(keyword, KEY_PRINT_PCT_HOM))
			okFlag[PRINT_PCT_HOM] = read_parm_int(parmsfp, &printPctHomFlag, 
				0, 1, keyword);

		/* Read the alignment format variable */
		if (0 == strcmp(keyword, KEY_PRINT_FORMAT))
			okFlag[PRINT_FORMAT] = read_parm_int(parmsfp, &printFormat, 
				0, 1, keyword);

		/* Read the flag which determines ordering
		   of sequences for printing the multiple alignment */
		if (0 == strcmp(keyword, KEY_PRINT_ORDER))
			okFlag[PRINT_ORDER] = read_parm_int(parmsfp, &printOrderFlag, 
				0, 1, keyword);

		/* Read mult alignment consensus percent */
		if (0 == strcmp(keyword, KEY_CONSENSUS_PCT))
			okFlag[CONSENSUS_PCT] = read_parm_int(parmsfp, &consensusPercent,
				0, 100, keyword);

		/* Read output line length for alignment */
		if (0 == strcmp(keyword, KEY_RES_LINE_LEN))
			okFlag[RES_LINE_LEN] = read_parm_int(parmsfp, &resLineLen,
				MIN_RES_LINE_LEN, MAX_RES_LINE_LEN, keyword);

		/* Read the gui/ascii version caller */
		if (0 == strcmp(keyword, KEY_ASCII_FLAG))
			okFlag[ASCII_FLAG] = read_parm_int(parmsfp, &asciiFlag, 
				0, 1, keyword);

	}

	/* now is a good time to process user entered sequence files */
	okFlag[DATA_FILE] = mget_seqs(dataFile, stdout, mySeq,
		fileType, MAX_MULT_ALIGN, MIN_RES, MAX_RES, &multAlignNum);

	check_xalign_parms(okFlag);

}	/* read_parms_file */


/*************************************************************/
/* checks that all the parameters have been read in.  
*/
static void check_xalign_parms(okFlag)
int *okFlag;
{
	int i;
	char *aStr, sVal[MAX_LINE_LEN];

	for (i=0; i<PARM_NUM; i++) {
		aStr = NULL;
		if (okFlag[i] == FALSE) {
			switch (i) {
				case SCOR_PARMS:		aStr = KEY_SCOR_PARMS; break;
				case SCOR_MATRIX:		aStr = KEY_SCOR_MATRIX; break;
				case SS_SCOR_MATRIX:	aStr = KEY_SS_SCOR_MATRIX; break;
				case GAP_PEN:			aStr = KEY_GAP_PEN; break;
				case GAP_PEN_SIZE:		aStr = KEY_GAP_PEN_SIZE; break;
				case MID_GAP_PEN:		aStr = KEY_MID_GAP_PEN; break;
				case END_GAP_PEN:		aStr = KEY_END_GAP_PEN; break;
				case BLANK_MATCH:		aStr = KEY_BLANK_MATCH; break;
				case STRUCT_BIAS:		aStr = KEY_STRUCT_BIAS; break;
				case INC_GAP_PEN:		aStr = KEY_INC_GAP_PEN; break;
				case NEAR_MATCH:		aStr = KEY_NEAR_MATCH; break;
				case PRINT_PCT_HOM:		aStr = KEY_PRINT_PCT_HOM; break;
				case CONSENSUS_PCT:		aStr = KEY_CONSENSUS_PCT; break;
				case PRINT_ORDER:		aStr = KEY_PRINT_ORDER; break;
				case DATA_FILE:			aStr = KEY_DATA_FILE; break;
				case DATA_FORMAT:		aStr = KEY_DATA_FORMAT; break;
				case SEQ_SELECT:		aStr = KEY_SEQ_SELECT; break;
			}
			if (asciiFlag == TRUE) {
				switch (i) {
					case RES_LINE_LEN:		aStr = KEY_RES_LINE_LEN; break;
					case PRINT_PAIRWISE:	aStr = KEY_PRINT_PAIRWISE; break;
					case PRINT_FORMAT:		aStr = KEY_PRINT_FORMAT; break;
				}
			}
		}
		if (aStr != NULL) {
			sprintf(sVal, "Missing Parameter: %s\n", aStr);
			show_err(TRUE, "check_xalign_parms", sVal);
		}
	}

}	/* check_xalign_parms */


/*************************************************************/
/* Reads scoring matrix parameters 
*/
static void read_mat_parms(fp, okFlag)
FILE *fp;
int	*okFlag;
{
	int len;
	char keyword[MAX_LINE_LEN];

	len = strlen("KEY");
	while (TRUE == read_parm_key(fp, "KEY", len, keyword)) {

		/* Read primary sequence scoring matrix */
		if (0 == strcmp(keyword, KEY_SCOR_MATRIX))
			okFlag[SCOR_MATRIX] = read_matrix_file(fp, wtMatrix);

		/* Read secondary structure scoring matrix */
		if (0 == strcmp(keyword, KEY_SS_SCOR_MATRIX))
			okFlag[SS_SCOR_MATRIX] = read_matrix_file(fp, ssMatrix);

		/* Read gap penalty parameter */
		if (0 == strcmp(keyword, KEY_GAP_PEN))
			okFlag[GAP_PEN] = read_parm_int(fp, &gapPen, 
				MIN_GAP_PEN, MAX_GAP_PEN, keyword);

		/* Read gap size penalty for forming alignments */
		if (0 == strcmp(keyword, KEY_GAP_PEN_SIZE))
			okFlag[GAP_PEN_SIZE] = read_parm_int(fp, &gapSizePen,
				MIN_GAP_SIZE_PEN, MAX_GAP_SIZE_PEN, keyword);

		/* Read near match score for amino acids */
		if (0 == strcmp(keyword, KEY_NEAR_MATCH))
			okFlag[NEAR_MATCH] = read_parm_int(fp, &nearMatchScore,
				MIN_WT_MAT_VAL, MAX_WT_MAT_VAL, keyword);

		/* Read blank match score for amino acids */
		if (0 == strcmp(keyword, KEY_BLANK_MATCH))
			okFlag[BLANK_MATCH] = read_parm_int(fp, &blankMatch,
				MIN_GAP_PEN, MAX_GAP_PEN, keyword);

		/* Read penalty for gapping in middle of structure */
		if (0 == strcmp(keyword, KEY_MID_GAP_PEN))
			okFlag[MID_GAP_PEN] = read_parm_int(fp, &midStructPen, 
				MIN_GAP_PEN, MAX_GAP_PEN, keyword);

		/* Read penalty for gapping at end of a structure */
		if (0 == strcmp(keyword, KEY_END_GAP_PEN))
			okFlag[END_GAP_PEN] = read_parm_int(fp, &endStructPen, 
				MIN_GAP_PEN, MAX_GAP_PEN, keyword);

	}

}	/* read_mat_parms */


/*
	file: print.c
	This file contains high level print routines for alignments. 
*/


/*************************************************************/
/* Print info about this run. 
*/
static void print_run(fp)
FILE *fp;
{
	char *flag;

	print_divide(fp, '-');

	fprintf(fp, "   Program.......: xalign (version %s)\n", VERSION);
	fprintf(fp, "   Description...: multiple sequence alignment program\n");
	fprintf(fp, "   Date..........: %s", get_date());
	fprintf(fp, "\n");

	fprintf(fp, "   P  Scoring Mat: %s\n", wtMatFname);

	fprintf(fp, "   Gap Penalty...: %-3d", gapPen);
	fprintf(fp, "  Structure Bias: %-3d", structBias);
	fprintf(fp, "  Line Length...: %-3d\n", resLineLen);

	fprintf(fp, "   Gap Size Pen..: %-3d", gapSizePen);
	fprintf(fp, "  IncGap Percent: %-3d", incGapPen);
	flag = printPctHomFlag == 0? "FALSE": "TRUE ";	
	fprintf(fp, "  Print Pct Hom.: %s\n", flag);

	fprintf(fp, "   Mid Struct Pen: %-3d", midStructPen);
	fprintf(fp, "  NearMatchScore: %-3d", nearMatchScore);
	flag = printOrderFlag == 0? "TRUE ": "FALSE";	
	fprintf(fp, "  Print Sorted..: %s\n", flag);

	fprintf(fp, "   End Struct Pen: %-3d", endStructPen);
	fprintf(fp, "  BlankMatchScor: %-3d", blankMatch);
	flag = printPairsFlag == 0? "FALSE": "TRUE ";	
	fprintf(fp, "  Print Pairwise: %s\n", flag);

	fprintf(fp, "   Consensus %%...: %-3d\n", consensusPercent);
	fprintf(fp, "\n");

	print_divide(fp, '-');
	fflush(fp);

}	/* print_run */


/*************************************************************/
/* prints the body of the multiple alignment to output.
*/
static void print_pair_align(fp, p)
FILE *fp;
ALIGN_REC *p;
{
	register int i, seqIx, len, rix, rLen;
	int pResCount[2];

	rix = 0;
	len = p->len;
	pResCount[0] = pResCount[1] = 0;

	for (i=0; i<2; i++) {
		seqIx = p->aSeq[i].dix;
		fprintf(fp, " Sequence %d ", i+1);
		fprintf(fp, "ID=%s  ", mySeq[seqIx].acc);
		fprintf(fp, "%s\n", mySeq[seqIx].title);
	}
	fprintf(fp, " Percent Sequence Homology: %6.2f\n", p->pctScore);
	fprintf(fp, " Alignment Score..........: %6.2f\n", p->alignScore);

/* print each section of the alignment until done */

	while (len > 0) {
		rLen = MIN(len, resLineLen);
		len -= rLen;
		putc('\n', fp);

		seqIx = p->aSeq[0].dix;
		print_res_line(fp, seqIx, p->aSeq[0].pRes, rix, rLen, &pResCount[0]);
		if (mySeq[seqIx].sFlag == TRUE)
			print_struct_line(fp, p->aSeq[0].sRes, rix, rLen);

		print_match_line(fp, p->aSeq[0].pRes, p->aSeq[1].pRes, rix, rLen);

		seqIx = p->aSeq[1].dix;
		print_res_line(fp, seqIx, p->aSeq[1].pRes, rix, rLen, &pResCount[1]);
		if (mySeq[seqIx].sFlag == TRUE)
			print_struct_line(fp, p->aSeq[1].sRes, rix, rLen);

		rix += resLineLen;
	}

	print_divide(fp, '-');
	fflush(fp);

}	/* print_pair_align */


/*************************************************************/
/* identify the sequences in the alignment
*/
static void mprint_seq_names(fp, orderArr)
FILE *fp;
int *orderArr;
{
	register int i, j, ix, seqIx;

	fprintf(fp, "\n");
	fprintf(fp, "Sequences\n");
	fprintf(fp, "*********\n\n");
	fprintf(fp, "   Num   Acc#       Sequence Title\n");

	for (i=0; i<multAlignNum; i++) {

		ix = get_malign_index(orderArr, i, printOrderFlag);
		seqIx = orderArr[ix];

		fprintf(fp, "   %3d   ", i+1);
		fprintf(fp, "%s   ", mySeq[seqIx].acc);

		for (j=strlen(mySeq[seqIx].acc); j<=MAX_ACC_LEN-1; j++)
			putc(' ', fp);

		fprintf(fp, "%s\n", mySeq[seqIx].title);

	}

	fprintf(fp, "\n");
	fflush(fp);

}	/* mprint_seq_names */


/*************************************************************/
/* print the percent sequence homology table 
*/
static void mprint_pct_hom_table(fp, scores, orderArr)
FILE *fp;
double scores[MAX_MULT_ALIGN][MAX_MULT_ALIGN];
int *orderArr;
{
	register int i, j;

	fprintf(fp, "\n");
	fprintf(fp, "Pairwise Percent Sequence Homology\n");
	fprintf(fp, "**********************************\n\n");
	fprintf(fp, "          ");

	for (i=0; i<multAlignNum; i++) {
		fprintf(fp, " ");
		print_seq_id(fp, mySeq[orderArr[i]].acc);
	}
	fprintf(fp, "\n");

	for (i=0; i<multAlignNum; i++) {
		fprintf(fp, " ");
		print_seq_id(fp, mySeq[orderArr[i]].acc);
		fprintf(fp, " ");
		for (j=0; j<multAlignNum; j++) {
			if (i == j)
				fprintf(fp, " %8.2f", 100.00);
			else
				fprintf(fp, " %8.2f", scores[orderArr[i]][orderArr[j]]);
		}
		fprintf(fp, "\n");
	}

	fprintf(fp, "\n");
	fflush(fp);

}	/* mprint_pct_hom_table */


/*************************************************************/
/* prints the order in how the sequences are aligned and the 
** mult align score.
*/
static void mprint_stats(fp, mar)
FILE *fp;
ALIGN_REC *mar;
{

	fprintf(fp, "\n");
	fprintf(fp, "Statistics\n");
	fprintf(fp, "**********\n\n");
	fprintf(fp, "   Total length for the multiple alignment: %d\n", mar->len);
	if (mar->endAlign > mar->startAlign) {
		fprintf(fp, "   Region where sequences overlap: %d - %d (length = %d)",
		mar->startAlign, mar->endAlign, mar->endAlign - mar->startAlign + 1);
	} else {
		fprintf(fp, "   No Region where all sequences overlap.");
	}
	fprintf(fp, "\n");
	fprintf(fp, "   Percent Sequence Homology in overlap region: %-8.2f\n",
		mar->pctScore);

	fprintf(fp, "\n");
	fflush(fp);

}	/* mprint_stats */


/*************************************************************/
/* prints the body of the multiple alignment to output.
*/
static void mprint_align(fp, mar, orderArr)
FILE *fp;
ALIGN_REC *mar;
int *orderArr;
{
	register int i, len, rix, rLen, ix, seqIx;
	double consensusThres;
	int pResCount[MAX_MULT_ALIGN];

	fprintf(fp, "\n");
	fprintf(fp, "Multiple Alignment\n");
	fprintf(fp, "******************\n");

	if (printFormat == 1) {
		for (i=0; i<multAlignNum; i++) {
			ix = get_malign_index(orderArr, i, printOrderFlag);
			seqIx = orderArr[ix];
			print_sequence(fp, &mySeq[seqIx], mar, ix);
		}
	}

/* print alignments in text format */
	else {
		rix = 0;
		len = mar->len;
		consensusThres = 0.01 * consensusPercent * mar->num;
		for (i=0; i<mar->num; i++)
			pResCount[i] = 0;

/* print each section of the alignment until done */
		while (len > 0) {
			rLen = MIN(len, resLineLen);
			len -= rLen;
			putc('\n', fp);
			mprint_seq_set(fp, mar, orderArr, rix, rLen,
				pResCount, consensusThres);
			rix += resLineLen;
		}
	}

}	/* mprint_align */


/*************************************************************/
/* prints one section of the aligned sequences including
** consensus sequence
*/
static void mprint_seq_set(fp, mar, orderArr, rix, rLen, pResCount, cThres)
FILE *fp;
ALIGN_REC *mar;
int *orderArr, rix, rLen, *pResCount;
double cThres;
{
	register int i, ix, seqIx;

	for (i=0; i<mar->num; i++) {

		ix = get_malign_index(orderArr, i, printOrderFlag);
		seqIx = mar->aSeq[ix].dix;

		print_res_line(fp, seqIx, mar->aSeq[ix].pRes, rix, rLen,
			&pResCount[seqIx]);
		if (mySeq[seqIx].sFlag == TRUE)
			print_struct_line(fp, mar->aSeq[ix].sRes, rix, rLen);

	}

	print_consensus_line(fp, mar, cThres, rix, rLen);

}	/* mprint_seq_set */


/*************************************************************/
/* return index of sequence to print when input sequence order
** is to be maintained.
*/
static int get_malign_index(orderArr, iNum, orderFlag)
int *orderArr;
int iNum, orderFlag;
{
	register int j;

/* sequences to be displayed in order of homology */
	if (orderFlag == 0)
		return(iNum);

/* sequences to be displayed in input sequence order */
	for (j=0; j<multAlignNum; j++)
		if (iNum == orderArr[j])
			break;

	if (j == multAlignNum)
		show_err(TRUE, "get_malign_index", "Program check, see programmer.");

	return(j);

}	/* get_malign_index */


/*
	file: uprint.c
	This file contains generic print routines for alignments. 
*/


/*************************************************************/
/* prints one output line of residues
*/
static void print_res_line(fp, ix, pRes, rix, rLen, pResCount)
FILE *fp;
char *pRes;
int ix, rix, rLen, *pResCount;
{
	register int j, ch;

	putc(' ', fp);
	print_seq_id(fp, mySeq[ix].acc);
	putc(':', fp);
	putc(' ', fp);

	if (rLen == 0)
		return;

	for (j=0; j<rLen; j++) {
		ch = pRes[rix++];
		if (ch == ' ')
			putc(ALIGN_GAP_CHAR, fp);
		else {
			putc(ch, fp);
			(*pResCount)++;
		}
	}

	while (j < resLineLen) {
		putc(' ', fp);
		j++;
	}

	fprintf(fp, "%7d\n", mySeq[ix].resId[*pResCount]);
	fflush(fp);

}	/* print_res_line */


/*************************************************************/
/* prints one output line of secondary structure from a sequence
*/
static void print_struct_line(fp, sStruct, six, rLen)
FILE *fp;
char *sStruct;
int six, rLen;
{
	register int j, ch;

	putc(' ', fp);
	print_seq_id(fp, "Struct");
	putc(':', fp);
	putc(' ', fp);

	if (rLen == 0)
		return;

	for (j=0; j<rLen; j++) {
		ch = sStruct[six++];
		if (ch == ' ')
			putc(ALIGN_GAP_CHAR, fp);
		else
			putc(AA_LC(ch), fp);
	}

	fprintf(fp, "\n");
	fflush(fp);

}	/* print_struct_line */


/*************************************************************/
/* prints one output line of match summary data. This indicates
** if the residues are a match, near match, or not a match.
*/
static void print_match_line(fp, qRes, dRes, rix, rLen)
FILE *fp;
char *qRes, *dRes;
int rix, rLen;
{
	register int j, qNum, dNum;

	fprintf(fp, "    Match: ");

	for (j=0; j<rLen; j++) {
		qNum = AA_IX(qRes[rix]);
		dNum = AA_IX(dRes[rix]);
		rix++;
		if ((qNum < CODES) && (qNum >= 0) &&
			(dNum < CODES) && (dNum >= 0)) {
			if (qNum == dNum)
				putc(AA_MATCH, fp);
			else if (wtMatrix[qNum*WT_MAT_DIM + dNum] >= nearMatchScore)
				putc(AA_NEAR_MATCH, fp);
			else
				putc(AA_NOT_MATCH, fp);
		}
		else
			putc(AA_NOT_MATCH, fp);
	}

	fprintf(fp, "\n");
	fflush(fp);

}	/* print_match_line */


/*************************************************************/
/* prints one output line for the consensus sequence.
*/
static void print_consensus_line(fp, mar, threshold, rix, rLen)
FILE *fp;
double threshold;
int rix, rLen;
ALIGN_REC *mar;
{
	register int i, j, ch;
	int bestScor, bestIx;
	int consensus[CODES];

	putc(' ', fp);
	print_seq_id(fp, "Summary");
	putc(':', fp);
	putc(' ', fp);

	for (i=0; i<rLen; i++) {

		for (j=0; j<CODES; j++)
			consensus[j] = 0;

		for (j=0; j<mar->num; j++) {
			ch = mar->aSeq[j].pRes[rix];
			if (AA_CVALID(ch) == TRUE)
				consensus[AA_IX(ch)]++;
		}

		bestScor = 0;
		bestIx = 0;
		for (j=0; j<CODES; j++) {
			if (consensus[j] >= bestScor) {
				bestScor = consensus[j];
				bestIx = j;
			}
		}

		if (bestScor >= threshold)
			ch = AA_CH(bestIx);
		else
			ch = NO_CONSENSUS;

		putc(ch, fp);
		rix++;
	}

	fprintf(fp, "\n");
	fflush(fp);

}	/* print_consensus_line */


/*************************************************************/
/* prints the sequence id code
*/
static void print_seq_id(fp, idName)
FILE *fp;
char *idName;
{

	fprintf(fp, "%8s", idName);

}	/* print_seq_id */


/*************************************************************/
/* print separator line on output 
*/
static void print_divide(fp, ch)
FILE *fp;
char ch;
{
	int i, len;

	len = resLineLen + 30;

	putc('\n', fp);
	for (i=0; i<len; i++)
		putc(ch, fp);
	putc('\n', fp);
	putc('\n', fp);

}	/* print_divide */
