/**************************************************************************

  pepmake.c
     Generate a PDB file from a file of internal angles and coordinates.

  Date: 
     March, 2003

  Author:
     Aaron Davidson @ BTI; Adapted from code by Nelson Young.
	 Updated by David Arndt in October-December 2006.

 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <time.h>

#include "util.h" 
#include "vec3D.h"
#include "pdbfile.h"
#include "pepmake.h"
#include "rama.h"
#include "preditor_parser.h"
#include "template.h"
#include "phipsi.h"


/* Function prototypes */

static void applyTemplate(ResData *res);
static void rotatePhiPsi(int num, ResData *residues, int onlyBackbone);
static void nextPeptide(ResData *res, int num, ResData *residues, int onlyBackbone);
static void initResData(ResData *r);
static int closeGapCCD(ResData *residues, int totalRes, GAP_CLOSURE_PARAMS gap_params);
static int residueHasIntendedHbond(ResData *residues, int resNum);
static int adjustSegmentCCD(ResData *residues, int numRes, int N_anchor_index, int C_anchor_index, int allowAnglesToBecomeWorse, double **original_angles,
		int mode, Vec3D target_CA, Vec3D target_C1, Vec3D target_N, double maxSumOfSquaredDistancesThreshold, int rebuild_after_gap,
		int begin_rigid_seg_index, int begin_hbonds_index, int end_hbonds_index);
static void setCorrectedAngles(ResData *residues, int numRes, int startIndex, int endIndex, int rama_score_threshold);
static int readResidue(ResData *res, FILE *fp);
static void compute_angles(ResData *res, int n);
static void parseHbondsFile(char* inputHbondsFilename, ResData *residues, int totalRes);
static void pepmake(int haveAnglesFile, char *inputAnglesFilename, int haveResCoordsFile, char *resCoordsFilename,
		char *inputPDBFilename, char *outputPDBFilename, int fixTorsionAngles, GAP_CLOSURE_PARAMS gap_params,
		int formHbonds, char* inputHbondsFilename);
static void usage();

/* Global variables */

int havePDBFile; // Whether the user has specified an input PDB file
int rotateChi = TRUE; // Whether to adjust chi angles when building a peptide.
int ignoreDBBackbone = FALSE;	// whether to ignore the databse backbone atoms

/* Lookup tables for N-CA-C1 angles, for general residues (the 20 standard amino
 * acids, minus Gly and Pro), Gly, and Pro.  These tables are used to lookup the
 * N-CA-C1 angle based on the phi and psi angle.
 * Element [0][0] corresponds to the overall average angle;
 * the first row and column correspond to the lookups for phi and psi
 * individually. To perform an angle lookup, it is advised to use one of the
 * functions lookup_N_CA_C1_angle, lookup_N_CA_C1_angle_from_phi,
 * lookup_N_CA_C1_angle_from_psi, or lookup_N_CA_C1_angle_baseline.
 */
double general_N_CA_C1_angle_lookup[37][37];
double gly_N_CA_C1_angle_lookup[37][37];
double pro_N_CA_C1_angle_lookup[37][37];
double xpr_N_CA_C1_angle_lookup[37][37];

/* Variables */

/** options **/
int sideChain = TRUE; 			// attach a side chain
int printH = FALSE;				// print HN atoms
int adjustOmega = TRUE; // Whether to adjust the omega angles.

/* default atom positions for a backbone. */
static Vec3D template[] = {
	{0.0, 0.0, 0.0},		  /* CA */
	{1.53, 0.0, 0.0},		  /* C  */
	{2.1686, 1.0629, 0.0},    /* 0  */
	{2.0669, -1.2059, 0.0},   /* N  */
	{3.5188, -1.4359, 0.0},   /* CA2 */
	{1.5223, -2.0446, 0.0},   /* HN */
};


/* Print the atoms in PDB format to the given file.
	Written by Shan and Raj.
	Updated & modified by David Arndt in November 2006, November 2007. */
void printPDBAtoms(FILE *out, ResData *residues, int numRes) {
	int n,c=1;
	int i;
	int printedHA, printedHNs;
	ResData *res = &residues[0];
	ResData *nextres = &residues[1];
	ResData *prevres = '\0';
	
	// Print first N atom.
	if (res->atoms[Np][0] != ATOM_COORD_NA) {
		fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
			(c++), res->names[Np], res->resName, res->id, res->atoms[Np][0], res->atoms[Np][1], res->atoms[Np][2]);
	}
	
	for (n=0; n<numRes; n++) {
		res = &residues[n];
		if (n > 0) {
			prevres = &residues[n-1];
		}
		nextres = &residues[n+1];
		printedHA = FALSE;
		printedHNs = FALSE;
		
		// Output atoms:
		
		if (res->atoms[CA][0] != ATOM_COORD_NA) {
			fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n", (c++), res->names[CA], res->resName, res->id, res->atoms[CA][0], res->atoms[CA][1], res->atoms[CA][2]);
		}
		
		if (res->atoms[C1][0] != ATOM_COORD_NA) {
			fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n", (c++), res->names[C1], res->resName, res->id, res->atoms[C1][0], res->atoms[C1][1], res->atoms[C1][2]);	
		}
		
		if (res->atoms[O][0] != ATOM_COORD_NA) {
			fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n", (c++), res->names[O], res->resName, res->id, res->atoms[O][0], res->atoms[O][1], res->atoms[O][2]);
		}
		
		if (strcmp(res->resName, "GLY")!=0 && res->atoms[CB][0] != ATOM_COORD_NA) {
			fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n", (c++), res->names[CB], res->resName, res->id, res->atoms[CB][0], res->atoms[CB][1], res->atoms[CB][2]);
		}
		
		// Print out sidechain non-hydrogen atoms and the OXT atom (if present).
		for (i = CG; i <= res->max_atom_index; i++) {
			if (res->names[i][0] != '\0' && res->names[i][1] != 'H') {
				if (sideChain || i == CG || strncmp(res->names[i], " OXT", 4) == 0) { // Sidechain atom.
					if (res->atoms[i][0] != ATOM_COORD_NA) {
						fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
							(c++), res->names[i], res->resName, res->id, res->atoms[i][0], res->atoms[i][1], res->atoms[i][2]);
					}
				}
			}
		}
		
		// Print out the HN atom(s) and HA atom(s), if present.
		if (n > 0) {
			if (strcmp(res->resName, "PRO") != 0) {
				if (prevres->atoms[HN][0] != ATOM_COORD_NA) {
					fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
						(c++), prevres->names[HN], res->resName, res->id, prevres->atoms[HN][0], prevres->atoms[HN][1], prevres->atoms[HN][2]);
				}
			}
		}
		else {
			for (i = HNp; i <= HNp3; i++) {
				if (res->atoms[i][0] != ATOM_COORD_NA) {
					fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
						(c++), res->names[i], res->resName, res->id, res->atoms[i][0], res->atoms[i][1], res->atoms[i][2]);
				}
			}
		}
		
		if (res->atoms[HA][0] != ATOM_COORD_NA) {
			fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
				(c++), res->names[HA], res->resName, res->id, res->atoms[HA][0], res->atoms[HA][1], res->atoms[HA][2]);
		}
		if (strcmp(res->resName, "GLY")==0) {
			if (res->atoms[HA2][0] != ATOM_COORD_NA) {
				fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n", (c++), res->names[HA2], res->resName, res->id, res->atoms[HA2][0], res->atoms[HA2][1], res->atoms[HA2][2]);
			}
		}
		
		// Print out the sidechain hydrogen atoms.
		for (i = CG; i <= res->max_atom_index; i++) {
			if (res->names[i][0] != '\0' && res->names[i][1] == 'H') {
				if (sideChain) { // Sidechain atom.
					if (res->atoms[i][0] != ATOM_COORD_NA) {
						fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
							(c++), res->names[i], res->resName, res->id, res->atoms[i][0], res->atoms[i][1], res->atoms[i][2]);
					}
				}
			}
		}
		
		// Print backbone N atom of next residue.
		if (n != (numRes-1)) { // If this is not the last residue.
			if (nextres->fixed_position == FALSE) {
				if (res->atoms[N][0] != ATOM_COORD_NA) {
					fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
						(c++), nextres->names[N], nextres->resName, nextres->id, res->atoms[N][0], res->atoms[N][1], res->atoms[N][2]);
				}
			}
			else {
				if (nextres->atoms[Np][0] != ATOM_COORD_NA) {
					fprintf(out, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
						(c++), nextres->names[N], nextres->resName, nextres->id, nextres->atoms[Np][0], nextres->atoms[Np][1], nextres->atoms[Np][2]);
				}
			}
		}
		else {
			fprintf(out, "TER  %6d %4s %s %5d %-11s %-7s %-7s                     \n",
				(c++), "    ", res->resName, res->id, "", "", "");
		}
		
	}
}//*** END printPDBAtoms....


/*
 * Output the given structure to the file with the given name.
 *
 * output_filename: The name of the file to which to write.
 * residues: Array of ResData structs representing the structure.
 * numRes: The number of residues in the structure.
 */
void printPDBAtomsFilename(char *output_filename, ResData *residues, int numRes) {
	FILE *fp = fopen(output_filename, "w");
	printPDBAtoms(fp, residues, numRes);
	fclose(fp);
}


/*
 * Given a 4-letter atom name abbreviation, this function returns the equivalent index
 * as defined in pepmake.h. Only "standard" atom names are accepted. If the atom name is
 * non-standard or does not have an index set aside in pepmake.h, this function returns -1.
 * The 4-letter atom name abbreviations occupy character columns 13-16 in a PDB file, as
 * note here: http://www.biochem.ucl.ac.uk/~roman/procheck/manual/manappb.html
 */
int getAtomIndexFromAtomName(char *name) {
	
	if      (strncmp(name, " N  ", 4) == 0) { return Np; }
	else if (strncmp(name, " CA ", 4) == 0) { return CA; }
	else if (strncmp(name, " C  ", 4) == 0) { return C1; }
	else if (strncmp(name, " O  ", 4) == 0) { return O; }
	else if (strncmp(name, " H  ", 4) == 0) { return HN; }
	else if (strncmp(name, "1H  ", 4) == 0) { return HNp; }
	else if (strncmp(name, "2H  ", 4) == 0) { return HNp2; }
	else if (strncmp(name, "3H  ", 4) == 0) { return HNp3; }
	else if (strncmp(name, " HA ", 4) == 0) { return HA; }
	else if (strncmp(name, "1HA ", 4) == 0) { return HA; }
	else if (strncmp(name, "2HA ", 4) == 0) { return HA2; }
	else if (strncmp(name, " CB ", 4) == 0) { return CB; }
	else if (strncmp(name, " CG ", 4) == 0) { return CG; }
	else if (strncmp(name, " CG1", 4) == 0) { return CG; }
	else if (strncmp(name, " CG2", 4) == 0) { return CG; }
	else if (strncmp(name, " OG ", 4) == 0) { return CG; }
	else if (strncmp(name, " OG1", 4) == 0) { return CG; }
	else if (strncmp(name, " OG2", 4) == 0) { return CG; }
	else if (strncmp(name, " SG ", 4) == 0) { return CG; }
	else {
		return -1;
	}
	
}


/*
 * Given an atom index as defined in pepmake.h, this function returns the 4-letter
 * atom name abbreviation for the index (or the most common atom name for a couple
 * indices). If an index is not defined, the empty string is returned.
 */
char *getAtomNameFromAtomIndex(int index) {
	
	switch(index) {
		case Np	 : return " N  ";
		case CA	 : return " CA ";
		case C1	 : return " C  ";
		case O	 : return " O  ";
		case N	 : return " N  ";
		case CA2 : return " CA ";
		case HN	 : return " H  ";
		case HNp : return "1H  ";
		case HNp2: return "2H  ";
		case HNp3: return "3H  ";
		case HA  : return " HA "; // Or "1HA "...
		case CB  : return " CB "; // Or "2HA "...
		case CG  : return " CG "; // Or others...
		default  : return "";
	}
	
}


/*************************************************************/
/* Apply the saved template of this residue's coordinates.   */
/* We rotate the residue into the canonical form, then swap  */
/* in the new coordinates, and then reverse all of the       */
/* rotations that were used to canonisize the coordinates    */
/*************************************************************/
static void applyTemplate(ResData *res) {
	int i;
	Vec3D xAxis = {1,0,0}; // a unit vector
	Vec3D T; // a temp vector
	double thetaZ, thetaY, thetaX;
	Vec3D transVec;
	
	// Translate to origin with CA at origin.
	copyVector(res->atoms[CA], transVec); // store for later
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			difference(res->atoms[i], transVec, res->atoms[i]);
		}
	}

	// compute angle from X axis
	setVector(T, res->atoms[C1][0], res->atoms[C1][1], 0);
	thetaZ = angle(T, xAxis);
	if (res->atoms[C1][1] > 0) thetaZ = 360 - thetaZ;
	// rotate atoms around Z axis
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateZ(res->atoms[i], thetaZ);
		}
	}
	
	// compute angle from X axis
	setVector(T, res->atoms[C1][0], 0, res->atoms[C1][2]);
	thetaY = angle(T, xAxis);
	if (res->atoms[C1][2] < 0) thetaY = 360 - thetaY;
	// rotate atoms around Y-axis
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateY(res->atoms[i], thetaY);
		}
	}
	
	// compute angle for N
	thetaX = atan(res->atoms[N][2] / res->atoms[N][1]);
	thetaX = 180 - (thetaX * 180.0/PI);
	if (res->atoms[N][1] < 0) thetaX = 180+thetaX;
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateX(res->atoms[i], thetaX);
		}
	}
	
	for (i=0; i <= res->max_atom_index; i++) {
		//printf("%d :  %lf %lf %lf\n", 
		//	i, res->atoms[i][0], res->atoms[i][1], res->atoms[i][2]);
		copyVector(res->template[i], res->atoms[i]);
	}
	
	// Initialize the other atoms.
	for (i = res->max_atom_index + 1; i < MAX_ATOMS; i++) {
		setVector(res->atoms[i], ATOM_COORD_NA, ATOM_COORD_NA, ATOM_COORD_NA);
	}

	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateX(res->atoms[i], -1*thetaX);
		}
	}
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateY(res->atoms[i], -1*thetaY);
		}
	}
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			rotateZ(res->atoms[i], -1*thetaZ);
		}
	}
	
	for (i=0; i <= res->max_atom_index; i++) {
		if (res->atoms[i][0] != ATOM_COORD_NA) {
			summation(res->atoms[i], transVec, res->atoms[i]);
		} 
	} 
}


/*************************************************************/
/* For debug. Outputs coordinates of the given vector/point. */
/*************************************************************/
void outputDebugCoords(char *filename, char *resName, int id, Vec3D *coords, int outputCB, int outputPRO) {
	FILE *fp_debug;
	int c = 1;
	
	fp_debug = fopen(filename, "w");
	
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " N  ", resName, id, coords[Np][0], coords[Np][1], coords[Np][2]);
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " CA ", resName, id, coords[CA][0], coords[CA][1], coords[CA][2]);
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " C  ", resName, id, coords[C1][0], coords[C1][1], coords[C1][2]);
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " O  ", resName, id, coords[O][0], coords[O][1], coords[O][2]);
	if (outputCB) {
		fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
			(c++), " CB ", resName, id, coords[CB][0], coords[CB][1], coords[CB][2]);
		
		// For PRO only
		fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
			(c++), " CG ", resName, id, coords[CG][0], coords[CG][1], coords[CG][2]);
		if (outputPRO) {
			fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
				(c++), " CD ", resName, id, coords[CG+1][0], coords[CG+1][1], coords[CG+1][2]);
			//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[Np][0], coords[Np][1], coords[Np][2]);
			//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[CG+1][0], coords[CG+1][1], coords[CG+1][2]);
			//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[CG][0], coords[CG][1], coords[CG][2]);
		}
		//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[CB][0], coords[CB][1], coords[CB][2]);
		
		//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[CA][0], coords[CA][1], coords[CA][2]);
	}
	//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[C1][0], coords[C1][1], coords[C1][2]);
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " N  ", "UNK", id+1, coords[N][0], coords[N][1], coords[N][2]);
	fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
		(c++), " CA ", "UNK", id+1, coords[CA2][0], coords[CA2][1], coords[CA2][2]);
	//fprintf(fp_debug, "%lf\t%lf\t%lf\n", coords[N][0], coords[N][1], coords[N][2]);
	if (coords[HN][0] != ATOM_COORD_NA) {
		fprintf(fp_debug, "ATOM %6d %4s %s %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
			(c++), " H  ", "UNK", id+1, coords[HN][0], coords[HN][1], coords[HN][2]);
	}
	
	fclose(fp_debug);
	
	/*
	// CA-C1 bond length print out.
	Vec3D CA_C1;
	difference(coords[CA], coords[C1], CA_C1);
	double mag = sqrt(dotproduct(CA_C1,CA_C1));
	printf("When outputting %s, CA-C1 bond length is %lf\n", filename, mag);
	*/
}


/*************************************************************/
/* Rotate the current residue by phi & psi. This function    */
/* also moves the residue from its old position (the same    */
/* position as the previous residue) to its new position.    */
/*                                                           */
/* Set rotateChi to TRUE to perform rotation of the chi      */
/* angle. Set onlyBackbone to TRUE to only manipulate        */
/* backhone atoms.                                           */
/*                                                           */
/* This version of the function is a complete reimplemention */
/* done by David Arndt in October-November 2006.             */
/*************************************************************/
static void rotatePhiPsi(int num, ResData *residues, int onlyBackbone) {
	int i;
	ResData *res = &residues[num];
	ResData *prev_res = '\0';
	double N_CA_C1_angle;
	Vec3D N_CA_axis; // Vector parallel to N-CA2 bond in previous residue.
	Vec3D CA_C1_axis; // Vector parallel to CA-C1 bond in current residue.
	Vec3D L;
	
	// debug variables
	int debug = FALSE; // If TRUE, assumes existence of directory "coords_for_Grapher".
	int debugNumToPrint = 0; // Which residue index to print for debug info.
	int debugOutputPRO = (strcmp(res->resName, "PRO") == 0);
	int debugPrintCB = TRUE;
	double intended_phi, intended_psi;
	FILE *fp_debug;
	
	if (num > 0) {
		prev_res = &residues[num-1];
	}
	
	//debug
	if (debug) {
		intended_phi = res->phi;
		intended_psi = res->psi;
		if (num == debugNumToPrint) {
			printf("Saving debug info for residue %d.\n", num);
			printf("Intended rotation: phi = %lf, psi = %lf, chi = %lf\n", res->phi, res->psi, res->chi);
		}
		if (num == debugNumToPrint && num > 0) {
			outputDebugCoords("coords_for_Grapher/00_prev.pdb", 
				prev_res->resName, prev_res->id, prev_res->atoms, debugPrintCB, debugOutputPRO);
		}
		if (num == debugNumToPrint && num > 0) {
			Vec3D prev_trans[MAX_ATOMS]; // Previous residue translated to origin by CA2. For debug only.
			
			// Translate previous residue to the origin.
			for (i=0; i <= prev_res->max_atom_index; i++) {
				if (prev_res->atoms[i][0] != ATOM_COORD_NA) {
					difference(prev_res->atoms[i], prev_res->atoms[CA2], prev_trans[i]);
				}
			}
			
			outputDebugCoords("coords_for_Grapher/01_prev_trans.pdb",
				prev_res->resName, prev_res->id, prev_trans, debugPrintCB, debugOutputPRO);
		}
		if (num == debugNumToPrint) {
			outputDebugCoords("coords_for_Grapher/02_init.pdb",
				res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
		}
	}
	
	difference(res->atoms[C1], res->atoms[CA], CA_C1_axis);
	unitVector(CA_C1_axis, CA_C1_axis);
	if (num > 0) {
		difference(prev_res->atoms[CA2], prev_res->atoms[N], N_CA_axis);
		unitVector(N_CA_axis, N_CA_axis);
	}
	else {
		N_CA_axis[0] = 1.0; N_CA_axis[1] = 0.0; N_CA_axis[2] = 0.0;
	}
	
	//debug
	if (debug) {
		if (num == debugNumToPrint) {
			
			crossproduct(CA_C1_axis, N_CA_axis, L); // Do this ahead of time for debugging purposes.
			unitVector(L, L);
			
			fp_debug = fopen("coords_for_Grapher/03_vec.txt","w");
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", 0.0, 0.0, 0.0);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", CA_C1_axis[0], CA_C1_axis[1], CA_C1_axis[2]);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", 0.0, 0.0, 0.0);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", N_CA_axis[0], N_CA_axis[1], N_CA_axis[2]);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", 0.0, 0.0, 0.0);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", L[0], L[1], L[2]);
			fclose(fp_debug);
		}
	}
	
	// Translate current residue to the origin with CA at origin.
	Vec3D old_CA;
	copyVector(res->atoms[CA], old_CA);
	if (!onlyBackbone) {
		for (i=0; i <= res->max_atom_index; i++) {
			if (res->atoms[i][0] != ATOM_COORD_NA) {
				difference(res->atoms[i], old_CA, res->atoms[i]);
			}
		}
	}
	else {
		for (i=0; i<=CA2; i++) {
			difference(res->atoms[i], old_CA, res->atoms[i]);
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/04_to_origin.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// Rotate the current residue so that its CA-C1 bond is parallel to the N-CA2 bond of the previous residue.
	crossproduct(CA_C1_axis, N_CA_axis, L);
	if ( !(L[0] == 0 && L[1] == 0 && L[2] == 0) && num > 0) {
		// Only do the rotation if L is non-zero. If it is zero, the bonds are already parallel, and we avoid
		// a call to rotateCC that would set all atom coordinates to (0, 0, 0).
		
		unitVector(L, L);
		
		double temp_angle = angle(N_CA_axis, CA_C1_axis);
		if (!onlyBackbone) {
			for (i=0; i <= res->max_atom_index; i++) {
				if (res->atoms[i][0] != ATOM_COORD_NA) {
					rotateCC(L, res->atoms[i], temp_angle);
				}
			}
		}
		else {
			for (i=0; i<=CA2; i++) {
				rotateCC(L, res->atoms[i], temp_angle);
			}
		}
		
		// We want the C1-N-CA2 of the previous residue to be in the same plane as CA-C1-N of the current residue,
		// but they may be off slightly, so we need to do a rotation of the current residue to straighten it.
		
		Vec3D C1_N_prev, C1_N_curr, cp1, cp2, cp12;
		difference(prev_res->atoms[N], prev_res->atoms[C1], C1_N_prev);
		difference(res->atoms[N], res->atoms[C1], C1_N_curr);
		crossproduct(C1_N_prev, N_CA_axis, cp1);
		//printf("%lf\t%lf\t%lf\n", cp1[0], cp1[1], cp1[2]);
		unitVector(cp1, cp1);
		crossproduct(C1_N_curr, N_CA_axis, cp2);
		//printf("%lf\t%lf\t%lf\n", cp2[0], cp2[1], cp2[2]);
		unitVector(cp2, cp2);
		crossproduct(cp1, cp2, cp12);
		double C1_N_CA_C1_N_angle = angle(cp1, cp2);
		if (dotproduct(N_CA_axis, cp12) > 0.0) {
			//printf("   *dotproduct(N_CA_axis, cp12) > 0.0\n");
			C1_N_CA_C1_N_angle = -C1_N_CA_C1_N_angle;
		}
		
		//printf("   About to apply rotation by C1-N-CA-C1-N angle: %lf\n", C1_N_CA_C1_N_angle);
		if (!onlyBackbone) {
			for (i=0; i <= res->max_atom_index; i++) {
				if (res->atoms[i][0] != ATOM_COORD_NA) {
					rotateCC(N_CA_axis, res->atoms[i], C1_N_CA_C1_N_angle);
				}
			}
		}
		else {
			for (i=0; i<=CA2; i++) {
				rotateCC(N_CA_axis, res->atoms[i], C1_N_CA_C1_N_angle);
			}
		}
				
		// Debug. Test angle after applying rotation.
		/*
		difference(prev_res->atoms[N], prev_res->atoms[C1], C1_N_prev);
		difference(res->atoms[N], res->atoms[C1], C1_N_curr);
		crossproduct(C1_N_prev, N_CA_axis, cp1);
		unitVector(cp1, cp1);
		crossproduct(C1_N_curr, N_CA_axis, cp2);
		unitVector(cp2, cp2);
		crossproduct(cp1, cp2, cp12);
		C1_N_CA_C1_N_angle = angle(cp1, cp2);
		if (dotproduct(N_CA_axis, cp12) > 0.0) {
			//printf("   *dotproduct(N_CA_axis, cp12) > 0.0\n");
			C1_N_CA_C1_N_angle = -C1_N_CA_C1_N_angle;
		}
		printf("   After rotation C1-N-CA-C1-N angle: %lf\n", C1_N_CA_C1_N_angle);
		*/
		
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/05_N_CA_par_CA_C1.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// We assume that psi == 180 degrees. That is, the R-group of the current residue is oriented in such a way
	// that the bond from the first CA of the current residue leading to N is in the same plane.
	
	// Before this rotation, the current residue's N-CA-C1 angle (N from the previous residue, CA and C1 from the
	// current residue) is 180 degrees. Rotate the current residue so that this angle is equal to N_CA_C1_angle.
	Vec3D C1_N_axis_temp; // C1-N axis of current residue
	Vec3D rotate_axis_temp;
	difference(res->atoms[C1], res->atoms[N], C1_N_axis_temp);
	crossproduct(N_CA_axis, C1_N_axis_temp, rotate_axis_temp);
	unitVector(rotate_axis_temp, rotate_axis_temp);
	N_CA_C1_angle = res->N_CA_C1_angle;
	
	// debug
	if (debug) {
		printf("N_CA_C1_angle = %lf\n", N_CA_C1_angle);
	}
	
	if (!onlyBackbone) {
		for (i=0; i <= res->max_atom_index; i++) {
			if (res->atoms[i][0] != ATOM_COORD_NA) {
				rotateCC(rotate_axis_temp, res->atoms[i], 180 - N_CA_C1_angle);
			}
		}
	}
	else {
		for (i=0; i<=CA2; i++) {
			rotateCC(rotate_axis_temp, res->atoms[i], 180 - N_CA_C1_angle);
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/06_after_rotate_N_CA_C1_angle.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// Re-calculate CA_C1_axis for the current residue; it has changes because of rotations of the current residue.
	difference(res->atoms[C1], res->atoms[CA], CA_C1_axis);
	unitVector(CA_C1_axis, CA_C1_axis);
	
	// Perform psi rotation. Note that we only rotate the backbone atoms after the CA-C1 bond.
	// (We omit rotating C1 because we know it will stay in the same place).
	if (res->psi != ANGLE_NA) {
		rotateCC(CA_C1_axis, res->atoms[O], res->psi - 180);
		rotateCC(CA_C1_axis, res->atoms[N], res->psi - 180);
		rotateCC(CA_C1_axis, res->atoms[CA2], res->psi - 180);
		if (!onlyBackbone) {
			if (res->atoms[HN][0] != ATOM_COORD_NA) {
				rotateCC(CA_C1_axis, res->atoms[HN], res->psi - 180);
			}
		}
		// Also rotate the OXT atom (which should only appear in the last residue).
		for (i=CG; i <= res->max_atom_index; i++) {
			if (strcmp(res->names[i], " OXT")==0) {
				rotateCC(CA_C1_axis, res->atoms[i], res->psi - 180);
				break;
			}
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/07_after_psi.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// Perform phi rotation. Note that we do not rotate Np, HNp, HNp2, or HNp3 (nor CA because we know it
	// will stay in the same place).
	if (res->phi != ANGLE_NA) {
		if (!onlyBackbone) {
			for (i=C1; i<=HN; i++) {
				if (res->atoms[i][0] != ATOM_COORD_NA) {
					rotateCC(N_CA_axis, res->atoms[i], res->phi);
				}
			}
			for (i=HA; i <= res->max_atom_index; i++) {
				if (res->atoms[i][0] != ATOM_COORD_NA) {
					rotateCC(N_CA_axis, res->atoms[i], res->phi);
				}
			}
		}
		else {
			for (i=C1; i<=CA2; i++) {
				rotateCC(N_CA_axis, res->atoms[i], res->phi);
			}
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/08_after_phi.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// Perform chi rotation.
	if ( rotateChi && !onlyBackbone && strcmp(res->resName, "GLY")!=0 && strcmp(res->resName, "ALA")!=0 &&
		res->atoms[CB][0] != ATOM_COORD_NA && res->atoms[CG][0] != ATOM_COORD_NA ) {
		
		double previous_chi;
		
		if (num > 0) {
			Vec3D prev_trans[MAX_ATOMS]; // Previous residue translated to origin with CA2 at origin.
			
			// Translate previous residue to the origin. (Store result in prev_trans).
			for (i=0; i <= prev_res->max_atom_index; i++) {
				difference(prev_res->atoms[i], prev_res->atoms[CA2], prev_trans[i]);
			}
			
			previous_chi = torsion(prev_trans[N], res->atoms[CA], res->atoms[CB], res->atoms[CG]);
			if (debug) {
				printf("   torsion of %lf from:\n", previous_chi);
				printf("   N: (%lf, %lf, %lf), ", prev_trans[N][0], prev_trans[N][1], prev_trans[N][2]);
				printf("CA: (%lf, %lf, %lf), ", res->atoms[CA][0], res->atoms[CA][1], res->atoms[CA][2]);
				printf("CB: (%lf, %lf, %lf), ", res->atoms[CB][0], res->atoms[CB][1], res->atoms[CB][2]);
				printf("CG: (%lf, %lf, %lf)\n", res->atoms[CG][0], res->atoms[CG][1], res->atoms[CG][2]);
			}
		}
		else {
			previous_chi = torsion(res->atoms[Np], res->atoms[CA], res->atoms[CB], res->atoms[CG]);
			if (debug) {
				printf("   torsion of %lf from:\n", previous_chi);
				printf("   Np: (%lf, %lf, %lf), ", res->atoms[Np][0], res->atoms[Np][1], res->atoms[Np][2]);
				printf("CA: (%lf, %lf, %lf), ", res->atoms[CA][0], res->atoms[CA][1], res->atoms[CA][2]);
				printf("CB: (%lf, %lf, %lf), ", res->atoms[CB][0], res->atoms[CB][1], res->atoms[CB][2]);
				printf("CG: (%lf, %lf, %lf)\n", res->atoms[CG][0], res->atoms[CG][1], res->atoms[CG][2]);
			}
		}
		
		Vec3D CA_CB_axis;
		difference(res->atoms[CB], res->atoms[CA], CA_CB_axis);
		unitVector(CA_CB_axis, CA_CB_axis);
		
		for (i=CG; i <= res->max_atom_index; i++) { // We assume that all entries in this range are for valid atoms.
			if (strcmp(res->names[i], " OXT")!=0) { // Do not rotate the OXT atom (which should only appear in the last residue).
				rotateCC(CA_CB_axis, res->atoms[i], res->chi - previous_chi);
			}
		}
		
		// debug
		if (debug) {
			double achieved_chi;
			if (num > 0) {
				Vec3D prev_trans[MAX_ATOMS]; // Previous residue translated to origin with CA2 at origin.
				
				// Translate previous residue to the origin. (Store result in prev_trans).
				for (i=0; i <= prev_res->max_atom_index; i++) {
					difference(prev_res->atoms[i], prev_res->atoms[CA2], prev_trans[i]);
				}
				
				achieved_chi = torsion(prev_trans[N], res->atoms[CA], res->atoms[CB], res->atoms[CG]);
			}
			else {
				achieved_chi = torsion(res->atoms[Np], res->atoms[CA], res->atoms[CB], res->atoms[CG]);
			}
			printf("   chi: prev = %lf, intended = %lf, achieved = %lf\n", previous_chi, res->chi, achieved_chi);
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/09_after_chi.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	// Translate current residue so that its first CA overlaps with CA2 of the previous residue.
	Vec3D prev_CA2;
	if (num > 0) {
		copyVector(prev_res->atoms[CA2], prev_CA2);
	}
	else {
		prev_CA2[0] = 0.0; prev_CA2[1] = 0.0; prev_CA2[2] = 0.0;
	}
	if (!onlyBackbone) {
		for (i=0; i <= res->max_atom_index; i++) {
			if (res->atoms[i][0] != ATOM_COORD_NA) {
				summation(res->atoms[i], prev_CA2, res->atoms[i]);
			}
		}
	}
	else {
		for (i=0; i<=CA2; i++) {
			summation(res->atoms[i], prev_CA2, res->atoms[i]);
		}
	}
	
	//debug
	if (debug && num == debugNumToPrint) {
		outputDebugCoords("coords_for_Grapher/10_done.pdb", res->resName, res->id, res->atoms, debugPrintCB, debugOutputPRO);
	}
	
	//debug
	if (debug) {
		
		// Determine actual phi and psi achieved.
		if (num == debugNumToPrint && num > 0) {
			
			// Print the atom coordinates used for computation of achieved phi to a file.
			fp_debug = fopen("coords_for_Grapher/phi_atoms.txt","w");
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", prev_res->atoms[C1][0], prev_res->atoms[C1][1], prev_res->atoms[C1][2]);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", prev_res->atoms[N][0], prev_res->atoms[N][1], prev_res->atoms[N][2]);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", prev_res->atoms[CA2][0], prev_res->atoms[CA2][1], prev_res->atoms[CA2][2]);
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", res->atoms[CA][0], res->atoms[CA][1], res->atoms[CA][2]); // Should be the same as prev's CA2.
			fprintf(fp_debug, "%lf\t%lf\t%lf\n", res->atoms[C1][0], res->atoms[C1][1], res->atoms[C1][2]);
			fclose(fp_debug);
			
			// Compute achieved phi and psi.
			
			double phi_achieved = torsion(prev_res->atoms[C1], prev_res->atoms[N], res->atoms[CA], res->atoms[C1]);
			printf("Acheived phi = %lf\n", phi_achieved);
			
			double psi_achieved = torsion(prev_res->atoms[N], res->atoms[CA], res->atoms[C1], res->atoms[N]);
			printf("Acheived psi = %lf\n", psi_achieved);
		}
		
		// Print all angle comparisons.
		if (num > 0) {
			double phi_achieved = torsion(prev_res->atoms[C1], prev_res->atoms[N], res->atoms[CA], res->atoms[C1]);
			//printf("Acheived phi = %lf\n", phi_achieved);
			
			double psi_achieved = torsion(prev_res->atoms[N], res->atoms[CA], res->atoms[C1], res->atoms[N]);
			//printf("Acheived psi = %lf\n", psi_achieved);
			
			double chi_achieved;
			if ( num > 0 && !(strcmp(res->resName, "ALA") == 0 || strcmp(res->resName, "GLY") == 0) &&
					res->atoms[CB][0] != ATOM_COORD_NA && res->atoms[CG][0] != ATOM_COORD_NA ) { // If this residue has a CG atom.
				
				chi_achieved = torsion(prev_res->atoms[N], res->atoms[CA], res->atoms[CB], res->atoms[CG]);
				printf("%d (%s, %d): phi %lf - %lf = %lf, psi %lf - %lf = %lf, chi %lf - %lf = %lf\n", num, res->resName, res->max_atom_index,
					intended_phi, phi_achieved, intended_phi - phi_achieved,
					intended_psi, psi_achieved, intended_psi - psi_achieved,
					res->chi, chi_achieved, res->chi - chi_achieved);
			}
			else {
				printf("%d (%s, %d): phi %lf - %lf = %lf, psi %lf - %lf = %lf, chi N/A\n", num, res->resName, res->max_atom_index,
					intended_phi, phi_achieved, intended_phi - phi_achieved,
					intended_psi, psi_achieved, intended_psi - psi_achieved);
			}
		}
		
		printf("\n");
	}
	
}


/*************************************************************/
/* Copy the position of src to be the position of dst        */
/*************************************************************/
void copyPosition(ResData *src, ResData *dst) {
	int i;
	for (i=0; i <= src->max_atom_index; i++) {
		copyVector(src->atoms[i], dst->atoms[i]);
	}
}


/*
 * Copies the data for a single ResData struct from the source to the
 * destination.
 *
 * copy_hbonds: Boolean. Whether to copy hydrogen bond records.
 */
void copySingleResData(ResData *src, ResData *dst, int copy_hbonds) {
	int j;
	dst->phi = src->phi;
	dst->psi = src->psi;
	dst->omega = src->omega;
	dst->chi = src->chi;
	dst->N_CA_C1_angle = src->N_CA_C1_angle;
	dst->id = src->id;
	dst->max_atom_index = src->max_atom_index;
	for(j=0; j <= src->max_atom_index; j++){
		copyVector(src->atoms[j], dst->atoms[j]);
		copyVector(src->template[j], dst->template[j]);
		strncpy(dst->names[j], src->names[j], 5);
	}
	strncpy(dst->resName, src->resName, 4);
	dst->fixed_position = src->fixed_position;
	if (copy_hbonds) {
		for (j = 0; j < 2; j++) {
			dst->hbonds_acceptor[j] = src->hbonds_acceptor[j];
			dst->hbonds_donor[j] = src->hbonds_donor[j];
			dst->hbonds_intended_acc[j] = src->hbonds_intended_acc[j];
			dst->hbonds_intended_don[j] = src->hbonds_intended_don[j];
		}
	}
}


/*
 * Copies the data for multiple ResData struct from the source to the
 * destination.
 *
 * n: Number of ResData structs in both src and dst.
 * copy_hbonds: Boolean. Whether to copy hydrogen bond records.
 */
void copyMultipleResData(ResData *src, ResData *dst, int n, int copy_hbonds) {
	int i;
	
	for (i = 0; i < n; i++) {
		copySingleResData(&src[i], &dst[i], copy_hbonds);
	}
}


/*************************************************************/
/* Add the next residue to the peptide chain, and carry out  */
/* the appropriate rotations and translations.               */
/* Set onlyBackbone to TRUE to only perform operations on    */
/* backbone atoms.                                           */
/*************************************************************/
static void nextPeptide(ResData *res, int num, ResData *residues, int onlyBackbone) {
	Vec3D CN;

	if (num > 0) {
		copyPosition(&residues[num-1], res);
	}
	
	applyTemplate(res);
	
	// Perform omega rotation for the peptide bond in the current residue.
	// This will potentially result in the relocation of the HN
	// and CA2 atoms in the current residue. This needs to be done
	// when reading residues from the template database, which have omega
	// angles that do not agree with those provided in the angles file. If,
	// on the other hand, PepMake has been given a PDB file, PepMake has
	// called phipsi with the --db option to generate a residue coordinates
	// file that already has the correct omega rotations implicit in the coordinates.
	// However, if we will be adjusting the omega angles (such as with repairAngles())
	// or if PepMake is used as part of another program such as GAFolder, then
	// we may want to create a peptide with omega angles that differ from those
	// of the given PDB file, and the code below will adjust the omega angles of
	// the peptide.
	
	if (adjustOmega) {
		double omega = res->omega;
		//int debugNumToPrint = 1;
		
		//if (res->atoms[CA][0] != ATOM_COORD_NA && res->atoms[C1][0] != ATOM_COORD_NA &&
		//	res->atoms[N][0] != ATOM_COORD_NA && res->atoms[CA2][0] != ATOM_COORD_NA) {
			
			//debug
			//if (num == debugNumToPrint) {
			//	outputDebugCoords("coords_for_Grapher/before_omega.txt", res->atoms, TRUE, FALSE);
			//}
			
			// determine omega currently present in the residue template
			double previous_omega = torsion(res->atoms[CA], res->atoms[C1], res->atoms[N], res->atoms[CA2]);
			
			difference(res->atoms[N], res->atoms[C1], CN);
			unitVector(CN, CN);
			//printf("CN: (%lf, %lf, %lf)\n", CN[0], CN[1], CN[2]); // debug
			
			// Translate to origin - but only translate the atoms that we will rotate.
			Vec3D C1_vec;
			copyVector(res->atoms[C1], C1_vec);
			//for (i=0; i <= res->max_atom_index; i++) {
			//	difference(res->atoms[i], C1_vec, res->atoms[i]);
			//}
			difference(res->atoms[CA2], C1_vec, res->atoms[CA2]);
			if (!onlyBackbone && res->atoms[HN][0] != ATOM_COORD_NA) {
				difference(res->atoms[HN], C1_vec, res->atoms[HN]);
			}
			
			//debug
			//if (num == debugNumToPrint) {
			//	outputDebugCoords("coords_for_Grapher/omega_to_origin.txt", res->atoms, TRUE, FALSE);
			//}
			
			// perform omega rotation
			rotateCC(CN, res->atoms[CA2], omega - previous_omega);
			if (!onlyBackbone && res->atoms[HN][0] != ATOM_COORD_NA) {
				rotateCC(CN, res->atoms[HN], omega - previous_omega);
			}
			
			// translate back to original position
			//for (i=0; i <= res->max_atom_index; i++) {
			//	summation(res->atoms[i], C1_vec, res->atoms[i]);
			//}
			summation(res->atoms[CA2], C1_vec, res->atoms[CA2]);
			if (!onlyBackbone && res->atoms[HN][0] != ATOM_COORD_NA) {
				summation(res->atoms[HN], C1_vec, res->atoms[HN]);
			}
			
			//debug
			//if (num == debugNumToPrint) {
			//	outputDebugCoords("coords_for_Grapher/after_omega.txt", res->atoms, TRUE, FALSE);
			//}
			
			//debug
			//double resulting_omega = torsion(res->atoms[CA], res->atoms[C1], res->atoms[N], res->atoms[CA2]);
			//printf("Performing omega rotation of %lf for residue %d (previous omega = %lf, resulting omega = %lf)\n", omega, num, previous_omega, resulting_omega);
		//}
	}
	
	rotatePhiPsi(num, residues, onlyBackbone);
}


/*************************************************************/
/* Initialize the residue atom coordinates to the default    */
/* template positions.                                       */
/*************************************************************/
static void initResData(ResData *r) {
	int i;
	for (i=0; i<=CA2; i++) { 
		copyVector(template[i], r->atoms[i]);
	}
}


/*
 * Extracts the sequence from the given array of ResData structs and writes it to 'seq' as
 * a null-terminated string of single-letter amino acid abbreviations.
 */
void get_seq_from_ResData(ResData *res, int n, char *seq) {
	int i;
	char c;
	
	for (i = 0; i < n; i++) {
		if ((c = res_str_to_char(res[i].resName)) != '?') {
			seq[i] = c;
		}
		else {
			fprintf(stderr, "ERROR: Unrecognized residue name \"%3s\" (residue ID %d).\n", res[n].resName, res[n].id);
			exit(1);
		}
	}
	seq[n] = '\0';
}


/*************************************************************/
/* Returns the N_CA_C1_angle for the given phi and psi       */
/* angles by performing a table lookup.                      */
/* whichLookupTable should refer to one of                   */
/* general_N_CA_C1_angle_lookup, gly_N_CA_C1_angle_lookup,   */
/* pro_N_CA_C1_angle_lookup, or xpr_N_CA_C1_angle_lookup.    */
/*************************************************************/
double lookup_N_CA_C1_angle(double whichLookupTable[37][37], double phi, double psi) {
	
	if (phi < -180.0) {
		phi += 360.0;
	}
	else if (phi > 180.0) {
		phi -= 360.0;
	}
	
	if (psi < -180.0) {
		psi += 360.0;
	}
	else if (psi > 180.0) {
		psi -= 360.0;
	}
	
	int phi_index = ((int) ((phi+180.0)/10.0)) + 1;
	int psi_index = ((int) ((psi+180.0)/10.0)) + 1;
	
	return whichLookupTable[phi_index][psi_index];
}


/*************************************************************/
/* Returns the N_CA_C1_angle for the given phi angle (with   */
/* psi unspecified) by performing a table lookup.            */
/*************************************************************/
double lookup_N_CA_C1_angle_from_phi(double whichLookupTable[37][37], double phi) {
	if (phi < -180.0) {
		phi += 360.0;
	}
	else if (phi > 180.0) {
		phi -= 360.0;
	}
	
	int phi_index = ((int) ((phi+180.0)/10.0)) + 1;
	int psi_index = 0;
	
	return whichLookupTable[phi_index][psi_index];
}


/*************************************************************/
/* Returns the N_CA_C1_angle for the given psi angle (with   */
/* phi unspecified) by performing a table lookup.            */
/*************************************************************/
double lookup_N_CA_C1_angle_from_psi(double whichLookupTable[37][37], double psi) {
	if (psi < -180.0) {
		psi += 360.0;
	}
	else if (psi > 180.0) {
		psi -= 360.0;
	}
	
	int phi_index = 0;
	int psi_index = ((int) ((psi+180.0)/10.0)) + 1;
	
	return whichLookupTable[phi_index][psi_index];
}


/*************************************************************/
/* Returns the baseline N_CA_C1_angle for all phi and psi    */
/* angles by performing a table lookup.                      */
/*************************************************************/
double lookup_N_CA_C1_angle_baseline(double whichLookupTable[37][37]) {
	return whichLookupTable[0][0];
}


/*************************************************************/
/* Utility function for use in the                           */
/* read_N_CA_C1_angle_lookup_file function. Converts the     */
/* given char into an angle deviation (in degrees),          */
/* according to the codes described in Figs. 4 and 5 of:     */
/*                                                           */
/* Karplus PA. Experimentally observed                       */
/* conformation-dependent geometry and hidden strain in      */
/* proteins. Protein Sci. 1996 Jul;5(7):1406-20.             */
/*************************************************************/
double convert_char_to_N_CA_C1_angle_deviation(char givenChar) {
	
	char tempString[2];
	tempString[1] = '\0';
	
	if (givenChar == '-') {
		return -6.0;
	}
	else if (givenChar == '=') {
		return -5.0;
	}
	else if (givenChar == 'A') {
		return 5.0;
	}
	else if (givenChar == 'B') {
		return 6.0;
	}
	else if (givenChar == 'C') {
		return 7.0;
	}
	else if (givenChar == 'D') {
		return 8.0;
	}
	else if (givenChar == 'E') {
		return 9.0;
	}
	else {
		// Assume char is a decimal number from 1 to 9 inclusive.
		tempString[0] = givenChar;
		return atof(tempString) -5.0; // Subtract 5.0 becasue '5' corresponds to a 0.0 degree deviation.
	}
}


/******************************************************************************************************************/
/* Reads the N-CA-C1 angle lookup file with the given filename, parses it and converts the data, and populates    */
/* the given array (one of general_N_CA_C1_angle_lookup, gly_N_CA_C1_angle_lookup, pro_N_CA_C1_angle_lookup, or   */
/* xpr_N_CA_C1_angle_lookup. These lookup tables are used to choose default N-CA-C1 angles based on the current   */
/* phi and psi angles. Karplus (see reference below) showed that the N-CA-C1 angle depends primarily on the phi   */
/* and psi angles, and the data in the given file (from the directory N_CA_C_angle_lookup) is based on data from  */
/* Karplus's paper.                                                                                               */
/*                                                                                                                */
/* More specifically, the files in the N_CA_C_angle_lookup directory corresponds to Fig. 5A, 5E, and 5G in the    */
/* paper, as well as data for Proline from the electronic data supplement for the paper. Data for all four files  */
/* was obtained from the electronic data supplement for this paper, from:                                         */
/*                                                                                                                */
/*    http://www.proteinscience.org/cgi/content/full/5/7/1406/DC1                                                 */
/*                                                                                                                */
/* The file "general.txt" was downloaded from here, and the data for param 10 (the N-CA-C1 angle) was copied out  */
/* into its own file, N_CA_C1_angle_lookup.txt.                                                                   */
/*                                                                                                                */
/* Karplus PA. Experimentally observed conformation-dependent geometry and hidden strain in proteins. Protein     */
/* Sci. 1996 Jul;5(7):1406-20.                                                                                    */
/******************************************************************************************************************/
void read_N_CA_C1_angle_lookup_file(char *filename, double targetArray[37][37]) {
	FILE *fp;
	int maxline = 40; // max chars per line
	char currentline[maxline];
	char currentChar;
	int charNum;
	int phi_index, psi_index; // indices in N_CA_C1_angle_lookup
	char lookup_table_chars[37][37]; // Lookup table with elements read in as chars.
	
	if ( (fp = fopen(filename, "r")) ) {
		
		// Skip the first 2 lines.
		fgets(currentline, maxline, fp);
		fgets(currentline, maxline, fp);
		
		// Read the data into lookup_table_chars first.
		
		// Read first line of data.
		psi_index = 0;
		fgets(currentline, maxline, fp);
		for (charNum = 0; charNum < 36; charNum++) {
			currentChar = currentline[charNum];
			phi_index = charNum + 1;
			lookup_table_chars[phi_index][psi_index] = currentChar;
		}
		lookup_table_chars[0][psi_index] = currentline[37];
		
		// Read the rest of the lines, indexing the array in reverse order because the psi angles in the
		// file go from 180 to -180, but in the array they will correspond to -180 to 180.
		for (psi_index = 36; psi_index > 0; psi_index--) {
			
			fgets(currentline, maxline, fp);
			
			for (charNum = 0; charNum < 36; charNum++) {
				currentChar = currentline[charNum];
				phi_index = charNum + 1;
				lookup_table_chars[phi_index][psi_index] = currentChar;
			}
			lookup_table_chars[0][psi_index] = currentline[37];
		}
		
		fclose(fp);
		
		// Convert the data in char format in lookup_table_chars into double format
		// representing the actual N_CA_C1 angles for N_CA_C1_angle_lookup.
		for (psi_index = 0; psi_index < 37; psi_index++) {
			
			for (phi_index = 0; phi_index < 37; phi_index++) {
				currentChar = lookup_table_chars[phi_index][psi_index];
				char chosenChar;
				double N_CA_C1_angle = 0.0; // This value should be overwritten.
				
				if (currentChar == 'x' || currentChar == '.') {
					char phi_axis_char = lookup_table_chars[phi_index][0];
					char psi_axis_char = lookup_table_chars[0][psi_index];
					if ( (phi_axis_char == 'x' || phi_axis_char == '.') &&
							(psi_axis_char == 'x' || psi_axis_char == '.') ) {
						chosenChar = lookup_table_chars[0][0];
						N_CA_C1_angle = convert_char_to_N_CA_C1_angle_deviation(chosenChar) + 110.0;
					}
					else if ( (phi_axis_char == 'x' || phi_axis_char == '.') &&
							!(psi_axis_char == 'x' || psi_axis_char == '.') ) {
						chosenChar = psi_axis_char;
						N_CA_C1_angle = convert_char_to_N_CA_C1_angle_deviation(chosenChar) + 110.0;
					}
					else if ( !(phi_axis_char == 'x' || phi_axis_char == '.') &&
							(psi_axis_char == 'x' || psi_axis_char == '.') ) {
						chosenChar = phi_axis_char;
						N_CA_C1_angle = convert_char_to_N_CA_C1_angle_deviation(chosenChar) + 110.0;
					}
					else if ( !(phi_axis_char == 'x' || phi_axis_char == '.') &&
							!(psi_axis_char == 'x' || psi_axis_char == '.') ) {
						N_CA_C1_angle = (convert_char_to_N_CA_C1_angle_deviation(phi_axis_char) +
							convert_char_to_N_CA_C1_angle_deviation(psi_axis_char))/2.0 + 110.0;
					}
				}
				else {
					chosenChar = currentChar;
					N_CA_C1_angle = convert_char_to_N_CA_C1_angle_deviation(chosenChar) + 110.0;
				}
				
				//N_CA_C1_angle_lookup[phi_index][psi_index] = N_CA_C1_angle;
				targetArray[phi_index][psi_index] = N_CA_C1_angle;
				
				//debug
				//printf("(%d, %d): %lf\n", phi_index, psi_index, N_CA_C1_angle);
				
			}
		}
		
	} else {
		fprintf(stderr, "Cannot open %s\n", filename);
		exit(1);
	}
	
}


/*
 * Predicts the N-CA-C1 angle for a single residue based on its phi and
 * psi angles. This function assumes that the ResData's name and phi and
 * psi angles are already set. For more information and a reference,
 * see the comments for the function read_N_CA_C1_angle_lookup_file.
 *
 * next_res_is_Pro: Boolean. Whether the residue following the given
 *    residue is a Proline residue.
 */
void predict_N_CA_C1_angle(ResData *res, int next_res_is_Pro) {
	
	if (next_res_is_Pro) {
		// If this residue is not the last residue and the following residue is a Pro residue.
		res->N_CA_C1_angle = lookup_N_CA_C1_angle(xpr_N_CA_C1_angle_lookup, res->phi, res->psi);
	}
	else if (strcmp(res->resName, "GLY") == 0) {
		res->N_CA_C1_angle = lookup_N_CA_C1_angle(gly_N_CA_C1_angle_lookup, res->phi, res->psi);
	}
	else if (strcmp(res->resName, "PRO") == 0) {
		res->N_CA_C1_angle = lookup_N_CA_C1_angle(pro_N_CA_C1_angle_lookup, res->phi, res->psi);
	}
	else {
		res->N_CA_C1_angle = lookup_N_CA_C1_angle(general_N_CA_C1_angle_lookup, res->phi, res->psi);
	}
	
	/* Uncomment one of these below to set a default N_CA_C1_angle for all residues. */
	//res->N_CA_C1_angle = 109.5;
	//res->N_CA_C1_angle = 111.0;
}


/*
 * Predicts and sets the N-CA-C1 angle for all n residues in the given peptide
 * based on their phi and psi angles. This function assumes that each ResData's
 * name and phi and psi angles are already set. For more information and a
 * reference, see the comments for the function read_N_CA_C1_angle_lookup_file.
 */    
void predict_N_CA_C1_angles(ResData *residues, int n) {
	int i;
	int next_res_is_Pro;
	
	// Do a table lookup for the N-CA-C1 angles based on the given phi and psi angles.
	// For more information and a reference, see the comments for the function read_N_CA_C1_angle_lookup_file.
	for (i=0; i<n; i++) {
		if (i + 1 < n && strcmp((&residues[i+1])->resName, "PRO") == 0) {
			next_res_is_Pro = TRUE;
		}
		else {
			next_res_is_Pro = FALSE;
		}
		predict_N_CA_C1_angle(&residues[i], next_res_is_Pro);
	}
	
}


/*
 * For debug only. Counts the number of residues with Rama scores 3, 2, 1, and 0, and appends
 * the counts to the file with name debugFilename.
 *
 * read_rama_tables() should be called before calling this function.
 */
void debugOutputRamaScores(ResData *residues, int numRes, char *debugFilename) {
	int i;
	double phi, psi, omega;
	int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
	int omega_score; // 1 is good, 0 is bad.
	int ramaCount[4] = {0, 0, 0, 0}; // ramaCount[i] is the number of scores with Rama score i
	int omegaCount[2] = {0, 0}; // omegaCount[i] is the number of scores with omega score i
	
	// count number of residues with each category of Rama score
	for (i = 0; i < numRes; i++) {
		phi = residues[i].phi;
		psi = residues[i].psi;
		omega = residues[i].omega;
		
		rama_score = lookupRamaScore(residues, i, numRes);
		
		if (!(omega >= 170.0 || omega <= -170.0 || (omega <= 10.0 && omega >= -10.0))) {
			omega_score = 0; // bad
		}
		else {
			omega_score = 1; // good
		}
		
		ramaCount[rama_score]++;
		omegaCount[omega_score]++;
	}
	
	FILE *scores_fp;
	scores_fp = fopen(debugFilename, "a");
	
	fprintf(scores_fp, "%d\t%d\t%d\t%d\t%d\t%d\n", ramaCount[3], ramaCount[2], ramaCount[1], ramaCount[0], omegaCount[1], omegaCount[0]);
	
	fclose(scores_fp);
	
}


/*
 * Repair the geometry of the O atoms in the template member variables in the given array of
 * ResData structs. (The reason is that Homodeller had a bug that misplaced O atoms).
 */
void repair_O_geometry(ResData *res, int numRes) {
	int i;
	Vec3D CA_C1;
	Vec3D N_C1;
	Vec3D ideal_O; // ideal location of the current CO atom
	//double C1_O_dist_orig;
	//double C1_O_dist_new;
	double O_O_diff;
	
	double ideal_C1_O_dist = 1.24;
		// Ideal distance between C1 and O atoms, based on mean distance in 1UBQ.
	double dist_tolerance = 0.34;
		// If original O atom is within this distance from ideal O atom location, do not change the location.
	
	for (i = 0; i < numRes; i++) {
		
		//C1_O_dist_orig = distance(res[i].template[C1], res[i].template[O]);
		
		difference(res[i].template[C1], res[i].template[CA], CA_C1);
		normalize(CA_C1, CA_C1);
		difference(res[i].template[C1], res[i].template[N], N_C1);
		normalize(N_C1, N_C1);
		summation(CA_C1, N_C1, ideal_O);
		normalize(ideal_O, ideal_O);
		scale(ideal_O, ideal_C1_O_dist);
		summation(res[i].template[C1], ideal_O, ideal_O);
		
		O_O_diff = distance(res[i].template[O], ideal_O);
		if (O_O_diff > dist_tolerance) {
			copyVector(ideal_O, res[i].template[O]);
		}
		
		// debug
		//C1_O_dist_new = distance(res[i].template[C1], res[i].template[O]);
		//fprintf(stderr, "%lf %lf %lf\n", C1_O_dist_orig, C1_O_dist_new, O_O_diff);
	}
	
}


/*
 * Scans all residues in the given peptide. If any of them have Ramachandran or omega angle
 * violations, modifySegmentAngles() is called to attempt to repair them.
 *
 * residues: Array of numRes ResData structs, one for each residue.
 * numRes: Number of residues in the peptide.
 * secStr: Optional. Null-terminated string of length numRes representing the best guess as
 *    to the secondary structure of the peptide. e.g. CCCCCHHHHHCCCCCCCBBBBBBBCCCC...
 */
void repairAngles(ResData *residues, int numRes, char *secStr) {
	int i, j;
	double phi, psi, omega;
	int rama_score; // 0, 1, 2, or 3. 3 is good, 0 is very bad (disallowed).
	int omega_score; // 1 is good, 0 is bad.
	int N_anchor_index; // N-terminal anchor for the current CCD segment.
	int C_anchor_index; // C-terminal anchor for the current CCD segment.
	int min_N_anchor_index;
	int min_C_anchor_index;
	//int rama_score_threshold;
	//int seg_length;
		// The length of the peptide segment to which we will apply the CCD algorithm,
		// including the N-terminal anchor and the C-terminal anchor.
	//int currentSegmentClosed;
	int modify_current_segment;
	
	// variable for debugging
	int numSegments = 0; // Number of segments we attempted to close.
	int segmentsClosed = 0; // Number of segments successfully closed.
	int segementsClosedWith2 = 0; // With rama_score_threshold of 2.
	int segementsClosedWith3 = 0; // With rama_score_threshold of 3.
	//clock_t start_clock = clock(); // For timing.
	
	//debugOutputRamaScores(residues, numRes, "debug_rama_scores_before_fix.txt");
	
	
	for (i = 0; i < numRes; i++) {
		phi = residues[i].phi;
		psi = residues[i].psi;
		omega = residues[i].omega;
		
		rama_score = lookupRamaScore(residues, i, numRes);
		
		// Omega stuff
		
		if (!(omega >= 170.0 || omega <= -170.0 || (omega <= 10.0 && omega >= -10.0))) {
			omega_score = 0; // bad
		}
		else {
			omega_score = 1; // good
		}
		
		//printf("%d %s: phi = %lf, psi = %lf, omega = %lf, score = %d, om = %d\n",
		//	i, residues[i].resName, residues[i].phi, residues[i].psi, residues[i].omega, rama_score, omega_score);
		
		if ( (omega_score == 0) || (rama_score < 3) ) {
			
			if (secStr == NULL || secStr[i] == 'C') { // Only try to repair angles for coil residues.
				
				if (i <= 1 && !(i == 1 && secStr != NULL && secStr[0] != 'C')) {
					// The problematic angle is near the N-terminus of the peptide, so we will simply adjust the angle
					// and let the N-terminus of the peptide be repositioned relative to the rest of the peptide.
					
					// debug
					//printf("Correcting angles for residue %d\n", i);
					
					setCorrectedAngles(residues, numRes, i, i, 3);
					buildPeptide(residues, numRes);
				}
				else if (i == numRes - 1) {
					// The problematic angle is at the C-terminus of the peptide, so we will simply adjust the angle
					// and let the C-terminus of the peptide be repositioned relative to the rest of the peptide.
					
					// debug
					//printf("Correcting angles for residue %d\n", i);
					
					setCorrectedAngles(residues, numRes, i, i, 3);
					rebuildPeptideSegment(residues, i, numRes - 1, FALSE);
				}
				else {
					// Use the CCD algorithm to try to fix the angles.
					// Note that the segment to be adjusted cannot include the first residue. (See "if (i <= 1)" above).
					
					/*
					// Try to fix the current problem angles by applying the CCD algorithm with different parameters.
					// If it fails with certain parameters, try again with other less stringent parameters. Start with
					// lower lengths for CCD segments (which are less likely to cause damage to the structure, but are
					// less likely to succeed) and progress to longer segment lengths, while trying to put the phi and
					// psi angles into the phi/psi core. Then, if there is still no success, try this again but only
					// trying to put phi and psi angles into the allowed region of the Ramachandran plots.
					//
					// Testing on one set of proteins seems to indicate that this variable-length CCD segment approach
					// results in less damage to the structure while resulting in nearly identical improvements in
					// phi, psi and omega angles, compared with a fixed-length segment approach.
					currentSegmentClosed = FALSE;
					for (rama_score_threshold = 3; rama_score_threshold >= 2; rama_score_threshold--) {
						
						if (rama_score == 2 && rama_score_threshold == 2) {
							// No use applying CCD in this situation.
							break;
						}
						
						for (seg_length = 6; seg_length <= 8; seg_length++) {
							
							if (i+seg_length-2 < numRes) {
								// Use a normal CCD segment positions.
								N_anchor_index = i-1;
								C_anchor_index = i+seg_length-2;
							}
							else {
								// Use a shorter CCD segment length with the C-terminal anchor being the last residue of the peptide.
								C_anchor_index = numRes - 1;
								N_anchor_index = numRes - seg_length;
							}
							
							// Try with a rama_score_threshold of 3. If that does not work, try again with a rama_score_threshold of 2.
							if (modifySegmentAngles(residues, numRes, N_anchor_index, C_anchor_index, rama_score_threshold)) {
								if (rama_score_threshold == 3) {
									segementsClosedWith3++;
								}
								else { // if (rama_score_threshold == 2)
									segementsClosedWith2++;
								}
								segmentsClosed++;
								
								// debug
								//printf("Segment closed with seg_length == %d\n", seg_length);
								
								currentSegmentClosed = TRUE;
								break;
							}
							
						}
						
						if (currentSegmentClosed) {
							break;
						}
					}
					*/
					
					// Older strategy that uses a fixed CCD segment length:
					
					if (i+CCD_SEG_LENGTH-2 < numRes) {
						// Use a normal CCD segment positions.
						N_anchor_index = i-1;
						C_anchor_index = i+CCD_SEG_LENGTH-2;
					}
					else {
						// Use a shorter CCD segment length with the C-terminal anchor being the last residue of the peptide.
						C_anchor_index = numRes - 1;
						N_anchor_index = numRes - CCD_SEG_LENGTH;
					}
					
					// Adjust anchor indices based on secondary structure information, if present. We allow only the anchor
					// residues to be non-loop (coil) residues.
					if (secStr != NULL) {
						// Scan to the right of residue i to find the minimum C-anchor index.
						min_C_anchor_index = C_anchor_index;
						for (j = i+1; j < C_anchor_index; j++) {
							if (secStr[j] != 'C') {
								min_C_anchor_index = j;
								break;
							}
						}
						
						// If the above scan forces us to move the C-anchor left, try adjusting the N-anchor left as well.
						if (min_C_anchor_index < C_anchor_index) {
							min_N_anchor_index = N_anchor_index;
							for (j = i-1; j > 0 && j > N_anchor_index - (C_anchor_index - min_C_anchor_index); j--) {
								if (secStr[j] != 'C') {
									min_N_anchor_index = j;
									break;
								}
							}
							
							// debug
							//fprintf(stderr, "Changing (%d %d) to (%d %d).\n",
							//	N_anchor_index, C_anchor_index, min_N_anchor_index, min_C_anchor_index);
							
							N_anchor_index = min_N_anchor_index;
							C_anchor_index = min_C_anchor_index;
						}
					}
					
					modify_current_segment = TRUE;
					
					// Make sure the segment does not span a gap in the structure.
					for (j = N_anchor_index+1; j <= C_anchor_index; j++) {
						if (residues[j].fixed_position == TRUE) {
							modify_current_segment = FALSE;
							
							// debug
							//fprintf(stderr, "Aborting from %d to %d: residue ID %d fixed.\n",
							//	N_anchor_index, C_anchor_index, residues[j].id);
							
							break;
						}
					}
					
					// Do not try segments that are too short.
					if (!(C_anchor_index - N_anchor_index + 1 >= MIN_CCD_SEG_LENGTH)) {
						modify_current_segment = FALSE;
					}
					
					if (modify_current_segment) {
						numSegments++;
						
						// Try with a rama_score_threshold of 3. If that does not work, try again with a rama_score_threshold of 2.
						if (modifySegmentAngles(residues, numRes, N_anchor_index, C_anchor_index, 3, MODIFY_SEG_MODE_REPAIR, NULL)) {
							segementsClosedWith3++;
							segmentsClosed++;
						}
						else if (modifySegmentAngles(residues, numRes, N_anchor_index, C_anchor_index, 2, MODIFY_SEG_MODE_REPAIR, NULL)) {
							segementsClosedWith2++;
							segmentsClosed++;
						}
					}
					
				}
			}
		}
		
	}
	
	// timing
	//clock_t end_clock = clock();
	
	// debug
	printf("Angle repair: Closed %d of %d segments (%d closed with Rama 2, %d closed with Rama 3).\n",
		segmentsClosed, numSegments, segementsClosedWith2, segementsClosedWith3);
	
	//fprintf(stderr,"\nfixAngles(): Time taken by the processor = %g sec.\n",
	//	(double)(end_clock - start_clock) / (double)CLOCKS_PER_SEC);
	
	//debugOutputRamaScores(residues, numRes, "debug_rama_scores_after_fix.txt");
}

/*
 * If run with 'mode' == MODIFY_SEG_MODE_REPAIR, this function applies the Cyclic
 * Coordinate Descent (CCD) algorithm to attempt to fix problematic phi, psi, and omega
 * angles in the peptide segment from indices N_anchor_index to C_anchor_index. The
 * algorithm tries to adjust all phi and psi angles so that they have a score of at
 * least rama_score_threshold. Returns TRUE if the CCD algorithm was successfully
 * applied. If the algorithm fails, the peptide segment is restored to its original
 * configuration and FALSE is returned.
 *
 * The operation in the other modes is similar, except that the step in which torsion
 * angles are repaired in replaced with different angle modifications. If 'mode' is
 * MODIFY_SEG_MODE_MUTATE, a random torsion angle at index N_anchor_index+1 is mutated.
 * If 'mode' is MODIFY_SEG_MODE_PREDITOR, the angles of the segment are set to those
 * predicted by Preditor. In all modes, segment closure is then attempted with CCD
 * as described above.
 *
 * mode: One of MODIFY_SEG_MODE_REPAIR, MODIFY_SEG_MODE_MUTATE, MODIFY_SEG_MODE_PREDITOR.
 * constraints: Array of ResConstraint structs, one for each residue.
 */
int modifySegmentAngles(ResData *residues, int numRes, int N_anchor_index, int C_anchor_index, int rama_score_threshold,
	int mode, ResConstraint *constraints) {
	
	int i, j;
	double **original_angles;
	double **corrected_angles;
		// Angles after calling setCorrectedAngles(), but before calling adjustSegmentCCD(). This information is used by
		// adjustSegmentCCD() only for the purpose of determining the Rama quality score of residues before calling adjustSegmentCCD().
	Vec3D target_CA, target_C1, target_N; // Original C-anchor atom coordinates.
	int segmentAdjustSucceeded;
	
	// variables used for debugging
	int debug = FALSE;
	Vec3D curr_CA, curr_C1, curr_N; // Movable C-terminal residue atom coordinates.
	double sumOfSquaredDistances;
		// Between current C-terminal residue atom coordinates and original C-anchor atom coordinates.
		// Same as S in the paper by Canutescu and Dunbrack.
	
	
	// debug
	if (debug) {printf("\nAdjusting residues with indices %d to %d:\n", N_anchor_index, C_anchor_index); }
	
	
	// Record the initial values for the dihedral angles in case the CCD algorithm does not work and
	// we need to restore the original values.
	
	// Allocate memory for original_angles.
	original_angles = malloc((C_anchor_index - N_anchor_index - 1) * sizeof(double *)); assert(original_angles != NULL);
	for (j = 0; j < (C_anchor_index - N_anchor_index - 1); j++) {
		original_angles[j] = malloc(3 * sizeof(double)); assert(original_angles[j] != NULL);
	}
	
	// Store original angles.
	for (i = N_anchor_index + 1; i < C_anchor_index; i++) {
		j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
		original_angles[j][0] = residues[i].phi;
		original_angles[j][1] = residues[i].psi;
		original_angles[j][2] = residues[i].omega;
	}
	
	// Record the initial positions of the C-anchor backbone atoms, which will be our target positions.
	
	copyVector(residues[C_anchor_index].atoms[CA], target_CA);
	copyVector(residues[C_anchor_index].atoms[C1], target_C1);
	copyVector(residues[C_anchor_index].atoms[N], target_N);
	
	// Modify the segment angles in some way. This will relocate the movable C-terminal residue of the segment.
	
	if (mode == MODIFY_SEG_MODE_REPAIR) {
		// Go through the residues of the current segment and repair any "bad" angles.
		setCorrectedAngles(residues, numRes, N_anchor_index + 1, C_anchor_index - 1, rama_score_threshold);
	}
	else if (mode == MODIFY_SEG_MODE_MUTATE) {
		// Mutate a random angle at residue N_anchor_index + 1, setting it to anywhere in its allowed range.
		
		if (N_anchor_index + 1 >= C_anchor_index) {
			fprintf(stderr, "ERROR: N_anchor_index + 1 >= C_anchor_index in modifySegmentAngles() in pepmake.c.\n"); exit(1);
		}
		
		double randomChoice; // What type of dihedral angle will we change?
		randomChoice = randDouble(3);
		i = N_anchor_index + 1;
		
		if (randomChoice < 1.0) {
			// Mutate phi angle.
			residues[i].phi = getRandomAngleInInterval(constraints[i].phi_min, constraints[i].phi_max);
		} else if (randomChoice < 2.0) {
			// Mutate psi angle.
			residues[i].psi = getRandomAngleInInterval(constraints[i].psi_min, constraints[i].psi_max);
		}
		else { // if (randomChoice < 3.0)
			// Mutate omega angle.
			residues[i].omega = getRandomAngleInInterval(constraints[i].omega_min, constraints[i].omega_max);
		}
		
	}
	else if (mode == MODIFY_SEG_MODE_PREDITOR) {
		// Change the torsion angles in the segment to those predicted by Preditor.
		
		for (i = N_anchor_index + 1; i < C_anchor_index; i++) {
			
			if (constraints[i].phi_preditor != PREDITOR_ANGLE_NA) {
				residues[i].phi = constraints[i].phi_preditor;
			}
			
			if (constraints[i].psi_preditor != PREDITOR_ANGLE_NA) {
				residues[i].psi = constraints[i].psi_preditor;
			}
			
			if (constraints[i].omega_preditor != PREDITOR_ANGLE_NA) {
				residues[i].omega = constraints[i].omega_preditor;
			}
		}
		
	}
	else {
		fprintf(stderr, "ERROR: Invalid value for 'mode' parameter in modifySegmentAngles() in pepmake.c.\n"); exit(1);
	}
	
	// Allocate memory for corrected_angles.
	corrected_angles = malloc((C_anchor_index - N_anchor_index - 1) * sizeof(double *)); assert(corrected_angles != NULL);
	for (j = 0; j < (C_anchor_index - N_anchor_index - 1); j++) {
		corrected_angles[j] = malloc(3 * sizeof(double)); assert(corrected_angles[j] != NULL);
	}
	
	// Store corrected angles.
	for (i = N_anchor_index + 1; i < C_anchor_index; i++) {
		j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
		corrected_angles[j][0] = residues[i].phi;
		corrected_angles[j][1] = residues[i].psi;
		corrected_angles[j][2] = residues[i].omega;
	}
	
	
	if (debug) { printf("------------ Beginning loop closure ------------\n"); }
	
	// Apply CCD algorithm to iteratively adjust segment angles to bring the movable
	// C-terminal residue to the same position as the original C-anchor.
	
	// Rebuild the backbone atoms of the current segment of the peptide using the repaired angles.
	rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, TRUE);
	
	
	segmentAdjustSucceeded = adjustSegmentCCD(residues, numRes, N_anchor_index, C_anchor_index, FALSE, corrected_angles, CCD_MODE_CLOSE_GAP, target_CA, target_C1, target_N,
		0.0, TRUE, 0, 0, 0);
	
	if (!segmentAdjustSucceeded) {
		
		if (debug) {
			printf("Restoring original angles yields C-terminal coordinates as follows:\n");
		}
		
		// Restore original angles.
		for (i = N_anchor_index + 1; i < C_anchor_index; i++) {
			j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
			residues[i].phi = original_angles[j][0];
			residues[i].psi = original_angles[j][1];
			residues[i].omega = original_angles[j][2];
		}
		
		// Rebuild the current segment of the peptide using the original angles.
		rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, FALSE);
		
		// debug
		if (debug) {
			copyVector(residues[C_anchor_index].atoms[CA], curr_CA);
			copyVector(residues[C_anchor_index].atoms[C1], curr_C1);
			copyVector(residues[C_anchor_index].atoms[N], curr_N);
			sumOfSquaredDistances = 
				distance(curr_CA, target_CA) * distance(curr_CA, target_CA)
				+ distance(curr_C1, target_C1) * distance(curr_C1, target_C1)
				+ distance(curr_N, target_N) * distance(curr_N, target_N);
			printf("curr: (%lf, %lf, %lf), (%lf, %lf, %lf), (%lf, %lf, %lf)\n", curr_CA[0], curr_CA[1], curr_CA[2],
				curr_C1[0], curr_C1[1], curr_C1[2], curr_N[0], curr_N[1], curr_N[2]);
			printf("Restored sumOfSquaredDistances = %lf\n", sumOfSquaredDistances);
		}
		
	}
	
	// free memory
	for (j = 0; j < (C_anchor_index - N_anchor_index - 1); j++) {
		free(original_angles[j]);
	}
	free(original_angles);
	
	for (j = 0; j < (C_anchor_index - N_anchor_index - 1); j++) {
		free(corrected_angles[j]);
	}
	free(corrected_angles);
	
	return segmentAdjustSucceeded;
}


/*
 * Attempts to close a gap in the peptide. gap_params.N_anchor_res_id and gap_params.C_anchor_res_id, or 
 * gap_params.N_anchor_res_index and gap_params.C_anchor_res_index, must specify the N-anchor and C-anchor for
 * use in the CCD algorithm. If the N-anchor is not specified, a few different N-anchor values will be tried;
 * otherwise, only the given N-anchor value will be tried. If given, the N-anchor should be several residues
 * before the gap. The C-anchor should be right next to the gap on the C-terminal side. e.g.
 *
 *      -----------------------------N--------    C--------------------
 *                                   |            |
 *                                N-anchor     C-anchor
 *
 * If the gap closure fails, the original angles and atom coordinates will be restored.
 * This function returns TRUE if the gap was closed, FALSE otherwise.
 */
static int closeGapCCD(ResData *residues, int totalRes, GAP_CLOSURE_PARAMS gap_params) {
	int i, j;
	int N_anchor_index = -1, C_anchor_index = -1; // -1 means not found.
	int min_N_anchor_index; // Minimum N-anchor index that we may try.
	int initial_N_anchor_index;
	double **original_angles;
	int original_max_atom_index_C_minus_2; // Original max_atom_index value for the residue 2 residues before the C-anchor.
	int original_max_atom_index_C_minus_1; // Original max_atom_index value for the residue 1 residue before the C-anchor.
	Vec3D original_template_C_minus_2[MAX_ATOMS]; // Original template data for the residue 2 residues before the C-anchor.
	Vec3D original_template_C_minus_1[MAX_ATOMS]; // Original template data for the residue 1 residue before the C-anchor.
	Vec3D original_atoms_C_anchor[MAX_ATOMS]; // Original atom data for the C-anchor.
	int original_C_anchor_fixed;
		// Original value of C-anchor's fixed_position variable, i.e. whether the original C-anchor had a fixed position
	Vec3D target_CA, target_C1, target_N; // Original C-anchor atom coordinates.
	int segmentAdjustSucceeded;
	
	int debug = FALSE;
	
	// Determine the N-anchor index and C-anchor index, which are NOT the same as corresponding residue IDs.
	
	if (gap_params.C_anchor_res_index != -1) {
		C_anchor_index = gap_params.C_anchor_res_index;
		
		if (gap_params.N_anchor_res_index == -1) {
			min_N_anchor_index = C_anchor_index - 9;
			if (min_N_anchor_index < 1) {
				min_N_anchor_index = 1;
			}
		}
		else {
			N_anchor_index = gap_params.N_anchor_res_index;
			min_N_anchor_index = N_anchor_index;
		}
	}
	else {
		
		for (i = 0; i < totalRes; i++) {
			if (residues[i].id == gap_params.C_anchor_res_id) {
				C_anchor_index = i;
				break;
			}
		}
		
		if (gap_params.N_anchor_res_id == RESIDUE_ID_NOT_AVAILABLE) {
			min_N_anchor_index = C_anchor_index - 9;
			if (min_N_anchor_index < 1) {
				min_N_anchor_index = 1;
			}
		}
		else {
			for (i = 0; i < totalRes; i++) {
				if (residues[i].id == gap_params.N_anchor_res_id) {
					N_anchor_index = i;
					break;
				}
			}
			
			min_N_anchor_index = N_anchor_index;
		}
	}
	
	
	// Test input.
	
	if (gap_params.C_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) {
		// User specified by residues IDs.
		
		if ((gap_params.N_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) && (N_anchor_index == -1)) {
			fprintf(stderr, "Given residue ID for gap closure not found in peptide. res_id_N = %d\n", gap_params.N_anchor_res_id);
			exit(1);
		}
		else if (C_anchor_index == -1) {
			fprintf(stderr, "Given residue ID for gap closure not found in peptide. res_id_C = %d\n", gap_params.C_anchor_res_id);
			exit(1);
		}
		else if ((gap_params.N_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) && !(C_anchor_index > N_anchor_index + 1)) {
			fprintf(stderr, "Gap closure: Given C-anchor must be closer to the C-terminal end than the N-anchor,\n");
			fprintf(stderr, "and there must be at least 1 residue between the two anchors.\n");
			exit(1);
		}
		else if ((gap_params.N_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) && (N_anchor_index == 0)) {
			fprintf(stderr, "Gap closure: N-anchor cannot be first residue in the peptide.\n");
			exit(1);
		}
		else if ((gap_params.N_anchor_res_id == RESIDUE_ID_NOT_AVAILABLE) && (C_anchor_index < 3)) {
			fprintf(stderr, "Gap closure: C-anchor cannot be one of the first 3 residues in the peptide.\n");
			exit(1);
		}
		
		if (gap_params.N_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) {
			for (i = N_anchor_index; i < C_anchor_index; i++) {
				if (residues[i].fixed_position == TRUE) {
					fprintf(stderr, "Gap closure: Gap does not immediately precede C-anchor.\n");
					exit(1);
				}
			}
		}
	}
	else {
		// User specified by residue indices.
		
		if ((gap_params.N_anchor_res_index != -1) && (N_anchor_index == -1)) {
			if (gap_params.N_anchor_res_index < 0) {
				fprintf(stderr, "Residue index for gap closure cannot be negative. res_index_N = %d\n", gap_params.N_anchor_res_index);
			}
			else {
				fprintf(stderr, "Residue index for gap closure is too large. res_index_N = %d\n", gap_params.N_anchor_res_index);
			}
			exit(1);
		}
		else if (C_anchor_index == -1) {
			if (gap_params.C_anchor_res_index < 0) {
				fprintf(stderr, "Residue index for gap closure cannot be negative. res_index_C = %d\n", gap_params.C_anchor_res_index);
			}
			else {
				fprintf(stderr, "Residue index for gap closure is too large. res_index_C = %d\n", gap_params.C_anchor_res_index);
			}
			exit(1);
		}
		else if ((gap_params.N_anchor_res_index != -1) && !(C_anchor_index > N_anchor_index + 1)) {
			fprintf(stderr, "Gap closure: Given C-anchor must be closer to the C-terminal end than the N-anchor,\n");
			fprintf(stderr, "and there must be at least 1 residue between the two anchors.\n");
			exit(1);
		}
		else if ((gap_params.N_anchor_res_index != -1) && (N_anchor_index == 0)) {
			fprintf(stderr, "Gap closure: N-anchor cannot be first residue in the peptide.\n");
			exit(1);
		}
		else if ((gap_params.N_anchor_res_index == -1) && (C_anchor_index < 3)) {
			fprintf(stderr, "Gap closure: C-anchor cannot be one of the first 3 residues in the peptide.\n");
			exit(1);
		}
		
		if (gap_params.N_anchor_res_index != -1) {
			for (i = N_anchor_index; i < C_anchor_index; i++) {
				if (residues[i].fixed_position == TRUE) {
					fprintf(stderr, "Gap closure: Gap does not immediately precede C-anchor.\n");
					exit(1);
				}
			}
		}
	}
	
	
	if (residues[C_anchor_index].atoms[CA][0] == ATOM_COORD_NA) {
		fprintf(stderr, "Gap closure: C-anchor is missing C-alpha atom.\n");
		exit(1);
	}
	else if (residues[C_anchor_index].atoms[C1][0] == ATOM_COORD_NA) {
		fprintf(stderr, "Gap closure: C-anchor is missing carbonyl C atom.\n");
		exit(1);
	}
	else if (residues[C_anchor_index].atoms[N][0] == ATOM_COORD_NA) {
		fprintf(stderr, "Gap closure: Residue %3s %d is missing N atom.\n",
			residues[C_anchor_index+1].resName, residues[C_anchor_index+1].id);
		exit(1);
	}
	
	// debug
	if (debug) {
		printf("gap_params.N_anchor_res_id = %d, gap_params.C_anchor_res_id = %d, N_anchor_index = %d, C_anchor_index = %d\n",
			gap_params.N_anchor_res_id, gap_params.C_anchor_res_id, N_anchor_index, C_anchor_index);
	}
	
	// Record the initial values for the dihedral angles in case the CCD algorithm does not work and
	// we need to restore the original values.
	
	// Allocate memory for original_angles.
	original_angles = malloc((C_anchor_index - min_N_anchor_index - 1) * sizeof(double *)); assert(original_angles != NULL);
	for (j = 0; j < (C_anchor_index - min_N_anchor_index - 1); j++) {
		original_angles[j] = malloc(3 * sizeof(double)); assert(original_angles[j] != NULL);
	}
	
	// Store original angles.
	for (i = min_N_anchor_index + 1; i < C_anchor_index; i++) {
		j = i - min_N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
		original_angles[j][0] = residues[i].phi;
		original_angles[j][1] = residues[i].psi;
		original_angles[j][2] = residues[i].omega;
	}
	
	// Store original max_atom_index values and template data for the 2 residues before the C-anchor.
	
	original_max_atom_index_C_minus_2 = residues[C_anchor_index - 2].max_atom_index;
	original_max_atom_index_C_minus_1 = residues[C_anchor_index - 1].max_atom_index;
	for (i = 0; i < MAX_ATOMS; i++) {
		copyVector(residues[C_anchor_index - 2].template[i], original_template_C_minus_2[i]);
		copyVector(residues[C_anchor_index - 1].template[i], original_template_C_minus_1[i]);
	}
	
	// Record the initial positions of the C-anchor backbone atoms, which will be our target positions.
	
	copyVector(residues[C_anchor_index].atoms[CA], target_CA);
	copyVector(residues[C_anchor_index].atoms[C1], target_C1);
	copyVector(residues[C_anchor_index].atoms[N], target_N);
	
	// Record the original positions of the C-anchor atoms.
	for (i = 0; i < MAX_ATOMS; i++) {
		copyVector(residues[C_anchor_index ].atoms[i], original_atoms_C_anchor[i]);
	}
	
	// Record original fixed_position value for the C-anchor, and set a new one for current use.
	
	original_C_anchor_fixed = residues[C_anchor_index].fixed_position;
	residues[C_anchor_index].fixed_position = FALSE;
	
	// Build the C-anchor onto the N-terminal side of the gap.
	
	readTemplate(&residues[C_anchor_index - 2], ignoreDBBackbone, FALSE, FALSE);
	readTemplate(&residues[C_anchor_index - 1], ignoreDBBackbone, FALSE, FALSE);
		// Assume that the 2 residues before the gap have bad coordinate information, and load new info from the template database.
	rebuildPeptideSegment(residues, C_anchor_index - 2, C_anchor_index, FALSE);
	
	// Try to close the gap.
	
	segmentAdjustSucceeded = FALSE;
	if (N_anchor_index != -1) {
		segmentAdjustSucceeded = adjustSegmentCCD(residues, totalRes, N_anchor_index, C_anchor_index, TRUE, original_angles, CCD_MODE_CLOSE_GAP, target_CA, target_C1, target_N,
			gap_params.max_threshold, FALSE, 0, 0, 0);
	}
	else {
		// No N-anchor was given, so try a few different N-anchor values.
		initial_N_anchor_index = C_anchor_index - 5;
		if (initial_N_anchor_index < 1) {
			initial_N_anchor_index = 1;
		}
		for (N_anchor_index = initial_N_anchor_index; N_anchor_index >= min_N_anchor_index; N_anchor_index--) {
			
			// Make sure the segment does not span any fixed-position residues.
			int skip = FALSE;
			for (i = N_anchor_index; i < C_anchor_index; i++) {
				if (residues[i].fixed_position == TRUE) {
					if (debug) {
						printf("Residue %3s %d (index %d) has a fixed position.\n", residues[i].resName, residues[i].id, i);
					}
					skip = TRUE;
					break;
				}
			}
			if (skip) {
				if (debug) {
					printf("Skipping with N_anchor_index = %d\n", N_anchor_index);
				}
				// Try a different N-anchor index.
				continue;
			}
			
			segmentAdjustSucceeded = adjustSegmentCCD(residues, totalRes, N_anchor_index, C_anchor_index, TRUE, original_angles, CCD_MODE_CLOSE_GAP, target_CA, target_C1, target_N,
				gap_params.max_threshold, FALSE, 0, 0, 0);
			if (segmentAdjustSucceeded) {
				if (debug) { printf("Gap closed with N_anchor_index = %d\n", N_anchor_index); }
				break;
			}
			else {
				if (debug) {
					printf("Failed gap closure with N_anchor_index = %d\n", N_anchor_index);
				}
				
				// Restore original angles.
				for (i = min_N_anchor_index + 1; i < C_anchor_index; i++) {
					j = i - min_N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
					residues[i].phi = original_angles[j][0];
					residues[i].psi = original_angles[j][1];
					residues[i].omega = original_angles[j][2];
				}
				
				// Rebuild the current segment of the peptide using the original angles.
				rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, FALSE);
					// We use N_anchor_index instead of min_N_anchor_index because we know we did not modify the peptide below N_anchor_index.
			}
		}
	}
	
	
	if (!segmentAdjustSucceeded) {
		
		// Restore original max_atom_index values and template data for the 2 residues before the C-anchor.
		
		residues[C_anchor_index - 2].max_atom_index = original_max_atom_index_C_minus_2;
		residues[C_anchor_index - 1].max_atom_index = original_max_atom_index_C_minus_1;
		for (i = 0; i < MAX_ATOMS; i++) {
			copyVector(original_template_C_minus_2[i], residues[C_anchor_index - 2].template[i]);
			copyVector(original_template_C_minus_1[i], residues[C_anchor_index - 1].template[i]);
		}
		
		// Restore original C-anchor atom coordinates.
		
		for (i = 0; i < MAX_ATOMS; i++) {
			copyVector(original_atoms_C_anchor[i], residues[C_anchor_index].atoms[i]);
		}
		
		// Restore fixed_position value for the C-anchor.
		residues[C_anchor_index].fixed_position = original_C_anchor_fixed;
		
		// Restore original angles.
		for (i = min_N_anchor_index + 1; i < C_anchor_index; i++) {
			j = i - min_N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
			residues[i].phi = original_angles[j][0];
			residues[i].psi = original_angles[j][1];
			residues[i].omega = original_angles[j][2];
		}
		
		// Rebuild the current segment of the peptide using the original angles.
		int startIndex = N_anchor_index;
			// We use N_anchor_index instead of min_N_anchor_index because we know we did not modify the peptide below N_anchor_index.
			// That is, unless N_anchor_index < min_N_anchor_index...
		if (N_anchor_index < min_N_anchor_index) {
			startIndex = min_N_anchor_index;
		}
		rebuildPeptideSegment(residues, startIndex, C_anchor_index, FALSE);
	}
	
	// free memory
	for (j = 0; j < (C_anchor_index - min_N_anchor_index - 1); j++) {
		free(original_angles[j]);
	}
	free(original_angles);
	
	return segmentAdjustSucceeded;
}


/*
 * Tries to align the hydrogen bonds of the given peptide, based on the intended hydrogen bonds specified by
 * the user. This function only aligns Beta sheets or other hydrogen bonds that do not involve are pair of
 * residues that are close to one another along the sequence; Alpha helices are not aligned. The data for
 * the intended hydrogen bonds should already be set in the hbonds_intended_acc and hbonds_intended_don
 * member variables of each ResData struct in the peptide.
 *
 * This function scans along the peptide to identify stretches of the peptide with residues that have
 * intended hydrogen bonds. Segments are allowed to contain "gaps" of MAX_GAP_ALONG_BETA_STRAND residues
 * without intended hydrogen bonds. For example, for 1xqq, suppose we specify intended hydrogen bonds as
 * follows (acceptor residues on the left, donor residues on the right):
 *
 *    MET 1          VAL 17
 *    GLN 2          GLU 64
 *    ILE 3          LEU 15
 *    PHE 4          LEU 67
 *    VAL 5          ILE 13
 *    LYS 6          LEU 69
 *    THR 7          GLY 10
 *    LYS 11         THR 7 
 *    ILE 13         VAL 5 
 *    LEU 15         ILE 3 
 *    SER 65         PHE 4
 *    LEU 67         LYS 6
 *
 * Let a dot (.) represent a residue that is either an acceptor or a donor in an intended hydrogen bond. Then
 * we get:
 *
 *    MQIFVKTLTGKTITLEVEPSDTIENVKAKIQDKEGIPPDQQRLIFAGKQLEDGRTLSDYNIQKESTLHLVLRLRGG
 *    .......  .. . . .                                              .. . .
 * 
 *    -------  --------                                              ------
 *
 * If MAX_GAP_ALONG_BETA_STRAND is 1, then the algorithm will identify 3 segments, denoted here by dashes. This
 * function will then try to align the hydrogen bonds for each of these 3 segments.
 *
 * Different strategies are used for segments that are near the N-terminus of the protein, near the C-terminus
 * of the protein, and in the rest of the protein. For a segment near the N-terminus (such as the first one in
 * the example above), the algorithm designates a flexible region of up to HBOND_ALIGN_FLEXIBLE_REGION_LENGTH
 * residues to the right of segment. However, this fexible region cannot contain any intended hydrogen bonds,
 * and may therefore be shorter. The N-terminal anchor is designated as the last residue in the hydrogen bonds
 * segment. Residues after the flexible region are designated as part of the rigid segment, and the C-terminal
 * anchor is the last residue in the peptide. For first segment in the example given above:
 * 
 *    MQIFVKTLTGKTITLEVEPSDTIENVKAKIQDKEGIPPDQQRLIFAGKQLEDGRTLSDYNIQKESTLHLVLRLRGG
 *    .......  .. . . .                                              .. . .
 * 
 *    -------~~rrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr
 *          N                                                                    C
 *
 * where - indicates the hydrogen bond segment
 *       N indicates the N-terminal anchor
 *       ~ indicates the flexible region
 *       r indicates the flexible region
 *       C indicates the C-terminal anchor
 *
 * The alignment algorithm called by this function will keep the hydrogen bond segment in place, and apply CCD to
 * change the torsion angles within the flexible region such that hydrogen bonds between the rigid segment and the
 * hydrogen bonds segment are aligned. (This equivalent to keeping the rigid segment in place and moving the
 * hydrogen bonds segment). It is possible that there will be a small number residues on the very N-terminal tip
 * of the peptide that do not have intended hydrogen bonds, but they will remain in place and the algorithm will
 * do essentially the same thing.
 *
 * When a hydrogen bonds segment is in the middle of the peptide, a different strategy is used. The algorithm
 * placed two flexible regions of up to HBOND_ALIGN_FLEXIBLE_REGION_LENGTH residues to the left and right of the
 * hydrogen bonds segment. The N-terminal anchor will be the residue to the left of the 1st flexible region, and
 * and C-terminal anchor will be the residue to the right of the 2nd flexible region. A 2nd N-terminal anchor will
 * be designated as the last residue in the hydrogen bonds segment. For the second segment in the example above, if
 * we use a HBOND_ALIGN_FLEXIBLE_REGION_LENGTH value of 5:
 *
 *    MQIFVKTLTGKTITLEVEPSDTIENVKAKIQDKEGIPPDQQRLIFAGKQLEDGRTLSDYNIQKESTLHLVLRLRGG
 *    .......  .. . . .                                              .. . .
 * 
 *           ~~--------~~~~~
 *             rrrrrrrrrrrrrr
 *          N         2     C
 *
 *
 * where "2" represents the 2nd N-terminal anchor.
 *
 * The 1st flexible region is shorter because of the intended hydrogen bonds of the first hydrogen bonds segment.
 *
 * The alignment then proceeeds in two stages. In the first stage, the 1st N-terminal anchor is used, and the first
 * flexible region's torsion angles are adjusted with CCD to move the rigid segment's intended hydrogen bonds with
 * their partners in the rest of the peptide. This results in the dislocation of the rigid segment, and a gap formed
 * between the C-terminal anchor and the following residues. In the second stage, the 2nd N-terminal anchor is used,
 * and the 2nd flexible region is adjusted with CCD so that the C-terminal anchor residue is moved back to the
 * original C-terminal anchor position to close the gap.
 *
 * When a hydrogen bonds segment is near the C-terminal end of the peptide, the alignment strategy proceeds as in the
 * case when the hydrogen bonds segment is in the middle of the peptide, but only for the first stage. In this stage
 * the rigid segment proceeed to the C-terminal end of the peptide, and the C-terminal anchor is the last residue of
 * the peptide. Thus this C-terminal segment will be dislocated relative the the rest of the peptide.
 *
 * When aligning a hydrogen bonds segment, the alignment may not succeed in aligning the hydrogen bonds to below the
 * sum-of-squared-distances threshold. But we will accept whatever the alignment managed to produce. If a gap closure
 * stage fails, however, we will restore the entire hydrogen bonds segment, including both flexible regions, such that
 * they are restored to what their positions were before any hydrogen bonds alignment was attempted.
 */
void alignBetaSheets1(ResData *residues, int totalRes) {
	int i, j;
	int resIndex;
	int N_anchor_index = -1, C_anchor_index = -1;
	int second_N_anchor_index = -1;
		// The N-anchor index for the second segment (not used if we're adjusting a segment near one of the ends of the peptide).
	int new_segment_found = FALSE; // Whether we have found a new segment for which we will align the hydrogen bonds.
	int segment_is_near_N_terminus = FALSE;
	int segment_is_near_C_terminus = FALSE;
	int begin_rigid_seg_index = -1;
		// The index of the residue that will be the beginning of the rigid segment. -1 means
		// that we have not yet found a residue with an 'intended' hydrogen bond to start the
		// rigid segment.
	int begin_hbonds_index, end_hbonds_index;
		// The first and last residues with hydrogen bonds that we want to optimize in the current segment.
		// Normally this region is identical to the rigid region, but the exception is when the segment is near
		// the N-terminus, in which case these variable will indicate the first several residues, while the
		// rigid region will be all of the peptide excluding some residues near the N-terminus.
	int non_bonded_res_count = 0;
		// The number of residues counted (since the counter was reset) that do not have intended
		// hydrogen bonds.
	double **original_angles;
	Vec3D dummy; // Dummy variable; not used.
	Vec3D target_CA, target_C1, target_N; // C-anchor target position for if we need to close a gap created by aligning hydrogen bonds.
	int segmentAdjustSucceeded;
	int try_segment_adjustment; // Boolean.
	
	int debug = FALSE;
	
	// Record the initial values for the dihedral angles in case the CCD algorithm does not work and
	// we need to restore the original values.
	
	// Allocate memory for original_angles. We only need to record the angles for the (1 or 2) flexible
	// regions.
	original_angles = malloc((HBOND_ALIGN_FLEXIBLE_REGION_LENGTH*2) * sizeof(double *)); assert(original_angles != NULL);
	for (j = 0; j < (HBOND_ALIGN_FLEXIBLE_REGION_LENGTH*2); j++) {
		original_angles[j] = malloc(3 * sizeof(double)); assert(original_angles[j] != NULL);
	}
	
	// Search for segments for hydrogen bond alignment, and align the hydrogen bonds.
	
	for (resIndex = 0; resIndex < totalRes; resIndex++) {
		
		if (residueHasIntendedHbond(residues, resIndex)) {
			if (!new_segment_found) {
				new_segment_found = TRUE;
				
				//if (non_bonded_res_count <= HBOND_ALIGN_FLEXIBLE_REGION_LENGTH) {
				if (resIndex <= HBOND_ALIGN_FLEXIBLE_REGION_LENGTH) {
					// The new segment is near the N-terminus of the peptide.
					segment_is_near_N_terminus = TRUE;
				}
				else {
					// The new segment is in the middle or near the end of the peptide.
					if (non_bonded_res_count < HBOND_ALIGN_FLEXIBLE_REGION_LENGTH) {
						N_anchor_index = resIndex - non_bonded_res_count - 1;
					}
					else {
						N_anchor_index = resIndex - HBOND_ALIGN_FLEXIBLE_REGION_LENGTH - 1;
					}
					
					if (N_anchor_index < 1) { // N_anchor_index cannot be <= 0
						N_anchor_index = 1;
					}
					
					begin_rigid_seg_index = resIndex;
				}
			}
			
			non_bonded_res_count = 0;
		}
		else {
			non_bonded_res_count++;
		}
		
		if (new_segment_found && ((non_bonded_res_count > MAX_GAP_ALONG_BETA_STRAND) || (resIndex == totalRes-1)) ) {
			
			if (segment_is_near_N_terminus) {
				N_anchor_index = resIndex - non_bonded_res_count; // Last residue seen with intended hbond.
				
				// Scan ahead to see where intended hydrogen bonds are. We will try to achieve a begin_rigid_seg_index
				// such that the flexible region will be of length HBOND_ALIGN_FLEXIBLE_REGION_LENGTH, but if we find
				// any intended hydrogen bonds before we reach that point, we will use a begin_rigid_seg_index that
				// involves a shorter flexible segment.
				begin_rigid_seg_index = N_anchor_index + HBOND_ALIGN_FLEXIBLE_REGION_LENGTH + 1;
				for (i = 1; i < HBOND_ALIGN_FLEXIBLE_REGION_LENGTH - non_bonded_res_count; i++) {
					if (residueHasIntendedHbond(residues, resIndex + i)) {
						begin_rigid_seg_index = resIndex + i;
						break;
					}
				}
				
				C_anchor_index = totalRes - 1;
				
				begin_hbonds_index = 0;
				end_hbonds_index = N_anchor_index;
			}
			else {
				
				C_anchor_index = resIndex - non_bonded_res_count + HBOND_ALIGN_FLEXIBLE_REGION_LENGTH + 1;
				
				if (C_anchor_index > totalRes - 1) {
					segment_is_near_C_terminus = TRUE;
					C_anchor_index = totalRes - 1;
				}
				else {
					
					for (i = 1; i < HBOND_ALIGN_FLEXIBLE_REGION_LENGTH - non_bonded_res_count; i++) {
						if (residueHasIntendedHbond(residues, resIndex + i)) {
							C_anchor_index = resIndex + i;
							break;
						}
					}
				}
				
				begin_hbonds_index = begin_rigid_seg_index;
				end_hbonds_index = C_anchor_index;
			}
			
			// Now we know the N_anchor_index, C_anchor_index, and begin_rigid_seg_index for the current segment.
			
			// Test to make sure the segment does not span a gap in the structure. This test was added Apr 1,
			// 2008 by David Arndt, but was not carefully tested.
			try_segment_adjustment = TRUE;
			for  (i = N_anchor_index; i <= C_anchor_index; i++) {
				if (residues[i].fixed_position == TRUE) {
					try_segment_adjustment = FALSE;
					break;
				}
			}
			
			if (try_segment_adjustment) {
				
				if (debug) {
					printf("\n");
					printf("-------------------------- New segment --------------------------\n");
					printf("Segment: N_anchor_index = %d, C_anchor_index = %d, begin_rigid_seg_index = %d\n",
						N_anchor_index, C_anchor_index, begin_rigid_seg_index);
					printf("Segment: N-anchor = %s %d, C-anchor = %s %d, begin_rigid_seg = %s %d\n",
						residues[N_anchor_index].resName, residues[N_anchor_index].id,
						residues[C_anchor_index].resName, residues[C_anchor_index].id,
						residues[begin_rigid_seg_index].resName, residues[begin_rigid_seg_index].id);
				}
				
				
				if (!segment_is_near_C_terminus && !segment_is_near_N_terminus) {
					second_N_anchor_index = resIndex - non_bonded_res_count; // Last residue seen with intended hbond.
					
					copyVector(residues[C_anchor_index].atoms[CA], target_CA);
					copyVector(residues[C_anchor_index].atoms[C1], target_C1);
					copyVector(residues[C_anchor_index].atoms[N], target_N);
				}
				
				// Store original angles.
				// The first HBOND_ALIGN_FLEXIBLE_REGION_LENGTH rows are reserved for the first flexible region,
				// and the remaining HBOND_ALIGN_FLEXIBLE_REGION_LENGTH rows are reserved for the second flexible
				// region (which only applies when the segment is not near one of the ends of the peptide.
				for (i = N_anchor_index + 1; i < begin_rigid_seg_index; i++) {
					j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
					original_angles[j][0] = residues[i].phi;
					original_angles[j][1] = residues[i].psi;
					original_angles[j][2] = residues[i].omega;
				}
				if (!segment_is_near_C_terminus && !segment_is_near_N_terminus) {
					for (i = second_N_anchor_index + 1; i < C_anchor_index; i++) {
						j = i - (second_N_anchor_index + 1) + HBOND_ALIGN_FLEXIBLE_REGION_LENGTH;
							// j here is the index into original_angles, starting from HBOND_ALIGN_FLEXIBLE_REGION_LENGTH.
						original_angles[j][0] = residues[i].phi;
						original_angles[j][1] = residues[i].psi;
						original_angles[j][2] = residues[i].omega;
					}
				}
				
				
				// Adjust the segment to align the hydrogen bonds.
				
				segmentAdjustSucceeded = adjustSegmentCCD(residues, totalRes, N_anchor_index, C_anchor_index, FALSE, original_angles, CCD_MODE_ALIGN_HBONDS,
						dummy, dummy, dummy, 0.0, TRUE, begin_rigid_seg_index, begin_hbonds_index, end_hbonds_index);
				
				if (!segmentAdjustSucceeded) {
					if (debug) {
						printf("Hydrogen bonds not aligned below threshold, but we will use what was aligned anyway.\n");
					}
				}
				
				// If needed, close the gap created by aligning the hydrogen bonds.
				
				if (!segment_is_near_C_terminus && !segment_is_near_N_terminus) {
					// Adjusting the segment to align hydrogen bonds has created a gap in the peptide, so we need to
					// adjust the segment following the first segment to close the gap.
					
					if (debug) {
						printf("\n");
						printf("Segment: second_N_anchor_index = %d, C_anchor_index = %d\n",
							second_N_anchor_index, C_anchor_index);
						printf("Segment: second N-anchor = %s %d, C-anchor = %s %d\n",
							residues[second_N_anchor_index].resName, residues[second_N_anchor_index].id,
							residues[C_anchor_index].resName, residues[C_anchor_index].id);
					}
					
					// Adjust segment to close the gap.
					
					segmentAdjustSucceeded = adjustSegmentCCD(residues, totalRes, second_N_anchor_index, C_anchor_index, TRUE, &original_angles[HBOND_ALIGN_FLEXIBLE_REGION_LENGTH], CCD_MODE_CLOSE_GAP,
						target_CA, target_C1, target_N, 0.0, TRUE, 0, 0, 0);
					
					if (!segmentAdjustSucceeded) {
						if (debug) {
							printf("** Gap closure failed! ** Segment will be restored to state before hydrogen bond alignment.\n");
						}
					}
					
					if (!segmentAdjustSucceeded) {
						// We failed to close the gap, so we will all of the original angles, undoing the changes we made in attempting
						// to close the gap as well as in aligning the hydrogen bonds.
						
						// Restore original angles (including for the first segment where we attempted to align the hydrogen bonds).
						for (i = N_anchor_index + 1; i < begin_rigid_seg_index; i++) {
							j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
							residues[i].phi = original_angles[j][0];
							residues[i].psi = original_angles[j][1];
							residues[i].omega = original_angles[j][2];
						}
						for (i = second_N_anchor_index + 1; i < C_anchor_index; i++) {
							j = i - (second_N_anchor_index + 1) + HBOND_ALIGN_FLEXIBLE_REGION_LENGTH;
								// j here is the index into original_angles, starting from HBOND_ALIGN_FLEXIBLE_REGION_LENGTH.
							residues[i].phi = original_angles[j][0];
							residues[i].psi = original_angles[j][1];
							residues[i].omega = original_angles[j][2];
						}
						
						// Rebuild the current segment of the peptide using the original angles.
						rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, FALSE);
						
					}
					
				}
			}
			
			new_segment_found = FALSE;
			segment_is_near_N_terminus = FALSE;
			segment_is_near_C_terminus = FALSE;
		}
		
	}
	
	// free memory
	for (j = 0; j < (HBOND_ALIGN_FLEXIBLE_REGION_LENGTH*2); j++) {
		free(original_angles[j]);
	}
	free(original_angles);
	
}



/*
 * Returns TRUE if the residue with index resNum has any 'intended' hydrogen bonds, FALSE otherwise.
 */
static int residueHasIntendedHbond(ResData *residues, int resNum) {
		int j;
		for (j = 0; j < 2; j++) {
			if (residues[resNum].hbonds_intended_acc[j].partner != -1) {
				// Residue resNum contains an 'intended' hydrogen bond that the user wants formed or optimized.
				return TRUE;
			}
			if (residues[resNum].hbonds_intended_don[j].partner != -1) {
				// Residue resNum contains an 'intended' hydrogen bond that the user wants formed or optimized.
				return TRUE;
			}
		}
		return FALSE;
}


/*
 * Applies the Cyclic Coordinate Descent (CCD) algorithm to relocate a segment of the
 * peptide from index N_anchor_index to C_anchor_index. If the algorithm fails to adjust
 * the segment to an acceptable position, the function returns false, but does not try
 * to restore the peptide segment to its original position.
 *
 * The 'mode' variable can be one of CCD_MODE_CLOSE_GAP or CCD_MODE_ALIGN_HBONDS. If 'mode'
 * is CCD_MODE_CLOSE_GAP, then this function will try to close the gap in the peptide by
 * repositioning the C-anchor of the given peptide segment so that it aligns with the target
 * coordinates given by the variables target_CA, target_C1, and target_N. The requirements
 * for alignment are very stringent in this case. The parameters begin_rigid_seg_index,
 * begin_hbonds_index, and end_hbonds_index are ignored (although begin_rigid_seg_index is
 * given the same value as C_anchor_index and is used).
 * 
 * If 'mode' is CCD_MODE_ALIGN_HBONDS, then this function will try to reposition the segment
 * so that its 'intended' hydrogen bonds, as indicated in the hbonds_intended_acc and
 * hbonds_intended_don member variables of the ResData structs of the segment, are formed
 * into proper hydrogen bonds. If the hydrogen bonds are already formed, the function will
 * adjust the segment to optimize the positions of the hydrogen bonds. Because hydrogen bond
 * geometry is flexible, the alignment requirements will not be very stringent in this case.
 * The parameters target_CA, target_C1, and target_N are ignored.
 *
 * original_angles is an array of the original torsion angles of the segment before any CCD
 * was applied. This information is used here only to determine the original Rama quality
 * scores of residues in the segment.
 *
 * If 'mode' is CCD_MODE_CLOSE_GAP, maxSumOfSquaredDistancesThreshold may be used to relax
 * how stringently we require the movable C-anchor to overlap with the original C-anchor.
 * The algorithm will first try to close the gap using the default threshold, but if it
 * cannot, the threshold maxSumOfSquaredDistancesThreshold will be considered acceptable.
 * If maxSumOfSquaredDistancesThreshold is <= 0.0, it will be ignored.
 *
 * If 'mode' is CCD_MODE_CLOSE_GAP and the gap is successfully closed and rebuild_after_gap
 * is TRUE, the peptide from the C-anchor to the C-terminus will be rebuilt. This may be
 * desired since the C-anchor may be slightly misaligned after gap closure, resulting in
 * incorrect geometry at the C-anchor. If rebuild_after_gap is FALSE, this rebuilding will
 * not be done.
 *
 * This implementation of the CCD algorithm is similar to and based upon the CCD algorithm
 * explained in:
 * Canutescu AA, Dunbrack RL Jr. Cyclic coordinate descent: A robotics algorithm for
 * protein loop closure. Protein Sci. 2003 May;12(5):963-72.
 */
static int adjustSegmentCCD(ResData *residues, int numRes, int N_anchor_index, int C_anchor_index, int allowAnglesToBecomeWorse, double **original_angles,
		int mode, Vec3D target_CA, Vec3D target_C1, Vec3D target_N, double maxSumOfSquaredDistancesThreshold, int rebuild_after_gap,
		int begin_rigid_seg_index, int begin_hbonds_index, int end_hbonds_index) {
	
	int i, j, whichAngleType;
	int num_passes; // Number of passes of the CCD algorithm through the whole segment.
	int closureFailed = 0; // Whether the CCD algorithm failed in closing the segment.
	int applyRotation;
		// Boolean. Whether to apply the rotation for the current angle (set to FALSE if rotation is by
		// 0.0 degrees, to avoid useless computation).
	double phi, psi, omega;
	double desired_phi, desired_psi, desired_omega;
		// Ideal angles for CCD rotation, unadjusted based on the constraints of peptide geometry.
	int maxRamaQualityDecrease = 0;
		// The maximum amount by which the Rama score for a new angle is allowed to decrease compared to the original Rama score.
		// See comments for getClosestAllowedAngle() in rama.c for a fuller explanation.
	double sumOfSquaredDistances;
		// Between current C-terminal residue atom coordinates and original C-anchor atom coordinates.
		// Same as S in the paper by Canutescu and Dunbrack.
	double sumOfSquaredDistancesThreshold;
		// When sumOfSquaredDistances is below or equal to this value, the alignment points are considered
		// to be sufficiently well aligned with their targets. 
	Vec3D rotationAxis;
	Vec3D basePoint;
		// Location of the N-terminal atom of the bond along which we are currently rotating.
	Vec3D temp;
	double tempDouble;
	double tan_alpha_nominator, tan_alpha_denominator;
	double tan_alpha;
	double alpha_rad; // The alpha angle (in radians) derived from tan_alpha;
	double alpha;
		// Alpha angle in degrees. This is the ideal amount by which we should change the current angle (phi, psi, or omega),
		// before taking into account constraints on those angles by peptide geometry.
	double second_deriv; // Second derivate of S with respect to theta.
	CCD_alignment_point *align_pts; // All of the alignment points for CCD.
	CCD_alignment_point *curr_align_pt;
	int total_align_pts;
	int point_num;
	int partner;
	
	
	int convergence_check_period = 10; // Number of iterations between checks to see how quickly sumOfSquaredDistances is decreasing.
	double olderSumOfSquaredDistances = -1.0; // The sumOfSquaredDistances from several (convergence_check_period) iterations ago.
	double slope, intercept; // Used to test how quickly sumOfSquaredDistances is decreasing.
	int estimated_passes; // Rough estimate of how many passes will have elapsed when sumOfSquaredDistances achieves the threshold. 
	
	
	// variables used for debugging
	int debug = FALSE;
	
	
	if ( (mode == CCD_MODE_CLOSE_GAP) && (maxSumOfSquaredDistancesThreshold > 0.0) &&
			(maxSumOfSquaredDistancesThreshold < CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD) ) {
		fprintf(stderr,
			"Illegal argument: maxSumOfSquaredDistancesThreshold < CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD\n");
		exit(1);
	}
	
	
	// Set up alignment points.
	
	if (mode == CCD_MODE_CLOSE_GAP) {
		align_pts = malloc(3 * sizeof(CCD_alignment_point)); assert(align_pts != NULL);
		
		for (point_num = 0; point_num < 3; point_num++) {
			align_pts[point_num].resIndex = C_anchor_index;
			align_pts[point_num].point_type = CCD_POINT_TYPE_C_ANCHOR;
		}
		
		copyVector(residues[C_anchor_index].atoms[CA], align_pts[0].current);
		copyVector(residues[C_anchor_index].atoms[C1], align_pts[1].current);
		copyVector(residues[C_anchor_index].atoms[N], align_pts[2].current);
		
		copyVector(target_CA, align_pts[0].target);
		copyVector(target_C1, align_pts[1].target);
		copyVector(target_N, align_pts[2].target);
		
		total_align_pts = 3;
		
		sumOfSquaredDistancesThreshold = CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD;
		begin_rigid_seg_index = C_anchor_index;
	}
	else if (mode == CCD_MODE_ALIGN_HBONDS) {
		
		// Count the number of intended hydrogen bonds from begin_hbonds_index to end_hbonds_index, inclusive.
		// Twice this number will be the number of alignment points.
		
		total_align_pts = 0;
		
		for (i = begin_hbonds_index; i <= end_hbonds_index; i++) {
			for (j = 0; j < 2; j++) {
				partner = residues[i].hbonds_intended_acc[j].partner;
				if ((partner != -1) && !((partner >= begin_hbonds_index) && (partner <= end_hbonds_index))) {
						// Make sure that the partner is not also in the hydrogen bonds alignment region.
					total_align_pts++;
				}
			}
			for (j = 0; j < 2; j++) {
				partner = residues[i].hbonds_intended_don[j].partner;
				if ((partner != -1) && !((partner >= begin_hbonds_index) && (partner <= end_hbonds_index))) {
						// Make sure that the partner is not also in the hydrogen bonds alignment region.
					total_align_pts++;
				}
			}
		}
		
		if (total_align_pts == 0) {
			// No points to align.
			return TRUE;
		}
		
		total_align_pts *= 2;
		
		// Allocate memory for the alignment points.
		
		align_pts = malloc(total_align_pts * sizeof(CCD_alignment_point)); assert(align_pts != NULL);
		
		// Fill in the alignment points data.
		
		point_num = 0;
		
		for (i = begin_hbonds_index; i <= end_hbonds_index; i++) {
			
			// For each intended hydrogen bond found, calculate the 2 alignment points for the hydrogen bond.
			// For now, we position the ideal alignment points for the acceptor O atom and the donor H atom
			// such that the two atoms are HBOND_IDEAL_O_H_DIST Angstroms apart, with the H atom lying on the
			// C-O axis (and further away from the C atom), and with the O atom lying on the N-H axis (and
			// further away from the N atom). To do the calculations, we will use "phantom" points to
			// represent the ideal positions of certain atoms.
			// 
			// (Note that in reality, it is more ideal for the N-H bond to be pointing along one of the O
			// atom's two lone pair orbitals; perhaps this could be imrpoved in the future).
			//
			//      acceptor  C==O.....H--N  donor
			//                      |
			//                    hydrogen bond
			
			// With residue i as acceptor:
			for (j = 0; j < 2; j++) {
				partner = residues[i].hbonds_intended_acc[j].partner;
				if ((partner != -1) && (partner != 0) && !((partner >= begin_hbonds_index) && (partner <= end_hbonds_index))) {
						// Make sure that the partner is not also in the hydrogen bonds alignment region.
						// Also, the partner (since it is a donor) cannot be the first residue in the peptide because we do
						// not want to worry about that special case.
					
					align_pts[point_num].resIndex = i;
					align_pts[point_num].point_type = CCD_POINT_TYPE_HBOND_ACC;
					
					align_pts[point_num+1].resIndex = i;
					align_pts[point_num+1].point_type = CCD_POINT_TYPE_HBOND_ACC;
					
					// Note: We use the index (partner-1) to refer to the ResData struct that has the coordinates of the
					// HN and N atoms involved in the hydrogen bond.
					
					if (begin_hbonds_index == 0) {
						// The segment containing the hydrogen bonds we want to align is near the N-terminus.
						// Therefore a short N-terminal segment containing this hydrogen bonds region will remain
						// stationary, while the rest of the peptide will be moved to try to align the hydrogen bonds.
						
						// Current O atom (phantom)
						difference(residues[partner-1].atoms[HN], residues[partner-1].atoms[N], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[partner-1].atoms[HN], align_pts[point_num].current);
						
						// Target O atom
						copyVector(residues[i].atoms[O], align_pts[point_num].target);
						
						// Current H atom
						copyVector(residues[partner-1].atoms[HN], align_pts[point_num+1].current);
						
						// Target H atom (phantom)
						difference(residues[i].atoms[O], residues[i].atoms[C1], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[i].atoms[O], align_pts[point_num+1].target);
					}
					else {
						// The segment containing the hydrogen bonds we want to align is in the middle of the
						// peptide or near the C-terminal end. Therefore will will move this hydrogen bonds region,
						// while most of the rest of the peptide will remain stationary.
						
						// Current O atom
						copyVector(residues[i].atoms[O], align_pts[point_num].current);
						
						// Target O atom (phantom)
						difference(residues[partner-1].atoms[HN], residues[partner-1].atoms[N], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[partner-1].atoms[HN], align_pts[point_num].target);
						
						// Current H atom (phantom)
						difference(residues[i].atoms[O], residues[i].atoms[C1], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[i].atoms[O], align_pts[point_num+1].current);
						
						// Target H atom
						copyVector(residues[partner-1].atoms[HN], align_pts[point_num+1].target);
					}
					
					point_num += 2;
				}
			}
			
			// With residue i as donor:
			for (j = 0; j < 2; j++) {
				partner = residues[i].hbonds_intended_don[j].partner;
				if ((partner != -1) && (i != 0) && !((partner >= begin_hbonds_index) && (partner <= end_hbonds_index))) {
						// Make sure that the partner is not also in the hydrogen bonds alignment region.
						// Also, residue i (since it is the donor) cannot be the first residue in the peptide because we do
						// not want to worry about that special case.
					
					align_pts[point_num].resIndex = i;
					align_pts[point_num].point_type = CCD_POINT_TYPE_HBOND_DON;
					
					align_pts[point_num+1].resIndex = i;
					align_pts[point_num+1].point_type = CCD_POINT_TYPE_HBOND_DON;
					
					// Note: We use the index (i-1) to refer to the ResData struct that has the coordinates of the
					// HN and N atoms involved in the hydrogen bond.
					
					if (begin_hbonds_index == 0) {
						// The segment containing the hydrogen bonds we want to align is near the N-terminus.
						// Therefore a short N-terminal segment containing this hydrogen bonds region will remain
						// stationary, while the rest of the peptide will be moved to try to align the hydrogen bonds.
						
						// Current O atom
						copyVector(residues[partner].atoms[O], align_pts[point_num].current);
						
						// Target O atom (phantom)
						difference(residues[i-1].atoms[HN], residues[i-1].atoms[N], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[i-1].atoms[HN], align_pts[point_num].target);
						
						// Current H atom (phantom)
						difference(residues[partner].atoms[O], residues[partner].atoms[C1], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[partner].atoms[O], align_pts[point_num+1].current);
						
						// Target H atom
						copyVector(residues[i-1].atoms[HN], align_pts[point_num+1].target);
					}
					else {
						// The segment containing the hydrogen bonds we want to align is in the middle of the
						// peptide or near the C-terminal end. Therefore will will move this hydrogen bonds region,
						// while most of the rest of the peptide will remain stationary.
						
						// Current O atom (phantom)
						difference(residues[i-1].atoms[HN], residues[i-1].atoms[N], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[i-1].atoms[HN], align_pts[point_num].current);
						
						// Target O atom
						copyVector(residues[partner].atoms[O], align_pts[point_num].target);
						
						// Current H atom
						copyVector(residues[i-1].atoms[HN], align_pts[point_num+1].current);
						
						// Target H atom (phantom)
						difference(residues[partner].atoms[O], residues[partner].atoms[C1], temp);
						unitVector(temp, temp);
						scale(temp, HBOND_IDEAL_O_H_DIST);
						summation(temp, residues[partner].atoms[O], align_pts[point_num+1].target);
					}
					
					point_num += 2;
				}
			}
			
		}
		
		// Determine sumOfSquaredDistancesThreshold from the pre-defined RMSD threshold.
		
		sumOfSquaredDistancesThreshold = (HBOND_ALIGN_RMSD_THRESHOLD * HBOND_ALIGN_RMSD_THRESHOLD)*((double) (total_align_pts/2));
	}
	else {
		fprintf(stderr, "Illegal mode given to adjustSegmentCCD() in pepmake.c\n");
		exit(1);
	}
	
	// debug - print alignment point targets, pretending they are water molecules.
	// Print in PDB format so that the printout can be pasted onto the end of a PDB file and viewed in a PDB structure viewer.
	/*
	for (point_num = 0; point_num < total_align_pts; point_num++) {
		
		curr_align_pt = &align_pts[point_num];
		
		printf("HETATM %4d  O   HOH %5d    %8.3lf%8.3lf%8.3lf  1.00  1.00      TAG\n",
			point_num, point_num, curr_align_pt->target[0], curr_align_pt->target[1], curr_align_pt->target[2]);
		
	}
	*/
	
	sumOfSquaredDistances = 0.0;
	for (point_num = 0; point_num < total_align_pts; point_num++) {
		tempDouble = distance(align_pts[point_num].current, align_pts[point_num].target);
		sumOfSquaredDistances += tempDouble * tempDouble;
	}
	
	
	num_passes = 0;
	while ( num_passes < 20 || sumOfSquaredDistances > sumOfSquaredDistancesThreshold) {
		// Perform at least a certain number of passes.
		
		
		// Check if we should increase maxRamaQualityDecrease. We increase this value in a stepwise manner:
		// First we try several iterations with maxRamaQualityDecrease == 0 (Rama scores not allowed to get
		// worse), then several iterations with maxRamaQualityDecrease == 1, and then the rest with
		// maxRamaQualityDecrease == 2. This allows the CCD algorithm to perform large initial angle rotations
		// with strong angle constraints, and then try more permissive constraints if the strong constraints
		// are too restrictive.
		if (allowAnglesToBecomeWorse) {
			if (num_passes == 50) {
				maxRamaQualityDecrease = 1;
			}
			else if (num_passes == 100) {
				maxRamaQualityDecrease = 2;
			}
		}
		
		
		// If loop closure has not succeeded after a certain number of passes, give up. The calling function may then
		// decide to restore the original angles if desired.
		if (num_passes >= CCD_MAX_ITER) {
			
			if ( (mode == CCD_MODE_CLOSE_GAP) && (maxSumOfSquaredDistancesThreshold > 0.0) &&
					(sumOfSquaredDistances <= maxSumOfSquaredDistancesThreshold)) {
				// The calling function supplied maxSumOfSquaredDistances, and we have least achived this threshold.
				// We we have successfully closed the gap.
				if (debug) {
					printf("Success with relaxed threshold (sumOfSquaredDistances = %lf, relaxed threshold = %lf).\n",
						sumOfSquaredDistances, maxSumOfSquaredDistancesThreshold);
				}
			}
			else {
				closureFailed = TRUE;
				
				if (debug) {
					printf("**LOOP CLOSURE FAILED** (sumOfSquaredDistances = %lf, threshold = %lf).\n", sumOfSquaredDistances, sumOfSquaredDistancesThreshold);
				}
			}
			
			break; // Break out of while loop.
		}
		else if ( (mode == CCD_MODE_CLOSE_GAP) && (num_passes % convergence_check_period == 0) ) {
			// Periodically, we will check to see how quickly our sumOfSquaredDistances is decreasing. If we estimate that
			// it will take too long to close the gap in the segment, we will abort the loop closure.
			//
			// Here we extrapolate linearly, based on the current sumOfSquaredDistances and the value measured last time we
			// checked, to estimate the number of passes that will have occurred when sumOfSquaredDistances will be below
			// the threshold. In reality, it appears that sumOfSquaredDistances approaches 0 (or some value greater than 0)
			// asymptotically, so the linear extrapolation will estimate a value for the number of passes that is lower than
			// the real value. So if this estimate is greater than our maximum number of passes, we should abort the loop
			// closure attempt since we can be confident that the real number of passes will be even greater. Also, it may
			// be that the peptide segment will never close (sumOfSquaredDistances is asymptotically approaching a value that
			// is greater than the threshold), in which case our linear extrapolation will predict loop closure after a very
			// large number of passes, and we will abort the loop closure.
			//
			// For the math, we imagine the x-axis as num_passes and the y-axis as sumOfSquaredDistances.
			
			if ( !( (!allowAnglesToBecomeWorse && num_passes < 20) || (allowAnglesToBecomeWorse && num_passes < 120) ) ) {
				
				double threshold_for_extrapolation = sumOfSquaredDistancesThreshold;
				if ((mode == CCD_MODE_CLOSE_GAP) && (maxSumOfSquaredDistancesThreshold > 0.0)) {
					threshold_for_extrapolation = maxSumOfSquaredDistancesThreshold;
				}
				
				if ( !((mode == CCD_MODE_CLOSE_GAP) && (maxSumOfSquaredDistancesThreshold > 0.0)) &&
						sumOfSquaredDistances == olderSumOfSquaredDistances) {
					// Only use this check if maxSumOfSquaredDistancesThreshold is not specified.
					closureFailed = TRUE;
					//printf("Oh no, they are the same!\n"); // debug
				}
				//else if (sumOfSquaredDistances > olderSumOfSquaredDistances) {
				//  // This should never happen.
				//	closureFailed = TRUE;
				//	printf("Yikes, it got worse!\n"); // debug
				//}
				else {
					slope = (sumOfSquaredDistances - olderSumOfSquaredDistances)/((double) convergence_check_period);
						// slope should be negative.
					intercept = sumOfSquaredDistances - slope*((double) num_passes);
					estimated_passes = (int) ((threshold_for_extrapolation - intercept)/slope);
					
					//debug
					//printf(" olderSumOfSquaredDistances = %lf, sumOfSquaredDistances = %lf, num_passes = %d\n",
					//	olderSumOfSquaredDistances, sumOfSquaredDistances, num_passes);
					//printf(" slope = %lf, intercept = %lf\n", slope, intercept);
					//printf("   estimated_passes: %d\n", estimated_passes);
					
					if (estimated_passes >= CCD_MAX_ITER) {
						closureFailed = TRUE;
						//printf(">Abort because of linear extrapolation!\n"); // debug
					}
				}
				
				if (closureFailed == TRUE) {
					if (debug) {
						printf("**LOOP CLOSURE FAILED**: aborted after %d passes (sumOfSquaredDistances = %lf, threshold = %lf).\n", num_passes,
							sumOfSquaredDistances, threshold_for_extrapolation);
					}
					
					break; // Break out of while loop.
				}
			}
			
			olderSumOfSquaredDistances = sumOfSquaredDistances;
		}
		
		
		// Iterate through the segment residues and apply CCD algorithm.
		
		for (i = N_anchor_index + 1; i < begin_rigid_seg_index; i++) {
			
			// Rebuild the backbone of the ith residue of the peptide. In the for loop immediately following,
			// we will adjust the backbone atoms in the ith residue.
			rebuildPeptideSegment(residues, i, i, TRUE);
			
			for (whichAngleType = 0; whichAngleType < 3; whichAngleType++) {
				// whichAngleType can be 0 (phi), 1 (psi) or 2 (omega).
				
				applyRotation = TRUE;
				
				// Determine the rotation axis.
				
				if (whichAngleType == ANGLE_TYPE_PHI) {
					difference(residues[i].atoms[CA], residues[i-1].atoms[N], rotationAxis);
					unitVector(rotationAxis, rotationAxis);
					copyVector(residues[i-1].atoms[N], basePoint);
				}
				else if (whichAngleType == ANGLE_TYPE_PSI) {
					difference(residues[i].atoms[C1], residues[i].atoms[CA], rotationAxis);
					unitVector(rotationAxis, rotationAxis);
					copyVector(residues[i].atoms[CA], basePoint);
				}
				else { // if (whichAngleType == ANGLE_TYPE_OMEGA)
					difference(residues[i].atoms[N], residues[i].atoms[C1], rotationAxis);
					unitVector(rotationAxis, rotationAxis);
					copyVector(residues[i].atoms[C1], basePoint);
				}
				
				tan_alpha_nominator = 0.0;
				tan_alpha_denominator = 0.0;
				
				for (point_num = 0; point_num < total_align_pts; point_num++) {
					
					curr_align_pt = &align_pts[point_num];
					
					// Calculate vec_O.
					difference(curr_align_pt->current, basePoint, curr_align_pt->curr_translated);
					copyVector(rotationAxis, temp);
					scale(temp, dotproduct(curr_align_pt->curr_translated, rotationAxis) / dotproduct(rotationAxis, rotationAxis));
					summation(temp, basePoint, curr_align_pt->vec_O); // Translate vec_O into normal peptide coordinate space.
					
					// Calculate vec_r.
					difference(curr_align_pt->current, curr_align_pt->vec_O, curr_align_pt->vec_r);
					
					// Calculate vec_r_unit.
					unitVector(curr_align_pt->vec_r, curr_align_pt->vec_r_unit);
					
					// Calculate mag_r.
					curr_align_pt->mag_r = distance(curr_align_pt->current, curr_align_pt->vec_O);
					
					// Calculate vec_s.
					crossproduct(curr_align_pt->vec_r_unit, rotationAxis, curr_align_pt->vec_s);
					
					// Calculate vec_f.
					difference(curr_align_pt->target, curr_align_pt->vec_O, curr_align_pt->vec_f);
					
					// As we loop, calculate the nominator and denominator that will be used to calculate tan_alpha.
					tan_alpha_nominator += dotproduct(curr_align_pt->vec_f, curr_align_pt->vec_s)*(curr_align_pt->mag_r);
					tan_alpha_denominator += dotproduct(curr_align_pt->vec_f, curr_align_pt->vec_r_unit)*(curr_align_pt->mag_r);
				}
				
				// Calculate tan(alpha).
				
				tan_alpha = tan_alpha_nominator / tan_alpha_denominator;
				
				alpha_rad = atan(tan_alpha);
				
				// Calculate alpha. Use the second derivative test to determine which alpha angle is the right one.
				
				second_deriv = 0.0;
				for (point_num = 0; point_num < total_align_pts; point_num++) {
					curr_align_pt = &align_pts[point_num];
					
					second_deriv += cos(alpha_rad)*dotproduct(curr_align_pt->vec_f, curr_align_pt->vec_r_unit)*(curr_align_pt->mag_r)
						+ sin(alpha_rad)*dotproduct(curr_align_pt->vec_f, curr_align_pt->vec_s)*(curr_align_pt->mag_r);
				}
				second_deriv *= 2.0;
				
				if (second_deriv < 0.0) {
					// Current alpha_rad is incorrect; change it by pi radians.
					if (alpha_rad >= 0.0) {
						alpha_rad -= PI;
					}
					else {
						alpha_rad += PI;
					}
				}
				
				alpha = alpha_rad * 180.0 / PI; // Convert to degrees.
				
				// Adjust the angle just in case rounding error has placed it outside the range -180.0 to 180.0.
				if (alpha < -180.0) {
					alpha += 360.0;
				}
				else if (alpha > 180.0) {
					alpha -= 360.0;
				}
				
				// Test the alpha angle and, if necessary, adjust it based on constraints imposed by peptide geometry. Then apply
				// the alpha angle.
				
				if (whichAngleType == ANGLE_TYPE_PHI) {
					desired_phi = residues[i].phi;
					
					desired_phi -= alpha;
					if (desired_phi < -180.0) {
						desired_phi += 360.0;
					}
					else if (desired_phi > 180.0) {
						desired_phi -= 360.0;
					}
					
					if (allowAnglesToBecomeWorse) {
						// Here we use the original phi and psi angles (before CCD) as parameters so that in
						// successive iterations the angles will not become worse more than once decrease in
						// quality by more than maxRamaQualityDecrease.
						j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
						phi = getClosestAllowedAngle(residues, i, numRes, desired_phi, ANGLE_TYPE_PHI,
							original_angles[j][0], original_angles[j][1], maxRamaQualityDecrease);
					}
					else {
						// Here we use the current phi and psi angles (resulting from the previous iteration).
						// Angles cannot become worse, but they may possibly have gotten better in a previous
						// iteration, and we want to prevent them from becoming worse again, even if this means
						// that they end up with the same quality as earlier. This places an additional
						// constraint on the CCD algorithm, but it was observed with a test set of proteins
						// that the resulting Ramachandran angles in the segment are better with this strategy,
						// along with slightly better Bump scores and RMSDs that are only slightly worse.
						phi = getClosestAllowedAngle(residues, i, numRes, desired_phi, ANGLE_TYPE_PHI,
							residues[i].phi, residues[i].psi, maxRamaQualityDecrease);
					}
					
					if (fabs(phi - residues[i].phi) > ANGLE_CHANGE_THRESHOLD) {
						
						// debug
						if (debug) {
							if (desired_phi != phi) {
								printf("%s %d: constraints changed phi: %lf --> %lf (original = %lf)\n", residues[i].resName, i,
									desired_phi, phi, residues[i].phi);
								printf("   staring at phi = %lf (phiindex = %d), psi = %lf (psiindex = %d)\n",
									residues[i].phi, getPhiIndex(residues[i].phi), residues[i].psi, getPsiIndex(residues[i].psi));
								printf("   ending at phi = %lf (phiindex = %d), desired_phi = %lf (phiindex = %d)\n",
									phi, getPhiIndex(phi), desired_phi, getPhiIndex(desired_phi));
							}
						}
						
						alpha = phi - residues[i].phi; // Set the constrained alpha value, which will be used to calculate curr_CA etc.
						
						residues[i].phi = phi;
							// Set the new angle, which will be used when rebuildPeptideSegment() is called after the
							// segment is closed.
					}
					else {
						// The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
						// for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
						// computation.
						/*
						if (debug) {
							if (phi != residues[i].phi) {
								printf("Did not rotate by %lf degrees.\n", fabs(phi - residues[i].phi));
							}
						}
						*/
						applyRotation = FALSE;
					}
				}
				else if (whichAngleType == ANGLE_TYPE_PSI) {
					desired_psi = residues[i].psi;
					
					desired_psi -= alpha;
					if (desired_psi < -180.0) {
						desired_psi += 360.0;
					}
					else if (desired_psi > 180.0) {
						desired_psi -= 360.0;
					}
					
					if (allowAnglesToBecomeWorse) {
						// Here we use the original phi and psi angles (before CCD) as parameters so that in
						// successive iterations the angles will not become worse more than once decrease in
						// quality by more than maxRamaQualityDecrease.
						j = i - N_anchor_index - 1; // j here is the index into original_angles, starting from 0.
						psi = getClosestAllowedAngle(residues, i, numRes, desired_psi, ANGLE_TYPE_PSI,
							original_angles[j][0], original_angles[j][1], maxRamaQualityDecrease);
					}
					else {
						// Here we use the current phi and psi angles (resulting from the previous iteration).
						// Angles cannot become worse, but they may possibly have gotten better in a previous
						// iteration, and we want to prevent them from becoming worse again, even if this means
						// that they end up with the same quality as earlier. This places an additional
						// constraint on the CCD algorithm, but it was observed with a test set of proteins
						// that the resulting Ramachandran angles in the segment are better with this strategy,
						// along with slightly better Bump scores and RMSDs that are only slightly worse.
						psi = getClosestAllowedAngle(residues, i, numRes, desired_psi, ANGLE_TYPE_PSI,
							residues[i].phi, residues[i].psi, maxRamaQualityDecrease);
					}
					
					if (fabs(psi - residues[i].psi) > ANGLE_CHANGE_THRESHOLD) {
						
						// debug
						if (debug) {
							if (desired_psi != psi) {
								printf("%s %d: constraints changed psi: %lf --> %lf (original = %lf)\n", residues[i].resName, i,
									desired_psi, psi, residues[i].psi);
								printf("   staring at phi = %lf (phiindex = %d), psi = %lf (psiindex = %d)\n",
									residues[i].phi, getPhiIndex(residues[i].phi), residues[i].psi, getPsiIndex(residues[i].psi));
								printf("   ending at psi = %lf (psiindex = %d), desired_psi = %lf (psiindex = %d)\n",
									psi, getPsiIndex(psi), desired_psi, getPsiIndex(desired_psi));
							}
						}
						
						alpha = psi - residues[i].psi; // Set the constrained alpha value, which will be used to calculate curr_CA etc.
						
						residues[i].psi = psi;
							// Set the new angle, which will be used when rebuildPeptideSegment() is called after the
							// segment is closed.
					}
					else {
						// The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
						// for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
						// computation.
						/*
						if (debug) {
							if (psi != residues[i].psi) {
								printf("Did not rotate by %lf degrees.\n", fabs(psi - residues[i].psi));
							}
						}
						*/
						applyRotation = FALSE;
					}
				}
				else { // whichAngleType == ANGLE_TYPE_OMEGA
					desired_omega = residues[i].omega;
					
					desired_omega -= alpha;
					if (desired_omega < -180.0) {
						desired_omega += 360.0;
					}
					else if (desired_omega > 180.0) {
						desired_omega -= 360.0;
					}
					
					omega = getClosestAllowedAngle(residues, i, numRes, desired_omega, ANGLE_TYPE_OMEGA,
						0.0, 0.0, maxRamaQualityDecrease);
					
					if (fabs(omega - residues[i].omega) > ANGLE_CHANGE_THRESHOLD) {
						
						// debug
						if (debug) {
							if (desired_omega != omega) {
								printf("%s %d: constraints changed omega: %lf --> %lf (original = %lf)\n", residues[i].resName, i,
									desired_omega, omega, residues[i].omega);
							}
						}
						
						alpha = omega - residues[i].omega; // Set the constrained alpha value, which will be used to calculate curr_CA etc.
						
						residues[i].omega = omega;
							// Set the new angle, which will be used when rebuildPeptideSegment() is called after the
							// segment is closed.
					}
					else {
						// The constraints from getClosestAllowedAngle resulted in no significant change to the angle,
						// for an effective alpha of 0.0 (or nearly 0.0). Make a note of this so that we can avoid needless
						// computation.
						/*
						if (debug) {
							if (omega != residues[i].omega) {
								printf("Did not rotate by %lf degrees.\n", fabs(omega - residues[i].omega));
							}
						}
						*/
						applyRotation = FALSE;
					}
				}
				
				if (applyRotation) {
					
					// Reposition the backbone atoms of the current ith residue according to the
					// change specified by alpha.
					if (whichAngleType == ANGLE_TYPE_PHI) {
						
						difference(residues[i].atoms[C1], basePoint, residues[i].atoms[C1]);
						difference(residues[i].atoms[N], basePoint, residues[i].atoms[N]);
						difference(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]);
						
						rotateCC(rotationAxis, residues[i].atoms[C1], alpha);
						rotateCC(rotationAxis, residues[i].atoms[N], alpha);
						rotateCC(rotationAxis, residues[i].atoms[CA2], alpha);
						
						summation(residues[i].atoms[C1], basePoint, residues[i].atoms[C1]); // Translate back into normal peptide coordinate space.
						summation(residues[i].atoms[N], basePoint, residues[i].atoms[N]); // Translate back into normal peptide coordinate space.
						summation(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]); // Translate back into normal peptide coordinate space.
					}
					else if (whichAngleType == ANGLE_TYPE_PSI) {
						
						difference(residues[i].atoms[N], basePoint, residues[i].atoms[N]);
						difference(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]);
						
						rotateCC(rotationAxis, residues[i].atoms[N], alpha);
						rotateCC(rotationAxis, residues[i].atoms[CA2], alpha);
						
						summation(residues[i].atoms[N], basePoint, residues[i].atoms[N]); // Translate back into normal peptide coordinate space.
						summation(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]); // Translate back into normal peptide coordinate space.
					}
					else { // whichAngleType == ANGLE_TYPE_OMEGA
						
						difference(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]);
						
						rotateCC(rotationAxis, residues[i].atoms[CA2], alpha);
						
						summation(residues[i].atoms[CA2], basePoint, residues[i].atoms[CA2]); // Translate back into normal peptide coordinate space.
					}
					
					
					// Calculate coordinates of the alignment points directly from alpha.
					
					for (point_num = 0; point_num < total_align_pts; point_num++) {
						
						curr_align_pt = &align_pts[point_num];
						
						difference(curr_align_pt->current, basePoint, curr_align_pt->curr_translated);
						rotateCC(rotationAxis, curr_align_pt->curr_translated, alpha);
						summation(curr_align_pt->curr_translated, basePoint, curr_align_pt->current); // Translate back into normal peptide coordinate space.
					}
					
				}
				
				// Only for debug
				/*
				sumOfSquaredDistances = 0.0;
				for (point_num = 0; point_num < total_align_pts; point_num++) {
					tempDouble = distance(align_pts[point_num].current, align_pts[point_num].target);
					sumOfSquaredDistances += tempDouble * tempDouble;
				}
				
				printf("sumOfSquaredDistances = %lf\n", sumOfSquaredDistances);
				*/
				
			}
			
		}
		
		// After each pass of the CCD algorithm through the whole segment, measure the sum of squared
		// distances between the current C-terminal residue atom coordinates and original C-anchor
		// atom coordinates. Apply the CCD algorithm to the segment repeatedly until the difference
		// is below a certain tolerance threshold where the two sets of coordinates almost perfectly
		// overlap.
		
		sumOfSquaredDistances = 0.0;
		for (point_num = 0; point_num < total_align_pts; point_num++) {
			tempDouble = distance(align_pts[point_num].current, align_pts[point_num].target);
			sumOfSquaredDistances += tempDouble * tempDouble;
		}
		
		num_passes++;
		
		//printf("   num_passes = %d, sumOfSquaredDistances = %lf\n", num_passes, sumOfSquaredDistances);
	}
	
	if (mode == CCD_MODE_CLOSE_GAP) {
		if (!closureFailed) {
			
			if (debug) {
				printf("Loop closed in %d passes.\n", num_passes);
			}
			
			// Loop closure has succeeded. Since our calls above to rebuildPeptideSegment set onlyBackbone
			// to TRUE, the non-backbone atoms are in the wrong positions. Therefore we should rebuild
			// the current peptide segment with onlyBackbone set to FALSE to get all atoms in the right
			// place.
			
			if (rebuild_after_gap) {
				// The C-terminal coordinates of the residue at C_anchor_index have been slightly changed
				// as a result of the CCD algorithm, which may slightly distort the structure from the
				// point of this segment to the C-terminus of the entire peptide. We will rebuild the
				// portion of the structure after the current segment to ensure correct local geometry.
				// However, this may cause global distortion in geometry because the portion after the
				// C-anchor will move. If the threshold for gap closure is very low, however (close to
				// the default value of 0.000001), this global distortion will be minimal.
				
				rebuildPeptideSegment(residues, N_anchor_index, numRes-1, FALSE);
			}
			else {
				// The C-terminal coordinates of the residue at C_anchor_index have been slightly changed
				// as a result of the CCD algorithm, which may slightly distort the structure from the
				// point of this segment to the C-terminus of the entire peptide. However, rebuild_after_gap
				// is FALSE, so we will only rebuild the segment from the N-anchor to the C-anchor. This
				// may result in local distortion of backbone geometry at the C-anchor residue.
				
				rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, FALSE);
			}
		}
	}
	else if (mode == CCD_MODE_ALIGN_HBONDS) {
		// We have finished aligning hydrogen bonds, and we may or may not have "failed" to achieve
		// sumOfSquaredDistances <= sumOfSquaredDistancesThreshold. But we have at least tried to move the
		// segment so that the alignment points are satisfied as well as they can be. In any case, the
		// segment we have adjusted has the non-backbone atoms in the wrong positions, so we should rebuild
		// the current peptide segment with onlyBackbone set to FALSE to get all atoms in the right place.
		// If the given segment is in the middle of the peptide, then there will most likely be a gap in
		// the structure between the residue at C_anchor_index and the following residue. We will not try
		// to fix that problem here; the calling function is responsible for repairing this gap.
		
		rebuildPeptideSegment(residues, N_anchor_index, C_anchor_index, FALSE);
	}
	
	// free memory
	free(align_pts);
	
	
	return !closureFailed;
}


/*************************************************************/
/* Sets the angles in the given segment of the peptide to    */
/* to the closest allowed corrected angles. For residues     */
/* startIndex to endIndex in the given ResData array, the    */
/* phi and psi angles are adjusted to the closest allowed    */
/* (phi, psi) pair that is in the region class specified by  */
/* rama_score_threshold:                                     */
/*                                                           */
/*   3 = angles adjusted to phi/psi core region              */
/*   2 = angles adjusted to allowed region (or core region   */
/*       if a core region is closer)                         */
/*   1 = angles adjusted to at least the minimally allowed   */
/*       region                                              */
/*                                                           */
/* In addition, omega angles are adjusted to the closest     */
/* acceptable omega angles. This function only sets the phi, */
/* psi and omega member variables of the ResData structs,    */
/* and does not recalculate any atom coordinates.            */
/*************************************************************/
static void setCorrectedAngles(ResData *residues, int numRes, int startIndex, int endIndex, int rama_score_threshold) {
	
	int i;
	double phi, psi, omega;
	int debug = FALSE;
	
	if (debug) { printf("Fixing angles for segment:\n"); }
	
	for (i = startIndex; i <= endIndex; i++) {
		
		// Fix phi and psi.
		
		findClosestAllowedPhiPsiPair(residues, i, numRes, rama_score_threshold, &phi, &psi);
		
		// debug
		if (debug) {
			if (phi != residues[i].phi || psi != residues[i].psi) {
				printf("%s %d: fixing phi and psi: (%lf, %lf) --> (%lf, %lf)\n", residues[i].resName, i,
					residues[i].phi, residues[i].psi, phi, psi);
				printf("   indices: (%d, %d) --> (%d, %d)\n", getPhiIndex(residues[i].phi), getPsiIndex(residues[i].psi),
					getPhiIndex(phi), getPsiIndex(psi));
			}
		}
		
		residues[i].phi = phi;
		residues[i].psi = psi;
		
		// Fix omega.
		
		omega = residues[i].omega;
		
		if (omega < -180.0) {
			omega += 360.0;
		}
		else if (omega > 180.0) {
			omega -= 360.0;
		}
		
		if (omega < 170.0 && omega >= 90.0) {
			omega = 170.0 + ANGLE_SNAP_MARGIN;
		}
		else if (omega > -170.0 && omega <= -90.0) {
			omega = -170.0 - ANGLE_SNAP_MARGIN;
		}
		else if (omega > 10.0 && omega <= 90.0) {
			omega = 10.0 - ANGLE_SNAP_MARGIN;
		}
		else if (omega < -10.0 && omega >= -90.0) {
			omega = -10.0 + ANGLE_SNAP_MARGIN;
		}
		
		if (debug) {
			if (omega != residues[i].omega) {
				printf("%s %d: fixing omega: %lf --> %lf\n", residues[i].resName, i, residues[i].omega, omega);
			}
		}
		
		residues[i].omega = omega;
		
	}
	
}


/*************************************************************/
/* Reads through the residue coordinates file indicated by   */
/* the given filename, and read the data into the "names"    */
/* and "template" fields of each residue's ResData struct.   */
/*************************************************************/
void readResCoordsFile(ResData *residues, int totalRes, char *resCoordsFilename) {
	FILE *fp;
	ResData *res;
	double chi, N_CA_C1_angle;
	int resNum, atomNum;
	int n;
	char name[5];
	double x,y,z;
	char resName[4];
	char line[64];
	int atom_index;
	int i, j;
	int readN, readCA, readCG;
	
	int debug = FALSE;
	
	
	if ( (fp = fopen(resCoordsFilename, "r")) ) {
		
		for (resNum=0; resNum<totalRes; resNum++) {
			res = &residues[resNum];
			
			// initialize
			for (i = 0; i < MAX_ATOMS; i++) {
				setVector(res->template[i], ATOM_COORD_NA, ATOM_COORD_NA, ATOM_COORD_NA);
				res->names[i][0] = (char) NULL;
				res->fixed_position = FALSE;
			}
			
			
			if (fgets(line, 64, fp) != NULL) {
				sscanf(line, "%s %lf %lf %d", resName, &chi, &N_CA_C1_angle, &n);
			}
			
			//r = fscanf(fp, "%s %lf %lf %d", resName, &chi, &N_CA_C1_angle, &n);
			if (debug) printf("chi=%lf, N_CA_C1_angle=%lf, n=%d\n", chi, N_CA_C1_angle, n);
			res->chi = chi;
			res->N_CA_C1_angle = N_CA_C1_angle;
			
			res->max_atom_index = 0; // Initialize this here; it will be adjusted below.
			
			j = CG;
			readN = FALSE;
			readCA = FALSE;
			readCG = FALSE;
			for (atomNum=0; atomNum<n; atomNum++) {
				
				if (fgets(line, 64, fp) != NULL) {
					strncpy(name, &(line[1]), 4);
						// On each line the first char is a space, then next 4 are the atom name.
					name[4] = (char) NULL;
					
					sscanf(&(line[5]), "%lf%lf%lf", &x, &y, &z);
					
					
					atom_index = getAtomIndexFromAtomName(name);
					
					if (atom_index == CG) {
						if (readCG == FALSE) {
							readCG = TRUE;
						}
						else {
							// Only the first gamma atom can occupy index CG.
							atom_index = -1;
						}
					}
					
					if (atom_index == -1) {
						j++;
						atom_index = j;
					}
					else if (atom_index == Np) {
						if (readN == FALSE) {
							readN = TRUE;
						}
						else {
							atom_index = N;
						}
					}
					else if (atom_index == CA) {
						if (readCA == FALSE) {
							readCA = TRUE;
						}
						else {
							atom_index = CA2;
						}
					}
					
					strncpy(res->names[atom_index], name, 5);
					res->template[atom_index][0] = x;
					res->template[atom_index][1] = y;
					res->template[atom_index][2] = z;
					
					// Adjust res->max_atom_index.
					if (atom_index > res->max_atom_index) {
						res->max_atom_index = atom_index;
					}
				}
				
			}
			
		}
		fclose(fp);
	} else {
		fprintf(stderr, "Cannot open %s\n", resCoordsFilename);
		exit(1);
	}
	
	//debug
	if (debug) {
		printf("Printing back what we read in:\n");
		for (resNum=0; resNum<totalRes; resNum++) {
			res = &residues[resNum];
			
			printf("max_atom_index = %d, chi = %lf\n", res->max_atom_index, res->chi);
			
			for (atomNum=0; atomNum <= res->max_atom_index; atomNum++) {
				printf("  %s %lf %lf %lf\n", res->names[atomNum], res->template[atomNum][0], res->template[atomNum][1], res->template[atomNum][2]);
			}
		}
	}
	
}


/*************************************************************/
/* Read the next residue from the file. Returns false if the */
/* residue could not be read (usually end of file)           */
/*************************************************************/
static int readResidue(ResData *res, FILE *fp) {
	int rc, num;

	/* read the residue & phi psi omega angles */
	rc = fscanf(fp, "%d%s%lf%lf%lf%lf%d", 
		&res->id, res->resName, &res->phi, &res->psi, 
		&res->omega, &res->chi, &num);
	if (rc != 7) return FALSE;
	return TRUE; 
}


/*************************************************************/
/* Opens the file and counts the number of residues entries  */
/* there are with dihedral angles, etc...                    */
/*************************************************************/
int countResidues(char *fname) {
	int n = 0;
	ResData res;
	FILE *fp = fopen(fname, "r");
	if (fp != NULL) {
	 
		while (!feof(fp)) {
		  if (readResidue(&res, fp)) {
				n++;
			} else {
				fclose(fp);
				return n;
			}
		}
		fclose(fp);
	}
	return n;
}

/*************************************************************/
/* Rebuilds a segment of the peptide from startIndex to      */
/* endIndex. Assumes that the atoms for the residue at index */
/* (startIndex - 1) have already been set, for example by a  */
/* previous call to buildPeptide. If onlyBackbone is TRUE,   */
/* only the backbone atoms will be manipulated.              */
/*************************************************************/
void rebuildPeptideSegment(ResData *residues, int startIndex, int endIndex, int onlyBackbone) {
	int i;
	
	if (startIndex <= 0) {
		fprintf(stderr, "Error: rebuildPeptideSegment called with startIndex <= 0\n");
		exit(-1);
	}
	else if (startIndex > endIndex) {
		fprintf(stderr, "Error: rebuildPeptideSegment called with startIndex > endIndex\n");
		exit(-1);
	}
	
	for (i = startIndex; i <= endIndex; i++) {
		ResData *res = &residues[i];
		
		if (residues[i].fixed_position == FALSE) {
			nextPeptide(res, i, residues, onlyBackbone);
		}
	}
}

/* iteratively assembles the peptide */
void buildPeptide(ResData *residues, int numRes) {
	int i;
	
	// Need to ensure that some coordinates are set for the first residue before calling nextPeptide.
	if (residues[0].fixed_position == FALSE) {
		initResData(&residues[0]);
	}
	
	for (i=0; i<numRes; i++) {
		ResData *res = &residues[i];
		
		if (residues[i].fixed_position == FALSE) {
			nextPeptide(res, i, residues, FALSE);
		}
	}
}


/*
 * Loads the residue template data (i.e. the coordinates of each residue in canonical form)
 * into each ResData struct.
 *
 * If usePDBFile is TRUE and createResCoordsFile is TRUE, then this data is derived from
 * the PDB file with name inputPDBFilename (by calling phispi with the --db option) and
 * reading the residue coordinates from the resulting file.
 *
 * If usePDBFile is TRUE and createResCoordsFile is FALSE, then this function assumes that
 * 'pdb' is a PDBFile struct that have been run through parsePDBFile(), and the data will
 * be loaded from here without using an intermediate residue coordinates file.
 *
 * If usePDBFile is FALSE and haveResCoordsFile is TRUE, the residue coordinates are loaded
 * from the given residue coordinates file 'resCoordsFilename'.
 *
 * If usePDBFile is FALSE and haveResCoordsFile is FALSE,  the residue coordinates are
 * loaded from the residue template database, and certain atoms for the first and last
 * residues are given default positions.
 */
void loadResidueTemplates(char *inputPDBFilename, PDBFile pdb, int usePDBFile, int createResCoordsFile,
		int haveResCoordsFile, char *resCoordsFilename, ResData *residues, int n) {
	
	int i = 0;
	
	if (usePDBFile) {
		if (createResCoordsFile) {
			// Call phipsi with the --db flag to generate a residue coordinates file based on the PDB file.
			// These coordinates will then be used instead of the default generic coordinates for the
			// backbone and side chains.
			resCoordsFilename = createResidueCoordinatesFile(inputPDBFilename);
			readResCoordsFile(residues, n, resCoordsFilename);
		}
		else {
			// Load the data directly from the PDB file into the ResData array.
			PDBtoResData(pdb, residues);
		}
	}
	else if (haveResCoordsFile) {
		// Use the coordinates in the given residue coordinates file instead of the default generic
		// coordinates for the backbone and side chains.
		readResCoordsFile(residues, n, resCoordsFilename);
	}
	else {
		for (i=0; i<n; i++) {
			ResData *res = &residues[i];
			
			/* read atom coordinates for the residue template */
			readTemplate(res, ignoreDBBackbone, i == 0, i == n-1);
		}
		
	}
	
}


/*
 * Given a PDBFile struct from a parsed PDB file, this function loads the structure
 * data into the given array of pdb.numRes ResData structs. Sets the IDs, residues
 * names, 'atom', 'names' and 'template' arrays, and max_atom_index and fixed_position
 * values. Computes and sets the angles as well. If there are any missing atoms, the
 * angles that depend on them will be set to ANGLE_NA.
 */
void PDBtoResData(PDBFile pdb, ResData *res) {
	
	int i, j;
	int resNumOfPreviousAtom = -1000000, resIndex = -1;
	
	// Initialize ResData atom locations and fixed_position member variables.
	for (i = 0; i < pdb.numRes; i++) {
		for (j = 0; j < MAX_ATOMS; j++) {
			setVector(res[i].atoms[j], ATOM_COORD_NA, ATOM_COORD_NA, ATOM_COORD_NA);
		}
		res[i].fixed_position = FALSE;
	}
	
	for (i = 0; i < pdb.numAtoms; i++) {
		
		if (pdb.atoms[i].resNum != resNumOfPreviousAtom) {
			resNumOfPreviousAtom = pdb.atoms[i].resNum;
			resIndex++;
			
			res[resIndex].id = pdb.atoms[i].resNum;
			strcpy(res[resIndex].resName, pdb.atoms[i].resName);
			
			parseResidueFromPDBFile(&pdb, i, res[resIndex].atoms, res[resIndex].template,
				res[resIndex].names, NULL, &(res[resIndex].max_atom_index) );
			
			
			// Set fixed_position member variable. If the current residue is the first
			// residue in the peptide, or if it is the first residue after a gap in the
			// structure, then it should have a fixed position.
			
			if (resIndex == 0) {
				res[resIndex].fixed_position = TRUE;
				
				// debug
				//fprintf(stderr, "Residue %3s %d is fixed because it is the first residue.\n",
				//	res[resIndex].resName, res[resIndex].id);
			}
			else{
				
				if (	res[resIndex-1].atoms[CA][0] == ATOM_COORD_NA ||
						res[resIndex-1].atoms[C1][0] == ATOM_COORD_NA ||
						res[resIndex-1].atoms[N][0] == ATOM_COORD_NA) {
					res[resIndex].fixed_position = TRUE;
					
					// debug
					//fprintf(stderr, "Residue %3s %d is fixed due to previous residue missing atoms.\n",
					//	res[resIndex].resName, res[resIndex].id);
				}
				else if (res[resIndex].atoms[CA][0] != ATOM_COORD_NA) {
					if (distance(res[resIndex-1].atoms[CA], res[resIndex].atoms[CA]) > CA_CA_GAP_THRESHOLD) {
						res[resIndex].fixed_position = TRUE;
						
						// debug
						//fprintf(stderr, "Residue %3s %d is fixed due to CA-CA distance.\n",
						//	res[resIndex].resName, res[resIndex].id);
					}
				}
				
			}
			
			// debug
			//if (res[resIndex].fixed_position == TRUE) {
			//	fprintf(stderr, "Residue %3s %d is fixed.\n", res[resIndex].resName, res[resIndex].id);
			//}
			
		}
		
	}
	
	compute_angles(res, resIndex+1);
}


/*
 * Computes the phi, psi, omega, chi, and N-CA-C1 angles for the given ResData
 * array of size n, based on the coordinates in the 'atom' member variables of
 * each ResData struct. If there are any missing atoms, the angles that
 * depend on them will be set to ANGLE_NA.
 */
static void compute_angles(ResData *res, int n) {
	int i, j;
	Vec3D CA_N, CA_C1;
	
	for (i = 0; i < n; i++) {
			
			// phi
			if (i == 0) {
				res[i].phi = ANGLE_NA;
			}
			else {
				if (	res[i-1].atoms[C1][0] != ATOM_COORD_NA &&
						res[i-1].atoms[N][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA) {
					res[i].phi = torsion(res[i-1].atoms[C1], res[i-1].atoms[N],
						res[i].atoms[CA], res[i].atoms[C1]);
				}
				else {
					res[i].phi = ANGLE_NA;
				}
			}
			
			// psi
			if (i == 0) {
				if (	res[i].atoms[Np][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA &&
						res[i].atoms[N][0] != ATOM_COORD_NA) {
					res[i].psi = torsion(res[i].atoms[Np], res[i].atoms[CA],
						res[i].atoms[C1], res[i].atoms[N]);
				}
				else {
					res[i].psi = ANGLE_NA;
				}
			}
			else if (i == n-1) {
				// Last residue.
				
				// Look for " OXT" atom in last residue.
				int OXT_index = -1;
				for (j = HN; j <= res[i].max_atom_index; j++) {
					if (strncmp(res[i].names[j], " OXT", 4) == 0) {
						OXT_index = j;
					}
				}
				
				if (OXT_index >= 0) {
					// Use the OXT atom to measure psi, only for the final residue.
					if (	res[i-1].atoms[N][0] != ATOM_COORD_NA &&
							res[i].atoms[CA][0] != ATOM_COORD_NA &&
							res[i].atoms[C1][0] != ATOM_COORD_NA &&
							res[i].atoms[OXT_index][0] != ATOM_COORD_NA) {
						res[i].psi = torsion(res[i-1].atoms[N], res[i].atoms[CA],
							res[i].atoms[C1], res[i].atoms[OXT_index]);
					}
					else {
						res[i].psi = ANGLE_NA;
					}
				}
				else {
					// Final residue does not have an OXT atom listed in the PDB file.
					// Use the location of the O atom to measure psi.
					if (	res[i-1].atoms[N][0] != ATOM_COORD_NA &&
							res[i].atoms[CA][0] != ATOM_COORD_NA &&
							res[i].atoms[C1][0] != ATOM_COORD_NA &&
							res[i].atoms[O][0] != ATOM_COORD_NA) {
						res[i].psi = torsion(res[i-1].atoms[N], res[i].atoms[CA],
							res[i].atoms[C1], res[i].atoms[O]);
						
						if (res[i].psi > 0.0) {
							res[i].psi -= 180.0;
						}
						else {
							res[i].psi += 180.0;
						}
					}
					else {
						res[i].psi = ANGLE_NA;
					}
				}
			}
			else {
				if (	res[i-1].atoms[N][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA &&
						res[i].atoms[N][0] != ATOM_COORD_NA) {
					res[i].psi = torsion(res[i-1].atoms[N], res[i].atoms[CA],
						res[i].atoms[C1], res[i].atoms[N]);
				}
				else if (	res[i].atoms[Np][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA &&
						res[i].atoms[N][0] != ATOM_COORD_NA) {
					res[i].psi = torsion(res[i].atoms[Np], res[i].atoms[CA],
						res[i].atoms[C1], res[i].atoms[N]);
				}
				else {
					res[i].psi = ANGLE_NA;
				}
			}
			
			// omega
			if (i < n-1) {
				if (	res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA &&
						res[i].atoms[N][0] != ATOM_COORD_NA &&
						res[i].atoms[CA2][0] != ATOM_COORD_NA) {
					res[i].omega = torsion(res[i].atoms[CA], res[i].atoms[C1],
						res[i].atoms[N], res[i].atoms[CA2]);
				}
				else {
					res[i].omega = ANGLE_NA;
				}
			}
			
			// chi
			if (strcmp(res[i].resName, "GLY") != 0 && strcmp(res[i].resName, "ALA") != 0) {
				if (i == 0) {
					if (	res[i].atoms[Np][0] != ATOM_COORD_NA &&
							res[i].atoms[CA][0] != ATOM_COORD_NA &&
							res[i].atoms[CB][0] != ATOM_COORD_NA &&
							res[i].atoms[CG][0] != ATOM_COORD_NA) {
						res[i].chi = torsion(res[i].atoms[Np], res[i].atoms[CA],
							res[i].atoms[CB], res[i].atoms[CG]);
					}
					else {
						res[i].chi = ANGLE_NA;
					}
				}
				else {
					if (	res[i-1].atoms[N][0] != ATOM_COORD_NA &&
							res[i].atoms[CA][0] != ATOM_COORD_NA &&
							res[i].atoms[CB][0] != ATOM_COORD_NA &&
							res[i].atoms[CG][0] != ATOM_COORD_NA) {
						res[i].chi = torsion(res[i-1].atoms[N], res[i].atoms[CA],
							res[i].atoms[CB], res[i].atoms[CG]);
					}
					else {
						res[i].chi = ANGLE_NA;
					}
				}
			}
			else {
				res[i].chi = ANGLE_NA;
			}
			
			// N-CA-C1 angle
			if (i == 0) {
				
				if (	res[i].atoms[Np][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA) {
					
					difference(res[i].atoms[Np], res[i].atoms[CA], CA_N);
					difference(res[i].atoms[C1], res[i].atoms[CA], CA_C1);
					res[i].N_CA_C1_angle = angle(CA_N, CA_C1);
				}
				else {
					res[i].N_CA_C1_angle = ANGLE_NA;
				}
			}
			else {
				
				if (	res[i-1].atoms[N][0] != ATOM_COORD_NA &&
						res[i].atoms[CA][0] != ATOM_COORD_NA &&
						res[i].atoms[C1][0] != ATOM_COORD_NA) {
					
					difference(res[i-1].atoms[N], res[i].atoms[CA], CA_N);
					difference(res[i].atoms[C1], res[i].atoms[CA], CA_C1);
					res[i].N_CA_C1_angle = angle(CA_N, CA_C1);
				}
				else {
					res[i].N_CA_C1_angle = ANGLE_NA;
				}
			}
			
			
			// debug
			//fprintf(stderr, "%3s %d:\t%lf\t%lf\t%lf\t%lf\n",
			//	res[i].resName, res[i].id, res[i].phi, res[i].psi, res[i].omega, res[i].chi);
	}
}


/*************************************************************/
/* Reads the angles for every residue from the given angles  */
/* file and into the ResData structs for each residue.       */
/*************************************************************/
void readAngles(char *inputAnglesFilename, ResData *residues, int n) {
	int i = 0;
	FILE *fp = fopen(inputAnglesFilename, "r");
	
	if (fp != NULL) {
		for (i=0; i<n; i++) {
			ResData *res = &residues[i];
			if (!readResidue(res, fp)) {
				fprintf(stderr, "Read residue failed\n");
			}
			
			/*
			printf( "%d: %s %lf %lf %lf\n\t", i, res->resName, res->phi, res->psi, res->omega);
			for (rc=0; rc <= res->max_atom_index; rc++) {
				printf( "%s, ", res->names[rc]);
			}	
			printf("\n");
			*/
		}
		fclose(fp);
		
	} else {
		fprintf(stderr, "ERROR: Could not open dihedral angles file \"%s\".\n", inputAnglesFilename);
		exit(1);
	}
}

/*************************************************************/
/* Calls phipsi with the --db flag to generate a file of     */
/* residue coordinates from the PDB file with the given      */
/* name. The residue coordinates file contains the           */
/* coordinates of each atom in each residue, for both the    */
/* backbone and the side chains, and the coordinates for     */
/* each residue are canonical form. Returns a pointer to the */
/* name of the residue coordinates file.                     */
/*************************************************************/
void * createResidueCoordinatesFile(char *inputPDBFilename) {
	char *call_phipsi, *phipsi_output;
	
	phipsi_output = (char*)malloc( (strlen(inputPDBFilename)+strlen(".rescoords")+1)*sizeof(char) );
	sprintf(phipsi_output,"%s.rescoords",inputPDBFilename);
	
	call_phipsi = (char*)malloc( (strlen("./phipsi --db ")+strlen(inputPDBFilename)+strlen(" > ")+strlen(phipsi_output)+1)*sizeof(char) );
	if(!call_phipsi){printf("Cannot open phipsi file for writing to.\n");exit(-1);}
	sprintf(call_phipsi, "./phipsi --db %s > %s",inputPDBFilename,phipsi_output);
	//printf("call_phipsi = %s\n",call_phipsi); // debug
	if (system(call_phipsi) != 0) {
		fprintf(stderr, "ERROR: Call to phipsi failed.\n");
		exit(-1);
	}
	free(call_phipsi);
	
	return phipsi_output;
}

/*************************************************************/
/* Calls phipsi to generate a dihedral angles file from the  */
/* PDB file with the given name. Returns a pointer to the    */
/* name of the dihedral angles file.                         */
/*************************************************************/
void * createAnglesFile(char *inputPDBFilename) {
	char *call_phipsi, *phipsi_output;
	
	phipsi_output = (char*)malloc( (strlen(inputPDBFilename)+strlen(".angles")+1)*sizeof(char) );
	sprintf(phipsi_output,"%s.angles",inputPDBFilename);
	
	call_phipsi = (char*)malloc( (strlen("./phipsi ")+strlen(inputPDBFilename)+strlen(" > ")+strlen(phipsi_output)+1)*sizeof(char) );
	if(!call_phipsi){printf("Cannot open phipsi file for writing to.\n");exit(-1);}
	sprintf(call_phipsi, "./phipsi %s > %s",inputPDBFilename,phipsi_output);
	//printf("call_phipsi = %s\n",call_phipsi); // debug
	if (system(call_phipsi) != 0) {
		fprintf(stderr, "ERROR: Call to phipsi failed.\n");
		exit(-1);
	}
	free(call_phipsi);
	
	return phipsi_output;
}

/*
 * Parses the hydrogen bonds file with the name given by inputHbondsFilename, and sets the
 * hbonds_intended_acc and hbonds_intended_don variables in the ResData structs in the ResData
 * array 'residues'.
 *
 * The format of the hydrogen bonds file is a series of lines, e.g.
 *		5 6 # comment
 *		-4 -3
 *		LEU 13 ARG 16
 *		TYR98 GLU99
 *		
 *		23 24
 *		+67 +68
 *		45 46 47 # the 47 is ignored
 *		56 57
 *
 * The first number encountered on a line is the residue ID of the acceptor, and the second number
 * encountered is the residue ID of the donor. Anything after a '#' sign on a line is ignored.
 * Letters can be added (e.g. for residue names) for human-readability, but these will be ignored
 * by the program. Negative residue IDs are allowed. Lines with no numbers or one number are ignored,
 * whereas on lines with more than two numbers only the first two numbers are used.
 *
 * Only hydrogen bonds involving only backbone atoms are being specified. A residue can be specified
 * to participate as an acceptor in up to 2 hydrogen bonds, and as a donor in up to 2 hydrogen bonds.
 * The file only represents hydrogen bonds that the user wants the program to try to form or optimize;
 * other hydrogen bonds may be formed in addition to the ones specified in the file.
 */
static void parseHbondsFile(char* inputHbondsFilename, ResData *residues, int totalRes) {
	int i, j;
	FILE *fp;
	char line[MAX_LINE_LEN];
	char *comment_ptr;
	int acceptor_id = -1000000, donor_id = -1000000; // The residues IDs of the acceptor and donor of the hydrogen bond.
	int acceptor_index = -1, donor_index = -1; // The indices of the acceptor and donor of the hydrogen bond.
	
	char c;
	int num_start_index = -1; // Index where a number starts. -1 mean none found so far.
	int num_found;
	int found_acceptor;
	int found_hbond;
	
	// Initialize the hbonds_intended_acc and hbonds_intended_don variables the peptide.
	
	for (i = 0; i < totalRes; i++) {
		for (j = 0; j < 2; j++) {
			residues[i].hbonds_intended_acc[j].partner = -1;
			residues[i].hbonds_intended_acc[j].distance = -1;
			residues[i].hbonds_intended_don[j].partner = -1;
			residues[i].hbonds_intended_don[j].distance = -1;
		}
	}
	
	// Parse the hydrogen bonds file.
	
	fp = fopen(inputHbondsFilename, "r");
	if (fp != NULL) {
		
		while (fgets(line, MAX_LINE_LEN, fp) != NULL) {
			if ((comment_ptr = strchr(line, '#')) != NULL) { // Look for first '#' sign, indicating a comment.
				*comment_ptr = '\0'; // If '#' was found, overwrite it with a NULL byte, terminating the string.
			}
			
			i = 0;
			num_start_index = -1;
			found_acceptor = FALSE;
			found_hbond = FALSE;
			while ((c = line[i]) != '\0') {
				
				if (num_start_index == -1) {
					if ( (c >= '0' && c <= '9') || ((c == '-' || c == '+') && (line[i+1] >= '0' && line[i+1] <= '9')) ) {
						num_start_index = i;
					}
				}
				else if ( !(c >= '0' && c <= '9') ) {
					sscanf(&line[num_start_index], "%d", &num_found);
					num_start_index = -1;
					
					if (!found_acceptor) {
						acceptor_id = num_found;
						found_acceptor = TRUE;
					}
					else {
						donor_id = num_found;
						found_hbond = TRUE;
						break;
					}
				}
				
				i++;
			}
			
			if (found_hbond) {
				// A hydrogen bonds is specified with the residue with ID acceptor_id as the acceptor,
				// and the residue with ID donor_id as the donor.
				
				//printf("Hbond: %d to %d\n", acceptor_id, donor_id);
				
				// Determine the indices of the acceptor and donor, which are NOT the same as the residue IDs.
				
				acceptor_index = -1;
				donor_index = -1;
				
				for (i = 0; i < totalRes; i++) {
					if (residues[i].id == acceptor_id) {
						acceptor_index = i;
						break;
					}
				}
				
				for (i = 0; i < totalRes; i++) {
					if (residues[i].id == donor_id) {
						donor_index = i;
						break;
					}
				}
				
				if (acceptor_index == -1 || donor_index == -1) {
					fprintf(stderr, "ERROR: Hydrogen bond in hydrogen bonds file uses residue IDs that do not exist in the peptide.\n");
					fprintf(stderr, "Acceptor residue ID %d, Donor residue ID %d\n", acceptor_id, donor_id);
					exit(1);
				}
				
				// debug
				//printf("   Acceptor index %d, Donor index %d\n", acceptor_index, donor_index);
				
				
				// Set the variables for this intended hydrogen bond.
				
				for (j = 0; j < 2; j++) {
					if (residues[acceptor_index].hbonds_intended_acc[j].partner == -1) {
						residues[acceptor_index].hbonds_intended_acc[j].partner = donor_index;
						break;
					}
					if (j == 1) {
						fprintf(stderr, "ERROR: A single residue cannot be the acceptor of more than 2 hydrogen bonds.\n");
						fprintf(stderr, "Problem in hydrogen bonds file. Problem residue ID is %d.\n", acceptor_id);
						exit(1);
					}
				}
				
				for (j = 0; j < 2; j++) {
					if (residues[donor_index].hbonds_intended_don[j].partner == -1) {
						residues[donor_index].hbonds_intended_don[j].partner = acceptor_index;
						break;
					}
					if (j == 1) {
						fprintf(stderr, "ERROR: A single residue cannot be the donor in more than 2 hydrogen bonds.\n");
						fprintf(stderr, "Problem in hydrogen bonds file. Problem residue ID is %d.\n", donor_id);
						exit(1);
					}
				}
				
			}
			
		}
		
		fclose(fp);
		
	} else {
		fprintf(stderr, "ERROR: Could not open hydrogen bonds file \"%s\".\n", inputHbondsFilename);
	}
}


/*************************************************************/
/* PepMake reads a file of dihedral angles and distances and */
/* builds a PDB file from the representation.                */
/*************************************************************/
static void pepmake(int haveAnglesFile, char *inputAnglesFilename, int haveResCoordsFile, char *resCoordsFilename,
		char *inputPDBFilename, char *outputPDBFilename, int fixTorsionAngles, GAP_CLOSURE_PARAMS gap_params,
		int formHbonds, char* inputHbondsFilename) {
	
	int n;
	FILE *fp_write;
	PDBFile pdb;
	int rc;
	
	if (havePDBFile) {
		
		rc = parsePDBFile(inputPDBFilename,&pdb, '?');
		if (rc != TRUE) {
			fprintf(stderr, "Error (%d): PepMake could not parse PDB file %s.\n", rc, inputPDBFilename);
			exit(-1);
		}
		
		n=pdb.numRes; // Both values are the number of residues in the peptide.
		
	}
	else {
		// No input PDB file was given. Count the number of residues based on the input angles file.
		n = countResidues(inputAnglesFilename);
	}
	
	fprintf(stderr, "Residues = %d\n", n);
	if (n > 0) {
		ResData *residues = SAFEMALLOC(sizeof(ResData)*n);
		
		read_rama_tables();
		
		read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", general_N_CA_C1_angle_lookup);
		read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/general_N_CA_C_angle.txt", gly_N_CA_C1_angle_lookup);
		read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/pro_N_CA_C_angle.txt", pro_N_CA_C1_angle_lookup);
		read_N_CA_C1_angle_lookup_file("N_CA_C_angle_lookup/xpr_N_CA_C_angle.txt", xpr_N_CA_C1_angle_lookup);
		
		if (havePDBFile) {
			// Since we have an input PDB file, loadResidueTemplates() will load the PDB coordinates from
			// the PDB file and set the angles based on these coordinates. Then, if we have also been given
			// an angles file, we will substitute with the angles from the angles file.
			
			loadResidueTemplates(inputPDBFilename, pdb, havePDBFile, FALSE, haveResCoordsFile, resCoordsFilename, residues, n);
			
			if (haveAnglesFile) {
				readAngles(inputAnglesFilename, residues, n);
			}
		}
		else {
			// No input PDB file was given. Read the angles from the input angles file, and then call
			// loadResidueTemplates(), which will load coordinate from the residue coordinates file or
			// from the residue templates without changing the angles.
			
			readAngles(inputAnglesFilename, residues, n);
			predict_N_CA_C1_angles(residues, n);
			
			loadResidueTemplates(inputPDBFilename, pdb, havePDBFile, FALSE, haveResCoordsFile, resCoordsFilename, residues, n);
		}
		
		repair_O_geometry(residues, n);
		
		buildPeptide(residues, n);
		
		if (gap_params.close_gap) {
			if ( !closeGapCCD(residues, n, gap_params) ) {
				printf("Gap closure failed.\n"); // Homodeller will test for this, so do not change it!
			}
		}
		
		if (formHbonds) {
			parseHbondsFile(inputHbondsFilename, residues, n);
			alignBetaSheets1(residues, n);
		}
		
		if (fixTorsionAngles) {
			repairAngles(residues, n, NULL);
		}
		
		fp_write = fopen(outputPDBFilename,"w");
		printPDBAtoms(fp_write, residues, n);
		fclose(fp_write);
		
		SAFEFREE(residues);
	}
	
}


/*************************************************************/
/* Prints info about how to use the program.                 */
/*************************************************************/
static void usage() {
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "   pepmake [--printHN] [--fixbb] [-t]\n");
	fprintf(stderr, "   [<< -g [res_index_N] res_index_C > | < -gid [res_id_N] res_id_C >>\n");
	fprintf(stderr, "   [-gt gap_threshold]] [-h hbonds_list_file] [-a dihedral.angles.txt]\n");
	fprintf(stderr, "   [-r residue_coords_file] [-pdb pdbfile.pdb] [-o output.pdb]\n\n");
	fprintf(stderr, "At least one input file must be given, either a dihedral angles file or a PDB\n");
	fprintf(stderr, "file, or both. Alternatively, an angles file can be given along with a\n");
	fprintf(stderr, "residue coordinates file. A residues coordinates file should have the same\n");
	fprintf(stderr, "format as the file produced by phipsi when using phipsi's --db option.\n\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, " --printHN\n");
	fprintf(stderr, "   prints HN atoms in the output PDB file\n");
	fprintf(stderr, " --fixbb\n");
	fprintf(stderr, "   Uses single fixed template for all backbone atoms\n");
	fprintf(stderr, " -t\n");
	fprintf(stderr, "   Fix problematic torsion angles (phi, psi, omega) using Cyclic Coordinate\n");
	fprintf(stderr, "   Descent algorithm.\n");
	fprintf(stderr, " -g [res_index_N] res_index_C\n");
	fprintf(stderr, "   Attempt to close a gap in the structure. e.g.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "         -----------------------------N--------    C--------------------\n");
	fprintf(stderr, "                                      |            |                    \n");
	fprintf(stderr, "                                   N-anchor     C-anchor                \n");
	fprintf(stderr, "\n");
	fprintf(stderr, "   res_index_C must be the index of the C-anchor, i.e. the residue just\n");
	fprintf(stderr, "   after the gap. res_index_N is the index of the N-anchor, which should be\n");
	fprintf(stderr, "   several residues to the left; if omitted, the program will try a few\n");
	fprintf(stderr, "   different N-anchor values. The \"index\" number is counted from the first\n");
	fprintf(stderr, "   residue in the peptide, which is 0. The N-anchor and C-anchor residues will\n");
	fprintf(stderr, "   stay in fixed positions, and the residues in between will be adjusted --\n");
	fprintf(stderr, "   except that slight distortion of backbone geometry may occur at the\n");
	fprintf(stderr, "   C-anchor, especially if the gap threshold is set high with the -gt option.\n");
	fprintf(stderr, "   The gap closure algorithm uses the Cyclic Coordinate Descent algorithm.\n");
	fprintf(stderr, "   If gap closure fails, \"Gap closure failed\" will be printed to standard\n");
	fprintf(stderr, "   output.\n");
	fprintf(stderr, " -gid [res_id_N] res_id_C\n");
	fprintf(stderr, "   Attempt to close a gap in the structure, using residue IDs to indicate\n");
	fprintf(stderr, "   residues.\n");
	fprintf(stderr, " -gt gap_threshold\n");
	fprintf(stderr, "   Allow the sum of squared distances threshold for gap closure to be relaxed\n");
	fprintf(stderr, "   up to gap_threshold. gap_threshold must be greater than the default strict\n");
	fprintf(stderr, "   threshold of %lf.\n", CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD);
	fprintf(stderr, " -h hbonds_list_file\n");
	fprintf(stderr, "   Try to form or optimize the backbone hydrogen bonds listed in the given\n");
	fprintf(stderr, "   file using Cyclic Coordinate Descent algorithm.\n");
	fprintf(stderr, "   Format: The first number in each line is the ID number of the acceptor\n");
	fprintf(stderr, "   residue, and the second number is the ID number of the donor residue.\n");
	fprintf(stderr, "   Anything after a # is ignored. e.g.\n");
	fprintf(stderr, "      7 24 # Residue 7 is the aceptor, residue 24 is the donor.\n");
	fprintf(stderr, "      LEU 9 ARG 34 # Letters ignored, but added for convenience.\n");
}

/*************************************************************/
/* The main entry point                                      */
/*************************************************************/
int main_pepmake(int argc, char **argv) {
	char *inputHbondsFilename = '\0';
 	char *inputAnglesFilename = '\0';
	char *inputResCoordsFilename ='\0';
	char *inputPDBFilename = '\0';
	char *outputPDBFilename = "OUTPUT.pdb";
	int haveAnglesFile = FALSE;
	int haveResCoordsFile = FALSE;
	havePDBFile = FALSE;
	int fixTorsionAngles = FALSE;
	int formHbonds = FALSE;
	int i;
	
	GAP_CLOSURE_PARAMS gap_params;
	gap_params.close_gap = FALSE;
	gap_params.N_anchor_res_index = -1;
	gap_params.C_anchor_res_index = -1;
	gap_params.N_anchor_res_id = RESIDUE_ID_NOT_AVAILABLE;
	gap_params.C_anchor_res_id = RESIDUE_ID_NOT_AVAILABLE;
	gap_params.max_threshold = 0.0;
	
	if (argc == 1) {
		fprintf(stderr, "**** PepMake 2.0, based on BTI PepMake ****\n");
	}
	
	if (argc < 2) {
		usage();
		exit(1);
	}
	
	printH = FALSE;
	ignoreDBBackbone = FALSE;
	
	
	for (i = 1; i < argc; i++) {
     	if (strcmp(argv[i], "--printHN") == 0) {
     		printH = TRUE;
		} else if (strcmp(argv[i], "--fixbb") == 0) {
     		ignoreDBBackbone = TRUE;
		} else if (strcmp(argv[i], "-t") == 0) {
     		fixTorsionAngles = TRUE;
		} else if (strcmp(argv[i], "-g") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				if ( sscanf(argv[i], "%d", &gap_params.N_anchor_res_index) != 1 ) {
					fprintf(stderr, "Residue indices after -g option must be integers.\n");
					usage();
					exit(1);
				}
				if (i + 1 < argc) { // If there is room for at least 1 additional arg
					i++;
					if ( sscanf(argv[i], "%d", &gap_params.C_anchor_res_index) != 1 ) {
						// It's not a number given for the -g option.
						// The first number must have been the C-anchor residue index.
						i--;
						gap_params.C_anchor_res_index = gap_params.N_anchor_res_index;
						gap_params.N_anchor_res_index = -1;
					}
				}
				else {
					// The first number must have been the C-anchor residue index.
					gap_params.C_anchor_res_index = gap_params.N_anchor_res_index;
					gap_params.N_anchor_res_index = -1;
				}
				
				gap_params.close_gap = TRUE;
			}
			else {
				fprintf(stderr, "res_index_N and res_index_C must be given after -g option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-gid") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				if ( sscanf(argv[i], "%d", &gap_params.N_anchor_res_id) != 1 ) {
					fprintf(stderr, "Residue IDs after -gid option must be integers.\n");
					usage();
					exit(1);
				}
				if (i + 1 < argc) { // If there is room for at least 1 additional arg
					i++;
					if ( sscanf(argv[i], "%d", &gap_params.C_anchor_res_id) != 1 ) {
						// It's not a number given for the -g option.
						// The first number must have been the C-anchor residue ID.
						i--;
						gap_params.C_anchor_res_id = gap_params.N_anchor_res_id;
						gap_params.N_anchor_res_id = RESIDUE_ID_NOT_AVAILABLE;
					}
				}
				else {
					// The first number must have been the C-anchor residue ID.
					gap_params.C_anchor_res_id = gap_params.N_anchor_res_id;
					gap_params.N_anchor_res_id = RESIDUE_ID_NOT_AVAILABLE;
				}
				gap_params.close_gap = TRUE;
			}
			else {
				fprintf(stderr, "res_id_N and res_id_C must be given after -gid option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-gt") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				if ( sscanf(argv[i], "%lf", &gap_params.max_threshold) != 1 ) {
					fprintf(stderr, "Gap threshold after -gt option must be a number.\n");
					usage();
					exit(1);
				}
				
				if (gap_params.max_threshold <= 0.0) {
					fprintf(stderr, "Gap threshold must be > 0.\n");
					usage();
					exit(1);
				}
				else if (gap_params.max_threshold <= CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD) {
					fprintf(stderr, "Gap threshold must be > default value of %lf.\n", CCD_DEFAULT_SUM_OF_SQUARED_DISTANCES_THRESHOLD);
					usage();
					exit(1);
				}
			}
			else {
				fprintf(stderr, "Gap threshold must be specified after -gt option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-h") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				inputHbondsFilename = argv[i];
				check_file_readability(inputHbondsFilename);
				formHbonds = TRUE;
				printf("input hydrogen bonds file = %s\n", inputHbondsFilename);
			}
			else {
				fprintf(stderr, "Hydrogen bonds file must be specified after -h option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-a") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				inputAnglesFilename = argv[i];
				check_file_readability(inputAnglesFilename);
				haveAnglesFile = TRUE;
				printf("input angles file = %s\n", inputAnglesFilename);
			}
			else {
				fprintf(stderr, "Angles file must be specified after -a option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-r") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				inputResCoordsFilename = argv[i];
				check_file_readability(inputResCoordsFilename);
				haveResCoordsFile = TRUE;
				printf("input residue coordinates file = %s\n", inputResCoordsFilename);
			}
			else {
				fprintf(stderr, "Residue coordinates file must be specified after -r option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-pdb") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				inputPDBFilename = argv[i];
				check_file_readability(inputPDBFilename);
				havePDBFile = TRUE;
				printf("input PDB file = %s\n", inputPDBFilename);
			}
			else {
				fprintf(stderr, "PDB file must be specified after -pdb option.\n");
				usage();
				exit(1);
			}
		} else if (strcmp(argv[i], "-o") == 0) {
			if (i + 1 < argc) { // if arg i is not the last arg
				i++;
				outputPDBFilename = argv[i];
			}
			else {
				fprintf(stderr, "Output PDB filename must be specified after -o option.\n");
				usage();
				exit(1);
			}
		} else {
			fprintf(stderr, "Unrecognized Option: [%s]\n", argv[i]);
			usage();
			exit(1);
		}
	}
	
	if (haveResCoordsFile && havePDBFile) {
		fprintf(stderr, "PDB file cannot be given along with residue coordinates file.\n");
		usage();
		exit(1);
	}
	else if (haveResCoordsFile && !haveAnglesFile) {
		fprintf(stderr, "Angles file must be given if residue coordinates file if given.\n");
		usage();
		exit(1);
	}
	else if (!(haveAnglesFile || havePDBFile)) {
		fprintf(stderr, "Must specify at least one input file.\n");
		usage();
		exit(1);
	}
	
	if (havePDBFile && !fixTorsionAngles && !formHbonds && !gap_params.close_gap) {
		// Since we have a PDB file, we will use the omega angles implicit in the PDB file, and
		// we do not have to spend time adjusting omega angles. This ONLY applies if we will
		// not be changing the omega angles in any way (as we will, for example, if fixTorsionAngles
		// is true, formHbonds is true, or if buildPeptide is called from another program such as
		// GAFolder).
		adjustOmega = FALSE;
	}
	
	if (havePDBFile) {
		// Do not adjust chi angles if we have been given a PDB file.
		rotateChi = FALSE;
	}
	
	if (!gap_params.close_gap && gap_params.max_threshold != 0.0) {
		fprintf(stderr, "You cannot specify the gap threshold without -g or -gid options.\n");
		usage();
		exit(1);
	}
	else if (gap_params.close_gap && gap_params.C_anchor_res_index != -1 &&
			gap_params.C_anchor_res_id != RESIDUE_ID_NOT_AVAILABLE) {
		fprintf(stderr, "You cannot use both -g and -gid options.\n");
		usage();
		exit(1);
	}
	
	pepmake(haveAnglesFile, inputAnglesFilename, haveResCoordsFile, inputResCoordsFilename, inputPDBFilename, outputPDBFilename,
		fixTorsionAngles, gap_params, formHbonds, inputHbondsFilename);
	
	printf("The pdb file is written to %s\n\n", outputPDBFilename);
	return 0;
}
